Package: progresstestcpp
Title: Testing Terminal Progress Bars
Version: 1.0.0
Authors@R: person("Gabor", "Csardi", , "csardi.gabor@gmail.com", role = c("aut", "cre"))
Description: Test progress bars from the cli package, C++ version.
License: MIT + file LICENSE
Imports:
    cli
LinkingTo:
    cli,
    cpp11
RoxygenNote: 7.1.1.9001
Encoding: UTF-8

Package: cli
Title: Helpers for Developing Command Line Interfaces
Version: 3.6.2
Authors@R: c(
    person("Gábor", "Csárdi", , "csardi.gabor@gmail.com", role = c("aut", "cre")),
    person("Hadley", "Wickham", role = "ctb"),
    person("Kirill", "Müller", role = "ctb"),
    person("Salim", "Brüggemann", , "salim-b@pm.me", role = "ctb",
           comment = c(ORCID = "0000-0002-5329-5987")),
    person("Posit Software, PBC", role = c("cph", "fnd"))
  )
Description: A suite of tools to build attractive command line interfaces
    ('CLIs'), from semantic elements: headings, lists, alerts, paragraphs,
    etc. Supports custom themes via a 'CSS'-like language. It also
    contains a number of lower level 'CLI' elements: rules, boxes, trees,
    and 'Unicode' symbols with 'ASCII' alternatives. It support ANSI
    colors and text styles as well.
License: MIT + file LICENSE
URL: https://cli.r-lib.org, https://github.com/r-lib/cli
BugReports: https://github.com/r-lib/cli/issues
Depends: R (>= 3.4)
Imports: utils
Suggests: callr, covr, crayon, digest, glue (>= 1.6.0), grDevices,
        htmltools, htmlwidgets, knitr, methods, mockery, processx, ps
        (>= 1.3.4.9000), rlang (>= 1.0.2.9003), rmarkdown, rprojroot,
        rstudioapi, testthat, tibble, whoami, withr
Config/Needs/website: r-lib/asciicast, bench, brio, cpp11, decor, desc,
        fansi, prettyunits, sessioninfo, tidyverse/tidytemplate,
        usethis, vctrs
Config/testthat/edition: 3
Encoding: UTF-8
RoxygenNote: 7.2.3
NeedsCompilation: yes
Packaged: 2023-12-10 22:38:10 UTC; gaborcsardi
Author: Gábor Csárdi [aut, cre],
  Hadley Wickham [ctb],
  Kirill Müller [ctb],
  Salim Brüggemann [ctb] (<https://orcid.org/0000-0002-5329-5987>),
  Posit Software, PBC [cph, fnd]
Maintainer: Gábor Csárdi <csardi.gabor@gmail.com>
Repository: CRAN
Date/Publication: 2023-12-11 07:40:13 UTC

Package: progresstest
Title: Testing Terminal Progress Bars
Version: 1.0.0
Authors@R: person("Gabor", "Csardi", , "csardi.gabor@gmail.com", role = c("aut", "cre"))
Description: Test progress bars from the cli package.
License: MIT + file LICENSE
Imports:
    cli
LinkingTo:
    cli
RoxygenNote: 7.1.1.9001
Encoding: UTF-8

# Generated by roxygen2: do not edit by hand

export(test_baseline)
export(test_cli)
export(test_cli_unroll)
export(test_modulo)
export(testx)
useDynLib(progresstest, .registration = TRUE, .fixes = "c_")

# Generated by roxygen2: do not edit by hand

S3method("[",cli_doc)
S3method(as.character,cli_no)
S3method(as.character,cli_noprint)
S3method(as.character,cli_sitrep)
S3method(cli_format,character)
S3method(cli_format,default)
S3method(cli_format,numeric)
S3method(format,cli_ansi_html_style)
S3method(format,cli_diff_chr)
S3method(format,cli_diff_str)
S3method(format,cli_doc)
S3method(format,cli_progress_demo)
S3method(format,cli_sitrep)
S3method(format,cli_spark)
S3method(print,cli_ansi_html_style)
S3method(print,cli_ansi_string)
S3method(print,cli_ansi_style)
S3method(print,cli_boxx)
S3method(print,cli_diff_chr)
S3method(print,cli_doc)
S3method(print,cli_progress_demo)
S3method(print,cli_rule)
S3method(print,cli_sitrep)
S3method(print,cli_spark)
S3method(print,cli_spinner)
S3method(print,cli_tree)
export("__cli_update_due")
export(ansi_align)
export(ansi_chartr)
export(ansi_collapse)
export(ansi_columns)
export(ansi_grep)
export(ansi_grepl)
export(ansi_has_any)
export(ansi_has_hyperlink_support)
export(ansi_hide_cursor)
export(ansi_html)
export(ansi_html_style)
export(ansi_hyperlink_types)
export(ansi_nchar)
export(ansi_nzchar)
export(ansi_palette_show)
export(ansi_palettes)
export(ansi_regex)
export(ansi_show_cursor)
export(ansi_simplify)
export(ansi_string)
export(ansi_strip)
export(ansi_strsplit)
export(ansi_strtrim)
export(ansi_strwrap)
export(ansi_substr)
export(ansi_substring)
export(ansi_tolower)
export(ansi_toupper)
export(ansi_trimws)
export(ansi_with_hidden_cursor)
export(bg_black)
export(bg_blue)
export(bg_br_black)
export(bg_br_blue)
export(bg_br_cyan)
export(bg_br_green)
export(bg_br_magenta)
export(bg_br_red)
export(bg_br_white)
export(bg_br_yellow)
export(bg_cyan)
export(bg_green)
export(bg_magenta)
export(bg_none)
export(bg_red)
export(bg_white)
export(bg_yellow)
export(boxx)
export(builtin_theme)
export(cat_boxx)
export(cat_bullet)
export(cat_line)
export(cat_print)
export(cat_rule)
export(ccli_tick_reset)
export(cli)
export(cli_abort)
export(cli_alert)
export(cli_alert_danger)
export(cli_alert_info)
export(cli_alert_success)
export(cli_alert_warning)
export(cli_blockquote)
export(cli_bullets)
export(cli_bullets_raw)
export(cli_code)
export(cli_div)
export(cli_dl)
export(cli_end)
export(cli_fmt)
export(cli_format)
export(cli_format_method)
export(cli_h1)
export(cli_h2)
export(cli_h3)
export(cli_inform)
export(cli_li)
export(cli_list_themes)
export(cli_ol)
export(cli_output_connection)
export(cli_par)
export(cli_process_done)
export(cli_process_failed)
export(cli_process_start)
export(cli_progress_along)
export(cli_progress_bar)
export(cli_progress_builtin_handlers)
export(cli_progress_cleanup)
export(cli_progress_demo)
export(cli_progress_done)
export(cli_progress_message)
export(cli_progress_num)
export(cli_progress_output)
export(cli_progress_step)
export(cli_progress_styles)
export(cli_progress_update)
export(cli_rule)
export(cli_sitrep)
export(cli_status)
export(cli_status_clear)
export(cli_status_update)
export(cli_text)
export(cli_tick_reset)
export(cli_ul)
export(cli_vec)
export(cli_verbatim)
export(cli_warn)
export(code_highlight)
export(code_theme_list)
export(col_black)
export(col_blue)
export(col_br_black)
export(col_br_blue)
export(col_br_cyan)
export(col_br_green)
export(col_br_magenta)
export(col_br_red)
export(col_br_white)
export(col_br_yellow)
export(col_cyan)
export(col_green)
export(col_grey)
export(col_magenta)
export(col_none)
export(col_red)
export(col_silver)
export(col_white)
export(col_yellow)
export(combine_ansi_styles)
export(console_width)
export(default_app)
export(demo_spinners)
export(diff_chr)
export(diff_str)
export(format_bullets_raw)
export(format_error)
export(format_inline)
export(format_message)
export(format_warning)
export(get_spinner)
export(has_keypress_support)
export(hash_animal)
export(hash_emoji)
export(hash_file_md5)
export(hash_file_sha1)
export(hash_file_sha256)
export(hash_file_xxhash)
export(hash_file_xxhash64)
export(hash_md5)
export(hash_obj_animal)
export(hash_obj_emoji)
export(hash_obj_md5)
export(hash_obj_sha1)
export(hash_obj_sha256)
export(hash_obj_xxhash)
export(hash_obj_xxhash64)
export(hash_raw_animal)
export(hash_raw_emoji)
export(hash_raw_md5)
export(hash_raw_sha1)
export(hash_raw_sha256)
export(hash_raw_xxhash)
export(hash_raw_xxhash64)
export(hash_sha1)
export(hash_sha256)
export(hash_xxhash)
export(hash_xxhash64)
export(is_ansi_tty)
export(is_dynamic_tty)
export(is_utf8_output)
export(keypress)
export(list_border_styles)
export(list_spinners)
export(list_symbols)
export(make_ansi_style)
export(make_spinner)
export(no)
export(num_ansi_colors)
export(pb_bar)
export(pb_current)
export(pb_current_bytes)
export(pb_elapsed)
export(pb_elapsed_clock)
export(pb_elapsed_raw)
export(pb_eta)
export(pb_eta_raw)
export(pb_eta_str)
export(pb_extra)
export(pb_id)
export(pb_name)
export(pb_percent)
export(pb_pid)
export(pb_rate)
export(pb_rate_bytes)
export(pb_rate_raw)
export(pb_spin)
export(pb_status)
export(pb_timestamp)
export(pb_total)
export(pb_total_bytes)
export(pluralize)
export(pretty_print_code)
export(qty)
export(rule)
export(ruler)
export(simple_theme)
export(spark_bar)
export(spark_line)
export(start_app)
export(stop_app)
export(style_blurred)
export(style_bold)
export(style_dim)
export(style_hidden)
export(style_hyperlink)
export(style_inverse)
export(style_italic)
export(style_no_bg_color)
export(style_no_blurred)
export(style_no_bold)
export(style_no_color)
export(style_no_dim)
export(style_no_hidden)
export(style_no_inverse)
export(style_no_italic)
export(style_no_strikethrough)
export(style_no_underline)
export(style_reset)
export(style_strikethrough)
export(style_underline)
export(symbol)
export(test_that_cli)
export(ticking)
export(tree)
export(truecolor)
export(utf8_graphemes)
export(utf8_nchar)
export(utf8_substr)
export(vt_output)
importFrom(utils,getParseData)
importFrom(utils,getSrcref)
useDynLib(cli, .registration=TRUE)

<a href="#"><img align="right" src="./docs/static/logo.png" width="20%"></a>

# Eask
> CLI for building, runing, testing, and managing your Emacs Lisp dependencies

[![License: GPL v3](https://img.shields.io/badge/License-GPL%20v3-green.svg)](https://www.gnu.org/licenses/gpl-3.0)
[![Emacs Version](https://img.shields.io/badge/Emacs-26.1+-7F5AB6.svg?logo=gnu%20emacs&logoColor=white)](https://www.gnu.org/software/emacs/download.html)
[![Release](https://img.shields.io/github/release/emacs-eask/cli.svg?logo=github)](https://github.com/emacs-eask/cli/releases/latest)
[![](https://img.shields.io/discord/1131434607213023262?label=Discord&logo=discord&logoColor=white&color=7289DA)](https://discord.gg/E9zzjWGfFD)

Eask was built to use as a package development tool in your Elisp packages. But
now, Eask supports various types of Emacs Lisp tasks. It can be used in three
major ways:

1. Dev tool for Elisp packages
2. Dependency management for your configuration
3. Run elisp programs for all other purposes

So what are the major differences between Eask and other build tools like
[Cask][], [makem.sh][], and [Eldev][], other than the things above?

Good question! Eask is more than a build tool now, it can be used for various
purposes! But here are Eask aims to be:

- **Consistent** enough to sandbox across all systems
- **General** enough to have Emacsers frequently used commands (`byte-compile`, `checkdoc`, etc)
- **Robust** enough to provide useful results even in the presence of user errors
- **Dependency-free** so that the tool can be run on any platform

*P.S. See [Why Eask?](https://emacs-eask.github.io/#-why-eask) for more detailed
information.*

## 🔗 Links
> 💡 `node` is not required to use Eask!

- [Documentation](https://emacs-eask.github.io/)
- [Installation](https://emacs-eask.github.io/Getting-Started/Install-Eask/)
- [Command-line interface](https://emacs-eask.github.io/Getting-Started/Commands-and-options/)
- [Examples](https://emacs-eask.github.io/Examples/Real-project-examples/)
- [FAQ](https://emacs-eask.github.io/FAQ/)

## 🧪 Testing

###### Documentation

| Description                            | Done | Status                                                                                                                                          |
|----------------------------------------|------|-------------------------------------------------------------------------------------------------------------------------------------------------|
| Keep the documentation page up to date | ✔    | [![Docs](https://github.com/emacs-eask/cli/actions/workflows/docs.yml/badge.svg)](https://github.com/emacs-eask/cli/actions/workflows/docs.yml) |

###### Development

| Description                                | Done | Status                                                                                                                                                   |
|--------------------------------------------|------|----------------------------------------------------------------------------------------------------------------------------------------------------------|
| Compile source and check redefined         | ✔    | [![Compile](https://github.com/emacs-eask/cli/actions/workflows/compile.yml/badge.svg)](https://github.com/emacs-eask/cli/actions/workflows/compile.yml) |
| Compatibility check for each Emacs version | ✔    | [![Compat](https://github.com/emacs-eask/cli/actions/workflows/compat.yml/badge.svg)](https://github.com/emacs-eask/cli/actions/workflows/compat.yml)    |
| Build executables                          | ✔    | [![Build](https://github.com/emacs-eask/cli/actions/workflows/build.yml/badge.svg)](https://github.com/emacs-eask/cli/actions/workflows/build.yml)       |

###### Commands

| Description                                  | Done | Status                                                                                                                                                                              |
|----------------------------------------------|------|-------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------|
| Test commands in global (`~/.eask/`) mode    | ✔    | [![Global](https://github.com/emacs-eask/cli/actions/workflows/global.yml/badge.svg)](https://github.com/emacs-eask/cli/actions/workflows/global.yml)                               |
| Test commands in config (`~/.emacs.d/`) mode | ✔    | [![Confg](https://github.com/emacs-eask/cli/actions/workflows/config.yml/badge.svg)](https://github.com/emacs-eask/cli/actions/workflows/config.yml)                                |
| Test commands in development (`./`) mode     | ✔    | [![Local](https://github.com/emacs-eask/cli/actions/workflows/local.yml/badge.svg)](https://github.com/emacs-eask/cli/actions/workflows/local.yml)                                  |
| Test install packages                        | ✔    | [![Install](https://github.com/emacs-eask/cli/actions/workflows/install.yml/badge.svg)](https://github.com/emacs-eask/cli/actions/workflows/install.yml)                            |
| Test link packages                           | ✔    | [![Link](https://github.com/emacs-eask/cli/actions/workflows/link.yml/badge.svg)](https://github.com/emacs-eask/cli/actions/workflows/link.yml)                                     |
| Test `docker` command                        | ✔    | [![Docker](https://github.com/emacs-eask/cli/actions/workflows/docker.yml/badge.svg)](https://github.com/emacs-eask/cli/actions/workflows/docker.yml)                               |
| Test `exec` command                          | ✔    | [![Exec](https://github.com/emacs-eask/cli/actions/workflows/exec.yml/badge.svg)](https://github.com/emacs-eask/cli/actions/workflows/exec.yml)                                     |
| Test `emacs` command                         | ✔    | [![Emacs](https://github.com/emacs-eask/cli/actions/workflows/emacs.yml/badge.svg)](https://github.com/emacs-eask/cli/actions/workflows/emacs.yml)                                  |
| Test search packages                         | ✔    | [![Search](https://github.com/emacs-eask/cli/actions/workflows/search.yml/badge.svg)](https://github.com/emacs-eask/cli/actions/workflows/search.yml)                               |
| Test upgrade and check outdated packages     | ✔    | [![Outdated_Upgrade](https://github.com/emacs-eask/cli/actions/workflows/outdated_upgrade.yml/badge.svg)](https://github.com/emacs-eask/cli/actions/workflows/outdated_upgrade.yml) |
| `Eask`-file checker                          | ✔    | [![Checker](https://github.com/emacs-eask/cli/actions/workflows/checker.yml/badge.svg)](https://github.com/emacs-eask/cli/actions/workflows/checker.yml)                            |

###### Options

| Description          | Done | Status                                                                                                                                                   |
|----------------------|------|----------------------------------------------------------------------------------------------------------------------------------------------------------|
| Test option switches | ✔    | [![Options](https://github.com/emacs-eask/cli/actions/workflows/options.yml/badge.svg)](https://github.com/emacs-eask/cli/actions/workflows/options.yml) |

###### Test

| Description               | Done | Status                                                                                                                                                                           |
|---------------------------|------|----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------|
| Test `ert` command        | ✔    | [![Test ert](https://github.com/emacs-eask/cli/actions/workflows/test_ert.yml/badge.svg)](https://github.com/emacs-eask/cli/actions/workflows/test_ert.yml)                      |
| Test `ert-runner` command | ✔    | [![Test ert-runner](https://github.com/emacs-eask/cli/actions/workflows/test_ert-runner.yml/badge.svg)](https://github.com/emacs-eask/cli/actions/workflows/test_ert-runner.yml) |
| Test `buttercup` command  | ✔    | [![Test buttercup](https://github.com/emacs-eask/cli/actions/workflows/test_buttercup.yml/badge.svg)](https://github.com/emacs-eask/cli/actions/workflows/test_buttercup.yml)    |

###### Others

| Description | Done | Status                                                                                                                                                            |
|-------------|------|-------------------------------------------------------------------------------------------------------------------------------------------------------------------|
| Webinstall  | ✔    | [![Webinstall](https://github.com/emacs-eask/cli/actions/workflows/webinstall.yml/badge.svg)](https://github.com/emacs-eask/cli/actions/workflows/webinstall.yml) |

[Cask]: https://github.com/cask/cask
[makem.sh]: https://github.com/alphapapa/makem.sh
[Eldev]: https://github.com/doublep/eldev

# Confluent CLI

[![Release](https://img.shields.io/github/v/release/confluentinc/cli)](https://github.com/confluentinc/cli/releases/latest)
[![Build Status](https://semaphore.ci.confluent.io/badges/cli/branches/main.svg?style=shields&key=36d1298e-932a-4d04-8cd0-2483a2a6ab85)](https://semaphore.ci.confluent.io/projects/cli)

The Confluent CLI lets you manage your Confluent Cloud and Confluent Platform deployments, right from the terminal.

## Documentation

The [Confluent CLI Overview](https://docs.confluent.io/confluent-cli/current/overview.html) shows how to get started with the Confluent CLI.

The [Confluent CLI Command Reference](https://docs.confluent.io/confluent-cli/current/command-reference/index.html)
contains information on command arguments and flags, and is programmatically generated from this repository.

## Contributing

All contributions are appreciated, no matter how small!
When opening a PR, please make sure to follow our [contribution guide](CONTRIBUTING.md).

## Installation

The Confluent CLI is available to install for macOS, Linux, and Windows.

#### Homebrew

Install the latest version of `confluent` to `/usr/local/bin`:

    brew install confluentinc/tap/cli

#### APT (Ubuntu and Debian)

Install the latest version of `confluent` to `/usr/bin` (requires `glibc 2.17` or above for `amd64` and `glibc 2.27` or above for `arm64`):

    wget -qO - https://packages.confluent.io/confluent-cli/deb/archive.key | sudo apt-key add -
    sudo apt install software-properties-common
    sudo add-apt-repository "deb https://packages.confluent.io/confluent-cli/deb stable main"
    sudo apt update && sudo apt install confluent-cli

#### YUM (RHEL and CentOS)

Install the latest version of `confluent` to `/usr/bin` (requires `glibc 2.17` or above for `amd64` and `glibc 2.27` or above for `arm64`):

    sudo rpm --import https://packages.confluent.io/confluent-cli/rpm/archive.key
    sudo yum install yum-utils
    sudo yum-config-manager --add-repo https://packages.confluent.io/confluent-cli/rpm/confluent-cli.repo
    sudo yum clean all && sudo yum install confluent-cli

#### Windows

1. Download the latest Windows ZIP file from https://github.com/confluentinc/cli/releases/latest
2. Unzip `confluent_X.X.X_windows_amd64.zip`
3. Run `confluent.exe`

#### Docker

Pull the latest version:

    docker pull confluentinc/confluent-cli:latest

Pull `confluent` v3.6.0:

    docker pull confluentinc/confluent-cli:3.6.0

### Building from Source

    make
    dist/confluent_$(go env GOOS)_$(go env GOARCH)/confluent -h

#### Cross Compile for Other Platforms

Cross compilation from a Darwin/amd64 machine to Darwin/arm64, Linux/amd64 and Windows/amd64 platforms is supported.
To build for Darwin/arm64, run the following:

    GOARCH=arm64 make cross-build

To build for Linux (glibc or musl), install cross compiler `musl-cross` with Homebrew:

    brew install FiloSottile/musl-cross/musl-cross
    GOOS=linux make cross-build

To build for Windows/amd64, install `mingw-w64` compilers with Homebrew:

    brew install mingw-w64
    GOOS=windows make cross-build

Cross compilation from an M1 Macbook (Darwin/arm64) to other platforms is also supported.

#### Troubleshooting

Please update your system to MacOS 11.0 or later if you are building on Darwin/arm64.

cli
================

<!-- README.md is generated from README.Rmd. Please edit that file -->

> Helpers for Developing Command Line Interfaces

<!-- badges: start -->

[![R-CMD-check](https://github.com/r-lib/cli/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/r-lib/cli/actions/workflows/R-CMD-check.yaml)
[![](https://www.r-pkg.org/badges/version/cli)](https://www.r-pkg.org/pkg/cli)
[![CRAN RStudio mirror
downloads](https://cranlogs.r-pkg.org/badges/cli)](https://www.r-pkg.org/pkg/cli)
[![Codecov test
coverage](https://codecov.io/gh/r-lib/cli/graph/badge.svg)](https://app.codecov.io/gh/r-lib/cli)
<!-- badges: end -->

A suite of tools to build attractive command line interfaces (CLIs),
from semantic elements: headers, lists, alerts, paragraphs, etc.
Supports theming via a CSS-like language. It also contains a number of
lower level CLI elements: rules, boxes, trees, and Unicode symbols with
ASCII alternatives. It supports ANSI markup for terminal colors and font
styles.

------------------------------------------------------------------------

# Features

- Build a CLI using semantic elements: headings, lists, alerts,
  paragraphs.
- Theming via a CSS-like language.
- Terminal colors and font styles.
- All cli text can contain interpreted string literals, via the
  [glue](https://github.com/tidyverse/glue) package.
- Progress bars from R and C code.
- Error and warning messages with rich text formatting.
- Support for pluralized messages.
- ANSI styled string manipulation.

# Installation

Install the stable version from CRAN:

``` r
install.packages("cli")
```

Install the development version from GitHub:

``` r
pak::pak("r-lib/cli")
```

# Short tour

Some of the more commonly used cli elements, and features.

## Short alert messages

One liner messages to inform or warn.

``` r
pkgs <- c("foo", "bar", "foobar")
cli_alert_success("Downloaded {length(pkgs)} packages.")
```

<picture>
<source media="(prefers-color-scheme: dark)" srcset="man/figures/README/alert-success-dark.svg">
<img src="man/figures/README/alert-success.svg" width="100%" />
</picture>

``` r
db_url <- "example.com:port"
cli_alert_info("Reopened database {.url {db_url}}.")
```

<picture>
<source media="(prefers-color-scheme: dark)" srcset="man/figures/README/alert-info-dark.svg">
<img src="man/figures/README/alert-info.svg" width="100%" /> </picture>

``` r
cli_alert_warning("Cannot reach GitHub, using local database cache.")
```

<picture>
<source media="(prefers-color-scheme: dark)" srcset="man/figures/README/alert-warning-dark.svg">
<img src="man/figures/README/alert-warning.svg" width="100%" />
</picture>

``` r
cli_alert_danger("Failed to connect to database.")
```

<picture>
<source media="(prefers-color-scheme: dark)" srcset="man/figures/README/alert-danger-dark.svg">
<img src="man/figures/README/alert-danger.svg" width="100%" />
</picture>

``` r
cli_alert("A generic alert")
```

<picture>
<source media="(prefers-color-scheme: dark)" srcset="man/figures/README/alert-dark.svg">
<img src="man/figures/README/alert.svg" width="100%" /> </picture>

## Headings

Three levels of headings.

``` r
cli_h1("Heading 1")
```

<picture>
<source media="(prefers-color-scheme: dark)" srcset="man/figures/README/h1-dark.svg">
<img src="man/figures/README/h1.svg" width="100%" /> </picture>

``` r
cli_h2("Heading 2")
```

<picture>
<source media="(prefers-color-scheme: dark)" srcset="man/figures/README/h2-dark.svg">
<img src="man/figures/README/h2.svg" width="100%" /> </picture>

``` r
cli_h3("Heading 3")
```

<picture>
<source media="(prefers-color-scheme: dark)" srcset="man/figures/README/h3-dark.svg">
<img src="man/figures/README/h3.svg" width="100%" /> </picture>

## Lists

Ordered, unordered and description lists, that can be nested.

``` r
fun <- function() {
  cli_ol()
  cli_li("Item 1")
  ulid <- cli_ul()
  cli_li("Subitem 1")
  cli_li("Subitem 2")
  cli_end(ulid)
  cli_li("Item 2")
  cli_end()
}
fun()
```

<picture>
<source media="(prefers-color-scheme: dark)" srcset="man/figures/README/lists-dark.svg">
<img src="man/figures/README/lists.svg" width="100%" /> </picture>

## Themes

Theming via a CSS-like language.

``` r
fun <- function() {
  cli_div(theme = list(span.emph = list(color = "orange")))
  cli_text("This is very {.emph important}")
  cli_end()
  cli_text("Back to the {.emph previous theme}")
}
fun()
```

<picture>
<source media="(prefers-color-scheme: dark)" srcset="man/figures/README/themes-dark.svg">
<img src="man/figures/README/themes.svg" width="100%" /> </picture>

## Command substitution

Automatic command substitution via the
[glue](https://github.com/tidyverse/glue) package.

``` r
size <- 123143123
dt <- 1.3454
cli_alert_info(c(
  "Downloaded {prettyunits::pretty_bytes(size)} in ",
  "{prettyunits::pretty_sec(dt)}"))
```

<picture>
<source media="(prefers-color-scheme: dark)" srcset="man/figures/README/glue-dark.svg">
<img src="man/figures/README/glue.svg" width="100%" /> </picture>

## Pluralization

Pluralization support.

``` r
nfiles <- 3
ndirs <- 1
cli_alert_info("Found {nfiles} file{?s} and {ndirs} director{?y/ies}.")
```

<picture>
<source media="(prefers-color-scheme: dark)" srcset="man/figures/README/plurals-dark.svg">
<img src="man/figures/README/plurals.svg" width="100%" /> </picture>

## Progress bars

``` r
clean <- function() {
  cli_progress_bar("Cleaning data", total = 100)
  for (i in 1:100) {
    Sys.sleep(5/100)
    cli_progress_update()
  }
}
clean()
```

<picture>
<source media="(prefers-color-scheme: dark)" srcset="man/figures/README/progress-dark.svg">
<img src="man/figures/README/progress.svg" width="100%" /> </picture>

# Documentation

See at
[`https://cli.r-lib.org/`](https://cli.r-lib.org/reference/index.html)
and also in the installed package: `help(package = "cli")`.

# Code of Conduct

Please note that the cli project is released with a [Contributor Code of
Conduct](https://cli.r-lib.org/dev/CODE_OF_CONDUCT.html). By
contributing to this project, you agree to abide by its terms.

# License

MIT © Posit Software, PBC

[![Github Workflow build on master](https://github.com/bitwarden/cli/actions/workflows/build.yml/badge.svg?branch=master)](https://github.com/bitwarden/cli/actions/workflows/build.yml?query=branch:master)
[![Join the chat at https://gitter.im/bitwarden/Lobby](https://badges.gitter.im/bitwarden/Lobby.svg)](https://gitter.im/bitwarden/Lobby)

# Bitwarden Command-line Interface

[![Platforms](https://imgur.com/AnTLX0S.png "Platforms")](https://help.bitwarden.com/article/cli/#download--install)

The Bitwarden CLI is a powerful, full-featured command-line interface (CLI) tool to access and manage a Bitwarden vault. The CLI is written with TypeScript and Node.js and can be run on Windows, macOS, and Linux distributions.

![CLI](https://raw.githubusercontent.com/bitwarden/brand/master/screenshots/cli-macos.png "CLI")

## Download/Install

You can install the Bitwarden CLI multiple different ways:

**NPM**

If you already have the Node.js runtime installed on your system, you can install the CLI using NPM. NPM makes it easy to keep your installation updated and should be the preferred installation method if you are already using Node.js.

```bash
npm install -g @bitwarden/cli
```

**Native Executable**

We provide natively packaged versions of the CLI for each platform which have no requirements on installing the Node.js runtime. You can obtain these from the [downloads section](https://help.bitwarden.com/article/cli/#download--install) in the documentation.

**Other Package Managers**

- [Chocolatey](https://chocolatey.org/packages/bitwarden-cli)
  ```powershell
  choco install bitwarden-cli
  ```
- [Homebrew](https://formulae.brew.sh/formula/bitwarden-cli)
  ```bash
  brew install bitwarden-cli
  ```
- [Snap](https://snapcraft.io/bw)
  ```bash
  sudo snap install bw
  ```

## Documentation

The Bitwarden CLI is self-documented with `--help` content and examples for every command. You should start exploring the CLI by using the global `--help` option:

```bash
bw --help
```

This option will list all available commands that you can use with the CLI.

Additionally, you can run the `--help` option on a specific command to learn more about it:

```bash
bw list --help
bw create --help
```

**Detailed Documentation**

We provide detailed documentation and examples for using the CLI in our help center at https://help.bitwarden.com/article/cli/.

## Build/Run

**Requirements**

- [Node.js](https://nodejs.org) v16.13.1.
  - Testing is done against Node 16, other versions may work, but are not guaranteed.
- NPM v8

**Run the app**

```bash
npm install
npm run sub:init # initialize the git submodule for jslib
npm run build:watch
```

You can then run commands from the `./build` folder:

```bash
node ./build/bw.js login
```

## We're Hiring!

Interested in contributing in a big way? Consider joining our team! We're hiring for many positions. Please take a look at our [Careers page](https://bitwarden.com/careers/) to see what opportunities are currently open as well as what it's like to work at Bitwarden.

## Contribute

Code contributions are welcome! Please commit any pull requests against the `master` branch. Learn more about how to contribute by reading the [`CONTRIBUTING.md`](CONTRIBUTING.md) file.

Security audits and feedback are welcome. Please open an issue or email us privately if the report is sensitive in nature. You can read our security policy in the [`SECURITY.md`](SECURITY.md) file.

## Prettier

We recently migrated to using Prettier as code formatter. All previous branches will need to updated to avoid large merge conflicts using the following steps:

1. Check out your local Branch
2. Run `git merge ec53a16c005e0dd9aef6845c18811e8b14067168`
3. Resolve any merge conflicts, commit.
4. Run `npm run prettier`
5. Commit
6. Run `git merge -Xours 910b4a24e649f21acbf4da5b2d422b121d514bd5`
7. Push

### Git blame

We also recommend that you configure git to ignore the prettier revision using:

```bash
git config blame.ignoreRevsFile .git-blame-ignore-revs
```

<a href="#"><img align="right" src="./docs/static/logo.png" width="20%"></a>
# Eask
> CLI for building, running, testing, and managing your Emacs Lisp dependencies

[![License: GPL v3](https://img.shields.io/badge/License-GPL%20v3-green.svg)](https://www.gnu.org/licenses/gpl-3.0)
[![Emacs Version](https://img.shields.io/badge/Emacs-26.1+-7F5AB6.svg?logo=gnu%20emacs&logoColor=white)](https://www.gnu.org/software/emacs/download.html)
[![Release](https://img.shields.io/github/release/emacs-eask/cli.svg?logo=github)](https://github.com/emacs-eask/cli/releases/latest)
[![Discord](https://img.shields.io/discord/1131434607213023262?label=Discord&logo=discord&logoColor=white&color=7289DA)](https://discord.gg/E9zzjWGfFD)

Eask was originally designed as a package development tool for Elisp projects.
However, it has since expanded to support a wide range of Emacs Lisp tasks.
It can now be used in three major ways:

1. As a development tool for Elisp packages.
2. For managing dependencies in your Emacs configuration.
3. To run Elisp programs for a variety of purposes (essentially functioning as a runtime).

With these capabilities in mind, what sets Eask apart from other build tools
like [Cask][], [makem.sh][], and [Eldev][]?

Great question! Eask has evolved beyond just a build tool—it serves multiple purposes! Here’s what Eask aims to be:

- **Consistent**: Provides a reliable sandboxing environment across all systems.
- **Versatile**: Includes commonly used Emacs commands like `byte-compilation`, `checkdoc`, and more.
- **Robust**: Delivers useful results even when user errors occur.
- **Lightweight**: Runs on any platform without dependencies.

*📝 P.S. See [Why Eask?](https://emacs-eask.github.io/Getting-Started/Introduction/#-why-eask) for more detailed
information.*

## 🔗 Links
> 💡 [`node`][node] is not required to use Eask!

- [Documentation](https://emacs-eask.github.io/)
- [Installation](https://emacs-eask.github.io/Getting-Started/Install-Eask/)
- [Command-line interface](https://emacs-eask.github.io/Getting-Started/Commands-and-options/)
- [Examples](https://emacs-eask.github.io/Examples/Real-project-examples/)
- [FAQ](https://emacs-eask.github.io/FAQ/)

## 🧪 Testing

We have incorporated a range of tests to ensure Eask remains stable throughout its release cycle.

###### Documentation

| Description                            | Done | Status                                                                                                                                          |
|----------------------------------------|------|-------------------------------------------------------------------------------------------------------------------------------------------------|
| Keep the documentation page up to date | ✔    | [![Docs](https://github.com/emacs-eask/cli/actions/workflows/docs.yml/badge.svg)](https://github.com/emacs-eask/cli/actions/workflows/docs.yml) |

###### Development

| Description                                | Done | Status                                                                                                                                                   |
|--------------------------------------------|------|----------------------------------------------------------------------------------------------------------------------------------------------------------|
| Compile source and check redefined         | ✔    | [![Compile](https://github.com/emacs-eask/cli/actions/workflows/compile.yml/badge.svg)](https://github.com/emacs-eask/cli/actions/workflows/compile.yml) |
| Compatibility check for each Emacs version | ✔    | [![Compat](https://github.com/emacs-eask/cli/actions/workflows/compat.yml/badge.svg)](https://github.com/emacs-eask/cli/actions/workflows/compat.yml)    |
| Build executables                          | ✔    | [![Build](https://github.com/emacs-eask/cli/actions/workflows/build.yml/badge.svg)](https://github.com/emacs-eask/cli/actions/workflows/build.yml)       |

###### Commands

| Description                                  | Done | Status                                                                                                                                                                              |
|----------------------------------------------|------|-------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------|
| Test commands in global (`~/.eask/`) mode    | ✔    | [![Global](https://github.com/emacs-eask/cli/actions/workflows/global.yml/badge.svg)](https://github.com/emacs-eask/cli/actions/workflows/global.yml)                               |
| Test commands in config (`~/.emacs.d/`) mode | ✔    | [![Confg](https://github.com/emacs-eask/cli/actions/workflows/config.yml/badge.svg)](https://github.com/emacs-eask/cli/actions/workflows/config.yml)                                |
| Test commands in development (`./`) mode     | ✔    | [![Local](https://github.com/emacs-eask/cli/actions/workflows/local.yml/badge.svg)](https://github.com/emacs-eask/cli/actions/workflows/local.yml)                                  |
| Test install packages                        | ✔    | [![Install](https://github.com/emacs-eask/cli/actions/workflows/install.yml/badge.svg)](https://github.com/emacs-eask/cli/actions/workflows/install.yml)                            |
| Test link packages                           | ✔    | [![Link](https://github.com/emacs-eask/cli/actions/workflows/link.yml/badge.svg)](https://github.com/emacs-eask/cli/actions/workflows/link.yml)                                     |
| Test `analyze` command / `Eask`-file checker | ✔    | [![Analyze](https://github.com/emacs-eask/cli/actions/workflows/analyze.yml/badge.svg)](https://github.com/emacs-eask/cli/actions/workflows/analyze.yml)                            |
| Test `docker` command                        | ✔    | [![Docker](https://github.com/emacs-eask/cli/actions/workflows/docker.yml/badge.svg)](https://github.com/emacs-eask/cli/actions/workflows/docker.yml)                               |
| Test `exec` command                          | ✔    | [![Exec](https://github.com/emacs-eask/cli/actions/workflows/exec.yml/badge.svg)](https://github.com/emacs-eask/cli/actions/workflows/exec.yml)                                     |
| Test `emacs` command                         | ✔    | [![Emacs](https://github.com/emacs-eask/cli/actions/workflows/emacs.yml/badge.svg)](https://github.com/emacs-eask/cli/actions/workflows/emacs.yml)                                  |
| Test search packages                         | ✔    | [![Search](https://github.com/emacs-eask/cli/actions/workflows/search.yml/badge.svg)](https://github.com/emacs-eask/cli/actions/workflows/search.yml)                               |
| Test upgrade and check outdated packages     | ✔    | [![Outdated_Upgrade](https://github.com/emacs-eask/cli/actions/workflows/outdated_upgrade.yml/badge.svg)](https://github.com/emacs-eask/cli/actions/workflows/outdated_upgrade.yml) |
| Test `upgrade-eask` command                  | ✔    | [![Upgrade Eask](https://github.com/emacs-eask/cli/actions/workflows/upgrade-eask.yml/badge.svg)](https://github.com/emacs-eask/cli/actions/workflows/upgrade-eask.yml)             |

###### Options

| Description          | Done | Status                                                                                                                                                   |
|----------------------|------|----------------------------------------------------------------------------------------------------------------------------------------------------------|
| Test option switches | ✔    | [![Options](https://github.com/emacs-eask/cli/actions/workflows/options.yml/badge.svg)](https://github.com/emacs-eask/cli/actions/workflows/options.yml) |

###### Test

| Description               | Done | Status                                                                                                                                                                           |
|---------------------------|------|----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------|
| Test `ert` command        | ✔    | [![Test ert](https://github.com/emacs-eask/cli/actions/workflows/test_ert.yml/badge.svg)](https://github.com/emacs-eask/cli/actions/workflows/test_ert.yml)                      |
| Test `ert-runner` command | ✔    | [![Test ert-runner](https://github.com/emacs-eask/cli/actions/workflows/test_ert-runner.yml/badge.svg)](https://github.com/emacs-eask/cli/actions/workflows/test_ert-runner.yml) |
| Test `buttercup` command  | ✔    | [![Test buttercup](https://github.com/emacs-eask/cli/actions/workflows/test_buttercup.yml/badge.svg)](https://github.com/emacs-eask/cli/actions/workflows/test_buttercup.yml)    |
| Test `ecukes` command     | ✔    | [![Test ecukes](https://github.com/emacs-eask/cli/actions/workflows/test_ecukes.yml/badge.svg)](https://github.com/emacs-eask/cli/actions/workflows/test_ecukes.yml)             |

###### Others

| Description | Done | Status                                                                                                                                                               |
|-------------|------|----------------------------------------------------------------------------------------------------------------------------------------------------------------------|
| Webinstall  | ✔    | [![Webinstall](https://github.com/emacs-eask/cli/actions/workflows/webinstall.yml/badge.svg)](https://github.com/emacs-eask/cli/actions/workflows/webinstall.yml)    |
| Exit Status | ✔    | [![Exit Status](https://github.com/emacs-eask/cli/actions/workflows/exit_status.yml/badge.svg)](https://github.com/emacs-eask/cli/actions/workflows/exit_status.yml) |

## ⚜️ License

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

See [`LICENSE`](./LICENSE) for details.


<!-- Links -->

[Cask]: https://github.com/cask/cask
[makem.sh]: https://github.com/alphapapa/makem.sh
[Eldev]: https://github.com/doublep/eldev

[node]: https://nodejs.org/

CLI.pm	- Simple command line interpreter

CLI.pm is a module to create simple command line programs with simple
variable and command settings. It supports settig simple variables,
arrays of values and running commands which use those variables.

The documentation is almost non-existant. This started as a private
project but has been placed on CPAN in case it is usable by others. If
it is the author will attempt to improve the doco. Look at test.pl for
an example of usage. interactive.pl is the same code but runs
interactively for better feeling for how it works.

Requirements:
-------------

Perl and the Astro::Time module from CPAN


Where can I get it from?
------------------------

  http://search.cpan.org/~cphil/

Installation:
------------

 % perl Makefile.PL
 % make
 % make test
 % make install

Changes:
--------

See the Changes file

Author:
-------

Chris Phillips, CSIRO, Australia Telescope National Facility

 Chris.Phillips@csiro.au

Acknowledgment:
---------------

Copyright
---------

This module is copyright (C) 2009-2015 Chris Phillips.  All rights
reserved.  This program is free software; you can redistribute it
and/or modify it under the same terms as Perl itself.

### Systemd unit files for `step`

For renewal timer documentation, see [Automating Certificate Renewal](https://smallstep.com/docs/step-ca/renewal)

For convenience, these files are redirect targets in the files.smallstep.com S3 bucket.
If you change their locations here, please update the redirects in that bucket.

# Errors

First, read the [pkg/output/README.md](Output) guide for instructions on formatting user-facing strings.

## Creating an Error
### Message Format
Error message format
- short one line description describing what the error is
- do not capitalize the first letter (unless the first word is a name)
- do not add full stop at the end
- the variable name must end with *ErrorMsg*

Any suggestions of the type of actions the users should take, or any longer description should be kept in the suggestions section.
Suggestion message is not required.
Suggestions format
- full sentence
- capitalize first letter
- end with a full stop
- separate suggestions using a newline; do not add a newline at the end of the last sentence
- for sentences ending in a url, put a space between the url and the fullstop
- the variable name must end with *Suggestions*

examples:
```
> ./confluent ksql cluster create kk --cluster lkc-asfdsaf
Error: Kafka cluster "lkc-asfdsaf" not found

Suggestions:
    List Kafka clusters with `confluent kafka cluster list`.
```
```
Error: no API key selected for resource "lkc-dvnr7"

Suggestions:
    Select an API key for resource "lkc-dvnr7" with `confluent api-key use <API_KEY>`.
    To do so, you must have either already created or stored an API key for the resource.
    To create an API key use `confluent api-key create --resource lkc-dvnr7`.
    To store an existing API key use `confluent api-key store --resource lkc-dvnr7`.
```
```
Error: Kafka cluster "lkc-yydnp" not ready

Suggestions:
    It may take up to 5 minutes for a recently created Kafka cluster to be ready.
```

## Initializing the errors
There are four ways to create errors for the CLI.

1. All basic errors will fall under this category. Use one of the error intializing functions in the errors package (errors.New, errors.Errorf, errors.Wrap, errors.Wrapf), with an error message defined in [error_message.go](error_message.go). The name of the variable must end with *ErrorMsg*.
```
fmt.Errorf(errors.AuthorizeAccountsErrorMsg, accountsStr)
```
2. For errors with suggestions, define the error message, and suggestions message next to each other in [error_message.go](error_message.go). The messages must have the same name with different ending following the naming convention (i.e. *ErrorMsg* and *Suggestions*). Then either `errors.NewErrorWithSuggestion` or `errors.NewWrapErrorWithSuggestion` is used to initialize the error.

```
ResolvingConfigPathErrorMsg        = "error resolving the config filepath at \"%s\" has occurred"
ResolvingConfigPathSuggestions     = "Try moving the config file to a different location."

return "", errors.NewErrorWithSuggestions(fmt.Sprintf(errors.ResolvingConfigPathErrorMsg, c.Filename), errors.ResolvingConfigPathSuggestions)
```

```
LookUpRoleErrorMsg              = "failed to look up role \"%s\""
LookUpRoleSuggestions           = "To check for valid roles, use `confluent iam rbac role list`."

return errors.NewWrapErrorWithSuggestions(err, fmt.Sprintf(errors.LookUpRoleErrorMsg, roleName), errors.LookUpRoleSuggestions)
```

3. If you know that your error will be used in many places, or needs to be caught downstream and processed later, you can define a typed error by implenting the `CLITypedError` interface
```
type CLITypedError interface {
	error
	UserFacingError() error
}
```
  See [typed.go](typed.go) for examples and interface definition.

4. For important errors that are thrown by external packages that need to be caught and translated, define a catcher in [catcher.go](catcher.go). The catcher can then be either inserted anyhwere in the CLI repo, or simply in `handleErrors` function in [handle.go](handle.go).

## HandleCommon
`errors.HandleCommon` is called for every cobra RunE or PrerunE command, to handle common logic required for all errors, including turning off the usage message.

## Non-error communcation
For all non-error messages, define the string variables in [strings.go](strings.go), with variables name ending with *Msg*. The only exception being that warning messages are defined in [warning_message.go](warning_message.go) instead.

# How to add a new Command

Before making any changes, please consult the [CLI style guide](https://github.com/urfave/cli)!

### Package Layout

The [`urfave/cli`](https://github.com/urfave/cli) package that forms the basis
of Step CLI supports N-levels of command hierarchy. Each level of the hierarchy
should exist within its own package if possible. For example, `version` and
`help` exist inside their own packages inside the top-level `command` package.

Any package used by a command but does not contain explicit business logic
directly related to the command should exist in the top-level of this
repository. For example, the `github.com/smallstep/cli/flags` and
`go.step.sm/cli-utils/errs` package are used by many different commands and
contain functionality for defining flags and creating/manipulating errors.

### Adding a Command

Once you figured out where to add the command inside the package hierarchy you
must register the command. This way the command *can* be made available if
desired inside the `cmd/step/main.go`.

An example of defining a command and registering it:

```golang
package validate

import (
  "github.com/urfave/cli"

  "github.com/smallstep/cli/command"
  "github.com/smallstep/cli/flags"
)

func init() {
  cmd := cli.Command{
    Name: "validate",
    Usage: "Returns whether or not the provided token is valid",
    Flags: []cli.Flag{
      flags.Token("The one-time token value to validate"),
    },
    Action: validate,
  }

  command.Register(validate)
}
```

Once this is done, you then must import the pkg inside `cmd/step/main.go` so
the packages `init` method is run appropriately. This only needs to be done for
top-level commands.

```golang
package main

import (
  "github.com/urfave/cli"

  _ "github.com/smallstep/cli/validate"
)
```

This will ensure that the `smallstep/cli/validate` package is initialized
and thus registered with the `smallstep/cli/command`.

### Usage, Flags, Errors, and Prompts

There are three packages which contain functionality to make writing commands easier:

- `github.com/smallstep/cli/usage`
- `github.com/smallstep/cli/flags`
- `github.com/smallstep/cli/prompts`
- `go.step.sm/cli-utils/errs`

The usage package is used to extend the default documentation provided by
`urfave/cli` by enabling us to document arguments, whether they are optional or
required, and ensuring they're printed out as a part of the `step help` or
`step <command> -h` flow. If you need to add a different type of annotation to
document an argument just add it to the `usage.Argument` struct!

When you add a flag, look into the pre-existing ones inside the `flags`
package. Could you use one of the pre-existing flags in order to reduce
duplication? If not, make sure to add a flag so it could be used in future!

The `errs` package contains functionality for defining and working with errors
to ensure they are mutated properly into a `urfave/cli.ExitError` which ensures
the process returns an appropriate exit code on termination. When you create an
error, consider whether or not it's general and could be predefined inside the
`errs` package. Errors that are specific to the command itself should exist
only inside that commands respective package.

The `prompts` package is a small wrapper around the various different types of
prompts used by the commands. If you need a new prompt, consider adding a new
function to this package to tailor the prompt for the step cli. This way other
commands can adopt the step aesthetic as new functionality is introduced.

### Hiding a Command

Sometimes it's desirable to prevent a command from showing up in the help menu
because it's been deprecated *or* it's not ready for users to leverage. This
can be achieved by setting the `Hidden` property on the `cli.Command` struct to
`true`.

# Powershell scripts

Please note that these scripts are referenced on the `files.smallstep.com` S3 website bucket as redirects to `raw.githubusercontent.com`. If you move them, please update the S3 redirects.

Blackfriday [![Build Status](https://travis-ci.org/russross/blackfriday.svg?branch=master)](https://travis-ci.org/russross/blackfriday)
===========

Blackfriday is a [Markdown][1] processor implemented in [Go][2]. It
is paranoid about its input (so you can safely feed it user-supplied
data), it is fast, it supports common extensions (tables, smart
punctuation substitutions, etc.), and it is safe for all utf-8
(unicode) input.

HTML output is currently supported, along with Smartypants
extensions.

It started as a translation from C of [Sundown][3].


Installation
------------

Blackfriday is compatible with any modern Go release. With Go 1.7 and git
installed:

    go get gopkg.in/russross/blackfriday.v2

will download, compile, and install the package into your `$GOPATH`
directory hierarchy. Alternatively, you can achieve the same if you
import it into a project:

    import "gopkg.in/russross/blackfriday.v2"

and `go get` without parameters.


Versions
--------

Currently maintained and recommended version of Blackfriday is `v2`. It's being
developed on its own branch: https://github.com/russross/blackfriday/v2. You
should install and import it via [gopkg.in][6] at
`gopkg.in/russross/blackfriday.v2`.

Version 2 offers a number of improvements over v1:

* Cleaned up API
* A separate call to [`Parse`][4], which produces an abstract syntax tree for
  the document
* Latest bug fixes
* Flexibility to easily add your own rendering extensions

Potential drawbacks:

* Our benchmarks show v2 to be slightly slower than v1. Currently in the
  ballpark of around 15%.
* API breakage. If you can't afford modifying your code to adhere to the new API
  and don't care too much about the new features, v2 is probably not for you.
* Several bug fixes are trailing behind and still need to be forward-ported to
  v2. See issue [#348](https://github.com/russross/blackfriday/issues/348) for
  tracking.

Usage
-----

For the most sensible markdown processing, it is as simple as getting your input
into a byte slice and calling:

```go
output := blackfriday.Run(input)
```

Your input will be parsed and the output rendered with a set of most popular
extensions enabled. If you want the most basic feature set, corresponding with
the bare Markdown specification, use:

```go
output := blackfriday.Run(input, blackfriday.WithNoExtensions())
```

### Sanitize untrusted content

Blackfriday itself does nothing to protect against malicious content. If you are
dealing with user-supplied markdown, we recommend running Blackfriday's output
through HTML sanitizer such as [Bluemonday][5].

Here's an example of simple usage of Blackfriday together with Bluemonday:

```go
import (
    "github.com/microcosm-cc/bluemonday"
    "github.com/russross/blackfriday"
)

// ...
unsafe := blackfriday.Run(input)
html := bluemonday.UGCPolicy().SanitizeBytes(unsafe)
```

### Custom options

If you want to customize the set of options, use `blackfriday.WithExtensions`,
`blackfriday.WithRenderer` and `blackfriday.WithRefOverride`.

You can also check out `blackfriday-tool` for a more complete example
of how to use it. Download and install it using:

    go get github.com/russross/blackfriday-tool

This is a simple command-line tool that allows you to process a
markdown file using a standalone program.  You can also browse the
source directly on github if you are just looking for some example
code:

* <http://github.com/russross/blackfriday-tool>

Note that if you have not already done so, installing
`blackfriday-tool` will be sufficient to download and install
blackfriday in addition to the tool itself. The tool binary will be
installed in `$GOPATH/bin`.  This is a statically-linked binary that
can be copied to wherever you need it without worrying about
dependencies and library versions.


Features
--------

All features of Sundown are supported, including:

*   **Compatibility**. The Markdown v1.0.3 test suite passes with
    the `--tidy` option.  Without `--tidy`, the differences are
    mostly in whitespace and entity escaping, where blackfriday is
    more consistent and cleaner.

*   **Common extensions**, including table support, fenced code
    blocks, autolinks, strikethroughs, non-strict emphasis, etc.

*   **Safety**. Blackfriday is paranoid when parsing, making it safe
    to feed untrusted user input without fear of bad things
    happening. The test suite stress tests this and there are no
    known inputs that make it crash.  If you find one, please let me
    know and send me the input that does it.

    NOTE: "safety" in this context means *runtime safety only*. In order to
    protect yourself against JavaScript injection in untrusted content, see
    [this example](https://github.com/russross/blackfriday#sanitize-untrusted-content).

*   **Fast processing**. It is fast enough to render on-demand in
    most web applications without having to cache the output.

*   **Thread safety**. You can run multiple parsers in different
    goroutines without ill effect. There is no dependence on global
    shared state.

*   **Minimal dependencies**. Blackfriday only depends on standard
    library packages in Go. The source code is pretty
    self-contained, so it is easy to add to any project, including
    Google App Engine projects.

*   **Standards compliant**. Output successfully validates using the
    W3C validation tool for HTML 4.01 and XHTML 1.0 Transitional.


Extensions
----------

In addition to the standard markdown syntax, this package
implements the following extensions:

*   **Intra-word emphasis suppression**. The `_` character is
    commonly used inside words when discussing code, so having
    markdown interpret it as an emphasis command is usually the
    wrong thing. Blackfriday lets you treat all emphasis markers as
    normal characters when they occur inside a word.

*   **Tables**. Tables can be created by drawing them in the input
    using a simple syntax:

    ```
    Name    | Age
    --------|------
    Bob     | 27
    Alice   | 23
    ```

*   **Fenced code blocks**. In addition to the normal 4-space
    indentation to mark code blocks, you can explicitly mark them
    and supply a language (to make syntax highlighting simple). Just
    mark it like this:

        ```go
        func getTrue() bool {
            return true
        }
        ```

    You can use 3 or more backticks to mark the beginning of the
    block, and the same number to mark the end of the block.

*   **Definition lists**. A simple definition list is made of a single-line
    term followed by a colon and the definition for that term.

        Cat
        : Fluffy animal everyone likes
        
        Internet
        : Vector of transmission for pictures of cats

    Terms must be separated from the previous definition by a blank line.

*   **Footnotes**. A marker in the text that will become a superscript number;
    a footnote definition that will be placed in a list of footnotes at the
    end of the document. A footnote looks like this:

        This is a footnote.[^1]
        
        [^1]: the footnote text.

*   **Autolinking**. Blackfriday can find URLs that have not been
    explicitly marked as links and turn them into links.

*   **Strikethrough**. Use two tildes (`~~`) to mark text that
    should be crossed out.

*   **Hard line breaks**. With this extension enabled newlines in the input
    translate into line breaks in the output. This extension is off by default.

*   **Smart quotes**. Smartypants-style punctuation substitution is
    supported, turning normal double- and single-quote marks into
    curly quotes, etc.

*   **LaTeX-style dash parsing** is an additional option, where `--`
    is translated into `&ndash;`, and `---` is translated into
    `&mdash;`. This differs from most smartypants processors, which
    turn a single hyphen into an ndash and a double hyphen into an
    mdash.

*   **Smart fractions**, where anything that looks like a fraction
    is translated into suitable HTML (instead of just a few special
    cases like most smartypants processors). For example, `4/5`
    becomes `<sup>4</sup>&frasl;<sub>5</sub>`, which renders as
    <sup>4</sup>&frasl;<sub>5</sub>.


Other renderers
---------------

Blackfriday is structured to allow alternative rendering engines. Here
are a few of note:

*   [github_flavored_markdown](https://godoc.org/github.com/shurcooL/github_flavored_markdown):
    provides a GitHub Flavored Markdown renderer with fenced code block
    highlighting, clickable heading anchor links.

    It's not customizable, and its goal is to produce HTML output
    equivalent to the [GitHub Markdown API endpoint](https://developer.github.com/v3/markdown/#render-a-markdown-document-in-raw-mode),
    except the rendering is performed locally.

*   [markdownfmt](https://github.com/shurcooL/markdownfmt): like gofmt,
    but for markdown.

*   [LaTeX output](https://bitbucket.org/ambrevar/blackfriday-latex):
    renders output as LaTeX.


Todo
----

*   More unit testing
*   Improve unicode support. It does not understand all unicode
    rules (about what constitutes a letter, a punctuation symbol,
    etc.), so it may fail to detect word boundaries correctly in
    some instances. It is safe on all utf-8 input.


License
-------

[Blackfriday is distributed under the Simplified BSD License](LICENSE.txt)


   [1]: https://daringfireball.net/projects/markdown/ "Markdown"
   [2]: https://golang.org/ "Go Language"
   [3]: https://github.com/vmg/sundown "Sundown"
   [4]: https://godoc.org/gopkg.in/russross/blackfriday.v2#Parse "Parse func"
   [5]: https://github.com/microcosm-cc/bluemonday "Bluemonday"
   [6]: https://labix.org/gopkg.in "gopkg.in"

#' SHA-256 hash
#'
#' Calculate the SHA-256 hash of each element of a character vector.
#'
#' @param x Character vector. If not a character vector, then
#' [as.character()] is used to try to coerce it into one. `NA` entries
#' will have an `NA` hash.
#' @return `hash_sha256()` returns a character vector of hexadecimal
#' SHA-256 hashes.

#' @family hash functions
#'
#' @export
#' @examples
#' hash_sha256(c("foo", NA, "bar", ""))

hash_sha256 <- function(x) {
  if (!is.character(x)) x <- as.character(x)
  na <- is.na(x)
  x[na] <- NA_character_
  x[!na] <- .Call(clic_sha256, x[!na])
  x
}

#' @export
#' @rdname hash_sha256
#' @details `hash_raw_sha256()` calculates the SHA-256 hash of the bytes
#' of a raw vector.
#'
#' @return `hash_raw_sha256()` returns a character scalar.

hash_raw_sha256 <- function(x) {
  stopifnot(is.raw(x))
  .Call(clic_sha256_raw, x)
}

#' @export
#' @rdname hash_sha256
#' @param serialize_version Workspace format version to use, see
#' [base::serialize()].
#' @details `hash_obj_sha256()` calculates the SHA-256 hash of an R
#' object. The object is serialized into a binary vector first.
#'
#' @return `hash_obj_sha256()` returns a character scalar.

hash_obj_sha256 <- function(x, serialize_version = 2) {
  sr <- serialize(x, NULL, version = serialize_version)[-(1:14)]
  hash_raw_sha256(sr)
}

#' @export
#' @rdname hash_sha256
#' @param paths Character vector of file names.
#' @details `hash_file_sha256()` calculates the SHA-256 hash of one or
#' more files.
#'
#' @return `hash_file_sha256()` returns a character vector of SHA-256
#' hashes.

hash_file_sha256 <- function(paths) {
  if (!is.character(paths)) paths <- as.character(paths)
  paths <- normalizePath(paths, mustWork = FALSE)
  if (is_windows()) {
    paths <- enc2utf8(paths)
  } else {
    paths <- enc2native(paths)
  }
  .Call(clic_sha256_file, paths)
}

#' SHA-1 hash
#'
#' Calculate the SHA-1 hash of each element of a character vector.
#'
#' @param x Character vector. If not a character vector, then
#' [as.character()] is used to try to coerce it into one. `NA` entries
#' will have an `NA` hash.
#' @return `hash_sha1()` returns a character vector of hexadecimal
#' SHA-1 hashes.

#' @family hash functions
#'
#' @export
#' @examples
#' hash_sha1(c("foo", NA, "bar", ""))

hash_sha1 <- function(x) {
  if (!is.character(x)) x <- as.character(x)
  na <- is.na(x)
  x[na] <- NA_character_
  x[!na] <- .Call(clic_sha1, x[!na])
  x
}

#' @export
#' @rdname hash_sha1
#' @details `hash_raw_sha1()` calculates the SHA-1 hash of the bytes
#' of a raw vector.
#'
#' @return `hash_raw_sha1()` returns a character scalar.

hash_raw_sha1 <- function(x) {
  stopifnot(is.raw(x))
  .Call(clic_sha1_raw, x)
}

#' @export
#' @rdname hash_sha1
#' @param serialize_version Workspace format version to use, see
#' [base::serialize()].
#' @details `hash_obj_sha1()` calculates the SHA-1 hash of an R
#' object. The object is serialized into a binary vector first.
#'
#' @return `hash_obj_sha1()` returns a character scalar.

hash_obj_sha1 <- function(x, serialize_version = 2) {
  sr <- serialize(x, NULL, version = serialize_version)[-(1:14)]
  hash_raw_sha1(sr)
}

#' @export
#' @rdname hash_sha1
#' @param paths Character vector of file names.
#' @details `hash_file_sha1()` calculates the SHA-1 hash of one or
#' more files.
#'
#' @return `hash_file_sha1()` returns a character vector of SHA-1
#' hashes.

hash_file_sha1 <- function(paths) {
  if (!is.character(paths)) paths <- as.character(paths)
  paths <- normalizePath(paths, mustWork = FALSE)
  if (is_windows()) {
    paths <- enc2utf8(paths)
  } else {
    paths <- enc2native(paths)
  }
  .Call(clic_sha1_file, paths)
}

#' MD5 hash
#'
#' Calculate the MD5 hash of each element of a character vector.
#'
#' @param x Character vector. If not a character vector, then
#' [as.character()] is used to try to coerce it into one. `NA` entries
#' will have an `NA` hash.
#' @return `hash_md5()` returns a character vector of hexadecimal MD5
#' hashes.
#'
#' @family hash functions
#' @seealso [tools::md5sum()] for a base R MD5 function that works on
#' files.
#'
#' @export
#' @examples
#' hash_md5(c("foo", NA, "bar", ""))

hash_md5 <- function(x) {
  if (!is.character(x)) x <- as.character(x)
  na <- is.na(x)
  x[na] <- NA_character_
  x[!na] <- .Call(clic_md5, x[!na])
  x
}

#' @export
#' @rdname hash_md5
#' @details `hash_raw_md5()` calculates the MD5 hash of the bytes
#' of a raw vector.
#'
#' @return `hash_raw_md5()` returns a character scalar.

hash_raw_md5 <- function(x) {
  stopifnot(is.raw(x))
  .Call(clic_md5_raw, x)
}

#' @export
#' @rdname hash_md5
#' @param serialize_version Workspace format version to use, see
#' [base::serialize()].
#' @details `hash_obj_md5()` calculates the MD5 hash of an R
#' object. The object is serialized into a binary vector first.
#'
#' @return `hash_obj_md5()` returns a character scalar.

hash_obj_md5 <- function(x, serialize_version = 2) {
  sr <- serialize(x, NULL, version = serialize_version)[-(1:14)]
  hash_raw_md5(sr)
}

#' @export
#' @rdname hash_md5
#' @param paths Character vector of file names.
#' @details `hash_file_md5()` calculates the MD5 hash of one or more
#' files.

hash_file_md5 <- function(paths) {
  if (!is.character(paths)) paths <- as.character(paths)
  paths <- normalizePath(paths, mustWork = FALSE)
  if (is_windows()) {
    paths <- enc2utf8(paths)
  } else {
    paths <- enc2native(paths)
  }
  .Call(clic_md5_file, paths)
}

#' Emoji hash
#'
#' @details
#' It uses the first 13 hexadecimal characters (out of the 32) of the MD5
#' hash of the input, and converts them into an emoji representation.
#' It uses a manually selected subset of all emojis, that tend to be
#' displayed correctly.
#'
#' ## Number of possible hash values
#'
#' ```{r include = FALSE}
#' hf <- function(size) {
#'   format(nrow(emojis)**size, big.mark = ",", scientific = FALSE)
#' }
#' ```
#'
#' cli uses `r nrow(emojis)` possible emojis. This is the number of
#' different hashes you can get for different values of `size`:
#'
#' | `size` | size of hash table space |
#' | -----: | -----------------------: |
#' | 1      | `r hf(1)`                |
#' | 2      | `r hf(2)`                |
#' | 3      | `r hf(3)`                |
#' | 4      | `r hf(4)`                |
#'
#' @param x Character vector. `NA` entries will have an `NA` hash.
#' @param size Number of emojis to use in a hash. Currently it has to
#'   be from 1 through 4.
#' @return `hash_emoji()` returns a data frame with columns
#'   * `hash`: the emoji hash, a string of the requested size.
#'   * `emojis`: list column with the emoji characters in character
#'     vectors. Note that an emoji might have multiple code points.
#'   * `text`: text representation of `hash`, comma separated.
#'   * `names`: list column with the text representations of `emojis`, in
#'     character vectors.
#'
#' @family hash functions
#' @seealso the emoji package for a comprehensive list of emojis
#' @export
#' @examples
#' hash_emoji(c("foo", NA, "bar", ""))$text
#'
#' # if you increase `size`, the shorter hash is a prefix of the longer:
#' hash_emoji("foobar", 1)$text
#' hash_emoji("foobar", 2)$text
#' hash_emoji("foobar", 3)$text
#' hash_emoji("foobar", 4)$text

hash_emoji <- function(x, size = 3) {
  # our integer arithmetic does not work if size > 4
  if (!is.character(x)) x <- as.character(x)
  stopifnot(
    is.character(x),
    is_count(size),
    size >= 1 && size <= 4
  )

  hashes <- lapply(x, hash_emoji1, size = size)
  emojis <- lapply(hashes, "[[", "emoji")
  names <- lapply(hashes, "[[", "names")

  data.frame(
    stringsAsFactors = FALSE,
    hash = vapply(emojis, hash_collapse, character(1)),
    emojis = I(emojis),
    text = vapply(names, hash_collapse, character(1), sep = ", "),
    names = I(names)
  )
}

hash_collapse <- function(x, sep = "") {
  if (anyNA(x)) {
    NA_character_
  } else {
    paste(x, collapse = sep)
  }
}

hash_emoji1 <- function(x, size = 3) {
  if (is.na(x)) {
    return(list(
      emoji = rep(NA_character_, size),
      names = rep(NA_character_, size)
    ))
  }

  md5 <- hash_md5(x)
  hash_emoji1_transform(md5, size)
}

hash_emoji1_transform <- function(md5, size) {
  md513 <- substr(md5, 1, 13)
  mdint <- as.integer(as.hexmode(strsplit(md513, "", fixed = TRUE)[[1]]))
  hash <- sum(mdint * 16^(0:12))

  base <- nrow(emojis)
  ehash <- hash %% (base ** size)
  digits <- integer()
  while (ehash > 0) {
    digits <-  c(digits, ehash %% base)
    ehash <- ehash %/% base
  }
  digits <- c(digits, rep(0, 10))[1:size]

  nms <- emojis$name[digits + 1]
  emo <- emojis$emoji[digits + 1]

  list(
    emoji = emo,
    names = nms
  )
}

#' @export
#' @rdname hash_emoji
#' @details `hash_raw_emoji()` calculates the emoji hash of the bytes
#' of a raw vector.
#'
#' @return `hash_raw_emoji()` and `hash_obj_emoji()` return a list with
#' entries:
#' * `hash`: the emoji hash, a string of requested size,
#' * `emojis`: the individual emoji characters in a character vector,
#' * `text`: text representation of `hash`, comma separated,
#' * `names`: names of the emojis, in a character vector.

hash_raw_emoji <- function(x, size = 3) {
  stopifnot(is.raw(x))
  md5 <- hash_raw_md5(x)
  emo <- hash_emoji1_transform(md5, size)

  list(
    hash = hash_collapse(emo$emoji),
    emojis = emo$emoji,
    text = hash_collapse(emo$emoji, sep = ", "),
    names = emo$names
  )
}

#' @export
#' @rdname hash_emoji
#' @param serialize_version Workspace format version to use, see
#' [base::serialize()].
#' @details `hash_obj_emoji()` calculates the emoji hash of an R
#' object. The object is serialized into a binary vector first.

hash_obj_emoji <- function(x, size = 3, serialize_version = 2) {
  sr <- serialize(x, NULL, version = serialize_version)[-(1:14)]
  hash_raw_emoji(sr, size = size)
}

#' Adjective-animal hash
#'
#' @details
#' It uses the first 13 hexadecimal characters (out of the 32) of the MD5
#' hash of the input, and converts them into an adjective-animal form to
#' create a human readable hash.
#'
#' ## Number of possible hash values
#'
#' ```{r include = FALSE}
#' hf <- function(n_adj) {
#'   format(
#'     length(gfycat_adjectives) ** n_adj * length(gfycat_animals),
#'     big.mark = ",",
#'     scientific = FALSE
#'   )
#' }
#' ```
#'
#' `hash_animal()` uses `r length(gfycat_animals)` animal names and
#' `r length(gfycat_adjectives)` different adjectives. The number of
#' different hashes you can get for different values of `n_adj`:
#'
#' | `n_adj` | size of the hash table space |
#' | ------: | ---------------------------: |
#' | 0       | `r hf(0)`                    |
#' | 1       | `r hf(1)`                    |
#' | 2       | `r hf(2)`                    |
#' | 3       | `r hf(3)`                    |
#'
#' ## Source
#'
#' The list of adjectives and animals comes from the ids package,
#' and in turn from
#' <https://github.com/a-type/adjective-adjective-animal>, and
#' from `https://gfycat.com` (now gone).
#'
#' @param x Character vector. `NA` entries will have an `NA` hash.
#' @param n_adj Number of adjectives to use. It must be from 0 through 3.
#' @return A data frame with columns
#'   * `hash`: the hash value, a string.
#'   * `words`: list column with the adjectives and the animal name in a
#'     character vector.
#'
#' @family hash functions
#' @seealso the ids package for generating random adjective-animal ids
#'
#' @export
#' @examples
#' hash_animal(c("foo", "bar"))
#'
#' # if you increase `n_adj`, the shorter hash is a suffix of the longer:
#' hash_animal("cli package", 0)$hash
#' hash_animal("cli package", 1)$hash
#' hash_animal("cli package", 2)$hash
#' hash_animal("cli package", 3)$hash

hash_animal <- function(x, n_adj = 2) {
  if (!is.character(x)) x <- as.character(x)
  stopifnot(
    is.character(x),
    is_count(n_adj),
    n_adj >= 0 && n_adj <= 3
  )

  hashes <- lapply(x, hash_animal1, n_adj = n_adj)

  data.frame(
    stringsAsFactors = FALSE,
    hash = vapply(hashes, hash_collapse, character(1), sep = " "),
    words = I(hashes)
  )
}

hash_animal1 <- function(x, n_adj = 2) {
  if (is.na(x)) {
    return(rep(NA_character_, n_adj + 1))
  }

  md5 <- hash_md5(x)
  hash_animal1_transform(md5, n_adj)
}

hash_animal1_transform <- function(md5, n_adj) {
  md513 <- substr(md5, 1, 13)
  mdint <- as.integer(as.hexmode(strsplit(md513, "", fixed = TRUE)[[1]]))
  hash <- sum(mdint * 16^(0:12))

  len_ani <- length(gfycat_animals)
  len_adj <- length(gfycat_adjectives)
  ehash <- hash %% (len_adj ** n_adj * len_ani)
  digits <- ehash %% len_ani
  ehash <- ehash %/% len_ani

  while (ehash > 0) {
    digits <- c(digits, ehash %% len_adj)
    ehash <- ehash %/% len_adj
  }
  digits <- c(digits, rep(0, 10))[1:(n_adj + 1)]
  digits <- rev(digits)

  c(
    gfycat_adjectives[digits[-length(digits)] + 1],
    gfycat_animals[digits[length(digits)] + 1]
  )
}

#' @export
#' @rdname hash_animal
#' @details `hash_raw_animal()` calculates the adjective-animal hash of
#' the bytes of a raw vector.
#'
#' @return `hash_raw_animal()` and `hash_obj_animal()` return a list
#' with entries:
#' * `hash`: the hash value, a string,
#' * `words: the adjectives and the animal name in a character vector.

hash_raw_animal <- function(x, n_adj = 2) {
  stopifnot(is.raw(x))
  md5 <- hash_raw_md5(x)
  hash <- hash_animal1_transform(md5, n_adj)

  list(
    hash = hash_collapse(hash, sep = ", "),
    words = hash
  )
}

#' @export
#' @rdname hash_animal
#' @param serialize_version Workspace format version to use, see
#' [base::serialize()].
#' @details `hash_obj_animal()` calculates the adjective-animal hash of
#' an R object. The object is serialized into a binary vector first.

hash_obj_animal <- function(x, n_adj = 2, serialize_version = 2) {
  sr <- serialize(x, NULL, version = serialize_version)[-(1:14)]
  hash_raw_animal(sr, n_adj = n_adj)
}

#' xxHash
#'
#' Extremely fast hash algorithm.
#'
#' @param x Character vector. If not a character vector, then
#' [as.character()] is used to try to coerce it into one. `NA` entries
#' will have an `NA` hash.
#' @return `hash_xxhash()` returns a character vector of hexadecimal
#' xxHash hashes.
#'
#' @family hash functions
#'
#' @export
#' @examples
#' hash_xxhash(c("foo", NA, "bar", ""))

hash_xxhash <- function(x) {
  if (!is.character(x)) x <- as.character(x)
  na <- is.na(x)
  x[na] <- NA_character_
  x[!na] <- .Call(clic_xxhash, x[!na])
  x
}

#' @export
#' @rdname hash_xxhash
#' @details `hash_raw_xxhash()` calculates the xxHash hash of the bytes
#' of a raw vector.
#' @return `hash_raw_xxhash()` returns a character scalar.

hash_raw_xxhash <- function(x) {
  stopifnot(is.raw(x))
  .Call(clic_xxhash_raw, x)
}

#' @export
#' @rdname hash_xxhash
#' @param serialize_version Workspace format version to use, see
#' [base::serialize()].
#' @details `hash_obj_xxhash()` calculates the xxHash hash of an R
#' object. The object is serialized into a binary vector first.
#' @return `hash_obj_xxhash()` returns a character scalar.

hash_obj_xxhash <- function(x, serialize_version = 2) {
  sr <- serialize(x, NULL, version = serialize_version)[-(1:14)]
  hash_raw_xxhash(sr)
}

#' @export
#' @rdname hash_xxhash
#' @param paths Character vector of file names.
#' @details `hash_file_xxhash()` calculates the xxHash hash of one or
#' more files.
#'
#' @return `hash_file_xxhash()` returns a character vector of xxHash
#' hashes.

hash_file_xxhash <- function(paths) {
  if (!is.character(paths)) paths <- as.character(paths)
  paths <- normalizePath(paths, mustWork = FALSE)
  if (is_windows()) {
    paths <- enc2utf8(paths)
  } else {
    paths <- enc2native(paths)
  }
  .Call(clic_xxhash_file, paths)
}

#' @export
#' @rdname hash_xxhash
#' @details The `64` functions caculate the 64 bit variant
#' of xxHash. Otherwise they work the same.

hash_xxhash64 <- function(x) {
  if (!is.character(x)) x <- as.character(x)
  na <- is.na(x)
  x[na] <- NA_character_
  x[!na] <- .Call(clic_xxhash64, x[!na])
  x
}

#' @export
#' @rdname hash_xxhash

hash_raw_xxhash64 <- function(x) {
  stopifnot(is.raw(x))
  .Call(clic_xxhash64_raw, x)
}

#' @export
#' @rdname hash_xxhash

hash_obj_xxhash64 <- function(x, serialize_version = 2) {
  sr <- serialize(x, NULL, version = serialize_version)[-(1:14)]
  hash_raw_xxhash64(sr)
}

#' @export
#' @rdname hash_xxhash

hash_file_xxhash64 <- function(paths) {
  if (!is.character(paths)) paths <- as.character(paths)
  paths <- normalizePath(paths, mustWork = FALSE)
  if (is_windows()) {
    paths <- enc2utf8(paths)
  } else {
    paths <- enc2native(paths)
  }
  .Call(clic_xxhash64_file, paths)
}

test_that("hash_sha256", {
  dig <- function(x) {
    digest::digest(x, serialize = FALSE, algo = "sha256")
  }

  cases <- list(
    list(character(), character()),
    list("", dig("")),
    list("x", dig("x")),
    list(NA_character_, NA_character_),
    list(NA, NA_character_),
    list(
      c(NA, "", "foo", NA),
      c(NA, dig(""), dig("foo"), NA)
    )
  )

  for (case in cases) {
    expect_equal(hash_sha256(case[[1]]), case[[2]])
  }
})

test_that("hash_raw_sha256", {
  dig <- function(x) {
    digest::digest(x, serialize = FALSE, algo = "sha256")
  }

  cases <- list(
    list(raw(), dig(raw())),
    list(as.raw(0), dig(as.raw(0))),
    list(charToRaw("foobar"), dig("foobar"))
  )

  for (case in cases) {
    expect_equal(hash_raw_sha256(case[[1]]), case[[2]])
  }
})

test_that("hash_obj_sha256", {
  dig <- function(x) {
    digest::digest(x, serializeVersion = 2, algo = "sha256")
  }

  cases <- list(
    "",
    raw(0),
    1:10,
    mtcars
  )

  for (case in cases) {
    expect_equal(hash_obj_sha256(case), dig(case))
  }
})

test_that("hash_file_sha256", {
  dig <- function(x) {
    digest::digest(file = x, algo = "sha256")
  }

  f <- test_path("test-hash.R")
  expect_equal(
    hash_file_sha256(character()),
    character()
  )

  expect_equal(hash_file_sha256(f), dig(f))
})

test_that("hash_md5", {
  dig <- function(x) {
    digest::digest(x, serialize = FALSE, algo = "md5")
  }

  cases <- list(
    list(character(), character()),
    list("", dig("")),
    list("x", dig("x")),
    list(NA_character_, NA_character_),
    list(NA, NA_character_),
    list(
      c(NA, "", "foo", NA),
      c(NA, dig(""), dig("foo"), NA)
    )
  )

  for (case in cases) {
    expect_equal(hash_md5(case[[1]]), case[[2]])
  }
})

test_that("hash_raw_md5", {
  dig <- function(x) {
    digest::digest(x, serialize = FALSE, algo = "md5")
  }

  cases <- list(
    list(raw(), dig(raw())),
    list(as.raw(0), dig(as.raw(0))),
    list(charToRaw("foobar"), dig("foobar"))
  )

  for (case in cases) {
    expect_equal(hash_raw_md5(case[[1]]), case[[2]])
  }
})

test_that("hash_obj_md5", {
  dig <- function(x) {
    digest::digest(x, serializeVersion = 2, algo = "md5")
  }

  cases <- list(
    "",
    raw(0),
    1:10,
    mtcars
  )

  for (case in cases) {
    expect_equal(hash_obj_md5(case), dig(case))
  }
})

test_that("hash_file_md5", {
  dig <- function(x) {
    digest::digest(file = x, algo = "md5")
  }

  f <- test_path("test-hash.R")
  expect_equal(
    hash_file_md5(character()),
    character()
  )

  expect_equal(hash_file_md5(f), dig(f))
  expect_equal(hash_file_md5(c(f, f)), c(dig(f), dig(f)))
})

test_that("hash_emoji", {
  expect_snapshot({
    hash_emoji(character())$names
    hash_emoji("")$names
    hash_emoji("x")$names
    hash_emoji(NA_character_)$names
    hash_emoji(NA)$names
    hash_emoji(c(NA, "", "foo", NA))$names
  })
})

test_that("hash_raw_emoji", {
  expect_snapshot({
    hash_raw_emoji(raw())$names
    hash_raw_emoji(as.raw(0))$names
    hash_raw_emoji(charToRaw("foobar"))$names
  })
})

test_that("hash_obj_emoji", {
  expect_snapshot({
    hash_obj_emoji("")$names
    hash_obj_emoji(raw(0))$names
    hash_obj_emoji(1:10)$names
    hash_obj_emoji(mtcars)$names
  })
})

test_that("hash_animal", {
  expect_snapshot({
    hash_animal(character())$words
    hash_animal("")$words
    hash_animal("x")$words
    hash_animal(NA_character_)$words
    hash_animal(NA)$words
    hash_animal(c(NA, "", "foo", NA))$words
  })
})

test_that("hash_raw_animal", {
  expect_snapshot({
    hash_raw_animal(raw())$words
    hash_raw_animal(as.raw(0))$words
    hash_raw_animal(charToRaw("foobar"))$words
  })
})

test_that("hash_obj_animal", {
  expect_snapshot({
    hash_obj_animal("")$words
    hash_obj_animal(raw(0))$words
    hash_obj_animal(1:10)$words
    hash_obj_animal(mtcars)$words
  })
})

test_that("hash_xxhash", {
  expect_snapshot({
    hash_xxhash(letters[1:5])
    hash_xxhash(c("a", NA, "b"))
    hash_xxhash64(letters[1:5])
    hash_raw_xxhash(charToRaw("a"))
    hash_raw_xxhash64(charToRaw("a"))
    hash_obj_xxhash(raw(0))
    hash_obj_xxhash64(raw(0))
  })
  tmp <- tempfile()
  on.exit(unlink(tmp), add = TRUE)
  writeBin(charToRaw("a"), tmp)
  expect_snapshot({
    hash_file_xxhash(tmp)
    hash_file_xxhash64(tmp)
  })
})

test_that("hash_sha256", {
  dig <- function(x) {
    digest::digest(x, serialize = FALSE, algo = "sha256")
  }

  cases <- list(
    list(character(), character()),
    list("", dig("")),
    list("x", dig("x")),
    list(NA_character_, NA_character_),
    list(NA, NA_character_),
    list(
      c(NA, "", "foo", NA),
      c(NA, dig(""), dig("foo"), NA)
    )
  )

  for (case in cases) {
    expect_equal(hash_sha256(case[[1]]), case[[2]])
  }
})

test_that("hash_raw_sha256", {
  dig <- function(x) {
    digest::digest(x, serialize = FALSE, algo = "sha256")
  }

  cases <- list(
    list(raw(), dig(raw())),
    list(as.raw(0), dig(as.raw(0))),
    list(charToRaw("foobar"), dig("foobar"))
  )

  for (case in cases) {
    expect_equal(hash_raw_sha256(case[[1]]), case[[2]])
  }
})

test_that("hash_obj_sha256", {
  dig <- function(x) {
    digest::digest(x, serializeVersion = 2, algo = "sha256")
  }

  cases <- list(
    "",
    raw(0),
    1:10,
    mtcars
  )

  for (case in cases) {
    expect_equal(hash_obj_sha256(case), dig(case))
  }
})

test_that("hash_file_sha256", {
  dig <- function(x) {
    digest::digest(file = x, algo = "sha256")
  }

  f <- test_path("test-hash.R")
  expect_equal(
    hash_file_sha256(character()),
    character()
  )

  expect_equal(hash_file_sha256(f), dig(f))
})

test_that("hash_md5", {
  dig <- function(x) {
    digest::digest(x, serialize = FALSE, algo = "md5")
  }

  cases <- list(
    list(character(), character()),
    list("", dig("")),
    list("x", dig("x")),
    list(NA_character_, NA_character_),
    list(NA, NA_character_),
    list(
      c(NA, "", "foo", NA),
      c(NA, dig(""), dig("foo"), NA)
    )
  )

  for (case in cases) {
    expect_equal(hash_md5(case[[1]]), case[[2]])
  }
})

test_that("hash_raw_md5", {
  dig <- function(x) {
    digest::digest(x, serialize = FALSE, algo = "md5")
  }

  cases <- list(
    list(raw(), dig(raw())),
    list(as.raw(0), dig(as.raw(0))),
    list(charToRaw("foobar"), dig("foobar"))
  )

  for (case in cases) {
    expect_equal(hash_raw_md5(case[[1]]), case[[2]])
  }
})

test_that("hash_obj_md5", {
  dig <- function(x) {
    digest::digest(x, serializeVersion = 2, algo = "md5")
  }

  cases <- list(
    "",
    raw(0),
    1:10,
    mtcars
  )

  for (case in cases) {
    expect_equal(hash_obj_md5(case), dig(case))
  }
})

test_that("hash_emoji", {
  expect_snapshot({
    hash_emoji(character())$names
    hash_emoji("")$names
    hash_emoji("x")$names
    hash_emoji(NA_character_)$names
    hash_emoji(NA)$names
    hash_emoji(c(NA, "", "foo", NA))$names
  })
})

test_that("hash_raw_emoji", {
  expect_snapshot({
    hash_raw_emoji(raw())$names
    hash_raw_emoji(as.raw(0))$names
    hash_raw_emoji(charToRaw("foobar"))$names
  })
})

test_that("hash_obj_emoji", {
  expect_snapshot({
    hash_obj_emoji("")$names
    hash_obj_emoji(raw(0))$names
    hash_obj_emoji(1:10)$names
    hash_obj_emoji(mtcars)$names
  })
})

test_that("hash_animal", {
  expect_snapshot({
    hash_animal(character())$words
    hash_animal("")$words
    hash_animal("x")$words
    hash_animal(NA_character_)$words
    hash_animal(NA)$words
    hash_animal(c(NA, "", "foo", NA))$words
  })
})

test_that("hash_raw_animal", {
  expect_snapshot({
    hash_raw_animal(raw())$words
    hash_raw_animal(as.raw(0))$words
    hash_raw_animal(charToRaw("foobar"))$words
  })
})

test_that("hash_obj_animal", {
  expect_snapshot({
    hash_obj_animal("")$words
    hash_obj_animal(raw(0))$words
    hash_obj_animal(1:10)$words
    hash_obj_animal(mtcars)$words
  })
})

#' @useDynLib progresstestcpp, .registration = TRUE
NULL

#' @export

test_baseline <- function() {
  test_baseline_()
}

#' @export

test_cli <- function() {
  test_cli_()
}

#' @export

test_template <- function() {
  test_template_()
}

#' @useDynLib cli, .registration=TRUE
NULL

## nocov start

dummy <- function() { }

cli_timer_dynamic <- 200L
cli_timer_non_dynamic <- 3000L

clienv <- new.env(parent = emptyenv())
clienv$pid <- Sys.getpid()
clienv$globalenv <- format(.GlobalEnv)
clienv$status <- list()
clienv$progress <- list()
clienv$progress_ids <- list()

clienv$load_time <- NULL
clienv$speed_time <- 1.0
clienv$tick_time <- 200L

task_callback <- NULL

clienv$unloaded <- FALSE

rstudio_r_fix <- 0

.onLoad <- function(libname, pkgname) {

  err$onload_hook()

  # Try to restore cursor as much as we can
  if (Sys.getenv("R_CLI_HIDE_CURSOR") != "false" && isatty(stdout())) {
    reg.finalizer(clienv, function(e) cli::ansi_show_cursor(), TRUE)
    task_callback <<- addTaskCallback(
      function(...) { cli::ansi_show_cursor(); TRUE },
      "cli-show-cursor"
    )
  }

  # https://github.com/r-lib/cli/issues/352
  rstudio_r_fix <<- (Sys.getenv("RSTUDIO") == 1) + 0L

  pkgenv <- environment(dummy)

  clienv$load_time <- Sys.time()

  clienv$speed_time <- as.double(Sys.getenv("CLI_SPEED_TIME", "1.0"))

  tt <- as.integer(Sys.getenv("CLI_TICK_TIME", NA_character_))
  if (is.na(tt)) {
    tt <- if (interactive() || is_dynamic_tty()) {
      cli_timer_dynamic
    } else {
      cli_timer_non_dynamic
    }
  }

  clienv$tick_time <- as.integer(tt)
  .Call(
    clic_start_thread,
    pkgenv,
    clienv$tick_time,
    clienv$speed_time
  )

  # For valgrind: https://github.com/r-lib/cli/issues/311
  reg.finalizer(asNamespace("cli"), function(x) x$unload(), TRUE)

  if (getRversion() >= "3.5.0") {
    `__cli_update_due` <<- .Call(clic_make_timer);
  } else {
    rm("__cli_update_due", envir = pkgenv)
    makeActiveBinding(
      "__cli_update_due",
      function() .Call(clic_update_due),
      pkgenv
    )
  }

  ccli_tick_reset <<- clic_tick_reset

  makeActiveBinding(
    "symbol",
    function() {
      ## If `cli.unicode` is set we use that
      opt <- getOption("cli.unicode",  NULL)
      if (!is.null(opt)) {
        if (isTRUE(opt)) {
          return(symbol_utf8)
        } else {
          return(symbol_ascii)
        }
      }

      ## Otherwise we try to auto-detect
      rst <- rstudio$detect()$type
      rok <- c("rstudio_console", "rstudio_console_starting")
      if (is_utf8_output()) {
        symbol_utf8
      } else if (is_latex_output()) {
        symbol_ascii
      } else {
        symbol_ascii
      }
    },
    pkgenv
  )

  makeActiveBinding("pb_bar",            cli__pb_bar,           pkgenv)
  makeActiveBinding("pb_current",        cli__pb_current,       pkgenv)
  makeActiveBinding("pb_current_bytes",  cli__pb_current_bytes, pkgenv)
  makeActiveBinding("pb_elapsed",        cli__pb_elapsed,       pkgenv)
  makeActiveBinding("pb_elapsed_clock",  cli__pb_elapsed_clock, pkgenv)
  makeActiveBinding("pb_elapsed_raw",    cli__pb_elapsed_raw,   pkgenv)
  makeActiveBinding("pb_eta",            cli__pb_eta,           pkgenv)
  makeActiveBinding("pb_eta_raw",        cli__pb_eta_raw,       pkgenv)
  makeActiveBinding("pb_eta_str",        cli__pb_eta_str,       pkgenv)
  makeActiveBinding("pb_extra",          cli__pb_extra,         pkgenv)
  makeActiveBinding("pb_id",             cli__pb_id,            pkgenv)
  makeActiveBinding("pb_name",           cli__pb_name,          pkgenv)
  makeActiveBinding("pb_percent",        cli__pb_percent,       pkgenv)
  makeActiveBinding("pb_pid",            cli__pb_pid,           pkgenv)
  makeActiveBinding("pb_rate",           cli__pb_rate,          pkgenv)
  makeActiveBinding("pb_rate_raw",       cli__pb_rate_raw,      pkgenv)
  makeActiveBinding("pb_rate_bytes",     cli__pb_rate_bytes,    pkgenv)
  makeActiveBinding("pb_spin",           cli__pb_spin,          pkgenv)
  makeActiveBinding("pb_status",         cli__pb_status,        pkgenv)
  makeActiveBinding("pb_timestamp",      cli__pb_timestamp,     pkgenv)
  makeActiveBinding("pb_total",          cli__pb_total,         pkgenv)
  makeActiveBinding("pb_total_bytes",    cli__pb_total_bytes,   pkgenv)

  if (is.null(getOption("callr.condition_handler_cli_message"))) {
    options(callr.condition_handler_cli_message = cli__default_handler)
  }
}

unload <- function() {
  if (!clienv$unloaded) .Call(clic_unload)
  clienv$unloaded <- TRUE
}

.onUnload <- function(libpath) {
  tryCatch(removeTaskCallback(task_callback), error = function(e) NULL)
  tryCatch(cli_progress_cleanup(), error = function(e) NULL)
  tryCatch(ansi_show_cursor(), error = function(e) NULL)
  unload()
}

## nocov end

start_app()
on.exit(stop_app(), add = TRUE)

test_that_cli("bullets", {
  expect_snapshot(cli_bullets(c(
          "noindent",
    " " = "space",
    "v" = "success",
    "x" = "danger",
    "!" = "warning",
    "i" = "info",
    "*" = "bullet",
    ">" = "arrow"
  )))
})

test_that_cli("bullets glue", {
  expect_snapshot(cli_bullets(c(
          "noindent {.key {1:3}}",
    " " = "space {.key {1:3}}",
    "v" = "success {.key {1:3}}",
    "x" = "danger {.key {1:3}}",
    "!" = "warning {.key {1:3}}",
    "i" = "info {.key {1:3}}",
    "*" = "bullet {.key {1:3}}",
    ">" = "arrow {.key {1:3}}"
  )))
})

test_that_cli("bullets wrapping", {
  txt <- strrep("This is some text that is longer than the width. ", 3)
  expect_snapshot(cli_bullets(c(
    txt,
    " " = txt,
    "v" = txt,
    "x" = txt,
    "!" = txt,
    "i" = txt,
    "*" = txt,
    ">" = txt
  )))
})

#' ANSI colored text
#'
#' cli has a number of functions to color and style text at the command
#' line. They provide a more modern interface than the crayon package.
#'
#' The `col_*` functions change the (foreground) color to the text.
#' These are the eight original ANSI colors. Note that in some terminals,
#' they might actually look differently, as terminals have their own
#' settings for how to show them. `col_none()` is the default color, this
#' is useful in a substring of a colored string.
#'
#' The `col_br_*` functions are bright versions of the eight ANSI colors.
#' Note that on some terminal configurations and themes they might be the
#' same as the non-bright colors.
#'
#' The `bg_*` functions change the background color of the text.
#' These are the eight original ANSI background colors. These, too, can
#' vary in appearance, depending on terminal settings. `bg_none()` the
#' the default background color, this is useful in a substring of a
#' background-colored string.
#'
#' The `bg_br_*` functions are the bright versions of the eight ANSI
#' background colors. Note that on some terminal configurations and themes
#' they might be the same as the non-bright colors.
#'
#' The `style_*` functions apply other styling to the text. The currently
#' supported styling functions are:
#' * `style_reset()` to remove any style, including color,
#' * `style_bold()` for boldface / strong text, although some terminals
#'   show a bright, high intensity text instead,
#' * `style_dim()` (or `style_blurred()` reduced intensity text.
#' * `style_italic()` (not widely supported).
#' * `style_underline()`,
#' * `style_inverse()`,
#' * `style_hidden()`,
#' * `style_strikethrough()` (not widely supported).
#'
#' The style functions take any number of character vectors as arguments,
#' and they concatenate them using `paste0()` before adding the style.
#'
#' Styles can also be nested, and then inner style takes precedence, see
#' examples below.
#'
#' Sometimes you want to revert back to the default text color, in the
#' middle of colored text, or you want to have a normal font in the middle
#' of italic text. You can use the `style_no_*` functions for this. Every
#' `style_*()` function has a `style_no_*()` pair, which defends its
#' argument from taking on the style. See examples below.
#'
#' @param ... Character strings, they will be pasted together with
#'   `paste0()`, before applying the style function.
#' @return An ANSI string (class `cli_ansi_string`), that contains ANSI
#'   sequences, if the current platform supports them. You can simply
#'   use `cat()` to print them to the terminal.
#'
#' @family ANSI styling
#' @name ansi-styles
#' @examples
#' col_blue("Hello ", "world!")
#' cat(col_blue("Hello ", "world!"))
#'
#' cat("... to highlight the", col_red("search term"),
#'     "in a block of text\n")
#'
#' ## Style stack properly
#' cat(col_green(
#'  "I am a green line ",
#'  col_blue(style_underline(style_bold("with a blue substring"))),
#'  " that becomes green again!"
#' ))
#'
#' error <- combine_ansi_styles("red", "bold")
#' warn <- combine_ansi_styles("magenta", "underline")
#' note <- col_cyan
#' cat(error("Error: subscript out of bounds!\n"))
#' cat(warn("Warning: shorter argument was recycled.\n"))
#' cat(note("Note: no such directory.\n"))
#'
#' # style_no_* functions, note that the color is not removed
#' style_italic(col_green(paste0(
#'   "italic before, ",
#'   style_no_italic("normal here, "),
#'   "italic after"
#' )))
#'
#' # avoiding  color for substring
#' style_italic(col_red(paste(
#'   "red before",
#'   col_none("not red between"),
#'   "red after"
#' )))
NULL

#' @export
#' @name ansi-styles
bg_black    <- create_ansi_style("bg_black")
#' @export
#' @name ansi-styles
bg_blue     <- create_ansi_style("bg_blue")
#' @export
#' @name ansi-styles
bg_cyan     <- create_ansi_style("bg_cyan")
#' @export
#' @name ansi-styles
bg_green    <- create_ansi_style("bg_green")
#' @export
#' @name ansi-styles
bg_magenta  <- create_ansi_style("bg_magenta")
#' @export
#' @name ansi-styles
bg_red      <- create_ansi_style("bg_red")
#' @export
#' @name ansi-styles
bg_white    <- create_ansi_style("bg_white")
#' @export
#' @name ansi-styles
bg_yellow   <- create_ansi_style("bg_yellow")
#' @export
#' @name ansi-styles
bg_none <- create_ansi_style("no_bg_color")

#' @export
#' @name ansi-styles
bg_br_black    <- create_ansi_style("bg_br_black")
#' @export
#' @name ansi-styles
bg_br_blue     <- create_ansi_style("bg_br_blue")
#' @export
#' @name ansi-styles
bg_br_cyan     <- create_ansi_style("bg_br_cyan")
#' @export
#' @name ansi-styles
bg_br_green    <- create_ansi_style("bg_br_green")
#' @export
#' @name ansi-styles
bg_br_magenta  <- create_ansi_style("bg_br_magenta")
#' @export
#' @name ansi-styles
bg_br_red      <- create_ansi_style("bg_br_red")
#' @export
#' @name ansi-styles
bg_br_white    <- create_ansi_style("bg_br_white")
#' @export
#' @name ansi-styles
bg_br_yellow   <- create_ansi_style("bg_br_yellow")

#' @export
#' @name ansi-styles
col_black   <- create_ansi_style("black")
#' @export
#' @name ansi-styles
col_blue    <- create_ansi_style("blue")
#' @export
#' @name ansi-styles
col_cyan    <- create_ansi_style("cyan")
#' @export
#' @name ansi-styles
col_green   <- create_ansi_style("green")
#' @export
#' @name ansi-styles
col_magenta <- create_ansi_style("magenta")
#' @export
#' @name ansi-styles
col_red     <- create_ansi_style("red")
#' @export
#' @name ansi-styles
col_white   <- create_ansi_style("white")
#' @export
#' @name ansi-styles
col_yellow  <- create_ansi_style("yellow")
#' @export
#' @name ansi-styles
col_grey    <- create_ansi_style("silver")
#' @export
#' @name ansi-styles
col_silver  <- create_ansi_style("silver")
#' @export
#' @name ansi-styles
col_none <- create_ansi_style("no_color")

#' @export
#' @name ansi-styles
col_br_black   <- create_ansi_style("br_black")
#' @export
#' @name ansi-styles
col_br_blue    <- create_ansi_style("br_blue")
#' @export
#' @name ansi-styles
col_br_cyan    <- create_ansi_style("br_cyan")
#' @export
#' @name ansi-styles
col_br_green   <- create_ansi_style("br_green")
#' @export
#' @name ansi-styles
col_br_magenta <- create_ansi_style("br_magenta")
#' @export
#' @name ansi-styles
col_br_red     <- create_ansi_style("br_red")
#' @export
#' @name ansi-styles
col_br_white   <- create_ansi_style("br_white")
#' @export
#' @name ansi-styles
col_br_yellow  <- create_ansi_style("br_yellow")

#' @export
#' @name ansi-styles
style_dim           <- create_ansi_style("blurred")
#' @export
#' @name ansi-styles
style_blurred       <- create_ansi_style("blurred")
#' @export
#' @name ansi-styles
style_bold          <- create_ansi_style("bold")
#' @export
#' @name ansi-styles
style_hidden        <- create_ansi_style("hidden")
#' @export
#' @name ansi-styles
style_inverse       <- create_ansi_style("inverse")
#' @export
#' @name ansi-styles
style_italic        <- create_ansi_style("italic")
#' @export
#' @name ansi-styles
style_reset         <- create_ansi_style("reset")
#' @export
#' @name ansi-styles
style_strikethrough <- create_ansi_style("strikethrough")
#' @export
#' @name ansi-styles
style_underline     <- create_ansi_style("underline")

#' @export
#' @name ansi-styles
style_no_bold <- create_ansi_style("no_bold")
#' @export
#' @name ansi-styles
style_no_blurred <- create_ansi_style("no_blurred")
#' @export
#' @name ansi-styles
style_no_dim <- create_ansi_style("no_blurred")
#' @export
#' @name ansi-styles
style_no_italic <- create_ansi_style("no_italic")
#' @export
#' @name ansi-styles
style_no_underline <- create_ansi_style("no_underline")
#' @export
#' @name ansi-styles
style_no_inverse <- create_ansi_style("no_inverse")
#' @export
#' @name ansi-styles
style_no_hidden <- create_ansi_style("no_hidden")
#' @export
#' @name ansi-styles
style_no_strikethrough <- create_ansi_style("no_strikethrough")
#' @export
#' @name ansi-styles
style_no_color <- create_ansi_style("no_color")
#' @export
#' @name ansi-styles
style_no_bg_color <- create_ansi_style("no_bg_color")

# Generated by cpp11: do not edit by hand

test_baseline_ <- function() {
  .Call(`_progresstestcpp_test_baseline_`)
}

test_cli_ <- function() {
  .Call(`_progresstestcpp_test_cli_`)
}

test_template_ <- function() {
  .Call(`_progresstestcpp_test_template_`)
}

cli_server_default <- function(msg) {
  cli_server_default_safe(msg)
}

cli_server_default_safe <- function(msg) {
  type <- as.character(msg$type)[1]
  app <- default_app() %||% start_app(.auto_close = FALSE)
  do.call(app[[type]], msg$args)
}

cli_server_callr_handler <- function(msg) {
  cli_server_default(msg)
}

rule_class <- function(x) {
  structure(x, class = c("cli_rule", "rule", "cli_ansi_string", "ansi_string", "character"))
}

capture_msgs <- function(expr) {
  msgs <- character()
  i <- 0
  suppressMessages(withCallingHandlers(
    expr,
    message = function(e) msgs[[i <<- i + 1]] <<- conditionMessage(e)))
  paste0(msgs, collapse = "")
}

capture_cli_messages <- function(expr) {
  msgs <- character()
  withCallingHandlers(
    expr,
    cliMessage = function(e) {
      msgs <<- c(msgs, conditionMessage(e))
      invokeRestart("muffleMessage")
    }
  )
  msgs
}

capt <- function(expr, print_it = TRUE) {
  pr <- if (print_it) print else identity
  paste(capture.output(pr(expr)), collapse = "\n")
}

capt0 <- function(expr, strip_style = FALSE) {
  out <- capture_msgs(expr)
  if  (strip_style) ansi_strip(out) else out
}

local_cli_config <- function(unicode = FALSE, dynamic = FALSE,
                             ansi = FALSE, num_colors = 1,
                             .local_envir = parent.frame()) {
  withr::local_options(
    cli.dynamic = dynamic,
    cli.ansi = ansi,
    cli.unicode = unicode,
    crayon.enabled = num_colors > 1,
    crayon.colors = num_colors,
    .local_envir = .local_envir
  )
  withr::local_envvar(
    PKG_OMIT_TIMES = "true",
    PKG_OMIT_SIZES = "true",
    .local_envir = .local_envir
  )
}

test_style <- function() {
  list(
    ".testcli h1" = list(
      "font-weight" = "bold",
      "font-style" = "italic",
      "margin-top" = 1,
      "margin-bottom" = 1),
    ".testcli h2" = list(
      "font-weight" = "bold",
      "margin-top" = 1,
      "margin-bottom" = 1),
    ".testcli h3" = list(
      "text-decoration" = "underline",
      "margin-top" = 1)
  )
}

fix_times <- function(out) {
  out <- sub("[(][ ]*[.0-9]+ [Mk]B/s[)]", "(8.5 MB/s)", out)
  out <- sub("[(][.0-9]+/s[)]", "(100/s)", out)
  out <- sub(" [.0-9]+(ms|s|m)", " 3ms", out)
  out <- sub("ETA:[ ]*[.0-9]+m?s", "ETA:  1s", out)
  out <- gsub("\\[[.0-9]+m?s\\]", "[1s]", out)
  out
}

fix_logger_output <- function(lines) {
  sub(
    paste0(
      "^[0-9][0-9][0-9][0-9]-[0-9][0-9]-[0-9][0-9]T",
      "[0-9][0-9]:[0-9][0-9]:[0-9][0-9]\\+00:00 ",
      "cli-[0-9]+-[0-9]+ "
    ),
    "2021-06-18T00:09:14+00:00 cli-36434-1 ",
    lines
  )
}

make_c_function <- function(file = NULL,
                            code = NULL,
                            args = character(),
                            type = c(".c", ".cpp"),
                            header = NULL,
                            linkingto = packageName(),
                            quiet = Sys.getenv("TESTTHAT") == "true") {
  type <- match.arg(type)

  # Create source file
  dir.create(dir <- tempfile())
  if (is.null(file)) {
    lines <- create_c_function_call(code, args, header = header)
  } else {
    lines <- readLines(file)
  }
  src <- basename(tempfile(fileext = type))
  writeLines(lines, file.path(dir, src))

  # Compile
  cflags <- ""
  for (pkg in linkingto) {
    pkgdir <- file.path(find.package(pkg), "include")
    lcldir <- file.path(find.package(pkg), "inst", "include")
    cflags <- paste(cflags, "-I", pkgdir, "-I", lcldir)
  }
  env <- c(PKG_CFLAGS = cflags)
  callr::rcmd(
    "SHLIB",
    src,
    wd = file.path(dir),
    env = env,
    echo = !quiet,
    show = !quiet
  )

  # Load DLL
  dllfile <- file.path(dir, sub("[.]c(pp)?$", .Platform$dynlib.ext, src))
  dll <- dyn.load(dllfile, local = TRUE, now = TRUE)

  # TODO: finalizer to unload/delete

  dll
}

create_c_function_call <- function(code, args, header = NULL) {
  c(
    "#include <Rinternals.h>",
    header,
    "SEXP tmp_c_function(",
    if (length(args) > 0) paste0("SEXP ", args, collapse = ", "),
    ") {",
    code,
    "}\n"
  )
}

win2unix <- function (str) {
  gsub("\r\n", "\n", str, fixed = TRUE, useBytes = TRUE)
}

st_from_bel <- function(x) {
  gsub("\007", "\033\\", x, fixed = TRUE)
}

st_to_bel <- function(x) {
  gsub("\033\\", "\007", x, fixed = TRUE)
}

test_package_root <- function() {
  x <- tryCatch(
    rprojroot::find_package_root_file(),
    error = function(e) NULL)

  if (!is.null(x)) return(x)

  pkg <- testthat::testing_package()
  x <- tryCatch(
    rprojroot::find_package_root_file(
      path = file.path("..", "..", "00_pkg_src", pkg)),
    error = function(e) NULL)

  if (!is.null(x)) return(x)

  stop("Cannot find package root")
}

sanitize_wd <- function(x) {
  wd <- paste0("file://", getwd())
  gsub(wd, "file:///testthat/home", x, fixed = TRUE)
}

sanitize_home <- function(x) {
  home <- paste0("file://", path.expand("~"))
  gsub(home, "file:///my/home", x, fixed = TRUE)
}

sanitize_srcref <- function(x) {
  gsub(" at .*.R:[0-9]+:[0-9]+", "", x)
}

sanitize_call <- function(x) {
  gsub(" in `.*`", "", x)
}

r_pty <- function(.envir = parent.frame()) {
  skip_on_cran()
  # TODO: why does this fail on the CI, in covr
  if (Sys.getenv("R_COVR") == "true" &&
      isTRUE(as.logical(Sys.getenv("CI")))) {
    skip("fails on CI in covr")
  }
  if (!Sys.info()[["sysname"]] %in% c("Darwin", "Linux")) skip("Needs Linux or macOS")

  r <- file.path(R.home("bin"), "R")
  p <- processx::process$new(
    r,
    c("-q", "--slave", "--vanilla"),
    pty = TRUE,
    env = c("current", R_CLI_HIDE_CURSOR = "false", R_LIBS = .libPaths()[1])
  )

  defer({
    close(p$get_input_connection())
    p$wait(1000)
    p$kill()
  }, envir = .envir)

  p$poll_io(1000)
  p$read_output()
  p
}

context("boxes")

test_that("empty label", {
  out <- capt(boxx(""))
  exp <- rebox(
    "┌──────┐",
    "│      │",
    "│      │",
    "│      │",
    "└──────┘")
  expect_equal(out, exp)
})

test_that("empty label 2", {
  out <- capt(boxx(character()))
  exp <- rebox(
    "┌──────┐",
    "│      │",
    "│      │",
    "└──────┘")
  expect_equal(out, exp)
})

test_that("label", {
  out <- capt(boxx("label"))
  exp <- rebox(
    "┌───────────┐",
    "│           │",
    "│   label   │",
    "│           │",
    "└───────────┘")
  expect_equal(out, exp)
})

test_that("label vector", {
  out <- capt(boxx(c("label", "l2")))
  exp <- rebox(
    "┌───────────┐",
    "│           │",
    "│   label   │",
    "│   l2      │",
    "│           │",
    "└───────────┘")
  expect_equal(out, exp)
})

test_that("border style", {
  out <- capt(boxx("label", border_style = "classic"))
  exp <- rebox(
    "+-----------+",
    "|           |",
    "|   label   |",
    "|           |",
    "+-----------+")
  expect_equal(out, exp)
})

test_that("padding", {
  out <- capt(boxx("label", padding = 2))
  exp <- rebox(
    "┌─────────────────┐",
    "│                 │",
    "│                 │",
    "│      label      │",
    "│                 │",
    "│                 │",
    "└─────────────────┘")
  expect_equal(out, exp)

  out <- capt(boxx("label", padding = c(1,2,1,2)))
  exp <- rebox(
    "┌─────────┐",
    "│         │",
    "│  label  │",
    "│         │",
    "└─────────┘")
  expect_equal(out, exp)

  out <- capt(boxx("label", padding = c(1,2,0,2)))
  exp <- rebox(
    "┌─────────┐",
    "│  label  │",
    "│         │",
    "└─────────┘")
  expect_equal(out, exp)

  out <- capt(boxx("label", padding = c(1,2,0,0)))
  exp <- rebox(
    "┌───────┐",
    "│  label│",
    "│       │",
    "└───────┘")
  expect_equal(out, exp)
})

test_that("margin", {
  out <- capt(boxx("label", margin = 1))
  exp <- rebox("",
    "   ┌───────────┐",
    "   │           │",
    "   │   label   │",
    "   │           │",
    "   └───────────┘", "")
  expect_equal(out, exp)

  out <- capt(boxx("label", margin = c(1,2,3,4)))
  exp <- rebox("", "", "",
    "  ┌───────────┐",
    "  │           │",
    "  │   label   │",
    "  │           │",
    "  └───────────┘", "")
  expect_equal(out, exp)

  out <- capt(boxx("label", margin = c(0,1,2,0)))
  exp <- rebox("", "",
    " ┌───────────┐",
    " │           │",
    " │   label   │",
    " │           │",
    " └───────────┘")
  expect_equal(out, exp)
})

test_that("float", {
  out <- capt(boxx("label", float = "center", width = 20))
  exp <- rebox(
    "    ┌───────────┐",
    "    │           │",
    "    │   label   │",
    "    │           │",
    "    └───────────┘")
  expect_equal(out, exp)

  out <- capt(boxx("label", float = "right", width = 20))
  exp <- rebox(
    "       ┌───────────┐",
    "       │           │",
    "       │   label   │",
    "       │           │",
    "       └───────────┘")
  expect_equal(out, exp)
})

test_that("background_col", {
  withr::with_options(
    list(crayon.enabled = TRUE, crayon.colors = 256), {
      bx <- boxx("label", background_col = "red")
      expect_true(crayon::has_style(bx))
      out <- capt_cat(crayon::strip_style(unclass(bx)))
      exp <- rebox(
        "┌───────────┐",
        "│           │",
        "│   label   │",
        "│           │",
        "└───────────┘")
      expect_equal(out, exp)

      bx <- boxx("label", background_col = crayon::red)
      expect_true(crayon::has_style(bx))
      out <- capt_cat(crayon::strip_style(unclass(bx)))
      exp <- rebox(
        "┌───────────┐",
        "│           │",
        "│   label   │",
        "│           │",
        "└───────────┘")
      expect_equal(out, exp)
    }
  )
})

test_that("border_col", {
  withr::with_options(
    list(crayon.enabled = TRUE, crayon.colors = 256), {
      bx <- boxx("label", border_col = "red")
      expect_true(crayon::has_style(bx))
      out <- capt_cat(crayon::strip_style(unclass(bx)))
      exp <- rebox(
        "┌───────────┐",
        "│           │",
        "│   label   │",
        "│           │",
        "└───────────┘")
      expect_equal(out, exp)

      bx <- boxx("label", border_col = crayon::red)
      expect_true(crayon::has_style(bx))
      out <- capt_cat(crayon::strip_style(unclass(bx)))
      exp <- rebox(
        "┌───────────┐",
        "│           │",
        "│   label   │",
        "│           │",
        "└───────────┘")
      expect_equal(out, exp)
    }
  )
})

test_that("align", {
  out <- capt(boxx(c("label", "l2"), align = "center"))
  exp <- rebox(
    "┌───────────┐",
    "│           │",
    "│   label   │",
    "│     l2    │",
    "│           │",
    "└───────────┘")
  expect_equal(out, exp)

  out <- capt(boxx(c("label", "l2"), align = "right"))
  exp <- rebox(
    "┌───────────┐",
    "│           │",
    "│   label   │",
    "│      l2   │",
    "│           │",
    "└───────────┘")
  expect_equal(out, exp)
})

test_that("empty input", {
  expect_snapshot(
    vt_output("", width = 20, height = 2)$segment
  )
})

test_that("raw input", {
  expect_snapshot(
    vt_output(charToRaw("foobar"), width = 20, height = 2)$segment
  )
})

test_that("overflow", {
  expect_snapshot(
    vt_output(strrep("1234567890", 2), width = 19, height = 2)$segment
  )
})

test_that("control characters", {
  expect_snapshot(
    vt_output("foo\nbar", width = 20, height = 2)$segment
  )
  expect_snapshot(
    vt_output("foobar\rbaz", width = 20, height = 2)$segment
  )
})

test_that("scroll up", {
  expect_snapshot(
    vt_output(strrep("1234567890", 5), width = 20, height = 2)$segment
  )
  expect_snapshot(
    vt_output(paste0(1:10, "\n"), width = 10, height = 5)$segment
  )
})

test_that_cli(configs = "ansi", "ANSI SGR", {
  expect_snapshot(
    vt_output("12\033[31m34\033[1m56\033[39m78\033[21m90", width = 20, height = 2)
  )

  expect_snapshot(
    vt_output(style_bold("I'm bold"), width = 20, height = 2)
  )

  expect_snapshot(
    vt_output(style_italic("I'm italic"), width = 20, height = 2)
  )

  expect_snapshot(
    vt_output(style_underline("I'm underlined"), width = 20, height = 2)
  )

  expect_snapshot(
    vt_output(style_strikethrough("I'm strikethrough"), width = 20, height = 2)
  )

  expect_snapshot(
    vt_output(style_inverse("I'm inverse"), width = 20, height = 2)
  )
})

test_that("hyperlinks", {
  withr::local_options(cli.hyperlink = TRUE)
  expect_snapshot({
    link <- style_hyperlink("text", "url")
    vt_output(c("pre ", st_from_bel(link), " post"), width = 20, height = 2)
  })
  expect_snapshot({
    link <- style_hyperlink("text", "url", params = c("f" = "x", "g" = "y"))
    vt_output(c("pre ", st_from_bel(link), " post"), width = 20, height = 2)
  })
})

test_that("erase in line", {
  expect_snapshot({
    vt_output("foobar\033[3D\033[K", width = 10, height = 2)$segment
    vt_output("foobar\033[3D\033[0K", width = 10, height = 2)$segment
    vt_output("foobar\033[3D\033[1K", width = 10, height = 2)$segment
    vt_output("foobar\033[3D\033[2K", width = 10, height = 2)$segment
  })
})

test_that("erase in screen", {
  expect_snapshot({
    vt_output("foo\nfoobar\nfoobar2\033[A\033[4D\033[J", width = 10, height = 4)$segment
    vt_output("foo\nfoobar\nfoobar2\033[A\033[4D\033[0J", width = 10, height = 4)$segment
    vt_output("foo\nfoobar\nfoobar2\033[A\033[4D\033[1J", width = 10, height = 4)$segment
    vt_output("foo\nfoobar\nfoobar2\033[A\033[4D\033[2Jx", width = 10, height = 4)$segment
    vt_output("foo\nfoobar\nfoobar2\033[A\033[4D\033[3Jx", width = 10, height = 4)$segment
  })
})

test_that("colors", {
  expect_equal(vt_output("\033[30mcolored\033[39m")$color[1], "0")
  expect_equal(vt_output("\033[37mcolored\033[39m")$color[1], "7")
  expect_equal(vt_output("\033[90mcolored\033[39m")$color[1], "8")
  expect_equal(vt_output("\033[97mcolored\033[39m")$color[1], "15")

  expect_equal(vt_output("\033[40mcolored\033[39m")$background_color[1], "0")
  expect_equal(vt_output("\033[47mcolored\033[39m")$background_color[1], "7")
  expect_equal(vt_output("\033[100mcolored\033[39m")$background_color[1], "8")
  expect_equal(vt_output("\033[107mcolored\033[39m")$background_color[1], "15")

  expect_equal(vt_output("\033[38;5;100mcolored\033[39m")$color[1], "100")
  expect_equal(vt_output("\033[48;5;110mcolored\033[39m")$background_color[1], "110")

  expect_equal(vt_output("\033[38;2;1;2;3mcolored\033[39m")$color[1], "#010203")
  expect_equal(vt_output("\033[48;2;4;5;6mcolored\033[39m")$background_color[1], "#040506")
})

test_that("No leftover SVG figures", {
  skip_on_cran()
  skip_on_covr()
  pkg_dir <- test_package_root()
  figs <- dir(file.path(pkg_dir, "man", "figures"), pattern = "[.]svg$")
  rd <- dir(file.path(pkg_dir, "man"), pattern = "[.]Rd$", full.names = TRUE)
  rd_figs <- unlist(lapply(rd, function(x) {
    grep("\\\\figure\\{", readLines(x), value = TRUE)
  }))
  rd_figs <- sub("^.*\\\\figure\\{(.*[.]svg).*$", "\\1", rd_figs)

  expect_equal(
    sort(figs),
    sort(unique(rd_figs))
  )

  figs2 <- dir(file.path(pkg_dir, "man", "figures", "README"), pattern = "[.]svg$")
  readme <- file.path(pkg_dir, "README.md")
  readme_figs <- grep("man/figures/", readLines(readme), value = TRUE)
  readme_figs <- sub("^.*man/figures/README/(.*[.]svg).*$", "\\1", readme_figs)

  expect_equal(
    sort(figs2),
    sort(unique(readme_figs))
  )
})

context("tree")

test_that("tree", {
  data <- data.frame(
    stringsAsFactors = FALSE,
    package = c("processx", "backports", "assertthat", "Matrix",
      "magrittr", "rprojroot", "clisymbols", "prettyunits", "withr",
      "desc", "igraph", "R6", "crayon", "debugme", "digest", "irlba",
      "rcmdcheck", "callr", "pkgconfig", "lattice"),
    dependencies = I(list(
      c("assertthat", "crayon", "debugme", "R6"), character(0),
      character(0), "lattice", character(0), "backports", character(0),
      c("magrittr", "assertthat"), character(0),
      c("assertthat", "R6", "crayon", "rprojroot"),
      c("irlba", "magrittr", "Matrix", "pkgconfig"), character(0),
      character(0), "crayon", character(0), "Matrix",
      c("callr", "clisymbols", "crayon", "desc", "digest", "prettyunits",
        "R6", "rprojroot", "withr"),
      c("processx", "R6"), character(0), character(0)
    ))
  )
  out <- capt(tree(data))
  exp <- rebox(mode = "tree",
    "processx",
    "├─assertthat",
    "├─crayon",
    "├─debugme",
    "│ └─crayon",
    "└─R6"
  )
  expect_equal(out, exp)

  out <- capt(tree(data, root = "desc"))
  exp <- rebox(mode = "tree",
    "desc",
    "├─assertthat",
    "├─R6",
    "├─crayon",
    "└─rprojroot",
    "  └─backports")
  expect_equal(out, exp)

  # Check that trees with apparent circularity error nicely
  data <- data.frame(
    stringsAsFactors = FALSE,
    X = c("a", "b", "c","d", "e", "f", "g", "h", "j"),
    Y = I(list(
      c("b", "e", "f"),
      c("d", "g"),
      character(0),
      c("a", "h"),
      character(0),
      character(0),
      character(0),
      c("j"),
      character(0)
    ))
  )

  expect_warning(tree(data), "Endless loop found in tree: a -> b -> d -> a")

})

test_that("trimming", {

  pkgdeps <- list(
    "dplyr@0.8.3" = c("assertthat@0.2.1", "glue@1.3.1", "magrittr@1.5",
      "R6@2.4.0", "Rcpp@1.0.2", "rlang@0.4.0", "tibble@2.1.3",
      "tidyselect@0.2.5"),
    "assertthat@0.2.1" = character(),
    "glue@1.3.1" = character(),
    "magrittr@1.5" = character(),
    "pkgconfig@2.0.3" = character(),
    "R6@2.4.0" = character(),
    "Rcpp@1.0.2" = character(),
    "rlang@0.4.0" = character(),
    "tibble@2.1.3" = c("cli@1.1.0", "crayon@1.3.4", "fansi@0.4.0",
       "pillar@1.4.2", "pkgconfig@2.0.3", "rlang@0.4.0"),
    "cli@1.1.0" = c("assertthat@0.2.1", "crayon@1.3.4"),
    "crayon@1.3.4" = character(),
    "fansi@0.4.0" = character(),
    "pillar@1.4.2" = c("cli@1.1.0", "crayon@1.3.4", "fansi@0.4.0",
       "rlang@0.4.0", "utf8@1.1.4", "vctrs@0.2.0"),
    "utf8@1.1.4" = character(),
    "vctrs@0.2.0" = c("backports@1.1.5", "ellipsis@0.3.0",
       "digest@0.6.21", "glue@1.3.1", "rlang@0.4.0", "zeallot@0.1.0"),
    "backports@1.1.5" = character(),
    "ellipsis@0.3.0" = c("rlang@0.4.0"),
    "digest@0.6.21" = character(),
    "glue@1.3.1" = character(),
    "zeallot@0.1.0" = character(),
    "tidyselect@0.2.5" = c("glue@1.3.1", "purrr@1.3.1", "rlang@0.4.0",
       "Rcpp@1.0.2"),
    "purrr@0.3.3" = c("magrittr@1.5", "rlang@0.4.0")
  )

  pkgs <- data.frame(
    stringsAsFactors = FALSE,
    name = names(pkgdeps),
    deps = I(unname(pkgdeps))
  )
  pkgs$label <- pkgs$name
  pkgs$trimmed <- paste(pkgs$name, " (trimmed)")

  out <- capt(tree(pkgs, trim = TRUE))

  exp <- rebox(mode = "tree",
    "dplyr@0.8.3",
    "├─assertthat@0.2.1",
    "├─glue@1.3.1",
    "├─magrittr@1.5",
    "├─R6@2.4.0",
    "├─Rcpp@1.0.2",
    "├─rlang@0.4.0",
    "├─tibble@2.1.3",
    "│ ├─cli@1.1.0",
    "│ │ ├─assertthat@0.2.1  (trimmed)",
    "│ │ └─crayon@1.3.4",
    "│ ├─crayon@1.3.4  (trimmed)",
    "│ ├─fansi@0.4.0",
    "│ ├─pillar@1.4.2",
    "│ │ ├─cli@1.1.0  (trimmed)",
    "│ │ ├─crayon@1.3.4  (trimmed)",
    "│ │ ├─fansi@0.4.0  (trimmed)",
    "│ │ ├─rlang@0.4.0  (trimmed)",
    "│ │ ├─utf8@1.1.4",
    "│ │ └─vctrs@0.2.0",
    "│ │   ├─backports@1.1.5",
    "│ │   ├─ellipsis@0.3.0",
    "│ │   │ └─rlang@0.4.0  (trimmed)",
    "│ │   ├─digest@0.6.21",
    "│ │   ├─glue@1.3.1  (trimmed)",
    "│ │   ├─rlang@0.4.0  (trimmed)",
    "│ │   └─zeallot@0.1.0",
    "│ ├─pkgconfig@2.0.3",
    "│ └─rlang@0.4.0  (trimmed)",
    "└─tidyselect@0.2.5",
    "  ├─glue@1.3.1  (trimmed)",
    "  ├─rlang@0.4.0  (trimmed)",
    "  └─Rcpp@1.0.2  (trimmed)"
  )

  expect_equal(out, exp)
})

test_that("control characaters", {
  skip_on_cran()
  p <- r_pty()

  expect_snapshot(
    for (code in c(1:2, 4:6, 8:14, 16L, 20L, 21L, 23L, 27L, 127L)) {
      p$write_input("cli::keypress()\n")
      Sys.sleep(0.1)
      p$write_input(as.raw(code))
      p$poll_io(1000)
      cat(p$read_output())
    }
  )
})

test_that("write ahead", {
  skip_on_cran()
  p <- r_pty()
  expect_snapshot({
    p$write_input("{ Sys.sleep(0.5); cli::keypress() }\nX")
    p$poll_io(1000)
    cat(p$read_output())
  })
})

test_that("arrows, etc", {
  skip_on_cran()
  p <- r_pty()
  keys <- paste0("\033", c(
    "[A", "[C", "[D", "[F", "[H", "-",
    "OA", "OB", "OC", "OD", "OF", "OH", "-",
    "[1~", "[2~", "[3~", "[4~", "[5~", "[6~", "-",
    "[[5~", "[[6~", "-",
    "[[7~", "[[8~", "-",
    "OP", "OQ", "OR", "OS", "-",
    "[15~", "[17~", "[18~", "[19~", "[20~", "[21~", "[23~", "[24~", "-",
    "[11~", "[12~", "[13~", "[14~", "-",
    ""
  ))
  keys[keys == "\033-"] <- "-"
  expect_snapshot({
    for (key in keys) {
      p$write_input("cli::keypress()\n")
      p$write_input(key)
      p$poll_io(1000)
      cat(p$read_output())
    }
  })
})

test_that("nonblocking", {
  skip_on_cran()
  p <- r_pty()
  expect_snapshot({
    p$write_input("cli::keypress(block = FALSE)\n")
    p$poll_io(1000)
    cat(p$read_output())
  })
  expect_snapshot({
    p$write_input("{ Sys.sleep(0.5); cli::keypress() }\nX")
    p$poll_io(1000)
    cat(p$read_output())
  })
})

test_that("empty input", {
  expect_snapshot(
    vt_output("", width = 20, height = 2)$segment
  )
})

test_that("raw input", {
  expect_snapshot(
    vt_output(charToRaw("foobar"), width = 20, height = 2)$segment
  )
})

test_that("overflow", {
  expect_snapshot(
    vt_output(strrep("1234567890", 2), width = 19, height = 2)$segment
  )
})

test_that("control characters", {
  expect_snapshot(
    vt_output("foo\nbar", width = 20, height = 2)$segment
  )
  expect_snapshot(
    vt_output("foobar\rbaz", width = 20, height = 2)$segment
  )
})

test_that("scroll up", {
  expect_snapshot(
    vt_output(strrep("1234567890", 5), width = 20, height = 2)$segment
  )
  expect_snapshot(
    vt_output(paste0(1:10, "\n"), width = 10, height = 5)$segment
  )
})

test_that_cli(config = "ansi", "ANSI SGR", {
  expect_snapshot(
    vt_output("12\033[31m34\033[1m56\033[39m78\033[21m90", width = 20, height = 2)
  )

  expect_snapshot(
    vt_output(style_bold("I'm bold"), width = 20, height = 2)
  )

  expect_snapshot(
    vt_output(style_italic("I'm italic"), width = 20, height = 2)
  )

  expect_snapshot(
    vt_output(style_underline("I'm underlined"), width = 20, height = 2)
  )

  expect_snapshot(
    vt_output(style_strikethrough("I'm strikethrough"), width = 20, height = 2)
  )

  expect_snapshot(
    vt_output(style_inverse("I'm inverse"), width = 20, height = 2)
  )
})

test_that("hyperlinks", {
  withr::local_options(cli.hyperlink = TRUE)
  expect_snapshot({
    link <- style_hyperlink("text", "url")
    vt_output(c("pre ", st_from_bel(link), " post"), width = 20, height = 2)
  })
  expect_snapshot({
    link <- style_hyperlink("text", "url", params = c("f" = "x", "g" = "y"))
    vt_output(c("pre ", st_from_bel(link), " post"), width = 20, height = 2)
  })
})

test_that("erase in line", {
  expect_snapshot({
    vt_output("foobar\033[3D\033[K", width = 10, height = 2)$segment
    vt_output("foobar\033[3D\033[0K", width = 10, height = 2)$segment
    vt_output("foobar\033[3D\033[1K", width = 10, height = 2)$segment
    vt_output("foobar\033[3D\033[2K", width = 10, height = 2)$segment
  })
})

test_that("erase in screen", {
  expect_snapshot({
    vt_output("foo\nfoobar\nfoobar2\033[A\033[4D\033[J", width = 10, height = 4)$segment
    vt_output("foo\nfoobar\nfoobar2\033[A\033[4D\033[0J", width = 10, height = 4)$segment
    vt_output("foo\nfoobar\nfoobar2\033[A\033[4D\033[1J", width = 10, height = 4)$segment
    vt_output("foo\nfoobar\nfoobar2\033[A\033[4D\033[2Jx", width = 10, height = 4)$segment
    vt_output("foo\nfoobar\nfoobar2\033[A\033[4D\033[3Jx", width = 10, height = 4)$segment
  })
})

test_that("colors", {
  expect_equal(vt_output("\033[30mcolored\033[39m")$color[1], "0")
  expect_equal(vt_output("\033[37mcolored\033[39m")$color[1], "7")
  expect_equal(vt_output("\033[90mcolored\033[39m")$color[1], "8")
  expect_equal(vt_output("\033[97mcolored\033[39m")$color[1], "15")

  expect_equal(vt_output("\033[40mcolored\033[39m")$background_color[1], "0")
  expect_equal(vt_output("\033[47mcolored\033[39m")$background_color[1], "7")
  expect_equal(vt_output("\033[100mcolored\033[39m")$background_color[1], "8")
  expect_equal(vt_output("\033[107mcolored\033[39m")$background_color[1], "15")

  expect_equal(vt_output("\033[38;5;100mcolored\033[39m")$color[1], "100")
  expect_equal(vt_output("\033[48;5;110mcolored\033[39m")$background_color[1], "110")

  expect_equal(vt_output("\033[38;2;1;2;3mcolored\033[39m")$color[1], "#010203")
  expect_equal(vt_output("\033[48;2;4;5;6mcolored\033[39m")$background_color[1], "#040506")
})

test_that("control characters", {
  skip_on_cran()
  p <- r_pty()

  expect_snapshot(
    for (code in c(1:2, 4:6, 8:14, 16L, 20L, 21L, 23L, 27L, 127L)) {
      p$write_input("cli::keypress()\n")
      Sys.sleep(0.1)
      p$write_input(as.raw(code))
      p$poll_io(1000)
      cat(p$read_output())
    }
  )
})

test_that("write ahead", {
  skip_on_cran()
  p <- r_pty()
  expect_snapshot({
    p$write_input("{ Sys.sleep(0.5); cli::keypress() }\nX")
    p$poll_io(1000)
    cat(p$read_output())
  })
})

test_that("arrows, etc", {
  skip_on_cran()
  p <- r_pty()
  keys <- paste0(
    "\033",
    c(
      "[A",
      "[C",
      "[D",
      "[F",
      "[H",
      "-",
      "OA",
      "OB",
      "OC",
      "OD",
      "OF",
      "OH",
      "-",
      "[1~",
      "[2~",
      "[3~",
      "[4~",
      "[5~",
      "[6~",
      "-",
      "[[5~",
      "[[6~",
      "-",
      "[[7~",
      "[[8~",
      "-",
      "OP",
      "OQ",
      "OR",
      "OS",
      "-",
      "[15~",
      "[17~",
      "[18~",
      "[19~",
      "[20~",
      "[21~",
      "[23~",
      "[24~",
      "-",
      "[11~",
      "[12~",
      "[13~",
      "[14~",
      "-",
      ""
    )
  )
  keys[keys == "\033-"] <- "-"
  expect_snapshot({
    for (key in keys) {
      p$write_input("cli::keypress()\n")
      p$write_input(key)
      p$poll_io(1000)
      cat(p$read_output())
    }
  })
})

test_that("nonblocking", {
  skip_on_cran()
  p <- r_pty()
  expect_snapshot({
    p$write_input("cli::keypress(block = FALSE)\n")
    p$poll_io(1000)
    cat(p$read_output())
  })
  expect_snapshot({
    p$write_input("{ Sys.sleep(0.5); cli::keypress() }\nX")
    p$poll_io(1000)
    cat(p$read_output())
  })
})

# # Standalone file for controlling when and where to build vignettes ----
#
# The canonical location of this file is in the asciicast package:
# https://github.com/r-lib/asciicast/master/R/lazyrmd.R
#
# This standalone file provides a vignette builder that gives you more
# control about when, where and how the vignettes of an R package will be
# built. Possible use cases:
# * do not rebuild vignettes on CRAN
# * do not rebuild vignettes on a CI if the build time dependencies are
#   not available.
# * avoid rebuilding long running vignettes.
# * fully static vignettes that never rebuild.
# * etc.
#
# ## API
#
# ```
# onload_hook(local = "if-newer", ci = TRUE, cran = "no-code")
# build(mode = c("force", "if-newer"))
# ```
#
# ## Usage
#
# To use this standalone in your package, do the following:
# 1. Copy this file into your package, without changes.
# 2. Call `lazyrmd$onload_hook()` from the `.onLoad()` function of your
#    package. For example:
#    ```
#    .onLoad <- function(libname, pkgname) {
#      lazyrmd$onload_hook(
#        local = FALSE,
#        ci = function() is_recording_supported(),
#        cran = "no-code"
#      )
#    }
#    ```
# 3. Add the package itself as a vignette builder, in `DESCRIPTION`.
# 4. Also in `DESCRIPTION`, add knitr and rmarkdown to `Suggests`, if you
#    want to install them automatically on the CI, and/or you want to
#    build vignettes in CRAN.
# 5. Change the builder of the vignettes that you want to build lazily,
#    in the YAML header. E.g.:
#    ```
#    vignette: >
#    %\VignetteIndexEntry{asciicast example vignette}
#    %\VignetteEngine{asciicast::lazyrmd}
#    %\VignetteEncoding{UTF-8}
#    ```
#
# ## NEWS:
#
# ### 1.0.0 -- 2019-12-01
#
# * First release.

lazyrmd <- local({

  # config ---------------------------------------------------------------

  # Need to do this before we mess up the environment
  package_name <- utils::packageName()
  lazy_env <- environment()
  parent.env(lazy_env) <- baseenv()

  re_lazy_rmd <- "[.]Rmd$"

  config = list()

  # API ------------------------------------------------------------------

  #' Configure the lazy vignette builder
  #'
  #' @noRd
  #' @param local Whether/when to build vignettes for local builds.
  #' @param ci Whether/when to build vignettes on the CI(s).
  #' @param cran Whether/when to build vignettes on CRAN.
  #'
  #' Possible values for `local`, `ci` and `cran` are:
  #' * `TRUE`: always (re)build the vignettes.
  #' * `FALSE`: never (re)build the vignettes.
  #' * `"if-newer"`: Only re(build) the vignettes if the Rmd file is newer.
  #' * `"no-code"`: never rebuild the vignettes, and make sure that the
  #'                `.R` file only contains dummy code. (This is to make
  #'                sure that the `.R` file runs on CRAN.)
  #' * a function object. This is called without arguments and if it
  #'   returns an `isTRUE()` value, then the vignette is built.

  onload_hook <- function(local = "if-newer", ci = TRUE, cran = "no-code") {
    config <<- list(local = local, ci = ci, cran = cran, forced = NULL)
    tools::vignetteEngine(
      "lazyrmd",
      weave = lazy_weave,
      tangle = lazy_tangle,
      pattern = re_lazy_rmd,
      package = package_name
    )
  }

  #' Build vignettes
  #'
  #' This is a utility function, and you would probably not use it in the
  #' package code itself. It can force the re-building of all vignettes.
  #' This function ignores the vignette configuration that was created with
  #' `onload_hook()`.
  #'
  #' @param mode `"force"` forces vignette re-building, even if the `.Rmd`
  #'   file is older than the output file(s). `"if-newer"` will only
  #'   re-build if the `.Rmd` file is newer.
  #' @return The return value of [tools::buildVignettes()].

  build <- function(mode = c("force", "if-newer")) {
    mode <- match.arg(mode)
    config$forced <<- if (mode == "force") TRUE else mode
    env_save <- Sys.getenv("LAZY_RMD_FORCED", NA_character_)
    if (is.na(env_save)) {
      on.exit(Sys.unsetenv("LAZY_RMD_FORCED"), add = TRUE)
    } else {
      on.exit(Sys.setenv(LAZY_RMD_FORCED = env_save), add = TRUE)
    }
    Sys.setenv(LAZY_RMD_FORCED = "true")
    tools::buildVignettes(dir = ".", tangle = TRUE)
  }

  # internals ------------------------------------------------------------

  #' This is called by [tools::buildVignettes], as the weave function of
  #' the lazyrmd vignette engine

  lazy_weave <- function(file, ...) {
    output <- sub(re_lazy_rmd, ".html", file)
    env <- detect_env()
    conf <- config[[env]]

    if (! should_build(file, output, conf)) {
      create_if_needed(output, conf)
      return(output)
    }

    weave(file, ...)
  }

  weave <- function(file, driver, syntax, encoding = "UTF-8",
                    quiet = TRUE, ...) {

    opts <- options(markdown.HTML.header = NULL)
    on.exit({
      knitr::opts_chunk$restore()
      knitr::knit_hooks$restore()
      options(opts)
    }, add = TRUE)

    knitr::opts_chunk$set(error = FALSE)

    # This will create the `.R` file as well. Then it might get overwritten
    # by a tangle step, if any. Typically it won't because it will be
    # newer than the `.Rmd` file.
    knitr::knit_hooks$set(purl = knitr::hook_purl)

    rmarkdown::render(
      file,
      encoding = encoding,
      quiet = quiet,
      envir = globalenv(),
      ...
    )
  }

  #' This is called by [tools::buildVignettes], as the tangle function of
  #' the lazyrmd vignette engine

  lazy_tangle <- function(file, ...) {
    output <- sub(re_lazy_rmd, ".R", file)
    env <- detect_env()
    conf <- config[[env]]

    if (! should_build(file, output, conf)) {
      create_if_needed(output, conf)
      return(output)
    }

    tangle(file, ...)
  }

  tangle <- function(file, ..., encoding = "UTF-8", quiet = FALSE) {
    output <- sub(re_lazy_rmd, ".R", file)
    knitr::purl(file, encoding = encoding, quiet = quiet, ...)
  }

  is_ci <- function() {
    isTRUE(as.logical(Sys.getenv("CI")))
  }

  #' If NOT_CRAN is set to a true value, then we are local
  #' Otherwise we are local if not running inside `R CMD check`.

  is_local <- function() {
    if (isTRUE(as.logical(Sys.getenv("NOT_CRAN")))) return(TRUE)
    Sys.getenv("_R_CHECK_PACKAGE_NAME_", "") == ""
  }

  detect_env <- function() {
    # Order matters here, is_local() == TRUE on the CI as well, usually
    if (!is.na(Sys.getenv("LAZY_RMD_FORCED", NA_character_))) return("forced")
    if (is_ci()) return("ci")
    if (is_local()) return("local")
    "cran"
  }

  should_build <- function(input, output, conf) {
    if (isTRUE(conf)) return(TRUE)
    if (identical(conf, FALSE)) return(FALSE)
    if (is.function(conf)) return(isTRUE(conf()))
    if (identical(conf, "if-newer")) {
      return(!file.exists(output) || file_mtime(input) > file_mtime(output))
    }
    FALSE
  }

  file_mtime <- function(...) {
    file.info(..., extra_cols = FALSE)$mtime
  }

  create_if_needed <- function(path, conf) {
    if (!file.exists(path)) file.create(path)
    if (identical(conf, "no-code")) {
      code_file <- sub("\\.html$", ".R", path)
      cat("# Dummy file for static vignette\n", file = code_file)
    }
    Sys.setFileTime(path, Sys.time())
  }

  structure(
    list(
      .internal = lazy_env,
      onload_hook = onload_hook,
      build = build
    ),
    class = c("standalone_lazyrmd", "standalone")
  )
})

test_that_cli("make_progress_bar", {
  withr::local_options(
    cli.progress_bar_style = NULL,
    cli.progress_bar_style_unicode = NULL,
    cli.progress_bar_style_ascii = NULL
  )
  expect_snapshot(make_progress_bar(.5))
})

test_that_cli(configs = "fancy", "cli_progress_styles", {
  withr::local_options(
    cli.progress_bar_style_unicode = NULL,
    cli.progress_bar_style_ascii = NULL
  )

  withr::local_options(cli.progress_bar_style = "classic")
  expect_snapshot(make_progress_bar(.5))

  withr::local_options(cli.progress_bar_style = "squares")
  expect_snapshot(make_progress_bar(.5))

  withr::local_options(cli.progress_bar_style = "dot")
  expect_snapshot(make_progress_bar(.5))

  withr::local_options(cli.progress_bar_style = "fillsquares")
  expect_snapshot(make_progress_bar(.5))

  withr::local_options(cli.progress_bar_style = "bar")
  expect_snapshot(make_progress_bar(.5))
})

test_that_cli(configs = c("plain", "unicode"), "custom style", {
  mybar <- list(complete = "X", incomplete = "O", current = ">")
  withr::local_options(
    cli.progress_bar_style = mybar,
    cli.progress_bar_style_unicode = NULL,
    cli.progress_bar_style_ascii = NULL
  )
  expect_snapshot(make_progress_bar(.5))
})

#' Draw a tree
#'
#' Draw a tree using box drawing characters. Unicode characters are
#' used if available. (Set the `cli.unicode` option if auto-detection
#' fails.)
#'
#' A node might appear multiple times in the tree, or might not appear
#' at all.
#'
#' @param data Data frame that contains the tree structure.
#'   The first column is an id, and the second column is a list column,
#'   that contains the ids of the child nodes. The optional third column
#'   may contain the text to print to annotate the node.
#' @param root The name of the root node.
#' @param style Optional box style list.
#' @param width Maximum width of the output. Defaults to the `width`
#'   option, see [base::options()].
#' @param trim Whether to avoid traversing the same nodes multiple times.
#'   If `TRUE` and `data` has a `trimmed` column, then that is used for
#'   printing repeated noded.
#' @return Character vector, the lines of the tree drawing.
#'
#' @export
#' @examples
#' data <- data.frame(
#'   stringsAsFactors = FALSE,
#'   package = c("processx", "backports", "assertthat", "Matrix",
#'     "magrittr", "rprojroot", "clisymbols", "prettyunits", "withr",
#'     "desc", "igraph", "R6", "crayon", "debugme", "digest", "irlba",
#'     "rcmdcheck", "callr", "pkgconfig", "lattice"),
#'   dependencies = I(list(
#'     c("assertthat", "crayon", "debugme", "R6"), character(0),
#'     character(0), "lattice", character(0), "backports", character(0),
#'     c("magrittr", "assertthat"), character(0),
#'     c("assertthat", "R6", "crayon", "rprojroot"),
#'     c("irlba", "magrittr", "Matrix", "pkgconfig"), character(0),
#'     character(0), "crayon", character(0), "Matrix",
#'     c("callr", "clisymbols", "crayon", "desc", "digest", "prettyunits",
#'       "R6", "rprojroot", "withr"),
#'     c("processx", "R6"), character(0), character(0)
#'   ))
#' )
#' tree(data)
#' tree(data, root = "rcmdcheck")
#'
#' # Colored nodes
#' data$label <- paste(data$package,
#'   style_dim(paste0("(", c("2.0.0.1", "1.1.1", "0.2.0", "1.2-11",
#'     "1.5", "1.2", "1.2.0", "1.0.2", "2.0.0", "1.1.1.9000", "1.1.2",
#'     "2.2.2", "1.3.4", "1.0.2", "0.6.12", "2.2.1", "1.2.1.9002",
#'     "1.0.0.9000", "2.0.1", "0.20-35"), ")"))
#'   )
#' roots <- ! data$package %in% unlist(data$dependencies)
#' data$label[roots] <- col_cyan(style_italic(data$label[roots]))
#' tree(data)
#' tree(data, root = "rcmdcheck")
#'
#' # Trimming
#' pkgdeps <- list(
#'   "dplyr@0.8.3" = c("assertthat@0.2.1", "glue@1.3.1", "magrittr@1.5",
#'     "R6@2.4.0", "Rcpp@1.0.2", "rlang@0.4.0", "tibble@2.1.3",
#'     "tidyselect@0.2.5"),
#'   "assertthat@0.2.1" = character(),
#'   "glue@1.3.1" = character(),
#'   "magrittr@1.5" = character(),
#'   "pkgconfig@2.0.3" = character(),
#'   "R6@2.4.0" = character(),
#'   "Rcpp@1.0.2" = character(),
#'   "rlang@0.4.0" = character(),
#'   "tibble@2.1.3" = c("cli@1.1.0", "crayon@1.3.4", "fansi@0.4.0",
#'      "pillar@1.4.2", "pkgconfig@2.0.3", "rlang@0.4.0"),
#'   "cli@1.1.0" = c("assertthat@0.2.1", "crayon@1.3.4"),
#'   "crayon@1.3.4" = character(),
#'   "fansi@0.4.0" = character(),
#'   "pillar@1.4.2" = c("cli@1.1.0", "crayon@1.3.4", "fansi@0.4.0",
#'      "rlang@0.4.0", "utf8@1.1.4", "vctrs@0.2.0"),
#'   "utf8@1.1.4" = character(),
#'   "vctrs@0.2.0" = c("backports@1.1.5", "ellipsis@0.3.0",
#'      "digest@0.6.21", "glue@1.3.1", "rlang@0.4.0", "zeallot@0.1.0"),
#'   "backports@1.1.5" = character(),
#'   "ellipsis@0.3.0" = c("rlang@0.4.0"),
#'   "digest@0.6.21" = character(),
#'   "glue@1.3.1" = character(),
#'   "zeallot@0.1.0" = character(),
#'   "tidyselect@0.2.5" = c("glue@1.3.1", "purrr@1.3.1", "rlang@0.4.0",
#'      "Rcpp@1.0.2"),
#'   "purrr@0.3.3" = c("magrittr@1.5", "rlang@0.4.0")
#' )
#'
#' pkgs <- data.frame(
#'   stringsAsFactors = FALSE,
#'   name = names(pkgdeps),
#'   deps = I(unname(pkgdeps))
#' )
#'
#' tree(pkgs)
#' tree(pkgs, trim = TRUE)
#'
#' # Mark the trimmed nodes
#' pkgs$label <- pkgs$name
#' pkgs$trimmed <- paste(pkgs$name, " (trimmed)")
#' tree(pkgs, trim = TRUE)

tree <- function(data, root = data[[1]][[1]], style = NULL,
                 width = console_width(), trim = FALSE) {
  assert_that(
    is.data.frame(data), ncol(data) >= 2,
    is_string(root),
    is.null(style) || (is_tree_style(style)),
    is_count(width)
  )

  style <- style %||% box_chars()

  labels <- if (ncol(data) >= 3) data[[3]] else data[[1]]
  trimlabs <- data$trimmed %||% labels

  seen <- character()
  res <- character()

  pt <- function(root, n = integer(), mx = integer(), used = character()) {

    num_root <- match(root, data[[1]])
    if (is.na(num_root)) return()

    level <- length(n) - 1
    prefix <- vcapply(seq_along(n), function(i) {
      if (n[i] < mx[i]) {
        if (i == length(n)) {
          paste0(style$j, style$h)
        } else {
          paste0(style$v, " ")
        }
      } else if (n[i] == mx[i] && i == length(n)) {
        paste0(style$l, style$h)
      } else {
        "  "
      }
    })

    root_seen <- root %in% seen
    root_lab <- if (trim && root_seen) trimlabs[[num_root]] else labels[[num_root]]
    res <<- c(res, paste0(paste(prefix, collapse = ""), root_lab))

    # Detect infinite loops
    if (!trim && root %in% used) {
      warning(call. = FALSE,
              "Endless loop found in tree: ",
              paste0(c(used, root), collapse = " -> "))
    } else if (! trim || ! root_seen) {
      seen <<- c(seen, root)
      children <- data[[2]][[num_root]]
      for (d in seq_along(children)) {
        pt(children[[d]], c(n, d), c(mx, length(children)), c(used, root))
      }
    }
  }

  if (nrow(data)) pt(root)

  res <- ansi_substr(res, 1, width)

  class(res) <- unique(c("tree", class(res), "character"))
  res
}

box_chars <- function() {
  if (is_utf8_output()) {
    list(
      "h" = "\u2500",                   # horizontal
      "v" = "\u2502",                   # vertical
      "l" = "\u2514",                   # leaf
      "j" = "\u251C"                    # junction
    )
  } else {
    list(
      "h" = "-",                        # horizontal
      "v" = "|",                        # vertical
      "l" = "\\",                       # leaf
      "j" = "+"                         # junction
    )
  }
}

#' @importFrom methods setOldClass

setOldClass(c("tree", "character"))

#' @export

print.tree <- function(x, ..., sep = "\n") {
  cat(x, ..., sep = sep)
  invisible(x)
}

test_that_cli("tree", {
  data <- data.frame(
    stringsAsFactors = FALSE,
    package = c("processx", "backports", "assertthat", "Matrix",
      "magrittr", "rprojroot", "clisymbols", "prettyunits", "withr",
      "desc", "igraph", "R6", "crayon", "debugme", "digest", "irlba",
      "rcmdcheck", "callr", "pkgconfig", "lattice"),
    dependencies = I(list(
      c("assertthat", "crayon", "debugme", "R6"), character(0),
      character(0), "lattice", character(0), "backports", character(0),
      c("magrittr", "assertthat"), character(0),
      c("assertthat", "R6", "crayon", "rprojroot"),
      c("irlba", "magrittr", "Matrix", "pkgconfig"), character(0),
      character(0), "crayon", character(0), "Matrix",
      c("callr", "clisymbols", "crayon", "desc", "digest", "prettyunits",
        "R6", "rprojroot", "withr"),
      c("processx", "R6"), character(0), character(0)
    ))
  )

  expect_snapshot(tree(data))

  expect_snapshot(tree(data, root = "desc"))

  # Check that trees with apparent circularity error nicely
  data <- data.frame(
    stringsAsFactors = FALSE,
    X = c("a", "b", "c","d", "e", "f", "g", "h", "j"),
    Y = I(list(
      c("b", "e", "f"),
      c("d", "g"),
      character(0),
      c("a", "h"),
      character(0),
      character(0),
      character(0),
      c("j"),
      character(0)
    ))
  )

  expect_warning(tree(data), "Endless loop found in tree: a -> b -> d -> a")
})

test_that_cli("trimming", {

  pkgdeps <- list(
    "dplyr@0.8.3" = c("assertthat@0.2.1", "glue@1.3.1", "magrittr@1.5",
      "R6@2.4.0", "Rcpp@1.0.2", "rlang@0.4.0", "tibble@2.1.3",
      "tidyselect@0.2.5"),
    "assertthat@0.2.1" = character(),
    "glue@1.3.1" = character(),
    "magrittr@1.5" = character(),
    "pkgconfig@2.0.3" = character(),
    "R6@2.4.0" = character(),
    "Rcpp@1.0.2" = character(),
    "rlang@0.4.0" = character(),
    "tibble@2.1.3" = c("cli@1.1.0", "crayon@1.3.4", "fansi@0.4.0",
       "pillar@1.4.2", "pkgconfig@2.0.3", "rlang@0.4.0"),
    "cli@1.1.0" = c("assertthat@0.2.1", "crayon@1.3.4"),
    "crayon@1.3.4" = character(),
    "fansi@0.4.0" = character(),
    "pillar@1.4.2" = c("cli@1.1.0", "crayon@1.3.4", "fansi@0.4.0",
       "rlang@0.4.0", "utf8@1.1.4", "vctrs@0.2.0"),
    "utf8@1.1.4" = character(),
    "vctrs@0.2.0" = c("backports@1.1.5", "ellipsis@0.3.0",
       "digest@0.6.21", "glue@1.3.1", "rlang@0.4.0", "zeallot@0.1.0"),
    "backports@1.1.5" = character(),
    "ellipsis@0.3.0" = c("rlang@0.4.0"),
    "digest@0.6.21" = character(),
    "glue@1.3.1" = character(),
    "zeallot@0.1.0" = character(),
    "tidyselect@0.2.5" = c("glue@1.3.1", "purrr@1.3.1", "rlang@0.4.0",
       "Rcpp@1.0.2"),
    "purrr@0.3.3" = c("magrittr@1.5", "rlang@0.4.0")
  )

  pkgs <- data.frame(
    stringsAsFactors = FALSE,
    name = names(pkgdeps),
    deps = I(unname(pkgdeps))
  )
  pkgs$label <- pkgs$name
  pkgs$trimmed <- paste(pkgs$name, " (trimmed)")

  expect_snapshot(tree(pkgs, trim = TRUE))
})

test_that("no warning for tibbles", {
  data <- tibble::tibble(
    package = c("A", "B", "C"),
    dependencies = c("B", "C", "")
  )

  expect_silent(cli::tree(data))
})

library(testthat)
library(cli)

## Run the tests in fancy mode and non-fancy mode as well
## Also, run them in latin1 encoding as well, this is for Unix,
## because Windows encoding names are different.

withr::with_options(
  list(cli.unicode = FALSE),
  test_check("cli")
)

withr::with_options(
  list(cli.unicode = TRUE),
  test_check("cli")
)

has_locale <- function(l) {
  has <- TRUE
  tryCatch(
    withr::with_locale(c(LC_CTYPE = l), "foobar"),
    warning = function(w) has <<- FALSE,
    error = function(e) has <<- FALSE
  )
  has
}

if (l10n_info()$`UTF-8` && has_locale("en_US.ISO8859-1")) {
  withr::with_locale(
    c(LC_CTYPE = "en_US.ISO8859-1"),
    test_check("cli")
  )
}

#' cli situation report
#'
#' Contains currently:
#' * `cli_unicode_option`: whether the `cli.unicode` option is set and its
#'   value. See [is_utf8_output()].
#' * `symbol_charset`: the selected character set for [symbol], UTF-8,
#'   Windows, or ASCII.
#' * `console_utf8`: whether the console supports UTF-8. See
#'   [base::l10n_info()].
#' * `latex_active`: whether we are inside knitr, creating a LaTeX
#'   document.
#' * `num_colors`: number of ANSI colors. See [num_ansi_colors()].
#' * `console_with`: detected console width.
#'
#' @return Named list with entries listed above. It has a `cli_sitrep`
#' class, with a `print()` and `format()` method.
#'
#' @export
#' @examples
#' cli_sitrep()

cli_sitrep <- function() {
  structure(
    list(
      cli_unicode_option = getOption("cli.unicode", NULL),
      symbol_charset = get_active_symbol_set(),
      console_utf8 = l10n_info()$`UTF-8`,
      latex_active = is_latex_output(),
      num_colors = num_ansi_colors(),
      console_width = console_width()),
    class = "cli_sitrep")
}

#' @export

print.cli_sitrep <- function(x, ...) {
  cat(format(x, ...), sep = "\n")
  invisible(x)
}

get_active_symbol_set <- function() {
  if (identical(symbol, symbol_utf8)) {
    "UTF-8"
  } else {
    "ASCII (non UTF-8)"
  }
}

#' @export

format.cli_sitrep <- function(x, ...) {
  fmt_names <- format(names(x))
  fmt_vals  <- vapply(x, format, character(1))
  paste0("- ", fmt_names, " : ", fmt_vals)
}

#' @export

as.character.cli_sitrep <- function(x, ...) {
  "<cli sitrep>"
}

test_that("sitrep runs", {
  expect_true(is.list(cli_sitrep()))
  expect_true(is.character(format(cli_sitrep())))
  out <- capture_output(print(cli_sitrep()))
  expect_true(all(grepl("^- ", out)))
})

test_that("get_active_symbol_set", {
  withr::with_options(list(cli.unicode = TRUE), {
    expect_true(get_active_symbol_set() %in% c("UTF-8", "RStudio (UTF-8)"))
  })
  withr::with_options(list(cli.unicode = FALSE), {
    set <- get_active_symbol_set()
    expect_equal(set, "ASCII (non UTF-8)")
  })
})

context("sitrep")

test_that("sitrep runs", {
  expect_true(is.list(cli_sitrep()))
  expect_true(is.character(format(cli_sitrep())))
  out <- capture_output(print(cli_sitrep()))
  expect_true(all(grepl("^- ", out)))
})

test_that("get_active_symbol_set", {
  withr::with_options(list(cli.unicode = TRUE), {
    expect_true(get_active_symbol_set() %in% c("UTF-8", "RStudio (UTF-8)"))
  })
  withr::with_options(list(cli.unicode = FALSE), {
    set <- get_active_symbol_set()
    expect_equal(set, "ASCII (non UTF-8)")
  })
})

#' Create a format method for an object using cli tools
#'
#' This method can be typically used in `format()` S3 methods. Then the
#' `print()` method of the class can be easily defined in terms of such a
#' `format()` method. See examples below.
#'
#' @param expr Expression that calls `cli_*` methods, [base::cat()] or
#' [base::print()] to format an object's printout.
#' @param theme Theme to use for the formatting.
#' @return Character vector, one element for each line of the printout.
#'
#' @export
#' @examples
#'
#' # Let's create format and print methods for a new S3 class that
#' # represents the an installed R package: `r_package`
#'
#' # An `r_package` will contain the DESCRIPTION metadata of the package
#' # and also its installation path.
#' new_r_package <- function(pkg) {
#'   tryCatch(
#'     desc <- packageDescription(pkg),
#'     warning = function(e) stop("Cannot find R package `", pkg, "`")
#'   )
#'   file <- dirname(attr(desc, "file"))
#'   if (basename(file) != pkg) file <- dirname(file)
#'   structure(
#'     list(desc = unclass(desc), lib = dirname(file)),
#'     class = "r_package"
#'   )
#' }
#'
#' format.r_package <- function(x, ...) {
#'   cli_format_method({
#'     cli_h1("{.pkg {x$desc$Package}} {cli::symbol$line} {x$desc$Title}")
#'     cli_text("{x$desc$Description}")
#'     cli_ul(c(
#'       "Version: {x$desc$Version}",
#'       if (!is.null(x$desc$Maintainer)) "Maintainer: {x$desc$Maintainer}",
#'       "License: {x$desc$License}"
#'     ))
#'     if (!is.na(x$desc$URL)) cli_text("See more at {.url {x$desc$URL}}")
#'   })
#' }
#'
#' # Now the print method is easy:
#' print.r_package <- function(x, ...) {
#'   cat(format(x, ...), sep = "\n")
#' }
#'
#' # Try it out
#' new_r_package("cli")
#'
#' # The formatting of the output depends on the current theme:
#' opt <- options(cli.theme = simple_theme())
#' print(new_r_package("cli"))
#' options(opt)  # <- restore theme

cli_format_method <- function(expr, theme = getOption("cli.theme")) {

  # This is not needed for cli, but needed for sink() and crayon
  nc <- num_ansi_colors()
  opts <- options(
    cli.num_colors = nc,
    crayon.enabled = nc > 1,
    crayon.colors = nc
  )
  on.exit(options(opts), add = TRUE)

  # Redirect everything to the connection
  con <- textConnection(NULL, open = "w", local = TRUE, encoding = "bytes")
  sink(con)
  on.exit(sink(NULL), add = TRUE)
  on.exit(close(con), add = TRUE)
  start_app(theme = theme, output = con)

  # Run the code
  withCallingHandlers(
    expr,
    cli_message = function(msg) {
      withCallingHandlers(
        cli_server_default(msg),
        cliMessage = function(msg2) {
          cat(conditionMessage(msg2), file = con, sep = "")
          invokeRestart("muffleMessage")
        }
      )
      invokeRestart("cli_message_handled")
    }
  )

  # Collect the output
  textConnectionValue(con)
}

# This is based on rlang:::obj_type_friendly, but adapted to cli

friendly_type <- local({

friendly_type <- function(x, value = TRUE, length = FALSE) {
  if (is_missing(x)) {
    return("absent")
  }

  if (is.object(x)) {
    if (inherits(x, "quosure")) {
      return("a {.cls quosure} object")
    } else if (identical(class(x), "data.frame")) {
      return("a data frame")
    } else if (identical(class(x), c("tbl_df", "tbl", "data.frame"))) {
      return("a tibble")
    } else {
      # this is sometimes wrong for 'h', but ce la vie
      fst <- tolower(substr(class(x)[1], 1, 1))
      prop <- if (fst %in% c("a", "e", "i", "o", "u")) {
        "an"
      } else {
        "a"
      }
      return(paste0(prop, " {.cls {class(x)[1]}} object"))
    }
  }

  if (!is_vector(x)) {
    return(as_friendly_type(typeof(x)))
  }

  n_dim <- length(dim(x))

  if (value && !n_dim) {
    if (is_na(x)) {
      return(switch(
        typeof(x),
        logical = "{.code NA}",
        integer = "an integer {.code NA}",
        double = "a numeric {.code NA}",
        complex = "a complex {.code NA}",
        character = "a character {.code NA}",
        typeof(x)
      ))
    }
    if (length(x) == 1 && !is_list(x)) {
      return(switch(
        typeof(x),
        logical = if (x) "{.code TRUE}" else "{.code FALSE}",
        integer = "an integer",
        double = "a number",
        complex = "a complex number",
        character = if (nzchar(x)) "a string" else "{.code \"\"}",
        raw = "a raw value",
        sprintf("a %s value", typeof(x))
      ))
    }
    if (length(x) == 0) {
      return(switch(
        typeof(x),
        logical = "an empty logical vector",
        integer = "an empty integer vector",
        double = "an empty numeric vector",
        complex = "an empty complex vector",
        character = "an empty character vector",
        raw = "an empty raw vector",
        list = "an empty list",
        sprintf("a %s of length one", typeof(x))
      ))
    }
  }

  type <- friendly_vector_type(typeof(x), n_dim)

  if (length && !n_dim) {
    type <- paste0(type, sprintf(" of length %s", length(x)))
  }

  type
}

friendly_vector_type <- function(type, n_dim) {
  if (type == "list") {
    if (n_dim < 2) {
      return("a list")
    } else if (n_dim == 2) {
      return("a list matrix")
    } else {
      return("a list array")
    }
  }

  type <- switch(
    type,
    logical = "a logical %s",
    integer = "an integer %s",
    numeric = ,
    double = "a double %s",
    complex = "a complex %s",
    character = "a character %s",
    raw = "a raw %s",
    type = paste0("a ", type, " %s")
  )

  if (n_dim < 2) {
    kind <- "vector"
  } else if (n_dim == 2) {
    kind <- "matrix"
  } else {
    kind <- "array"
  }
  sprintf(type, kind)
}

as_friendly_type <- function(type) {
  switch(
    type,

    list = "a list",

    NULL = "NULL",
    environment = "an environment",
    externalptr = "a pointer",
    weakref = "a weak reference",
    S4 = "an S4 object",

    name = ,
    symbol = "a symbol",
    language = "a call",
    pairlist = "a pairlist node",
    expression = "an expression vector",

    char = "an internal string",
    promise = "an internal promise",
    ... = "an internal dots object",
    any = "an internal {.code any} object",
    bytecode = "an internal bytecode object",

    primitive = ,
    builtin = ,
    special = "a primitive function",
    closure = "a function",

    type
  )
}

is_missing <- function(x) {
  missing(x) || identical(x, quote(expr = ))
}

is_vector <- function(x) {
  t <- typeof(x)
  t %in% c(
    "logical",
    "integer",
    "double",
    "complex",
    "character",
    "raw",
    "list"
  )
}

is_scalar_vector <- function(x) {
  is_vector(x) && length(x) == 1L
}

is_na <- function(x) {
  is_scalar_vector(x) && is.na(x)
}

is_list <- function(x) {
  typeof(x) == "list"
}

list(
  .internal = environment(),
  friendly_type = friendly_type
)

})

typename <- friendly_type$friendly_type

re_table <- function(...) {
  lapply(gregexpr(...), function(x) {
    res <- cbind(
      start = x,
      end = x + attr(x, "match.length") - 1,
      length = attr(x, "match.length")
    )
    res <- res[res[, "start"] != -1, , drop=FALSE]
  })
}

## Create the non-matching table from the matching table

non_matching <- function(table, str, empty = FALSE) {
  mapply(table, str, SIMPLIFY = FALSE, FUN = function(t, s) {
    if (! nrow(t)) {
      cbind(start = 1, end = base::nchar(s), length = base::nchar(s))
    } else {
      start <- c(1, t[, "end"] + 1)
      end <- c(t[, "start"] - 1, base::nchar(s))
      res <- cbind(start = start, end = end, length = end - start + 1)
      if (!empty) res[ res[, "length"] != 0, , drop = FALSE ] else res
    }
  })
}

myseq <- function(from, to, by = 1) {
  stopifnot(by != 0)
  if (by > 0) {
    if (to < from) {
      integer()
    } else {
      seq(from, to, by = by)
    }
  } else {
    if (to > from) {
      integer()
    } else {
      seq(from, to, by = by)
    }
  }
}

`%:%` <- myseq

test_that("re_table", {
  withr::local_options(
    cli.num_colors = 256,
    cli.hyperlink = TRUE
  )
  txt <- paste0(
    "this is some text ",
    col_red("red"),
    " some more text ",
    col_green("green"),
    " then some ",
    style_hyperlink("text", "https://example.com")
  )
  tbl <- re_table(ansi_regex(), txt)[[1]]
  tbl2 <- cbind(tbl, c(text = substring(txt, tbl[, "start"], tbl[, "end"])))
  expect_snapshot(tbl2)
  expect_snapshot(non_matching(list(tbl), txt))
})

test_that("re_table special cases", {
  withr::local_options(
    cli.num_colors = 256,
    cli.hyperlink = TRUE
  )
  txt <- "foobar"
  tbl <- re_table(ansi_regex(), txt)[[1]]
  expect_snapshot(tbl)
  expect_snapshot(non_matching(list(tbl), txt))
  expect_snapshot(non_matching(list(tbl), txt, empty = TRUE))

  txt <- col_red("foobar")
  tbl <- re_table(ansi_regex(), txt)[[1]]
  expect_snapshot(tbl)
  expect_snapshot(non_matching(list(tbl), txt))
  expect_snapshot(non_matching(list(tbl), txt, empty = TRUE))

  txt <- paste0("foo ", col_red(""), " bar")
  tbl <- re_table(ansi_regex(), txt)[[1]]
  expect_snapshot(tbl)
  expect_snapshot(non_matching(list(tbl), txt))
  expect_snapshot(non_matching(list(tbl), txt, empty = TRUE))
})

test_that("myseq", {
  expect_snapshot({
    myseq(1, 5)
    myseq(1, 1)
    myseq(1, 0)
    myseq(1, 5, 2)
    myseq(1, 6, 2)
    myseq(1, 1, 2)
    myseq(1, 2, -1)
    myseq(10, 1, -1)
    myseq(10, 1, -2)
    myseq(1, 5, -2)
  })
})

test_that_cli(configs = c("plain", "unicode"), "empty label", {
  expect_snapshot(boxx(""))
})

test_that_cli(configs = c("plain", "unicode"), "empty label 2", {
  expect_snapshot(boxx(character()))
})

test_that_cli(configs = c("plain", "unicode"), "label", {
  expect_snapshot(boxx("label"))
})

test_that_cli(configs = c("plain", "unicode"), "label vector", {
  expect_snapshot(boxx(c("label", "l2")))
})

test_that_cli(configs = c("plain", "unicode"), "border style", {
  expect_snapshot(boxx("label", border_style = "classic"))
})

test_that_cli(configs = c("plain", "unicode"), "padding", {
  expect_snapshot(boxx("label", padding = 2))
  expect_snapshot(boxx("label", padding = c(1,2,1,2)))
  expect_snapshot(boxx("label", padding = c(1,2,0,2)))
  expect_snapshot(boxx("label", padding = c(1,2,0,0)))
})

test_that_cli(configs = c("plain", "unicode"), "margin", {
  expect_snapshot(boxx("label", margin = 1))
  expect_snapshot(boxx("label", margin = c(1,2,3,4)))
  expect_snapshot(boxx("label", margin = c(0,1,2,0)))
})

test_that_cli(configs = c("plain", "unicode"), "float", {
  expect_snapshot(boxx("label", float = "center", width = 20))
  expect_snapshot(boxx("label", float = "right", width = 20))
})

test_that_cli("background_col", {
  expect_snapshot(boxx("label", background_col = "red"))
  expect_snapshot(boxx("label", background_col = col_red))
})

test_that_cli("border_col", {
  expect_snapshot(boxx("label", border_col = "red"))
  expect_snapshot(boxx("label", border_col = col_red))
})

test_that_cli(configs = c("plain", "unicode"), "align", {
  expect_snapshot(boxx(c("label", "l2"), align = "center"))
  expect_snapshot(boxx(c("label", "l2"), align = "right"))
})

test_that_cli(configs = c("plain", "unicode"), "header", {
  expect_snapshot(boxx("foobar", header = "foo"))
})

test_that_cli(configs = c("plain", "unicode"), "footer", {
  expect_snapshot(boxx("foobar", footer = "foo"))
})

is_interactive <- function() {
  opt <- getOption("rlib_interactive")
  if (isTRUE(opt)) {
    TRUE
  } else if (identical(opt, FALSE)) {
    FALSE
  } else if (tolower(getOption("knitr.in.progress", "false")) == "true") {
    FALSE
  } else if (tolower(getOption("rstudio.notebook.executing", "false")) == "true") {
    FALSE
  } else if (identical(Sys.getenv("TESTTHAT"), "true")) {
    FALSE
  } else {
    interactive()
  }
}

#' The connection option that cli would use
#'
#' Note that this only refers to the current R process. If the output
#' is produced in another process, then it is not relevant.
#'
#' In interactive sessions the standard output is chosen, othrwise the
#' standard error is used. This is to avoid painting output messages red
#' in the R GUIs.
#'
#' @return Connection object.
#'
#' @export

cli_output_connection <- function() {
  if (is_interactive() && no_sink()) stdout() else stderr()
}

no_sink <- function() {
  sink.number() == 0 && sink.number("message") == 2
}

is_stdout <- function(stream) {
  identical(stream, stdout()) && sink.number() == 0
}

is_stderr <- function(stream) {
  identical(stream, stderr()) && sink.number("message") == 2
}

is_stdx <- function(stream){
  is_stdout(stream) || is_stderr(stream)
}

is_rstudio_dynamic_tty <- function(stream) {
  rstudio$detect()[["dynamic_tty"]] &&
    (is_stdout(stream) || is_stderr(stream))
}

is_rapp <- function() {
  Sys.getenv("R_GUI_APP_VERSION") != ""
}

is_rapp_stdx <- function(stream) {
  interactive() &&
    is_rapp() &&
    (is_stdout(stream) || is_stderr(stream))
}

is_emacs <- function() {
  Sys.getenv("EMACS") != "" || Sys.getenv("INSIDE_EMACS") != ""
}

is_rkward <- function() {
  "rkward" %in% (.packages())
}

is_rkward_stdx <- function(stream) {
  interactive() &&
    is_rkward() &&
    (is_stdout(stream) || is_stderr(stream))
}

#' Detect whether a stream supports `\\r` (Carriage return)
#'
#' In a terminal, `\\r` moves the cursor to the first position of the
#' same line. It is also supported by most R IDEs. `\\r` is typically
#' used to achive a more dynamic, less cluttered user interface, e.g.
#' to create progress bars.
#'
#' If the output is directed to a file, then `\\r` characters are typically
#' unwanted. This function detects if `\\r` can be used for the given
#' stream or not.
#'
#' The detection mechanism is as follows:
#' 1. If the `cli.dynamic` option is set to `TRUE`, `TRUE` is returned.
#' 2. If the `cli.dynamic` option is set to anything else, `FALSE` is
#'    returned.
#' 3. If the `R_CLI_DYNAMIC` environment variable is not empty and set to
#'    the string `"true"`, `"TRUE"` or `"True"`, `TRUE` is returned.
#' 4. If `R_CLI_DYNAMIC` is not empty and set to anything else, `FALSE` is
#'    returned.
#' 5. If the stream is a terminal, then `TRUE` is returned.
#' 6. If the stream is the standard output or error within RStudio,
#'    the macOS R app, or RKWard IDE, `TRUE` is returned.
#' 7. Otherwise `FALSE` is returned.
#'
#' @param stream The stream to inspect or manipulate, an R connection
#' object. It can also be a string, one of `"auto"`, `"message"`,
#' `"stdout"`, `"stderr"`. `"auto"` will select `stdout()` if the session is
#' interactive and there are no sinks, otherwise it will select `stderr()`.
#'
#' @family terminal capabilities
#' @export
#' @examples
#' is_dynamic_tty()
#' is_dynamic_tty(stdout())

is_dynamic_tty <- function(stream = "auto") {

  stream <- get_real_output(stream)

  ## Option?
  if (!is.null(x <- getOption("cli.dynamic"))) {
    return(isTRUE(x))
  }

  ## Env var?
  if ((x <- Sys.getenv("R_CLI_DYNAMIC", "")) != "") {
    return(isTRUE(as.logical(x)))
  }

  ## Autodetect...
  ## RGui has isatty(stdout()) and isatty(stderr()), so we don't need
  ## to check that explicitly
  isatty(stream) ||
    is_rstudio_dynamic_tty(stream) ||
    is_rapp_stdx(stream) ||
    is_rkward_stdx(stream)
}

ANSI_ESC <- "\u001B["
ANSI_HIDE_CURSOR <- paste0(ANSI_ESC, "?25l")
ANSI_SHOW_CURSOR <- paste0(ANSI_ESC, "?25h")
ANSI_EL <- paste0(ANSI_ESC, "K")

#' Detect if a stream support ANSI escape characters
#'
#' We check that all of the following hold:
#' * The stream is a terminal.
#' * The platform is Unix.
#' * R is not running inside R.app (the macOS GUI).
#' * R is not running inside RStudio.
#' * R is not running inside Emacs.
#' * The terminal is not "dumb".
#' * `stream` is either the standard output or the standard error stream.
#'
#' @inheritParams is_dynamic_tty
#' @return `TRUE` or `FALSE`.
#'
#' @family terminal capabilities
#' @export
#' @examples
#' is_ansi_tty()

is_ansi_tty <- function(stream = "auto") {

  stream <- get_real_output(stream)

  # Option takes precedence
  opt <- getOption("cli.ansi")
  if (isTRUE(opt)) {
    return(TRUE)
  } else if (identical(opt, FALSE)) {
    return(FALSE)
  }

  # RStudio is handled separately
  if (rstudio$detect()[["ansi_tty"]] && is_stdx(stream)) return(TRUE)

  isatty(stream) &&
    .Platform$OS.type == "unix" &&
    !is_rapp() &&
    !is_emacs() &&
    Sys.getenv("TERM", "") != "dumb" &&
    is_stdx(stream)
}

#' Hide/show cursor in a terminal
#'
#' This only works in terminal emulators. In other environments, it
#' does nothing.
#'
#' `ansi_hide_cursor()` hides the cursor.
#'
#' `ansi_show_cursor()` shows the cursor.
#'
#' `ansi_with_hidden_cursor()` temporarily hides the cursor for
#' evaluating an expression.
#'
#' @inheritParams is_dynamic_tty
#' @param expr R expression to evaluate.
#'
#' @family terminal capabiltiies
#' @family low level ANSI functions
#' @export

ansi_hide_cursor <- function(stream = "auto") {
  stream <- get_real_output(stream)
  if (is_ansi_tty(stream)) cat(ANSI_HIDE_CURSOR, file = stream)
}

#' @export
#' @name ansi_hide_cursor

ansi_show_cursor <- function(stream = "auto") {
  stream <- get_real_output(stream)
  if (is_ansi_tty(stream)) cat(ANSI_SHOW_CURSOR, file = stream)
}

#' @export
#' @name ansi_hide_cursor

ansi_with_hidden_cursor <- function(expr, stream = "auto") {
  stream <- get_real_output(stream)
  ansi_hide_cursor(stream)
  on.exit(ansi_show_cursor(), add = TRUE)
  expr
}

#' Determine the width of the console
#'
#' It uses the `cli.width` option, if set. Otherwise it tries to
#' determine the size of the terminal or console window.
#'
#' These are the exact rules:
#' * If the `cli.width` option is set to a positive integer, it is used.
#' * If the `cli.width` option is set, but it is not a positive integer,
#'   and error is thrown.
#'
#' Then we try to determine the size of the terminal or console window:
#' * If we are not in RStudio, or we are in an RStudio temrinal,
#'   then we try to use the `ps::ps_tty_size()` function to query the
#'   terminal size. This might fail if ps is not installed or R is not
#'   running in a terminal, but failures are ignored.
#' * If we are in the RStudio build pane, then the `RSTUDIO_CONSOLE_WIDTH`
#'   environment variable is used. If the build pane is resized, then this
#'   environment variable is not accurate any more, and the output might
#'   get garbled.
#' * We are _not_ using the `RSTUDIO_CONSOLE_WIDTH` environment variable
#'   if we are in the RStudio console.
#'
#' If we cannot determine the size of the terminal or console window, then
#' we use the `width` option. If the `width` option is not set, then
#' we return 80L.
#'
#' @return Integer scalar, the console with, in number of characters.
#'
#' @export
#' @examples
#' console_width()

console_width <- function() {
  # cli.width option always takes priotity
  cwopt <- getOption("cli.width")
  if (!is.null(cwopt)) {
    if (!is.numeric(cwopt)) stop("options(\"cli.width\") must be integer")
    if (length(cwopt) != 1) stop("options(\"cli.width\") must be a scalar")
    if (!is.finite(cwopt)) stop("options(\"cli.width\") must be finite")
    if (is.na(cwopt)) stop("options(\"cli.width\") cannot be NA")
    cwopti <- as.integer(cwopt)
    if (cwopti <= 0) stop("options(\"cli.width\") must be a positive integer")
    return(cwopti)
  }

  # detect if in RStudio
  rs <- rstudio$detect()
  if (rs$type == "not_rstudio") {
    # maybe a terminal?
    width <- terminal_width()

  } else if (rs$type == "rstudio_console_starting") {
    # there isn't much we can do here, options and env vars are not set
    width <- NULL

  } else if (rs$type == "rstudio_console") {
    # will just use getOption("width"), in case the user changed it,
    # and ignore the RSTUDIO_CONSOLE_WIDTH env var
    width <- NULL

  } else if (rs$type == "rstudio_build_pane") {
    # RStudio explicitly sets this for build pane processes
    # It is only good when the build starts, but we cannot do better
    width <- rs_console_width()

  } else if (rs$type == "rstudio_terminal") {
    # Can also be a subprocess of the terminal, with a pty,
    # but that's fine, the pty should have a width set up.
    # We do not fall back to the RSTUDIO_CONSOLE_WIDTH env var,
    # because the user might have changed options("width") and the env
    # var is only good when the terminal starts, anyway.
    width <- terminal_width()

  } else { # rstudio_subprocess
    width <- NULL
  }

  # If not set, then use the option
  width <- width %||% getOption("width") %||% 80L

  width
}

terminal_width <- function() {
  w <- tryCatch(ps::ps_tty_size()[["width"]], error = function(e) NULL)
  # this is probably a pty that does not set the width, use st sensible
  if (!is.null(w) && w == 0) w <- 80L
  w
}

rs_console_width <- function() {
  ev <- as.integer(Sys.getenv("RSTUDIO_CONSOLE_WIDTH", ""))[1]
  if (!is.na(ev)) ev else NULL
}

test_that("errors", {
  withr::local_options(cli.width = letters)
  expect_snapshot_error(
    console_width()
  )

  withr::local_options(cli.width = NA_integer_)
  expect_snapshot_error(
    console_width()
  )

  withr::local_options(cli.width = -100L)
  expect_snapshot_error(
    console_width()
  )
})

#! /usr/bin/env Rscript

setup_app <- function() {
  theme <- list(
    "url" = list(color = "blue"),
    ".pkg" = list(color = "orange"),
    "it" = list("margin-bottom" = 1))
  start_app(theme = theme, output = "stdout")
}

load_packages <- function() {
  tryCatch({
    library(cli)
    library(httr)
    library(jsonlite)
    library(prettyunits)
    library(glue)
    library(parsedate)
    library(docopt) },
    error = function(e) {
      cli_alert_danger(
            "The {.pkg glue}, {.pkg httr}, {.pkg jsonlite}, {.pkg prettyunits},",
            " {.pkg parsedate} and {.pkg docopt} packages are needed!")
      q(save = "no", status = 1)
    })
}

news <- function(n = 10, day = FALSE, week = FALSE, since = NULL,
                 reverse = FALSE) {

  load_packages()
  setup_app()
  
  result <- if (day)
    news_day()
  else if (week)
    news_week()
  else if (!is.null(since))
    news_since(since)
  else
    news_n(as.numeric(n))

  if (reverse) result <- rev(result)

  format_results(result)
  invisible()
}

news_day <- function() {
  date <- format_iso_8601(Sys.time() - as.difftime(1, units="days"))
  ep <- glue("/-/pkgreleases?descending=true&endkey=%22{date}%22")
  do_query(ep)
}

news_week <- function() {
  date <- format_iso_8601(Sys.time() - as.difftime(7, units="days"))
  ep <- glue("/-/pkgreleases?descending=true&endkey=%22{date}%22")
  do_query(ep)
}

news_since <- function(since) {
  date <- format_iso_8601(parse_date(since))
  ep <- glue("/-/pkgreleases?descending=true&endkey=%22{date}%22")
  do_query(ep)
}

news_n <- function(n) {
  ep <- glue("/-/pkgreleases?limit={n}&descending=true")
  do_query(ep)
}

do_query <- function(ep) {
  base <- "https://crandb.r-pkg.org"
  url <- glue("{base}{ep}")
  response <- GET(url)
  stop_for_status(response)
  fromJSON(content(response, as = "text"), simplifyVector = FALSE)
}

format_results <- function(results) {
  cli_div(theme = list(ul = list("list-style-type" = "")))
  cli_ol()
  lapply(results, format_result)
}

parse_arguments <- function() {

  "Usage:
  news.R [-r | --reverse] [-n num ]
  news.R [-r | --reverse] --day | --week | --since date
  news.R [-h | --help]

Options:
  -n num        Show the last 'n' releases [default: 10].
  --day         Show releases in the last 24 hours
  --week        Show relaases in the last 7 * 24 hours
  --since date  Show releases since 'date'
  -r --reverse  Reverse the order, show older on top
  -h --help     Print this help message

New package releases on CRAN
" -> doc

  docopt(doc)
}

format_result <- function(result) {
  pkg <- result$package
  ago <- vague_dt(Sys.time() - parse_iso_8601(result$date))
  cli_li()
  cli_text("{.pkg {pkg$Package}} {pkg$Version} --
           {ago} by {.emph {pkg$Maintainer}}")
  cli_text("{pkg$Title}")
  cli_text("{.url https://r-pkg.org/pkg/{pkg$Package}}")
}

if (is.null(sys.calls())) {
  load_packages()
  opts <- parse_arguments()
  news(opts$n, opts$day, opts$week, opts$since, opts$reverse)
}

cray_wrapper_fun <- function(style) {
  style
  fun <- function(...) {
    txt <- paste0(...)
    structure(style(txt), class = "ansi_string")
  }
  class(fun) <- "ansi_style"
  attr(fun, "crayon_style") <- style
  fun
}

cray_wrapper <- function(name) {
  cray_wrapper_fun(asNamespace("crayon")[[name]])
}

#' @export

print.ansi_string <- function(x, ...) {
  cat("<ansi_string>\n")
  if (length(x)) {
    cat(format(paste0("[", seq_along(x), "] ", format(x))), sep = "\n")
  }
  invisible(x)
}

#' Create a new ANSI style
#'
#' Create a function that can be used to add ANSI styles to text.
#' All arguments are passed to [crayon::make_style()], but see the
#' Details below.
#'
#' @param ... The style to create. See details and examples below.
#' @param bg Whether the color applies to the background.
#' @param grey Whether to specifically create a grey color.
#'   This flag is included, because ANSI 256 has a finer color scale
#'   for greys, then the usual 0:5 scale for red, green and blue components.
#'   It is only used for RGB color specifications (either numerically
#'   or via a hexa string), and it is ignored on eigth color ANSI
#'   terminals.
#' @param colors Number of colors, detected automatically
#'   by default.
#' @return A function that can be used to color (style) strings.
#'
#' @details
#' The styles (elements of `...`) can be any of the
#' following:
#' * An R color name, see [grDevices::colors()].
#' * A 6- or 8-digit hexa color string, e.g. `#ff0000` means
#'   red. Transparency (alpha channel) values are ignored.
#' * A one-column matrix with three rows for the red, green
#'   and blue channels, as returned by [grDevices::col2rgb()].
#'
#' `make_ansistyle()` detects the number of colors to use
#' automatically (this can be overridden using the `colors`
#' argument). If the number of colors is less than 256 (detected or given),
#' then it falls back to the color in the ANSI eight color mode that
#' is closest to the specified (RGB or R) color.
#'
#'
#' @family ANSI styling
#' @export
#' @examples
#' make_ansi_style("orange")
#' make_ansi_style("#123456")
#' make_ansi_style("orange", bg = TRUE)
#'
#' orange <- make_ansi_style("orange")
#' orange("foobar")
#' cat(orange("foobar"))

make_ansi_style <- function(..., bg = FALSE, grey = FALSE,
                            colors = crayon::num_colors()) {

  dots <- lapply(list(...), function(x) {
    if (identical(x, "dim")) return("blurred") else x
  })
  args <- c(dots, list(bg = bg , grey = grey, colors = colors))
  style <- do.call(crayon::make_style, args)
  cray_wrapper_fun(style)
}

#' @export

print.ansi_style <- function(x, ...) {
  cat("<ansi_style>\n")
  cat(x("Example output"))
  cat("\n")
  invisible(x)
}


#' Combine two or more ANSI styles
#'
#' Combine two or more styles or style functions into a new style function
#' that can be called on strings to style them.
#'
#' It does not usually make sense to combine two foreground
#' colors (or two background colors), because only the first one
#' applied will be used.
#'
#' It does make sense to combine different kind of styles,
#' e.g. background color, foreground color, bold font.
#'
#' @param ... The styles to combine. For character strings, the
#'   [make_ansi_style()] function is used to create a style first.
#'   They will be applied from right to left.
#' @return The combined style function.
#'
#' @family ANSI styling
#' @export
#' @examples
#' ## Use style names
#' alert <- combine_ansi_styles("bold", "red4")
#' cat(alert("Warning!"), "\n")
#'
#' ## Or style functions
#' alert <- combine_ansi_styles(style_bold, col_red, bg_cyan)
#' cat(alert("Warning!"), "\n")
#'
#' ## Combine a composite style
#' alert <- combine_ansi_styles(
#'   "bold",
#'   combine_ansi_styles("red", bg_cyan))
#' cat(alert("Warning!"), "\n")

combine_ansi_styles <- function(...) {
  args <- list(...)
  args <- lapply(args, function(x) {
    if (inherits(x, "ansi_style")) attr(x, "crayon_style") else x
  })
  style <- do.call(crayon::combine_styles, args)
  cray_wrapper_fun(style)
}

#' ANSI colored text
#'
#' cli has a number of functions to color and style text at the command
#' line. These all use the crayon package under the hood, but provide a
#' slightly simpler interface.
#'
#' The `col_*` functions change the (foreground) color to the text.
#' These are the eight original ANSI colors. Note that in some terminals,
#' they might actually look differently, as terminals have their own
#' settings for how to show them.
#'
#' The `bg_*` functions change the background color of the text.
#' These are the eight original ANSI background colors. These, too, can
#' vary in appearence, depending on terminal settings.
#'
#' The `style_*` functions apply other styling to the text. The currently
#' supported styling funtions are:
#' * `style_reset()` to remove any style, including color,
#' * `style_bold()` for boldface / strong text, although some terminals
#'   show a bright, high intensity text instead,
#' * `style_dim()` (or `style_blurred()` reduced intensity text.
#' * `style_italic()` (not widely supported).
#' * `style_underline()`,
#' * `style_inverse()`,
#' * `style_hidden()`,
#' * `style_strikethrough() (not widely supported).
#'
#' The style functions take any number of character vectors as arguments,
#' and they concatenate them using `paste0()` before adding the style.
#'
#' Styles can also be nested, and then inner style takes precedence, see
#' examples below.
#'
#' @param ... Character strings, they will be pasted together with
#'   `paste0()`, before applying the style function.
#' @return An ANSI string (class `ansi_string`), that contains ANSI
#'   sequences, if the current platform supports them. You can simply
#'   use `cat()` to print them to the terminal.
#'
#' @family ANSI styling
#' @name ansi-styles
#' @examples
#' col_blue("Hello ", "world!")
#' cat(col_blue("Hello ", "world!"))
#'
#' cat("... to highlight the", col_red("search term"),
#'     "in a block of text\n")
#'
#' ## Style stack properly
#' cat(col_green(
#'  "I am a green line ",
#'  col_blue(style_underline(style_bold("with a blue substring"))),
#'  " that becomes green again!"
#' ))
#'
#' error <- combine_ansi_styles("red", "bold")
#' warn <- combine_ansi_styles("magenta", "underline")
#' note <- col_cyan
#' cat(error("Error: subscript out of bounds!\n"))
#' cat(warn("Warning: shorter argument was recycled.\n"))
#' cat(note("Note: no such directory.\n"))
#'
NULL

#' @export
#' @name ansi-styles
bg_black    <- cray_wrapper("bgBlack")
#' @export
#' @name ansi-styles
bg_blue     <- cray_wrapper("bgBlue")
#' @export
#' @name ansi-styles
bg_cyan     <- cray_wrapper("bgCyan")
#' @export
#' @name ansi-styles
bg_green    <- cray_wrapper("bgGreen")
#' @export
#' @name ansi-styles
bg_magenta  <- cray_wrapper("bgMagenta")
#' @export
#' @name ansi-styles
bg_red      <- cray_wrapper("bgRed")
#' @export
#' @name ansi-styles
bg_white    <- cray_wrapper("bgWhite")
#' @export
#' @name ansi-styles
bg_yellow   <- cray_wrapper("bgYellow")

#' @export
#' @name ansi-styles
col_black   <- cray_wrapper("black")
#' @export
#' @name ansi-styles
col_blue    <- cray_wrapper("blue")
#' @export
#' @name ansi-styles
col_cyan    <- cray_wrapper("cyan")
#' @export
#' @name ansi-styles
col_green   <- cray_wrapper("green")
#' @export
#' @name ansi-styles
col_magenta <- cray_wrapper("magenta")
#' @export
#' @name ansi-styles
col_red     <- cray_wrapper("red")
#' @export
#' @name ansi-styles
col_white   <- cray_wrapper("white")
#' @export
#' @name ansi-styles
col_yellow  <- cray_wrapper("yellow")
#' @export
#' @name ansi-styles
col_grey    <- cray_wrapper("silver")
#' @export
#' @name ansi-styles
col_silver  <- cray_wrapper("silver")

#' @export
#' @name ansi-styles
style_dim           <- cray_wrapper("blurred")
#' @export
#' @name ansi-styles
style_blurred       <- cray_wrapper("blurred")
#' @export
#' @name ansi-styles
style_bold          <- cray_wrapper("bold")
#' @export
#' @name ansi-styles
style_hidden        <- cray_wrapper("hidden")
#' @export
#' @name ansi-styles
style_inverse       <- cray_wrapper("inverse")
#' @export
#' @name ansi-styles
style_italic        <- cray_wrapper("italic")
#' @export
#' @name ansi-styles
style_reset         <- cray_wrapper("reset")
#' @export
#' @name ansi-styles
style_strikethrough <- cray_wrapper("strikethrough")
#' @export
#' @name ansi-styles
style_underline     <- cray_wrapper("underline")

test_that("one style", {
  testthat::skip_on_covr() # because we are comparing functions
  expect_equal(
    combine_ansi_styles(col_red),
    col_red,
    ignore_function_env = TRUE
  )
  expect_equal(
    combine_ansi_styles(style_bold),
    style_bold,
    ignore_function_env = TRUE
  )
})

test_that_cli(configs = c("plain", "ansi"), "style objects", {
  expect_equal(
    combine_ansi_styles(col_red, style_bold)("blah"),
    col_red(style_bold("blah"))
  )
  expect_equal(
    combine_ansi_styles(col_red, style_bold, style_underline)("foo"),
    col_red(style_bold(style_underline("foo")))
  )
})

test_that_cli(configs = c("plain", "ansi"), "create styles on the fly", {
  expect_equal(
    combine_ansi_styles("darkolivegreen", style_bold)("blah"),
    make_ansi_style("darkolivegreen")((style_bold("blah")))
  )
  expect_equal(
    combine_ansi_styles(style_bold, "darkolivegreen", style_underline)("foo"),
    style_bold(make_ansi_style("darkolivegreen")(style_underline("foo")))
  )
})

context("make_ansi_style")

test_that("make_style without name", {
  pink <- make_ansi_style("pink")
  expect_true(inherits(pink, "ansi_style"))
})

test_that("hexa color regex works", {
  positive <- c("#000000", "#ffffff", "#0f0f0f", "#f0f0f0",
                "#00000000", "#ffffffff", "#0f0f0f00", "#f0f0f055")

  negative <- c("", "#12345", "123456", "1234567", "12345678",
                "#1234567", "#1234ffg", "#gggggx", "foo#123456",
                "foo#123456bar")

  for (color in positive) {
    expect_true(grepl(hash_color_regex, color))
    expect_true(grepl(hash_color_regex, toupper(color)))
  }

  for (color in negative) {
    expect_false(grepl(hash_color_regex, color))
    expect_false(grepl(hash_color_regex, toupper(color)))
  }
})

test_that("we fall back for ANSI 8 if needed", {
  yellow3 <- make_ansi_style("yellow3", colors = 8)
  expect_equal(yellow3("foobar"), col_yellow("foobar"))
})

test_that("we can create a style from an R color", {
  red4 <- make_ansi_style("red4")
  red_text <- red4("text")
  expect_true(num_ansi_colors() == 1 || ansi_has_any(red_text))
})

context("ansi combine_ansi_styles")

test_that("one style", {
  testthat::skip_on_covr()
  expect_equal(combine_ansi_styles(col_red), col_red)
  expect_equal(combine_ansi_styles(style_bold), style_bold)
})

test_that("style objects", {
  withr::with_options(
    list(cli.num_colots = 256L), {
      expect_equal(
        combine_ansi_styles(col_red, style_bold)("blah"),
        col_red(style_bold("blah"))
      )
      expect_equal(
        combine_ansi_styles(col_red, style_bold, style_underline)("foo"),
        col_red(style_bold(style_underline("foo")))
      )
    }
  )
})

test_that("create styles on the fly", {
  withr::with_options(
    list(cli.num_colors = 256L), {
      expect_equal(
        combine_ansi_styles("darkolivegreen", style_bold)("blah"),
        make_ansi_style("darkolivegreen")((style_bold("blah")))
      )
      expect_equal(
        combine_ansi_styles(style_bold, "darkolivegreen", style_underline)("foo"),
        style_bold(make_ansi_style("darkolivegreen")(style_underline("foo")))
      )
    }
  )
})

test_that("Classes", {
  expect_equal(class(style_underline("foo")), c("cli_ansi_string", "ansi_string", "character"))
})

test_that("Coloring and highlighting works", {
  local_reproducible_output(crayon = TRUE)
  expect_equal(c(style_underline("foo")), "\u001b[4mfoo\u001b[24m")
  expect_equal(c(col_red("foo")), "\u001b[31mfoo\u001b[39m")
  expect_equal(c(bg_red("foo")), "\u001b[41mfoo\u001b[49m")
})

test_that("Applying multiple styles at once works", {
  local_reproducible_output(crayon = TRUE)
  st <- combine_ansi_styles(col_red, bg_green, "underline")
  expect_equal(
    c(st("foo")),
    "\u001b[31m\u001b[42m\u001b[4mfoo\u001b[24m\u001b[49m\u001b[39m")
  st <- combine_ansi_styles(style_underline, "red", bg_green)
  expect_equal(
    c(st("foo")),
    "\u001b[4m\u001b[31m\u001b[42mfoo\u001b[49m\u001b[39m\u001b[24m")
})

test_that("Nested styles are supported", {
  local_reproducible_output(crayon = TRUE)
  st <- combine_ansi_styles(style_underline, bg_blue)
  expect_equal(
    c(col_red("foo", st("bar"), "!")),
    "\u001b[31mfoo\u001b[4m\u001b[44mbar\u001b[49m\u001b[24m!\u001b[39m")
})

test_that("Nested styles of the same type are supported", {
  local_reproducible_output(crayon = TRUE)
  expect_equal(
    c(col_red("a", col_blue("b", col_green("c"), "b"), "c")),
    "\u001b[31ma\u001b[34mb\u001b[32mc\u001b[34mb\u001b[31mc\u001b[39m")
})

test_that("Reset all styles", {
  local_reproducible_output(crayon = TRUE)
  st <- combine_ansi_styles("red", bg_green, "underline")
  ok <- c(
    paste0(
      "\033[0m\033[31m\033[42m\033[4mfoo\033[24m\033[49m\033[39m",
      "foo\033[0m\033[22m\033[23m\033[24m\033[27m\033[28m",
      "\033[29m\033[39m\033[49m"),
    paste0("\u001b[0m\u001b[31m\u001b[42m\u001b[4mfoo\u001b[24m\u001b[49m",
           "\u001b[39mfoo\u001b[0m")
  )
  expect_true(style_reset(st("foo"), "foo") %in% ok)
})

test_that("Variable number of arguments", {
  local_reproducible_output(crayon = TRUE)
  expect_equal(c(col_red("foo", "bar")), "\u001b[31mfoobar\u001b[39m")
})

# modified from https://shiny.rstudio.com/articles/progress.html

# This app has a custom format string

library(cli)

# !!! You don't need these in real code.
# cli.progress_show_after makes sure that we see the progress bar from
# the beginning, not only after a delay.
options(cli.progress_show_after = 0)

# !!! You don't need these in real code.
# This also requests logging the progress bar to the standard output,
# in the console.
options(cli.progress_handlers_only = c("shiny", "logger"))

server <- function(input, output) {
  output$plot <- renderPlot({
    input$goPlot # Re-run when button is clicked
    
    # Create 0-row data frame which will be used to store data
    dat <- data.frame(x = numeric(0), y = numeric(0))
    
    # Number of times we'll go through the loop
    n <- 10

    cli_progress_bar(total = n, "Rendering plot", format = "Starting part {i}")
    for (i in 1:n) {
      # Each time through the loop, add another row of data. This is
      # a stand-in for a long-running computation.
      dat <- rbind(dat, data.frame(x = rnorm(1), y = rnorm(1)))

      cli_progress_update()

      # Pause for 0.1 seconds to simulate a long computation.
      Sys.sleep(0.5)
    }

    plot(dat$x, dat$y)
  })
}

ui <- shinyUI(basicPage(
  plotOutput('plot', width = "300px", height = "300px"),
  actionButton('goPlot', 'Go plot')
))

shinyApp(ui = ui, server = server)

is_yes <- function(x) {
  tolower(x) %in% c("true", "yes", "y", "t", "1")
}

format_iso_8601 <- function(p) {
  format(p, "%Y-%m-%dT%H:%M:%S+00:00")
}

has_packages <- function(pkgs) {
  all(vapply(pkgs, requireNamespace, logical(1), quietly = TRUE))
}

cli_escape <- function(x) {
  x <- gsub("{", "{{", x, fixed = TRUE)
  x <- gsub("}", "}}", x, fixed = TRUE)
  x
}

# missing from older R

isFALSE <- function (x) {
  is.logical(x) && length(x) == 1L && !is.na(x) && !x
}

get_ppid <- function() {
  .Call(clic_getppid)
}

box_styles <- function() {

  styles <- list(
    single = list(
      top_left     = "\u250c",
      top_right    = "\u2510",
      bottom_right = "\u2518",
      bottom_left  = "\u2514",
      vertical    = "\u2502",
      horizontal  = "\u2500"
    ),
    double = list(
      top_left = "\u2554",
      top_right = "\u2557",
      bottom_right = "\u255d",
      bottom_left = "\u255a",
      vertical = "\u2551",
      horizontal = "\u2550"
    ),
    round= list(
      top_left = "\u256d",
      top_right = "\u256e",
      bottom_right = "\u256f",
      bottom_left = "\u2570",
      vertical = "\u2502",
      horizontal = "\u2500"
    ),
    "single-double" = list(
      top_left = "\u2553",
      top_right = "\u2556",
      bottom_right = "\u255c",
      bottom_left = "\u2559",
      vertical = "\u2551",
      horizontal = "\u2500"
    ),
    "double-single" = list(
      top_left = "\u2552",
      top_right = "\u2555",
      bottom_right = "\u255b",
      bottom_left = "\u2558",
      vertical = "\u2502",
      horizontal = "\u2550"
    ),
    classic = list(
      top_left = "+",
      top_right = "+",
      bottom_right = "+",
      bottom_left = "+",
      vertical = "|",
      horizontal = "-"
    ),
    none = list(
      top_left = " ",
      top_right = " ",
      bottom_right = " ",
      bottom_left = " ",
      vertical = " ",
      horizontal = " "
    )
  )

  ## If the platform is not UTF-8, then we replace the styles that have
  ## Unicode characters, with the classic style.

  if (!is_utf8_output()) {
    for (n in setdiff(names(styles), c("classic", "none"))) {
      styles[[n]] <- styles[["classic"]]
    }
  }

  do.call(rbind, styles)
}

#' @export
#' @rdname boxx

list_border_styles <- function() {
  rownames(box_styles())
}

ansi_builtin_styles <- list(
  reset = list(0, c(0, 22, 23, 24, 27, 28, 29, 39, 49)),
  bold = list(1, 22), # 21 isn't widely supported and 22 does the same thing
  blurred = list(2, 22),
  italic = list(3, 23),
  underline = list(4, 24),
  inverse = list(7, 27),
  hidden = list(8, 28),
  strikethrough = list(9, 29),

  black = list(30, 39),
  red = list(31, 39),
  green = list(32, 39),
  yellow = list(33, 39),
  blue = list(34, 39),
  magenta = list(35, 39),
  cyan = list(36, 39),
  white = list(37, 39),
  silver = list(90, 39),

  bg_black = list(40, 49),
  bg_red = list(41, 49),
  bg_green = list(42, 49),
  bg_yellow = list(43, 49),
  bg_blue = list(44, 49),
  bg_magenta = list(45, 49),
  bg_cyan = list(46, 49),
  bg_white = list(47, 49)
)

is_builtin_style <- function(x) {
  is_string(x) && x %in% names(ansi_builtin_styles)
}

ansi_fg_r <- c(
  "black" = "black",
  "red" = "red",
  "green" = "green",
  "yellow" = "yellow",
  "blue" = "blue",
  "magenta" = "magenta",
  "cyan" = "cyan",
  "white" = "white",
  "silver" = "grey"
)

#' @importFrom grDevices col2rgb

ansi_fg_rgb <- col2rgb(ansi_fg_r)

ansi_bg_r <- c(
  "bg_black" = "black",
  "bg_red" = "red",
  "bg_green" = "green",
  "bg_yellow" = "yellow",
  "bg_blue" = "blue",
  "bg_magenta" = "magenta",
  "bg_cyan" = "cyan",
  "bg_white" = "white"
)

ansi_bg_rgb <- col2rgb(ansi_bg_r)

ansi_style_str <- function(x) {
  paste0("\u001b[", x, "m", collapse = "")
}

create_ansi_style_tag <- function(name, open, close) {
  structure(
    list(list(open = open, close = close)),
    names = name
  )
}

create_ansi_style_fun <- function(styles) {
  fun <- eval(substitute(function(...) {
    mystyles <- .styles
    txt <- paste0(...)
    if (num_ansi_colors() > 1) {
      for (st in rev(mystyles)) {
        txt <- paste0(
          st$open,
          gsub(st$close, st$open, txt, fixed = TRUE),
          st$close
        )
      }
    }
    class(txt) <- c("ansi_string", "character")
    txt
  }, list(.styles = styles)))

  class(fun) <- "ansi_style"
  attr(fun, "_styles") <- styles
  fun
}

create_ansi_style <- function(name, open = NULL, close = NULL) {
  open <- open %||% ansi_style_str(ansi_builtin_styles[[name]][[1]])
  close <- close %||% ansi_style_str(ansi_builtin_styles[[name]][[2]])
  style <- create_ansi_style_tag(name, open, close)
  create_ansi_style_fun(style)
}

#' @export

print.ansi_string <- function(x, ...) {
  cat("<ansi_string>\n")
  if (length(x)) {
    cat(format(paste0("[", seq_along(x), "] ", format(x))), sep = "\n")
  }
  invisible(x)
}

#' @export

print.ansi_style <- function(x, ...) {
  cat("<ansi_style>\n")
  cat(x("Example output"))
  cat("\n")
  invisible(x)
}

#' Create a new ANSI style
#'
#' Create a function that can be used to add ANSI styles to text.
#'
#' @param ... The style to create. See details and examples below.
#' @param bg Whether the color applies to the background.
#' @param grey Whether to specifically create a grey color.
#'   This flag is included, because ANSI 256 has a finer color scale
#'   for greys, then the usual 0:5 scale for red, green and blue components.
#'   It is only used for RGB color specifications (either numerically
#'   or via a hexa string), and it is ignored on eigth color ANSI
#'   terminals.
#' @param colors Number of colors, detected automatically
#'   by default.
#' @return A function that can be used to color (style) strings.
#'
#' @details
#' The `...` style argument can be any of the following:
#' * A cli ANSI style function of class `ansi_style`. This is returned
#'   as is, without looking at the other arguments.
#' * An R color name, see [grDevices::colors()].
#' * A 6- or 8-digit hexa color string, e.g. `#ff0000` means
#'   red. Transparency (alpha channel) values are ignored.
#' * A one-column matrix with three rows for the red, green
#'   and blue channels, as returned by [grDevices::col2rgb()].
#'
#' `make_ansi_style()` detects the number of colors to use
#' automatically (this can be overridden using the `colors`
#' argument). If the number of colors is less than 256 (detected or given),
#' then it falls back to the color in the ANSI eight color mode that
#' is closest to the specified (RGB or R) color.
#'
#'
#' @family ANSI styling
#' @export
#' @examples
#' make_ansi_style("orange")
#' make_ansi_style("#123456")
#' make_ansi_style("orange", bg = TRUE)
#'
#' orange <- make_ansi_style("orange")
#' orange("foobar")
#' cat(orange("foobar"))

make_ansi_style <- function(..., bg = FALSE, grey = FALSE,
                            colors = num_ansi_colors()) {

  style <- list(...)[[1]]
  if (inherits(style, "ansi_style")) return(style)
  if (inherits(style, "crayon")) {
    return(create_ansi_style_fun(attr(style, "_styles")))
  }

  if (identical(style, "dim")) style <- "blurred"

  orig_style_name <- style_name <- names(args)[1]

  stopifnot(is.character(style) && length(style) == 1 ||
            is_rgb_matrix(style) && ncol(style) == 1,
            is.logical(bg) && length(bg) == 1,
            is.numeric(colors) && length(colors) == 1)

  ansi_seqs <- if (is_builtin_style(style)) {
    if (bg && substr(style, 1, 3) != "bg_") {
      style <- paste0("bg_", style)
    }
    if (is.null(style_name)) style_name <- style
    ansi_builtin_styles[[style]]

  } else if (is_r_color(style)) {
    if (is.null(style_name)) style_name <- style
    ansi_style_from_r_color(style, bg, colors, grey)

  } else if (is_rgb_matrix(style)) {
    ansi_style_from_rgb(style, bg, colors, grey)

  } else {
    stop("Unknown style specification: ", style)
  }

  create_ansi_style(style_name, ansi_seqs$open, ansi_seqs$close)
}

hash_color_regex <- "^#([A-Fa-f0-9]{6}|[A-Fa-f0-9]{8})$"

#' @importFrom grDevices colors

is_r_color <- function(x) {
  x %in% colors() || grepl(hash_color_regex, x)
}

is_rgb_matrix <- function(x) {
  is.matrix(x) && is.numeric(x) && (nrow(x) == 3 || nrow(x) == 4)
}

ansi_style_from_r_color <- function(color, bg, num_colors, grey) {
  ansi_style_from_rgb(col2rgb(color), bg, num_colors, grey)
}

ansi_style_8_from_rgb <- function(rgb, bg) {
  ansi_cols <- if (bg) ansi_bg_rgb else ansi_fg_rgb
  dist <- colSums((ansi_cols - as.vector(rgb)) ^ 2 )
  builtin_name <- names(which.min(dist))[1]
  btn <- ansi_builtin_styles[[builtin_name]]
  list(open = ansi_style_str(btn[[1]]), close = ansi_style_str(btn[[2]]))
}

ansi_style_from_rgb <- function(rgb, bg, num_colors, grey) {
  if (num_colors < 256) { return(ansi_style_8_from_rgb(rgb, bg)) }
  ansi256(rgb, bg, grey)
}

# nocov start
fgcodes <- c(paste0('\x1b[38;5;', 0:255, 'm'), '\x1b[39m')
bgcodes <- c(paste0('\x1b[48;5;', 0:255, 'm'), '\x1b[49m')

rgb_index <- 17:232
gray_index <- 233:256
reset_index <- 257
#nocov end

ansi_scale <- function(x, from = c(0, 255), to = c(0, 5), round = TRUE) {
  y <- (x - from[1]) /
    (from[2] - from[1]) *
    (to[2] - to[1]) +
    to[1]

  if (round) {
    round(y)
  } else {
    y
  }
}

ansi256 <- function(rgb, bg = FALSE, grey = FALSE) {
  codes <- if (bg) bgcodes else fgcodes
  if (grey) {
    ## Gray
    list(
      open = codes[gray_index][ansi_scale(rgb[1], to = c(0, 23)) + 1],
      close = codes[reset_index]
    )

  } else {
    ## Not gray
    list(
      open = codes[ansi256_rgb_index(rgb[1L], rgb[2L], rgb[3L])],
      close = codes[reset_index]
    )
  }
}

## This is based off the algorithm in the ruby "paint" gem, as
## implemented in rainbowrite.
ansi256_rgb_index <- function(red, green, blue) {
  gray_possible <- TRUE
  sep <- 42.5
  while (gray_possible) {
    if (red < sep || green < sep || blue < sep) {
      gray <- red < sep && green < sep && blue < sep
      gray_possible <- FALSE
    }
    sep <- sep + 42.5
  }

  ## NOTE: The +1 here translates from base0 to base1 for the index
  ## that does the same.  Not ideal, but that does get the escape
  ## characters in nicely.
  if (gray) {
    232 + round((red + green + blue) / 33) + 1
  } else {
    16 + sum(floor(6 * c(red, green, blue) / 256) * c(36, 6, 1)) + 1
  }
}

#' Combine two or more ANSI styles
#'
#' Combine two or more styles or style functions into a new style function
#' that can be called on strings to style them.
#'
#' It does not usually make sense to combine two foreground
#' colors (or two background colors), because only the first one
#' applied will be used.
#'
#' It does make sense to combine different kind of styles,
#' e.g. background color, foreground color, bold font.
#'
#' @param ... The styles to combine. For character strings, the
#'   [make_ansi_style()] function is used to create a style first.
#'   They will be applied from right to left.
#' @return The combined style function.
#'
#' @family ANSI styling
#' @export
#' @examples
#' ## Use style names
#' alert <- combine_ansi_styles("bold", "red4")
#' cat(alert("Warning!"), "\n")
#'
#' ## Or style functions
#' alert <- combine_ansi_styles(style_bold, col_red, bg_cyan)
#' cat(alert("Warning!"), "\n")
#'
#' ## Combine a composite style
#' alert <- combine_ansi_styles(
#'   "bold",
#'   combine_ansi_styles("red", bg_cyan))
#' cat(alert("Warning!"), "\n")

combine_ansi_styles <- function(...) {
  styles <- lapply(
    list(...),
    function(x) attr(make_ansi_style(x), "_styles")
  )
  styles <- unlist(styles, recursive = FALSE)
  create_ansi_style_fun(styles)
}

test_that("Classes", {
  expect_equal(class(style_underline("foo")), c("cli_ansi_string", "ansi_string", "character"))
})

test_that("Coloring and highlighting works", {
  local_reproducible_output(crayon = TRUE)
  expect_equal(c(style_underline("foo")), "\u001b[4mfoo\u001b[24m")
  expect_equal(c(col_red("foo")), "\u001b[31mfoo\u001b[39m")
  expect_equal(c(bg_red("foo")), "\u001b[41mfoo\u001b[49m")
})

test_that("Applying multiple styles at once works", {
  local_reproducible_output(crayon = TRUE)
  st <- combine_ansi_styles(col_red, bg_green, "underline")
  expect_equal(
    c(st("foo")),
    "\u001b[31m\u001b[42m\u001b[4mfoo\u001b[24m\u001b[49m\u001b[39m")
  st <- combine_ansi_styles(style_underline, "red", bg_green)
  expect_equal(
    c(st("foo")),
    "\u001b[4m\u001b[31m\u001b[42mfoo\u001b[49m\u001b[39m\u001b[24m")
})

test_that("Nested styles are supported", {
  local_reproducible_output(crayon = TRUE)
  st <- combine_ansi_styles(style_underline, bg_blue)
  expect_equal(
    c(col_red("foo", st("bar"), "!")),
    "\u001b[31mfoo\u001b[4m\u001b[44mbar\u001b[49m\u001b[24m!\u001b[39m")
})

test_that("Nested styles of the same type are supported", {
  local_reproducible_output(crayon = TRUE)
  expect_equal(
    c(col_red("a", col_blue("b", col_green("c"), "b"), "c")),
    "\u001b[31ma\u001b[34mb\u001b[32mc\u001b[34mb\u001b[31mc\u001b[39m")
})

test_that("Reset all styles", {
  local_reproducible_output(crayon = TRUE)
  st <- combine_ansi_styles("red", bg_green, "underline")
  ok <- c(
    paste0(
      "\033[0m\033[31m\033[42m\033[4mfoo\033[24m\033[49m\033[39m",
      "foo\033[0m\033[22m\033[23m\033[24m\033[27m\033[28m",
      "\033[29m\033[39m\033[49m"),
    paste0("\u001b[0m\u001b[31m\u001b[42m\u001b[4mfoo\u001b[24m\u001b[49m",
           "\u001b[39mfoo\u001b[0m")
  )
  expect_true(style_reset(st("foo"), "foo") %in% ok)
})

test_that("Variable number of arguments", {
  local_reproducible_output(crayon = TRUE)
  expect_equal(c(col_red("foo", "bar")), "\u001b[31mfoobar\u001b[39m")
})

test_that("print.cli_ansi_style", {
  expect_snapshot(
    print(col_red)
  )
})

test_that("print.cli_ansi_string", {
  withr::local_options(cli.num_colors = 256)
  expect_snapshot(
    print(col_red("red"))
  )
})

test_that("ansi-scale", {
  expect_snapshot({
    ansi_scale(c(0,0,0))
    ansi_scale(c(255,100,0))
    ansi_scale(c(255,100,0), round = FALSE)
  })
})

test_that("zero length vectors", {
  withr::local_options(cli.num_colors = 1)
  expect_equal(length(col_cyan(character())), 0)
  expect_equal(length(bg_cyan(character())), 0)
  expect_equal(length(col_br_cyan(character())), 0)
  expect_equal(length(bg_br_cyan(character())), 0)

  withr::local_options(cli.num_colors = 8)
  expect_equal(length(col_cyan(character())), 0)
  expect_equal(length(bg_cyan(character())), 0)
  expect_equal(length(col_br_cyan(character())), 0)
  expect_equal(length(bg_br_cyan(character())), 0)

  withr::local_options(cli.num_colors = 256)
  expect_equal(length(col_cyan(character())), 0)
  expect_equal(length(bg_cyan(character())), 0)
  expect_equal(length(col_br_cyan(character())), 0)
  expect_equal(length(bg_br_cyan(character())), 0)

  withr::local_options(cli.num_colors = truecolor)
  expect_equal(length(col_cyan(character())), 0)
  expect_equal(length(bg_cyan(character())), 0)
  expect_equal(length(col_br_cyan(character())), 0)
  expect_equal(length(bg_br_cyan(character())), 0)
})

test_that_cli(configs = c("plain", "unicode"), "spark_bar", {
  expect_snapshot({
    spark_bar(seq(0, 1, length.out = 8))
    spark_bar(c(0, NA, 0.5, NA, 1))
  })
})

test_that_cli(configs = c("plain", "unicode"), "spark_line", {
  expect_snapshot({
    spark_line(seq(0, 1, length.out = 10))
  })
})

#' @useDynLib progresstest, .registration = TRUE, .fixes = "c_"
NULL

#' @export

test_baseline <- function() {
  .Call(c_test0)
}

#' @export

test_modulo <- function(progress = FALSE) {
  .Call(c_test00, progress)
}

#' @export

test_cli <- function() {
  .Call(c_test1)
}

#' @export

test_cli_unroll <- function() {
  this <- "test2"
  .Call(c_test2)
}

#' @export

testx <- function() {
  this <- "testx"
  .Call(c_testx)
}

test3 <- function() {
  this <- "test3"
  test2()
}

call_with_cleanup <- function(ptr, ...) {
  .Call(c_cleancall_call, pairlist(ptr, ...), parent.frame())
}

testc <- function() {
  this <- "testc"
  call_with_cleanup(c_testc)
}

#' Update the status bar (superseded)
#'
#' @description
#' **The `cli_status_*()` functions are superseded by
#' the [cli_progress_message()] and [cli_progress_step()] functions,
#' because they have a better default behavior.**
#'
#' The status bar is the last line of the terminal. cli apps can use this
#' to show status information, progress bars, etc. The status bar is kept
#' intact by all semantic cli output.
#'
#' @details
#' Use [cli_status_clear()] to clear the status bar.
#'
#' Often status messages are associated with processes. E.g. the app starts
#' downloading a large file, so it sets the status bar accordingly. Once the
#' download is done (or has failed), the app typically updates the status bar
#' again. cli automates much of this, via the `msg_done`, `msg_failed`, and
#' `.auto_result` arguments. See examples below.
#'
#' @param msg The text to show, a character vector. It will be
#'   collapsed into a single string, and the first line is kept and cut to
#'   [console_width()]. The message is often associated with the start of
#'   a calculation.
#' @param msg_done The message to use when the message is cleared, when
#'   the calculation finishes successfully. If `.auto_close` is `TRUE`
#'   and `.auto_result` is `"done"`, then this is printed automatically
#'   when the calling function (or `.envir`) finishes.
#' @param msg_failed The message to use when the message is cleared, when
#'   the calculation finishes unsuccessfully. If `.auto_close` is `TRUE`
#'   and `.auto_result` is `"failed"`, then this is printed automatically
#'   when the calling function (or `.envir`) finishes.
#' @param .keep What to do when this status bar is cleared. If `TRUE` then
#'   the content of this status bar is kept, as regular cli output (the
#'   screen is scrolled up if needed). If `FALSE`, then this status bar
#'   is deleted.
#' @param .auto_close Whether to clear the status bar when the calling
#'   function finishes (or `.envir` is removed from the stack, if
#'   specified).
#' @param .envir Environment to evaluate the glue expressions in. It is
#'   also used to auto-clear the status bar if `.auto_close` is `TRUE`.
#' @param .auto_result What to do when auto-closing the status bar.
#' @return The id of the new status bar container element, invisibly.
#'
#' @seealso Status bars support [inline markup][inline-markup].
#' @seealso The [cli_progress_message()] and [cli_progress_step()]
#'   functions, for a superior API.
#' @family status bar
#' @family functions supporting inline markup
#' @export

cli_status <- function(msg, msg_done = paste(msg, "... done"),
                       msg_failed = paste(msg, "... failed"),
                       .keep = FALSE, .auto_close = TRUE,
                       .envir = parent.frame(),
                       .auto_result = c("clear", "done", "failed", "auto")) {

  id <- new_uuid()
  cli__message(
    "status",
    list(
      id = id,
      msg = glue_cmd(msg, .envir = .envir),
      msg_done = glue_cmd(msg_done, .envir = .envir),
      msg_failed = glue_cmd(msg_failed, .envir = .envir),
      keep = .keep,
      auto_result = match.arg(.auto_result),
      globalenv = identical(.envir, .GlobalEnv)
    ),
    .auto_close = .auto_close,
    .envir = .envir
  )

  invisible(id)
}

#' Clear the status bar (superseded)
#'
#' @description
#' **The `cli_status_*()` functions are superseded by
#' the [cli_progress_message()] and [cli_progress_step()] functions,
#' because they have a better default behavior.**
#'
#' Clear the status bar
#'
#' @param id Id of the status bar container to clear. If `id` is not the id
#'   of the current status bar (because it was overwritten by another
#'   status bar container), then the status bar is not cleared. If `NULL`
#'   (the default) then the status bar is always cleared.
#' @param result Whether to show a message for success or failure or just
#'   clear the status bar.
#' @param msg_done If not `NULL`, then the message to use for successful
#'   process termination. This overrides the message given when the status
#'   bar was created.
#' @param msg_failed If not `NULL`, then the message to use for failed
#'   process termination. This overrides the message give when the status
#'   bar was created.
#' @inheritParams cli_status
#'
#' @family status bar
#' @seealso The [cli_progress_message()] and [cli_progress_step()]
#'   functions, for a superior API.
#' @export

cli_status_clear <- function(id = NULL, result = c("clear", "done", "failed"),
                             msg_done = NULL, msg_failed = NULL,
                             .envir = parent.frame()) {

  cli__message(
    "status_clear",
    list(
      id = id %||% NA_character_,
      result = match.arg(result[1], c("clear", "done", "failed", "auto")),
      msg_done = if (!is.null(msg_done)) glue_cmd(msg_done, .envir = .envir),
      msg_failed = if (!is.null(msg_failed)) glue_cmd(msg_failed, .envir = .envir)
    )
  )
}

#' Update the status bar (superseded)
#'
#' @description
#' **The `cli_status_*()` functions are superseded by
#' the [cli_progress_message()] and [cli_progress_step()] functions,
#' because they have a better default behavior.**
#'
#' Update the status bar
#'
#' @param msg Text to update the status bar with. `NULL` if you don't want
#'   to change it.
#' @param msg_done Updated "done" message. `NULL` if you don't want to
#'   change it.
#' @param msg_failed Updated "failed" message. `NULL` if you don't want to
#'   change it.
#' @param id Id of the status bar to update. Defaults to the current
#'   status bar container.
#' @param .envir Environment to evaluate the glue expressions in.
#' @return Id of the status bar container.
#'
#' @seealso This function supports [inline markup][inline-markup].
#' @seealso The [cli_progress_message()] and [cli_progress_step()]
#'   functions, for a superior API.
#' @family status bar
#' @family functions supporting inline markup
#' @export

cli_status_update <- function(id = NULL, msg = NULL, msg_done = NULL,
                              msg_failed = NULL, .envir = parent.frame()) {
  cli__message(
    "status_update",
    list(
      msg = if (!is.null(msg)) glue_cmd(msg, .envir = .envir),
      msg_done = if (!is.null(msg_done)) glue_cmd(msg_done, .envir = .envir),
      msg_failed = if (!is.null(msg_failed)) glue_cmd(msg_failed, .envir = .envir),
      id = id %||% NA_character_
    )
  )
}

#' Indicate the start and termination of some computation in the status bar
#' (superseded)
#'
#' @description
#' **The `cli_process_*()` functions are superseded by
#' the [cli_progress_message()] and [cli_progress_step()] functions,
#' because they have a better default behavior.**
#'
#' Typically you call `cli_process_start()` to start the process, and then
#' `cli_process_done()` when it is done. If an error happens before
#' `cli_process_done()` is called, then cli automatically shows the message
#' for unsuccessful termination.
#'
#' @details
#' If you handle the errors of the process or computation, then you can do
#' the opposite: call `cli_process_start()` with `on_exit = "done"`, and
#' in the error handler call `cli_process_failed()`. cli will automatically
#' call `cli_process_done()` on successful termination, when the calling
#' function finishes.
#'
#' See examples below.
#'
#' @param msg The message to show to indicate the start of the process or
#'   computation. It will be collapsed into a single string, and the first
#'   line is kept and cut to [console_width()].
#' @param msg_done The message to use for successful termination.
#' @param msg_failed The message to use for unsuccessful termination.
#' @param on_exit Whether this process should fail or terminate
#'   successfully when the calling function (or the environment in `.envir`)
#'   exits.
#' @param msg_class The style class to add to the message. Use an empty
#'   string to suppress styling.
#' @param done_class The style class to add to the successful termination
#'   message. Use an empty string to suppress styling.a
#' @param failed_class The style class to add to the unsuccessful
#'   termination message. Use an empty string to suppress styling.a
#' @inheritParams cli_status
#' @return Id of the status bar container.
#'
#' @seealso This function supports [inline markup][inline-markup].
#' @seealso The [cli_progress_message()] and [cli_progress_step()]
#'   functions, for a superior API.
#' @family status bar
#' @family functions supporting inline markup
#' @export
#' @examples
#'
#' ## Failure by default
#' fun <- function() {
#'   cli_process_start("Calculating")
#'   if (interactive()) Sys.sleep(1)
#'   if (runif(1) < 0.5) stop("Failed")
#'   cli_process_done()
#' }
#' tryCatch(fun(), error = function(err) err)
#'
#' ## Success by default
#' fun2 <- function() {
#'   cli_process_start("Calculating", on_exit = "done")
#'   tryCatch({
#'     if (interactive()) Sys.sleep(1)
#'     if (runif(1) < 0.5) stop("Failed")
#'   }, error = function(err) cli_process_failed())
#' }
#' fun2()

cli_process_start <- function(msg, msg_done = paste(msg, "... done"),
                              msg_failed = paste(msg, "... failed"),
                              on_exit = c("auto", "failed", "done"),
                              msg_class = "alert-info",
                              done_class = "alert-success",
                              failed_class = "alert-danger",
                              .auto_close = TRUE, .envir = parent.frame()) {

  # Force the defaults, because we might modify msg
  msg_done
  msg_failed

  if (length(msg_class) > 0 && msg_class != "") {
    msg <- paste0("{.", msg_class, " ", msg, "}")
  }
  if (length(done_class) > 0 && done_class != "") {
    msg_done <- paste0("{.", done_class, " ", msg_done, "}")
  }
  if (length(failed_class) > 0 && failed_class != "") {
    msg_failed <- paste0("{.", failed_class, " ", msg_failed, "}")
  }

  cli_status(msg, msg_done, msg_failed, .auto_close = .auto_close,
             .envir = .envir, .auto_result = match.arg(on_exit))
}

#' @param id Id of the status bar container to clear. If `id` is not the id
#'   of the current status bar (because it was overwritten by another
#'   status bar container), then the status bar is not cleared. If `NULL`
#'   (the default) then the status bar is always cleared.
#'
#' @rdname cli_process_start
#' @export

cli_process_done <- function(id = NULL, msg_done = NULL,
                             .envir = parent.frame(),
                             done_class = "alert-success") {

  if (!is.null(msg_done) && length(done_class) > 0 && done_class != "") {
    msg_done <- paste0("{.", done_class, " ", msg_done, "}")
  }
  cli_status_clear(id, result = "done", msg_done = msg_done, .envir = .envir)
}

#' @rdname cli_process_start
#' @export

cli_process_failed <- function(id = NULL, msg = NULL, msg_failed = NULL,
                               .envir = parent.frame(),
                               failed_class = "alert-danger") {
  if (!is.null(msg_failed) && length(failed_class) > 0 &&
      failed_class != "") {
    msg_failed <- paste0("{.", failed_class, " ", msg_failed, "}")
  }
  cli_status_clear(
    id,
    result = "failed",
    msg_failed = msg_failed,
    .envir = .envir
  )
}

# -----------------------------------------------------------------------

clii_status <- function(app, id, msg, msg_done, msg_failed, keep,
                        auto_result, globalenv) {

  app$status_bar[[id]] <- list(
    content = "",
    msg_done = msg_done,
    msg_failed = msg_failed,
    keep = keep,
    auto_result = auto_result
  )
  if (isTRUE(getOption("cli.hide_cursor", TRUE)) && !isTRUE(globalenv)) {
    ansi_hide_cursor(app$output)
  }
  clii_status_update(app, id, msg, msg_done = NULL, msg_failed = NULL)
}

clii_status_clear <- function(app, id, result, msg_done, msg_failed) {
  ## If NA then the most recent one
  if (is.na(id)) id <- names(app$status_bar)[1]

  ## If no active status bar, then ignore
  if (is.null(id) || is.na(id)) return(invisible())
  if (! id %in% names(app$status_bar)) return(invisible())

  if (result == "auto") {
    r1 <- random_marker
    if (identical(returnValue(r1), r1)) {
      result <- "failed"
    } else {
      result <- "done"
    }
  }

  if (result == "done") {
    msg <- msg_done %||% app$status_bar[[id]]$msg_done
    clii_status_update(app, id, msg, NULL, NULL)
    app$status_bar[[id]]$keep <- TRUE
  } else if (result == "failed") {
    msg <- msg_failed %||% app$status_bar[[id]]$msg_failed
    clii_status_update(app, id, msg, NULL, NULL)
    app$status_bar[[id]]$keep <- TRUE
  }

  if (names(app$status_bar)[1] == id) {
    ## This is the active one
    if (app$status_bar[[id]]$keep) {
      ## Keep? Just emit it
      app$cat("\n")

    } else {
      ## Not keep? Remove it
      clii__clear_status_bar(app)
    }
    if (isTRUE(getOption("cli.hide_cursor", TRUE))) {
      ansi_show_cursor(app$output)
    }

  } else {
    if (app$status_bar[[id]]$keep) {
      ## Keep?
      clii__clear_status_bar(app)
      app$cat(paste0(app$status_bar[[id]]$content, "\n"))
      app$cat(paste0(app$status_bar[[1]]$content, "\r"))

    } else {
      ## Not keep? Nothing to output
    }
  }

  ## Remove
  app$status_bar[[id]] <- NULL

  ## Switch to the previous one
  if (length(app$status_bar)) {
    app$cat(paste0(app$status_bar[[1]]$content, "\r"))
  }
}

clii_status_update <- function(app, id, msg, msg_done, msg_failed) {
  ## If NA then the most recent one
  if (is.na(id)) id <- names(app$status_bar)[1]

  ## If no active status bar, then ignore
  if (is.na(id)) return(invisible())

  ## Update messages
  if (!is.null(msg_done)) app$status_bar[[id]]$msg_done <- msg_done
  if (!is.null(msg_failed)) app$status_bar[[id]]$msg_failed <- msg_failed

  ## Do we have a new message?
  if (is.null(msg)) return(invisible())

  ## Do we need to clear the current content?
  current <- paste0("", app$status_bar[[1]]$content)

  ## Format the line
  content <- ""
  fmsg <- app$inline(msg)
  cfmsg <- ansi_strtrim(fmsg, width = app$get_width())
  content <- strsplit(cfmsg, "\r?\n")[[1]][1]
  if (is.na(content)) content <- ""

  ## Update status bar, put it in front
  app$status_bar[[id]]$content <- content
  app$status_bar <- c(
    app$status_bar[id],
    app$status_bar[setdiff(names(app$status_bar), id)])

  ## New content, if it is an ANSI terminal we'll overwrite and clear
  ## until the end of the line. Otherwise we add some space characters
  ## to the content to make sure we clear up residual content.
  output <- get_real_output(app$output)
  if (is_ansi_tty(output)) {
    app$cat(paste0("\r", content, ANSI_EL, "\r"))
  } else if (is_dynamic_tty(output)) {
    nsp <- max(ansi_nchar(current) - ansi_nchar(content), 0)
    app$cat(paste0("\r", content, strrep(" ", nsp), "\r"))
  } else {
    app$cat(paste0(content, "\n"))
  }

  ## Reset timer
  .Call(clic_tick_reset)

  invisible()
}

clii__clear_status_bar <- function(app) {
  output <- get_real_output(app$output)
  if (is_ansi_tty(output)) {
    app$cat(paste0("\r", ANSI_EL))
  } else if (is_dynamic_tty(output)) {
    text <- app$status_bar[[1]]$content
    len <- ansi_nchar(text, type = "width")
    app$cat(paste0("\r", strrep(" ", len + rstudio_r_fix), "\r"))
  }
}

start_app()
on.exit(stop_app(), add = TRUE)

# We can't easily use snapshot tests here, because they don't capture \r

test_that("create and clear", {
  f <- function() {
    cli_status("* This is the current status")
    cli_status_clear()
  }
  out <- ansi_strip(capt0(f()))
  expect_match(out, "* This is the current status", fixed = TRUE)
  out <- ansi_strip(capt0(f()))
  expect_match(out, "* This is the current status", fixed = TRUE)
})

test_that("output while status bar is active", {
  withr::local_options(list(cli.ansi = FALSE, cli.dynamic = TRUE))
  f <- function() {
    cli_text("out1")
    sb <- cli_status("status1")
    cli_text("out2")
    cli_status_update("status2", id = sb)
  }
  out <- ansi_strip(capt0(f()))
  expect_equal(out, paste0(
    "out1\n",
    "\rstatus1\r",
    "\r       \rout2\nstatus1\r",
    "\rstatus2\r",
    "\r       \r"))
})

test_that("interpolation", {
  withr::local_options(list(cli.ansi = FALSE, cli.dynamic = TRUE))
  f <- function() {
    cli_div(theme = list("span.pkg" = list("before" = "{", after = "}")))
    cli_status("You see 1+1={1+1}, this is {.pkg cli}")
    cli_status_clear()
  }
  out <- ansi_strip(capt0(f()))
  expect_equal(out, paste0(
    "\rYou see 1+1=2, this is {cli}\r",
    "\r                            \r"))
})

test_that("update", {
  withr::local_options(list(cli.ansi = FALSE, cli.dynamic = TRUE))
  f <- function() {
    cli_text("out1")
    sb <- cli_status("status1")
    cli_status_update("status2", id = sb)
  }
  out <- ansi_strip(capt0(f()))
  expect_equal(out, paste0(
    "out1\n",
    "\rstatus1\r",
    "\rstatus2\r",
    "\r       \r"))
})

test_that("keep", {
  withr::local_options(list(cli.ansi = FALSE, cli.dynamic = TRUE))
  f <- function() {
    cli_status("* This is the current status", .keep = TRUE)
    cli_status_clear()
  }
  out <- ansi_strip(capt0(f()))
  expect_equal(out, "\r* This is the current status\r\n")
})

test_that("multiple status bars", {
  withr::local_options(list(cli.ansi = FALSE, cli.dynamic = TRUE))
  f <- function() {
    sb1 <- cli_status("status1")
    cli_text("text1")
    sb2 <- cli_status("status2")
    cli_text("text2")
    cli_status_clear(sb2)
    cli_text("text3")
  }
  out <- ansi_strip(capt0(f()))
  expect_equal(out, paste0(
    "\rstatus1\r",
    "\r       \rtext1\nstatus1\r",      # emit text1, restore status1
    "\rstatus2\r",                      # show status2
    "\r       \rtext2\nstatus2\r",      # emit text2, restore status2
    "\r       \rstatus1\r",             # clear status2, restore status1
    "\r       \rtext3\nstatus1\r",      # emit text3, restore status1
    "\r       \r"))                     # (auto)clear status1
})

test_that("truncating", {
  withr::local_options(list(
    cli.ansi = FALSE,
    cli.dynamic = TRUE,
    cli.unicode = FALSE
  ))
  f <- function() {
    withr::local_options(list(cli.width = 40))
    txt <- "Eiusmod enim mollit aute aliquip Lorem sunt cupidatat."
    cli_status(c(txt, txt))
  }
  out <- ansi_strip(capt0(f()))
  expect_equal(out, paste0(
    "\rEiusmod enim mollit aute aliquip Lore...\r",
    "\r                                        \r"))
})

test_that("ansi colors and clearing", {
  withr::local_options(list(
    cli.num_colors = 256L,
    cli.ansi = FALSE,
    cli.dynamic = TRUE
  ))
  f <- function() {
    withr::local_options(list(num_ansi_colors = 256L))
    cli_status(col_red("This is red"))
    cli_status_clear()
  }
  out <- capt0(f())
  expect_match(out, "\033[31m", fixed = TRUE)
  expect_match(out, "\r           \r", fixed = TRUE)
})

test_that("theming status bar", {
  f <- function() {
    cli_text("out1")
    sb <- cli_status("{.alert-info status1}")
    cli_text("out2")
    cli_status_update("status2", id = sb)
  }
  out <- ansi_strip(capt0(f()))
  out2 <- ansi_strip(capt0(cli_alert_info("status1")))
  expect_match(out, str_trim(out2), fixed = TRUE)
})

test_that("successful termination", {
  withr::local_options(list(cli.ansi = FALSE, cli.dynamic = TRUE))
  f <- function() {
    cli_text("out1")
    sb <- cli_status("status1")
    cli_text("out2")
    cli_status_clear(result = "done")
  }
  out <- ansi_strip(capt0(f()))
  expect_equal(out, paste0(
    "out1\n",
    "\rstatus1\r",
    "\r       \rout2\nstatus1\r",
    "\rstatus1 ... done\r\n"
  ))
})

test_that("terminate with failed", {
  withr::local_options(list(cli.ansi = FALSE, cli.dynamic = TRUE))
  f <- function() {
    cli_text("out1")
    sb <- cli_status("status1")
    cli_text("out2")
    cli_status_clear(result = "failed")
  }
  out <- ansi_strip(capt0(f()))
  expect_equal(out, paste0(
    "out1\n",
    "\rstatus1\r",
    "\r       \rout2\nstatus1\r",
    "\rstatus1 ... failed\r\n"
  ))
})

test_that("auto close with success", {
  withr::local_options(list(cli.ansi = FALSE, cli.dynamic = TRUE))
  f <- function() {
    cli_text("out1")
    sb <- cli_status("status1", .auto_result = "done")
    cli_text("out2")
  }
  out <- ansi_strip(capt0(f()))
  expect_equal(out, paste0(
    "out1\n",
    "\rstatus1\r",
    "\r       \rout2\nstatus1\r",
    "\rstatus1 ... done\r\n"
  ))
})

test_that("auto close wtih failure", {
  withr::local_options(list(cli.ansi = FALSE, cli.dynamic = TRUE))
  f <- function() {
    cli_text("out1")
    sb <- cli_status("status1", .auto_result = "failed")
    if (is_interactive()) Sys.sleep(2)
    cli_text("out2")
    if (is_interactive()) Sys.sleep(2)
  }
  out <- ansi_strip(capt0(f()))
  expect_equal(out, paste0(
    "out1\n",
    "\rstatus1\r",
    "\r       \rout2\nstatus1\r",
    "\rstatus1 ... failed\r\n"
  ))
})

test_that("auto close with styling", {
  f <- function() {
    cli_text("out1")
    sb <- cli_status(
      msg = "{.alert-info status1}",
      msg_done = "{.alert-success status1 ... done}",
      msg_failed = "{.alert-danger status1 ... failed}",
      .auto_result = "failed"
    )
    if (is_interactive()) Sys.sleep(1)
    cli_text("out2")
    if (is_interactive()) Sys.sleep(1)
  }
  out <- ansi_strip(capt0(f()))
  expect_match(out, "status1 ... failed", fixed = TRUE)

  f2 <- function() {
    cli_text("out1")
    sb <- cli_status(
      msg = "{.alert-info status1}",
      msg_done = "{.alert-success status1 ... done}",
      msg_failed = "{.alert-danger status1 ... failed}",
      .auto_result = "done"
    )
    if (is_interactive()) Sys.sleep(1)
    cli_text("out2")
    if (is_interactive()) Sys.sleep(1)
  }
  out2 <- ansi_strip(capt0(f2()))
  expect_match(out2, "status1 ... done", fixed = TRUE)
})

test_that("process auto close with success", {
  f <- function() {
    cli_text("out1")
    sb <- cli_process_start("status1", on_exit = "done")
    cli_text("out2")
  }
  out <- ansi_strip(capt0(f()))
  expect_match(out, "status1 ... done")
})

test_that("process auto close with failure", {
  f <- function() {
    cli_text("out1")
    sb <- cli_process_start("status1", on_exit = "failed")
    if (is_interactive()) Sys.sleep(2)
    cli_text("out2")
    if (is_interactive()) Sys.sleep(2)
  }
  out <- ansi_strip(capt0(f()))
  expect_match(out, "status1 ... failed")
})

test_that("Multiple spaces are no condensed in a status bar", {
  f <- function() {
    cli_status("* This  is the current  status")
    cli_status_clear()
  }
  out <- ansi_strip(capt0(f()))
  expect_match(out, "* This  is the current  status", fixed = TRUE)
  out <- ansi_strip(capt0(f()))
  expect_match(out, "* This  is the current  status", fixed = TRUE)
})

test_that("Emojis are cleaned up properly", {
  skip_on_os("windows")
  withr::local_options(list(cli.ansi = FALSE, cli.dynamic = TRUE))
  f <- function() {
    cli_text("out1")
    sb <- cli_status("\U0001F477")
    cli_text("out2")
    cli_status_update("\u2728", id = sb)
  }
  out <- ansi_strip(capt0(f()))
  exps <- c(
    paste0(
      "out1\n",
      "\r\U0001F477\r",
      "\r  \rout2\n\U0001F477\r",
      "\r\u2728\r",
      "\r  \r"),
    paste0(
      "out1\n",
      "\r<U+0001F477>\r",
      "\r  \rout2\n<U+0001F477>\r",
      "\r<U+2728>\r",
      "\r  \r")
  )
  expect_true(out %in% exps)
})

test_that("auto-close with done or failure", {
  withr::local_options(list(
    cli.ansi = FALSE,
    cli.dynamic = TRUE,
    cli.unicode = FALSE
  ))
  f <- function() {
    cli_text("out1")
    sb <- cli_process_start("status1")
    cli_text("out2")
  }
  out <- ansi_strip(capt0(f()))
  expect_match(out, "status1 ... done")

  out <- ansi_strip(capt0(f()))
  expect_match(out, "status1 ... done")

  # This fails on older R versions, only if f2() is tryCatch()-ed.
  if (getRversion() < "3.5.0") skip("Needs R 3.5.0")

  f2 <- function() {
    cli_text("out1")
    sb <- cli_process_start("status1")
    cli_text("out2")
    stop("oops")
  }

  out2 <- ansi_strip(capt0(tryCatch(f2(), error = function(err) NULL)))
  expect_match(out2, fixed = TRUE, paste0(
    "out1\n",
    "\ri status1\r",
    "\r         \r",
    "out2\n",
    "i status1\r",
    "\rx status1 ... failed\r\n"
  ))
})

context("cli status bar")

setup(start_app())
teardown(stop_app())

test_that("create and clear", {
  f <- function() {
    cli_status("* This is the current status")
    cli_status_clear()
  }
  out <- crayon::strip_style(capt0(f()))
  expect_match(out, "* This is the current status", fixed = TRUE)
  out <- crayon::strip_style(capt0(f()))
  expect_match(out, "* This is the current status", fixed = TRUE)
})

test_that("output while status bar is active", {
  f <- function() {
    cli_text("out1")
    sb <- cli_status("status1")
    cli_text("out2")
    cli_status_update("status2", id = sb)
  }
  out <- crayon::strip_style(capt0(f()))
  expect_equal(out, paste0(
    "out1\n",
    "\r\rstatus1",
    "\r       \rout2\nstatus1",
    "\r       \rstatus2",
    "\r       \r"))
})

test_that("interpolation", {
  f <- function() {
    cli_div(theme = list("span.pkg" = list("before" = "{", after = "}")))
    cli_status("You see 1+1={1+1}, this is {.pkg cli}")
    cli_status_clear()
  }
  out <- crayon::strip_style(capt0(f()))
  expect_equal(out, paste0(
    "\r\rYou see 1+1=2, this is {cli}",
    "\r                            \r"))
})

test_that("update", {
  f <- function() {
    cli_text("out1")
    sb <- cli_status("status1")
    cli_status_update("status2", id = sb)
  }
  out <- crayon::strip_style(capt0(f()))
  expect_equal(out, paste0(
    "out1\n",
    "\r\rstatus1",
    "\r       \rstatus2",
    "\r       \r"))
})

test_that("keep", {
  f <- function() {
    cli_status("* This is the current status", .keep = TRUE)
    cli_status_clear()
  }
  out <- crayon::strip_style(capt0(f()))
  expect_equal(out, "\r\r* This is the current status\n")
})

test_that("multiple status bars", {
  f <- function() {
    sb1 <- cli_status("status1")
    cli_text("text1")
    sb2 <- cli_status("status2")
    cli_text("text2")
    cli_status_clear(sb2)
    cli_text("text3")
  }
  out <- crayon::strip_style(capt0(f()))
  expect_equal(out, paste0(
    "\r\rstatus1",
    "\r       \rtext1\nstatus1",        # emit text1, restore status1
    "\r       \rstatus2",               # show status2
    "\r       \rtext2\nstatus2",        # emit text2, restore status2
    "\r       \rstatus1",               # clear status2, restore status1
    "\r       \rtext3\nstatus1",        # emit text3, restore status1
    "\r       \r"))                     # (auto)clear status1
})

test_that("truncating", {
  f <- function() {
    withr::local_options(list(cli.width = 40))
    txt <- "Eiusmod enim mollit aute aliquip Lorem sunt cupidatat."
    cli_status(c(txt, txt))
  }
  out <- crayon::strip_style(capt0(f()))
  expect_equal(out, paste0(
    "\r\rEiusmod enim mollit aute aliquip Lorem",
    "\r                                      \r"))
})

test_that("ansi colors and clearing", {
  f <- function() {
    withr::local_options(list(crayon.enabled = TRUE, crayon.colors = 256))
    crayon::num_colors(forget = TRUE)
    cli_status(col_red("This is red"))
    cli_status_clear()
  }
  out <- capt0(f())
  expect_match(out, "\033[31m", fixed = TRUE)
  expect_match(out, "\r           \r", fixed = TRUE)
})

test_that("theming status bar", {
  f <- function() {
    cli_text("out1")
    sb <- cli_status("{.alert-info status1}")
    cli_text("out2")
    cli_status_update("status2", id = sb)
  }
  out <- crayon::strip_style(capt0(f()))
  out2 <- crayon::strip_style(capt0(cli_alert_info("status1")))
  expect_match(out, str_trim(out2), fixed = TRUE)
})

test_that("successful termination", {
  f <- function() {
    cli_text("out1")
    sb <- cli_status("status1")
    cli_text("out2")
    cli_status_clear(result = "done")
  }
  out <- crayon::strip_style(capt0(f()))
  expect_equal(out, paste0(
    "out1\n",
    "\r\rstatus1",
    "\r       \rout2\nstatus1",
    "\r       \rstatus1 ... done\n"
  ))
})

test_that("terminate with failed", {
  f <- function() {
    cli_text("out1")
    sb <- cli_status("status1")
    cli_text("out2")
    cli_status_clear(result = "failed")
  }
  out <- crayon::strip_style(capt0(f()))
  expect_equal(out, paste0(
    "out1\n",
    "\r\rstatus1",
    "\r       \rout2\nstatus1",
    "\r       \rstatus1 ... failed\n"
  ))
})

test_that("auto close with success", {
  f <- function() {
    cli_text("out1")
    sb <- cli_status("status1", .auto_result = "done")
    cli_text("out2")
  }
  out <- crayon::strip_style(capt0(f()))
  expect_equal(out, paste0(
    "out1\n",
    "\r\rstatus1",
    "\r       \rout2\nstatus1",
    "\r       \rstatus1 ... done\n"
  ))
})

test_that("auto close wtih failure", {
  f <- function() {
    cli_text("out1")
    sb <- cli_status("status1", .auto_result = "failed")
    if (is_interactive()) Sys.sleep(2)
    cli_text("out2")
    if (is_interactive()) Sys.sleep(2)
  }
  out <- crayon::strip_style(capt0(f()))
  expect_equal(out, paste0(
    "out1\n",
    "\r\rstatus1",
    "\r       \rout2\nstatus1",
    "\r       \rstatus1 ... failed\n"
  ))
})

test_that("auto close with styling", {
  f <- function() {
    cli_text("out1")
    sb <- cli_status(
      msg = "{.alert-info status1}",
      msg_done = "{.alert-success status1 ... done}",
      msg_failed = "{.alert-danger status1 ... failed}",
      .auto_result = "failed"
    )
    if (is_interactive()) Sys.sleep(1)
    cli_text("out2")
    if (is_interactive()) Sys.sleep(1)
  }
  out <- crayon::strip_style(capt0(f()))
  expect_match(out, "status1 ... failed\n")

  f2 <- function() {
    cli_text("out1")
    sb <- cli_status(
      msg = "{.alert-info status1}",
      msg_done = "{.alert-success status1 ... done}",
      msg_failed = "{.alert-danger status1 ... failed}",
      .auto_result = "done"
    )
    if (is_interactive()) Sys.sleep(1)
    cli_text("out2")
    if (is_interactive()) Sys.sleep(1)
  }
  out2 <- crayon::strip_style(capt0(f2()))
  expect_match(out2, "status1 ... done\n")
})

test_that("process auto close with success", {
  f <- function() {
    cli_text("out1")
    sb <- cli_process_start("status1", on_exit = "done")
    cli_text("out2")
  }
  out <- crayon::strip_style(capt0(f()))
  expect_match(out, "status1 ... done")
})

test_that("process auto close with failure", {
  f <- function() {
    cli_text("out1")
    sb <- cli_process_start("status1", on_exit = "failed")
    if (is_interactive()) Sys.sleep(2)
    cli_text("out2")
    if (is_interactive()) Sys.sleep(2)
  }
  out <- crayon::strip_style(capt0(f()))
  expect_match(out, "status1 ... failed")
})

`%||%` <- function(l, r) if (is.null(l)) r else l

new_class <- function(class_name, ...) {
  structure(as.environment(list(...)), class = class_name)
}

make_space <- function(len) {
  strrep(" ", len)
}

strrep <- function(x, times) {
  x <- as.character(x)
  if (length(x) == 0L) return(x)
  r <- .mapply(
    function(x, times) {
      if (is.na(x) || is.na(times)) return(NA_character_)
      if (times <= 0L) return("")
      paste0(replicate(times, x), collapse = "")
    },
    list(x = x, times = times),
    MoreArgs = list()
  )

  res <- unlist(r, use.names = FALSE)
  Encoding(res) <- Encoding(x)
  res
}

is_latex_output <- function() {
  if (!("knitr" %in% loadedNamespaces())) return(FALSE)
  get("is_latex_output", asNamespace("knitr"))()
}

is_windows <-  function() {
  .Platform$OS.type == "windows"
}

apply_style <- function(text, style, bg = FALSE) {
  if (identical(text, ""))
    return(text)

  if (is.function(style)) {
    style(text)
  } else if (is.character(style)) {
    make_ansi_style(style, bg = bg)(text)
  } else if (is.null(style)) {
    text
  } else {
    throw(cli_error(
      "{.arg style} must be a color name or an ANSI style function",
      "i" = "{.arg style} is {.typeof {style}}"
    ))
  }
}

vcapply <- function(X, FUN, ..., USE.NAMES = TRUE) {
  vapply(X, FUN, FUN.VALUE = character(1), ..., USE.NAMES = USE.NAMES)
}

viapply <- function(X, FUN, ..., USE.NAMES = TRUE) {
  vapply(X, FUN, FUN.VALUE = integer(1), ..., USE.NAMES = USE.NAMES)
}

vlapply <- function(X, FUN, ..., USE.NAMES = TRUE) {
  vapply(X, FUN, FUN.VALUE = logical(1), ..., USE.NAMES = USE.NAMES)
}

rpad <- function(x, width = NULL) {
  if (!length(x)) return(x)
  w <- nchar(x, type = "width")
  if (is.null(width)) width <- max(w)
  paste0(x, strrep(" ", pmax(width - w, 0)))
}

lpad <- function(x, width = NULL) {
  if (!length(x)) return(x)
  w <- nchar(x, type = "width")
  if (is.null(width)) width <- max(w)
  paste0(strrep(" ", pmax(width - w, 0)), x)
}

tail_na <- function(x, n = 1) {
  utils::tail(c(rep(NA, n), x), n)
}

dedent <- function(x, n = 2) {
  first_n_char <- strsplit(ansi_substr(x, 1, n), "", fixed = TRUE)[[1]]
  n_space <- cumsum(first_n_char == " ")
  d_n_space <- diff(c(0, n_space))
  first_not_space <- utils::head(c(which(d_n_space == 0), n + 1), 1)
  ansi_substr(x, first_not_space, nchar(x))
}

new_uuid <- (function() {
  cnt <- 0
  function() {
    cnt <<- cnt + 1
    paste0("cli-", clienv$pid, "-", cnt)
  }
})()

na.omit <- function(x) {
  if (is.atomic(x)) x[!is.na(x)] else x
}

last <- function(x) {
  utils::tail(x, 1)[[1]]
}

str_tail <- function(x) {
  substr(x, 2, nchar(x))
}

push <- function(l, el, name = NULL) {
  c(l, structure(list(el), names = name))
}

try_silently <- function(expr) {
  suppressWarnings(tryCatch(expr, error = function(x) x))
}

random_id <- local({
  i <- 0
  function() {
    i <<- i + 1
    paste0("FCkNXbE-", i)
  }
})

random_marker <- "ImzV8dciA4cn4POI"

str_trim <- function(x) {
  sub("^\\s+", "", sub("\\s+$", "", x))
}

last_character <- function(x) {
  substr(x, nchar(x), nchar(x))
}

first_character <- function(x) {
  substr(x, 1, 1)
}

second_character <- function(x) {
  substr(x, 2, 2)
}

is_alnum <- function(x, ok = "") {
  grepl(paste0("^[[:alnum:]/_.", ok, "]*$"), x)
}

os_type <- function() {
  .Platform$OS.type
}

leading_space <- function(x) {
  sub("^([\\s\u00a0]*).*$", "\\1", x, perl = TRUE)
}

trailing_space <- function(x) {
  sub("^.*[^\\s\u00a0]([\\s\u00a0]*)$", "\\1", x, perl = TRUE)
}

get_rstudio_theme <- function() {
  suppressWarnings(rstudioapi::getThemeInfo())
}

# ansi_strtrim might not support NAs

abbrev <- function(x, len = 10) {
  # this is better than strtrim() because it adds ...
  ansi_strtrim(x, len)
}

context("utils")

test_that("make_space", {
  expect_equal(make_space(0), "")
  expect_equal(make_space(1), " ")
  expect_equal(make_space(5), "     ")
})

test_that("apply_style", {
  expect_error(
    apply_style("text", raw(0)),
    "Not a colour name or ANSI style"
  )
})

test_that("viapply", {
  expect_equal(
    viapply(c("foo", "foobar"), length),
    vapply(c("foo", "foobar"), length, integer(1))
  )

  expect_equal(
    viapply(character(), length),
    vapply(character(), length, integer(1))
  )
})

test_that("ruler", {
  out <- capt00(ruler(20))
  exp <- rebox(
    "----+----1----+----2",
    "12345678901234567890"
  )
  expect_equal(ansi_strip(out), exp)
})

test_that("rpad", {
  expect_equal(rpad(character()), character())
  expect_equal(rpad("foo"), "foo")
  expect_equal(rpad(c("foo", "foobar")), c("foo   ", "foobar"))
})

test_that("lpad", {
  expect_equal(lpad(character()), character())
  expect_equal(lpad("foo"), "foo")
  expect_equal(lpad(c("foo", "foobar")), c("   foo", "foobar"))
})

test_that("is_utf8_output", {

  mockery::stub(
    is_utf8_output, "l10n_info",
    list(MBCS = TRUE, `UTF-8` = TRUE, `Latin-1` = FALSE)
  )
  withr::with_options(
    list(cli.unicode = NULL),
    expect_true(is_utf8_output())
  )

  mockery::stub(
    is_utf8_output, "l10n_info",
    list(MBCS = FALSE, `UTF-8` = FALSE, `Latin-1` = TRUE)
  )
  withr::with_options(
    list(cli.unicode = NULL),
    expect_false(is_utf8_output())
  )
})

test_that("is_latex_output", {

  mockery::stub(is_latex_output, "loadedNamespaces", "foobar")
  expect_false(is_latex_output())

  mockery::stub(is_latex_output, "loadedNamespaces", "knitr")
  mockery::stub(
    is_latex_output, "get",
    function(x, ...) {
      if (x == "is_latex_output") {
        function() TRUE
      } else {
        base::get(x, ...)
      }
    }
  )
  expect_true(is_latex_output())
})

test_that("dedent", {
  cases <- list(
    list("", 0, ""),
    list("", 1, ""),
    list("", 2, ""),
    list("x", 0, "x"),
    list("x", 1, "x"),
    list("x", 2, "x"),
    list("xx", 0, "xx"),
    list("xx", 1, "xx"),
    list("xx", 2, "xx"),
    list("foobar", 0, "foobar"),
    list("foobar", 1, "foobar"),
    list("foobar", 2, "foobar"),

    list(" ", 0, " "),
    list(" ", 1, ""),
    list(" ", 2, ""),
    list("  ", 0, "  "),
    list("  ", 1, " "),
    list("  ", 2, ""),
    list(" x", 0, " x"),
    list(" x", 1, "x"),
    list(" x", 2, "x"),
    list("  x", 0, "  x"),
    list("  x", 1, " x"),
    list("  x", 2, "x"),

    list(" x  y", 3, "x  y"),
    list(" x  y", 4, "x  y"),
    list(" x  y", 5, "x  y"),
    list(" x  ", 3, "x  "),
    list(" x  ", 4, "x  "),
    list(" x  ", 5, "x  ")
  )

  for (c in cases) expect_identical(dedent(c[[1]], c[[2]]), ansi_string(c[[3]]))
})

test_that("tail_na", {
  cases <- list(
    list(1:4, 4L),
    list(1, 1),
    list(double(), NA_real_),
    list(character(), NA_character_)
  )

  for (i in seq_along(cases)) {
    c <- cases[[i]]
    expect_identical(tail_na(c[[1]]), c[[2]], info = i)
  }

  cases2 <- list(
    list(1:4, 2, 3:4),
    list(1, 2, c(NA_real_, 1)),
    list(double(), 2, c(NA_real_, NA_real_)),
    list(character(), 2, c(NA_character_, NA_character_))
  )

  for (i in seq_along(cases2)) {
    c <- cases2[[i]]
    expect_identical(tail_na(c[[1]], c[[2]]), c[[3]], info = i)
  }
})

test_that("is_windows", {
  expect_equal(is_windows(), .Platform$OS.type == "windows")
})

test_that("make_space", {
  expect_equal(make_space(0), "")
  expect_equal(make_space(1), " ")
  expect_equal(make_space(5), "     ")
})

test_that("apply_style", {
  expect_error(
    apply_style("text", raw(0)),
    "must be a color name or an ANSI style function"
  )
  expect_equal(
    apply_style("foo", function(x) toupper(x)),
    "FOO"
  )

  withr::local_options(cli.num_ansi_colors = truecolor)
  expect_equal(
    apply_style("foo", "red"),
    col_red("foo")
  )
})

test_that("viapply", {
  expect_equal(
    viapply(c("foo", "foobar"), length),
    vapply(c("foo", "foobar"), length, integer(1))
  )

  expect_equal(
    viapply(character(), length),
    vapply(character(), length, integer(1))
  )
})

test_that("ruler", {
  expect_snapshot(
    ruler(20)
  )
})

test_that("rpad", {
  expect_equal(rpad(character()), character())
  expect_equal(rpad("foo"), "foo")
  expect_equal(rpad(c("foo", "foobar")), c("foo   ", "foobar"))
})

test_that("lpad", {
  expect_equal(lpad(character()), character())
  expect_equal(lpad("foo"), "foo")
  expect_equal(lpad(c("foo", "foobar")), c("   foo", "foobar"))
})

test_that("is_utf8_output", {

  local_mocked_bindings(
    l10n_info = function() list(MBCS = TRUE, `UTF-8` = TRUE, `Latin-1` = FALSE)
  )
  withr::with_options(
    list(cli.unicode = NULL),
    expect_true(is_utf8_output())
  )

  local_mocked_bindings(
    l10n_info = function() list(MBCS = FALSE, `UTF-8` = FALSE, `Latin-1` = TRUE)
  )
  withr::with_options(
    list(cli.unicode = NULL),
    expect_false(is_utf8_output())
  )
})

test_that("is_latex_output", {

  local_mocked_bindings(loadedNamespaces = function() "foobar")
  expect_false(is_latex_output())

  local_mocked_bindings(
    loadedNamespaces = function() "knitr",
    get = function(x, ...) {
      if (x == "is_latex_output") {
        function() TRUE
      } else {
        base::get(x, ...)
      }
    }
  )
  expect_true(is_latex_output())
})

test_that("dedent", {
  cases <- list(
    list("", 0, ""),
    list("", 1, ""),
    list("", 2, ""),
    list("x", 0, "x"),
    list("x", 1, "x"),
    list("x", 2, "x"),
    list("xx", 0, "xx"),
    list("xx", 1, "xx"),
    list("xx", 2, "xx"),
    list("foobar", 0, "foobar"),
    list("foobar", 1, "foobar"),
    list("foobar", 2, "foobar"),

    list(" ", 0, " "),
    list(" ", 1, ""),
    list(" ", 2, ""),
    list("  ", 0, "  "),
    list("  ", 1, " "),
    list("  ", 2, ""),
    list(" x", 0, " x"),
    list(" x", 1, "x"),
    list(" x", 2, "x"),
    list("  x", 0, "  x"),
    list("  x", 1, " x"),
    list("  x", 2, "x"),

    list(" x  y", 3, "x  y"),
    list(" x  y", 4, "x  y"),
    list(" x  y", 5, "x  y"),
    list(" x  ", 3, "x  "),
    list(" x  ", 4, "x  "),
    list(" x  ", 5, "x  ")
  )

  for (c in cases) expect_identical(dedent(c[[1]], c[[2]]), ansi_string(c[[3]]))
})

test_that("tail_na", {
  cases <- list(
    list(1:4, 4L),
    list(1, 1),
    list(double(), NA_real_),
    list(character(), NA_character_)
  )

  for (i in seq_along(cases)) {
    c <- cases[[i]]
    expect_identical(tail_na(c[[1]]), c[[2]], info = i)
  }

  cases2 <- list(
    list(1:4, 2, 3:4),
    list(1, 2, c(NA_real_, 1)),
    list(double(), 2, c(NA_real_, NA_real_)),
    list(character(), 2, c(NA_character_, NA_character_))
  )

  for (i in seq_along(cases2)) {
    c <- cases2[[i]]
    expect_identical(tail_na(c[[1]], c[[2]]), c[[3]], info = i)
  }
})

test_that("get_ppid", {
  expect_equal(
    ps::ps_ppid(),
    get_ppid()
  )
})

test_that("na.omit", {
  expect_snapshot({
    na.omit(character())
    na.omit(integer())
    na.omit(1:5)
    na.omit(c(1,NA,2,NA))
    na.omit(c(NA_integer_, NA_integer_))
    na.omit(list(1,2,3))
  })
})

test_that("get_rstudio_theme", {
  local_mocked_bindings(
    getThemeInfo = function() function(...) warning("just a word"), .package = "rstudioapi"
  )
  expect_silent(get_rstudio_theme())
})

test_that("try_silently", {
  expect_silent(
    try_silently(1:10)
  )
  expect_s3_class(
    try_silently(stop("not this")),
    "error"
  )
})

test_that("str_trim", {
  expect_snapshot({
    str_trim("foo")
    str_trim(character())
    str_trim("   foo")
    str_trim("foo  ")
    str_trim("   foo  ")
    str_trim(c(NA_character_, " foo  ", NA_character_, "  bar  "))
  })
})

test_that("leading_space", {
  testthat::local_reproducible_output(unicode = TRUE)
  expect_snapshot({
    paste0("-", leading_space("foo"), "-")
    paste0("-", leading_space("  foo"), "-")
    paste0("-", leading_space("  foo  "), "-")
    paste0("-", leading_space(" \t foo  "), "-")
    paste0("-", leading_space("\u00a0foo  "), "-")
    paste0("-", leading_space(" \u00a0 foo  "), "-")
  })
})

test_that("trailing_space", {
  testthat::local_reproducible_output(unicode = TRUE)
  expect_snapshot({
    paste0("-", trailing_space("foo"), "-")
    paste0("-", trailing_space("foo  "), "-")
    paste0("-", trailing_space("  foo  "), "-")
    paste0("-", trailing_space("  foo \t "), "-")
    paste0("-", trailing_space("  foo\u00a0"), "-")
    paste0("-", trailing_space(" \u00a0 foo \u00a0 "), "-")
  })
})

test_that("abbrev", {
  expect_snapshot({
    abbrev("123456789012345")
    abbrev("12345678901")
    abbrev("1234567890")
    abbrev("123456789")
    abbrev("12345")
    abbrev("1")
    abbrev("")
    abbrev("\033[31m1234567890\033[39m")
    abbrev(c("\033[31m1234567890\033[39m", "", "1234567890123"), 5)
    abbrev(rep("\033[31m1234567890\033[39m", 5), 5)
  })
})

test_style <- function() {
  list(
    ".testcli h1" = list(
      "font-weight" = "bold",
      "font-style" = "italic",
      "margin-top" = 1,
      "margin-bottom" = 1),
    ".testcli h2" = list(
      "font-weight" = "bold",
      "margin-top" = 1,
      "margin-bottom" = 1),
    ".testcli h3" = list(
      "text-decoration" = "underline",
      "margin-top" = 1)
  )
}

capture_messages <- function(expr) {
  msgs <- character()
  i <- 0
  suppressMessages(withCallingHandlers(
    expr,
    message = function(e) msgs[[i <<- i + 1]] <<- conditionMessage(e)))
  paste0(msgs, collapse = "")
}

capt <- function(expr, print_it = TRUE) {
  pr <- if (print_it) print else identity
  paste(capture.output(pr(expr)), collapse = "\n")
}

capt0 <- function(expr, strip_style = FALSE) {
  out <- capture_messages(expr)
  if  (strip_style) ansi_strip(out) else out
}

call_if_fun <- function(x) {
  if (is.function(x)) x() else x
}

clii__xtext <- function(app, text, .list, indent, padding, ln = TRUE, wrap = TRUE) {
  style <- app$get_current_style()
  text <- app$inline(text, .list = .list)
  exdent <- style$`text-exdent` %||% 0L

  esc <- function(x) gsub(" ", "\u00a0", x, fixed = TRUE)

  bef <- call_if_fun(style$before)
  if (!is.null(bef)) text[1] <- paste0(esc(bef), text[1])
  aft <- call_if_fun(style$after)
  if (!is.null(aft)) text[length(text)] <- paste0(text[length(text)], esc(aft))

  if (!is.null(style$fmt)) text <- style$fmt(text)

  if (wrap) {
    text <- ansi_strwrap(
      text,
      exdent = exdent,
      width = app$get_width(extra = padding)
    )
  } else {
    text <- ansi_simplify(text)
  }

  app$cat_ln(text, indent = indent, padding)
  invisible(app)
}

clii__get_width <- function(app, extra) {
  style <- app$get_current_style()
  left <- style$`margin-left` %||% 0 + style$`padding-left` %||% 0
  right <- style$`margin-right` %||% 0 + style$`padding-right` %||% 0
  console_width() - left - right - extra
}

clii__cat <- function(app, lines) {
  clii__message(lines, appendLF = FALSE, output = app$output, signal = app$signal)
}

clii__cat_ln <- function(app, lines, indent, padding) {
  if (!is.null(item <- app$state$delayed_item)) {
    app$state$delayed_item <- NULL
    return(app$item_text(item$type, NULL, item$cnt_id, .list = lines))
  }

  style <- app$get_current_style()

  ## left margin
  left <- padding + (style$`margin-left` %||% 0) + (style$`padding-left` %||% 0)
  if (length(lines) && left) lines <- paste0(strrep(" ", left), lines)

  ## indent or negative indent
  if (length(lines)) {
    if (indent < 0) {
      lines[1] <- dedent(lines[1], - indent)
    } else if (indent > 0) {
      lines[1] <- paste0(strrep(" ", indent), lines[1])
    }
  }

  ## zero out margin
  app$margin <- 0

  signal <- !identical(app$signal, FALSE)
  if (signal && length(app$status_bar)) clii__clear_status_bar(app)
  app$cat(paste0(paste0(lines, "\n"), collapse = ""))
  if (signal && length(app$status_bar)) {
    app$cat(paste0(app$status_bar[[1]]$content, "\r"))
  }
}

clii__vspace <- function(app, n) {
  if (app$margin < n) {
    sp <- strrep("\n", n - app$margin)
    signal <- !identical(app$signal, FALSE)
    if (signal && length(app$status_bar)) clii__clear_status_bar(app)
    clii__message(sp, appendLF = FALSE, output = app$output, signal = app$signal)
    app$margin <- n
    if (signal && length(app$status_bar)) {
      app$cat(paste0(app$status_bar[[1]]$content, "\r"))
    }
  }
}

get_real_output <- function(output) {
  if (! inherits(output, "connection")) {
    output <- switch(
      output,
      "auto" = cli_output_connection(),
      "message" = ,
      "stderr" = stderr(),
      "stdout" = stdout()
    )
  }
  output
}

clii__message <- function(..., domain = NA, appendLF = TRUE,
                          output = stderr(), signal = TRUE) {

  msg <- .makeMessage(..., domain = domain, appendLF = appendLF)
  output <- get_real_output(output)

  # to avoid non-breaking spaces in files, if output is redirected
  msg <- gsub("\u00a0", " ", msg, fixed = TRUE)

  if (identical(signal, FALSE)) {
    safe_cat0(msg, file = output)

  } else {
    withRestarts(muffleMessage = function() NULL, {
      cond <- simpleMessage(msg)
      class(cond) <- c("cliMessage", class(cond))
      signalCondition(cond)
      safe_cat0(msg, file = output)
    })
  }
}

safe_cat0 <- function(x, file) {
  if (inherits(file, "rawConnection")) {
    x <- enc2utf8(x)
    writeBin(charToRaw(x), file)
  } else {
    cat(x, file = file, sep = "")
  }
}

#' CLI pluralization
#'
#' @name pluralization
#' @family pluralization
#' @includeRmd man/chunks/pluralization.Rmd
NULL

make_quantity <- function(object) {
  val <- if (is.numeric(object)) {
    stopifnot(length(object) == 1)
    as.integer(object)
  } else {
    length(object)
  }
}

#' Pluralization helper functions
#'
#' @rdname pluralization-helpers
#' @param expr For `no()` it is an expression that is printed as "no" in
#'   cli expressions, it is interpreted as a zero quantity. For `qty()`
#'   an expression that sets the pluralization quantity without printing
#'   anything. See examples below.
#'
#' @export
#' @family pluralization

no <- function(expr) {
  stopifnot(is.numeric(expr), length(expr) == 1, !is.na(expr))
  structure(
    expr,
    class = "cli_no"
  )
}

#' @export

as.character.cli_no <- function(x, ...) {
  if (make_quantity(x) == 0) "no" else as.character(unclass(x))
}

#' @rdname pluralization-helpers
#' @export

qty <- function(expr) {
  structure(
    make_quantity(expr),
    class = "cli_noprint"
  )
}

#' @export

as.character.cli_noprint <- function(x, ...) {
  ""
}

parse_plural <- function(code, values) {
  # If we have the quantity already, then process it now.
  # Otherwise we put in a marker for it, and request post-processing.
  qty <- make_quantity(values$qty)
  if (!is.na(qty)) {
    process_plural(qty, code)
  } else {
    values$postprocess <- TRUE
    id <- random_id()
    values$pmarkers[[id]] <- code
    id
  }
}

process_plural <- function(qty, code) {
  parts <- strsplit(str_tail(code), "/", fixed = TRUE)[[1]]
  if (last_character(code) == "/") parts <- c(parts, "")
  if (length(parts) == 1) {
    if (qty != 1) parts[1] else ""
  } else if (length(parts) == 2) {
    if (qty == 1) parts[1] else parts[2]
  } else if (length(parts) == 3) {
    if (qty == 0) {
      parts[1]
    } else if (qty == 1) {
      parts[2]
    } else {
      parts[3]
    }
  } else {
    stop("Invalid pluralization directive: `", code, "`")
  }
}

post_process_plurals <- function(str, values) {
  if (!values$postprocess) return(str)
  if (values$num_subst == 0) {
    stop("Cannot pluralize without a quantity")
  }
  if (values$num_subst != 1) {
    stop("Multiple quantities for pluralization")
  }

  qty <- make_quantity(values$qty)
  for (i in seq_along(values$pmarkers)) {
    mark <- values$pmarkers[i]
    str <- sub(names(mark), process_plural(qty, mark[[1]]), str)
  }

  str
}

#' String templating with pluralization
#'
#' `pluralize()` is similar to [glue::glue()], with two differences:
#' * It supports cli's [pluralization] syntax, using `{?}` markers.
#' * It collapses substituted vectors into a comma separated string.
#'
#' See [pluralization] and some examples below.
#'
#' @param ...,.envir,.transformer All arguments are passed to [glue::glue()].
#'
#' @export
#' @family pluralization
#' @examples
#' # Regular plurals
#' nfile <- 0; pluralize("Found {nfile} file{?s}.")
#' nfile <- 1; pluralize("Found {nfile} file{?s}.")
#' nfile <- 2; pluralize("Found {nfile} file{?s}.")
#'
#' # Irregular plurals
#' ndir <- 1; pluralize("Found {ndir} director{?y/ies}.")
#' ndir <- 5; pluralize("Found {ndir} director{?y/ies}.")
#'
#' # Use 'no' instead of zero
#' nfile <- 0; pluralize("Found {no(nfile)} file{?s}.")
#' nfile <- 1; pluralize("Found {no(nfile)} file{?s}.")
#' nfile <- 2; pluralize("Found {no(nfile)} file{?s}.")
#'
#' # Use the length of character vectors
#' pkgs <- "pkg1"
#' pluralize("Will remove the {pkgs} package{?s}.")
#' pkgs <- c("pkg1", "pkg2", "pkg3")
#' pluralize("Will remove the {pkgs} package{?s}.")
#'
#' pkgs <- character()
#' pluralize("Will remove {?no/the/the} {pkgs} package{?s}.")
#' pkgs <- c("pkg1", "pkg2", "pkg3")
#' pluralize("Will remove {?no/the/the} {pkgs} package{?s}.")
#'
#' # Multiple quantities
#' nfiles <- 3; ndirs <- 1
#' pluralize("Found {nfiles} file{?s} and {ndirs} director{?y/ies}")
#'
#' # Explicit quantities
#' nupd <- 3; ntotal <- 10
#' cli_text("{nupd}/{ntotal} {qty(nupd)} file{?s} {?needs/need} updates")

pluralize <- function(..., .envir = parent.frame(),
                      .transformer = glue::identity_transformer) {

  values <- new.env(parent = emptyenv())
  values$empty <- random_id()
  values$qty <- values$empty
  values$num_subst <- 0L
  values$postprocess <- FALSE
  values$pmarkers <- list()

  tf <- function(text, envir) {
    if (substr(text, 1, 1) == "?") {
      if (identical(values$qty, values$empty)) {
        values$postprocess <- TRUE
        id <- random_id()
        values$pmarkers[[id]] <- text
        return(id)
      } else {
        return(process_plural(make_quantity(values$qty), text))
      }

    } else {
      values$num_subst <- values$num_subst + 1
      qty <- .transformer(text, envir)
      values$qty <- qty
      return(inline_collapse(qty))
    }
  }

  raw <- glue::glue(..., .envir = .envir, .transformer = tf)
  post_process_plurals(raw, values)
}

start_app()
on.exit(stop_app(), add = TRUE)

test_that("simplest", {
  expect_snapshot({
    for (n in 0:2) cli_text("{n} package{?s}")
    for (n in 0:2) print(pluralize("{n} package{?s}"))
  })
})

test_that("irregular", {
  expect_snapshot({
    for (n in 0:2) cli_text("{n} dictionar{?y/ies}")
    for (n in 0:2) print(pluralize("{n} dictionar{?y/ies}"))
  })
})

test_that("multiple substitutions", {
  expect_snapshot({
    for (n in 0:2) cli_text("{n} package{?s} {?is/are} ...")
    for (n in 0:2) print(pluralize("{n} package{?s} {?is/are} ..."))
  })
})

test_that("multiple quantities", {
  expect_snapshot({
    for (m in 0:2) for (n in 0:2) cli_text("{m} package{?s} and {n} folder{?s}")
    for (m in 0:2) for (n in 0:2) print(pluralize("{m} package{?s} and {n} folder{?s}"))
  })
})

test_that("no()", {
  expect_snapshot({
    for (n in 0:2) cli_text("{no(n)} package{?s}")
    for (n in 0:2) print(pluralize("{no(n)} package{?s}"))
  })
})

test_that("set qty() explicitly", {
  expect_snapshot({
    for (n in 0:2) cli_text("{qty(n)}There {?is/are} {n} package{?s}")
    for (n in 0:2) print(pluralize("{qty(n)}There {?is/are} {n} package{?s}"))
  })
})

test_that("collapsing vectors", {
  expect_snapshot({
    pkgs <- function(n) glue::glue("pkg{seq_len(n)}")
    for (n in 1:3) cli_text("The {pkgs(n)} package{?s}")
    for (n in 1:3) print(pluralize("The {pkgs(n)} package{?s}"))
  })
})

test_that("pluralization and style", {
  expect_snapshot({
    special_style <- list(span.foo = list(before = "<", after = ">"))
    cli_div(theme = special_style)
    for (n in 0:2) cli_text("{n} {.foo package{?s}}")
  })

  expect_snapshot({
    pkgs <- function(n) glue::glue("pkg{seq_len(n)}")
    for (n in 1:3) cli_text("The {.foo {pkgs(n)}} package{?s}")
  })
})

test_that("post-processing", {
  expect_snapshot({
    for (n in 0:2) cli_text("Package{?s}: {n}")
  })

  expect_snapshot({
    pkgs <- function(n) glue::glue("pkg{seq_len(n)}")
    for (n in 1:2) cli_text("Package{?s}: {pkgs(n)}")
    for (n in 1:2) print(pluralize("Package{?s}: {pkgs(n)}"))
  })
})

test_that("post-processing errors", {
  expect_error(
    cli_text("package{?s}"),
    "Cannot pluralize without a quantity"
  )
  expect_error(
    pluralize("package{?s}"),
    "Cannot pluralize without a quantity"
  )
  expect_error(
    cli_text("package{?s} {5} {10}"),
    "Multiple quantities for pluralization"
  )
  expect_error(
    pluralize("package{?s} {5} {10}"),
    "Multiple quantities for pluralization"
  )
})

test_that("issue 158", {
  expect_snapshot({
    print(pluralize("{0} word{?A/B/}"))
    print(pluralize("{1} word{?A/B/}"))
    print(pluralize("{9} word{?A/B/}"))
  })
})

test_that("Edge cases for pluralize() (#701)", {
  expect_snapshot({
    # Should not be pluralized
    print(pluralize("{NA} file{?s} expected"))
    print(pluralize("{NA_character_} file{?s} expected"))

    # Should be pluralized
    print(pluralize("{NA_real_} file{?s} expected"))
    print(pluralize("{NA_integer_} file{?s} expected"))
    print(pluralize("{NaN} file{?s} expected"))
    print(pluralize("{Inf} file{?s} expected"))
    print(pluralize("{-Inf} file{?s} expected"))
  })

  expect_snapshot({
    # Should not be pluralized
    print(pluralize("Found {NA} director{?y/ies}."))
    print(pluralize("Found {NA_character_} director{?y/ies}."))

    # Should be pluralized
    print(pluralize("Found {NA_real_} director{?y/ies}."))
    print(pluralize("Found {NA_integer_} director{?y/ies}."))
    print(pluralize("Found {NaN} director{?y/ies}."))
    print(pluralize("Found {Inf} director{?y/ies}."))
    print(pluralize("Found {-Inf} director{?y/ies}."))
  })

  expect_snapshot({
    # Should not be pluralized
    print(pluralize("Will remove {?no/the/the} {NA} package{?s}."))
    print(pluralize("Will remove {?no/the/the} {NA_character_} package{?s}."))

    # Should be pluralized
    print(pluralize("Will remove {?no/the/the} {NA_real_} package{?s}."))
    print(pluralize("Will remove {?no/the/the} {NA_integer_} package{?s}."))
    print(pluralize("Will remove {?no/the/the} {NaN} package{?s}."))
    print(pluralize("Will remove {?no/the/the} {Inf} package{?s}."))
    print(pluralize("Will remove {?no/the/the} {-Inf} package{?s}."))
  })
})

start_app()
on.exit(stop_app(), add = TRUE)

test_that("simplest", {
  expect_snapshot({
    for (n in 0:2) cli_text("{n} package{?s}")
    for (n in 0:2) print(pluralize("{n} package{?s}"))
  })
})

test_that("irregular", {
  expect_snapshot({
    for (n in 0:2) cli_text("{n} dictionar{?y/ies}")
    for (n in 0:2) print(pluralize("{n} dictionar{?y/ies}"))
  })
})

test_that("multiple substitutions", {
  expect_snapshot({
    for (n in 0:2) cli_text("{n} package{?s} {?is/are} ...")
    for (n in 0:2) print(pluralize("{n} package{?s} {?is/are} ..."))
  })
})

test_that("multiple quantities", {
  expect_snapshot({
    for (m in 0:2) for (n in 0:2) cli_text("{m} package{?s} and {n} folder{?s}")
    for (m in 0:2) for (n in 0:2) print(pluralize("{m} package{?s} and {n} folder{?s}"))
  })
})

test_that("no()", {
  expect_snapshot({
    for (n in 0:2) cli_text("{no(n)} package{?s}")
    for (n in 0:2) print(pluralize("{no(n)} package{?s}"))
  })
})

test_that("set qty() explicitly", {
  expect_snapshot({
    for (n in 0:2) cli_text("{qty(n)}There {?is/are} {n} package{?s}")
    for (n in 0:2) print(pluralize("{qty(n)}There {?is/are} {n} package{?s}"))
  })
})

test_that("collapsing vectors", {
  expect_snapshot({
    pkgs <- function(n) glue::glue("pkg{seq_len(n)}")
    for (n in 1:3) cli_text("The {pkgs(n)} package{?s}")
    for (n in 1:3) print(pluralize("The {pkgs(n)} package{?s}"))
  })
})

test_that("pluralization and style", {
  expect_snapshot({
    special_style <- list(span.foo = list(before = "<", after = ">"))
    cli_div(theme = special_style)
    for (n in 0:2) cli_text("{n} {.foo package{?s}}")
  })

  expect_snapshot({
    pkgs <- function(n) glue::glue("pkg{seq_len(n)}")
    for (n in 1:3) cli_text("The {.foo {pkgs(n)}} package{?s}")
  })
})

test_that("post-processing", {
  expect_snapshot({
    for (n in 0:2) cli_text("Package{?s}: {n}")
  })

  expect_snapshot({
    pkgs <- function(n) glue::glue("pkg{seq_len(n)}")
    for (n in 1:2) cli_text("Package{?s}: {pkgs(n)}")
    for (n in 1:2) print(pluralize("Package{?s}: {pkgs(n)}"))
  })
})

test_that("post-processing errors", {
  expect_error(
    cli_text("package{?s}"),
    "Cannot pluralize without a quantity"
  )
  expect_error(
    pluralize("package{?s}"),
    "Cannot pluralize without a quantity"
  )
  expect_error(
    cli_text("package{?s} {5} {10}"),
    "Multiple quantities for pluralization"
  )
  expect_error(
    pluralize("package{?s} {5} {10}"),
    "Multiple quantities for pluralization"
  )
})

test_that("issue 158", {
  expect_snapshot({
    print(pluralize("{0} word{?A/B/}"))
    print(pluralize("{1} word{?A/B/}"))
    print(pluralize("{9} word{?A/B/}"))
  })
})

progress_c_update <- function(pb, auto_done = TRUE) {
  cli_tick_reset()

  caller <- pb$caller %||% sys.frame(sys.nframe() - 1L)

  pb$tick <- pb$tick + 1L

  if (is.null(pb$format)) {
    pb$format <- pb__default_format(pb$type, pb$total)
  }

  if (pb$auto_terminate && auto_done && !is.na(pb$total) &&
      pb$current == pb$total) {
    progress_c_done(pb, caller = caller)
    return(NULL)
  }

  opt <- options(cli__pb = pb)
  on.exit(options(opt), add = TRUE)

  handlers <- cli_progress_select_handlers(pb, caller)
  if (is.null(pb$added)) {
    pb$added <- TRUE
    for (h in handlers) {
      if ("add" %in% names(h)) h$add(pb, .envir = caller)
    }
  }

  for (h in handlers) {
    if ("set" %in% names(h)) h$set(pb, .envir = caller)
  }

  NULL
}

progress_c_done <- function(pb, caller = NULL) {
  if (isTRUE(pb$done)) return()

  caller <- caller %||% pb$caller %||% sys.frame(sys.nframe() - 1L)

  opt <- options(cli__pb = pb)
  on.exit(options(opt), add = TRUE)

  handlers <- cli_progress_select_handlers()
  for (h in handlers) {
    if ("complete" %in% names(h)) {
      h$complete(pb, .envir = caller, result = "done")
    }
  }

  if (!is.null(pb$id)) clienv$progress[[pb$id]] <- NULL
  if (!is.null(pb$envkey)) clienv$progress_ids[[pb$envkey]] <- NULL

  pb$done <- TRUE

  NULL
}

#' Whether cli is emitting UTF-8 characters
#'
#' UTF-8 cli characters can be turned on by setting the `cli.unicode`
#' option to `TRUE`. They can be turned off by setting if to `FALSE`.
#' If this option is not set, then [base::l10n_info()] is used to detect
#' UTF-8 support.
#'
#' @return Flag, whether cli uses UTF-8 characters.
#'
#' @export

is_utf8_output <- function() {
  opt <- getOption("cli.unicode", NULL)
  if (! is.null(opt)) {
    isTRUE(opt)
  } else {
    l10n_info()$`UTF-8` && !is_latex_output()
  }
}

#' Count the number of characters in a character vector
#'
#' By default it counts Unicode grapheme clusters, instead of code points.
#'
#' @param x Character vector, it is converted to UTF-8.
#' @param type Whether to count graphemes (characters), code points,
#'   bytes, or calculate the display width of the string.
#' @return Numeric vector, the length of the strings in the character
#'   vector.
#'
#' @family UTF-8 string manipulation
#' @export
#' @examples
#' # Grapheme example, emoji with combining characters. This is a single
#' # grapheme, consisting of five Unicode code points:
#' # * `\U0001f477` is the construction worker emoji
#' # * `\U0001f3fb` is emoji modifier that changes the skin color
#' # * `\u200d` is the zero width joiner
#' # * `\u2640` is the female sign
#' # * `\ufe0f` is variation selector 16, requesting an emoji style glyph
#' emo <- "\U0001f477\U0001f3fb\u200d\u2640\ufe0f"
#' cat(emo)
#'
#' utf8_nchar(emo, "chars") # = graphemes
#' utf8_nchar(emo, "bytes")
#' utf8_nchar(emo, "width")
#' utf8_nchar(emo, "codepoints")
#'
#' # For comparision, the output for width depends on the R version used:
#' nchar(emo, "chars")
#' nchar(emo, "bytes")
#' nchar(emo, "width")

utf8_nchar <- function(x, type = c("chars", "bytes", "width", "graphemes",
                                   "codepoints")) {

  type <- match.arg(type)
  if (type == "chars") type <- "graphemes"

  x <- enc2utf8(x)

  if (type == "width") {
    .Call(clic_utf8_display_width, x)

  } else if (type == "graphemes") {
    .Call(clic_utf8_nchar_graphemes, x)

  } else if (type == "codepoints") {
    base::nchar(x, allowNA = FALSE, keepNA = TRUE, type = "chars")

  } else { # bytes
    base::nchar(x, allowNA = FALSE, keepNA = TRUE, type = "bytes")
  }
}

#' Substring of an UTF-8 string
#'
#' This function uses grapheme clusters instead of Unicode code points in
#' UTF-8 strings.
#'
#' @param x Character vector.
#' @param start Starting index or indices, recycled to match the length
#'   of `x`.
#' @param stop Ending index or indices, recycled to match the length of
#'   `x`.
#' @return Character vector of the same length as `x`, containing
#'   the requested substrings.
#'
#' @family UTF-8 string manipulation
#' @export
#' @examples
#' # Five grapheme clusters, select the middle three
#' str <- paste0(
#'   "\U0001f477\U0001f3ff\u200d\u2640\ufe0f",
#'   "\U0001f477\U0001f3ff",
#'   "\U0001f477\u200d\u2640\ufe0f",
#'   "\U0001f477\U0001f3fb",
#'   "\U0001f477\U0001f3ff")
#' cat(str)
#' str24 <- utf8_substr(str, 2, 4)
#' cat(str24)

utf8_substr <- function(x, start, stop) {
  if (!is.character(x)) x <- as.character(x)
  if (!is.numeric(start) || !is.numeric(stop)) {
    throw(cli_error(
      "{.arg start} and {.arg stop} must be numeric vectors",
      "i" = if (!is.numeric(start)) "{.arg start} is {.typeof {start}}",
      "i" = if (!is.numeric(stop))  "{.arg stop}  is {.typeof {stop}}"
    ))
  }
  start2 <- suppressWarnings(as.integer(start))
  stop2 <- suppressWarnings(as.integer(stop))
  if (!length(start2) || !length(stop2)) {
    throw(cli_error(
      "{.arg start} and {.arg stop} must have at least length 1",
      "i" = if (!length(start2)) "{.arg start} has length 0",
      "i" = if (!length(stop2))  "{.arg stop}  has length 0"
    ))
  }
  x <- enc2utf8(x)

  # TODO: better recycling
  start2 <- rep_len(start2, length(x))
  stop2 <- rep_len(stop2, length(x))
  .Call(clic_utf8_substr, x, start2, stop2)
}

#' Break an UTF-8 character vector into grapheme clusters
#'
#' @param x Character vector.
#' @return List of characters vectors, the grapheme clusters of the input
#'   string.
#'
#' @family UTF-8 string manipulation
#' @export
#' @examples
#' # Five grapheme clusters
#' str <- paste0(
#'   "\U0001f477\U0001f3ff\u200d\u2640\ufe0f",
#'   "\U0001f477\U0001f3ff",
#'   "\U0001f477\u200d\u2640\ufe0f",
#'   "\U0001f477\U0001f3fb",
#'   "\U0001f477\U0001f3ff")
#' cat(str, "\n")
#' chrs <- utf8_graphemes(str)

utf8_graphemes <- function(x) {
  if (!is.character(x)) x <- as.character(x)
  x <- enc2utf8(x)
  .Call(clic_utf8_graphemes, x)
}

operator_tokens <- function() {
  c(
    "'-'", "'+'", "'!'", "'~'", "'?'", "':'", "'*'", "'/'", "'^'",
    "SPECIAL", "LT", "GT", "EQ", "GE", "LE", "AND", "AND2", "OR", "OR2",
    "LEFT_ASSIGN", "RIGHT_ASSIGN", "'$'", "'@'", "EQ_ASSIGN", "PIPE"
  )
}

reserved_words <- function() {
  c("FUNCTION", "'\\\\'", "IF", "ELSE",
    "REPEAT", "WHILE", "FOR", "IN", "NEXT", "BREAK")
}


#' Syntax highlight R code
#'
#' @details
#' See [code_theme_list()] for the default syntax highlighting theme and
#' how to change it.
#'
#' If `code` does not parse, then it is returned unchanged and a
#' `cli_parse_failure` condition is thrown. Note that this is not an error,
#' and the condition is ignored, unless explicitly caught.
#'
#' @param code Character vector, each element is one line of code.
#' @param code_theme Theme see [code_theme_list()].
#' @param envir Environment to look up function calls for hyperlinks.
#'   If `NULL`, then the global search path is used.
#' @return Character vector, the highlighted code.
#'
#' @family syntax highlighting
#' @importFrom utils getSrcref getParseData
#' @export
#' @examples
#' code_highlight(deparse(ls))
#' cat(code_highlight(deparse(ls)), sep = "\n")

code_highlight <- function(code, code_theme = NULL, envir = NULL) {

  code_theme <- code_theme %||% code_theme_default()

  parsed <- tryCatch(
    parse(text = code, keep.source = TRUE),
    error = function(e) e
  )

  if (inherits(parsed, "error")) {
    cnd <- structure(
      list(message = conditionMessage(parsed), code = code),
      class = c("cli_parse_failure", "condition")
    )
    signalCondition(cnd)
    return(code)
  }

  theme <- code_theme_make(code_theme)
  data <- getParseData(parsed, includeText = NA)

  hitext <- data$text

  cnv <- function(x) do.call(combine_ansi_styles, as.list(x))

  brackettheme <- lapply(theme$bracket, cnv)
  theme <- theme[names(theme) != "bracket"]
  theme <- structure(lapply(theme, cnv), names = names(theme))

  ## Reserved words if else repeat while function for in next break
  if (!is.null(theme$reserved)) {
    reserved <- data$token %in% reserved_words()
    hitext[reserved] <- theme$reserved(data$text[reserved])
  }

  ## Numeric constants, including NAs, NaN and Inf
  if (!is.null(theme$number)) {
    num_const <- data$token == "NUM_CONST"
    hitext[num_const] <- theme$number(data$text[num_const])
  }

  ## NULL
  if (!is.null(theme$null)) {
    null <- data$token == "NULL_CONST"
    hitext[null] <- theme$null(data$text[null])
  }

  ## Operators
  if (!is.null(theme$operator)) {
    operator <- data$token %in% operator_tokens()
    hitext[operator] <- theme$operator(data$text[operator])
  }

  ## Function calls
  fun_call <- data$token == "SYMBOL_FUNCTION_CALL"
  if (ansi_hyperlink_types()$help) {
    hitext[fun_call] <- pretty_fun_link(data, fun_call, envir)
  }
  if (!is.null(theme$call)) {
    hitext[fun_call] <- theme$call(hitext[fun_call])
  }

  ## Strings
  if (!is.null(theme$string)) {
    string <- data$token == "STR_CONST"
    reserved <- theme$reserved %||% function(x) x
    raw <- substr(data$text[string], 1, 1) == "r"
    hitext[string][raw] <- paste0(
      rep(reserved("r"), sum(raw)),
      theme$string(substr(data$text[string][raw], 2, nchar(data$text[string][raw])))
    )
    hitext[string][!raw] <- theme$string(data$text[string][!raw])
  }

  ## Comments
  if (!is.null(theme$comment)) {
    comment <- data$token == "COMMENT"
    hitext[comment] <- theme$comment(data$text[comment])
  }

  ## Brackets
  if (length(brackettheme)) {
    bracket <- data$token %in% bracket_tokens()
    hitext[bracket] <- color_brackets(data$text[bracket], brackettheme)
  }

  do_subst(code, data, hitext)
}

do_subst <- function(code, pdata, hitext) {

  pdata$hitext <- hitext

  ## Need to do this line by line. TODO: multiline stuff might be broken
  vapply(seq_along(code), FUN.VALUE = character(1), function(no) {
    my <- pdata[pdata$line1 == no & pdata$line2 == no,, drop = FALSE]
    replace_in_place(code[no], my$col1, my$col2, my$hitext)
  })
}

open_brackets <- function() {
  c("(", "{", "[")
}

close_brackets <- function(){
  c(")", "}", "]")
}

bracket_tokens <- function() {
  s <- c(open_brackets(), close_brackets())
  c(paste0("'", s, "'"), "LBB")
}

apply_color <- function(x, lvl, l){
  k <- (lvl - 1) %% length(l) + 1
  l[[k]](x)
}

#' Colored brackets
#'
#' Add color to brackets. Brackets will be coloured consecutively with the
#' colors provided in \code{color_seq} by scope.
#'
#' @param x a character vector of brackets consisting of a valid sequence of any
#'   of the following: \code{'[[', '[', ']', '(', ')', '{', '}'}
#' @param color_seq a list of functions that take and return a character scalar. The
#' ordering defines the sequence of color functions to apply to a given scope level.
#' Color functions are recycled when the scope level exceeds the length of \code{color_seq}
#'
#' @details Meant for coloring brackets encountered within \code{highlight}.
#'   Note that occurrences of 'orphan' brackets are not taken into account
#'   mainly due to the fact that cases such as
#'
#'   \code{foo <- function(x){ `[[`(x, 1) }}
#'
#'   will either be converted to
#'
#'   \code{foo <- function(x){ x[[1]] }}
#'
#'   before the brackets are coloured if passed in as
#'   \code{highlight(deparse(foo))} or will be identified as a
#'   'SYMBOL_FUNCTION_CALL' token instead of 'LBB' if passed in as
#'
#'   \code{highlight("foo <- function(x){ `[[`(x, 1) }")}
#'
#'   Similarly, invalid code that would lead to orphaned brackets is not taken
#'   into account as this would be caught before hand in \code{highlight}.
#'
#' @noRd

color_brackets <- function(x, color_seq = list(col_yellow, col_blue, col_cyan)) {
  stopifnot(vapply(color_seq, is.function, logical(1)))
  open <- c(open_brackets(), "[[")
  o <- character()
  lvl <- 0
  i <- 1
  while (i <= length(x)) {

    if (x[i] %in% open) {
      o[length(o) + 1] <- x[i]
      lvl <- lvl + 1
      x[i] <- apply_color(x[i], lvl, color_seq)
      i <- i + 1
      next
    }

    j <- nchar(o[length(o)])
    x[i:(i + j - 1)] <-
      apply_color(x[i:(i + j - 1)], lvl, color_seq)
    i <- i + j
    lvl <- lvl - 1
    o <- o[-length(o)]
  }
  x
}

replace_in_place <- function(str, start, end, replacement) {

  stopifnot(
    length(str) == 1,
    length(start) == length(end),
    length(end) == length(replacement)
  )

  keep <- substring(str, c(1, end + 1), c(start - 1, nchar(str)))

  pieces <- character(length(replacement) * 2 + 1)

  even <- seq_along(replacement) * 2
  odd <- c(1, even + 1)
  pieces[even] <- replacement
  pieces[odd] <- keep

  paste0(pieces, collapse = "")
}

code_theme_default <- function() {
  opt <- code_theme_opt("cli.code_theme")
  if (!is.null(opt)) return(opt)

  rs <- rstudio_detect()
  if (rs$type %in% c("rstudio_console", "rstudio_console_starting")) {
    opt <- code_theme_opt("cli.code_theme_rstudio")
    if (!is.null(opt)) return(opt)
    if (requireNamespace("rstudioapi", quietly = TRUE)) {
      return(code_theme_default_rstudio())
    }
  }

  opt <- code_theme_opt("cli.code_theme_terminal")
  if (!is.null(opt)) return(opt)
  code_theme_default_term()
}

code_theme_opt <- function(option) {
  theme <- getOption(option)
  if (is.null(theme)) return(NULL)

  code_theme_make(theme)
}

code_theme_make <- function(theme) {
  if (is.list(theme)) return(theme)
  if (is_string(theme)) {
    if (theme %in% names(rstudio_themes)) return(rstudio_themes[[theme]])
    lcs <- gsub(" ", "_", tolower(names(rstudio_themes)), fixed = TRUE)
    if (theme %in% lcs) return(rstudio_themes[[ match(theme, lcs)[1] ]])
    warning("Unknown cli code theme: `", theme, "`.")
    return(NULL)
  }
  warning("Invalid cli code theme, see documentation")
  NULL
}

code_theme_default_rstudio <- function() {
  theme <- get_rstudio_theme()$editor
  if (! theme %in% names(rstudio_themes)) {
    if (!getOption("cli.ignore_unknown_rstudio_theme", FALSE)) {
      warning(
        "cli does not know this RStudio theme: '", theme, "'.",
        "\nSet `options(cli.ignore_unknown_rstudio_theme = TRUE)` ",
        "to suppress this warning"
      )
    }
    return(code_theme_default_term())
  }
  rstudio_themes[[theme]]
}

code_theme_default_term <- function() {
  "Solarized Dark"
}

#' Syntax highlighting themes
#'
#' @description
#' `code_theme_list()` lists the built-in code themes.
#'
#' # Code themes
#' A theme is a list of character vectors, except for `bracket`, see below.
#' Each character vector must contain RGB colors (e.g. `"#a9a9a9"`),
#' and cli styles, e.g. `"bold"`. Entries in the list:
#' * `reserved`: reserved words
#' * `number`: numeric literals
#' * `null`: the `NULL` constant
#' * `operator`: operators, including assignment
#' * `call`: function calls
#' * `string`: character literals
#' * `comment`: comments
#' * `bracket`: brackets: \code{(){}[]} This is a list of character vectors,
#'   to create "rainbow" brackets. It is recycled for deeply nested lists.
#'
#' # The default code theme
#'
#' In RStudio, it matches the current theme of the IDE.
#'
#' You can use three options to customize the code theme:
#' * If `cli.code_theme` is set, it is used.
#' * Otherwise if R is running in RStudio and `cli.code_theme_rstudio` is
#'   set, then it is used.
#' * Otherwise if T is not running in RStudio and `cli.code_theme_terminal`
#'   is set, then it is used.
#'
#' You can set these options to the name of a built-in theme, or to list
#' that specifies a custom theme. See [code_theme_list()] for the list
#' of the built-in themes.
#'
#' @return Character vector of the built-in code theme names.
#'
#' @family syntax highlighting
#' @export
#' @examples
#' code_theme_list()
#' code_highlight(deparse(get), code_theme = "Solarized Dark")

code_theme_list <- function() {
  names(rstudio_themes)
}

pretty_print_function <- function(x, useSource = TRUE, code_theme = NULL, ...) {
  if (num_ansi_colors() == 1L) return(base::print.function(x, useSource))

  srcref <- getSrcref(x)
  src <- if (useSource && ! is.null(srcref)) {
    as.character(srcref)
  } else {
    deparse(x)
  }

  err <- FALSE
  hisrc <- tryCatch(
    code_highlight(src, code_theme = code_theme, envir = environment(x)),
    error = function(e) err <<- TRUE)
  if (err) return(base::print.function(x, useSource))

  ## Environment of the function
  hisrc <- c(hisrc, utils::capture.output(print(environment(x))))

  cat(hisrc, sep = "\n")
  invisible(x)
}

#' Turn on pretty-printing functions at the R console
#'
#' Defines a print method for functions, in the current session, that supports
#' syntax highlighting.
#'
#' The new print method takes priority over the built-in one. Use
#' [base::suppressMessages()] to suppress the alert message.
#'
#' @export

pretty_print_code <- function() {
  registerS3method("print", "function", pretty_print_function, asNamespace("cli"))
  cli::cli_alert_success("Registered pretty printing function method")
}

pretty_fun_link <- function(data, fun_call, envir) {
  sprt <- ansi_hyperlink_types()$help
  wch <- which(fun_call)
  txt <- data$text[wch]
  if (! sprt || length(wch) == 0) return(txt)

  scheme <- if (identical(attr(sprt, "type"), "rstudio")) {
    "ide:help"
  } else {
    "x-r-help"
  }

  pkg <- vcapply(wch, function(idx) {
    prt <- data$parent[idx]
    sgs <- which(data$parent == prt)
    # not a pkg::fun call?
    if (length(sgs) != 3 || data$token[sgs[1]] != "SYMBOL_PACKAGE" ||
        data$token[sgs[2]] != "NS_GET") {
      # note: we do not process ::: which would be NS_GET_INT
      find_function_symbol(data$text[idx], envir %||% .GlobalEnv)
    } else {
      data$text[sgs[1]]
    }
  })

  wlnk <- which(!is.na(pkg))
  txt[wlnk] <- style_hyperlink(
    text = txt[wlnk],
    url = paste0(scheme, ":", pkg[wlnk], "::", txt[wlnk])
  )

  txt
}

find_function_symbol <- function(name, envir = .GlobalEnv) {
  empty <- emptyenv()
  while (!identical(envir, empty)) {
    if (exists(name, envir = envir, inherits = FALSE, mode = "function")) {
      env_name <- environmentName(envir)
      if (grepl("package:", env_name, fixed = TRUE)) {
        env_name <- sub("^package:", "", env_name)
      }
      if (grepl("imports:", env_name, fixed = TRUE)) {
        env_name <- environmentName(environment(get(name, envir)))
      }
      if (grepl("package:", env_name, fixed = TRUE)) {
        env_name <- sub("^package:", "", env_name)
      }
      if (env_name %in% c("", "R_GlobalEnv")) {
        env_name <- NA_character_
      }
      return(env_name)
    } else {
      envir <- parent.env(envir)
    }
  }
  NA_character_
}

col_seq <- list(function(x)
  paste0("1", x),
  function(x)
    paste0("2", x),
  function(x)
    paste0("3", x))

test_that("bracket highlighting", {
  # [](){}
  expect_equal(color_brackets(c("[", "]", "(", ")", "{", "}"), col_seq),
               c("1[", "1]", "1(", "1)", "1{", "1}"))
  
  # [({[({})]})]
  expect_equal(
    color_brackets(c(
      "[", "(", "{", "[", "(", "{", "}", ")", "]", "}", ")", "]"
    ),
    col_seq),
    c(
      "1[",
      "2(",
      "3{",
      "1[",
      "2(",
      "3{",
      "3}",
      "2)",
      "1]",
      "3}",
      "2)",
      "1]"
    )
  )
  
  # [[ [] ]][[ ()() ]]
  expect_equal(
    color_brackets(
      c("[[", "[", "]", "]", "]", "[[", "(", ")", "(", ")", "]", "]"),
      col_seq
    ),
    c(
      "1[[",
      "2[",
      "2]",
      "1]",
      "1]",
      "1[[",
      "2(",
      "2)",
      "2(",
      "2)",
      "1]",
      "1]"
    )
  )
})

test_that_cli(configs = c("plain", "ansi"), "reserved", {
  expect_snapshot({
    cat(code_highlight("function () { }", list(reserved = "bold")))
    cat(code_highlight("if (1) NULL else NULL", list(reserved = "bold")))
    cat(code_highlight("repeat {}", list(reserved = "bold")))
    cat(code_highlight("while (1) {}", list(reserved = "bold")))
    cat(code_highlight("for (i in x) next", list(reserved = "bold")))
    cat(code_highlight("for (i in x) break", list(reserved = "bold")))
  })
})

test_that_cli(configs = c("plain", "ansi"), "number", {
  expect_snapshot({
    cat(code_highlight("1 + 1.0 + -1 + 2L + Inf", list(number = "bold")))
    cat(code_highlight(
      "NA + NA_real_ + NA_integer_ + NA_character_",
      list(number = "bold")
    ))
    cat(code_highlight("TRUE + FALSE", list(number = "bold")))
  })
})

test_that_cli(configs = c("plain", "ansi"), "null", {
  expect_snapshot({
    cat(code_highlight("NULL", list(null = "bold")))
  })
})

test_that_cli(configs = c("plain", "ansi"), "operator", {

  expect_snapshot({
    cat(code_highlight("~ ! 1 - 2 + 3:4 * 5 / 6 ^ 7", list(operator = "bold")))
    cat(code_highlight(
      "? 1 %% 2 %+% 2 < 3 & 4 > 5 && 6 == 7 | 8 <= 9 || 10 >= 11",
      list(operator = "bold")
    ))
    cat(code_highlight(
      "a <- 10; 20 -> b; c = 30; a$b; a@b",
      list(operator = "bold")
    ))
  })
})

test_that_cli(configs = c("plain", "ansi"), "call", {
  expect_snapshot({
    cat(code_highlight("ls(2)", list(call = "bold")))
  })
})

test_that_cli(configs = c("plain", "ansi"), "string", {
  expect_snapshot({
    cat(code_highlight("'s' + \"s\"", list(string = "bold")))
  })
})

test_that_cli(configs = c("plain", "ansi"), "comment", {
  expect_snapshot({
    cat(code_highlight(c("# COM", " ls() ## ANOT"), list(comment = "bold")))
  })
})

test_that_cli(configs = c("plain", "ansi"), "bracket", {
  expect_snapshot({
    cat(code_highlight("foo <- function(x){x}", list(bracket = list("bold"))))
  })
})

test_that("replace_in_place", {
  expect_equal(
    replace_in_place("1234567890", c(2, 6), c(5, 8), c("foobar", "xxx")),
    "1foobarxxx90"
  )

  expect_equal(
    replace_in_place("1234567890", c(1, 5), c(6, 10), c("A", "B")),
    "AB"
  )
})

test_that("replace_in_place corner cases", {
  expect_equal(
    replace_in_place("foobar", integer(), integer(), character()),
    "foobar"
  )

  expect_equal(
    replace_in_place("12345", 1L, 5L, "no!"),
    "no!"
  )

  expect_equal(
    replace_in_place("12345", 1:5, 1:5, letters[1:5]),
    "abcde"
  )
})

test_that_cli(configs = c("plain", "ansi"), "parse errors", {
  expect_equal(
    code_highlight("not good!!!"), "not good!!!"
  )
  cnd <- NULL
  withCallingHandlers(
    expect_equal(
      code_highlight("not good!!!"), "not good!!!"
    ),
    cli_parse_failure = function(e) cnd <<- e
  )
  expect_s3_class(cnd, "cli_parse_failure")
  expect_s3_class(cnd, "condition")
  expect_equal(cnd$code, "not good!!!")
})

test_that("code themes", {
  withr::local_options(cli.code_theme = "solarized_dark")
  expect_equal(code_theme_default()$reserved, "#859900")

  withr::local_options(cli.code_theme = NULL)
  withr::local_envvar(RSTUDIO = "0")
  withr::local_options(cli.code_theme_terminal = "solarized_light")
  expect_equal(code_theme_default()$reserved, "#859900")

  local_mocked_bindings(
    rstudio_detect = function() list(type = "rstudio_console")
  )
  withr::local_options(cli.code_theme_rstudio = "Xcode")
  expect_equal(code_theme_default()$reserved, "#C800A4")

  withr::local_options(cli.code_theme_rstudio = NULL)
  skip_if_not_installed("rstudioapi")
  local_mocked_bindings(code_theme_default_rstudio = function() "foo")
  expect_equal(code_theme_default(), "foo")
})

test_that("code themes 2", {
  withr::local_options(cli.code_theme = "nope!!")
  expect_warning(code_theme_default(), "Unknown cli code theme")

  withr::local_options(cli.code_theme = 111)
  expect_warning(code_theme_default(), "Invalid cli code theme")
})

test_that("code_theme_default_rstudio", {
  local_mocked_bindings(
    get_rstudio_theme = function() list(editor = "Solarized Dark")
  )
  expect_equal(code_theme_default_rstudio()$reserved, "#859900")

  local_mocked_bindings(
    get_rstudio_theme = function() list(editor = "Not really")
  )
  expect_warning(
    cth <- code_theme_default_rstudio(),
    "cli does not know this RStudio theme"
  )
  expect_equal(cth, code_theme_default_term())
})

test_that("code_theme_list", {
  expect_snapshot(code_theme_list())
})

test_that_cli(configs = "ansi", "new language features, raw strings", {
  if (getRversion() < "4.0.1") { expect_true(TRUE); return() }
  expect_snapshot(
    cat(code_highlight(
      '"old" + r"("new""")"',
      list(string = "bold", reserved = "italic")
    ))
  )
})

test_that_cli(configs = "ansi", "new language features, pipe", {
  if (getRversion() < "4.1.0") { expect_true(TRUE); return() }
  expect_snapshot(
    cat(code_highlight('dir() |> toupper()', list(operator = "bold")))
  )
})

test_that_cli(configs = "ansi", "new language features, lambda functions", {
  if (getRversion() < "4.1.0") { expect_true(TRUE); return() }
  expect_snapshot(
    cat(code_highlight('\\(x) x * 2', list(reserved = "bold")))
  )
})

test_that("code_highlight() works on long strings and symbols", {
  expect_true(
    grepl(
      strrep("-", 1000),
      code_highlight(paste0("foo('", strrep("-", 1000), "')"))
    )
  )

  expect_true(
    grepl(
      strrep("-", 1000),
      code_highlight(paste0("foo(`", strrep("-", 1000), "`)"))
    )
  )
  expect_true(
    grepl(
      strrep("-", 1000),
      code_highlight(paste0("a$`", strrep("-", 1000), "`"))
    )
  )

  expect_true(
    grepl(
      strrep("-", 1000),
      code_highlight(paste0("`", strrep("-", 1000), "`$a"))
    )
  )
})

col_seq <- list(function(x)
  paste0("1", x),
  function(x)
    paste0("2", x),
  function(x)
    paste0("3", x))

test_that("bracket highlighting", {
  # [](){}
  expect_equal(color_brackets(c("[", "]", "(", ")", "{", "}"), col_seq),
               c("1[", "1]", "1(", "1)", "1{", "1}"))
  
  # [({[({})]})]
  expect_equal(
    color_brackets(c(
      "[", "(", "{", "[", "(", "{", "}", ")", "]", "}", ")", "]"
    ),
    col_seq),
    c(
      "1[",
      "2(",
      "3{",
      "1[",
      "2(",
      "3{",
      "3}",
      "2)",
      "1]",
      "3}",
      "2)",
      "1]"
    )
  )
  
  # [[ [] ]][[ ()() ]]
  expect_equal(
    color_brackets(
      c("[[", "[", "]", "]", "]", "[[", "(", ")", "(", ")", "]", "]"),
      col_seq
    ),
    c(
      "1[[",
      "2[",
      "2]",
      "1]",
      "1]",
      "1[[",
      "2(",
      "2)",
      "2(",
      "2)",
      "1]",
      "1]"
    )
  )
})

test_that_cli(configs = c("plain", "ansi"), "reserved", {
  expect_snapshot({
    cat(code_highlight("function () { }", list(reserved = "bold")))
    cat(code_highlight("if (1) NULL else NULL", list(reserved = "bold")))
    cat(code_highlight("repeat {}", list(reserved = "bold")))
    cat(code_highlight("while (1) {}", list(reserved = "bold")))
    cat(code_highlight("for (i in x) next", list(reserved = "bold")))
    cat(code_highlight("for (i in x) break", list(reserved = "bold")))
  })
})

test_that_cli(configs = c("plain", "ansi"), "number", {
  expect_snapshot({
    cat(code_highlight("1 + 1.0 + -1 + 2L + Inf", list(number = "bold")))
    cat(code_highlight(
      "NA + NA_real_ + NA_integer_ + NA_character_",
      list(number = "bold")
    ))
    cat(code_highlight("TRUE + FALSE", list(number = "bold")))
  })
})

test_that_cli(configs = c("plain", "ansi"), "null", {
  expect_snapshot({
    cat(code_highlight("NULL", list(null = "bold")))
  })
})

test_that_cli(configs = c("plain", "ansi"), "operator", {

  expect_snapshot({
    cat(code_highlight("~ ! 1 - 2 + 3:4 * 5 / 6 ^ 7", list(operator = "bold")))
    cat(code_highlight(
      "? 1 %% 2 %+% 2 < 3 & 4 > 5 && 6 == 7 | 8 <= 9 || 10 >= 11",
      list(operator = "bold")
    ))
    cat(code_highlight(
      "a <- 10; 20 -> b; c = 30; a$b; a@b",
      list(operator = "bold")
    ))
  })
})

test_that_cli(configs = c("plain", "ansi"), "call", {
  expect_snapshot({
    cat(code_highlight("ls(2)", list(call = "bold")))
  })
})

test_that_cli(configs = c("plain", "ansi"), "string", {
  expect_snapshot({
    cat(code_highlight("'s' + \"s\"", list(string = "bold")))
  })
})

test_that_cli(configs = c("plain", "ansi"), "comment", {
  expect_snapshot({
    cat(code_highlight(c("# COM", " ls() ## ANOT"), list(comment = "bold")))
  })
})

test_that_cli(configs = c("plain", "ansi"), "bracket", {
  expect_snapshot({
    cat(code_highlight("foo <- function(x){x}", list(bracket = list("bold"))))
  })
})

test_that("replace_in_place", {
  expect_equal(
    replace_in_place("1234567890", c(2, 6), c(5, 8), c("foobar", "xxx")),
    "1foobarxxx90"
  )

  expect_equal(
    replace_in_place("1234567890", c(1, 5), c(6, 10), c("A", "B")),
    "AB"
  )
})

test_that("replace_in_place corner cases", {
  expect_equal(
    replace_in_place("foobar", integer(), integer(), character()),
    "foobar"
  )

  expect_equal(
    replace_in_place("12345", 1L, 5L, "no!"),
    "no!"
  )

  expect_equal(
    replace_in_place("12345", 1:5, 1:5, letters[1:5]),
    "abcde"
  )
})

test_that_cli(configs = c("plain", "ansi"), "parse errors", {
  expect_equal(
    code_highlight("not good!!!"), "not good!!!"
  )
  cnd <- NULL
  withCallingHandlers(
    expect_equal(
      code_highlight("not good!!!"), "not good!!!"
    ),
    cli_parse_failure = function(e) cnd <<- e
  )
  expect_s3_class(cnd, "cli_parse_failure")
  expect_s3_class(cnd, "condition")
  expect_equal(cnd$code, "not good!!!")
})

test_that("code themes", {
  withr::local_options(cli.code_theme = "solarized_dark")
  expect_equal(code_theme_default()$reserved, "#859900")

  withr::local_options(cli.code_theme = NULL)
  withr::local_envvar(RSTUDIO = "0")
  withr::local_options(cli.code_theme_terminal = "solarized_light")
  expect_equal(code_theme_default()$reserved, "#859900")

  local_mocked_bindings(
    rstudio_detect = function() list(type = "rstudio_console")
  )
  withr::local_options(cli.code_theme_rstudio = "Xcode")
  expect_equal(code_theme_default()$reserved, "#C800A4")

  withr::local_options(cli.code_theme_rstudio = NULL)
  skip_if_not_installed("rstudioapi")
  local_mocked_bindings(code_theme_default_rstudio = function() "foo")
  expect_equal(code_theme_default(), "foo")
})

test_that("code themes 2", {
  withr::local_options(cli.code_theme = "nope!!")
  expect_warning(code_theme_default(), "Unknown cli code theme")

  withr::local_options(cli.code_theme = 111)
  expect_warning(code_theme_default(), "Invalid cli code theme")
})

test_that("code_theme_default_rstudio", {
  local_mocked_bindings(
    get_rstudio_theme = function() list(editor = "Solarized Dark")
  )
  expect_equal(code_theme_default_rstudio()$reserved, "#859900")

  local_mocked_bindings(
    get_rstudio_theme = function() list(editor = "Not really")
  )
  expect_warning(
    cth <- code_theme_default_rstudio(),
    "cli does not know this RStudio theme"
  )
  expect_equal(cth, code_theme_default_term())
})

test_that("code_theme_list", {
  expect_snapshot(code_theme_list())
})

test_that_cli(configs = "ansi", "new language features, raw strings", {
  if (getRversion() < "4.0.1") { expect_true(TRUE); return() }
  expect_snapshot(
    cat(code_highlight(
      '"old" + r"("new""")"',
      list(string = "bold", reserved = "italic")
    ))
  )
})

test_that_cli(configs = "ansi", "new language features, pipe", {
  if (getRversion() < "4.1.0") { expect_true(TRUE); return() }
  expect_snapshot(
    cat(code_highlight('dir() |> toupper()', list(operator = "bold")))
  )
})

test_that_cli(configs = "ansi", "new language features, lambda functions", {
  if (getRversion() < "4.1.0") { expect_true(TRUE); return() }
  expect_snapshot(
    cat(code_highlight('\\(x) x * 2', list(reserved = "bold")))
  )
})

test_that_cli(configs = c("plain", "unicode"), "spark_bar", {
  expect_snapshot({
    spark_bar(seq(0, 1, length = 8))
    spark_bar(c(0, NA, 0.5, NA, 1))
  })
})

test_that_cli(configs = c("plain", "unicode"), "spark_line", {
  expect_snapshot({
    spark_line(seq(0, 1, length = 10))
  })
})

#' Draw a banner-like box in the console
#'
#' @details
#'
#' ## Defaults
#'
#' ```{asciicast box-default}
#' boxx("Hello there!")
#' ```
#'
#' ## Change border style
#'
#' ```{asciicast box-border}
#' boxx("Hello there!", border_style = "double")
#' ```
#'
#' ## Multiple lines
#'
#' ```{asciicast box-lines}
#' boxx(c("Hello", "there!"), padding = 1)
#' ```
#'
#' ## Padding
#'
#' ```{asciicast box-padding}
#' boxx("Hello there!", padding = 1)
#' boxx("Hello there!", padding = c(1, 5, 1, 5))
#' ```
#'
#' ## Floating
#'
#' ```{asciicast box-float}
#' boxx("Hello there!", padding = 1, float = "center")
#' boxx("Hello there!", padding = 1, float = "right")
#' ```
#'
#' ## Text color
#'
#' ```{asciicast box-text-color}
#' boxx(col_cyan("Hello there!"), padding = 1, float = "center")
#' ```
#'
#' ## Background color
#'
#' ```{asciicast box-bg-color}
#' boxx("Hello there!", padding = 1, background_col = "brown")
#' boxx("Hello there!", padding = 1, background_col = bg_red)
#' ```
#'
#' ## Border color
#'
#' ```{asciicast box-border-color}
#' boxx("Hello there!", padding = 1, border_col = "green")
#' boxx("Hello there!", padding = 1, border_col = col_red)
#' ```
#'
#' ## Label alignment
#'
#' ```{asciicast box-label-align}
#' boxx(c("Hi", "there", "you!"), padding = 1, align = "left")
#' boxx(c("Hi", "there", "you!"), padding = 1, align = "center")
#' boxx(c("Hi", "there", "you!"), padding = 1, align = "right")
#' ```
#'
#' ## A very customized box
#'
#' ```{asciicast box-custom}
#' star <- symbol$star
#' label <- c(paste(star, "Hello", star), "  there!")
#' boxx(
#'   col_white(label),
#'   border_style="round",
#'   padding = 1,
#'   float = "center",
#'   border_col = "tomato3",
#'   background_col="darkolivegreen"
#' )
#' ```
#'
#' @param label Label to show, a character vector. Each element will be
#'   in a new line. You can color it using the `col_*`, `bg_*` and
#'   `style_*` functions, see [ANSI styles][ansi-styles] and the examples
#'   below.
#' @param header Text to show on top border of the box. If too long,
#'   it will be cut.
#' @param footer Text to show on the bottom border of the box. If too long,
#'   it will be cut.
#' @param border_style String that specifies the border style.
#'   `list_border_styles` lists all current styles.
#' @param padding Padding within the box. Either an integer vector of
#'   four numbers (bottom, left, top, right), or a single number `x`, which
#'   is interpreted as `c(x, 3*x, x, 3*x)`.
#' @param margin Margin around the box. Either an integer vector of four
#'   numbers (bottom, left, top, right), or a single number `x`, which is
#'   interpreted as `c(x, 3*x, x, 3*x)`.
#' @param float Whether to display the box on the `"left"`, `"center"`, or
#'   the `"right"` of the screen.
#' @param background_col Background color of the inside of the box.
#'   Either a style function (see [ANSI styles][ansi-styles]), or a color
#'   name which will be used in [make_ansi_style()] to create a
#'   *background* style (i.e. `bg = TRUE` is used).
#' @param col Color of text, and default border color. Either a style
#'   function (see [ANSI styles][ansi-styles]) or a color name that is
#'   passed to [make_ansi_style()].
#' @param border_col Color of the border. Either a style function
#'   (see [ANSI styles][ansi-styles]) or a color name that is passed to
#'   [make_ansi_style()].
#' @param align Alignment of the label within the box: `"left"`,
#'   `"center"`, or `"right"`.
#' @param width Width of the screen, defaults to [console_width()].
#'
#' @section About fonts and terminal settings:
#' The boxes might or might not look great in your terminal, depending
#' on the box style you use and the font the terminal uses. We found that
#' the Menlo font looks nice in most terminals an also in Emacs.
#'
#' RStudio currently has a line height greater than one for console output,
#' which makes the boxes ugly.
#'
#' @export

boxx <- function(label, header = "", footer = "",
                 border_style = "single", padding = 1, margin = 0,
                 float = c("left", "center", "right"),
                 col = NULL, background_col = NULL, border_col = col,
                 align = c("left", "center", "right"),
                 width = console_width()) {

  label <- apply_style(as.character(label), col)
  widest <- max(ansi_nchar(label, "width"), 0)

  stopifnot(
    is_border_style(border_style),
    is_padding_or_margin(padding),
    is_padding_or_margin(margin)
  )
  float <- match.arg(float)
  align <- match.arg(align)

  if (length(padding) == 1) {
    padding <- c(padding, padding * 3, padding, padding * 3)
  }
  if (length(margin) == 1) {
    margin <- c(margin, margin * 3, margin, margin * 3)
  }

  label <- ansi_align(label, align = align, width = widest)
  content_width <- widest + padding[2] + padding[4]

  mar_left <- if (float == "center") {
    make_space((width - content_width) / 2)
  } else if (float == "right") {
    make_space(max(width - content_width - 2, 0))
  } else {
    make_space(margin[2])
  }

  color_border <- function(x) apply_style(x, border_col)
  color_content <- function(x) apply_style(x, background_col, bg = TRUE)

  label <- c(rep("", padding[3]), label, rep("", padding[1]))

  chars <- box_styles()[border_style, ]

  pad_left <- make_space(padding[2])
  pad_right <- make_space(
    content_width - ansi_nchar(label, "width") - padding[2]
  )

  if (header != "") {
    header <- paste0(" ", ansi_strtrim(header, content_width - 2), " ")
  }
  hdw <- ansi_nchar(header, "width")
  if (footer != "") {
    footer <- paste0(" ", ansi_strtrim(footer, content_width - 2), " ")
  }
  ftw <- ansi_nchar(footer, "width")

  hdline <- paste0(header, strrep(chars$horizontal, content_width - hdw))
  top <- color_border(paste0(
    strrep("\n", margin[3]),
    mar_left, chars$top_left, hdline, chars$top_right
  ))
  ftline <- paste0(strrep(chars$horizontal, content_width - ftw), footer)
  bottom <- color_border(paste0(
    mar_left, chars$bottom_left, ftline, chars$bottom_right,
    strrep("\n", margin[1])
  ))
  side <- color_border(chars$vertical)

  middle <- paste0(mar_left, side,
                   color_content(paste0(pad_left, label, pad_right)), side)

  box <- paste0(top, "\n", paste0(middle, collapse = "\n"), "\n", bottom)

  class(box) <- unique(c("cli_boxx", "boxx", class(box), "character"))
  box
}

methods::setOldClass(c("cli_boxx", "character"))

#' @export

print.cli_boxx <- function(x, ..., sep = "\n") {
  cat(x, ..., sep = sep)
  invisible(x)
}

#' Add a progress bar to a mapping function or for loop
#'
#' @description
#' Note that this function is currently experimental!
#'
#' Use `cli_progress_along()` in a mapping function or in a for loop, to add a
#' progress bar. It uses [cli_progress_bar()] internally.
#'
#' @details
#'
#' ## `for` loop
#'
#' A `for` loop with `cli_progress_along()` looks like this:
#'
#' ```r
#' for (i in cli_progress_along(seq)) {
#'   ...
#' }
#' ```
#'
#' A complete example:
#'
#' ```{asciicast progress-along-1}
#' #| asciicast_at = "all",
#' #| asciicast_knitr_output = "svg",
#' #| asciciast_cursor = FALSE
#' clifun <- function() {
#'   for (i in cli_progress_along(1:100, "Downloading")) {
#'      Sys.sleep(4/100)
#'   }
#' }
#' clifun()
#' ```
#'
#' ## `lapply()` and other mapping functions
#'
#' They will look like this:
#'
#' ```r
#' lapply(cli_progress_along(X), function(i) ...)
#' ```
#'
#' A complete example:
#'
#' ```{asciicast progress-along-2}
#' #| asciicast_at = "all",
#' #| asciicast_knitr_output = "svg",
#' #| asciicast_cursor = FALSE
#' res <- lapply(cli_progress_along(1:100, "Downloading"), function(i) {
#'   Sys.sleep(4/100)
#' })
#' ```
#'
#' ## Custom format string
#'
#' ```{asciicast progress-along-3}
#' #| asciicast_at = "all",
#' #| asciicast_knitr_output = "svg",
#' #| asciicast_cursor = FALSE
#' clifun <- function() {
#'   for (i in cli_progress_along(1:100,
#'       format = "Downloading data file {cli::pb_current}")) {
#'      Sys.sleep(4/100)
#'   }
#' }
#' clifun()
#' ```
#'
#' ## Breaking out of loops
#'
#' Note that if you use `break` in the `for` loop, you probably want to
#' terminate the progress bar explicitly when breaking out of the loop,
#' or right after the loop:
#'
#' ```r
#' for (i in cli_progress_along(seq)) {
#'   ...
#'   if (cond) cli_progress_done() && break
#'   ...
#' }
#' ```
#'
#' @param x Sequence to add the progress bar to.
#' @param name Name of the progress bar, a label, passed to
#'   [cli_progress_bar()].
#' @param total Passed to [cli_progress_bar()].
#' @param ... Passed to [cli_progress_bar()].
#' @param .envir Passed to [cli_progress_bar()].
#'
#' @return An index vector from 1 to `length(x)` that triggers progress
#' updates as you iterate over it.
#'
#' @seealso This function supports [inline markup][inline-markup].
#' @seealso [cli_progress_bar()] and the traditional progress bar API.
#' @family progress bar functions
#' @family functions supporting inline markup
#' @export

cli_progress_along <- function(x,
                       name = NULL,
                       total = length(x),
                       ...,
                       .envir = parent.frame()) {

  name; total; .envir; list(...)

  if (getRversion() < "3.5.0") return(seq_along(x))
  id <- cli_progress_bar(name = name, total = total, ...,
                         .auto_close = FALSE, .envir = .envir)
  closeenv <- sys.frame(-1)
  if (format(closeenv) != clienv$globalenv) {
    defer(
      cli_progress_done(id = id, .envir = .envir, result = "auto"),
      envir = closeenv
    )
  }
  sax <- seq_along(x)
  clienv$progress[[id]]$caller <- .envir
  .Call(clic_progress_along, sax, clienv$progress[[id]])
}

progress_altrep_update <- function(pb) {
  tryCatch({
    cli_tick_reset()
    caller <- pb$caller
    pb$tick <- pb$tick + 1L

    if (is.null(pb$format)) {
      pb$format <- pb__default_format(pb$type, pb$total)
    }

    opt <- options(cli__pb = pb)
    on.exit(options(opt), add = TRUE)

    handlers <- cli_progress_select_handlers(pb, caller)
    if (is.null(pb$added)) {
      pb$added <- TRUE
      for (h in handlers) {
        if ("add" %in% names(h)) h$add(pb, .envir = caller)
      }
    } else {
      for (h in handlers) {
        if ("set" %in% names(h)) h$set(pb, .envir = caller)
      }
    }
  }, error = function(err) {
    if (!isTRUE(pb$warned)) {
      warning("cli progress bar update failed: ", conditionMessage(err),
              immediate. = TRUE)
    }
    pb$warned <- TRUE
  })

  NULL
}

#' List the currently active themes
#'
#' If there is no active app, then it calls [start_app()].
#'
#' @return A list of data frames with the active themes.
#' Each data frame row is a style that applies to selected CLI tree nodes.
#' Each data frame has columns:
#' * `selector`: The original CSS-like selector string. See [themes].
#' * `parsed`: The parsed selector, as used by cli for matching to nodes.
#' * `style`: The original style.
#' * `cnt`: The id of the container the style is currently applied to, or
#'   `NA` if the style is not used.
#'
#' @export
#' @seealso [themes]

cli_list_themes <- function() {
  app <- default_app() %||% start_app()
  app$list_themes()
}

clii_list_themes <- function(app) {
  app$themes
}

clii_add_theme <- function(app, theme) {
  id <- new_uuid()
  app$themes <-
    c(app$themes, structure(list(theme_create(theme)), names = id))
  id
}

clii_remove_theme <- function(app, id) {
  if (! id %in% names(app$themes)) return(invisible(FALSE))
  app$themes[[id]] <- NULL
  invisible(TRUE)
}

#' The built-in CLI theme
#'
#' This theme is always active, and it is at the bottom of the theme
#' stack. See [themes].
#'
#' # Showcase
#'
#' ```{asciicast builtin-theme}
#' cli_h1("Heading 1")
#' cli_h2("Heading 2")
#' cli_h3("Heading 3")
#'
#' cli_par()
#' cli_alert_danger("Danger alert")
#' cli_alert_warning("Warning alert")
#' cli_alert_info("Info alert")
#' cli_alert_success("Success alert")
#' cli_alert("Alert for starting a process or computation",
#'   class = "alert-start")
#' cli_end()
#'
#' cli_text("Packages and versions: {.pkg cli} {.version 1.0.0}.")
#' cli_text("Time intervals: {.timestamp 3.4s}")
#'
#' cli_text("{.emph Emphasis} and  {.strong strong emphasis}")
#'
#' cli_text("This is a piece of code: {.code sum(x) / length(x)}")
#' cli_text("Function names: {.fn cli::simple_theme}")
#'
#' cli_text("Files: {.file /usr/bin/env}")
#' cli_text("URLs: {.url https://r-project.org}")
#'
#' cli_h2("Longer code chunk")
#' cli_par(class = "code R")
#' cli_verbatim(
#'   '# window functions are useful for grouped mutates',
#'   'mtcars %>%',
#'   '  group_by(cyl) %>%',
#'   '  mutate(rank = min_rank(desc(mpg)))')
#' ```
#'
#' @seealso [themes], [simple_theme()].
#' @return A named list, a CLI theme.
#'
#' @param dark Whether to use a dark theme. The `cli.theme_dark` option
#'   can be used to request a dark theme explicitly. If this is not set,
#'   or set to `"auto"`, then cli tries to detect a dark theme, this
#'   works in recent RStudio versions and in iTerm on macOS.
#' @export

builtin_theme <- function(dark = getOption("cli.theme_dark", "auto")) {

  dark <- detect_dark_theme(dark)

  list(
    body = list(
      "class-map" = list(
        fs_path = "file",
        "cli-progress-bar" = "progress-bar"
      )
    ),

    h1 = list(
      "font-weight" = "bold",
      "margin-top" = 1,
      "margin-bottom" = 0,
      fmt = function(x) cli::rule(x, line_col = "cyan")),
    h2 = list(
      "font-weight" = "bold",
      "margin-top" = 1,
      "margin-bottom" = 1,
      fmt = function(x) paste0(symbol$line, symbol$line, " ", x, " ",
                               symbol$line, symbol$line)),
    h3 = list(
      "margin-top" = 1,
      fmt = function(x) paste0(symbol$line, symbol$line, " ", x, " ")),

    ".alert" = list(
      before = function() paste0(symbol$arrow_right, " ")
    ),
    ".alert-success" = list(
      before = function() paste0(col_green(symbol$tick), " ")
    ),
    ".alert-danger" = list(
      before = function() paste0(col_red(symbol$cross), " ")
    ),
    ".alert-warning" = list(
      before = function() paste0(col_yellow("!"), " ")
    ),
    ".alert-info" = list(
      before = function() paste0(col_cyan(symbol$info), " ")
    ),

    ".bullets .bullet-empty" = list(),
    ".bullets .bullet-space" = list("margin-left" = 2),
    ".bullets .bullet-v" = list(
      "text-exdent" = 2,
      before = function(x) paste0(col_green(symbol$tick), " ")
    ),
    ".bullets .bullet-x" = list(
      "text-exdent" = 2,
      before = function(x) paste0(col_red(symbol$cross), " ")
    ),
    ".bullets .bullet-!" = list(
      "text-exdent" = 2,
      before = function(x) paste0(col_yellow("!"), " ")
    ),
    ".bullets .bullet-i" = list(
      "text-exdent" = 2,
      before = function(x) paste0(col_cyan(symbol$info), " ")
    ),
    ".bullets .bullet-*" = list(
      "text-exdent" = 2,
      before = function(x) paste0(col_cyan(symbol$bullet), " ")
    ),
    ".bullets .bullet->" = list(
      "text-exdent" = 2,
      before = function(x) paste0(symbol$arrow_right, " ")
    ),
    ".bullets .bullet-1" = list(
    ),

    par = list("margin-top" = 0, "margin-bottom" = 1),
    ul = list(
      "list-style-type" = function() symbol$bullet
    ),

    # these are tags in HTML, but in cli they are inline
    span.dt = list(postfix = ": "),
    span.dd = list(),

    # This means that list elements have a margin, if they are nested
    "ul ul li" = list("margin-left" = 2),
    "ul ol li" = list("margin-left" = 2),
    "ul dl li" = list("margin-left" = 2),
    "ol ul li" = list("margin-left" = 2),
    "ol ol li" = list("margin-left" = 2),
    "ol dl li" = list("margin-left" = 2),
    "ol ul li" = list("margin-left" = 2),
    "ol ol li" = list("margin-left" = 2),
    "ol dl li" = list("margin-left" = 2),

    blockquote = list("padding-left" = 4L, "padding-right" = 10L,
                      "font-style" = "italic", "margin-top" = 1L,
                      "margin-bottom" = 1L,
                      before = function() symbol$dquote_left,
                      after = function() symbol$dquote_right),
    "blockquote cite" = list(
      before = function() paste0(symbol$em_dash, " "),
      "font-style" = "italic",
      "font-weight" = "bold"
    ),

    .code = list(fmt = format_code(dark)),
    .code.R = list(fmt = format_r_code(dark)),

    span.emph = list("font-style" = "italic"),
    span.strong = list("font-weight" = "bold"),
    span.code = theme_code_tick(dark),

    span.q   = list(fmt = quote_weird_name2),
    span.pkg = list(color = "blue"),
    span.fn = theme_function(dark),
    span.fun = theme_function(dark),
    span.arg = theme_code_tick(dark),
    span.kbd = list(before = "[", after = "]", color = "blue"),
    span.key = list(before = "[", after = "]", color = "blue"),
    span.file = theme_file(),
    span.path = theme_file(),
    span.email = list(
      color = "blue",
      transform = function(x) make_link(x, type = "email"),
      fmt = quote_weird_name
    ),
    span.url = list(
      before = "<", after = ">",
      color = "blue", "font-style" = "italic",
      transform = function(x) make_link(x, type = "url")
    ),
    span.href = list(
      transform = function(x) make_link(x, type = "href")
    ),
    span.help = list(
      transform = function(x) make_link(x, type = "help")
    ),
    span.topic = list(
      transform = function(x) make_link(x, type = "topic")
    ),
    span.vignette = list(
      transform = function(x) make_link(x, type = "vignette")
    ),
    span.run = list(
      transform = function(x) make_link(x, type = "run")
    ),
    span.var = theme_code_tick(dark),
    span.col = theme_code_tick(dark),
    span.str = list(fmt = encode_string),
    span.envvar = theme_code_tick(dark),
    span.val = list(
      transform = function(x, ...) cli_format(x, ...),
      color = "blue"
    ),
    span.field = list(color = "green"),
    span.cls = list(collapse = "/", color = "blue", before = "<", after = ">"),
    "span.progress-bar" = list(
      transform = theme_progress_bar,
      color = "green"
    ),
    span.obj_type_friendly = list(
      transform = function(x) format_inline(typename(x))
    ),
    span.type = list(
      transform = function(x) format_inline(typename(x))
    ),
    span.or = list("vec-sep2" = " or ", "vec-last" = ", or "),
    span.timestamp = list(before = "[", after = "]", color = "grey")
  )
}

encode_string <- function(x) {
  encodeString(x, quote = "\"")
}

quote_weird_name0 <- function(x) {
  x <- gsub(" ", "\u00a0", x, fixed = TRUE)
  x2 <- ansi_strip(x)

  fc <- first_character(x2)
  sc <- second_character(x2)
  lc <- last_character(x2)

  wfst <- !is_alnum(fc, ok = "~") || (fc == "~" && !is_alnum(sc))
  wlst <- !is_alnum(lc)

  if (wfst || wlst) {
    lsp <- leading_space(x2)
    tsp <- trailing_space(x2)
    if (nzchar(lsp)) {
      x <- paste0(
        bg_blue(lsp),
        ansi_substr(x, nchar(lsp) + 1, ansi_nchar(x))
      )
    }
    if (nzchar(tsp)) {
      x <- paste0(
        ansi_substr(x, 1, ansi_nchar(x) - nchar(tsp)),
        bg_blue(tsp)
      )
    }
  }

  list(x, wfst || wlst)
}

quote_weird_name <- function(x) {
  x2 <- quote_weird_name0(x)
  if (x2[[2]] || num_ansi_colors() == 1) {
    x2[[1]] <- paste0("'", x2[[1]], "'")
  }
  x2[[1]]
}

quote_weird_name2 <- function(x) {
  x2 <- quote_weird_name0(x)
  paste0("\"", x2[[1]], "\"")
}

theme_progress_bar <- function(x, app, style) {
  make_progress_bar(x$current / x$total, style = style)
}

detect_dark_theme <- function(dark) {
  tryCatch({
    if (dark == "auto") {
      dark <- if (Sys.getenv("RSTUDIO", "0") == "1") {
        get_rstudio_theme()$dark
      } else if (is_iterm()) {
        is_iterm_dark()
      } else if (is_emacs()) {
        Sys.getenv("ESS_BACKGROUND_MODE", "light") == "dark"
      } else {
        FALSE
      }
    }
  }, error = function(e) FALSE)

  isTRUE(dark)
}

theme_code <- function(dark) {
  list()
}

tick_formatter <- function(x) {
  tt <- grepl("`", x, fixed = TRUE) + 1L
  t1 <- c("`", "`` ")[tt]
  t2 <- c("`", " ``")[tt]
  paste0(t1, x, t2)
}

tick_formatter_fun <- function(x) {
  tt <- grepl("`", x, fixed = TRUE) + 1L
  t1 <- c("`", "`` ")[tt]
  t2 <- c("()`", " ()``")[tt]
  paste0(t1, x, t2)
}

theme_code_tick <- function(dark) {
  utils::modifyList(
    theme_code(dark),
    list(transform = tick_formatter)
  )
}

theme_function <- function(dark) {
  utils::modifyList(
    theme_code(dark),
    list(transform = function(x) tick_formatter_fun(make_link(x, type = "fun")))
  )
}

theme_file <- function() {
  list(
    color = "blue",
    transform = function(x) make_link(x, type = "file"),
    fmt = quote_weird_name
  )
}

format_r_code <- function(dark) {
  function(x) {
    x <- ansi_strip(x)
    lines <- unlist(strsplit(x, "\n", fixed = TRUE))
    code_highlight(lines)
  }
}

format_code <- function(dark) {
  function(x) {
    unlist(strsplit(x, "\n", fixed = TRUE))
  }
}

theme_create <- function(theme) {
  mtheme <- theme
  mtheme[] <- lapply(mtheme, create_formatter)
  selectors <- names(theme)
  res <- data.frame(
    stringsAsFactors = FALSE,
    selector = as.character(selectors),
    parsed = I(lapply(selectors, parse_selector) %||% list()),
    style = I(mtheme %||% list()),
    cnt = rep(NA_character_, length(selectors))
  )

  rownames(res) <- NULL
  res
}

create_formatter <- function(x) {
  is_bold <- identical(x[["font-weight"]], "bold")
  is_italic <- identical(x[["font-style"]], "italic")
  is_underline <- identical(x[["text-decoration"]], "underline")
  is_color <- "color" %in% names(x)
  is_bg_color <- "background-color" %in% names(x)

  if (!is_bold && !is_italic && !is_underline && !is_color
      && !is_bg_color) return(x)

  if (is_color && is.null(x[["color"]])) {
    x[["color"]] <- "none"
  }
  if (is_bg_color && is.null(x[["background-color"]])) {
    x[["background-color"]] <- "none"
  }

  fmt <- c(
    if (is_bold) list(style_bold),
    if (is_italic) list(style_italic),
    if (is_underline) list(style_underline),
    if (is_color) make_ansi_style(x[["color"]]),
    if (is_bg_color) make_ansi_style(x[["background-color"]], bg = TRUE)
  )

  new_fmt <- do.call(combine_ansi_styles, fmt)

  if (is.null(x[["fmt"]])) {
    x[["fmt"]] <- new_fmt
  } else {
    orig_fmt <- x[["fmt"]]
    x[["fmt"]] <- function(x) orig_fmt(new_fmt(x))
  }

  x
}

merge_embedded_styles <- function(old, new) {
  # before and after is not inherited, fmt is not inherited, either
  # side margins are additive, class mappings are merged
  # rest is updated, counter is reset, prefix and postfix are merged
  old$before <- old$after <- old$fmt <- NULL
  old$transform <- NULL

  # these will be applied on the container, so we don't need them inside
  old$color <- old$`background-color` <- NULL

  top <- new$`margin-top` %||% 0L
  bottom <- new$`margin-bottom` %||% 0L
  left <- (old$`margin-left` %||% 0L) + (new$`margin-left` %||% 0L)
  right <- (old$`margin-right` %||% 0L) + (new$`margin-right` %||% 0L)

  prefix <- paste0(old$prefix, new$prefix)
  postfix <- paste0(new$postfix, old$postfix)

  map <- utils::modifyList(old$`class-map` %||% list(), new$`class-map` %||% list())

  start <- new$start %||% 1L

  mrg <- utils::modifyList(old, new)
  mrg[c("margin-top", "margin-bottom", "margin-left", "margin-right",
        "start", "class-map", "prefix", "postfix")] <-
    list(top, bottom, left, right, start, map, prefix, postfix)

  ## Formatter needs to be re-generated
  create_formatter(mrg)
}

#' Parse a CSS3-like selector
#'
#' This is the rather small subset of CSS3 that is supported:
#'
#' Selectors:
#'
#' * Type selectors, e.g. `input` selects all `<input>` elements.
#' * Class selectors, e.g. `.index` selects any element that has a class
#'   of "index".
#' * ID selector. `#toc` will match the element that has the ID `"toc"`.
#'
#' Combinators:
#'
#' * Descendant combinator, i.e. the space, that combinator selects nodes
#'   that are descendants of the first element. E.g. `div span` will match
#'   all `<span>` elements that are inside a `<div>` element.
#'
#' @param x CSS3-like selector string.
#'
#' @keywords internal

parse_selector <- function(x) {
  lapply(strsplit(x, " ", fixed = TRUE)[[1]], parse_selector_node)
}

parse_selector_node <- function(x) {

  parse_ids <- function(y) {
    r <- strsplit(y, "#", fixed = TRUE)[[1]]
    if (length(r) > 1) r[-1] <- paste0("#", r[-1])
    r
  }

  parts <- strsplit(x, ".", fixed = TRUE)[[1]]
  if (length(parts) > 1) parts[-1] <- paste0(".", parts[-1])
  parts <- unlist(lapply(parts, parse_ids))
  parts <- parts[parts != ""]

  m_cls <- grepl("^\\.", parts)
  m_ids <- grepl("^#", parts)

  list(tag = as.character(unique(parts[!m_cls & !m_ids])),
       class = str_tail(unique(parts[m_cls])),
       id = str_tail(unique(parts[m_ids])))
}

#' Match a selector node to a container
#'
#' @param node Selector node, as parsed by `parse_selector_node()`.
#' @param cnt Container node, has elements `tag`, `id`, `class`.
#'
#' The selector node matches the container, if all these hold:
#'
#' * The id of the selector is missing or unique.
#' * The tag of the selector is missing or unique.
#' * The id of the container is missing or unique.
#' * The tag of the container is unique.
#' * If the selector specifies an id, it matches the id of the container.
#' * If the selector specifies a tag, it matches the tag of the container.
#' * If the selector specifies class names, the container has all these
#'   classes.
#'
#' @keywords internal

match_selector_node <- function(node, cnt) {
  if (length(node$id) > 1 || length(cnt$id) > 1) return(FALSE)
  if (length(node$tag) > 1 || length(cnt$tag) > 1) return(FALSE)
  all(node$id %in% cnt$id) &&
    all(node$tag %in% cnt$tag) &&
    all(node$class %in% cnt$class)
}

#' Match a selector to a container stack
#'
#' @param sels A list of selector nodes.
#' @param cnts A list of container nodes.
#'
#' The last selector in the list must match the last container, so we
#' do the matching from the back. This is because we use this function
#' to calculate the style of newly encountered containers.
#'
#' @keywords internal

match_selector <- function(sels, cnts) {
  sptr <- length(sels)
  cptr <- length(cnts)

  # Last selector must match the last container
  if (sptr == 0 || sptr > cptr) return(FALSE)
  match <- match_selector_node(sels[[sptr]], cnts[[cptr]])
  if (!match) return (FALSE)

  # Plus the rest should match somehow
  sptr <- sptr - 1L
  cptr <- cptr - 1L
  while (sptr != 0L && sptr <= cptr) {
    match <- match_selector_node(sels[[sptr]], cnts[[cptr]])
    if (match) {
      sptr <- sptr - 1L
      cptr <- cptr - 1L
    } else {
      cptr <- cptr - 1L
    }
  }

  sptr == 0
}

test_that("parse_selector_node", {

  empty <- list(tag = character(), class = character(), id = character())

  cases <- list(
    list("", empty),
    list("tag", list(tag = "tag")),
    list(".class", list(class = "class")),
    list("#id", list(id = "id")),
    list("tag.class", list(tag = "tag", class = "class")),
    list("tag.c1.c2.c3", list(tag = "tag", class = c("c1", "c2", "c3"))),
    list("tag#id", list(tag = "tag", id = "id")),
    list("tag#id.class", list(tag = "tag", class = "class", id = "id")),
    list("tag.class#id", list(tag = "tag", class = "class", id = "id")),
    list("#id.class", list(class = "class", id = "id")),
    list("#id.c1.c2", list(class = c("c1", "c2"), id = "id")),
    list("#id1#id2", list(id = c("id1", "id2")))
  )

  for (c in cases) {
    exp <- modifyList(empty, c[[2]])
    expect_identical(parse_selector_node(c[[1]]), exp, info = c[[1]])
  }
})

test_that("parse_selector", {

  empty <- list(tag = character(), class = character(), id = character())

  cases <- list(
    list("", list()),
    list("foo", list(list(tag = "foo"))),
    list("foo bar", list(list(tag = "foo"), list(tag = "bar"))),
    list("foo.c1 bar.c2",
         list(list(tag = "foo", class = "c1"),
              list(tag = "bar", class = "c2"))),
    list("#i1 tag #i2 .cl",
         list(list(id = "i1"), list(tag = "tag"), list(id = "i2"),
              list(class = "cl")))
  )

  for (c in cases) {
    exp <- lapply(c[[2]], function(x) modifyList(empty, x))
    expect_identical(parse_selector(c[[1]]), exp, info = c[[1]])
  }
})

test_that("match_selector_node", {

  default <- list(tag = "mytag", class = character(), id = "myid")

  pos <- list(
    list("foo", list(tag = "foo", class = "class"), id = "id"),
    list(".class", list(tag = "foo", class = "class"), id = "id"),
    list("foo.class", list(tag = "foo", class = "class", id = "id")),
    list("#id", list(tag = "foo", class = "class", id = "id")),
    list(".c", list(class = c("c", "d", "e")))
  )

  for (c in pos) {
    sel <- parse_selector_node(c[[1]])
    cnt <- modifyList(default, c[[2]])
    expect_true(match_selector_node(sel, cnt), info = c[[1]])
  }

  neg <- list(
    list("foo", list()),
    list(".class", list()),
    list("#id", list()),
    list("foo.class", list()),
    list("foo.c1.c2", list(tag = "foo", class = c("c1", "c3")))
  )

  for (c in neg) {
    sel <- parse_selector_node(c[[1]])
    cnt <- modifyList(default, c[[2]])
    expect_false(match_selector_node(sel, cnt), info = c[[1]])
  }
  
})

test_that("match_selector", {

  default <- list(tag = "mytag", class = character(), id = "myid")

  pos <- list(
    list("foo bar", list(list(tag = "foo"), list(tag = "bar"))),
    list("bar", list(list(tag = "foo"), list(tag = "bar"))),
    list(".class", list(list(tag = "x"), list(class = "class"))),
    list(".c1",
         list(list(tag = "x"), list(class = "c"), list(class = "c1")))
  )

  for (c in pos) {
    sels <- parse_selector(c[[1]])
    cnts <- lapply(c[[2]], function(x) modifyList(default, x))
    expect_true(match_selector(sels, cnts), info = c[[1]])
  }
  
  neg <- list(
    list("foo bar", list(list(tag = "foo"), list(tag = "ba"))),
    list("foo bar", list(list(tag = "foo"), list(class = "bar"))),
    list(".class", list(list(tag = "x"), list(class = "class1"))),
    list(".c1", list(list(tag = "x"), list(class = "c")))
  )

  for (c in neg) {
    sels <- parse_selector(c[[1]])
    cnts <- lapply(c[[2]], function(x) modifyList(default, x))
    expect_false(match_selector(sels, cnts), info = c[[1]])
  }
})

context("css parsing")

test_that("parse_selector_node", {

  empty <- list(tag = character(), class = character(), id = character())

  cases <- list(
    list("", empty),
    list("tag", list(tag = "tag")),
    list(".class", list(class = "class")),
    list("#id", list(id = "id")),
    list("tag.class", list(tag = "tag", class = "class")),
    list("tag.c1.c2.c3", list(tag = "tag", class = c("c1", "c2", "c3"))),
    list("tag#id", list(tag = "tag", id = "id")),
    list("tag#id.class", list(tag = "tag", class = "class", id = "id")),
    list("tag.class#id", list(tag = "tag", class = "class", id = "id")),
    list("#id.class", list(class = "class", id = "id")),
    list("#id.c1.c2", list(class = c("c1", "c2"), id = "id")),
    list("#id1#id2", list(id = c("id1", "id2")))
  )

  for (c in cases) {
    exp <- modifyList(empty, c[[2]])
    expect_identical(parse_selector_node(c[[1]]), exp, info = c[[1]])
  }
})

test_that("parse_selector", {

  empty <- list(tag = character(), class = character(), id = character())

  cases <- list(
    list("", list()),
    list("foo", list(list(tag = "foo"))),
    list("foo bar", list(list(tag = "foo"), list(tag = "bar"))),
    list("foo.c1 bar.c2",
         list(list(tag = "foo", class = "c1"),
              list(tag = "bar", class = "c2"))),
    list("#i1 tag #i2 .cl",
         list(list(id = "i1"), list(tag = "tag"), list(id = "i2"),
              list(class = "cl")))
  )

  for (c in cases) {
    exp <- lapply(c[[2]], function(x) modifyList(empty, x))
    expect_identical(parse_selector(c[[1]]), exp, info = c[[1]])
  }
})

test_that("match_selector_node", {

  default <- list(tag = "mytag", class = character(), id = "myid")

  pos <- list(
    list("foo", list(tag = "foo", class = "class"), id = "id"),
    list(".class", list(tag = "foo", class = "class"), id = "id"),
    list("foo.class", list(tag = "foo", class = "class", id = "id")),
    list("#id", list(tag = "foo", class = "class", id = "id")),
    list(".c", list(class = c("c", "d", "e")))
  )

  for (c in pos) {
    sel <- parse_selector_node(c[[1]])
    cnt <- modifyList(default, c[[2]])
    expect_true(match_selector_node(sel, cnt), info = c[[1]])
  }

  neg <- list(
    list("foo", list()),
    list(".class", list()),
    list("#id", list()),
    list("foo.class", list()),
    list("foo.c1.c2", list(tag = "foo", class = c("c1", "c3")))
  )

  for (c in neg) {
    sel <- parse_selector_node(c[[1]])
    cnt <- modifyList(default, c[[2]])
    expect_false(match_selector_node(sel, cnt), info = c[[1]])
  }
  
})

test_that("match_selector", {

  default <- list(tag = "mytag", class = character(), id = "myid")

  pos <- list(
    list("foo bar", list(list(tag = "foo"), list(tag = "bar"))),
    list("bar", list(list(tag = "foo"), list(tag = "bar"))),
    list(".class", list(list(tag = "x"), list(class = "class"))),
    list(".c1",
         list(list(tag = "x"), list(class = "c"), list(class = "c1")))
  )

  for (c in pos) {
    sels <- parse_selector(c[[1]])
    cnts <- lapply(c[[2]], function(x) modifyList(default, x))
    expect_true(match_selector(sels, cnts), info = c[[1]])
  }
  
  neg <- list(
    list("foo bar", list(list(tag = "foo"), list(tag = "ba"))),
    list("foo bar", list(list(tag = "foo"), list(class = "bar"))),
    list(".class", list(list(tag = "x"), list(class = "class1"))),
    list(".c1", list(list(tag = "x"), list(class = "c")))
  )

  for (c in neg) {
    sels <- parse_selector(c[[1]])
    cnts <- lapply(c[[2]], function(x) modifyList(default, x))
    expect_false(match_selector(sels, cnts), info = c[[1]])
  }
})

lorem_words <- c(
  "ad",
  "adipisicing",
  "aliqua",
  "aliquip",
  "amet",
  "anim",
  "aute",
  "cillum",
  "commodo",
  "consectetur",
  "consequat",
  "culpa",
  "cupidatat",
  "deserunt",
  "do",
  "dolor",
  "dolore",
  "duis",
  "ea",
  "eiusmod",
  "elit",
  "enim",
  "esse",
  "est",
  "et",
  "eu",
  "ex",
  "excepteur",
  "exercitation",
  "fugiat",
  "id",
  "in",
  "incididunt",
  "ipsum",
  "irure",
  "labore",
  "laboris",
  "laborum",
  "Lorem",
  "magna",
  "minim",
  "mollit",
  "nisi",
  "non",
  "nostrud",
  "nulla",
  "occaecat",
  "officia",
  "pariatur",
  "proident",
  "qui",
  "quis",
  "reprehenderit",
  "sint",
  "sit",
  "sunt",
  "tempor",
  "ullamco",
  "ut",
  "velit",
  "veniam",
  "voluptate"
)

lorem_ipsum <- function(paragraphs = 1, par_sentence_range = 5:10,
                        sentence_word_range = 5:15) {
  vcapply(
    1:paragraphs,
    function(x, ...) lorem_paragraph(...),
    par_sentence_range = par_sentence_range,
    sentence_word_range = sentence_word_range
  )
}

lorem_paragraph <- function(par_sentence_range, sentence_word_range) {
  num <- sample(par_sentence_range, 1)
  paste(
    collapse = " ",
    vcapply(
      1:num,
      function(x, ...) lorem_sentence(...),
      sentence_word_range = sentence_word_range
    )
  )
}

lorem_sentence <- function(sentence_word_range) {
  num <- sample(sentence_word_range, 1)
  words <- sample(lorem_words, num, replace = TRUE)
  words[1] <- capitalize(words[1])
  paste0(paste(words, collapse = " "), ".")
}

capitalize <- function(x) {
  substr(x, 1, 1) <- toupper(substr(x, 1, 1))
  x
}

add_child <- function(x, tag, ...) {
  push(x, list(tag = tag, ...))
}

clii__container_start <- function(app, tag, class = NULL,
                                  id = NULL, theme = NULL) {

  id <- id %||% new_uuid()
  if (!length(class)) class <- ""
  class <- setdiff(unique(strsplit(class, " ", fixed = TRUE)[[1]]), "")

  app$doc <- add_child(app$doc, tag, id = id, class = class,
                           theme = theme)

  ## Go over all themes, and collect the selectors that match the
  ## current element
  new_sels <- list()
  for (t in seq_along(app$themes)) {
    theme <- app$themes[[t]]
    for (i in seq_len(nrow(theme))) {
      if (match_selector(theme$parsed[[i]], app$doc)) {
        app$themes[[t]]$cnt[i] <- id
        new_sels <- utils::modifyList(new_sels, theme$style[[i]])
      }
    }
  }
  new_style <- merge_embedded_styles(last(app$styles) %||% list(), new_sels)
  app$styles <- push(app$styles, new_style, name = id)

  ## Top margin, if any
  app$vspace(new_style$`margin-top` %||% 0)

  invisible(id)
}

clii__container_end <- function(app, id) {
  debug <- is_yes(Sys.getenv("CLI_DEBUG_BAD_END", ""))

  ## Defaults to last container
  if (is.null(id) || is.na(id)) id <- last(app$doc)$id

  ## Do not remove the <body>
  if (id == "body") {
    if (debug) warning("No cli container to close")
    return(invisible(app))
  }

  ## Do we have 'id' at all?
  wh <- which(vlapply(app$doc, function(x) identical(x$id, id)))[1]
  if (is.na(wh)) {
    if (debug) warning("Can't find cli container '", id, "' to close")
    return(invisible(app))
  }

  ## ids to remove
  del_ids <- unlist(lapply(utils::tail(app$doc, - (wh - 1L)), "[[", "id"))

  ## themes to remove
  del_thm <- unlist(lapply(utils::tail(app$doc, - (wh - 1L)), "[[", "theme"))

  ## Remove the whole subtree of 'cnt'
  app$doc <- utils::head(app$doc, wh - 1L)

  ## Bottom margin
  del_from <- match(id, names(app$styles))
  bottom <- max(viapply(
    app$styles[del_from:length(app$styles)],
    function(x) as.integer(x$`margin-bottom` %||% 0L)
  ))
  app$vspace(bottom)

  ## Remove styles
  app$styles <- utils::head(app$styles, del_from - 1L)

  ## Remove claimed styles that are not used any more
  for (t in seq_along(app$themes)) {
    m <- app$themes[[t]]$cnt %in% del_ids
    app$themes[[t]]$cnt[m] <- NA_character_
  }

  ## Remove themes
  app$themes <- app$themes[setdiff(names(app$themes), del_thm)]

  invisible(app)
}

## div --------------------------------------------------------------

clii_div <- function(app, id, class, theme) {
  theme_id <- app$add_theme(theme)
  clii__container_start(app, "div", class, id, theme = theme_id)
  id
}

## Paragraph --------------------------------------------------------

clii_par <- function(app, id, class) {
  clii__container_start(app, "par", class, id)
}

## Lists ------------------------------------------------------------

clii_ul <- function(app, items, id, class, .close) {
  id <- clii__container_start(app, "ul", id = id, class = class)
  if (length(items)) { app$li(items); if (.close) app$end(id) }
  invisible(id)
}

clii_ol <- function(app, items, id, class, .close) {
  id <- clii__container_start(app, "ol", id = id, class = class)
  if (length(items)) { app$li(items); if (.close) app$end(id) }
  invisible(id)
}

clii_dl <- function(app, items, labels, id, class, .close) {
  id <- clii__container_start(app, "dl", id = id, class = class)
  if (length(items)) { app$li(items, labels); if (.close) app$end(id) }
  invisible(id)
}

clii_li <- function(app, items, labels, id, class) {
  id <- id %||% new_uuid()

  ## check the last active list container
  last <- length(app$doc)
  while (! app$doc[[last]]$tag %in% c("ul", "ol", "dl", "body")) {
    last <- last - 1L
  }

  ## if not the last container, close the ones below it
  if (app$doc[[last]]$tag != "body" &&
      last != length(app$doc)) {
    app$end(app$doc[[last + 1L]]$id)
  }

  ## if none, then create an ul container
  if (app$doc[[last]]$tag == "body") {
    cnt_id <- app$ul()
    type <- "ul"
  } else {
    cnt_id <- app$doc[[last]]$id
    type <- app$doc[[last]]$tag
  }

  if (length(items) > 0) {
    for (i in seq_along(items)) {
      id <- clii__container_start(app, "li", id = id, class = class)
      app$item_text(type, labels[[i]], cnt_id, items[[i]])
      if (i < length(items)) app$end(id)
    }
  } else {
    app$delayed_item <- list(type = type, cnt_id = cnt_id)
    id <- clii__container_start(app, "li", id = id, class = class)
  }

  invisible(id)
}

clii__item_text <- function(app, type, name, cnt_id, text, .list) {

  style <- app$get_current_style()
  cnt_style <- app$styles[[cnt_id]]

  head <- if (type == "dl") name else glue_delay(name)

  head$str <- if (type == "ul") {
    paste0(call_if_fun(style$`list-style-type`) %||% "*", " ")
  } else if (type == "ol") {
    res <- paste0(cnt_style$start %||% 1L, ". ")
    app$styles[[cnt_id]]$start <- (cnt_style$start %||% 1L) + 1L
    res
  } else if (type == "dl") {
    mrk <- text$values$marker
    text$str <- paste0("<", mrk, ".dd ", text$str, mrk, ">")
    mrk2 <- head$values$marker
    paste0("<", mrk2, ".dt ", head$str, mrk2, ">")
  }

  app$xtext(
    .list = c(list(head), list(text), .list),
    indent = - (style$`padding-left` %||% 0),
    padding = (cnt_style$`padding-left` %||% 0)
  )
}

## Close container(s) -----------------------------------------------

clii_end <- function(app, id) {
  clii__container_end(app, id)
}

#' Labels a character vector as containing ANSI control codes.
#'
#' This function sets the class of its argument, activating
#' ANSI-string-specific methods such as for printing.
#'
#' @param x A character vector or something that can be
#'   coerced into one.
#' @return A `cli_ansi_string` object, a subclass of
#'   `character`, with the same length and contents
#'   as `x`.
#' @family low level ANSI functions
#' @export
ansi_string <- function(x) {
  if (!is.character(x)) x <- as.character(x)
  x <- enc2utf8(x)
  class(x) <- unique(c("cli_ansi_string", "ansi_string", class(x), "character"))
  x
}

#' Perl compatible regular expression that matches ANSI escape
#' sequences
#'
#' Don't forget to use `perl = TRUE` when using this with [grepl()] and
#' friends.
#'
#' @return String scalar, the regular expression.
#'
#' @family low level ANSI functions
#' @export

ansi_regex <- function() {
  paste0(
    "(?:(?:\\x{001b}\\[)|\\x{009b})",
    "(?:(?:[0-9]{1,3})?(?:(?:;[0-9]{0,3})*)?[A-M|f-m])",
    "|\\x{001b}[A-M]",
    # this is for hyperlinks, we must be non-greedy
    "|\\x{001b}\\]8;.*?;.*?\\x{001b}\\\\",
    "|\\x{001b}\\]8;.*?;.*?\\x{0007}"
  )
}

#' Check if a string has some ANSI styling
#'
#' @param string The string to check. It can also be a character
#'   vector.
#' @param sgr Whether to look for SGR (styling) control sequences.
#' @param csi Whether to look for non-SGR control sequences.
#' @param link Whether to look for ANSI hyperlinks.
#' @return Logical vector, `TRUE` for the strings that have some
#'   ANSI styling.
#'
#' @family low level ANSI functions
#' @export
#' @examples
#' ## The second one has style if ANSI colors are supported
#' ansi_has_any("foobar")
#' ansi_has_any(col_red("foobar"))

ansi_has_any <- function(string, sgr = TRUE, csi = TRUE, link = TRUE) {
  if (!is.character(string)) string <- as.character(string)
  string <- enc2utf8(string)
  stopifnot(
    is_flag(sgr),
    is_flag(csi),
    is_flag(link)
  )
  .Call(clic_ansi_has_any, string, sgr, csi, link)
}

#' Remove ANSI escape sequences from a string
#'
#' The input may be of class `cli_ansi_string` class, this is also dropped
#' from the result.
#'
#' @param string The input string.
#' @param sgr Whether to remove for SGR (styling) control sequences.
#' @param csi Whether to remove for non-SGR control sequences.
#' @param link Whether to remove ANSI hyperlinks.
#' @return The cleaned up string. Note that `ansi_strip()` always drops
#' the `cli_ansi_string` class, even if `sgr` and sci` are `FALSE`.
#'
#' @family low level ANSI functions
#' @export
#' @examples
#' ansi_strip(col_red("foobar")) == "foobar"

ansi_strip <- function(string, sgr = TRUE, csi = TRUE, link = TRUE) {
  if (!is.character(string)) string <- as.character(string)
  string <- enc2utf8(string)
  stopifnot(
    is_flag(sgr),
    is_flag(csi),
    is_flag(link)
  )
  clean <- .Call(clic_ansi_strip, string, sgr, csi, link)
  class(clean) <- setdiff(class(clean), c("cli_ansi_string", "ansi_string"))
  clean
}

#' Count number of characters in an ANSI colored string
#'
#' This is a color-aware counterpart of [utf8_nchar()]. By default it
#' counts Unicode grapheme clusters, instead of code points.
#'
#' @param x Character vector, potentially ANSI styled, or a vector to be
#'   coerced to character. If it converted to UTF-8.
#' @param type Whether to count graphemes (characters), code points,
#'   bytes, or calculate the display width of the string.
#' @return Numeric vector, the length of the strings in the character
#'   vector.
#'
#' @family ANSI string operations
#' @export
#' @examples
#' str <- paste(
#'   col_red("red"),
#'   "default",
#'   col_green("green")
#' )
#'
#' cat(str, "\n")
#' nchar(str)
#' ansi_nchar(str)
#' nchar(ansi_strip(str))

ansi_nchar <- function(x,
                       type = c("chars", "bytes", "width", "graphemes",
                                "codepoints")) {
  type <- match.arg(type)
  if (type == "chars") type <- "graphemes"
  type <- match(type, c("graphemes", "bytes", "width", "codepoints"))
  if (!is.character(x)) x <- as.character(x)
  x <- enc2utf8(x)
  .Call(clic_ansi_nchar, x, type)
}

#' Substring(s) of an ANSI colored string
#'
#' This is a color-aware counterpart of [base::substr()].
#' It works exactly like the original, but keeps the colors
#' in the substrings. The ANSI escape sequences are ignored when
#' calculating the positions within the string.
#'
#' @param x Character vector, potentially ANSI styled, or a vector to
#'   coerced to character.
#' @param start Starting index or indices, recycled to match the length
#'   of `x`.
#' @param stop Ending index or indices, recycled to match the length
#'   of `x`.
#' @return Character vector of the same length as `x`, containing
#'   the requested substrings. ANSI styles are retained.
#'
#' @family ANSI string operations
#' @export
#' @examples
#' str <- paste(
#'   col_red("red"),
#'   "default",
#'   col_green("green")
#' )
#'
#' cat(str, "\n")
#' cat(ansi_substr(str, 1, 5), "\n")
#' cat(ansi_substr(str, 1, 15), "\n")
#' cat(ansi_substr(str, 3, 7), "\n")
#'
#' substr(ansi_strip(str), 1, 5)
#' substr(ansi_strip(str), 1, 15)
#' substr(ansi_strip(str), 3, 7)
#'
#' str2 <- paste(
#'   "another",
#'   col_red("multi-", style_underline("style")),
#'   "text"
#' )
#'
#' cat(str2, "\n")
#' cat(ansi_substr(c(str, str2), c(3,5), c(7, 18)), sep = "\n")
#' substr(ansi_strip(c(str, str2)), c(3,5), c(7, 18))

ansi_substr <- function(x, start, stop) {
  if (!is.character(x)) x <- as.character(x)
  if (!length(x)) return(ansi_string(x))
  start <- suppressWarnings(as.integer(start))
  stop <- suppressWarnings(as.integer(stop))
  if (!length(start) || !length(stop)) {
    throw(cli_error(
      "{.code ansi_substr()} must have non-empty {.arg start} and {.arg stop} arguments",
      "i" = if (!length(start)) "{.arg start} has length {length(start)}",
      "i" = if (!length(stop)) "{.arg stop} has length {length(stop)}"
    ))
  }
  nastart <- anyNA(start)
  nastop <- anyNA(stop)
  if (nastart || nastop) {
    throw(cli_error(
      "{.arg start} and {.arg stop} must not have {.code NA} values",
      "i" = if (nastart) paste(
              "{.arg start} has {sum(is.na(start))}",
              "{.code NA} value{?s}, after coercion to integer"),
      "i" = if (nastop) paste(
              "{.arg stop} has {sum(is.na(stop))} {.code NA} value{?s},",
              "after coercion to integer")
    ))
  }
  x <- enc2utf8(x)
  start <- rep_len(start, length(x))
  stop <- rep_len(stop, length(x))
  .Call(clic_ansi_substr, x, start, stop)
}

#' Substring(s) of an ANSI colored string
#'
#' This is the color-aware counterpart of [base::substring()].
#' It works exactly like the original, but keeps the colors in the
#' substrings. The ANSI escape sequences are ignored when
#' calculating the positions within the string.
#'
#' @param text Character vector, potentially ANSI styled, or a vector to
#'   coerced to character. It is recycled to the longest of `first`
#'   and `last`.
#' @param first Starting index or indices, recycled to match the length
#'   of `x`.
#' @param last Ending index or indices, recycled to match the length
#'   of `x`.
#' @return Character vector of the same length as `x`, containing
#'   the requested substrings. ANSI styles are retained.
#'
#' @family ANSI string operations
#' @export
#' @examples
#' str <- paste(
#'   col_red("red"),
#'   "default",
#'   col_green("green")
#' )
#'
#' cat(str, "\n")
#' cat(ansi_substring(str, 1, 5), "\n")
#' cat(ansi_substring(str, 1, 15), "\n")
#' cat(ansi_substring(str, 3, 7), "\n")
#'
#' substring(ansi_strip(str), 1, 5)
#' substring(ansi_strip(str), 1, 15)
#' substring(ansi_strip(str), 3, 7)
#'
#' str2 <- paste(
#'   "another",
#'   col_red("multi-", style_underline("style")),
#'   "text"
#' )
#'
#' cat(str2, "\n")
#' cat(ansi_substring(str2, c(3,5), c(7, 18)), sep = "\n")
#' substring(ansi_strip(str2), c(3,5), c(7, 18))

ansi_substring <- function(text, first, last = 1000000L) {
  if (!is.character(text)) text <- as.character(text)
  n <- max(lt <- length(text), length(first), length(last))
  if (lt && lt < n) text <- rep_len(text, length.out = n)
  text <- enc2utf8(text)
  first <- rep_len(as.integer(first), n)
  last <- rep_len(as.integer(last), n)
  .Call(clic_ansi_substr, text, first, last)
}


#' Split an ANSI colored string
#'
#' This is the color-aware counterpart of [base::strsplit()].
#' It works almost exactly like the original, but keeps the colors in the
#' substrings.
#'
#' @param x Character vector, potentially ANSI styled, or a vector to
#'   coerced to character.
#' @param split Character vector of length 1 (or object which can be coerced to
#'   such) containing regular expression(s) (unless `fixed = TRUE`) to use
#'   for splitting.  If empty matches occur, in particular if `split` has
#'   zero characters, `x` is split into single characters.
#' @param ... Extra arguments are passed to `base::strsplit()`.
#' @return A list of the same length as `x`, the `i`-th element of
#'   which contains the vector of splits of `x[i]`. ANSI styles are
#'   retained.
#'
#' @family ANSI string operations
#' @export
#' @examples
#' str <- paste0(
#'   col_red("I am red---"),
#'   col_green("and I am green-"),
#'   style_underline("I underlined")
#' )
#'
#' cat(str, "\n")
#'
#' # split at dashes, keep color
#' cat(ansi_strsplit(str, "[-]+")[[1]], sep = "\n")
#' strsplit(ansi_strip(str), "[-]+")
#'
#' # split to characters, keep color
#' cat(ansi_strsplit(str, "")[[1]], "\n", sep = " ")
#' strsplit(ansi_strip(str), "")

ansi_strsplit <- function(x, split, ...) {
  split <- try(as.character(split), silent = TRUE)
  if (inherits(split, "try-error") || !is.character(split) || length(split) > 1L) {
    throw(cli_error(
      "{.arg split} must be character of length <= 1, or must coerce to that",
      i = "{.arg split} is (or was coerced to) {.type {split}}"
    ))
  }
  if (!is.character(x)) x <- as.character(x)
  x <- enc2utf8(x)
  if(!length(split)) split <- ""
  plain <- ansi_strip(x)
  splits <- re_table(split, plain, ...)
  chunks <- non_matching(splits, plain, empty = TRUE)
  # silently recycle `split`; doesn't matter currently since we don't support
  # split longer than 1, but might in future
  split.r <- rep(split, length.out=length(x))
  # Drop empty chunks to align with `substr` behavior
  chunks <- lapply(
    seq_along(chunks),
    function(i) {
      y <- chunks[[i]]
      # empty split means drop empty first match
      if(nrow(y) && !nzchar(split.r[[i]]) && !utils::head(y, 1L)[, "length"]) {
        y <- y[-1L, , drop=FALSE]
      }
      # drop empty last matches
      if(nrow(y) && !utils::tail(y, 1L)[, "length"]) y[-nrow(y), , drop=FALSE] else y
    }
  )
  zero.chunks <- !vapply(chunks, nrow, integer(1L))
  # Pull out zero chunks from colored string b/c ansi_substring won't work
  # with them
  res <- vector("list", length(chunks))
  res[zero.chunks] <- list(character(0L))
  res[!zero.chunks] <- mapply(
    chunks[!zero.chunks], x[!zero.chunks], SIMPLIFY = FALSE,
    FUN = function(tab, xx) ansi_substring(xx, tab[, "start"], tab[, "end"])
  )
  lapply(res, ansi_string)
}

#' Align an ANSI colored string
#'
#' @details
#'
#' ```{asciicast ansi-align}
#' str <- c(
#'   col_red("This is red"),
#'   style_bold("This is bold")
#' )
#' astr <- ansi_align(str, width = 30)
#' boxx(astr)
#' ```
#'
#' ```{asciicast ansi-align-center}
#' str <- c(
#'   col_red("This is red"),
#'   style_bold("This is bold")
#' )
#' astr <- ansi_align(str, align = "center", width = 30)
#' boxx(astr)
#' ```
#'
#' ```{asciicast ansi-align-right}
#' str <- c(
#'   col_red("This is red"),
#'   style_bold("This is bold")
#' )
#' astr <- ansi_align(str, align = "right", width = 30)
#' boxx(astr)
#' ```
#'
#' @param text The character vector to align.
#' @param width Width of the field to align in.
#' @param align Whether to align `"left"`, `"center"` or `"right"`.
#' @param type Passed on to [ansi_nchar()] and there to [nchar()]
#' @return The aligned character vector.
#'
#' @family ANSI string operations
#' @export

# TODO: show wide Unicode charadcters, once they work in asciicast

ansi_align <- function(text, width = console_width(),
                      align = c("left", "center", "right"),
                      type = "width") {

  align <- match.arg(align)
  text <- enc2utf8(text)
  nc <- ansi_nchar(text, type = type)

  if (!length(text)) return(ansi_string(text))

  res <- if (align == "left") {
    paste0(text, make_space(width - nc))

  } else if (align == "center") {
    paste0(make_space(ceiling((width - nc) / 2)),
           text,
           make_space(floor((width - nc) / 2)))

  } else {
    paste0(make_space(width - nc), text)
  }

  ansi_string(res)
}

make_space <- function(num, filling = " ") {
  num <- pmax(0, num)
  res <- strrep(filling, num)
  Encoding(res) <- Encoding(filling)
  res
}

strrep <- function (x, times) {
  x = as.character(x)
  if (length(x) == 0L) return(x)

  mapply(
    function(x, times) {
      if (is.na(x) || is.na(times)) {
        NA_character_
      } else if (times <= 0L) {
        ""
      } else {
        paste0(rep(x, times), collapse = "")
      }
    },
    x, times,
    USE.NAMES = FALSE
  )
}

#' Remove leading and/or trailing whitespace from an ANSI string
#'
#' This function is similar to [base::trimws()] but works on ANSI strings,
#' and keeps color and other styling.
#'
#' @param x ANSI string vector.
#' @param which Whether to remove leading or trailing whitespace or both.
#' @return ANSI string, with the whitespace removed.
#'
#' @family ANSI string operations
#' @export
#' @examples
#' trimws(paste0("   ", col_red("I am red"), "   "))
#' ansi_trimws(paste0("   ", col_red("I am red"), "   "))
#' trimws(col_red("   I am red   "))
#' ansi_trimws(col_red("   I am red   "))

ansi_trimws <- function(x, which = c("both", "left", "right")) {

  if (!is.character(x)) x <- as.character(x)
  which <- match.arg(which)
  x <- enc2utf8(x)
  if (!length(x)) return(ansi_string(x))

  sl <- 0L
  if (which %in% c("both", "left")) {
    xs <- ansi_strip(x)
    xl <- trimws(xs, "left")
    nxs <- nchar(xs)
    sl <- nxs - nchar(xl)
  }

  rl <- 0L
  if (which %in% c("both", "right")) {
    xs <- ansi_strip(x)
    xr <- trimws(xs, "right")
    nxs <- nchar(xs)
    rl <- nxs - nchar(xr)
  }

  if (any(sl > 0L | rl > 0L)) {
    start <- rep_len(1L + sl, length(x))
    x <- .Call(clic_ansi_substr, x, start, ansi_nchar(x) - rl)
  }

  ansi_string(x)
}

#' Wrap an ANSI styled string to a certain width
#'
#' This function is similar to [base::strwrap()], but works on ANSI
#' styled strings, and leaves the styling intact.
#'
#' @param x ANSI string.
#' @param width Width to wrap to.
#' @param indent Indentation of the first line of each paragraph.
#' @param exdent Indentation of the subsequent lines of each paragraph.
#' @param simplify Whether to return all wrapped strings in a single
#'   character vector, or wrap each element of `x` independently and return
#'   a list.
#' @return If `simplify` is `FALSE`, then a list of character vectors,
#'   each an ANSI string. Otherwise a single ANSI string vector.
#'
#' @family ANSI string operations
#' @export
#' @examples
#' text <- cli:::lorem_ipsum()
#' # Highlight some words, that start with 's'
#' rexp <- gregexpr("\\b([sS][a-zA-Z]+)\\b", text)
#' regmatches(text, rexp) <- lapply(regmatches(text, rexp), col_red)
#' cat(text)
#'
#' wrp <- ansi_strwrap(text, width = 40)
#' cat(wrp, sep = "\n")

ansi_strwrap <- function(x, width = console_width(), indent = 0,
                         exdent = 0, simplify = TRUE) {

  if (!is.character(x)) x <- as.character(x)
  x <- enc2utf8(x)
  if (length(x) == 0) {
    return(ansi_string(x))
  }
  if (length(x) > 1) {
    wrp <- lapply(x, ansi_strwrap, width = width, indent = indent,
                  exdent = exdent, simplify = FALSE)
    if (simplify) wrp <- ansi_string(unlist(wrp))
    return(wrp)
  }

  # Workaround for bad Unicode width
  x <- unicode_pre(x)

  # Form feeds are forced line breaks
  # R 4.2 removes the \f after <https://github.com/wch/r-source/commit/101b142d04dd5456a2039d54de9483240bcc1512>
  # se we need to put in a random marker instead
  mark <- "yShtnpteEk"
  smark <- paste0("\n\n", mark, "\n\n")
  x <- gsub_("\f", smark, x, fixed = TRUE, useBytes = TRUE)
  fix_ff <- function(x) {
    xs <- ansi_strip(x)
    rem <- which(xs == mark)
    if (length(rem)) {
      x <- x[-c(rem - 1, rem + 1)]
      xs <- xs[-c(rem - 1, rem + 1)]
      if (xs[length(xs)] == mark) {
        x <- c(x, mark)
        xs <- c(xs, mark)
      }
      if (length(x) >= 2 && x[1] == "" && xs[2] == mark) {
        x <- x[-1]
        xs <- xs[-1]
      }
      # At this point, we have as many marks as many newlines we need
      # But (except for the begnning) we need one less empty lines than
      # newlines, because an empty line corresponds to two newlines at
      # the end of a non-empty line.
      del <- which(xs[-1] == mark & xs[-length(xs)] != mark) + 1L
      if (length(del) > 0) {
        x <- x[-del]
        xs <- xs[-del]
      }
      x[xs == mark] <- ""
      x
    } else {
      x
    }

  }

  # First we need to remove the multiple spaces, to make it easier to
  # map the strings later on. We do this per paragraph, to keep paragraphs.
  pars <- strsplit(x, "\n[ \t\n]*\n", perl = TRUE)
  pars <- lapply(pars, ansi_trimws)

  # Within paragraphs, replace multiple spaces with one, except when there
  # were two spaces at the end of a sentence, where we keep two.
  # This does not work well, when some space is inside an ANSI tag, and
  # some is outside, but for now, we'll live with this limitation.
  pars <- lapply(pars, function(s) {
    # First replace multiple spaces that are not at the end of a sentence
    s <- gsub("(?<![.!?])[ \t\n][ \t\n]*", " ", s, perl = TRUE)
    # Handle multiple spaces at the end of a sentence
    s <- gsub("(?<=[.!?])[ \t\n][ \t\n][ \t\n]*", "  ", s, perl = TRUE)
    # Handle simple space at the end of a sentence
    gsub("(?<=[.!?])[ \t\n]", " ", s, perl = TRUE)
  })

  # Put them back together
  xx <- vcapply(pars, function(s) paste(s, collapse = "\n\n"))

  xs <- ansi_strip(xx)
  xw0 <- base::strwrap(xs, width = width, indent = indent, exdent = exdent)
  if (xs == xx) return(ansi_string(unicode_post(fix_ff(xw0))))

  xw <- trimws(xw0, "left")
  indent <- nchar(xw0) - nchar(xw)

  # Now map the positions from xw back to xs by going over both in parallel
  splits <- 1L
  drop <- integer()
  xslen <- nchar(xs)
  xsidx <- 1L
  xwlen <- nchar(xw[1])
  xwidx <- c(1L, 1L)

  while (xsidx <= xslen) {
    xsc <- substr(xs, xsidx, xsidx)
    xwc <- substr(xw[xwidx[1]], xwidx[2], xwidx[2])
    if (is.na(xwc)) {
      # colored trailing white space in input?
      xsidx <- xsidx + 1L
    } else if (xsc == xwc) {
      xsidx <- xsidx + 1L
      xwidx[2] <- xwidx[2] + 1L
    } else if (xsc %in% c(" ", "\n", "\t")) {
      drop <- c(drop, xsidx)
      xsidx <- xsidx + 1L
    } else if (xwc == " ") {
      xwidx[2] <- xwidx[2] + 1L
    } else {
      throw(cli_error("Internal error in {.fun cli::ansi_strwrap}")) # nocov
    }

    while (xsidx <= xslen && xwidx[1] <= length(xw) && xwidx[2] > xwlen) {
      splits <- c(splits, xsidx)
      xwidx[1] <- xwidx[1] + 1L
      xwidx[2] <- 1L
      xwlen <- nchar(xw[xwidx[1]])
    }
  }
  splits <- c(splits, xsidx)

  wrp <- vcapply(seq_along(splits[-1]), function(i) {
    from <- splits[i]
    to <- splits[i + 1L] - 1L
    while (from %in% drop) from <- from + 1L
    .Call(clic_ansi_substr, xx, from, to)
  })

  indent <- strrep(" ", indent)
  ansi_string(unicode_post(fix_ff(paste0(indent, wrp))))
}

#' Truncate an ANSI string
#'
#' This function is similar to [base::strtrim()], but works correctly with
#' ANSI styled strings. It also adds `...` (or the corresponding Unicode
#' character if Unicode characters are allowed) to the end of truncated
#' strings.
#'
#' Note: `ansi_strtrim()` does not support NA values currently.
#'
#' @param x Character vector of ANSI strings.
#' @param width The width to truncate to.
#' @param ellipsis The string to append to truncated strings. Supply an
#'   empty string if you don't want a marker.
#'
#' @family ANSI string operations
#' @export
#' @examples
#' text <- cli::col_red(cli:::lorem_ipsum())
#' ansi_strtrim(c(text, "foobar"), 40)

ansi_strtrim <- function(x, width = console_width(),
                         ellipsis = symbol$ellipsis) {

  if (width < 0) {
    throw(cli_error(
      "{.arg width} must be non-negative in {.fun cli::ansi_strtrim}."
    ))
  }

  x <- enc2utf8(x)

  # Unicode width notes. We have nothing to fix here, because we'll just
  # use ansi_substr() and ansi_nchar(), which work correctly with wide
  # characters.

  # if ellipsis is already longer than width, then we just return that
  tw <- ansi_nchar(ellipsis, "width")
  if (tw == width) {
    x[] <- ellipsis
    return(x)
  } else if (tw > width) {
    x[] <- ansi_strtrim(ellipsis, width, ellipsis = "")
    return(x)
  }

  # First we cut according to _characters_. This might be too wide if we
  # have wide characters.
  lx <- length(x)
  xt <- .Call(clic_ansi_substr, x, rep(1L, lx), rep(as.integer(width), lx))

  # If there was a cut, or xt is too wide (using _width_!), that's bad
  # We keep the initial bad ones, these are the ones that need an ellipsis.
  # Then we keep chopping off single characters from the too wide ones,
  # until they are narrow enough.
  if (ansi_nzchar(ellipsis)) {
    bad0 <- bad <- !is.na(x) &
      (ansi_strip(xt) != ansi_strip(x) | ansi_nchar(xt, "width") > width)
  } else {
    # if ellipsis is zero length, then the truncated ones are not bad
    bad0 <- bad <- !is.na(x) & ansi_nchar(xt, "width") > width
  }

  while (any(bad)) {
    xt[bad] <- .Call(
      clic_ansi_substr,
      xt[bad],
      rep(1L, sum(bad)),
      ansi_nchar(xt[bad]) - 1L
    )
    bad <- ansi_nchar(xt, "width") > width - tw
  }

  xt[bad0] <- paste0(xt[bad0], ellipsis)
  xt
}

#' Format a character vector in multiple columns
#'
#' This function helps with multi-column output of ANSI styles strings.
#' It works well together with [boxx()], see the example below.
#'
#' If a string does not fit into the specified `width`, it will be
#' truncated using [ansi_strtrim()].
#'
#' ```{asciicast ansi-column}
#' fmt <- ansi_columns(
#'   paste(col_red("foo"), 1:10),
#'   width = 50,
#'   fill = "rows",
#'   max_cols=10,
#'   align = "center",
#'   sep = "   "
#' )
#' boxx(fmt, padding = c(0,1,0,1), header = col_cyan("Columns"))
#' ```
#'
#' @param text Character vector to format. Each element will formatted
#'   as a cell of a table.
#' @param width Width of the screen.
#' @param sep Separator between the columns. It may have ANSI styles.
#' @param fill Whether to fill the columns row-wise or column-wise.
#' @param max_cols Maximum number of columns to use. Will not use more,
#'   even if there is space for it.
#' @param align Alignment within the columns.
#' @param type Passed to [ansi_nchar()] and [ansi_align()]. Most probably
#'   you want the default, `"width"`.
#' @inheritParams ansi_strtrim
#' @return ANSI string vector.
#'
#' @family ANSI string operations
#' @export

ansi_columns <- function(text, width = console_width(), sep = " ",
                         fill = c("rows", "cols"), max_cols = 4,
                         align = c("left", "center", "right"),
                         type = "width", ellipsis = symbol$ellipsis) {

  fill <- match.arg(fill)
  align <- match.arg(align)

  text <- enc2utf8(text)

  if (length(text) == 0) return(ansi_string(text))

  swdh <- ansi_nchar(sep, type = "width")
  twdh <- max(ansi_nchar(text, type = type)) + swdh
  cols <- min(floor(width / twdh), max_cols)
  if (cols == 0) {
    cols <- 1
    text <- ansi_strtrim(text, width = width, ellipsis = ellipsis)
  }

  len <- length(text)
  extra <- ceiling(len / cols) * cols - len
  text <- c(text, rep("", extra))
  tm <- matrix(text, byrow = fill == "rows", ncol = cols)

  colwdh <- diff(c(0, round((width / cols)  * (1:cols))))
  for (c in seq_len(ncol(tm))) {
    tm[, c] <- ansi_align(
      paste0(tm[, c], if (cols > 1) sep),
      colwdh[c],
      align = align,
      type = type
    )
  }

  clp <- apply(tm, 1, paste0, collapse = "")
  ansi_string(clp)
}

#' ANSI character translation and case folding
#'
#' There functions are similar to [toupper()], [tolower()] and
#' [chartr()], but they keep the ANSI colors of the string.
#'
#' @inheritParams base::chartr
#' @param x Input string. May have ANSI colors and styles.
#' @return Character vector of the same length as `x`, containing
#'   the translated strings. ANSI styles are retained.
#'
#' @family ANSI string operations
#' @export
#' @examples
#' ansi_toupper(col_red("Uppercase"))
#'
#' ansi_tolower(col_red("LowerCase"))
#'
#' x <- paste0(col_green("MiXeD"), col_red(" cAsE 123"))
#' ansi_chartr("iXs", "why", x)

ansi_toupper <- function(x) {
  ansi_convert(x, toupper)
}

#' @family ANSI string operations
#' @export
#' @rdname ansi_toupper

ansi_tolower <- function(x) {
  ansi_convert(x, tolower)
}

#' @family ANSI string operations
#' @export
#' @rdname ansi_toupper

ansi_chartr <- function(old, new, x) {
  ansi_convert(x, chartr, old, new)
}

ansi_convert <- function(x, converter, ...) {
  x <- enc2utf8(x)
  ansi <- re_table(ansi_regex(), x)
  text <- non_matching(ansi, x, empty=TRUE)
  out <- mapply(x, text, USE.NAMES = FALSE, FUN = function(x1, t1) {
    t1 <- t1[t1[,1] <= t1[,2], , drop = FALSE]
    for (i in seq_len(nrow(t1))) {
      substring(x1, t1[i, 1], t1[i, 2]) <-
        converter(x = substring(x1, t1[i, 1], t1[i, 2]), ...)
    }
    x1
  })

  ansi_string(out)
}

#' Simplify ANSI styling tags
#'
#' It creates an equivalent, but possibly shorter ANSI styled string, by
#' removing duplicate and empty tags.
#'
#' @param x Input string
#' @param csi What to do with non-SGR ANSI sequences, either `"keep"`,
#'   or `"drop"` them.
#' @return Simplified `cli_ansi_string` vector.
#'
#' @export

ansi_simplify <- function(x, csi = c("keep", "drop")) {
  if (!is.character(x)) x <- as.character(x)
  csi <- match.arg(csi)
  x <- enc2utf8(x)
  .Call(clic_ansi_simplify, x, csi == "keep")
}

#' Convert ANSI styled text to HTML
#'
#' @param x Input character vector.
#' @param escape_reserved Whether to escape characters that are reserved
#'   in HTML (`&`, `<` and `>`).
#' @param csi What to do with non-SGR ANSI sequences, either `"keep"`,
#'   or `"drop"` them.
#' @return Character vector of HTML.
#'
#' @family ANSI to HTML conversion
#' @export
#' @examplesIf cli:::has_packages(c("htmltools", "withr"))
#' ## Syntax highlight the source code of an R function with ANSI tags,
#' ## and export it to a HTML file.
#' code <- withr::with_options(
#'   list(ansi.num_colors = 256),
#'   code_highlight(format(ansi_html))
#' )
#' hcode <- paste(ansi_html(code), collapse = "\n")
#' css <- paste(format(ansi_html_style()), collapse=  "\n")
#' page <- htmltools::tagList(
#'   htmltools::tags$head(htmltools::tags$style(css)),
#'   htmltools::tags$pre(htmltools::HTML(hcode))
#' )
#'
#' if (interactive()) htmltools::html_print(page)

ansi_html <- function(x, escape_reserved = TRUE, csi = c("drop", "keep")) {
  if (!is.character(x)) x <- as.character(x)
  csi <- match.arg(csi)
  x <- enc2utf8(x)
  if (escape_reserved) {
    x <- gsub_("&", "&amp;", x, fixed = TRUE, useBytes = TRUE)
    x <- gsub_("<", "&lt;",  x, fixed = TRUE, useBytes = TRUE)
    x <- gsub_(">", "&gt;",  x, fixed = TRUE, useBytes = TRUE)
  }
  .Call(clic_ansi_html, x, csi == "keep")
}

#' CSS styles for the output of `ansi_html()`
#'
#'
#'
#' @param colors Whether or not to include colors. `FALSE` will not include
#'   colors, `TRUE` or `8` will include eight colors (plus their bright
#'   variants), `256` will include 256 colors.
#' @param palette Character scalar, palette to use for the first eight colors
#'   plus their bright variants. Terminals define these colors differently,
#'   and cli includes a couple of examples. Sources of palettes:
#'   * https://en.wikipedia.org/wiki/ANSI_escape_code#3-bit_and_4-bit
#'   * iTerm2 builtin palettes
#'   * <https://github.com/sindresorhus/iterm2-snazzy>
#' @return Named list of CSS declaration blocks, where the names are
#'   CSS selectors. It has a `format()` and `print()` methods, which you
#'   can use to write the output to a CSS or HTML file.
#'
#' @family ANSI to HTML conversion
#' @export
#' @examples
#' ansi_html_style(colors = FALSE)
#' ansi_html_style(colors = 8, palette = "iterm-snazzy")

ansi_html_style <- function(colors = TRUE, palette = NULL) {
  if (is.character(palette)) {
    palette <- match.arg(palette)
    palette <- as.list(ansi_palettes[palette, ])
  }

  stopifnot(
    isTRUE(colors) || identical(colors, FALSE) ||
      (is_count(colors) && colors %in% c(8,256)),
    is_string(palette) || is.list(palette) && length(palette) == 16
  )

  ret <- list(
    ".ansi-bold"       = "{ font-weight: bold;             }",
    # .ansi-faint ???
    ".ansi-italic"     = "{ font-style: italic;            }",
    ".ansi-underline"  = "{ text-decoration: underline;    }",
    ".ansi-blink"      = "{ text-decoration: blink;        }",
    # .ansi-inverse ???
    ".ansi-hide"       = "{ visibility: hidden;            }",
    ".ansi-crossedout" = "{ text-decoration: line-through; }",
    ".ansi-link:hover" = "{ text-decoration: underline;    }"
  )

  if (!identical(colors, FALSE)) {
    fg <- structure(
      names = paste0(".ansi-color-", 0:15),
      paste0("{ color: ", palette, " }")
    )
    bg <- structure(
      names = paste0(".ansi-bg-color-", 0:15),
      paste0("{ background-color: ", palette, " }")
    )
    ret <- c(ret, fg, bg)
  }

  if (isTRUE(colors) || colors == 256) {
    grid <- expand.grid(r = 0:5, g = 0:5, b = 0:5)
    num <- 16 + 36 * grid$r + 6 * grid$g + grid$b
    cols <- grDevices::rgb(grid$r, grid$g, grid$b, maxColorValue = 5)
    fg256 <- structure(
      names = paste0(".ansi-color-", num),
      paste0("{ color: ", tolower(cols), " }")
    )
    bg256 <- structure(
      names = paste0(".ansi-bg-color-", num),
      paste0("{ background-color: ", tolower(cols), " }")
    )
    gr <- seq(1, 24)
    grcols <- grDevices::rgb(gr, gr, gr, maxColorValue = 25)
    fggrey <- structure(
      names = paste0(".ansi-color-", 232:255),
      paste0("{ color: ", tolower(grcols), " }")
    )
    bggrey <- structure(
      names = paste0(".ansi-bg-color-", 232:255),
      paste0("{ background-color: ", tolower(grcols), " }")
    )
    ret <- c(ret, fg256, fggrey, bg256, bggrey)
  }

  class(ret) <- "cli_ansi_html_style"
  ret
}

# This avoids duplication, but messes up the source ref of the function...
formals(ansi_html_style)$palette <- c("vscode", setdiff(rownames(ansi_palettes), "vscode"))
attr(body(ansi_html_style), "srcref") <- NULL
attr(body(ansi_html_style), "wholeSrcref") <- NULL
attr(body(ansi_html_style), "srcfile") <- NULL

#' @export

format.cli_ansi_html_style <- function(x, ...) {
  paste0(format(names(x)), " ", x)
}

#' @export

print.cli_ansi_html_style <- function(x, ...) {
  cat(format(x, ...), sep = "\n")
}

#' Like [base::grep()] and [base::grepl()], but for ANSI strings
#'
#' First ANSI sequences will be stripped with [ansi_strip()], both
#'
#' Note that these functions work on code points (or bytes if
#' `useBytes = TRUE`), and not graphemes.
#'
#' Unlike [base::grep()] and [base::grepl()] these functions do not special
#' case factors.
#'
#' Both `pattern` and `x` are converted to UTF-8.
#'
#' @param pattern Character scalar, regular expression or fixed string
#'   (if `fixed = TRUE`), the pattern to search for. Other objects will be
#'   coerced using [as.character()].
#' @param x Character vector to search in. Other objects will be coerced
#'   using [as.character()].
#' @param ignore.case,perl,value Passed to [base::grep()].
#' @param ... Extra arguments are passed to [base::grep()] or [base::grepl()].
#' @return The same as [base::grep()] and [base::grepl()], respectively.
#'
#' @export
#' @examples
#' red_needle <- col_red("needle")
#' haystack <- c("foo", "needle", "foo")
#' green_haystack <- col_green(haystack)
#' ansi_grepl(red_needle, haystack)
#' ansi_grepl(red_needle, green_haystack)

ansi_grep <- function(pattern, x, ignore.case = FALSE, perl = FALSE,
                      value = FALSE, ...) {

  # if value = FALSE, then we want to return the original values as
  # ansi strings, so we need to special case that
  if (value) {
    idx <- ansi_grep(pattern, x, ignore.case = ignore.case, perl = perl,
                     value = FALSE, ...)
    ansi_string(x[idx])
  } else {
    ansi_grep_internal(grep, pattern, x, ignore.case = ignore.case,
                       perl = perl, value = value, ...)
  }
}

#' @rdname ansi_grep
#' @export

ansi_grepl <- function(pattern, x, ...) {
  ansi_grep_internal(grepl, pattern, x, ...)
}

ansi_grep_internal <- function(fun, pattern, x, ...) {
  pattern <- ansi_strip(pattern)
  x <- ansi_strip(x)
  fun(pattern, x, ...)
}

#' Like [base::nzchar()], but for ANSI strings
#'
#' @param x Charcater vector. Other objects are coarced using
#'   [base::as.character()].
#' @param ... Passed to [base::nzchar()].
#' @export
#' @examples
#' ansi_nzchar("")
#' ansi_nzchar(col_red(""))

ansi_nzchar <- function(x, ...) {
  x <- ansi_strip(x)
  nzchar(x, ...)
}

test_that("cli_ansi_string", {
  right <- c("cli_ansi_string", "ansi_string", "character")
  expect_equal(class(ansi_string("foobar")), right)
  expect_equal(class(ansi_string(133)), right)
  expect_equal(class(ansi_string(ansi_string(134))), right)
})

test_that("ansi_regex", {
  # somewhat special sequences
  cases <- list(
    list("foo\033[39;49mbar", "foobar"),
    list("foo\033[1;3;4mbar", "foobar"),
    list("foo\033[1;;4mbar", "foobar"),
    list("foo\033[mbar", "foobar")
  )

  for (case in cases) {
    expect_equal(gsub(ansi_regex(), "", case[[1]], perl = TRUE), case[[2]])
  }
})

test_that("ansi_has_any works", {
  withr::local_options(list(
    cli.num_colors = 256L,
    cli.hyperlink = TRUE
  ))
  expect_false(ansi_has_any("foobar"))
  funcs <- ls(asNamespace("cli"), pattern = "^col_|^bg_|^style_")

  for (sym in setdiff(funcs, "style_hyperlink")) {
    fun <- get(sym, envir = asNamespace("cli"))
    expect_true(ansi_has_any(fun("foo", "bar")), label = sym)
  }
})

test_that("ansi_strip works", {
  withr::local_options(list(
    cli.num_colors = 256L,
    cli.hyperlink = TRUE
  ))
  expect_equal("", ansi_strip(""))
  expect_equal("foobar", ansi_strip("foobar"))
  expect_equal(
    "foobar",
    ansi_strip(col_red(style_underline(style_bold("foobar"))))
  )

  funcs <- ls(asNamespace("cli"), pattern = "^col_|^bg_|^style_")
  for (sym in setdiff(funcs, "style_hyperlink")) {
    fun <- get(sym, envir = asNamespace("cli"))
    ans <- if (sym == "style_hyperlink") "foo" else "foobar"
    expect_equal(ans, ansi_strip(fun("foo", "bar")))
  }
})

str <- c("",
         "plain",
         "\033[31m",
         "\033[39m",
         "\033[31mred\033[39m",
         "\033[31mred\033[39m\033[31mred\033[39m",
         "foo\033[31mred\033[39m",
         "\033[31mred\033[39mfoo")

test_that("ansi_nchar", {
  withr::local_options(list(cli.num_colors = 256L))
  for (s in str) {
    expect_equal(ansi_nchar(s), nchar(ansi_strip(s)), info = s)
  }
})

test_that("ansi_nchar wide characters", {
  expect_equal(ansi_nchar("\u231a", "width"), 2L)
})

test_that("ansi_substr bad input", {
  expect_snapshot(
    error = TRUE,
    ansi_substr("foobar", NULL, 10)
  )
  expect_snapshot(
    error = TRUE,
    ansi_substr("foobar", 10, NULL)
  )
  expect_snapshot(
    error = TRUE,
    ansi_substr("foobar", "bad", "bad")
  )
})

test_that("ansi_substr", {
  withr::local_options(list(cli.num_colors = 256L))
  for (s in str) {
    for (i in 1 %:% ansi_nchar(s)) {
      for (j in i %:% ansi_nchar(s)) {
        expect_equal(ansi_strip(ansi_substr(s, i, j)),
                     substr(ansi_strip(s), i, j), info = paste(s, i, j))
      }
    }
  }
})

test_that("ansi_substr keeps color", {
  withr::local_options(list(cli.num_colors = 256L))
  expect_equal(
    ansi_substr("\033[31mred\033[39m", 1, 1),
    ansi_string("\033[31mr\033[39m")
  )
  expect_equal(
    ansi_substr("foo\033[31mred\033[39m", 4, 4),
    ansi_string("\033[31mr\033[39m")
  )
  expect_equal(
    ansi_substr("foo\033[31mred\033[39mbar", 4, 4),
    ansi_string("\033[31mr\033[39m")
  )
  expect_equal(
    ansi_substr("\033[31mred\033[39mfoo\033[31mred\033[39mbar", 7, 7),
    ansi_string("\033[31mr\033[39m")
  )
})

test_that("ansi_substr, start after string end", {
  withr::local_options(list(cli.num_colors = 256L))
  expect_equal(ansi_substr("red", 4, 4), ansi_string(""))
  expect_equal(ansi_substr("red", 4, 5), ansi_string(""))
  expect_equal(ansi_strip(ansi_substr("\033[31mred\033[39m", 4, 4)), "")
  expect_equal(ansi_strip(ansi_substr("\033[31mred\033[39m", 4, 5)), "")

  expect_equal(ansi_substr("red", 3, 4), ansi_string("d"))
  expect_equal(ansi_substr("red", 3, 5), ansi_string("d"))
  expect_equal(ansi_strip(ansi_substr("\033[31mred\033[39m", 3, 4)), "d")
  expect_equal(ansi_strip(ansi_substr("\033[31mred\033[39m", 3, 5)), "d")
})

test_that("ansi_substr, multiple strings", {
  withr::local_options(list(cli.num_colors = 256L))
  set.seed(42)
  for (i in 1:100) {
    strs <- sample(str, 4)
    num_starts <- sample(1:5, 1)
    num_stops <- sample(1:5, 1)
    starts <- sample(1:5, num_starts, replace = TRUE)
    stops <- sample(1:5, num_stops, replace = TRUE)
    r1 <- ansi_strip(ansi_substr(strs, starts, stops))
    r2 <- substr(ansi_strip(strs), starts, stops)
    expect_equal(r1, r2)
  }
})

test_that("ansi_substr corner cases", {
  withr::local_options(list(cli.num_colors = 256L))
  # Zero length input

  c0 <- character(0L)
  o0 <- structure(list(), class="abc")
  co0 <- structure(character(0L), class="abc")
  expect_identical(ansi_substr(c0, 1, 1), ansi_string(substr(c0, 1, 1)))
  expect_identical(ansi_substr(o0, 1, 1), ansi_string(substr(o0, 1, 1)))
  expect_identical(ansi_substr(co0, 1, 1), ansi_string(substr(co0, 1, 1)))

  expect_identical(
    ansi_substring(c0, 1, 1),
    ansi_string(substring(c0, 1, 1))
  )
  expect_identical(
    ansi_substring(o0, 1, 1),
    ansi_string(substring(o0, 1, 1))
  )
  expect_identical(
    ansi_substring(co0, 1, 1),
    ansi_string(substring(co0, 1, 1))
  )

  # Character start/stop
  expect_identical(
    ansi_substr("abc", "1", 1),
    ansi_string(substr("abc", "1", 1))
  )
  expect_identical(
    ansi_substr("abc", 1, "1"),
    ansi_string(substr("abc", 1, "1"))
  )

  # non-numeric arguments cause errors; NOTE: this actually "works"
  # with 'substr' but not implemented in 'ansi_substr'
  expect_snapshot(
    error = TRUE,
    ansi_substr("abc", "hello", 1)
  )
})

test_that("ansi_substring", {
  withr::local_options(list(cli.num_colors = 256L))
  for (s in str) {
    for (i in 1 %:% ansi_nchar(s)) {
      for (j in i %:% ansi_nchar(s)) {
        expect_equal(ansi_strip(ansi_substring(s, i, j)),
                     substring(ansi_strip(s), i, j), info = paste(s, i, j))
      }
    }
  }
})

test_that("ansi_substring, multiple strings", {
  withr::local_options(list(cli.num_colors = 256L))
  set.seed(42)
  for (i in 1:100) {
    strs <- sample(str, 4)
    num_starts <- sample(1:5, 1)
    num_stops <- sample(1:5, 1)
    starts <- sample(1:5, num_starts, replace = TRUE)
    stops <- sample(1:5, num_stops, replace = TRUE)
    r1 <- ansi_strip(ansi_substring(strs, starts, stops))
    r2 <- substring(ansi_strip(strs), starts, stops)
    expect_equal(r1, r2)
  }
})

test_that("ansi_substring corner cases", {
  withr::local_options(list(cli.num_colors = 256L))
  # Zero length input

  c0 <- character(0L)
  o0 <- structure(list(), class="abc")
  co0 <- structure(character(0L), class="abc")
  expect_identical(
    ansi_substring(c0, 1, 1),
    ansi_string(substring(c0, 1, 1))
  )
  expect_identical(
    ansi_substring(o0, 1, 1),
    ansi_string(substring(o0, 1, 1))
  )
  expect_identical(
    ansi_substring(co0, 1, 1),
    ansi_string(substring(co0, 1, 1))
  )
})

test_that("ansi_strsplit", {
  withr::local_options(list(cli.num_colors = 256L))
  red <- "\033[31mred\033[39m"

  str <- "plain-plain"
  expect_equal(
    ansi_strsplit(str, "-"),
    lapply(strsplit(str, "-"), ansi_string)
  )

  str <- paste0(red, "-plain")
  expect_equal(
    ansi_strip(ansi_strsplit(str, "-")[[1]]),
    strsplit(ansi_strip(str), "-")[[1]]
  )

  expect_equal(
    ansi_strsplit(str, "e"),
    list(ansi_string(c("\033[31mr\033[39m", "\033[31md\033[39m-plain")))
  )

  str <- paste0(red, "-", red, "-", red)
  expect_equal(
    ansi_strip(ansi_strsplit(str, "-")[[1]]),
    strsplit(ansi_strip(str), "-")[[1]]
  )

  # with leading and trailing separators
  str.2 <- paste0("-", red, "-", red, "-", red, "-")
  expect_equal(ansi_strip(ansi_strsplit(str.2, "-")[[1]]),
               strsplit(ansi_strip(str.2), "-")[[1]])

  # greater than length 1
  str.3 <- paste0("-", c(col_green("hello"), col_red("goodbye")), "-world-")
  expect_equal(ansi_strip(unlist(ansi_strsplit(str.3, "-"))),
               unlist(strsplit(ansi_strip(str.3), "-")))
})

test_that("ansi_strsplit multiple strings", {
  withr::local_options(list(cli.num_colors = 256L))
  red <- "\033[31mred\033[39m"
  str <- c(
    paste0("plain-plain-", red, "-plain-", red),
    paste0(red, "-", red),
    red
  )

  r1 <- lapply(ansi_strsplit(str, "-"), ansi_strip)
  r2 <- strsplit(ansi_strip(str), "-")
  expect_equal(r1, r2)
})

test_that("ansi_strsplit edge cases", {
  withr::local_options(list(cli.num_colors = 256L))
  expect_equal(ansi_strsplit("", "-"), list(ansi_string(character(0L))))
  expect_equal(
    ansi_strip(ansi_strsplit("\033[31m\033[39m", "-")[[1]]), character(0L)
  )

  # special cases
  expect_equal(ansi_strsplit("", ""), lapply(strsplit("", ""), ansi_string))
  expect_equal(
    ansi_strsplit("a", "a"),
    lapply(strsplit("a", "a"), ansi_string)
  )

  # this following test isn't working yet
  expect_equal(
    ansi_strsplit("a", ""),
    lapply(strsplit("a", ""), ansi_string)
  )
  expect_equal(
    ansi_strsplit("", "a"),
    lapply(strsplit("", "a"), ansi_string)
  )

  # Longer strings
  expect_identical(
    ansi_strsplit(c("", "a", "aa"), "a"),
    lapply(strsplit(c("", "a", "aa"), "a"), ansi_string)
  )
  expect_identical(
    ansi_strsplit(c("abaa", "ababza"), "b."),
    lapply(strsplit(c("abaa", "ababza"), "b."), ansi_string)
  )
})

test_that("Weird length 'split'", {
  withr::local_options(list(cli.num_colors = 256L))
  expect_snapshot(
    error = TRUE,
    ansi_strsplit(c("ab", "bd"), c("b", "d"))
  )
  expect_identical(
    ansi_strsplit("ab", NULL),
    lapply(strsplit("ab", NULL), ansi_string)
  )
  expect_identical(
    ansi_strsplit("ab", character(0L)),
    lapply(strsplit("ab", character(0L)), ansi_string)
  )
})

test_that("ansi_align", {
  withr::local_options(list(cli.num_colors = 256L))
  expect_equal(ansi_align(character()), ansi_string(character()))
  expect_equal(ansi_align("", 0), ansi_string(""))
  expect_equal(ansi_align(" ", 0), ansi_string(" "))
  expect_equal(ansi_align(" ", 1), ansi_string(" "))
  expect_equal(ansi_align(" ", 2), ansi_string("  "))
  expect_equal(ansi_align("a", 1), ansi_string("a"))
  expect_equal(ansi_align(letters, 1), ansi_string(letters))
  expect_equal(ansi_align(letters, 0), ansi_string(letters))
  expect_equal(ansi_align(letters, -1), ansi_string(letters))

  expect_equal(ansi_align(letters, 2), ansi_string(paste0(letters, " ")))
  expect_equal(
    ansi_align(letters, 3, "center"),
    ansi_string(paste0(" ", letters, " "))
  )
  expect_equal(
    ansi_align(letters, 2, "right"),
    ansi_string(paste0(" ", letters))
  )

  expect_equal(
    ansi_align(c("foo", "foobar", "", "a"), 6, "left"),
    ansi_string(c("foo   ", "foobar", "      ", "a     ")))

  expect_equal(
    ansi_align(c("foo", "foobar", "", "a"), 6, "center"),
    ansi_string(c("  foo ", "foobar", "      ", "   a  ")))

  expect_equal(
    ansi_align(c("foo", "foobar", "", "a"), 6, "right"),
    ansi_string(c("   foo", "foobar", "      ", "     a")))

  # #54: alignment of wide characters
  expect_equal(
    ansi_align(c("foo", "\u6210\u4ea4\u65e5", "", "a"), 6, "left"),
    ansi_string(c("foo   ", "\u6210\u4ea4\u65e5", "      ", "a     ")))

  expect_equal(
    ansi_align(c("foo", "\u6210\u4ea4\u65e5", "", "a"), 6, "center"),
    ansi_string(c("  foo ", "\u6210\u4ea4\u65e5", "      ", "   a  ")))

  expect_equal(
    ansi_align(c("foo", "\u6210\u4ea4\u65e5", "", "a"), 6, "right"),
    ansi_string(c("   foo", "\u6210\u4ea4\u65e5", "      ", "     a")))
})

test_that("stripping hyperlinks", {
  withr::local_options(list(cli.hyperlink = TRUE))
  x <- unclass(style_hyperlink("foo", "https://r-pkg.org"))
  expect_equal(
    ansi_strip(paste0("1111-", x, "-2222-", x, "-333")),
    "1111-foo-2222-foo-333"
  )
})

test_that("ansi_trimws", {
  cases <- list(
    list(character(), ansi_string(character())),
    list(1, ansi_string(1)),
    list("", ansi_string("")),
    list("foo", ansi_string("foo")),
    list("  foo  ", ansi_string("foo")),
    list(c("foo", "bar"), ansi_string(c("foo", "bar"))),
    list(col_red(c("  colored  ")), ansi_string(col_red("colored"))),
    list(
      paste0("   ", col_red(c("  colored  ")), "   "),
      ansi_string(col_red("colored")))
  )

  for (case in cases) expect_equal(ansi_trimws(case[[1]]), case[[2]])

  cases_left <- list(
    list(character(), ansi_string(character())),
    list(1, ansi_string(1)),
    list("", ansi_string("")),
    list("foo", ansi_string("foo")),
    list("  foo  ", ansi_string("foo  ")),
    list(c("foo", "bar"), ansi_string(c("foo", "bar"))),
    list(col_red(c("  colored  ")), ansi_string(col_red("colored  "))),
    list(
      paste0("   ", col_red(c("  colored  ")), "   "),
      ansi_string(paste0(col_red("colored  "), "   ")))
  )

  for (case in cases_left) {
    expect_equal(ansi_trimws(case[[1]], "left"), case[[2]])
  }

  cases_right <- list(
    list(character(), ansi_string(character())),
    list(1, ansi_string(1)),
    list("", ansi_string("")),
    list("foo", ansi_string("foo")),
    list("  foo  ", ansi_string("  foo")),
    list(c("  foo  ", "  bar  "), ansi_string(c("  foo", "  bar"))),
    list(c("foo", "bar"), ansi_string(c("foo", "bar"))),
    list(col_red(c("  colored  ")), ansi_string(col_red("  colored"))),
    list(
      paste0("   ", col_red(c("  colored  ")), "   "),
      ansi_string(paste0("   ", col_red("  colored"))))
  )

  for (case in cases_right) {
    expect_equal(ansi_trimws(case[[1]], "right"), case[[2]])
  }
})

test_that("ansi_strwrap simple", {
  cases = list(
    list(character(), character()),
    list("", ""),
    list("foo", "foo"),
    list(1, "1"),
    list(c("foo", "bar"), c("foo", "bar"))
  )

  for (case in cases) {
    expect_equal(ansi_strwrap(case[[1]], 20), ansi_string(case[[2]]))
  }
})

test_that("ansi_strwrap simple styled", {
  cases = list(
    list(col_red("foo"), col_red("foo")),
    list(col_red(c("foo", "bar")), col_red(c("foo", "bar"))),
    list(
      paste0("foo", bg_red("   "), "bar"),
      paste0("foo", bg_red(" "), "bar")
    )
  )

  for (case in cases) {
    expect_equal(ansi_strwrap(case[[1]], 20), ansi_string(case[[2]]))
  }
})

test_that("ansi_strwrap", {
  txt0 <- glue::glue_col("
    {col_red Velit occaecat} quis culpa occaecat.  {col_green Pariatur} \\
    ad veniam pariatur {bg_blue consectetur}.  Dolore aliquip et \\
    {style_underline consequat Lorem consectetur} dolor.  Irure id velit \\
    proident elit veniam eu exercitation nisi laboris officia.     Qui \\
    {col_red sunt      occaecat} cillum {col_red sit    commodo sit.    \\
    Culpa} aliquip et consectetur ullamco aliqua Lorem laborum dolore.    ")

  txt <- paste0(txt0, "\n\t  \n", txt0)
  expect_equal(
    ansi_strip(ansi_strwrap(txt0, 40)),
    strwrap(ansi_strip(txt0), 40)
  )
  expect_equal(
    ansi_strip(ansi_strwrap(txt, 40)),
    strwrap(ansi_strip(txt), 40)
  )
  expect_equal(
    ansi_strip(ansi_strwrap(txt, 40, indent = 2)),
    strwrap(ansi_strip(txt), 40, indent = 2)
  )
  expect_equal(
    ansi_strip(ansi_strwrap(txt, 40, exdent = 2)),
    strwrap(ansi_strip(txt), 40, exdent = 2)
  )
  expect_equal(
    ansi_strip(ansi_strwrap(txt, 40, indent = 2, exdent = 4)),
    strwrap(ansi_strip(txt), 40, indent = 2, exdent = 4)
  )
})

test_that("ansi_strwrap double width", {
  expect_equal(
    ansi_strwrap("\U0001F477 1 2 3", 4),
    ansi_string(c("\U0001f477", "1 2", "3"))
  )
})

test_that("ansi_strwrap newlines", {
  expect_equal(
    ansi_strwrap("\033[32mv\033[39m hello world.\nxxx"),
    ansi_string("\033[32mv\033[39m hello world. xxx")
  )
})

test_that("ansi_strwrap and \f edge cases", {
  expect_equal(
    ansi_strwrap("\033[32mfoo\fbar\033[39m"),
    ansi_string(c("\033[32mfoo\033[39m", "\033[32mbar\033[39m"))
  )
  expect_equal(
    ansi_strwrap("\033[32m\ffoo\f\033[39m"),
    ansi_string(c("", "\033[32mfoo\033[39m", ""))
  )
})

test_that_cli(configs = c("plain", "ansi"), "ansi_strtrim", {
  withr::local_options(c(cli.unicode = FALSE))
  setup_unicode_width_fix()
  cases <- list(
    list("", ansi_string("")),
    list("1", ansi_string("1")),
    list("123456789", ansi_string("123456789")),
    list("1234567890", ansi_string("1234567890")),
    list("12345678901", ansi_string("1234567...")),
    list(
      strrep("\u231A", 6),
      ansi_string(paste0(strrep("\u231A", 3), "..."))),
    list(col_red("1"), col_red("1")),
    list(c("foo", NA, col_red("bar")), ansi_string(c("foo", NA, col_red("bar"))))
  )

  for (case in cases) expect_equal(ansi_strtrim(case[[1]], 10), case[[2]])
})

test_that("ansi_strtrim with zero-length ellipsis", {
  expect_snapshot({
    ansi_strtrim("12345", 1, ellipsis = "")
    ansi_strtrim("12345", 3, ellipsis = "")
    ansi_strtrim("12345", 5, ellipsis = "")
  })
})

test_that("ansi_strtrim errors", {
  expect_snapshot(
    error = TRUE,
    ansi_strtrim("foobar", -1)
  )
})

test_that("ansi_strtrim edge cases", {
  expect_snapshot({
    ansi_strtrim("foobar", width = 3, ellipsis = "...")
    ansi_strtrim("foobar", width = 2, ellipsis = "...")
    ansi_strtrim("foobar", width = 1, ellipsis = "...")
    ansi_strtrim("foobar", width = 0, ellipsis = "...")
  })
})

test_that("ansi_columns", {
  withr::local_options(c(cli.unicode = FALSE))
  local_edition(3)

  expect_equal(ansi_columns(character()), ansi_string(character()))

  expect_snapshot_output(
    cat(ansi_columns(paste("foo", 1:10), width = 40), sep = "\n")
  )
  txt60 <- strrep("1234567890", 6)
  expect_snapshot_output(
    cat(ansi_columns(txt60, width = 15), sep = "\n")
  )
})

test_that_cli(configs = c("plain", "ansi"), "ansi_toupper", {
  x <- paste0(
    col_red("Red "),
    "normal ",
    style_bold(col_green("green"))
  )
  expect_snapshot(local({
    cat_line(x)
    cat_line(ansi_toupper(x))
  }))
})

test_that_cli(configs = c("plain", "ansi"), "ansi_tolower", {
  x <- paste0(
    col_red("Red "),
    "NORMAL ",
    style_bold(col_green("grEeN"))
  )
  expect_snapshot(local({
    cat_line(x)
    cat_line(ansi_tolower(x))
  }))
})

test_that_cli(configs = c("plain", "ansi"), "ansi_chartr", {
  x <- paste0(
    col_red("Red "),
    "normal ",
    style_bold(col_green("green"))
  )
  expect_snapshot(local({
    cat_line(x)
    cat_line(ansi_chartr(" R_", "-r*", x))
  }))
})

test_that("very long strings", {
  withr::local_options(cli.num_colors = 256)
  str <- strrep("1234 ", 1000)
  expect_equal(
    ansi_simplify(col_red(col_red(str))),
    col_red(str)
  )

  str2 <- paste0(col_green(str), col_red(str))
  expect_equal(
    ansi_simplify(col_green(str2)),
    ansi_string(str2)
  )
})

test_that("RGB colors", {
  cases <- list(
    list("\033[38;5;123mfoo\033[39m", "\033[38;5;123mfoo\033[39m"),
    list("\033[38;2;1;2;3mfoo\033[39m", "\033[38;2;1;2;3mfoo\033[39m"),
    list("\033[38;2;mfoo\033[39m", "\033[38;2;0;0;0mfoo\033[39m"),
    list("\033[48;5;123mfoo\033[49m", "\033[48;5;123mfoo\033[49m"),
    list("\033[48;2;1;2;3mfoo\033[49m", "\033[48;2;1;2;3mfoo\033[49m"),
    list("\033[48;2;mfoo\033[49m", "\033[48;2;0;0;0mfoo\033[49m"),
    ## bad tags are skipped completely
    list("\033[48;4;12mfoo\033[49m", "foo")
  )

  for (c in cases) {
    expect_equal(ansi_simplify(c[[1]]), ansi_string(c[[2]]))
  }
})

test_that("0m closing tag", {
  cases <- list(
    list("\033[1mfoo\033[0m", "\033[1mfoo\033[22m"),
    list("\033[1mfoo\033[m", "\033[1mfoo\033[22m")
  )

  for (c in cases) {
    expect_equal(ansi_simplify(c[[1]]), ansi_string(c[[2]]))
  }
})

test_that("various tags", {
  cases <- list(
    list("\033[1m\033[22m\033[1mfoo",   "\033[1mfoo\033[22m"),
    list("\033[2m\033[22m\033[2mfoo",   "\033[2mfoo\033[22m"),
    list("\033[3m\033[23m\033[3mfoo",   "\033[3mfoo\033[23m"),
    list("\033[4m\033[24m\033[4mfoo",   "\033[4mfoo\033[24m"),
    list("\033[5m\033[25m\033[5mfoo",   "\033[5mfoo\033[25m"),
    list("\033[7m\033[27m\033[7mfoo",   "\033[7mfoo\033[27m"),
    list("\033[8m\033[28m\033[8mfoo",   "\033[8mfoo\033[28m"),
    list("\033[9m\033[29m\033[9mfoo",   "\033[9mfoo\033[29m"),
    list("\033[30m\033[39m\033[30mfoo", "\033[30mfoo\033[39m"),
    list("\033[31m\033[39m\033[31mfoo", "\033[31mfoo\033[39m"),
    list("\033[32m\033[39m\033[32mfoo", "\033[32mfoo\033[39m"),
    list("\033[33m\033[39m\033[33mfoo", "\033[33mfoo\033[39m"),
    list("\033[34m\033[39m\033[34mfoo", "\033[34mfoo\033[39m"),
    list("\033[35m\033[39m\033[35mfoo", "\033[35mfoo\033[39m"),
    list("\033[36m\033[39m\033[36mfoo", "\033[36mfoo\033[39m"),
    list("\033[37m\033[39m\033[37mfoo", "\033[37mfoo\033[39m"),
    list("\033[90m\033[39m\033[90mfoo", "\033[90mfoo\033[39m"),
    list("\033[91m\033[39m\033[91mfoo", "\033[91mfoo\033[39m"),
    list("\033[92m\033[39m\033[92mfoo", "\033[92mfoo\033[39m"),
    list("\033[93m\033[39m\033[93mfoo", "\033[93mfoo\033[39m"),
    list("\033[94m\033[39m\033[94mfoo", "\033[94mfoo\033[39m"),
    list("\033[95m\033[39m\033[95mfoo", "\033[95mfoo\033[39m"),
    list("\033[96m\033[39m\033[96mfoo", "\033[96mfoo\033[39m"),
    list("\033[97m\033[39m\033[97mfoo", "\033[97mfoo\033[39m"),
    list("\033[40m\033[49m\033[40mfoo", "\033[40mfoo\033[49m"),
    list("\033[41m\033[49m\033[41mfoo", "\033[41mfoo\033[49m"),
    list("\033[42m\033[49m\033[42mfoo", "\033[42mfoo\033[49m"),
    list("\033[43m\033[49m\033[43mfoo", "\033[43mfoo\033[49m"),
    list("\033[44m\033[49m\033[44mfoo", "\033[44mfoo\033[49m"),
    list("\033[45m\033[49m\033[45mfoo", "\033[45mfoo\033[49m"),
    list("\033[46m\033[49m\033[46mfoo", "\033[46mfoo\033[49m"),
    list("\033[47m\033[49m\033[47mfoo", "\033[47mfoo\033[49m"),
    list("\033[100m\033[49m\033[100mfoo", "\033[100mfoo\033[49m"),
    list("\033[101m\033[49m\033[101mfoo", "\033[101mfoo\033[49m"),
    list("\033[102m\033[49m\033[102mfoo", "\033[102mfoo\033[49m"),
    list("\033[103m\033[49m\033[103mfoo", "\033[103mfoo\033[49m"),
    list("\033[104m\033[49m\033[104mfoo", "\033[104mfoo\033[49m"),
    list("\033[105m\033[49m\033[105mfoo", "\033[105mfoo\033[49m"),
    list("\033[106m\033[49m\033[106mfoo", "\033[106mfoo\033[49m"),
    list("\033[107m\033[49m\033[107mfoo", "\033[107mfoo\033[49m")
  )

  for (c in cases) {
    expect_equal(ansi_simplify(c[[1]]), ansi_string(c[[2]]))
  }
})

test_that("unknown tags are kept as is, and [0m is also kept for then", {
  cases <- list(
    list("\033[11mfoo\033[0m", "\033[11mfoo\033[0m")
  )
  for (c in cases) {
    expect_equal(ansi_simplify(c[[1]]), ansi_string(c[[2]]))
  }
})

test_that("simplify w/o tags", {
  cases <- list(
    list(c("", "foo", "\033[1mbold"), c("", "foo", "\033[1mbold\033[22m"))
  )
  for (c in cases) {
    expect_equal(ansi_simplify(c[[1]]), ansi_string(c[[2]]))
  }
})

test_that("CSI sequences", {
  expect_equal(
    ansi_simplify("foo\033[10Abar", csi = "drop"),
    ansi_string("foobar")
  )
  expect_equal(
    ansi_simplify("\033[1mfoo\033[0m\033[10Abar", csi = "drop"),
    ansi_string("\033[1mfoo\033[22mbar")
  )
  expect_equal(
    ansi_simplify("foo\033[10Abar", csi = "keep"),
    ansi_string("foo\033[10Abar")
  )
  expect_equal(
    ansi_simplify("\033[1mfoo\033[0m\033[10Abar", csi = "keep"),
    ansi_string("\033[1mfoo\033[10A\033[22mbar")
  )
  expect_equal(
    ansi_strip("\033[1mfoo\033[10Abar\033[0m", csi = TRUE, sgr = FALSE),
    "\033[1mfoobar\033[0m"
  )
  expect_equal(
    ansi_strip("\033[1mfoo\033[10Abar\033[0m", csi = FALSE, sgr = TRUE),
    "foo\033[10Abar"
  )
})

test_that("ansi_has_any", {
  T <- TRUE
  F <- FALSE
  expect_false(ansi_has_any("foobar",                sgr = T, csi = T))
  expect_true (ansi_has_any("\033[1mfoobar",         sgr = T, csi = T))
  expect_true (ansi_has_any("\033[10Afoobar",        sgr = T, csi = T))
  expect_true (ansi_has_any("\033[10A\033[1mfoobar", sgr = T, csi = T))
  expect_false(ansi_has_any("foobar",                sgr = T, csi = F))
  expect_true (ansi_has_any("\033[1mfoobar",         sgr = T, csi = F))
  expect_false(ansi_has_any("\033[10Afoobar",        sgr = T, csi = F))
  expect_true (ansi_has_any("\033[10A\033[1mfoobar", sgr = T, csi = F))
  expect_false(ansi_has_any("foobar",                sgr = F, csi = T))
  expect_false(ansi_has_any("\033[1mfoobar",         sgr = F, csi = T))
  expect_true (ansi_has_any("\033[10Afoobar",        sgr = F, csi = T))
  expect_true (ansi_has_any("\033[10A\033[1mfoobar", sgr = F, csi = T))
  expect_false(ansi_has_any("foobar",                sgr = F, csi = F))
  expect_false(ansi_has_any("\033[1mfoobar",         sgr = F, csi = F))
  expect_false(ansi_has_any("\033[10Afoobar",        sgr = F, csi = F))
  expect_false(ansi_has_any("\033[10A\033[1mfoobar", sgr = F, csi = F))
})

test_that("NA", {
  T <- TRUE
  F <- FALSE
  s <- c("foo", NA, "bar", "\033[1mfoobar")

  expect_equal(
    ansi_simplify(s),
    ansi_string(c("foo", NA, "bar", "\033[1mfoobar\033[22m"))
  )
  expect_equal(is.na(ansi_simplify(s)), c(F, T, F, F))

  expect_equal(
    ansi_substr(s, 1, 2),
    ansi_string(c("fo", NA, "ba", "\033[1mfo\033[22m"))
  )
  expect_equal(is.na(ansi_substr(s, 1, 2)), c(F, T, F, F))

  expect_snapshot(ansi_html(s))
  expect_equal(is.na(ansi_html(s)), c(F, T, F, F))

  expect_equal(ansi_has_any(s), c(F, NA, F, T))

  expect_equal(ansi_strip(s), c("foo", NA, "bar", "foobar"))
  expect_equal(is.na(ansi_strip(s)), c(F, T, F, F))
})

test_that("strrep", {
  expect_equal(strrep(character(), 5), character())
  expect_equal(strrep("a", 5), "aaaaa")
})

test_that("convert to character", {
  expect_false(ansi_has_any(123))
  expect_equal(ansi_strip(123), "123")
  expect_equal(ansi_substr(1234, 2, 3), ansi_string("23"))
  expect_equal(ansi_substring(1234, 2, 3), ansi_string("23"))
  expect_equal(ansi_strsplit(1234, 2), list(ansi_string(c("1", "34"))))
  expect_equal(ansi_trimws(123), ansi_string("123"))
  expect_equal(ansi_strwrap(123), ansi_string("123"))
  expect_equal(ansi_simplify(123), ansi_string("123"))
  expect_equal(ansi_html(123), "123")
  expect_equal(ansi_nchar(123), 3)
})

test_that("ansi_nchar", {
  b <- col_red(c(
    "\U0001f477\U0001f3fb",
    "\U0001f477\U0001f3fc\u200d\u2642\ufe0f",
    "\U0001f477\U0001f3fd\u200d\u2640\ufe0f",
    "\U0001f477",
    "\U0001f477\U0001f3fe\u200d\u2640\ufe0f"
  ))
  expect_equal(ansi_nchar(b), rep(1L, 5))
  expect_equal(ansi_nchar(b, "chars"), rep(1L, 5))
  expect_equal(ansi_nchar(b, "bytes"), c(8L, 17L, 17L, 4L, 17L))
  expect_equal(ansi_nchar(b, "width"), rep(2L, 5))
  expect_equal(ansi_nchar(b, "graphemes"), rep(1L, 5))
  expect_equal(ansi_nchar(b, "codepoints"), c(2L, 5L, 5L, 1L, 5L))

  bb <- paste0(b, collapse = "")
  expect_equal(ansi_nchar(bb), sum(rep(1L, 5)))
  expect_equal(ansi_nchar(bb, "chars"), sum(rep(1L, 5)))
  expect_equal(ansi_nchar(bb, "bytes"), sum(c(8L, 17L, 17L, 4L, 17L)))
  expect_equal(ansi_nchar(bb, "width"), sum(rep(2L, 5)))
  expect_equal(ansi_nchar(bb, "graphemes"), sum(rep(1L, 5)))
  expect_equal(ansi_nchar(bb, "codepoints"), sum(c(2L, 5L, 5L, 1L, 5L)))

  expect_equal(ansi_nchar(character(), "chars"), integer())
  expect_equal(ansi_nchar(character(), "bytes"), integer())
  expect_equal(ansi_nchar(character(), "width"), integer())
  expect_equal(ansi_nchar(character(), "graphemes"), integer())
  expect_equal(ansi_nchar(character(), "codepoints"), integer())

  expect_equal(ansi_nchar("", "chars"), 0L)
  expect_equal(ansi_nchar("", "bytes"), 0L)
  expect_equal(ansi_nchar("", "width"), 0L)
  expect_equal(ansi_nchar("", "graphemes"), 0L)
  expect_equal(ansi_nchar("", "codepoints"), 0L)
})

test_that("ansi_align with graphemes", {
  withr::local_options(cli.num_colors = 256)

  b <- col_red(c(
    "\U0001f477\U0001f3fb",
    "\U0001f477\U0001f3fc\u200d\u2642\ufe0f",
    "\U0001f477\U0001f3fd\u200d\u2640\ufe0f",
    "\U0001f477",
    "\U0001f477\U0001f3fe\u200d\u2640\ufe0f"
  ))

  expect_equal(
    ansi_align(b, width = 10, align = "left"),
    ansi_string(paste0(b, strrep(" ", 8)))
  )
  expect_equal(
    ansi_align(b, width = 10, align = "center"),
    ansi_string(paste0(strrep(" ", 4), b, strrep(" ", 4)))
  )
  expect_equal(
    ansi_align(b, width = 10, align = "right"),
    ansi_string(paste0(strrep(" ", 8), b))
  )
})

test_that("ansi_columns with graphemes", {
  withr::local_options(cli.num_colors = 256)

  b <- col_red(c(
    "\U0001f477\U0001f3fb",
    "\U0001f477\U0001f3fc\u200d\u2642\ufe0f",
    "\U0001f477\U0001f3fd\u200d\u2640\ufe0f",
    "\U0001f477",
    "\U0001f477\U0001f3fe\u200d\u2640\ufe0f"
  ))

  expect_equal(
    ansi_columns(b, width = 6),
    ansi_string(c(
      paste0(b[1], " ", b[2], " "),
      paste0(b[3], " ", b[4], " "),
      paste0(b[5], "    ")
    ))
  )
})

test_that("ansi_substr with graphemes", {
  withr::local_options(cli.num_colors = 256)

  b <- col_red(c(
    "\U0001f477\U0001f3fb",
    "\U0001f477\U0001f3fc\u200d\u2642\ufe0f",
    "\U0001f477\U0001f3fd\u200d\u2640\ufe0f",
    "\U0001f477",
    "\U0001f477\U0001f3fe\u200d\u2640\ufe0f"
  ))
  bb <- paste0(b, collapse = "")

  expect_equal(
    ansi_substr(bb, 2, 4),
    col_red(paste0(
      "\U0001f477\U0001f3fc\u200d\u2642\ufe0f",
      "\U0001f477\U0001f3fd\u200d\u2640\ufe0f",
      "\U0001f477"
    ))
  )
})

test_that("ansi_grep", {
  withr::local_options(cli.num_colors = 256, cli.hyperlink = TRUE)
  red_needle <- col_red("needle")
  haystack <- c("foo", "needle", "foo")
  green_haystack <- col_green(haystack)
  expect_equal(
    ansi_grep(red_needle, green_haystack),
    2L
  )
  expect_equal(
    ansi_grep(red_needle, haystack),
    2L
  )
  expect_equal(
    ansi_grep(red_needle, green_haystack, value = TRUE),
    ansi_string(green_haystack[2])
  )
  expect_equal(
    ansi_grep("nope", haystack),
    integer()
  )
  expect_equal(
    ansi_grep("nope", haystack, value = TRUE),
    ansi_string(character())
  )
  expect_equal(
    ansi_grep("nope", character()),
    integer()
  )
  expect_equal(
    ansi_grep("nope", character(), value = TRUE),
    ansi_string(character())
  )
})

test_that("ansi_grepl", {
  withr::local_options(cli.num_colors = 256, cli.hyperlink = TRUE)
  red_needle <- col_red("needle")
  haystack <- c("foo", "needle", "foo")
  green_haystack <- col_green(haystack)
  expect_equal(
    ansi_grepl(red_needle, green_haystack),
    c(FALSE, TRUE, FALSE)
  )
  expect_equal(
    ansi_grepl(red_needle, haystack),
    c(FALSE, TRUE, FALSE)
  )
  expect_equal(
    ansi_grepl("nope", haystack),
    c(FALSE, FALSE, FALSE)
  )
  expect_equal(
    ansi_grepl("nope", character()),
    logical()
  )
})

test_that("ansi_nzchar", {
  withr::local_options(cli.num_colors = 256, cli.hyperlink = TRUE)

  expect_equal(ansi_nzchar(""), FALSE)
  expect_equal(ansi_nzchar(c("", "")), c(FALSE, FALSE))
  expect_equal(ansi_nzchar(c("", "yes")), c(FALSE, TRUE))
  expect_equal(ansi_nzchar(c("yes", "")), c(TRUE, FALSE))
  expect_equal(ansi_nzchar(c("yes", "yes")), c(TRUE, TRUE))

  expect_equal(ansi_nzchar(col_red("")), FALSE)
  expect_equal(ansi_nzchar(col_red(c("", ""))), c(FALSE, FALSE))
  expect_equal(ansi_nzchar(col_red(c("", "yes"))), c(FALSE, TRUE))
  expect_equal(ansi_nzchar(col_red(c("yes", ""))), c(TRUE, FALSE))
  expect_equal(ansi_nzchar(col_red(c("yes", "yes"))), c(TRUE, TRUE))

  expect_equal(ansi_nzchar(style_hyperlink("text", "https://x.xom")), TRUE)
  expect_equal(ansi_nzchar(style_hyperlink("", "https://x.xom")), FALSE)

  expect_equal(ansi_nzchar(NA), nzchar(NA))
  expect_equal(ansi_nzchar(NA, keepNA = TRUE), nzchar(NA, keepNA = TRUE))
})

test_that("ansi_align", {
  txt0 <- "\033]8;;https://ex.com\007te\033]8;;\007"
  txt1 <- st_from_bel(txt0)
  txt <- paste0(txt0, txt1)
  space3 <- strrep(" ", 3)
  space6 <- strrep(" ", 6)
  expect_equal(
    ansi_align(txt, 10),
    ansi_string(paste0(txt, space6))
  )

  expect_equal(
    ansi_align(txt, 10, "center"),
    ansi_string(paste0(space3, txt, space3))
  )

  expect_equal(
    ansi_align(txt, 10, "right"),
    ansi_string(paste0(space6, txt))
  )
})

test_that("ansi_chartr", {
  txt0 <- "1\033]8;;https://ex.com\007text\033]8;;\0072"
  txt1 <- st_from_bel(txt0)
  txt <- paste0(txt0, txt1)
  expect_equal(
    ansi_chartr("12x0", "34yo", txt),
    ansi_string(paste0(
      "3\033]8;;https://ex.com\007teyt\033]8;;\0074",
      st_from_bel("3\033]8;;https://ex.com\007teyt\033]8;;\0074")
    ))
  )
})

test_that("ansi_columns", {
  txt <- "\033]8;;https://ex.com\007text\033]8;;\007"

  expect_equal(
    ansi_strip(ansi_columns(rep(txt, 4), 10)),
    c("text text ", "text text ")
  )
})

test_that("ansi_has_any", {
  txt <- "1\033]8;;https://ex.com\007text\033]8;;\0072"
  expect_true(ansi_has_any(txt))
  expect_true(ansi_has_any(txt, sgr = FALSE, csi = FALSE))
  expect_false(ansi_has_any(txt, link = FALSE))
})

test_that("ansi_html", {
  txt <- paste0(
    "1",
    "\033[1m\033]8;;https://ex.com\007",
    "text",
    "\033]8;;\007",
    "\033[22m",
    "2"
  )

  expect_equal(
    ansi_html(txt),
    paste0(
      "1",
      "<a class=\"ansi-link\" href=\"https://ex.com\">",
      "<span class=\"ansi ansi-bold\">",
      "text",
      "</span>",
      "</a>",
      "2"
    )
  )
})

test_that("ansi_nchar", {
  cases <- list(
    list("\033]8;;https://ex.com\007text\033]8;;\007", 4),
    list("\033]8;x=1:y=2;https://ex.com\007text\033]8;;\007", 4),
    list("\033]8;;https://ex.com\007text\033]8;;\007", 4),
    list("\033[1m\033]8;;https://ex.com\007text\033]8;;\007\033[22m", 4)
  )

  for (c in cases) {
    expect_equal(ansi_nchar(c[[1]]), c[[2]])
  }
})

test_that("ansi_regex", {
  cases <- c(
    "1\033]8;;https://ex.com\007text\033]8;;\0072",
    "1\033]8;x=1:y=2;https://ex.com\007text\033]8;;\0072",
    "1\033]8;;https://ex.com\007text\033]8;;\0072",
    "1\033[1m\033]8;;https://ex.com\007text\033]8;;\007\033[22m2"
  )

  for (case in cases) {
    expect_equal(gsub(ansi_regex(), "", case, perl = TRUE), "1text2")
  }
})

test_that("ansi_simplify", {
  txt <- "1\033[1m\033]8;;https://ex.com\007text\033]8;;\007\033[22m2"
  expect_equal(
    ansi_simplify(txt),
    ansi_string(
      "1\033]8;;https://ex.com\007\033[1mtext\033[22m\033]8;;\0072"
    )
  )
})

test_that("ansi_strip", {
  cases <- c(
    "1\033]8;;https://ex.com\007text\033]8;;\0072",
    "1\033]8;x=1:y=2;https://ex.com\007text\033]8;;\0072",
    "1\033]8;;https://ex.com\007text\033]8;;\0072",
    "1\033[1m\033]8;;https://ex.com\007text\033]8;;\007\033[22m2"
  )

  for (case in cases) {
    expect_equal(ansi_strip(case), "1text2")
  }

  txt <- "1\033[1m\033]8;;https://ex.com\007text\033]8;;\007\033[22m2"
  expect_equal(
    ansi_strip(txt, link = FALSE),
    "1\033]8;;https://ex.com\007text\033]8;;\0072"
  )
  expect_equal(
    ansi_strip(txt, sgr = FALSE),
    "1\033[1mtext\033[22m2"
  )
})

test_that("ansi_strsplit", {
  txt <- "1\033]8;;https://ex.com\007te;xt\033]8;;\0072"
  expect_equal(
    ansi_strsplit(txt, ";"),
    list(ansi_string(c(
      "1\033]8;;https://ex.com\007te\033]8;;\007",
      "\033]8;;https://ex.com\007xt\033]8;;\0072"
    )))
  )
})

test_that("ansi_strtrim", {
  txt <- "1\033]8;;https://ex.com\007text\033]8;;\0072"
  expect_equal(
    ansi_strtrim(txt, 6),
    ansi_string(txt)
  )
  expect_equal(
    ansi_strtrim(txt, 4),
    ansi_string(paste0(ansi_substr(txt, 1, 1), "..."))
  )
})

test_that("ansi_strwrap", {
  txt <- "1\033]8;;https://ex.com\007text\033]8;;\0072"
  wrp <- ansi_strwrap(strrep(paste0(txt, " "), 10), 15)
  expect_equal(
    wrp,
    ansi_string(rep(paste0(txt, " ", txt), 5))
  )
})

test_that("ansi_substr", {
  txt <- "1\033[1m\033]8;;https://ex.com\007text\033]8;;\007\033[22m2"

  cases <- list(
    list(1, 3, "1\033]8;;https://ex.com\007\033[1mte\033[22m\033]8;;\007"),
    list(4, 6, "\033]8;;https://ex.com\007\033[1mxt\033[22m\033]8;;\0072"),
    list(3, 4, "\033]8;;https://ex.com\007\033[1mex\033[22m\033]8;;\007"),
    list(1, 10, "1\033]8;;https://ex.com\007\033[1mtext\033[22m\033]8;;\0072")
  )

  for (c in cases) {
    expect_equal(ansi_substr(txt, c[[1]], c[[2]]), ansi_string(c[[3]]))
  }
})

test_that("ansi_substring", {
  txt <- "1\033[1m\033]8;;https://ex.com\007text\033]8;;\007\033[22m2"

  cases <- list(
    list(1, 3, "1\033]8;;https://ex.com\007\033[1mte\033[22m\033]8;;\007"),
    list(4, 6, "\033]8;;https://ex.com\007\033[1mxt\033[22m\033]8;;\0072"),
    list(3, 4, "\033]8;;https://ex.com\007\033[1mex\033[22m\033]8;;\007"),
    list(1, 10, "1\033]8;;https://ex.com\007\033[1mtext\033[22m\033]8;;\0072")
  )

  for (c in cases) {
    expect_equal(ansi_substring(txt, c[[1]], c[[2]]), ansi_string(c[[3]]))
  }
})

test_that("ansi_tolower", {
  txt <- "Pre\033]8;;https://ex.com\007tEXt\033]8;;\007PoST"
  expect_equal(
    ansi_tolower(txt),
    ansi_string("pre\033]8;;https://ex.com\007text\033]8;;\007post")
  )
})

test_that("ansi_toupper", {
  txt <- "Pre\033]8;;https://ex.com\007tEXt\033]8;;\007PoST"
  expect_equal(
    ansi_toupper(txt),
    ansi_string("PRE\033]8;;https://ex.com\007TEXT\033]8;;\007POST")
  )
})

test_that("ansi_trimws", {
  txt <- "1\033[1m\033]8;;https://ex.com\007text\033]8;;\007\033[22m2"
  expect_equal(
    ansi_trimws(paste0("   ", txt, "   ")),
    ansi_simplify(txt)
  )
  expect_equal(
    ansi_trimws(txt),
    ansi_string(txt)
  )
})

test_that("unknown hyperlink type", {
  expect_snapshot(
    error = TRUE,
    make_link("this", "foobar")
  )
})

test_that("iterm file links", {
  withr::local_envvar(R_CLI_HYPERLINK_STYLE = "iterm")
  withr::local_envvar(R_CLI_HYPERLINK_FILE_URL_FORMAT = NA_character_)
  withr::local_options(cli.hyperlink = TRUE)
  expect_snapshot({
    cli::cli_text("{.file /path/to/file:10}")
    cli::cli_text("{.file /path/to/file:10:20}")
  })
})

test_that("rstudio links", {
  local_clean_cli_context()
  withr::local_envvar(
    RSTUDIO = "1",
    RSTUDIO_SESSION_PID = Sys.getpid(),
    RSTUDIO_CHILD_PROCESS_PANE = "build",
    RSTUDIO_CLI_HYPERLINKS = "1"
  )
  withr::local_options(
    cli.hyperlink = TRUE,
    cli.hyperlink_help = TRUE,
    cli.hyperlink_run = TRUE,
    cli.hyperlink_vignette = TRUE
  )
  expect_snapshot(
    cli::cli_text("{.fun pkg::fun}")
  )
  expect_snapshot(
    cli::cli_text("{.help fun}")
  )
  expect_snapshot(
    cli::cli_text("{.run package::func()}")
  )
  expect_snapshot(
    cli::cli_text("{.vignette package::title}")
  )
  expect_snapshot(
    cli::cli_text("{.topic pkg::topic}")
  )
})

test_that("ST hyperlinks", {
  withr::local_envvar(R_CLI_HYPERLINK_MODE = "posix")
  withr::local_options(cli.hyperlink = TRUE)
  expect_snapshot(
    cat(style_hyperlink("text", "https://example.com"))
  )
})

test_that("ansi_has_hyperlink_support", {
  local_clean_cli_context()

  # force with env var
  withr::with_envvar(list(R_CLI_HYPERLINKS = "true"),
    expect_true(ansi_has_hyperlink_support())
  )

  # if no ansi support, then no
  local_mocked_bindings(num_ansi_colors = function() 256L)
  expect_false(ansi_has_hyperlink_support())

  # are we in rstudio with support?
  local_mocked_bindings(
    num_ansi_colors = function() 257L,
    rstudio_detect = function() list(type = "rstudio_console", hyperlink = TRUE)
  )
  expect_true(ansi_has_hyperlink_support())
})

test_that("ansi_has_hyperlink_support 2", {
  local_clean_cli_context()
  local_mocked_bindings(
    num_ansi_colors = function() 256L,
    isatty = function(...) FALSE
  )
  expect_false(ansi_has_hyperlink_support())
})

test_that("ansi_has_hyperlink_support 3", {
  local_clean_cli_context()
  local_mocked_bindings(
    num_ansi_colors = function() 256L,
    isatty = function(...) TRUE,
    is_windows = function() TRUE
  )
  withr::local_envvar(WT_SESSION = "4c464723-f51f-4612-83f7-31e1c75abd83")
  expect_true(ansi_has_hyperlink_support())
})

test_that("ansi_has_hyperlink_support 4", {
  local_clean_cli_context()
  local_mocked_bindings(
    num_ansi_colors = function() 256L,
    isatty = function(...) TRUE
  )

  withr::local_envvar("CI" = "true")
  expect_false(ansi_has_hyperlink_support())

  withr::local_envvar("CI" = NA_character_, TEAMCITY_VERSION = "1")
  expect_false(ansi_has_hyperlink_support())
})

test_that("ansi_has_hyperlink_support 5", {
  local_clean_cli_context()
  local_mocked_bindings(
    num_ansi_colors = function() 256L,
    isatty = function(...) TRUE
  )

  withr::local_envvar(
    TERM_PROGRAM = "iTerm.app",
    TERM_PROGRAM_VERSION = "3.4.16"
  )
  expect_true(ansi_has_hyperlink_support())
})

test_that("ansi_has_hyperlink_support 5", {
  local_clean_cli_context()
  local_mocked_bindings(
    num_ansi_colors = function() 256L,
    isatty = function(...) TRUE
  )

  withr::local_envvar(VTE_VERSION = "0.51.1")
  expect_true(ansi_has_hyperlink_support())

  withr::local_envvar(VTE_VERSION = "5110")
  expect_true(ansi_has_hyperlink_support())

  withr::local_envvar(VTE_VERSION = "foo")
  expect_false(ansi_has_hyperlink_support())
})

test_that("ansi_hyperlink_types", {
  local_clean_cli_context()
  withr::local_envvar(
    R_CLI_HYPERLINKS = "true",
    R_CLI_HYPERLINK_RUN = "true"
  )
  expect_true(ansi_hyperlink_types()[["run"]])
})

test_that("get_config_chr() consults option, env var, then its default", {
  local_clean_cli_context()

  key <- "hyperlink_TYPE_url_format"

  expect_null(get_config_chr(key))

  withr::local_envvar(R_CLI_HYPERLINK_TYPE_URL_FORMAT = "envvar")
  expect_equal(get_config_chr(key), "envvar")

  withr::local_options(cli.hyperlink_type_url_format = "option")
  expect_equal(get_config_chr(key), "option")
})

test_that("get_config_chr() errors if option is not NULL or string", {
  withr::local_options(cli.something = FALSE)

  expect_error(get_config_chr("something"), "is_string")
})

test_that("get_hyperlink_format() delivers custom format", {
  local_clean_cli_context()

  withr::local_options(
    cli.hyperlink_run = TRUE,
    cli.hyperlink_help = TRUE,
    cli.hyperlink_vignette = TRUE
  )

  # env var is consulted after option, so start with env var
  withr::local_envvar(
    R_CLI_HYPERLINK_RUN_URL_FORMAT = "envvar{code}",
    R_CLI_HYPERLINK_HELP_URL_FORMAT = "envvar{topic}",
    R_CLI_HYPERLINK_VIGNETTE_URL_FORMAT = "envvar{vignette}"
  )

  expect_equal(get_hyperlink_format("run"), "envvar{code}")
  expect_equal(get_hyperlink_format("help"), "envvar{topic}")
  expect_equal(get_hyperlink_format("vignette"), "envvar{vignette}")

  withr::local_options(
    cli.hyperlink_run_url_format = "option{code}",
    cli.hyperlink_help_url_format = "option{topic}",
    cli.hyperlink_vignette_url_format = "option{vignette}"
  )

  expect_equal(get_hyperlink_format("run"), "option{code}")
  expect_equal(get_hyperlink_format("help"), "option{topic}")
  expect_equal(get_hyperlink_format("vignette"), "option{vignette}")
})

test_that("parse_file_link_params(), typical input", {
  expect_equal(
    parse_file_link_params("some/path.ext"),
    list(
      path = "some/path.ext",
      line = NULL,
      column = NULL
    )
  )
  expect_equal(
    parse_file_link_params("some/path.ext:14"),
    list(
      path = "some/path.ext",
      line = "14",
      column = NULL
    )
  )
  expect_equal(
    parse_file_link_params("some/path.ext:14:23"),
    list(
      path = "some/path.ext",
      line = "14",
      column = "23"
    )
  )
})

test_that("parse_file_link_params(), weird trailing colons", {
  expect_equal(
    parse_file_link_params("some/path.ext:"),
    list(
      path = "some/path.ext",
      line = NULL,
      column = NULL
    )
  )
  expect_equal(
    parse_file_link_params("some/path.ext::"),
    list(
      path = "some/path.ext",
      line = NULL,
      column = NULL
    )
  )
  expect_equal(
    parse_file_link_params("some/path.ext:14:"),
    list(
      path = "some/path.ext",
      line = "14",
      column = NULL
    )
  )
})

test_that("interpolate_parts(), more or less data in `params`", {
  fmt <- "whatever/{path}#@${line}^&*{column}"
  params <- list(path = "some/path.ext", line = "14", column = "23")

  expect_equal(
    interpolate_parts(fmt, params),
    "whatever/some/path.ext#@$14^&*23"
  )

  params <- list(path = "some/path.ext", line = "14", column = NULL)
  expect_equal(
    interpolate_parts(fmt, params),
    "whatever/some/path.ext#@$14"
  )

  params <- list(path = "some/path.ext", line = NULL, column = NULL)
  expect_equal(
    interpolate_parts(fmt, params),
    "whatever/some/path.ext"
  )
})

test_that("interpolate_parts(), format only has `path`", {
  fmt <- "whatever/{path}"
  params <- list(path = "some/path.ext", line = "14", column = "23")
  expect_equal(
    interpolate_parts(fmt, params),
    "whatever/some/path.ext"
  )
})

test_that("construct_file_link() works with custom format and an absolute path", {
  withr::local_options(
    "cli.hyperlink_file_url_format" = "positron://file{path}:{line}:{column}"
  )

  expect_equal(
    construct_file_link(list(path = "/absolute/path")),
    list(url = "positron://file/absolute/path")
  )
  expect_equal(
    construct_file_link(list(path = "/absolute/path", line = "12")),
    list(url = "positron://file/absolute/path:12")
  )
  expect_equal(
    construct_file_link(list(path = "/absolute/path", line = "12", column = "5")),
    list(url = "positron://file/absolute/path:12:5")
  )

  local_mocked_bindings(is_windows = function() TRUE)
  expect_equal(
    construct_file_link(list(path = "c:/absolute/path")),
    list(url = "positron://file/c:/absolute/path")
  )
})

test_that("construct_file_link() works with custom format and a relative path", {
  withr::local_options(
    "cli.hyperlink_file_url_format" = "positron://file{path}:{line}:{column}"
  )

  # inspired by test helpers `sanitize_wd()` and `sanitize_home()`, but these
  # don't prefix the pattern-to-replace with `file://`
  sanitize_dir <- function(x, what = c("wd", "home")) {
    what <- match.arg(what)
    pattern <- switch(what, wd = getwd(), home = path.expand("~"))
    if (is_windows()) {
      pattern <- paste0("/", pattern)
    }
    replacement <- switch(what, wd = "/working/directory", home = "/my/home")
    sub(pattern, replacement, x$url, fixed = TRUE)
  }

  expect_equal(
    sanitize_dir(construct_file_link(list(path = "relative/path")), what = "wd"),
    "positron://file/working/directory/relative/path"
  )
  expect_equal(
    sanitize_dir(construct_file_link(list(path = "relative/path:12")), what = "wd"),
    "positron://file/working/directory/relative/path:12"
  )
  expect_equal(
    sanitize_dir(construct_file_link(list(path = "relative/path:12:5")), what = "wd"),
    "positron://file/working/directory/relative/path:12:5"
  )

  expect_equal(
    sanitize_dir(construct_file_link(list(path = "./relative/path")), what = "wd"),
    "positron://file/working/directory/./relative/path"
  )
  expect_equal(
    sanitize_dir(construct_file_link(list(path = "./relative/path:12")), what = "wd"),
    "positron://file/working/directory/./relative/path:12"
  )
  expect_equal(
    sanitize_dir(construct_file_link(list(path = "./relative/path:12:5")), what = "wd"),
    "positron://file/working/directory/./relative/path:12:5"
  )

  expect_equal(
    sanitize_dir(construct_file_link(list(path = "~/relative/path")), what = "home"),
    "positron://file/my/home/relative/path"
  )
  expect_equal(
    sanitize_dir(construct_file_link(list(path = "~/relative/path:17")), what = "home"),
    "positron://file/my/home/relative/path:17"
  )
  expect_equal(
    sanitize_dir(construct_file_link(list(path = "~/relative/path:17:22")), what = "home"),
    "positron://file/my/home/relative/path:17:22"
  )
})

test_that("construct_file_link() works with custom format and input starting with 'file://'", {
  withr::local_options(
    "cli.hyperlink_file_url_format" = "positron://file{path}:{line}:{column}"
  )

  expect_equal(
    construct_file_link(list(path = "file:///absolute/path")),
    list(url = "positron://file/absolute/path")
  )
  expect_equal(
    construct_file_link(list(path = "file:///absolute/path", line = "12", column = "5")),
    list(url = "positron://file/absolute/path:12:5")
  )
})

test_that("make_style without name", {
  pink <- make_ansi_style("pink")
  expect_true(inherits(pink, "cli_ansi_style"))
})

test_that("hexa color regex works", {
  positive <- c("#000000", "#ffffff", "#0f0f0f", "#f0f0f0",
                "#00000000", "#ffffffff", "#0f0f0f00", "#f0f0f055")

  negative <- c("", "#12345", "123456", "1234567", "12345678",
                "#1234567", "#1234ffg", "#gggggx", "foo#123456",
                "foo#123456bar")

  for (color in positive) {
    expect_true(grepl(hash_color_regex, color))
    expect_true(grepl(hash_color_regex, toupper(color)))
  }

  for (color in negative) {
    expect_false(grepl(hash_color_regex, color))
    expect_false(grepl(hash_color_regex, toupper(color)))
  }
})

test_that("we fall back for ANSI 8 if needed", {
  yellow3 <- make_ansi_style("yellow3", colors = 8)
  expect_equal(yellow3("foobar"), col_yellow("foobar"))
})

test_that("we can create a style from an R color", {
  red4 <- make_ansi_style("red4")
  red_text <- red4("text")
  expect_true(num_ansi_colors() == 1 || ansi_has_any(red_text))
})

test_that("errors", {
  expect_snapshot(
    error = TRUE,
    make_ansi_style(1:10)
  )
})

test_that("make_ansi_style", {
  withr::local_options(cli.num_colors = 256)
  cf <- crayon::red
  expect_equal(
    make_ansi_style(cf)("foo"),
    col_red("foo")
  )

  expect_equal(
    make_ansi_style("dim")("foo"),
    style_blurred("foo")
  )

  expect_equal(
    make_ansi_style("red", bg = TRUE)("red"),
    bg_red("red")
  )

  expect_equal(
    make_ansi_style(cbind(c(200, 200, 100)), grey = TRUE)("200-200-100"),
    ansi_string("\033[38;5;250m200-200-100\033[39m")
  )

  withr::local_options(cli.num_colors = truecolor)
  expect_equal(
    make_ansi_style(cbind(c(200, 200, 100)))("200-200-100"),
    ansi_string("\033[38;2;200;200;100m200-200-100\033[39m")
  )

  expect_equal(
    make_ansi_style(cbind(c(200, 200, 100)), bg = TRUE)("200-200-100"),
    ansi_string("\033[48;2;200;200;100m200-200-100\033[49m")
  )

  # errors
  expect_snapshot(
    error = TRUE,
    make_ansi_style(1:10)
  )

  expect_snapshot(
    error = TRUE,
    make_ansi_style("foobar")
  )  
})

test_that("cli_ansi_string", {
  right <- c("cli_ansi_string", "ansi_string", "character")
  expect_equal(class(ansi_string("foobar")), right)
  expect_equal(class(ansi_string(133)), right)
  expect_equal(class(ansi_string(ansi_string(134))), right)
})

test_that("ansi_regex", {
  # somewhat special sequences
  cases <- list(
    list("foo\033[39;49mbar", "foobar"),
    list("foo\033[1;3;4mbar", "foobar"),
    list("foo\033[1;;4mbar", "foobar"),
    list("foo\033[mbar", "foobar")
  )

  for (case in cases) {
    expect_equal(gsub(ansi_regex(), "", case[[1]], perl = TRUE), case[[2]])
  }
})

test_that("ansi_has_any works", {
  withr::local_options(list(
    cli.num_colors = 256L,
    cli.hyperlink = TRUE
  ))
  expect_false(ansi_has_any("foobar"))
  funcs <- ls(asNamespace("cli"), pattern = "^col_|^bg_|^style_")

  for (sym in setdiff(funcs, "style_hyperlink")) {
    fun <- get(sym, envir = asNamespace("cli"))
    expect_true(ansi_has_any(fun("foo", "bar")), label = sym)
  }
})

test_that("ansi_strip works", {
  withr::local_options(list(
    cli.num_colors = 256L,
    cli.hyperlink = TRUE
  ))
  expect_equal("", ansi_strip(""))
  expect_equal("foobar", ansi_strip("foobar"))
  expect_equal(
    "foobar",
    ansi_strip(col_red(style_underline(style_bold("foobar"))))
  )

  funcs <- ls(asNamespace("cli"), pattern = "^col_|^bg_|^style_")
  for (sym in setdiff(funcs, "style_hyperlink")) {
    fun <- get(sym, envir = asNamespace("cli"))
    ans <- if (sym == "style_hyperlink") "foo" else "foobar"
    expect_equal(ans, ansi_strip(fun("foo", "bar")))
  }
})

str <- c("",
         "plain",
         "\033[31m",
         "\033[39m",
         "\033[31mred\033[39m",
         "\033[31mred\033[39m\033[31mred\033[39m",
         "foo\033[31mred\033[39m",
         "\033[31mred\033[39mfoo")

test_that("ansi_nchar", {
  withr::local_options(list(cli.num_colors = 256L))
  for (s in str) {
    expect_equal(ansi_nchar(s), nchar(ansi_strip(s)), info = s)
  }
})

test_that("ansi_nchar wide characters", {
  expect_equal(ansi_nchar("\u231a", "width"), 2L)
})

test_that("ansi_substr bad input", {
  expect_error(
    ansi_substr("foobar", NULL, 10),
    "invalid substring arguments"
  )
  expect_error(
    ansi_substr("foobar", 10, NULL),
    "invalid substring arguments"
  )
})

test_that("ansi_substr", {
  withr::local_options(list(cli.num_colors = 256L))
  for (s in str) {
    for (i in 1 %:% ansi_nchar(s)) {
      for (j in i %:% ansi_nchar(s)) {
        expect_equal(ansi_strip(ansi_substr(s, i, j)),
                     substr(ansi_strip(s), i, j), info = paste(s, i, j))
      }
    }
  }
})

test_that("ansi_substr keeps color", {
  withr::local_options(list(cli.num_colors = 256L))
  expect_equal(
    ansi_substr("\033[31mred\033[39m", 1, 1),
    ansi_string("\033[31mr\033[39m")
  )
  expect_equal(
    ansi_substr("foo\033[31mred\033[39m", 4, 4),
    ansi_string("\033[31mr\033[39m")
  )
  expect_equal(
    ansi_substr("foo\033[31mred\033[39mbar", 4, 4),
    ansi_string("\033[31mr\033[39m")
  )
  expect_equal(
    ansi_substr("\033[31mred\033[39mfoo\033[31mred\033[39mbar", 7, 7),
    ansi_string("\033[31mr\033[39m")
  )
})

test_that("ansi_substr, start after string end", {
  withr::local_options(list(cli.num_colors = 256L))
  expect_equal(ansi_substr("red", 4, 4), ansi_string(""))
  expect_equal(ansi_substr("red", 4, 5), ansi_string(""))
  expect_equal(ansi_strip(ansi_substr("\033[31mred\033[39m", 4, 4)), "")
  expect_equal(ansi_strip(ansi_substr("\033[31mred\033[39m", 4, 5)), "")

  expect_equal(ansi_substr("red", 3, 4), ansi_string("d"))
  expect_equal(ansi_substr("red", 3, 5), ansi_string("d"))
  expect_equal(ansi_strip(ansi_substr("\033[31mred\033[39m", 3, 4)), "d")
  expect_equal(ansi_strip(ansi_substr("\033[31mred\033[39m", 3, 5)), "d")
})

test_that("ansi_substr, multiple strings", {
  withr::local_options(list(cli.num_colors = 256L))
  set.seed(42)
  for (i in 1:100) {
    strs <- sample(str, 4)
    num_starts <- sample(1:5, 1)
    num_stops <- sample(1:5, 1)
    starts <- sample(1:5, num_starts, replace = TRUE)
    stops <- sample(1:5, num_stops, replace = TRUE)
    r1 <- ansi_strip(ansi_substr(strs, starts, stops))
    r2 <- substr(ansi_strip(strs), starts, stops)
    expect_equal(r1, r2)
  }
})

test_that("ansi_substr corner cases", {
  withr::local_options(list(cli.num_colors = 256L))
  # Zero length input

  c0 <- character(0L)
  o0 <- structure(list(), class="abc")
  co0 <- structure(character(0L), class="abc")
  expect_identical(ansi_substr(c0, 1, 1), ansi_string(substr(c0, 1, 1)))
  expect_identical(ansi_substr(o0, 1, 1), ansi_string(substr(o0, 1, 1)))
  expect_identical(ansi_substr(co0, 1, 1), ansi_string(substr(co0, 1, 1)))

  expect_identical(
    ansi_substring(c0, 1, 1),
    ansi_string(substring(c0, 1, 1))
  )
  expect_identical(
    ansi_substring(o0, 1, 1),
    ansi_string(substring(o0, 1, 1))
  )
  expect_identical(
    ansi_substring(co0, 1, 1),
    ansi_string(substring(co0, 1, 1))
  )

  # Character start/stop
  expect_identical(
    ansi_substr("abc", "1", 1),
    ansi_string(substr("abc", "1", 1))
  )
  expect_identical(
    ansi_substr("abc", 1, "1"),
    ansi_string(substr("abc", 1, "1"))
  )

  # non-numeric arguments cause errors; NOTE: this actually "works"
  # with 'substr' but not implemented in 'ansi_substr'
  suppressWarnings(
    expect_error(ansi_substr("abc", "hello", 1), "non-numeric")
  )

})

test_that("ansi_substring", {
  withr::local_options(list(cli.num_colors = 256L))
  for (s in str) {
    for (i in 1 %:% ansi_nchar(s)) {
      for (j in i %:% ansi_nchar(s)) {
        expect_equal(ansi_strip(ansi_substring(s, i, j)),
                     substring(ansi_strip(s), i, j), info = paste(s, i, j))
      }
    }
  }
})

test_that("ansi_substring, multiple strings", {
  withr::local_options(list(cli.num_colors = 256L))
  set.seed(42)
  for (i in 1:100) {
    strs <- sample(str, 4)
    num_starts <- sample(1:5, 1)
    num_stops <- sample(1:5, 1)
    starts <- sample(1:5, num_starts, replace = TRUE)
    stops <- sample(1:5, num_stops, replace = TRUE)
    r1 <- ansi_strip(ansi_substring(strs, starts, stops))
    r2 <- substring(ansi_strip(strs), starts, stops)
    expect_equal(r1, r2)
  }
})

test_that("ansi_substring corner cases", {
  withr::local_options(list(cli.num_colors = 256L))
  # Zero length input

  c0 <- character(0L)
  o0 <- structure(list(), class="abc")
  co0 <- structure(character(0L), class="abc")
  expect_identical(
    ansi_substring(c0, 1, 1),
    ansi_string(substring(c0, 1, 1))
  )
  expect_identical(
    ansi_substring(o0, 1, 1),
    ansi_string(substring(o0, 1, 1))
  )
  expect_identical(
    ansi_substring(co0, 1, 1),
    ansi_string(substring(co0, 1, 1))
  )
})

test_that("ansi_strsplit", {
  withr::local_options(list(cli.num_colors = 256L))
  red <- "\033[31mred\033[39m"

  str <- "plain-plain"
  expect_equal(
    ansi_strsplit(str, "-"),
    lapply(strsplit(str, "-"), ansi_string)
  )

  str <- paste0(red, "-plain")
  expect_equal(
    ansi_strip(ansi_strsplit(str, "-")[[1]]),
    strsplit(ansi_strip(str), "-")[[1]]
  )

  expect_equal(
    ansi_strsplit(str, "e"),
    list(ansi_string(c("\033[31mr\033[39m", "\033[31md\033[39m-plain")))
  )

  str <- paste0(red, "-", red, "-", red)
  expect_equal(
    ansi_strip(ansi_strsplit(str, "-")[[1]]),
    strsplit(ansi_strip(str), "-")[[1]]
  )

  # with leading and trailing separators
  str.2 <- paste0("-", red, "-", red, "-", red, "-")
  expect_equal(ansi_strip(ansi_strsplit(str.2, "-")[[1]]),
               strsplit(ansi_strip(str.2), "-")[[1]])

  # greater than length 1
  str.3 <- paste0("-", c(col_green("hello"), col_red("goodbye")), "-world-")
  expect_equal(ansi_strip(unlist(ansi_strsplit(str.3, "-"))),
               unlist(strsplit(ansi_strip(str.3), "-")))
})

test_that("ansi_strsplit multiple strings", {
  withr::local_options(list(cli.num_colors = 256L))
  red <- "\033[31mred\033[39m"
  str <- c(
    paste0("plain-plain-", red, "-plain-", red),
    paste0(red, "-", red),
    red
  )

  r1 <- lapply(ansi_strsplit(str, "-"), ansi_strip)
  r2 <- strsplit(ansi_strip(str), "-")
  expect_equal(r1, r2)
})

test_that("ansi_strsplit edge cases", {
  withr::local_options(list(cli.num_colors = 256L))
  expect_equal(ansi_strsplit("", "-"), list(ansi_string(character(0L))))
  expect_equal(
    ansi_strip(ansi_strsplit("\033[31m\033[39m", "-")[[1]]), character(0L)
  )

  # special cases
  expect_equal(ansi_strsplit("", ""), lapply(strsplit("", ""), ansi_string))
  expect_equal(
    ansi_strsplit("a", "a"),
    lapply(strsplit("a", "a"), ansi_string)
  )

  # this following test isn't working yet
  expect_equal(
    ansi_strsplit("a", ""),
    lapply(strsplit("a", ""), ansi_string)
  )
  expect_equal(
    ansi_strsplit("", "a"),
    lapply(strsplit("", "a"), ansi_string)
  )

  # Longer strings
  expect_identical(
    ansi_strsplit(c("", "a", "aa"), "a"),
    lapply(strsplit(c("", "a", "aa"), "a"), ansi_string)
  )
  expect_identical(
    ansi_strsplit(c("abaa", "ababza"), "b."),
    lapply(strsplit(c("abaa", "ababza"), "b."), ansi_string)
  )
})

test_that("Weird length 'split'", {
  withr::local_options(list(cli.num_colors = 256L))
  expect_error(
    ansi_strsplit(c("ab", "bd"), c("b", "d")),
    "must be character"
  )
  expect_identical(
    ansi_strsplit("ab", NULL),
    lapply(strsplit("ab", NULL), ansi_string)
  )
  expect_identical(
    ansi_strsplit("ab", character(0L)),
    lapply(strsplit("ab", character(0L)), ansi_string)
  )
})

test_that("ansi_align", {
  withr::local_options(list(cli.num_colors = 256L))
  expect_equal(ansi_align(character()), ansi_string(character()))
  expect_equal(ansi_align("", 0), ansi_string(""))
  expect_equal(ansi_align(" ", 0), ansi_string(" "))
  expect_equal(ansi_align(" ", 1), ansi_string(" "))
  expect_equal(ansi_align(" ", 2), ansi_string("  "))
  expect_equal(ansi_align("a", 1), ansi_string("a"))
  expect_equal(ansi_align(letters, 1), ansi_string(letters))
  expect_equal(ansi_align(letters, 0), ansi_string(letters))
  expect_equal(ansi_align(letters, -1), ansi_string(letters))

  expect_equal(ansi_align(letters, 2), ansi_string(paste0(letters, " ")))
  expect_equal(
    ansi_align(letters, 3, "center"),
    ansi_string(paste0(" ", letters, " "))
  )
  expect_equal(
    ansi_align(letters, 2, "right"),
    ansi_string(paste0(" ", letters))
  )

  expect_equal(
    ansi_align(c("foo", "foobar", "", "a"), 6, "left"),
    ansi_string(c("foo   ", "foobar", "      ", "a     ")))

  expect_equal(
    ansi_align(c("foo", "foobar", "", "a"), 6, "center"),
    ansi_string(c("  foo ", "foobar", "      ", "   a  ")))

  expect_equal(
    ansi_align(c("foo", "foobar", "", "a"), 6, "right"),
    ansi_string(c("   foo", "foobar", "      ", "     a")))

  # #54: alignment of wide characters
  expect_equal(
    ansi_align(c("foo", "\u6210\u4ea4\u65e5", "", "a"), 6, "left"),
    ansi_string(c("foo   ", "\u6210\u4ea4\u65e5", "      ", "a     ")))

  expect_equal(
    ansi_align(c("foo", "\u6210\u4ea4\u65e5", "", "a"), 6, "center"),
    ansi_string(c("  foo ", "\u6210\u4ea4\u65e5", "      ", "   a  ")))

  expect_equal(
    ansi_align(c("foo", "\u6210\u4ea4\u65e5", "", "a"), 6, "right"),
    ansi_string(c("   foo", "\u6210\u4ea4\u65e5", "      ", "     a")))
})

test_that("stripping hyperlinks", {
  skip("Temporarily defunct")
  withr::local_options(list(cli.hyperlink = TRUE))
  x <- unclass(style_hyperlink("foo", "https://r-pkg.org"))
  expect_equal(
    ansi_strip(paste0("1111-", x, "-2222-", x, "-333")),
    "1111-foo-2222-foo-333"
  )
})

test_that("ansi_trimws", {
  cases <- list(
    list(character(), ansi_string(character())),
    list(1, ansi_string(1)),
    list("", ansi_string("")),
    list("foo", ansi_string("foo")),
    list("  foo  ", ansi_string("foo")),
    list(c("foo", "bar"), ansi_string(c("foo", "bar"))),
    list(col_red(c("  colored  ")), ansi_string(col_red("colored"))),
    list(
      paste0("   ", col_red(c("  colored  ")), "   "),
      ansi_string(col_red("colored")))
  )

  for (case in cases) expect_equal(ansi_trimws(case[[1]]), case[[2]])

  cases_left <- list(
    list(character(), ansi_string(character())),
    list(1, ansi_string(1)),
    list("", ansi_string("")),
    list("foo", ansi_string("foo")),
    list("  foo  ", ansi_string("foo  ")),
    list(c("foo", "bar"), ansi_string(c("foo", "bar"))),
    list(col_red(c("  colored  ")), ansi_string(col_red("colored  "))),
    list(
      paste0("   ", col_red(c("  colored  ")), "   "),
      ansi_string(paste0(col_red("colored  "), "   ")))
  )

  for (case in cases_left) {
    expect_equal(ansi_trimws(case[[1]], "left"), case[[2]])
  }

  cases_right <- list(
    list(character(), ansi_string(character())),
    list(1, ansi_string(1)),
    list("", ansi_string("")),
    list("foo", ansi_string("foo")),
    list("  foo  ", ansi_string("  foo")),
    list(c("  foo  ", "  bar  "), ansi_string(c("  foo", "  bar"))),
    list(c("foo", "bar"), ansi_string(c("foo", "bar"))),
    list(col_red(c("  colored  ")), ansi_string(col_red("  colored"))),
    list(
      paste0("   ", col_red(c("  colored  ")), "   "),
      ansi_string(paste0("   ", col_red("  colored"))))
  )

  for (case in cases_right) {
    expect_equal(ansi_trimws(case[[1]], "right"), case[[2]])
  }
})

test_that("ansi_strwrap simple", {
  cases = list(
    list(character(), character()),
    list("", ""),
    list("foo", "foo"),
    list(1, "1"),
    list(c("foo", "bar"), c("foo", "bar"))
  )

  for (case in cases) {
    expect_equal(ansi_strwrap(case[[1]], 20), ansi_string(case[[2]]))
  }
})

test_that("ansi_strwrap simple styled", {
  cases = list(
    list(col_red("foo"), col_red("foo")),
    list(col_red(c("foo", "bar")), col_red(c("foo", "bar")))
  )

  for (case in cases) {
    expect_equal(ansi_strwrap(case[[1]], 20), ansi_string(case[[2]]))
  }
})

test_that("ansi_strwrap", {
  txt0 <- glue::glue_col("
    {col_red Velit occaecat} quis culpa occaecat.  {col_green Pariatur} \\
    ad veniam pariatur {bg_blue consectetur}.  Dolore aliquip et \\
    {style_underline consequat Lorem consectetur} dolor.  Irure id velit \\
    proident elit veniam eu exercitation nisi laboris officia.     Qui \\
    {col_red sunt      occaecat} cillum {col_red sit    commodo sit.    \\
    Culpa} aliquip et consectetur ullamco aliqua Lorem laborum dolore.    ")

  txt <- paste0(txt0, "\n\t  \n", txt0)
  expect_equal(
    ansi_strip(ansi_strwrap(txt0, 40)),
    strwrap(ansi_strip(txt0), 40)
  )
  expect_equal(
    ansi_strip(ansi_strwrap(txt, 40)),
    strwrap(ansi_strip(txt), 40)
  )
  expect_equal(
    ansi_strip(ansi_strwrap(txt, 40, indent = 2)),
    strwrap(ansi_strip(txt), 40, indent = 2)
  )
  expect_equal(
    ansi_strip(ansi_strwrap(txt, 40, exdent = 2)),
    strwrap(ansi_strip(txt), 40, exdent = 2)
  )
  expect_equal(
    ansi_strip(ansi_strwrap(txt, 40, indent = 2, exdent = 4)),
    strwrap(ansi_strip(txt), 40, indent = 2, exdent = 4)
  )
})

test_that("ansi_strwrap double width", {
  expect_equal(
    ansi_strwrap("\U0001F477 1 2 3", 4),
    ansi_string(c("\U0001f477", "1 2", "3"))
  )
})

test_that("ansi_strwrap newlines", {
  expect_equal(
    ansi_strwrap("\033[32mv\033[39m hello world.\nxxx"),
    ansi_string("\033[32mv\033[39m hello world. xxx")
  )
})

test_that("ansi_strwrap and \f edge cases", {
  expect_equal(
    ansi_strwrap("\033[32mfoo\fbar\033[39m"),
    ansi_string(c("\033[32mfoo\033[39m", "\033[32mbar\033[39m"))
  )
  expect_equal(
    ansi_strwrap("\033[32m\ffoo\f\033[39m"),
    ansi_string(c("", "\033[32mfoo\033[39m"))
  )
})

test_that_cli(configs = c("plain", "ansi"), "ansi_strtrim", {
  withr::local_options(c(cli.unicode = FALSE))
  setup_unicode_width_fix()
  cases <- list(
    list("", ansi_string("")),
    list("1", ansi_string("1")),
    list("123456789", ansi_string("123456789")),
    list("1234567890", ansi_string("1234567890")),
    list("12345678901", ansi_string("1234567...")),
    list(
      strrep("\u231A", 6),
      ansi_string(paste0(strrep("\u231A", 3), "..."))),
    list(col_red("1"), col_red("1")),
    list(c("foo", NA, col_red("bar")), ansi_string(c("foo", NA, col_red("bar"))))
  )

  for (case in cases) expect_equal(ansi_strtrim(case[[1]], 10), case[[2]])
})

test_that("ansi_columns", {
  withr::local_options(c(cli.unicode = FALSE))
  local_edition(3)

  expect_equal(ansi_columns(character()), ansi_string(character()))

  expect_snapshot_output(
    cat(ansi_columns(paste("foo", 1:10), width = 40), sep = "\n")
  )
  txt60 <- strrep("1234567890", 6)
  expect_snapshot_output(
    cat(ansi_columns(txt60, width = 15), sep = "\n")
  )
})

test_that_cli(configs = c("plain", "ansi"), "ansi_toupper", {
  x <- paste0(
    col_red("Red "),
    "normal ",
    style_bold(col_green("green"))
  )
  expect_snapshot(local({
    cat_line(x)
    cat_line(ansi_toupper(x))
  }))
})

test_that_cli(configs = c("plain", "ansi"), "ansi_tolower", {
  x <- paste0(
    col_red("Red "),
    "NORMAL ",
    style_bold(col_green("grEeN"))
  )
  expect_snapshot(local({
    cat_line(x)
    cat_line(ansi_tolower(x))
  }))
})

test_that_cli(configs = c("plain", "ansi"), "ansi_chartr", {
  x <- paste0(
    col_red("Red "),
    "normal ",
    style_bold(col_green("green"))
  )
  expect_snapshot(local({
    cat_line(x)
    cat_line(ansi_chartr(" R_", "-r*", x))
  }))
})

test_that("ansi_align", {
  txt0 <- "\033]8;;https://ex.com\007te\033]8;;\007"
  txt1 <- st_from_bel(txt0)
  txt <- paste0(txt0, txt1)
  space3 <- strrep(" ", 3)
  space6 <- strrep(" ", 6)
  expect_equal(
    ansi_align(txt, 10),
    ansi_string(paste0(txt, space6))
  )

  expect_equal(
    ansi_align(txt, 10, "center"),
    ansi_string(paste0(space3, txt, space3))
  )

  expect_equal(
    ansi_align(txt, 10, "right"),
    ansi_string(paste0(space6, txt))
  )
})

test_that("ansi_chartr", {
  txt0 <- "1\033]8;;https://ex.com\007text\033]8;;\0072"
  txt1 <- st_from_bel(txt0)
  txt <- paste0(txt0, txt1)
  expect_equal(
    ansi_chartr("12x0", "34yo", txt),
    ansi_string(paste0(
      "3\033]8;;https://ex.com\007teyt\033]8;;\0074",
      st_from_bel("3\033]8;;https://ex.com\007teyt\033]8;;\0074")
    ))
  )
})

test_that("ansi_columns", {
  txt <- "\033]8;;https://ex.com\007text\033]8;;\007"

  expect_equal(
    ansi_strip(ansi_columns(rep(txt, 4), 10)),
    c("text text ", "text text ")
  )
})

test_that("ansi_has_any", {
  txt <- "1\033]8;;https://ex.com\007text\033]8;;\0072"
  expect_true(ansi_has_any(txt))
  expect_true(ansi_has_any(txt, sgr = FALSE, csi = FALSE))
  expect_false(ansi_has_any(txt, link = FALSE))
})

test_that("ansi_html", {
  txt <- paste0(
    "1",
    "\033[1m\033]8;;https://ex.com\007",
    "text",
    "\033]8;;\007",
    "\033[22m",
    "2"
  )

  expect_equal(
    ansi_html(txt),
    paste0(
      "1",
      "<a class=\"ansi-link\" href=\"https://ex.com\">",
      "<span class=\"ansi ansi-bold\">",
      "text",
      "</span>",
      "</a>",
      "2"
    )
  )
})

test_that("ansi_nchar", {
  cases <- list(
    list("\033]8;;https://ex.com\007text\033]8;;\007", 4),
    list("\033]8;x=1:y=2;https://ex.com\007text\033]8;;\007", 4),
    list("\033]8;;https://ex.com\007text\033]8;;\007", 4),
    list("\033[1m\033]8;;https://ex.com\007text\033]8;;\007\033[22m", 4)
  )

  for (c in cases) {
    expect_equal(ansi_nchar(c[[1]]), c[[2]])
  }
})

test_that("ansi_regex", {
  cases <- c(
    "1\033]8;;https://ex.com\007text\033]8;;\0072",
    "1\033]8;x=1:y=2;https://ex.com\007text\033]8;;\0072",
    "1\033]8;;https://ex.com\007text\033]8;;\0072",
    "1\033[1m\033]8;;https://ex.com\007text\033]8;;\007\033[22m2"
  )

  for (case in cases) {
    expect_equal(gsub(ansi_regex(), "", case, perl = TRUE), "1text2")
  }
})

test_that("ansi_simplify", {
  txt <- "1\033[1m\033]8;;https://ex.com\007text\033]8;;\007\033[22m2"
  expect_equal(
    ansi_simplify(txt),
    ansi_string(
      "1\033]8;;https://ex.com\007\033[1mtext\033[22m\033]8;;\0072"
    )
  )
})

test_that("ansi_strip", {
  cases <- c(
    "1\033]8;;https://ex.com\007text\033]8;;\0072",
    "1\033]8;x=1:y=2;https://ex.com\007text\033]8;;\0072",
    "1\033]8;;https://ex.com\007text\033]8;;\0072",
    "1\033[1m\033]8;;https://ex.com\007text\033]8;;\007\033[22m2"
  )

  for (case in cases) {
    expect_equal(ansi_strip(case), "1text2")
  }

  txt <- "1\033[1m\033]8;;https://ex.com\007text\033]8;;\007\033[22m2"
  expect_equal(
    ansi_strip(txt, link = FALSE),
    "1\033]8;;https://ex.com\007text\033]8;;\0072"
  )
  expect_equal(
    ansi_strip(txt, sgr = FALSE),
    "1\033[1mtext\033[22m2"
  )
})

test_that("ansi_strsplit", {
  txt <- "1\033]8;;https://ex.com\007te;xt\033]8;;\0072"
  expect_equal(
    ansi_strsplit(txt, ";"),
    list(ansi_string(c(
      "1\033]8;;https://ex.com\007te\033]8;;\007",
      "\033]8;;https://ex.com\007xt\033]8;;\0072"
    )))
  )
})

test_that("ansi_strtrim", {
  txt <- "1\033]8;;https://ex.com\007text\033]8;;\0072"
  expect_equal(
    ansi_strtrim(txt, 6),
    ansi_string(txt)
  )
  expect_equal(
    ansi_strtrim(txt, 4),
    ansi_string(paste0(ansi_substr(txt, 1, 1), "..."))
  )
})

test_that("ansi_strwrap", {
  txt <- "1\033]8;;https://ex.com\007text\033]8;;\0072"
  wrp <- ansi_strwrap(strrep(paste0(txt, " "), 10), 15)
  expect_equal(
    wrp,
    ansi_string(rep(paste0(txt, " ", txt), 5))
  )
})

test_that("ansi_substr", {
  txt <- "1\033[1m\033]8;;https://ex.com\007text\033]8;;\007\033[22m2"

  cases <- list(
    list(1, 3, "1\033]8;;https://ex.com\007\033[1mte\033[22m\033]8;;\007"),
    list(4, 6, "\033]8;;https://ex.com\007\033[1mxt\033[22m\033]8;;\0072"),
    list(3, 4, "\033]8;;https://ex.com\007\033[1mex\033[22m\033]8;;\007"),
    list(1, 10, "1\033]8;;https://ex.com\007\033[1mtext\033[22m\033]8;;\0072")
  )

  for (c in cases) {
    expect_equal(ansi_substr(txt, c[[1]], c[[2]]), ansi_string(c[[3]]))
  }
})

test_that("ansi_substring", {
  txt <- "1\033[1m\033]8;;https://ex.com\007text\033]8;;\007\033[22m2"

  cases <- list(
    list(1, 3, "1\033]8;;https://ex.com\007\033[1mte\033[22m\033]8;;\007"),
    list(4, 6, "\033]8;;https://ex.com\007\033[1mxt\033[22m\033]8;;\0072"),
    list(3, 4, "\033]8;;https://ex.com\007\033[1mex\033[22m\033]8;;\007"),
    list(1, 10, "1\033]8;;https://ex.com\007\033[1mtext\033[22m\033]8;;\0072")
  )

  for (c in cases) {
    expect_equal(ansi_substring(txt, c[[1]], c[[2]]), ansi_string(c[[3]]))
  }
})

test_that("ansi_tolower", {
  txt <- "Pre\033]8;;https://ex.com\007tEXt\033]8;;\007PoST"
  expect_equal(
    ansi_tolower(txt),
    ansi_string("pre\033]8;;https://ex.com\007text\033]8;;\007post")
  )
})

test_that("ansi_toupper", {
  txt <- "Pre\033]8;;https://ex.com\007tEXt\033]8;;\007PoST"
  expect_equal(
    ansi_toupper(txt),
    ansi_string("PRE\033]8;;https://ex.com\007TEXT\033]8;;\007POST")
  )
})

test_that("ansi_trimws", {
  txt <- "1\033[1m\033]8;;https://ex.com\007text\033]8;;\007\033[22m2"
  expect_equal(
    ansi_trimws(paste0("   ", txt, "   ")),
    ansi_simplify(txt)
  )
  expect_equal(
    ansi_trimws(txt),
    ansi_string(txt)
  )
})

test_that("unknown hyperlink type", {
  expect_snapshot(
    error = TRUE,
    make_link("this", "foobar")
  )
})

test_that("iterm file links", {
  withr::local_envvar(R_CLI_HYPERLINK_STYLE = "iterm")
  withr::local_options(cli.hyperlink = TRUE)
  expect_snapshot({
    cli::cli_text("{.file /path/to/file:10}")
    cli::cli_text("{.file /path/to/file:10:20}")
  })
})

test_that("rstudio links", {
  withr::local_envvar(
    RSTUDIO = "1",
    RSTUDIO_SESSION_PID = Sys.getpid(),
    RSTUDIO_CHILD_PROCESS_PANE = "build",
    RSTUDIO_CLI_HYPERLINKS = "1"
  )
  withr::local_options(
    cli.hyperlink = TRUE,
    cli.hyperlink_help = TRUE,
    cli.hyperlink_run = TRUE,
    cli.hyperlink_vignette = TRUE
  )
  expect_snapshot(
    cli::cli_text("{.fun pkg::fun}")
  )
  expect_snapshot(
    cli::cli_text("{.help fun}")
  )
  expect_snapshot(
    cli::cli_text("{.run package::func()}")
  )
  expect_snapshot(
    cli::cli_text("{.vignette package::title}")
  )
  expect_snapshot(
    cli::cli_text("{.topic pkg::topic}")
  )
})

test_that("ST hyperlinks", {
  withr::local_envvar(R_CLI_HYPERLINK_MODE = "posix")
  withr::local_options(cli.hyperlink = TRUE)
  expect_snapshot(
    cat(style_hyperlink("text", "https://example.com"))
  )
})

test_that("ansi_has_hyperlink_support", {
  local_clean_cli_context()

  # force with env var
  withr::with_envvar(list(R_CLI_HYPERLINKS = "true"),
    expect_true(ansi_has_hyperlink_support())
  )

  # if no ansi support, then no
  mockery::stub(ansi_has_hyperlink_support, "num_ansi_colors", 256L)
  expect_false(ansi_has_hyperlink_support())

  # are we in rstudio with support?
  mockery::stub(ansi_has_hyperlink_support, "num_ansi_colors", 256L)
  mockery::stub(ansi_has_hyperlink_support, "rstudio_detect",
                list(type = "rstudio_console", hyperlink = TRUE))
  expect_true(ansi_has_hyperlink_support())
})

test_that("ansi_has_hyperlink_support 2", {
  local_clean_cli_context()
  mockery::stub(ansi_has_hyperlink_support, "num_ansi_colors", 256L)

  mockery::stub(ansi_has_hyperlink_support, "isatty", FALSE)
  expect_false(ansi_has_hyperlink_support())
})

test_that("ansi_has_hyperlink_support 3", {
  local_clean_cli_context()
  mockery::stub(ansi_has_hyperlink_support, "num_ansi_colors", 256L)

  mockery::stub(ansi_has_hyperlink_support, "isatty", TRUE)
  mockery::stub(ansi_has_hyperlink_support, "is_windows", TRUE)
  withr::local_envvar(WT_SESSION = "4c464723-f51f-4612-83f7-31e1c75abd83")
  expect_true(ansi_has_hyperlink_support())
})

test_that("ansi_has_hyperlink_support 4", {
  local_clean_cli_context()
  mockery::stub(ansi_has_hyperlink_support, "num_ansi_colors", 256L)
  mockery::stub(ansi_has_hyperlink_support, "isatty", TRUE)

  withr::local_envvar("CI" = "true")
  expect_false(ansi_has_hyperlink_support())

  withr::local_envvar("CI" = NA_character_, TEAMCITY_VERSION = "1")
  expect_false(ansi_has_hyperlink_support())
})

test_that("ansi_has_hyperlink_support 5", {
  local_clean_cli_context()
  mockery::stub(ansi_has_hyperlink_support, "num_ansi_colors", 256L)
  mockery::stub(ansi_has_hyperlink_support, "isatty", TRUE)

  withr::local_envvar(
    TERM_PROGRAM = "iTerm.app",
    TERM_PROGRAM_VERSION = "3.4.16"
  )
  expect_true(ansi_has_hyperlink_support())
})

test_that("ansi_has_hyperlink_support 5", {
  local_clean_cli_context()
  mockery::stub(ansi_has_hyperlink_support, "num_ansi_colors", 256L)
  mockery::stub(ansi_has_hyperlink_support, "isatty", TRUE)

  withr::local_envvar(VTE_VERSION = "0.51.1")
  expect_true(ansi_has_hyperlink_support())

  withr::local_envvar(VTE_VERSION = "5110")
  expect_true(ansi_has_hyperlink_support())

  withr::local_envvar(VTE_VERSION = "foo")
  expect_false(ansi_has_hyperlink_support())
})

test_that("ansi_hyperlink_types", {
  local_clean_cli_context()
  withr::local_envvar(
    R_CLI_HYPERLINKS = "true",
    R_CLI_HYPERLINK_RUN = "true"
  )
  expect_true(ansi_hyperlink_types()[["run"]])
})

# To create the data:

update_wide_unicode_data <- function() {
  tab <- read.delim(
    stringsAsFactors = FALSE,
    "https://unicode.org/Public/UNIDATA/EastAsianWidth.txt",
    comment.char = "#",
    sep = ";",
    strip.white = TRUE,
    header = FALSE
  )

  # Keep wide ones
  wide <- tab$V1[tab$V2 == "W"]
  first <- sub("\\.\\..*$", "", wide)
  range <- sub("^([0-9A-F]+)\\.\\.([0-9A-F]+)$", "\\\\U\\1-\\\\U\\2", wide)
  range <- sub("^([0-9A-F]+)$", "\\\\U\\1", range)

  wide_chars <- data.frame(
    stringsAsFactors = FALSE,
    test = sapply(parse(text = paste0('"', "\\U", first, '"')), eval),
    regex = sapply(parse(text = paste0('"', range, '"')), eval)
  )

  env <- new.env(parent = emptyenv())
  load("R/sysdata.rda", envir = env)

  env$wide_chars <- wide_chars
  save(list = ls(env), file = "R/sysdata.rda", envir = env, version = 2)
}

#' A simple CLI theme
#'
#' To use this theme, you can set it as the `cli.theme` option.
#' Note that this is in addition to the builtin theme, which is still in
#' effect.
#'
#' ```r
#' options(cli.theme = cli::simple_theme())
#' ```
#'
#' and then CLI apps started after this will use it as the default theme.
#' You can also use it temporarily, in a div element:
#'
#' ```r
#' cli_div(theme = cli::simple_theme())
#' ```
#'
#' # Showcase
#'
#' ```{asciicast simple-theme}
#' show <- cli_div(theme = cli::simple_theme())
#'
#' cli_h1("Heading 1")
#' cli_h2("Heading 2")
#' cli_h3("Heading 3")
#'
#' cli_par()
#' cli_alert_danger("Danger alert")
#' cli_alert_warning("Warning alert")
#' cli_alert_info("Info alert")
#' cli_alert_success("Success alert")
#' cli_alert("Alert for starting a process or computation",
#'   class = "alert-start")
#' cli_end()
#'
#' cli_text("Packages and versions: {.pkg cli} {.version 1.0.0}.")
#' cli_text("Time intervals: {.timestamp 3.4s}")
#'
#' cli_text("{.emph Emphasis} and  {.strong strong emphasis}")
#'
#' cli_text("This is a piece of code: {.code sum(x) / length(x)}")
#' cli_text("Function names: {.fn cli::simple_theme}")
#'
#' cli_text("Files: {.file /usr/bin/env}")
#' cli_text("URLs: {.url https://r-project.org}")
#'
#' cli_h2("Longer code chunk")
#' cli_par(class = "code R")
#' cli_verbatim(
#'   '# window functions are useful for grouped mutates',
#'   'mtcars %>%',
#'   '  group_by(cyl) %>%',
#'   '  mutate(rank = min_rank(desc(mpg)))')
#'
#' cli_end(show)
#' ```
#'
#' @param dark Whether the theme should be optimized for a dark
#'   background. If `"auto"`, then cli will try to detect this.
#'   Detection usually works in recent RStudio versions, and in iTerm
#'   on macOS, but not on other platforms.
#'
#' @seealso [themes], [builtin_theme()].
#' @export

simple_theme <- function(dark = getOption("cli.theme_dark", "auto")) {

  dark <- detect_dark_theme(dark)

  list(
    h1 = list(
      "margin-top" = 1,
      "margin-bottom" = 0,
      color = "cyan",
      fmt = function(x) cli::rule(x, line_col = "cyan")),

    h2 = list(
      "margin-top" = 1,
      "margin-bottom" = 0,
      color = "cyan",
      fmt = function(x) paste0(symbol$line, " ", x, " ", symbol$line, symbol$line)),

    h3 = list(
      "margin-top" = 1,
      "margin-bottom" = 0,
      color = "cyan"),

    par = list("margin-top" = 0, "margin-bottom" = 1),

    ".alert-danger" = list(
      "background-color" = "red",
      color = "white",
      before = function() paste0(symbol$cross, " ")
    ),

    ".alert-warning" = list(
      color = "orange",
      "font-weight" = "bold",
      before = paste0("!", " ")
    ),

    ".alert-success" = list(
      before = function() paste0(col_green(symbol$tick), " ")
    ),
    ".alert-info" = list(
      before = function() paste0(col_cyan(symbol$info), " ")
    ),

    ".alert-start" = list(
      before = function() paste0(symbol$arrow_right, " ")),

    span.pkg = list(
      color = "blue",
      "font-weight" = "bold"),
    span.version = list(color = "blue"),

    span.emph = simple_theme_emph(),
    span.strong = list("font-weight" = "bold", "font-style" = "italic"),

    span.fun = utils::modifyList(simple_theme_code(dark), list(after = "()")),
    span.fn = utils::modifyList(simple_theme_code(dark), list(after = "()")),
    span.arg = simple_theme_code(dark),
    span.kbd = utils::modifyList(simple_theme_code(dark),
                          list(before = "<", after = ">")),
    span.key = utils::modifyList(simple_theme_code(dark),
                          list(before = "<", after = ">")),
    span.file = simple_theme_file(),
    span.path = simple_theme_file(),
    span.email = simple_theme_url(),
    span.url = utils::modifyList(simple_theme_url(),
                          list(before = "<", after = ">")),
    span.var = simple_theme_code(dark),
    span.envvar = simple_theme_code(dark),

    span.timestamp = list(before = "[", after = "]", color = "grey")
  )
}

simple_theme_emph <- function() {
  list("font-style" = "italic")
}

simple_theme_url <- function() {
  list(color = "blue")
}

simple_theme_code <- function(dark) {
  if (dark) {
    list("background-color" = "#232323", color = "#f0f0f0")
  } else{
    list("background-color" = "#f8f8f8", color = "#202020")
  }
}

simple_theme_file <- function() {
  list(color = "blue")
}

simple_theme_r_code <- function(dark) {
  dark <- dark
  style <- if (dark) {
    make_ansi_style("#f0f0f0")
  } else {
    make_ansi_style("#202020")
  }
  function(x) {
    x <- ansi_strip(x)
    lines <- strsplit(x, "\n", fixed = TRUE)[[1]]
    fmd <- code_highlight(lines)
    style(fmd)
  }
}

is_iterm <- function() {
  isatty(stdout()) && Sys.getenv("TERM_PROGRAM", "") == "iTerm.app"
}

is_iterm_dark <- function() {
  if (is.null(clienv[["is_iterm_dark"]])) {
    clienv$is_iterm_dark <-
      tryCatch(
        error = function(x) FALSE, {
          osa <- '
            tell application "iTerm2"
              tell current session of current window
                get background color
              end tell
            end tell
          '
          out <- suppressWarnings(system2(
            "osascript",
            c("-e", shQuote(osa)),
            stdout = TRUE,
            stderr = TRUE
          ))
          nums <- scan(text = gsub(",", "", out, fixed = TRUE), quiet = TRUE)
          mean(nums) < 20000
        })
  }
  clienv[["is_iterm_dark"]]
}

#! /usr/bin/env Rscript

## To get the pkgcache package:
## source("https://install-github.me/r-lib/pkgcache")

setup_app <- function() {
  theme <- list(
    "url" = list(color = "blue"),
    ".pkg" = list(color = "orange"))
  start_app(theme = theme, output = "stdout")
}

load_packages <- function() {
  tryCatch(suppressPackageStartupMessages({
    library(cli)
    library(pkgcache)
    library(docopt) }),
    error = function(e) {
      cli_alert_danger("The {.pkg pkgcache} and {.pkg docopt} packages are needed!")
      q(save = "no", status = 1)
    })
}

outdated <- function(lib = NULL, notcran = FALSE) {
  load_packages()
  setup_app()
  if (is.null(lib)) lib <- .libPaths()[1]
  inst <- utils::installed.packages(lib = lib)
  cli_alert_info("Getting repository metadata")
  repo <- meta_cache_list(rownames(inst))

  if (!notcran) inst <- inst[inst[, "Package"] %in% repo$package, ]

  for (i in seq_len(nrow(inst))) {
    pkg <- inst[i, "Package"]
    iver <- inst[i, "Version"]

    if (! pkg %in% repo$package) {
      cli_alert_info("{.pkg {pkg}}: \tnot a CRAN/BioC package")
      next
    }

    rpkg <- repo[repo$package == pkg, ]
    newer <- rpkg[package_version(rpkg$version) > iver, ]
    if (!nrow(newer)) next

    nver <- package_version(newer$version)
    mnver <- max(nver)
    newest <- newer[mnver == nver, ]
    bin <- if (any(newest$platform != "source")) "bin" else ""
    src <- if (any(newest$platform == "source")) "src" else ""

    cli_alert_danger("{.emph {pkg}}")
    cli_alert_danger(
          "{.pkg {pkg}} \t{iver} {symbol$arrow_right} {mnver}  {.emph ({bin} {src})}")
  }
}

parse_arguments <- function() {
  "Usage:
  outdated.R [-l lib] [-x]
  outdated.R -h | --help

Options:
  -x         Print not CRAN/BioC packages as well
  -l lib     Library directory, default is first directory in the lib path
  -h --help  Print this help message

Check for outdated packages in a package library.
  " -> doc

  docopt(doc)
}

if (is.null(sys.calls())) {
  load_packages()
  opts <- parse_arguments()
  outdated(opts$l, opts$x)
}

# modified from https://shiny.rstudio.com/articles/progress.html

# Nested progress bars

library(cli)

# !!! You don't need these in real code.
# cli.progress_show_after makes sure that we see the progress bar from
# the beginning, not only after a delay.
options(cli.progress_show_after = 0)

# !!! You don't need these in real code.
# This also requests logging the progress bar to the standard output,
# in the console.
options(cli.progress_handlers_only = c("shiny", "logger"))

server <- function(input, output) {
  output$plot <- renderPlot({
    input$goPlot # Re-run when button is clicked
    
    # Create 0-row data frame which will be used to store data
    dat <- data.frame(x = numeric(0), y = numeric(0))
    
    # Number of times we'll go through the loop
    n <- 10

    # Set auto-terminate to FALSE, as we do not want the progress bar
    # terminated when we reach step 10, as at that point we are only at
    # the beginning of the iteration
    cli_progress_bar(
      "Rendering plot",
      total = n,
      format = "Starting {i}"
    )
    for (i in 1:n) {
      cli_progress_update(set = i - 1)
      # Each time through the loop, add another row of data. This is
      # a stand-in for a long-running computation.
      dat <- rbind(dat, data.frame(x = rnorm(1), y = rnorm(1)))
      render_plot_detail(i)
    }

    plot(dat$x, dat$y)
  })
}

ui <- shinyUI(basicPage(
  plotOutput('plot', width = "300px", height = "300px"),
  actionButton('goPlot', 'Go plot')
))

render_plot_detail <- function(i) {
  cli_progress_bar(total = 10, paste("Detailing", i))
  for (i in 1:10) {
    cli_progress_update()
    Sys.sleep(0.1)
  }
}

shinyApp(ui = ui, server = server)

make_line <- function(x, char = symbol$line, col = NULL) {

  ## Easiest to handle this specially
  if (x <= 0) return("")

  cw <- ansi_nchar(char, "width")

  ## We handle the simple case differently, to make it faster
  if (cw == 1) {
    line <- paste(rep(char, x), collapse = "")
  } else {
    line <- ansi_substr(paste(rep(char, ceiling(x / cw)), collapse = ""), 1, x)
  }

  unclass(apply_style(line, col))
}

#' Make a rule with one or two text labels
#'
#' @description
#' The rule can include either a centered text label, or labels on the
#' left and right side.
#'
#' To color the labels, use the functions `col_*`, `bg_*` and `style_*`
#' functions, see [ANSI styles][ansi-styles], and the examples below.
#' To color the line, either these functions directly, or the `line_col`
#' option.
#'
#'
#' @details
#' ## Simple rule
#'
#' ```{asciicast rule-simple}
#' rule()
#' ```
#'
#' ## Line styles
#' Some strings for the `line` argument are interpreted specially:
#'
#' * `"single"`: (same as `1`), a single line,
#' * `"double"`: (same as `2`), a double line,
#' * `"bar1"`, `"bar2"`, `"bar3"`, etc., `"bar8"` uses varying height bars.
#'
#' ### Double rule
#'
#' ```{asciicast rule-double}
#' rule(line = 2)
#' ```
#'
#' ### Bars
#'
#' ```{asciicast rule-bars}
#' rule(line = "bar2")
#' rule(line = "bar5")
#' ```
#'
#' ### Custom lines
#'
#' ```{asciicast rule-custom-line}
#' rule(center = "TITLE", line = "~")
#' ```
#'
#' ```{asciicast rule-custom-line-2}
#' rule(center = "TITLE", line = col_blue("~-"))
#' ```
#'
#' ```{asciicast rule-custom-line-3}
#' rule(center = bg_red(" ", symbol$star, "TITLE",
#'   symbol$star, " "),
#'   line = "\u2582",
#'   line_col = "orange")
#' ```
#'
#' ## Left label
#'
#' ```{asciicast rule-left-label}
#' rule(left = "Results")
#' ```
#'
#' ## Centered label
#'
#' ```{asciicast rule-center-label}
#' rule(center = " * RESULTS * ")
#' ```
#'
#' ## Colored labels
#'
#' ```{asciicast rule-colored-label}
#' rule(center = col_red(" * RESULTS * "))
#' ```
#'
#' ## Colored line
#'
#' ```{asciicast rule-colored-line}
#' rule(center = col_red(" * RESULTS * "), line_col = "red")
#' ```
#'
#' @param left Label to show on the left. It interferes with the `center`
#'   label, only at most one of them can be present.
#' @param center Label to show at the center. It interferes  with the
#'   `left` and `right` labels.
#' @param right Label to show on the right. It interferes with the `center`
#'   label, only at most one of them can be present.
#' @param line The character or string that is used to draw the line.
#'   It can also `1` or `2`, to request a single line (Unicode, if
#'   available), or a double line. Some strings are interpreted specially,
#'   see *Line styles* below.
#' @param col Color of text, and default line color. Either an ANSI style
#'   function (see [ANSI styles][ansi-styles]), or a color name that is
#'   passed to [make_ansi_style()].
#' @param line_col,background_col Either a color name (used in
#'   [make_ansi_style()]), or a style function (see
#'   [ANSI styles][ansi-styles]), to color the line and background.
#' @param width Width of the rule. Defaults to the `width` option, see
#'   [base::options()].
#' @return Character scalar, the rule.
#'
#' @export

rule <- function(left = "", center = "", right = "", line = 1,
                 col = NULL, line_col = col, background_col = NULL,
                 width = console_width()) {

  try_silently(left <- as.character(left))
  try_silently(center <- as.character(center))
  try_silently(right <- as.character(right))

  stopifnot(
    is_string(left),
    is_string(center),
    is_string(right),
    is_string(line) || line == 1 || line == 2,
    is_col(col),
    is_col(line_col),
    is_count(width)
  )

  left <- apply_style(left, col)
  center <- apply_style(center, col)
  right <- apply_style(right, col)

  options <- as.list(environment())
  options$line <- get_line_char(options$line)

  res <- if (ansi_nchar(center)) {
    if (ansi_nchar(left) || ansi_nchar(right)) {
      stop(sQuote("center"), " cannot be specified with ", sQuote("left"),
           " or ", sQuote("right"))
    }
    rule_center(options)

  } else if (ansi_nchar(left) && ansi_nchar(right)) {
    rule_left_right(options)

  } else if (ansi_nchar(left)) {
    rule_left(options)

  } else if (ansi_nchar(right)) {
    rule_right(options)

  } else {
    rule_line(options)
  }

  res <- ansi_substr(res, 1, width)
  res <- apply_style(res, background_col, bg = TRUE)

  class(res) <- unique(c("cli_rule", "rule", class(res), "character"))
  res
}

get_line_char <- function(line) {
  if (identical(line, 1) || identical(line, 1L) || identical(line, "single")) {
    symbol$line

  } else if (identical(line, 2) || identical(line, 2L) || identical(line, "double")) {
    symbol$double_line

  } else if (length(line) == 1 && line %in% paste0("bar", 1:8)) {
    bars <- structure(
      paste0("lower_block_", 1:8),
      names = paste0("bar", 1:8)
    )
    symbol[[ bars[[line]] ]]

  } else {
    paste(as.character(line), collapse = "")
  }
}

rule_line <- function(o) {
  make_line(o$width, o$line, o$line_col)
}

rule_center <- function(o) {

  o$center <- ansi_substring(o$center, 1, o$width - 4)
  o$center <- paste0(" ", o$center, " ")
  ncc <- ansi_nchar(o$center, "width")

  ndashes <- o$width - ncc

  paste0(
    make_line(ceiling(ndashes / 2), o$line, o$line_col),
    o$center,
    make_line(floor(ndashes / 2), o$line, o$line_col)
  )
}

rule_left <- function(o) {
  ncl <- ansi_nchar(o$left, "width")

  paste0(
    make_line(2, get_line_char(o$line), o$line_col),
    " ", o$left, " ",
    make_line(o$width - ncl - 4, o$line, o$line_col)
  )
}

rule_right <- function(o) {
  ncr <- ansi_nchar(o$right, "width")

  paste0(
    make_line(o$width - ncr - 4, o$line, o$line_col),
    " ", o$right, " ",
    make_line(2, o$line, o$line_col)
  )
}

rule_left_right <- function(o) {

  ncl <- ansi_nchar(o$left, "width")
  ncr <- ansi_nchar(o$right,  "width")

  ## -- (ncl) -- (ncr) --
  if (ncl + ncr + 10 > o$width) return(rule_left(o))

  paste0(
    make_line(2, o$line, o$line_col),
    " ", o$left, " ",
    make_line(o$width - ncl - ncr - 8, o$line, o$line_col),
    " ", o$right, " ",
    make_line(2, o$line, o$line_col)
  )
}

methods::setOldClass(c("cli_rule", "character"))

#' @export

print.cli_rule <- function(x, ..., sep = "\n") {
  cat(x, ..., sep = sep)
  invisible(x)
}

test_that_cli("make_line", {

  expect_equal(make_line(1, "-"), "-")
  expect_equal(make_line(0, "-"), "")
  expect_equal(make_line(2, "-"), "--")
  expect_equal(make_line(10, "-"), "----------")

  expect_equal(make_line(2, "12"), "12")
  expect_equal(make_line(0, "12"), "")
  expect_equal(make_line(1, "12"), "1")
  expect_equal(make_line(9, "12"), "121212121")
  expect_equal(make_line(10, "12"), "1212121212")
})

test_that("width option", {
  expect_equal(
    rule(width = 11, line = "-"),
    rule_class("-----------")
  )
})

test_that("left label", {

  expect_equal(
    rule("label", width = 12, line = "-"),
    rule_class("-- label ---")
  )
  expect_equal(
    rule("l", width = 12, line = "-"),
    rule_class("-- l -------")
  )
  expect_equal(
    rule("label", width = 9, line = "-"),
    rule_class("-- label ")
  )
  expect_equal(
    rule("label", width = 8, line = "-"),
    rule_class("-- label")
  )
  expect_equal(
    rule("label", width = 6, line = "-"),
    rule_class("-- lab")
  )
})

test_that("centered label", {

  expect_error(
    rule(left = "label", center = "label"),
    "cannot be specified"
  )
  expect_error(
    rule(center = "label", right = "label"),
    "cannot be specified"
  )

  expect_equal(
    rule(center = "label", width = 13, line = "-"),
    rule_class("--- label ---")
  )
  expect_equal(
    rule(center = "label", width = 14, line = "-"),
    rule_class("---- label ---")
  )
  expect_equal(
    rule(center = "label", width = 9, line = "-"),
    rule_class("- label -")
  )
  expect_equal(
    rule(center = "label", width = 8, line = "-"),
    rule_class("- labe -")
  )
  expect_equal(
    rule(center = "label", width = 7, line = "-"),
    rule_class("- lab -")
  )
})

test_that("right label", {
  expect_equal(
    rule(right = "label", width = 12, line = "-"),
    rule_class("--- label --")
  )
  expect_equal(
    rule(right = "l", width = 12, line = "-"),
    rule_class("------- l --")
  )
  expect_equal(
    rule(right = "label", width = 9, line = "-"),
    rule_class(" label --")
  )
  expect_equal(
    rule(right = "label", width = 8, line = "-"),
    rule_class(" label -")
  )
  expect_equal(
    rule(right = "label", width = 6, line = "-"),
    rule_class(" label")
  )
  expect_equal(
    rule(right = "label", width = 5, line = "-"),
    rule_class(" labe")
  )
  expect_equal(
    rule(right = "label", width = 4, line = "-"),
    rule_class(" lab")
  )
})

test_that("line_col", {

  withr::with_options(
    list(cli.num_colors = 256L), {
      expect_true(ansi_has_any(
        rule(line_col = "red")
      ))
      expect_true(ansi_has_any(
        rule(left = "left", line_col = "red")
      ))
      expect_true(ansi_has_any(
        rule(left = "left", right = "right", line_col = "red")
      ))
      expect_true(ansi_has_any(
        rule(center = "center", line_col = "red")
      ))
      expect_true(ansi_has_any(
        rule(right = "right", line_col = "red")
      ))

      expect_true(ansi_has_any(
        rule(line_col = col_red)
      ))
    }
  )
})

test_that_cli("get_line_char", {
  expect_equal(get_line_char(1), cli::symbol$line)
  expect_equal(get_line_char(2), cli::symbol$double_line)

  expect_equal(get_line_char("bar1"), cli::symbol$lower_block_1)
  expect_equal(get_line_char("bar2"), cli::symbol$lower_block_2)
  expect_equal(get_line_char("bar3"), cli::symbol$lower_block_3)
  expect_equal(get_line_char("bar4"), cli::symbol$lower_block_4)
  expect_equal(get_line_char("bar5"), cli::symbol$lower_block_5)
  expect_equal(get_line_char("bar6"), cli::symbol$lower_block_6)
  expect_equal(get_line_char("bar7"), cli::symbol$lower_block_7)
  expect_equal(get_line_char("bar8"), cli::symbol$lower_block_8)

  expect_equal(get_line_char("xxx"), "xxx")
  expect_equal(get_line_char(c("x", "y", "z")), "xyz")
})

test_that("print.cli_rule", {
  withr::local_options(cli.width = 20)
  expect_snapshot(rule("foo"))
})

context("rules")

test_that("make_line", {

  expect_equal(make_line(1, "-"), "-")
  expect_equal(make_line(0, "-"), "")
  expect_equal(make_line(2, "-"), "--")
  expect_equal(make_line(10, "-"), "----------")

  expect_equal(make_line(2, "12"), "12")
  expect_equal(make_line(0, "12"), "")
  expect_equal(make_line(1, "12"), "1")
  expect_equal(make_line(9, "12"), "121212121")
  expect_equal(make_line(10, "12"), "1212121212")
})

test_that("width option", {
  expect_equal(
    rule(width = 11, line = "-"),
    rule_class("-----------")
  )
})

test_that("left label", {

  expect_equal(
    rule("label", width = 12, line = "-"),
    rule_class("-- label ---")
  )
  expect_equal(
    rule("l", width = 12, line = "-"),
    rule_class("-- l -------")
  )
  expect_equal(
    rule("label", width = 9, line = "-"),
    rule_class("-- label ")
  )
  expect_equal(
    rule("label", width = 8, line = "-"),
    rule_class("-- label")
  )
  expect_equal(
    rule("label", width = 6, line = "-"),
    rule_class("-- lab")
  )
})

test_that("centered label", {

  expect_error(
    rule(left = "label", center = "label"),
    "cannot be specified"
  )
  expect_error(
    rule(center = "label", right = "label"),
    "cannot be specified"
  )

  expect_equal(
    rule(center = "label", width = 13, line = "-"),
    rule_class("--- label ---")
  )
  expect_equal(
    rule(center = "label", width = 14, line = "-"),
    rule_class("---- label ---")
  )
  expect_equal(
    rule(center = "label", width = 9, line = "-"),
    rule_class("- label -")
  )
  expect_equal(
    rule(center = "label", width = 8, line = "-"),
    rule_class("- labe -")
  )
  expect_equal(
    rule(center = "label", width = 7, line = "-"),
    rule_class("- lab -")
  )
})

test_that("right label", {
  expect_equal(
    rule(right = "label", width = 12, line = "-"),
    rule_class("--- label --")
  )
  expect_equal(
    rule(right = "l", width = 12, line = "-"),
    rule_class("------- l --")
  )
  expect_equal(
    rule(right = "label", width = 9, line = "-"),
    rule_class(" label --")
  )
  expect_equal(
    rule(right = "label", width = 8, line = "-"),
    rule_class(" label -")
  )
  expect_equal(
    rule(right = "label", width = 6, line = "-"),
    rule_class(" label")
  )
  expect_equal(
    rule(right = "label", width = 5, line = "-"),
    rule_class(" labe")
  )
  expect_equal(
    rule(right = "label", width = 4, line = "-"),
    rule_class(" lab")
  )
})

test_that("line_col", {

  withr::with_options(
    list(cli.num_colors = 256L), {
      expect_true(ansi_has_any(
        rule(line_col = "red")
      ))
      expect_true(ansi_has_any(
        rule(left = "left", line_col = "red")
      ))
      expect_true(ansi_has_any(
        rule(left = "left", right = "right", line_col = "red")
      ))
      expect_true(ansi_has_any(
        rule(center = "center", line_col = "red")
      ))
      expect_true(ansi_has_any(
        rule(right = "right", line_col = "red")
      ))

      expect_true(ansi_has_any(
        rule(line_col = col_red)
      ))
    }
  )
})

test_that("get_line_char", {
  expect_equal(get_line_char(1), cli::symbol$line)
  expect_equal(get_line_char(2), cli::symbol$double_line)

  expect_equal(get_line_char("bar1"), cli::symbol$lower_block_1)
  expect_equal(get_line_char("bar2"), cli::symbol$lower_block_2)
  expect_equal(get_line_char("bar3"), cli::symbol$lower_block_3)
  expect_equal(get_line_char("bar4"), cli::symbol$lower_block_4)
  expect_equal(get_line_char("bar5"), cli::symbol$lower_block_5)
  expect_equal(get_line_char("bar6"), cli::symbol$lower_block_6)
  expect_equal(get_line_char("bar7"), cli::symbol$lower_block_7)
  expect_equal(get_line_char("bar8"), cli::symbol$lower_block_8)

  expect_equal(get_line_char("xxx"), "xxx")
  expect_equal(get_line_char(c("x", "y", "z")), "xyz")
})

test_that("print.rule", {
  withr::with_options(
    list(cli.width = 20), {
      out <- capt(rule("foo"))
      exp <- rebox("── foo ─────────────")
      expect_equal(out, exp)
    }
  )
})

# modified from https://shiny.rstudio.com/articles/progress.html

# This example uses `cli_progress_along()`.

library(cli)

# !!! You don't need these in real code.
# cli.progress_show_after makes sure that we see the progress bar from
# the beginning, not only after a delay.
options(cli.progress_show_after = 0)

# !!! You don't need these in real code.
# This also requests logging the progress bar to the standard output,
# in the console.
options(cli.progress_handlers_only = c("shiny", "logger"))

server <- function(input, output) {
  output$plot <- renderPlot({
    input$goPlot # Re-run when button is clicked

    # Create 0-row data frame which will be used to store data
    dat <- data.frame(x = numeric(0), y = numeric(0))

    # Number of times we'll go through the loop
    n <- 10

    for (i in cli_progress_along(1:n, "Rendering", format = "At {i}")) {
      # Each time through the loop, add another row of data. This is
      # a stand-in for a long-running computation.
      dat <- rbind(dat, data.frame(x = rnorm(1), y = rnorm(1)))

      # Pause for 0.1 seconds to simulate a long computation.
      Sys.sleep(0.5)
    }

    plot(dat$x, dat$y)
  })
}

ui <- shinyUI(basicPage(
  plotOutput('plot', width = "300px", height = "300px"),
  actionButton('goPlot', 'Go plot')
))

shinyApp(ui = ui, server = server)

#' Debug cli internals
#'
#' Return the current state of a cli app. It includes the currently
#' open tags, their ids, classes and their computed styles.
#'
#' The returned data frame has a print method, and if you want to create
#' a plain data frame from it, index it with an empty bracket:
#' `cli_debug_doc()[]`.
#'
#' To see all currently active themes, use `app$themes`, e.g. for the
#' default app: `default_app()$themes`.
#'
#' @param app The cli app to debug. Defaults to the current app.
#'   if there is no app, then it creates one by calling [start_app()].
#' @return Data frame with columns: `tag`, `id`, `class` (space separated),
#'   theme (id of the theme the element added), `styles` (computed styles
#'   for the element).
#'
#' @seealso [cli_sitrep()]. To debug containers, you can set the
#' `CLI-DEBUG_BAD_END` environment variable to `true`, and then cli will
#' warn when it cannot find the specified container to close (or any
#' contained at all).
#'
#' @examples
#' \dontrun{
#' cli_debug_doc()
#'
#' olid <- cli_ol()
#' cli_li()
#' cli_debug_doc()
#' cli_debug_doc()[]
#'
#' cli_end(olid)
#' cli_debug_doc()
#' }

cli_debug_doc <- function(app = default_app() %||% start_app()) {
  tgs <- vcapply(app$doc, "[[", "tag")
  ids <- vcapply(app$doc, "[[", "id")
  cls <- vcapply(app$doc, function(x) paste(x$class, collapse = " "))
  thm <- lapply(app$doc, function(x) x$theme)

  df <- data.frame(
    stringsAsFactors = FALSE,
    tag    = tgs,
    id     = ids,
    class  = cls,
    theme  = I(as.list(thm)),
    styles = I(as.list(unname(app$styles)))
  )

  class(df) <- c("cli_doc", class(df))
  df
}

#' @export

format.cli_doc <- function(x, ...) {
  nz <- nrow(x) > 0
  c("<cli document>",
    paste0(
      if (nz) "<",
      x$tag,
      if (nz) " id=\"", x$id, if (nz) "\"",
      ifelse (x$class == "", "", paste0(" class=\"", x$class, "\"")),
      if (nz) ">",
      ifelse (vlapply(x$theme, is.null), "", " +theme")
    )
  )
}

#' @export

print.cli_doc <- function(x, ...) {
  cat(format(x, ...), sep = "\n")
  invisible(x)
}

#' @export

`[.cli_doc` <- function(x, ...) {
  class(x) <- setdiff(class(x), "cli_doc")
  NextMethod()
}

#' @title About inline markup in the semantic cli
#'
#' @description
#' To learn how to use cli’s semantic markup, start with the ‘Building
#' a semantic CLI’ article at <https://cli.r-lib.org>.
#'
#' @section Command substitution:
#'
#' All text emitted by cli supports glue interpolation. Expressions
#' enclosed by braces will be evaluated as R code. See [glue::glue()] for
#' details.
#'
#' In addition to regular glue interpolation, cli can also add classes
#' to parts of the text, and these classes can be used in themes. For
#' example
#'
#' ```{asciicast inline-text}
#' cli_text("This is {.emph important}.")
#' ```
#'
#' adds a class to the "important" word, class `"emph"`. Note that in this
#' case the string within the braces is usually not a valid R expression.
#' If you want to mix classes with interpolation, add another pair of
#' braces:
#'
#' ```{asciicast inline-text-2}
#' adjective <- "great"
#' cli_text("This is {.emph {adjective}}.")
#' ```
#'
#' An inline class will always create a `span` element internally. So in
#' themes, you can use the `span.emph` CSS selector to change how inline
#' text is emphasized:
#'
#' ```{asciicast inline-text-3}
#' cli_div(theme = list(span.emph = list(color = "red")))
#' adjective <- "nice and red"
#' cli_text("This is {.emph {adjective}}.")
#' ```
#'
#' @section Classes:
#'
#' The default theme defines the following inline classes:
#' * `arg` for a function argument.
#' * `cls` for an S3, S4, R6 or other class name.
#' * `code` for a piece of code.
#' * `dt` is used for the terms in a definition list ([cli_dl()]).
#' * `dd` is used for the descriptions in a definition list ([cli_dl()]).
#' * `email` for an email address.
#'   If the terminal supports ANSI hyperlinks (e.g. RStudio, iTerm2, etc.),
#'   then cli creates a clickable link.
#'   See [links] for more information about cli hyperlinks.
#' * `emph` for emphasized text.
#' * `envvar` for the name of an environment variable.
#' * `field` for a generic field, e.g. in a named list.
#' * `file` for a file name. If the terminal supports ANSI hyperlinks (e.g.
#'   RStudio, iTerm2, etc.), then cli creates a clickable link that opens
#'   the file in RStudio or with the default app for the file type.
#'   See [links] for more information about cli hyperlinks.
#' * `fn` for a function name. If it is in the `package::function_name`
#'   form, and the terminal supports ANSI hyperlinks (e.g. RStudio,
#'   iTerm2, etc.), then cli creates a clickable link.
#'   See [links] for more information about cli hyperlinks.
#' * `fun` same as `fn`.
#' * `help` is a help page of a _function_.
#'   If the terminal supports ANSI hyperlinks to help pages (e.g. RStudio),
#'   then cli creates a clickable link. It supports link text.
#'   See [links] for more information about cli hyperlinks.
#' * `href` creates a hyperlink, potentially with a link text.
#'   If the terminal supports ANSI hyperlinks (e.g. RStudio, iTerm2, etc.),
#'   then cli creates a clickable link.
#'   See [links] for more information about cli hyperlinks.
#' * `kbd` for a keyboard key.
#' * `key` same as `kbd`.
#' * `obj_type_friendly` formats the type of an R object in a readable way,
#'   and it should be used with `{}`, see an example below.
#' * `or` changes the string that separates the last two elements of
#'   collapsed vectors (see below) from "and" to "or".
#' * `path` for a path (the same as `file` in the default theme).
#' * `pkg` for a package name.
#' * `run` is an R expression, that is potentially clickable if the terminal
#'   supports ANSI hyperlinks to runnable code (e.g. RStudio).
#'   It supports link text. See [links] for more information about cli hyperlinks.
#' * `str` for a double quoted string escaped by [base::encodeString()].
#' * `strong` for strong importance.
#' * `topic` is a help page of a _topic_.
#'   If the terminal supports ANSI hyperlinks to help pages (e.g. RStudio),
#'   then cli creates a clickable link. It supports link text.
#'   See [links] for more information about cli hyperlinks.
#' * `type` formats the type of an R object in a readable way, and it
#'   should be used with `{}`, see an example below.
#' * `url` for a URL. If the terminal supports ANSI hyperlinks (e.g.
#'   RStudio, iTerm2, etc.), then cli creates a clickable link.
#'   See [links] for more information about cli hyperlinks.
#' * `var` for a variable name.
#' * `val` for a generic "value".
#' * `vignette` is a vignette.
#'   If the terminal supports ANSI hyperlinks to help pages (e.g. RStudio),
#'   then cli creates a clickable link. It supports link text.
#'   See [links] for more information about cli hyperlinks.
#'
#' ```{asciicast inline-examples}
#' ul <- cli_ul()
#' cli_li("{.emph Emphasized} text.")
#' cli_li("{.strong Strong} importance.")
#' cli_li("A piece of code: {.code sum(a) / length(a)}.")
#' cli_li("A package name: {.pkg cli}.")
#' cli_li("A function name: {.fn cli_text}.")
#' cli_li("A keyboard key: press {.kbd ENTER}.")
#' cli_li("A file name: {.file /usr/bin/env}.")
#' cli_li("An email address: {.email bugs.bunny@acme.com}.")
#' cli_li("A URL: {.url https://example.com}.")
#' cli_li("An environment variable: {.envvar R_LIBS}.")
#' cli_li("`mtcars` is {.obj_type_friendly {mtcars}}")
#' cli_end(ul)
#' ```
#'
#' You can add new classes by defining them in the theme, and then using
#' them.
#'
#' ```{asciicast inline-newclass}
#' cli_div(theme = list(
#'   span.myclass = list(color = "lightgrey"),
#'   "span.myclass" = list(before = "<<"),
#'   "span.myclass" = list(after = ">>")))
#' cli_text("This is {.myclass in angle brackets}.")
#' cli_end()
#' ```
#'
#' ## Highlighting weird-looking values
#'
#' Often it is useful to highlight a weird file or path name, e.g. one
#' that starts or ends with space characters. The built-in theme does this
#' for `.file`, `.path` and `.email` by default. You can highlight
#' any string inline by adding the `.q` class to it.
#'
#' The current highlighting algorithm
#' * adds single quotes to the string if it does not start or end with an
#'   alphanumeric character, underscore, dot or forward slash.
#' * Highlights the background colors of leading and trailing spaces on
#'   terminals that support ANSI colors.
#'
#' @section Collapsing inline vectors:
#'
#' When cli performs inline text formatting, it automatically collapses
#' glue substitutions, after formatting. This is handy to create lists of
#' files, packages, etc.
#'
#' ```{asciicast inline-collapse}
#' pkgs <- c("pkg1", "pkg2", "pkg3")
#' cli_text("Packages: {pkgs}.")
#' cli_text("Packages: {.pkg {pkgs}}.")
#' ```
#'
#' Class names are collapsed differently by default
#'
#' ```{asciicast inline-collapse-2}
#' x <- Sys.time()
#' cli_text("Hey, {.var x} has class {.cls {class(x)}}.")
#' ```
#'
#' By default cli truncates long vectors. The truncation limit is by default
#' twenty elements, but you can change it with the `vec-trunc` style.
#'
#' ```{asciicast inline-collapse-trunc}
#' nms <- cli_vec(names(mtcars), list("vec-trunc" = 5))
#' cli_text("Column names: {nms}.")
#' ```
#'
#' @section Formatting values:
#'
#' The `val` inline class formats values. By default (c.f. the built-in
#' theme), it calls the [cli_format()] generic function, with the current
#' style as the argument. See [cli_format()] for examples.
#'
#' `str` is for formatting strings, it uses [base::encodeString()] with
#' double quotes.
#'
#' @section Escaping `{` and `}`:
#'
#' It might happen that you want to pass a string to `cli_*` functions,
#' and you do _not_ want command substitution in that string, because it
#' might contain `{` and `}` characters. The simplest solution for this is
#' to refer to the string from a template:
#'
#' ```{asciicast inline-escape}
#' msg <- "Error in if (ncol(dat$y)) {: argument is of length zero"
#' cli_alert_warning("{msg}")
#' ```
#'
#' If you want to explicitly escape `{` and `}` characters, just double
#' them:
#'
#' ```{asciicast inline-escape-2}
#' cli_alert_warning("A warning with {{ braces }}.")
#' ```
#'
#' See also examples below.
#'
#' @section Pluralization:
#'
#' All cli commands that emit text support pluralization. Some examples:
#'
#' ```{asciicast inline-plural}
#' ndirs <- 1
#' nfiles <- 13
#' pkgs <- c("pkg1", "pkg2", "pkg3")
#' cli_alert_info("Found {ndirs} director{?y/ies} and {nfiles} file{?s}.")
#' cli_text("Will install {length(pkgs)} package{?s}: {.pkg {pkgs}}")
#' ```
#'
#' See [pluralization] for details.
#'
#' @section Wrapping:
#'
#' Most cli containers wrap the text to width the container's width,
#' while observing margins requested by the theme.
#'
#' To avoid a line break, you can use the UTF_8 non-breaking space
#' character: `\u00a0`. cli will not break a line here.
#'
#' To force a line break, insert a form feed character: `\f` or
#' `\u000c`. cli will insert a line break there.
#'
#' @name inline-markup
NULL

#' About cli containers
#'
#' Container elements may contain other elements. Currently the following
#' commands create container elements: [cli_div()], [cli_par()], the list
#' elements: [cli_ul()], [cli_ol()], [cli_dl()], and list items are
#' containers as well: [cli_li()].
#'
#' ## Themes
#'
#' A container can add a new theme, which is removed when the container
#' exits.
#'
#' ```{asciicast cnt-theme}
#' d <- cli_div(theme = list(h1 = list(color = "blue",
#'                                     "font-weight" = "bold")))
#' cli_h1("Custom title")
#' cli_end(d)
#' ```
#'
#' ## Auto-closing
#'
#' Container elements are closed with [cli_end()]. For convenience,
#' by default they are closed automatically when the function that created
#' them terminated (either regularly or with an error). The default
#' behavior can be changed with the `.auto_close` argument.
#'
#' ```{asciicast cnt-auto-close}
#' div <- function() {
#'   cli_div(class = "tmp", theme = list(.tmp = list(color = "yellow")))
#'   cli_text("This is yellow")
#' }
#' div()
#' cli_text("This is not yellow any more")
#' ```
#'
#' ## Debugging
#'
#' You can use the internal `cli:::cli_debug_doc()` function to see the
#' currently open containers.
#'
#' ```{asciicast cnt-debug, echo = -1}
#' stop_app()
#' fun <- function() {
#'   cli_div(id = "mydiv")
#'   cli_par(class = "myclass")
#'   cli:::cli_debug_doc()
#' }
#' fun()
#' ```
#'
#' @name containers
NULL

#' About cli themes
#'
#' CLI elements can be styled via a CSS-like language of selectors and
#' properties. Only a small subset of CSS3 is supported, and
#' a lot visual properties cannot be implemented on a terminal, so these
#' will be ignored as well.
#'
#' @section Adding themes:
#' The style of an element is calculated from themes from four sources.
#' These form a stack, and the themes on the top of the stack take
#' precedence, over themes in the bottom.
#'
#' 1. The cli package has a built-in theme. This is always active.
#'    See [builtin_theme()].
#' 2. When an app object is created via [start_app()], the caller can
#'    specify a theme, that is added to theme stack. If no theme is
#'    specified for [start_app()], the content of the `cli.theme` option
#'    is used. Removed when the corresponding app stops.
#' 3. The user may specify a theme in the `cli.user_theme` option. This
#'    is added to the stack _after_ the app's theme (step 2.), so it can
#'    override its settings. Removed when the app that added it stops.
#' 4. Themes specified explicitly in [cli_div()] elements. These are
#'    removed from the theme stack, when the corresponding [cli_div()]
#'    elements are closed.
#'
#' @section Writing themes:
#' A theme is a named list of lists. The name of each entry is a CSS
#' selector. Only a subset of CSS is supported:
#' * Type selectors, e.g. `input` selects all `<input>` elements.
#' * Class selectors, e.g. `.index` selects any element that has a class
#'   of "index".
#' * ID selector. `#toc` will match the element that has the ID "toc".
#' * The descendant combinator, i.e. the space, that selects nodes
#'   that are descendants of the first element. E.g. `div span` will match
#'   all `<span>` elements that are inside a `<div>` element.
#'
#' The content of a theme list entry is another named list, where the
#' names are CSS properties, e.g. `color`, or `font-weight` or
#' `margin-left`, and the list entries themselves define the values of
#' the properties. See [builtin_theme()] and [simple_theme()] for examples.
#'
#' @section Formatter callbacks:
#' For flexibility, themes may also define formatter functions, with
#' property name `fmt`. These will be called once the other styles are
#' applied to an element. They are only called on elements that produce
#' output, i.e. _not_ on container elements.
#'
#' @section Supported properties:
#' Right now only a limited set of properties are supported. These include
#' left, right, top and bottom margins, background and foreground colors,
#' bold and italic fonts, underlined text. The `before` and `after`
#' properties are supported to insert text before and after the
#' content of the element.
#'
#' The current list of properties:
#'
#' * `after`: A string literal to insert after the element. It can also be
#'   a function that returns a string literal. Supported by all inline
#'   elements, list items, alerts and rules.
#' * `background-color`: An R color name, or HTML hexadecimal color.
#'   It can be applied to most elements (inline elements, rules, text,
#'   etc.), but the background of containers is not colored properly
#'   currently.
#' * `before`: A string literal to insert before the element. It can also be
#'   a function that returns a string literal. Supported by all inline
#'   elements, list items, alerts and rules.
#' * `class-map`: Its value can be a named list, and it specifies how
#'   R (S3) class names are mapped to cli class names. E.g.
#'   `list(fs_path = "file")` specifies that `fs_path` objects (from the fs
#'   package) should always print as `.file` objects in cli.
#' * `color`: Text color, an R color name or a HTML hexadecimal color. It
#'   can be applied to most elements that are printed.
#' * `collapse`: Specifies how to collapse a vector, before applying
#'   styling. If a character string, then that is used as the separator.
#'   If a function, then it is called, with the vector as the only
#'   argument.
#' * `digits`: Number of digits after the decimal point for numeric inline
#'   element of class `.val`.
#' * `fmt`: Generic formatter function that takes an input text and returns
#'   formatted text. Can be applied to most elements. If colors are in use,
#'   the input text provided to `fmt` already includes ANSI sequences.
#' * `font-style`: If `"italic"` then the text is printed as cursive.
#' * `font-weight`: If `"bold"`, then the text is printed in boldface.
#' * `line-type`: Line type for [cli_rule()].
#' * `list-style-type`: String literal or functions that returns a string
#'   literal, to be used as a list item marker in un-ordered lists.
#' * `margin-bottom`, `margin-left`, `margin-right`, `margin-top`: Margins.
#' * `padding-left`, `padding-right`: This is currently used the same way
#'   as the margins, but this might change later.
#' * `start`: Integer number, the first element in an ordered list.
#' * `string-quote`: Quoting character for inline elements of class `.val`.
#' * `text-decoration`: If `"underline"`, then underlined text is created.
#' * `text-exdent`: Amount of indentation from the second line of wrapped
#'    text.
#' * `transform`: A function to call on glue substitutions, before
#'   collapsing them. Note that `transform` is applied prior to
#'   implementing color via ANSI sequences.
#' * `vec-last`: The last separator when collapsing vectors.
#' * `vec-sep`: The separator to use when collapsing vectors.
#' * `vec-sep2`: The separator to use for two elements when collapsing
#'   vectors. If not set, then `vec-sep` is used for these as well.
#' * `vec-trunc`: Vectors longer than this will be truncated. Defaults to
#'   100.
#' * `vec-trunc-style`: Select between two ways of collapsing vectors:
#'   - `"both-ends"` is the current default and it shows the beginning and
#'     the end of the vector.
#'   - `"head"` only shows the beginning of the vector.
#'
#' More properties might be added later. If you think that a property is
#' not applied properly to an element, please open an issue about it in
#' the cli issue tracker.
#'
#' @section Examples:
#' Color of headings, that are only active in paragraphs with an
#' 'output' class:
#' ```
#' list(
#'   "par.output h1" = list("background-color" = "red", color = "#e0e0e0"),
#'   "par.output h2" = list("background-color" = "orange", color = "#e0e0e0"),
#'   "par.output h3" = list("background-color" = "blue", color = "#e0e0e0")
#' )
#' ```
#'
#' Create a custom alert type:
#' ```
#' list(
#'   ".alert-start" = list(before = symbol$play),
#'   ".alert-stop"  = list(before = symbol$stop)
#' )
#' ```
#' @name themes

# TODO: examples

NULL

#' cli hyperlinks
#'
#' @description
#' Certain cli styles create clickable links, if your IDE or terminal
#' supports them.
#'
#' # Note: hyperlinks are currently experimental
#'
#' The details of the styles that create hyperlinks will prrobably change
#' in the near future, based on user feedback.
#'
#' # About the links in this manual page
#'
#' The hyperlinks that are included in this manual are demonstrative
#' only, except for the `https:` links. They look like a hyperlink, and
#' you can click on them, but they do nothing. I.e. a `.run` link will
#' not run the linked expression if you click on it.
#'
#' # Hyperlink Support
#'
#' As of today, the latest release of RStudio (version v2022.07.0+548)
#' supports all hyperlink types discussed here. Certain terminals, e.g.
#' iTerm on macOS, Linux terminals based on VTE (GNOME terminal) support
#' `.href`, `.email` and `.file` links.
#'
#' You can use [ansi_has_hyperlink_support()] to check if your terminal or
#' IDE has hyperlink support in general, and [ansi_hyperlink_types()] to
#' check if various types of hyperlinks are supported.
#'
#' If your hyperlink support is not detected properly in your IDE or
#' terminal, please open a cli issue at
#' <https://github.com/r-lib/cli/issues>.
#'
#' ```{asciicast links-setup, include = FALSE, cache = FALSE}
#' options(
#'   cli.hyperlink = TRUE,
#'   cli.hyperlink_run = TRUE,
#'   cli.hyperlink_help = TRUE,
#'   cli.hyperlink_vignette = TRUE
#' )
#' ```
#'
#' # Link text
#'
#' Before we delve into the various types of hyperlinks, a general comment
#' about link texts. Some link styles support a custom link text:
#'
#' * `.href`
#' * `.help`
#' * `.topic`
#' * `.vignette`
#' * `.run`
#'
#' Others, i.e. `.email`, `.file`, `.fun` and `.url` do not support custom
#' link text.
#'
#' The generic syntax for link text is the same as for Markdown hyperlinks:
#' ```
#' {.style [link text](url)}
#' ```
#'
#' ## Vectorization
#'
#' Note that it is not possible to add link text to a vector of URLs. E.g.
#' this will create a list of three URLs, all clickable:
#'
#' ```{asciicast link-example}
#' urls <- paste0("https://httpbin.org/status/", c(200, 403, 404))
#' cli::cli_text("Some httpbin URLs: {.url {urls}}.")
#' ```
#' But it is not possible to use a different link text for them.
#'
#' ## What if hyperlinks are not available?
#'
#' If ANSI hyperlinks are not available, then the link text for of these
#' styles outputs both the link text and the URL in a (hopefully) helpful
#' way. See examples below.
#'
#' # URLs
#'
#' There are two cli styles to link to generic URLs. `.url` does not
#' allow custom link text, but `\href` does.
#'
#' ```{asciicast links-url-1}
#' cli_text(
#'   "See the cli homepage at {.url https://cli.r-lib.org} for details."
#' )
#' ```
#'
#'```{asciicast links-url-2}
#' cli_text(
#'   "See the {.href [cli homepage](https://cli.r-lib.org)} for details."
#' )
#' ```
#'
#' ## Without hyperlink support
#'
#' This is how these links look without hyperlink support:
#'
#' ```{asciicast links-url-3}
#' local({
#'   withr::local_options(cli.hyperlink = FALSE)
#'   cli_text(
#'     "See the cli homepage at {.url https://cli.r-lib.org} for details."
#'   )
#'   cli_text(
#'     "See the {.href [cli homepage](https://cli.r-lib.org)} for details."
#'   )
#' })
#' ```
#'
#' ## URL encoding
#'
#' Note that cli does not encode the url, so you might need to call
#' `utils::URLencode()` on it, especially, if it is substituted in
#' via `{}`.
#'
#' ```{asciicast links-url-4}
#' weirdurl <- utils::URLencode("https://example.com/has some spaces")
#' cli_text("See more at {.url {weirdurl}}.")
#' ```
#'
#' # Files
#'
#' The `.file` style now automatically creates a `file:` hyperlink.
#' Because `file:` hyperlinks must contain an absolute path, cli tries to
#' convert relative paths, and paths starting with `~` to aboslute path.
#'
#' ```{asciicast links-file-1}
#' cli_text("... edit your {.file ~/.Rprofile} file.")
#' ```
#'
#' ## Link text
#'
#' `.file` cannot use a custom link text. If you custom link text, then
#' you can use `.href` with a `file:` URL.
#'
#' ```{asciicast links-file-2}
#' prof <- path.expand("~/.Rprofile")
#' cli_text("... edit your {.href [R profile](file://{prof})}.")
#' ```
#'
#' ## Line and column numbers
#'
#' You may add a line number to a file name, separated by `:`. Handlers
#' typically place the cursor at that line after opening the file.
#' You may also add a column number, after the line number, separated by
#' another `:`.
#'
#' ```{asciicast links-file-3}
#' cli_text("... see line 5 in {.file ~/.Rprofile:5}.")
#' ```
#'
#' ## Default handler
#'
#' In RStudio `file:` URLs open within RStudio. If you click on a file
#' link outside of RStudio, typically the operating system is consulted
#' for the application to open it.
#'
#' ## Without hyperlink support
#'
#' One issue with using `.href` file files is that it does not look great
#' if hyperlinks are not available. This will be improved in the future:
#'
#' ```{asciicast links-file-4}
#' local({
#'   withr::local_options(cli.hyperlink = FALSE)
#'   prof <- path.expand("~/.Rprofile")
#'   cli_text("... edit your {.href [R profile](file://{prof})}.")
#' })
#' ```
#'
#' # Links to the manual
#'
#' `.fun` automatically creates links to the manual page of the function,
#' provided the function name is in the `packagename::functionname` form:
#'
#' ```{asciicast links-fun-1}
#' cli::cli_text("... see {.fun stats::lm} to learn more.")
#' ```
#'
#' ## Link text
#'
#' For a custom link text, use `.help` instead of `.fun`.
#'
#' ```{asciicast links-fun-2}
#' cli::cli_text("... see {.help [{.fun lm}](stats::lm)} to learn more.")
#' ```
#'
#' ## Without hyperlink support
#'
#' The same message without hyperlink support looks like this:
#'
#' ```{asciicast links-fun-3}
#' local({
#'   withr::local_options(cli.hyperlink = FALSE)
#'   cli::cli_text("... see {.help [{.fun lm}](stats::lm)} to learn more.")
#' })
#' ```
#'
#' ## Topics
#'
#' To link to a help topic that is not a function, use `.topic`:
#'
#' ```{asciicast links-topic}
#' cli::cli_text("... the tibble options at {.topic tibble::tibble_options}.")
#' ```
#'
#' `.topic` support link text.
#'
#' ## Vignettes
#'
#' To link to a vignette, use `.vignette`:
#'
#' ```{asciicast links-vignette}
#' cli::cli_text("... see the {.vignette tibble::types} vignette.")
#' ```
#'
#' # Click to run code
#'
#' RStudio also supports a special link type that runs R code in the
#' current R session upon clicking.
#'
#' You can create these links with `.run`:
#'
#' ```{asciicast links-run}
#' cli::cli_text("Run {.run testthat::snapshot_review()} to review")
#' ```
#'
#' ## Link text
#'
#' Sometimes you want to show a slightly different expression in the link,
#' than the one that is evaluated. E.g. the evaluated expression probably
#' needs to qualify packages with `::`, but you might not want to show this:
#'
#' ```{asciicast links-run-2}
#' cli::cli_text(
#'   "Run {.run [snapshot_review()](testthat::snapshot_review())} to review"
#' )
#' ```
#'
#' ## Security considerations
#'
#' To make `.run` hyperlinks more secure, RStudio will not run code
#'
#' * that is not in the `pkg::fun(args)` form,
#' * if `args` contains `(`, `)` or `;`,
#' * if it calls a core package (base, stats, etc.),
#' * if it calls a package that is not loaded, and it is not one of
#'   testthat, devtools, usethis, rlang, pkgload, or pkgdown which are explicitly allowed.
#'
#' When RStudio does not run a `.run` hyperlink, then it shows the code
#' and the user can copy and paste it to the console, if they consider
#' it safe to run.
#'
#' Note that depending on your version of RStudio, the behavior can change.
#' @name links
NULL

format_num <- local({
  pretty_num <- function(number, style = c("default", "nopad", "6")) {
    style <- switch(
      match.arg(style),
      "default" = pretty_num_default,
      "nopad" = pretty_num_nopad,
      "6" = pretty_num_6
    )

    style(number)
  }

  compute_num <- function(number, smallest_prefix = "y") {
    prefixes0 <- c(
      "y",
      "z",
      "a",
      "f",
      "p",
      "n",
      "u",
      "m",
      "",
      "k",
      "M",
      "G",
      "T",
      "P",
      "E",
      "Z",
      "Y"
    )
    zeroshif0 <- 9L

    stopifnot(
      is.numeric(number),
      is.character(smallest_prefix),
      length(smallest_prefix) == 1,
      !is.na(smallest_prefix),
      smallest_prefix %in% prefixes0
    )

    limits <- c(999950 * 1000^(seq_len(length(prefixes0)) - (zeroshif0 + 1L)))
    nrow <- length(limits)
    low <- match(smallest_prefix, prefixes0)
    zeroshift <- zeroshif0 + 1L - low
    prefixes <- prefixes0[low:length(prefixes0)]
    limits <- limits[low:nrow]
    nrow <- nrow - low + 1

    neg <- number < 0 & !is.na(number)
    number <- abs(number)
    mat <- matrix(
      rep(number, each = nrow),
      nrow = nrow,
      ncol = length(number)
    )
    mat2 <- matrix(mat < limits, nrow = nrow, ncol = length(number))
    exponent <- nrow - colSums(mat2) - (zeroshift - 1L)
    in_range <- function(exponent) {
      max(
        min(exponent, nrow - zeroshift, na.rm = FALSE),
        1L - zeroshift,
        na.rm = TRUE
      )
    }
    if (length(exponent)) {
      exponent <- sapply(exponent, in_range)
    }
    res <- number / 1000^exponent
    prefix <- prefixes[exponent + zeroshift]

    ## Zero number
    res[number == 0] <- 0
    prefix[number == 0] <- prefixes[zeroshift]

    ## NA and NaN number
    res[is.na(number)] <- NA_real_
    res[is.nan(number)] <- NaN
    prefix[is.na(number)] <- "" # prefixes0[low] is meaningless    # Includes NaN as well

    data.frame(
      stringsAsFactors = FALSE,
      amount = res,
      prefix = prefix,
      negative = neg
    )
  }

  pretty_num_default <- function(number) {
    szs <- compute_num(number)
    amt <- szs$amount
    sep <- " "

    ## String. For fractions we always show two fraction digits
    res <- character(length(amt))
    int <- is.na(amt) | abs(amt - as.integer(amt)) <= .Machine$double.eps
    res[int] <- format(
      ifelse(szs$negative[int], -1, 1) * amt[int],
      scientific = FALSE
    )
    res[!int] <- sprintf("%.2f", ifelse(szs$negative[!int], -1, 1) * amt[!int])

    format(paste(res, szs$prefix, sep = sep), justify = "right")
  }

  pretty_num_nopad <- function(number) {
    sub("^\\s+", "", pretty_num_default(number))
  }

  pretty_num_6 <- function(number) {
    szs <- compute_num(number, smallest_prefix = "y")
    amt <- round(szs$amount, 2)
    sep <- " "

    na <- is.na(amt)
    nan <- is.nan(amt)
    neg <- !na & !nan & szs$negative
    l10p <- !na & !nan & !neg & amt < 10
    l100p <- !na & !nan & !neg & amt >= 10 & amt < 100
    b100p <- !na & !nan & !neg & amt >= 100
    l10n <- !na & !nan & neg & amt < 10
    l100n <- !na & !nan & neg & amt >= 10 & amt < 100
    b100n <- !na & !nan & neg & amt >= 100

    famt <- character(length(amt))
    famt[na] <- "  NA"
    famt[nan] <- " NaN"
    famt[l10p] <- sprintf("%.2f", amt[l10p])
    famt[l100p] <- sprintf("%.1f", amt[l100p])
    famt[b100p] <- sprintf(" %.0f", amt[b100p])
    famt[l10n] <- sprintf("-%.1f", amt[l10n])
    famt[l100n] <- sprintf(" -%.0f", amt[l100n])
    famt[b100n] <- sprintf("-%.0f", amt[b100n])

    sub(" $", "  ", paste0(famt, sep, szs$prefix))
  }

  structure(
    list(
      .internal = environment(),
      pretty_num = pretty_num,
      compute_num = compute_num
    ),
    class = c("standalone_num", "standalone")
  )
})

# from pillar

#' Draw a sparkline bar graph with unicode block characters
#'
#' @description
#' Rendered using [block elements](https://en.wikipedia.org/wiki/Block_Elements).
#' In most common fixed width fonts these are rendered wider than regular
#' characters which means they are not suitable if you need precise alignment.
#'
#' You might want to avoid sparklines on non-UTF-8 systems, because they
#' do not look good. You can use [is_utf8_output()] to test for support
#' for them.
#'
#' @details
#'
#' ```{asciicast spark-bar-1}
#' x <- seq(0, 1, length = 6)
#' spark_bar(x)
#' ```
#'
#' ```{asciicast spark-bar-2}
#' x <- seq(0, 1, length = 6)
#' spark_bar(sample(x))
#' ```
#'
#' ```{asciicast spark-bar-3}
#' spark_bar(seq(0, 1, length = 8))
#' ```
#'
#' `NA`s are left out:
#'
#' ```{asciicast spark-bar-na}
#' spark_bar(c(0, NA, 0.5, NA, 1))
#' ```
#'
#' @param x A numeric vector between 0 and 1
#' @export
#' @seealso [spark_line()]

spark_bar <- function(x) {
  stopifnot(is.numeric(x))

  chars <- spark_bar_chars(x)

  structure(
    paste0(chars, collapse = ""),
    class = c("cli_spark_bar", "cli_spark")
  )
}

#' @export

format.cli_spark <- function(x, ...) {
  unclass(x)
}

#' @export

print.cli_spark <- function(x, ...) {
  cat(format(x, ...), sep = "\n")
}

spark_bar_chars <- function(x, bars = NULL) {

  if (is.null(bars)) {
    if (is_utf8_output()) {
      bars <- vapply(0x2581:0x2588, intToUtf8, character(1))
    } else {
      bars <- c("_", ",", "*", "#")
    }
  }

  factor <- cut(
    x,
    breaks = seq(0, 1, length.out = length(bars) + 1),
    labels = bars,
    include.lowest = TRUE
  )
  chars <- as.character(factor)
  chars[is.na(chars)] <- " "

  chars
}

#' Draw a sparkline line graph with Braille characters.
#'
#' You might want to avoid sparklines on non-UTF-8 systems, because they
#' do not look good. You can use [is_utf8_output()] to test for support
#' for them.
#'
#' @details
#'
#' ```{asciicast spark-line}
#' x <- seq(0, 1, length = 10)
#' spark_line(x)
#' ```
#'
#' @inheritParams spark_bar
#' @export
#' @seealso [spark_bar()]

spark_line <- function(x) {
  stopifnot(is.numeric(x))

  if (length(x) %% 2 == 1) {
    x <- c(x, NA)
  }

  if (is_utf8_output()) {
    y <- findInterval(x, seq(0, 1, length.out = 5), all.inside = TRUE)
    ind <- matrix(y, ncol = 2, byrow = TRUE)
    ind[, 2] <- ind[, 2] + 4
    chars <- apply(ind, 1, braille)
  } else {
    ind <- matrix(x, ncol = 2, byrow = TRUE)
    bars <- c("_", ",", "-", "^")
    chars <- spark_bar_chars(apply(ind, 1, mean), bars)
  }

  structure(
    paste0(chars, collapse = ""),
    class = c("cli_spark_line", "cli_spark")
  )
}

# https://en.wikipedia.org/wiki/Braille_Patterns
braille <- function(x) {
  # remap to braille sequence
  x <- c(7L, 3L, 2L, 1L, 8L, 6L, 5L, 4L)[x]

  raised <- 1:8 %in% x
  binary <- raised * 2^(0:7)

  # offset in hex is 2800
  val <- 10240 + sum(raised * 2^(0:7))

  intToUtf8(val)
}

# ------------------------------------------------------------------------

#' cli progress handlers
#'
#' The progress handler(s) to use can be selected with global options.
#'
#' There are three options that specify which handlers will be selected,
#' but most of the time you only need to use one of them. You can set these
#' options to a character vector, the names of the built-in cli handlers you
#' want to use:
#'
#' * If `cli.progress_handlers_only` is set, then these handlers are used,
#'   without considering others and without checking if they are able to
#'   handle a progress bar. This option is mainly intended for testing
#'   purposes.
#' * The handlers named in `cli.progress_handlers` are checked if they are
#'   able to handle the progress bar, and from the ones that are, the first
#'   one is selected. This is usually the option that the end use would want
#'   to set.
#' * The handlers named in `cli.progress_handlers_force` are always appended
#'   to the ones selected via `cli.progress_handlers`. This option is useful
#'   to add an additional handler, e.g. a logger that writes to a file.
#'
#' # The built-in progress handlers
#'
#' ### `cli`
#'
#' Use cli's internal status bar, the last line of the screen, to show the
#' progress bar. This handler is always able to handle all progress bars.
#'
#' ### `logger`
#'
#' Log progress updates to the screen, with one line for each update and with
#' time stamps. This handler is always able to handle all progress bars.
#'
#' ### `progressr`
#'
#' Use the progressr package to create
#' progress bars. This handler is always able to handle all progress bars.
#' (The progressr package needs to be installed.)
#'
#' ### `rstudio`
#'
#' Use [RStudio's job panel](https://posit.co/blog/rstudio-1-2-jobs/)
#' to show the progress bars. This handler is available at the RStudio console,
#' in recent versions of RStudio.
#'
#' ### `say`
#'
#' Use the macOS `say` command to announce progress events in speech (type
#' `man say` on a terminal for more info). Set the `cli.progress_say_frequency`
#' option to set the minimum delay between `say` invocations, the default is
#' three seconds. This handler is available on macOS, if the `say` command is
#' on the path.
#'
#' The external command and its arguments can be configured with options:
#'
#' * `cli_progress_say_args`: command line arguments, e.g. you can use this
#'   to select a voice on macOS,
#' * `cli_progress_say_command`: external command to run,
#' * `cli_progress_say_frequency`: wait at least this many seconds between
#'   calling the external command.
#'
#' ### `shiny`
#'
#' Use [shiny's progress bars](https://shiny.rstudio.com/articles/progress.html).
#' This handler is available if a shiny app is running.
#'
#' @return `cli_progress_builtin_handlers()` returns the names of the
#' currently supported progress handlers.
#'
#' @family progress bar functions
#' @export

# TODO: examples

cli_progress_builtin_handlers <- function() {
  names(builtin_handlers())
}

cli_progress_select_handlers <- function(bar, .envir) {
  hnd <- getOption("cli.progress_handlers", c("shiny", "cli"))
  frc <- getOption("cli.progress_handlers_force")
  onl <- getOption("cli.progress_handlers_only")

  bin <- builtin_handlers()
  if (!is.null(onl)) return(bin[onl])

  hnd_imp <- bin[hnd]
  hnd_able <- Filter(function(h) is.null(h$able) || h$able(bar, .envir), hnd_imp)
  if (length(hnd_able) > 1) hnd_able <- hnd_able[1]

  c(hnd_able, bin[frc])
}

# ------------------------------------------------------------------------

builtin_handler_cli <- list(
  add = function(bar, .envir) {
    bar$cli_statusbar <- cli_status(
      bar$format,
      msg_done = bar$format_done %||% bar$format,
      msg_failed = bar$format_failed %||% bar$format,
      .auto_close = FALSE,
      .envir = .envir,
      )
    bar$last_shown <- bar$current
    bar$justadded <- TRUE
  },

  set = function(bar, .envir) {
    if (isTRUE(bar$justadded)) {
      bar$justadded <- FALSE
      return()
    }
    bar$last_shown <- bar$current
    cli_status_update(id = bar$cli_statusbar, bar$format, .envir = .envir)
  },

  complete = function(bar, .envir, result) {
    if (isTRUE(bar$added)) {
      if (bar$clear) {
        # Show the full bar non-dynamic ttys
        if (!is_dynamic_tty() &&
            !identical(bar$last_shown, bar$current)) {
          cli_status_update(id = bar$cli_statusbar, bar$format, .envir = .envir)
        }
        cli_status_clear(bar$cli_statusbar, result = "clear", .envir = .envir)
      } else {
        if (result == "done" && !is.na(bar$total)) bar$current <- bar$total
        cli_status_clear(
          bar$cli_statusbar,
          result = result,
          msg_done = bar$format_done %||% bar$format,
          msg_failed = bar$format_failed %||% bar$format,
          .envir = .envir
        )
      }
    }
    bar$cli_statusbar <- TRUE
  },

  output = function(bar, .envir, text) {
    cli_verbatim(text)
  }
)

# ------------------------------------------------------------------------

builtin_handler_progressr <- list(
  add = function(bar, .envir) {
    steps <- if (is.na(bar$total)) 0 else bar$total
    bar$progressr_progressor <- asNamespace("progressr")$progressor(
      steps,
      auto_finish = FALSE,
      on_exit = TRUE,
      envir = .envir,
      label = bar$name %||% NA_character_
    )
  },

  set = function(bar, .envir) {
    if (!is.null(bar$progressr_progressor)) {
      bar$progressr_progressor(step = bar$current)
    }
  },

  complete = function(bar, .envir, result) {
    if (!is.null(bar$progressr_progressor)) {
      bar$progressr_progressor(step = bar$current, type = "finish")
    }
  },

  output = function(bar, .envir, text) {
    if (!is.null(bar$progressr_progressor)) {
      bar$progressr_progressor(message = text, class = "sticky", amount = 0)
    }
  }
)

# ------------------------------------------------------------------------

logger_out <- function(bar, event) {
  cat(sep = "", format_iso_8601(Sys.time()), " ", bar$id, " ",
      bar$current, "/", bar$total, " ", event, "\n")
}

builtin_handler_logger <- list(
  create = function(bar, .envir) {
    logger_out(bar, "created")
  },

  add = function(bar, .envir) {
    logger_out(bar, "added")
  },

  set = function(bar, .envir) {
    logger_out(bar, "updated")
  },

  complete = function(bar, .envir, result) {
    logger_out(bar, paste0("terminated (", result, ")"))
  },

  output = function(bar, .envir, text) {
    logger_out(bar, text)
  }
)

# ------------------------------------------------------------------------

say_out <- function(text) {
  say <- getOption("cli.progress_say_command", "say")
  args <- getOption("cli.progress_say_args", character())
  processx::process$new(say, c(args, text))
}

say_update <- function(bar) {
  now <- .Call(clic_get_time)
  freq <- getOption("cli.progress_say_frequency", 3.0)
  if (is.null(bar$say_last) || now - bar$say_last > freq) {
    txt <- if (is.na(bar$total)) bar$current else cli__pb_percent(bar)
    bar$say_proc <- say_out(txt)
    bar$say_last <- now
  }
}

builtin_handler_say <- list(
  able = function(bar, .envir) {
    if (!is.null(getOption("cli.progress_say_command"))) return(TRUE)
    Sys.info()[["sysname"]] == "Darwin" && Sys.which("say") != ""
  },

  add = function(bar, .envir) {
    ## Nothing to do here
  },

  set = function(bar, .envir) {
    say_update(bar)
  },

  complete = function(bar, .envir, result) {
    if (!is.null(bar$say_proc)) {
      bar$say_proc$kill()
      say_out("done")
    }
  }
)

# ------------------------------------------------------------------------

builtin_handler_rstudio <- list(
  able = function(bar, .envir) {
    tryCatch(
      .Platform$GUI == "Rstudio" && rstudioapi::isAvailable(),
      error = function(err) FALSE
    )
  },

  add = function(bar, .envir) {
    total <- if (is.na(bar$total)) 0L else as.integer(bar$total)
    bar$status <- bar$status %||% ""
    rstudio_id <- rstudioapi::jobAdd(
      name = bar$name %||% "",
      status = bar$status,
      progressUnits = total,
      running = TRUE,
      show = FALSE
    )
    # so the name is not duplicated in the format string as well
    if (is.na(bar$total)) bar$name <- NULL
    bar$rstudio_id <- rstudio_id
    bar$rstudio_status <- bar$status
  },

  set = function(bar, .envir) {
    if (!is.na(bar$total)) {
      rstudioapi::jobSetProgress(bar$rstudio_id, bar$current)
      if (bar$rstudio_status != bar$status) {
        rstudioapi::jobSetStatus(bar$rstudio_id, bar$status)
        bar$rstudio_status <- bar$status
      }
    } else {
      status <- cli_fmt(
        cli_text(bar$format, .envir = .envir),
        collapse = TRUE,
        strip_newline = TRUE
      )
      rstudioapi::jobSetStatus(bar$rstudio_id, status)
    }
  },

  complete = function(bar, .envir, results) {
    if (!is.null(bar$rstudio_id)) {
      rstudioapi::jobRemove(bar$rstudio_id)
    }
  },

  output = function(bar, .envir, text) {
    rstudioapi::jobAddOutput(bar$rstudio_id, text)
  }
)

# ------------------------------------------------------------------------

shiny_detail <- function(bar, .envir) {
  status <- if (is.null(bar$format_orig)) {
    bar$status %||% ""
  } else {
    cli_fmt(
      cli_text(bar$format, .envir = .envir),
      collapse = TRUE,
      strip_newline = TRUE
    )
  }
  output <- bar$shiny_output %||% ""
  paste0(
    status,
    if (status != "" && output != "") "\n",
    output
  )
}

last_lines <- function(txt, keep = 5) {
  txt <- sub("^\n*", "", txt)
  txt <- sub("\n*$", "", txt)
  lines <- strwrap(txt, width = 40)
  paste(utils::tail(lines, keep), collapse = "\n")
}

builtin_handler_shiny <- list(
  able = function(bar, .envir) {
    "shiny" %in% loadedNamespaces() && asNamespace("shiny")$isRunning()
  },

  add = function(bar, .envir) {
    bar$shiny_progress <- asNamespace("shiny")$Progress$new(
      asNamespace("shiny")$getDefaultReactiveDomain(),
      min = 0,
      max = bar$total
    )
    bar$shiny_progress$set(
      message = bar$name %||% "",
      detail = shiny_detail(bar, .envir)
    )
  },

  set = function(bar, .envir) {
    bar$shiny_progress$set(
      value = bar$current,
      detail = shiny_detail(bar, .envir)
    )
  },

  complete = function(bar, .envir, results) {
    if (!is.null(bar$shiny_progress)) {
      bar$shiny_progress$set(
        value = bar$current,
        detail = shiny_detail(bar, .envir)
      )
      bar$shiny_progress$close()
    }
    bar$shiny_progress <- NULL
  },

  output = function(bar, .envir, text) {
    bar$shiny_output <-
      last_lines(paste0(bar$shiny_output, " \u2022 ", text))
    bar$shiny_progress$set(
      value = bar$current,
      detail = shiny_detail(bar, .envir)
    )
  }
)

# ------------------------------------------------------------------------

builtin_handlers <- function() {
  list(
    cli = builtin_handler_cli,
    logger = builtin_handler_logger,
    progressr = builtin_handler_progressr,
    rstudio = builtin_handler_rstudio,
    say = builtin_handler_say,
    shiny = builtin_handler_shiny
  )
}

test_that("say_out", {
  px <- asNamespace("processx")$get_tool("px")
  tmp <- tempfile("cli-test-")
  on.exit(unlink(tmp), add = TRUE)
  withr::local_options(
    cli.progress_say_command = px,
    cli.progress_say_args = c("writefile", tmp)
  )
  p <- say_out(" 70%")
  expect_s3_class(p, "process")
  p$wait(3000)
  expect_false(p$is_alive())
  p$kill()
  expect_equal(readLines(tmp, warn = FALSE), " 70%")
})

test_that("say_update", {
  withr::local_options(cli.progress_say_frequency = 1e-9)
  local_mocked_bindings(say_out = function(text) text)
  bar <- new.env(parent = emptyenv())
  bar$current <- 10
  bar$total <- NA
  say_update(bar)
  expect_equal(bar$say_proc, 10)

  bar$total <- 50
  say_update(bar)
  expect_equal(bar$say_proc, " 20%")
})

test_that("cli_progress_builtin_handlers", {
  expect_true(is.character(cli_progress_builtin_handlers()))
  expect_true(all(
    c("cli", "shiny", "rstudio") %in% cli_progress_builtin_handlers()
  ))
})

test_that("cli_progress_select_handlers #1", {
  # only handlers
  withr::local_options(
    "cli.progress_handlers" = c("foo", "bar"),
    "cli.progress_handlers_force" = c("forced"),
    "cli.progress_handlers_only" = "logger"
  )
  expect_equal(
    names(cli_progress_select_handlers(list(), environment())),
    "logger"
  )
})

test_that("cli_progress_select_handlers #2", {
  # auto-select
  withr::local_options(
    "cli.progress_handlers" = c("foo", "bar", "baz"),
    "cli.progress_handlers_force" = NULL,
    "cli.progress_handlers_only" = NULL
  )
  fake <- list(
    foo = list(able = function(...) FALSE),
    bar = list(),
    baz = list(),
    forced = list()
  )
  local_mocked_bindings(builtin_handlers = function() fake)
  expect_equal(cli_progress_select_handlers(), fake["bar"])
})

test_that("cli_progress_select_handlers #3", {
  # auto-select
  withr::local_options(
    "cli.progress_handlers" = c("foo", "bar", "baz"),
    "cli.progress_handlers_force" = c("forced"),
    "cli.progress_handlers_only" = NULL
  )
  fake <- list(
    foo = list(able = function(...) FALSE),
    bar = list(able = function(...) TRUE),
    baz = list(),
    forced = list()
  )
  local_mocked_bindings(builtin_handlers = function() fake)
  expect_equal(cli_progress_select_handlers(), fake[c("bar", "forced")])
})

test_that("builtin_handlers", {
  expect_true(is.list(builtin_handlers()))
  expect_true(all(c("cli", "shiny", "rstudio") %in% names(builtin_handlers())))
})

test_that("loggerr_out", {
  bar <- new.env(parent = emptyenv())
  bar$id <- "id"
  bar$current <- 13
  bar$total <- 113
  local_mocked_bindings(Sys.time = function() .POSIXct(1623325865, tz = "CET"))
  expect_snapshot(logger_out(bar, "updated"))
})

test_that("cli_progress_builtin_handlers", {
  expect_true(is.character(cli_progress_builtin_handlers()))
  expect_true(all(
    c("cli", "shiny", "rstudio") %in% cli_progress_builtin_handlers()
  ))
})

test_that("cli_progress_select_handlers #1", {
  # only handlers
  withr::local_options(
    "cli.progress_handlers" = c("foo", "bar"),
    "cli.progress_handlers_force" = c("forced"),
    "cli.progress_handlers_only" = "logger"
  )
  expect_equal(
    names(cli_progress_select_handlers(list(), environment())),
    "logger"
  )
})

test_that("cli_progress_select_handlers #2", {
  # auto-select
  withr::local_options(
    "cli.progress_handlers" = c("foo", "bar", "baz"),
    "cli.progress_handlers_force" = NULL,
    "cli.progress_handlers_only" = NULL
  )
  fake <- list(
    foo = list(able = function(...) FALSE),
    bar = list(),
    baz = list(),
    forced = list()
  )
  mockery::stub(cli_progress_select_handlers, "builtin_handlers", fake)
  expect_equal(cli_progress_select_handlers(), fake["bar"])
})

test_that("cli_progress_select_handlers #3", {
  # auto-select
  withr::local_options(
    "cli.progress_handlers" = c("foo", "bar", "baz"),
    "cli.progress_handlers_force" = c("forced"),
    "cli.progress_handlers_only" = NULL
  )
  fake <- list(
    foo = list(able = function(...) FALSE),
    bar = list(able = function(...) TRUE),
    baz = list(),
    forced = list()
  )
  mockery::stub(cli_progress_select_handlers, "builtin_handlers", fake)
  expect_equal(cli_progress_select_handlers(), fake[c("bar", "forced")])
})

test_that("builtin_handlers", {
  expect_true(is.list(builtin_handlers()))
  expect_true(all(c("cli", "shiny", "rstudio") %in% names(builtin_handlers())))
})

test_that("say_out", {
  px <- asNamespace("processx")$get_tool("px")
  tmp <- tempfile("cli-test-")
  on.exit(unlink(tmp), add = TRUE)
  withr::local_options(
    cli.progress_say_command = px,
    cli.progress_say_args = c("writefile", tmp)
  )
  p <- say_out(" 70%")
  expect_s3_class(p, "process")
  p$wait(3000)
  expect_false(p$is_alive())
  p$kill()
  expect_equal(readLines(tmp, warn = FALSE), " 70%")
})

test_that("say_update", {
  withr::local_options(cli.progress_say_frequency = 1e-9)
  mockery::stub(say_update, "say_out", function(text) text)
  bar <- new.env(parent = emptyenv())
  bar$current <- 10
  bar$total <- NA
  say_update(bar)
  expect_equal(bar$say_proc, 10)

  bar$total <- 50
  say_update(bar)
  expect_equal(bar$say_proc, " 20%")
})

test_that("loggerr_out", {
  bar <- new.env(parent = emptyenv())
  bar$id <- "id"
  bar$current <- 13
  bar$total <- 113
  mockery::stub(logger_out, "Sys.time", .POSIXct(1623325865, tz = "CET"))
  expect_snapshot(logger_out(bar, "updated"))
})

# ------------------------------------------------------------------------

#' @title Progress bar variables
#'
#' @details These variables can be used in cli progress bar format
#' strings. They are calculated on demand. To use a variable, e.g. `pb_bar`
#' in a package, you either need to to import `pb_bar` from cli, or use
#' the qualified form in the format string: `cli::pb_bar`.
#'
#' Similarly, in R scripts, you can use `pb_bar` after `library(cli)`,
#' or `cli::pb_bar` if you do not attach the cli package.
#'
#' @family progress bar functions
#' @name progress-variables
NULL

#' @name progress-variables
#' @export pb_bar
#' @usage NULL
#' @aliases pb_bar
#'
#' @details
#' ### `pb_bar`
#'
#' Creates a visual progress bar. If the number of total units
#' is unknown, then it will return an empty string.
#'
#' ```{asciicast progress-var-bar, echo = 2:5}
#' x <- invisible(quote(
#' cli_progress_bar(
#'   total = 100,
#'   format = "Fitting model {cli::pb_bar} {cli::pb_percent}"
#' )
#' ))
#' cli:::var_helper(x, current = 66)
#' ```

cli__pb_bar <- function(pb = getOption("cli__pb")) {
  if (is.null(pb)) return("")
  if (is.na(pb$total)) return("")
  structure(
    list(current = pb$current, total = pb$total),
    class = "cli-progress-bar"
  )
}

#' @name progress-variables
#' @export pb_current
#' @usage NULL
#' @aliases pb_current
#'
#' @details
#' ### `pb_current`
#'
#' The number of current progress units.
#'
#' ```{asciicast progress-var-current, echo = 2:5}
#' x <- invisible(quote(
#' cli_progress_bar(
#'   total = 100,
#'   format = "{cli::pb_spin} Reading file {cli::pb_current}/{cli::pb_total}"
#' )
#' ))
#' cli:::var_helper(x, current = 66)
#' ```

cli__pb_current <- function(pb = getOption("cli__pb")) {
  if (is.null(pb)) return("")
  pb$current
}

#' @name progress-variables
#' @export pb_current_bytes
#' @usage NULL
#' @aliases pb_current_bytes
#'
#' @details
#' ### `pb_current_bytes`
#'
#' The number of current progress units formatted as bytes.
#' The output has a constant width of six characters.
#'
#' ```{asciicast progress-var-current-bytes, echo = 2:4}
#' x <- invisible(quote(
#' cli_progress_bar(
#'   format = "Got {cli::pb_current_bytes} in {cli::pb_elapsed}"
#' )
#' ))
#' cli:::var_helper2(x, current = 1024 * 512, delay = 5)
#' ```

cli__pb_current_bytes <- function(pb = getOption("cli__pb")) {
  if (is.null(pb)) return("")
  format_bytes$pretty_bytes(pb$current, style = "6")
}

#' @name progress-variables
#' @export pb_elapsed
#' @usage NULL
#' @aliases pb_elapsed
#'
#' @details
#' ### `pb_elapsed`
#'
#' The elapsed time since the start of the progress bar. The time is
#' measured since the progress bar was created with [cli_progress_bar()]
#' or similar.
#'
#' ```{asciicast progress-var-elapsed, echo = 2:5}
#' x <- invisible(quote(
#' cli_progress_bar(
#'   total = 100,
#'   format = "{cli::pb_bar} {cli::pb_percent} [{cli::pb_elapsed}]"
#' )
#' ))
#' cli:::var_helper2(x, current = 65, delay = 5)
#' ```

cli__pb_elapsed <- function(pb = getOption("cli__pb")) {
  if (is.null(pb)) return("")
  secs <- (.Call(clic_get_time) - pb$start) * clienv$speed_time
  format_time$pretty_sec(secs)
}

#' @name progress-variables
#' @export pb_elapsed_clock
#' @usage NULL
#' @aliases pb_elapsed_clock
#'
#' @details
#' ### `pb_elapsed_clock`
#'
#' The elapsed time, in `hh::mm::ss` format.
#'
#' ```{asciicast progress-var-elapsed-clock, echo = 2:5}
#' x <- invisible(quote(
#' cli_progress_bar(
#'   total = 100,
#'   format = "{cli::pb_bar} {cli::pb_percent} [{cli::pb_elapsed_clock}]"
#' )
#' ))
#' cli:::var_helper2(x, current = 65, delay = 5)
#' ```

cli__pb_elapsed_clock <- function(pb = getOption("cli__pb")) {
  if (is.null(pb)) return("")
  s <- (.Call(clic_get_time) - pb$start) * clienv$speed_time
  hours <- floor(s / 3600)
  minutes <- floor((s / 60) %% 60)
  seconds <- round(s %% 60, 1)
  paste0(
    formatC(hours, width = 2, flag = "0"),
    ":",
    formatC(minutes, width = 2, flag = "0"),
    ":",
    formatC(seconds, width = 2, flag = "0")
  )
}

#' @name progress-variables
#' @export pb_elapsed_raw
#' @usage NULL
#' @aliases pb_elapsed_raw
#'
#' @details
#' ### `pb_elapsed_raw`
#'
#' The number of seconds since the start of the progress bar.
#'
#' ```{asciicast progress-var-elapsed-raw, echo = 2:5}
#' x <- invisible(quote(
#' cli_progress_bar(
#'   total = 100,
#'   format = "{cli::pb_bar} {cli::pb_percent} [{round(cli::pb_elapsed_raw)}s]"
#' )
#' ))
#' cli:::var_helper2(x, current = 65, delay = 5)
#' ```

cli__pb_elapsed_raw <- function(pb = getOption("cli__pb")) {
  if (is.null(pb)) return("")
  (.Call(clic_get_time) - pb$start) * clienv$speed_time
}

#' @name progress-variables
#' @export pb_eta
#' @usage NULL
#' @aliases pb_eta
#'
#' @details
#' ### `pb_eta`
#'
#' The estimated time until the end of the progress bar,
#' in human readable form.
#'
#' ```{asciicast progress-var-eta, echo = 2:5}
#' x <- invisible(quote(
#' cli_progress_bar(
#'   total = 100,
#'   format = "{cli::pb_bar} {cli::pb_percent} | ETA: {cli::pb_eta}"
#' )
#' ))
#' cli:::var_helper2(x, current = 65, delay = 5)
#' ```

cli__pb_eta <- function(pb = getOption("cli__pb")) {
  if (is.null(pb)) return("")
  eta <- cli__pb_eta_raw(pb)
  if (is.na(eta)) {
    "?"
  } else {
    format_time_ago$vague_dt(eta, format = "terse")
  }
}

#' @name progress-variables
#' @export pb_eta_raw
#' @usage NULL
#' @aliases pb_eta_raw
#'
#' @details
#' ### `pb_eta_raw`
#'
#' The estimated time until the end of the progress
#' bar, in seconds. This is useful if you want to adjust the default
#' `pb_eta` display.
#'
#' ```{asciicast progress-var-eta-raw, echo = 2:5}
#' x <- invisible(quote(
#' cli_progress_bar(
#'   total = 100,
#'   format = "{cli::pb_bar} {cli::pb_percent} | ETA: {round(cli::pb_eta_raw)}s"
#' )
#' ))
#' cli:::var_helper2(x, current = 65, delay = 5)
#' ```

cli__pb_eta_raw <- function(pb = getOption("cli__pb")) {
  if (is.null(pb)) return("")
  if (is.na(pb$total)) return(NA_real_)
  if (pb$current == pb$total) return(as.difftime(0, units = "secs"))
  if (pb$current == 0L) return(NA_real_)
  elapsed <- (.Call(clic_get_time) - pb$start) * clienv$speed_time
  as.difftime(elapsed * (pb$total / pb$current - 1.0), units = "secs")
}

#' @name progress-variables
#' @export pb_eta_str
#' @usage NULL
#' @aliases pb_eta_str
#'
#' @details
#' ### `pb_eta_str`
#'
#' The estimated time until the end of the progress bar.
#' It includes the `"ETA:"` prefix. It is only shown if the time can be
#' estimated, otherwise it is the empty string.
#'
#' ```{asciicast progress-var-eta-str, echo = 2:5}
#' x <- invisible(quote(
#' cli_progress_bar(
#'   total = 100,
#'   format = "{cli::pb_bar} {cli::pb_percent} | {cli::pb_eta_str}"
#' )
#' ))
#' cli:::var_helper2(x, current = 65, delay = 5)
#' ```

cli__pb_eta_str <- function(pb = getOption("cli__pb")) {
  if (is.null(pb)) return("")
  eta <- cli__pb_eta(pb)
  if (eta != "?") paste0("ETA: ", eta) else ""
}

#' @name progress-variables
#' @export pb_extra
#' @usage NULL
#' @aliases pb_extra
#'
#' @details
#' ### `pb_extra`
#'
#' `pb_extra` can be used to access extra data, see the `extra` argument
#' of `cli_progress_bar()` and `cli_progress_update()`.
#'
#' ```{asciicast progress-var-extra, echo = 2:6}
#' x <- invisible(quote(
#' cli_progress_bar(
#'   total = 100,
#'   extra = list(user = whoami::username()),
#'   format = "Cleaning cache for user '{cli::pb_extra$user}': {cli::pb_current_bytes}"
#' )
#' ))
#' cli:::var_helper(x, current = 1024 * 1024 * 154)
#' ```

cli__pb_extra <- function(pb = getOption("cli__pb")) {
  if (is.null(pb)) return("")
  pb$extra
}

#' @name progress-variables
#' @export pb_id
#' @usage NULL
#' @aliases pb_id
#'
#' @details
#' ### `pb_id`
#'
#' The id of the progress bar. The id has the format
#' `cli-<pid>-<counter>` where `<pid>` is the process id, and
#' `<counter>` is an integer counter that is incremented every time
#' cli needs a new unique id.
#'
#' This is useful for debugging progress bars.
#'
#' ```{asciicast progress-var-id, echo = 2:4}
#' x <- invisible(quote(
#' cli_progress_bar(
#'   format = "Progress bar '{cli::pb_id}' is at {cli::pb_current}"
#' )
#' ))
#' cli:::var_helper(x, current = 64)
#' ```

cli__pb_id <- function(pb = getOption("cli__pb")) {
  if (is.null(pb)) return("")
  pb$id
}

#' @name progress-variables
#' @export pb_name
#' @usage NULL
#' @aliases pb_name
#'
#' @details
#' ### `pb_name`
#'
#' The name of the progress bar. This is supplied by the
#' developer, and it is by default the empty string. A space character
#' is added to non-empty names.
#'
#' ```{asciicast progress-var-name, echo = 2:6}
#' x <- invisible(quote(
#' cli_progress_bar(
#'   name = "Loading training data",
#'   total = 100,
#'   format = "{cli::pb_name} {cli::pb_bar} {cli::pb_percent}"
#' )
#' ))
#' cli:::var_helper(x, current = 66)
#' ```
#'

cli__pb_name <- function(pb = getOption("cli__pb")) {
  if (is.null(pb)) return("")
  if (!is.null(pb$name)) {
    paste0(pb$name, " ")
  } else {
    ""
  }
}

#' @name progress-variables
#' @export pb_percent
#' @usage NULL
#' @aliases pb_percent
#'
#' @details
#' ### `pb_percent`
#'
#' The percentage of the progress bar, always formatted
#' in three characters plus the percentage sign. If the total number of
#' units is unknown, then it is `" NA%"`.
#'
#' ```{asciicast progress-var-percent, echo = 2:5}
#' x <- invisible(quote(
#' cli_progress_bar(
#'   total = 100,
#'   format = "{cli::pb_bar} {cli::pb_percent}"
#' )
#' ))
#' cli:::var_helper(x, current = 66)
#' ```

cli__pb_percent <- function(pb = getOption("cli__pb")) {
  if (is.null(pb)) return("")
  paste0(format(round(pb$current / pb$total * 100), width = 3), "%")
}

#' @name progress-variables
#' @export pb_pid
#' @usage NULL
#' @aliases pb_pid
#'
#' @details
#' ### `pb_pid`
#'
#' The integer process id of the progress bar. This is useful if you are
#' aggregating logging output or progress results from multiple processes.

cli__pb_pid <- function(pb = getOption("cli__pb")) {
  if (is.null(pb)) return("")
  pb$pid %||% Sys.getpid()
}

#' @name progress-variables
#' @export pb_rate
#' @usage NULL
#' @aliases pb_rate
#'
#' @details
#' ### `pb_rate`
#'
#' The progress rate, in number of units per second, formatted in a string.
#'
#' ```{asciicast progress-var-rate, echo = 2:5}
#' x <- invisible(quote(
#' cli_progress_bar(
#'   total = 156,
#'   format = "Reading input files {pb_current}/{pb_total} [{pb_rate}]"
#' )
#' ))
#' cli:::var_helper2(x, current = 67, delay = 5)
#' ```

cli__pb_rate <- function(pb = getOption("cli__pb")) {
  if (is.null(pb)) return("")
  rate <- cli__pb_rate_raw(pb)
  if (is.nan(rate) || is.na(rate) || is.infinite(rate)) return("?/s")
  paste0(format(rate, digits = 2), "/s")
}

#' @name progress-variables
#' @export pb_rate_raw
#' @usage NULL
#' @aliases pb_rate_raw
#'
#' @details
#' ### `pb_rate_raw`
#'
#' The raw progress rate, in number of units per second.
#'
#' ```{asciicast progress-var-rate-raw, echo = 2:5}
#' x <- invisible(quote(
#' cli_progress_bar(
#'   total = 156,
#'   format = "Reading input files {pb_current}/{pb_total} [{round(pb_rate_raw)}/s]"
#' )
#' ))
#' cli:::var_helper2(x, current = 67, delay = 5)
#' ```

cli__pb_rate_raw <- function(pb = getOption("cli__pb")) {
  if (is.null(pb)) return("")
  elapsed <- cli__pb_elapsed_raw(pb)
  pb$current / elapsed
}

#' @name progress-variables
#' @export pb_rate_bytes
#' @usage NULL
#' @aliases pb_rate_bytes
#'
#' @details
#' ### `pb_rate_bytes`
#'
#' The progress rate, formatted as bytes per second, in human readable form.
#'
#' ```{asciicast progress-var-rate-bytes, echo = 2:7}
#' x <- invisible(quote(
#' cli_progress_bar(
#'   total = 256 * 1024 * 1014,
#'   format = paste0(
#'     "Reading data {pb_current_bytes}/{pb_total_bytes} ",
#'     "[{ansi_trimws(pb_rate_bytes)}]"
#'   )
#' )
#' ))
#' cli:::var_helper2(x, current = 67 * 1024 * 1024, delay = 5)
#' ```

cli__pb_rate_bytes <- function(pb = getOption("cli__pb")) {
  if (is.null(pb)) return("")
  rate <- cli__pb_rate_raw(pb)
  paste0(
    format_bytes$pretty_bytes(rate, style = "6"),
    "/s"
  )
}

#' @name progress-variables
#' @export pb_spin
#' @usage NULL
#' @aliases pb_spin
#'
#' @details
#' ### `pb_spin`
#'
#' A spinner. The default spinner is selected via a [get_spinner()] call.
#'
#' ```{asciicast progress-var-current, echo = 2:5}
#' x <- invisible(quote(
#' cli_progress_bar(
#'   total = 100,
#'   format = "{cli::pb_spin} Reading file {cli::pb_current}/{cli::pb_total}"
#' )
#' ))
#' cli:::var_helper(x, current = 66)
#' ```

cli__pb_spin <- function(pb = getOption("cli__pb")) {
  if (is.null(pb)) return("")

  pb$spinner <- pb$spinner %||% get_spinner()
  nx <- pb$tick %% length(pb$spinner$frames) + 1L
  pb$spinner$frames[[nx]]
}

#' @name progress-variables
#' @export pb_status
#' @usage NULL
#' @aliases pb_status
#'
#' @details
#' ### `pb_status`
#'
#' The status string of the progress bar. By default this is an empty
#' string, but it is possible to set it in [cli_progress_bar()]
#' and `cli_progress_update()].
#'
#' ```{asciicast progress-var-status, echo = 2}
#' x <- invisible(quote(
#' cli_progress_bar(status = "Connecting...")
#' ))
#' cli:::var_helper(x, current = 0, delay = 1)
#' ```

cli__pb_status <- function(pb = getOption("cli__pb")) {
  if (is.null(pb)) return("")
  if (!is.null(pb$status)) {
    paste0(pb$status, " ")
  } else {
    ""
  }
}

#' @name progress-variables
#' @export pb_timestamp
#' @usage NULL
#' @aliases pb_timestamp
#'
#' @details
#' ### `pb_timestamp`
#'
#' A time stamp for the current time in ISO 8601 format.
#'
#' ```{asciicast progress-var-timestamp, echo = 2:4}
#' x <- invisible(quote(
#' cli_progress_bar(
#'   "Loading training data files",
#'   format = "{pb_timestamp} {pb_current} ({pb_rate})"
#' )
#' ))
#' cli:::var_helper(x, current = 125, delay = 5)
#' ```

cli__pb_timestamp <- function(pb = getOption("cli__pb")) {
  if (is.null(pb)) return("")
  st <- Sys.time()
  if (clienv$speed_time != 1.0) {
    st <- clienv$load_time + (st - clienv$load_time) * clienv$speed_time
  }
  format_iso_8601(st)
}

#' @name progress-variables
#' @export pb_total
#' @usage NULL
#' @aliases pb_total
#'
#' @details
#' ### `pb_total`
#'
#' The total number of progress units, or `NA` if the number of units is
#' unknown.
#'
#' ```{asciicast progress-var-current, echo = 2:5}
#' x <- invisible(quote(
#' cli_progress_bar(
#'   total = 100,
#'   format = "{cli::pb_spin} Reading file {cli::pb_current}/{cli::pb_total}"
#' )
#' ))
#' cli:::var_helper(x, current = 66)
#' ```

cli__pb_total <- function(pb = getOption("cli__pb")) {
  if (is.null(pb)) return("")
  pb$total
}

#' @name progress-variables
#' @export pb_total_bytes
#' @usage NULL
#' @aliases pb_total_bytes
#'
#' @details
#' ### `pb_total_bytes`
#'
#' The total number of progress units, formatted as
#' bytes, in a human readable format.
#'
#' ```{asciicast progress-var-rate-bytes, echo = 2:7}
#' x <- invisible(quote(
#' cli_progress_bar(
#'   total = 256 * 1024 * 1014,
#'   format = paste0(
#'     "Reading data {pb_current_bytes}/{pb_total_bytes} ",
#'     "[{ansi_trimws(pb_rate_bytes)}]"
#'   )
#' )
#' ))
#' cli:::var_helper2(x, current = 67 * 1024 * 1024, delay = 5)
#' ```

cli__pb_total_bytes <- function(pb = getOption("cli__pb")) {
  if (is.null(pb)) return("")
  format_bytes$pretty_bytes(pb$total, style = "6")
}

# ------------------------------------------------------------------------

var_helper <- function(expr, current = 66, delay = 1) {
  expr[[1]] <- quote(cli_progress_demo)
  expr$at <- current
  expr$start <- as.difftime(delay, units = "secs")
  eval(expr)
}

var_helper2 <- function(expr, clear = TRUE, delay = 0, ...) {
  expr$.envir <- environment()
  id <- eval(expr, envir = new.env())
  bar <- clienv$progress[[id]]
  bar$start <- bar$start - delay
  bar$clear <- clear
  args <- list(...)
  for (i in seq_along(args)) bar[[names(args)[i]]] <- args[[i]]
  cli_progress_update(force = TRUE, id = id)
  cat("\n")
  suppressMessages(cli_progress_done(id = id))
}

# ------------------------------------------------------------------------

#' cli progress bar demo
#'
#' Useful for experimenting with format strings and for documentation.
#' It creates a progress bar, iterates it until it terminates and saves the
#' progress updates.
#'
#' @param name Passed to [cli_progress_bar()].
#' @param status Passed to [cli_progress_bar()].
#' @param type Passed to [cli_progress_bar()].
#' @param total Passed to [cli_progress_bar()].
#' @param .envir Passed to [cli_progress_bar()].
#' @param ... Passed to [cli_progress_bar()].
#' @param at The number of progress units to show and capture the progress
#'   bar at. If `NULL`, then a sequence of states is generated to show the
#'   progress from beginning to end.
#' @param show_after Delay to show the progress bar. Overrides the
#'   `cli.progress_show_after` option.
#' @param live Whether to show the progress bat on the screen, or just
#'   return the recorded updates. Defaults to the value of the
#'   `cli.progress_demo_live` options. If unset, then it is `TRUE` in
#'   interactive sessions.
#' @param delay Delay between progress bar updates.
#' @param start Time to subtract from the start time, to simulate a
#'   progress bar that takes longer to run.
#'
#' @return List with class `cli_progress_demo`, which has a print and a
#' format method for pretty printing. The `lines` entry contains the
#' output lines, each corresponding to one update.
#'
#' @export

# TODO: examples

cli_progress_demo <- function(name = NULL, status = NULL,
                              type = c("iterator", "tasks",
                                       "download", "custom"),
                              total = NA,
                              .envir = parent.frame(),
                              ...,
                              at = if (is_interactive()) NULL else 50,
                              show_after = 0,
                              live = NULL,
                              delay = 0,
                              start = as.difftime(5, units = "secs")) {

  opt <- options(cli.progress_show_after = show_after)
  on.exit(options(opt), add = TRUE)

  live <- live %||%
    getOption("cli.progress_demo_live") %||%
    is_interactive()

  id <- cli_progress_bar(
    name = name,
    status = status,
    type = type,
    total = total,
    ...,
    .envir = .envir,
    current = FALSE
  )
  bar <- clienv$progress[[id]]
  bar$start <- bar$start - as.double(start, units = "secs")

  last <- is.null(at)
  if (is.null(at)) {
    if (is.na(total)) {
      at <- 1:5
    } else {
      at <- seq_len(total)
    }
  }

  output <- file(open = "w+b")
  on.exit(close(output), add = TRUE)
  size <- 0L

  withCallingHandlers({
    for (crnt in at) {
      cli_progress_update(set = crnt, id = id, force = TRUE, .envir = .envir)
      if (delay > 0) Sys.sleep(delay)
    }
    if (last) {
      cli_progress_done(id = id, .envir = .envir)
    } else {
      suppressMessages(cli_progress_done(id = id, .envir = .envir))
    }
  }, cliMessage = function(msg) {
    cat(file = output, msg$message)
    size <<- size + nchar(msg$message, type = "bytes")
    if (!live) invokeRestart("muffleMessage")
  })

  lines <- readChar(output, size, useBytes = TRUE)
  lines <- sub_("^\r\r*", "", lines, useBytes = TRUE)
  lines <- sub_("\r\r*$", "", lines, useBytes = TRUE)
  lines <- gsub_("\r\r*", "\r", lines, useBytes = TRUE)
  lines <- strsplit_(lines, "[\r\n]", useBytes = TRUE)[[1]]

  res <- structure(
    list(lines = lines),
    class = "cli_progress_demo"
  )

  if (live) invisible(res) else res
}

#' @export

format.cli_progress_demo <- function(x, ...) {
  x$lines
}

#' @export

print.cli_progress_demo <- function(x, ...) {
  cat(format(x, ...), sep = "\n")
}

test_that("cli_progress_demo", {
  withr::local_options(cli.ansi = TRUE)
  out <- cli_progress_demo(live = FALSE, at = 50)
  out$lines <- fix_times(out$lines)
  expect_snapshot(out)

  out <- cli_progress_demo(live = FALSE, at = NULL, total = 3)
  out$lines <- fix_times(out$lines)
  expect_snapshot(out)

  out <- cli_progress_demo(live = FALSE, at = NULL, total = NA, delay = 0.001)
  out$lines <- fix_times(out$lines)
  expect_snapshot(out)

  fun <- function() {
    options(cli.progress_handlers_only = "cli")
    cli_progress_demo(live = TRUE, at = NULL, total = 3)
  }

  msgs <- capture_cli_messages(out <- fun())
  expect_snapshot(out)
  expect_snapshot(msgs)
})

test_that("pb_bar", {
  expect_equal(cli__pb_bar(NULL), "")
  withr::local_options(cli__pb = list(current = 15, total = NA))
  expect_snapshot(cli_text("-{cli::pb_bar}-"))
  withr::local_options(cli__pb = list(current = 15, total = 30))
  expect_snapshot(cli_text("-{cli::pb_bar}-"))
})

test_that("pb_current", {
  expect_equal(cli__pb_current(NULL), "")
  withr::local_options(cli__pb = list(current = 15))
  expect_equal(cli::pb_current, 15)
})

test_that("pb_current_bytes", {
  expect_equal(cli__pb_current_bytes(NULL), "")
  expect_snapshot({
    cli__pb_current_bytes(list(current = 0))
    cli__pb_current_bytes(list(current = 1))
    cli__pb_current_bytes(list(current = 1000))
    cli__pb_current_bytes(list(current = 1000 * 23))
    cli__pb_current_bytes(list(current = 1000 * 1000 * 23))
  })
})

test_that("pb_elapsed", {
  expect_equal(cli__pb_elapsed(NULL), "")
  withr::local_options(cli__pb = list(start = 0, speed_time = 1))
  local_mocked_bindings(.Call = function(...) 1)
  expect_snapshot(cli__pb_elapsed())
  local_mocked_bindings(.Call = function(...) 21)
  expect_snapshot(cli__pb_elapsed())
  local_mocked_bindings(.Call = function(...) 58)
  expect_snapshot(cli__pb_elapsed())
  local_mocked_bindings(.Call = function(...) 60 * 65)
  expect_snapshot(cli__pb_elapsed())
})

test_that("pb_elapsed_clock", {
  expect_equal(cli__pb_elapsed_clock(NULL), "")
  withr::local_options(cli__pb = list(start = 0, speed_time = 1))
  local_mocked_bindings(.Call = function(...) 1)
  expect_snapshot(cli__pb_elapsed_clock())
  local_mocked_bindings(.Call = function(...) 21)
  expect_snapshot(cli__pb_elapsed_clock())
  local_mocked_bindings(.Call = function(...) 58)
  expect_snapshot(cli__pb_elapsed_clock())
  local_mocked_bindings(.Call = function(...) 60 * 65)
  expect_snapshot(cli__pb_elapsed_clock())
})

test_that("pb_elapsed_raw", {
  expect_equal(cli__pb_elapsed_raw(NULL), "")
  withr::local_options(cli__pb = list(start = 0, speed_time = 1))
  local_mocked_bindings(.Call = function(...) 1)
  expect_snapshot(cli__pb_elapsed_raw())
  local_mocked_bindings(.Call = function(...) 21)
  expect_snapshot(cli__pb_elapsed_raw())
  local_mocked_bindings(.Call = function(...) 58)
  expect_snapshot(cli__pb_elapsed_raw())
  local_mocked_bindings(.Call = function(...) 60 * 65)
  expect_snapshot(cli__pb_elapsed_raw())
})

test_that("pb_eta", {
  expect_equal(cli__pb_eta(NULL), "")
  local_mocked_bindings(cli__pb_eta_raw = function(...) NA_real_)
  expect_snapshot(cli__pb_eta(list()))
  local_mocked_bindings(cli__pb_eta_raw = function(...) as.difftime(12, units = "secs"))
  expect_snapshot(cli__pb_eta(list()))
})

test_that("pb_eta_raw", {
  expect_equal(cli__pb_eta_raw(NULL), "")
  withr::local_options(cli__pb = list(total = NA))
  expect_identical(cli__pb_eta_raw(), NA_real_)
  local_mocked_bindings(.Call = function(...) 0)
  withr::local_options(cli__pb = list(start = -10, total = 100, current = 0))
  expect_snapshot(cli__pb_eta_raw())
  withr::local_options(cli__pb = list(start = -10, total = 100, current = 20))
  expect_snapshot(cli__pb_eta_raw())
  withr::local_options(cli__pb = list(start = -40, total = 100, current = 80))
  expect_snapshot(cli__pb_eta_raw())
  withr::local_options(cli__pb = list(start = -50, total = 100, current = 100))
  expect_snapshot(cli__pb_eta_raw())
})

test_that("pb_eta_str", {
  expect_equal(cli__pb_eta_str(NULL), "")
  withr::local_options(cli__pb = list(total = NA))
  expect_identical(cli__pb_eta_str(), "")
  local_mocked_bindings(cli__pb_eta = function(...) "?")
  expect_snapshot(cli__pb_eta_str(list()))
  local_mocked_bindings(cli__pb_eta = function(...) " 1s")
  expect_snapshot(cli__pb_eta_str(list()))
})

test_that("pb_extra", {
  expect_equal(cli__pb_extra(NULL), "")
  expect_equal(cli__pb_extra(list(extra = list(a = 1))), list(a = 1))
})

test_that("pb_id", {
  expect_equal(cli__pb_id(NULL), "")
  expect_equal(cli__pb_id(list(id = 123)), 123)
})

test_that("pb_name", {
  expect_equal(cli__pb_name(NULL), "")
  expect_equal(cli__pb_name(list(name = NULL)), "")
  expect_equal(cli__pb_name(list(name = "foo")), "foo ")
})

test_that("pb_percent", {
  expect_equal(cli__pb_percent(NULL), "")
  expect_snapshot({
    cli__pb_percent(list(current = 0, total = 99))
    cli__pb_percent(list(current = 5, total = 99))
    cli__pb_percent(list(current = 10, total = 99))
    cli__pb_percent(list(current = 25, total = 99))
    cli__pb_percent(list(current = 99, total = 99))
    cli__pb_percent(list(current = 100, total = 99))
  })
})

test_that("pb_pid", {
  expect_equal(cli__pb_pid(NULL), "")
  expect_equal(cli__pb_pid(list()), Sys.getpid())
  expect_equal(cli__pb_pid(list(pid = 100)), 100)
})

test_that("pb_rate", {
  expect_equal(cli__pb_rate(NULL), "")
  local_mocked_bindings(cli__pb_rate_raw = function(...) NaN)
  expect_snapshot(cli__pb_rate(list()))
  local_mocked_bindings(cli__pb_rate_raw = function(...) Inf)
  expect_snapshot(cli__pb_rate(list()))
  local_mocked_bindings(cli__pb_rate_raw = function(...) 1 / 10)
  expect_snapshot(cli__pb_rate(list()))
  local_mocked_bindings(cli__pb_rate_raw = function(...) 12.4)
  expect_snapshot(cli__pb_rate(list()))
})

test_that("pb_rate_raw", {
  expect_equal(cli__pb_rate_raw(NULL), "")
  local_mocked_bindings(cli__pb_elapsed_raw = function(...) this)
  this <- 0
  expect_equal(cli__pb_rate_raw(list(current = 0)), NaN)
  expect_equal(cli__pb_rate_raw(list(current = 23)), Inf)
  this <- 1
  expect_equal(cli__pb_rate_raw(list(current = 23)), 23)
  this <- 10
  expect_equal(cli__pb_rate_raw(list(current = 1)), 1/10)
})

test_that("pb_rate_bytes", {
  expect_equal(cli__pb_rate_bytes(NULL), "")
  local_mocked_bindings(cli__pb_rate_raw = function(...) NaN)
  expect_snapshot(cli__pb_rate_bytes(list()))
  local_mocked_bindings(cli__pb_rate_raw = function(...) Inf)
  expect_snapshot(cli__pb_rate_bytes(list()))
  local_mocked_bindings(cli__pb_rate_raw = function(...) 0)
  expect_snapshot(cli__pb_rate_bytes(list()))
  local_mocked_bindings(cli__pb_rate_raw = function(...) 1024)
  expect_snapshot(cli__pb_rate_bytes(list()))
  local_mocked_bindings(cli__pb_rate_raw = function(...) 1024 * 23)
  expect_snapshot(cli__pb_rate_bytes(list()))
  local_mocked_bindings(cli__pb_rate_raw = function(...) 1024 * 1024 * 23)
  expect_snapshot(cli__pb_rate_bytes(list()))
})

test_that("pb_spin", {
  expect_equal(cli__pb_spin(NULL), "")
  withr::local_options(cli.spinner = "line")
  withr::local_options(cli__pb = list(tick = 1))
  expect_snapshot(cli_text("-{cli::pb_spin}-{cli::pb_spin}-"))
  withr::local_options(cli__pb = list(tick = 2))
  expect_snapshot(cli_text("-{cli::pb_spin}-{cli::pb_spin}-"))
  withr::local_options(cli__pb = list(tick = 10))
  expect_snapshot(cli_text("-{cli::pb_spin}-{cli::pb_spin}-"))
})

test_that("pb_status", {
  expect_equal(cli__pb_status(NULL), "")
  expect_equal(cli__pb_status(NULL), "")
  expect_equal(cli__pb_status(list(status = NULL)), "")
  expect_equal(cli__pb_status(list(status = "foo")), "foo ")
})

test_that("pb_timestamp", {
  expect_equal(cli__pb_timestamp(NULL), "")
  fake <- .POSIXct(1623974954, tz = "GMT")
  local_mocked_bindings(Sys.time = function() fake)
  expect_snapshot(cli__pb_timestamp(list()))

  backup <- mget(c("load_time", "speed_time"), clienv)
  on.exit({
    clienv$load_time <- backup$load_time
    clienv$speed_time <- backup$speed_time
  }, add = TRUE)

  clienv$load_time <- fake - 10
  clienv$speed_time <- 3.0
  expect_snapshot(cli__pb_timestamp(list()))
})

test_that("pb_total", {
  expect_equal(cli__pb_total(NULL), "")
  expect_equal(cli__pb_total(list(total = 101)), 101)
})

test_that("pb_total_bytes", {
  expect_equal(cli__pb_total_bytes(NULL), "")
  expect_snapshot({
    cli__pb_total_bytes(list(total = 0))
    cli__pb_total_bytes(list(total = 1))
    cli__pb_total_bytes(list(total = 1000))
    cli__pb_total_bytes(list(total = 1000 * 23))
    cli__pb_total_bytes(list(total = 1000 * 1000 * 23))
  })
})

test_that("cli_progress_demo", {
  withr::local_options(cli.ansi = TRUE)
  out <- cli_progress_demo(live = FALSE, at = 50)
  out$lines <- fix_times(out$lines)
  expect_snapshot(out)

  out <- cli_progress_demo(live = FALSE, at = NULL, total = 3)
  out$lines <- fix_times(out$lines)
  expect_snapshot(out)

  out <- cli_progress_demo(live = FALSE, at = NULL, total = NA, delay = 0.001)
  out$lines <- fix_times(out$lines)
  expect_snapshot(out)

  fun <- function() {
    options(cli.progress_handlers_only = "cli")
    cli_progress_demo(live = TRUE, at = NULL, total = 3)
  }

  msgs <- capture_cli_messages(out <- fun())
  expect_snapshot(out)
  expect_snapshot(msgs)
})

test_that("pb_bar", {
  expect_equal(cli__pb_bar(NULL), "")
  withr::local_options(cli__pb = list(current = 15, total = NA))
  expect_snapshot(cli_text("-{cli::pb_bar}-"))
  withr::local_options(cli__pb = list(current = 15, total = 30))
  expect_snapshot(cli_text("-{cli::pb_bar}-"))
})

test_that("pb_current", {
  expect_equal(cli__pb_current(NULL), "")
  withr::local_options(cli__pb = list(current = 15))
  expect_equal(cli::pb_current, 15)
})

test_that("pb_current_bytes", {
  expect_equal(cli__pb_current_bytes(NULL), "")
  expect_snapshot({
    cli__pb_current_bytes(list(current = 0))
    cli__pb_current_bytes(list(current = 1))
    cli__pb_current_bytes(list(current = 1000))
    cli__pb_current_bytes(list(current = 1000 * 23))
    cli__pb_current_bytes(list(current = 1000 * 1000 * 23))
  })
})

test_that("pb_elapsed", {
  expect_equal(cli__pb_elapsed(NULL), "")
  withr::local_options(cli__pb = list(start = 0, speed_time = 1))
  mockery::stub(cli__pb_elapsed, ".Call", 1)
  expect_snapshot(cli__pb_elapsed())
  mockery::stub(cli__pb_elapsed, ".Call", 21)
  expect_snapshot(cli__pb_elapsed())
  mockery::stub(cli__pb_elapsed, ".Call", 58)
  expect_snapshot(cli__pb_elapsed())
  mockery::stub(cli__pb_elapsed, ".Call", 60 * 65)
  expect_snapshot(cli__pb_elapsed())
})

test_that("pb_elapsed_clock", {
  expect_equal(cli__pb_elapsed_clock(NULL), "")
  withr::local_options(cli__pb = list(start = 0, speed_time = 1))
  mockery::stub(cli__pb_elapsed_clock, ".Call", 1)
  expect_snapshot(cli__pb_elapsed_clock())
  mockery::stub(cli__pb_elapsed_clock, ".Call", 21)
  expect_snapshot(cli__pb_elapsed_clock())
  mockery::stub(cli__pb_elapsed_clock, ".Call", 58)
  expect_snapshot(cli__pb_elapsed_clock())
  mockery::stub(cli__pb_elapsed_clock, ".Call", 60 * 65)
  expect_snapshot(cli__pb_elapsed_clock())
})

test_that("pb_elapsed_raw", {
  expect_equal(cli__pb_elapsed_raw(NULL), "")
  withr::local_options(cli__pb = list(start = 0, speed_time = 1))
  mockery::stub(cli__pb_elapsed_raw, ".Call", 1)
  expect_snapshot(cli__pb_elapsed_raw())
  mockery::stub(cli__pb_elapsed_raw, ".Call", 21)
  expect_snapshot(cli__pb_elapsed_raw())
  mockery::stub(cli__pb_elapsed_raw, ".Call", 58)
  expect_snapshot(cli__pb_elapsed_raw())
  mockery::stub(cli__pb_elapsed_raw, ".Call", 60 * 65)
  expect_snapshot(cli__pb_elapsed_raw())
})

test_that("pb_eta", {
  expect_equal(cli__pb_eta(NULL), "")
  mockery::stub(cli__pb_eta, "cli__pb_eta_raw", NA_real_)
  expect_snapshot(cli__pb_eta(list()))
  mockery::stub(cli__pb_eta, "cli__pb_eta_raw", as.difftime(12, units = "secs"))
  expect_snapshot(cli__pb_eta(list()))
})

test_that("pb_eta_raw", {
  expect_equal(cli__pb_eta_raw(NULL), "")
  withr::local_options(cli__pb = list(total = NA))
  expect_identical(cli__pb_eta_raw(), NA_real_)
  mockery::stub(cli__pb_eta_raw, ".Call", 0)
  withr::local_options(cli__pb = list(start = -10, total = 100, current = 0))
  expect_snapshot(cli__pb_eta_raw())
  withr::local_options(cli__pb = list(start = -10, total = 100, current = 20))
  expect_snapshot(cli__pb_eta_raw())
  withr::local_options(cli__pb = list(start = -40, total = 100, current = 80))
  expect_snapshot(cli__pb_eta_raw())
  withr::local_options(cli__pb = list(start = -50, total = 100, current = 100))
  expect_snapshot(cli__pb_eta_raw())
})

test_that("pb_eta_str", {
  expect_equal(cli__pb_eta_str(NULL), "")
  withr::local_options(cli__pb = list(total = NA))
  expect_identical(cli__pb_eta_str(), "")
  mockery::stub(cli__pb_eta_str, "cli__pb_eta", "?")
  expect_snapshot(cli__pb_eta_str(list()))
  mockery::stub(cli__pb_eta_str, "cli__pb_eta", " 1s")
  expect_snapshot(cli__pb_eta_str(list()))
})

test_that("pb_extra", {
  expect_equal(cli__pb_extra(NULL), "")
  expect_equal(cli__pb_extra(list(extra = list(a = 1))), list(a = 1))
})

test_that("pb_id", {
  expect_equal(cli__pb_id(NULL), "")
  expect_equal(cli__pb_id(list(id = 123)), 123)
})

test_that("pb_name", {
  expect_equal(cli__pb_name(NULL), "")
  expect_equal(cli__pb_name(list(name = NULL)), "")
  expect_equal(cli__pb_name(list(name = "foo")), "foo ")
})

test_that("pb_percent", {
  expect_equal(cli__pb_percent(NULL), "")
  expect_snapshot({
    cli__pb_percent(list(current = 0, total = 99))
    cli__pb_percent(list(current = 5, total = 99))
    cli__pb_percent(list(current = 10, total = 99))
    cli__pb_percent(list(current = 25, total = 99))
    cli__pb_percent(list(current = 99, total = 99))
    cli__pb_percent(list(current = 100, total = 99))
  })
})

test_that("pb_pid", {
  expect_equal(cli__pb_pid(NULL), "")
  expect_equal(cli__pb_pid(list()), Sys.getpid())
  expect_equal(cli__pb_pid(list(pid = 100)), 100)
})

test_that("pb_rate", {
  expect_equal(cli__pb_rate(NULL), "")
  mockery::stub(cli__pb_rate, "cli__pb_rate_raw", NaN)
  expect_snapshot(cli__pb_rate(list()))
  mockery::stub(cli__pb_rate, "cli__pb_rate_raw", Inf)
  expect_snapshot(cli__pb_rate(list()))
  mockery::stub(cli__pb_rate, "cli__pb_rate_raw", 1/10)
  expect_snapshot(cli__pb_rate(list()))
  mockery::stub(cli__pb_rate, "cli__pb_rate_raw", 12.4)
  expect_snapshot(cli__pb_rate(list()))
})

test_that("pb_rate_raw", {
  expect_equal(cli__pb_rate_raw(NULL), "")
  mockery::stub(cli__pb_rate_raw, "cli__pb_elapsed_raw", function(...) this)
  this <- 0
  expect_equal(cli__pb_rate_raw(list(current = 0)), NaN)
  expect_equal(cli__pb_rate_raw(list(current = 23)), Inf)
  this <- 1
  expect_equal(cli__pb_rate_raw(list(current = 23)), 23)
  this <- 10
  expect_equal(cli__pb_rate_raw(list(current = 1)), 1/10)
})

test_that("pb_rate_bytes", {
  expect_equal(cli__pb_rate_bytes(NULL), "")
  mockery::stub(cli__pb_rate_bytes, "cli__pb_rate_raw", NaN)
  expect_snapshot(cli__pb_rate_bytes(list()))
  mockery::stub(cli__pb_rate_bytes, "cli__pb_rate_raw", Inf)
  expect_snapshot(cli__pb_rate_bytes(list()))
  mockery::stub(cli__pb_rate_bytes, "cli__pb_rate_raw", 0)
  expect_snapshot(cli__pb_rate_bytes(list()))
  mockery::stub(cli__pb_rate_bytes, "cli__pb_rate_raw", 1024)
  expect_snapshot(cli__pb_rate_bytes(list()))
  mockery::stub(cli__pb_rate_bytes, "cli__pb_rate_raw", 1024 * 23)
  expect_snapshot(cli__pb_rate_bytes(list()))
  mockery::stub(cli__pb_rate_bytes, "cli__pb_rate_raw", 1024 * 1024 * 23)
  expect_snapshot(cli__pb_rate_bytes(list()))
})

test_that("pb_spin", {
  expect_equal(cli__pb_spin(NULL), "")
  withr::local_options(cli.spinner = "line")
  withr::local_options(cli__pb = list(tick = 1))
  expect_snapshot(cli_text("-{cli::pb_spin}-{cli::pb_spin}-"))
  withr::local_options(cli__pb = list(tick = 2))
  expect_snapshot(cli_text("-{cli::pb_spin}-{cli::pb_spin}-"))
  withr::local_options(cli__pb = list(tick = 10))
  expect_snapshot(cli_text("-{cli::pb_spin}-{cli::pb_spin}-"))
})

test_that("pb_status", {
  expect_equal(cli__pb_status(NULL), "")
  expect_equal(cli__pb_status(NULL), "")
  expect_equal(cli__pb_status(list(status = NULL)), "")
  expect_equal(cli__pb_status(list(status = "foo")), "foo ")
})

test_that("pb_timestamp", {
  expect_equal(cli__pb_timestamp(NULL), "")
  fake <- .POSIXct(1623974954, tz = "GMT")
  mockery::stub(cli__pb_timestamp, "Sys.time", fake)
  expect_snapshot(cli__pb_timestamp(list()))

  backup <- mget(c("load_time", "speed_time"), clienv)
  on.exit({
    clienv$load_time <- backup$load_time
    clienv$speed_time <- backup$speed_time
  }, add = TRUE)

  clienv$load_time <- fake - 10
  clienv$speed_time <- 3.0
  expect_snapshot(cli__pb_timestamp(list()))
})

test_that("pb_total", {
  expect_equal(cli__pb_total(NULL), "")
  expect_equal(cli__pb_total(list(total = 101)), 101)
})

test_that("pb_total_bytes", {
  expect_equal(cli__pb_total_bytes(NULL), "")
  expect_snapshot({
    cli__pb_total_bytes(list(total = 0))
    cli__pb_total_bytes(list(total = 1))
    cli__pb_total_bytes(list(total = 1000))
    cli__pb_total_bytes(list(total = 1000 * 23))
    cli__pb_total_bytes(list(total = 1000 * 1000 * 23))
  })
})

#' Auto-linking existing styles
#'
#' They keep formatting. It is not possible to use a different link text
#' with them. We could add link text support, but theming is applied to the
#' result of these tags, and it would look weird for link text. (I.e. if
#' there is link text you don't want to append `()` to the function name,
#' etc.)
#'
#' N | Goal                                        | Input                           |Links to (link text is always the verbatim content, styled)
#' --|---------------------------------------------|---------------------------------|---------------------------------------------------------------------
#' 1 | auto-link emails                            | `{.email foo@bar.com}`          | `mailto:foo@bar.com`
#' 2 | auto-link file                              | `{.file path/file}`             | `file:///abs/path/dile`
#' 3 | auto-link file with line and column numbers | `{.file /abs/path:line:col}`    | `file:///abs/path:line:col`, `params = list(line = line, col = col)`
#' 4 | auto-link function                          | `{.fun pkg::fun}`               | `x-r-help:pkg::fun`
#' 5 | mention function w/o package                | `{.fun fun}`                    | no link is created for this form
#' 6 | auto-link url                               | `{.url url}`                    | `url`
#'
#' ## New styles to create links
#'
#' These all have link text support, via the `[text](link)` markdown syntax.
#'
#' N | Goal                                        | Input                           | Link text         | Links to                 | Non-link form
#' --|---------------------------------------------|---------------------------------|-------------------|--------------------------|------------------------------------
#' 7 | link qualified function name to help        | `{.help pkg::fun}`              | `{.fun pkg::fun}` | `x-r-help:pkg::fun`      | `{.fun ?pkg::fun}`
#' 8 | link to function with link text             | `{.help [text](pkg::fun)}`      | `text`            | `x-r-help:pkg::fun`      | `text ({.fun pkg::fun})`
#' 9 | link to topic                               | `{.topic pkg::topic}`           | `pkg::topic`      | `x-r-help:pkg::topic`    | `{.code pkg::topic}`
#' 10| link to topic with link text                | `{.topic [text](pkg::topic)}`   | `text`            | `x-r-help:pkg::topic`    | `text ({.code pkg::topic})`
#' 11| link url                                    | `{.href url}`                   | `{.url url}`      | `url`                    | `{.url url}`
#' 12| link url with link text                     | `{.href [text](url)}`           | `text`            | `url`                    | `text ({.url url})`
#' 13| link running expr                           | `{.run expr}`                   | `{.code expr}`    | `x-r-run:expr`           | `{.code expr}`
#' 14| link running expr, show code                | `{.run [code](expr)}`           | `{.code code}`    | `x-r-run:expr`           | `{.code expr}`
#' 15| link to vignette                            | `{.vignette pkg::name}`         | `pkg::name`       | `x-r-vignette:pkg::name` | `{.code vignette(pkg::name)}`
#' 16| link to vignette with link text             | `{.vignette [text](pkg::name)}` | `text`            | `x-r-vignette:pkg::name` | `text ({.code vignette(pkg::name)})`
#'
#' @name cli-links
#' @noRd
NULL

make_link <- function(txt, type = c("email", "file", "fun", "help", "href",
                                    "run", "topic", "url", "vignette")) {
  type <- match.arg(type)

  switch(
    type,
    email    = make_link_email(txt),
    file     = make_link_file(txt),
    fun      = make_link_fun(txt),
    help     = make_link_help(txt),
    href     = make_link_href(txt),
    run      = make_link_run(txt),
    topic    = make_link_topic(txt),
    url      = make_link_url(txt),
    vignette = make_link_vignette(txt),
    throw(cli_error("Unknown hyperlink type: {.code {type}}, internal cli error")) # nocov
  )
}

# -- {.email} -------------------------------------------------------------

make_link_email <- function(txt) {
  style_hyperlink(txt, paste0("mailto:", txt))
}

# -- {.file} and {.path} --------------------------------------------------

# if txt already contains a hyperlink, then we do not add another link
# this is needed because some packages, e.g. roxygen2 currently create
# links to files manually:
# https://github.com/r-lib/roxygen2/blob/3ddfd7f2e35c3a71d5705ab4f49e851cd8da306d/R/utils.R#L91

make_link_file <- function(txt) {
  ret <- txt
  linked <- grepl("\007|\033\\\\", txt)
  ret[!linked] <- vcapply(which(!linked), function(i) {
    params <- parse_file_link_params(txt[i])
    link <- construct_file_link(params)
    style_hyperlink(
      txt[i],
      link$url,
      params = link$params
    )
  })
  ret
}

parse_file_link_params <- function(txt) {
  pattern <- "^(?<path>.*?)(?::(?<line>\\d*))?(?::(?<column>\\d*))?$"
  matches <- re_match(txt, pattern)
  ret <- as.list(matches)
  ret[!nzchar(ret)] <- list(NULL)
  if (is.null(ret[["path"]])) ret[["path"]] <- ""
  ret
}

construct_file_link <- function(params) {
  fmt <- get_config_chr("hyperlink_file_url_format")

  if (is.null(fmt)) {
    return(construct_file_link_OG(params))
  }

  params$path <- sub("^file://", "", params$path)
  params$path <- path.expand(params$path)

  looks_absolute <- function(path) {
    grepl("^/", params$path) || (is_windows() && grepl("^[a-zA-Z]:", params$path))
  }
  if (!looks_absolute(params$path)) {
    params$path <- file.path(getwd(), params$path)
  }
  if (!grepl("^/", params$path)) {
    params$path <- paste0("/", params$path)
  }

  res <- interpolate_parts(fmt, params)
  list(url = res)
}

# the order of operations is very intentional and important:
# column, then line, then path
# relates to how interpolate_part() works
interpolate_parts <- function(fmt, params) {
  res <- interpolate_part(fmt, "column", params$column)
  res <- interpolate_part(res, "line", params$line)
  interpolate_part(res, "path", params$path)
}

# interpolate a part, if possible
# if no placeholder for part, this is a no-op
# if placeholder exists, but no value to fill, remove placeholder (and everything after it!)
interpolate_part <- function(fmt, part = c("column", "line", "path"), value = NULL) {
  part <- match.arg(part)
  re <- glue(
    "^(?<before>.*)(?<part>\\{<<<part>>>\\})(?<after>.*?)$",
    .open = "<<<", .close = ">>>"
  )
  m <- re_match(fmt, re)

  if (is.na(m$part) || !nzchar(m$part)) {
    return(fmt)
  }

  if (is.null(value) || !nzchar(value)) {
    return(sub("}[^}]*$", "}", m$before))
  }

  paste0(m$before, value, m$after)
}

# handle the iterm and RStudio cases, which predated the notion of configuring
# the file hyperlink format
construct_file_link_OG <- function(params) {
  params$path <- abs_path(params$path)

  if (Sys.getenv("R_CLI_HYPERLINK_STYLE") == "iterm") {
    fmt <- "{path}#{line}:{column}"
    res <- interpolate_parts(fmt, params)
    return(list(url = res))
  }

  # RStudio takes line and col via params
  loc <- if (is.null(params$line)) {
    NULL
  } else {
    list(line = params$line, col = params$column %||% 1)
  }

  list(url = params$path, params = loc)
}

abs_path <- function(x) {
  x <- path.expand(x)
  vcapply(x, abs_path1, USE.NAMES = FALSE)
}

abs_path1 <- function(x) {
  if (grepl("^file://", x)) return(x)
  if (grepl("^/", x)) return(paste0("file://", x))
  if (is_windows() && grepl("^[a-zA-Z]:", x)) return(paste0("file://", x))
  paste0("file://", file.path(getwd(), x))
}

# -- {.fun} ---------------------------------------------------------------

make_link_fun <- function(txt) {
  tolink <- grepl("::", txt, fixed = TRUE)
  linked <- grepl("\007|\033\\\\", txt)
  todo <- tolink & !linked
  if (!any(todo)) return(txt)

  sprt <- ansi_hyperlink_types()$help
  if (!sprt) {
    return(txt)
  }

  fmt <- get_hyperlink_format("help")
  # the format has a placeholder for 'topic'
  topic <- txt[todo]
  done <- style_hyperlink(text = topic, url = glue(fmt))

  txt[todo] <- done

  txt
}

# -- {.help} --------------------------------------------------------------

make_link_help <- function(txt) {
  mch <- re_match(txt, "^\\[(?<text>.*)\\]\\((?<url>.*)\\)$")
  text <- ifelse(is.na(mch$text), txt, mch$text)
  topic <- ifelse(is.na(mch$url), txt, mch$url)

  sprt <- ansi_hyperlink_types()$help
  if (!sprt) {
    topic2 <- vcapply(topic, function(x) format_inline("{.fun ?{x}}"))
    return(ifelse(text == topic, topic2, paste0(text, " (", topic2, ")")))
  }

  fmt <- get_hyperlink_format("help")
  style_hyperlink(text = text, url = glue(fmt))
}

# -- {.href} --------------------------------------------------------------

make_link_href <- function(txt) {
  mch <- re_match(txt, "^\\[(?<text>.*)\\]\\((?<url>.*)\\)$")
  text <- ifelse(is.na(mch$text), txt, mch$text)
  url <- ifelse(is.na(mch$url), txt, mch$url)
  if (ansi_has_hyperlink_support()) {
    link <- style_hyperlink(text = text, url = url)
    style <- is.na(mch$text)
    link[style] <- vcapply(
      url[style],
      function(url1) format_inline("{.url {url1}}")
    )
    link
  } else {
    url2 <- vcapply(url, function(url1) format_inline("{.url {url1}}"))
    ifelse(text == url, url2, paste0(text, " (", url2, ")"))
  }
}

# -- {.run} ---------------------------------------------------------------

make_link_run <- function(txt) {
  mch <- re_match(txt, "^\\[(?<text>.*)\\]\\((?<url>.*)\\)$")
  text <- ifelse(is.na(mch$text), txt, mch$text)
  code <- ifelse(is.na(mch$url), txt, mch$url)

  sprt <- ansi_hyperlink_types()$run
  if (!sprt) {
    return(vcapply(text, function(code1) format_inline("{.code {code1}}")))
  }

  fmt <- get_hyperlink_format("run")
  style_hyperlink(text = text, url = glue(fmt))
}

# -- {.topic} -------------------------------------------------------------

make_link_topic <- function(txt) {
  mch <- re_match(txt, "^\\[(?<text>.*)\\]\\((?<url>.*)\\)$")
  text <- ifelse(is.na(mch$text), txt, mch$text)
  topic <- ifelse(is.na(mch$url), txt, mch$url)

  sprt <- ansi_hyperlink_types()$help
  if (!sprt) {
    topic2 <- vcapply(topic, function(x) format_inline("{.code ?{x}}"))
    return(ifelse(text == topic, topic2, paste0(text, " (", topic2, ")")))
  }

  fmt <- get_hyperlink_format("help")
  style_hyperlink(text = text, url = glue(fmt))
}

# -- {.url} ---------------------------------------------------------------

make_link_url <- function(txt) {
  linked <- grepl("\007|\033\\\\", txt)
  if (all(linked)) return(txt)
  txt[!linked] <- style_hyperlink(txt[!linked], txt[!linked])
  txt
}

# -- {.vignette} ----------------------------------------------------------

make_link_vignette <- function(txt) {
  mch <- re_match(txt, "^\\[(?<text>.*)\\]\\((?<url>.*)\\)$")
  text <- ifelse(is.na(mch$text), txt, mch$text)
  vignette <- ifelse(is.na(mch$url), txt, mch$url)

  sprt <- ansi_hyperlink_types()$vignette
  if (!sprt) {
    vignette2 <- vcapply(vignette, function(x) format_inline("{.code vignette({x})}"))
    return(ifelse(text == vignette, vignette2, paste0(text, " (", vignette2, ")")))
  }

  fmt <- get_hyperlink_format("vignette")
  style_hyperlink(text = text, url = glue(fmt))
}

#' Terminal Hyperlinks
#'
#' `ansi_hyperlink()` creates an ANSI hyperlink.
#'
#' @details
#' This function is currently experimental. In particular, many of the
#' `ansi_*()` functions do not support it properly.
#'
#' `ansi_has_hyperlink_support()` checks if the current `stdout()`
#' supports hyperlinks.
#'
#' See also
#' <https://gist.github.com/egmontkob/eb114294efbcd5adb1944c9f3cb5feda>.
#'
#' @param text Text to show. `text` and `url` are recycled to match their
#'   length, via a `paste0()` call.
#' @param url URL to link to.
#' @param params A named character vector of additional parameters, or `NULL`.
#' @return Styled `cli_ansi_string` for `style_hyperlink()`.
#'   Logical scalar for `ansi_has_hyperlink_support()`.
#'
#' @export
#' @examples
#' cat("This is an", style_hyperlink("R", "https://r-project.org"), "link.\n")

style_hyperlink <- function(text, url, params = NULL) {
  params <- if (length(params)) {
    paste(
      names(params), "=", params,
      collapse = ":"
    )
  }

  if (Sys.getenv("R_CLI_HYPERLINK_MODE") == "posix") {
    ST <- "\033\\"
  } else {
    ST <- "\u0007"
  }

  out <- if (ansi_has_hyperlink_support()) {
    paste0("\u001B]8;", params, ";", url, ST, text, "\u001B]8;;", ST)
  } else {
    text
  }

  class(out) <- c("cli_ansi_string", "ansi_string", "character")
  out
}

#' @export
#' @rdname style_hyperlink
#' @examples
#' ansi_has_hyperlink_support()

ansi_has_hyperlink_support <- function() {

  ## Hyperlinks forced?
  enabled <- getOption("cli.hyperlink", getOption("crayon.hyperlink"))
  if (!is.null(enabled)) { return(isTRUE(enabled)) }

  ## forced by environment variable
  enabled <- Sys.getenv("R_CLI_HYPERLINKS", "")
  if (isTRUE(as.logical(enabled))){ return(TRUE) }

  ## If ANSI support is off, then this is off as well
  if (num_ansi_colors() == 1) return(FALSE)

  ## Are we in RStudio?
  rstudio <- rstudio_detect()
  if (rstudio$type != "not_rstudio") { return(rstudio$hyperlink) }

  ## Are we in a terminal? No?
  if (!isatty(stdout())) { return(FALSE) }

  ## Are we in a windows terminal?
  if (is_windows() && Sys.getenv("WT_SESSION") != "")  { return(TRUE) }

  ## Better to avoid it in CIs
  if (nzchar(Sys.getenv("CI")) ||
      nzchar(Sys.getenv("TEAMCITY_VERSION"))) { return(FALSE) }

  ## iTerm
  if (nzchar(TERM_PROGRAM <- Sys.getenv("TERM_PROGRAM"))) {
    version <- package_version(
      Sys.getenv("TERM_PROGRAM_VERSION"),
      strict = FALSE)

    if (TERM_PROGRAM == "iTerm.app") {
      if (!is.na(version) && version >= "3.1") return(TRUE)
    }
  }

  if (nzchar(VTE_VERSION <- Sys.getenv("VTE_VERSION"))) {
    # See #441 -- some apparent heterogeneity in how the version gets
    #   encoded to this env variable. Accept either form.
    if (grepl("^\\d{4}$", VTE_VERSION)) {
      VTE_VERSION <- sprintf("%2.02f", as.numeric(VTE_VERSION) / 100)
      VTE_VERSION <- package_version(list(major = "0", minor = VTE_VERSION))
    } else {
      VTE_VERSION <- package_version(VTE_VERSION, strict = FALSE)
      if (is.na(VTE_VERSION)) {
        VTE_VERSION <- package_version("0.1.0")
      }
    }
    if (VTE_VERSION >= "0.50.1") return(TRUE)
  }

  FALSE
}


#' @details
#' `ansi_hyperlink_types()` checks if current `stdout()` supports various
#' types of hyperlinks. It returns a list with entries `href`, `run`,
#' `help` and `vignettes`.
#'
#' @rdname style_hyperlink
#' @export

ansi_hyperlink_types <- function() {

  get_config <- function(x, default = NULL) {
    opt <- getOption(paste0("cli.", tolower(x)))
    if (!is.null(opt)) return(isTRUE(opt))

    env <- Sys.getenv(paste0("R_CLI_", toupper(x)), NA_character_)
    if (!is.na(env)) return(isTRUE(as.logical(env)))

    default
  }

  rs <- rstudio_detect()
  has <- ansi_has_hyperlink_support()

  # they are on by default in RStudio, but not otherwise
  run <- get_config("hyperlink_run", default = rs$hyperlink)
  hlp <- get_config("hyperlink_help", default = rs$hyperlink)
  vgn <- get_config("hyperlink_vignette", default = rs$hyperlink)

  if (!has) {
    list(
      href = FALSE,
      run = FALSE,
      help = FALSE,
      vignette = FALSE
    )

  } else if (isTRUE(rs$hyperlink)) {
    list(
      href = TRUE,
      run = structure(run, type = "rstudio"),
      help = structure(hlp, type = "rstudio"),
      vignette = structure(vgn, type = "rstudio")
    )

  } else {
    list(
      href = TRUE,
      run = structure(run, type = "standard"),
      help = structure(hlp, type = "standard"),
      vignette = structure(vgn, type = "standard")
    )
  }
}

get_hyperlink_format <- function(type = c("run", "help", "vignette")) {
  type <- match.arg(type)

  key <- glue("hyperlink_{type}_url_format")
  sprt <- ansi_hyperlink_types()[[type]]

  custom_fmt <- get_config_chr(key)
  if (is.null(custom_fmt)) {
    if (identical(attr(sprt, "type"), "rstudio")) {
      fmt_type <- "rstudio"
    } else {
      fmt_type <- "standard"
    }
  } else {
    fmt_type <- "custom"
  }

  variable <- c(run = "code", help = "topic", vignette = "vignette")
  fmt <- switch(
    fmt_type,
    custom = custom_fmt,
    rstudio = glue("ide:{type}:{{{variable[type]}}}"),
    standard = glue("x-r-{type}:{{{variable[type]}}}")
  )
  fmt
}

get_config_chr <- function(x, default = NULL) {
  opt <- getOption(paste0("cli.", tolower(x)))
  if (!is.null(opt)) {
    stopifnot(is_string(opt))
    return(opt)
  }

  env <- Sys.getenv(paste0("R_CLI_", toupper(x)), NA_character_)
  if (!is.na(env)) return(env)

  default
}

# -- {.email} -------------------------------------------------------------

test_that_cli(configs = c("plain", "fancy"), links = c("all", "none"),
              "{.email}", {

  expect_snapshot({
    cli_text("{.email bugs.bunny@acme.com}")
  })
})

test_that_cli(configs = c("plain", "fancy"), links = c("all", "none"),
              "{.email} vectors", {

  expect_snapshot({
    emails <- paste0("bugs.bunny-", 1:3, "@acme.com")
    cli_text("{.email {emails}}")
  })
})

# -- {.file} and {.path} --------------------------------------------------

test_that_cli(configs = c("plain", "fancy"), links = c("all", "none"),
              "{.file} and {.path}", {

  withr::local_envvar(R_CLI_HYPERLINK_STYLE = NA_character_)

  # absolute path
  expect_snapshot({
    cli_text("{.file /absolute/path}")
    cli_text("{.file file:///absolute/path}")
    cli_text("{.path /absolute/path}")
    cli_text("{.path file:///absolute/path}")
  })

  # relative path
  expect_snapshot({
    cli_text("{.file relative/path}")
    cli_text("{.file ./relative/path}")
    cli_text("{.path relative/path}")
    cli_text("{.path ./relative/path}")
  }, transform = sanitize_wd)

  # ~
  expect_snapshot({
    cli_text("{.file ~/relative/path}")
    cli_text("{.path ~/relative/path}")
  }, transform = sanitize_home)

  # vectorized
  expect_snapshot({
    paths <- c("~/foo", "bar", "file:///abs")
    cli_text("{.file {paths}}")
  }, transform = function(x) sanitize_home(sanitize_wd(x)))

  # weird names
  expect_snapshot({
    paths <- c("foo  ", " bar ", "file:///a bs ")
    cli_text("{.file {paths}}")
  }, transform = sanitize_wd)

  # hand created hyperlink is skipped
  expect_snapshot({
    name <- cli::style_hyperlink("/foo/bar", "/foo/bar")
    cli_text("{.file {name}}")
  })

  # line numbers
  expect_snapshot({
    cli_text("{.file /absolute/path:12}")
    cli_text("{.file file:///absolute/path:5}")
    cli_text("{.path /absolute/path:123}")
    cli_text("{.path file:///absolute/path:51}")
  })
  expect_snapshot({
    cli_text("{.file relative/path:12}")
    cli_text("{.file ./relative/path:5}")
    cli_text("{.path relative/path:123}")
    cli_text("{.path ./relative/path:51}")
  }, transform = sanitize_wd)
  expect_snapshot({
    cli_text("{.file ~/relative/path:12}")
    cli_text("{.path ~/relative/path:5}")
  }, transform = sanitize_home)

  # line and column numbers
  expect_snapshot({
    cli_text("{.file /absolute/path:12:5}")
    cli_text("{.file file:///absolute/path:5:100}")
    cli_text("{.path /absolute/path:123:1}")
    cli_text("{.path file:///absolute/path:51:6}")
  })
  expect_snapshot({
    cli_text("{.file relative/path:12:13}")
    cli_text("{.file ./relative/path:5:20}")
    cli_text("{.path relative/path:123:21}")
    cli_text("{.path ./relative/path:51:2}")
  }, transform = sanitize_wd)
  expect_snapshot({
    cli_text("{.file ~/relative/path:12:23}")
    cli_text("{.path ~/relative/path:5:2}")
  }, transform = sanitize_home)
  expect_snapshot({
    paths <- c("~/foo", "bar:10", "file:///abs:10:20")
    cli_text("{.file {paths}}")
  }, transform = function(x) sanitize_home(sanitize_wd(x)))

  local_mocked_bindings(is_windows = function() TRUE)
  expect_equal(
    abs_path1("c:/foo/bar"),
    "file://c:/foo/bar"
  )
})

# -- {.fun} ---------------------------------------------------------------

test_that_cli(configs = "plain", links = c("all", "none"),
              "{.fun}", {

  expect_snapshot({
    cli_text("{.fun myfun}")
    cli_text("{.fun mypackage::myfun}")
  })

  expect_snapshot({
    funs <- paste0("mypkg::myfun", 1:3)
    cli_text("{.fun {funs}}")
  })
})

test_that_cli(configs = "plain", links = "all", "turning off help", {
  withr::local_options(cli.hyperlink_help = FALSE)
  expect_snapshot({
    cli_text("{.fun pkg::func}")
  })
})

test_that_cli(configs = "plain", links = "all", ".fun with custom format", {
  withr::local_options(cli.hyperlink_help_url_format = "aaa-{topic}-zzz")
  expect_snapshot({
    cli_text("{.fun pkg::func}")
  })
})

# -- {.help} --------------------------------------------------------------

test_that_cli(configs = "plain", links = c("all", "none"),
              "{.help}", {

  expect_snapshot({
    cli_text("{.help pkg::fun}")
    cli_text("{.help [link text](pkg::fun)}")
  })

  expect_snapshot({
    funcs <- paste0("pkg::fun", 1:3)
    cli_text("{.help {funcs}}")
  })
})

test_that_cli(configs = "plain", links = "all", ".help with custom format", {
  withr::local_options(cli.hyperlink_help_url_format = "aaa-{topic}-zzz")
  expect_snapshot({
    cli_text("{.help pkg::fun}")
  })
})

# -- {.href} --------------------------------------------------------------

test_that_cli(configs = "plain", links = c("all", "none"),
              "{.href}", {
  expect_snapshot({
    cli_text("{.href https://cli.r-lib.org}")
    cli_text("{.href [linktext](https://cli.r-lib.org)}")
    cli_text("{.href [link text](https://cli.r-lib.org)}")
  })
})

test_that_cli(configs = "plain", links = c("all", "none"),
              "{.href} vectors", {
  expect_snapshot({
    url <- paste0("https://cli.r-lib.org/", 1:3)
    cli_text("{.href {url}}")
  })
})

# -- {.run} ---------------------------------------------------------------

test_that_cli(configs = "plain", links = c("all", "none"),
              "{.run}", {
  expect_snapshot({
    cli_text("{.run pkg::fun(param)}")
    cli_text("{.run [run(p1, p2)](pkg::fun(p1, p2, other = 'foo'))}")
  })
})

test_that_cli(configs = "plain", links = c("all", "none"),
              "{.run} vectors", {
  expect_snapshot({
    codes <- paste0("pkg::fun", 1:3, "()")
    cli_text("{.run {codes}}")
  })
})

test_that_cli(configs = "plain", links = "all", ".run with custom format", {
  withr::local_options(cli.hyperlink_run_url_format = "aaa-{code}-zzz")
  expect_snapshot({
    cli_text("{.run devtools::document()}")
  })
})

# -- {.topic} -------------------------------------------------------------

test_that_cli(configs = "plain", links = c("all", "none"),
              "{.topic}", {

  expect_snapshot({
    cli_text("{.topic pkg::topic}")
    cli_text("{.topic [link text](pkg::topic)}")
  })

  expect_snapshot({
    topics <- paste0("pkg::topic", 1:3)
    cli_text("{.topic {topics}}")
  })
})

test_that_cli(configs = "plain", links = "all", ".topic with custom format", {
  withr::local_options(cli.hyperlink_help_url_format = "aaa-{topic}-zzz")
  expect_snapshot({
    cli_text("{.topic pkg::fun}")
  })
})

# -- {.url} ---------------------------------------------------------------

test_that_cli(configs = c("plain", "fancy"), links = c("all", "none"),
              "{.url}", {
  expect_snapshot({
    cli_text("{.url https://cli.r-lib.org}")
  })
})

test_that_cli(configs = c("plain", "fancy"), links = c("all", "none"),
              "{.url} vector", {
  expect_snapshot({
    urls <- paste0("https://cli.r-lib.org/", 1:3)
    cli_text("{.url {urls}}")
  })
})

test_that_cli(configs = "plain", links = "all", "linked {.url}", {
  expect_snapshot({
    link <- c(
      "https://cli.r-lib.org",
      style_hyperlink("text", "https://cli.r-lib.org")
    )
    cli_text("{.url {link}}")
  })
})

test_that("make_link_url", {
  withr::local_options(cli.hyperlink = TRUE)
  x <- style_hyperlink(paste0("foo", 1:3), paste0("https://foo.bar/", 1:3))
  expect_equal(make_link_url(x), x)
})

# -- {.vignette} ----------------------------------------------------------

test_that_cli(configs = "plain", links = c("all", "none"),
              "{.vignette}", {

  expect_snapshot({
    cli_text("{.vignette pkg::name}")
    cli_text("{.vignette [link text](pkg::name)}")
  })

  expect_snapshot({
    vignettes <- paste0("pkg::topic", 1:3)
    cli_text("{.vignette {vignettes}}")
  })
})

test_that_cli(configs = "plain", links = "all", ".vignette with custom format", {
  withr::local_options(cli.hyperlink_vignette_url_format = "aaa-{vignette}-zzz")
  expect_snapshot({
    cli_text("{.vignette pkgdown::accessibility}")
  })
})

# -- {.email} -------------------------------------------------------------

test_that_cli(
  configs = c("plain", "fancy"),
  links = c("all", "none"),
  "{.email}",
  {
    expect_snapshot({
      cli_text("{.email bugs.bunny@acme.com}")
    })
  }
)

test_that_cli(
  configs = c("plain", "fancy"),
  links = c("all", "none"),
  "{.email} vectors",
  {
    expect_snapshot({
      emails <- paste0("bugs.bunny-", 1:3, "@acme.com")
      cli_text("{.email {emails}}")
    })
  }
)

# -- {.file} and {.path} --------------------------------------------------

test_that_cli(
  configs = c("plain", "fancy"),
  links = c("all", "none"),
  "{.file} and {.path}",
  {
    withr::local_envvar(R_CLI_HYPERLINK_STYLE = NA_character_)

    # absolute path
    expect_snapshot({
      cli_text("{.file /absolute/path}")
      cli_text("{.file file:///absolute/path}")
      cli_text("{.path /absolute/path}")
      cli_text("{.path file:///absolute/path}")
    })

    # relative path
    expect_snapshot(
      {
        cli_text("{.file relative/path}")
        cli_text("{.file ./relative/path}")
        cli_text("{.path relative/path}")
        cli_text("{.path ./relative/path}")
      },
      transform = sanitize_wd
    )

    # ~
    expect_snapshot(
      {
        cli_text("{.file ~/relative/path}")
        cli_text("{.path ~/relative/path}")
      },
      transform = sanitize_home
    )

    # vectorized
    expect_snapshot(
      {
        paths <- c("~/foo", "bar", "file:///abs")
        cli_text("{.file {paths}}")
      },
      transform = function(x) sanitize_home(sanitize_wd(x))
    )

    # weird names
    expect_snapshot(
      {
        paths <- c("foo  ", " bar ", "file:///a bs ")
        cli_text("{.file {paths}}")
      },
      transform = sanitize_wd
    )

    # hand created hyperlink is skipped
    expect_snapshot({
      name <- cli::style_hyperlink("/foo/bar", "/foo/bar")
      cli_text("{.file {name}}")
    })

    # line numbers
    expect_snapshot({
      cli_text("{.file /absolute/path:12}")
      cli_text("{.file file:///absolute/path:5}")
      cli_text("{.path /absolute/path:123}")
      cli_text("{.path file:///absolute/path:51}")
    })
    expect_snapshot(
      {
        cli_text("{.file relative/path:12}")
        cli_text("{.file ./relative/path:5}")
        cli_text("{.path relative/path:123}")
        cli_text("{.path ./relative/path:51}")
      },
      transform = sanitize_wd
    )
    expect_snapshot(
      {
        cli_text("{.file ~/relative/path:12}")
        cli_text("{.path ~/relative/path:5}")
      },
      transform = sanitize_home
    )

    # line and column numbers
    expect_snapshot({
      cli_text("{.file /absolute/path:12:5}")
      cli_text("{.file file:///absolute/path:5:100}")
      cli_text("{.path /absolute/path:123:1}")
      cli_text("{.path file:///absolute/path:51:6}")
    })
    expect_snapshot(
      {
        cli_text("{.file relative/path:12:13}")
        cli_text("{.file ./relative/path:5:20}")
        cli_text("{.path relative/path:123:21}")
        cli_text("{.path ./relative/path:51:2}")
      },
      transform = sanitize_wd
    )
    expect_snapshot(
      {
        cli_text("{.file ~/relative/path:12:23}")
        cli_text("{.path ~/relative/path:5:2}")
      },
      transform = sanitize_home
    )
    expect_snapshot(
      {
        paths <- c("~/foo", "bar:10", "file:///abs:10:20")
        cli_text("{.file {paths}}")
      },
      transform = function(x) sanitize_home(sanitize_wd(x))
    )

    local_mocked_bindings(is_windows = function() TRUE)
    expect_equal(
      abs_path1("c:/foo/bar"),
      "file://c:/foo/bar"
    )
  }
)

# -- {.fun} ---------------------------------------------------------------

test_that_cli(configs = "plain", links = c("all", "none"), "{.fun}", {
  expect_snapshot({
    cli_text("{.fun myfun}")
    cli_text("{.fun mypackage::myfun}")
  })

  expect_snapshot({
    funs <- paste0("mypkg::myfun", 1:3)
    cli_text("{.fun {funs}}")
  })
})

test_that_cli(configs = "plain", links = "all", "turning off help", {
  withr::local_options(cli.hyperlink_help = FALSE)
  expect_snapshot({
    cli_text("{.fun pkg::func}")
  })
})

test_that_cli(configs = "plain", links = "all", ".fun with custom format", {
  withr::local_options(cli.hyperlink_help_url_format = "aaa-{topic}-zzz")
  expect_snapshot({
    cli_text("{.fun pkg::func}")
  })
})

# -- {.help} --------------------------------------------------------------

test_that_cli(configs = "plain", links = c("all", "none"), "{.help}", {
  expect_snapshot({
    cli_text("{.help pkg::fun}")
    cli_text("{.help [link text](pkg::fun)}")
  })

  expect_snapshot({
    funcs <- paste0("pkg::fun", 1:3)
    cli_text("{.help {funcs}}")
  })
})

test_that_cli(configs = "plain", links = "all", ".help with custom format", {
  withr::local_options(cli.hyperlink_help_url_format = "aaa-{topic}-zzz")
  expect_snapshot({
    cli_text("{.help pkg::fun}")
  })
})

# -- {.href} --------------------------------------------------------------

test_that_cli(configs = "plain", links = c("all", "none"), "{.href}", {
  expect_snapshot({
    cli_text("{.href https://cli.r-lib.org}")
    cli_text("{.href [linktext](https://cli.r-lib.org)}")
    cli_text("{.href [link text](https://cli.r-lib.org)}")
  })
})

test_that_cli(configs = "plain", links = c("all", "none"), "{.href} vectors", {
  expect_snapshot({
    url <- paste0("https://cli.r-lib.org/", 1:3)
    cli_text("{.href {url}}")
  })
})

# -- {.run} ---------------------------------------------------------------

test_that_cli(configs = "plain", links = c("all", "none"), "{.run}", {
  expect_snapshot({
    cli_text("{.run pkg::fun(param)}")
    cli_text("{.run [run(p1, p2)](pkg::fun(p1, p2, other = 'foo'))}")
  })
})

test_that_cli(configs = "plain", links = c("all", "none"), "{.run} vectors", {
  expect_snapshot({
    codes <- paste0("pkg::fun", 1:3, "()")
    cli_text("{.run {codes}}")
  })
})

test_that_cli(configs = "plain", links = "all", ".run with custom format", {
  withr::local_options(cli.hyperlink_run_url_format = "aaa-{code}-zzz")
  expect_snapshot({
    cli_text("{.run devtools::document()}")
  })
})

# -- {.topic} -------------------------------------------------------------

test_that_cli(configs = "plain", links = c("all", "none"), "{.topic}", {
  expect_snapshot({
    cli_text("{.topic pkg::topic}")
    cli_text("{.topic [link text](pkg::topic)}")
  })

  expect_snapshot({
    topics <- paste0("pkg::topic", 1:3)
    cli_text("{.topic {topics}}")
  })
})

test_that_cli(configs = "plain", links = "all", ".topic with custom format", {
  withr::local_options(cli.hyperlink_help_url_format = "aaa-{topic}-zzz")
  expect_snapshot({
    cli_text("{.topic pkg::fun}")
  })
})

# -- {.url} ---------------------------------------------------------------

test_that_cli(
  configs = c("plain", "fancy"),
  links = c("all", "none"),
  "{.url}",
  {
    expect_snapshot({
      cli_text("{.url https://cli.r-lib.org}")
    })
  }
)

test_that_cli(
  configs = c("plain", "fancy"),
  links = c("all", "none"),
  "{.url} vector",
  {
    expect_snapshot({
      urls <- paste0("https://cli.r-lib.org/", 1:3)
      cli_text("{.url {urls}}")
    })
  }
)

test_that_cli(configs = "plain", links = "all", "linked {.url}", {
  expect_snapshot({
    link <- c(
      "https://cli.r-lib.org",
      style_hyperlink("text", "https://cli.r-lib.org")
    )
    cli_text("{.url {link}}")
  })
})

test_that("make_link_url", {
  withr::local_options(cli.hyperlink = TRUE)
  x <- style_hyperlink(paste0("foo", 1:3), paste0("https://foo.bar/", 1:3))
  expect_equal(make_link_url(x), x)
})

# -- {.vignette} ----------------------------------------------------------

test_that_cli(configs = "plain", links = c("all", "none"), "{.vignette}", {
  expect_snapshot({
    cli_text("{.vignette pkg::name}")
    cli_text("{.vignette [link text](pkg::name)}")
  })

  expect_snapshot({
    vignettes <- paste0("pkg::topic", 1:3)
    cli_text("{.vignette {vignettes}}")
  })
})

test_that_cli(
  configs = "plain",
  links = "all",
  ".vignette with custom format",
  {
    withr::local_options(
      cli.hyperlink_vignette_url_format = "aaa-{vignette}-zzz"
    )
    expect_snapshot({
      cli_text("{.vignette pkgdown::accessibility}")
    })
  }
)

format_time <- local({

  parse_ms <- function(ms) {
    stopifnot(is.numeric(ms))

    data.frame(
      days = floor(ms / 86400000),
      hours = floor((ms / 3600000) %% 24),
      minutes = floor((ms / 60000) %% 60),
      seconds = round((ms / 1000) %% 60, 1)
    )
  }

  first_positive <- function(x) which(x > 0)[1]

  trim <- function (x) gsub("^\\s+|\\s+$", "", x)

  pretty_ms <- function(ms, compact = FALSE) {

    stopifnot(is.numeric(ms))

    parsed <- t(parse_ms(ms))

    if (compact) {
      units <- c("d", "h", "m", "s")
      parsed2 <- parsed
      parsed2[] <- paste0(parsed, units)
      idx <- cbind(
        apply(parsed, 2, first_positive),
        seq_len(length(ms))
      )
      tmp <- paste0("~", parsed2[idx])

      # handle NAs
      tmp[is.na(parsed2[idx])] <- NA_character_
      tmp

    } else {

      ## Exact for small ones
      exact            <- paste0(ceiling(ms), "ms")
      exact[is.na(ms)] <- NA_character_

      ## Approximate for others, in seconds
      merge_pieces <- function(pieces) {
        ## handle NAs
        if (all(is.na(pieces))) {
          return(NA_character_)
        }

        ## handle non-NAs
        paste0(
          if (pieces[1]) paste0(pieces[1], "d "),
          if (pieces[2]) paste0(pieces[2], "h "),
          if (pieces[3]) paste0(pieces[3], "m "),
          if (pieces[4]) paste0(pieces[4], "s ")
        )
      }
      approx <- trim(apply(parsed, 2, merge_pieces))

      ifelse(ms < 1000, exact, approx)
    }
  }

  pretty_sec <- function(sec, compact = FALSE) {
    pretty_ms(sec * 1000, compact = compact)
  }

  pretty_dt <- function(dt, compact = FALSE) {

    assert_diff_time(dt)

    units(dt) <- "secs"

    pretty_sec(as.vector(dt), compact = compact)
  }

  structure(
    list(
      .internal = environment(),
      pretty_ms = pretty_ms,
      pretty_sec = pretty_sec,
      pretty_dt = pretty_dt
    ),
    class = c("standalone_time", "standalone")
  )
})

make_progress_bar <- function(percent, width = 30, style = list()) {
  complete_len <- round(width * percent)

  def <- default_progress_style()
  chr_complete <- style[["progress-complete"]] %||% def[["complete"]]
  chr_incomplete <- style[["progress-incomplete"]] %||% def[["incomplete"]]
  chr_current <- style[["progress-current"]] %||% def[["current"]]

  complete <- paste(rep(chr_complete, complete_len), collapse = "")
  current <- if (percent == 100) chr_complete else chr_current
  incomplete <- paste(rep(chr_incomplete, width - complete_len), collapse = "")
  paste0(complete, current, incomplete, " ")
}

default_progress_style <- function() {
  opt <- progress_style(getOption("cli.progress_bar_style"))
  if (is_utf8_output()) {
    opu <- progress_style(getOption("cli.progress_bar_style_unicode"))
    list(
      complete = opu$complete %||% opt$complete %||% "\u25A0",
      current = opu$current %||% opt$current %||% opu$complete %||%
        opt$complete %||% "\u25A0",
      incomplete = opu$incomplete %||% opt$incomplete %||% "\u00a0"
    )
  } else {
    opa <- progress_style(getOption("cli.progress_bar_style_ascii"))
    list(
      complete = opa$complete %||% opt$complete %||% "=",
      current = opa$current %||% opt$current %||% opa$complete %||%
        opt$complete %||% ">",
      incomplete = opa$incomplete %||% opt$incomplete %||% "-"
    )
  }
}

progress_style <- function(x) {
  if (is.null(x)) return(x)
  if (is_string(x)) return(cli_progress_styles()[[x]])
  x
}

#' List of built-in cli progress styles
#'
#' The following options are used to select a style:
#' * `cli_progress_bar_style`
#' * `cli_progress_bar_style_ascii`
#' * `cli_progress_bar_style_unicode`
#'
#' On Unicode terminals (if [is_utf8_output()] is `TRUE`), the
#' `cli_progress_bar_style_unicode` and `cli_progress_bar_style`
#' options are used.
#'
#' On ASCII terminals (if [is_utf8_output()] is `FALSE`), the
#' `cli_pgoress_bar_style_ascii` and `cli_progress_bar_style` options
#' are are used.
#'
#' ```{asciicast progress-style}
#' for (style in names(cli_progress_styles())) {
#'   options(cli.progress_bar_style = style)
#'   label <- ansi_align(paste0("Style '", style, "'"), 20)
#'   print(cli_progress_demo(label, live = FALSE, at = 66, total = 100))
#' }
#' options(cli.progress_var_style = NULL)
#' ```
#'
#' @return A named list with sublists containing elements
#' `complete`, `incomplete` and potentially `current`.
#'
#' @family progress bar functions
#' @export

cli_progress_styles <- function() {
  list(
    classic = list(
      complete = "#",
      incomplete = "\u00a0"
    ),
    squares = list(
      complete = "\u25a0",
      incomplete = "\u00a0"
    ),
    dot = list(
      complete = col_grey("\u2500"),
      incomplete = col_grey("\u2500"),
      current = col_red(symbol$record)
    ),
    fillsquares = list(
      complete = "\u25a0",
      incomplete = col_grey("\u25a1")
    ),
    bar = list(
      complete = "\u2588",
      incomplete = col_grey("\u2588")
    )
  )
}

cliapp <- function(theme = getOption("cli.theme"),
                   user_theme = getOption("cli.user_theme"),
                   output = c("auto", "message", "stdout", "stderr")) {

  app <- new_class(
    "cliapp",

    new = function(theme, user_theme, output)
      clii_init(app, theme, user_theme, output),

    ## Meta
    meta = function(...) {
      txt <- cli__fmt(list(...), collapse = TRUE, app = app)
      clii__message(txt, appendLF = FALSE, output = app$output, signal = app$signal)
    },

    ## Themes
    list_themes = function()
      clii_list_themes(app),
    add_theme = function(theme)
      clii_add_theme(app, theme),
    remove_theme = function(id)
      clii_remove_theme(app, id),

    ## Close container(s)
    end = function(id = NULL)
      clii_end(app, id),

    ## Generic container
    div = function(id = NULL, class = NULL, theme = NULL)
      clii_div(app, id, class, theme),

    ## Paragraphs
    par = function(id = NULL, class = NULL)
      clii_par(app, id, class),

    ## Text, wrapped
    text = function(text)
      clii_text(app, text),

    ## Text, not wrapped
    inline_text = function(text)
      clii_inline_text(app, text),

    ## Text, not wrapped, verbatim
    verbatim = function(...)
      clii_verbatim(app, ...),

    ## Markdow(ish) text, wrapped: emphasis, strong emphasis, links, code
    md_text = function(...)
      clii_md_text(app, ...),

    ## Headings
    h1 = function(text, id = NULL, class = NULL)
      clii_h1(app, text, id, class),
    h2 = function(text, id = NULL, class = NULL)
      clii_h2(app, text, id, class),
    h3 = function(text, id = NULL, class = NULL)
      clii_h3(app, text, id, class),

    ## Block quote
    blockquote = function(quote, citation = NULL, id, class = NULL)
      clii_blockquote(app, quote, citation, id, class),

    ## Lists
    ul = function(items = NULL, id = NULL, class = NULL, .close = TRUE)
      clii_ul(app, items, id, class, .close),
    ol = function(items = NULL, id = NULL, class = NULL, .close = TRUE)
      clii_ol(app, items, id, class, .close),
    dl = function(items = NULL, labels = NULL, id = NULL, class = NULL, .close = TRUE)
      clii_dl(app, items, labels, id, class, .close),
    li = function(items = NULL, labels = NULL, id = NULL, class = NULL)
      clii_li(app, items, labels, id, class),

    ## Tables
    table = function(cells, id = NULL, class = NULL)
      clii_table(app, cells, class),

    ## Alerts
    alert = function(text, id = NULL, class = NULL, wrap = FALSE)
      clii_alert(app, "alert", text, id, class, wrap),
    alert_success = function(text, id = NULL, class = NULL, wrap = FALSE)
      clii_alert(app, "alert-success", text, id, class, wrap),
    alert_danger = function(text, id = NULL, class = NULL, wrap = FALSE)
      clii_alert(app, "alert-danger", text, id, class, wrap),
    alert_warning = function(text, id = NULL, class = NULL, wrap = FALSE)
      clii_alert(app, "alert-warning", text, id, class, wrap),
    alert_info = function(text, id = NULL, class = NULL, wrap = FALSE)
      clii_alert(app, "alert-info", text, id, class, wrap),

    ## Bullets
    bullets = function(text, id = NULL, class = NULL)
      clii_bullets(app, text, id, class),

    ## Horizontal rule
    rule = function(left, center, right, id = NULL)
      clii_rule(app, left, center, right, id),

    ## Status bar
    status = function(id = NULL, msg, msg_done = NULL, msg_failed = NULL,
                      keep = FALSE, auto_result = "clear", globalenv = FALSE)
      clii_status(app, id, msg, msg_done, msg_failed, keep, auto_result, globalenv),
    status_clear = function(id = NULL, result, msg_done = NULL, msg_failed = NULL)
      clii_status_clear(app, id, result, msg_done, msg_failed),
    status_update = function(id = NULL, msg, msg_done = NULL, msg_failed = NULL)
      clii_status_update(app, id, msg, msg_done, msg_failed),

    doc = NULL,
    themes = NULL,
    styles = NULL,
    delayed_item = NULL,
    status_bar = list(),

    margin = 0,
    output = NULL,

    get_current_style = function()
      utils::tail(app$styles, 1)[[1]],

    xtext = function(text = NULL, .list = NULL, indent = 0, padding = 0, wrap = TRUE)
      clii__xtext(app, text, .list = .list, indent = indent,
                  padding = padding, wrap = wrap),

    vspace = function(n = 1)
      clii__vspace(app, n),

    inline = function(text = NULL, .list = NULL)
      clii__inline(app, text, .list = .list),

    item_text = function(type, name, cnt_id, items = list(), .list = NULL)
      clii__item_text(app, type, name, cnt_id, items, .list = .list),

    get_width = function(extra = 0)
      clii__get_width(app, extra),
    cat = function(lines)
      clii__cat(app, lines),
    cat_ln = function(lines, indent = 0, padding = 0)
      clii__cat_ln(app, lines, indent, padding)
  )

  if (! inherits(output, "connection")) output <- match.arg(output)
  app$new(theme, user_theme, output)

  app
}

clii_init <- function(app, theme, user_theme, output) {
  app$doc <- list()
  app$output <- output
  app$styles <- NULL

  if (Sys.getenv("CLI_NO_BUILTIN_THEME", "") != "true") {
    app$add_theme(builtin_theme())
  }
  app$add_theme(theme)
  app$add_theme(user_theme)

  clii__container_start(app, "body", id = "body")

  invisible(app)
}

## Text -------------------------------------------------------------

clii_text <- function(app, text) {
  app$xtext(text)
}

clii_inline_text <- function(app, text) {
  app$xtext(text, wrap = FALSE)
}

clii_verbatim <- function(app, ..., .envir) {
  style <- app$get_current_style()
  text <- unlist(strsplit(unlist(list(...)), "\n", fixed = TRUE))
  if (!is.null(style$fmt)) text <- style$fmt(text)
  app$cat_ln(text)
  invisible(app)
}

clii_md_text <- function(app, ...) {
  throw(cli_error("Markdown text is not implemented yet"))
}

## Headings ----------------------------------------------------------

clii_h1 <- function(app, text, id, class) {
  clii__heading(app, "h1", text, id, class)
}

clii_h2 <- function(app, text, id, class) {
  clii__heading(app, "h2", text, id, class)
}

clii_h3 <- function(app, text, id, class) {
  clii__heading(app, "h3", text, id, class)
}

clii__heading <- function(app, type, text, id, class) {
  id <- new_uuid()
  clii__container_start(app, type, id = id, class = class)
  on.exit(clii__container_end(app, id), add = TRUE)
  text <- app$inline(text)
  style <- app$get_current_style()
  if (is.function(style$fmt)) text <- style$fmt(text)
  app$cat_ln(text)
  invisible(app)
}

## Block quote ------------------------------------------------------

clii_blockquote <- function(app, quote, citation, id, class) {
  c1 <- clii__container_start(app, "blockquote", id = id, class = class)
  on.exit(clii__container_end(app, id), add = TRUE)
  app$xtext(quote)

  c2 <- clii__container_start(app, "cite", id = new_uuid())
  app$xtext(citation)
}

## Table ------------------------------------------------------------

clii_table <- function(app, cells, id, class) {
  throw(cli_error("Tables are not implemented yet"))
}

## Rule -------------------------------------------------------------

clii_rule <- function(app, left, center, right, id) {
  left <- app$inline(left)
  center <- app$inline(center)
  right <- app$inline(right)
  clii__container_start(app, "rule", id = id)
  on.exit(clii__container_end(app, id), add = TRUE)
  style <- app$get_current_style()
  before <- call_if_fun(style$before) %||% ""
  after <- call_if_fun(style$after) %||% ""
  width <- console_width() - ansi_nchar(before) - ansi_nchar(after)
  text <- rule(left, center, right, line = style$`line-type` %||% 1)
  text[1] <- paste0(before, text[1])
  text[length(text)] <- paste0(text[length(text)], after)
  if (is.function(style$fmt)) text <- style$fmt(text)
  app$cat_ln(text)
}

## Alerts -----------------------------------------------------------

clii_alert <- function(app, type, text, id, class, wrap) {
  clii__container_start(app, "div", id = id,
                       class = paste(class, "alert", type))
  on.exit(clii__container_end(app, id), add = TRUE)
  if (wrap) {
    app$xtext(text)
  } else {
    text <- app$inline(text)
    style <- app$get_current_style()
    before <- call_if_fun(style$before) %||% ""
    after <- call_if_fun(style$after) %||% ""
    before <- gsub(" ", "\u00a0", before, fixed = TRUE)
    after <- gsub(" ", "\u00a0", after, fixed = TRUE)
    text[1] <- paste0(before, text[1])
    text[length(text)] <- paste0(text[length(text)], after)
    if (is.function(style$fmt)) text <- style$fmt(text)
    app$cat_ln(text)
  }
}

## Bullets -------------------------------------------------------------

clii_bullets <- function(app, text, id, class) {
  clii__container_start(app, "div", id = id, class = paste("memo bullets", class))
  on.exit(clii__container_end(app, id), add = TRUE)

  # Normalize names a bit, so we can use them as class names
  nms <- as.character(names(text))
  length(nms) <- length(text)
  nms[is.na(nms) | nms == ""] <- "empty"
  nms[nms == " "] <- "space"
  nms <- gsub(" ", "-", nms, fixed = TRUE)
  # cls is vectorized here (!)
  cls <- paste0("bullet memo-item bullet-", nms, " memo-item=", nms)

  lapply(seq_along(text), function(i) {
    iid <- new_uuid()
    clii__container_start(app, "div", id = iid, class = cls[i])
    on.exit(clii__container_end(app, iid), add = TRUE)
    app$text(text[[i]])
  })

  invisible()
}

# Compared to glue::glue(), these are fixed:
# - .sep = ""
# - .trim = TRUE
# - .null = character()
# - .literal = TRUE
# - .comment = ""
#
# we also don't allow passing in data as arguments, and `text` is
# a single argument, no need to `paste()` etc.

glue <- function(text, .envir = parent.frame(),
                 .transformer = identity_transformer,
                 .open = "{", .close = "}", .cli = FALSE, .trim = TRUE) {

  text <- paste0(text, collapse = "")

  if (length(text) < 1L) {
    return(text)
  }

  if (is.na(text)) {
    return(text)
  }

  if (.trim) {
    text <- trim(text)
  }

  f <- function(expr) {
    eval_func <- as.character(.transformer(expr, .envir) %||% character())
  }

  res <- .Call(glue_, text, f, .open, .close, .cli)

  res <- drop_null(res)
  if (any(lengths(res) == 0L)) {
    return(character(0L))
  }

  res[] <- lapply(res, function(x) replace(x, is.na(x), "NA"))

  do.call(paste0, res)
}

count_brace_exp <- function(text, .open = "{", .close = "}") {
  cnt <- 0L
  trans <- function(text, envir) {
    cnt <<- cnt + 1L
    ""
  }
  glue(text, .transformer = trans, .open = .open, .close = .close)
  cnt
}

identity_transformer <- function(text, envir) {
  eval(parse(text = text, keep.source = FALSE), envir)
}

drop_null <- function(x) {
  x[!vapply(x, is.null, logical(1L))]
}

#' Collapse a vector into a string scalar
#'
#' @description
#' Features:
#'
#' - custom separator (`sep`),
#' - custom separator for length-two input (`sep2`),
#' - custom last separator (`last`),
#' - adds ellipsis to truncated strings,
#' - uses Unicode ellipsis character on UTF-8 console,
#' - can collapse "from both ends", with `style = "both-ends"`,
#' - can consider a limit for the display width of the result, in characters,
#' - handles ANSI control sequences correctly when measuring display width.
#'
#' @param x Character vector, or an object with an `as.character()` method
#' to collapse.
#' @param sep Separator. A character string.
#' @param sep2 Separator for the special case that `x` contains only two
#' elements. A character string. Defaults to the value of `last` without the
#' serial comma.
#' @param last Last separator, if there is no truncation. E.g. use
#' `", and "` for the [serial
#' comma](https://en.wikipedia.org/wiki/Serial_comma). A character string.
#' @param trunc Maximum number of elements to show. For `style = "head"`
#' at least `trunc = 1` is used. For `style = "both-ends"` at least
#' `trunc = 5` is used, even if a smaller number is specified.
#' @param width Limit for the display width of the result, in characters.
#' This is a hard limit, and the output will never exceed it.
#' This argument is not implemented for the `"both-ends"` style, which
#' always uses `Inf`, with a warning if a finite `width` value is set.
#' @param ellipsis Character string to use at the place of the truncation.
#' By default, the Unicode ellipsis character is used if the console is
#' UTF-8, and three dots otherwise.
#' @param style Truncation style:
#' * `both-ends`: the default, shows the beginning and end of the vector,
#'   and skips elements in the middle if needed.
#' * `head`: shows the beginning of the vector, and skips elements at the
#'   end, if needed.
#' @return Character scalar. It is `NA_character_` if any elements in `x`
#' are `NA`.
#'
#' @seealso `glue_collapse` in the glue package inspired this function.
#' @export
#' @examples
#' ansi_collapse(letters)
#'
#' # truncate
#' ansi_collapse(letters, trunc = 5)
#'
#' # head style
#' ansi_collapse(letters, trunc = 5, style = "head")

ansi_collapse <- function(x, sep = ", ", sep2 = sub("^,", "", last), last = ", and ",
                          trunc = Inf, width = Inf, ellipsis = symbol$ellipsis,
                          style = c("both-ends", "head")) {

  style <- match.arg(style)
  switch(
    style,
    "both-ends" = collapse_both_ends(
      x, sep, sep2, last, trunc, width, ellipsis
    ),
    "head" = collapse_head(x, sep, sep2, last, trunc, width, ellipsis)
  )
}

collapse_head_notrim <- function(x, trunc, sep, sep2, last, ellipsis) {

  lnx <- length(x)

  if (lnx == 1L) return(x)
  if (lnx == 2L) return(paste0(x, collapse = sep2))
  if (lnx <= trunc) {
    # no truncation
    return(paste0(
      paste(x[1:(lnx - 1L)], collapse = sep),
      last,
      x[lnx]
    ))
  } else {
    # truncation, no need for 'last'
    return(paste0(
      paste(x[1:trunc], collapse = sep),
      sep,
      ellipsis
    ))
  }
}

collapse_head <- function(x, sep, sep2, last, trunc, width, ellipsis) {

  trunc <- max(trunc, 1L)
  x <- as.character(x)
  lnx <- length(x)

  # special cases that do not need trimming
  if (lnx == 0L) {
    return("")
  } else if (anyNA(x)) {
    return(NA_character_)
  }

  # easier case, no width trimming
  if (width == Inf) {
    return(collapse_head_notrim(x, trunc, sep, sep2, last, ellipsis))
  }

  # complex case, with width wrapping
  # first we truncate
  tcd <- lnx > trunc
  if (tcd) x <- x[1:trunc]

  # then we calculate the width w/o trimming
  wx    <- ansi_nchar(x)
  wsep  <- ansi_nchar(sep, "width")
  wsep2 <- ansi_nchar(sep2, "width")
  wlast <- ansi_nchar(last, "width")
  well  <- ansi_nchar(ellipsis, "width")
  if (!tcd) {
    # x[1]
    # x[1] and x[2]
    # x[1], x[2], and x[3]
    nsep  <- if (lnx > 2L) lnx - 2L else 0L
    nsep2 <- if (lnx == 2L) 1L else 0L
    nlast <- if (lnx > 2L) 1L else 0L
    wtot  <- sum(wx) + nsep * wsep + nsep2 * wsep2 + nlast * wlast
    if (wtot <= width) {
      if (lnx == 1L) {
        return(x)
      } else if (lnx == 2L) {
        return(paste0(x, collapse = sep2))
      } else {
        return(paste0(
          paste(x[1:(lnx - 1L)], collapse = sep),
          last,
          x[lnx]
        ))
      }
    }

  } else {
    # x[1], x[2], x[trunc], ...
    wtot <- sum(wx) + trunc * wsep + well
    if (wtot <= width) {
      return(paste0(
        paste(x, collapse = sep),
        sep,
        ellipsis
      ))
    }
  }

  # we need to find the longest possible truncation for the form
  # x[1], x[2], x[trunc], ...
  # each item is wx + wsep wide, so we search how many fits, with ellipsis
  last <- function(x) if (length(x) >= 1L) x[length(x)] else x[NA_integer_]
  trunc <- last(which(cumsum(wx + wsep) + well <= width))

  # not even one element fits
  if (is.na(trunc)) {
    if (well > width) {
      return(ansi_strtrim(ellipsis, width, ellipsis = ""))
    } else if (well == width) {
      return(ellipsis)
    } else if (well + wsep >= width) {
      return(paste0(ansi_strtrim(x[1L], width, ellipsis = ""), ellipsis))
    } else {
      return(paste0(
        ansi_strtrim(x[1L], max(width - well - wsep, 0L), ellipsis = ellipsis),
        sep,
        ellipsis
      ))
    }
  }

  return(paste0(
    paste(x[1:trunc], collapse = sep),
    sep,
    ellipsis
  ))
}

collapse_both_ends <- function(x, sep, sep2, last, trunc, width, ellipsis) {

  if (width != Inf) {
    warning(format_warning(c(
      "!" = "finite {.arg width} is not implemented in {.fun cli::ansi_collapse}.",
      "i" = "{.code width = Inf} is used instead."
    )))
    width <- Inf
  }

  # we always list at least 5 elements
  trunc <- max(trunc, 5L)
  trunc <- min(trunc, length(x))
  if (length(x) <= 5L || length(x) <= trunc) {
    return(collapse_head(x, sep, sep2, last, trunc = trunc, width, ellipsis))
  }

  # we have at least six elements in the vector
  # 1, 2, 3, ..., 9, and 10
  x <- as.character(c(x[1:(trunc - 2L)], x[length(x) - 1L], x[length(x)]))
  paste0(
    c(x[1:(trunc - 2L)], ellipsis, paste0(x[trunc - 1L], last, x[trunc])),
    collapse = sep
  )
}

trim <- function(x) {
  has_newline <- function(x) any(grepl("\\n", x))
  if (length(x) == 0L || !has_newline(x)) {
    return(x)
  }
  .Call(trim_, x)
}

json <- "https://raw.githubusercontent.com/sindresorhus/cli-spinners/45cef9dff64ac5e36b46a194c68bccba448899ac/spinners.json"
parsed <- jsonlite::fromJSON(json, simplifyVector = TRUE)
pasis <- lapply(parsed, function(x) { x$frames <- I(x$frames); x })
pdt <- as.data.frame(do.call(rbind, pasis))
pdt$name <- rownames(pdt)
rownames(pdt) <- NULL
spinners <- pdt[, c("name", "interval", "frames")]
usethis::use_data(spinners, internal = TRUE)
spinners <- rbind(
  spinners,
  list(name = "growVeriticalDotsLR", interval = 80, frames = strsplit("⠀⡀⣀⣄⣤⣦⣶⣷⣿⣾⣶⣴⣤⣠⣀⢀", "", fixed = TRUE)),
  list(name = "growVeriticalDotsRL", interval = 80, frames = strsplit("⠀⢀⣀⣠⣤⣴⣶⣾⣿⣷⣶⣦⣤⣄⣀⡀", "", fixed = TRUE)),
  list(name = "growVeriticalDotsLL", interval = 80, frames = strsplit("⠀⡀⣀⣄⣤⣦⣶⣷⣿⣷⣶⣦⣤⣄⣀⡀", "", fixed = TRUE)),
  list(name = "growVeriticalDotsRR", interval = 80, frames = strsplit("⠀⡀⣀⣠⣤⣴⣶⣾⣿⣾⣶⣴⣤⣠⣀⢀", "", fixed = TRUE))
)

get_palette_color <- function(style, colors = num_ansi_colors()) {
  opt <- getOption("cli.palette")
  if (is.null(opt) || colors < 256) return(style)
  cache_palette_color(opt, style$palette, colors)
}

palette_cache <- new.env(parent = emptyenv())

cache_palette_color <- function(pal, idx, colors = num_ansi_colors()) {
  if (is_string(pal)) {
    if (! pal %in% rownames(ansi_palettes)) {
      opt <- options(cli.palette = NULL)
      defer(options(opt))
      throw(cli_error(
        "Cannot find cli ANSI palette {.val {pal}}",
        "i" = "Know palettes are {.val {rownames(cli::ansi_palettes)}}.",
        "i" = "Maybe the {.code cli.palette} option is incorrect?"
      ))
    }
    pal <- ansi_palettes[pal, ]
  }

  bg <- idx < 0
  idx <- abs(idx)
  col <- pal[[idx]]

  colkey <- as.character(colors)
  key <- paste0(col, bg)
  if (key %in% names(palette_cache[[colkey]])) {
    return(palette_cache[[colkey]][[key]])
  }

  val <- ansi_style_from_r_color(
    col,
    bg = bg,
    colors,
    grey = FALSE
  )

  if (is.null(palette_cache[[colkey]])) {
    palette_cache[[colkey]] <- new.env(parent = emptyenv())
  }
  palette_cache[[colkey]][[key]] <- val

  return(val)
}

#' @details
#' `truecolor` is an integer constant for the number of 24 bit ANSI colors.
#'
#' @format `truecolor` is an integer scalar.
#'
#' @export
#' @rdname ansi_palettes

truecolor <- as.integer(256 ^ 3)

#' ANSI colors palettes
#'
#' If your platform supports at least 256 colors, then you can configure
#' the colors that cli uses for the eight base and the eight bright colors.
#' (I.e. the colors of [col_black()], [col_red()], and [col_br_black()],
#' [col_br_red()], etc.
#'
#' To customize the default palette, set the `cli.palette` option to the
#' name of a built-in palette (see `ansi_palettes()`), or the list of
#' 16 colors. Colors can be specified with RGB colors strings:
#' `#rrggbb` or R color names (see the output of [grDevices::colors()]).
#'
#' For example, you can put this in your R profile:
#' ```r
#' options(cli.palette = "vscode")
#' ```
#'
#' It is currently not possible to configure the background colors
#' separately, these will be always the same as the foreground colors.
#'
#' If your platform only has 256 colors, then the colors specified in the
#' palette have to be interpolated. On true color platforms they RGB
#' values are used as-is.
#'
#' `ansi_palettes` is a data frame of the built-in palettes, each row
#' is one palette.
#'
#' `ansi_palette_show()` shows the colors of an ANSI palette on the screen.
#'
#' @format `ansi_palettes` is a data frame with one row for each palette,
#'   and one column for each base ANSI color. `attr(ansi_palettes, "info")`
#'   contains a list with information about each palette.
#'
#' @export
#' @examples
#' ansi_palettes
#' ansi_palette_show("dichro", colors = truecolor)

ansi_palettes <- rbind(
  utils::read.table(
    "tools/ansi-palettes.txt",
    comment.char = ";",
    stringsAsFactors = FALSE
  ),
  utils::read.table(
    "tools/ansi-iterm-palettes.txt",
    comment.char = ";",
    stringsAsFactors = FALSE
  )
)

attr(ansi_palettes, "info") <-
  list(
    dichro = paste(
      "Colorblind friendly palette, from",
      "https://github.com/romainl/vim-dichromatic#dichromatic."
    ),
    vga = paste(
      "Typical colors that are used when booting PCs and leaving them in",
      "text mode, which used a 16-entry color table. The colors are",
      "different in the EGA/VGA graphic modes.",
      "From https://en.wikipedia.org/wiki/ANSI_escape_code#SGR."
    ),
    winxp = paste(
      "Windows XP Console. Seen in Windows XP through Windows 8.1.",
      "From https://en.wikipedia.org/wiki/ANSI_escape_code#SGR."
    ),
    vscode = paste(
      "Visual Studio Debug console, 'Dark+' theme.",
      "From https://en.wikipedia.org/wiki/ANSI_escape_code#SGR."
    ),
    win10 = paste0(
      "Campbell theme, used as of Windows 10 version 1709. Also used",
      "by PowerShell 6.",
      "From https://en.wikipedia.org/wiki/ANSI_escape_code#SGR."
    ),
    macos = paste0(
      "Terminal.app in macOS",
      "From https://en.wikipedia.org/wiki/ANSI_escape_code#SGR."
    ),
    putty = paste0(
      "From https://en.wikipedia.org/wiki/ANSI_escape_code#SGR."
    ),
    mirc = paste0(
      "From https://en.wikipedia.org/wiki/ANSI_escape_code#SGR."
    ),
    xterm = paste0(
      "From https://en.wikipedia.org/wiki/ANSI_escape_code#SGR."
    ),
    ubuntu = paste0(
      "For virtual terminals, from /etc/vtrgb.",
      "From https://en.wikipedia.org/wiki/ANSI_escape_code#SGR."
    ),
    eclipse = paste0(
      "From https://en.wikipedia.org/wiki/ANSI_escape_code#SGR."
    ),
    iterm = "Built-in iTerm2 theme.",
    "iterm-pastel" = "Built-In iTerm2 theme.",
    "iterm-smoooooth" = "Built-In iTerm2 theme.",
    "iterm-snazzy" = "From https://github.com/sindresorhus/iterm2-snazzy.",
    "iterm-solarized" = "Built-In iTerm2 theme.",
    "iterm-tango" = "Built-In iTerm2 theme."
  )

#' @param palette The palette to show, in the same format as for the
#'   `cli.palette` option, so it can be the name of a built-in palette,
#'   of a list of 16 colors.
#' @param colors Number of ANSI colors to use the show the palette. If the
#'   platform does not have sufficient support, the output might have
#'   a lower color resolution. Without color support it will have no color
#'   at all.
#' @param rows The number of colored rows to print.
#' @return `ansi_palette_show` returns a character vector, the rows that
#'   are printed to the screen, invisibly.
#'
#' @export
#' @rdname ansi_palettes

ansi_palette_show <- function(palette = NULL, colors = num_ansi_colors(),
                              rows = 4) {
  opts <- options(
    cli.palette = palette %||% getOption("cli.palette"),
    cli.num_colors = colors
  )
  on.exit(options(opts), add = TRUE)

  blk <- strrep(symbol$lower_block_8, 4)
  blks <- c(
    "blck" = col_black(blk),
    "red " = col_red(blk),
    "grn " = col_green(blk),
    "yllw" = col_yellow(blk),
    "blue" = col_blue(blk),
    "mgnt" = col_magenta(blk),
    "cyan" = col_cyan(blk),
    "whte" = col_white(blk),
    "blck" = col_br_black(blk),
    "red " = col_br_red(blk),
    "grn " = col_br_green(blk),
    "yllw" = col_br_yellow(blk),
    "blue" = col_br_blue(blk),
    "mgnt" = col_br_magenta(blk),
    "cyan" = col_br_cyan(blk),
    "whte" = col_br_white(blk)
  )

  join <- function(x) {
    paste0(
      paste(x[1:8], collapse = " "),
      "  ",
      paste(x[9:16], collapse = " ")
    )
  }

  nms <- join(names(blks))
  str <- join(blks)

  out <- c(
    paste(strrep(" ", 52), "bright variants"),
    nms,
    "",
    rep(str, rows)
  )

  cat_line(out)
  invisible(out)
}

test_that("ansi_palette_show", {
  local_clean_cli_context()
  expect_snapshot(
    ansi_palette_show(colors = truecolor)
  )

  withr::local_options(cli.palette = "iterm-snazzy")
  expect_snapshot(
    ansi_palette_show(colors = truecolor)
  )
})

test_that("error", {
  expect_snapshot(
    error = TRUE,
    withr::with_options(
      list(cli.palette = "foobar12"),
      ansi_palette_show(colors = 256)
    )
  )
})

test_that("custom palettes", {
  withr::local_options(
    cli.num_colors = 256,
    cli.palette = "iterm-snazzy"
  )
  expect_snapshot({
    col_black("black")
    col_red("red")
    col_green("green")
    col_yellow("yellow")
    col_blue("blue")
    col_magenta("magenta")
    col_cyan("cyan")
    col_white("white")

    col_br_black("br_black")
    col_br_red("br_red")
    col_br_green("br_green")
    col_br_yellow("br_yellow")
    col_br_blue("br_blue")
    col_br_magenta("br_magenta")
    col_br_cyan("br_cyan")
    col_br_white("br_white")
  })
})

is_string <- function(x) {
  is.character(x) && length(x) == 1 && !is.na(x)
}

is_flag <- function(x) {
  is.logical(x) && length(x) == 1 && !is.na(x)
}

is_border_style <- function(x) {
  is_string(x) && x %in% rownames(box_styles())
}

is_padding_or_margin <- function(x) {
  is.numeric(x) && length(x) %in% c(1, 4) && !anyNA(x) &&
    all(as.integer(x) == x)
}

is_col <- function(x) {
  is.null(x) || is_string(x) || is.function(x)
}

is_count <- function(x) {
  is.numeric(x) && length(x) == 1 && !is.na(x) && as.integer(x) == x &&
    x >= 0
}

is_tree_style <- function(x) {
  is.list(x) &&
    length(x) == 4 &&
    !is.null(names(x)) &&
    all(sort(names(x)) == sort(c("h", "v", "l", "j"))) &&
    all(sapply(x, is_string))
}

is_named <- function(x) {
  !is.null(names(x))
}

test_that("is_string", {

  strings <- list("foo", "", "111", "1", "-", "NA")
  not_strings <- list(1, character(), NA_character_, NA,
                      c("foo", NA), c("1", "2"), NULL)

  for (p in strings) {
    expect_true(is_string(p))
    expect_silent(stopifnot(is_string(p)))
  }

  for (n in not_strings) {
    expect_false(is_string(n))
    expect_error(
      stopifnot(is_string(n)),
      "is_string(n) is not TRUE",
      fixed = TRUE
    )
  }
})

test_that("is_border_style", {
  expect_true(is_border_style(rownames(box_styles())[1]))
  expect_false(is_border_style("blahblahxxx"))

  expect_silent(stopifnot(is_border_style(rownames(box_styles())[1])))
  expect_error(
    stopifnot(is_border_style("blahblahxxx")),
    "is_border_style(\"blahblahxxx\") is not TRUE",
    fixed = TRUE
  )
})

test_that("is_padding_or_margin", {
  good <- list(1, 0, 0L, 1L, 237, c(1,2,3,4), c(0,0,0,0), rep(1L, 4))
  bad <- list(numeric(), integer(), c(1,2), c(1L, 2L, 3L), 1:5,
              "1", c("1", "2", "3", "1"), NA, NA_real_, NA_integer_,
              c(1,2,NA,1), c(1L,NA,3L))

  for (g in good) {
    expect_true(is_padding_or_margin(g))
    expect_silent(stopifnot(is_padding_or_margin(g)))
  }
  for (b in bad) {
    expect_false(is_padding_or_margin(b))
    expect_error(
      stopifnot(is_padding_or_margin(b)),
      "is_padding_or_margin(b) is not TRUE",
      fixed = TRUE
    )
  }
})

test_that("is_col", {
  good <- list("red", "orange", NULL, col_red)
  bad <- list(c("red", "orange"), character(), NA_character_)

  for (g in good) {
    expect_true(is_col(g))
    expect_silent(stopifnot(is_col(g)))
  }
  for (b in bad) {
    expect_false(is_col(b))
    expect_error(
      stopifnot(is_col(b)),
      "is_col(b) is not TRUE",
      fixed = TRUE
    )
  }
})

test_that("is_count", {

  counts <- list(1, 1L, 0, 0L, 42, 42L)
  not_counts <- list(c(1, 2), numeric(), NA_integer_, NA_real_, NA, 1.1,
                     NULL, "1")

  for (c in counts) {
    expect_true(is_count(c))
    expect_silent(stopifnot(is_count(c)))
  }

  for (n in not_counts) {
    expect_false(is_count(n))
    expect_error(
      stopifnot(is_count(n)),
      "is_count(n) is not TRUE",
      fixed = TRUE
    )
  }
})

test_that("is_tree_style", {
  good <- list(
    list(h = "1", v = "2", l = "3", j = "4"),
    list(j = "4", v = "2", h = "1", l = "3")
  )
  bad <- list(
    NULL,
    1:4,
    c(h = "1", v = "2", l = "3", j = "4"),
    list(h = "1", v = "2", l = "3", j = "4", x = "10"),
    list(h = "1", v = c("2", "3"), l = "3", j = "4"),
    list(h = "1", v = "2", l = character(), j = "4"),
    list(h = "1", v = "2", l = 3, j = "4"),
    list("1", v = "2", l = "3", j = "4"),
    list("1", "2", "3", "4")
  )

  for (x in good) expect_true (is_tree_style(x))
  for (x in bad ) expect_false(is_tree_style(x))
})

test_that("is_string", {
  strings <- list("foo", "", "111", "1", "-", "NA")
  not_strings <- list(
    1,
    character(),
    NA_character_,
    NA,
    c("foo", NA),
    c("1", "2"),
    NULL
  )

  for (p in strings) {
    expect_true(is_string(p))
    expect_silent(stopifnot(is_string(p)))
  }

  for (n in not_strings) {
    expect_false(is_string(n))
    expect_snapshot(error = TRUE, stopifnot(is_string(n)))
  }
})

test_that("is_border_style", {
  expect_true(is_border_style(rownames(box_styles())[1]))
  expect_false(is_border_style("blahblahxxx"))

  expect_silent(stopifnot(is_border_style(rownames(box_styles())[1])))
  expect_snapshot(error = TRUE, stopifnot(is_border_style("blahblahxxx")))
})

test_that("is_padding_or_margin", {
  good <- list(1, 0, 0L, 1L, 237, c(1, 2, 3, 4), c(0, 0, 0, 0), rep(1L, 4))
  bad <- list(
    numeric(),
    integer(),
    c(1, 2),
    c(1L, 2L, 3L),
    1:5,
    "1",
    c("1", "2", "3", "1"),
    NA,
    NA_real_,
    NA_integer_,
    c(1, 2, NA, 1),
    c(1L, NA, 3L)
  )

  for (g in good) {
    expect_true(is_padding_or_margin(g))
    expect_silent(stopifnot(is_padding_or_margin(g)))
  }
  for (b in bad) {
    expect_false(is_padding_or_margin(b))
    expect_snapshot(error = TRUE, stopifnot(is_padding_or_margin(b)))
  }
})

test_that("is_col", {
  good <- list("red", "orange", NULL, col_red)
  bad <- list(c("red", "orange"), character(), NA_character_)

  for (g in good) {
    expect_true(is_col(g))
    expect_silent(stopifnot(is_col(g)))
  }
  for (b in bad) {
    expect_false(is_col(b))
    expect_snapshot(error = TRUE, stopifnot(is_col(b)))
  }
})

test_that("is_count", {
  counts <- list(1, 1L, 0, 0L, 42, 42L)
  not_counts <- list(
    c(1, 2),
    numeric(),
    NA_integer_,
    NA_real_,
    NA,
    1.1,
    NULL,
    "1"
  )

  for (c in counts) {
    expect_true(is_count(c))
    expect_silent(stopifnot(is_count(c)))
  }

  for (n in not_counts) {
    expect_false(is_count(n))
    expect_snapshot(error = TRUE, stopifnot(is_count(n)))
  }
})

test_that("is_tree_style", {
  good <- list(
    list(h = "1", v = "2", l = "3", j = "4"),
    list(j = "4", v = "2", h = "1", l = "3")
  )
  bad <- list(
    NULL,
    1:4,
    c(h = "1", v = "2", l = "3", j = "4"),
    list(h = "1", v = "2", l = "3", j = "4", x = "10"),
    list(h = "1", v = c("2", "3"), l = "3", j = "4"),
    list(h = "1", v = "2", l = character(), j = "4"),
    list(h = "1", v = "2", l = 3, j = "4"),
    list("1", v = "2", l = "3", j = "4"),
    list("1", "2", "3", "4")
  )

  for (x in good) expect_true(is_tree_style(x))
  for (x in bad) expect_false(is_tree_style(x))
})

# NOTE: patched to use `cli_error()`

# nocov start --- compat-defer ---
#
# This drop-in file implements withr::defer(). Please find the most
# recent version in withr's repository.
#
# 2022-03-03
# * Support for `source()` and `knitr::knit()`
# * Handlers are now stored in environments instead of lists to avoid
#   infinite recursion issues.
# * The handler list is now soft-namespaced.

defer <- function(expr, envir = parent.frame(), priority = c("first", "last")) {
}

local({
  defer <<- defer <- function(
    expr,
    envir = parent.frame(),
    priority = c("first", "last")
  ) {
    priority <- match.arg(priority)
    invisible(
      add_handler(
        envir,
        handler = new_handler(substitute(expr), parent.frame()),
        front = priority == "first"
      )
    )
  }

  new_handler <- function(expr, envir) {
    hnd <- new.env(FALSE, size = 2)
    hnd[["expr"]] <- expr
    hnd[["envir"]] <- envir
    hnd
  }

  add_handler <- function(
    envir,
    handler,
    front,
    frames = as.list(sys.frames()),
    calls = as.list(sys.calls())
  ) {
    envir <- exit_frame(envir, frames, calls)

    if (front) {
      handlers <- c(list(handler), get_handlers(envir))
    } else {
      handlers <- c(get_handlers(envir), list(handler))
    }

    set_handlers(envir, handlers, frames = frames, calls = calls)
    handler
  }

  set_handlers <- function(envir, handlers, frames, calls) {
    if (is.null(get_handlers(envir))) {
      # Ensure that list of handlers called when environment "ends"
      setup_handlers(envir)
    }

    attr(envir, "withr_handlers") <- handlers
  }

  # Evaluate `frames` lazily
  setup_handlers <- function(
    envir,
    frames = as.list(sys.frames()),
    calls = as.list(sys.calls())
  ) {
    if (is_top_level_global_env(envir, frames)) {
      # For session scopes we use reg.finalizer()
      if (is_interactive()) {
        message(
          sprintf("Setting global deferred event(s).\n"),
          "i These will be run:\n",
          "  * Automatically, when the R session ends.\n",
          "  * On demand, if you call `withr::deferred_run()`.\n",
          "i Use `withr::deferred_clear()` to clear them without executing."
        )
      }
      reg.finalizer(envir, function(env) deferred_run(env), onexit = TRUE)
    } else {
      # for everything else we use on.exit()

      call <- make_call(execute_handlers, envir)
      # We have to use do.call here instead of eval because of the way on.exit
      # determines its evaluation context
      # (https://stat.ethz.ch/pipermail/r-devel/2013-November/067867.html)

      do.call(base::on.exit, list(call, TRUE), envir = envir)
    }
  }

  exit_frame <- function(
    envir,
    frames = as.list(sys.frames()),
    calls = as.list(sys.calls())
  ) {
    frame_loc <- frame_loc(envir, frames)
    if (!frame_loc) {
      return(envir)
    }

    if (in_knitr(envir)) {
      out <- knitr_frame(envir, frames, calls, frame_loc)
      if (!is.null(out)) {
        return(out)
      }
    }

    out <- source_frame(envir, frames, calls, frame_loc)
    if (!is.null(out)) {
      return(out)
    }

    envir
  }

  knitr_frame <- function(envir, frames, calls, frame_loc) {
    knitr_ns <- asNamespace("knitr")

    # This doesn't handle correctly the recursive case (knitr called
    # within a chunk). Handling this would be a little fiddly for an
    # uncommon edge case.
    for (i in seq(1, frame_loc)) {
      if (identical(topenv(frames[[i]]), knitr_ns)) {
        return(frames[[i]])
      }
    }

    NULL
  }

  source_frame <- function(envir, frames, calls, frame_loc) {
    i <- frame_loc

    if (i < 4) {
      return(NULL)
    }

    is_call <- function(x, fn) {
      is.call(x) && identical(x[[1]], fn)
    }
    calls <- as.list(calls)

    if (!is_call(calls[[i - 3]], quote(source))) {
      return(NULL)
    }
    if (!is_call(calls[[i - 2]], quote(withVisible))) {
      return(NULL)
    }
    if (!is_call(calls[[i - 1]], quote(eval))) {
      return(NULL)
    }
    if (!is_call(calls[[i - 0]], quote(eval))) {
      return(NULL)
    }

    frames[[i - 3]]
  }

  frame_loc <- function(envir, frames) {
    n <- length(frames)
    if (!n) {
      return(0)
    }

    for (i in seq_along(frames)) {
      if (identical(frames[[n - i + 1]], envir)) {
        return(n - i + 1)
      }
    }

    0
  }

  in_knitr <- function(envir) {
    knitr_in_progress() && identical(knitr::knit_global(), envir)
  }

  is_top_level_global_env <- function(envir, frames) {
    if (!identical(envir, globalenv())) {
      return(FALSE)
    }

    # Check if another global environment is on the stack
    !any(vapply(frames, identical, NA, globalenv()))
  }

  get_handlers <- function(envir) {
    attr(envir, "withr_handlers")
  }

  execute_handlers <- function(envir) {
    handlers <- get_handlers(envir)
    errors <- list()
    for (handler in handlers) {
      tryCatch(eval(handler$expr, handler$envir), error = function(e) {
        errors[[length(errors) + 1]] <<- e
      })
    }
    attr(envir, "withr_handlers") <- NULL

    for (error in errors) {
      stop(error) %??%
        cli_error("Error in a deferred {.code on.exit()} clause")
    }
  }

  make_call <- function(...) {
    as.call(list(...))
  }

  # base implementation of rlang::is_interactive()
  is_interactive <- function() {
    opt <- getOption("rlang_interactive")
    if (!is.null(opt)) {
      return(opt)
    }
    if (knitr_in_progress()) {
      return(FALSE)
    }
    if (identical(Sys.getenv("TESTTHAT"), "true")) {
      return(FALSE)
    }
    interactive()
  }

  knitr_in_progress <- function() {
    isTRUE(getOption("knitr.in.progress"))
  }
}) # defer() namespace

# nocov end

cliappenv <- new.env()
cliappenv$stack <- list()
cliappenv$pid <- Sys.getpid()

#' Start, stop, query the default cli application
#'
#' `start_app` creates an app, and places it on the top of the app stack.
#'
#' `stop_app` removes the top app, or multiple apps from the app stack.
#'
#' `default_app` returns the default app, the one on the top of the stack.
#'
#' @param theme Theme to use.
#' @param output How to print the output.
#' @param .auto_close Whether to stop the app, when the calling frame
#'   is destroyed.
#' @param .envir The environment to use, instead of the calling frame,
#'   to trigger the stop of the app.
#' @param app App to stop. If `NULL`, the current default app is stopped.
#'   Otherwise we find the supplied app in the app stack, and remote it,
#'   together with all the apps above it.
#' @return
#'   `start_app` returns the new app, `default_app` returns the default app.
#'   `stop_app` does not return anything.
#'
#' @export

start_app <- function(theme = getOption("cli.theme"),
                      output = c("auto", "message", "stdout", "stderr"),
                      .auto_close = TRUE, .envir = parent.frame()) {

  if (! inherits(output, "connection")) output <- match.arg(output)

  app <- cliapp(
    theme = theme,
    user_theme = getOption("cli.user_theme"),
    output = output
  )
  cliappenv$stack[[length(cliappenv$stack) + 1]] <- app

  if (.auto_close && !identical(.envir, globalenv())) {
    defer(stop_app(app = app), envir = .envir, priority = "first")
  }

  invisible(app)
}

#' @export
#' @name start_app

stop_app <- function(app = NULL) {
  if (is.null(app)) {
    cliappenv$stack <- utils::head(cliappenv$stack, -1)

  } else {
    if (!inherits(app, "cliapp")) stop("Not a CLI app")
    ndl <- format.default(app)
    nms <- vapply(cliappenv$stack, format.default, character(1))
    if (! ndl %in% nms) {
      warning("No app to end")
      return()
    }
    wh <- which(nms == ndl)[1]
    cliappenv$stack <- utils::head(cliappenv$stack, wh - 1)
  }

  invisible()
}

#' @export
#' @name start_app

default_app <- function() {
  top <- utils::tail(cliappenv$stack, 1)
  if (length(top)) top[[1]] else NULL
}

start_app()
on.exit(stop_app(), add = TRUE)

test_that("collapsing without formatting, n>3", {
  expect_snapshot({
    pkgs <- paste0("pkg", 1:5)
    cli_text("Packages: {pkgs}.")
  })
})

test_that("collapsing without formatting, n<3", {
  expect_snapshot({
    pkgs <- paste0("pkg", 1:2)
    cli_text("Packages: {pkgs}.")
  })
})

test_that("collapsing with formatting", {
  expect_snapshot(local({
    cli_div(theme = list(".pkg" = list(fmt = function(x) paste0(x, " (P)"))))
    pkgs <- paste0("pkg", 1:5)
    cli_text("Packages: {.pkg {pkgs}}.")
  }))
})

test_that("collapsing with formatting, custom seps", {
  expect_snapshot(local({
    cli_div(theme = list(div = list("vec-sep" = " ... ")))
    pkgs <- paste0("pkg", 1:5)
    cli_text("Packages: {.pkg {pkgs}}.")
  }))
})

test_that("collapsing a cli_vec", {
  expect_snapshot({
    pkgs <- cli_vec(
      paste0("pkg", 1:5),
      style = list("vec-sep" = " & ", "vec-last" = " & ")
    )
    cli_text("Packages: {pkgs}.")
  })
})

test_that_cli(configs = c("plain", "ansi"), "collapsing a cli_vec with styling", {
  expect_snapshot(local({
    cli_div(theme = list(body = list("vec-sep" = " ... ")))
    pkgs <- cli_vec(
      paste0("pkg", 1:5),
      style = list("vec-sep" = " & ", "vec-last" = " & ", color = "blue")
    )
    cli_text("Packages: {pkgs}.")
  }))
})

test_that("head", {
  v <- function(n, t = 5) {
    cli_vec(
      seq_len(n),
      style = list("vec-trunc-style" = "head", "vec-trunc" = t)
    )
  }
  expect_snapshot({
    cli_text("{v(0,1)}")
    cli_text("{v(1,1)}")
    cli_text("{v(2,1)}")
    cli_text("{v(3,1)}")
    cli_text("{v(4,1)}")

    cli_text("{v(0,2)}")
    cli_text("{v(1,2)}")
    cli_text("{v(2,2)}")
    cli_text("{v(3,2)}")
    cli_text("{v(4,2)}")

    cli_text("{v(0,3)}")
    cli_text("{v(1,3)}")
    cli_text("{v(2,3)}")
    cli_text("{v(3,3)}")
    cli_text("{v(4,3)}")

    cli_text("{v(0,4)}")
    cli_text("{v(1,4)}")
    cli_text("{v(2,4)}")
    cli_text("{v(3,4)}")
    cli_text("{v(4,4)}")

    cli_text("{v(0,5)}")
    cli_text("{v(1,5)}")
    cli_text("{v(2,5)}")
    cli_text("{v(3,5)}")
    cli_text("{v(4,5)}")

    cli_text("{v(10,5)}")
  })
})

test_that("both-ends", {
  v <- function(n, t = 5) {
    cli_vec(
      seq_len(n),
      style = list("vec-trunc-style" = "both-ends", "vec-trunc" = t)
    )
  }
  expect_snapshot({
    cli_text("{v(0,1)}")
    cli_text("{v(1,1)}")
    cli_text("{v(2,1)}")
    cli_text("{v(3,1)}")
    cli_text("{v(4,1)}")
    cli_text("{v(5,1)}")
    cli_text("{v(6,1)}")
    cli_text("{v(7,1)}")
    cli_text("{v(10,1)}")
  })
})

test_that_cli(configs = c("plain", "ansi"), "both-ends with formatting", {
  v <- function(n, t = 5) {
    cli_vec(
      seq_len(n),
      style = list("vec-trunc-style" = "both-ends", "vec-trunc" = t)
    )
  }
  expect_snapshot({
    cli_text("{.val {v(0,1)}}")
    cli_text("{.val {v(1,1)}}")
    cli_text("{.val {v(2,1)}}")
    cli_text("{.val {v(3,1)}}")
    cli_text("{.val {v(4,1)}}")
    cli_text("{.val {v(5,1)}}")
    cli_text("{.val {v(6,1)}}")
    cli_text("{.val {v(7,1)}}")
    cli_text("{.val {v(10,1)}}")

    cli_text("{.val {v(10,6)}}")
    cli_text("{.val {v(10,10)}}")
    cli_text("{.val {v(11,10)}}")
  })
})

test_that("ansi_collapse", {
  l10 <- letters[1:10]
  expect_snapshot({
    ansi_collapse(l10)
    ansi_collapse(l10, trunc = 6)
    ansi_collapse(l10, trunc = 5)
    ansi_collapse(l10, trunc = 4)
    ansi_collapse(l10, trunc = 1)
    ansi_collapse(l10, sep = "; ")
    ansi_collapse(l10, sep = "; ", last = "; or ")
    ansi_collapse(l10, sep = "; ")
    ansi_collapse(l10, sep = "; ", last = "; or ", trunc = 6)

    ansi_collapse(l10, style = "head")
    ansi_collapse(l10, trunc = 6, style = "head")
    ansi_collapse(l10, trunc = 5, style = "head")
    ansi_collapse(l10, trunc = 4, style = "head")
    ansi_collapse(l10, trunc = 1, style = "head")
    ansi_collapse(l10, sep = "; ", style = "head")
    ansi_collapse(l10, sep = "; ", last = "; or ", style = "head")
    ansi_collapse(l10, sep = "; ", last = "; or ", trunc = 6, style = "head")
  })
})

test_that("ansi_collapse with width trimming", {
  l10 <- letters[1:10]
  expect_snapshot({
    ansi_collapse(l10, width = 1, style = "head")
    ansi_collapse(l10, width = 2, style = "head")
    ansi_collapse(l10, width = 3, style = "head")
    ansi_collapse(l10, width = 4, style = "head")
    ansi_collapse(l10, width = 5, style = "head")
    ansi_collapse(l10, width = 6, style = "head")
    ansi_collapse(l10, width = 7, style = "head")
    ansi_collapse(l10, width = 8, style = "head")
    ansi_collapse(l10, width = 9, style = "head")
    ansi_collapse(l10, width = 30, style = "head")
    ansi_collapse(l10, width = 31, style = "head")
    ansi_collapse(l10, width = 32, style = "head")
    ansi_collapse(l10, width = 40, style = "head")
  })

  expect_snapshot({
    ansi_collapse(l10, width = 10, style = "both-ends")
  })
})

test_that("ansi_collapse produces consistent truncation results", {
  expect_equal(ansi_collapse(1:2, trunc = 1, style = "head"),
               ansi_collapse(1:2, trunc = 0, style = "head"))
})

test_that("ansi_collapse uses `sep2` for length-two inputs", {
  expect_equal(ansi_collapse(1:2),
               "1 and 2")
  expect_equal(ansi_collapse(1:2, trunc = 2, style = "head"),
               "1 and 2")
})

test_that("Avoid duplication of length 1 vecs when width set (#590)", {
  expect_equal(ansi_collapse(1), "1")
  expect_equal(ansi_collapse(1, style = "head"), "1")
  expect_equal(ansi_collapse(1, style = "head", width = 70), "1")
  expect_equal(ansi_collapse(1, style = "head", last = " and again "), "1")
  expect_equal(ansi_collapse(1, style = "head", width = 70, last = " and again "), "1")
})

test_that("Issue #681", {
  # sep2 takes precedence
  expect_equal(ansi_collapse(1:2, sep2 = " and ", last = " or "), "1 and 2")
  expect_equal(ansi_collapse(1:2, sep2 = " xor ", last = " or "), "1 xor 2")
  # default for sep2 is last without the Oxford comma
  expect_equal(ansi_collapse(1:3, last = ", or "), "1, 2, or 3")
  expect_equal(ansi_collapse(1:2, last = ", or "), "1 or 2")
  expect_equal(ansi_collapse(1:2, last = " or "), "1 or 2")

  expect_snapshot({
    v <- cli::cli_vec(
      c("foo", "bar", "foobar"),
      style = list("vec-last" = ", or ")
    )
    cli::cli_text("Must be one of: {v}.")

    v <- cli::cli_vec(
      c("foo", "bar"),
      style = list("vec-last" = " or ")
    )
    cli::cli_text("Must be one of: {v}.")

    v <- cli::cli_vec(
      c("foo", "bar"),
      style = list("vec-last" = " or ", "vec-sep2" = " xor ")
    )
    cli::cli_text("Must be one of: {v}.")
  })
})

withr::local_envvar(CLI_NO_BUILTIN_THEME = "true")
withr::local_options(cli.theme = NULL, cli.user_theme = NULL)
start_app()
on.exit(stop_app(), add = TRUE)

test_that_cli(configs = c("plain", "ansi"), "inline classes", {
  classes <- c(
    "emph", "strong", "code", "pkg", "fun", "arg", "key", "file", "path",
    "email", "url", "var", "envvar", "cls")

  do <- function(class) {

    special_style <- structure(
      list(
        list(color = "cyan"),
        list(before = "<<<"),
        list(after =">>>")),
      names = c(
        paste0("span.", class),
        paste0("span.", class),
        paste0("span.", class)
      )
    )

    cli_div(theme = special_style)
    txt <- glue::glue("This is {.<class> it} really",
                      .open = "<", .close = ">")
    cli_text(txt)
  }

  expect_snapshot(
    invisible(lapply(classes, do))
  )
})

test_that("{{ and }} can be used for comments", {
  expect_snapshot(local({
    cli_text("Escaping {{ works")
    cli_text("Escaping }} works")
    cli_text("Escaping {{ and }} works")
    cli_text("Escaping {{{{ works")
    cli_text("Escaping }}}} works")
    cli_text("Escaping {{{{ and }} works")
    cli_text("Escaping {{{{ and }}}} works")
    cli_text("Escaping {{ and }}}} works")
  }))
})

test_that("no glue substitution in expressions that evaluate to a string", {
  expect_snapshot(local({
    msg <- "Message with special characters like } { }} {{"
    cli_text("{msg}")
    cli_text("{.emph {msg}}")
  }))
})

test_that("S3 class is used for styling", {
  expect_snapshot(local({
    cli_div(
      theme = list(
        div = list("class-map" = list("foo" = "bar")),
        ".bar" = list(before = "::"))
    )
    obj <- structure("yep", class = "foo")
    cli_text("This is {obj}.")
  }))
})

context("verbatim text")

setup(start_app())
teardown(stop_app())

test_that("verbatim text is correctly styled", {
  theme <- list(.padded = list("margin-left" = 4))
  cli_div(class = "padded", theme = theme)

  lines <- c("first", "second", "third")

  out0 <- capt0(cli_verbatim(lines))
  out1 <- capt0(cli_verbatim(paste0(lines, collapse = "\n")))

  # expect the margin being applied to all lines
  expected <- paste0(paste0("    ", lines, collapse = "\n"), "\n")

  expect_equal(out0, expected)
  expect_equal(out1, expected)
})

test_that("stop_app() errors", {
  expect_snapshot(
    error = TRUE,
    stop_app(1:10)
  )
})

test_that("warning if inactive app", {
  app <- start_app(.auto_close = FALSE)
  stop_app(app)
  expect_snapshot(
    stop_app(app)
  )
})

start_app()
on.exit(stop_app(), add = TRUE)

test_that("verbatim text is correctly styled", {
  expect_snapshot(local({
    theme <- list(.padded = list("margin-left" = 4))
    cli_div(class = "padded", theme = theme)

    lines <- c("first", "second", "third")

    cli_verbatim(lines)
    cli_verbatim(paste0(lines, collapse = "\n"))
  }))
})

start_app()
on.exit(stop_app(), add = TRUE)

test_that("cliapp output auto", {
  skip_on_cran()

  txt <- "Stay calm. This is a test."
  script <- tempfile("clitest1-", fileext = ".R")
  on.exit(unlink(script, recursive = TRUE), add = TRUE)

  # stderr if not interactive ----------------

  code <- substitute(env = list(txt = txt), {
    options(rlib_interactive = FALSE, cli.num_colors = 1L)
    cli::cli_text(txt)
  })
  cat(deparse(code), file = script, sep = "\n")

  out <- callr::rscript(script, show = FALSE, fail_on_status = FALSE)
  expect_true(out$stderr %in% paste0(txt, c("\n", "\r\n")))
  expect_equal(out$stdout, "")

  # stdout if interactive --------------------

  code <- substitute(env = list(txt = txt), {
    options(rlib_interactive = TRUE, cli.num_colors = 1L)
    cli::cli_text(txt)
  })
  cat(deparse(code), file = script, sep = "\n")

  out <- callr::rscript(script, show = FALSE, fail_on_status = FALSE)
  expect_true(out$stdout %in% paste0(txt, c("\n", "\r\n")))
  expect_equal(out$stderr, "")

  # choose explicitly -----------------------

  txt2 <- "Don't move"
  code <- substitute(env = list(txt = txt, txt2 = txt2), {
    options(rlib_interactive = FALSE, cli_num_colors = 1L)
    cli::start_app(output = "stderr")
    cli::cli_text(txt)
    cli::stop_app()
    options(rlib_interactive = TRUE)
    cli::start_app(output = "stdout")
    cli::cli_text(txt2)
    cli::stop_app()
  })
  cat(deparse(code), file = script, sep = "\n")

  out <- callr::rscript(script, show = FALSE, fail_on_status = FALSE)
  expect_true(out$stderr %in% paste0(txt, c("\n", "\r\n")))
  expect_true(out$stdout %in% paste0(txt2, c("\n", "\r\n")))
})

test_that("can also use a connection", {
  skip_on_cran()

  txt <- "Stay calm. This is a test."
  script <- tempfile("clitest2-", fileext = ".R")
  on.exit(unlink(script, recursive = TRUE), add = TRUE)

  code <- substitute(env = list(txt = txt), {
    options(cli.num_colors = 1L)
    con <- textConnection(NULL, open = "w", local = TRUE)
    cli::start_app(output = con)
    cli::cli_text(txt)
    cli::stop_app()
    flush(con)
    cat("output:", textConnectionValue(con), "\n", sep = "")
  })
  cat(deparse(code), file = script, sep = "\n")

  out <- callr::rscript(script, show = FALSE, fail_on_status = FALSE)
  expect_true(out$stdout %in% paste0("output:", txt, c("\n", "\r\n")))
  expect_equal(out$stderr, "")
})

test_that("message if there is a sink", {
  # if there is an output sink, non-interactive
  msgs <- NULL
  tmp <- NULL
  fun <- function() {
    sink(tmp <<- tempfile("clisinktest-"))
    on.exit(sink(NULL), add = TRUE)
    cli_text("Hola")
  }
  withCallingHandlers(
    fun(),
    cliMessage = function(m) {
      msgs <<- c(msgs, list(m))
      invokeRestart("muffleMessage")
    }
  )
  expect_equal(msgs[[1]]$message, "Hola\n")
  expect_equal(length(readLines(tmp)), 0)

  # if there is a message sink, non-interactive
  msgs <- NULL
  tmp <- tempfile("clisinktest2-")
  con <- file(tmp, open = "w+")
  on.exit(close(con), add = TRUE)
  fun <- function() {
    sink(con, type = "message")
    on.exit(sink(NULL, type = "message"), add = TRUE)
    cat("this\n", file = stderr())
    cli_text("Hola")
  }
  withCallingHandlers(
    fun(),
    cliMessage = function(m) {
      msgs <<- c(msgs, list(m))
      invokeRestart("muffleMessage")
    }
  )
  expect_equal(msgs[[1]]$message, "Hola\n")
  expect_equal(readLines(tmp), "this")

  withr::local_options(rlib_interactive = TRUE)

  # if there is an output sink, interactive
  msgs <- NULL
  tmp <- NULL
  fun <- function() {
    sink(tmp <<- tempfile("clisinktest3-"))
    on.exit(sink(NULL), add = TRUE)
    cli_text("Hola")
  }
  withCallingHandlers(
    fun(),
    cliMessage = function(m) {
      msgs <<- c(msgs, list(m))
      invokeRestart("muffleMessage")
    }
  )
  expect_equal(msgs[[1]]$message, "Hola\n")
  expect_equal(length(readLines(tmp)), 0)

  # if there is a message sink, interactive
  msgs <- NULL
  tmp <- tempfile("clisinktest4-")
  con2 <- file(tmp, open = "w+")
  on.exit(close(con2), add = TRUE)
  fun <- function() {
    sink(con2, type = "message")
    on.exit(sink(NULL, type = "message"), add = TRUE)
    cat("this\n", file = stderr())
    cli_text("Hola")
  }
  withCallingHandlers(
    fun(),
    cliMessage = function(m) {
      msgs <<- c(msgs, list(m))
      invokeRestart("muffleMessage")
    }
  )
  expect_equal(msgs[[1]]$message, "Hola\n")
  expect_equal(readLines(tmp), "this")
})

start_app()
on.exit(stop_app(), add = TRUE)

test_that("cliapp output auto", {
  skip_on_cran()

  txt <- "Stay calm. This is a test."
  script <- tempfile(fileext = ".R")
  on.exit(unlink(script, recursive = TRUE), add = TRUE)

  # stderr if not interactive ----------------

  code <- substitute(env = list(txt = txt), {
    options(rlib_interactive = FALSE, cli.num_colors = 1L)
    cli::cli_text(txt)
  })
  cat(deparse(code), file = script, sep = "\n")

  out <- callr::rscript(script, show = FALSE, fail_on_status = FALSE)
  expect_true(out$stderr %in% paste0(txt, c("\n", "\r\n")))
  expect_equal(out$stdout, "")

  # stdout if interactive --------------------

  code <- substitute(env = list(txt = txt), {
    options(rlib_interactive = TRUE, cli.num_colors = 1L)
    cli::cli_text(txt)
  })
  cat(deparse(code), file = script, sep = "\n")

  out <- callr::rscript(script, show = FALSE, fail_on_status = FALSE)
  expect_true(out$stdout %in% paste0(txt, c("\n", "\r\n")))
  expect_equal(out$stderr, "")

  # choose explicitly -----------------------

  txt2 <- "Don't move"
  code <- substitute(env = list(txt = txt, txt2 = txt2), {
    options(rlib_interactive = FALSE, cli_num_colors = 1L)
    cli::start_app(output = "stderr")
    cli::cli_text(txt)
    cli::stop_app()
    options(rlib_interactive = TRUE)
    cli::start_app(output = "stdout")
    cli::cli_text(txt2)
    cli::stop_app()
  })
  cat(deparse(code), file = script, sep = "\n")

  out <- callr::rscript(script, show = FALSE, fail_on_status = FALSE)
  expect_true(out$stderr %in% paste0(txt, c("\n", "\r\n")))
  expect_true(out$stdout %in% paste0(txt2, c("\n", "\r\n")))
})

test_that("can also use a connection", {
  skip_on_cran()

  txt <- "Stay calm. This is a test."
  script <- tempfile(fileext = ".R")
  on.exit(unlink(script, recursive = TRUE), add = TRUE)

  code <- substitute(env = list(txt = txt), {
    options(cli.num_colors = 1L)
    con <- textConnection(NULL, open = "w", local = TRUE)
    cli::start_app(output = con)
    cli::cli_text(txt)
    cli::stop_app()
    flush(con)
    cat("output:", textConnectionValue(con), "\n", sep = "")
  })
  cat(deparse(code), file = script, sep = "\n")

  out <- callr::rscript(script, show = FALSE, fail_on_status = FALSE)
  expect_true(out$stdout %in% paste0("output:", txt, c("\n", "\r\n")))
  expect_equal(out$stderr, "")
})

test_that("message if there is a sink", {
  # if there is an output sink, non-interactive
  msgs <- NULL
  tmp <- NULL
  fun <- function() {
    sink(tmp <<- tempfile())
    on.exit(sink(NULL), add = TRUE)
    cli_text("Hola")
  }
  withCallingHandlers(
    fun(),
    cliMessage = function(m) {
      msgs <<- c(msgs, list(m))
      invokeRestart("muffleMessage")
    }
  )
  expect_equal(msgs[[1]]$message, "Hola\n")
  expect_equal(length(readLines(tmp)), 0)

  # if there is a message sink, non-interactive
  msgs <- NULL
  tmp <- tempfile()
  con <- file(tmp, open = "w+")
  fun <- function() {
    sink(con, type = "message")
    on.exit(sink(NULL, type = "message"), add = TRUE)
    cat("this\n", file = stderr())
    cli_text("Hola")
  }
  withCallingHandlers(
    fun(),
    cliMessage = function(m) {
      msgs <<- c(msgs, list(m))
      invokeRestart("muffleMessage")
    }
  )
  expect_equal(msgs[[1]]$message, "Hola\n")
  expect_equal(readLines(tmp), "this")

  withr::local_options(rlib_interactive = TRUE)

  # if there is an output sink, interactive
  msgs <- NULL
  tmp <- NULL
  fun <- function() {
    sink(tmp <<- tempfile())
    on.exit(sink(NULL), add = TRUE)
    cli_text("Hola")
  }
  withCallingHandlers(
    fun(),
    cliMessage = function(m) {
      msgs <<- c(msgs, list(m))
      invokeRestart("muffleMessage")
    }
  )
  expect_equal(msgs[[1]]$message, "Hola\n")
  expect_equal(length(readLines(tmp)), 0)

  # if there is a message sink, interactive
  msgs <- NULL
  tmp <- tempfile()
  con <- file(tmp, open = "w+")
  fun <- function() {
    sink(con, type = "message")
    on.exit(sink(NULL, type = "message"), add = TRUE)
    cat("this\n", file = stderr())
    cli_text("Hola")
  }
  withCallingHandlers(
    fun(),
    cliMessage = function(m) {
      msgs <<- c(msgs, list(m))
      invokeRestart("muffleMessage")
    }
  )
  expect_equal(msgs[[1]]$message, "Hola\n")
  expect_equal(readLines(tmp), "this")
})

withr::local_envvar(CLI_NO_BUILTIN_THEME = "true")
withr::local_options(cli.theme = NULL, cli.user_theme = NULL)
start_app()
on.exit(stop_app(), add = TRUE)

test_that_cli(config = c("plain", "ansi"), "inline classes", {
  classes <- c(
    "emph", "strong", "code", "pkg", "fun", "arg", "key", "file", "path",
    "email", "url", "var", "envvar", "cls")

  do <- function(class) {

    special_style <- structure(
      list(
        list(color = "cyan"),
        list(before = "<<<"),
        list(after =">>>")),
      names = c(
        paste0("span.", class),
        paste0("span.", class),
        paste0("span.", class)
      )
    )

    cli_div(theme = special_style)
    txt <- glue::glue("This is {.<class> it} really",
                      .open = "<", .close = ">")
    cli_text(txt)
  }

  expect_snapshot(
    invisible(lapply(classes, do))
  )
})

test_that("{{ and }} can be used for comments", {
  expect_snapshot(local({
    cli_text("Escaping {{ works")
    cli_text("Escaping }} works")
    cli_text("Escaping {{ and }} works")
    cli_text("Escaping {{{{ works")
    cli_text("Escaping }}}} works")
    cli_text("Escaping {{{{ and }} works")
    cli_text("Escaping {{{{ and }}}} works")
    cli_text("Escaping {{ and }}}} works")
  }))
})

test_that("no glue substitution in expressions that evaluate to a string", {
  expect_snapshot(local({
    msg <- "Message with special characters like } { }} {{"
    cli_text("{msg}")
    cli_text("{.emph {msg}}")
  }))
})

test_that("S3 class is used for styling", {
  expect_snapshot(local({
    cli_div(
      theme = list(
        div = list("class-map" = list("foo" = "bar")),
        ".bar" = list(before = "::"))
    )
    obj <- structure("yep", class = "foo")
    cli_text("This is {obj}.")
  }))
})

start_app()
on.exit(stop_app(), add = TRUE)

test_that("collapsing without formatting, n>3", {
  expect_snapshot({
    pkgs <- paste0("pkg", 1:5)
    cli_text("Packages: {pkgs}.")
  })
})

test_that("collapsing without formatting, n<3", {
  expect_snapshot({
    pkgs <- paste0("pkg", 1:2)
    cli_text("Packages: {pkgs}.")
  })
})

test_that("collapsing with formatting", {
  expect_snapshot(local({
    cli_div(theme = list(".pkg" = list(fmt = function(x) paste0(x, " (P)"))))
    pkgs <- paste0("pkg", 1:5)
    cli_text("Packages: {.pkg {pkgs}}.")
  }))
})

test_that("collapsing with formatting, custom seps", {
  expect_snapshot(local({
    cli_div(theme = list(div = list("vec-sep" = " ... ")))
    pkgs <- paste0("pkg", 1:5)
    cli_text("Packages: {.pkg {pkgs}}.")
  }))
})

test_that("collapsing a cli_vec", {
  expect_snapshot({
    pkgs <- cli_vec(
      paste0("pkg", 1:5),
      style = list("vec-sep" = " & ", "vec-last" = " & ")
    )
    cli_text("Packages: {pkgs}.")
  })
})

test_that_cli(configs = c("plain", "ansi"), "collapsing a cli_vec with styling", {
  expect_snapshot(local({
    cli_div(theme = list(body = list("vec-sep" = " ... ")))
    pkgs <- cli_vec(
      paste0("pkg", 1:5),
      style = list("vec-sep" = " & ", "vec-last" = " & ", color = "blue")
    )
    cli_text("Packages: {pkgs}.")
  }))
})

test_that("head", {
  v <- function(n, t = 5) {
    cli_vec(
      seq_len(n),
      style = list("vec-trunc-style" = "head", "vec-trunc" = t)
    )
  }
  expect_snapshot({
    cli_text("{v(0,1)}")
    cli_text("{v(1,1)}")
    cli_text("{v(2,1)}")
    cli_text("{v(3,1)}")
    cli_text("{v(4,1)}")

    cli_text("{v(0,2)}")
    cli_text("{v(1,2)}")
    cli_text("{v(2,2)}")
    cli_text("{v(3,2)}")
    cli_text("{v(4,2)}")

    cli_text("{v(0,3)}")
    cli_text("{v(1,3)}")
    cli_text("{v(2,3)}")
    cli_text("{v(3,3)}")
    cli_text("{v(4,3)}")

    cli_text("{v(0,4)}")
    cli_text("{v(1,4)}")
    cli_text("{v(2,4)}")
    cli_text("{v(3,4)}")
    cli_text("{v(4,4)}")

    cli_text("{v(0,5)}")
    cli_text("{v(1,5)}")
    cli_text("{v(2,5)}")
    cli_text("{v(3,5)}")
    cli_text("{v(4,5)}")

    cli_text("{v(10,5)}")
  })
})

test_that("both-ends", {
  v <- function(n, t = 5) {
    cli_vec(
      seq_len(n),
      style = list("vec-trunc-style" = "both-ends", "vec-trunc" = t)
    )
  }
  expect_snapshot({
    cli_text("{v(0,1)}")
    cli_text("{v(1,1)}")
    cli_text("{v(2,1)}")
    cli_text("{v(3,1)}")
    cli_text("{v(4,1)}")
    cli_text("{v(5,1)}")
    cli_text("{v(6,1)}")
    cli_text("{v(7,1)}")
    cli_text("{v(10,1)}")
  })
})

test_that_cli(config = c("plain", "ansi"), "both-ends with formatting", {
  v <- function(n, t = 5) {
    cli_vec(
      seq_len(n),
      style = list("vec-trunc-style" = "both-ends", "vec-trunc" = t)
    )
  }
  expect_snapshot({
    cli_text("{.val {v(0,1)}}")
    cli_text("{.val {v(1,1)}}")
    cli_text("{.val {v(2,1)}}")
    cli_text("{.val {v(3,1)}}")
    cli_text("{.val {v(4,1)}}")
    cli_text("{.val {v(5,1)}}")
    cli_text("{.val {v(6,1)}}")
    cli_text("{.val {v(7,1)}}")
    cli_text("{.val {v(10,1)}}")

    cli_text("{.val {v(10,6)}}")
    cli_text("{.val {v(10,10)}}")
    cli_text("{.val {v(11,10)}}")
  })
})

test_that("ansi_collapse", {
  l10 <- letters[1:10]
  expect_snapshot({
    ansi_collapse(l10)
    ansi_collapse(l10, trunc = 6)
    ansi_collapse(l10, trunc = 5)
    ansi_collapse(l10, trunc = 4)
    ansi_collapse(l10, trunc = 1)
    ansi_collapse(l10, sep = "; ")
    ansi_collapse(l10, sep = "; ", last = "; or ")
    ansi_collapse(l10, sep = "; ")
    ansi_collapse(l10, sep = "; ", last = "; or ", trunc = 6)

    ansi_collapse(l10, style = "head")
    ansi_collapse(l10, trunc = 6, style = "head")
    ansi_collapse(l10, trunc = 5, style = "head")
    ansi_collapse(l10, trunc = 4, style = "head")
    ansi_collapse(l10, trunc = 1, style = "head")
    ansi_collapse(l10, sep = "; ", style = "head")
    ansi_collapse(l10, sep = "; ", last = "; or ", style = "head")
    ansi_collapse(l10, sep = "; ", last = "; or ", trunc = 6, style = "head")
  })
})

test_that("ansi_collapse with width trimming", {
  l10 <- letters[1:10]
  expect_snapshot({
    ansi_collapse(l10, width = 1, style = "head")
    ansi_collapse(l10, width = 2, style = "head")
    ansi_collapse(l10, width = 3, style = "head")
    ansi_collapse(l10, width = 4, style = "head")
    ansi_collapse(l10, width = 5, style = "head")
    ansi_collapse(l10, width = 6, style = "head")
    ansi_collapse(l10, width = 7, style = "head")
    ansi_collapse(l10, width = 8, style = "head")
    ansi_collapse(l10, width = 9, style = "head")
    ansi_collapse(l10, width = 30, style = "head")
    ansi_collapse(l10, width = 31, style = "head")
    ansi_collapse(l10, width = 32, style = "head")
    ansi_collapse(l10, width = 40, style = "head")
  })

  expect_snapshot({
    ansi_collapse(l10, width = 10, style = "both-ends")
  })
})

test_that("ansi_collapse produces consistent truncation results", {
  expect_equal(ansi_collapse(1:2, trunc = 1, style = "head"),
               ansi_collapse(1:2, trunc = 0, style = "head"))
})

test_that("ansi_collapse uses `sep2` for length-two inputs", {
  expect_equal(ansi_collapse(1:2),
               "1 and 2")
  expect_equal(ansi_collapse(1:2, trunc = 2, style = "head"),
               "1 and 2")
})

start_app()
on.exit(stop_app(), add = TRUE)

test_that("glue errors", {
  skip_if_no_srcrefs()
  expect_snapshot(error = TRUE, {
    cli_h1("foo { asdfasdfasdf } bar")
    cli_text("foo {cmd {dsfsdf()}}")
  })
})

#' Signal an error, warning or message with a cli formatted
#' message
#'
#' These functions let you create error, warning or diagnostic
#' messages with cli formatting, including inline styling,
#' pluralization and glue substitutions.
#'
#' @details
#'
#' ```{asciicast cli-abort}
#' n <- "boo"
#' cli_abort(c(
#'         "{.var n} must be a numeric vector",
#'   "x" = "You've supplied a {.cls {class(n)}} vector."
#' ))
#' ```
#'
#' ```{asciicast cli-abort-2}
#' len <- 26
#' idx <- 100
#' cli_abort(c(
#'         "Must index an existing element:",
#'   "i" = "There {?is/are} {len} element{?s}.",
#'   "x" = "You've tried to subset element {idx}."
#' ))
#' ```
#'
#' @param message It is formatted via a call to [cli_bullets()].
#' @param ... Passed to [rlang::abort()], [rlang::warn()] or
#'   [rlang::inform()].
#' @param .envir Environment to evaluate the glue expressions in.
#' @inheritParams rlang::abort
#'
#' @seealso These functions support [inline markup][inline-markup].
#' @family functions supporting inline markup
#' @export

cli_abort <- function(message,
                      ...,
                      call = .envir,
                      .envir = parent.frame(),
                      .frame = .envir) {
  message[] <- vcapply(message, format_inline, .envir = .envir)
  rlang::abort(
    message,
    ...,
    call = call,
    use_cli_format = TRUE,
    .frame = .frame
  )
}

#' @rdname cli_abort
#' @export

cli_warn <- function(message, ..., .envir = parent.frame()) {
  rlang::warn(
    format_warning(message, .envir = .envir),
    ...
  )
}

#' @rdname cli_abort
#' @export

cli_inform <- function(message, ..., .envir = parent.frame()) {
  rlang::inform(
    format_message(message, .envir = .envir),
    ...
  )
}

test_that_cli("cli_abort", {
  withr::local_options(cli.theme_dark = FALSE)
  expect_snapshot(error = TRUE, local({
    n <- "boo"
    cli_abort(c(
            "{.var n} must be a numeric vector",
      "x" = "You've supplied a {.cls {class(n)}} vector."
    ))
  }))

  expect_snapshot(error = TRUE, local({
    len <- 26
    idx <- 100
    cli_abort(c(
            "Must index an existing element:",
      "i" = "There {?is/are} {len} element{?s}.",
      "x" = "You've tried to subset element {idx}."
    ))
  }))

  n <- "boo"
  err <- tryCatch(
    cli_abort(c(
      "{.var n} must be a numeric vector",
      "x" = "You've supplied a {.cls {class(n)}} vector."
    )),
    error = function(e) e
  )
  expect_snapshot(c(err$message, err$body))
})

test_that_cli("cli_warn", {
  skip_if_not_installed("rlang", "1.0.0")

  withr::local_options(cli.theme_dark = FALSE)
  expect_snapshot({
    n <- "boo"
    cli_warn(c(
            "{.var n} must be a numeric vector",
      "x" = "You've supplied a {.cls {class(n)}} vector."
    ))
  })

  expect_snapshot(local({
    len <- 26
    idx <- 100
    cli_warn(c(
            "Must index an existing element:",
      "i" = "There {?is/are} {len} element{?s}.",
      "x" = "You've tried to subset element {idx}."
    ))
  }))
})

test_that_cli("cli_inform", {
  skip_if_not_installed("rlang", "1.0.0")

  withr::local_options(cli.ansi = FALSE)
  expect_snapshot({
    n <- "boo"
    cli_inform(c(
            "{.var n} must be a numeric vector",
      "x" = "You've supplied a {.cls {class(n)}} vector."
    ))
  })

  expect_snapshot(local({
    len <- 26
    idx <- 100
    cli_inform(c(
            "Must index an existing element:",
      "i" = "There {?is/are} {len} element{?s}.",
      "x" = "You've tried to subset element {idx}."
    ))
  }))
})

test_that("cli_abort width in RStudio", {
  # this is to fix breakage with new testthat
  withr::local_options(cli.condition_width = getOption("cli.width"))
  local_mocked_bindings(rstudio_detect = function() list(type = "rstudio_console"))
  withr::local_rng_version("3.5.0")
  set.seed(42)
  expect_snapshot(error = TRUE, local({
    len <- 26
    idx <- 100
    cli_abort(c(
            lorem_ipsum(1, 3),
      "i" = lorem_ipsum(1, 3),
      "x" = lorem_ipsum(1, 3)
    ))
  }))
})

test_that_cli(configs = "ansi", "color in RStudio", {
  local_mocked_bindings(
    rstudio_detect = function() list(type = "rstudio_console", num_colors = 256),
    get_rstudio_theme = function() list(foreground = "rgb(0, 0, 0)")
  )
  expect_snapshot({
    col <- get_rstudio_fg_color0()
    cat(col("this is the new color"))
  })

  local_mocked_bindings(get_rstudio_theme = function() list())
  expect_null(get_rstudio_fg_color0())

  local_mocked_bindings(
    rstudio_detect = function() list(type = "rstudio_console", num_colors = 1)
  )
  expect_null(get_rstudio_fg_color0())
})

test_that_cli(configs = "ansi", "update_rstudio_color", {
  local_mocked_bindings(
    get_rstudio_fg_color = function() make_ansi_style("#008800")
  )
  expect_snapshot(cat(update_rstudio_color("color me interested")))
})

test_that("cli_abort() captures correct call and backtrace", {
  skip_on_cran()
  rlang::local_options(
    rlang_trace_top_env = environment(),
    rlang_trace_format_srcrefs = FALSE
  )

  f <- function() g()
  g <- function() h()
  h <- function() cli::cli_abort("foo")

  expect_snapshot({
    print(expect_error(f()))
  }, variant = paste0("rlang-", packageVersion("rlang")))

  classed_stop <- function(message, env = parent.frame()) {
    cli::cli_abort(
      message,
      .envir = env,
      class = "cli_my_class"
    )
  }
  h <- function(x) {
    if (!length(x)) {
      classed_stop("{.arg x} can't be empty.")
    }
  }

  f <- function(x) g(x)
  g <- function(x) h(x)

  expect_snapshot({
    print(expect_error(f(list())))
  }, variant = paste0("rlang-", packageVersion("rlang")))
})

test_that("cli_abort(.internal = TRUE) reports the correct function (r-lib/rlang#1386)", {
  skip_on_cran()
  fn <- function() {
    cli::cli_abort("Message.", .internal = TRUE)
  }
  environment(fn) <- rlang::ns_env("base")

  # Should mention an internal error in the `base` package
  expect_snapshot({
    (expect_error(fn()))
  }, variant = paste0("rlang-", packageVersion("rlang")))
})

test_that_cli("cli_abort", {
  withr::local_options(cli.theme_dark = FALSE)
  expect_snapshot(
    error = TRUE,
    local({
      n <- "boo"
      cli_abort(c(
        "{.var n} must be a numeric vector",
        "x" = "You've supplied a {.cls {class(n)}} vector."
      ))
    })
  )

  expect_snapshot(
    error = TRUE,
    local({
      len <- 26
      idx <- 100
      cli_abort(c(
        "Must index an existing element:",
        "i" = "There {?is/are} {len} element{?s}.",
        "x" = "You've tried to subset element {idx}."
      ))
    })
  )

  n <- "boo"
  err <- tryCatch(
    cli_abort(c(
      "{.var n} must be a numeric vector",
      "x" = "You've supplied a {.cls {class(n)}} vector."
    )),
    error = function(e) e
  )
  expect_snapshot(c(err$message, err$body))
})

test_that_cli("cli_warn", {
  skip_if_not_installed("rlang", "1.0.0")

  withr::local_options(cli.theme_dark = FALSE)
  expect_snapshot({
    n <- "boo"
    cli_warn(c(
      "{.var n} must be a numeric vector",
      "x" = "You've supplied a {.cls {class(n)}} vector."
    ))
  })

  expect_snapshot(local({
    len <- 26
    idx <- 100
    cli_warn(c(
      "Must index an existing element:",
      "i" = "There {?is/are} {len} element{?s}.",
      "x" = "You've tried to subset element {idx}."
    ))
  }))
})

test_that_cli("cli_inform", {
  skip_if_not_installed("rlang", "1.0.0")

  withr::local_options(cli.ansi = FALSE)
  expect_snapshot({
    n <- "boo"
    cli_inform(c(
      "{.var n} must be a numeric vector",
      "x" = "You've supplied a {.cls {class(n)}} vector."
    ))
  })

  expect_snapshot(local({
    len <- 26
    idx <- 100
    cli_inform(c(
      "Must index an existing element:",
      "i" = "There {?is/are} {len} element{?s}.",
      "x" = "You've tried to subset element {idx}."
    ))
  }))
})

test_that("cli_abort width in RStudio", {
  # this is to fix breakage with new testthat
  withr::local_options(cli.condition_width = getOption("cli.width"))
  local_mocked_bindings(
    rstudio_detect = function() list(type = "rstudio_console")
  )
  withr::local_rng_version("3.5.0")
  set.seed(42)
  expect_snapshot(
    error = TRUE,
    local({
      len <- 26
      idx <- 100
      cli_abort(c(
        lorem_ipsum(1, 3),
        "i" = lorem_ipsum(1, 3),
        "x" = lorem_ipsum(1, 3)
      ))
    })
  )
})

test_that_cli(configs = "ansi", "color in RStudio", {
  local_mocked_bindings(
    rstudio_detect = function()
      list(type = "rstudio_console", num_colors = 256),
    get_rstudio_theme = function() list(foreground = "rgb(0, 0, 0)")
  )
  expect_snapshot({
    col <- get_rstudio_fg_color0()
    cat(col("this is the new color"))
  })

  local_mocked_bindings(get_rstudio_theme = function() list())
  expect_null(get_rstudio_fg_color0())

  local_mocked_bindings(
    rstudio_detect = function() list(type = "rstudio_console", num_colors = 1)
  )
  expect_null(get_rstudio_fg_color0())
})

test_that_cli(configs = "ansi", "update_rstudio_color", {
  local_mocked_bindings(
    get_rstudio_fg_color = function() make_ansi_style("#008800")
  )
  expect_snapshot(cat(update_rstudio_color("color me interested")))
})

#' Format a value for printing
#'
#' This function can be used directly, or via the `{.val ...}` inline
#' style. `{.val {expr}}` calls `cli_format()` automatically on the value
#' of `expr`, before styling and collapsing it.
#'
#' ## Default style
#'
#' ```{asciicast cli-format-default}
#' months <- month.name[1:3]
#' cli_text("{.val {months}}")
#' ```
#'
#' ```{asciicast cli-format-num}
#' nums <- 1:5 / 7
#' cli_text("{.val {nums}}")
#' ```
#'
#' ## Styling with themes
#'
#' ```{asciicast cli-format-theme}
#' nums <- 1:5 / 7
#' divid <- cli_div(theme = list(.val = list(digits = 3)))
#' cli_text("{.val {nums}}")
#' cli_end(divid)
#' ```
#'
#' It is possible to define new S3 methods for `cli_format` and then
#' these will be used automatically for `{.val ...}` expressions.
#'
#' ```{asciicast cli-format-class}
#' cli_format.month <- function(x, style = NULL, ...) {
#'   x <- encodeString(substr(x, 1, 3), quote = "\"")
#'   NextMethod("cli_format")
#' }
#' registerS3method("cli_format", "month", cli_format.month)
#' months <- structure(month.name[1:3], class = "month")
#' cli_text("{.val {months}}")
#' ```
#'
#' @param x The object to format.
#' @param style List of formatting options, see the individual methods
#'   for the style options they support.
#' @param ... Additional arguments for methods.
#'
#' @export
#' @seealso [cli_vec()]

cli_format <- function(x, style = NULL, ...) {
  if (is.null(style) && !is.null(default_app())) {
    style <- default_app()$get_current_style()
    cli_format(x, style, ...)
  } else {
    UseMethod("cli_format")
  }
}

#' @rdname cli_format
#' @export

cli_format.default <- function(x, style = NULL, ...) {
  x
}

#' * Styles for character vectors:
#'   - `string-quote` is the quoting character for [encodeString()].
#'
#' @rdname cli_format
#' @export

cli_format.character <- function(x, style = NULL, ...) {
  quote <- style$`string-quote` %||% style$string_quote %||% "\""
  encodeString(x, quote = quote)
}

#' * Styles for numeric vectors:
#'   - `digits` is the number of digits to print after the decimal point.
#'
#' @rdname cli_format
#' @export

cli_format.numeric <- function(x, style = NULL, ...) {
  digits <- style$digits
  if (!is.null(digits)) x <- round(x, digits)
  x
}

#' Add custom cli style to a vector
#'
#' @details
#' You can use this function to change the default parameters of
#' collapsing the vector into a string, see an example below.
#'
#' The style is added as an attribute, so operations that remove
#' attributes will remove the style as well.
#'
#' ## Custom collapsing separator
#'
#' ```{asciicast cli-vec}
#' v <- cli_vec(
#'   c("foo", "bar", "foobar"),
#'   style = list("vec-sep" = " & ", "vec-last" = " & ")
#' )
#' cli_text("My list: {v}.")
#' ```
#'
#' ## Custom truncation
#'
#' ```{asciicast cli-vec-2}
#' x <- cli_vec(names(mtcars), list("vec-trunc" = 3))
#' cli_text("Column names: {x}.")
#' ```
#'
#' @param x Vector that will be collapsed by cli.
#' @param style Style to apply to the vector. It is used as a theme on
#' a `span` element that is created for the vector. You can set `vec-sep`
#' and `vec-last` to modify the separator and the last separator.
#'
#' @export
#' @seealso [cli_format()]

cli_vec <- function(x, style = list()) {
  attr(x, "cli_style") <- style
  x
}

#' @export

ticking <- function(cond, name = NULL, ..., .envir = parent.frame()) {
  val <- force(cond)

  new <- is.null(clienv$progress_ids[[format(.envir)]])

  if (new && val) cli_progress_bar(name = name, ..., .envir = .envir)

  if (val) {
    cli_progress_update(.envir = .envir)
  } else {
    cli_progress_done(.envir = .envir)
  }

  val
}

library(css)

rstudio_theme_details_map <- list(
  "ambiance" = list(name = "Ambiance", isDark = TRUE),
  "chaos" = list(name = "Chaos", isDark = TRUE),
  "chrome" = list(name = "Chrome", isDark = FALSE),
  "clouds" = list(name = "Clouds", isDark = FALSE),
  "clouds_midnight" = list(name = "Clouds Midnight", isDark = TRUE),
  "cobalt" = list(name = "Cobalt", isDark = TRUE),
  "crimson_editor" = list(name = "Crimson Editor", isDark = FALSE),
  "dawn" = list(name = "Dawn", isDark = FALSE),
  "dracula" = list(name = "Dracula", isDark = TRUE),
  "dreamweaver" = list(name = "Dreamweaver", isDark = FALSE),
  "eclipse" = list(name = "Eclipse", isDark = FALSE),
  "idle_fingers" = list(name = "Idle Fingers", isDark = TRUE),
  "katzenmilch" = list(name = "Katzenmilch", isDark = FALSE),
  "kr_theme" = list(name = "Kr Theme", isDark = TRUE),
  "material" = list(name = "Material", isDark = TRUE),
  "merbivore" = list(name = "Merbivore", isDark = TRUE),
  "merbivore_soft" = list(name = "Merbivore Soft", isDark = TRUE),
  "mono_industrial" = list(name = "Mono Industrial", isDark = TRUE),
  "monokai" = list(name = "Monokai", isDark = TRUE),
  "pastel_on_dark" = list(name = "Pastel On Dark", isDark = TRUE),
  "solarized_dark" = list(name = "Solarized Dark", isDark = TRUE),
  "solarized_light" = list(name = "Solarized Light", isDark = FALSE),
  "textmate" = list(name = "Textmate (default)", isDark = FALSE),
  "tomorrow" = list(name = "Tomorrow", isDark = FALSE),
  "tomorrow_night" = list(name = "Tomorrow Night", isDark = TRUE),
  "tomorrow_night_blue" = list(name = "Tomorrow Night Blue", isDark = TRUE),
  "tomorrow_night_bright" = list(name = "Tomorrow Night Bright", isDark = TRUE),
  "tomorrow_night_eighties" = list(name = "Tomorrow Night 80s", isDark = TRUE),
  "twilight" = list(name = "Twilight", isDark = TRUE),
  "vibrant_ink" = list(name = "Vibrant Ink", isDark = TRUE),
  "xcode" = list(name = "Xcode", isDark = FALSE)
)

rstudio_theme_url_template <- paste0(
  "https://raw.githubusercontent.com/",
  "rstudio/rstudio/master/src/cpp/session/resources/themes/%s.rstheme"
)

## A set of operator colors to use, for each theme. Should match the name
## of the theme file in ace.
## We need to explicity set themes that should be overridden with the default
## vaue to NULL
operator_theme_map <- list(
   "solarized_light" = "#93A1A1",
   "solarized_dark" = "#B58900",
   "twilight" = "#7587A6",
   "idle_fingers" = "#6892B2",
   "clouds_midnight" = "#A53553",
   "cobalt" = "#BED6FF",
   "kr_theme" = "#A56464",
   "clouds" = NULL,
   "dawn" = NULL,
   "eclipse" = NULL,
   "katzenmilch" = NULL,
   "merbivore" = NULL,
   "merbivore_soft" = NULL,
   "monokai" = NULL,
   "pastel_on_dark" = NULL,
   "vibrant_ink" = NULL,
   "xcode" = NULL
)

## Similarly, colors for keywords that we might override.
keyword_theme_map <- list(
   "eclipse" = "#800080",
   "clouds" = "#800080"
)

# Needs https://github.com/romainfrancois/css

rstudio_css <- function(theme) {
  message("Downloading theme '", theme, "'")
  url <- sprintf(rstudio_theme_url_template, theme)
  css <- asNamespace("css")$read_css(url)

  g <- function(sel) {
    col <- vapply(
      sel,
      FUN.VALUE = "",
      function(sel1) {
        tail(c(NA_character_, css$value[grepl(sel1, css$rule) &
                                        css$property == "color"]), 1)
      }
    )

    col <- na.omit(col)[1]

    if (is.na(col) || length(col) != 1) {
      stop("Cannot get '", sel, "' from theme '", theme, "'")
    }

    col <- sub("[ ]+!important", "", col)

    ## Three digit colors are not handled by cli...
    if (grepl("^#[0-9a-fA-F][0-9a-fA-F][0-9a-fA-F]$", col)) {
      col <- paste(rep(strsplit(col, "", fixed = TRUE)[[1]], c(1, 2, 2, 2)), collapse = "")
    }
    ## rgb () form
    if (grepl("^rgb", col)) {
      col <- gsub("[^0-9 ]", "", col)
      col <- do.call(rgb, as.list(scan(text = col, quiet = TRUE) / 255))
    }
    unname(col)
  }

  if (theme %in% names(keyword_theme_map)) {
    kw <- keyword_theme_map[[theme]]
  } else {
    kw <- g("\\.ace_keyword$|\\.ace_keyword,")
  }

  if (theme %in% names(operator_theme_map)) {
    if (is.null(operator_theme_map[[theme]])) {
      if (rstudio_theme_details_map[[theme]]$isDark) {
        op <- "#aaaaaa"
      } else {
        op <- "#888888"
      }
    } else {
      op <- operator_theme_map[[theme]]
    }
  } else {
    op <- g(c(
      "^\\.ace_keyword\\.ace_operator$",
      "^\\.ace_constant\\.ace_language$",
      "^\\.ace_variable\\.ace_language$",
      "\\.ace_constant,",
      "\\.ace_constant\\.ace_buildin"
    ))
  }

  list(
    reserved_ = kw,
    number_ = g(c("\\.ace_constant\\.ace_numeric$", "\\.ace_constant,")),
    null_ = g(c("\\.ace_constant\\.ace_language$",
                "\\.ace_variable\\.ace_language$",
                "\\.ace_constant,",
                "\\.ace_constant\\.ace_buildin")),
    operator_ = op,
    call_ = NA_character_,
    string_ = g("\\.ace_string$|\\.ace_string,"),
    comment_ = g("\\.ace_comment$|\\.ace_comment,"),
    bracket_ = g(c("\\.ace_paren\\.ace_keyword\\.ace_operator",
                   "\\.ace_keyword\\.ace_operator",
                   "\\.ace_keyword"))
  )
}

create_rstudio_data <- function() {

  themes <- lapply(names(rstudio_theme_details_map), rstudio_css)
  names(themes) <- names(rstudio_theme_details_map)

  # Some substitutions
  for (nm in names(operator_theme_map)) {
    if (is.null(operator_theme_map[[nm]])) {

    } else {
      themes[[nm]]$null_ <- operator_theme_map[[nm]]
      themes[[nm]]$operator_ <- operator_theme_map[[nm]]
      themes[[nm]]$bracket_ <- operator_theme_map[[nm]]
    }
  }

  for (nm in names(keyword_theme_map)) {
    themes[[nm]]$reserved_ <- keyword_theme_map[[nm]]
  }

  # check that cli can handle these, otherwise it will error
  lapply(na.omit(unlist(themes)), cli::make_ansi_style)

  themes2 <- lapply(themes, function(theme) {
    list(
      reserved = theme$reserved_,
      number   = theme$number_,
      null     = theme$null_,
      operator = theme$operator_,
      call     = "bold",
      string   = theme$string_,
      comment  = theme$comment_,
      bracket  = list(theme$bracket_, "yellow", "blue", "cyan")
    )
  })

  names(themes2) <- vapply(rstudio_theme_details_map, "[[", "", "name")

  message("Saving sysdata.rda... ", appendLF = FALSE)
  sysenv <- new.env(parent = emptyenv())
  load(file.path("R", "sysdata.rda"), envir = sysenv)
  sysenv$rstudio_themes <- themes2
  save(
    list = ls(sysenv),
    envir = sysenv,
    file = file.path("R", "sysdata.rda"),
    version = 2
  )
  message("done.")

  invisible(themes2)
}

if (is.null(sys.calls())) {
  create_rstudio_data()
}

#' `cat()` helpers
#'
#' These helpers provide useful wrappers around [cat()]: most importantly
#' they all set `sep = ""`, and `cat_line()` automatically adds a newline.
#'
#' @export
#' @param ... For `cat_line()` and `cat_bullet()`, pasted together with
#'   `collapse = "\n"`. For `cat_rule()` and `cat_boxx()` passed on to
#'   [rule()] and [boxx()] respectively.
#' @param bullet Name of bullet character. Indexes into [symbol]
#' @param col,background_col,bullet_col Colors for text, background, and
#'   bullets respectively.
#' @param x An object to print.
#' @param file Output destination. Defaults to standard output.
#' @examples
#' cat_line("This is ", "a ", "line of text.", col = "red")
#' cat_bullet(letters[1:5])
#' cat_bullet(letters[1:5], bullet = "tick", bullet_col = "green")
#' cat_rule()
cat_line <- function(..., col = NULL, background_col = NULL, file = stdout()) {
  out <- paste0(..., collapse = "\n")
  out <- apply_style(out, col)
  out <- apply_style(out, background_col, bg = TRUE)

  cat(out, "\n", sep = "", file = file, append = TRUE)
}

#' @export
#' @rdname cat_line
cat_bullet <- function(..., col = NULL, background_col = NULL,
                       bullet = "bullet", bullet_col = NULL,
                       file = stdout()) {
  out <- apply_style(paste0(...), col)
  bullet <- apply_style(symbol[[bullet]], bullet_col)

  cat_line(paste(bullet, out), background_col = background_col, file = file)
}

#' @export
#' @rdname cat_line
cat_boxx <- function(..., file = stdout()) {
  cat_line(boxx(...), file = file)
}

#' @export
#' @rdname cat_line
cat_rule <- function(..., file = stdout()) {
  cat_line(rule(...), file = file)
}

#' @export
#' @rdname cat_line
cat_print <- function(x, file = "") {
  if (!identical(file, "")) {
    sink(file)
    on.exit(sink(NULL))
  }

  print(x)
}

test_that("cat_line", {
  expect_snapshot(
    cat_line("This is ", "a ", "line of text.")
  )

  tmp <- tempfile()
  on.exit(unlink(tmp), add = TRUE)
  cat_line("This is ", "a ", "line of text.", file = tmp)
  exp <- "This is a line of text."
  expect_equal(readLines(tmp, warn = FALSE), exp)

  local_reproducible_output(crayon = TRUE)

  expect_snapshot({
    cat_line("This is ", "a ", "line of text.", col = "red")
    cat_line("This is ", "a ", "line of text.", background_col = "green")
  })
})

test_that_cli(configs = c("plain", "unicode"), "cat_bullet", {
  expect_snapshot({
    cat_bullet(letters[1:5])
  })
})

test_that_cli(configs = c("plain", "unicode"), "cat_boxx", {
  expect_snapshot({
    cat_boxx("foo")
  })
})

test_that_cli(configs = c("plain", "unicode"), "cat_rule", {
  expect_snapshot(local({
    withr::local_options(cli.width = 20)
    cat_rule("title")
  }))
})

test_that_cli(configs = c("plain", "unicode"), "cat_print", {
  expect_snapshot({
    cat_print(boxx(""))
  })

  expect_snapshot(local({
    tmp <- tempfile()
    on.exit(unlink(tmp), add = TRUE)
    expect_silent(cat_print(boxx(""), file = tmp))
    cat(readLines(tmp, warn = FALSE), sep = "\n")
  }))
})

context("cat helpers")

test_that("cat_line", {
  out <- capt00(cat_line("This is ", "a ", "line of text."))
  exp <- rebox("This is a line of text.")
  expect_equal(out, exp)

  withr::with_options(
    list(cli.num_colors = 256L), {
      out <- capt00(cat_line("This is ", "a ", "line of text.", col = "red"))
      expect_true(ansi_has_any(out))
      exp <- rebox("This is a line of text.")
      expect_equal(ansi_strip(out), exp)

      out <- capt00(cat_line("This is ", "a ", "line of text.",
                            background_col = "green"))
      expect_true(ansi_has_any(out))
      exp <- rebox("This is a line of text.")
      expect_equal(ansi_strip(out), exp)
    }
  )

  tmp <- tempfile()
  on.exit(unlink(tmp), add = TRUE)
  cat_line("This is ", "a ", "line of text.", file = tmp)
  exp <- rebox("This is a line of text.")
  expect_equal(readLines(tmp, warn = FALSE), exp)
})

test_that("cat_bullet", {
  out <- capt00(cat_bullet(letters[1:5]))
  exp <- rebox("● a\n● b\n● c\n● d\n● e")
  expect_equal(out, exp)
})

test_that("cat_boxx", {
  out <- capt00(cat_boxx("foo"))
  exp <- rebox(
    "┌─────────┐",
    "│         │",
    "│   foo   │",
    "│         │",
    "└─────────┘")
  expect_equal(out, exp)
})

test_that("cat_rule", {
  out <- withr::with_options(
    c(cli.width = 20),
    capt00(cat_rule("title"))
  )
  exp <- rebox("── title ───────────")
  expect_equal(out, exp)
})

test_that("cat_print", {
  out <- capt00(cat_print(boxx("")))
  exp <- rebox(
    "┌──────┐",
    "│      │",
    "│      │",
    "│      │",
    "└──────┘")
  expect_equal(out, exp)

  tmp <- tempfile()
  on.exit(unlink(tmp), add = TRUE)
  expect_silent(cat_print(boxx(""), file = tmp))
  expect_equal(paste(readLines(tmp, warn = FALSE), collapse = "\n"), exp)
})

format_bytes <- local({

  pretty_bytes <- function(bytes, style = c("default", "nopad", "6")) {

    style <- switch(
      match.arg(style),
      "default" = pretty_bytes_default,
      "nopad" = pretty_bytes_nopad,
      "6" = pretty_bytes_6
    )

    style(bytes)
  }

  compute_bytes <- function(bytes, smallest_unit = "B") {
    units0 <- c("B", "kB", "MB", "GB", "TB", "PB", "EB", "ZB", "YB")

    stopifnot(
      is.numeric(bytes),
      is.character(smallest_unit),
      length(smallest_unit) == 1,
      !is.na(smallest_unit),
      smallest_unit %in% units0
    )

    limits <- c(1000, 999950 * 1000 ^ (seq_len(length(units0) - 2) - 1))
    low <- match(smallest_unit, units0)
    units <- units0[low:length(units0)]
    limits <- limits[low:length(limits)]

    neg <- bytes < 0 & !is.na(bytes)
    bytes <- abs(bytes)

    mat <- matrix(
      rep(bytes, each = length(limits)),
      nrow = length(limits),
      ncol = length(bytes)
    )
    mat2 <- matrix(mat < limits, nrow  = length(limits), ncol = length(bytes))
    exponent <- length(limits) - colSums(mat2) + low - 1L
    res <- bytes / 1000 ^ exponent
    unit <- units[exponent - low + 2L]

    ## Zero bytes
    res[bytes == 0] <- 0
    unit[bytes == 0] <- units[1]

    ## NA and NaN bytes
    res[is.na(bytes)] <- NA_real_
    res[is.nan(bytes)] <- NaN
    unit[is.na(bytes)] <- units0[low]     # Includes NaN as well

    data.frame(
      stringsAsFactors = FALSE,
      amount = res,
      unit = unit,
      negative = neg
    )
  }

  pretty_bytes_default <- function(bytes) {
    szs <- compute_bytes(bytes)
    amt <- szs$amount

    ## String. For fractions we always show two fraction digits
    res <- character(length(amt))
    int <- is.na(amt) | amt == as.integer(amt)
    res[int] <- format(
      ifelse(szs$negative[int], -1, 1) * amt[int],
      scientific = FALSE
    )
    res[!int] <- sprintf("%.2f", ifelse(szs$negative[!int], -1, 1) * amt[!int])

    format(paste(res, szs$unit), justify = "right")
  }

  pretty_bytes_nopad <- function(bytes) {
    sub("^\\s+", "", pretty_bytes_default(bytes))
  }

  pretty_bytes_6 <- function(bytes) {
    szs <- compute_bytes(bytes, smallest_unit = "kB")
    amt <- szs$amount

    na   <- is.na(amt)
    nan  <- is.nan(amt)
    neg  <- !na & !nan & szs$negative
    l10  <- !na & !nan & !neg & amt < 10
    l100 <- !na & !nan & !neg & amt >= 10 & amt < 100
    b100 <- !na & !nan & !neg & amt >= 100

    szs$unit[neg] <- "kB"

    famt <- character(length(amt))
    famt[na] <- " NA"
    famt[nan] <- "NaN"
    famt[neg] <- "< 0"
    famt[l10] <- sprintf("%.1f", amt[l10])
    famt[l100] <- sprintf(" %.0f", amt[l100])
    famt[b100] <- sprintf("%.0f", amt[b100])

    paste0(famt, " ", szs$unit)
  }

  structure(
    list(
      .internal     = environment(),
      pretty_bytes  = pretty_bytes,
      compute_bytes = compute_bytes
    ),
    class = c("standalone_bytes", "standalone")
  )
})

#' cli progress bars
#'
#' @description
#' This is the reference manual of the three functions that create,
#' update and terminate progress bars. For a tutorial see the
#' [cli progress bars](https://cli.r-lib.org/articles/progress.html).
#'
#' `cli_progress_bar()` creates a new progress bar.
#'
#' @details
#'
#' ## Basic usage
#'
#' `cli_progress_bar()` creates a progress bar, `cli_progress_update()`
#' updates an existing progress bar, and `cli_progress_done()` terminates
#' it.
#'
#' It is good practice to always set the `name` argument, to make the
#' progress bar more informative.
#'
#' ```{asciicast progress-1}
#' #| asciicast_at = "all",
#' #| asciicast_knitr_output = "svg",
#' #| asciicast_cursor = FALSE
#' clean <- function() {
#'   cli_progress_bar("Cleaning data", total = 100)
#'   for (i in 1:100) {
#'     Sys.sleep(5/100)
#'     cli_progress_update()
#'   }
#'   cli_progress_done()
#' }
#' clean()
#' ```
#'
#' ## Progress bar types
#'
#' There are three builtin types of progress bars, and a custom type.
#'
#' ```{asciicast progress-tasks}
#' #| asciicast_at = "all",
#' #| asciicast_knitr_output = "svg",
#' #| asciicast_cursor = FALSE
#' tasks <- function() {
#'   cli_progress_bar("Tasks", total = 3, type = "tasks")
#'   for (i in 1:3) {
#'     Sys.sleep(1)
#'     cli_progress_update()
#'   }
#'   cli_progress_done()
#' }
#' tasks()
#' ```
#'
#' ## Unknown `total`
#'
#' If `total` is not known, then cli shows a different progress bar.
#' Note that you can also set `total` in `cli_progress_update()`, if it
#' not known when the progress bar is created, but you learn it later.
#'
#' ```{asciicast progress-natotal}
#' #| asciicast_at = "all",
#' #| asciicast_knitr_output = "svg",
#' #| asciicast_cursor = FALSE
#' nototal <- function() {
#'   cli_progress_bar("Parameter tuning")
#'   for (i in 1:100) {
#'     Sys.sleep(3/100)
#'     cli_progress_update()
#'   }
#'   cli_progress_done()
#' }
#' nototal()
#' ```
#'
#' ## Clearing the progress bar
#'
#' By default cli removes terminated progress bars from the screen, if
#' the terminal supports this. If you want to change this, use the
#' `clear` argument of `cli_progress_bar()`, or the `cli.progress_clear`
#' global option (see [cli-config]) to change this.
#'
#' (In the cli documentation we usually set `cli.progress_clear` to `FALSE`,
#' so users can see how finished progress bars look.)
#'
#' In this example the first progress bar is cleared, the second is not.
#'
#' ```{asciicast progress-clear}
#' #| asciicast_at = "all",
#' #| asciicast_knitr_output = "svg",
#' #| asciicast_cursor = FALSE
#' fun <- function() {
#'   cli_progress_bar("Data cleaning", total = 100, clear = TRUE)
#'   for (i in 1:100) {
#'     Sys.sleep(3/100)
#'     cli_progress_update()
#'   }
#'   cli_progress_bar("Parameter tuning", total = 100, clear = FALSE)
#'   for (i in 1:100) {
#'     Sys.sleep(3/100)
#'     cli_progress_update()
#'   }
#' }
#' fun()
#' ```
#'
#' ## Initial delay
#'
#' Updating a progress bar on the screen is costly, so cli tries to avoid
#' it for quick loops. By default a progress bar is only shown after two
#' seconds, or after half of that if less than 50% of the iterations are
#' complete. You can change the two second default with the
#' `cli.progress_show_after` global option (see [cli-config]).
#'
#' (In the cli documentation we usually set `cli.progress_show_after` to
#' `0` (zero seconds), so progress bars are shown immediately.)
#'
#' In this example we only show the progress bar after one second, because
#' more than 50% of the iterations remain after one second.
#'
#' ```{asciicast progress-after}
#' #| asciicast_at = "all",
#' #| asciicast_knitr_output = "svg",
#' #| asciicast_cursor = FALSE
#' fun <- function() {
#'   cli_alert("Starting now, at {Sys.time()}")
#'   cli_progress_bar(
#'     total = 100,
#'     format = "{cli::pb_bar} {pb_percent} @ {Sys.time()}"
#'   )
#'   for (i in 1:100) {
#'     Sys.sleep(4/100)
#'     cli_progress_update()
#'   }
#' }
#' options(cli.progress_show_after = 2)
#' fun()
#' ```
#'
#' ```{asciicast, include = FALSE, cache = FALSE}
#' # reset to our default
#' options(cli.progress_show_after = 0)
#' ```
#'
#' ## The _current_ progress bar
#'
#' By default cli sets the new progress bar as the _current_ progress bar
#' of the calling function. The current progress bar is the default one
#' in cli progress bar operations. E.g. if no progress bar id is supplied
#' in `cli_progress_update()`, then the current progress bar is updated.
#'
#' Every function can only have a single _current_ progress bar, and if a
#' new one is created, then the previous one (if any) is automatically
#' terminated. The current progress bar is also terminated when the function
#' that created it exits. Thanks to these rules, most often you don't need
#' to explicitly deal with progress bar ids, and you don't need to
#' explicitly call `cli_progress_done()`:
#'
#' ```{asciicast progress-current}
#' #| asciicast_at = "all",
#' #| asciicast_knitr_output = "svg",
#' #| asciicast_cursor = FALSE
#' fun <- function() {
#'   cli_progress_bar("First step ", total = 100)
#'   for (i in 1:100) {
#'     Sys.sleep(2/100)
#'     cli_progress_update()
#'   }
#'   cli_progress_bar("Second step", total = 100)
#'   for (i in 1:100) {
#'     Sys.sleep(2/100)
#'     cli_progress_update()
#'   }
#' }
#' fun()
#' ```
#'
#' ## cli output while the progress bar is active
#'
#' cli allows emitting regular cli output (alerts, headers, lists, etc.)
#' while a progress bar is active. On terminals that support this, cli
#' will remove the progress bar temporarily, emit the output, and then
#' restores the progress bar.
#'
#' ```{asciicast progress-output}
#' #| asciicast_at = "all",
#' #| asciicast_knitr_output = "svg",
#' #| asciicast_cursor = FALSE
#' fun <- function() {
#'   cli_alert_info("Before the progress bar")
#'   cli_progress_bar("Calculating", total = 100)
#'   for (i in 1:50) {
#'     Sys.sleep(4/100)
#'     cli_progress_update()
#'   }
#'   cli_alert_info("Already half way!")
#'   for (i in 1:50) {
#'     Sys.sleep(4/100)
#'     cli_progress_update()
#'   }
#'   cli_alert_info("All done")
#' }
#' fun()
#' ```
#'
#' See also [cli_progress_output()], which sends text for the current
#' progress handler. E.g. in a Shiny app it will send the output to the
#' Shiny progress bar, as opposed to the `cli_alert()` etc. cli functions
#' which will print the text to the console.
#'
#' ## Custom formats
#'
#' In addition to the builtin types, you can also specify a custom
#' format string. In this case [progress variables][progress-variables]
#' are probably useful to avoid calculating some progress bar quantities
#' like the elapsed time, of the ETA manually. You can also use your own
#' variables in the calling function:
#'
#' ```{asciicast progress-format}
#' #| asciicast_at = "all",
#' #| asciicast_knitr_output = "svg",
#' #| asciicast_cursor = FALSE
#' fun <- function(urls) {
#'   cli_progress_bar(
#'     format = paste0(
#'       "{pb_spin} Downloading {.path {basename(url)}} ",
#'       "[{pb_current}/{pb_total}]   ETA:{pb_eta}"
#'     ),
#'     format_done = paste0(
#'       "{col_green(symbol$tick)} Downloaded {pb_total} files ",
#'       "in {pb_elapsed}."
#'     ),,
#'     total = length(urls)
#'   )
#'   for (url in urls) {
#'     cli_progress_update()
#'     Sys.sleep(5/10)
#'   }
#' }
#' fun(paste0("https://acme.com/data-", 1:10, ".zip"))
#' ```
#'
#' @param name This is typically used as a label, and should be short,
#'   at most 20 characters.
#' @param status Initial status of the progress bar. If not empty, then
#'   it is typically shown after the label.
#' @param type Type of the progress bar. It is used to select a default
#'   display if `format` is not specified. Currently supported types:
#'   * `iterator`: e.g. a for loop or a mapping function,
#'   * `tasks`: a (typically small) number of tasks,
#'   * `download`: download of one file,
#'   * `custom`: custom type, `format` must not be `NULL` for this type.
#' @param total Total number of progress units, or `NA` if it is unknown.
#'   `cli_progress_update()` can update the total number of units. This is
#'   handy if you don't know the size of a download at the beginning, and
#'   also in some other cases. If `format` is set to `NULL`, `format` (plus
#'   `format_done` and `format_failed`) will be updated when you change
#'   `total` from `NA` to a number. I.e. default format strings will be
#'   updated, custom ones won't be.
#' @param format Format string. It has to be specified for custom progress
#'   bars, otherwise it is optional, and a default display is selected
#'   based on the progress bat type and whether the number of total units
#'   is known. Format strings may contain glue substitution, the support
#'   pluralization and cli styling. See [progress-variables] for special
#'   variables that you can use in the custom format.
#' @param format_done Format string for successful termination. By default
#'   the same as `format`.
#' @param format_failed Format string for unsuccessful termination. By
#'   default the same as `format`.
#' @param clear Whether to remove the progress bar from the screen after
#'   it has terminated. Defaults to the `cli.progress_clear` option, or
#'   `TRUE` if unset.
#' @param current Whether to use this progress bar as the current progress
#'   bar of the calling function. See more at 'The current progress bar'
#'   below.
#' @param auto_terminate Whether to terminate the progress bar if the
#'   number of current units reaches the number of total units.
#' @param extra Extra data to add to the progress bar. This can be
#'   used in custom format strings for example. It should be a named list.
#'   `cli_progress_update()` can update the extra data. Often you can get
#'   away with referring to local variables in the format string, and
#'   then you don't need to use this argument. Explicitly including these
#'   constants or variables in `extra` can result in cleaner code. In
#'   the rare cases when you need to refer to the same progress bar from
#'   multiple functions, and you can them to `extra`.
#' @param .auto_close Whether to terminate the progress bar when the
#'   calling function (or the one with execution environment in `.envir`
#'   exits. (Auto termination does not work for progress bars created
#'   from the global environment, e.g. from a script.)
#' @param .envir The environment to use for auto-termination and for glue
#'   substitution. It is also used to find and set the current progress bar.
#'
#' @return `cli_progress_bar()` returns the id of the new progress bar.
#' The id is a string constant.
#'
#' @seealso These functions support [inline markup][inline-markup].
#' @seealso [cli_progress_message()] and [cli_progress_step()] for simpler
#'   progress messages.
#' @family progress bar functions
#' @family functions supporting inline markup
#' @aliases __cli_update_due cli_tick_reset ccli_tick_reset ticking
#' @export

cli_progress_bar <- function(name = NULL,
                             status = NULL,
                             type = c("iterator", "tasks", "download",
                                      "custom"),
                             total = NA,
                             format = NULL,
                             format_done = NULL,
                             format_failed = NULL,
                             clear = getOption("cli.progress_clear", TRUE),
                             current = TRUE,
                             auto_terminate = type != "download",
                             extra = NULL,
                             .auto_close = TRUE,
                             .envir = parent.frame()) {

  start <- .Call(clic_get_time)
  id <- new_uuid()
  envkey <- format(.envir)
  type <- match.arg(type)
  if (type == "custom" && is.null(format)) {
    stop("Need to specify format if `type == \"custom\"")
  }

  ## If `total` is infinite, use behavior seen when `total` is NA
  if (is.infinite(total)) {
    total <- NA
  }

  ## If changes, synchronize with C API in progress.c
  bar <- new.env(parent = emptyenv())
  bar$id <- id
  bar$name <- name
  bar$status <- status
  bar$type <- match.arg(type)
  bar$total <- total
  bar$show_after <- start + getOption("cli.progress_show_after", 2)
  bar$show_50 <- start + getOption("cli.progress_show_after", 2) / 2
  bar$format_orig <- bar$format <- format
  bar$format_done_orig <- bar$format_done <- format_done %||% format
  bar$format_failed_orig <- bar$format_failed <- format_failed %||% format
  bar$clear <- clear
  bar$auto_terminate <- auto_terminate
  bar$envkey <- if (current) envkey else NULL
  bar$current <- 0L
  bar$start <- start
  bar$tick <- 0L
  bar$extra <- extra
  clienv$progress[[id]] <- bar
  if (current) {
    if (!is.null(clienv$progress_ids[[envkey]])) {
      cli_progress_done(clienv$progress_ids[[envkey]], .envir = .envir, result = "done")
    }
    clienv$progress_ids[[envkey]] <- id
  }

  if (.auto_close && envkey != clienv$globalenv) {
    defer(
      cli_progress_done(id = id, .envir = .envir, result = "auto"),
      envir = .envir
    )
  }

  opt <- options(cli__pb = bar)
  on.exit(options(opt), add = TRUE)

  bar$handlers <- cli_progress_select_handlers(bar, .envir)
  for (h in bar$handlers) {
    if ("create" %in% names(h)) h$create(bar, .envir = .envir)
  }

  invisible(id)
}

#' @description
#' `cli_progress_update()` updates the state of a progress bar, and
#' potentially the display as well.
#'
#' @param inc Increment in progress units. This is ignored if `set` is
#'   not `NULL`.
#' @param set Set the current number of progress units to this value.
#'   Ignored if `NULL`.
#' @param status New status string of the progress bar, if not `NULL`.
#' @param id Progress bar to update or terminate. If `NULL`, then the
#'   current progress bar of the calling function (or `.envir` if
#'   specified) is updated or terminated.
#' @param force Whether to force a display update, even if no update is
#'   due.
#'
#' @return `cli_progress_update()` returns the id of the progress bar,
#' invisibly.
#'
#' @name cli_progress_bar
#' @export

cli_progress_update <- function(inc = NULL, set = NULL, total = NULL,
                                status = NULL, extra = NULL,
                                id = NULL, force = FALSE,
                                .envir = parent.frame()) {

  id <- id %||% clienv$progress_ids[[format(.envir)]]
  if (is.null(id)) {
    envkey <- format(.envir)
    stop("Cannot find current progress bar for `", envkey, "`")
  }
  pb <- clienv$progress[[id]]
  if (is.null(pb)) stop("Cannot find progress bar `", id, "`")

  if (!is.null(status)) pb$status <- status

  if (!is.null(extra)) pb$extra <- utils::modifyList(pb$extra, extra)

  if (!is.null(set)) {
    pb$current <- set
  } else {
    inc <- inc %||% 1L
    pb$current <- pb$current + inc
  }

  if (!is.null(total)) {
    if (is.na(pb$total) != is.na(total) ||
        (!is.na(total) && pb$total != total)) {
      pb$total <- total
      if (!is.null(pb$format) && is.null(pb$format_orig)) {
        pb$format <- pb__default_format(pb$type, pb$total)
        pb$format_done <- pb$format_done_orig %||% pb$format
        pb$format_failed <- pb$format_failed_orig %||% pb$format
      }
    }
  }

  if (pb$auto_terminate && !is.na(pb$total) && pb$current == pb$total) {
    cli_progress_done(id, .envir = .envir, result = "done")
    return(invisible(id))
  }

  now <- .Call(clic_get_time)
  upd <- .Call(clic_update_due)
  if (force || (upd && now > pb$show_after) ||
      (!is.na(pb$total) && upd && now > pb$show_50 && pb$current <= pb$total / 2)) {
    if (upd) cli_tick_reset()
    pb$tick <- pb$tick + 1L

    if (is.null(pb$format)) {
      pb$format <- pb__default_format(pb$type, pb$total)
      pb$format_done <- pb$format_done_orig %||% pb$format
      pb$format_failed <- pb$format_failed_orig %||% pb$format
    }

    opt <- options(cli__pb = pb)
    on.exit(options(opt), add = TRUE)

    if (is.null(pb$added)) {
      pb$added <- TRUE
      for (h in pb$handlers) {
        if ("add" %in% names(h)) h$add(pb, .envir = .envir)
      }
    }

    for (h in pb$handlers) {
      if ("set" %in% names(h)) h$set(pb, .envir = .envir)
    }
  }

  # Return TRUE, to allow cli_progress_update() && break in loops
  invisible(id)
}

#' @description
#' `cli_progress_done()` terminates a progress bar.
#'
#' @param result String to select successful or unsuccessful termination.
#'   It is only used if the progress bar is not cleared from the screen.
#'   It can be one of `"done"`, `"failed"`, `"clear"`, and `"auto"`.
#'
#' @return `cli_progress_done()` returns `TRUE`, invisibly, always.
#'
#' @name cli_progress_bar
#' @export

cli_progress_done <- function(id = NULL, .envir = parent.frame(),
                              result = "done") {
  envkey <- format(.envir)
  id <- id %||% clienv$progress_ids[[envkey]]
  if (is.null(id)) return(invisible(TRUE))
  pb <- clienv$progress[[id]]
  if (is.null(pb)) return(invisible(TRUE))

  opt <- options(cli__pb = pb)
  on.exit(options(opt), add = TRUE)

  if (result == "auto") {
    r1 <- random_marker
    if (identical(returnValue(r1), r1)) {
      result <- "failed"
    } else {
      result <- "done"
    }
  }

  for (h in pb$handlers) {
    if ("complete" %in% names(h)) {
      h$complete(pb, .envir = .envir, result = result)
    }
  }

  clienv$progress[[id]] <- NULL
  if (!is.null(pb$envkey)) clienv$progress_ids[[pb$envkey]] <- NULL

  invisible(TRUE)
}

#' Add text output to a progress bar
#'
#' The text is calculated via [cli_text()], so all cli features can be
#' used here, including progress variables.
#'
#' The text is passed to the progress handler(s), that may or may not be
#' able to print it.
#'
#' ```{asciicast progress-output2}
#' #| asciicast_at = "all",
#' #| asciicast_knitr_output = "svg",
#' #| asciicast_cursor = FALSE
#' fun <- function() {
#'   cli_alert_info("Before the progress bar")
#'   cli_progress_bar("Calculating", total = 100)
#'   for (i in 1:50) {
#'     Sys.sleep(4/100)
#'     cli_progress_update()
#'   }
#'   cli_progress_output("Already half way!")
#'   for (i in 1:50) {
#'     Sys.sleep(4/100)
#'     cli_progress_update()
#'   }
#'   cli_alert_info("All done")
#' }
#' fun()
#' ```
#'
#' @param text Text to output. It is formatted via [cli_text()].
#' @param id Progress bar id. The default is the current progress bar.
#' @param .envir Environment to use for glue interpolation of `text`.
#' @return `TRUE`, always.
#'
#' @seealso This function supports [inline markup][inline-markup].
#' @family progress bar functions
#' @family functions supporting inline markup
#' @export

cli_progress_output <- function(text, id = NULL, .envir = parent.frame()) {
  envkey <- format(.envir)
  id <- id %||% clienv$progress_ids[[envkey]]
  if (is.null(id)) {
    stop("Cannot find current progress bar for `", envkey, "`")
  }
  pb <- clienv$progress[[id]]
  if (is.null(pb)) stop("Cannot find progress bar `", id, "`")

  txt <- cli_fmt(cli_text(text, .envir = .envir))
  for (h in pb$handlers) {
    if ("output" %in% names(h)) {
      h$output(pb, .envir = .envir, text = txt)
    }
  }

  invisible(TRUE)
}

# ------------------------------------------------------------------------

#' Simplified cli progress messages
#'
#' @description This is a simplified progress bar, a single (dynamic)
#' message, without progress units.
#'
#' @details `cli_progress_message()` always shows the message, even if no
#' update is due. When the progress message is terminated, it is removed
#' from the screen by default.
#'
#' Note that the message can be dynamic: if you update it with
#' [cli_progress_update()], then cli uses the current values in the string
#' substitutions.
#'
#' ```{asciicast progress-message}
#' #| echo = c(-2, -3),
#' #| asciicast_at = "all",
#' #| asciicast_knitr_output = "svg",
#' #| asciicast_cursor = FALSE
#' fun <- function() {
#'   opts <- options(cli.progress_clear = TRUE)
#'   on.exit(options(opts), add = TRUE)
#'   cli_progress_message("Task one is running...")
#'   Sys.sleep(2)
#'
#'   cli_progress_message("Task two is running...")
#'   Sys.sleep(2)
#'
#'   step <- 1L
#'   cli_progress_message("Task three is underway: step {step}")
#'   for (step in 1:5) {
#'     Sys.sleep(0.5)
#'     cli_progress_update()
#'   }
#' }
#' fun()
#' ```
#'
#' @param msg Message to show. It may contain glue substitution and cli
#'   styling. It can be updated via [cli_progress_update()], as usual.
#' @param current Passed to [cli_progress_bar()].
#' @param .auto_close Passed to [cli_progress_bar()].
#' @param .envir Passed to [cli_progress_bar()].
#' @param ... Passed to [cli_progress_bar()].
#'
#' @return The id of the new progress bar.
#'
#' @seealso This function supports [inline markup][inline-markup].
#' @seealso [cli_progress_bar()] for the complete progress bar API.
#'   [cli_progress_step()] for a similar display that is styled by default.
#' @family progress bar functions
#' @family functions supporting inline markup
#' @export

cli_progress_message <- function(msg,
                                 current = TRUE,
                                 .auto_close = TRUE,
                                 .envir = parent.frame(),
                                 ...) {

  id <- cli_progress_bar(
    type = "custom",
    format = msg,
    current = current,
    .auto_close = .auto_close,
    .envir = .envir,
    ...
  )

  cli_progress_update(id = id, force = TRUE, .envir = .envir)

  invisible(id)
}

# ------------------------------------------------------------------------

#' Simplified cli progress messages, with styling
#'
#' @description This is a simplified progress bar, a single (dynamic)
#' message, without progress units.
#'
#' @details `cli_progress_step()` always shows the progress message,
#' even if no update is due.
#'
#' ## Basic use
#'
#' ```{asciicast progress-step}
#' #| asciicast_at = "all",
#' #| asciicast_knitr_output = "svg",
#' #| asciicast_cursor = FALSE
#' f <- function() {
#'   cli_progress_step("Downloading data")
#'   Sys.sleep(2)
#'   cli_progress_step("Importing data")
#'   Sys.sleep(1)
#'   cli_progress_step("Cleaning data")
#'   Sys.sleep(2)
#'   cli_progress_step("Fitting model")
#'   Sys.sleep(3)
#' }
#' f()
#' ```
#'
#' ## Spinner
#'
#' You can add a spinner to some or all steps with `spinner = TRUE`,
#' but note that this will only work if you call [cli_progress_update()]
#' regularly.
#'
#' ```{asciicast progress-step-spin}
#' #| asciicast_at = "all",
#' #| asciicast_knitr_output = "svg",
#' #| asciicast_cursor = FALSE
#' f <- function() {
#'   cli_progress_step("Downloading data", spinner = TRUE)
#'   for (i in 1:100) { Sys.sleep(2/100); cli_progress_update() }
#'   cli_progress_step("Importing data")
#'   Sys.sleep(1)
#'   cli_progress_step("Cleaning data")
#'   Sys.sleep(2)
#'   cli_progress_step("Fitting model", spinner = TRUE)
#'   for (i in 1:100) { Sys.sleep(3/100); cli_progress_update() }
#' }
#' f()
#' ```
#'
#' ## Dynamic messages
#'
#' You can make the step messages dynamic, using glue templates.
#' Since `cli_progress_step()` show that message immediately, we need
#' to initialize `msg` first.
#'
#' ```{asciicast progress-step-dynamic}
#' #| asciicast_at = "all",
#' #| asciicast_knitr_output = "svg",
#' #| asciicast_cursor = FALSE
#' f <- function() {
#'   msg <- ""
#'   cli_progress_step("Downloading data{msg}", spinner = TRUE)
#'   for (i in 1:100) {
#'     Sys.sleep(2/100)
#'     msg <- glue::glue(", got file {i}/100")
#'     cli_progress_update()
#'   }
#'   cli_progress_step("Importing data")
#'   Sys.sleep(1)
#'   cli_progress_step("Cleaning data")
#'   Sys.sleep(2)
#'   cli_progress_step("Fitting model", spinner = TRUE)
#'   for (i in 1:100) { Sys.sleep(3/100); cli_progress_update() }
#' }
#' f()
#' ```
#'
#' ## Termination messages
#'
#' You can specify a different message for successful and/or
#' unsuccessful termination:
#'
#' ```{asciicast progress-step-msg}
#' #| error = FALSE,
#' #| asciicast_at = "all",
#' #| asciicast_knitr_output = "svg",
#' #| asciicast_cursor = FALSE
#' f <- function() {
#'   size <- 0L
#'   cli_progress_step(
#'     "Downloading data.",
#'     msg_done = "Downloaded {prettyunits::pretty_bytes(size)}.",
#'     spinner = TRUE
#'   )
#'   for (i in 1:100) {
#'     Sys.sleep(3/100)
#'     size <- size + 8192
#'     cli_progress_update()
#'   }
#' }
#' f()
#' ```
#'
#' @param msg Message to show. It may contain glue substitution and cli
#'   styling. It can be updated via [cli_progress_update()], as usual.
#'   It is style as a cli info alert (see [cli_alert_info()]).
#' @param msg_done Message to show on successful termination. By default
#'   this it is the same as `msg` and it is styled as a cli success alert
#'   (see [cli_alert_success()]).
#' @param msg_failed Message to show on unsuccessful termination. By
#'   default it is the same as `msg` and it is styled as a cli danger alert
#'   (see [cli_alert_danger()]).
#' @param spinner Whether to show a spinner at the beginning of the line.
#'   To make the spinner spin, you'll need to call `cli_progress_update()`
#'   regularly.
#' @param class cli class to add to the message. By default there is no
#'   class for steps with a spinner.
#' @param current Passed to [cli_progress_bar()].
#' @param .auto_close Passed to [cli_progress_bar()].
#' @param .envir Passed to [cli_progress_bar()].
#' @param ... Passed to [cli_progress_bar()].
#'
#' @seealso This function supports [inline markup][inline-markup].
#' @family progress bar functions
#' @family functions supporting inline markup
#' @export

cli_progress_step <- function(msg,
                              msg_done = msg,
                              msg_failed = msg,
                              spinner = FALSE,
                              class = if (!spinner) ".alert-info",
                              current = TRUE,
                              .auto_close = TRUE,
                              .envir = parent.frame(),
                              ...) {

  format <- paste0(
    if (!is.null(class)) paste0("{", class, " "),
    if (spinner) "{cli::pb_spin} ",
    msg,
    if (!is.null(class)) "}"
  )
  ts <- " {.timestamp {cli::pb_elapsed}}"
  format_done <- paste0("{.alert-success ", msg_done, ts, "}")
  format_failed <- paste0("{.alert-danger ", msg_failed, ts, "}")

  opt <- options(cli.progress_show_after = 0)
  on.exit(options(opt), add = TRUE)
  id <- cli_progress_bar(
    type = "custom",
    format = format,
    format_done = format_done,
    format_failed = format_failed,
    clear = FALSE,
    current = current,
    .auto_close = .auto_close,
    .envir = .envir,
    ...
  )

  cli_progress_update(id = id, force = TRUE, .envir = .envir)

  invisible(id)
}

# ------------------------------------------------------------------------

pb__default_format <- function(type, total) {
  if (type == "iterator") {
    if (!is.na(total)) {
      opt <- getOption("cli.progress_format_iterator")
      if (!is.null(opt)) return(opt)
      paste0(
        "{cli::pb_name}{cli::pb_bar} {cli::pb_percent} | {cli::pb_status}",
        "ETA: {cli::pb_eta}"
      )
    } else {
      opt <- getOption("cli.progress_format_iterator_nototal") %||%
        getOption("cli.progress_format_iterator")
      if (!is.null(opt)) return(opt)
      paste0(
        "{cli::pb_spin} {cli::pb_name}{cli::pb_status}",
        "{cli::pb_current} done ({cli::pb_rate}) | {cli::pb_elapsed}"
      )
    }

  } else if (type == "tasks") {
    if (!is.na(total)) {
      opt <- getOption("cli.progress_format_tasks")
      if (!is.null(opt)) return(opt)
      paste0(
        "{cli::pb_spin} {cli::pb_current}/{cli::pb_total} ",
        "ETA: {cli::pb_eta} | {cli::pb_name}{cli::pb_status}"
      )
    } else {
      opt <- getOption("cli.progress_format_tasks_nototal") %||%
        getOption("cli.progress_format_tasks")
      if (!is.null(opt)) return(opt)
      paste0(
        "{cli::pb_spin} {cli::pb_name}{cli::pb_status}",
        "{cli::pb_current} done ({cli::pb_rate}) | {cli::pb_elapsed}"
      )
    }

  } else if (type == "download") {
    if (!is.na(total)) {
      opt <- getOption("cli.progress_format_download")
      if (!is.null(opt)) return(opt)
      paste0(
        "{cli::pb_name}{cli::pb_status}{cli::pb_bar}| ",
        "{cli::pb_current_bytes}/{cli::pb_total_bytes} {cli::pb_eta_str}"
      )
    } else {
      opt <- getOption("cli.progress_format_download_nototal") %||%
        getOption("cli.progress_format_download")
      if (!is.null(opt)) return(opt)
      paste0(
        "{cli::pb_name}{cli::pb_status}{cli::pb_spin} ",
        "{cli::pb_current_bytes} ({cli::pb_rate_bytes}) | {cli::pb_elapsed}"
      )
    }
  }
}

test_that("cli_progress_bar", {
  withr::local_options(cli.dynamic = FALSE, cli.ansi = FALSE)
  fun <- function() {
    bar <- cli_progress_bar(
      name = "name",
      status = "status",
      format = "{cli::pb_spin} {cli::pb_name}{cli::pb_status}{cli::pb_current}"
    )
    cli_progress_update(force = TRUE)
    cli_progress_done(id = bar)
  }
  expect_snapshot(capture_cli_messages(fun()))
})

test_that("custom format needs a format string", {
  expect_error(cli_progress_bar(type = "custom"), "Need to specify format")
})

test_that("removes previous progress bar", {
  withr::local_options(cli.dynamic = FALSE, cli.ansi = FALSE)
  fun <- function() {
    bar <- cli_progress_bar(format = "first", format_done = "first done", clear = FALSE)
    cli_progress_update(force = TRUE)
    bar2 <- cli_progress_bar(format = "second", format_done = "second done", clear = FALSE)
    cli_progress_update(force = TRUE)
  }

  expect_snapshot(capture_cli_messages(fun()))
})

test_that("backend methods are called", {
  withr::local_options(cli.progress_handlers_only = "logger")
  fun <- function() {
    cli_progress_bar()
    cli_progress_update(force = TRUE)
    cli_progress_done()
  }
  out <- capture_output(fun())
  expect_match(out, "created.*added.*updated.*terminated")
})

test_that("update errors if no progress bar", {
  fun <- function() {
    cli_progress_update()
  }
  expect_error(fun(), "Cannot find current progress bar")

  fun <- function() {
    cli_progress_output("boo")
  }
  expect_error(fun(), "Cannot find current progress bar")

  envkey <- NULL
  fun <- function() {
    envkey <<- format(environment())
    clienv$progress_ids[[envkey]] <- "foobar"
    cli_progress_update()
  }
  expect_error(fun(), "Cannot find progress bar")

  envkey <- NULL
  fun <- function() {
    envkey <<- format(environment())
    clienv$progress_ids[[envkey]] <- "foobar"
    cli_progress_output("booboo")
  }
  expect_error(fun(), "Cannot find progress bar")

  clienv$progress_ids[[envkey]] <- NULL
})

test_that("cli_progress_update can update status", {
  withr::local_options(cli.dynamic = FALSE, cli.ansi = FALSE)
  fun <- function() {
    bar <- cli_progress_bar(
      name = "name",
      status = "status",
      format = "{cli::pb_spin} {cli::pb_name}{cli::pb_status}{cli::pb_current}"
    )
    cli_progress_update(force = TRUE)
    cli_progress_update(status = "new status", force = TRUE)
    cli_progress_done(id = bar)
  }
  expect_snapshot(capture_cli_messages(fun()))
})

test_that("cli_progress_update can update extra data", {
  withr::local_options(cli.dynamic = FALSE, cli.ansi = FALSE)
  fun <- function() {
    bar <- cli_progress_bar(
      format = "Extra: {cli::pb_extra$foo}",
      extra = list(foo = "bar")
    )
    cli_progress_update(force = TRUE)
    cli_progress_update(extra = list(foo = "baz"), force = TRUE)
    cli_progress_done(id = bar)
  }
  expect_snapshot(capture_cli_messages(fun()))
})

test_that("update set", {
  withr::local_options(cli.dynamic = FALSE, cli.ansi = FALSE)
  fun <- function() {
    bar <- cli_progress_bar(
      name = "name",
      total = 100,
      format = "{cli::pb_name}{cli::pb_status}{cli::pb_current}/{cli::pb_total}"
    )
    cli_progress_update(force = TRUE)
    cli_progress_update(force = TRUE, set = 50)
    cli_progress_done(id = bar)
  }
  expect_snapshot(capture_cli_messages(fun()))
})

test_that("format changes if we (un)learn total", {
  withr::local_options(cli.dynamic = FALSE, cli.ansi = FALSE)
  fun <- function() {
    bar <- cli_progress_bar(
      name = "name",
      total = NA,
    )
    cli_progress_update(force = TRUE)
    cli_progress_update(force = TRUE, set = 50, total = 100)
    cli_progress_update(force = TRUE, set = 75, total = NA)
    cli_progress_done(id = bar)
  }
  out <- fix_times(capture_cli_messages(fun()))
  expect_snapshot(out)
})

test_that("auto-terminate", {
  withr::local_options(cli.dynamic = FALSE, cli.ansi = FALSE)
  fun <- function() {
    bar <- cli_progress_bar(total = 10, format = "first", format_done = "first done", clear = FALSE)
    cli_progress_update(force = TRUE)
    cli_progress_update(force = TRUE, set = 10)
    cli_text("First is done by now.\n")
    bar2 <- cli_progress_bar(format = "second", format_done = "second done", clear = FALSE)
    cli_progress_update(force = TRUE)
  }

  expect_snapshot(capture_cli_messages(fun()))
})

test_that("done does nothing if no progress bar", {
  fun <- function() {
    cli_progress_done()
  }
  expect_true(fun())
})

test_that("cli_progress_output", {
  skip_if_not_installed("testthat", "3.1.2")
  withr::local_options(cli.dynamic = TRUE, cli.ansi = TRUE)
  fun <- function() {
    bar <- cli_progress_bar(total = 10, format = "first")
    cli_progress_update(force = TRUE)
    cli_progress_output("just {1} text{?s}")
    cli_progress_update(force = TRUE)
  }

  expect_snapshot(capture_cli_messages(fun()))

  withr::local_options(cli.dynamic = TRUE, cli.ansi = FALSE)
  expect_snapshot(capture_cli_messages(fun()))
})

test_that("cli_progress_bar handles Inf like NA", {
  withr::local_options(cli.dynamic = FALSE, cli.ansi = FALSE)
  fun <- function(total) {
    bar <- cli_progress_bar(
      name = "name",
      format = "{cli::pb_name}{cli::pb_current}",
      total = total
    )
    cli_progress_update(force = TRUE)
    cli_progress_done(id = bar)
  }
  expect_equal(capture_cli_messages(fun(total = NA)), capture_cli_messages(fun(total = Inf)))
})

test_that("cli_progress_bar", {
  withr::local_options(cli.dynamic = FALSE, cli.ansi = FALSE)
  fun <- function() {
    bar <- cli_progress_bar(
      name = "name",
      status = "status",
      format = "{cli::pb_spin} {cli::pb_name}{cli::pb_status}{cli::pb_current}"
    )
    cli_progress_update(force = TRUE)
    cli_progress_done(id = bar)
  }
  expect_snapshot(capture_cli_messages(fun()))
})

test_that("custom format needs a format string", {
  expect_error(cli_progress_bar(type = "custom"), "Need to specify format")
})

test_that("removes previous progress bar", {
  withr::local_options(cli.dynamic = FALSE, cli.ansi = FALSE)
  fun <- function() {
    bar <- cli_progress_bar(format = "first", format_done = "first done", clear = FALSE)
    cli_progress_update(force = TRUE)
    bar2 <- cli_progress_bar(format = "second", format_done = "second done", clear = FALSE)
    cli_progress_update(force = TRUE)
  }

  expect_snapshot(capture_cli_messages(fun()))
})

test_that("backend methods are called", {
  withr::local_options(cli.progress_handlers_only = "logger")
  fun <- function() {
    cli_progress_bar()
    cli_progress_update(force = TRUE)
    cli_progress_done()
  }
  out <- capture_output(fun())
  expect_match(out, "created.*added.*updated.*terminated")
})

test_that("update errors if no progress bar", {
  fun <- function() {
    cli_progress_update()
  }
  expect_error(fun(), "Cannot find current progress bar")

  fun <- function() {
    cli_progress_output("boo")
  }
  expect_error(fun(), "Cannot find current progress bar")
  
  envkey <- NULL
  fun <- function() {
    envkey <<- format(environment())
    clienv$progress_ids[[envkey]] <- "foobar"
    cli_progress_update()
  }
  expect_error(fun(), "Cannot find progress bar")

  envkey <- NULL
  fun <- function() {
    envkey <<- format(environment())
    clienv$progress_ids[[envkey]] <- "foobar"
    cli_progress_output("booboo")
  }
  expect_error(fun(), "Cannot find progress bar")

  clienv$progress_ids[[envkey]] <- NULL
})

test_that("cli_progress_update can update status", {
  withr::local_options(cli.dynamic = FALSE, cli.ansi = FALSE)
  fun <- function() {
    bar <- cli_progress_bar(
      name = "name",
      status = "status",
      format = "{cli::pb_spin} {cli::pb_name}{cli::pb_status}{cli::pb_current}"
    )
    cli_progress_update(force = TRUE)
    cli_progress_update(status = "new status", force = TRUE)
    cli_progress_done(id = bar)
  }
  expect_snapshot(capture_cli_messages(fun()))
})

test_that("cli_progress_update can update extra data", {
  withr::local_options(cli.dynamic = FALSE, cli.ansi = FALSE)
  fun <- function() {
    bar <- cli_progress_bar(
      format = "Extra: {cli::pb_extra$foo}",
      extra = list(foo = "bar")
    )
    cli_progress_update(force = TRUE)
    cli_progress_update(extra = list(foo = "baz"), force = TRUE)
    cli_progress_done(id = bar)
  }
  expect_snapshot(capture_cli_messages(fun()))
})

test_that("update set", {
  withr::local_options(cli.dynamic = FALSE, cli.ansi = FALSE)
  fun <- function() {
    bar <- cli_progress_bar(
      name = "name",
      total = 100,
      format = "{cli::pb_name}{cli::pb_status}{cli::pb_current}/{cli::pb_total}"
    )
    cli_progress_update(force = TRUE)
    cli_progress_update(force = TRUE, set = 50)
    cli_progress_done(id = bar)
  }
  expect_snapshot(capture_cli_messages(fun()))
})

test_that("format changes if we (un)learn total", {
  withr::local_options(cli.dynamic = FALSE, cli.ansi = FALSE)
  fun <- function() {
    bar <- cli_progress_bar(
      name = "name",
      total = NA,
    )
    cli_progress_update(force = TRUE)
    cli_progress_update(force = TRUE, set = 50, total = 100)
    cli_progress_update(force = TRUE, set = 75, total = NA)
    cli_progress_done(id = bar)
  }
  out <- fix_times(capture_cli_messages(fun()))
  expect_snapshot(out)
})

test_that("auto-terminate", {
  withr::local_options(cli.dynamic = FALSE, cli.ansi = FALSE)
  fun <- function() {
    bar <- cli_progress_bar(total = 10, format = "first", format_done = "first done", clear = FALSE)
    cli_progress_update(force = TRUE)
    cli_progress_update(force = TRUE, set = 10)
    cli_text("First is done by now.\n")
    bar2 <- cli_progress_bar(format = "second", format_done = "second done", clear = FALSE)
    cli_progress_update(force = TRUE)
  }

  expect_snapshot(capture_cli_messages(fun()))
})

test_that("done does nothing if no progress bar", {
  fun <- function() {
    cli_progress_done()
  }
  expect_true(fun())
})

test_that("cli_progress_output", {
  if (packageVersion("testthat") <= "3.1.1") skip("needs newer testthat")
  withr::local_options(cli.dynamic = TRUE, cli.ansi = TRUE)
  fun <- function() {
    bar <- cli_progress_bar(total = 10, format = "first")
    cli_progress_update(force = TRUE)
    cli_progress_output("just {1} text{?s}")
    cli_progress_update(force = TRUE)
  }

  expect_snapshot(capture_cli_messages(fun()))

  withr::local_options(cli.dynamic = TRUE, cli.ansi = FALSE)
  expect_snapshot(capture_cli_messages(fun()))  
})

#' Print the helpful ruler to the screen
#'
#' @export
#' @param width Ruler width.
#' @examples
#' ruler()

ruler <- function(width = console_width()) {
  x <- seq_len(width)
  y <- rep("-", length(x))

  y[x %% 5 == 0] <- "+"
  y[x %% 10 == 0] <- style_bold(as.character((x[x %% 10 == 0] %/% 10) %% 10))

  cat(y, "\n", sep = "")
  cat(x %% 10, "\n", sep = "")
}

# keep encoding, even if useBytes = TRUE

sub_ <- function(pattern, replacement, x, ...) {
  enc <- Encoding(x)
  ret <- sub(pattern, replacement, x, ...)
  if (length(ret)) Encoding(ret) <- enc
  ret
}

gsub_ <- function(pattern, replacement, x, ...) {
  enc <- Encoding(x)
  ret <- gsub(pattern, replacement, x, ...)
  if (length(ret)) Encoding(ret) <- enc
  ret
}

strsplit_ <- function(x, ...) {
  enc <- Encoding(x)
  ret <- strsplit(x, ...)
  for (i in seq_along(ret)) {
    Encoding(ret[[i]]) <- enc[i]
  }
  ret
}

#' @export
`__cli_update_due` <- FALSE

#' @export

cli_tick_reset <- function() {
  .Call(clic_tick_reset)
}

#' @export

ccli_tick_reset <- NULL

cli_tick_set <- function(tick_time = NULL, speed_time = NULL) {
  tick_time <- tick_time %||% clienv$tick_time
  speed_time <- speed_time %||% clienv$speed_time

  clienv$speed_time <- as.double(speed_time)
  clienv$tick_time <- as.integer(tick_time)
  .Call(clic_tick_set, clienv$tick_time, clienv$speed_time)
  invisible()
}

cli_tick_pause <- function(state = TRUE) {
  .Call(clic_tick_pause, state)
}

cli_tick_resume <- function(state = TRUE) {
  .Call(clic_tick_resume, state)
}

cli_with_ticks <- function(expr) {
  on.exit(cli_tick_resume(TRUE), add = TRUE)
  opts <- options(cli.progress_show_after = 0)
  on.exit(options(opts), add = TRUE)
  cli_tick_pause(TRUE)
  expr
}

cli_without_ticks <- function(expr) {
  on.exit(cli_tick_resume(TRUE), add = TRUE)
  opts <- options(cli.progress_show_after = 0)
  on.exit(options(opts), add = TRUE)
  cli_tick_pause(FALSE)
  expr
}

test_that("ALTREP methods", {
  expect_equal(length(`__cli_update_due`), 1L)

  expect_silent({
    tim <- `__cli_update_due`
    tim[1] <- FALSE
  })

  expect_silent({
    `__cli_update_due`[1]
  })
})

test_that("cli_tick_set", {
  skip_on_cran()
  expect_silent(cli_tick_set())
})

#' Working around the bad Unicode character widths
#'
#' R 3.6.2 and also the coming 3.6.3 and 4.0.0 versions use the Unicode 8
#' standard to calculate the display width of Unicode characters.
#' Unfortunately the widths of most emojis are incorrect in this standard,
#' and width 1 is reported instead of the correct 2 value.
#'
#' cli implements a workaround for this. The package contains a table that
#' contains all Unicode ranges that have wide characters (display width 2).
#'
#' On first use of one of the workaround wrappers (in `ansi_nchar()`, etc.)
#' we check what the current version of R thinks about the width of these
#' characters, and then create a regex that matches the ones that R
#' is wrong about (`re_bad_char_width`).
#'
#' Then we use this regex to duplicate all of the problematic characters
#' in the input string to the wrapper function, before calling the real
#' string manipulation function (`nchar()`, `strwrap()`) etc. At end we
#' undo the duplication before we return the result.
#'
#' This workaround is fine for `nchar()` and `strwrap()`, and consequently
#' `ansi_align()` and `ansi_strtrim()` as well.
#'
#' The rest of the `ansi_*()` functions work on characters, and do not
#' deal with character width.
#'
#' @keywords internal
#' @name unicode-width-workaround
NULL

setup_unicode_width_fix <- function() {
  bad <- base::nchar(wide_chars$test, type = "width") == 1
  re <- paste0(wide_chars$regex[bad], collapse = "")
  clienv$re_bad_char_width <- paste0("([", re, "])")
  clienv$re_bad_char_width_fix <- paste0("([", re, "])\\1")
}

unicode_pre <- function(x) {
  if (is.null(clienv$re_bad_char_width)) setup_unicode_width_fix()
  if (clienv$re_bad_char_width != "([])") {
    x <- gsub(clienv$re_bad_char_width, "\\1\\1", x, perl = TRUE)
  }
  x
}

unicode_post <- function(x) {
  if (is.null(clienv$re_bad_char_width)) setup_unicode_width_fix()
  if (clienv$re_bad_char_width != "([])") {
    x <- gsub(clienv$re_bad_char_width_fix, "\\1", x, perl = TRUE)
  }
  x
}

#' Read a single keypress at the terminal
#'
#' It currently only works at Linux/Unix and OSX terminals,
#' and at the Windows command line. see \code{\link{has_keypress_support}}.
#'
#' The following special keys are supported:
#' * Arrow keys: 'up', 'down', 'right', 'left'.
#' * Function keys: from 'f1' to 'f12'.
#' * Others: 'home', 'end', 'insert', 'delete', 'pageup', 'pagedown',
#'     'tab', 'enter', 'backspace' (same as 'delete' on OSX keyboards),
#'     'escape'.
#' * Control with one of the following keys: 'a', 'b', 'c', 'd', 'e', 'f',
#'     'h', 'k', 'l', 'n', 'p', 't', 'u', 'w'.
#'
#' @param block Whether to wait for a key press, if there is none
#'   available now.
#' @return The key pressed, a character scalar. For non-blocking reads
#'   `NA` is returned if no keys are available.
#'
#' @family keypress function
#' @export
#' @examplesIf FALSE
#' x <- keypress()
#' cat("You pressed key", x, "\n")

keypress <- function(block = TRUE) {
  if (!has_keypress_support()) {
    stop("Your platform/terminal does not support `keypress()`.")
  }
  block <- as.logical(block)
  if (length(block) != 1) stop("'block' must be a logical scalar")
  ret <- .Call(cli_keypress, block)
  if (ret == "none") NA_character_ else ret
}

#' Check if the current platform/terminal supports reading
#' single keys.
#'
#' @details
#' Supported platforms:
#' * Terminals in Windows and Unix.
#' * RStudio terminal.
#'
#' Not supported:
#' * RStudio (if not in the RStudio terminal).
#' * R.app on macOS.
#' * Rgui on Windows.
#' * Emacs ESS.
#' * Others.
#'
#' @return Whether there is support for waiting for individual
#' keypressses.
#'
#' @family keypress function
#' @export
#' @examples
#' has_keypress_support()

has_keypress_support <- function() {
  ## Supported if we have a terminal or RStudio terminal.
  ## Not supported otherwise in RStudio, R.app, Rgui or Emacs

  rs <- rstudio$detect()

  if (rs$type != "not_rstudio") {
    rs$has_canonical_mode

  } else {
    isatty(stdin()) &&
      Sys.getenv("R_GUI_APP_VERSION") == "" &&
      .Platform$GUI != "Rgui" &&
      ! identical(getOption("STERM"), "iESS") &&
      Sys.getenv("EMACS") != "t" &&
      Sys.getenv("TERM") != "dumb"
  }
}

# modified from https://shiny.rstudio.com/articles/progress.html

library(cli)

# !!! You don't need these in real code.
# cli.progress_show_after makes sure that we see the progress bar from
# the beginning, not only after a delay.
options(cli.progress_show_after = 0)

# !!! You don't need this in real code.
options(cli.progress_handlers_only = "shiny")

server <- function(input, output) {
  output$plot <- renderPlot({
    input$goPlot # Re-run when button is clicked

    # Create 0-row data frame which will be used to store data
    dat <- data.frame(x = numeric(0), y = numeric(0))

    # Number of times we'll go through the loop
    n <- 10

    cli_progress_bar(total = n, "Rendering plot")
    for (i in 1:n) {
      # Each time through the loop, add another row of data. This is
      # a stand-in for a long-running computation.
      dat <- rbind(dat, data.frame(x = rnorm(1), y = rnorm(1)))

      cli_progress_update(status = paste("Doing part", i))

      # Pause for 0.5 seconds to simulate a long computation.
      # Produce some extra output
      Sys.sleep(0.5)
      msg <- paste(strwrap(cli:::lorem_ipsum(1, 1)), collapse = "\n")
      if (i != n) cli_progress_output(msg)
      Sys.sleep(0.5)
    }

    plot(dat$x, dat$y)
  })
}

ui <- shinyUI(basicPage(
  plotOutput('plot', width = "300px", height = "300px"),
  actionButton('goPlot', 'Go plot')
))

shinyApp(ui = ui, server = server)

re_match <- function(text, pattern, perl = TRUE, ...) {

  stopifnot(is.character(pattern), length(pattern) == 1, !is.na(pattern))
  text <- as.character(text)

  match <- regexpr(pattern, text, perl = perl, ...)

  start  <- as.vector(match)
  length <- attr(match, "match.length")
  end    <- start + length - 1L

  matchstr <- substring(text, start, end)
  matchstr[ start == -1 ] <- NA_character_

  empty <- data.frame(stringsAsFactors = FALSE, .text = text)[, numeric()]
  res <- list(match = !is.na(matchstr), groups = empty)

  if (!is.null(attr(match, "capture.start"))) {

    gstart  <- attr(match, "capture.start")
    glength <- attr(match, "capture.length")
    gend    <- gstart + glength - 1L

    groupstr <- substring(text, gstart, gend)
    groupstr[ gstart == -1 ] <- NA_character_
    dim(groupstr) <- dim(gstart)

    res$groups <- cbind(groupstr, res$groups, stringsAsFactors = FALSE)
    names(res$groups) <- attr(match, "capture.names")
  }

  res$groups
}

#' Compare two character vectors elementwise
#'
#' Its printed output is similar to calling `diff -u` at the command
#' line.
#'
#' @param old First character vector.
#' @param new Second character vector.
#' @param max_dist Maximum distance to consider, or `Inf` for no limit.
#'   If the LCS edit distance is larger than this, then the function
#'   throws an error with class `"cli_diff_max_dist"`. (If you specify
#'   `Inf` the real limit is `.Machine$integer.max` but to reach this the
#'   function would have to run a very long time.)
#' @return A list that is a `cli_diff_chr` object, with a `format()` and a
#' `print()` method. You can also access its members:
#'  * `old` and `new` are the original inputs,
#'  * `lcs` is a data frame of LCS edit that transform `old` into `new`.
#'
#' The `lcs` data frame has the following columns:
#' * `operation`: one of `"match"`, `"delete"` or `"insert"`.
#' * `offset`: offset in `old` for matches and deletions, offset in `new`
#'   for insertions.
#' * `length`: length of the operation, i.e. number of matching, deleted
#'   or inserted elements.
#' * `old_offset`: offset in `old` _after_ the operation.
#' * `new_offset`: offset in `new` _after_ the operation.
#'
#' @family diff functions in cli
#' @seealso The diffobj package for a much more comprehensive set of
#' `diff`-like tools.
#' @export
#' @examples
#' letters2 <- c("P", "R", "E", letters, "P", "O", "S", "T")
#' letters2[11:16] <- c("M", "I", "D", "D", "L", "E")
#' diff_chr(letters, letters2)

diff_chr <- function(old, new, max_dist = Inf) {
  stopifnot(
    is.character(old),
    is.character(new),
    max_dist == Inf || is_count(max_dist)
  )
  max_dist2 <- as_max_dist(max_dist)

  lcs <- .Call(clic_diff_chr, old, new, max_dist2)

  if (max_dist2 != 0 && lcs[[4]] == max_dist2) {
    throw(cli_error(
      "Diff edit distance is larger than the limit.",
      "i" = "The edit distance limit is {max_dist}.",
      .data = list(max_dist = max_dist),
      .class = "cli_diff_max_dist"
    ))
  }

  op <- c("match", "delete", "insert")[lcs[[1]]]
  lcs <- data.frame(
    stringsAsFactors = FALSE,
    operation = op,
    offset = lcs[[2]],
    length = lcs[[3]],
    old_offset = cumsum(ifelse(op == "insert", 0, lcs[[3]])),
    new_offset = cumsum(ifelse(op == "delete", 0, lcs[[3]]))
  )

  ret <- structure(
    list(old = old, new = new, lcs = lcs),
    class = c("cli_diff_chr", "cli_diff", "list")
  )

  ret
}

#' Compare two character strings, character by character
#'
#' Characters are defined by UTF-8 graphemes.
#'
#' @param old First string, must not be `NA`.
#' @param new Second string, must not be `NA`.
#' @inheritParams diff_chr
#' @return A list that is a `cli_diff_str` object and also a
#'   `cli_diff_chr` object, see [diff_str] for the details about its
#'   structure.
#'
#' @family diff functions in cli
#' @seealso The diffobj package for a much more comprehensive set of
#' `diff`-like tools.
#'
#' @export
#' @examples
#' str1 <- "abcdefghijklmnopqrstuvwxyz"
#' str2 <- "PREabcdefgMIDDLEnopqrstuvwxyzPOST"
#' diff_str(str1, str2)

diff_str <- function(old, new, max_dist = Inf) {
  stopifnot(
    is_string(old),
    is_string(new)
    # max_dist is checked in diff_chr
  )

  old1 <- utf8_graphemes(old)[[1]]
  new1 <- utf8_graphemes(new)[[1]]

  ret <- diff_chr(old1, new1, max_dist)

  class(ret) <- c("cli_diff_str", class(ret))

  ret
}

#' @export

format.cli_diff_chr <- function(x, context = 3L, ...) {
  stopifnot(context == Inf || is_count(context))
  if (length(list(...)) > 0) {
    warning("Extra arguments were ignored in `format.cli_diff_chr()`.")
  }

  chunks <- get_diff_chunks(x$lcs, context = context)
  out <- lapply(
    seq_len(nrow(chunks)),
    format_chunk,
    x = x,
    chunks = chunks,
    context = context
  )

  ret <- as.character(unlist(out))
  if (context == Inf && length(ret) > 0) ret <- ret[-1]

  ret
}

get_diff_chunks <- function(lcs, context = 3L) {
  # the number of chunks is the number of non-matching sequences if
  # context == 0, but short matching parts do not separate chunks
  runs <- rle(lcs$operation != "match" | lcs$length <= 2 * context)
  nchunks <- sum(runs$values)

  # special case for a single match chunk
  if (nrow(lcs) == 1 && lcs$operation == "match") {
    nchunks <- if (context == Inf) 1 else 0
  }

  chunks <- data.frame(
    op_begin   = integer(nchunks),    # first op in chunk
    op_length  = integer(nchunks),    # number of operations in chunk
    old_begin  = integer(nchunks),    # first line from `old` in chunk
    old_length = integer(nchunks),    # number of lines from `old` in chunk
    new_begin  = integer(nchunks),    # first line from `new` in chunk
    new_length = integer(nchunks)     # number of lines from `new` in chunk
  )

  if (nchunks == 0) return(chunks)

  # infer some data about the original diff input
  old_off <- c(0, lcs$old_offset)
  new_off <- c(0, lcs$new_offset)
  old_size <- old_off[length(old_off)]
  new_size <- new_off[length(new_off)]
  old_empty <- old_size == 0
  new_empty <- new_size == 0

  # avoid working with Inf
  if (context == Inf) context <- max(old_size, new_size)

  # chunk starts at operation number sum(length) before it, plus 1, but
  # at the end we change this to include the context chunks are well
  chunks$op_begin  <- c(0, cumsum(runs$lengths))[which(runs$values)] + 1
  chunks$op_length <- runs$lengths[runs$values]

  # `old` positions are from `old_off`, but need to fix the boundaries
  chunks$old_begin <- old_off[chunks$op_begin] - context + 1
  chunks$old_begin[chunks$old_begin <= 1] <- if (old_empty) 0 else 1
  old_end <- old_off[chunks$op_begin + chunks$op_length] + context
  old_end[old_end > old_size] <- old_size
  chunks$old_length <- old_end - chunks$old_begin + 1

  # `new` positions are similar
  chunks$new_begin <- new_off[chunks$op_begin] - context + 1
  chunks$new_begin[chunks$new_begin <= 1] <- if (new_empty) 0 else 1
  new_end <- new_off[chunks$op_begin + chunks$op_length] + context
  new_end[new_end > new_size] <- new_size
  chunks$new_length <- new_end - chunks$new_begin + 1

  # change to include context chunks
  if (context > 0) {
    # calculae the end before updating the begin
    op_end <- chunks$op_begin + chunks$op_length - 1 + 1
    op_end[op_end > nrow(lcs)] <- nrow(lcs)
    chunks$op_begin <- chunks$op_begin - 1
    chunks$op_begin[chunks$op_begin == 0] <- 1
    chunks$op_length <- op_end - chunks$op_begin + 1
  }

  chunks
}

format_chunk <- function(x, chunks, num, context) {
  hdr <- paste0(
    "@@ -",
    chunks$old_begin[num],
    if ((l <- chunks$old_length[num]) != 1) paste0(",", l),
    " +",
    chunks$new_begin[num],
    if ((l <- chunks$new_length[num]) != 1) paste0(",", l),
    " @@"
  )

  from <- chunks$op_begin[num]
  to <- chunks$op_begin[num] + chunks$op_length[num] - 1

  lines <- lapply(from:to, function(i) {
    op <- x$lcs$operation[i]
    off <- x$lcs$offset[i]
    len <- x$lcs$length[i]
    if (op == "match") {
      if (len > context) {
        if (i == from) {
          # start later
          off <- off + len - context
          len <- context
        } else {
          # finish earlier
          len <- context
        }
      }
      paste0(" ", x$old[off + 1:len])

    } else if (op == "delete") {
      col_blue(paste0("-", x$old[off + 1:len]))

    } else if (op == "insert") {
      col_green(paste0("+", x$new[off + 1:len]))
    }
  })
  c(hdr, lines)
}

#' @export

print.cli_diff_chr <- function(x, ...) {
  writeLines(format(x, ...))
}

#' @export

format.cli_diff_str <- function(x, ...) {
  if (length(list(...)) > 0) {
    warning("Extra arguments were ignored in `format.cli_diff_chr()`.")
  }

  if (num_ansi_colors() == 1) {
    format_diff_str_nocolor(x, ...)
  } else {
    format_diff_str_color(x, ...)
  }
}

format_diff_str_color <- function(x, ...) {
  out <- lapply(seq_len(nrow(x$lcs)), function(i) {
    op <- x$lcs$operation[i]
    off <- x$lcs$offset[i]
    len <- x$lcs$length[i]
    if (op == "match") {
      paste0(x$old[off + 1:len], collapse = "")

    } else if (op == "delete") {
      bg_blue(col_black(paste0(x$old[off + 1:len], collapse = "")))

    } else if (op == "insert") {
      bg_green(col_black(paste0(x$new[off + 1:len], collapse = "")))
    }
  })

  paste(out, collapse = "")
}

format_diff_str_nocolor <- function(x, ...) {
  out <- lapply(seq_len(nrow(x$lcs)), function(i) {
    op <- x$lcs$operation[i]
    off <- x$lcs$offset[i]
    len <- x$lcs$length[i]
    if (op == "match") {
      paste0(x$old[off + 1:len], collapse = "")

    } else if (op == "delete") {
      paste0(c("[-", x$old[off + 1:len], "-]"), collapse = "")

    } else if (op == "insert") {
      paste0(c("{+", x$new[off + 1:len], "+}"), collapse = "")
    }
  })

  paste(out, collapse = "")
}

as_max_dist <- function(max_dist) {
  if (max_dist == Inf) {
    0L
  } else {
    as.integer(max_dist + 1L)
  }
}

test_that("diff_chr", {
  # Something simple first
  a <- as.character(c(1,1,1,1,1,1,1,2,3,4,4,4,4,4,4,4,5))
  b <- as.character(c(1,1,1,1,1,1,1,2,10,4,4,4,4,4,4,4,6,7,5))
  d <- diff_chr(a, b)
  expect_snapshot(d$lcs)
  d <- diff_chr(b, a)
  expect_snapshot(d$lcs)
})

test_that_cli(configs = c("plain", "ansi"), "diff_chr", {
  # Something simple first
  a <- as.character(c(1,1,1,1,1,1,1,2,3,4,4,4,4,4,4,4,5))
  b <- as.character(c(1,1,1,1,1,1,1,2,10,4,4,4,4,4,4,4,6,7,5))
  d <- diff_chr(a, b)
  expect_snapshot(d)
  expect_snapshot(d$lcs)
})

test_that("diff_chr edge cases", {
  expect_snapshot(diff_chr(character(), character()))
  expect_snapshot(diff_chr(character(), character())$lcs)
  expect_snapshot(diff_chr("a", character()))
  expect_snapshot(diff_chr(character(), "b"))
  expect_snapshot(diff_chr("a", "a"))
  expect_snapshot(diff_chr(letters, letters))
  expect_snapshot(diff_chr(c("a", NA, "a2"), "b"))
  expect_snapshot(diff_chr(NA_character_, "NA"))
})

test_that("format.cli_diff_chr context", {
  # Something simple first
  a <- as.character(c(1,1,1,1,1,1,1,2,3,4,4,4,4,4,4,4,5))
  b <- as.character(c(1,1,1,1,1,1,1,2,10,4,4,4,4,4,4,4,6,7,5))
  d <- diff_chr(a, b)
  expect_snapshot(print(d, context = 1))
  expect_snapshot(print(d, context = 0))
  expect_snapshot(print(d, context = Inf))
  d2 <- diff_chr(c("foo", "bar"), c("foo", "bar"))
  expect_snapshot(print(d2, context = Inf))
})

test_that_cli(configs = c("plain", "ansi"), "diff_str", {
  str1 <- "abcdefghijklmnopqrstuvwxyz"
  str2 <- "PREabcdefgMIDDLEnopqrstuvwxyzPOST"
  d <- diff_str(str1, str2)
  expect_snapshot(d)
})

test_that("warnings and errors", {
  expect_error(diff_chr(1:10, 1:10), "is.character")
  expect_error(
    format(diff_chr("foo", "bar"), context = -1),
    "is_count"
  )
  expect_warning(
    format(diff_chr("foo", "bar"), what = 1, is = 2, this = 3),
    "Extra arguments"
  )
  expect_warning(
    format(diff_str("foo", "bar"), what = 1, is = 2, this = 3),
    "Extra arguments"
  )
})

test_that("max_diff", {
  expect_snapshot_error(
    class = "cli_diff_max_dist",
    diff_chr("a", c("a", "b"), 0)
  )

  expect_silent(diff_chr(c("a", "c"), c("a", "b"), 2))

  expect_snapshot_error(
    class = "cli_diff_max_dist",
    diff_chr(c("a", "c"), c("a", "b"), 1)
  )
})

test_that("diff_chr", {
  # Something simple first
  a <- as.character(c(1,1,1,1,1,1,1,2,3,4,4,4,4,4,4,4,5))
  b <- as.character(c(1,1,1,1,1,1,1,2,10,4,4,4,4,4,4,4,6,7,5))
  d <- diff_chr(a, b)
  expect_snapshot(d$lcs)
  d <- diff_chr(b, a)
  expect_snapshot(d$lcs)
})

test_that_cli(configs = c("plain", "ansi"), "diff_chr", {
  # Something simple first
  a <- as.character(c(1,1,1,1,1,1,1,2,3,4,4,4,4,4,4,4,5))
  b <- as.character(c(1,1,1,1,1,1,1,2,10,4,4,4,4,4,4,4,6,7,5))
  d <- diff_chr(a, b)
  expect_snapshot(d)
  expect_snapshot(d$lcs)
})

test_that("diff_chr edge cases", {
  expect_snapshot(diff_chr(character(), character()))
  expect_snapshot(diff_chr(character(), character())$lcs)
  expect_snapshot(diff_chr("a", character()))
  expect_snapshot(diff_chr(character(), "b"))
  expect_snapshot(diff_chr("a", "a"))
  expect_snapshot(diff_chr(letters, letters))
  expect_snapshot(diff_chr(c("a", NA, "a2"), "b"))
  expect_snapshot(diff_chr(NA_character_, "NA"))
})

test_that("format.cli_diff_chr context", {
  # Something simple first
  a <- as.character(c(1,1,1,1,1,1,1,2,3,4,4,4,4,4,4,4,5))
  b <- as.character(c(1,1,1,1,1,1,1,2,10,4,4,4,4,4,4,4,6,7,5))
  d <- diff_chr(a, b)
  expect_snapshot(print(d, context = 1))
  expect_snapshot(print(d, context = 0))
  expect_snapshot(print(d, context = Inf))
  d2 <- diff_chr(c("foo", "bar"), c("foo", "bar"))
  expect_snapshot(print(d2, context = Inf))
})

test_that_cli(config = c("plain", "ansi"), "diff_str", {
  str1 <- "abcdefghijklmnopqrstuvwxyz"
  str2 <- "PREabcdefgMIDDLEnopqrstuvwxyzPOST"
  d <- diff_str(str1, str2)
  expect_snapshot(d)
})

test_that("warnings and errors", {
  expect_error(diff_chr(1:10, 1:10), "is.character")
  expect_error(
    format(diff_chr("foo", "bar"), context = -1),
    "is_count"
  )
  expect_warning(
    format(diff_chr("foo", "bar"), what = 1, is = 2, this = 3),
    "Extra arguments"
  )
  expect_warning(
    format(diff_str("foo", "bar"), what = 1, is = 2, this = 3),
    "Extra arguments"
  )
})

test_that("max_diff", {
  err <- tryCatch(
    diff_chr("a", c("a", "b"), 0),
    error = function(e) e
  )
  expect_s3_class(err, "cli_diff_max_dist")

  expect_silent(diff_chr(c("a", "c"), c("a", "b"), 2))
  err <- tryCatch(
    diff_chr(c("a", "c"), c("a", "b"), 1),
    error = function(e) e
  )
  expect_s3_class(err, "cli_diff_max_dist")
})

## See tools/spinners.R for how the RDS file is created

#' Character vector to put a spinner on the screen
#'
#' `cli` contains many different spinners, you choose one according to your
#' taste.
#'
#' ```{asciicast get-spinner}
#' #| asciicast_at = "all",
#' #| asciicast_knitr_output = "svg",
#' #| asciicast_cursor = FALSE
#' options(cli.spinner = "hearts")
#' fun <- function() {
#'   cli_progress_bar("Spinning")
#'   for (i in 1:100) {
#'     Sys.sleep(4/100)
#'     cli_progress_update()
#'   }
#' }
#' fun()
#' options(cli.spinner = NULL)
#' ```
#'
#' @param which The name of the chosen spinner. If `NULL`, then the default
#'   is used, which can be customized via the `cli.spinner_unicode`,
#'   `cli.spinner_ascii` and `cli.spinner` options. (The latter applies to
#'   both Unicode and ASCII displays. These options can be set to the name
#'   of a built-in spinner, or to a list that has an entry called `frames`,
#'   a character vector of frames.
#' @return A list with entries: `name`, `interval`: the suggested update
#'   interval in milliseconds and `frames`: the character vector of the
#'   spinner's frames.
#'
#' @family spinners
#' @export

get_spinner <- function(which = NULL) {
  stopifnot(is.null(which) || is_string(which) || is.list(which))

  if (is.null(which)) {
    if (is_utf8_output()) {
      which <-
        getOption("cli.spinner_unicode") %||%
        getOption("cli.spinner") %||%
        "dots"
    } else {
      which <-
        getOption("cli.spinner_ascii") %||%
        getOption("cli.spinner") %||%
        "line"
    }
  }

  if (is.character(which)) {
    row <- match(which, spinners$name)
    which <- list(
      name = which,
      interval = spinners$interval[[row]],
      frames = spinners$frames[[row]])
  }

  if (!is.character(which$frames)) {
    stop("Spinner frames must be a character vector")
  }

  which$name <- which$name %||% NA_character_
  which$interval <- which$interval %||% 100L

  which
}

#' List all available spinners
#'
#' @return Character vector of all available spinner names.
#'
#' @family spinners
#' @export
#' @examples
#' list_spinners()
#' get_spinner(list_spinners()[1])

list_spinners <- function() {
  spinners$name
}

#' Create a spinner
#'
#' @param template A template string, that will contain the spinner. The
#'   spinner itself will be substituted for `{spin}`. See example below.
#' @param stream The stream to use for the spinner. Typically this is
#'   standard error, or maybe the standard output stream.
#'   It can also be a string, one of `"auto"`, `"message"`, `"stdout"`,
#'   `"stderr"`. `"auto"` will select `stdout()` if the session is
#'   interactive and there are no sinks, otherwise it will select
#'   `stderr()`.
#' @param static What to do if the terminal does not support dynamic
#'   displays:
#'   * `"dots"`: show a dot for each `$spin()` call.
#'   * `"print"`: just print the frames of the spinner, one after another.
#'   * `"print_line"`: print the frames of the spinner, each on its own line.
#'   * `"silent"` do not print anything, just the `template`.
#' @inheritParams get_spinner
#' @return A `cli_spinner` object, which is a list of functions. See
#'   its methods below.
#'
#' `cli_spinner` methods:
#' * `$spin()`: output the next frame of the spinner.
#' * `$finish()`: terminate the spinner. Depending on terminal capabilities
#'   this removes the spinner from the screen. Spinners can be reused,
#'   you can start calling the `$spin()` method again.
#'
#' All methods return the spinner object itself, invisibly.
#'
#' The spinner is automatically throttled to its ideal update frequency.
#'
#' @section Examples:
#'
#' ## Default spinner
#'

#' ```{asciicast make-spinner-default}
#' #| asciicast_at = "all",
#' #| asciicast_knitr_output = "svg",
#' #| asciicast_cursor = FALSE,
#' #| asciicast_end_wait = 0
#' sp1 <- make_spinner()
#' fun_with_spinner <- function() {
#'   lapply(1:100, function(x) { sp1$spin(); Sys.sleep(0.05) })
#'   sp1$finish()
#' }
#' ansi_with_hidden_cursor(fun_with_spinner())
#' ```
#'
#' ## Spinner with a template
#'
#' ```{asciicast make-spinner-template}
#' #| asciicast_at = "all",
#' #| asciicast_knitr_output = "svg",
#' #| asciicast_cursor = FALSE,
#' #| asciicast_end_wait = 0
#' sp2 <- make_spinner(template = "Computing {spin}")
#' fun_with_spinner2 <- function() {
#'   lapply(1:100, function(x) { sp2$spin(); Sys.sleep(0.05) })
#'   sp2$finish()
#' }
#' ansi_with_hidden_cursor(fun_with_spinner2())
#' ```
#'
#' ## Custom spinner
#'
#' ```{asciicast make-spinner-custom}
#' #| asciicast_at = "all",
#' #| asciicast_knitr_output = "svg",
#' #| asciicast_cursor = FALSE
#' sp3 <- make_spinner("simpleDotsScrolling", template = "Downloading {spin}")
#' fun_with_spinner3 <- function() {
#'   lapply(1:100, function(x) { sp3$spin(); Sys.sleep(0.05) })
#'   sp3$finish()
#' }
#' ansi_with_hidden_cursor(fun_with_spinner3())
#' ```
#'
#' @family spinners
#' @export

make_spinner <- function(which = NULL, stream = "auto", template = "{spin}",
                         static = c("dots", "print", "print_line",
                                    "silent")) {

  stopifnot(
    inherits(stream, "connection") || is_string(stream),
    is_string(template))

  c_stream <- get_real_output(stream)
  c_spinner <- get_spinner(which)
  c_template <- template
  c_static <- match.arg(static)
  c_state <- 1L
  c_first <- TRUE
  c_col <- 1L
  c_width <- 0L
  c_last <- Sys.time() - as.difftime(1, units = "secs")
  c_int <- as.difftime(c_spinner$interval / 1000, units = "secs")

  c_res <- list()

  throttle <- function() Sys.time() - c_last < c_int
  clear_line <- function() {
    str <- paste0(c("\r", rep(" ", c_width), "\r"), collapse = "")
    cat(str, file = c_stream)
  }
  inc <- function() {
    c_state <<- c_state + 1L
    c_first <<- FALSE
    if (c_state > length(c_spinner$frames)) c_state <<- 1L
    c_last <<- Sys.time()
    invisible(c_res)
  }

  c_res$finish <- function() {
    c_state <<- 1L
    c_first <<- TRUE
    c_col <<- 1L
    c_last <<- Sys.time()
    if (is_dynamic_tty(c_stream)) clear_line() else cat("\n", file = c_stream)
    invisible(c_res)
  }

  if (is_dynamic_tty(c_stream)) {
    c_res$spin <- function(template = NULL) {
      if (!is.null(template)) c_template <<- template
      if (throttle()) return()
      line <- sub("{spin}", c_spinner$frames[[c_state]], c_template,
                  fixed = TRUE)
      line_width <- ansi_nchar(line)
      if (is_ansi_tty(c_stream)) {
        cat("\r", line, ANSI_EL, sep = "", file = c_stream)
      } else {
        # extra padding in case the line width has changed
        # so that we don't get any garbage in the output
        padding <- if (line_width < c_width) {
          paste0(rep(" ", line_width), collapse = "")
        } else {
          ""
        }
        cat("\r", line, padding, sep = "", file = c_stream)
      }
      # save the new line width
      c_width <<- line_width
      inc()
    }

  } else {
    if (c_static == "dots") {
      c_res$spin <- function(template = NULL) {
        if (!is.null(template)) c_template <<- template
        if (c_first) cat(template, "\n", sep = "", file = c_stream)
        if (throttle()) return()
        cat(".", file = c_stream)
        c_col <<- c_col + 1L
        if (c_col == console_width()) {
          cat("\n", file = c_stream)
          c_col <<- 1L
        }
        inc()
      }
    } else if (c_static == "print") {
      c_res$spin <- function(template = NULL) {
        if (!is.null(template)) c_template <<- template
        if (throttle()) return()
        line <- sub("{spin}", c_spinner$frames[[c_state]], c_template,
                    fixed = TRUE)
        cat(line, file = c_stream)
        inc()
      }
    } else if (c_static == "print_line") {
      c_res$spin <- function(template = NULL) {
        if (!is.null(template)) c_template <<- template
        if (throttle()) return()
        line <- sub("{spin}", c_spinner$frames[[c_state]], c_template,
                    fixed = TRUE)
        cat(line, "\n", sep = "", file = c_stream)
        inc()
      }
    } else if (c_static == "silent") {
      c_res$spin <- function(template = NULL) {
        if (!is.null(template)) c_template <<- template
        if (throttle()) return()
        inc()
      }
    }
  }

  class(c_res) <- "cli_spinner"
  c_res
}

#' @export

print.cli_spinner <- function(x, ...) {
  cat("<cli_spinner>\n")
  invisible(x)
}

## nocov start

#' Show a demo of some (by default all) spinners
#'
#' Each spinner is shown for about 2-3 seconds.
#'
#' @details
#'
#' ```{asciicast demo-spinners}
#' #| asciicast_at = "all",
#' #| asciicast_knitr_output = "svg",
#' #| asciicast_cursor = FALSE,
#' #| asciicast_end_wait = 0
#' demo_spinners("clock")
#' ```
#'
#' @param which Character vector, which spinners to demo.
#'
#' @family spinners
#' @export

demo_spinners <- function(which = NULL) {
  stopifnot(is.null(which) || is.character(which))

  all <- list_spinners()
  which <- which %||% all

  if (length(bad <- setdiff(which, all))) {
    stop("Unknown spinners: ", paste(bad, collapse = ", "))
  }

  for (w in which) {
    sp <- get_spinner(w)
    interval <- sp$interval / 1000
    frames <- sp$frames
    cycles <- max(round(2.5 / ((length(frames) - 1) * interval)), 1)
    for (i in 1:(length(frames) * cycles) - 1) {
      fr <- unclass(frames[i %% length(frames) + 1])
      cat("\r", rpad(fr, width = 10), w, sep = "")
      Sys.sleep(interval)
    }
    cat("\n")
  }
}

demo_spinners_terminal <- function(ticks = 100 * 3000) {
  up <- function(n) cat(paste0("\u001B[", n, "A"))
  show <- function() cat("\u001b[?25h")
  hide <- function() cat("\u001b[?25l")

  on.exit(show(), add = TRUE)

  names <- unlist(spinners$name)
  frames <- spinners$frames
  intervals <- unlist(spinners$interval)
  num_frames <- viapply(frames, length)
  spin_width <- viapply(frames, function(x) max(nchar(x, type = "width")))
  name_width <- nchar(names, type = "width")
  col_width <- spin_width + max(name_width) + 3
  col1_width <- max(col_width[1:(length(col_width)/2)])

  frames <- mapply(
    frames,
    names,
    FUN = function(f, n) {
      rpad(paste(lpad(n, max(name_width) + 2), f), col1_width)
    }
  )

  hide()

  for (tick in 0:ticks) {
    tic <- Sys.time()
    wframe <- trunc(tick / intervals) %% num_frames + 1
    sp <- mapply(frames, wframe, FUN = "[")

    sp2 <- paste(
      sep = "  ",
      sp[1:(length(sp) / 2)],
      sp[(length(sp) / 2 + 1):length(sp)]
    )

    cat(sp2, sep = "\n")
    up(length(sp2))
    took <- Sys.time() - tic
    togo <- as.difftime(1/1000, units = "secs") - took
    if (togo > 0) Sys.sleep(togo)
  }

}

## nocov end

test_that("get_spinner", {
  if (is_utf8_output()) {
    expect_equal(get_spinner()$name, "dots")
  } else {
    expect_equal(get_spinner()$name, "line")
  }
})

test_that("list_spinners", {
  ls <- list_spinners()
  expect_true(is.character(ls))
  expect_true("dots" %in% ls)
  expect_true("line" %in% ls)
})

context("spinners")

test_that("get_spinner", {
  if (is_utf8_output()) {
    expect_equal(get_spinner()$name, "dots")
  } else {
    expect_equal(get_spinner()$name, "line")
  }
})

test_that("list_spinners", {
  ls <- list_spinners()
  expect_true(is.character(ls))
  expect_true("dots" %in% ls)
  expect_true("line" %in% ls)
})

#' Compose multiple cli functions
#'
#' `cli()` will record all `cli_*` calls in `expr`, and emit them together
#' in a single message. This is useful if you want to built a larger
#' piece of output from multiple `cli_*` calls.
#'
#' Use this function to build a more complex piece of CLI that would not
#' make sense to show in pieces.
#'
#' ```{asciicast cli-cli}
#' cli({
#'   cli_h1("Title")
#'   cli_h2("Subtitle")
#'   cli_ul(c("this", "that", "end"))
#' })
#' ```
#'
#' @param expr Expression that contains `cli_*` calls. Their output is
#' collected and sent as a single message.
#' @return Nothing.
#'
#' @export

cli <- function(expr) {
  cond <- cli__message_create("meta", cli__rec(expr))
  # cli() might be nested
  record <- getOption("cli.record")
  if (is.null(record)) {
    cli__message_emit(cond)
  } else {
    cli_recorded[[record]] <- c(cli_recorded[[record]], list(cond))
  }
  invisible()
}

cli__rec <- function(expr) {
  id <- new_uuid()
  cli_recorded[[id]] <- list()
  on.exit(rm(list = id, envir = cli_recorded), add = TRUE)
  old <- options(cli.record = id)
  on.exit(options(old), add = TRUE)
  expr
  cli_recorded[[id]]
}

cli__fmt <- function(record, collapse = FALSE, strip_newline = FALSE,
                     app = NULL) {
  app <- app %||% default_app() %||% start_app(.auto_close = FALSE)

  old <- app$output
  oldsig <- app$signal
  on.exit(app$output <- old, add = TRUE)
  on.exit(app$signal <- oldsig, add = TRUE)
  out <- rawConnection(raw(1000), open = "wb")
  on.exit(close(out), add = TRUE)
  app$output <- out
  app$signal <- FALSE

  for (msg in record) {
    do.call(app[[msg$type]], msg$args)
  }

  txt <- rawToChar(rawConnectionValue(out))
  Encoding(txt) <- "UTF-8"
  if (!collapse) {
    txt <- unlist(strsplit(txt, "\n", fixed = TRUE))
  } else if (strip_newline) {
    txt <- substr(txt, 1, nchar(txt) - 1L)
  }
  txt
}

#' Capture the output of cli functions instead of printing it
#'
#' @param expr Expression to evaluate, containing `cli_*()` calls,
#'   typically.
#' @param collapse Whether to collapse the output into a single character
#'   scalar, or return a character vector with one element for each line.
#' @param strip_newline Whether to strip the trailing newline.
#'
#' @export
#' @examples
#' cli_fmt({
#'   cli_alert_info("Loading data file")
#'   cli_alert_success("Loaded data file")
#' })

cli_fmt <- function(expr, collapse = FALSE, strip_newline = FALSE) {
  rec <- cli__rec(expr)
  cli__fmt(rec, collapse, strip_newline)
}

#' Format and returns a line of text
#'
#' You can use this function to format a line of cli text, without emitting
#' it to the screen. It uses [cli_text()] internally.
#'
#' `format_inline()` performs no width-wrapping.
#'
#' @param ... Passed to [cli_text()].
#' @param .envir Environment to evaluate the expressions in.
#' @param collapse Whether to collapse the result if it has multiple
#'   lines, e.g. because of `\f` characters.
#' @param keep_whitespace Whether to keep all whitepace (spaces, newlines
#'   and form feeds) as is in the input.
#' @return Character scalar, the formatted string.
#'
#' @seealso This function supports [inline markup][inline-markup].
#' @family functions supporting inline markup
#' @export
#' @examples
#' format_inline("A message for {.emph later}, thanks {.fn format_inline}.")

format_inline <- function(..., .envir = parent.frame(), collapse = TRUE,
                          keep_whitespace = TRUE) {
  opts <- options(cli.width = Inf)
  on.exit(options(opts), add = TRUE)
  fun <- if (keep_whitespace) cli_inline else cli_text
  cli_fmt(
    fun(..., .envir = .envir),
    collapse = collapse,
    strip_newline = TRUE
  )
}

#' CLI text
#'
#' Write some text to the screen. This function is most appropriate for
#' longer paragraphs. See [cli_alert()] for shorter status messages.
#'
#' @details
#'
#' ## Text wrapping
#'
#' Text is wrapped to the console width, see [console_width()].
#'
#' ```{asciicast cli-text}
#' cli_text(cli:::lorem_ipsum())
#' ```
#'
#' ## New lines
#'
#' A `cli_text()` call always appends a newline character to the end.
#'
#' ```{asciicast cli-text-newline}
#' cli_text("First line.")
#' cli_text("Second line.")
#' ```
#'
#' ## Styling
#'
#' You can use [inline markup][inline-markup], as usual.
#'
#' ```{asciicast cli-text-markup}
#' cli_text("The {.fn cli_text} function in the {.pkg cli} package.")
#' ```
#'
#' ## Interpolation
#'
#' String interpolation via glue works as usual. Interpolated vectors
#' are collapsed.
#'
#' ```{asciicast cli-text-glue}
#' pos <- c(5, 14, 25, 26)
#' cli_text("We have {length(pos)} missing measurements: {pos}.")
#' ```
#'
#' ## Styling and interpolation
#'
#' Use double braces to combine styling and string interpolation.
#'
#' ```{asciicast cli-text-glue-style}
#' fun <- "cli-text"
#' pkg <- "cli"
#' cli_text("The {.fn {fun}} function in the {.pkg {pkg}} package.")
#' ```
#'
#' ## Multiple arguments
#'
#' Arguments are concatenated.
#'
#' ```{asciicast cli-text-concat}
#' cli_text(c("This ", "will ", "all "), "be ", "one ", "sentence.")
#' ```
#'
#' ## Containers
#'
#' You can use `cli_text()` within cli [containers].
#'
#' ```{asciicast cli-text-containers}
#' ul <- cli_ul()
#' cli_li("First item.")
#' cli_text("Still the {.emph first} item")
#' cli_li("Second item.")
#' cli_text("Still the {.emph second} item")
#' cli_end(ul)
#' ```
#'
#' @param ... The text to show, in character vectors. They will be
#'   concatenated into a single string. Newlines are _not_ preserved.
#' @param .envir Environment to evaluate the glue expressions in.
#'
#' @seealso This function supports [inline markup][inline-markup].
#' @family functions supporting inline markup
#' @export

cli_text <- function(..., .envir = parent.frame()) {
  cli__message("text", list(text = glue_cmd(..., .envir = .envir, .call = sys.call())))
}

cli_inline <- function(..., .envir = parent.frame()) {
  cli__message(
    "inline_text",
    list(
      text = glue_cmd(..., .envir = .envir, .call = sys.call(), .trim = FALSE)
    )
  )
}

#' CLI verbatim text
#'
#' It is not wrapped, but printed as is. Long lines will overflow.
#' No glue substitution is performed on verbatim text.
#'
#' @details
#'
#' ## Line breaks
#'
#' ```{asciicast cli-verbatim}
#' cli_verbatim("This has\nthree\nlines,")
#' ```
#'
#' ## Special characters
#'
#' No glue substitution happens here.
#'
#' ```{asciicast cli-verbatim-2}
#' cli_verbatim("No string {interpolation} or {.emph styling} here")
#' ```
#'
#' @param ... The text to show, in character vectors. Each element is
#'   printed on a new line.
#' @param .envir Environment to evaluate the glue expressions in.
#'
#' @seealso [cli_code()] for printing R or other source code.
#' @export

cli_verbatim <- function(..., .envir = parent.frame()) {
  cli__message("verbatim", c(list(...), list(.envir = .envir)))
}

#' CLI headings
#'
#' cli has three levels of headings.
#'
#' @details
#'
#' This is how the headings look with the default builtin theme.
#'
#' ```{asciicast, cli-h1}
#' cli_h1("Header {.emph 1}")
#' cli_h2("Header {.emph 2}")
#' cli_h3("Header {.emph 3}")
#' ```
#'
#' @param text Text of the heading. It can contain inline markup.
#' @param id Id of the heading element, string. It can be used in themes.
#' @param class Class of the heading element, string. It can be used in
#'   themes.
#' @param .envir Environment to evaluate the glue expressions in.
#'
#' @seealso These functions supports [inline markup][inline-markup].
#' @family functions supporting inline markup
#' @export

cli_h1 <- function(text, id = NULL, class = NULL, .envir = parent.frame()) {
  cli__message(
    "h1",
    list(
      text = glue_cmd(text, .envir = .envir, .call = sys.call()),
      id = id,
      class = class
    )
  )
}

#' @rdname cli_h1
#' @export

cli_h2 <- function(text, id = NULL, class = NULL, .envir = parent.frame()) {
  cli__message(
    "h2",
    list(
      text = glue_cmd(text, .envir = .envir, .call = sys.call()),
      id = id,
      class = class
    )
  )
}

#' @rdname cli_h1
#' @export

cli_h3 <- function(text, id = NULL, class = NULL, .envir = parent.frame()) {
  cli__message(
    "h3",
    list(
      text = glue_cmd(text, .envir = .envir, .call = sys.call()),
      id = id,
      class = class
    )
  )
}

#' Generic CLI container
#'
#' See [containers]. A `cli_div` container is special, because it may
#' add new themes, that are valid within the container.
#'
#' @details
#'
#' ## Custom themes
#'
#' ```{asciicast cli-div}
#' d <- cli_div(theme = list(h1 = list(color = "cyan",
#'                                     "font-weight" = "bold")))
#' cli_h1("Custom title")
#' cli_end(d)
#' ```
#'
#' ## Auto-closing
#'
#' By default a `cli_div()` is closed automatically when the calling
#' frame exits.
#'
#' ```{asciicast cli-div-close}
#' div <- function() {
#'   cli_div(class = "tmp", theme = list(.tmp = list(color = "yellow")))
#'   cli_text("This is yellow")
#' }
#' div()
#' cli_text("This is not yellow any more")
#' ```
#'
#' @param id Element id, a string. If `NULL`, then a new id is generated
#'   and returned.
#' @param class Class name, sting. Can be used in themes.
#' @param theme A custom theme for the container. See [themes].
#' @param .auto_close Whether to close the container, when the calling
#'   function finishes (or `.envir` is removed, if specified).
#' @param .envir Environment to evaluate the glue expressions in. It is
#'   also used to auto-close the container if `.auto_close` is `TRUE`.
#' @return The id of the new container element, invisibly.
#'
#' @export

cli_div <- function(id = NULL, class = NULL, theme = NULL,
                    .auto_close = TRUE, .envir = parent.frame()) {
  cli__message("div", list(id = id, class = class, theme = theme),
               .auto_close = .auto_close, .envir = .envir)
}

#' CLI paragraph
#'
#' The builtin theme leaves an empty line between paragraphs.
#' See also [containers].
#'
#' ```{asciicast cli-par}
#' clifun <- function() {
#'   cli_par()
#'   cli_text(cli:::lorem_ipsum())
#' }
#' clifun()
#' clifun()
#' ```
#'
#' @param id Element id, a string. If `NULL`, then a new id is generated
#'   and returned.
#' @param class Class name, sting. Can be used in themes.
#' @inheritParams cli_div
#' @return The id of the new container element, invisibly.
#'
#' @export

cli_par <- function(id = NULL, class = NULL, .auto_close = TRUE,
                    .envir = parent.frame()) {
  cli__message("par", list(id = id, class = class),
               .auto_close = .auto_close, .envir = .envir)
}

#' Close a CLI container
#'
#' Containers aut0-close by default, but sometimes you need to explicitly
#' close them. Closing a container also closes all of its nested
#' containers.
#'
#' @details
#'
#' ## Explicit closing
#'
#' ```{asciicast cli-end}
#' cnt <- cli_par()
#' cli_text("First paragraph.")
#' cli_end(cnt)
#' cnt <- cli_par()
#' cli_text("Second paragraph.")
#' cli_end(cnt)
#' ```
#'
#' ## Closing a stack of containers
#'
#' ```{asciicast cli-end-many}
#' list <- cli_ul()
#' cli_li("Item one:")
#' cli_li("Item two:")
#' cli_par()
#' cli_text("Still item two.")
#' cli_end(list)
#' cli_text("Not in the list any more")
#' ```
#'
#' ## Omitting `id`
#'
#' If `id` is omitted, the container that was opened last will be closed.
#'
#' ```{asciicast cli-end-noid}
#' cli_par()
#' cli_text("First paragraph")
#' cli_end()
#' cli_par()
#' cli_text("Second paragraph")
#' cli_end()
#' ```
#'
#' ## Debugging containers
#'
#' You can use the internal `cli:::cli_debug_doc()` function to see the
#' currently open containers.
#'
#' ```{asciicast cli-end-debug}
#' fun <- function() {
#'   cli_div(id = "mydiv")
#'   cli_par(class = "myclass")
#'   cli:::cli_debug_doc()
#' }
#' fun()
#' ```
#'
#' @param id Id of the container to close. If missing, the current
#' container is closed, if any.
#'
#' @export

cli_end <- function(id = NULL) {
  cli__message("end", list(id = id %||% NA_character_))
}

#' Unordered CLI list
#'
#' An unordered list is a container, see [containers].
#'
#' @details
#'
#' ## Adding all items at once
#'
#' ```{asciicast cli-ul}
#' fun <- function() {
#'   cli_ul(c("one", "two", "three"))
#' }
#' fun()
#' ```
#'
#' ## Adding items one by one
#'
#' ```{asciicast cli-ul-2}
#' fun <- function() {
#'   cli_ul()
#'   cli_li("{.emph one}")
#'   cli_li("{.emph two}")
#'   cli_li("{.emph three}")
#'   cli_end()
#' }
#' fun()
#' ```
#'
#' @param items If not `NULL`, then a character vector. Each element of
#'   the vector will be one list item, and the list container will be
#'   closed by default (see the `.close` argument).
#' @param id Id of the list container. Can be used for closing it with
#'   [cli_end()] or in themes. If `NULL`, then an id is generated and
#'   returned invisibly.
#' @param class Class of the list container. Can be used in themes.
#' @param .close Whether to close the list container if the `items` were
#'   specified. If `FALSE` then new items can be added to the list.
#' @inheritParams cli_div
#' @return The id of the new container element, invisibly.
#'
#' @seealso This function supports [inline markup][inline-markup].
#' @family functions supporting inline markup
#' @export


cli_ul <- function(items = NULL, id = NULL, class = NULL,
                   .close = TRUE, .auto_close = TRUE,
                   .envir = parent.frame()) {
  cli__message(
    "ul",
    list(
      items = lapply(items, glue_cmd, .envir = .envir, .call = sys.call()),
      id = id,
      class = class,
      .close = .close
    ),
    .auto_close = .auto_close, .envir = .envir
  )
}

#' Ordered CLI list
#'
#' An ordered list is a container, see [containers].
#'
#' @details
#'
#' ## Adding all items at once
#'
#' ```{asciicast cli-ol}
#' fun <- function() {
#'   cli_ol(c("one", "two", "three"))
#' }
#' fun()
#' ```
#'
#' ## Adding items one by one
#'
#' ```{asciicast cli-ol-2}
#' ## Adding items one by one
#' fun <- function() {
#'   cli_ol()
#'   cli_li("{.emph one}")
#'   cli_li("{.emph two}")
#'   cli_li("{.emph three}")
#'   cli_end()
#' }
#' fun()
#' ```
#'
#' ## Nested lists
#'
#' ```{asciicast cli-ol-3}
#' fun <- function() {
#'   cli_div(theme = list(ol = list("margin-left" = 2)))
#'   cli_ul()
#'   cli_li("one")
#'   cli_ol(c("foo", "bar", "foobar"))
#'   cli_li("two")
#'   cli_end()
#'   cli_end()
#' }
#' fun()
#' ```
#'
#' @inheritParams cli_ul
#' @return The id of the new container element, invisibly.
#'
#' @seealso This function supports [inline markup][inline-markup].
#' @family functions supporting inline markup
#' @export

cli_ol <- function(items = NULL, id = NULL, class = NULL,
                   .close = TRUE, .auto_close = TRUE,
                   .envir = parent.frame()) {
  cli__message(
    "ol",
    list(
      items = lapply(items, glue_cmd, .envir = .envir, .call = sys.call()),
      id = id,
      class = class,
      .close = .close
    ),
    .auto_close = .auto_close, .envir = .envir
  )
}

#' Definition list
#'
#' A definition list is a container, see [containers].
#'
#' @details
#'
#' ## All items at once
#'
#' ```{asciicast cli-dl}
#' fun <- function() {
#'   cli_dl(c(foo = "one", bar = "two", baz = "three"))
#' }
#' fun()
#' ```
#'
#' ## Items one by one
#'
#' ```{asciicast cli-dl-2}
#' fun <- function() {
#'   cli_dl()
#'   cli_li(c(foo = "{.emph one}"))
#'   cli_li(c(bar = "two"))
#'   cli_li(c(baz = "three"))
#' }
#' fun()
#' ```
#'
#' @param items Named character vector, or `NULL`. If not `NULL`, they
#'   are used as list items.
#' @param labels Item labels. Defaults the names in `items`.
#' @inheritParams cli_ul
#' @return The id of the new container element, invisibly.
#'
#' @seealso This function supports [inline markup][inline-markup].
#' @family functions supporting inline markup
#' @export

cli_dl <- function(items = NULL, labels = names(items), id = NULL,
                   class = NULL, .close = TRUE, .auto_close = TRUE,
                   .envir = parent.frame()) {
  if (!is.null(items) && !is_named(items)) {
    throw(cli_error(
      "{.arg items} must be a named character vector",
      "i" = if (!is_named(items)) "{.arg items} is not named"
    ))
  }

  cli__message(
    "dl",
    list(
      items = lapply(items, glue_cmd, .envir = .envir, .call = sys.call()),
      labels = if (!is.null(labels)) {
                 lapply(labels, glue_cmd, .envir = .envir, .call = sys.call())
               },
      id = id,
      class = class, .close = .close),
    .auto_close = .auto_close, .envir = .envir)
}

#' CLI list item(s)
#'
#' A list item is a container, see [containers].
#'
#' @details
#'
#' ## Nested lists
#'
#' ```{asciicast cli-li}
#' fun <- function() {
#'   ul <- cli_ul()
#'   cli_li("one:")
#'   cli_ol(letters[1:3])
#'   cli_li("two:")
#'   cli_li("three")
#'   cli_end(ul)
#' }
#' fun()
#' ```
#'
#' @param items Character vector of items, or `NULL`.
#' @param labels For definition lists the item labels.
#' @param id Id of the new container. Can be used for closing it with
#'   [cli_end()] or in themes. If `NULL`, then an id is generated and
#'   returned invisibly.
#' @param class Class of the item container. Can be used in themes.
#' @inheritParams cli_div
#' @return The id of the new container element, invisibly.
#'
#' @seealso This function supports [inline markup][inline-markup].
#' @family functions supporting inline markup
#' @export

cli_li <- function(items = NULL, labels = names(items), id = NULL,
                   class = NULL, .auto_close = TRUE,
                   .envir = parent.frame()) {
  cli__message(
    "li",
    list(
      items = lapply(items, glue_cmd, .envir = .envir, .call = sys.call()),
      labels = if (!is.null(labels)) {
                 lapply(labels, glue_cmd, .envir = .envir, .call = sys.call())
               },
      id = id,
      class = class),
    .auto_close = .auto_close, .envir = .envir)
}

#' CLI alerts
#'
#' Alerts are typically short status messages.
#'
#' @details
#'
#' ## Success
#'
#' ```{asciicast alert-success}
#' nbld <- 11
#' tbld <- prettyunits::pretty_sec(5.6)
#' cli_alert_success("Built {.emph {nbld}} status report{?s} in {tbld}.")
#' ```
#'
#' ## Info
#'
#' ```{asciicast alert-info}
#' cfl <- "~/.cache/files/latest.cache"
#' cli_alert_info("Updating cache file {.path {cfl}}.")
#' ```
#'
#' ## Warning
#'
#' ```{asciicast alert-warning}
#' cfl <- "~/.cache/files/latest.cache"
#' cli_alert_warning("Failed to update cache file {.path {cfl}}.")
#' ```
#'
#' ## Danger
#'
#' ```{asciicast alert-danger}
#' cfl <- "~/.config/report.yaml"
#' cli_alert_danger("Cannot validate config file at {.path {cfl}}.")
#' ```
#'
#' ## Text wrapping
#'
#' Alerts are printed without wrapping, unless you set `wrap = TRUE`:
#'
#' ```{asciicast alert-wrap, asciicast_rows = 4}
#' cli_alert_info("Data columns: {.val {names(mtcars)}}.")
#' cli_alert_info("Data columns: {.val {names(mtcars)}}.", wrap = TRUE)
#' ```
#'
#' @param text Text of the alert.
#' @param id Id of the alert element. Can be used in themes.
#' @param class Class of the alert element. Can be used in themes.
#' @param wrap Whether to auto-wrap the text of the alert.
#' @param .envir Environment to evaluate the glue expressions in.
#'
#' @seealso These functions supports [inline markup][inline-markup].
#' @family functions supporting inline markup
#' @export

cli_alert <- function(text, id = NULL, class = NULL, wrap = FALSE,
                      .envir = parent.frame()) {
  cli__message(
    "alert",
    list(
      text = glue_cmd(text, .envir = .envir, .call = sys.call()),
      id = id,
      class = class,
      wrap = wrap
    )
  )
}

#' @rdname cli_alert
#' @export

cli_alert_success <- function(text, id = NULL, class = NULL, wrap = FALSE,
                              .envir = parent.frame()) {
  cli__message(
    "alert_success",
    list(
      text = glue_cmd(text, .envir = .envir, .call = sys.call()),
      id = id,
      class = class,
      wrap = wrap
    )
  )
}

#' @rdname cli_alert
#' @export

cli_alert_danger <- function(text, id = NULL, class = NULL, wrap = FALSE,
                              .envir = parent.frame()) {
  cli__message(
    "alert_danger",
    list(
      text = glue_cmd(text, .envir = .envir, .call = sys.call()),
      id = id,
      class = class,
      wrap = wrap
    )
  )
}

#' @rdname cli_alert
#' @export

cli_alert_warning <- function(text, id = NULL, class = NULL, wrap = FALSE,
                               .envir = parent.frame()) {
  cli__message(
    "alert_warning",
    list(
      text = glue_cmd(text, .envir = .envir, .call = sys.call()),
      id = id,
      class = class,
      wrap = wrap
    )
  )
}

#' @rdname cli_alert
#' @export

cli_alert_info <- function(text, id = NULL, class = NULL, wrap = FALSE,
                            .envir = parent.frame()) {
  cli__message(
    "alert_info",
    list(
      text = glue_cmd(text, .envir = .envir, .call = sys.call()),
      id = id,
      class = class,
      wrap = wrap
    )
  )
}

#' CLI horizontal rule
#'
#' It can be used to separate parts of the output.
#'
#' @details
#'
#' ## Inline styling and interpolation
#'
#' ```{asciicast cli-rule}
#' pkg <- "mypackage"
#' cli_rule(left = "{.pkg {pkg}} results")
#' ```
#'
#' ## Theming
#'
#' The line style of the rule can be changed via the the `line-type`
#' property. Possible values are:
#'
#' * `"single"`: (same as `1`), a single line,
#' * `"double"`: (same as `2`), a double line,
#' * `"bar1"`, `"bar2"`, `"bar3"`, etc., `"bar8"` uses varying height bars.
#'
#' Colors and background colors can similarly changed via a theme.
#'
#' ```{asciicast cli-rule-line-type}
#' d <- cli_div(theme = list(rule = list(
#'   color = "cyan",
#'   "line-type" = "double")))
#' cli_rule("Summary", right = "{.pkg mypackage}")
#' cli_end(d)
#' ```
#'
#' @param .envir Environment to evaluate the glue expressions in.
#' @inheritParams rule
#' @inheritParams cli_div
#'
#' @seealso This function supports [inline markup][inline-markup].
#' @family functions supporting inline markup
#' @export

cli_rule <- function(left = "", center = "", right = "", id = NULL,
                     .envir = parent.frame()) {
  cli__message("rule", list(left = glue_cmd(left, .envir = .envir, .call = sys.call()),
                            center = glue_cmd(center, .envir = .envir, .call = sys.call()),
                            right = glue_cmd(right, .envir = .envir, .call = sys.call()),
                            id = id))
}

#' CLI block quote
#'
#' A section that is quoted from another source. It is typically indented.
#'
#' @details
#'
#' ```{asciicast cli-blockquote}
#' evil <- paste(
#'   "The real problem is that programmers have spent far too much time",
#'   "worrying about efficiency in the wrong places and at the wrong",
#'   "times; premature optimization is the root of all evil (or at least",
#'   "most of it) in programming.")
#' cli_blockquote(evil, citation = "Donald Ervin Knuth")
#' ```
#'
#' @param quote Text of the quotation.
#' @param citation Source of the quotation, typically a link or the name
#'   of a person.
#' @inheritParams cli_div
#'
#' @seealso This function supports [inline markup][inline-markup].
#' @family functions supporting inline markup
#' @export

cli_blockquote <- function(quote, citation = NULL, id = NULL,
                           class = NULL, .envir = parent.frame()) {
  cli__message(
    "blockquote",
    list(
      quote = glue_cmd(quote, .envir = .envir, .call = sys.call()),
      citation = glue_cmd(citation, .envir = .envir, .call = sys.call()),
      id = id,
      class = class
    )
  )
}

#' A block of code
#'
#' A helper function that creates a `div` with class `code` and then calls
#' `cli_verbatim()` to output code lines. The builtin theme formats these
#' containers specially. In particular, it adds syntax highlighting to
#' valid R code.
#'
#' @details
#'
#' ```{asciicast cli-code}
#' myfun <- function() {
#'   message("Just an example function")
#'   graphics::pairs(iris, col = 1:4)
#' }
#' cli_code(format(myfun))
#' ```
#'
#' @param lines Character vector, each line will be a line of code, and
#'   newline characters also create new lines. Note that _no_ glue
#'   substitution is performed on the code.
#' @param ... More character vectors, they are appended to `lines`.
#' @param language Programming language. This is also added as a class,
#'   in addition to `code`.
#' @param .auto_close Passed to `cli_div()` when creating the container of
#'   the code. By default the code container is closed after emitting
#'   `lines` and `...` via `cli_verbatim()`. You can keep that container
#'   open with `.auto_close` and/or `.envir`, and then calling
#'   `cli_verbatim()` to add (more) code. Note that the code will be
#'   formatted and syntax highlighted separately for each `cli_verbatim()`
#'   call.
#' @param .envir Passed to `cli_div()` when creating the container of the
#'   code.
#' @return The id of the container that contains the code.
#'
#' @export

cli_code <- function(lines = NULL, ..., language = "R",
                     .auto_close = TRUE, .envir = environment()) {
  lines <- c(lines, unlist(list(...)))
  id <- cli_div(
    class = paste("code", language),
    .auto_close = .auto_close,
    .envir = .envir
  )
  cli_verbatim(lines)
  invisible(id)
}

cli_recorded <- new.env(parent = emptyenv())

cli__message <- function(type, args, .auto_close = TRUE, .envir = NULL,
                         record = getOption("cli.record")) {

  if ("id" %in% names(args) && is.null(args$id)) args$id <- new_uuid()

  if (.auto_close && !is.null(.envir) && !identical(.envir, .GlobalEnv)) {
    if (type == "status") {
      defer(cli_status_clear(id = args$id, result = args$auto_result),
            envir = .envir, priority = "first")
    } else {
      defer(cli_end(id = args$id), envir = .envir, priority = "first")
    }
  }

  cond <- cli__message_create(type, args)

  if (is.null(record)) {
    cli__message_emit(cond)
  } else {
    cli_recorded[[record]] <- c(cli_recorded[[record]], list(cond))
  }

  invisible(args$id)
}

cli__message_create <- function(type, args) {
  cond <- list(message = paste("cli message", type),
               type = type, args = args, pid = clienv$pid)

  class(cond) <- c(
    getOption("cli.message_class"),
    "cli_message",
    "condition"
  )

  cond
}

cli__message_emit <- function(cond) {
  withRestarts(
  {
    signalCondition(cond)
    cli__default_handler(cond)
  },
  cli_message_handled = function() NULL)
}

cli__default_handler <- function(msg) {
  custom_handler <- getOption("cli.default_handler")

  if (is.function(custom_handler)) {
    custom_handler(msg)
  } else {
    cli_server_default(msg)
  }
}

start_app()
on.exit(stop_app(), add = TRUE)

test_that_cli(configs = c("plain", "unicode"), "ul", {
  expect_snapshot(local({
    lid <- cli_ul()
    cli_li("foo")
    cli_li(c("bar", "foobar"))
    cli_end(lid)
  }))
})

test_that_cli(configs = c("plain", "unicode"), "ol", {
  expect_snapshot(local({
    cli_div(theme = list(ol = list()))
    lid <- cli_ol()
    cli_li("foo")
    cli_li(c("bar", "foobar"))
    cli_end(lid)
  }))
})

test_that_cli(configs = c("plain", "unicode"), "ul ul", {
  expect_snapshot(local({
    cli_div(
      theme = list(
        "ul ul" = list("list-style-type" = "-", "margin-left" = 2)
      )
    )
    lid <- cli_ul()
    cli_li("1")
    lid2 <- cli_ul()
    cli_li("1 1")
    cli_li(c("1 2", "1 3"))
    cli_end(lid2)
    cli_li("2")
    cli_end(lid)
  }))
})

test_that_cli(configs = c("plain", "unicode"), "ul ol", {
  expect_snapshot(local({
    cli_div(
      theme = list(
        li = list("margin-left" = 2)
      )
    )
    lid <- cli_ul()
    cli_li("1")
    lid2 <- cli_ol()
    cli_li("1 1")
    cli_li(c("1 2", "1 3"))
    cli_end(lid2)
    cli_li("2")
    cli_end(lid)
  }))
})

test_that_cli(configs = c("plain", "unicode"), "ol ol", {
  expect_snapshot(local({
    cli_div(
      theme = list(
        "li" = list("margin-left" = 2),
        "li li" = list("margin-left" = 2)
      )
    )
    lid <- cli_ol()
    cli_li("1")
    lid2 <- cli_ol()
    cli_li("1 1")
    cli_li(c("1 2", "1 3"))
    cli_end(lid2)
    cli_li("2")
    cli_end(lid)
  }))
})

test_that_cli(configs = c("plain", "unicode"), "ol ul", {
  expect_snapshot(local({
    cli_div(
      theme = list(
        ul = list("margin-left" = 2)
      )
    )
    lid <- cli_ol()
    cli_li("1")
    lid2 <- cli_ul()
    cli_li("1 1")
    cli_li(c("1 2", "1 3"))
    cli_end(lid2)
    cli_li("2")
    cli_end(lid)
  }))
})

test_that_cli(configs = c("plain", "unicode"), "starting with an item", {
  expect_snapshot(local({
    cli_li("foo")
    cli_li(c("bar", "foobar"))
    cli_end()
    cli_end()
  }))
})

test_that_cli(configs = c("plain", "unicode"), "ol, with first item", {
  expect_snapshot(local({
    cli_div(theme = list(ol = list()))
    lid <- cli_ol("foo", .close = FALSE)
    cli_li(c("bar", "foobar"))
    cli_end(lid)
  }))
})

test_that_cli(configs = c("plain", "unicode"), "ul, with first item", {
  expect_snapshot(local({
    lid <- cli_ul("foo", .close = FALSE)
    cli_li(c("bar", "foobar"))
    cli_end(lid)
  }))
})

test_that_cli(configs = c("plain", "unicode"), "dl", {
  expect_snapshot(local({
    cli_div(theme = list(ul = list()))
    lid <- cli_dl()
    cli_li(c(this = "foo"))
    cli_li(c(that = "bar", other = "foobar"))
    cli_end(lid)
  }))
})

test_that_cli(configs = c("plain", "unicode"), "dl dl", {
  expect_snapshot(local({
    cli_div(
      theme = list(
        li = list("margin-left" = 2)
      )
    )
    lid <- cli_dl()
    cli_li(c(a = "1"))
    lid2 <- cli_dl()
    cli_li(c("a-a" = "1 1"))
    cli_li(c("a-b" = "1 2", "a-c" = "1 3"))
    cli_end(lid2)
    cli_li(c(b = "2"))
    cli_end(lid)
  }))
})

test_that_cli(configs = c("plain", "unicode"), "dl ol", {
  expect_snapshot(local({
    cli_div(
      theme = list(
        li = list("margin-left" = 2)
      )
    )
    lid <- cli_dl()
    cli_li(c(a = "1"))
    lid2 <- cli_ol()
    cli_li(c("1 1"))
    cli_li(c("1 2", "1 3"))
    cli_end(lid2)
    cli_li(c(b = "2"))
    cli_end(lid)
  }))
})

test_that_cli(configs = c("plain", "unicode"), "dl ul", {
  expect_snapshot(local({
    cli_div(
      theme = list(
        li = list("margin-left" = 2)
      )
    )
    lid <- cli_dl()
    cli_li(c(a = "1"))
    lid2 <- cli_ul()
    cli_li(c("1 1"))
    cli_li(c("1 2", "1 3"))
    cli_end(lid2)
    cli_li(c(b = "2"))
    cli_end(lid)
  }))
})

test_that_cli(configs = c("plain", "unicode"), "ol dl", {
  expect_snapshot(local({
    cli_div(
      theme = list(
        li = list("margin-left" = 2)
      )
    )
    lid <- cli_ol()
    cli_li("1")
    lid2 <- cli_dl()
    cli_li(c("a-a" = "1 1"))
    cli_li(c("a-b" = "1 2", "a-c" = "1 3"))
    cli_end(lid2)
    cli_li("2")
    cli_end(lid)
  }))
})

test_that_cli(configs = c("plain", "unicode"), "ul dl", {
  expect_snapshot(local({
    cli_div(
      theme = list(
        li = list("margin-left" = 2)
      )
    )
    lid <- cli_ul()
    cli_li("1")
    lid2 <- cli_dl()
    cli_li(c("a-a" = "1 1"))
    cli_li(c("a-b" = "1 2", "a-c" = "1 3"))
    cli_end(lid2)
    cli_li("2")
    cli_end(lid)
  }))
})

test_that_cli(configs = c("plain", "unicode"), "dl, with first item", {
  expect_snapshot(local({
    cli_div(theme = list(ul = list()))
    lid <- cli_dl(c(this = "foo"), .close = FALSE)
    cli_li(c(that = "bar", other = "foobar"))
    cli_end(lid)
  }))
})

test_that_cli(configs = "ansi", "styling pieces of a dl", {
  expect_snapshot(local({
    cli_div(
      theme = list(
        .dt = list(postfix = " -> "),
        .dd = list(color = "blue")
      )
    )
    cli_dl(c(foo = "bar", bar = "baz"))
  }))
})

test_that("cli_dl edge cases", {
  # invalid input
  expect_snapshot(
    error = TRUE,
    cli_dl("foo", "must be a named character vector")
  )

  # empty dd
  expect_snapshot(
    cli_dl(c(abc = "foo", empty = "", def = "bar"))
  )
})

test_that("cli_dl label style", {
  expect_snapshot(local({
    cli_dl(c(
      "{.code code}" = "{.code this is code too}",
      "{.str strin}" = "{.url https://x.com}"
    ))
  }))
})

start_app()
on.exit(stop_app(), add = TRUE)

test_that_cli(
  configs = c("plain", "ansi"),
  "quoting phrases that don't start or end with letter or number", {

    expect_snapshot(local({
      x0 <- "good-name"
      cli_text("The name is {.file {x0}}.")

      x <- "weird-name "
      cli_text("The name is {.file {x}}.")
      cli_text("The name is {.path {x}}.")
      cli_text("The name is {.email {x}}.")
    }))
  }
)

test_that_cli(configs = c("plain", "ansi"), "quoting weird names, still", {
  nb <- function(x) gsub("\u00a0", " ", x, fixed = TRUE)
  expect_snapshot(local({
    cat_line(nb(quote_weird_name("good")))
    cat_line(nb(quote_weird_name("  bad")))
    cat_line(nb(quote_weird_name("bad  ")))
    cat_line(nb(quote_weird_name("  bad  ")))
  }))
})

test_that_cli(configs = c("ansi"), "~/ files are not weird", {
  nb <- function(x) gsub("\u00a0", " ", x, fixed = TRUE)
  expect_snapshot(local({
    cat_line(nb(quote_weird_name("~/good")))
    cat_line(nb(quote_weird_name("~~bad")))
    cat_line(nb(quote_weird_name("bad~  ")))
    cat_line(nb(quote_weird_name(" ~ bad ~ ")))
  }))
})

test_that_cli("custom truncation", {
  expect_snapshot({
    x <- cli_vec(1:100, list("vec-trunc" = 5))
    cli_text("Some numbers: {x}.")
    cli_text("Some numbers: {.val {x}}.")
  })
})

test_that_cli(configs = c("plain", "ansi"), "collapsing class names", {
  expect_snapshot(local({
    cc <- c("one", "two")
    cli_text("this is a class: {.cls myclass}")
    cli_text("multiple classes: {.cls {cc}}")
  }))
})

test_that_cli(configs = c("plain", "ansi"), "transform", {
  expect_snapshot(local({
    cli_text("This is a {.field field} (before)")
    foo <- function(x) toupper(x)
    cli_div(theme = list(span.field = list(transform = foo)))
    cli_text("This is a {.field field} (during)")
    cli_end()
    cli_text("This is a {.field field} (after)")
  }))
})

test_that("cli_format", {
  expect_snapshot(
    cli_format(1:4/7, list(digits = 2))
  )
})

test_that("cli_format() is used for .val", {
  withr::local_options(cli.width = 60)
  withr::local_rng_version("3.3.0")
  set.seed(42)
  expect_snapshot(local({
    cli_div(theme = list(.val = list(digits = 2)))
    cli_text("Some random numbers: {.val {runif(4)}}.")
  }))
})

test_that(".q always double quotes", {
  expect_snapshot(
    cli_text("just a {.q string}, nothing more")
  )
})

test_that(".or", {
  expect_snapshot(
    cli_text("{.or {letters[1:5]}}")
  )
  expect_snapshot(
    cli_text("{.or {letters[1:2]}}")
  )
})

test_that("line breaks", {
  txt <- paste(
    "Cupidatat deserunt culpa enim deserunt minim aliqua tempor fugiat",
    "cupidatat laboris officia esse ex aliqua. Ullamco mollit adipisicing",
    "anim."
  )
  txt2 <- paste0(txt, "\f", txt)
  expect_snapshot(ansi_strwrap(txt2, width = 60))
})

test_that_cli(configs = "ansi", "double ticks", {
  x <- c("a", "`x`", "b")
  cli_div(theme = list(
    .code = list(color = "red"),
    .fun = list(color = "red")
  ))
  expect_snapshot(format_inline("{.code {x}}"))
  expect_snapshot(format_inline("{.fun {x}}"))
})

test_that("do not inherit 'transform' issue #422", {
  expect_snapshot({
    d <- deparse(c("cli", "glue"))
    cli::cli_alert_info("To install, run {.code install.packages({d})}")
  })

  expect_snapshot({
    cli::cli_text("{.code foo({1+1})}")
  })
})

test_that_cli(configs = c("ansi", "plain"), "no inherit color, issue #474", {
  expect_snapshot({
    cli::cli_text("pre {.val x {'foo'} y} post")
  })
})

test_that_cli(configs = c("ansi", "plain"), "\\f at the end, issue #491", {
  expect_snapshot({
    cli_fmt(cli::cli_text("{.val a}{.val b}"))
    cli_fmt(cli::cli_text("\f{.val a}{.val b}"))
    cli_fmt(cli::cli_text("\f\f{.val a}{.val b}"))
    cli_fmt(cli::cli_text("{.val a}\f{.val b}"))
    cli_fmt(cli::cli_text("{.val a}\f\f{.val b}"))
    cli_fmt(cli::cli_text("{.val a}{.val b}\f"))
    cli_fmt(cli::cli_text("{.val a}{.val b}\f\f"))
    cli_fmt(cli::cli_text("\f\f\f{.val a}\f\f\f{.val b}\f\f\f"))
  })
})

test_that("truncate vectors at 20", {
  expect_snapshot(
    cli::cli_text("Some letters: {letters}")
  )
})

test_that_cli(configs = "ansi", "brace expresssion edge cases", {
  foo <- "foo"
  bar <- "bar"
  expect_snapshot({
    cli_text("{.code {foo} and {bar}}")
    cli_text("{.emph {foo} and {bar}}")
    cli_text("{.q {foo} and {bar}}")
  })
})

test_that("various errors", {
  expect_snapshot_error(
    cli_text("xx {.foobar} yy")
  )
  expect_snapshot_error(
    cli_text("xx {.someverylong+expression} yy")
  )
  expect_snapshot(
    error = TRUE,
    cli_text("xx {__cannot-parse-this__} yy"),
    transform = sanitize_srcref,
    variant = if (getRversion() < "4.2.0") "old-r" else "new-r"
  )
  expect_snapshot(
    error = TRUE,
    cli_text("xx {1 + 'a'} yy"),
    transform = function(x) sanitize_call(sanitize_srcref(x))
  )
})

test_that("format_inline and newlines", {
  expect_snapshot({
    format_inline("foo\nbar")
    format_inline("\nfoo\n\nbar\n")
    format_inline("foo\fbar")
    format_inline("\ffoo\f\fbar\f")
  })

  expect_snapshot({
    format_inline("foo\nbar", keep_whitespace = FALSE)
    format_inline("\nfoo\n\nbar\n", keep_whitespace = FALSE)
    format_inline("foo\fbar", keep_whitespace = FALSE)
    format_inline("\ffoo\f\fbar\f", keep_whitespace = FALSE)
  })
})

start_app()
on.exit(stop_app(), add = TRUE)

test_that("generic", {
  expect_snapshot(local({
    cli_div(theme = list(".alert" = list(before = "GENERIC! ")))
    cli_alert("wow")
  }))
})

test_that_cli("success", {
  expect_snapshot(local({
    cli_alert_success("wow")
  }))
})

test_that_cli("danger", {
  expect_snapshot(local({
    cli_alert_danger("wow")
  }))
})

test_that_cli("warning", {
  expect_snapshot(local({
    cli_alert_warning("wow")
  }))
})

test_that_cli("info", {
  expect_snapshot(local({
    cli_alert_info("wow")
  }))
})

test_that("before and after can have spaces", {
  expect_snapshot(local({
    cli_div(theme = list(.alert = list(before = "x  ", after = "  x")))
    cli_alert("continuing that first alert", wrap = TRUE)
  }))
})

context("cli alerts")

setup(start_app())
teardown(stop_app())

test_that("generic", {
  cli_div(theme = list(".alert" = list(before = "GENERIC! ")))
  out <- capt0(cli_alert("wow"))
  expect_match(out, "GENERIC")
})

test_that("success", {
  cli_div(theme = list(".alert-success" = list(before = "SUCCESS! ")))
  out <- capt0(cli_alert_success("wow"))
  expect_match(out, "SUCCESS")
})

test_that("danger", {
  cli_div(theme = list(".alert-danger" = list(before = "DANGER! ")))
  out <- capt0(cli_alert_danger("wow"))
  expect_match(out, "DANGER")
})

test_that("warning", {
  cli_div(theme = list(".alert-warning" = list(before = "WARNING! ")))
  out <- capt0(cli_alert_warning("wow"))
  expect_match(out, "WARNING")
})

test_that("info", {
  cli_div(theme = list(".alert-info" = list(before = "INFO! ")))
  out <- capt0(cli_alert_info("wow"))
  expect_match(out, "INFO")
})

# We need an UTF-8 platform or a recent R version on Windows

utf8 <- l10n_info()$`UTF-8`
newwin <- .Platform$OS.type == "windows" && getRversion() >= "4.0.0"
if (!utf8 && !newwin) return()

test_that("UTF-8 output on Windows", {
  skip_on_cran()
  skip_on_covr()
  out <- r_utf8(function() {
    library(cli)
    options(cli.unicode = TRUE)
    options(cli.num_colors = 1)
    options(cli.width = 70)
    options(cli.dynamic = FALSE)
    Sys.setenv(RSTUDIO = NA)

    s1 <- "\u30DE\u30EB\u30C1\u30D0\u30A4\u30C8\u306E\u30BF\u30A4\u30C8\u30EB"
    s2 <- "\u00e1rv\u00edzt\u0171r\u0151 t\u00fck\u00f6rf\u00far\u00f3g\u00e9p"

    cli_h1("Alerts")
    cli_alert(s1)
    cli_alert_danger(s2)
    cli_alert_info(s1)
    cli_alert_success(s1)
    cli_alert_warning(s1)

    cli_h1("Block quote")
    cli_blockquote(s1, s2)

    cli_h1("Bullets")
    cli_bullets(c("*" = s1, "!" = s2))

    cli_h1("Code")
    cli_code(c(s1, s2))

    cli_h1("Lists")
    cli_dl(c(s1 = s1, s2 = s2))
    cli_ol(c(s1, s2))
    cli_ul(c(s1, s2))

    cli_h1("Headers")
    cli_h1(s1)
    cli_h2(s2)
    cli_h3(s1)

    cli_h1("Progress bars")
    cli_progress_step(s1)
    cli_progress_step(s2)
    cli_progress_done()

    cli_h1("Text")
    idx <- c(1L, 2L, 1L, 2L, 1L, 1L, 1L, 1L, 2L, 1L, 2L, 1L, 1L, 1L, 1L)
    cli_text(paste(c(s1, s2)[idx], collapse = " "))

    cli_h1("Verbatim")
    cli_verbatim(c(s1, s2))
  })

  out$stdout <- gsub("\\[[.0-9]+m?s\\]", "[1s]", out$stdout)

  tmp <- tempfile(fileext = ".txt")
  writeBin(charToRaw(out$stdout), tmp)
  expect_snapshot_file(tmp, name = "utf8-output.txt")
})

test_that("utf8_graphemes", {
  expect_equal(utf8_graphemes(character()), list())
  expect_equal(utf8_graphemes(""), list(character()))

  str <- c(
    NA,
    "",
    "alpha",
    "\U0001f477\U0001f3ff\u200d\u2640\ufe0f",
    "\U0001f477\U0001f3ff",
    "\U0001f477\u200d\u2640\ufe0f",
    "\U0001f477\U0001f3fb",
    "\U0001f477\U0001f3ff"
  )
  exp <- list(
    NA_character_,
    character(),
    c("a", "l", "p", "h", "a"),
    "\U0001f477\U0001f3ff\u200d\u2640\ufe0f",
    "\U0001f477\U0001f3ff",
    "\U0001f477\u200d\u2640\ufe0f",
    "\U0001f477\U0001f3fb",
    "\U0001f477\U0001f3ff"
  )
  expect_equal(utf8_graphemes(str), exp)

  str2 <- paste0(na.omit(str), collapse = "")
  exp2 <- list(na.omit(unlist(exp)))
  expect_equal(utf8_graphemes(str2), exp2)
})

test_that("errors", {
  expect_snapshot_error(
    fix_r_utf8_output("\x02\xff\xfe \x03\xff\xfe \x03\xff\xfe")
  )
})

start_app()
on.exit(stop_app(), add = TRUE)

test_that("deep lists ul", {
  test_ul = function(n = 2) {
    for(i in seq_len(n)) {
      cli::cli_ul()
      cli::cli_li(paste0("Level ",i))
    }
  }
  expect_snapshot(
    for (i in 1:4) test_ul(i)
  )
})

test_that("deep lists ol", {
  test_ol = function(n = 2) {
    for(i in seq_len(n)) {
      cli::cli_ol()
      cli::cli_li(paste0("Level ",i))
    }
  }
  expect_snapshot(
    for (i in 1:4) test_ol(i)
  )
})

test_that("deep lists ol ul", {
  test_ol_ul = function(n = 2) {
    for(i in seq_len(n)) {
      cli::cli_ol()
      cli::cli_li(paste0("Level ",2*i-1))

      cli::cli_ul()
      cli::cli_li(paste0("Level ",2*i))
    }
  }
  expect_snapshot(
    for (i in 1:4) test_ol_ul(i)
  )
})

test_that("deep lists ul ol", {
  test_ul_ol = function(n = 2) {
    for(i in seq_len(n)) {
      cli::cli_ul()
      cli::cli_li(paste0("Level ",2*i-1))

      cli::cli_ol()
      cli::cli_li(paste0("Level ",2*i))
    }
  }
  expect_snapshot(
    for (i in 1:4) test_ul_ol(i)
  )
})

start_app()
on.exit(stop_app(), add = TRUE)

test_that_cli("headers", {
  expect_snapshot(local({
    cli_div(class = "testcli", theme = test_style())
    cli_h1("HEADER")
    cli_h2("Header")
    cli_h3("Header")
    x <- "foobar"
    xx <- 100
    cli_h2("{xx}. header: {x}")
  }), variant = if (packageVersion("testthat") <= "3.1.4") "old" else "new")
})

test_that("issue #218", {
  expect_snapshot({
    cli_h1("one {1} two {2} three {3}")
    cli_h2("one {1} two {2} three {3}")
  }, variant = if (packageVersion("testthat") <= "3.1.4") "old" else "new")
})

test_that("custom handler works", {
  conds <- list()
  withr::with_options(
    list(cli.default_handler = function(msg) conds <<- c(conds, list(msg))),
    { cli_h1("title"); cli_h2("subtitle"); cli_text("text") }
  )
  expect_equal(length(conds), 3)
  lapply(conds, expect_s3_class, "cli_message")
  expect_equal(conds[[1]]$type, "h1")
  expect_equal(conds[[2]]$type, "h2")
  expect_equal(conds[[3]]$type, "text")
})

context("custom-handler")

test_that("custom handler works", {
  conds <- list()
  withr::with_options(
    list(cli.default_handler = function(msg) conds <<- c(conds, list(msg))),
    { cli_h1("title"); cli_h2("subtitle"); cli_text("text") }
  )
  expect_equal(length(conds), 3)
  lapply(conds, expect_s3_class, "cli_message")
  expect_equal(conds[[1]]$type, "h1")
  expect_equal(conds[[2]]$type, "h2")
  expect_equal(conds[[3]]$type, "text")
})

start_app()
on.exit(stop_app(), add = TRUE)

test_that("auto closing", {
  expect_snapshot(local({
    cli_div(theme = list(".xx .emph" = list(before = "itsu:")))
    f <- function() {
      cli_par(class = "xx")
      cli_text("foo {.emph blah} bar")
    }

    # this has the marker
    f()
    # but this does not, because the .xx par was closed
    cli_text("foo {.emph blah} bar")
  }))
})

test_that("opt out of auto closing", {
  expect_snapshot(local({
    cli_div(theme = list(".xx .emph" = list(before = "itsu:")))
    id <- NULL
    f <- function() {
      id <<- cli_par(class = "xx", .auto_close = FALSE)
      cli_text("foo {.emph blah} bar")
    }

    f()

    # Still active
    cli_text("foo {.emph blah} bar")

    ## close explicitly
    expect_false(is.null(id))
    cli_end(id)
    cli_text("foo {.emph blah} bar")
  }))
})

test_that("auto closing with special env", {
  expect_snapshot(local({
    cli_div(theme = list(".xx .emph" = list(before = "itsu:")))
    f <- function() {
      g()
      # Still active
      cli_text("foo {.emph blah} bar")
    }

    g <- function() {
      cli_par(class = "xx", .auto_close = TRUE, .envir = parent.frame())
      cli_text("foo {.emph blah} bar")
    }

    f()

    # Not active any more
    cli_text("foo {.emph blah} bar")
  }))
})

test_that("div with special style", {
  expect_snapshot({
    f <- function() {
      cli_div(theme = list(".xx .emph" = list(before = "itsu:")))
      cli_par(class = "xx")
      cli_text("foo {.emph blah} bar")
    }
    f()
    # Not active any more
    cli_text("foo {.emph blah} bar")
  })
})

test_that("margin is squashed", {
  skip_if_not_installed("testthat", "3.1.2")

  # expect_snapshot cuts off the trailing newline from the message it
  # seems, so instead of 4 empty lines, there will be only three
  expect_snapshot(local({
    cli_div(theme = list(par = list("margin-top" = 4, "margin-bottom" = 4)))
    cli_text("three lines")
    cli_par()
    cli_par()
    cli_par()
    cli_text("until here")
    cli_end()
    cli_end()
    cli_end()
    cli_par()
    cli_par()
    cli_par()
    cli_text("no space, still")
    cli_end()
    cli_end()
    cli_end()
    cli_text("three lines again")
  }))
})

test_that("before and after work properly", {
  expect_snapshot(local({
    cli_div(
      theme = list(
        "div.alert-success" = list(before ="!!!")
      ))
    cli_alert_success("{.pkg foobar} is good")
  }))
})

#' Test cli output with testthat
#'
#' Use this function in your testthat test files, to test cli output.
#' It requires testthat edition 3, and works best with snapshot tests.
#'
#' `test_that_cli()` calls [testthat::test_that()] multiple times, with
#' different cli configurations. This makes it simple to test cli output
#' with and without ANSI colors, with and without Unicode characters.
#'
#' Currently available configurations:
#' * `plain`: no ANSI colors, ASCII characters only.
#' * `ansi`: ANSI colors, ASCII characters only.
#' * `unicode`: no ANSI colors, Unicode characters.
#' * `fancy`; ANSI colors, Unicode characters.
#'
#' See examples below and in cli's own tests, e.g. in
#' <https://github.com/r-lib/cli/tree/main/tests/testthat>
#' and the corresponding snapshots at
#' <https://github.com/r-lib/cli/tree/main/tests/testthat/_snaps>
#'
#' ## Important note regarding Windows
#'
#' Because of base R's limitation to record Unicode characters on Windows,
#' we suggest that you record your snapshots on Unix, or you restrict
#' your tests to ASCII configurations.
#'
#' Unicode tests on Windows are automatically skipped by testthat
#' currently.
#'
#' @param desc Test description, passed to [testthat::test_that()], after
#' appending the name of the cli configuration to it.
#' @param code Test code, it is modified to set up the cli config, and
#' then passed to [testthat::test_that()]
#' @param configs cli configurations to test `code` with. The default is
#' `NULL`, which includes all possible configurations. It can also be a
#' character vector, to restrict the tests to some configurations only.
#' See available configurations below.
#' @param links Whether to run the code with various hyperlinks allowed.
#' If `NULL` then hyperlinks are turned off. Otherwise it can be a character
#' vector with possible hyperlink configurations:
#'   * `"all"`: turn on all hyperlinks,
#'   * `"none"`: turn off all hyperlinks.
#'
#' @export
#' @examples
#' # testthat cannot record or compare snapshots when you run these
#' # examples interactively, so you might want to copy them into a test
#' # file
#'
#' # Default configurations
#' cli::test_that_cli("success", {
#'   testthat::local_edition(3)
#'   testthat::expect_snapshot({
#'     cli::cli_alert_success("wow")
#'   })
#' })
#'
#' # Only use two configurations, because this output does not have colors
#' cli::test_that_cli(configs = c("plain", "unicode"), "cat_bullet", {
#'   testthat::local_edition(3)
#'   testthat::expect_snapshot({
#'     cli::cat_bullet(letters[1:5])
#'   })
#' })
#'
#' # You often need to evaluate all cli calls of a test case in the same
#' # environment. Use `local()` to do that:
#' cli::test_that_cli("theming", {
#'   testthat::local_edition(3)
#'   testthat::expect_snapshot(local({
#'     cli::cli_div(theme = list(".alert" = list(before = "!!! ")))
#'     cli::cli_alert("wow")
#'   }))
#' })

test_that_cli <- function(desc, code,
                          configs = c("plain", "ansi", "unicode", "fancy"),
                          links = NULL) {
  code <- substitute(code)

  configs <- apply(expand.grid(configs, links %||% ""), 1, paste, collapse = "-")
  configs <- sub("-$", "", configs)

  doconfigs <- list(
    list(id = "plain",   unicode = FALSE, num_colors =   1, links = FALSE),
    list(id = "ansi",    unicode = FALSE, num_colors = 256, links = FALSE),
    list(id = "unicode", unicode = TRUE,  num_colors =   1, links = FALSE),
    list(id = "fancy",   unicode = TRUE,  num_colors = 256, links = FALSE),

    list(id = "plain-none",   unicode = FALSE, num_colors =   1, links = FALSE),
    list(id = "ansi-none",    unicode = FALSE, num_colors = 256, links = FALSE),
    list(id = "unicode-none", unicode = TRUE,  num_colors =   1, links = FALSE),
    list(id = "fancy-none",   unicode = TRUE,  num_colors = 256, links = FALSE),

    list(id = "plain-all",   unicode = FALSE, num_colors =   1, links = TRUE),
    list(id = "ansi-all",    unicode = FALSE, num_colors = 256, links = TRUE),
    list(id = "unicode-all", unicode = TRUE,  num_colors =   1, links = TRUE),
    list(id = "fancy-all",   unicode = TRUE,  num_colors = 256, links = TRUE)
  )

  parent <- parent.frame()
  lapply(doconfigs, function(conf) {
    if (!is.null(configs) && ! conf$id %in% configs) return()
    code2 <- substitute({
      testthat::local_reproducible_output(
        crayon = num_colors > 1,
        unicode = unicode
      )
      withr::local_options(
        cli.hyperlink = links,
        cli.hyperlink_help = links,
        cli.hyperlink_run = links,
        cli.hyperlink_vignette = links,
        cli.hyperlink_file_url_format = NULL,
        cli.hyperlink_run_url_format = NULL,
        cli.hyperlink_help_url_format = NULL,
        cli.hyperlink_vignette_url_format = NULL
      )
      withr::local_envvar(
        R_CLI_HYPERLINK_FILE_URL_FORMAT = NA_character_,
        R_CLI_HYPERLINK_RUN_URL_FORMAT = NA_character_,
        R_CLI_HYPERLINK_HELP_URL_FORMAT = NA_character_,
        R_CLI_HYPERLINK_VIGNETTE_URL_FORMAT = NA_character_
      )
      code_
    }, c(conf, list(code_ = code)))
    desc2 <- paste0(desc, " [", conf$id, "]")
    test <- substitute(
      testthat::test_that(desc, code),
      list(desc = desc2, code = code2)
    )
    eval(test, envir = parent)
  })
}

local_clean_cli_context <- function(.local_envir = parent.frame()) {
  withr::local_options(
    .local_envir = .local_envir,
    cli.hyperlink = NULL,
    cli.hyperlink_run = NULL,
    cli.hyperlink_help = NULL,
    cli.hyperlink_vignette = NULL,
    cli.hyperlink_file_url_format = NULL,
    cli.hyperlink_run_url_format = NULL,
    cli.hyperlink_help_url_format = NULL,
    cli.hyperlink_vignette_url_format = NULL,
    cli.num_colors = NULL,
    cli.palette = NULL,
    crayon.enabled = NULL
  )
  withr::local_envvar(
    .local_envir = .local_envir,
    R_CLI_HYPERLINKS = NA_character_,
    R_CLI_HYPERLINK_RUN = NA_character_,
    R_CLI_HYPERLINK_HELP = NA_character_,
    R_CLI_HYPERLINK_VIGNETTE = NA_character_,
    R_CLI_HYPERLINK_FILE_URL_FORMAT = NA_character_,
    R_CLI_HYPERLINK_RUN_URL_FORMAT = NA_character_,
    R_CLI_HYPERLINK_HELP_URL_FORMAT = NA_character_,
    R_CLI_HYPERLINK_VIGNETTE_URL_FORMAT = NA_character_,
    RSTUDIO_CLI_HYPERLINKS = NA_character_,
    R_CLI_NUM_COLORS = NA_character_,
    NO_COLOR = NA_character_,
    WT_SESSION = NA_character_,
    CI = NA_character_,
    TEAMCITY_VERSION = NA_character_,
    TERM_PROGRAM = NA_character_,
    TERM_PROGRAM_VERSION = NA_character_,
    VTE_VERSION = NA_character_
  )
}

test_that("suppress output", {
  if (getRversion() >= "4.0.0") {
    cnd <- NULL
    tryCatch(
      suppressMessages(cli_text("foo"), "cliMessage"),
      message = function(cnd2) cnd <<- cnd2
    )
    expect_null(cnd)
  }

  mysuppress <- function(expr) {
    withCallingHandlers(
      expr,
      cliMessage = function(msg) invokeRestart("muffleMessage")
    )
  }

  cnd <- NULL
  tryCatch(
    mysuppress(cli_text("foo")),
    message = function(cnd2) cnd <<- cnd2
  )
  expect_null(cnd)
})

context("cli containers")

setup(start_app())
teardown(stop_app())

test_that("auto closing", {
  cli_div(theme = list(".xx .emph" = list(before = "itsu:")))
  id <- ""
  out <- ""
  f <- function() {
    capt0(id <<- cli_par(class = "xx"))
    out <<- capt0(cli_text("foo {.emph blah} bar"))
  }

  capt0(f())
  expect_match(out, "itsu:", fixed = TRUE)

  out <- capt0(cli_text("foo {.emph blah} bar"))
  expect_false(grepl("itsu:", out, fixed = TRUE))
})

test_that("opt out of auto closing", {
  cli_div(theme = list(".xx .emph" = list(before = "itsu:")))
  id <- NULL
  f <- function() {
    capt0(id <<- cli_par(class = "xx", .auto_close = FALSE))
    out <- capt0(cli_text("foo {.emph blah} bar"))
    expect_match(out, "itsu:", fixed = TRUE)
  }

  capt0(f())

  ## Still active
  out <- capt0(cli_text("foo {.emph blah} bar"))
  expect_match(out, "itsu:", fixed = TRUE)

  ## close explicitly
  expect_false(is.null(id))
  capt0(cli_end(id))

  out <- capt0(cli_text("foo {.emph blah} bar"))
  expect_false(grepl("itsu:", out, fixed = TRUE))
})

test_that("auto closing with special env", {
  cli_div(theme = list(".xx .emph" = list(before = "itsu:")))
  id <- NULL
  f <- function() {
    g()
    ## Still active
    out <- capt0(cli_text("foo {.emph blah} bar"))
    expect_match(out, "itsu:", fixed = TRUE)
  }

  g <- function() {
    capt0(id <<- cli_par(class = "xx", .auto_close = TRUE,
                        .envir = parent.frame()))
    out <- capt0(cli_text("foo {.emph blah} bar"))
    expect_match(out, "itsu:", fixed = TRUE)
  }

  capt0(f())

  ## Not active any more
  out <- capt0(cli_text("foo {.emph blah} bar"))
  expect_false(grepl("itsu:", out, fixed = TRUE))
})

test_that("div with special style", {
  f <- function() {
    cli_div(theme = list(".xx .emph" = list(before = "itsu:")))
    capt0(cli_par(class = "xx"))
    out <- capt0(cli_text("foo {.emph blah} bar"))
    expect_match(out, "itsu:", fixed = TRUE)
  }

  capt0(f())

  ## Not active any more
  out <- capt0(cli_text("foo {.emph blah} bar"))
  expect_false(grepl("itsu:", out, fixed = TRUE))
})

test_that("margin is squashed", {
  cli_div(theme = list(par = list("margin-top" = 3, "margin-bottom" = 3)))
  out <- capt0({ cli_par(); cli_par(); cli_par() }, strip_style = TRUE)
  expect_equal(out, "\n\n\n")
  out <- capt0({ cli_end(); cli_end(); cli_end() })
  expect_equal(out, "")

  out <- capt0({ cli_par(); cli_par(); cli_par() })
  expect_equal(out, "")
  capt0(cli_text(lorem_ipsum()))
  out <- capt0({ cli_end(); cli_end(); cli_end() }, strip_style = TRUE)
  expect_equal(out, "\n\n\n")
})

test_that("before and after work properly", {
  cli_div(theme = list(
    "div.alert-success" = list(before ="!!!")
  ))
  out <- capt0(cli_alert_success("{.pkg foobar} is good"))
  expect_match(out, "!!!", fixed = TRUE)
})

start_app()
on.exit(stop_app(), add = TRUE)

test_that("text is wrapped", {
  withr::local_options(cli.width = 60)
  withr::local_rng_version("3.3.0")
  set.seed(42)
  expect_snapshot(local({
    cli_div(class = "testcli", theme = test_style())
    cli_h1("Header")
    cli_text(lorem_ipsum())
  }))
})

test_that("verbatim text is not wrapped", {
  cli_div(class = "testcli", theme = test_style())
  withr::local_options(cli.width = 60)
  suppressMessages(cli_h1("Header"))
  txt <- strrep("1234567890 ", 20)
  out <- capt0(cli_verbatim(txt), strip_style = TRUE)
  expect_equal(out, paste0(txt, "\n"))
})

test_that("type style", {
  expect_snapshot({
    # objects
    cli_text("{.obj_type_friendly {mtcars}}")
    cli_text("{.obj_type_friendly {tibble::as_tibble(mtcars)}}")
    cli_text("{.obj_type_friendly {rlang::quo(1)}}")

    # matrices and arrays
    cli_text("{.obj_type_friendly {list()}}")
    cli_text("{.obj_type_friendly {matrix(list(1, 2))}}")
    cli_text("{.obj_type_friendly {array(list(1, 2, 3))}}")

    cli_text("{.obj_type_friendly {integer()}}")
    cli_text("{.obj_type_friendly {matrix(1:3)}}")
    cli_text("{.obj_type_friendly {array(1:3, dim = 1:3)}}")

    cli_text("{.obj_type_friendly {character()}}")
    cli_text("{.obj_type_friendly {matrix(letters)}}")
    cli_text("{.obj_type_friendly {array(letters[1:3], dim = 1:3)}}")

    # missing arguments
    # this does not actually work in {.obj_type_friendly} because of the eval rounds
    typename(quote(expr = ))

    # scalars
    cli_text("{.obj_type_friendly {NA}}")
    cli_text("{.obj_type_friendly {NA_integer_}}")
    cli_text("{.obj_type_friendly {NA_real_}}")
    cli_text("{.obj_type_friendly {NA_complex_}}")
    cli_text("{.obj_type_friendly {NA_character_}}")

    cli_text("{.obj_type_friendly {TRUE}}")
    cli_text("{.obj_type_friendly {FALSE}}")

    cli_text("{.obj_type_friendly {1L}}")
    cli_text("{.obj_type_friendly {1.0}}")
    cli_text("{.obj_type_friendly {1i}}")
    cli_text("{.obj_type_friendly {as.raw(1L)}}")

    cli_text("{.obj_type_friendly {'foo'}}")
    cli_text("{.obj_type_friendly {''}}")

    cli_text("{.obj_type_friendly {list(1)}}")

    cli_text("{.obj_type_friendly {matrix(NA)}}")
    cli_text("{.obj_type_friendly {matrix(1)}}")

    # empty vectors
    cli_text("{.obj_type_friendly {logical()}}")
    cli_text("{.obj_type_friendly {integer()}}")
    cli_text("{.obj_type_friendly {double()}}")
    cli_text("{.obj_type_friendly {complex()}}")
    cli_text("{.obj_type_friendly {character()}}")
    cli_text("{.obj_type_friendly {raw()}}")
    cli_text("{.obj_type_friendly {list()}}")

    # correct article
    cli_text("{.obj_type_friendly {structure(1, class = 'igraph')}}")
  })
})

# https://github.com/r-lib/cli/issues/370

test_that("glue quotes and comments", {
  expect_snapshot({
    cli_dl(
      c(
        "test_1" = "all good",
        "test_2" = "not #good"
      )
    )
    cli::cli_dl(c("test_3" = "no' good either"))
    cli::cli_dl(c("test_4" = "no\" good also"))
    cli::cli_text("{.url https://example.com/#section}")
    cli::cli_alert_success("Qapla'")
  })
})

test_that("quotes, etc. within expressions are still OK", {
  expect_snapshot({
    cli::cli_text("{.url URL} {x <- 'foo'; nchar(x)}")
    cli::cli_text("{.url URL} {x <- \"foo\"; nchar(x)}")
    cli::cli_text("{.url URL} {1 + 1 # + 1} {1 + 1}")
  })
})

test_that("{. } is always a style", {
  cmd <- paste0("{.", basename(tempfile()), " }")
  expect_silent(glue_cmd(cmd, .envir = environment()))
  .foo <- 100
  vals <- glue_cmd("{.foo }", .envir = environment())$values
  expect_false(any(grepl("^v[0-9]+$", names(vals))))
})

test_that("{ } is parsed with literal = FALSE", {
  expect_snapshot(
    format_message("{.emph {'{foo {}'}}")
  )
})

start_app()
on.exit(stop_app(), add = TRUE)

test_that_cli("meta basics", {
  expect_snapshot(
    cli::cli({
      message("This is before")
      cli_alert_info("First message")
      message("This as well")
      cli_alert_success("Success!")
    })
  )
})

test_that_cli("meta is single cli_message", {
  msgs <- list()
  withCallingHandlers(
    cli::cli({
      cli_alert_info("First message")
      cli_alert_success("Success!")
    }),
    cli_message = function(msg) {
      msgs <<- c(msgs, list(msg))
      invokeRestart("cli_message_handled")
    }
  )

  expect_equal(length(msgs), 1L)
  expect_snapshot(cli_server_default(msgs[[1]]))
})

test_that_cli("meta is single cliMessage", {
  msgs <- list()
  expect_snapshot(
    withCallingHandlers(
      cli::cli({
        cli_alert_info("First message")
        cli_alert_success("Success!")
      }),
      cliMessage = function(msg) {
        msgs <<- c(msgs, list(msg))
      }
    )
  )

  expect_equal(length(msgs), 1L)
})

test_that_cli("substitution", {
  expect_snapshot(local({
    x <- 1:3
    cli({
      title <- "My title"
      cli_h1("Title: {.emph {title}}")
      cli_text("And {.emph some} more: {.val {x}}")
    })
  }))
})

test_that("return values are ok when recording (#496)", {
  expect_snapshot(
    cli::cli({
      lid <- cli::cli_ul()
      cli::cli_li("a bullet")
      cli::cli_end(lid)
    })
  )
})

test_that("nested cli() (#497)", {
  expect_snapshot(
    cli::cli({
      cli::cli_h1("Header")
      cli::cli(cli::cli_text("Some text"))
      cli::cli_text("Some more text")
    })
  )
})

test_that("does not break", {
  testthat::local_reproducible_output(unicode = TRUE)
  expect_snapshot(local({
    withr::local_options(cli.width = 40)
    str30 <- "123456789 123456789 1234567890"
    cli_text(c(str30, "this\u00a0is\u00a0not\u00a0breaking"))
  }))
})

context("subprocess")

test_that("events are properly generated", {
  ## This needs callr >= 3.0.0.90001, which is not yet on CRAN
  if (packageVersion("callr") < "3.0.0.9001") skip("Need newer callr")

  do <- function() {
    options(cli.message_class = "callr_message")
    cli::cli_div()
    cli::cli_h1("title")
    cli::cli_text("text")
  }

  rs <- callr::r_session$new()
  on.exit(rs$kill(), add = TRUE)

  msgs <- list()
  handler <- function(msg) {
    msgs <<- c(msgs, list(msg))
    if (!is.null(findRestart("cli_message_handled"))) {
      invokeRestart("cli_message_handled")
    }
  }

  withCallingHandlers(
    rs$run(do),
    cli_message = handler)

  expect_equal(length(msgs), 4)
  lapply(msgs, expect_s3_class, "cli_message")
  expect_equal(msgs[[1]]$type, "div")
  expect_equal(msgs[[2]]$type, "h1")
  expect_equal(msgs[[3]]$type, "text")
  expect_equal(msgs[[4]]$type, "end")

  rs$close()
})

test_that("subprocess with default handler", {
  ## This needs callr >= 3.0.0.90001, which is not yet on CRAN
  if (packageVersion("callr") < "3.0.0.9001") skip("Need newer callr")

  do <- function() {
    options(cli.message_class = "callr_message")
    cli::cli_div()
    cli::cli_h1("title")
    cli::cli_text("text")
  }

  rs <- callr::r_session$new()
  on.exit(rs$kill(), add = TRUE)

  msgs <- list()
  withr::with_options(list(
    cli.default_handler = function(msg)  {
      msgs <<- c(msgs, list(msg))
      if (!is.null(findRestart("cli_message_handled"))) {
        invokeRestart("cli_message_handled")
      }
    }),
    rs$run(do)
  )

  expect_equal(length(msgs), 4)
  lapply(msgs, expect_s3_class, "cli_message")
  expect_equal(msgs[[1]]$type, "div")
  expect_equal(msgs[[2]]$type, "h1")
  expect_equal(msgs[[3]]$type, "text")
  expect_equal(msgs[[4]]$type, "end")

  rs$close()
})

test_that("output in child process", {
  ## This needs callr >= 3.0.0.90001, which is not yet on CRAN
  if (packageVersion("callr") < "3.0.0.9001") skip("Need newer callr")

  # We need to do our own condition handling, otherwise callr will
  # handle `cli_message` and copy it to the main process.
  # So on `cli_message` we just call the default handler, which will
  # call `message()`, and on `message` we'll copy the formatted message
  # to the main process.

  do <- function() {
    options(cli.num_colors = 256)
    withCallingHandlers({
        cli::start_app(theme = cli::simple_theme())
        cli::cli_h1("Title")
        cli::cli_text("This is generated in the {.emph subprocess}.")
        "foobar"
      },
      cli_message = function(msg) {
        withCallingHandlers({
          cli:::cli_server_default(msg)
          invokeRestart("cli_message_handled") },
          message = function(mmsg) {
            class(mmsg) <- c("callr_message", "message", "condition")
            signalCondition(mmsg)
        })
      }
    )
  }

  rs <- callr::r_session$new()
  on.exit(rs$kill(), add = TRUE)

  # Store the formatted messages from callr
  # We also need to muffle the default handler here

  msgs <- list()
  result <- withCallingHandlers(
    rs$run_with_output(do),
    callr_message = function(msg) {
      msgs <<- c(msgs, list(msg))
      if (!is.null(msg$muffle) && !is.null(findRestart(msg$muffle))) {
        invokeRestart(msg$muffle)
      }
    }
  )

  expect_equal(result$stdout, "")
  expect_equal(result$stderr, "")
  expect_identical(result$result, "foobar")
  expect_null(result$error)
  str <- paste(vcapply(msgs, "[[", "message"), collapse = "")
  expect_true(ansi_has_any(str))
  expect_match(str, "Title")
  expect_match(str, "This is generated")

  rs$close()
})

test_that("substitution in child process", {

  do <- function() {
    options(cli.message_class = "callr_message")
    cli::cli_text("This is process {Sys.getpid()}.")
  }

  rs <- callr::r_session$new()
  on.exit(rs$kill(), add = TRUE)

  out <- capt0(rs$run(do))
  expect_match(out, glue::glue("This is process {rs$get_pid()}"))

  rs$close()
})

start_app(.auto_close = TRUE)

test_that_cli("add/remove/list themes", {
  withr::local_rng_version("3.3.0")
  set.seed(24)

  start_app(.auto_close = TRUE)

  id <- default_app()$add_theme(list(".green" = list(color = "green")))
  expect_true(id %in% names(default_app()$list_themes()))

  expect_snapshot({
    cli_par(class = "green")
    cli_text(lorem_ipsum())
    cli_end()
  })

  default_app()$remove_theme(id)
  expect_false(id %in% names(default_app()$list_themes()))
})

test_that("default theme is valid", {
  expect_error({
    id <- default_app()$add_theme(builtin_theme())
    default_app()$remove_theme(id)
  }, NA)
})

test_that("explicit formatter is used, and combined", {
  id <- default_app()$add_theme(list(
    "span.emph" = list(
      fmt = function(x) paste0("(((", x, ")))"),
      before = "<<", after = ">>")
    ))
  on.exit(default_app()$remove_theme(id), add = TRUE)
  expect_snapshot(
    cli_text("this is {.emph it}, really")
  )
})

test_that("simple theme", {
  def <- simple_theme()
  expect_true(is.list(def))
  expect_false(is.null(names(def)))
  expect_true(all(names(def) != ""))
  expect_true(all(vlapply(def, is.list)))
})

test_that("user's override", {
  custom <- list(".alert" = list(before = "custom:"))
  override <- list(".alert" = list(after = "override:"))

  expect_snapshot(local({
    start_app(theme = custom, .auto_close = FALSE)
    cli_alert("Alert!")
    stop_app()

    withr::local_options(cli.user_theme = override)
    start_app(theme = custom, .auto_close = FALSE)
    cli_alert("Alert!")
    stop_app()
  }))
})

test_that("theme does not precompute Unicode symbols", {
  withr::local_options(cli.unicode = TRUE, cli.num_colors = 256L)
  start_app()
  msg <- NULL
  withCallingHandlers(
    cli_alert_success("ok"),
    cliMessage = function(m) {
      msg <<- m
      invokeRestart("muffleMessage")
    }
  )
  expect_true(ansi_has_any(msg$message))

  msg2 <- NULL
  withr::local_options(cli.unicode = FALSE, cli.num_colors = 1L)
  withCallingHandlers(
    cli_alert_success("ok2"),
    cliMessage = function(m) {
      msg2 <<- m
      invokeRestart("muffleMessage")
    }
  )
  expect_equal(msg2$message, "v ok2\n")
})

test_that("NULL will undo a style property", {
  expect_snapshot(local({
    cli_alert("this has an arrow")
    cli_div(theme = list(.alert = list(before = NULL)))
    cli_alert("this does not")
  }))
})

test_that_cli(configs = "ansi", "NULL will undo color", {
  expect_snapshot(local({
    cli_alert("{.emph {.val this is blue}}")
    cli_div(theme = list(span.val = list(color = NULL)))
    cli_alert("{.emph {.val this is not}}")
  }))
  expect_snapshot(local({
    cli_alert("{.emph {.val this is blue}}")
    cli_div(theme = list(span.val = list(color = "none")))
    cli_alert("{.emph {.val this is not}}")
  }))
})

withr::local_options(cli.theme = NULL, cli.user_theme = NULL)
withr::local_options(cli.theme_dark = FALSE, cli.num_colors = 256)
start_app()
on.exit(stop_app(), add = TRUE)

test_that_cli(configs = "ansi", "NULL will undo background color", {
  skip_if_not_installed("testthat", "3.1.2")
  expect_snapshot(local({
    cli_alert("{.emph {.code this has bg color}}")
    cli_div(theme = list(span = list("background-color" = NULL)))
    cli_alert("{.emph {.code this does not}}")
  }))
  expect_snapshot(local({
    cli_alert("{.emph {.code this has bg color}}")
    cli_div(theme = list(span = list("background-color" = "none")))
    cli_alert("{.emph {.code this does not}}")
  }))
})

test_that("events are properly generated", {
  skip_on_cran()
  ## This needs callr >= 3.0.0.90001, which is not yet on CRAN
  if (packageVersion("callr") < "3.0.0.9001") skip("Need newer callr")

  do <- function() {
    options(cli.message_class = "callr_message")
    cli::cli_div()
    cli::cli_h1("title")
    cli::cli_text("text")
  }

  rs <- callr::r_session$new()
  on.exit(rs$kill(), add = TRUE)

  msgs <- list()
  handler <- function(msg) {
    msgs <<- c(msgs, list(msg))
    if (!is.null(findRestart("cli_message_handled"))) {
      invokeRestart("cli_message_handled")
    }
    if (!is.null(findRestart("callr_r_session_muffle"))) {
      invokeRestart("callr_r_session_muffle")
    }
  }

  withCallingHandlers(
    rs$run(do),
    cli_message = handler)

  expect_equal(length(msgs), 4)
  lapply(msgs, expect_s3_class, "cli_message")
  expect_equal(msgs[[1]]$type, "div")
  expect_equal(msgs[[2]]$type, "h1")
  expect_equal(msgs[[3]]$type, "text")
  expect_equal(msgs[[4]]$type, "end")

  rs$close()
})

test_that("subprocess with default handler", {
  skip_on_cran()
  ## This needs callr >= 3.0.0.90001, which is not yet on CRAN
  if (packageVersion("callr") < "3.0.0.9001") skip("Need newer callr")

  do <- function() {
    options(cli.message_class = "callr_message")
    cli::cli_div()
    cli::cli_h1("title")
    cli::cli_text("text")
  }

  rs <- callr::r_session$new()
  on.exit(rs$kill(), add = TRUE)

  msgs <- list()
  withr::with_options(list(
    cli.default_handler = function(msg)  {
      msgs <<- c(msgs, list(msg))
      if (!is.null(findRestart("cli_message_handled"))) {
        invokeRestart("cli_message_handled")
      }
    }),
    rs$run(do)
  )

  expect_equal(length(msgs), 4)
  lapply(msgs, expect_s3_class, "cli_message")
  expect_equal(msgs[[1]]$type, "div")
  expect_equal(msgs[[2]]$type, "h1")
  expect_equal(msgs[[3]]$type, "text")
  expect_equal(msgs[[4]]$type, "end")

  rs$close()
})

test_that("output in child process", {
  skip_on_cran()
  ## This needs callr >= 3.0.0.90001, which is not yet on CRAN
  if (packageVersion("callr") < "3.0.0.9001") skip("Need newer callr")

  # We need to do our own condition handling, otherwise callr will
  # handle `cli_message` and copy it to the main process.
  # So on `cli_message` we just call the default handler, which will
  # call `message()`, and on `message` we'll copy the formatted message
  # to the main process.

  do <- function() {
    options(cli.num_colors = 256)
    withCallingHandlers({
        cli::start_app(theme = cli::simple_theme())
        cli::cli_h1("Title")
        cli::cli_text("This is generated in the {.emph subprocess}.")
        "foobar"
      },
      cli_message = function(msg) {
        withCallingHandlers({
          cli:::cli_server_default(msg)
          invokeRestart("cli_message_handled") },
          message = function(mmsg) {
            class(mmsg) <- c("callr_message", "message", "condition")
            signalCondition(mmsg)
        })
      }
    )
  }

  rs <- callr::r_session$new()
  on.exit(rs$kill(), add = TRUE)

  # Store the formatted messages from callr
  # We also need to muffle the default handler here

  msgs <- list()
  result <- withCallingHandlers(
    rs$run_with_output(do),
    callr_message = function(msg) {
      msgs <<- c(msgs, list(msg))
      if (!is.null(msg$muffle) && !is.null(findRestart(msg$muffle))) {
        invokeRestart(msg$muffle)
      }
    }
  )

  expect_equal(result$stdout, "")
  expect_equal(result$stderr, "")
  expect_identical(result$result, "foobar")
  expect_null(result$error)
  str <- paste(vcapply(msgs, "[[", "message"), collapse = "")
  expect_true(ansi_has_any(str))
  expect_match(str, "Title")
  expect_match(str, "This is generated")

  rs$close()
})

test_that("substitution in child process", {
  skip_on_cran()
  do <- function() {
    options(cli.message_class = "callr_message")
    cli::cli_text("This is process {Sys.getpid()}.")
  }

  rs <- callr::r_session$new()
  on.exit(rs$kill(), add = TRUE)

  out <- capt0(rs$run(do))
  expect_match(out, glue::glue("This is process {rs$get_pid()}"))

  rs$close()
})

start_app()
on.exit(stop_app(), add = TRUE)

test_that_cli("meta basics", {
  expect_snapshot(
    cli::cli({
      message("This is before")
      cli_alert_info("First message")
      message("This as well")
      cli_alert_success("Success!")
    })
  )
})

test_that_cli("meta is single cli_message", {
  msgs <- list()
  withCallingHandlers(
    cli::cli({
      cli_alert_info("First message")
      cli_alert_success("Success!")
    }),
    cli_message = function(msg) {
      msgs <<- c(msgs, list(msg))
      invokeRestart("cli_message_handled")
    }
  )

  expect_equal(length(msgs), 1L)
  expect_snapshot(cli_server_default(msgs[[1]]))
})

test_that_cli("meta is single cliMessage", {
  msgs <- list()
  expect_snapshot(
    withCallingHandlers(
      cli::cli({
        cli_alert_info("First message")
        cli_alert_success("Success!")
      }),
      cliMessage = function(msg) {
        msgs <<- c(msgs, list(msg))
      }
    )
  )

  expect_equal(length(msgs), 1L)
})

test_that_cli("substitution", {
  expect_snapshot(local({
    x <- 1:3
    cli({
      title <- "My title"
      cli_h1("Title: {.emph {title}}")
      cli_text("And {.emph some} more: {.val {x}}")
    })
  }))
})

start_app()
on.exit(stop_app(), add = TRUE)

test_that_cli(configs = c("plain", "unicode"), "ul", {
  expect_snapshot(local({
    lid <- cli_ul()
    cli_li("foo")
    cli_li(c("bar", "foobar"))
    cli_end(lid)
  }))
})

test_that_cli(configs = c("plain", "unicode"), "ol", {
  expect_snapshot(local({
    cli_div(theme = list(ol = list()))
    lid <- cli_ol()
    cli_li("foo")
    cli_li(c("bar", "foobar"))
    cli_end(lid)
  }))
})

test_that_cli(configs = c("plain", "unicode"), "ul ul", {
  expect_snapshot(local({
    cli_div(
      theme = list(
        "ul ul" = list("list-style-type" = "-", "margin-left" = 2)
      )
    )
    lid <- cli_ul()
    cli_li("1")
    lid2 <- cli_ul()
    cli_li("1 1")
    cli_li(c("1 2", "1 3"))
    cli_end(lid2)
    cli_li("2")
    cli_end(lid)
  }))
})

test_that_cli(configs = c("plain", "unicode"), "ul ol", {
  expect_snapshot(local({
    cli_div(
      theme = list(
        li = list("margin-left" = 2)
      )
    )
    lid <- cli_ul()
    cli_li("1")
    lid2 <- cli_ol()
    cli_li("1 1")
    cli_li(c("1 2", "1 3"))
    cli_end(lid2)
    cli_li("2")
    cli_end(lid)
  }))
})

test_that_cli(configs = c("plain", "unicode"), "ol ol", {
  expect_snapshot(local({
    cli_div(
      theme = list(
        "li" = list("margin-left" = 2),
        "li li" = list("margin-left" = 2)
      )
    )
    lid <- cli_ol()
    cli_li("1")
    lid2 <- cli_ol()
    cli_li("1 1")
    cli_li(c("1 2", "1 3"))
    cli_end(lid2)
    cli_li("2")
    cli_end(lid)
  }))
})

test_that_cli(configs = c("plain", "unicode"), "ol ul", {
  expect_snapshot(local({
    cli_div(
      theme = list(
        ul = list("margin-left" = 2)
      )
    )
    lid <- cli_ol()
    cli_li("1")
    lid2 <- cli_ul()
    cli_li("1 1")
    cli_li(c("1 2", "1 3"))
    cli_end(lid2)
    cli_li("2")
    cli_end(lid)
  }))
})

test_that_cli(configs = c("plain", "unicode"), "starting with an item", {
  expect_snapshot(local({
    cli_li("foo")
    cli_li(c("bar", "foobar"))
    cli_end()
    cli_end()
  }))
})

test_that_cli(configs = c("plain", "unicode"), "ol, with first item", {
  expect_snapshot(local({
    cli_div(theme = list(ol = list()))
    lid <- cli_ol("foo", .close = FALSE)
    cli_li(c("bar", "foobar"))
    cli_end(lid)
  }))
})

test_that_cli(configs = c("plain", "unicode"), "ul, with first item", {
  expect_snapshot(local({
    lid <- cli_ul("foo", .close = FALSE)
    cli_li(c("bar", "foobar"))
    cli_end(lid)
  }))
})

test_that_cli(configs = c("plain", "unicode"), "dl", {
  expect_snapshot(local({
    cli_div(theme = list(ul = list()))
    lid <- cli_dl()
    cli_li(c(this = "foo"))
    cli_li(c(that = "bar", other = "foobar"))
    cli_end(lid)
  }))
})

test_that_cli(configs = c("plain", "unicode"), "dl dl", {
  expect_snapshot(local({
    cli_div(
      theme = list(
        li = list("margin-left" = 2)
      )
    )
    lid <- cli_dl()
    cli_li(c(a = "1"))
    lid2 <- cli_dl()
    cli_li(c("a-a" = "1 1"))
    cli_li(c("a-b" = "1 2", "a-c" = "1 3"))
    cli_end(lid2)
    cli_li(c(b = "2"))
    cli_end(lid)
  }))
})

test_that_cli(configs = c("plain", "unicode"), "dl ol", {
  expect_snapshot(local({
    cli_div(
      theme = list(
        li = list("margin-left" = 2)
      )
    )
    lid <- cli_dl()
    cli_li(c(a = "1"))
    lid2 <- cli_ol()
    cli_li(c("1 1"))
    cli_li(c("1 2", "1 3"))
    cli_end(lid2)
    cli_li(c(b = "2"))
    cli_end(lid)
  }))
})

test_that_cli(configs = c("plain", "unicode"), "dl ul", {
  expect_snapshot(local({
    cli_div(
      theme = list(
        li = list("margin-left" = 2)
      )
    )
    lid <- cli_dl()
    cli_li(c(a = "1"))
    lid2 <- cli_ul()
    cli_li(c("1 1"))
    cli_li(c("1 2", "1 3"))
    cli_end(lid2)
    cli_li(c(b = "2"))
    cli_end(lid)
  }))
})

test_that_cli(configs = c("plain", "unicode"), "ol dl", {
  expect_snapshot(local({
    cli_div(
      theme = list(
        li = list("margin-left" = 2)
      )
    )
    lid <- cli_ol()
    cli_li("1")
    lid2 <- cli_dl()
    cli_li(c("a-a" = "1 1"))
    cli_li(c("a-b" = "1 2", "a-c" = "1 3"))
    cli_end(lid2)
    cli_li("2")
    cli_end(lid)
  }))
})

test_that_cli(configs = c("plain", "unicode"), "ul dl", {
  expect_snapshot(local({
    cli_div(
      theme = list(
        li = list("margin-left" = 2)
      )
    )
    lid <- cli_ul()
    cli_li("1")
    lid2 <- cli_dl()
    cli_li(c("a-a" = "1 1"))
    cli_li(c("a-b" = "1 2", "a-c" = "1 3"))
    cli_end(lid2)
    cli_li("2")
    cli_end(lid)
  }))
})

test_that_cli(configs = c("plain", "unicode"), "dl, with first item", {
  expect_snapshot(local({
    cli_div(theme = list(ul = list()))
    lid <- cli_dl(c(this = "foo"), .close = FALSE)
    cli_li(c(that = "bar", other = "foobar"))
    cli_end(lid)
  }))
})

test_that_cli(configs = "ansi", "styling pieces of a dl", {
  expect_snapshot(local({
    cli_div(
      theme = list(
        .dt = list(after = " -> "),
        .dd = list(color = "blue")
      )
    )
    cli_dl(c(foo = "bar", bar = "baz"))
  }))
})

test_that("cli_dl edge cases", {
  # invalid input
  expect_error(
    cli_dl("foo", "must be a named character vector")
  )

  # empty dd
  expect_snapshot(
    cli_dl(c(abc = "foo", empty = "", def = "bar"))
  )
})

# We need an UTF-8 platform or a recent R version on Windows

utf8 <- l10n_info()$`UTF-8`
newwin <- .Platform$OS.type == "windows" && getRversion() >= "4.0.0"
if (!utf8 && !newwin) return()

test_that("UTF-8 output on Windows", {
  skip_on_cran()
  out <- r_utf8(function() {
    library(cli)
    options(cli.unicode = TRUE)
    options(cli.num_colors = 1)
    options(cli.width = 70)
    options(cli.dynamic = FALSE)
    Sys.setenv(RSTUDIO = NA)

    s1 <- "\u30DE\u30EB\u30C1\u30D0\u30A4\u30C8\u306E\u30BF\u30A4\u30C8\u30EB"
    s2 <- "\u00e1rv\u00edzt\u0171r\u0151 t\u00fck\u00f6rf\u00far\u00f3g\u00e9p"

    cli_h1("Alerts")
    cli_alert(s1)
    cli_alert_danger(s2)
    cli_alert_info(s1)
    cli_alert_success(s1)
    cli_alert_warning(s1)

    cli_h1("Block quote")
    cli_blockquote(s1, s2)

    cli_h1("Bullets")
    cli_bullets(c("*" = s1, "!" = s2))

    cli_h1("Code")
    cli_code(c(s1, s2))

    cli_h1("Lists")
    cli_dl(c(s1 = s1, s2 = s2))
    cli_ol(c(s1, s2))
    cli_ul(c(s1, s2))

    cli_h1("Headers")
    cli_h1(s1)
    cli_h2(s2)
    cli_h3(s1)

    cli_h1("Progress bars")
    cli_progress_step(s1)
    cli_progress_step(s2)
    cli_progress_done()

    cli_h1("Text")
    idx <- c(1L, 2L, 1L, 2L, 1L, 1L, 1L, 1L, 2L, 1L, 2L, 1L, 1L, 1L, 1L)
    cli_text(paste(c(s1, s2)[idx], collapse = " "))

    cli_h1("Verbatim")
    cli_verbatim(c(s1, s2))
  })

  out$stdout <- gsub("\\[[.0-9]+m?s\\]", "[1s]", out$stdout)

  tmp <- tempfile(fileext = ".txt")
  writeBin(charToRaw(out$stdout), tmp)
  expect_snapshot_file(tmp, name = "utf8-output.txt")
})

test_that("utf8_graphemes", {
  expect_equal(utf8_graphemes(character()), list())
  expect_equal(utf8_graphemes(""), list(character()))

  str <- c(
    NA,
    "",
    "alpha",
    "\U0001f477\U0001f3ff\u200d\u2640\ufe0f",
    "\U0001f477\U0001f3ff",
    "\U0001f477\u200d\u2640\ufe0f",
    "\U0001f477\U0001f3fb",
    "\U0001f477\U0001f3ff"
  )
  exp <- list(
    NA_character_,
    character(),
    c("a", "l", "p", "h", "a"),
    "\U0001f477\U0001f3ff\u200d\u2640\ufe0f",
    "\U0001f477\U0001f3ff",
    "\U0001f477\u200d\u2640\ufe0f",
    "\U0001f477\U0001f3fb",
    "\U0001f477\U0001f3ff"
  )
  expect_equal(utf8_graphemes(str), exp)

  str2 <- paste0(na.omit(str), collapse = "")
  exp2 <- list(na.omit(unlist(exp)))
  expect_equal(utf8_graphemes(str2), exp2)
})

start_app()
on.exit(stop_app(), add = TRUE)

test_that_cli("headers", {
  expect_snapshot(
    local({
      cli_div(class = "testcli", theme = test_style())
      cli_h1("HEADER")
      cli_h2("Header")
      cli_h3("Header")
      x <- "foobar"
      xx <- 100
      cli_h2("{xx}. header: {x}")
    })
  )
})

test_that("issue #218", {
  expect_snapshot(
    {
      cli_h1("one {1} two {2} three {3}")
      cli_h2("one {1} two {2} three {3}")
    }
  )
})

# https://github.com/r-lib/cli/issues/370

test_that("glue quotes and comments", {
  expect_snapshot({
    cli_dl(
      c(
        "test_1" = "all good",
        "test_2" = "not #good"
      )
    )
#    cli::cli_dl(c("test_3" = "no' good either"))
#    cli::cli_dl(c("test_4" = "no\" good also"))
    cli::cli_text("{.url https://example.com/#section}")
#    cli::cli_alert_success("Qapla'")
  })
})

test_that("quotes, etc. within expressions are still OK", {
  expect_snapshot({
#    cli::cli_text("{.url URL} {x <- 'foo'; nchar(x)}")
#    cli::cli_text("{.url URL} {x <- \"foo\"; nchar(x)}")
    cli::cli_text("{.url URL} {1 + 1 # + 1} {1 + 1}")
  })
})

test_that("does not break", {
  expect_snapshot(local({
    withr::local_options(cli.width = 40)
    str30 <- "123456789 123456789 1234567890"
    cli_text(c(str30, "this\u00a0is\u00a0not\u00a0breaking"))
  }))
})

start_app()
on.exit(stop_app(), add = TRUE)

test_that_cli("add/remove/list themes", {
  local_rng_version("3.3.0")
  set.seed(24)

  id <- default_app()$add_theme(list(".green" = list(color = "green")))
  on.exit(default_app()$remove_theme(id), add = TRUE)
  expect_true(id %in% names(default_app()$list_themes()))

  expect_snapshot({
    cli_par(class = "green")
    cli_text(lorem_ipsum())
    cli_end()
  })

  default_app()$remove_theme(id)
  expect_false(id %in% names(default_app()$list_themes()))
})

test_that("default theme is valid", {
  expect_error({
    id <- default_app()$add_theme(builtin_theme())
    default_app()$remove_theme(id)
  }, NA)
})

test_that("explicit formatter is used, and combined", {
  id <- default_app()$add_theme(list(
    "span.emph" = list(
      fmt = function(x) paste0("(((", x, ")))"),
      before = "<<", after = ">>")
    ))
  on.exit(default_app()$remove_theme(id), add = TRUE)
  expect_snapshot(
    cli_text("this is {.emph it}, really")
  )
})

test_that("simple theme", {
  def <- simple_theme()
  expect_true(is.list(def))
  expect_false(is.null(names(def)))
  expect_true(all(names(def) != ""))
  expect_true(all(vlapply(def, is.list)))
})

test_that("user's override", {
  custom <- list(".alert" = list(before = "custom:"))
  override <- list(".alert" = list(after = "override:"))

  expect_snapshot(local({
    start_app(theme = custom, .auto_close = FALSE)
    cli_alert("Alert!")
    stop_app()

    withr::local_options(cli.user_theme = override)
    start_app(theme = custom, .auto_close = FALSE)
    cli_alert("Alert!")
    stop_app()
  }))
})

test_that("theme does not precompute Unicode symbols", {
  withr::local_options(cli.unicode = TRUE, cli.num_colors = 256L)
  start_app()
  msg <- NULL
  withCallingHandlers(
    cli_alert_success("ok"),
    cliMessage = function(m) {
      msg <<- m
      invokeRestart("muffleMessage")
    }
  )
  expect_true(ansi_has_any(msg$message))

  msg2 <- NULL
  withr::local_options(cli.unicode = FALSE, cli.num_colors = 1L)
  withCallingHandlers(
    cli_alert_success("ok2"),
    cliMessage = function(m) {
      msg2 <<- m
      invokeRestart("muffleMessage")
    }
  )
  expect_equal(msg2$message, "v ok2\n")
})

test_that("NULL will undo a style property", {
  expect_snapshot(local({
    cli_alert("this has an arrow")
    cli_div(theme = list(.alert = list(before = NULL)))
    cli_alert("this does not")
  }))
})

test_that_cli(configs = "ansi", "NULL will undo color", {
  expect_snapshot(local({
    cli_alert("{.emph {.val this is blue}}")
    cli_div(theme = list(span.val = list(color = NULL)))
    cli_alert("{.emph {.val this is not}}")
  }))
  expect_snapshot(local({
    cli_alert("{.emph {.val this is blue}}")
    cli_div(theme = list(span.val = list(color = "none")))
    cli_alert("{.emph {.val this is not}}")
  }))
})

withr::local_options(cli.theme = NULL, cli.user_theme = NULL)
withr::local_options(cli.theme_dark = FALSE, cli.num_colors = 256)
start_app()
on.exit(stop_app(), add = TRUE)

test_that_cli(configs = "ansi", "NULL will undo background color", {
  if (packageVersion("testthat") <= "3.1.1") skip("needs newer testthat")
  expect_snapshot(local({
    cli_alert("{.emph {.code this has bg color}}")
    cli_div(theme = list(span = list("background-color" = NULL)))
    cli_alert("{.emph {.code this does not}}")
  }))
  expect_snapshot(local({
    cli_alert("{.emph {.code this has bg color}}")
    cli_div(theme = list(span = list("background-color" = "none")))
    cli_alert("{.emph {.code this does not}}")
  }))
})

start_app()
on.exit(stop_app(), add = TRUE)

test_that("text is wrapped", {
  withr::local_options(cli.width = 60)
  local_rng_version("3.3.0")
  set.seed(42)
  expect_snapshot(local({
    cli_div(class = "testcli", theme = test_style())
    cli_h1("Header")
    cli_text(lorem_ipsum())
  }))
})

test_that("verbatim text is not wrapped", {
  cli_div(class = "testcli", theme = test_style())
  withr::local_options(cli.width = 60)
  suppressMessages(cli_h1("Header"))
  txt <- strrep("1234567890 ", 20)
  out <- capt0(cli_verbatim(txt), strip_style = TRUE)
  expect_equal(out, paste0(txt, "\n"))
})

start_app()
on.exit(stop_app(), add = TRUE)

test_that_cli(
  config = c("plain", "ansi"),
  "quoting phrases that don't start or end with letter or number", {

    expect_snapshot(local({
      x0 <- "good-name"
      cli_text("The name is {.file {x0}}.")

      x <- "weird-name "
      cli_text("The name is {.file {x}}.")
      cli_text("The name is {.path {x}}.")
      cli_text("The name is {.email {x}}.")
    }))
  }
)

test_that_cli(config = c("plain", "ansi"), "quoting weird names, still", {
  nb <- function(x) gsub("\u00a0", " ", x, fixed = TRUE)
  expect_snapshot(local({
    cat_line(nb(quote_weird_name("good")))
    cat_line(nb(quote_weird_name("  bad")))
    cat_line(nb(quote_weird_name("bad  ")))
    cat_line(nb(quote_weird_name("  bad  ")))
  }))
})

test_that_cli(config = c("ansi"), "~/ files are not weird", {
  nb <- function(x) gsub("\u00a0", " ", x, fixed = TRUE)
  expect_snapshot(local({
    cat_line(nb(quote_weird_name("~/good")))
    cat_line(nb(quote_weird_name("~~bad")))
    cat_line(nb(quote_weird_name("bad~  ")))
    cat_line(nb(quote_weird_name(" ~ bad ~ ")))
  }))
})

test_that_cli("custom truncation", {
  expect_snapshot({
    x <- cli_vec(1:100, list(vec_trunc = 5))
    cli_text("Some numbers: {x}.")
    cli_text("Some numbers: {.val {x}}.")
  })
})

test_that_cli(configs = c("plain", "ansi"), "collapsing class names", {
  expect_snapshot(local({
    cc <- c("one", "two")
    cli_text("this is a class: {.cls myclass}")
    cli_text("multiple classes: {.cls {cc}}")
  }))
})

test_that_cli(configs = c("plain", "ansi"), "transform", {
  expect_snapshot(local({
    cli_text("This is a {.field field} (before)")
    foo <- function(x) toupper(x)
    cli_div(theme = list(span.field = list(transform = foo)))
    cli_text("This is a {.field field} (during)")
    cli_end()
    cli_text("This is a {.field field} (after)")
  }))
})

test_that("cli_format", {
  expect_snapshot(
    cli_format(1:4/7, list(digits = 2))
  )
})

test_that("cli_format() is used for .val", {
  withr::local_options(cli.width = 60)
  local_rng_version("3.3.0")
  set.seed(42)
  expect_snapshot({
    cli_div(theme = list(.val = list(digits = 2)))
    cli_text("Some random numbers: {.val {runif(4)}}.")
  })
})

test_that(".q always double quotes", {
  expect_snapshot(
    cli_text("just a {.q string}, nothing more")
  )
})

test_that(".or", {
  expect_snapshot(
    cli_text("{.or {letters[1:5]}}")
  )
  expect_snapshot(
    cli_text("{.or {letters[1:2]}}")
  )
})

test_that("line breaks", {
  txt <- paste(
    "Cupidatat deserunt culpa enim deserunt minim aliqua tempor fugiat",
    "cupidatat laboris officia esse ex aliqua. Ullamco mollit adipisicing",
    "anim."
  )
  txt2 <- paste0(txt, "\f", txt)
  expect_snapshot(ansi_strwrap(txt2, width = 60))
})

test_that_cli(config = "ansi", "double ticks", {
  x <- c("a", "`x`", "b")
  cli_div(theme = list(
    .code = list(color = "red"),
    .fun = list(color = "red")
  ))
  expect_snapshot(format_inline("{.code {x}}"))
  expect_snapshot(format_inline("{.fun {x}}"))
})

parse_iterm <- function(path) {
  doc <- xml2::read_xml(path)
  elm <- xml2::xml_find_first(doc, "/plist/dict")
  key <- sapply(
    xml2::xml_find_all(elm, "/plist/dict/key"),
    xml2::xml_text
  )
  val <- xml2::xml_find_all(elm, "/plist/dict/dict")

  get <- function(k) {
    wh <- match(k, key)
    v <- val[[wh]]
    vks <- sapply(
      xml2::xml_find_all(v, "key"),
      xml2::xml_text
    )
    if (! "Color Space" %in% vks) stop("Unknown color space")
    chd <- xml2::xml_children(v)
    csp <- xml2::xml_text(
      chd[[which(vks == "Color Space") * 2]]
    )
    if (csp != "sRGB") stop("Color space is not sRGB")
    r <- as.numeric(xml2::xml_text(chd[[which(vks == "Red Component") * 2]]))
    g <- as.numeric(xml2::xml_text(chd[[which(vks == "Green Component") * 2]]))
    b <- as.numeric(xml2::xml_text(chd[[which(vks == "Blue Component") * 2]]))
    tolower(grDevices::rgb(r, g, b))
  }

  rn <- sub(".itermcolors", "", basename(path), fixed = TRUE)
  data.frame(
    stringsAsFactors = FALSE,
    row.names = paste0("iterm-", rn),
    black    = get("Ansi 0 Color"),
    red      = get("Ansi 1 Color"),
    green    = get("Ansi 2 Color"),
    yellow   = get("Ansi 3 Color"),
    blue     = get("Ansi 4 Color"),
    magenta  = get("Ansi 5 Color"),
    cyan     = get("Ansi 6 Color"),
    white    = get("Ansi 7 Color"),
    bblack   = get("Ansi 8 Color"),
    bred     = get("Ansi 9 Color"),
    bgreen   = get("Ansi 10 Color"),
    byellow  = get("Ansi 11 Color"),
    bblue    = get("Ansi 12 Color"),
    bmagenta = get("Ansi 13 Color"),
    bcyan    = get("Ansi 14 Color"),
    bwhite   = get("Ansi 15 Color")
  )
}

parse_all <- function(dir = "tools") {
  paths <- dir(dir, pattern = "[.]itermcolors$", full.names = TRUE)
  cols <- lapply(paths, parse_iterm)
  all <- do.call(rbind, cols)
  write.table(all, "tools/ansi-iterm-themes.txt", quote = FALSE)
}

if (is.null(sys.calls())) {
  parse_all()
}

cli_error <- function(..., .data = NULL, .class = NULL, .envir = parent.frame(),
                      call. = TRUE) {
  .hide_from_trace <- TRUE
  cnd <- new_error(
    call. = call.,
    format_error(
      .envir = .envir,
      c(
        ...
      )
    )
  )

  if (length(.data)) cnd[names(.data)] <- .data
  if (length(class)) class(cnd) <- c(.class, class(cnd))

  cnd
}

stop_if_not <- function(message, ..., .envir = parent.frame(),
                        call. = sys.call(-1)) {
  conds <- list(...)
  for (cond in conds) {
    if (!cond) {
      throw(
        new_error(format_error(.envir = .envir, message), call. = call.),
        frame = .envir
      )
    }
  }
}

`%??%` <- function(expr, err) {
  chain_error(expr, err, srcref = utils::getSrcref(sys.call()))
}

if (exists(".knitr_asciicast_process", envir = .GlobalEnv)) {
  rm(list = ".knitr_asciicast_process", envir = .GlobalEnv)
}

asciicast::init_knitr_engine(
  echo = TRUE,
  echo_input = FALSE,
  interactive = TRUE,
  timeout = as.integer(Sys.getenv("ASCIICAST_TIMEOUT", 10)),
  startup = quote({
    options(cli.width = 70)
    options(cli.progress_show_after = 0)
    options(cli.progress_clear = FALSE)
    library(cli)
    set.seed(1)
  })
)

knitr::opts_chunk$set(
  asciicast_knitr_output = "html",
  asciicast_include_style = FALSE,
  asciicast_theme = "pkgdown"
)

list(
  markdown = TRUE,
  knitr_chunk_options = list(
    cache = TRUE,
    cache_lazy = FALSE,
    cache.path = file.path(getwd(), "man/_cache/"),
    fig.path = file.path(getwd(), "man/figures"),
    error = TRUE
  ),
  restrict_image_formats = TRUE
)

rule_class <- function(x) {
  structure(x, class = c("cli_rule", "rule", "cli_ansi_string", "ansi_string", "character"))
}

capture_msgs <- function(expr) {
  msgs <- character()
  i <- 0
  suppressMessages(withCallingHandlers(
    expr,
    message = function(e) msgs[[i <<- i + 1]] <<- conditionMessage(e)))
  paste0(msgs, collapse = "")
}

capture_cli_messages <- function(expr) {
  msgs <- character()
  withCallingHandlers(
    expr,
    cliMessage = function(e) {
      msgs <<- c(msgs, conditionMessage(e))
      invokeRestart("muffleMessage")
    }
  )
  msgs
}

capt <- function(expr, print_it = TRUE) {
  pr <- if (print_it) print else identity
  paste(capture.output(pr(expr)), collapse = "\n")
}

capt0 <- function(expr, strip_style = FALSE) {
  out <- capture_msgs(expr)
  if  (strip_style) ansi_strip(out) else out
}

local_cli_config <- function(unicode = FALSE, dynamic = FALSE,
                             ansi = FALSE, num_colors = 1,
                             .local_envir = parent.frame()) {
  withr::local_options(
    cli.dynamic = dynamic,
    cli.ansi = ansi,
    cli.unicode = unicode,
    crayon.enabled = num_colors > 1,
    crayon.colors = num_colors,
    .local_envir = .local_envir
  )
  withr::local_envvar(
    PKG_OMIT_TIMES = "true",
    PKG_OMIT_SIZES = "true",
    .local_envir = .local_envir
  )
}

test_style <- function() {
  list(
    ".testcli h1" = list(
      "font-weight" = "bold",
      "font-style" = "italic",
      "margin-top" = 1,
      "margin-bottom" = 1),
    ".testcli h2" = list(
      "font-weight" = "bold",
      "margin-top" = 1,
      "margin-bottom" = 1),
    ".testcli h3" = list(
      "text-decoration" = "underline",
      "margin-top" = 1)
  )
}

# to work around https://github.com/r-lib/withr/issues/167
local_rng_version <- function(version, .local_envir = parent.frame()) {
  withr::defer(RNGversion(as.character(getRversion())), envir = .local_envir)
  suppressWarnings(RNGversion(version))
}

fix_times <- function(out) {
  out <- sub("[(][ ]*[.0-9]+ [Mk]B/s[)]", "(8.5 MB/s)", out)
  out <- sub("[(][.0-9]+/s[)]", "(100/s)", out)
  out <- sub(" [.0-9]+(ms|s|m)", " 3ms", out)
  out <- sub("ETA:[ ]*[.0-9]+m?s", "ETA:  1s", out)
  out <- gsub("\\[[.0-9]+m?s\\]", "[1s]", out)
  out
}

fix_logger_output <- function(lines) {
  sub(
    paste0(
      "^[0-9][0-9][0-9][0-9]-[0-9][0-9]-[0-9][0-9]T",
      "[0-9][0-9]:[0-9][0-9]:[0-9][0-9]\\+00:00 ",
      "cli-[0-9]+-[0-9]+ "
    ),
    "2021-06-18T00:09:14+00:00 cli-36434-1 ",
    lines
  )
}

make_c_function <- function(file = NULL,
                            code = NULL,
                            args = character(),
                            type = c(".c", ".cpp"),
                            header = NULL,
                            linkingto = packageName(),
                            quiet = Sys.getenv("TESTTHAT") == "true") {
  type <- match.arg(type)

  # Create source file
  dir.create(dir <- tempfile())
  if (is.null(file)) {
    lines <- create_c_function_call(code, args, header = header)
  } else {
    lines <- readLines(file)
  }
  src <- basename(tempfile(fileext = type))
  writeLines(lines, file.path(dir, src))

  # Compile
  cflags <- ""
  for (pkg in linkingto) {
    pkgdir <- file.path(find.package(pkg), "include")
    lcldir <- file.path(find.package(pkg), "inst", "include")
    cflags <- paste(cflags, "-I", pkgdir, "-I", lcldir)
  }
  env <- c(PKG_CFLAGS = cflags)
  callr::rcmd(
    "SHLIB",
    src,
    wd = file.path(dir),
    env = env,
    echo = !quiet,
    show = !quiet
  )

  # Load DLL
  dllfile <- file.path(dir, sub("[.]c(pp)?$", .Platform$dynlib.ext, src))
  dll <- dyn.load(dllfile, local = TRUE, now = TRUE)

  # TODO: finalizer to unload/delete

  dll
}

create_c_function_call <- function(code, args, header = NULL) {
  c(
    "#include <Rinternals.h>",
    header,
    "SEXP tmp_c_function(",
    if (length(args) > 0) paste0("SEXP ", args, collapse = ", "),
    ") {",
    code,
    "}\n"
  )
}

win2unix <- function (str) {
  gsub("\r\n", "\n", str, fixed = TRUE, useBytes = TRUE)
}

expect_snapshot <- function(...) {
  if (packageVersion("testthat") >= "3.1.1" &&
      packageVersion("testthat") < "3.1.1.9000") {
    skip("testthat bug with snapshots")
  }
  testthat::expect_snapshot(...)
}

test_that("cli_progress_message", {
  withr::local_options(cli.dynamic = FALSE, cli.ansi = FALSE)
  fun <- function() {
    cli_progress_message("Simplest progress 'bar', {.fun fn} {2} two{?s}")
  }
  expect_snapshot(capture_cli_messages(fun()))
})

test_that("cli_progress_message error", {
  # we need the env var as well, because the on.exit handler of the progress
  # bar might run after the on.exit handler that removes the `cli.dynamic`
  # option.
  withr::local_envvar(R_CLI_DYNAMIC = "false")
  fun <- function() {
    suppressWarnings(testthat::local_reproducible_output())
    options(cli.dynamic = FALSE, cli.ansi = FALSE)
    cli::cli_progress_message("Simplest progress 'bar', {.fun fn} {2} two{?s}")
    stop("oopsie")
  }

  outfile <- tempfile()
  on.exit(unlink(outfile), add = TRUE)
  expect_error(callr::r(fun, stdout = outfile, stderr = outfile), "oopsie")
  expect_snapshot(readLines(outfile))

  # we need the env var as well, because the on.exit handler of the progress
  # bar might run after the on.exit handler that removes the `cli.dynamic`
  # option.
  withr::local_envvar(R_CLI_DYNAMIC = "true")
  fun2 <- function() {
    suppressWarnings(testthat::local_reproducible_output())
    options(cli.dynamic = TRUE, cli.ansi = TRUE)
    cli::cli_progress_message("Simplest progress 'bar', {.fun fn} {2} two{?s}")
    stop("oopsie")
  }

  outfile <- tempfile()
  on.exit(unlink(outfile), add = TRUE)
  expect_error(callr::r(fun2, stdout = outfile, stderr = outfile), "oopsie")
  out <- rawToChar(readBin(outfile, "raw", 1000))
  expect_snapshot(win2unix(out))
})

start_app()
on.exit(stop_app(), add = TRUE)

test_that("cli_progress_step", {
  withr::local_options(cli.dynamic = TRUE, cli.ansi = TRUE)
  suppressWarnings(testthat::local_reproducible_output())
  fun <- function() {
    cli_progress_step("First step")
    cli_progress_step("Second step")
  }
  msgs <- fix_times(capture_cli_messages(fun()))
  expect_snapshot(msgs)
})

test_that("cli_progress_step error", {
  if (getRversion() < "3.5.0") skip("Needs R 3.5.0")
  fun <- function() {
    options(
      cli.dynamic = FALSE,
      cli.ansi = FALSE,
      cli.unicode = FALSE,
      cli.width = 80,
      width = 80,
      cli.num_colors = 1
    )
    cli::cli_progress_step("First step")
    cli::cli_progress_step("Second step")
    stop("oopsie")
  }

  outfile <- tempfile()
  on.exit(unlink(outfile), add = TRUE)
  expect_error(callr::r(fun, stdout = outfile, stderr = "2>&1"), "oopsie")
  out <- fix_times(rawToChar(readBin(outfile, "raw", 1000)))
  expect_snapshot(win2unix(out))
})

test_that("c api #1", {
  skip_on_cran()
  withr::local_options(cli.ansi = TRUE, cli.dynamic = TRUE)
  withr::local_options(cli.progress_handlers_only = "cli")

  dll <- make_c_function(test_path("progress-1.c"), linkingto = "cli")
  on.exit(dyn.unload(dll[["path"]]), add = TRUE)

  # simple crud
  out <- capture_cli_messages(cli_with_ticks(
    ret <- .Call(dll$clitest__progress_crud, NULL)
  ))
  out <- fix_times(out)
  expect_snapshot(out)

  # config can be FALSE
  out <- capture_cli_messages(cli_with_ticks(
    ret <- .Call(dll$clitest__progress_crud, FALSE)
  ))
  expect_false(ret)
  expect_equal(out, character())

  # config can override defaults
  out <- capture_cli_messages(cli_with_ticks(
    ret <- .Call(
      dll$clitest__progress_crud,
      list(format = "{cli::pb_current}/{cli::pb_total}")
    )
  ))
  expect_snapshot(out)

  # config must be a named list
  expect_error(
    .Call(dll$clitest__progress_crud, list(123)),
    "list elements must be named"
  )
  expect_error(
    .Call(dll$clitest__progress_crud, 100L),
    "Unknown cli progress bar configuation"
  )

  # config can be a progress bar name
  withr::local_options(
    cli.progress_format_iterator = "{cli::pb_name}{cli::pb_current}"
  )
  out <- capture_cli_messages(cli_with_ticks(
    ret <- .Call(dll$clitest__progress_crud, "me llama")
  ))
  expect_snapshot(out)

  # set various options
  out <- capture_cli_messages(cli_with_ticks(
    ret <- .Call(dll$clitest__progress_sets)
  ))
  expect_snapshot(out)

  # cli_progress_add
  out <- capture_cli_messages(cli_with_ticks(
    ret <- .Call(dll$clitest__progress_add)
  ))
  expect_snapshot(out)

  # cli_progress_num
  before <- .Call(dll$clitest__progress_num)
  bar <- cli_progress_bar()
  after <- .Call(dll$clitest__progress_num)
  cli_progress_done()
  expect_true(before + 1 == after)

  # cli_progress_update
  out <- capture_cli_messages(cli_without_ticks(
    ret <- .Call(dll$clitest__progress_update)
  ))
  expect_snapshot(out)
  out <- capture_cli_messages(cli_without_ticks(
    ret <- .Call(dll$clitest__progress_update2)
  ))
  expect_snapshot(out)
  out <- capture_cli_messages(cli_with_ticks(
    ret <- .Call(dll$clitest__progress_update3)
  ))
  expect_snapshot(out)

  # C progress bars have ids
  withr::local_options(cli.progress_handlers_only = "logger")
  out <- capture.output(cli_with_ticks(
    ret <- .Call(dll$clitest__progress_crud, NULL)
  ))
  out <- fix_logger_output(out)
  expect_snapshot(out)

  # cli_progress_sleep
  tic <- .Call(clic_get_time)
  .Call(dll$clitest__progress_sleep, 0L, 100L * 1000L * 1000L)
  toc <- .Call(clic_get_time)
  expect_true(toc - tic > 0.05)

  # progress vars in format_done
  withr::local_options(cli.progress_handlers_only = "cli")
  out <- capture_cli_messages(cli_with_ticks(
    ret <- .Call(
      dll$clitest__progress_crud,
      list(
        format = "{cli::pb_current}/{cli::pb_total}",
        format_done = "Just did {cli::pb_current} step{?s}.",
        clear = FALSE
      )
    )
  ))
  expect_snapshot(out)
})

test_that("c api #2", {
  skip_on_cran()
  dll <- make_c_function(test_path("progress-2.c"), linkingto = "cli")
  ret <- cli_with_ticks(.Call(dll$clitest__init_timer))
  expect_equal(ret, c(0L, 1L))
})

test_that("clic__find_var", {
  x <- 10
  expect_equal(.Call(clic__find_var, environment(), as.symbol("x")), 10)
  # not inherit
  env <- new.env(parent = environment())
  expect_error(
    .Call(clic__find_var, env, as.symbol("x")),
    "Cannot find variable"
  )
  expect_error(
    .Call(clic__find_var, environment(), as.symbol(basename(tempfile()))),
    "Cannot find variable"
  )
})

test_that("unloading stops the thread", {
  # It is important to skip this on CRAN, because in ASAN we do not
  # kill the tick thread on unload, because it triggers an ASAN crash,
  # which is similar to https://github.com/google/sanitizers/issues/1152
  skip_on_cran()
  fun <- function() {
    before <- ps::ps_num_threads()
    library(cli)
    between <- ps::ps_num_threads()
    unloadNamespace("cli")
    after <- ps::ps_num_threads()
    list(before = before, between = between, after = after)
  }

  out <- callr::r(fun)
  expect_equal(out$between, out$before + 1L)
  expect_equal(out$after, out$between - 1L)
})

test_that("ticking", {
  withr::local_options(
    cli.ansi = TRUE,
    cli.dynamic = TRUE,
    cli.progress_show_after = 0,
    cli.progress_handlers_only = "cli"
  )

  fun <- function() {
    i <- 0L
    while (ticking(i < 10L, total = 10L, name = "ticking", format = "{cli::pb_current}/{cli::pb_total}")) {
      i <- i + 1L
    }
  }
  
  out <- capture_cli_messages(cli_with_ticks(fun()))
  expect_snapshot(out)
})

test_that("cli_progress_along crud", {
  fun <- function() {
    sapply(cli_progress_along(letters), function(i) i)
  }
  capture_cli_messages(ret <- fun())
  expect_identical(ret, seq_along(letters))
})

test_that("progress bar terminated at mapping function exit", {
  fun <- function() {
    snap <- as.character(names(clienv$progress))
    sapply(cli_progress_along(letters), function(i) i)
    expect_identical(as.character(names(clienv$progress)), snap)
  }
  capture_cli_messages(fun())
})

test_that("interpolation uses the right env", {
  if (getRversion() < "3.5.0") skip("Needs ALTREP")
  fun <- function() {
    withr::local_options(
      cli.ansi = TRUE,
      cli.dynamic = TRUE,
      cli.progress_show_after = 0,
      cli.progress_handlers_only = "cli"
    )
    x <- 10
    sapply(cli_progress_along(1:5, format = "x: {x}"), function(i) i)
  }

  out <- capture_cli_messages(cli_with_ticks(fun()))
  expect_snapshot(out)
})

test_that("cli_progress_along", {
  if (getRversion() < "3.5.0") skip("Needs ALTREP")
  withr::local_envvar(CLI_NO_THREAD = "1")
  fun <- function() {
    withr::local_options(
      cli.ansi = TRUE,
      cli.dynamic = TRUE,
      cli.progress_show_after = 0,
      cli.progress_handlers_only = "logger"
    )
    vapply(cli::cli_progress_along(1:10), function(i) i, integer(1))
  }

  lines <- fix_logger_output(capture.output(cli_with_ticks(fun())))
  expect_snapshot(lines)
})

test_that("cli_progress_along error", {
  if (getRversion() < "3.5.0") skip("Needs ALTREP")
  withr::local_envvar(CLI_NO_THREAD = "1")
  fun <- function() {
    withr::local_options(
      cli.ansi = TRUE,
      cli.dynamic = TRUE,
      cli.progress_show_after = 0,
      cli.progress_handlers_only = "logger"
    )
    suppressWarnings(testthat::local_reproducible_output())
    lapply(
      cli::cli_progress_along(1:10, clear = FALSE),
      function(i) { if (i == 5) stop("oops") }
    )
  }

  outfile <- tempfile()
  expect_error(callr::r(fun, stdout = outfile, stderr = outfile))

  lines <- fix_logger_output(readLines(outfile))
  expect_snapshot(lines)
})

test_that("old R is just seq_along", {
  # It is tricky to check that we get seq_along(), because
  # identical(cli_progress_along(1:10), seq_along(1:10)) holds,
  # so we just check that no progress bar is created.
  local_mocked_bindings(getRversion = function() package_version("3.4.0"))
  snapshot <- names(clienv$progress)
  it <- cli_progress_along(1:10)
  expect_identical(snapshot, names(clienv$progress))
  expect_identical(it, seq_along(1:10))
})

test_that("error in handler is a single warning", {
  if (getRversion() < "3.5.0") skip("Needs ALTREP")
  fun <- function() {
    withr::local_options(
      cli.ansi = TRUE,
      cli.dynamic = TRUE,
      cli.progress_show_after = 0,
      cli.progress_handlers_only = "cli"
    )
    x <- 10
    sapply(cli_progress_along(1:5, format = "{1+''}"), function(i) i)
  }

  expect_snapshot(
    cli_with_ticks(fun()),
    transform = sanitize_srcref
  )
})

test_that("length 1 seq", {
  fun <- function() {
    sapply(cli_progress_along(1L), function(i) i)
  }
  capture_cli_messages(ret <- cli_with_ticks(fun()))
  expect_identical(ret, 1L)
})

test_that("ALTREP methods", {
  if (getRversion() < "3.5.0") skip("Needs ALTREP")
  seq <- cli_progress_along(1:10)
  expect_output(.Internal(inspect(seq)), "progress_along")

  expect_equal(is.unsorted(seq), FALSE)
  expect_equal(sum(seq), sum(1:10))

  seq <- cli_progress_along(letters)
  expect_equal(min(seq), 1L)
  expect_equal(max(seq), length(letters))

  z <- cli_progress_along(character())
  expect_equal(min(z), Inf)

  seq <- cli_progress_along(letters)
  expect_equal(.Call(clic_dataptr, seq), seq_along(letters) * 2)

  seq2 <- seq
  expect_silent(seq2[1] <- 100)
})

start_app()
on.exit(stop_app(), add = TRUE)

test_that("iterator", {
  withr::local_options(
    cli.dynamic = TRUE,
    cli.ansi = TRUE,
    cli.spinner = NULL,
    cli.spinner_unicode = NULL,
    cli.progress_format_iterator = NULL,
    cli.progress_format_iterator_nototal= NULL,
    cli.width = Inf
  )

  fun <- function() {
    cli_progress_bar(type = "iterator", total = 100)
    cli_progress_update(set = 50, force = TRUE)
  }
  out <- fix_times(capture_cli_messages(fun()))
  expect_snapshot(out)

  fun <- function() {
    cli_progress_bar(type = "iterator", total = NA)
    cli_progress_update(set = 50, force = TRUE)
  }
  out <- fix_times(capture_cli_messages(fun()))
  expect_snapshot(out)
})

test_that("tasks", {
  skip_on_cran()
  withr::local_options(
    cli.dynamic = TRUE,
    cli.ansi = TRUE,
    cli.spinner = NULL,
    cli.spinner_unicode = NULL,
    cli.progress_format_tasks = NULL,
    cli.progress_format_tasks_nototal= NULL,
    cli.width = Inf
  )

  fun <- function() {
    cli_progress_bar(type = "tasks", total = 100)
    cli_progress_update(set = 50, force = TRUE)
  }
  out <- fix_times(capture_cli_messages(fun()))
  expect_snapshot(out)

  fun <- function() {
    cli_progress_bar(type = "tasks", total = NA)
    cli_progress_update(set = 50, force = TRUE)
  }
  out <- fix_times(capture_cli_messages(fun()))
  expect_snapshot(out)
})

test_that("download", {
  withr::local_options(
    cli.dynamic = TRUE,
    cli.ansi = TRUE,
    cli.spinner = NULL,
    cli.spinner_unicode = NULL,
    cli.progress_format_download = NULL,
    cli.progress_format_download_nototal= NULL
  )

  fun <- function() {
    cli_progress_bar(type = "download", total = 1024 * 1024)
    cli_progress_update(set = 52 * 1024, force = TRUE)
  }
  out <- fix_times(capture_cli_messages(fun()))
  expect_snapshot(out)

  fun <- function() {
    cli_progress_bar(type = "download", total = NA)
    cli_progress_update(set = 52 * 1024, force = TRUE)
  }
  out <- fix_times(capture_cli_messages(fun()))
  expect_snapshot(out)
})

test_that("customize with options, iterator", {
  withr::local_options(
    cli.dynamic = TRUE,
    cli.ansi = TRUE,
    cli.progress_format_iterator = "new format {cli::pb_current}",
    cli.progress_format_iterator_nototal = NULL
  )

  fun <- function(type, total, set = 50) {
    cli_progress_bar(type = type, total = total)
    cli_progress_update(set = set, force = TRUE)
  }

  expect_snapshot(capture_cli_messages(fun("iterator", 100)))
  expect_snapshot(capture_cli_messages(fun("iterator", NA)))

  withr::local_options(
    cli.progress_format_iterator_nototal = "new too {cli::pb_current}"
  )

  expect_snapshot(capture_cli_messages(fun("iterator", 100)))
  expect_snapshot(capture_cli_messages(fun("iterator", NA)))
})

test_that("customize with options, tasks", {
  withr::local_options(
    cli.dynamic = TRUE,
    cli.ansi = TRUE,
    cli.progress_format_tasks = "new format {cli::pb_current}",
    cli.progress_format_tasks_nototal = NULL
  )

  fun <- function(type, total, set = 50) {
    cli_progress_bar(type = type, total = total)
    cli_progress_update(set = set, force = TRUE)
  }

  expect_snapshot(capture_cli_messages(fun("tasks", 100)))
  expect_snapshot(capture_cli_messages(fun("tasks", NA)))

  withr::local_options(
    cli.progress_format_tasks_nototal = "new too {cli::pb_current}"
  )

  expect_snapshot(capture_cli_messages(fun("tasks", 100)))
  expect_snapshot(capture_cli_messages(fun("tasks", NA)))
})

test_that("customize with options, download", {
  withr::local_options(
    cli.dynamic = TRUE,
    cli.ansi = TRUE,
    cli.progress_format_download = "new format {cli::pb_current_bytes}",
    cli.progress_format_download_nototal = NULL
  )

  fun <- function(type, total, set = 50) {
    cli_progress_bar(type = type, total = total)
    cli_progress_update(set = set, force = TRUE)
  }

  expect_snapshot(capture_cli_messages(fun("download", 1024 * 1024, 512 * 1024)))
  expect_snapshot(capture_cli_messages(fun("download", NA, 512 * 1024)))

  withr::local_options(
    cli.progress_format_download_nototal = "new too {cli::pb_current_bytes}"
  )

  expect_snapshot(capture_cli_messages(fun("download", 1024 * 1024, 512 * 1024)))
  expect_snapshot(capture_cli_messages(fun("download", NA, 512 * 1024)))
})

start_app()
on.exit(stop_app(), add = TRUE)

test_that("iterator", {
  withr::local_options(
    cli.dynamic = TRUE,
    cli.ansi = TRUE,
    cli.spinner = NULL,
    cli.spinner_unicode = NULL,
    cli.progress_format_iterator = NULL,
    cli.progress_format_iterator_nototal= NULL
  )

  fun <- function() {
    cli_progress_bar(type = "iterator", total = 100)
    cli_progress_update(set = 50, force = TRUE)
  }
  out <- fix_times(capture_cli_messages(fun()))
  expect_snapshot(out)

  fun <- function() {
    cli_progress_bar(type = "iterator", total = NA)
    cli_progress_update(set = 50, force = TRUE)
  }
  out <- fix_times(capture_cli_messages(fun()))
  expect_snapshot(out)
})

test_that("tasks", {
  withr::local_options(
    cli.dynamic = TRUE,
    cli.ansi = TRUE,
    cli.spinner = NULL,
    cli.spinner_unicode = NULL,
    cli.progress_format_tasks = NULL,
    cli.progress_format_tasks_nototal= NULL
  )

  fun <- function() {
    cli_progress_bar(type = "tasks", total = 100)
    cli_progress_update(set = 50, force = TRUE)
  }
  out <- fix_times(capture_cli_messages(fun()))
  expect_snapshot(out)

  fun <- function() {
    cli_progress_bar(type = "tasks", total = NA)
    cli_progress_update(set = 50, force = TRUE)
  }
  out <- fix_times(capture_cli_messages(fun()))
  expect_snapshot(out)
})

test_that("download", {
  withr::local_options(
    cli.dynamic = TRUE,
    cli.ansi = TRUE,
    cli.spinner = NULL,
    cli.spinner_unicode = NULL,
    cli.progress_format_download = NULL,
    cli.progress_format_download_nototal= NULL
  )

  fun <- function() {
    cli_progress_bar(type = "download", total = 1024 * 1024)
    cli_progress_update(set = 52 * 1024, force = TRUE)
  }
  out <- fix_times(capture_cli_messages(fun()))
  expect_snapshot(out)

  fun <- function() {
    cli_progress_bar(type = "download", total = NA)
    cli_progress_update(set = 52 * 1024, force = TRUE)
  }
  out <- fix_times(capture_cli_messages(fun()))
  expect_snapshot(out)
})

test_that("customize with options, iterator", {
  withr::local_options(
    cli.dynamic = TRUE,
    cli.ansi = TRUE,
    cli.progress_format_iterator = "new format {cli::pb_current}",
    cli.progress_format_iterator_nototal = NULL
  )

  fun <- function(type, total, set = 50) {
    cli_progress_bar(type = type, total = total)
    cli_progress_update(set = set, force = TRUE)
  }

  expect_snapshot(capture_cli_messages(fun("iterator", 100)))
  expect_snapshot(capture_cli_messages(fun("iterator", NA)))

  withr::local_options(
    cli.progress_format_iterator_nototal = "new too {cli::pb_current}"
  )

  expect_snapshot(capture_cli_messages(fun("iterator", 100)))
  expect_snapshot(capture_cli_messages(fun("iterator", NA)))
})

test_that("customize with options, tasks", {
  withr::local_options(
    cli.dynamic = TRUE,
    cli.ansi = TRUE,
    cli.progress_format_tasks = "new format {cli::pb_current}",
    cli.progress_format_tasks_nototal = NULL
  )

  fun <- function(type, total, set = 50) {
    cli_progress_bar(type = type, total = total)
    cli_progress_update(set = set, force = TRUE)
  }

  expect_snapshot(capture_cli_messages(fun("tasks", 100)))
  expect_snapshot(capture_cli_messages(fun("tasks", NA)))

  withr::local_options(
    cli.progress_format_tasks_nototal = "new too {cli::pb_current}"
  )

  expect_snapshot(capture_cli_messages(fun("tasks", 100)))
  expect_snapshot(capture_cli_messages(fun("tasks", NA)))
})

test_that("customize with options, download", {
  withr::local_options(
    cli.dynamic = TRUE,
    cli.ansi = TRUE,
    cli.progress_format_download = "new format {cli::pb_current_bytes}",
    cli.progress_format_download_nototal = NULL
  )

  fun <- function(type, total, set = 50) {
    cli_progress_bar(type = type, total = total)
    cli_progress_update(set = set, force = TRUE)
  }

  expect_snapshot(capture_cli_messages(fun("download", 1024 * 1024, 512 * 1024)))
  expect_snapshot(capture_cli_messages(fun("download", NA, 512 * 1024)))

  withr::local_options(
    cli.progress_format_download_nototal = "new too {cli::pb_current_bytes}"
  )

  expect_snapshot(capture_cli_messages(fun("download", 1024 * 1024, 512 * 1024)))
  expect_snapshot(capture_cli_messages(fun("download", NA, 512 * 1024)))
})

test_that("ticking", {
  withr::local_options(
    cli.ansi = TRUE,
    cli.dynamic = TRUE,
    cli.progress_show_after = 0,
    cli.progress_handlers_only = "cli"
  )

  fun <- function() {
    i <- 0L
    while (
      ticking(
        i < 10L,
        total = 10L,
        name = "ticking",
        format = "{cli::pb_current}/{cli::pb_total}"
      )
    ) {
      i <- i + 1L
    }
  }

  out <- capture_cli_messages(cli_with_ticks(fun()))
  expect_snapshot(out)
})

test_that("cli_progress_message", {
  withr::local_options(cli.dynamic = FALSE, cli.ansi = FALSE)
  fun <- function() {
    cli_progress_message("Simplest progress 'bar', {.fun fn} {2} two{?s}")
  }
  expect_snapshot(capture_cli_messages(fun()))
})

test_that("cli_progress_message error", {
  # we need the env var as well, because the on.exit handler of the progress
  # bar might run after the on.exit handler that removes the `cli.dynamic`
  # option.
  withr::local_envvar(R_CLI_DYNAMIC = "false")
  fun <- function() {
    withr::local_options(cli.dynamic = FALSE, cli.ansi = FALSE)
    suppressWarnings(testthat::local_reproducible_output())
    cli::cli_progress_message("Simplest progress 'bar', {.fun fn} {2} two{?s}")
    stop("oopsie")
  }

  outfile <- tempfile()
  on.exit(unlink(outfile), add = TRUE)
  expect_error(callr::r(fun, stdout = outfile, stderr = outfile), "oopsie")
  expect_snapshot(readLines(outfile))

  # we need the env var as well, because the on.exit handler of the progress
  # bar might run after the on.exit handler that removes the `cli.dynamic`
  # option.
  withr::local_envvar(R_CLI_DYNAMIC = "true")
  fun2 <- function() {
    withr::local_options(cli.dynamic = TRUE, cli.ansi = TRUE)
    suppressWarnings(testthat::local_reproducible_output())
    cli::cli_progress_message("Simplest progress 'bar', {.fun fn} {2} two{?s}")
    stop("oopsie")
  }

  outfile <- tempfile()
  on.exit(unlink(outfile), add = TRUE)
  expect_error(callr::r(fun2, stdout = outfile, stderr = outfile), "oopsie")
  out <- rawToChar(readBin(outfile, "raw", 1000))
  expect_snapshot(win2unix(out))
})

start_app()
on.exit(stop_app(), add = TRUE)

test_that("cli_progress_step", {
  withr::local_options(cli.dynamic = TRUE, cli.ansi = TRUE)
  suppressWarnings(testthat::local_reproducible_output())
  fun <- function() {
    cli_progress_step("First step")
    cli_progress_step("Second step")
  }
  msgs <- fix_times(capture_cli_messages(fun()))
  expect_snapshot(msgs)
})

test_that("cli_progress_step error", {
  if (getRversion() < "3.5.0") skip("Needs R 3.5.0")
  fun <- function() {
    withr::local_options(cli.dynamic = FALSE, cli.ansi = FALSE)
    suppressWarnings(testthat::local_reproducible_output())
    cli::cli_progress_step("First step")
    cli::cli_progress_step("Second step")
    stop("oopsie")
  }

  outfile <- tempfile()
  on.exit(unlink(outfile), add = TRUE)
  expect_error(callr::r(fun, stdout = outfile, stderr = "2>&1"), "oopsie")
  out <- fix_times(rawToChar(readBin(outfile, "raw", 1000)))
  expect_snapshot(win2unix(out))
})

test_that("cli_progress_along crud", {
  fun <- function() {
    sapply(cli_progress_along(letters), function(i) i)
  }
  capture_cli_messages(ret <- fun())
  expect_identical(ret, seq_along(letters))
})

test_that("progress bar terminated at mapping function exit", {
  fun <- function() {
    snap <- as.character(names(clienv$progress))
    sapply(cli_progress_along(letters), function(i) i)
    expect_identical(as.character(names(clienv$progress)), snap)
  }
  capture_cli_messages(fun())
})

test_that("interpolation uses the right env", {
  if (getRversion() < "3.5.0") skip("Needs ALTREP")
  fun <- function() {
    withr::local_options(
      cli.ansi = TRUE,
      cli.dynamic = TRUE,
      cli.progress_show_after = 0,
      cli.progress_handlers_only = "cli"
    )
    x <- 10
    sapply(cli_progress_along(1:5, format = "x: {x}"), function(i) i)
  }

  out <- capture_cli_messages(cli_with_ticks(fun()))
  expect_snapshot(out)
})

test_that("cli_progress_along", {
  if (getRversion() < "3.5.0") skip("Needs ALTREP")
  withr::local_envvar(CLI_NO_THREAD = "1")
  fun <- function() {
    withr::local_options(
      cli.ansi = TRUE,
      cli.dynamic = TRUE,
      cli.progress_show_after = 0,
      cli.progress_handlers_only = "logger"
    )
    vapply(cli::cli_progress_along(1:10), function(i) i, integer(1))
  }

  lines <- fix_logger_output(capture.output(cli_with_ticks(fun())))
  expect_snapshot(lines)
})

test_that("cli_progress_along error", {
  if (getRversion() < "3.5.0") skip("Needs ALTREP")
  withr::local_envvar(CLI_NO_THREAD = "1")
  fun <- function() {
    withr::local_options(
      cli.ansi = TRUE,
      cli.dynamic = TRUE,
      cli.progress_show_after = 0,
      cli.progress_handlers_only = "logger"
    )
    suppressWarnings(testthat::local_reproducible_output())
    lapply(
      cli::cli_progress_along(1:10, clear = FALSE),
      function(i) { if (i == 5) stop("oops") }
    )
  }

  outfile <- tempfile()
  expect_error(callr::r(fun, stdout = outfile, stderr = outfile))

  lines <- fix_logger_output(readLines(outfile))
  expect_snapshot(lines)
})

test_that("old R is just seq_along", {
  # It is tricky to check that we get seq_along(), because
  # identical(cli_progress_along(1:10), seq_along(1:10)) holds,
  # so we just check that no progress bar is created.
  mockery::stub(cli_progress_along, "getRversion", package_version("3.4.0"))
  snapshot <- names(clienv$progress)
  it <- cli_progress_along(1:10)
  expect_identical(snapshot, names(clienv$progress))
  expect_identical(it, seq_along(1:10))
})

test_that("error in handler is a single warning", {
  if (getRversion() < "3.5.0") skip("Needs ALTREP")
  fun <- function() {
    withr::local_options(
      cli.ansi = TRUE,
      cli.dynamic = TRUE,
      cli.progress_show_after = 0,
      cli.progress_handlers_only = "cli"
    )
    x <- 10
    sapply(cli_progress_along(1:5, format = "{1+''}"), function(i) i)
  }

  expect_snapshot(cli_with_ticks(fun()))
})

test_that("length 1 seq", {
  fun <- function() {
    sapply(cli_progress_along(1L), function(i) i)
  }
  capture_cli_messages(ret <- cli_with_ticks(fun()))
  expect_identical(ret, 1L)
})

test_that("ALTREP methods", {
  if (getRversion() < "3.5.0") skip("Needs ALTREP")
  seq <- cli_progress_along(1:10)
  expect_output(.Internal(inspect(seq)), "progress_along")

  expect_equal(is.unsorted(seq), FALSE)
  expect_equal(sum(seq), sum(1:10))

  seq <- cli_progress_along(letters)
  expect_equal(min(seq), 1L)
  expect_equal(max(seq), length(letters))

  z <- cli_progress_along(character())
  expect_equal(min(z), Inf)

  seq <- cli_progress_along(letters)
  expect_equal(.Call(clic_dataptr, seq), seq_along(letters) * 2)

  seq2 <- seq
  expect_silent(seq2[1] <- 100)
})

#' List of items
#'
#' It is often useful to print out a list of items, tasks a function or
#' package performs, or a list of notes.
#'
#' @details
#'
#' Items may be formatted differently, e.g. they can have a prefix symbol.
#' Formatting is specified by the names of `text`, and can be themed.
#' cli creates a `div` element of class `bullets` for the whole bullet list.
#' Each item is another `div` element of class `bullet-<name>`, where
#' `<name>` is the name of the entry in `text`. Entries in `text` without
#' a name create a `div` element of class `bullet-empty`, and if the
#' name is a single space character, the class is `bullet-space`.
#'
#' The built-in theme defines the following item types:
#' * No name: Item without a prefix.
#' * ` `: Indented item.
#' * `*`: Item with a bullet.
#' * `>`: Item with an arrow or pointer.
#' * `v`: Item with a green "tick" symbol, like [cli_alert_success()].
#' * `x`: Item with a ref cross, like [cli_alert_danger()].
#' * `!`: Item with a yellow exclamation mark, like [cli_alert_warning()].
#' * `i`: Info item, like [cli_alert_info()].
#'
#' You can define new item type by simply defining theming for the
#' corresponding `bullet-<name>` classes.
#'
#' ```{asciicast cli-bullets}
#' cli_bullets(c(
#'         "noindent",
#'   " " = "indent",
#'   "*" = "bullet",
#'   ">" = "arrow",
#'   "v" = "success",
#'   "x" = "danger",
#'   "!" = "warning",
#'   "i" = "info"
#' ))
#' ```
#'
#' @param text Character vector of items. See details below on how names
#' are interpreted.
#' @param id Optional id of the `div.bullets` element, can be used in themes.
#' @param class Optional additional class(es) for the `div.bullets` element.
#' @param .envir Environment to evaluate the glue expressions in.
#'
#' @seealso This function supports [inline markup][inline-markup].
#' @family functions supporting inline markup
#' @export

cli_bullets <- function(text, id = NULL, class = NULL,
                     .envir = parent.frame()) {
  cli__message(
    "bullets",
    list(
      text = structure(
        lapply(text, glue_cmd, .envir = .envir),
        names = names(text)
      ),
      id = id,
      class = class
    )
  )
}

#' List of verbatim items
#'
#' `cli_format_bullets_raw()` is similar to [cli_bullets()], but it does
#' not perform any inline styling or glue substitutions in the input.
#'
#' `format_bullets_raw()` returns the output instead of printing it.
#'
#' @param text Character vector of items. See details below on how names
#' are interpreted.
#' @param id Optional id of the `div.bullets` element, can be used in themes.
#' @param class Optional additional class(es) for the `div.bullets` element.
#'
#' @seealso These functions support [inline markup][inline-markup].
#' @seealso See [cli_bullets()] for examples.
#' @family functions supporting inline markup
#' @export

cli_bullets_raw <- function(text, id = NULL, class = NULL) {
  text <- cli_escape(text)
  cli__message(
    "bullets",
    list(
      text = structure(
        lapply(text, glue_no_cmd),
        names = names(text)
      ),
      id = id,
      class = class
    )
  )
}

#' @rdname cli_bullets_raw
#' @export

format_bullets_raw <- function(text, id = NULL, class = NULL) {
  cli_fmt(cli_bullets_raw(text, id, class))
}

#' Frequently Asked Questions
#'
#' @name faq
#' @includeRmd man/chunks/FAQ.Rmd
NULL


docs_progress_c_api <- function() {
  if (file.exists("inst/include/cli/progress.h")) {
    lines <- readLines("inst/include/cli/progress.h")
  } else {
    lines <- readLines("../inst/include/cli/progress.h")
  }

  ## Remove non-matching lines, but leave an empty line between blocks
  ptn <- "^//'[ ]?"
  mtch <- grepl(ptn, lines)
  lines[!mtch] <- ""
  prev <- c("", lines[-length(lines)])
  lines <- lines[mtch | prev != lines]

  ## Remove doc pattern
  lines <- sub(ptn, "", lines)

  tmp <- tempfile(fileext = ".Rmd")
  cat(lines, sep = "\n", file = tmp)
  tmp
}

#' @title The cli progress C API
#' @name progress-c
#' @section The cli progress C API:
#'
#' ```{r include = FALSE, cache = FALSE, child = cli:::docs_progress_c_api()}
#' ```
NULL

#' @title cli environment variables and options
#' @name cli-config
#'
#' @section User facing configuration:
#'
#' ```{r include = FALSE, child = "vignettes/cli-config-user.Rmd"}
#' ```
#'
#' @section Internal configuration:
#'
#' ```{r include = FALSE, child = "vignettes/cli-config-internal.Rmd"}
#' ```
NULL

#! /usr/bin/env Rscript

## To get the async package:
## source("https://install-github.me/r-lib/async")

setup_app <- function() {
  theme <- list("url" = list(color = "blue"))
  app <- cli::start_app(theme = theme, output = "stdout")
}

load_packages <- function() {
  tryCatch({
    library(cli)
    library(async)
    library(docopt) },
    error = function(e) {
      cli_alert_danger("The {.pkg async} and {.pkg docopt} packages are needed!")
      q(save = "no", status = 1)
    })
}

up <- function(urls, timeout = 5) {
  load_packages()
  setup_app()
  chk_url <- async(function(url, ...) {
    http_head(url, ...)$
      then(function(res) {
        if (res$status_code < 300) {
          cli_alert_success("{.url {url}} ({res$times[['total']]}s)")
        } else {
          cli_alert_danger("{.url {url}} (HTTP {res$status_code})")
        }
      })$
      catch(error = function(err) {
        e <- if (grepl("timed out", err$message)) "timed out" else "error"
        cli_alert_danger("{.url {url}} ({e})")
      })
  })

  invisible(synchronise(
    async_map(urls, chk_url, options = list(timeout = timeout))
  ))
}

parse_arguments <- function() {

  "Usage:
  up.R [-t timeout] [URLS ...]
  up.R -h | --help

Options:
  -t timeout   Timeout for giving up on a site, in seconds [default: 5].
  -h --help    Print this help message

Check if web sites are up.
" -> doc

  docopt(doc)
}
  
if (is.null(sys.calls())) {
  load_packages()
  opts <- parse_arguments()
  up(opts$URLS, timeout = as.numeric(opts$t))
}

#' Various handy symbols to use in a command line UI
#'
#' @usage
#' symbol
#'
#' @format A named list, see \code{names(symbol)} for all sign names.
#'
#' @details
#'
#' On Windows they have a fallback to less fancy symbols.
#'
#' `list_symbols()` prints a table with all symbols to the screen.
#'
#' @name symbol
#' @aliases symbol
#' @export symbol
#'
#' @examples
#' cat(symbol$tick, " SUCCESS\n", symbol$cross, " FAILURE\n", sep = "")
#'
#' ## All symbols
#' cat(paste(format(names(symbol), width = 20),
#'   unlist(symbol)), sep = "\n")
NULL

symbol_utf8 <- list(
  "tick" = '\u2714',
  "cross" = '\u2716',
  "star" = '\u2605',
  "square" = '\u2587',
  "square_small" = '\u25FB',
  "square_small_filled" = '\u25FC',
  "circle" = '\u25EF',
  "circle_filled" = '\u25C9',
  "circle_dotted" = '\u25CC',
  "circle_double" = '\u25CE',
  "circle_circle" = '\u24DE',
  "circle_cross" = '\u24E7',
  "circle_pipe" = '\u24be',
  "circle_question_mark" = '?\u20DD',
  "bullet" = '\u2022',
  "dot" = '\u2024',
  "line" = '\u2500',
  "double_line" = "\u2550",
  "ellipsis" = '\u2026',
  "continue" = '\u2026',
  "pointer" = '\u276F',
  "info" = '\u2139',
  "warning" = '\u26A0',
  "menu" = '\u2630',
  "smiley" = '\u263A',
  "mustache" = '\u0DF4',
  "heart" = '\u2665',
  "arrow_up" = '\u2191',
  "arrow_down" = '\u2193',
  "arrow_left" = '\u2190',
  "arrow_right" = '\u2192',
  "radio_on" = '\u25C9',
  "radio_off" = '\u25EF',
  "checkbox_on" = '\u2612',
  "checkbox_off" = '\u2610',
  "checkbox_circle_on" = '\u24E7',
  "checkbox_circle_off" = '\u24BE',
  "fancy_question_mark" = '\u2753',
  "neq" = "\u2260",
  "geq" = "\u2265",
  "leq" = "\u2264",
  "times" = "\u00d7",

  "upper_block_1" = "\u2594",
  "upper_block_4" = "\u2580",

  "lower_block_1" = "\u2581",
  "lower_block_2" = "\u2582",
  "lower_block_3" = "\u2583",
  "lower_block_4" = "\u2584",
  "lower_block_5" = "\u2585",
  "lower_block_6" = "\u2586",
  "lower_block_7" = "\u2587",
  "lower_block_8" = "\u2588",

  "full_block" = "\u2588",

  "sup_0" = "\u2070",
  "sup_1" = "\u00b9",
  "sup_2" = "\u00b2",
  "sup_3" = "\u00b3",
  "sup_4" = "\u2074",
  "sup_5" = "\u2075",
  "sup_6" = "\u2076",
  "sup_7" = "\u2077",
  "sup_8" = "\u2078",
  "sup_9" = "\u2079",

  "sup_minus" = "\u207b",
  "sup_plus" = "\u207a",

  "play" = "\u25b6",
  "stop" = "\u25a0",
  "record" = "\u25cf",

  "figure_dash" = "\u2012",
  "en_dash" = "\u2013",
  "em_dash" = "\u2014",

  "dquote_left" = "\u201c",
  "dquote_right" = "\u201d",
  "squote_left" = "\u2018",
  "squote_right" = "\u2019"
)

symbol_rstudio <- symbol_utf8
symbol_rstudio$tick <- "\u2713"
symbol_rstudio$cross <- "x"
symbol_rstudio$star <- "\u066d"
symbol_rstudio$square # ???
symbol_rstudio$square_small # ???
symbol_rstudio$suare_small_filled # ???
symbol_rstudio$circle_circle # ???
symbol_rstudio$circle_cross # ???
symbol_rstudio$circle_pipe # ???
symbol_rstudio$circle_question_mark # ???
symbol_rstudio$pointer <- ">"
symbol_rstudio$warning <- "!"
symbol_rstudio$menu # ???
symbol_rstudio$mustache # ???
symbol_rstudio$checkbox_circle_on # ???
symbol_rstudio$checkbox_circle_off # ???
symbol_rstudio$fancy_question_mark # ???

symbol_win <- list(
  "tick" = '\u221A',
  "cross" = 'x',
  "star" = '*',
  "square" = '\u2588',
  "square_small" = '[ ]',
  "square_small_filled" = '[\u2588]',
  "circle" = '( )',
  "circle_filled" = '(*)',
  "circle_dotted" = '( )',
  "circle_double" = '(o)',
  "circle_circle" = '(o)',
  "circle_cross" = '(x)',
  "circle_pipe" = '(|)',
  "circle_question_mark" = '(?)',
  "bullet" = '*',
  "dot" = '.',
  "line" = '-',
  "double_line" = "=",
  "ellipsis" = '...',
  "continue" = '~',
  "pointer" = '>',
  "info" = 'i',
  "warning" = '\u203C',
  "menu" = '\u2261',
  "smiley" = '\u263A',
  "mustache" = '\u250C\u2500\u2510',
  "heart" = '\u2665',
  "arrow_up" = '^',
  "arrow_down" = 'v',
  "arrow_left" = '<',
  "arrow_right" = '>',
  "radio_on" = '(*)',
  "radio_off" = '( )',
  "checkbox_on" = '[x]',
  "checkbox_off" = '[ ]',
  "checkbox_circle_on" = '(x)',
  "checkbox_circle_off" = '( )',
  "fancy_question_mark" = "(?)",
  "neq" = "!=",
  "geq" = ">=",
  "leq" = "<=",
  "times" = "x",

  "upper_block_1" = "^",
  "upper_block_4" = "^",

  "lower_block_1" = ".",
  "lower_block_2" = "_",
  "lower_block_3" = "_",
  "lower_block_4" = "=",
  "lower_block_5" = "=",
  "lower_block_6" = "*",
  "lower_block_7" = "\u2588",
  "lower_block_8" = "\u2588",

  "full_block" = "\u2588",

  "sup_0" = "0",
  "sup_1" = "1",
  "sup_2" = "2",
  "sup_3" = "3",
  "sup_4" = "4",
  "sup_5" = "5",
  "sup_6" = "6",
  "sup_7" = "7",
  "sup_8" = "8",
  "sup_9" = "9",

  "sup_minus" = "-",
  "sup_plus" = "+",

  "play" = ">",
  "stop" = "#",
  "record" = "o",

  "figure_dash" = "-",
  "en_dash" = "--",
  "em_dash" = "---",

  "dquote_left" = "\"",
  "dquote_right" = "\"",
  "squote_left" = "'",
  "squote_right" = "'"
)

symbol_ascii <- list(
  "tick" = 'v',
  "cross" = 'x',
  "star" = '*',
  "square" = '[ ]',
  "square_small" = '[ ]',
  "square_small_filled" = '[x]',
  "circle" = '( )',
  "circle_filled" = '(*)',
  "circle_dotted" = '( )',
  "circle_double" = '(o)',
  "circle_circle" = '(o)',
  "circle_cross" = '(x)',
  "circle_pipe" = '(|)',
  "circle_question_mark" = '(?)',
  "bullet" = '*',
  "dot" = '.',
  "line" = '-',
  "double_line" = "=",
  "ellipsis" = '...',
  "continue" = '~',
  "pointer" = '>',
  "info" = 'i',
  "warning" = '!',
  "menu" = '=',
  "smiley" = ':)',
  "mustache" = '/\\/',
  "heart" = '<3',
  "arrow_up" = '^',
  "arrow_down" = 'v',
  "arrow_left" = '<',
  "arrow_right" = '>',
  "radio_on" = '(*)',
  "radio_off" = '( )',
  "checkbox_on" = '[x]',
  "checkbox_off" = '[ ]',
  "checkbox_circle_on" = '(x)',
  "checkbox_circle_off" = '( )',
  "fancy_question_mark" = "(?)",
  "neq" = "!=",
  "geq" = ">=",
  "leq" = "<=",
  "times" = "x",

  "upper_block_1" = "^",
  "upper_block_4" = "^",

  "lower_block_1" = ".",
  "lower_block_2" = "_",
  "lower_block_3" = "_",
  "lower_block_4" = "=",
  "lower_block_5" = "=",
  "lower_block_6" = "*",
  "lower_block_7" = "#",
  "lower_block_8" = "#",

  "full_block" = "#",

  "sup_0" = "0",
  "sup_1" = "1",
  "sup_2" = "2",
  "sup_3" = "3",
  "sup_4" = "4",
  "sup_5" = "5",
  "sup_6" = "6",
  "sup_7" = "7",
  "sup_8" = "8",
  "sup_9" = "9",

  "sup_minus" = "-",
  "sup_plus" = "+",

  "play" = ">",
  "stop" = "#",
  "record" = "o",

  "figure_dash" = "-",
  "en_dash" = "--",
  "em_dash" = "---",

  "dquote_left" = "\"",
  "dquote_right" = "\"",
  "squote_left" = "'",
  "squote_right" = "'"
)

#' @export
#' @rdname symbol

list_symbols <- function() {
  rpad <- function(x, width) {
    w <- nchar(x, type = "width")
    paste0(x, strrep(" ", width - w))
  }
  chars <- rpad(paste0(symbol, "\t", names(symbol)), 25)
  if (length(chars) %% 2) chars <- c(chars, "")
  chars <- paste(
  sep = "   ",
  chars[1:(length(chars)/2)],
  chars[(length(chars)/2 + 1):length(chars)])
  cat(chars, sep = "\n")
}

#! /usr/bin/env Rscript

setup_app <- function() {
  theme <- list(
    "url" = list(color = "blue"),
    ".pkg" = list(color = "orange"))
  start_app(theme = theme, output = "stdout")
}

load_packages <- function() {
  tryCatch({
    library(cliapp)
    library(pkgsearch)
    library(docopt)
    library(prettyunits)
    error = function(e) {
      cli_alert_danger(
            "The {.pkg pkgsearch}, {.pkg prettyunits} and {.pkg docopt} packages are needed!")
      q(save = "no", status = 1)
    }
  })
}

search <- function(terms, from = 1, size = 5) {
  load_packages()
  setup_app()
  term <- paste(encodeString(quote = '"', terms), collapse = " ")
  result <- do_query(term, from = from, size = size)
  format_result(result, from = from, size = size)
  invisible()
}

`%||%` <- function(l, r) if (is.null(l)) r else l

do_query <- function(query, from, size) {
  cli_alert_info("Searching...")
  pkg_search(query, from = from, size = size)
}

format_result <- function(obj, from, size) {
  meta <- attr(obj, "metadata")
  if (!meta$total) {
    cli_alert_danger("No results :(")
    return()
  }

  cli_alert_success("Found {meta$total} packages in {pretty_ms(meta$took)}")
  cli_text()
  cli_div(theme = list(ul = list("list-style-type" = "")))
  cli_ol()

  lapply(seq_len(nrow(obj)), function(i) format_hit(obj[i,]))
}

format_hit <- function(hit) {
  ago <- vague_dt(Sys.time() - hit$date)
  cli_li()
  cli_text("{.pkg {hit$package}} {hit$version}  --
          {.emph {hit$title}}")
  cli_par()
  cli_text(hit$description)
  cli_text("{.emph {ago} by {hit$maintainer_name}}")
}

parse_arguments <- function() {
  "Usage:
    cransearch.R [-h | --help] [ -f from ] [ -n size ] <term>...

Options:
    -h --help  Print this help message
    -f first   First hit to include
    -n size    Number of hits to include

Seach for CRAN packages on r-pkg.org
  " -> doc
  docopt(doc)
}

if (is.null(sys.calls())) {
  load_packages()
  opts <- parse_arguments()
  search(opts$term,
         from = as.numeric(opts$f %||% 1),
         size = as.numeric(opts$n %||% 5))
}

format_time_ago <- local({

  e <- expression

  `%s%` <- function(lhs, rhs) {
    assert_string(lhs)
    do.call(
      sprintf,
      c(list(lhs), as.list(rhs))
    )
  }

  assert_string <- function(x) {
    stopifnot(is.character(x), length(x) == 1L)
  }

  assert_diff_time <- function(x) {
    stopifnot(inherits(x, "difftime"))
  }

  vague_dt_default <- list(
    list(c = e(seconds < 10), s = "moments ago"),
    list(c = e(seconds < 45), s = "less than a minute ago"),
    list(c = e(seconds < 90), s = "about a minute ago"),
    list(c = e(minutes < 45), s = e("%d minutes ago" %s% round(minutes))),
    list(c = e(minutes < 90), s = "about an hour ago"),
    list(c = e(hours < 24),   s = e("%d hours ago" %s% round(hours))),
    list(c = e(hours < 42),   s = "a day ago"),
    list(c = e(days < 30),    s = e("%d days ago" %s% round(days))),
    list(c = e(days < 45),    s = "about a month ago"),
    list(c = e(days < 335),   s = e("%d months ago" %s% round(days / 30))),
    list(c = e(years < 1.5),  s = "about a year ago"),
    list(c = TRUE,            s = e("%d years ago" %s% round(years)))
  )

  vague_dt_short <- list(
    list(c = e(seconds < 50), s = "<1 min"),
    list(c = e(minutes < 50), s = e("%d min" %s% round(minutes))),
    list(c = e(hours < 1.5),  s = "1 hour"),
    list(c = e(hours < 18),   s = e("%d hours" %s% round(hours))),
    list(c = e(hours < 42),   s = "1 day"),
    list(c = e(days < 30),    s = e("%d day" %s% round(days))),
    list(c = e(days < 45),    s = "1 mon"),
    list(c = e(days < 335),   s = e("%d mon" %s% round(days / 30))),
    list(c = e(years < 1.5),  s = "1 year"),
    list(c = TRUE,            s = e("%d years" %s% round(years)))
  )

  vague_dt_terse <- list(
    list(c = e(seconds < 50), s = e("%2ds" %s% round(seconds))),
    list(c = e(minutes < 50), s = e("%2dm" %s% round(minutes))),
    list(c = e(hours < 18),   s = e("%2dh" %s% round(hours))),
    list(c = e(days < 30),    s = e("%2dd" %s% round(days))),
    list(c = e(days < 335),   s = e("%2dM" %s% round(days / 30))),
    list(c = TRUE,            s = e("%2dy" %s% round(years)))
  )

  vague_dt_formats <- list(
    "default" = vague_dt_default,
    "short" = vague_dt_short,
    "terse" = vague_dt_terse
  )

  time_ago <- function(date, format = c("default", "short", "terse")) {

    date <- as.POSIXct(date)

    if (length(date) > 1) return(sapply(date, time_ago, format = format))

    seconds <- difftime(Sys.time(), date, units = "secs")

    vague_dt(seconds, format = format)
  }

  vague_dt <- function(dt, format = c("default", "short", "terse")) {

    assert_diff_time(dt)

    units(dt) <- "secs"
    seconds <- as.vector(dt)

    ## Simplest to quit here for empty input
    if (!length(seconds)) return(character())

    pieces <- list(
      minutes = seconds / 60,
      hours = seconds / 60 / 60,
      days = seconds / 60 / 60 / 24,
      years = seconds / 60 / 60 / 24 / 365.25
    )

    format <- match.arg(format)

    for (p in vague_dt_formats[[format]]) {
      if (eval(p$c, pieces)) return(eval(p$s, pieces))
    }
  }

  structure(
    list(
      .internal = environment(),
      time_ago = time_ago,
      vague_dt = vague_dt
    ),
    class = c("standalone_time_ago", "standalone")
  )
})

#' Format an error, warning or diagnostic message
#'
#' You can then throw this message with [stop()] or `rlang::abort()`.
#'
#' The messages can use inline styling, pluralization and glue
#' substitutions.
#'
#' ```{asciicast format-error}
#' n <- "boo"
#' stop(format_error(c(
#'         "{.var n} must be a numeric vector",
#'   "x" = "You've supplied a {.cls {class(n)}} vector."
#' )))
#' ```
#'
#' ```{asciicast format-error-2}
#' len <- 26
#' idx <- 100
#' stop(format_error(c(
#'         "Must index an existing element:",
#'   "i" = "There {?is/are} {len} element{?s}.",
#'   "x" = "You've tried to subset element {idx}."
#' )))
#' ```
#'
#' @param message It is formatted via a call to [cli_bullets()].
#' @param .envir Environment to evaluate the glue expressions in.
#'
#' @seealso These functions support [inline markup][inline-markup].
#' @family functions supporting inline markup
#' @export

format_error <- function(message, .envir = parent.frame()) {
  if (length(message) > 0 &&
      (is.null(names(message)) || names(message)[1] == "")) {
    # The default theme will make this bold
    names(message)[1] <- "1"
  }

  if (length(message) > 0) {
    message[1] <- paste0("Error: ", message[1])
  }

  rsconsole <- c("rstudio_console", "rstudio_console_starting")
  oldopt <- options(
    cli.width = getOption("cli.condition_width") %||% getOption("cli.width")
  )
  on.exit(options(oldopt), add =TRUE)

  # We need to create a frame here, so cli_div() is closed.
  # Cannot use local(), it does not work in snapshot tests, it potentially
  # has issues elsewhere as well.
  formatted1 <- cli_fmt((function() {
    cli_div(class = "cli_rlang cli_abort", theme = cnd_theme())
    cli_bullets(message, .envir = .envir)
  })(), collapse = TRUE, strip_newline = TRUE)

  # remove "Error: " that was only needed for the wrapping
  formatted1[1] <- sub("Error:[ ]?", "", formatted1[1])

  update_rstudio_color(formatted1)
}

#' @rdname format_error
#' @export

format_warning <- function(message, .envir = parent.frame()) {
  if (length(message) > 0 &&
      (is.null(names(message)) || names(message)[1] == "")) {
    # The default theme will make this bold
    names(message)[1] <- "1"
  }

  oldopt <- options(
    cli.width = getOption("cli.condition_width") %||% getOption("cli.width")
  )
  on.exit(options(oldopt), add = TRUE)

  formatted1 <- cli_fmt((function() {
    cli_div(class = "cli_rlang cli_warn", theme = cnd_theme())
    cli_bullets(message, .envir = .envir)
  })(), collapse = TRUE, strip_newline = TRUE)

  update_rstudio_color(formatted1)
}

#' @rdname format_error
#' @export

format_message <- function(message, .envir = parent.frame()) {
  oldopt <- options(
    cli.width = getOption("cli.condition_width") %||% getOption("cli.width")
  )
  on.exit(options(oldopt), add = TRUE)
  formatted1 <- cli_fmt((function() {
    cli_div(class = "cli_rlang cli_inform", theme = cnd_theme())
    cli_bullets(message, .envir = .envir)
  })(), collapse = TRUE, strip_newline = TRUE)
  update_rstudio_color(formatted1)
}

update_rstudio_color <- function(message) {
  rscol <- get_rstudio_fg_color()
  if (!is.null(rscol)) {
    # in RStudio we only need to change the color
    message[] <- rscol(message)
  } else {
    # in a terminal we need to undo the bold
    message <- paste0(style_bold(""), message)
  }
  message
}

get_rstudio_fg_color <- function() {
  tryCatch(
    get_rstudio_fg_color0(),
    error = function(e) NULL
  )
}

get_rstudio_fg_color0 <- function() {
  rs <- rstudio_detect()
  oktypes <- c("rstudio_console", "rstudio_console_starting")
  if (! rs$type %in% oktypes) return(NULL)
  if (rs$num_colors == 1) return(NULL)
  colstr <- get_rstudio_theme()$foreground
  if (is.null(colstr)) return(NULL)
  colstr0 <- substr(colstr, 5, nchar(colstr) - 1)
  rgbnum <- scan(text = colstr0, sep = ",", quiet = TRUE)
  rgb <- grDevices::rgb(rgbnum[1]/255, rgbnum[2]/255, rgbnum[3]/255)
  make_ansi_style(rgb)
}

rstudio_detect <- function() {
  rstudio$detect()
}

cnd_theme <- function() {
  list(
    ".cli_rlang .bullets .bullet-v" = list(
      before = function(x) paste0(col_green(cnd_symb("tick")), " ")
    ),
    ".bullets .bullet-x" = list(
      before = function(x) paste0(col_red(cnd_symb("cross")), " ")
    ),
    ".bullets .bullet-i" = list(
      before = function(x) paste0(col_cyan(cnd_symb("info")), " ")
    ),
    ".bullets .bullet-*" = list(
      before = function(x) paste0(col_cyan(cnd_symb("bullet")), " ")
    ),
    ".bullets .bullet->" = list(
      before = function(x) paste0(cnd_symb("arrow_right"), " ")
    )
  )
}

cnd_symb <- function(name) {
  opt <- getOption("cli.condition_unicode_bullets", NULL)
  if (isTRUE(opt)) {
    symbol_utf8[[name]]
  } else if (isFALSE(opt)) {
    symbol_ascii[[name]]
  } else {
    symbol[[name]]
  }
}

test_that_cli("format_error", {
  expect_snapshot(error = TRUE, local({
    n <- "boo"
    stop(format_error(c(
            "{.var n} must be a numeric vector",
      "x" = "You've supplied a {.cls {class(n)}} vector."
    )))
  }))

  expect_snapshot(error = TRUE, local({
    len <- 26
    idx <- 100
    stop(format_error(c(
            "Must index an existing element:",
      "i" = "There {?is/are} {len} element{?s}.",
      "x" = "You've tried to subset element {idx}."
    )))
  }))
})

test_that_cli("format_warning", {
  expect_snapshot({
    n <- "boo"
    warning(format_warning(c(
            "{.var n} must be a numeric vector",
      "x" = "You've supplied a {.cls {class(n)}} vector."
    )))
  })

  expect_snapshot(local({
    len <- 26
    idx <- 100
    warning(format_warning(c(
            "Must index an existing element:",
      "i" = "There {?is/are} {len} element{?s}.",
      "x" = "You've tried to subset element {idx}."
    )))
  }))
})

test_that_cli("format_message", {
  expect_snapshot({
    n <- "boo"
    message(format_message(c(
            "{.var n} must be a numeric vector",
      "x" = "You've supplied a {.cls {class(n)}} vector."
    )))
  })

  expect_snapshot(local({
    len <- 26
    idx <- 100
    message(format_message(c(
            "Must index an existing element:",
      "i" = "There {?is/are} {len} element{?s}.",
      "x" = "You've tried to subset element {idx}."
    )))
  }))
})

test_that_cli(configs = "ansi", "color in RStudio", {
  local_mocked_bindings(
    rstudio_detect = function() list(type = "rstudio_console", num_colors = 256),
    get_rstudio_theme = function() list(foreground = "rgb(0, 0, 0)")
  )
  expect_snapshot({
    col <- get_rstudio_fg_color0()
    cat(col("this is the new color"))
  })

  local_mocked_bindings(get_rstudio_theme = function() list())
  expect_null(get_rstudio_fg_color0())

  local_mocked_bindings(
    rstudio_detect = function() list(type = "rstudio_console", num_colors = 1)
  )
  expect_null(get_rstudio_fg_color0())
})

test_that_cli(configs = "ansi", "update_rstudio_color", {
  local_mocked_bindings(
    get_rstudio_fg_color = function() make_ansi_style("#008800")
  )
  expect_snapshot(cat(update_rstudio_color("color me interested")))
})

test_that("named first element", {
  expect_snapshot(
    format_error(c("*" = "foo", "*" = "bar"))
  )
  expect_snapshot(
    format_warning(c("*" = "foo", "*" = "bar"))
  )
})

test_that("no cli conditions are thrown", {
  cnd <- NULL
  withCallingHandlers({
    format_error("error")
    format_warning("warning")
    format_message("message")
  }, cli_message = function(cnd_) cnd <<- cnd_)

  expect_null(cnd)
})

test_that("cli.condition_width", {
  withr::local_options(cli.condition_width = 40, cli.num_colors = 1)
  msg <- strrep("1234567890 ", 8)
  expect_snapshot({
    format_error(msg)
    format_warning(msg)
    format_message(msg)
  })

  withr::local_options(cli.condition_width = Inf)
  expect_snapshot({
    format_error(msg)
    format_warning(msg)
    format_message(msg)
  })
})

test_that_cli("suppressing Unicode bullets", {
  withr::local_options(cli.condition_unicode_bullets = FALSE)
  expect_snapshot(error = TRUE, local({
    n <- "boo"
    stop(format_error(c(
            "{.var n} must be a numeric vector",
      "x" = "You've supplied a {.cls {class(n)}} vector.",
      "v" = "Success.",
      "i" = "Info.",
      "*" = "Bullet",
      ">" = "Arrow"
    )))
  }))
})

test_that("edge cases", {
  expect_equal(cli::format_error(""), "")
  expect_equal(cli::format_error(NULL), "")
  expect_equal(cli::format_error(character()), "")

  expect_equal(cli::format_warning(""), "")
  expect_equal(cli::format_warning(NULL), "")
  expect_equal(cli::format_warning(character()), "")

  expect_equal(cli::format_message(""), "")
  expect_equal(cli::format_message(NULL), "")
  expect_equal(cli::format_message(character()), "")
})

test_that_cli("format_error", {
  expect_snapshot(error = TRUE, local({
    n <- "boo"
    stop(format_error(c(
            "{.var n} must be a numeric vector",
      "x" = "You've supplied a {.cls {class(n)}} vector."
    )))
  }))

  expect_snapshot(error = TRUE, local({
    len <- 26
    idx <- 100
    stop(format_error(c(
            "Must index an existing element:",
      "i" = "There {?is/are} {len} element{?s}.",
      "x" = "You've tried to subset element {idx}."
    )))
  }))
})

test_that_cli("format_warning", {
  expect_snapshot({
    n <- "boo"
    warning(format_warning(c(
            "{.var n} must be a numeric vector",
      "x" = "You've supplied a {.cls {class(n)}} vector."
    )))
  })

  expect_snapshot(local({
    len <- 26
    idx <- 100
    warning(format_warning(c(
            "Must index an existing element:",
      "i" = "There {?is/are} {len} element{?s}.",
      "x" = "You've tried to subset element {idx}."
    )))
  }))
})

test_that_cli("format_message", {
  expect_snapshot({
    n <- "boo"
    message(format_message(c(
            "{.var n} must be a numeric vector",
      "x" = "You've supplied a {.cls {class(n)}} vector."
    )))
  })

  expect_snapshot(local({
    len <- 26
    idx <- 100
    message(format_message(c(
            "Must index an existing element:",
      "i" = "There {?is/are} {len} element{?s}.",
      "x" = "You've tried to subset element {idx}."
    )))
  }))
})

test_that("format_error width in RStudio", {
  mockery::stub(format_error, "rstudio_detect", list(type = "rstudio_console"))
  local_rng_version("3.3.0")
  set.seed(42)
  expect_snapshot(error = TRUE, local({
    len <- 26
    idx <- 100
    stop(format_error(c(
            lorem_ipsum(1, 3),
      "i" = lorem_ipsum(1, 3),
      "x" = lorem_ipsum(1, 3)
    )))
  }))
})

test_that_cli(config = "ansi", "color in RStudio", {
  mockery::stub(
    get_rstudio_fg_color0,
    "rstudio_detect",
    list(type = "rstudio_console", num_colors = 256)
  )
  mockery::stub(
    get_rstudio_fg_color0,
    "rstudioapi::getThemeInfo",
    list(foreground = "rgb(0, 0, 0)")
  )
  expect_snapshot({
    col <- get_rstudio_fg_color0()
    cat(col("this is the new color"))
  })

  mockery::stub(
    get_rstudio_fg_color0,
    "rstudioapi::getThemeInfo",
    list()
  )
  expect_null(get_rstudio_fg_color0())

  mockery::stub(
    get_rstudio_fg_color0,
    "rstudio_detect",
    list(type = "rstudio_console", num_colors = 1)
    )
  expect_null(get_rstudio_fg_color0())
})

test_that_cli(config = "ansi", "update_rstudio_color", {
  mockery::stub(
    update_rstudio_color,
    "get_rstudio_fg_color",
    function() make_ansi_style("#008800")
  )
  expect_snapshot(cat(update_rstudio_color("color me interested")))
})

test_that("named first element", {
  expect_snapshot(
    format_error(c("*" = "foo", "*" = "bar"))
  )
  expect_snapshot(
    format_warning(c("*" = "foo", "*" = "bar"))
  )
})

test_that("no cli conditions are thrown", {
  cnd <- NULL
  withCallingHandlers({
    format_error("error")
    format_warning("warning")
    format_message("message")
  }, cli_message = function(cnd_) cnd <<- cnd_)

  expect_null(cnd)
})

test_that("cli.condition_width", {
  withr::local_options(cli.condition_width = 40, cli.num_colors = 1)
  msg <- strrep("1234567890 ", 8)
  expect_snapshot({
    format_error(msg)
    format_warning(msg)
    format_message(msg)
  })

  withr::local_options(cli.condition_width = Inf)
  expect_snapshot({
    format_error(msg)
    format_warning(msg)
    format_message(msg)
  })
})

test_that_cli("suppressing Unicode bullets", {
  withr::local_options(cli.condition_unicode_bullets = FALSE)
  expect_snapshot(error = TRUE, local({
    n <- "boo"
    stop(format_error(c(
            "{.var n} must be a numeric vector",
      "x" = "You've supplied a {.cls {class(n)}} vector.",
      "v" = "Success.",
      "i" = "Info.",
      "*" = "Bullet",
      ">" = "Arrow"
    )))
  }))
})

test_that("edge cases", {
  expect_equal(cli::format_error(""), "")
})

if (getRversion() >= "2.15.1") utils::globalVariables("app")

inline_generic <- function(app, x, style) {

  if (is.character(x) && any(grepl("\n", x))) {
    if (getOption("cli.warn_inline_newlines", FALSE)) {
      warning("cli replaced newlines within {. ... } with spaces")
    }
    x <- gsub_("\n", " ", x, useBytes = TRUE, fixed = TRUE)
  }

  before <- call_if_fun(style$before)
  after <- call_if_fun(style$after)
  transform <- style$transform
  if (is.function(transform)) {
    if (length(formals(transform)) == 1) {
      x <- transform(x)
    } else {
      x <- transform(x, app = app, style = style)
    }
  }
  collapse <- style$collapse
  if (is.character(collapse)) {
    x <- paste0(x, collapse = collapse[1])
  }
  if (is.function(collapse)) {
    x <- collapse(x)
  }
  xx <- paste0(before, x, after)
  fmt <- style$fmt
  if (!is.null(fmt) && is.function(fmt)) {
    if (length(formals(fmt)) == 1) {
      xx <- vcapply(xx, fmt)
    } else {
      xx <- vcapply(xx, fmt, app = app, style = style)
    }
  }
  prefix <- call_if_fun(style$prefix)
  postfix <- call_if_fun(style$postfix)

  paste0(prefix, xx, postfix)
}

inline_collapse <- function(x, style = list()) {
  last <- style[["vec-last"]] %||% style[["vec_last"]] %||% ", and "
  sep <- style[["vec-sep"]] %||% style[["vec_sep"]] %||% ", "
  sep2 <- style[["vec-sep2"]] %||% style[["vec_sep2"]] %||% sub("^,", "", last)

  trunc <- style[["vec-trunc"]] %||% style[["vec_trunc"]] %||% 20L
  col_style <- style[["vec-trunc-style"]] %||% "both-ends"

  ansi_collapse(
    x,
    sep = sep,
    sep2 = sep2,
    last = last,
    trunc = trunc,
    style = col_style
  )
}

#' This glue transformer performs the inline styling of cli
#'
#' The two rules are the following:
#' * If `code` is a class expression (i.e. it starts with a dot),
#'   then we open a `<span>` with the right class and recurse.
#' * If `code` is not a class expression (i.e. it does not start with a
#'   dor), then we collapse vectors.
#'
#' The rest of the work is about:
#' * Making sure that the outside container's non-inherited style is _not_
#'   applied to the substitution. E.g. in
#'   ```r
#'   cli_h2("This is heading number {n}")
#'   ```
#'   The `before` and `after`, etc. style attributes should not be applied
#'   to `{n}`.
#' * While making sure that the inlide style's non-interited style is
#'   applied to brace expressions. I.e. in
#'   ```r
#'   cli_text("{.fun {x}}")
#'   ```
#'   every element of `x` should be formatted as a function name.
#' * Adding extra classes from the `class-map` to the right `<span>`.
#' * Adding extra styles from [cli_vec()] to the right `<span>`.
#'
#' A class expression is a brace expression if it is of the form:
#' ```
#' {.class {expr}}
#' ```
#' I.e. `{expr}` must be a single glue substitution which is not a class
#' expression. These are treated differently internally, because they do
#' collapsing, and the styling must be applied before collapsing. For other
#' expressions the styling is applied after collapsing. E.g.
#' ```r
#' cli_text("{.fun {1:3}}")
#' #> `1()`, `2()`, and `3()`
#'
#' cli_text("{.fun f{1}}")
#' #> `f1()`
#' ```
#' In the first case `.fun` is applied to each element of `1:3`, whereas
#' in the second case, the `{1}` (rather trivial) substitution is performed
#' first, and then `.fun` is applied to `"f1"`.
#'
#' See the rest of the comments inline (pun intended).
#'
#' @param code The text inside the `{...}` glue substitution.
#' @param envir Environment with the data to perform the styling.
#'   The actual substituted values have the form `v<x>`, where `<x>` is
#'   a number. The rest of the values are metadata. E.g. the app itself is
#'   added as `envir$app`.
#' @return The substituted and styles text, a character scalar.
#'
#' @noRd

inline_transformer <- function(code, envir) {
  app <- envir$app

  match <- regexpr(inline_regex(), code, perl = TRUE)
  has_style <- match != -1

  if (has_style) {
    # styling
    starts <- attr(match, "capture.start")
    ends <- starts + attr(match, "capture.length") - 1L
    captures <- substring(code, starts, ends)
    funname <- captures[[1]]
    text <- captures[[2]]

    id <- clii__container_start(app, "span", class = funname)
    on.exit(clii__container_end(app, id), add = TRUE)

    # If we don't have a brace expression, then we add another class-less
    # `<span>` here, because this will be the one replaced by the pure
    # substitutions (the other branch of the `if`). This ensures that
    # (non-inherited) styling will _not_ be applied before collapsing them,
    # but only to the whole non-brace expression. We don't need to end this
    # container, because the one above (`id`) will end this one as well.

    braceexp <- grepl("^[<][^.][^}]*[>]$", text) &&
      count_brace_exp(text, .open = "<", .close = ">") == 1
    if (!braceexp) {
      id2 <- clii__container_start(app, "span", class = NULL)
    }

    out <- glue(
      text,
      .envir = envir,
      .transformer = inline_transformer,
      .open = paste0("<", envir$marker),
      .close = paste0(envir$marker, ">")
    )

    # If we don't have a brace expression, then (non-inherited) styling was
    # not applied internally, and we need to apply it now. We also need to
    # end the dummy class-less `<span>` here, so we use the original styled
    # contained (`id`).

    if (!braceexp) {
      clii__container_end(app, id2)
      style <- app$get_current_style()
      out <- inline_generic(app, out, style)
    }

    out

  } else {
    # plain substitution
    expr <- parse(text = code, keep.source = FALSE)
    val <- eval(expr, envir = envir)

    # If we are inside another `<span>`, then we'll "replace" that with a
    # new one, so we can add extra classes (from `class-map`) and styles
    # (from [cli_vec()] to it. Replacing means that we look up the right
    # classes and the id, end the container, and create another one with
    # the same classes (+ from `class_map`), the same id, and potentially
    # the styles from [cli_vec()].
    #
    # If we are not inside another `<span>`, then we'll just add a
    # class-less span, so that the non-inherited styles (e.g. `before`) are
    # not used before collapsing.

    node <- utils::tail(app$doc, 1)[[1]]
    if (node$tag == "span") {
      class <- node$class
      id <- node$id
      clii__container_end(app, id = node$id)
    } else {
      class <- NULL
      id <- NULL
    }

    rcls <- class(val)
    stls <- app$get_current_style()$`class-map`
    cls <- na.omit(match(rcls, names(stls)))[1]
    if (!is.na(cls)) class <- c(class, stls[[cls]])

    vec_style <- attr(val, "cli_style")
    tid <- if (!is.null(vec_style)) {
      app$add_theme(list(span = vec_style))
    }

    id <- clii__container_start(
      app, "span", id = id,
      class = paste(class, collapse = " "), theme = tid
    )
    # We don't need to end the replacement container, that happens upstream.
    if (node$tag != "span") {
      on.exit(clii__container_end(app, id), add = TRUE)
    }

    style <- app$get_current_style()
    inline_collapse(
      inline_generic(app, val, style = style),
      style = style
    )
  }
}

clii__inline <- function(app, text, .list) {
  ## Inject that app, so we can style
  texts <- c(if (!is.null(text)) list(text), .list)
  out <- lapply(texts, function(t) {
    t$values$app <- app
    glue(
      t$str,
      .envir = t$values,
      .transformer = inline_transformer,
      .open = paste0("<", t$values$marker),
      .close = paste0(t$values$marker, ">"),
      .trim = FALSE
    )
  })
  paste(out, collapse = "")
}

inline_regex <- function() "(?s)^[.]([-[:alnum:]_]+)[[:space:]]+(.*)"

make_cmd_transformer <- function(values, .call = NULL) {
  values$marker <- random_id()
  values$qty <- NA_integer_
  values$num_subst <- 0L
  values$postprocess <- FALSE
  values$pmarkers <- list()

  # These are common because of purrr's default argument names, so we
  # hardcode them es exceptions. They are in packages
  # crossmap, crosstable, rstudio.prefs, rxode2, starter.
  # rxode2 has the other ones, and we should fix that in rxode2
  # the function calls are in the oolong packagee, need to fix this as well.
  exceptions <- c(
    ".x", ".y", ".",
    ".md", ".met", ".med", ".mul", ".muR", ".dir", ".muU",
    ".sym_flip(bool_word)", ".sym_flip(bool_topic)", ".sym_flip(bool_wsi)"
  )

  # it is not easy to do better than this, we would need to pass a call
  # down from the exported functions

  caller <- .call %||% sys.call(-1)
  function(code, envir) {
    first_char <- substr(code, 1, 1)

    # {?} pluralization
    if (first_char == "?") {
      parse_plural(code, values)

    # {.} cli style
    } else if (first_char == "." && ! code %in% exceptions) {
      m <- regexpr(inline_regex(), code, perl = TRUE)
      has_match <- m != -1
      if (!has_match) {
        throw(cli_error(
          call. = caller,
          "Invalid cli literal: {.code {{{abbrev(code, 10)}}}} starts with a dot.",
          "i" = "Interpreted literals must not start with a dot in cli >= 3.4.0.",
          "i" = paste("{.code {{}}} expressions starting with a dot are",
                      "now only used for cli styles."),
          "i" = paste("To avoid this error, put a space character after",
                      "the starting {.code {'{'}} or use parentheses:",
                      "{.code {{({abbrev(code, 10)})}}}.")
        ))
      }

      starts <- attr(m, "capture.start")
      ends <- starts + attr(m, "capture.length") - 1L
      captures <- substring(code, starts, ends)
      funname <- captures[[1]]
      text <- captures[[2]]

      out <- glue(
        text,
        .envir = envir,
        .transformer = sys.function(),
        .cli = TRUE
      )
      paste0("<", values$marker, ".", funname, " ", out, values$marker, ">")

    # {} plain substitution
    } else {
      expr <- parse(text = code, keep.source = FALSE) %??%
        cli_error(
          call. = caller,
          "Could not parse cli {.code {{}}} expression:
           {.code {abbrev(code, 20)}}."
        )
      res <- eval(expr, envir = envir) %??%
        cli_error(
          call. = caller,
          "Could not evaluate cli {.code {{}}} expression:
           {.code {abbrev(code, 20)}}."
        )

      id <- paste0("v", length(values))
      values[[id]] <- res
      values$qty <- if (length(res) == 0) 0 else res
      values$num_subst <- values$num_subst + 1L
      paste0("<", values$marker, id, values$marker, ">")
    }
  }
}

glue_cmd <- function(..., .envir, .call = sys.call(-1), .trim = TRUE) {
  str <- paste0(unlist(list(...), use.names = FALSE), collapse = "")
  values <- new.env(parent = emptyenv())
  transformer <- make_cmd_transformer(values, .call = .call)
  pstr <- glue(
    str,
    .envir = .envir,
    .transformer = transformer,
    .cli = TRUE,
    .trim = .trim
  )
  glue_delay(
    str = post_process_plurals(pstr, values),
    values = values
  )
}

glue_no_cmd <- function(...) {
  str <- paste0(unlist(list(...), use.names = FALSE), collapse = "")
  values <-new.env(parent = emptyenv())
  glue_delay(
    str = str,
    values = values
  )
}

glue_delay <- function(str, values = NULL) {
  structure(
    list(str = str, values = values),
    class = "cli_glue_delay"
  )
}

#' Simulate (a subset of) a VT-5xx ANSI terminal
#'
#' This is utility function that calculates the state of a VT-5xx screen
#' after a certain set of output.
#'
#' Currently it supports:
#'
#' - configurable terminal width and height
#' - ASCII printable characters.
#' - `\n`, `\r`.
#' - ANSI SGR colors, 8 color mode, 256 color mode and true color mode.
#' - Other ANSI SGR features: bold, italic, underline, strikethrough,
#'   blink, inverse.
#'
#' It does _not_ currently supports other features, mode notably:
#'
#' - Other ANSI control sequences and features. Other control sequences
#'   are silently ignored.
#' - Wide Unicode characters. Their width is not taken into account
#'   correctly.
#' - Unicode graphemes.
#'
#' @param output Character vector or raw vector. Character vectors are
#' collapsed (without a separater), and converted to a raw vector using
#' [base::charToRaw()].
#' @param width Terminal width.
#' @param height Terminal height.
#' @return Data frame with columns `lineno`, `segmentno`, `segment`,
#' `attributes`.
#'
#' @note
#' This function is experimental, and the virtual temrinal API will
#' likely change in future versions of cli.
#'
#' @export

vt_output <- function(output, width = 80L, height = 25L) {
  if (is.character(output)) {
    output <- charToRaw(paste(output, collapse = ""))
  }

  res <- .Call(
    clic_vt_output,
    output,
    as.integer(width),
    as.integer(height)
  )

  linksx <- vapply(res$links, intToUtf8, character(1))
  links <- sub("^[^;]*;", "", linksx)
  links_params <- sub(";[^;]*$", "", linksx)

  df <- data.frame(
    stringsAsFactors = FALSE,
    lineno = integer(),
    segmentno = integer(),
    segment = character(),
    attributes = character()
  )

  segments <- lapply(seq_along(res$lines), function(i) {
    line <- intToUtf8(res$lines[[i]])
    attr <- res$attrs[[i]]
    lgs <- rle(attr)
    clgs <- cumsum(c(0, lgs$lengths))
    segs <- mapply(clgs[-length(clgs)], clgs[-1], FUN = function(s, e) {
      utf8_substr(line, s + 1, e)
    })

    fg <- re_match(lgs$values, "fg:([0-9]+|#[0-9a-f]+);")[,1]
    bg <- re_match(lgs$values, "bg:([0-9]+|#[0-9a-f]+);")[,1]
    linkno <- as.integer(re_match(lgs$values, "link:([0-9]+);")[,1])
    link <- links[linkno]
    link_params <- links_params[linkno]

    data.frame(
      stringsAsFactors = FALSE,
      lineno = i,
      segmentno = seq_along(segments),
      segment = segs,
      bold = grepl("bold;", lgs$values, fixed = TRUE),
      italic = grepl("italic;", lgs$values, fixed = TRUE),
      underline = grepl("underline;", lgs$values, fixed = TRUE),
      strikethrough = grepl("strikethrough;", lgs$values, fixed = TRUE),
      blink = grepl("blink;", lgs$values, fixed = TRUE),
      inverse = grepl("inverse;", lgs$values, fixed = TRUE),
      color = fg,
      background_color = bg,
      link = link,
      link_params = link_params
    )
  })

  do.call(rbind, c(list(df), segments))
}

#' @title Progress bar utility functions.
#' @details `cli_progress_num()` returns the number of currently
#' active progress bars. (These do not currently include the progress
#' bars created in C/C++ code.)
#' @return `cli_progress_num()` returns an integer scalar.
#'
#' @rdname progress-utils
#' @family progress bar functions
#' @export

cli_progress_num <- function() {
  length(clienv$progress)
}

#' @details `cli_progress_cleanup()` terminates all active progress bars.
#' (It currently ignores progress bars created in the C/C++ code.)
#' @return `cli_progress_cleanup() does not return anything.
#'
#' @rdname progress-utils
#' @export

cli_progress_cleanup <- function() {
  while ((n <- cli_progress_num()) > 0) {
    cli_progress_done(clienv$progress[[n]]$id)
  }
  ansi_show_cursor()
  invisible()
}

should_run_progress_examples <- function() {
  if (is_rcmd_check()) return(FALSE)
  tolower(Sys.getenv("R_PROGRESS_NO_EXAMPLES")) != "true"
}

is_rcmd_check <- function() {
  if (identical(Sys.getenv("NOT_CRAN"), "true")) {
    FALSE
  } else {
    Sys.getenv("_R_CHECK_PACKAGE_NAME_", "") != ""
  }
}

test_that("cli_progress_num", {
  withr::local_options(cli.progress_handlers_only = "cli")
  fun <- function() {
    before <- cli_progress_num()
    cli_progress_bar()
    after <- cli_progress_num()
    expect_equal(before + 1, after)
  }

  capture_cli_messages(fun())
})

test_that("cli_progress_cleanup", {
  fun <- function() {
    num <- NULL
    cli::cli_progress_bar()
    fun2 <- function() {
      cli::cli_progress_bar()
      cli::cli_progress_cleanup()
      num <<- cli::cli_progress_num()
    }
    fun2()
  }

  out <- callr::r(fun)
  expect_equal(out, 0L)
})

test_that("should_run_progress_examples", {
  withr::local_envvar(NOT_CRAN = "true")
  expect_true(should_run_progress_examples())

  local_mocked_bindings(is_rcmd_check = function() TRUE)
  expect_false(should_run_progress_examples())
})

test_that("is_rcmd_check", {
  withr::local_envvar(NOT_CRAN = NA, "_R_CHECK_PACKAGE_NAME_" = NA)
  expect_false(is_rcmd_check())
  withr::local_envvar(NOT_CRAN = NA, "_R_CHECK_PACKAGE_NAME_" = "cli")
  expect_true(is_rcmd_check())
})

test_that("cli_progress_num", {
  withr::local_options(cli.progress_handlers_only = "cli")
  fun <- function() {
    before <- cli_progress_num()
    cli_progress_bar()
    after <- cli_progress_num()
    expect_equal(before + 1, after)
  }

  capture_cli_messages(fun())
})

test_that("cli_progress_cleanup", {
  fun <- function() {
    num <- NULL
    cli::cli_progress_bar()
    fun2 <- function() {
      cli::cli_progress_bar()
      cli::cli_progress_cleanup()
      num <<- cli::cli_progress_num()
    }
    fun2()
  }

  out <- callr::r(fun)
  expect_equal(out, 0L)
})

test_that("should_run_progress_examples", {
  withr::local_envvar(NOT_CRAN = "true")
  expect_true(should_run_progress_examples())

  mockery::stub(should_run_progress_examples, "is_rcmd_check", TRUE)
  expect_false(should_run_progress_examples())
})

test_that("is_rcmd_check", {
  withr::local_envvar(NOT_CRAN = NA, "_R_CHECK_PACKAGE_NAME_" = NA)
  expect_false(is_rcmd_check())
  withr::local_envvar(NOT_CRAN = NA, "_R_CHECK_PACKAGE_NAME_" = "cli")
  expect_true(is_rcmd_check())
})

#' Detect the number of ANSI colors to use
#'
#' @description
#' Certain Unix and Windows terminals, and also certain R GUIs, e.g.
#' RStudio, support styling terminal output using special control
#' sequences (ANSI sequences).
#'
#' `num_ansi_colors()` detects if the current R session supports ANSI
#' sequences, and if it does how many colors are supported.
#'
#' @param stream The stream that will be used for output, an R connection
#' object. It can also be a string, one of `"auto"`, `"message"`,
#' `"stdout"`, `"stderr"`. `"auto"` will select `stdout()` if the session is
#' interactive and there are no sinks, otherwise it will select `stderr()`.
#' @return Integer, the number of ANSI colors the current R session
#' supports for `stream`.
#'
#' @family ANSI styling
#' @export
#' @examples
#' num_ansi_colors()
#'
#' @details
#' The detection mechanism is quite involved and it is designed to work
#' out of the box on most systems. If it does not work on your system,
#' please report a bug. Setting options and environment variables to turn
#' on ANSI support is error prone, because they are inherited in other
#' environments, e.g. knitr, that might not have ANSI support.
#'
#' If you want to _turn off_ ANSI colors, set the `NO_COLOR` environment
#' variable to a non-empty value.
#'
#' The exact detection mechanism is as follows:

num_ansi_colors <- function(stream = "auto") {
  #' 1. If the `cli.num_colors` options is set, that is returned.

  opt <- getOption("cli.num_colors", NULL)
  if (!is.null(opt)) return(as.integer(opt))

  #' 1. If the `R_CLI_NUM_COLORS` environment variable is set to a
  #'    non-empty value, then it is used.

  if ((env <- Sys.getenv("R_CLI_NUM_COLORS", "")) != "") {
    return(as.integer(env))
  }

  #' 1. If the `crayon.enabled` option is set to `FALSE`, 1L is returned.
  #'    (This is for compatibility with code that uses the crayon package.)
  #' 1. If the `crayon.enabled` option is set to `TRUE` and the
  #'    `crayon.colors` option is not set, then the value of the
  #'    `cli.default_num_colors` option, or if it is unset, then 8L is
  #'    returned.
  #' 1. If the `crayon.enabled` option is set to `TRUE` and the
  #'    `crayon.colors` option is also set, then the latter is returned.
  #'    (This is for compatibility with code that uses the crayon package.)

  cray_opt_has <- getOption("crayon.enabled", NULL)
  cray_opt_num <- getOption("crayon.colors", NULL)
  if (!is.null(cray_opt_has) && !isTRUE(cray_opt_has)) return(1L)
  if (isTRUE(cray_opt_has) && !is.null(cray_opt_num)) {
    return(as.integer(cray_opt_num))
  }
  if (isTRUE(cray_opt_has) && is.null(cray_opt_num)) {
    default <- get_default_number_of_colors()
    return(default %||% 8L)
  }

  #' 1. If the `NO_COLOR` environment variable is set, then 1L is returned.

  if (!is.na(Sys.getenv("NO_COLOR", NA_character_))) return(1L)

  #' 1. If we are in knitr, then 1L is returned, to turn off colors in
  #'    `.Rmd` chunks.

  if (isTRUE(getOption("knitr.in.progress"))) return(1L)

  #' 1. If `stream` is `"auto"` (the default) and there is an active
  #'    sink (either for `"output"` or `"message"`), then we return 1L.
  #'    (In theory we would only need to check the stream that will be
  #'    be actually used, but there is no easy way to tell that.)
  if (stream == "auto" && !no_sink()) return(1L)

  # Defer computation on streams to speed up common case
  # when environment variables are set
  orig_stream <- stream
  stream <- get_real_output(stream)

  is_stdout <- is_stderr <- is_std <- FALSE
  std <- "nope"
  if (identical(stream, stdout())) {
    is_stdout <- is_std <- TRUE
    std <- "stdout"
  } else if (identical(stream, stderr())) {
    is_stderr <- is_std <- TRUE
    std <- "stderr"
  }

  #' 1. If `stream` is not `"auto"`, but it is `stderr()` and there is an
  #'    active sink for it, then 1L is returned.
  #'    (If a sink is active for "output", then R changes the `stdout()`
  #'    stream, so this check is not needed.)

  # If a sink is active for "message" (ie. stderr), then R does not update
  # the `stderr()` stream, so we need to catch this case.
  if (is_stderr && sink.number("message") != 2) return(1L)

  #' 1. If the `cli.default_num_colors` option is set, then we use that.

  dopt <- get_default_number_of_colors()
  if (!is.null(dopt)) return(as.integer(dopt))

  #' 1. If R is running inside RGui on Windows, or R.app on macOS, then we
  #'    return 1L.

  # RStudio sets GUI to RGui initially, so we'll handle that after RStudio.
  if (.Platform$GUI == "AQUA") return(1L)

  #' 1. If R is running inside RStudio, with color support, then the
  #'    appropriate number of colors is returned, usually 256L.

  rstudio <- rstudio$detect()
  rstudio_colors <- c(
    "rstudio_console",
    "rstudio_console_starting",
    "rstudio_build_pane",
    "rstudio_job"
  )
  if (is.na(rstudio$num_colors)) rstudio$num_colors <- 1L
  if (rstudio$type %in% rstudio_colors && is_std) {
    return(rstudio$num_colors)
  }

  # RGui? We need to do this after RStudio, because .Platform$GUI is
  # "Rgui" in RStudio when we are starting up
  if (.Platform$GUI == "Rgui") return(1L)

  #' 1. If R is running on Windows, inside an Emacs version that is recent
  #'    enough to support ANSI colors, then the value of the
  #'    `cli.default_num_colors` option, or if unset 8L is returned.
  #'    (On Windows, Emacs has `isatty(stdout()) == FALSE`, so we need to
  #'    check for this here before dealing with terminals.)

  # Windows Emacs? The top R process will have `--ess` in ESS, but the
  # subprocesses won't. (Without ESS subprocesses will also report 8L
  # colors, this is a problem, but we expect most people use ESS in Emacs.)
  if (os_type() == "windows" &&
      "--ess" %in% commandArgs() &&
      is_emacs_with_color()) {
    default <- get_default_number_of_colors()
    return(default %||% 8L)
  }

  #' 1. If `stream` is not the standard output or standard error  in a
  #'    terminal, then 1L is returned.

  if (!isatty(stream)) return(1L)
  if (!is_std) return(1L)

  #' 1. Otherwise we use and cache the result of the terminal color
  #'     detection (see below).

  # Otherwise use/set the cache
  if (is.null(clienv$num_colors)) clienv$num_colors <- list()
  clienv$num_colors[[std]] <- clienv$num_colors[[std]] %||% detect_tty_colors()
  clienv$num_colors[[std]]
}

#' @rdname num_ansi_colors
#' @details
#' The terminal color detection algorithm:

detect_tty_colors <- function() {

  default <- get_default_number_of_colors()

  #' 1. If the `COLORTERM` environment variable is set to `truecolor` or
  #'    `24bit`, then we return 16 million colors.
  #' 1. If the `COLORTERM` environment variable is set to anything else,
  #'    then we return the value of the `cli.num_default_colors` option,
  #'    8L if unset.

  ct <- Sys.getenv("COLORTERM", NA_character_)
  if (!is.na(ct)) {
    if (ct == "truecolor" || ct == "24bit") {
      return(truecolor)
    } else {
      return(default %||% 8L)
    }
  }

  #' 1. If R is running on Unix, inside an Emacs version that is recent
  #'    enough to support ANSI colors, then the value of the
  #'    `cli.default_num_colors` option is returned, or 8L if unset.

  if (os_type() == "unix" && is_emacs_with_color()) return(default %||% 8L)

  #' 1. If we are on Windows in an RStudio terminal, then apparently
  #'    we only have eight colors, but the `cli.default_num_colors` option
  #'    can be used to override this.

  win10 <- win10_build()
  if (os_type() == "windows" && win10 >= 10586 &&
      rstudio_detect()$type == "rstudio_terminal") {
    # this is rather weird, but echo turns on color support :D
    system2("cmd", c("/c", "echo 1 >NUL"))
    return(default %||% 8L)
  }

  #' 1. If we are in a recent enough Windows 10 terminal, then there
  #'    is either true color (from build 14931) or 256 color (from
  #'    build 10586) support. You can also use the `cli.default_num_colors`
  #'    option to override these.

  if (os_type() == "windows" && win10 >= 10586) {
    # this is rather weird, but echo turns on color support :D
    system2("cmd", c("/c", "echo 1 >NUL"))
    # https://devblogs.microsoft.com/commandline/24-bit-color-in-the-windows-console/
    if (win10 >= 14931) {
      return(default %||% truecolor)
    } else {
      return(default %||% 256L)
    }
  }

  if (os_type() == "windows") {

    #' 1. If we are on Windows, under ConEmu or cmder, or ANSICON is loaded,
    #'    then the value of `cli.default_num_colors`, or 8L if unset, is
    #'    returned.

    if (Sys.getenv("ConEmuANSI") == "ON" ||
        Sys.getenv("CMDER_ROOT") != "") {
      return(default %||% 8L)
    }
    if (Sys.getenv("ANSICON") != "") return(default %||% 8L)

    #' 1. Otherwise if we are on Windows, return 1L.

    return(1L)
  }

  #' 1. Otherwise we are on Unix and try to run `tput colors` to determine
  #'    the number of colors. If this succeeds, we return its return value.

  cols <- suppressWarnings(try(
    silent = TRUE,
    as.numeric(system("tput colors 2>/dev/null", intern = TRUE))[1]
  ))
  if (inherits(cols, "try-error") || !length(cols) || is.na(cols)) {
    return(guess_tty_colors())
  }
  if (cols %in% c(-1, 0, 1)) { return(1) }

  #'    If the `TERM` environment variable is `xterm` and `tput`
  #'    returned 8L, we return 256L, because xterm compatible terminals
  #'    tend to support 256 colors
  #'    (<https://github.com/r-lib/crayon/issues/17>)
  #'    You can override this with the `cli.default_num_colors` option.

  if (cols == 8 && identical(Sys.getenv("TERM"), "xterm")) {
    cols <- default %||% 256
  }

  #' 1. If `TERM` is set to `dumb`, we return 1L.
  #' 1. If `TERM` starts with `screen`, `xterm`, or `vt100`, we return 8L.
  #' 1. If `TERM` contains `color`, `ansi`, `cygwin` or `linux`, we return 8L.
  #' 1. Otherwise we return 1L.

  cols
}

get_default_number_of_colors <- function() {
  dft <- getOption("cli.default_num_colors")
  if (!is.null(dft)) {
    if (!is_count(dft)) {
      warning(
        "The `cli.default_num_colors` option must be an integer scalar"
      )
      dft <- NULL
    }
  }
  dft
}

guess_tty_colors <- function() {
  term <- Sys.getenv("TERM")
  if (term == "dumb") return (1L)

  if (grepl(
    "^screen|^xterm|^vt100|color|ansi|cygwin|linux",
    term,
    ignore.case = TRUE,
    perl = TRUE
  )) {
    8L
  } else {
    1L
  }
}

is_emacs_with_color <- function() {
  (Sys.getenv("EMACS") != "" || Sys.getenv("INSIDE_EMACS") != "") &&
    ! is.na(emacs_version()[1]) && emacs_version()[1] >= 23
}

emacs_version <- function() {
  ver <- Sys.getenv("INSIDE_EMACS")
  if (ver == "") return(NA_integer_)

  ver <- gsub("'", "", ver, fixed = TRUE)

  ver <- strsplit(ver, ",", fixed = TRUE)[[1]]
  ver <- strsplit(ver, ".", fixed = TRUE)[[1]]
  suppressWarnings(as.numeric(ver))
}

win10_build <- function() {
  os <- utils::sessionInfo()$running %||% ""
  if (!grepl("^Windows 10 ", os)) return(0L)
  mch <- re_match(os, "[(]build (?<build>[0-9]+)[)]")
  mch <- suppressWarnings(as.integer(mch))
  if (is.na(mch)) return(0L)
  mch
}

test_that("win10_build works for different osVersion", {
  local_mocked_bindings(
    sessionInfo = function() list(running = NULL), .package = "utils"
  )
  expect_identical(win10_build(), 0L)

  local_mocked_bindings(
    sessionInfo = function() list(running = "Debian GNU/Linux 11 (bullseye)"),
    .package = "utils"
  )
  expect_identical(win10_build(), 0L)

  local_mocked_bindings(
    sessionInfo = function() list(running = "Windows 10 x64 (build 16299)"),
    .package = "utils"
  )
  expect_identical(win10_build(), 16299L)
})

test_that("cli.default_num_colors #1", {

  # crayon.enabled
  withr::local_envvar(R_CLI_NUM_COLORS = NA_character_)
  withr::local_options(
    cli.num_colors = NULL,
    crayon.enabled = TRUE,
    crayon.colors = NULL,
    cli.default_num_colors = NULL
  )

  expect_equal(num_ansi_colors(), 8L)

  withr::local_options(cli.default_num_colors = 123L)
  expect_equal(num_ansi_colors(), 123L)
})

test_that("cli.default_num_colors #2", {

  # Windows emacs
  withr::local_envvar(
    R_CLI_NUM_COLORS = NA_character_,
    RSTUDIO = NA_character_
  )
  withr::local_options(
    cli.num_colors = NULL,
    crayon.enabled = NULL,
    crayon.colors = NULL,
    cli.default_num_colors = NULL
  )

  local_mocked_bindings(
    os_type = function() "windows",
    commandArgs = function() "--ess",
    is_emacs_with_color = function() TRUE
  )

  expect_equal(num_ansi_colors(), 8L)

  withr::local_options(cli.default_num_colors = 123L)
  expect_equal(num_ansi_colors(), 123L)
})

test_that("cli.default_num_colors #3", {

  # non-truecolor COLORMAP
  withr::local_envvar(COLORTERM = "other")
  withr::local_options(cli.default_num_colors = NULL)

  expect_equal(detect_tty_colors(), 8L)

  withr::local_options(cli.default_num_colors = 123L)
  expect_equal(detect_tty_colors(), 123L)
})

test_that("cli.default_num_colors #4", {

  # Unix emacs with color
  withr::local_envvar(COLORTERM = NA_character_)

  local_mocked_bindings(
    os_type = function() "unix",
    is_emacs_with_color = function() TRUE
  )

  withr::local_options(cli.default_num_colors = NULL)

  expect_equal(detect_tty_colors(), 8L)

  withr::local_options(cli.default_num_colors = 123L)
  expect_equal(detect_tty_colors(), 123L)
})

test_that("cli.default_num_colors #5", {

  # rstudio terminal on Windows
  withr::local_envvar(COLORTERM = NA_character_)

  local_mocked_bindings(
    os_type = function() "windows",
    win10_build = function() 10586,
    rstudio_detect = function() list(type = "rstudio_terminal"),
    system2 = function(...) TRUE
  )

  withr::local_options(cli.default_num_colors = NULL)
  expect_equal(detect_tty_colors(), 8L)

  withr::local_options(cli.default_num_colors = 123L)
  expect_equal(detect_tty_colors(), 123L)
})

test_that("cli.default_num_colors #6", {

  # Windows 10 terminal
  withr::local_envvar(COLORTERM = NA_character_)
  withr::local_options(cli.default_num_colors = NULL)

  local_mocked_bindings(
    os_type = function() "windows",
    win10_build = function() 10586,
    rstudio_detect = function() list(type = "not_rstudio"),
    system2 = function(...) TRUE
  )
  expect_equal(detect_tty_colors(), 256L)

  local_mocked_bindings(win10_build = function() 14931)
  expect_equal(detect_tty_colors(), truecolor)

  withr::local_options(cli.default_num_colors = 123L)
  expect_equal(detect_tty_colors(), 123L)
})

test_that("cli.default_num_colors #7", {

  # conemu or cmder
  withr::local_envvar(
    COLORTERM = NA_character_,
    ConEmuANSI = "ON"
  )
  withr::local_options(cli.default_num_colors = NULL)
  local_mocked_bindings(
    os_type = function() "windows",
    win10_build = function() 1
  )

  expect_equal(detect_tty_colors(), 8L)
  withr::local_options(cli.default_num_colors = 123L)
  expect_equal(detect_tty_colors(), 123L)
})

test_that("cli.default_num_colors #8", {

  # unix terminal, xterm
  withr::local_envvar(
    COLORTERM = NA_character_,
    TERM = "xterm"
  )

  local_mocked_bindings(
    os_type = function() "unix",
    is_emacs_with_color = function() FALSE,
    system = function(...) "8"
  )

  withr::local_options(cli.default_num_colors = NULL)
  expect_equal(detect_tty_colors(), 256L)

  withr::local_options(cli.default_num_colors = 123L)
  expect_equal(detect_tty_colors(), 123L)
})

test_that("ESS_BACKGROUND_MODE", {
  withr::local_envvar(
    RSTUDIO = NA_character_,
    ESS_BACKGROUND_MODE = NA_character_
  )

  local_mocked_bindings(
    is_iterm = function() FALSE,
    is_emacs = function() TRUE
  )

  expect_false(detect_dark_theme("auto"))

  withr::local_envvar(ESS_BACKGROUND_MODE = "dark")
  expect_true(detect_dark_theme("auto"))
})

test_that("emacs_version", {
  withr::local_envvar(INSIDE_EMACS = "")
  expect_true(is.na(emacs_version()))
  withr::local_envvar(INSIDE_EMACS = "foobar")
  expect_true(is.na(emacs_version()))

  withr::local_envvar(INSIDE_EMACS = "23.2.3")
  expect_equal(emacs_version(), c(23, 2, 3))
  withr::local_envvar(INSIDE_EMACS = "23.2.3,foobar")
  expect_equal(emacs_version(), c(23, 2, 3))
  withr::local_envvar(INSIDE_EMACS = "'23.2.3'")
  expect_equal(emacs_version(), c(23, 2, 3))
  withr::local_envvar(INSIDE_EMACS = "'23.2.3',foobar")
  expect_equal(emacs_version(), c(23, 2, 3))
})

test_that("win10_build works for different osVersion", {
    mockery::stub(
        win10_build, "utils::sessionInfo",
        list(running = NULL)
    )
    expect_identical(win10_build(), 0L)

    mockery::stub(
        win10_build, "utils::sessionInfo",
        list(running = "Debian GNU/Linux 11 (bullseye)")
    )
    expect_identical(win10_build(), 0L)

    mockery::stub(
        win10_build, "utils::sessionInfo",
        list(running = "Windows 10 x64 (build 16299)")
    )
    expect_identical(win10_build(), 16299L)
})

test_that("cli.default_num_colors #1", {

  # crayon.enabled
  withr::local_envvar(R_CLI_NUM_COLORS = NA_character_)
  withr::local_options(
    cli.num_colors = NULL,
    crayon.enabled = TRUE,
    crayon.colors = NULL,
    cli.default_num_colors = NULL
  )

  expect_equal(num_ansi_colors(), 8L)

  withr::local_options(cli.default_num_colors = 123L)
  expect_equal(num_ansi_colors(), 123L)
})

test_that("cli.default_num_colors #2", {

  # Windows emacs
  withr::local_envvar(
    R_CLI_NUM_COLORS = NA_character_,
    RSTUDIO = NA_character_
  )
  withr::local_options(
    cli.num_colors = NULL,
    crayon.enabled = NULL,
    crayon.colors = NULL,
    cli.default_num_colors = NULL
  )

  mockery::stub(num_ansi_colors, "os_type", "windows")
  mockery::stub(num_ansi_colors, "commandArgs", "--ess")
  mockery::stub(num_ansi_colors, "is_emacs_with_color", TRUE)

  expect_equal(num_ansi_colors(), 8L)

  withr::local_options(cli.default_num_colors = 123L)
  expect_equal(num_ansi_colors(), 123L)
})

test_that("cli.default_num_colors #3", {

  # non-truecolor COLORMAP
  withr::local_envvar(COLORTERM = "other")
  withr::local_options(cli.default_num_colors = NULL)

  expect_equal(detect_tty_colors(), 8L)

  withr::local_options(cli.default_num_colors = 123L)
  expect_equal(detect_tty_colors(), 123L)
})

test_that("cli.default_num_colors #4", {

  # Unix emacs with color
  withr::local_envvar(COLORTERM = NA_character_)

  mockery::stub(detect_tty_colors, "os_type", "unix")
  mockery::stub(detect_tty_colors, "is_emacs_with_color", TRUE)

  withr::local_options(cli.default_num_colors = NULL)

  expect_equal(detect_tty_colors(), 8L)

  withr::local_options(cli.default_num_colors = 123L)
  expect_equal(detect_tty_colors(), 123L)
})

test_that("cli.default_num_colors #5", {

  # rstudio terminal on Windows
  withr::local_envvar(COLORTERM = NA_character_)

  mockery::stub(detect_tty_colors, "os_type", "windows")
  mockery::stub(detect_tty_colors, "win10_build", 10586)
  mockery::stub(
    detect_tty_colors,
    "rstudio_detect",
    list(type = "rstudio_terminal")
  )
  mockery::stub(detect_tty_colors, "system2", TRUE)

  withr::local_options(cli.default_num_colors = NULL)
  expect_equal(detect_tty_colors(), 8L)

  withr::local_options(cli.default_num_colors = 123L)
  expect_equal(detect_tty_colors(), 123L)
})

test_that("cli.default_num_colors #6", {

  # Windows 10 terminal
  withr::local_envvar(COLORTERM = NA_character_)
  withr::local_options(cli.default_num_colors = NULL)

  mockery::stub(detect_tty_colors, "os_type", "windows")
  mockery::stub(detect_tty_colors, "win10_build", 10586)
  mockery::stub(
    detect_tty_colors,
    "rstudio_detect",
    list(type = "not_rstudio")
  )
  mockery::stub(detect_tty_colors, "system2", TRUE)
  expect_equal(detect_tty_colors(), 256L)

  mockery::stub(detect_tty_colors, "win10_build", 14931)
  expect_equal(detect_tty_colors(), truecolor)

  withr::local_options(cli.default_num_colors = 123L)
  expect_equal(detect_tty_colors(), 123L)
})

test_that("cli.default_num_colors #7", {

  # conemu or cmder
  withr::local_envvar(
    COLORTERM = NA_character_,
    ConEmuANSI = "ON"
  )
  withr::local_options(cli.default_num_colors = NULL)
  mockery::stub(detect_tty_colors, "os_type", "windows")
  mockery::stub(detect_tty_colors, "win10_build", 1)

  expect_equal(detect_tty_colors(), 8L)
  withr::local_options(cli.default_num_colors = 123L)
  expect_equal(detect_tty_colors(), 123L)
})

test_that("cli.default_num_colors #8", {

  # unix terminal, xterm
  withr::local_envvar(
    COLORTERM = NA_character_,
    TERM = "xterm"
  )

  mockery::stub(detect_tty_colors, "os_type", "unix")
  mockery::stub(detect_tty_colors, "is_emacs_with_color", FALSE)
  mockery::stub(detect_tty_colors, "system", "8")

  withr::local_options(cli.default_num_colors = NULL)
  expect_equal(detect_tty_colors(), 256L)

  withr::local_options(cli.default_num_colors = 123L)
  expect_equal(detect_tty_colors(), 123L)
})

test_that("ESS_BACKGROUND_MODE", {
  withr::local_envvar(
    RSTUDIO = NA_character_,
    ESS_BACKGROUND_MODE = NA_character_
  )

  mockery::stub(detect_dark_theme, "is_iterm", FALSE)
  mockery::stub(detect_dark_theme, "is_emacs", TRUE)

  expect_false(detect_dark_theme("auto"))

  withr::local_envvar(ESS_BACKGROUND_MODE = "dark")
  expect_true(detect_dark_theme("auto"))
})

#!/bin/sh

gpg --quiet --batch --yes --decrypt --passphrase="$SECRETS_PASSWORD" --output ./.github/secrets/hcloud_cli.p12 ./.github/secrets/hcloud_cli.p12.gpg

security create-keychain -p "" build.keychain
security import ./.github/secrets/hcloud_cli.p12 -t agg -k ~/Library/Keychains/build.keychain -P "$CERT_PASSWORD" -A

security list-keychains -s ~/Library/Keychains/build.keychain
security default-keychain -s ~/Library/Keychains/build.keychain
security unlock-keychain -p "" ~/Library/Keychains/build.keychain

security set-key-partition-list -S apple-tool:,apple: -s -k "" ~/Library/Keychains/build.keychain

#!/bin/bash -e

usage() {
  echo "usage: build.bash OS ARCH RELEASE" >&2
  exit 2
}

os="$1"
[ -z "$os" ] && usage

arch="$2"
[ -z "$arch" ] && usage

release="$3"
[ -z "$release" ] && usage

LD_FLAGS="-w -X github.com/hetznercloud/cli/internal/version.Version=$release"
GOOS=$os GOARCH=$arch go build -o ./dist/hcloud-$os-$arch-$release -ldflags "$LD_FLAGS" ./cmd/hcloud

#!/bin/bash
set -e

# This script is designed to live in a Confluent Platform tarball distribution.
# It depends on the CLI binaries for different OS/ARCH combinations living at
#   ../libexec/cli/${OS}_${ARCH}/confluent

is_supported_platform() {
  platform=$1
  found=1
  case "$platform" in
    linux/amd64) found=0 ;;
    linux/arm64) found=0 ;;
    darwin/amd64) found=0 ;;
    darwin/arm64) found=0 ;;
    windows/amd64) found=0 ;;
  esac
  case "$platform" in
    darwin/386) found=1 ;;
    linux/386) found=1 ;;
    windows/386) found=1 ;;
  esac
  return $found
}

check_platform() {
  if is_supported_platform "$PLATFORM"; then
    # optional logging goes here
    true
  else
    log_crit "platform $PLATFORM is not supported.  Make sure this script is up-to-date and file request at https://github.com/${PREFIX}/issues/new"
    exit 1
  fi
}

adjust_os() {
  # adjust archive name based on OS
  case ${OS} in
    amd64) OS=x86_64 ;;
    darwin) OS=darwin ;;
    linux) OS=linux ;;
    alpine) OS=alpine ;;
    windows) OS=windows ;;
  esac
  true
}

check_executable() {
  if [[ -f "${EXECUTABLE}" ]] ; then
    # optional logging goes here
    true
  else
    log_crit "executable $EXECUTABLE does not exist.  Make sure this script is up-to-date and resides at bin/confluent in the Confluent Platform release directory."
    exit 1
  fi
}

init_config() {
  mkdir -p ${HOME}/.confluent
  if [[ ! -f "${HOME}/.confluent/config.json" ]] ; then
    echo '{"disable_updates": true}' > "${HOME}/.confluent/config.json"
  fi
}

cat /dev/null <<EOF
------------------------------------------------------------------------
https://github.com/client9/shlib - portable posix shell functions
Public domain - http://unlicense.org
https://github.com/client9/shlib/blob/master/LICENSE.md
but credit (and pull requests) appreciated.
------------------------------------------------------------------------
EOF

echoerr() {
  echo "$@" 1>&2
}

log_prefix() {
  echo "$0"
}

_logp=6
log_set_priority() {
  _logp="$1"
}

log_priority() {
  if test -z "$1"; then
    echo "$_logp"
    return
  fi
  [ "$1" -le "$_logp" ]
}

log_tag() {
  case $1 in
    0) echo "emerg" ;;
    1) echo "alert" ;;
    2) echo "crit" ;;
    3) echo "err" ;;
    4) echo "warning" ;;
    5) echo "notice" ;;
    6) echo "info" ;;
    7) echo "debug" ;;
    *) echo "$1" ;;
  esac
}

log_debug() {
  log_priority 7 || return 0
  echoerr "$(log_prefix)" "$(log_tag 7)" "$@"
}

log_info() {
  log_priority 6 || return 0
  echoerr "$(log_prefix)" "$(log_tag 6)" "$@"
}

log_err() {
  log_priority 3 || return 0
  echoerr "$(log_prefix)" "$(log_tag 3)" "$@"
}

log_crit() {
  log_priority 2 || return 0
  echoerr "$(log_prefix)" "$(log_tag 2)" "$@"
}

uname_os() {
  os=$(uname -s | tr '[:upper:]' '[:lower:]')
  case "$os" in
    msys*) os="windows" ;;
    mingw*) os="windows" ;;
  esac
  echo "$os"
}

uname_arch() {
  arch=$(uname -m)
  case $arch in
    x86_64) arch="amd64" ;;
    arm64) arch="arm64" ;;
    aarch64) arch="arm64" ;;
    armv5*) arch="armv5" ;;
    armv6*) arch="armv6" ;;
    armv7*) arch="armv7" ;;
  esac
  echo ${arch}
}

uname_os_check() {
  os=$(uname_os)
  case "$os" in
    darwin) return 0 ;;
    dragonfly) return 0 ;;
    freebsd) return 0 ;;
    linux) return 0 ;;
    android) return 0 ;;
    nacl) return 0 ;;
    netbsd) return 0 ;;
    openbsd) return 0 ;;
    plan9) return 0 ;;
    solaris) return 0 ;;
    windows) return 0 ;;
  esac
  log_crit "uname_os_check '$(uname -s)' got converted to '$os' which is not a GOOS value. Please file bug at https://github.com/client9/shlib"
  return 1
}

uname_arch_check() {
  arch=$(uname_arch)
  case "$arch" in
    amd64) return 0 ;;
    arm64) return 0 ;;
    armv5) return 0 ;;
    armv6) return 0 ;;
    armv7) return 0 ;;
    ppc64) return 0 ;;
    ppc64le) return 0 ;;
    mips) return 0 ;;
    mipsle) return 0 ;;
    mips64) return 0 ;;
    mips64le) return 0 ;;
    s390x) return 0 ;;
    amd64p32) return 0 ;;
  esac
  log_crit "uname_arch_check '$(uname -m)' got converted to '$arch' which is not a GOARCH value.  Please file bug report at https://github.com/client9/shlib"
  return 1
}

cat /dev/null <<EOF
------------------------------------------------------------------------
End of functions from https://github.com/client9/shlib
------------------------------------------------------------------------
EOF

OWNER=confluentinc
REPO="cli"
BINARY=confluent
OS=$(uname_os)
ARCH=$(uname_arch)
PREFIX="${OWNER}/${REPO}"
EXE_PATH="${BASH_SOURCE%/*}/../libexec/cli"
PLATFORM="${OS}/${ARCH}"

# use in logging routines
log_prefix() {
	echo "${PREFIX}"
}

uname_os_check
uname_arch_check

check_platform

adjust_os

EXECUTABLE="${EXE_PATH}/${OS}_${ARCH}/${BINARY}"
if [[ "$OS" = "windows" ]]; then
  EXECUTABLE="${EXECUTABLE}.exe"
fi

check_executable
init_config

# call the underlying executable
${EXECUTABLE} "$@"

#/bin/sh
#/
#/ usage: ./bundle.sh [ --zip ] <output-dir> <release-dir> <version> <platform> <architecture> <executable-name>
#/
#/  Flags:
#/   --zip:         unique name of the module
#/   -h, --help:    print usage info
#/
#/  Positional Args:
#/   output-dir:
#/   release-dir:
#/   version:
#/   platform:
#/   architecture:
#/   executable-name:
#/
#/  Examples:
#/   ./bundle.sh ./output/binary/linux ./travis-releases 0.14.1-rc.5 linux amd64 step
#/   ./bundle.sh --zip ./output/binary/windows ./travis-releases 0.14.1-rc.5 windows amd64 step.exe
#/

usage() {
    cat "$0" | grep '^#/' | cut -c 4-
    echo "Exiting with $1"
    exit $1
}

ZIP=0
while [ $# -gt 0 ]
do
    case "$1" in
        "--zip") ZIP=1; shift 1;;
        "-h"|"--help") usage 0;;
        "--") break;;
        *) break;;
    esac
done
set -ex;

OUTPUT_DIR=$1
RELEASE_DIR=$2

STEP_VERSION=$3
STEP_PLATFORM=$4
STEP_ARCH=$5
STEP_EXEC_NAME=$6

BUNDLE_DIR=${OUTPUT_DIR}/bundle

mkdir -p "$BUNDLE_DIR" "$RELEASE_DIR"
TMP=$(mktemp -d "$BUNDLE_DIR/tmp.XXXX")
trap 'rm -rf "$TMP"; exit' EXIT INT QUIT TERM

stepName=step_${STEP_VERSION}
newdir="$TMP/${stepName}"
mkdir -p "$newdir/bin"

cp "$OUTPUT_DIR/bin/step" "$newdir/bin/${STEP_EXEC_NAME}"
cp README.md "$newdir"

if [ ${ZIP} -eq 0 ]; then
    NEW_BUNDLE="${RELEASE_DIR}/step_${STEP_PLATFORM}_${STEP_VERSION}_${STEP_ARCH}.tar.gz"

    rm -f "$NEW_BUNDLE"
    tar -zcvf "$NEW_BUNDLE" -C "$TMP" "${stepName}"
else
    NEW_BUNDLE="${RELEASE_DIR}/step_${STEP_PLATFORM}_${STEP_VERSION}_${STEP_ARCH}.zip"

    rm -f "${NEW_BUNDLE}"
    zip -jr "${NEW_BUNDLE}" "${TMP}"
fi

#!/bin/sh
set -e

# Code generated by godownloader on 2018-06-07T07:36:49Z.
# Manually modified to download from S3 for confluentinc/cli goreleaser config
#
# The major modifications include:
# * reworked `github_release` into `s3_release` function
# * updated `TARBALL_URL` and `CHECKSUM_URL` to point to S3 instead of GitHub API
# * added a new `-l` flag to list versions from S3, since we can't link to our (private) GitHub repo
# * extracted a `BINARY` variable instead of having binary names hardcoded in `execute`
# * updated version/tag handling of the `v` prefix; it's expected in GitHub and inconsistently used in S3
# * updated the usage message, logging, and file comments a bit

S3_URL=https://s3-us-west-2.amazonaws.com/confluent.cloud
S3_CONTENT_CHECK_URL="${S3_URL}?prefix="
if [ -n "$OVERRIDE_S3_FOLDER" ]
then
	S3_CONTENT_CHECK_URL="${S3_URL}?prefix=${OVERRIDE_S3_FOLDER}/"
	S3_URL=${S3_URL}/${OVERRIDE_S3_FOLDER}
fi

usage() {
  this=$1
  cat <<EOF
$this: download binaries for confluentinc/cli

Usage: $this [-b] bindir [-d] [tag] | -l
  -b sets bindir or installation directory, Defaults to ./bin
  -d turns on debug logging
   [tag] is a valid version tag shown with -l
   If tag is missing, then the latest will be used.

  -l returns a list of all available tags/versions

EOF
  exit 2
}

parse_args() {
  #BINDIR is ./bin unless set be ENV
  # over-ridden by flag below

  BINDIR=${BINDIR:-./bin}
  while getopts "b:ldh?" arg; do
    case "$arg" in
      b) BINDIR="$OPTARG" ;;
      l) s3_releases ; exit 0 ;;
      d) log_set_priority 10 ;;
      h | \?) usage "$0" ;;
    esac
  done
  shift $((OPTIND - 1))
  TAG=$1
}
# this function wraps all the destructive operations
# if a curl|bash cuts off the end of the script due to
# network, either nothing will happen or will syntax error
# out preventing half-done work
execute() {
  tmpdir=$(mktmpdir)
  log_debug "downloading files into ${tmpdir}"
  http_download "${tmpdir}/${TARBALL}" "${TARBALL_URL}" "Accept:application/octet-stream"
  http_download "${tmpdir}/${CHECKSUM}" "${CHECKSUM_URL}" "Accept:application/octet-stream"
  hash_sha256_verify "${tmpdir}/${TARBALL}" "${tmpdir}/${CHECKSUM}"
  srcdir="${tmpdir}/${BINARY}"
  rm -rf "${srcdir}"
  (cd "${tmpdir}" && untar "${TARBALL}")
  install -d "${BINDIR}"
  for binexe in "${BINARY}" ; do
    if [ "$OS" = "windows" ]; then
      binexe="${binexe}.exe"
    fi
    install "${srcdir}/${binexe}" "${BINDIR}/"
    log_info "NOTICE: see licenses located in ${tmpdir}/${BINARY}"
    log_info "installed ${BINDIR}/${binexe}"
    log_info "please ensure ${BINDIR} is in your PATH"
  done
}
is_supported_platform() {
  platform=$1
  found=1
  case "$platform" in
    alpine/amd64) found=0 ;;
    alpine/arm64) found=0 ;;
    linux/amd64) found=0 ;;
    linux/arm64) found=0 ;;
    darwin/amd64) found=0 ;;
    darwin/arm64) found=0 ;;
    windows/amd64) found=0 ;;
  esac
  return $found
}
check_platform() {
  if is_supported_platform "$PLATFORM"; then
    # optional logging goes here
    true
  else
    log_crit "platform $PLATFORM is not supported.  Please contact Confluent support if you believe this is a mistake."
    exit 1
  fi
}
tag_to_version() {
  if [ -z "${TAG}" ]; then
    log_info "checking S3 for latest tag"
  else
    log_info "checking S3 for tag '${TAG}'"
  fi
  REALTAG=$(s3_release "${TAG}") && true
  if test -z "$REALTAG"; then
    log_crit "unable to find '${TAG}' - use 'latest' or see https://docs.confluent.io/${PROJECT_NAME}/current/release-notes.html for available versions."
    exit 1
  fi
  # if version starts with 'v', don't remove it
  TAG="$REALTAG"
  VERSION=${TAG}
}
adjust_format() {
  # change format (tar.gz or zip) based on ARCH
  case ${ARCH} in
    windows) FORMAT=zip ;;
  esac
  true
}
adjust_os() {
  # adjust archive name based on OS
  case ${OS} in
    amd64) OS=x86_64 ;;
    darwin) OS=darwin ;;
    linux) OS=linux ;;
    alpine) OS=alpine ;;
    windows) OS=windows ;;
  esac
  true
}
s3_releases() {
  s3url="${S3_CONTENT_CHECK_URL}${PROJECT_NAME}/archives/&delimiter=/"
  xml=$(http_copy "$s3url")
  versions=$(echo "$xml" | sed -n 's/</\
</gp' | sed -n "s/<Prefix>${PROJECT_NAME}\/archives\/\(.*\)\//\1/p") || return 1
  test -z "$versions" && return 1
  echo "$versions" | sort --version-sort
}
s3_release() {
  version=$1
  test -z "$version" && version="latest"
  s3url="${S3_CONTENT_CHECK_URL}${PROJECT_NAME}/archives/${version#v}/&delimiter=/"
  xml=$(http_copy "$s3url")
  exists=$(echo "$xml" | grep "<Key>") || return 1
  test -z "$version" && return 1
  echo "$version"
}

cat /dev/null <<EOF
------------------------------------------------------------------------
https://github.com/client9/shlib - portable posix shell functions
Public domain - http://unlicense.org
https://github.com/client9/shlib/blob/master/LICENSE.md
but credit (and pull requests) appreciated.
------------------------------------------------------------------------
EOF
is_command() {
  command -v "$1" >/dev/null
}
echoerr() {
  echo "$@" 1>&2
}
log_prefix() {
  echo "$0"
}
_logp=6
log_set_priority() {
  _logp="$1"
}
log_priority() {
  if test -z "$1"; then
    echo "$_logp"
    return
  fi
  [ "$1" -le "$_logp" ]
}
log_tag() {
  case $1 in
    0) echo "emerg" ;;
    1) echo "alert" ;;
    2) echo "crit" ;;
    3) echo "err" ;;
    4) echo "warning" ;;
    5) echo "notice" ;;
    6) echo "info" ;;
    7) echo "debug" ;;
    *) echo "$1" ;;
  esac
}
log_debug() {
  log_priority 7 || return 0
  echoerr "$(log_prefix)" "$(log_tag 7)" "$@"
}
log_info() {
  log_priority 6 || return 0
  echoerr "$(log_prefix)" "$(log_tag 6)" "$@"
}
log_err() {
  log_priority 3 || return 0
  echoerr "$(log_prefix)" "$(log_tag 3)" "$@"
}
log_crit() {
  log_priority 2 || return 0
  echoerr "$(log_prefix)" "$(log_tag 2)" "$@"
}
uname_os() {
  os=$(uname -s | tr '[:upper:]' '[:lower:]')
  osid=$(awk -F= '$1=="ID" { print $2 ;}' /etc/os-release 2>/dev/null || true)
  case "$os" in
    msys*) os="windows" ;;
    mingw*) os="windows" ;;
    cygwin*) os="windows" ;;
  esac
  case "$osid" in
    alpine*) os="alpine" ;;
  esac
  echo "$os"
}
uname_arch() {
  arch=$(uname -m)
  case $arch in
    x86_64) arch="amd64" ;;
    arm64) arch="arm64" ;;
    aarch64) arch="arm64" ;;
    armv5*) arch="armv5" ;;
    armv6*) arch="armv6" ;;
    armv7*) arch="armv7" ;;
  esac
  echo ${arch}
}
uname_os_check() {
  os=$(uname_os)
  case "$os" in
    darwin) return 0 ;;
    dragonfly) return 0 ;;
    freebsd) return 0 ;;
    linux) return 0 ;;
    alpine) return 0 ;;
    android) return 0 ;;
    nacl) return 0 ;;
    netbsd) return 0 ;;
    openbsd) return 0 ;;
    plan9) return 0 ;;
    solaris) return 0 ;;
    windows) return 0 ;;
  esac
  log_crit "uname_os_check '$(uname -s)' got converted to '$os' which is not a GOOS value. Please file bug at https://github.com/client9/shlib"
  return 1
}
uname_arch_check() {
  arch=$(uname_arch)
  case "$arch" in
    amd64) return 0 ;;
    arm64) return 0 ;;
    armv5) return 0 ;;
    armv6) return 0 ;;
    armv7) return 0 ;;
    ppc64) return 0 ;;
    ppc64le) return 0 ;;
    mips) return 0 ;;
    mipsle) return 0 ;;
    mips64) return 0 ;;
    mips64le) return 0 ;;
    s390x) return 0 ;;
    amd64p32) return 0 ;;
  esac
  log_crit "uname_arch_check '$(uname -m)' got converted to '$arch' which is not a GOARCH value.  Please file bug report at https://github.com/client9/shlib"
  return 1
}
untar() {
  tarball=$1
  case "${tarball}" in
    *.tar.gz | *.tgz) tar -xzf "${tarball}" ;;
    *.tar) tar -xf "${tarball}" ;;
    *.zip) unzip "${tarball}" ;;
    *)
      log_err "untar unknown archive format for ${tarball}"
      return 1
      ;;
  esac
}
mktmpdir() {
  test -z "$TMPDIR" && TMPDIR="$(mktemp -d)"
  mkdir -p "${TMPDIR}"
  echo "${TMPDIR}"
}
http_download_curl() {
  local_file=$1
  source_url=$2
  header=$3
  if [ -z "$header" ]; then
    code=$(curl -n -w '%{http_code}' -sL -o "$local_file" "$source_url")
  else
    code=$(curl -n -w '%{http_code}' -sL -H "$header" -o "$local_file" "$source_url")
  fi
  if [ "$code" != "200" ]; then
    log_debug "http_download_curl received HTTP status $code"
    return 1
  fi
  return 0
}
http_download_wget() {
  local_file=$1
  source_url=$2
  header=$3
  if [ -z "$header" ]; then
    wget -q -O "$local_file" "$source_url"
  else
    wget -q --header "$header" -O "$local_file" "$source_url"
  fi
}
http_download() {
  log_debug "http_download $2"
  if is_command curl; then
    http_download_curl "$@"
    return
  elif is_command wget; then
    http_download_wget "$@"
    return
  fi
  log_crit "http_download unable to find wget or curl"
  return 1
}
http_copy() {
  tmp=$(mktemp)
  http_download "${tmp}" "$1" "$2" || return 1
  body=$(cat "$tmp")
  rm -f "${tmp}"
  echo "$body"
}
hash_sha256() {
  TARGET=${1:-/dev/stdin}
  if is_command gsha256sum; then
    hash=$(gsha256sum "$TARGET") || return 1
    echo "$hash" | cut -d ' ' -f 1
  elif is_command sha256sum; then
    hash=$(sha256sum "$TARGET") || return 1
    echo "$hash" | cut -d ' ' -f 1
  elif is_command shasum; then
    hash=$(shasum -a 256 "$TARGET" 2>/dev/null) || return 1
    echo "$hash" | cut -d ' ' -f 1
  elif is_command openssl; then
    hash=$(openssl -dst openssl dgst -sha256 "$TARGET") || return 1
    echo "$hash" | cut -d ' ' -f a
  else
    log_crit "hash_sha256 unable to find command to compute sha-256 hash"
    return 1
  fi
}
hash_sha256_verify() {
  TARGET=$1
  checksums=$2
  if [ -z "$checksums" ]; then
    log_err "hash_sha256_verify checksum file not specified in arg2"
    return 1
  fi
  BASENAME=${TARGET##*/}
  want=$(grep "${BASENAME}" "${checksums}" 2>/dev/null | tr '\t' ' ' | cut -d ' ' -f 1)
  if [ -z "$want" ]; then
    log_err "hash_sha256_verify unable to find checksum for '${TARGET}' in '${checksums}'"
    return 1
  fi
  got=$(hash_sha256 "$TARGET")
  if [ "$want" != "$got" ]; then
    log_err "hash_sha256_verify checksum for '$TARGET' did not verify ${want} vs $got"
    return 1
  fi
}
cat /dev/null <<EOF
------------------------------------------------------------------------
End of functions from https://github.com/client9/shlib
------------------------------------------------------------------------
EOF

PROJECT_NAME="confluent-cli"
OWNER=confluentinc
REPO="cli"
BINARY=confluent
FORMAT=tar.gz
OS=$(uname_os)
ARCH=$(uname_arch)
PREFIX="$OWNER/$REPO"

# use in logging routines
log_prefix() {
	echo "$PREFIX"
}
PLATFORM="${OS}/${ARCH}"

main() {
  uname_os_check
  uname_arch_check

  parse_args "$@"

  check_platform

  tag_to_version

  adjust_format

  adjust_os

  if [ ${OS} = "windows" ]; then
    FORMAT=zip ;
  fi

  log_info "found version: ${VERSION} for ${TAG}/${OS}/${ARCH}"

  # If < v3, archive version is prefixed with "v"
  VERSION=${VERSION#v}
  VERSION=$([ "${VERSION#3.}" = "${VERSION}" ] && echo "v${VERSION}" || echo "${VERSION}")
  VERSION=$([ "${VERSION}" = "vlatest" ] && echo "latest" || echo "${VERSION}")
  
  S3_ARCHIVES_URL=${S3_URL}/${PROJECT_NAME}/archives/${VERSION#v}
  NAME=${BINARY}_${VERSION}_${OS}_${ARCH}
  TARBALL=${NAME}.${FORMAT}
  TARBALL_URL=${S3_ARCHIVES_URL}/${TARBALL}
  CHECKSUM=${BINARY}_${VERSION}_checksums.txt
  CHECKSUM_URL=${S3_ARCHIVES_URL}/${CHECKSUM}

  execute
}

if [ "${TEST}" != "true" ]; then
  main "$@"
fi

#!/bin/sh
set -e
# Code generated by godownloader. DO NOT EDIT.
#

usage() {
  this=$1
  cat <<EOF
$this: download go binaries for golangci/golangci-lint

Usage: $this [-b] bindir [-d] [tag]
  -b sets bindir or installation directory, Defaults to ./bin
  -d turns on debug logging
   [tag] is a tag from
   https://github.com/golangci/golangci-lint/releases
   If tag is missing, then the latest will be used.

 Generated by godownloader
  https://github.com/goreleaser/godownloader

EOF
  exit 2
}

parse_args() {
  #BINDIR is ./bin unless set be ENV
  # over-ridden by flag below

  BINDIR=${BINDIR:-./bin}
  while getopts "b:dh?x" arg; do
    case "$arg" in
      b) BINDIR="$OPTARG" ;;
      d) log_set_priority 10 ;;
      h | \?) usage "$0" ;;
      x) set -x ;;
    esac
  done
  shift $((OPTIND - 1))
  TAG=$1
}
# this function wraps all the destructive operations
# if a curl|bash cuts off the end of the script due to
# network, either nothing will happen or will syntax error
# out preventing half-done work
execute() {
  tmpdir=$(mktemp -d)
  log_debug "downloading files into ${tmpdir}"
  http_download "${tmpdir}/${TARBALL}" "${TARBALL_URL}"
  http_download "${tmpdir}/${CHECKSUM}" "${CHECKSUM_URL}"
  hash_sha256_verify "${tmpdir}/${TARBALL}" "${tmpdir}/${CHECKSUM}"
  srcdir="${tmpdir}/${NAME}"
  rm -rf "${srcdir}"
  (cd "${tmpdir}" && untar "${TARBALL}")
  test ! -d "${BINDIR}" && install -d "${BINDIR}"
  for binexe in $BINARIES; do
    if [ "$OS" = "windows" ]; then
      binexe="${binexe}.exe"
    fi
    install "${srcdir}/${binexe}" "${BINDIR}/"
    log_info "installed ${BINDIR}/${binexe}"
  done
  rm -rf "${tmpdir}"
}
get_binaries() {
  case "$PLATFORM" in
    darwin/amd64) BINARIES="golangci-lint" ;;
    darwin/arm64) BINARIES="golangci-lint" ;;
    darwin/armv6) BINARIES="golangci-lint" ;;
    darwin/armv7) BINARIES="golangci-lint" ;;
    darwin/mips64) BINARIES="golangci-lint" ;;
    darwin/mips64le) BINARIES="golangci-lint" ;;
    darwin/ppc64le) BINARIES="golangci-lint" ;;
    darwin/s390x) BINARIES="golangci-lint" ;;
    freebsd/386) BINARIES="golangci-lint" ;;
    freebsd/amd64) BINARIES="golangci-lint" ;;
    freebsd/armv6) BINARIES="golangci-lint" ;;
    freebsd/armv7) BINARIES="golangci-lint" ;;
    freebsd/mips64) BINARIES="golangci-lint" ;;
    freebsd/mips64le) BINARIES="golangci-lint" ;;
    freebsd/ppc64le) BINARIES="golangci-lint" ;;
    freebsd/s390x) BINARIES="golangci-lint" ;;
    linux/386) BINARIES="golangci-lint" ;;
    linux/amd64) BINARIES="golangci-lint" ;;
    linux/arm64) BINARIES="golangci-lint" ;;
    linux/armv6) BINARIES="golangci-lint" ;;
    linux/armv7) BINARIES="golangci-lint" ;;
    linux/mips64) BINARIES="golangci-lint" ;;
    linux/mips64le) BINARIES="golangci-lint" ;;
    linux/ppc64le) BINARIES="golangci-lint" ;;
    linux/s390x) BINARIES="golangci-lint" ;;
    windows/386) BINARIES="golangci-lint" ;;
    windows/amd64) BINARIES="golangci-lint" ;;
    windows/arm64) BINARIES="golangci-lint" ;;
    windows/armv6) BINARIES="golangci-lint" ;;
    windows/armv7) BINARIES="golangci-lint" ;;
    windows/mips64) BINARIES="golangci-lint" ;;
    windows/mips64le) BINARIES="golangci-lint" ;;
    windows/ppc64le) BINARIES="golangci-lint" ;;
    windows/s390x) BINARIES="golangci-lint" ;;
    *)
      log_crit "platform $PLATFORM is not supported.  Make sure this script is up-to-date and file request at https://github.com/${PREFIX}/issues/new"
      exit 1
      ;;
  esac
}
tag_to_version() {
  if [ -z "${TAG}" ]; then
    log_info "checking GitHub for latest tag"
  else
    log_info "checking GitHub for tag '${TAG}'"
  fi
  REALTAG=$(github_release "$OWNER/$REPO" "${TAG}") && true
  if test -z "$REALTAG"; then
    log_crit "unable to find '${TAG}' - use 'latest' or see https://github.com/${PREFIX}/releases for details"
    exit 1
  fi
  # if version starts with 'v', remove it
  TAG="$REALTAG"
  VERSION=${TAG#v}
}
adjust_format() {
  # change format (tar.gz or zip) based on OS
  case ${OS} in
    windows) FORMAT=zip ;;
  esac
  true
}
adjust_os() {
  # adjust archive name based on OS
  true
}
adjust_arch() {
  # adjust archive name based on ARCH
  true
}

cat /dev/null <<EOF
------------------------------------------------------------------------
https://github.com/client9/shlib - portable posix shell functions
Public domain - http://unlicense.org
https://github.com/client9/shlib/blob/master/LICENSE.md
but credit (and pull requests) appreciated.
------------------------------------------------------------------------
EOF
is_command() {
  command -v "$1" >/dev/null
}
echoerr() {
  echo "$@" 1>&2
}
log_prefix() {
  echo "$0"
}
_logp=6
log_set_priority() {
  _logp="$1"
}
log_priority() {
  if test -z "$1"; then
    echo "$_logp"
    return
  fi
  [ "$1" -le "$_logp" ]
}
log_tag() {
  case $1 in
    0) echo "emerg" ;;
    1) echo "alert" ;;
    2) echo "crit" ;;
    3) echo "err" ;;
    4) echo "warning" ;;
    5) echo "notice" ;;
    6) echo "info" ;;
    7) echo "debug" ;;
    *) echo "$1" ;;
  esac
}
log_debug() {
  log_priority 7 || return 0
  echoerr "$(log_prefix)" "$(log_tag 7)" "$@"
}
log_info() {
  log_priority 6 || return 0
  echoerr "$(log_prefix)" "$(log_tag 6)" "$@"
}
log_err() {
  log_priority 3 || return 0
  echoerr "$(log_prefix)" "$(log_tag 3)" "$@"
}
log_crit() {
  log_priority 2 || return 0
  echoerr "$(log_prefix)" "$(log_tag 2)" "$@"
}
uname_os() {
  os=$(uname -s | tr '[:upper:]' '[:lower:]')
  case "$os" in
    cygwin_nt*) os="windows" ;;
    mingw*) os="windows" ;;
    msys_nt*) os="windows" ;;
  esac
  echo "$os"
}
uname_arch() {
  arch=$(uname -m)
  case $arch in
    x86_64) arch="amd64" ;;
    x86) arch="386" ;;
    i686) arch="386" ;;
    i386) arch="386" ;;
    aarch64) arch="arm64" ;;
    armv5*) arch="armv5" ;;
    armv6*) arch="armv6" ;;
    armv7*) arch="armv7" ;;
  esac
  echo ${arch}
}
uname_os_check() {
  os=$(uname_os)
  case "$os" in
    darwin) return 0 ;;
    dragonfly) return 0 ;;
    freebsd) return 0 ;;
    linux) return 0 ;;
    android) return 0 ;;
    nacl) return 0 ;;
    netbsd) return 0 ;;
    openbsd) return 0 ;;
    plan9) return 0 ;;
    solaris) return 0 ;;
    windows) return 0 ;;
  esac
  log_crit "uname_os_check '$(uname -s)' got converted to '$os' which is not a GOOS value. Please file bug at https://github.com/client9/shlib"
  return 1
}
uname_arch_check() {
  arch=$(uname_arch)
  case "$arch" in
    386) return 0 ;;
    amd64) return 0 ;;
    arm64) return 0 ;;
    armv5) return 0 ;;
    armv6) return 0 ;;
    armv7) return 0 ;;
    ppc64) return 0 ;;
    ppc64le) return 0 ;;
    mips) return 0 ;;
    mipsle) return 0 ;;
    mips64) return 0 ;;
    mips64le) return 0 ;;
    s390x) return 0 ;;
    amd64p32) return 0 ;;
  esac
  log_crit "uname_arch_check '$(uname -m)' got converted to '$arch' which is not a GOARCH value.  Please file bug report at https://github.com/client9/shlib"
  return 1
}
untar() {
  tarball=$1
  case "${tarball}" in
    *.tar.gz | *.tgz) tar --no-same-owner -xzf "${tarball}" ;;
    *.tar) tar --no-same-owner -xf "${tarball}" ;;
    *.zip) unzip "${tarball}" ;;
    *)
      log_err "untar unknown archive format for ${tarball}"
      return 1
      ;;
  esac
}
http_download_curl() {
  local_file=$1
  source_url=$2
  header=$3
  if [ -z "$header" ]; then
    code=$(curl -w '%{http_code}' -sL -o "$local_file" "$source_url")
  else
    code=$(curl -w '%{http_code}' -sL -H "$header" -o "$local_file" "$source_url")
  fi
  if [ "$code" != "200" ]; then
    log_debug "http_download_curl received HTTP status $code"
    return 1
  fi
  return 0
}
http_download_wget() {
  local_file=$1
  source_url=$2
  header=$3
  if [ -z "$header" ]; then
    wget -q -O "$local_file" "$source_url"
  else
    wget -q --header "$header" -O "$local_file" "$source_url"
  fi
}
http_download() {
  log_debug "http_download $2"
  if is_command curl; then
    http_download_curl "$@"
    return
  elif is_command wget; then
    http_download_wget "$@"
    return
  fi
  log_crit "http_download unable to find wget or curl"
  return 1
}
http_copy() {
  tmp=$(mktemp)
  http_download "${tmp}" "$1" "$2" || return 1
  body=$(cat "$tmp")
  rm -f "${tmp}"
  echo "$body"
}
github_release() {
  owner_repo=$1
  version=$2
  test -z "$version" && version="latest"
  giturl="https://github.com/${owner_repo}/releases/${version}"
  json=$(http_copy "$giturl" "Accept:application/json")
  test -z "$json" && return 1
  version=$(echo "$json" | tr -s '\n' ' ' | sed 's/.*"tag_name":"//' | sed 's/".*//')
  test -z "$version" && return 1
  echo "$version"
}
hash_sha256() {
  TARGET=${1:-/dev/stdin}
  if is_command gsha256sum; then
    hash=$(gsha256sum "$TARGET") || return 1
    echo "$hash" | cut -d ' ' -f 1
  elif is_command sha256sum; then
    hash=$(sha256sum "$TARGET") || return 1
    echo "$hash" | cut -d ' ' -f 1
  elif is_command shasum; then
    hash=$(shasum -a 256 "$TARGET" 2>/dev/null) || return 1
    echo "$hash" | cut -d ' ' -f 1
  elif is_command openssl; then
    hash=$(openssl -dst openssl dgst -sha256 "$TARGET") || return 1
    echo "$hash" | cut -d ' ' -f a
  else
    log_crit "hash_sha256 unable to find command to compute sha-256 hash"
    return 1
  fi
}
hash_sha256_verify() {
  TARGET=$1
  checksums=$2
  if [ -z "$checksums" ]; then
    log_err "hash_sha256_verify checksum file not specified in arg2"
    return 1
  fi
  BASENAME=${TARGET##*/}
  want=$(grep "${BASENAME}" "${checksums}" 2>/dev/null | tr '\t' ' ' | cut -d ' ' -f 1)
  if [ -z "$want" ]; then
    log_err "hash_sha256_verify unable to find checksum for '${TARGET}' in '${checksums}'"
    return 1
  fi
  got=$(hash_sha256 "$TARGET")
  if [ "$want" != "$got" ]; then
    log_err "hash_sha256_verify checksum for '$TARGET' did not verify ${want} vs $got"
    return 1
  fi
}
cat /dev/null <<EOF
------------------------------------------------------------------------
End of functions from https://github.com/client9/shlib
------------------------------------------------------------------------
EOF

PROJECT_NAME="golangci-lint"
OWNER=golangci
REPO="golangci-lint"
BINARY=golangci-lint
FORMAT=tar.gz
OS=$(uname_os)
ARCH=$(uname_arch)
PREFIX="$OWNER/$REPO"

# use in logging routines
log_prefix() {
	echo "$PREFIX"
}
PLATFORM="${OS}/${ARCH}"
GITHUB_DOWNLOAD=https://github.com/${OWNER}/${REPO}/releases/download

uname_os_check "$OS"
uname_arch_check "$ARCH"

parse_args "$@"

get_binaries

tag_to_version

adjust_format

adjust_os

adjust_arch

log_info "found version: ${VERSION} for ${TAG}/${OS}/${ARCH}"

NAME=${BINARY}-${VERSION}-${OS}-${ARCH}
TARBALL=${NAME}.${FORMAT}
TARBALL_URL=${GITHUB_DOWNLOAD}/${TAG}/${TARBALL}
CHECKSUM=${PROJECT_NAME}-${VERSION}-checksums.txt
CHECKSUM_URL=${GITHUB_DOWNLOAD}/${TAG}/${CHECKSUM}


execute

echo ""
$step_path = step.exe path --base
if ( Test-Path -Path $step_path -PathType Container ) {
	$Confirm = Read-Host "Remove user configuration directory '$step_path'? [Y/N]"
	if ($Confirm -match 'y') {
		echo "Removing user configuration directory: '$step_path'"
		rm -R $step_path
	}
}
$install_dir = join-path $env:ProgramW6432 "Smallstep Labs"
echo "Uninstalling step: '$install_dir'"
rm -R $install_dir
echo 'Removing step.exe from the Machine $Path'
$path = [System.Environment]::GetEnvironmentVariable("Path", "Machine")
$path = @($Env:Path.split(";") | where {$_ -ne $install_dir}) -join ";"
[System.Environment]::SetEnvironmentVariable("Path", $path, "Machine")
echo "Disabling ssh-agent..."
Stop-Service ssh-agent
Set-Service -Name ssh-agent -StartupType Disabled
Get-Service ssh-agent | Select Name, Status, StartType
echo ""
echo "Done!"

# Create install directory
echo ""
$install_dir = join-path $env:ProgramW6432 "Smallstep Labs"
echo "Installing step: '$install_dir'"
mkdir $install_dir >$null

# Determine filename to download
$response = Invoke-Webrequest -Uri "https://api.github.com/repos/smallstep/cli/releases/latest" -UseBasicParsing
$release = $response | ConvertFrom-Json

# Download the binary and uninstall script
echo ""
$exe_file = join-path $install_dir "step.exe"
curl.exe -L -o $exe_file "https://dl.step.sm/s3/cli/s3-windows-installer/step_latest_windows.exe"
$uninstall_file = join-path $install_dir "uninstall-step.ps1"
curl.exe -L -o $uninstall_file https://dl.step.sm/s3/cli/s3-windows-installer/uninstall-step.ps1
echo ""

echo 'Adding step.exe to the Machine $Path'
$path = [System.Environment]::GetEnvironmentVariable("Path", "Machine")
$path = @($Env:Path.split(";") | where {$_ -ne $install_dir}) -join ";"
[System.Environment]::SetEnvironmentVariable("Path", $path + ";$install_dir", "Machine")

echo ""
echo "Enabling ssh-agent..."
Set-Service -Name ssh-agent -StartupType Automatic
Start-Service ssh-agent
Get-Service ssh-agent | Select Name, Status, StartType
echo ""
echo "Please restart shell sessions to pickup path changes."

#!/bin/sh

updateAlternatives() {
	update-alternatives --install /usr/bin/step step /usr/bin/step-cli 50
}

updateCompletion() {
  /usr/bin/step completion bash > /usr/share/bash-completion/completions/step
  chmod 644 /usr/share/bash-completion/completions/step
}

cleanInstall() {
	updateAlternatives
	updateCompletion
}

upgrade() {
	updateAlternatives
	updateCompletion
}

action="$1"
if [ "$1" = "configure" ] && [ -z "$2" ]; then
	action="install"
elif [ "$1" = "configure" ] && [ -n "$2" ]; then
	action="upgrade"
fi

case "$action" in
	"1" | "install")
		cleanInstall
		;;
	"2" | "upgrade")
		upgrade
		;;
	*)
		cleanInstall
		;;
esac

#!/bin/sh

removeAlternatives() {
	update-alternatives --remove step /usr/bin/step-cli
}

upgrade() {
	:
}

remove() {
	removeAlternatives
}

action="$1"
if [ "$1" = "remove" ]; then
	action="remove"
elif [ "$1" = "upgrade" ] && [ -n "$2" ]; then
	action="upgrade"
elif [ "$1" = "disappear" ]; then
	action="remove"
fi

case "$action" in
	"0" | "remove")
		remove
		;;
	"1" | "upgrade")
		upgrade
		;;
	*)
		remove
		;;
esac

#!/usr/bin/env bash
set -e

# values set in test loop below
TEST_OS=
TEST_ARCH=
ARCHIVES_VERSION=$1

uname() {
  if [[ "$1" = "-s" ]]; then
    echo ${TEST_OS}
  elif [[ "$1" = "-m" ]]; then
    echo ${TEST_ARCH}
  fi
}

TEST=true . install.sh

# useful reference for valid uname system/machine pairs: https://en.wikipedia.org/wiki/Uname
for pair in Darwin,x86_64,darwin,amd64 Darwin,arm64,darwin,arm64 Linux,x86_64,linux,amd64 CYGWIN_NT-10.0,x86_64,windows,amd64; do
  # tip from https://stackoverflow.com/a/36393986/337735
  IFS=',' read TEST_OS TEST_ARCH EXPECT_OS EXPECT_ARCH <<< "${pair}"

  [[ "$(uname_os)" = "${EXPECT_OS}" ]] || ( echo "${TEST_OS}: got uname_os $(uname_os), want ${EXPECT_OS}" && exit 1 )
  [[ "$(uname_arch)" = "${EXPECT_ARCH}" ]] || ( echo "${TEST_OS}: got uname_arch $(uname_arch), want ${EXPECT_ARCH}" && exit 1 )

  uname_os_check
  uname_arch_check
done

binary="confluent"
TEST_OS=$(go env GOOS)
TEST_ARCH=$(go env GOARCH)
[[ -z "$ARCHIVES_VERSION" ]] && VERSION_TO_TEST="LATEST" || VERSION_TO_TEST=$ARCHIVES_VERSION
echo === TESTING installer script, VERSION: $VERSION_TO_TEST ===
output=$(./install.sh -d ${ARCHIVES_VERSION} 2>&1)
tmpdir=$(echo "${output}" | sed -n 's/.*licenses located in \(.*\)/\1/p')
echo "<install.sh output and debug log>:"
echo $output

ls "${tmpdir}" | grep -q "LICENSE" || ( echo "License file not found" && exit 1 )
[[ "$(find "${tmpdir}/legal/licenses" -type f | wc -l)" -ge 20 ]] || ( echo "Appears to be missing some licenses; found less than 20 in the tmp dir" && exit 1 )

./bin/${binary} -h 2>&1 >/dev/null | grep -q "Manage your .*" || ( echo "Unable to execute installed ${binary} CLI" && exit 1 )

echo "All tests passed!"

#!/bin/bash

export AWS_PROFILE=cc-production-1/prod-administrator
aws ecr get-login-password --region us-west-1 | docker login --username AWS --password-stdin 050879227952.dkr.ecr.us-west-1.amazonaws.com

new_tags="\nThe following tags ready to be pushed to ECR:\n"

arm64_tags=($(aws ecr describe-images --registry-id 050879227952 --repository-name confluentinc/cli-ubuntu-base-arm64 --query 'imageDetails[?imageTags.contains(@, `"arm64-latest"`)].imageTags' | jq '.[]' | jq -r '.[]'))
if [ ${#arm64_tags[@]} == 0 ]; then
  exit 1
elif [ ${#arm64_tags[@]} != 2 ]; then # check that there are exactly 2 tags for the image tagged w/ "arm64-latest"
  echo "error: there should be exactly two tags for the arm64-latest cli-ubuntu-base-arm64 image. Current tags are: \""${arm64_tags[@]}"\"; check the ECR"
  exit 1
else
  echo "Found tags \"${arm64_tags[@]}\" for confluentinc/cli-ubuntu-base-arm64"
fi

incremented_tag=""
for tag in ${arm64_tags[@]}; do
  if [ $tag != "arm64-latest" ]; then
    number_part=$(cut -f2 -d'-' <<< $tag) && \
    integer_tag=$(printf "%.0f" $number_part) && \
    incremented_tag+="arm64-$(($integer_tag + 1)).0" && \
    docker build --no-cache --file ./docker/Dockerfile_linux_arm64_base --tag "050879227952.dkr.ecr.us-west-1.amazonaws.com/confluentinc/cli-ubuntu-base-arm64:$incremented_tag" . && \
    docker tag "050879227952.dkr.ecr.us-west-1.amazonaws.com/confluentinc/cli-ubuntu-base-arm64:$incremented_tag" "050879227952.dkr.ecr.us-west-1.amazonaws.com/confluentinc/cli-ubuntu-base-arm64:arm64-latest" && \
    new_tags+="docker push 050879227952.dkr.ecr.us-west-1.amazonaws.com/confluentinc/cli-ubuntu-base-arm64:$incremented_tag\n" && \
    new_tags+="docker push 050879227952.dkr.ecr.us-west-1.amazonaws.com/confluentinc/cli-ubuntu-base-arm64:arm64-latest\n"
  fi
done

echo -e $new_tags

read -p "Push these tags to ECR? (y/n)"
if [ $REPLY == "y" ]; then
  docker push "050879227952.dkr.ecr.us-west-1.amazonaws.com/confluentinc/cli-ubuntu-base-arm64:$incremented_tag"
  docker push "050879227952.dkr.ecr.us-west-1.amazonaws.com/confluentinc/cli-ubuntu-base-arm64:arm64-latest"
fi

#!/bin/bash

export AWS_PROFILE=cc-production-1/prod-administrator
aws ecr get-login-password --region us-west-1 | docker login --username AWS --password-stdin 050879227952.dkr.ecr.us-west-1.amazonaws.com

new_tags="\nThe following tags ready to be pushed to ECR:\n"

amd64_tags=($(aws ecr describe-images --registry-id 050879227952 --repository-name confluentinc/cli-centos-base-amd64 --query 'imageDetails[?imageTags.contains(@, `"latest"`)].imageTags' | jq '.[]' | jq -r '.[]'))
if [ ${#amd64_tags[@]} == 0 ]; then
  exit 1
elif [ ${#amd64_tags[@]} != 2 ]; then # check that there are exactly 2 tags for the image tagged w/ "latest"
  echo "error: there should be exactly two tags for the latest cli-centos-base-amd64 image. Current tags are: \""${amd64_tags[@]}"\"; check the ECR"
  exit 1
else
  echo "Found tags \"${amd64_tags[@]}\" for confluentinc/cli-centos-base-amd64"
fi

incremented_tag=""
for tag in ${amd64_tags[@]}; do
  if [ $tag != "latest" ]; then
    integer_tag=$(printf "%.0f" $tag) && \
    incremented_tag+="$(($integer_tag + 1)).0" && \
    docker build --no-cache --file ./docker/Dockerfile_linux_amd64_base --tag "050879227952.dkr.ecr.us-west-1.amazonaws.com/confluentinc/cli-centos-base-amd64:$incremented_tag" . && \
    docker tag "050879227952.dkr.ecr.us-west-1.amazonaws.com/confluentinc/cli-centos-base-amd64:$incremented_tag" "050879227952.dkr.ecr.us-west-1.amazonaws.com/confluentinc/cli-centos-base-amd64:latest" && \
    new_tags+="050879227952.dkr.ecr.us-west-1.amazonaws.com/confluentinc/cli-centos-base-amd64:$incremented_tag\n" && \
    new_tags+="050879227952.dkr.ecr.us-west-1.amazonaws.com/confluentinc/cli-centos-base-amd64:latest\n"
  fi
done

echo -e $new_tags

read -p "Push these tags to ECR? (y/n)"
if [ $REPLY == "y" ]; then
  docker push "050879227952.dkr.ecr.us-west-1.amazonaws.com/confluentinc/cli-centos-base-amd64:$incremented_tag"
  docker push "050879227952.dkr.ecr.us-west-1.amazonaws.com/confluentinc/cli-centos-base-amd64:latest"
fi

# issue #154 [plain]

    Code
      cli_code("a\nb\nc")
    Message
      a
      b
      c

# issue #154 [ansi]

    Code
      cli_code("a\nb\nc")
    Message
      a
      b
      c

# issue #154 [unicode]

    Code
      cli_code("a\nb\nc")
    Message
      a
      b
      c

# issue #154 [fancy]

    Code
      cli_code("a\nb\nc")
    Message
      a
      b
      c

# stop_app() errors

    Code
      stop_app(1:10)
    Condition
      Error:
      ! `app` must be a CLI app
      i `app` is an integer vector

# warning if inactive app

    Code
      stop_app(app)
    Condition
      Warning in `stop_app()`:
      No app to end
    Output
      NULL

# ansi_html

    Code
      ansi_html(str)
    Output
       [1] "<span class=\"ansi ansi-bold\">bold</span>"                          
       [2] "<span class=\"ansi ansi-faint\">faint</span>"                        
       [3] "<span class=\"ansi ansi-italic\">italic</span>"                      
       [4] "<span class=\"ansi ansi-underline\">underline</span>"                
       [5] "<span class=\"ansi ansi-blink\">blink</span>"                        
       [6] "<span class=\"ansi ansi-inverse\">inverse</span>"                    
       [7] "<span class=\"ansi ansi-hide\">hide</span>"                          
       [8] "<span class=\"ansi ansi-crossedout\">crossedout</span>"              
       [9] "<span class=\"ansi ansi-color-0\">black</span>"                      
      [10] "<span class=\"ansi ansi-color-1\">red</span>"                        
      [11] "<span class=\"ansi ansi-color-2\">green</span>"                      
      [12] "<span class=\"ansi ansi-color-3\">yellow</span>"                     
      [13] "<span class=\"ansi ansi-color-4\">blue</span>"                       
      [14] "<span class=\"ansi ansi-color-5\">magenta</span>"                    
      [15] "<span class=\"ansi ansi-color-6\">cyan</span>"                       
      [16] "<span class=\"ansi ansi-color-7\">white</span>"                      
      [17] "<span class=\"ansi ansi-color-0\">bblack</span>"                     
      [18] "<span class=\"ansi ansi-color-1\">bred</span>"                       
      [19] "<span class=\"ansi ansi-color-2\">bgreen</span>"                     
      [20] "<span class=\"ansi ansi-color-3\">byellow</span>"                    
      [21] "<span class=\"ansi ansi-color-4\">bblue</span>"                      
      [22] "<span class=\"ansi ansi-color-5\">bmagenta</span>"                   
      [23] "<span class=\"ansi ansi-color-6\">bcyan</span>"                      
      [24] "<span class=\"ansi ansi-color-7\">bwhite</span>"                     
      [25] "<span class=\"ansi ansi-color-156\">color-156</span>"                
      [26] "<span class=\"ansi ansi-color-1-22-255\">color-1-22-255</span>"      
      [27] "<span class=\"ansi ansi-bg-color-0\">bg-black</span>"                
      [28] "<span class=\"ansi ansi-bg-color-1\">bg-red</span>"                  
      [29] "<span class=\"ansi ansi-bg-color-2\">bg-green</span>"                
      [30] "<span class=\"ansi ansi-bg-color-3\">bg-yellow</span>"               
      [31] "<span class=\"ansi ansi-bg-color-4\">bg-blue</span>"                 
      [32] "<span class=\"ansi ansi-bg-color-5\">bg-magenta</span>"              
      [33] "<span class=\"ansi ansi-bg-color-6\">bg-cyan</span>"                 
      [34] "<span class=\"ansi ansi-bg-color-7\">bg-white</span>"                
      [35] "<span class=\"ansi ansi-bg-color-0\">bg-bblack</span>"               
      [36] "<span class=\"ansi ansi-bg-color-1\">bg-bred</span>"                 
      [37] "<span class=\"ansi ansi-bg-color-2\">bg-bgreen</span>"               
      [38] "<span class=\"ansi ansi-bg-color-3\">bg-byellow</span>"              
      [39] "<span class=\"ansi ansi-bg-color-4\">bg-bblue</span>"                
      [40] "<span class=\"ansi ansi-bg-color-5\">bg-bmagenta</span>"             
      [41] "<span class=\"ansi ansi-bg-color-6\">bg-bcyan</span>"                
      [42] "<span class=\"ansi ansi-bg-color-7\">bg-bwhite</span>"               
      [43] "<span class=\"ansi ansi-bg-color-156\">bg-color-156</span>"          
      [44] "<span class=\"ansi ansi-bg-color-1-22-255\">bg-color-1-22-255</span>"

# multiple styles

    Code
      ansi_html("\033[1;2;35;45mmultiple")
    Output
      [1] "<span class=\"ansi ansi-bold ansi-faint ansi-color-5 ansi-bg-color-5\">multiple</span>"

# ansi_html_style

    Code
      ansi_html_style(colors = 8)
    Output
      .ansi-bold        { font-weight: bold;             }
      .ansi-italic      { font-style: italic;            }
      .ansi-underline   { text-decoration: underline;    }
      .ansi-blink       { text-decoration: blink;        }
      .ansi-hide        { visibility: hidden;            }
      .ansi-crossedout  { text-decoration: line-through; }
      .ansi-link:hover  { text-decoration: underline;    }
      .ansi-color-0     { color: #000000 }
      .ansi-color-1     { color: #cd3131 }
      .ansi-color-2     { color: #0dbc79 }
      .ansi-color-3     { color: #e5e510 }
      .ansi-color-4     { color: #2472c8 }
      .ansi-color-5     { color: #bc3fbc }
      .ansi-color-6     { color: #11a8cd }
      .ansi-color-7     { color: #e5e5e5 }
      .ansi-color-8     { color: #666666 }
      .ansi-color-9     { color: #f14c4c }
      .ansi-color-10    { color: #23d18b }
      .ansi-color-11    { color: #f5f543 }
      .ansi-color-12    { color: #3b8eea }
      .ansi-color-13    { color: #d670d6 }
      .ansi-color-14    { color: #29b8db }
      .ansi-color-15    { color: #e5e5e5 }
      .ansi-bg-color-0  { background-color: #000000 }
      .ansi-bg-color-1  { background-color: #cd3131 }
      .ansi-bg-color-2  { background-color: #0dbc79 }
      .ansi-bg-color-3  { background-color: #e5e510 }
      .ansi-bg-color-4  { background-color: #2472c8 }
      .ansi-bg-color-5  { background-color: #bc3fbc }
      .ansi-bg-color-6  { background-color: #11a8cd }
      .ansi-bg-color-7  { background-color: #e5e5e5 }
      .ansi-bg-color-8  { background-color: #666666 }
      .ansi-bg-color-9  { background-color: #f14c4c }
      .ansi-bg-color-10 { background-color: #23d18b }
      .ansi-bg-color-11 { background-color: #f5f543 }
      .ansi-bg-color-12 { background-color: #3b8eea }
      .ansi-bg-color-13 { background-color: #d670d6 }
      .ansi-bg-color-14 { background-color: #29b8db }
      .ansi-bg-color-15 { background-color: #e5e5e5 }

---

    Code
      ansi_html_style(colors = 256, palette = "ubuntu")
    Output
      .ansi-bold         { font-weight: bold;             }
      .ansi-italic       { font-style: italic;            }
      .ansi-underline    { text-decoration: underline;    }
      .ansi-blink        { text-decoration: blink;        }
      .ansi-hide         { visibility: hidden;            }
      .ansi-crossedout   { text-decoration: line-through; }
      .ansi-link:hover   { text-decoration: underline;    }
      .ansi-color-0      { color: #010101 }
      .ansi-color-1      { color: #de382b }
      .ansi-color-2      { color: #39b54a }
      .ansi-color-3      { color: #ffc706 }
      .ansi-color-4      { color: #006fb8 }
      .ansi-color-5      { color: #762671 }
      .ansi-color-6      { color: #2cb5e9 }
      .ansi-color-7      { color: #cccccc }
      .ansi-color-8      { color: #808080 }
      .ansi-color-9      { color: #ff0000 }
      .ansi-color-10     { color: #00ff00 }
      .ansi-color-11     { color: #ffff00 }
      .ansi-color-12     { color: #0000ff }
      .ansi-color-13     { color: #ff00ff }
      .ansi-color-14     { color: #00ffff }
      .ansi-color-15     { color: #ffffff }
      .ansi-bg-color-0   { background-color: #010101 }
      .ansi-bg-color-1   { background-color: #de382b }
      .ansi-bg-color-2   { background-color: #39b54a }
      .ansi-bg-color-3   { background-color: #ffc706 }
      .ansi-bg-color-4   { background-color: #006fb8 }
      .ansi-bg-color-5   { background-color: #762671 }
      .ansi-bg-color-6   { background-color: #2cb5e9 }
      .ansi-bg-color-7   { background-color: #cccccc }
      .ansi-bg-color-8   { background-color: #808080 }
      .ansi-bg-color-9   { background-color: #ff0000 }
      .ansi-bg-color-10  { background-color: #00ff00 }
      .ansi-bg-color-11  { background-color: #ffff00 }
      .ansi-bg-color-12  { background-color: #0000ff }
      .ansi-bg-color-13  { background-color: #ff00ff }
      .ansi-bg-color-14  { background-color: #00ffff }
      .ansi-bg-color-15  { background-color: #ffffff }
      .ansi-color-16     { color: #000000 }
      .ansi-color-52     { color: #330000 }
      .ansi-color-88     { color: #660000 }
      .ansi-color-124    { color: #990000 }
      .ansi-color-160    { color: #cc0000 }
      .ansi-color-196    { color: #ff0000 }
      .ansi-color-22     { color: #003300 }
      .ansi-color-58     { color: #333300 }
      .ansi-color-94     { color: #663300 }
      .ansi-color-130    { color: #993300 }
      .ansi-color-166    { color: #cc3300 }
      .ansi-color-202    { color: #ff3300 }
      .ansi-color-28     { color: #006600 }
      .ansi-color-64     { color: #336600 }
      .ansi-color-100    { color: #666600 }
      .ansi-color-136    { color: #996600 }
      .ansi-color-172    { color: #cc6600 }
      .ansi-color-208    { color: #ff6600 }
      .ansi-color-34     { color: #009900 }
      .ansi-color-70     { color: #339900 }
      .ansi-color-106    { color: #669900 }
      .ansi-color-142    { color: #999900 }
      .ansi-color-178    { color: #cc9900 }
      .ansi-color-214    { color: #ff9900 }
      .ansi-color-40     { color: #00cc00 }
      .ansi-color-76     { color: #33cc00 }
      .ansi-color-112    { color: #66cc00 }
      .ansi-color-148    { color: #99cc00 }
      .ansi-color-184    { color: #cccc00 }
      .ansi-color-220    { color: #ffcc00 }
      .ansi-color-46     { color: #00ff00 }
      .ansi-color-82     { color: #33ff00 }
      .ansi-color-118    { color: #66ff00 }
      .ansi-color-154    { color: #99ff00 }
      .ansi-color-190    { color: #ccff00 }
      .ansi-color-226    { color: #ffff00 }
      .ansi-color-17     { color: #000033 }
      .ansi-color-53     { color: #330033 }
      .ansi-color-89     { color: #660033 }
      .ansi-color-125    { color: #990033 }
      .ansi-color-161    { color: #cc0033 }
      .ansi-color-197    { color: #ff0033 }
      .ansi-color-23     { color: #003333 }
      .ansi-color-59     { color: #333333 }
      .ansi-color-95     { color: #663333 }
      .ansi-color-131    { color: #993333 }
      .ansi-color-167    { color: #cc3333 }
      .ansi-color-203    { color: #ff3333 }
      .ansi-color-29     { color: #006633 }
      .ansi-color-65     { color: #336633 }
      .ansi-color-101    { color: #666633 }
      .ansi-color-137    { color: #996633 }
      .ansi-color-173    { color: #cc6633 }
      .ansi-color-209    { color: #ff6633 }
      .ansi-color-35     { color: #009933 }
      .ansi-color-71     { color: #339933 }
      .ansi-color-107    { color: #669933 }
      .ansi-color-143    { color: #999933 }
      .ansi-color-179    { color: #cc9933 }
      .ansi-color-215    { color: #ff9933 }
      .ansi-color-41     { color: #00cc33 }
      .ansi-color-77     { color: #33cc33 }
      .ansi-color-113    { color: #66cc33 }
      .ansi-color-149    { color: #99cc33 }
      .ansi-color-185    { color: #cccc33 }
      .ansi-color-221    { color: #ffcc33 }
      .ansi-color-47     { color: #00ff33 }
      .ansi-color-83     { color: #33ff33 }
      .ansi-color-119    { color: #66ff33 }
      .ansi-color-155    { color: #99ff33 }
      .ansi-color-191    { color: #ccff33 }
      .ansi-color-227    { color: #ffff33 }
      .ansi-color-18     { color: #000066 }
      .ansi-color-54     { color: #330066 }
      .ansi-color-90     { color: #660066 }
      .ansi-color-126    { color: #990066 }
      .ansi-color-162    { color: #cc0066 }
      .ansi-color-198    { color: #ff0066 }
      .ansi-color-24     { color: #003366 }
      .ansi-color-60     { color: #333366 }
      .ansi-color-96     { color: #663366 }
      .ansi-color-132    { color: #993366 }
      .ansi-color-168    { color: #cc3366 }
      .ansi-color-204    { color: #ff3366 }
      .ansi-color-30     { color: #006666 }
      .ansi-color-66     { color: #336666 }
      .ansi-color-102    { color: #666666 }
      .ansi-color-138    { color: #996666 }
      .ansi-color-174    { color: #cc6666 }
      .ansi-color-210    { color: #ff6666 }
      .ansi-color-36     { color: #009966 }
      .ansi-color-72     { color: #339966 }
      .ansi-color-108    { color: #669966 }
      .ansi-color-144    { color: #999966 }
      .ansi-color-180    { color: #cc9966 }
      .ansi-color-216    { color: #ff9966 }
      .ansi-color-42     { color: #00cc66 }
      .ansi-color-78     { color: #33cc66 }
      .ansi-color-114    { color: #66cc66 }
      .ansi-color-150    { color: #99cc66 }
      .ansi-color-186    { color: #cccc66 }
      .ansi-color-222    { color: #ffcc66 }
      .ansi-color-48     { color: #00ff66 }
      .ansi-color-84     { color: #33ff66 }
      .ansi-color-120    { color: #66ff66 }
      .ansi-color-156    { color: #99ff66 }
      .ansi-color-192    { color: #ccff66 }
      .ansi-color-228    { color: #ffff66 }
      .ansi-color-19     { color: #000099 }
      .ansi-color-55     { color: #330099 }
      .ansi-color-91     { color: #660099 }
      .ansi-color-127    { color: #990099 }
      .ansi-color-163    { color: #cc0099 }
      .ansi-color-199    { color: #ff0099 }
      .ansi-color-25     { color: #003399 }
      .ansi-color-61     { color: #333399 }
      .ansi-color-97     { color: #663399 }
      .ansi-color-133    { color: #993399 }
      .ansi-color-169    { color: #cc3399 }
      .ansi-color-205    { color: #ff3399 }
      .ansi-color-31     { color: #006699 }
      .ansi-color-67     { color: #336699 }
      .ansi-color-103    { color: #666699 }
      .ansi-color-139    { color: #996699 }
      .ansi-color-175    { color: #cc6699 }
      .ansi-color-211    { color: #ff6699 }
      .ansi-color-37     { color: #009999 }
      .ansi-color-73     { color: #339999 }
      .ansi-color-109    { color: #669999 }
      .ansi-color-145    { color: #999999 }
      .ansi-color-181    { color: #cc9999 }
      .ansi-color-217    { color: #ff9999 }
      .ansi-color-43     { color: #00cc99 }
      .ansi-color-79     { color: #33cc99 }
      .ansi-color-115    { color: #66cc99 }
      .ansi-color-151    { color: #99cc99 }
      .ansi-color-187    { color: #cccc99 }
      .ansi-color-223    { color: #ffcc99 }
      .ansi-color-49     { color: #00ff99 }
      .ansi-color-85     { color: #33ff99 }
      .ansi-color-121    { color: #66ff99 }
      .ansi-color-157    { color: #99ff99 }
      .ansi-color-193    { color: #ccff99 }
      .ansi-color-229    { color: #ffff99 }
      .ansi-color-20     { color: #0000cc }
      .ansi-color-56     { color: #3300cc }
      .ansi-color-92     { color: #6600cc }
      .ansi-color-128    { color: #9900cc }
      .ansi-color-164    { color: #cc00cc }
      .ansi-color-200    { color: #ff00cc }
      .ansi-color-26     { color: #0033cc }
      .ansi-color-62     { color: #3333cc }
      .ansi-color-98     { color: #6633cc }
      .ansi-color-134    { color: #9933cc }
      .ansi-color-170    { color: #cc33cc }
      .ansi-color-206    { color: #ff33cc }
      .ansi-color-32     { color: #0066cc }
      .ansi-color-68     { color: #3366cc }
      .ansi-color-104    { color: #6666cc }
      .ansi-color-140    { color: #9966cc }
      .ansi-color-176    { color: #cc66cc }
      .ansi-color-212    { color: #ff66cc }
      .ansi-color-38     { color: #0099cc }
      .ansi-color-74     { color: #3399cc }
      .ansi-color-110    { color: #6699cc }
      .ansi-color-146    { color: #9999cc }
      .ansi-color-182    { color: #cc99cc }
      .ansi-color-218    { color: #ff99cc }
      .ansi-color-44     { color: #00cccc }
      .ansi-color-80     { color: #33cccc }
      .ansi-color-116    { color: #66cccc }
      .ansi-color-152    { color: #99cccc }
      .ansi-color-188    { color: #cccccc }
      .ansi-color-224    { color: #ffcccc }
      .ansi-color-50     { color: #00ffcc }
      .ansi-color-86     { color: #33ffcc }
      .ansi-color-122    { color: #66ffcc }
      .ansi-color-158    { color: #99ffcc }
      .ansi-color-194    { color: #ccffcc }
      .ansi-color-230    { color: #ffffcc }
      .ansi-color-21     { color: #0000ff }
      .ansi-color-57     { color: #3300ff }
      .ansi-color-93     { color: #6600ff }
      .ansi-color-129    { color: #9900ff }
      .ansi-color-165    { color: #cc00ff }
      .ansi-color-201    { color: #ff00ff }
      .ansi-color-27     { color: #0033ff }
      .ansi-color-63     { color: #3333ff }
      .ansi-color-99     { color: #6633ff }
      .ansi-color-135    { color: #9933ff }
      .ansi-color-171    { color: #cc33ff }
      .ansi-color-207    { color: #ff33ff }
      .ansi-color-33     { color: #0066ff }
      .ansi-color-69     { color: #3366ff }
      .ansi-color-105    { color: #6666ff }
      .ansi-color-141    { color: #9966ff }
      .ansi-color-177    { color: #cc66ff }
      .ansi-color-213    { color: #ff66ff }
      .ansi-color-39     { color: #0099ff }
      .ansi-color-75     { color: #3399ff }
      .ansi-color-111    { color: #6699ff }
      .ansi-color-147    { color: #9999ff }
      .ansi-color-183    { color: #cc99ff }
      .ansi-color-219    { color: #ff99ff }
      .ansi-color-45     { color: #00ccff }
      .ansi-color-81     { color: #33ccff }
      .ansi-color-117    { color: #66ccff }
      .ansi-color-153    { color: #99ccff }
      .ansi-color-189    { color: #ccccff }
      .ansi-color-225    { color: #ffccff }
      .ansi-color-51     { color: #00ffff }
      .ansi-color-87     { color: #33ffff }
      .ansi-color-123    { color: #66ffff }
      .ansi-color-159    { color: #99ffff }
      .ansi-color-195    { color: #ccffff }
      .ansi-color-231    { color: #ffffff }
      .ansi-color-232    { color: #0a0a0a }
      .ansi-color-233    { color: #141414 }
      .ansi-color-234    { color: #1f1f1f }
      .ansi-color-235    { color: #292929 }
      .ansi-color-236    { color: #333333 }
      .ansi-color-237    { color: #3d3d3d }
      .ansi-color-238    { color: #474747 }
      .ansi-color-239    { color: #525252 }
      .ansi-color-240    { color: #5c5c5c }
      .ansi-color-241    { color: #666666 }
      .ansi-color-242    { color: #707070 }
      .ansi-color-243    { color: #7a7a7a }
      .ansi-color-244    { color: #858585 }
      .ansi-color-245    { color: #8f8f8f }
      .ansi-color-246    { color: #999999 }
      .ansi-color-247    { color: #a3a3a3 }
      .ansi-color-248    { color: #adadad }
      .ansi-color-249    { color: #b8b8b8 }
      .ansi-color-250    { color: #c2c2c2 }
      .ansi-color-251    { color: #cccccc }
      .ansi-color-252    { color: #d6d6d6 }
      .ansi-color-253    { color: #e0e0e0 }
      .ansi-color-254    { color: #ebebeb }
      .ansi-color-255    { color: #f5f5f5 }
      .ansi-bg-color-16  { background-color: #000000 }
      .ansi-bg-color-52  { background-color: #330000 }
      .ansi-bg-color-88  { background-color: #660000 }
      .ansi-bg-color-124 { background-color: #990000 }
      .ansi-bg-color-160 { background-color: #cc0000 }
      .ansi-bg-color-196 { background-color: #ff0000 }
      .ansi-bg-color-22  { background-color: #003300 }
      .ansi-bg-color-58  { background-color: #333300 }
      .ansi-bg-color-94  { background-color: #663300 }
      .ansi-bg-color-130 { background-color: #993300 }
      .ansi-bg-color-166 { background-color: #cc3300 }
      .ansi-bg-color-202 { background-color: #ff3300 }
      .ansi-bg-color-28  { background-color: #006600 }
      .ansi-bg-color-64  { background-color: #336600 }
      .ansi-bg-color-100 { background-color: #666600 }
      .ansi-bg-color-136 { background-color: #996600 }
      .ansi-bg-color-172 { background-color: #cc6600 }
      .ansi-bg-color-208 { background-color: #ff6600 }
      .ansi-bg-color-34  { background-color: #009900 }
      .ansi-bg-color-70  { background-color: #339900 }
      .ansi-bg-color-106 { background-color: #669900 }
      .ansi-bg-color-142 { background-color: #999900 }
      .ansi-bg-color-178 { background-color: #cc9900 }
      .ansi-bg-color-214 { background-color: #ff9900 }
      .ansi-bg-color-40  { background-color: #00cc00 }
      .ansi-bg-color-76  { background-color: #33cc00 }
      .ansi-bg-color-112 { background-color: #66cc00 }
      .ansi-bg-color-148 { background-color: #99cc00 }
      .ansi-bg-color-184 { background-color: #cccc00 }
      .ansi-bg-color-220 { background-color: #ffcc00 }
      .ansi-bg-color-46  { background-color: #00ff00 }
      .ansi-bg-color-82  { background-color: #33ff00 }
      .ansi-bg-color-118 { background-color: #66ff00 }
      .ansi-bg-color-154 { background-color: #99ff00 }
      .ansi-bg-color-190 { background-color: #ccff00 }
      .ansi-bg-color-226 { background-color: #ffff00 }
      .ansi-bg-color-17  { background-color: #000033 }
      .ansi-bg-color-53  { background-color: #330033 }
      .ansi-bg-color-89  { background-color: #660033 }
      .ansi-bg-color-125 { background-color: #990033 }
      .ansi-bg-color-161 { background-color: #cc0033 }
      .ansi-bg-color-197 { background-color: #ff0033 }
      .ansi-bg-color-23  { background-color: #003333 }
      .ansi-bg-color-59  { background-color: #333333 }
      .ansi-bg-color-95  { background-color: #663333 }
      .ansi-bg-color-131 { background-color: #993333 }
      .ansi-bg-color-167 { background-color: #cc3333 }
      .ansi-bg-color-203 { background-color: #ff3333 }
      .ansi-bg-color-29  { background-color: #006633 }
      .ansi-bg-color-65  { background-color: #336633 }
      .ansi-bg-color-101 { background-color: #666633 }
      .ansi-bg-color-137 { background-color: #996633 }
      .ansi-bg-color-173 { background-color: #cc6633 }
      .ansi-bg-color-209 { background-color: #ff6633 }
      .ansi-bg-color-35  { background-color: #009933 }
      .ansi-bg-color-71  { background-color: #339933 }
      .ansi-bg-color-107 { background-color: #669933 }
      .ansi-bg-color-143 { background-color: #999933 }
      .ansi-bg-color-179 { background-color: #cc9933 }
      .ansi-bg-color-215 { background-color: #ff9933 }
      .ansi-bg-color-41  { background-color: #00cc33 }
      .ansi-bg-color-77  { background-color: #33cc33 }
      .ansi-bg-color-113 { background-color: #66cc33 }
      .ansi-bg-color-149 { background-color: #99cc33 }
      .ansi-bg-color-185 { background-color: #cccc33 }
      .ansi-bg-color-221 { background-color: #ffcc33 }
      .ansi-bg-color-47  { background-color: #00ff33 }
      .ansi-bg-color-83  { background-color: #33ff33 }
      .ansi-bg-color-119 { background-color: #66ff33 }
      .ansi-bg-color-155 { background-color: #99ff33 }
      .ansi-bg-color-191 { background-color: #ccff33 }
      .ansi-bg-color-227 { background-color: #ffff33 }
      .ansi-bg-color-18  { background-color: #000066 }
      .ansi-bg-color-54  { background-color: #330066 }
      .ansi-bg-color-90  { background-color: #660066 }
      .ansi-bg-color-126 { background-color: #990066 }
      .ansi-bg-color-162 { background-color: #cc0066 }
      .ansi-bg-color-198 { background-color: #ff0066 }
      .ansi-bg-color-24  { background-color: #003366 }
      .ansi-bg-color-60  { background-color: #333366 }
      .ansi-bg-color-96  { background-color: #663366 }
      .ansi-bg-color-132 { background-color: #993366 }
      .ansi-bg-color-168 { background-color: #cc3366 }
      .ansi-bg-color-204 { background-color: #ff3366 }
      .ansi-bg-color-30  { background-color: #006666 }
      .ansi-bg-color-66  { background-color: #336666 }
      .ansi-bg-color-102 { background-color: #666666 }
      .ansi-bg-color-138 { background-color: #996666 }
      .ansi-bg-color-174 { background-color: #cc6666 }
      .ansi-bg-color-210 { background-color: #ff6666 }
      .ansi-bg-color-36  { background-color: #009966 }
      .ansi-bg-color-72  { background-color: #339966 }
      .ansi-bg-color-108 { background-color: #669966 }
      .ansi-bg-color-144 { background-color: #999966 }
      .ansi-bg-color-180 { background-color: #cc9966 }
      .ansi-bg-color-216 { background-color: #ff9966 }
      .ansi-bg-color-42  { background-color: #00cc66 }
      .ansi-bg-color-78  { background-color: #33cc66 }
      .ansi-bg-color-114 { background-color: #66cc66 }
      .ansi-bg-color-150 { background-color: #99cc66 }
      .ansi-bg-color-186 { background-color: #cccc66 }
      .ansi-bg-color-222 { background-color: #ffcc66 }
      .ansi-bg-color-48  { background-color: #00ff66 }
      .ansi-bg-color-84  { background-color: #33ff66 }
      .ansi-bg-color-120 { background-color: #66ff66 }
      .ansi-bg-color-156 { background-color: #99ff66 }
      .ansi-bg-color-192 { background-color: #ccff66 }
      .ansi-bg-color-228 { background-color: #ffff66 }
      .ansi-bg-color-19  { background-color: #000099 }
      .ansi-bg-color-55  { background-color: #330099 }
      .ansi-bg-color-91  { background-color: #660099 }
      .ansi-bg-color-127 { background-color: #990099 }
      .ansi-bg-color-163 { background-color: #cc0099 }
      .ansi-bg-color-199 { background-color: #ff0099 }
      .ansi-bg-color-25  { background-color: #003399 }
      .ansi-bg-color-61  { background-color: #333399 }
      .ansi-bg-color-97  { background-color: #663399 }
      .ansi-bg-color-133 { background-color: #993399 }
      .ansi-bg-color-169 { background-color: #cc3399 }
      .ansi-bg-color-205 { background-color: #ff3399 }
      .ansi-bg-color-31  { background-color: #006699 }
      .ansi-bg-color-67  { background-color: #336699 }
      .ansi-bg-color-103 { background-color: #666699 }
      .ansi-bg-color-139 { background-color: #996699 }
      .ansi-bg-color-175 { background-color: #cc6699 }
      .ansi-bg-color-211 { background-color: #ff6699 }
      .ansi-bg-color-37  { background-color: #009999 }
      .ansi-bg-color-73  { background-color: #339999 }
      .ansi-bg-color-109 { background-color: #669999 }
      .ansi-bg-color-145 { background-color: #999999 }
      .ansi-bg-color-181 { background-color: #cc9999 }
      .ansi-bg-color-217 { background-color: #ff9999 }
      .ansi-bg-color-43  { background-color: #00cc99 }
      .ansi-bg-color-79  { background-color: #33cc99 }
      .ansi-bg-color-115 { background-color: #66cc99 }
      .ansi-bg-color-151 { background-color: #99cc99 }
      .ansi-bg-color-187 { background-color: #cccc99 }
      .ansi-bg-color-223 { background-color: #ffcc99 }
      .ansi-bg-color-49  { background-color: #00ff99 }
      .ansi-bg-color-85  { background-color: #33ff99 }
      .ansi-bg-color-121 { background-color: #66ff99 }
      .ansi-bg-color-157 { background-color: #99ff99 }
      .ansi-bg-color-193 { background-color: #ccff99 }
      .ansi-bg-color-229 { background-color: #ffff99 }
      .ansi-bg-color-20  { background-color: #0000cc }
      .ansi-bg-color-56  { background-color: #3300cc }
      .ansi-bg-color-92  { background-color: #6600cc }
      .ansi-bg-color-128 { background-color: #9900cc }
      .ansi-bg-color-164 { background-color: #cc00cc }
      .ansi-bg-color-200 { background-color: #ff00cc }
      .ansi-bg-color-26  { background-color: #0033cc }
      .ansi-bg-color-62  { background-color: #3333cc }
      .ansi-bg-color-98  { background-color: #6633cc }
      .ansi-bg-color-134 { background-color: #9933cc }
      .ansi-bg-color-170 { background-color: #cc33cc }
      .ansi-bg-color-206 { background-color: #ff33cc }
      .ansi-bg-color-32  { background-color: #0066cc }
      .ansi-bg-color-68  { background-color: #3366cc }
      .ansi-bg-color-104 { background-color: #6666cc }
      .ansi-bg-color-140 { background-color: #9966cc }
      .ansi-bg-color-176 { background-color: #cc66cc }
      .ansi-bg-color-212 { background-color: #ff66cc }
      .ansi-bg-color-38  { background-color: #0099cc }
      .ansi-bg-color-74  { background-color: #3399cc }
      .ansi-bg-color-110 { background-color: #6699cc }
      .ansi-bg-color-146 { background-color: #9999cc }
      .ansi-bg-color-182 { background-color: #cc99cc }
      .ansi-bg-color-218 { background-color: #ff99cc }
      .ansi-bg-color-44  { background-color: #00cccc }
      .ansi-bg-color-80  { background-color: #33cccc }
      .ansi-bg-color-116 { background-color: #66cccc }
      .ansi-bg-color-152 { background-color: #99cccc }
      .ansi-bg-color-188 { background-color: #cccccc }
      .ansi-bg-color-224 { background-color: #ffcccc }
      .ansi-bg-color-50  { background-color: #00ffcc }
      .ansi-bg-color-86  { background-color: #33ffcc }
      .ansi-bg-color-122 { background-color: #66ffcc }
      .ansi-bg-color-158 { background-color: #99ffcc }
      .ansi-bg-color-194 { background-color: #ccffcc }
      .ansi-bg-color-230 { background-color: #ffffcc }
      .ansi-bg-color-21  { background-color: #0000ff }
      .ansi-bg-color-57  { background-color: #3300ff }
      .ansi-bg-color-93  { background-color: #6600ff }
      .ansi-bg-color-129 { background-color: #9900ff }
      .ansi-bg-color-165 { background-color: #cc00ff }
      .ansi-bg-color-201 { background-color: #ff00ff }
      .ansi-bg-color-27  { background-color: #0033ff }
      .ansi-bg-color-63  { background-color: #3333ff }
      .ansi-bg-color-99  { background-color: #6633ff }
      .ansi-bg-color-135 { background-color: #9933ff }
      .ansi-bg-color-171 { background-color: #cc33ff }
      .ansi-bg-color-207 { background-color: #ff33ff }
      .ansi-bg-color-33  { background-color: #0066ff }
      .ansi-bg-color-69  { background-color: #3366ff }
      .ansi-bg-color-105 { background-color: #6666ff }
      .ansi-bg-color-141 { background-color: #9966ff }
      .ansi-bg-color-177 { background-color: #cc66ff }
      .ansi-bg-color-213 { background-color: #ff66ff }
      .ansi-bg-color-39  { background-color: #0099ff }
      .ansi-bg-color-75  { background-color: #3399ff }
      .ansi-bg-color-111 { background-color: #6699ff }
      .ansi-bg-color-147 { background-color: #9999ff }
      .ansi-bg-color-183 { background-color: #cc99ff }
      .ansi-bg-color-219 { background-color: #ff99ff }
      .ansi-bg-color-45  { background-color: #00ccff }
      .ansi-bg-color-81  { background-color: #33ccff }
      .ansi-bg-color-117 { background-color: #66ccff }
      .ansi-bg-color-153 { background-color: #99ccff }
      .ansi-bg-color-189 { background-color: #ccccff }
      .ansi-bg-color-225 { background-color: #ffccff }
      .ansi-bg-color-51  { background-color: #00ffff }
      .ansi-bg-color-87  { background-color: #33ffff }
      .ansi-bg-color-123 { background-color: #66ffff }
      .ansi-bg-color-159 { background-color: #99ffff }
      .ansi-bg-color-195 { background-color: #ccffff }
      .ansi-bg-color-231 { background-color: #ffffff }
      .ansi-bg-color-232 { background-color: #0a0a0a }
      .ansi-bg-color-233 { background-color: #141414 }
      .ansi-bg-color-234 { background-color: #1f1f1f }
      .ansi-bg-color-235 { background-color: #292929 }
      .ansi-bg-color-236 { background-color: #333333 }
      .ansi-bg-color-237 { background-color: #3d3d3d }
      .ansi-bg-color-238 { background-color: #474747 }
      .ansi-bg-color-239 { background-color: #525252 }
      .ansi-bg-color-240 { background-color: #5c5c5c }
      .ansi-bg-color-241 { background-color: #666666 }
      .ansi-bg-color-242 { background-color: #707070 }
      .ansi-bg-color-243 { background-color: #7a7a7a }
      .ansi-bg-color-244 { background-color: #858585 }
      .ansi-bg-color-245 { background-color: #8f8f8f }
      .ansi-bg-color-246 { background-color: #999999 }
      .ansi-bg-color-247 { background-color: #a3a3a3 }
      .ansi-bg-color-248 { background-color: #adadad }
      .ansi-bg-color-249 { background-color: #b8b8b8 }
      .ansi-bg-color-250 { background-color: #c2c2c2 }
      .ansi-bg-color-251 { background-color: #cccccc }
      .ansi-bg-color-252 { background-color: #d6d6d6 }
      .ansi-bg-color-253 { background-color: #e0e0e0 }
      .ansi-bg-color-254 { background-color: #ebebeb }
      .ansi-bg-color-255 { background-color: #f5f5f5 }

Release Notes
-------------
<!--
If this PR introduces any user-facing changes, please document them below. Please delete any unused section titles and placeholders.
Please match the style of previous release notes: https://docs.confluent.io/confluent-cli/current/release-notes.html
-->

Breaking Changes
- PLACEHOLDER

New Features
- PLACEHOLDER

Bug Fixes
- PLACEHOLDER

Checklist
---------
- [ ] Leave this box unchecked if features are not yet available in production

What
----
<!--
Briefly describe **what** you have changed and **why**.
Optionally include your implementation strategy.
-->

References
----------
<!--
Copy and paste links to tickets, related PRs, GitHub issues, etc.
-->

Test & Review
-------------
<!--
Has it been tested? How?
Copy and paste any instructions or steps that can save the reviewer time.
-->

# unknown hyperlink type

    Code
      make_link("this", "foobar")
    Condition
      Error in `match.arg()`:
      ! 'arg' should be one of "email", "file", "fun", "help", "href", "run", "topic", "url", "vignette"

# iterm file links

    Code
      cli::cli_text("{.file /path/to/file:10}")
    Message
      ']8;;file:///path/to/file#10/path/to/file:10]8;;'
    Code
      cli::cli_text("{.file /path/to/file:10:20}")
    Message
      ']8;;file:///path/to/file#10:20/path/to/file:10:20]8;;'

# rstudio links

    Code
      cli::cli_text("{.fun pkg::fun}")
    Message
      `]8;;ide:help:pkg::funpkg::fun]8;;()`

---

    Code
      cli::cli_text("{.help fun}")
    Message
      ]8;;ide:help:funfun]8;;

---

    Code
      cli::cli_text("{.run package::func()}")
    Message
      ]8;;ide:run:package::func()package::func()]8;;

---

    Code
      cli::cli_text("{.vignette package::title}")
    Message
      ]8;;ide:vignette:package::titlepackage::title]8;;

---

    Code
      cli::cli_text("{.topic pkg::topic}")
    Message
      ]8;;ide:help:pkg::topicpkg::topic]8;;

# ST hyperlinks

    Code
      cat(style_hyperlink("text", "https://example.com"))
    Output
      ]8;;https://example.com\text]8;;\

# iterator

    Code
      out
    Output
      [1] "\r===============>---------------   50% | ETA:  1s\033[K\r"
      [2] "\r\033[K"                                                  

---

    Code
      out
    Output
      [1] "\r\\ 50 done (100/s) | 3ms\033[K\r" "\r\033[K"                          

# tasks

    Code
      out
    Output
      [1] "\r\\ 50/100 ETA:  1s | \033[K\r" "\r\033[K"                       

---

    Code
      out
    Output
      [1] "\r\\ 50 done (100/s) | 3ms\033[K\r" "\r\033[K"                          

# download

    Code
      out
    Output
      [1] "\r==>---------------------------- |  53 kB/1.0 MB ETA:  1s\033[K\r"
      [2] "\r\033[K"                                                          

---

    Code
      out
    Output
      [1] "\r\\  53 kB (8.5 MB/s) | 3ms\033[K\r"
      [2] "\r\033[K"                            

# customize with options, iterator

    Code
      capture_cli_messages(fun("iterator", 100))
    Output
      [1] "\rnew format 50\033[K\r" "\r\033[K"               

---

    Code
      capture_cli_messages(fun("iterator", NA))
    Output
      [1] "\rnew format 50\033[K\r" "\r\033[K"               

---

    Code
      capture_cli_messages(fun("iterator", 100))
    Output
      [1] "\rnew format 50\033[K\r" "\r\033[K"               

---

    Code
      capture_cli_messages(fun("iterator", NA))
    Output
      [1] "\rnew too 50\033[K\r" "\r\033[K"            

# customize with options, tasks

    Code
      capture_cli_messages(fun("tasks", 100))
    Output
      [1] "\rnew format 50\033[K\r" "\r\033[K"               

---

    Code
      capture_cli_messages(fun("tasks", NA))
    Output
      [1] "\rnew format 50\033[K\r" "\r\033[K"               

---

    Code
      capture_cli_messages(fun("tasks", 100))
    Output
      [1] "\rnew format 50\033[K\r" "\r\033[K"               

---

    Code
      capture_cli_messages(fun("tasks", NA))
    Output
      [1] "\rnew too 50\033[K\r" "\r\033[K"            

# customize with options, download

    Code
      capture_cli_messages(fun("download", 1024 * 1024, 512 * 1024))
    Output
      [1] "\rnew format 524 kB\033[K\r" "\r\033[K"                   

---

    Code
      capture_cli_messages(fun("download", NA, 512 * 1024))
    Output
      [1] "\rnew format 524 kB\033[K\r" "\r\033[K"                   

---

    Code
      capture_cli_messages(fun("download", 1024 * 1024, 512 * 1024))
    Output
      [1] "\rnew format 524 kB\033[K\r" "\r\033[K"                   

---

    Code
      capture_cli_messages(fun("download", NA, 512 * 1024))
    Output
      [1] "\rnew too 524 kB\033[K\r" "\r\033[K"                

## Contributing

Thanks for your interest in contributing to the Confluent CLI!

### Development Environment

Start by following these steps to set up your computer for CLI development:

#### Go Version

We recommend you use [goenv](https://github.com/syndbg/goenv) to manage your Go versions.
There's a `.go-version` file in this repo with the exact version we use (and test against in CI).

We recommend cloning the `goenv` repo directly to ensure that you have access to the latest version of Go. If you've
already installed `goenv` with brew, uninstall it first:

    brew uninstall goenv

Now, clone the `goenv` repo:

    git clone https://github.com/syndbg/goenv.git ~/.goenv

Then, add the following to your shell profile:

    export GOENV_ROOT="$HOME/.goenv"
    export PATH="$GOENV_ROOT/bin:$PATH"
    eval "$(goenv init -)"
    export PATH="$PATH:$GOPATH/bin"

Finally, you can install the appropriate version of Go by running the following command inside the root directory of the repository:

    goenv install

#### Developing on MacOS

Our integration tests read a lot of files while they are running. On MacOS, the default maximum number of open files is
256, which is too small (you will see an error like `error retrieving command exit code` or `too many open files`).
Please run the following three commands *and then restart* for these changes to take effect:

    echo 'kern.maxfiles=20480' | sudo tee -a /etc/sysctl.conf
    echo -e 'limit maxfiles 8192 20480\nlimit maxproc 1000 2000' | sudo tee -a /etc/launchd.conf
    echo 'ulimit -n 4096' | sudo tee -a /etc/profile

#### Security

We use `pre-commit` hooks and `gitleaks` to prevent secrets from being committed to this repo. Please install `pre-commit` hooks (Note that the second command should be run inside the root directory of the repository):

    brew install pre-commit
    pre-commit install

### File Layout

This repo mostly follows the [Standard Go Project Layout](https://github.com/golang-standards/project-layout).
Here's the basic file structure:

    cli/
    ├─ cmd/
    │  ├─ confluent/
    │  │  ├─ main.go                            (entry point for the CLI binary)
    ├─ dist/
    │  ├─ confluent_<os>_<arch>/
    │  │  ├─ confluent                          (the CLI binary)
    ├─ internal/
    │  ├─ cmd/                                  (CLI commands)
    │  │  ├─ <command>/
    │  │  │  ├─ command.go                      (a top-level CLI command)
    │  │  │  ├─ command_<subcommand>.go         (a subcommand of a top-level CLI command)
    │  │  │  ├─ command_<subcommand>_onprem.go  (the on-prem version of the above command, if applicable)
    │  │  ├─ command.go                         (the root CLI command)
    │  ├─ pkg/
    ├─ test/                                    (integration tests)
    │  ├─ fixtures/
    │  │  ├─ output/
    │  │  │  ├─ <command>/                      (the golden files for a top-level CLI command)
    │  │  ├─ cli_test.go                        (entry point for all integration tests)
    │  │  ├─ <command>_test.go                  (the integration tests for a top-level CLI command)

### Testing

The CLI is tested with a combination of unit tests and integration tests. To run all tests:

    make test

#### Unit Tests

Unit tests exist in files ending with `_test.go`, and are located alongside the main source code files.
Unit tests should test small, isolated functions, and should not be unnecessarily complex (i.e. mocking backend calls or CLI commands).

To run the all unit tests:

    make unit-test

To run a subset of unit tests, you must specify the suite and optionally the name of a specific tests:

    # Run a suite of unit tests
    make unit-test UNIT_TEST_ARGS="-run TestApiTestSuite"

    # Run a specific unit test within a suite
    make unit-test UNIT_TEST_ARGS="-run TestApiTestSuite/TestCreateCloudAPIKey"

#### Integration Tests

The [test/](./test) directory contains our integration tests. These tests build the CLI binary and invoke commands on it.
These CLI integration tests roughly follow this [pattern](http://lucapette.me/writing-integration-tests-for-a-go-cli-application):

1. Run a test HTTP server to mock Confluent Cloud or the Confluent Platform Control Plane API.
2. Run a logical sequence of CLI commands.
3. Ensure that the output of these commands matches the corresponding golden files.

To update the golden files from the current output:

    make integration-test INTEGRATION_TEST_ARGS="-update"

To skip rebuilding the CLI, if it already exists in `dist/`:

    make integration-test INTEGRATION_TEST_ARGS="-no-rebuild"

To run a subset of integration tests, you must specify the suite and optionally the name of a specific test:

    # Run a suite of integration tests
    make integration-test INTEGRATION_TEST_ARGS="-run TestCLI/TestKafka"

    # Run a specific integration test within a suite
    make integration-test INTEGRATION_TEST_ARGS="-run TestCLI/TestKafka/kafka_cluster_--help"

### Example: Adding a New Command to the CLI

As a basic demonstration, we'll be implementing a command which prints the name of the CLI config file a specified number of times:

    $ confluent config describe 3
    ~/.confluent/config.json
    ~/.confluent/config.json
    ~/.confluent/config.json

#### Creating the Command

Like all other commands, this command will reside in `internal`. First, we must create a directory for this command:

    mkdir internal/config

Next, we create two files, one for the top-level command `config`, and another for `describe`.

`internal/config/command.go`:

```go
package config

import (
    "github.com/spf13/cobra"

    pcmd "github.com/confluentinc/cli/v3/pkg/cmd"
)

type command struct {
    *pcmd.CLICommand
}

func New(prerunner pcmd.PreRunner) *cobra.Command {
    cmd := &cobra.Command{Use: "config"}

    c := &command{pcmd.NewAnonymousCLICommand(cmd, prerunner)}

    cmd.AddCommand(c.newDescribeCommand())

    return cmd
}
```

`internal/config/command_describe.go`:

```go
package config

import (
    "strconv"

    "github.com/spf13/cobra"

    "github.com/confluentinc/cli/v3/pkg/errors"
    "github.com/confluentinc/cli/v3/pkg/utils"
)

func (c *command) newDescribeCommand() *cobra.Command {
    return &cobra.Command{
        Use:  "describe",
        Args: cobra.ExactArgs(1),
        RunE: c.describe,
    }
}

func (c *command) describe(_ *cobra.Command, args []string) error {
    filename := c.CLICommand.Config.Config.Filename
    if filename == "" {
        return fmt.Errorf("config file not found")
    }

    n, err := strconv.Atoi(args[0])
    if err != nil {
        return err
    }

    for i := 0; i < n; i++ {
        output.Println(filename)
    }
    return nil
}

```

#### Registering the Command

Finally, we must add the newly created `config` command as a child of the root command.
Add the following line to `internal/command.go`, and make sure to import its package:

    cmd.AddCommand(config.New(prerunner))

#### Running the Command

To build the CLI binary, run `make build`. After this, we can run our command in the following way, and see that it (hopefully) works!

    make build
    dist/confluent_<os>_<arch>/confluent config file describe 3

#### Integration Testing

There's not much code here to unit test, so we'll skip right to integration testing. Create the following file:

`test/config/config_test.go`:

```go
package test

func (s *CLITestSuite) TestConfigDescribe() {
	tests := []CLITest{
		{args: "config describe 3", fixture: "config/1.golden"},
	}

	for _, test := range tests {
		s.runConfluentTest(test)
	}
}
```

We'll also need to add the new golden file, `test/fixtures/output/config/1.golden`.
After running the command manually to ensure the output is correct, the content for the golden file can either be:

1. Copied directly from the terminal.
2. Updated automatically with `make integration-test INTEGRATION_TEST_ARGS="-run TestCLI/TestConfigDescribe -update"` (slow).

Now, run `make integration-test INTEGRATION_TEST_ARGS="-run TestCLI/TestConfigDescribe"` and verify that it works!

#### Add Autocompletion

Add support for autocompletion using `ValidArgsFunction` if applicable (for example, if the command takes resource IDs or resource names as arguments):

```go
func (c *command) newDescribeCommand() *cobra.Command {
    return &cobra.Command{
        Use:               "describe",
        Args:              cobra.ExactArgs(1),
        ValidArgsFunction: pcmd.NewValidArgsFunction(c.validArgs),
        RunE:              c.describe,
    }
}
```

See the [Autocompletion](pkg/cmd/AUTOCOMPLETION.md) resource for implementation details.

#### Adding a New Delete Command

For most resource types, a `delete` command should support multiple arguments. The exceptions are resources which do not have an ID (e.g. ACLs, role bindings) or unique resources (e.g. the Schema Registry cluster).
See [Supporting Multiple Deletion](pkg/deletion/README.md) for instructions on how to write such commands.

### Opening a PR

That's it! As you can see, the process of adding a new CLI command is pretty straightforward. You can open a PR if:
* You're able to build the CLI with `make build`.
* All unit and integration test pass with `make test`.
* Running `make lint` produces no linter errors.

Note: If there is a JIRA ticket associated with your PR, please format the PR description as "[CLI-1234] Description of PR".

### Detailed Implementation Guides

Please familiarize yourself with the following resources before writing your first CLI command:

* [Autocompletion](pkg/cmd/AUTOCOMPLETION.md)
* [Cloud and On-Prem Annotations](pkg/cmd/ANNOTATIONS.md)
* [Errors](pkg/errors/README.md)
* [Output](pkg/output/README.md)

# cat_line

    Code
      cat_line("This is ", "a ", "line of text.")
    Output
      This is a line of text.

---

    Code
      cat_line("This is ", "a ", "line of text.", col = "red")
    Output
      [31mThis is a line of text.[39m
    Code
      cat_line("This is ", "a ", "line of text.", background_col = "green")
    Output
      [42mThis is a line of text.[49m

# cat_bullet [plain]

    Code
      cat_bullet(letters[1:5])
    Output
      * a
      * b
      * c
      * d
      * e

# cat_bullet [unicode]

    Code
      cat_bullet(letters[1:5])
    Output
      • a
      • b
      • c
      • d
      • e

# cat_boxx [plain]

    Code
      cat_boxx("foo")
    Output
      +---------+
      |         |
      |   foo   |
      |         |
      +---------+

# cat_boxx [unicode]

    Code
      cat_boxx("foo")
    Output
      ┌─────────┐
      │         │
      │   foo   │
      │         │
      └─────────┘

# cat_rule [plain]

    Code
      local({
        withr::local_options(cli.width = 20)
        cat_rule("title")
      })
    Output
      -- title -----------

# cat_rule [unicode]

    Code
      local({
        withr::local_options(cli.width = 20)
        cat_rule("title")
      })
    Output
      ── title ───────────

# cat_print [plain]

    Code
      cat_print(boxx(""))
    Output
      +------+
      |      |
      |      |
      |      |
      +------+

---

    Code
      local({
        tmp <- tempfile()
        on.exit(unlink(tmp), add = TRUE)
        expect_silent(cat_print(boxx(""), file = tmp))
        cat(readLines(tmp, warn = FALSE), sep = "\n")
      })
    Output
      +------+
      |      |
      |      |
      |      |
      +------+

# cat_print [unicode]

    Code
      cat_print(boxx(""))
    Output
      ┌──────┐
      │      │
      │      │
      │      │
      └──────┘

---

    Code
      local({
        tmp <- tempfile()
        on.exit(unlink(tmp), add = TRUE)
        expect_silent(cat_print(boxx(""), file = tmp))
        cat(readLines(tmp, warn = FALSE), sep = "\n")
      })
    Output
      ┌──────┐
      │      │
      │      │
      │      │
      └──────┘

# headers [plain]

    Code
      local({
        cli_div(class = "testcli", theme = test_style())
        cli_h1("HEADER")
        cli_h2("Header")
        cli_h3("Header")
        x <- "foobar"
        xx <- 100
        cli_h2("{xx}. header: {x}")
      })
    Message
      
      HEADER
      
      Header
      
      Header
      
      100. header: foobar
      

# headers [ansi]

    Code
      local({
        cli_div(class = "testcli", theme = test_style())
        cli_h1("HEADER")
        cli_h2("Header")
        cli_h3("Header")
        x <- "foobar"
        xx <- 100
        cli_h2("{xx}. header: {x}")
      })
    Message
      
      [1m[3m[1m[3mHEADER[3m[1m[23m[22m
      
      [1m[1mHeader[1m[22m
      
      [4m[4mHeader[4m[24m
      
      [1m[1m[1m100[1m. header: [1mfoobar[1m[1m[22m
      

# headers [unicode]

    Code
      local({
        cli_div(class = "testcli", theme = test_style())
        cli_h1("HEADER")
        cli_h2("Header")
        cli_h3("Header")
        x <- "foobar"
        xx <- 100
        cli_h2("{xx}. header: {x}")
      })
    Message
      
      HEADER
      
      Header
      
      Header
      
      100. header: foobar
      

# headers [fancy]

    Code
      local({
        cli_div(class = "testcli", theme = test_style())
        cli_h1("HEADER")
        cli_h2("Header")
        cli_h3("Header")
        x <- "foobar"
        xx <- 100
        cli_h2("{xx}. header: {x}")
      })
    Message
      
      [1m[3m[1m[3mHEADER[3m[1m[23m[22m
      
      [1m[1mHeader[1m[22m
      
      [4m[4mHeader[4m[24m
      
      [1m[1m[1m100[1m. header: [1mfoobar[1m[1m[22m
      

# issue #218

    Code
      cli_h1("one {1} two {2} three {3}")
    Message
      
      -- one 1 two 2 three 3 ---------------------------------------------------------
    Code
      cli_h2("one {1} two {2} three {3}")
    Message
      
      -- one 1 two 2 three 3 --
      

# Autocompletion

We leverage the [Cobra Completions API](https://github.com/spf13/cobra/blob/master/shell_completions.md) to provide tab
autocompletion for all of our commands, arguments, and flags. Autocompletion is used extensively throughout the CLI, and
occasionally requires authentication, so we provide several helper functions to limit the amount of boilerplate.

Here's how you would use the `NewValidArgsFunction()` function to dynamically complete a command which accepts an API
key as an argument:

```go
pcmd.NewValidArgsFunction(c.validArgs)
```

```go
func (c *command) validArgs(cmd *cobra.Command, args []string) []string {
    if len(args) > 0 {
        return nil
    }

    if err := c.PersistentPreRunE(cmd, args); err != nil {
        return nil
    }

    return pcmd.AutocompleteApiKeys(c.V2Client)
}
```

Here's how you would use the `RegisterFlagCompletionFunc()` function to dynamically complete a flag which accepts an API
key as a value:

```go
pcmd.RegisterFlagCompletionFunc(cmd, "api-key", func(cmd *cobra.Command, args []string) []string {
    if err := c.PersistentPreRunE(cmd, args); err != nil {
        return nil
    }

    return pcmd.AutocompleteApiKeys(c.V2Client)
})
```

# Contributing to `step cli`

We welcome contributions to `step cli` of any kind including
documentation, themes, organization, tutorials, blog posts, bug reports,
issues, feature requests, feature implementations, pull requests, helping
to manage issues, etc.

## Table of Contents

* [Asking Support Questions](#asking-support-questions)
* [Reporting Issues](#reporting-issues)
* [Submitting Patches](#submitting-patches)
  * [Code Contribution Guidelines](#code-contribution-guidelines)
  * [Git Commit Message Guidelines](#git-commit-message-guidelines)

## Reporting Issues

If you believe you have found a defect in `step cli` or its
documentation, use the GitHub [issue
tracker](https://github.com/smallstep/cli/issues) to report the
problem. When reporting the issue, please provide the version of `step
cli` in use (`step version`) and your operating system.

## Asking Support Questions

Users and developers can ask questions on [GitHub Discussions](https://github.com/smallstep/certificates/discussions).

## Code Contribution

`step cli` aims to become a fully featured toolkit for cryptographic primitives.
We encourage all contributions that meet the following criteria:

* fit naturally into a toolkit for creating and working with cryptographic
primitives and higher order resources.
* strive not to break existing functionality.
* close or update an open [`step cli` issue](https://github.com/smallstep/cli/issues)

**Bug fixes are, of course, always welcome.**

## Submitting Patches

`step cli` welcomes all contributors and contributions. If you are
interested in helping with the project, please reach out to us or, better yet,
submit a PR :).

### Code Contribution Guidelines

Because we want to create the best possible product for our users and the best
contribution experience for our developers, we have a set of guidelines which
ensure that all contributions are acceptable. The guidelines are not intended
as a filter or barrier to participation. If you are unfamiliar with the
contribution process, the Smallstep team will guide you in order to get your
contribution in accordance with the guidelines.

To make the contribution process as seamless as possible, we ask for the following:

* Go ahead and fork the project and make your changes. We encourage pull
requests to allow for review and discussion of code changes.
* When you’re ready to create a pull request, be sure to:
    * Sign the [CLA](https://cla-assistant.io/smallstep/cli).
    * Have test cases for the new code. If you have questions about how to do
    this, please ask in your pull request.
    * Run `go fmt`.
    * Add documentation if you are adding new features or changing
    functionality.
    * Squash your commits into a single commit. `git rebase -i`. It’s okay to
    force update your pull request with `git push -f`.
    * Follow the **Git Commit Message Guidelines** below.

### Git Commit Message Guidelines

This [blog article](http://chris.beams.io/posts/git-commit/) is a good resource
for learning how to write good commit messages, the most important part being
that each commit message should have a title/subject in imperative mood
starting with a capital letter and no trailing period: *"Return error on wrong
use of the Paginator"*, **NOT** *"returning some error."*

Also, if your commit references one or more GitHub issues, always end your
commit message body with *See #1234* or *Fixes #1234*. Replace *1234* with the
GitHub issue ID. The last example will close the issue when the commit is
merged into *master*.

Please use a short and descriptive branch name, e.g. **NOT** "patch-1". It's
very common but creates a naming conflict each time when a submission is pulled
for a review.

An example:

```text
Add step certificate install

Add a command line utility for installing (and uninstalling) certificates to the
local system truststores. This should help developers with local development
flows.

Fixes #75
```

# diff_chr

    Code
      d$lcs
    Output
        operation offset length old_offset new_offset
      1     match      0      8          8          8
      2    delete      8      1          9          8
      3    insert      8      1          9          9
      4     match      9      7         16         16
      5    insert     16      2         16         18
      6     match     16      1         17         19

---

    Code
      d$lcs
    Output
        operation offset length old_offset new_offset
      1     match      0      8          8          8
      2    delete      8      1          9          8
      3    insert      8      1          9          9
      4     match      9      7         16         16
      5    delete     16      2         18         16
      6     match     18      1         19         17

# diff_chr [plain]

    Code
      d
    Output
      @@ -6,7 +6,7 @@
       1
       1
       2
      -3
      +10
       4
       4
       4
      @@ -14,4 +14,6 @@
       4
       4
       4
      +6
      +7
       5

---

    Code
      d$lcs
    Output
        operation offset length old_offset new_offset
      1     match      0      8          8          8
      2    delete      8      1          9          8
      3    insert      8      1          9          9
      4     match      9      7         16         16
      5    insert     16      2         16         18
      6     match     16      1         17         19

# diff_chr [ansi]

    Code
      d
    Output
      @@ -6,7 +6,7 @@
       1
       1
       2
      [34m-3[39m
      [32m+10[39m
       4
       4
       4
      @@ -14,4 +14,6 @@
       4
       4
       4
      [32m+6[39m
      [32m+7[39m
       5

---

    Code
      d$lcs
    Output
        operation offset length old_offset new_offset
      1     match      0      8          8          8
      2    delete      8      1          9          8
      3    insert      8      1          9          9
      4     match      9      7         16         16
      5    insert     16      2         16         18
      6     match     16      1         17         19

# diff_chr edge cases

    Code
      diff_chr(character(), character())

---

    Code
      diff_chr(character(), character())$lcs
    Output
      [1] operation  offset     length     old_offset new_offset
      <0 rows> (or 0-length row.names)

---

    Code
      diff_chr("a", character())
    Output
      @@ -1 +0 @@
      -a

---

    Code
      diff_chr(character(), "b")
    Output
      @@ -0 +1 @@
      +b

---

    Code
      diff_chr("a", "a")

---

    Code
      diff_chr(letters, letters)

---

    Code
      diff_chr(c("a", NA, "a2"), "b")
    Output
      @@ -1,3 +1 @@
      -a
      -NA
      -a2
      +b

---

    Code
      diff_chr(NA_character_, "NA")
    Output
      @@ -1 +1 @@
      -NA
      +NA

# format.cli_diff_chr context

    Code
      print(d, context = 1)
    Output
      @@ -8,3 +8,3 @@
       2
      -3
      +10
       4
      @@ -16,2 +16,4 @@
       4
      +6
      +7
       5

---

    Code
      print(d, context = 0)
    Output
      @@ -9 +9 @@
      -3
      +10
      @@ -17,0 +17,2 @@
      +6
      +7

---

    Code
      print(d, context = Inf)
    Output
       1
       1
       1
       1
       1
       1
       1
       2
      -3
      +10
       4
       4
       4
       4
       4
       4
       4
      +6
      +7
       5

---

    Code
      print(d2, context = Inf)
    Output
       foo
       bar

# diff_str [plain]

    Code
      d
    Output
      {+PRE+}abcdefg[-hijklm-]{+MIDDLE+}nopqrstuvwxyz{+POST+}

# diff_str [ansi]

    Code
      d
    Output
      [42m[30mPRE[39m[49mabcdefg[44m[30mhijklm[39m[49m[42m[30mMIDDLE[39m[49mnopqrstuvwxyz[42m[30mPOST[39m[49m

# max_diff

    ! Diff edit distance is larger than the limit.
    i The edit distance limit is 0.

---

    ! Diff edit distance is larger than the limit.
    i The edit distance limit is 1.

---
name: Documentation Request
about: Request documentation for a feature
title: '[docs]:'
labels: docs, needs triage
assignees: ''

---

## Hello!
<!-- Please leave this section as-is, it's designed to help others in the community know how to interact with our GitHub issues. -->

- Vote on this issue by adding a 👍 reaction
- If you want to document this feature, comment to let us know (we'll work with you on design, scheduling, etc.)

## Affected area/feature

<!---
Tell us which feature you'd like to see documented. 
 - Where would you like that documentation to live (command line usage output, website, github markdown on the repo)? 
- If there are specific attributes or options you'd like to see documented, please include those in the request.
-->

# add/remove/list themes [plain]

    Code
      cli_par(class = "green")
      cli_text(lorem_ipsum())
    Message
      Non incididunt mollit ullamco duis officia proident. Laborum esse mollit mollit
      eiusmod tempor cupidatat. In commodo anim irure nostrud. Deserunt nisi amet
      laborum magna aliqua. Do esse consectetur ut deserunt nulla Lorem non. Fugiat
      est dolore deserunt aliqua amet et esse dolore elit exercitation sint
      exercitation non ipsum.
    Code
      cli_end()
    Message
      

# add/remove/list themes [ansi]

    Code
      cli_par(class = "green")
      cli_text(lorem_ipsum())
    Message
      [32mNon incididunt mollit ullamco duis officia proident. Laborum esse mollit mollit[39m
      [32meiusmod tempor cupidatat. In commodo anim irure nostrud. Deserunt nisi amet[39m
      [32mlaborum magna aliqua. Do esse consectetur ut deserunt nulla Lorem non. Fugiat[39m
      [32mest dolore deserunt aliqua amet et esse dolore elit exercitation sint[39m
      [32mexercitation non ipsum.[39m
    Code
      cli_end()
    Message
      

# add/remove/list themes [unicode]

    Code
      cli_par(class = "green")
      cli_text(lorem_ipsum())
    Message
      Non incididunt mollit ullamco duis officia proident. Laborum esse mollit mollit
      eiusmod tempor cupidatat. In commodo anim irure nostrud. Deserunt nisi amet
      laborum magna aliqua. Do esse consectetur ut deserunt nulla Lorem non. Fugiat
      est dolore deserunt aliqua amet et esse dolore elit exercitation sint
      exercitation non ipsum.
    Code
      cli_end()
    Message
      

# add/remove/list themes [fancy]

    Code
      cli_par(class = "green")
      cli_text(lorem_ipsum())
    Message
      [32mNon incididunt mollit ullamco duis officia proident. Laborum esse mollit mollit[39m
      [32meiusmod tempor cupidatat. In commodo anim irure nostrud. Deserunt nisi amet[39m
      [32mlaborum magna aliqua. Do esse consectetur ut deserunt nulla Lorem non. Fugiat[39m
      [32mest dolore deserunt aliqua amet et esse dolore elit exercitation sint[39m
      [32mexercitation non ipsum.[39m
    Code
      cli_end()
    Message
      

# explicit formatter is used, and combined

    Code
      cli_text("this is {.emph it}, really")
    Message
      this is (((<<it>>))), really

# user's override

    Code
      local({
        start_app(theme = custom, .auto_close = FALSE)
        cli_alert("Alert!")
        stop_app()
        withr::local_options(cli.user_theme = override)
        start_app(theme = custom, .auto_close = FALSE)
        cli_alert("Alert!")
        stop_app()
      })
    Message
      custom:Alert!
      custom:Alert!override:

# NULL will undo a style property

    Code
      local({
        cli_alert("this has an arrow")
        cli_div(theme = list(.alert = list(before = NULL)))
        cli_alert("this does not")
      })
    Message
      > this has an arrow
      this does not

# NULL will undo color [ansi]

    Code
      local({
        cli_alert("{.emph {.val this is blue}}")
        cli_div(theme = list(span.val = list(color = NULL)))
        cli_alert("{.emph {.val this is not}}")
      })
    Message
      > [34m[3m[34m"this is blue"[34m[23m[39m
      > [0m[22m[23m[24m[27m[28m[29m[49m[3m[0m[22m[3m[24m[27m[28m[29m[49m"this is not"[0m[22m[23m[24m[27m[28m[29m[49m[23m[39m

---

    Code
      local({
        cli_alert("{.emph {.val this is blue}}")
        cli_div(theme = list(span.val = list(color = "none")))
        cli_alert("{.emph {.val this is not}}")
      })
    Message
      > [34m[3m[34m"this is blue"[34m[23m[39m
      > [0m[22m[23m[24m[27m[28m[29m[49m[3m[0m[22m[3m[24m[27m[28m[29m[49m"this is not"[0m[22m[23m[24m[27m[28m[29m[49m[23m[39m

# NULL will undo background color [ansi]

    Code
      local({
        cli_alert("{.emph {.code this has bg color}}")
        cli_div(theme = list(span = list(`background-color` = NULL)))
        cli_alert("{.emph {.code this does not}}")
      })
    Message
      > [3m`this has bg color`[23m
      > [0m[22m[23m[24m[27m[28m[29m[39m[3m[0m[22m[3m[24m[27m[28m[29m[39m`this does not`[0m[22m[23m[24m[27m[28m[29m[39m[23m[49m

---

    Code
      local({
        cli_alert("{.emph {.code this has bg color}}")
        cli_div(theme = list(span = list(`background-color` = "none")))
        cli_alert("{.emph {.code this does not}}")
      })
    Message
      > [3m`this has bg color`[23m
      > [0m[22m[23m[24m[27m[28m[29m[39m[3m[0m[22m[3m[24m[27m[28m[29m[39m`this does not`[0m[22m[23m[24m[27m[28m[29m[39m[23m[49m

---
name: Enhancement
about: Suggest an enhancement to step cli
title: ''
labels: enhancement, needs triage
assignees: ''

---

## Hello!
<!-- Please leave this section as-is, 
it's designed to help others in the community know how to interact with our GitHub issues. -->

- Vote on this issue by adding a 👍 reaction
- If you want to implement this feature, comment to let us know (we'll work with you on design, scheduling, etc.)

## Issue details

<!-- Enhancement requests are most helpful when they describe the problem you're having 
as well as articulating the potential solution you'd like to see built. -->

## Why is this needed?

<!-- Let us know why you think this enhancement would be good for the project or community. -->

# empty input

    Code
      vt_output("", width = 20, height = 2)$segment
    Output
      [1] "                    " "                    "

# raw input

    Code
      vt_output(charToRaw("foobar"), width = 20, height = 2)$segment
    Output
      [1] "foobar              " "                    "

# overflow

    Code
      vt_output(strrep("1234567890", 2), width = 19, height = 2)$segment
    Output
      [1] "1234567890123456789" "0                  "

# control characters

    Code
      vt_output("foo\nbar", width = 20, height = 2)$segment
    Output
      [1] "foo                 " "bar                 "

---

    Code
      vt_output("foobar\rbaz", width = 20, height = 2)$segment
    Output
      [1] "bazbar              " "                    "

# scroll up

    Code
      vt_output(strrep("1234567890", 5), width = 20, height = 2)$segment
    Output
      [1] "12345678901234567890" "1234567890          "

---

    Code
      vt_output(paste0(1:10, "\n"), width = 10, height = 5)$segment
    Output
      [1] "7         " "8         " "9         " "10        " "          "

# ANSI SGR [ansi]

    Code
      vt_output("12\033[31m34\033[1m56\033[39m78\033[21m90", width = 20, height = 2)
    Output
        lineno segmentno              segment  bold italic underline strikethrough
      1      1         1                   12 FALSE  FALSE     FALSE         FALSE
      2      1         1                   34 FALSE  FALSE     FALSE         FALSE
      3      1         1                   56  TRUE  FALSE     FALSE         FALSE
      4      1         1                   78  TRUE  FALSE     FALSE         FALSE
      5      1         1         90           FALSE  FALSE     FALSE         FALSE
      6      2         1                      FALSE  FALSE     FALSE         FALSE
        blink inverse color background_color link link_params
      1 FALSE   FALSE  <NA>             <NA> <NA>        <NA>
      2 FALSE   FALSE     1             <NA> <NA>        <NA>
      3 FALSE   FALSE     1             <NA> <NA>        <NA>
      4 FALSE   FALSE  <NA>             <NA> <NA>        <NA>
      5 FALSE   FALSE  <NA>             <NA> <NA>        <NA>
      6 FALSE   FALSE  <NA>             <NA> <NA>        <NA>

---

    Code
      vt_output(style_bold("I'm bold"), width = 20, height = 2)
    Output
        lineno segmentno              segment  bold italic underline strikethrough
      1      1         1             I'm bold  TRUE  FALSE     FALSE         FALSE
      2      1         1                      FALSE  FALSE     FALSE         FALSE
      3      2         1                      FALSE  FALSE     FALSE         FALSE
        blink inverse color background_color link link_params
      1 FALSE   FALSE  <NA>             <NA> <NA>        <NA>
      2 FALSE   FALSE  <NA>             <NA> <NA>        <NA>
      3 FALSE   FALSE  <NA>             <NA> <NA>        <NA>

---

    Code
      vt_output(style_italic("I'm italic"), width = 20, height = 2)
    Output
        lineno segmentno              segment  bold italic underline strikethrough
      1      1         1           I'm italic FALSE   TRUE     FALSE         FALSE
      2      1         1                      FALSE  FALSE     FALSE         FALSE
      3      2         1                      FALSE  FALSE     FALSE         FALSE
        blink inverse color background_color link link_params
      1 FALSE   FALSE  <NA>             <NA> <NA>        <NA>
      2 FALSE   FALSE  <NA>             <NA> <NA>        <NA>
      3 FALSE   FALSE  <NA>             <NA> <NA>        <NA>

---

    Code
      vt_output(style_underline("I'm underlined"), width = 20, height = 2)
    Output
        lineno segmentno              segment  bold italic underline strikethrough
      1      1         1       I'm underlined FALSE  FALSE      TRUE         FALSE
      2      1         1                      FALSE  FALSE     FALSE         FALSE
      3      2         1                      FALSE  FALSE     FALSE         FALSE
        blink inverse color background_color link link_params
      1 FALSE   FALSE  <NA>             <NA> <NA>        <NA>
      2 FALSE   FALSE  <NA>             <NA> <NA>        <NA>
      3 FALSE   FALSE  <NA>             <NA> <NA>        <NA>

---

    Code
      vt_output(style_strikethrough("I'm strikethrough"), width = 20, height = 2)
    Output
        lineno segmentno              segment  bold italic underline strikethrough
      1      1         1    I'm strikethrough FALSE  FALSE     FALSE          TRUE
      2      1         1                      FALSE  FALSE     FALSE         FALSE
      3      2         1                      FALSE  FALSE     FALSE         FALSE
        blink inverse color background_color link link_params
      1 FALSE   FALSE  <NA>             <NA> <NA>        <NA>
      2 FALSE   FALSE  <NA>             <NA> <NA>        <NA>
      3 FALSE   FALSE  <NA>             <NA> <NA>        <NA>

---

    Code
      vt_output(style_inverse("I'm inverse"), width = 20, height = 2)
    Output
        lineno segmentno              segment  bold italic underline strikethrough
      1      1         1          I'm inverse FALSE  FALSE     FALSE         FALSE
      2      1         1                      FALSE  FALSE     FALSE         FALSE
      3      2         1                      FALSE  FALSE     FALSE         FALSE
        blink inverse color background_color link link_params
      1 FALSE    TRUE  <NA>             <NA> <NA>        <NA>
      2 FALSE   FALSE  <NA>             <NA> <NA>        <NA>
      3 FALSE   FALSE  <NA>             <NA> <NA>        <NA>

# hyperlinks

    Code
      link <- style_hyperlink("text", "url")
      vt_output(c("pre ", st_from_bel(link), " post"), width = 20, height = 2)
    Output
        lineno segmentno              segment  bold italic underline strikethrough
      1      1         1                 pre  FALSE  FALSE     FALSE         FALSE
      2      1         1                 text FALSE  FALSE     FALSE         FALSE
      3      1         1          post        FALSE  FALSE     FALSE         FALSE
      4      2         1                      FALSE  FALSE     FALSE         FALSE
        blink inverse color background_color link link_params
      1 FALSE   FALSE  <NA>             <NA> <NA>        <NA>
      2 FALSE   FALSE  <NA>             <NA>  url            
      3 FALSE   FALSE  <NA>             <NA> <NA>        <NA>
      4 FALSE   FALSE  <NA>             <NA> <NA>        <NA>

---

    Code
      link <- style_hyperlink("text", "url", params = c(f = "x", g = "y"))
      vt_output(c("pre ", st_from_bel(link), " post"), width = 20, height = 2)
    Output
        lineno segmentno              segment  bold italic underline strikethrough
      1      1         1                 pre  FALSE  FALSE     FALSE         FALSE
      2      1         1                 text FALSE  FALSE     FALSE         FALSE
      3      1         1          post        FALSE  FALSE     FALSE         FALSE
      4      2         1                      FALSE  FALSE     FALSE         FALSE
        blink inverse color background_color link link_params
      1 FALSE   FALSE  <NA>             <NA> <NA>        <NA>
      2 FALSE   FALSE  <NA>             <NA>  url f = x:g = y
      3 FALSE   FALSE  <NA>             <NA> <NA>        <NA>
      4 FALSE   FALSE  <NA>             <NA> <NA>        <NA>

# erase in line

    Code
      vt_output("foobar\033[3D\033[K", width = 10, height = 2)$segment
    Output
      [1] "foo       " "          "
    Code
      vt_output("foobar\033[3D\033[0K", width = 10, height = 2)$segment
    Output
      [1] "foo       " "          "
    Code
      vt_output("foobar\033[3D\033[1K", width = 10, height = 2)$segment
    Output
      [1] "    ar    " "          "
    Code
      vt_output("foobar\033[3D\033[2K", width = 10, height = 2)$segment
    Output
      [1] "          " "          "

# erase in screen

    Code
      vt_output("foo\nfoobar\nfoobar2\033[A\033[4D\033[J", width = 10, height = 4)$
        segment
    Output
      [1] "foo       " "foo       " "          " "          "
    Code
      vt_output("foo\nfoobar\nfoobar2\033[A\033[4D\033[0J", width = 10, height = 4)$
        segment
    Output
      [1] "foo       " "foo       " "          " "          "
    Code
      vt_output("foo\nfoobar\nfoobar2\033[A\033[4D\033[1J", width = 10, height = 4)$
        segment
    Output
      [1] "          " "    ar    " "foobar2   " "          "
    Code
      vt_output("foo\nfoobar\nfoobar2\033[A\033[4D\033[2Jx", width = 10, height = 4)$
        segment
    Output
      [1] "          " "   x      " "          " "          "
    Code
      vt_output("foo\nfoobar\nfoobar2\033[A\033[4D\033[3Jx", width = 10, height = 4)$
        segment
    Output
      [1] "          " "   x      " "          " "          "

# control characaters

    Code
      for (code in c(1:2, 4:6, 8:14, 16L, 20L, 21L, 23L, 27L, 127L)) {
        p$write_input("cli::keypress()\n")
        Sys.sleep(0.1)
        p$write_input(as.raw(code))
        p$poll_io(1000)
        cat(p$read_output())
      }
    Output
      [1] "ctrl-a"
      [1] "ctrl-b"
      [1] "ctrl-d"
      [1] "ctrl-e"
      [1] "ctrl-f"
      [1] "ctrl-h"
      [1] "tab"
      [1] "enter"
      [1] "ctrl-k"
      [1] "ctrl-l"
      [1] "enter"
      [1] "ctrl-n"
      [1] "ctrl-p"
      [1] "ctrl-t"
      [1] "ctrl-u"
      [1] "ctrl-w"
      [1] "escape"
      [1] "backspace"

# write ahead

    Code
      p$write_input("{ Sys.sleep(0.5); cli::keypress() }\nX")
      p$poll_io(1000)
    Output
        output    error  process 
       "ready" "nopipe" "silent" 
    Code
      cat(p$read_output())
    Output
      [1] "X"

# arrows, etc

    Code
      for (key in keys) {
        p$write_input("cli::keypress()\n")
        p$write_input(key)
        p$poll_io(1000)
        cat(p$read_output())
      }
    Output
      [1] "up"
      [1] "right"
      [1] "left"
      [1] "end"
      [1] "home"
      [1] "-"
      [1] "up"
      [1] "down"
      [1] "right"
      [1] "left"
      [1] "end"
      [1] "home"
      [1] "-"
      [1] "home"
      [1] "insert"
      [1] "delete"
      [1] "end"
      [1] "pageup"
      [1] "pagedown"
      [1] "-"
      [1] "pageup"
      [1] "pagedown"
      [1] "-"
      [1] "home"
      [1] "end"
      [1] "-"
      [1] "f1"
      [1] "f2"
      [1] "f3"
      [1] "f4"
      [1] "-"
      [1] "f5"
      [1] "f6"
      [1] "f7"
      [1] "f8"
      [1] "f9"
      [1] "f10"
      [1] "f11"
      [1] "f12"
      [1] "-"
      [1] "f1"
      [1] "f2"
      [1] "f3"
      [1] "f4"
      [1] "-"
      [1] "escape"

# nonblocking

    Code
      p$write_input("cli::keypress(block = FALSE)\n")
      p$poll_io(1000)
    Output
        output    error  process 
       "ready" "nopipe" "silent" 
    Code
      cat(p$read_output())
    Output
      [1] NA

---

    Code
      p$write_input("{ Sys.sleep(0.5); cli::keypress() }\nX")
      p$poll_io(1000)
    Output
        output    error  process 
       "ready" "nopipe" "silent" 
    Code
      cat(p$read_output())
    Output
      [1] "X"

# spark_bar [plain]

    Code
      spark_bar(seq(0, 1, length.out = 8))
    Output
      __,,**##
    Code
      spark_bar(c(0, NA, 0.5, NA, 1))
    Output
      _ , #

# spark_bar [unicode]

    Code
      spark_bar(seq(0, 1, length.out = 8))
    Output
      ▁▂▃▄▅▆▇█
    Code
      spark_bar(c(0, NA, 0.5, NA, 1))
    Output
      ▁ ▄ █

# spark_line [plain]

    Code
      spark_line(seq(0, 1, length.out = 10))
    Output
      _,,-^

# spark_line [unicode]

    Code
      spark_line(seq(0, 1, length.out = 10))
    Output
      ⣀⡠⠔⠊⠉

# meta basics [plain]

    Code
      cli::cli({
        message("This is before")
        cli_alert_info("First message")
        message("This as well")
        cli_alert_success("Success!")
      })
    Message
      This is before
      This as well
      i First message
      v Success!

# meta basics [ansi]

    Code
      cli::cli({
        message("This is before")
        cli_alert_info("First message")
        message("This as well")
        cli_alert_success("Success!")
      })
    Message
      This is before
      This as well
      [36mi[39m First message
      [32mv[39m Success!

# meta basics [unicode]

    Code
      cli::cli({
        message("This is before")
        cli_alert_info("First message")
        message("This as well")
        cli_alert_success("Success!")
      })
    Message
      This is before
      This as well
      ℹ First message
      ✔ Success!

# meta basics [fancy]

    Code
      cli::cli({
        message("This is before")
        cli_alert_info("First message")
        message("This as well")
        cli_alert_success("Success!")
      })
    Message
      This is before
      This as well
      [36mℹ[39m First message
      [32m✔[39m Success!

# meta is single cli_message [plain]

    Code
      cli_server_default(msgs[[1]])
    Message
      i First message
      v Success!
    Output
      NULL

# meta is single cli_message [ansi]

    Code
      cli_server_default(msgs[[1]])
    Message
      [36mi[39m First message
      [32mv[39m Success!
    Output
      NULL

# meta is single cli_message [unicode]

    Code
      cli_server_default(msgs[[1]])
    Message
      ℹ First message
      ✔ Success!
    Output
      NULL

# meta is single cli_message [fancy]

    Code
      cli_server_default(msgs[[1]])
    Message
      [36mℹ[39m First message
      [32m✔[39m Success!
    Output
      NULL

# meta is single cliMessage [plain]

    Code
      withCallingHandlers(cli::cli({
        cli_alert_info("First message")
        cli_alert_success("Success!")
      }), cliMessage = function(msg) {
        msgs <<- c(msgs, list(msg))
      })
    Message
      i First message
      v Success!

# meta is single cliMessage [ansi]

    Code
      withCallingHandlers(cli::cli({
        cli_alert_info("First message")
        cli_alert_success("Success!")
      }), cliMessage = function(msg) {
        msgs <<- c(msgs, list(msg))
      })
    Message
      [36mi[39m First message
      [32mv[39m Success!

# meta is single cliMessage [unicode]

    Code
      withCallingHandlers(cli::cli({
        cli_alert_info("First message")
        cli_alert_success("Success!")
      }), cliMessage = function(msg) {
        msgs <<- c(msgs, list(msg))
      })
    Message
      ℹ First message
      ✔ Success!

# meta is single cliMessage [fancy]

    Code
      withCallingHandlers(cli::cli({
        cli_alert_info("First message")
        cli_alert_success("Success!")
      }), cliMessage = function(msg) {
        msgs <<- c(msgs, list(msg))
      })
    Message
      [36mℹ[39m First message
      [32m✔[39m Success!

# substitution [plain]

    Code
      local({
        x <- 1:3
        cli({
          title <- "My title"
          cli_h1("Title: {.emph {title}}")
          cli_text("And {.emph some} more: {.val {x}}")
        })
      })
    Message
      
      -- Title: My title -------------------------------------------------------------
      And some more: 1, 2, and 3

# substitution [ansi]

    Code
      local({
        x <- 1:3
        cli({
          title <- "My title"
          cli_h1("Title: {.emph {title}}")
          cli_text("And {.emph some} more: {.val {x}}")
        })
      })
    Message
      
      [36m--[39m [1mTitle: [3mMy title[23m[22m [36m-------------------------------------------------------------[39m
      And [3msome[23m more: [34m1[39m, [34m2[39m, and [34m3[39m

# substitution [unicode]

    Code
      local({
        x <- 1:3
        cli({
          title <- "My title"
          cli_h1("Title: {.emph {title}}")
          cli_text("And {.emph some} more: {.val {x}}")
        })
      })
    Message
      
      ── Title: My title ─────────────────────────────────────────────────────────────
      And some more: 1, 2, and 3

# substitution [fancy]

    Code
      local({
        x <- 1:3
        cli({
          title <- "My title"
          cli_h1("Title: {.emph {title}}")
          cli_text("And {.emph some} more: {.val {x}}")
        })
      })
    Message
      
      [36m──[39m [1mTitle: [3mMy title[23m[22m [36m─────────────────────────────────────────────────────────────[39m
      And [3msome[23m more: [34m1[39m, [34m2[39m, and [34m3[39m

# return values are ok when recording (#496)

    Code
      cli::cli({
        lid <- cli::cli_ul()
        cli::cli_li("a bullet")
        cli::cli_end(lid)
      })
    Message
      * a bullet

# nested cli() (#497)

    Code
      cli::cli({
        cli::cli_h1("Header")
        cli::cli(cli::cli_text("Some text"))
        cli::cli_text("Some more text")
      })
    Message
      
      -- Header ----------------------------------------------------------------------
      Some text
      Some more text

# text is wrapped

    Code
      local({
        cli_div(class = "testcli", theme = test_style())
        cli_h1("Header")
        cli_text(lorem_ipsum())
      })
    Message
      
      Header
      
      Duis quis magna incididunt nulla commodo minim non
      exercitation nostrud ullamco dolor exercitation ut veniam.
      Fugiat irure tempor commodo voluptate ut. In et tempor
      excepteur quis. Qui et nisi ad quis ad cupidatat tempor
      laborum est excepteur aliqua veniam. Velit sunt magna
      veniam Lorem elit enim et pariatur. Occaecat mollit
      consequat dolore in. Veniam officia labore reprehenderit
      culpa dolore quis nisi do aliqua commodo deserunt.
      Cupidatat nostrud ad est in ad laboris consectetur esse
      minim. Irure do anim anim ea mollit ad cupidatat ullamco
      ullamco nulla elit in. Lorem Lorem deserunt deserunt et ut
      velit nulla nulla ipsum ad laborum quis.

---
title: "CLI pluralization"
author: "Gábor Csárdi"
date: "`r Sys.Date()`"
output:
  rmarkdown::html_vignette:
    keep_md: true
vignette: >
  %\VignetteIndexEntry{CLI pluralization}
  %\VignetteEngine{cli::lazyrmd}
  %\VignetteEncoding{UTF-8}
---

```{r child = "../man/chunks/pluralization.Rmd"}
```

# glue quotes and comments

    Code
      cli_dl(c(test_1 = "all good", test_2 = "not #good"))
    Message
      test_1: all good
      test_2: not #good
    Code
      cli::cli_dl(c(test_3 = "no' good either"))
    Message
      test_3: no' good either
    Code
      cli::cli_dl(c(test_4 = "no\" good also"))
    Message
      test_4: no" good also
    Code
      cli::cli_text("{.url https://example.com/#section}")
    Message
      <https://example.com/#section>
    Code
      cli::cli_alert_success("Qapla'")
    Message
      v Qapla'

# quotes, etc. within expressions are still OK

    Code
      cli::cli_text("{.url URL} {x <- 'foo'; nchar(x)}")
    Message
      <URL> 3
    Code
      cli::cli_text("{.url URL} {x <- \"foo\"; nchar(x)}")
    Message
      <URL> 3
    Code
      cli::cli_text("{.url URL} {1 + 1 # + 1} {1 + 1}")
    Message
      <URL> 2 2

# { } is parsed with literal = FALSE

    Code
      format_message("{.emph {'{foo {}'}}")
    Output
      [1] "{foo {}"

# hash_emoji

    Code
      hash_emoji(character())$names
    Output
      list()
    Code
      hash_emoji("")$names
    Output
      [[1]]
      [1] "teacup without handle" "rhinoceros"            "flushed face"         
      
    Code
      hash_emoji("x")$names
    Output
      [[1]]
      [1] "mage: dark skin tone" "spoon"                "fuel pump"           
      
    Code
      hash_emoji(NA_character_)$names
    Output
      [[1]]
      [1] NA NA NA
      
    Code
      hash_emoji(NA)$names
    Output
      [[1]]
      [1] NA NA NA
      
    Code
      hash_emoji(c(NA, "", "foo", NA))$names
    Output
      [[1]]
      [1] NA NA NA
      
      [[2]]
      [1] "teacup without handle" "rhinoceros"            "flushed face"         
      
      [[3]]
      [1] "sun behind cloud"                    "raised back of hand: dark skin tone"
      [3] "children crossing"                  
      
      [[4]]
      [1] NA NA NA
      

# hash_raw_emoji

    Code
      hash_raw_emoji(raw())$names
    Output
      [1] "teacup without handle" "rhinoceros"            "flushed face"         
    Code
      hash_raw_emoji(as.raw(0))$names
    Output
      [1] "diamond with a dot"    "raised fist"           "face with thermometer"
    Code
      hash_raw_emoji(charToRaw("foobar"))$names
    Output
      [1] "fishing pole"     "money with wings" "eagle"           

# hash_obj_emoji

    Code
      hash_obj_emoji("")$names
    Output
      [1] "woman pilot: medium skin tone" "cat with wry smile"           
      [3] "couple with heart"            
    Code
      hash_obj_emoji(raw(0))$names
    Output
      [1] "fire"                                "man teacher: medium-light skin tone"
      [3] "carp streamer"                      
    Code
      hash_obj_emoji(1:10)$names
    Output
      [1] "backhand index pointing right: medium-light skin tone"
      [2] "woman office worker: medium skin tone"                
      [3] "tiger face"                                           
    Code
      hash_obj_emoji(mtcars)$names
    Output
      [1] "woman judge: medium skin tone" "pause button"                 
      [3] "flag: North Korea"            

# hash_animal

    Code
      hash_animal(character())$words
    Output
      list()
    Code
      hash_animal("")$words
    Output
      [[1]]
      [1] "dogtired"     "conventional" "olingo"      
      
    Code
      hash_animal("x")$words
    Output
      [[1]]
      [1] "goodhearted" "lovelorn"    "amphiuma"   
      
    Code
      hash_animal(NA_character_)$words
    Output
      [[1]]
      [1] NA NA NA
      
    Code
      hash_animal(NA)$words
    Output
      [[1]]
      [1] NA NA NA
      
    Code
      hash_animal(c(NA, "", "foo", NA))$words
    Output
      [[1]]
      [1] NA NA NA
      
      [[2]]
      [1] "dogtired"     "conventional" "olingo"      
      
      [[3]]
      [1] "sacrilegious" "diet"         "lion"        
      
      [[4]]
      [1] NA NA NA
      

# hash_raw_animal

    Code
      hash_raw_animal(raw())$words
    Output
      [1] "dogtired"     "conventional" "olingo"      
    Code
      hash_raw_animal(as.raw(0))$words
    Output
      [1] "scarabaeiform" "surly"         "goldfish"     
    Code
      hash_raw_animal(charToRaw("foobar"))$words
    Output
      [1] "unevolving" "degrading"  "harrier"   

# hash_obj_animal

    Code
      hash_obj_animal("")$words
    Output
      [1] "pathworky" "wondrous"  "minibeast"
    Code
      hash_obj_animal(raw(0))$words
    Output
      [1] "undestructive"   "unequal"         "indianglassfish"
    Code
      hash_obj_animal(1:10)$words
    Output
      [1] "benignant" "profound"  "ambushbug"
    Code
      hash_obj_animal(mtcars)$words
    Output
      [1] "novice"             "flushed"            "australianshelduck"

# hash_xxhash

    Code
      hash_xxhash(letters[1:5])
    Output
      [1] "a96faf705af16834e6c632b61e964e1f" "4b2212e31ac97fd4575a0b1c44d8843f"
      [3] "12d8bdd17f74de858c40219a46b9f81b" "56a841f9102d5ff745f80274c9c7a7ca"
      [5] "2d97a8f9e2edaaefe5e72e5e3bec4a78"
    Code
      hash_xxhash(c("a", NA, "b"))
    Output
      [1] "a96faf705af16834e6c632b61e964e1f" NA                                
      [3] "4b2212e31ac97fd4575a0b1c44d8843f"
    Code
      hash_xxhash64(letters[1:5])
    Output
      [1] "e6c632b61e964e1f" "575a0b1c44d8843f" "8c40219a46b9f81b" "45f80274c9c7a7ca"
      [5] "e5e72e5e3bec4a78"
    Code
      hash_raw_xxhash(charToRaw("a"))
    Output
      [1] "a96faf705af16834e6c632b61e964e1f"
    Code
      hash_raw_xxhash64(charToRaw("a"))
    Output
      [1] "e6c632b61e964e1f"
    Code
      hash_obj_xxhash(raw(0))
    Output
      [1] "64f55f347c3e13113cde6a6f033766e3"
    Code
      hash_obj_xxhash64(raw(0))
    Output
      [1] "0c9b74982308d8fd"

---

    Code
      hash_file_xxhash(tmp)
    Output
      [1] "a96faf705af16834e6c632b61e964e1f"
    Code
      hash_file_xxhash64(tmp)
    Output
      [1] "e6c632b61e964e1f"

# cli_progress_demo

    Code
      out
    Output
      \ 50 done (100/s) | 3ms[K

---

    Code
      out
    Output
      ==========>--------------------   33% | ETA:  1s[K
      ====================>----------   67% | ETA:  1s[K
      ==============================>  100% | ETA:  1s[K
      [K

---

    Code
      out
    Output
      \ 1 done (100/s) | 3ms[K
      | 2 done (100/s) | 3ms[K
      / 3 done (100/s) | 3ms[K
      - 4 done (100/s) | 3ms[K
      \ 5 done (100/s) | 3ms[K
      [K

---

    Code
      out
    Output
      ==========>--------------------   33% | ETA: 10s[K
      ====================>----------   67% | ETA:  3s[K
      ==============================>  100% | ETA:  0s[K
      [K

---

    Code
      msgs
    Output
      [1] "\r==========>--------------------   33% | ETA: 10s\033[K\r"
      [2] "\r====================>----------   67% | ETA:  3s\033[K\r"
      [3] "\r==============================>  100% | ETA:  0s\033[K\r"
      [4] "\r\033[K"                                                  

# pb_bar

    Code
      cli_text("-{cli::pb_bar}-")
    Message
      --

---

    Code
      cli_text("-{cli::pb_bar}-")
    Message
      -===============>--------------- -

# pb_current_bytes

    Code
      cli__pb_current_bytes(list(current = 0))
    Output
      [1] "0.0 kB"
    Code
      cli__pb_current_bytes(list(current = 1))
    Output
      [1] "0.0 kB"
    Code
      cli__pb_current_bytes(list(current = 1000))
    Output
      [1] "1.0 kB"
    Code
      cli__pb_current_bytes(list(current = 1000 * 23))
    Output
      [1] " 23 kB"
    Code
      cli__pb_current_bytes(list(current = 1000 * 1000 * 23))
    Output
      [1] " 23 MB"

# pb_elapsed

    Code
      cli__pb_elapsed()
    Output
      [1] "1s"

---

    Code
      cli__pb_elapsed()
    Output
      [1] "21s"

---

    Code
      cli__pb_elapsed()
    Output
      [1] "58s"

---

    Code
      cli__pb_elapsed()
    Output
      [1] "1h 5m"

# pb_elapsed_clock

    Code
      cli__pb_elapsed_clock()
    Output
      [1] "00:00:01"

---

    Code
      cli__pb_elapsed_clock()
    Output
      [1] "00:00:21"

---

    Code
      cli__pb_elapsed_clock()
    Output
      [1] "00:00:58"

---

    Code
      cli__pb_elapsed_clock()
    Output
      [1] "01:05:00"

# pb_elapsed_raw

    Code
      cli__pb_elapsed_raw()
    Output
      [1] 1

---

    Code
      cli__pb_elapsed_raw()
    Output
      [1] 21

---

    Code
      cli__pb_elapsed_raw()
    Output
      [1] 58

---

    Code
      cli__pb_elapsed_raw()
    Output
      [1] 3900

# pb_eta

    Code
      cli__pb_eta(list())
    Output
      [1] "?"

---

    Code
      cli__pb_eta(list())
    Output
      [1] "12s"

# pb_eta_raw

    Code
      cli__pb_eta_raw()
    Output
      [1] NA

---

    Code
      cli__pb_eta_raw()
    Output
      Time difference of 40 secs

---

    Code
      cli__pb_eta_raw()
    Output
      Time difference of 10 secs

---

    Code
      cli__pb_eta_raw()
    Output
      Time difference of 0 secs

# pb_eta_str

    Code
      cli__pb_eta_str(list())
    Output
      [1] ""

---

    Code
      cli__pb_eta_str(list())
    Output
      [1] "ETA:  1s"

# pb_percent

    Code
      cli__pb_percent(list(current = 0, total = 99))
    Output
      [1] "  0%"
    Code
      cli__pb_percent(list(current = 5, total = 99))
    Output
      [1] "  5%"
    Code
      cli__pb_percent(list(current = 10, total = 99))
    Output
      [1] " 10%"
    Code
      cli__pb_percent(list(current = 25, total = 99))
    Output
      [1] " 25%"
    Code
      cli__pb_percent(list(current = 99, total = 99))
    Output
      [1] "100%"
    Code
      cli__pb_percent(list(current = 100, total = 99))
    Output
      [1] "101%"

# pb_rate

    Code
      cli__pb_rate(list())
    Output
      [1] "?/s"

---

    Code
      cli__pb_rate(list())
    Output
      [1] "?/s"

---

    Code
      cli__pb_rate(list())
    Output
      [1] "0.1/s"

---

    Code
      cli__pb_rate(list())
    Output
      [1] "12/s"

# pb_rate_bytes

    Code
      cli__pb_rate_bytes(list())
    Output
      [1] "NaN kB/s"

---

    Code
      cli__pb_rate_bytes(list())
    Output
      [1] "Inf YB/s"

---

    Code
      cli__pb_rate_bytes(list())
    Output
      [1] "0.0 kB/s"

---

    Code
      cli__pb_rate_bytes(list())
    Output
      [1] "1.0 kB/s"

---

    Code
      cli__pb_rate_bytes(list())
    Output
      [1] " 24 kB/s"

---

    Code
      cli__pb_rate_bytes(list())
    Output
      [1] " 24 MB/s"

# pb_spin

    Code
      cli_text("-{cli::pb_spin}-{cli::pb_spin}-")
    Message
      -\-\-

---

    Code
      cli_text("-{cli::pb_spin}-{cli::pb_spin}-")
    Message
      -|-|-

---

    Code
      cli_text("-{cli::pb_spin}-{cli::pb_spin}-")
    Message
      -|-|-

# pb_timestamp

    Code
      cli__pb_timestamp(list())
    Output
      [1] "2021-06-18T00:09:14+00:00"

---

    Code
      cli__pb_timestamp(list())
    Output
      [1] "2021-06-18T00:09:34+00:00"

# pb_total_bytes

    Code
      cli__pb_total_bytes(list(total = 0))
    Output
      [1] "0.0 kB"
    Code
      cli__pb_total_bytes(list(total = 1))
    Output
      [1] "0.0 kB"
    Code
      cli__pb_total_bytes(list(total = 1000))
    Output
      [1] "1.0 kB"
    Code
      cli__pb_total_bytes(list(total = 1000 * 23))
    Output
      [1] " 23 kB"
    Code
      cli__pb_total_bytes(list(total = 1000 * 1000 * 23))
    Output
      [1] " 23 MB"

# bullets [plain]

    Code
      cli_bullets(c("noindent", ` ` = "space", v = "success", x = "danger", `!` = "warning",
        i = "info", `*` = "bullet", `>` = "arrow"))
    Message
      noindent
        space
      v success
      x danger
      ! warning
      i info
      * bullet
      > arrow

# bullets [ansi]

    Code
      cli_bullets(c("noindent", ` ` = "space", v = "success", x = "danger", `!` = "warning",
        i = "info", `*` = "bullet", `>` = "arrow"))
    Message
      noindent
        space
      [32mv[39m success
      [31mx[39m danger
      [33m![39m warning
      [36mi[39m info
      [36m*[39m bullet
      > arrow

# bullets [unicode]

    Code
      cli_bullets(c("noindent", ` ` = "space", v = "success", x = "danger", `!` = "warning",
        i = "info", `*` = "bullet", `>` = "arrow"))
    Message
      noindent
        space
      ✔ success
      ✖ danger
      ! warning
      ℹ info
      • bullet
      → arrow

# bullets [fancy]

    Code
      cli_bullets(c("noindent", ` ` = "space", v = "success", x = "danger", `!` = "warning",
        i = "info", `*` = "bullet", `>` = "arrow"))
    Message
      noindent
        space
      [32m✔[39m success
      [31m✖[39m danger
      [33m![39m warning
      [36mℹ[39m info
      [36m•[39m bullet
      → arrow

# bullets glue [plain]

    Code
      cli_bullets(c("noindent {.key {1:3}}", ` ` = "space {.key {1:3}}", v = "success {.key {1:3}}",
        x = "danger {.key {1:3}}", `!` = "warning {.key {1:3}}", i = "info {.key {1:3}}",
        `*` = "bullet {.key {1:3}}", `>` = "arrow {.key {1:3}}"))
    Message
      noindent [1], [2], and [3]
        space [1], [2], and [3]
      v success [1], [2], and [3]
      x danger [1], [2], and [3]
      ! warning [1], [2], and [3]
      i info [1], [2], and [3]
      * bullet [1], [2], and [3]
      > arrow [1], [2], and [3]

# bullets glue [ansi]

    Code
      cli_bullets(c("noindent {.key {1:3}}", ` ` = "space {.key {1:3}}", v = "success {.key {1:3}}",
        x = "danger {.key {1:3}}", `!` = "warning {.key {1:3}}", i = "info {.key {1:3}}",
        `*` = "bullet {.key {1:3}}", `>` = "arrow {.key {1:3}}"))
    Message
      noindent [34m[1][39m, [34m[2][39m, and [34m[3][39m
        space [34m[1][39m, [34m[2][39m, and [34m[3][39m
      [32mv[39m success [34m[1][39m, [34m[2][39m, and [34m[3][39m
      [31mx[39m danger [34m[1][39m, [34m[2][39m, and [34m[3][39m
      [33m![39m warning [34m[1][39m, [34m[2][39m, and [34m[3][39m
      [36mi[39m info [34m[1][39m, [34m[2][39m, and [34m[3][39m
      [36m*[39m bullet [34m[1][39m, [34m[2][39m, and [34m[3][39m
      > arrow [34m[1][39m, [34m[2][39m, and [34m[3][39m

# bullets glue [unicode]

    Code
      cli_bullets(c("noindent {.key {1:3}}", ` ` = "space {.key {1:3}}", v = "success {.key {1:3}}",
        x = "danger {.key {1:3}}", `!` = "warning {.key {1:3}}", i = "info {.key {1:3}}",
        `*` = "bullet {.key {1:3}}", `>` = "arrow {.key {1:3}}"))
    Message
      noindent [1], [2], and [3]
        space [1], [2], and [3]
      ✔ success [1], [2], and [3]
      ✖ danger [1], [2], and [3]
      ! warning [1], [2], and [3]
      ℹ info [1], [2], and [3]
      • bullet [1], [2], and [3]
      → arrow [1], [2], and [3]

# bullets glue [fancy]

    Code
      cli_bullets(c("noindent {.key {1:3}}", ` ` = "space {.key {1:3}}", v = "success {.key {1:3}}",
        x = "danger {.key {1:3}}", `!` = "warning {.key {1:3}}", i = "info {.key {1:3}}",
        `*` = "bullet {.key {1:3}}", `>` = "arrow {.key {1:3}}"))
    Message
      noindent [34m[1][39m, [34m[2][39m, and [34m[3][39m
        space [34m[1][39m, [34m[2][39m, and [34m[3][39m
      [32m✔[39m success [34m[1][39m, [34m[2][39m, and [34m[3][39m
      [31m✖[39m danger [34m[1][39m, [34m[2][39m, and [34m[3][39m
      [33m![39m warning [34m[1][39m, [34m[2][39m, and [34m[3][39m
      [36mℹ[39m info [34m[1][39m, [34m[2][39m, and [34m[3][39m
      [36m•[39m bullet [34m[1][39m, [34m[2][39m, and [34m[3][39m
      → arrow [34m[1][39m, [34m[2][39m, and [34m[3][39m

# bullets wrapping [plain]

    Code
      cli_bullets(c(txt, ` ` = txt, v = txt, x = txt, `!` = txt, i = txt, `*` = txt,
        `>` = txt))
    Message
      This is some text that is longer than the width. This is some text that is
      longer than the width. This is some text that is longer than the width.
        This is some text that is longer than the width. This is some text that is
        longer than the width. This is some text that is longer than the width.
      v This is some text that is longer than the width. This is some text that is
        longer than the width. This is some text that is longer than the width.
      x This is some text that is longer than the width. This is some text that is
        longer than the width. This is some text that is longer than the width.
      ! This is some text that is longer than the width. This is some text that is
        longer than the width. This is some text that is longer than the width.
      i This is some text that is longer than the width. This is some text that is
        longer than the width. This is some text that is longer than the width.
      * This is some text that is longer than the width. This is some text that is
        longer than the width. This is some text that is longer than the width.
      > This is some text that is longer than the width. This is some text that is
        longer than the width. This is some text that is longer than the width.

# bullets wrapping [ansi]

    Code
      cli_bullets(c(txt, ` ` = txt, v = txt, x = txt, `!` = txt, i = txt, `*` = txt,
        `>` = txt))
    Message
      This is some text that is longer than the width. This is some text that is
      longer than the width. This is some text that is longer than the width.
        This is some text that is longer than the width. This is some text that is
        longer than the width. This is some text that is longer than the width.
      [32mv[39m This is some text that is longer than the width. This is some text that is
        longer than the width. This is some text that is longer than the width.
      [31mx[39m This is some text that is longer than the width. This is some text that is
        longer than the width. This is some text that is longer than the width.
      [33m![39m This is some text that is longer than the width. This is some text that is
        longer than the width. This is some text that is longer than the width.
      [36mi[39m This is some text that is longer than the width. This is some text that is
        longer than the width. This is some text that is longer than the width.
      [36m*[39m This is some text that is longer than the width. This is some text that is
        longer than the width. This is some text that is longer than the width.
      > This is some text that is longer than the width. This is some text that is
        longer than the width. This is some text that is longer than the width.

# bullets wrapping [unicode]

    Code
      cli_bullets(c(txt, ` ` = txt, v = txt, x = txt, `!` = txt, i = txt, `*` = txt,
        `>` = txt))
    Message
      This is some text that is longer than the width. This is some text that is
      longer than the width. This is some text that is longer than the width.
        This is some text that is longer than the width. This is some text that is
        longer than the width. This is some text that is longer than the width.
      ✔ This is some text that is longer than the width. This is some text that is
        longer than the width. This is some text that is longer than the width.
      ✖ This is some text that is longer than the width. This is some text that is
        longer than the width. This is some text that is longer than the width.
      ! This is some text that is longer than the width. This is some text that is
        longer than the width. This is some text that is longer than the width.
      ℹ This is some text that is longer than the width. This is some text that is
        longer than the width. This is some text that is longer than the width.
      • This is some text that is longer than the width. This is some text that is
        longer than the width. This is some text that is longer than the width.
      → This is some text that is longer than the width. This is some text that is
        longer than the width. This is some text that is longer than the width.

# bullets wrapping [fancy]

    Code
      cli_bullets(c(txt, ` ` = txt, v = txt, x = txt, `!` = txt, i = txt, `*` = txt,
        `>` = txt))
    Message
      This is some text that is longer than the width. This is some text that is
      longer than the width. This is some text that is longer than the width.
        This is some text that is longer than the width. This is some text that is
        longer than the width. This is some text that is longer than the width.
      [32m✔[39m This is some text that is longer than the width. This is some text that is
        longer than the width. This is some text that is longer than the width.
      [31m✖[39m This is some text that is longer than the width. This is some text that is
        longer than the width. This is some text that is longer than the width.
      [33m![39m This is some text that is longer than the width. This is some text that is
        longer than the width. This is some text that is longer than the width.
      [36mℹ[39m This is some text that is longer than the width. This is some text that is
        longer than the width. This is some text that is longer than the width.
      [36m•[39m This is some text that is longer than the width. This is some text that is
        longer than the width. This is some text that is longer than the width.
      → This is some text that is longer than the width. This is some text that is
        longer than the width. This is some text that is longer than the width.

# deep lists ul

    Code
      for (i in 1:4) test_ul(i)
    Message
      * Level 1
      * Level 1
        * Level 2
      * Level 1
        * Level 2
          * Level 3
      * Level 1
        * Level 2
          * Level 3
            * Level 4

# deep lists ol

    Code
      for (i in 1:4) test_ol(i)
    Message
      1. Level 1
      1. Level 1
        1. Level 2
      1. Level 1
        1. Level 2
          1. Level 3
      1. Level 1
        1. Level 2
          1. Level 3
            1. Level 4

# deep lists ol ul

    Code
      for (i in 1:4) test_ol_ul(i)
    Message
      1. Level 1
        * Level 2
      1. Level 1
        * Level 2
          1. Level 3
            * Level 4
      1. Level 1
        * Level 2
          1. Level 3
            * Level 4
              1. Level 5
                * Level 6
      1. Level 1
        * Level 2
          1. Level 3
            * Level 4
              1. Level 5
                * Level 6
                  1. Level 7
                    * Level 8

# deep lists ul ol

    Code
      for (i in 1:4) test_ul_ol(i)
    Message
      * Level 1
        1. Level 2
      * Level 1
        1. Level 2
          * Level 3
            1. Level 4
      * Level 1
        1. Level 2
          * Level 3
            1. Level 4
              * Level 5
                1. Level 6
      * Level 1
        1. Level 2
          * Level 3
            1. Level 4
              * Level 5
                1. Level 6
                  * Level 7
                    1. Level 8

# cli_abort() captures correct call and backtrace

    Code
      print(expect_error(f()))
    Output
      <error/rlang_error>
      Error in `h()`:
      ! foo
      ---
      Backtrace:
           x
        1. +-base::print(expect_error(f()))
        2. +-testthat::expect_error(f())
        3. | \-testthat:::expect_condition_matching(...)
        4. |   \-testthat:::quasi_capture(...)
        5. |     +-testthat (local) .capture(...)
        6. |     | \-base::withCallingHandlers(...)
        7. |     \-rlang::eval_bare(quo_get_expr(.quo), quo_get_env(.quo))
        8. \-cli (local) f()
        9.   \-cli (local) g()
       10.     \-cli (local) h()

---

    Code
      print(expect_error(f(list())))
    Output
      <error/cli_my_class>
      Error in `h()`:
      ! `x` can't be empty.
      ---
      Backtrace:
           x
        1. +-base::print(expect_error(f(list())))
        2. +-testthat::expect_error(f(list()))
        3. | \-testthat:::expect_condition_matching(...)
        4. |   \-testthat:::quasi_capture(...)
        5. |     +-testthat (local) .capture(...)
        6. |     | \-base::withCallingHandlers(...)
        7. |     \-rlang::eval_bare(quo_get_expr(.quo), quo_get_env(.quo))
        8. \-cli (local) f(list())
        9.   \-cli (local) g(x)
       10.     \-cli (local) h(x)

# cli_abort(.internal = TRUE) reports the correct function (r-lib/rlang#1386)

    Code
      (expect_error(fn()))
    Output
      <error/rlang_error>
      Error in `fn()`:
      ! Message.
      i This is an internal error that was detected in the base package.

# collapsing without formatting, n>3

    Code
      pkgs <- paste0("pkg", 1:5)
      cli_text("Packages: {pkgs}.")
    Message
      Packages: pkg1, pkg2, pkg3, pkg4, and pkg5.

# collapsing without formatting, n<3

    Code
      pkgs <- paste0("pkg", 1:2)
      cli_text("Packages: {pkgs}.")
    Message
      Packages: pkg1 and pkg2.

# collapsing with formatting

    Code
      local({
        cli_div(theme = list(.pkg = list(fmt = function(x) paste0(x, " (P)"))))
        pkgs <- paste0("pkg", 1:5)
        cli_text("Packages: {.pkg {pkgs}}.")
      })
    Message
      Packages: pkg1 (P), pkg2 (P), pkg3 (P), pkg4 (P), and pkg5 (P).

# collapsing with formatting, custom seps

    Code
      local({
        cli_div(theme = list(div = list(`vec-sep` = " ... ")))
        pkgs <- paste0("pkg", 1:5)
        cli_text("Packages: {.pkg {pkgs}}.")
      })
    Message
      Packages: pkg1 ... pkg2 ... pkg3 ... pkg4, and pkg5.

# collapsing a cli_vec

    Code
      pkgs <- cli_vec(paste0("pkg", 1:5), style = list(`vec-sep` = " & ", `vec-last` = " & "))
      cli_text("Packages: {pkgs}.")
    Message
      Packages: pkg1 & pkg2 & pkg3 & pkg4 & pkg5.

# collapsing a cli_vec with styling [plain]

    Code
      local({
        cli_div(theme = list(body = list(`vec-sep` = " ... ")))
        pkgs <- cli_vec(paste0("pkg", 1:5), style = list(`vec-sep` = " & ",
          `vec-last` = " & ", color = "blue"))
        cli_text("Packages: {pkgs}.")
      })
    Message
      Packages: pkg1 & pkg2 & pkg3 & pkg4 & pkg5.

# collapsing a cli_vec with styling [ansi]

    Code
      local({
        cli_div(theme = list(body = list(`vec-sep` = " ... ")))
        pkgs <- cli_vec(paste0("pkg", 1:5), style = list(`vec-sep` = " & ",
          `vec-last` = " & ", color = "blue"))
        cli_text("Packages: {pkgs}.")
      })
    Message
      Packages: [34mpkg1[39m & [34mpkg2[39m & [34mpkg3[39m & [34mpkg4[39m & [34mpkg5[39m.

# head

    Code
      cli_text("{v(0,1)}")
    Message
      
    Code
      cli_text("{v(1,1)}")
    Message
      1
    Code
      cli_text("{v(2,1)}")
    Message
      1 and 2
    Code
      cli_text("{v(3,1)}")
    Message
      1, ...
    Code
      cli_text("{v(4,1)}")
    Message
      1, ...
    Code
      cli_text("{v(0,2)}")
    Message
      
    Code
      cli_text("{v(1,2)}")
    Message
      1
    Code
      cli_text("{v(2,2)}")
    Message
      1 and 2
    Code
      cli_text("{v(3,2)}")
    Message
      1, 2, ...
    Code
      cli_text("{v(4,2)}")
    Message
      1, 2, ...
    Code
      cli_text("{v(0,3)}")
    Message
      
    Code
      cli_text("{v(1,3)}")
    Message
      1
    Code
      cli_text("{v(2,3)}")
    Message
      1 and 2
    Code
      cli_text("{v(3,3)}")
    Message
      1, 2, and 3
    Code
      cli_text("{v(4,3)}")
    Message
      1, 2, 3, ...
    Code
      cli_text("{v(0,4)}")
    Message
      
    Code
      cli_text("{v(1,4)}")
    Message
      1
    Code
      cli_text("{v(2,4)}")
    Message
      1 and 2
    Code
      cli_text("{v(3,4)}")
    Message
      1, 2, and 3
    Code
      cli_text("{v(4,4)}")
    Message
      1, 2, 3, and 4
    Code
      cli_text("{v(0,5)}")
    Message
      
    Code
      cli_text("{v(1,5)}")
    Message
      1
    Code
      cli_text("{v(2,5)}")
    Message
      1 and 2
    Code
      cli_text("{v(3,5)}")
    Message
      1, 2, and 3
    Code
      cli_text("{v(4,5)}")
    Message
      1, 2, 3, and 4
    Code
      cli_text("{v(10,5)}")
    Message
      1, 2, 3, 4, 5, ...

# both-ends

    Code
      cli_text("{v(0,1)}")
    Message
      
    Code
      cli_text("{v(1,1)}")
    Message
      1
    Code
      cli_text("{v(2,1)}")
    Message
      1 and 2
    Code
      cli_text("{v(3,1)}")
    Message
      1, 2, and 3
    Code
      cli_text("{v(4,1)}")
    Message
      1, 2, 3, and 4
    Code
      cli_text("{v(5,1)}")
    Message
      1, 2, 3, 4, and 5
    Code
      cli_text("{v(6,1)}")
    Message
      1, 2, 3, ..., 5, and 6
    Code
      cli_text("{v(7,1)}")
    Message
      1, 2, 3, ..., 6, and 7
    Code
      cli_text("{v(10,1)}")
    Message
      1, 2, 3, ..., 9, and 10

# both-ends with formatting [plain]

    Code
      cli_text("{.val {v(0,1)}}")
    Message
      
    Code
      cli_text("{.val {v(1,1)}}")
    Message
      1
    Code
      cli_text("{.val {v(2,1)}}")
    Message
      1 and 2
    Code
      cli_text("{.val {v(3,1)}}")
    Message
      1, 2, and 3
    Code
      cli_text("{.val {v(4,1)}}")
    Message
      1, 2, 3, and 4
    Code
      cli_text("{.val {v(5,1)}}")
    Message
      1, 2, 3, 4, and 5
    Code
      cli_text("{.val {v(6,1)}}")
    Message
      1, 2, 3, ..., 5, and 6
    Code
      cli_text("{.val {v(7,1)}}")
    Message
      1, 2, 3, ..., 6, and 7
    Code
      cli_text("{.val {v(10,1)}}")
    Message
      1, 2, 3, ..., 9, and 10
    Code
      cli_text("{.val {v(10,6)}}")
    Message
      1, 2, 3, 4, ..., 9, and 10
    Code
      cli_text("{.val {v(10,10)}}")
    Message
      1, 2, 3, 4, 5, 6, 7, 8, 9, and 10
    Code
      cli_text("{.val {v(11,10)}}")
    Message
      1, 2, 3, 4, 5, 6, 7, 8, ..., 10, and 11

# both-ends with formatting [ansi]

    Code
      cli_text("{.val {v(0,1)}}")
    Message
      
    Code
      cli_text("{.val {v(1,1)}}")
    Message
      [34m1[39m
    Code
      cli_text("{.val {v(2,1)}}")
    Message
      [34m1[39m and [34m2[39m
    Code
      cli_text("{.val {v(3,1)}}")
    Message
      [34m1[39m, [34m2[39m, and [34m3[39m
    Code
      cli_text("{.val {v(4,1)}}")
    Message
      [34m1[39m, [34m2[39m, [34m3[39m, and [34m4[39m
    Code
      cli_text("{.val {v(5,1)}}")
    Message
      [34m1[39m, [34m2[39m, [34m3[39m, [34m4[39m, and [34m5[39m
    Code
      cli_text("{.val {v(6,1)}}")
    Message
      [34m1[39m, [34m2[39m, [34m3[39m, ..., [34m5[39m, and [34m6[39m
    Code
      cli_text("{.val {v(7,1)}}")
    Message
      [34m1[39m, [34m2[39m, [34m3[39m, ..., [34m6[39m, and [34m7[39m
    Code
      cli_text("{.val {v(10,1)}}")
    Message
      [34m1[39m, [34m2[39m, [34m3[39m, ..., [34m9[39m, and [34m10[39m
    Code
      cli_text("{.val {v(10,6)}}")
    Message
      [34m1[39m, [34m2[39m, [34m3[39m, [34m4[39m, ..., [34m9[39m, and [34m10[39m
    Code
      cli_text("{.val {v(10,10)}}")
    Message
      [34m1[39m, [34m2[39m, [34m3[39m, [34m4[39m, [34m5[39m, [34m6[39m, [34m7[39m, [34m8[39m, [34m9[39m, and [34m10[39m
    Code
      cli_text("{.val {v(11,10)}}")
    Message
      [34m1[39m, [34m2[39m, [34m3[39m, [34m4[39m, [34m5[39m, [34m6[39m, [34m7[39m, [34m8[39m, ..., [34m10[39m, and [34m11[39m

# ansi_collapse

    Code
      ansi_collapse(l10)
    Output
      [1] "a, b, c, d, e, f, g, h, i, and j"
    Code
      ansi_collapse(l10, trunc = 6)
    Output
      [1] "a, b, c, d, ..., i, and j"
    Code
      ansi_collapse(l10, trunc = 5)
    Output
      [1] "a, b, c, ..., i, and j"
    Code
      ansi_collapse(l10, trunc = 4)
    Output
      [1] "a, b, c, ..., i, and j"
    Code
      ansi_collapse(l10, trunc = 1)
    Output
      [1] "a, b, c, ..., i, and j"
    Code
      ansi_collapse(l10, sep = "; ")
    Output
      [1] "a; b; c; d; e; f; g; h; i, and j"
    Code
      ansi_collapse(l10, sep = "; ", last = "; or ")
    Output
      [1] "a; b; c; d; e; f; g; h; i; or j"
    Code
      ansi_collapse(l10, sep = "; ")
    Output
      [1] "a; b; c; d; e; f; g; h; i, and j"
    Code
      ansi_collapse(l10, sep = "; ", last = "; or ", trunc = 6)
    Output
      [1] "a; b; c; d; ...; i; or j"
    Code
      ansi_collapse(l10, style = "head")
    Output
      [1] "a, b, c, d, e, f, g, h, i, and j"
    Code
      ansi_collapse(l10, trunc = 6, style = "head")
    Output
      [1] "a, b, c, d, e, f, ..."
    Code
      ansi_collapse(l10, trunc = 5, style = "head")
    Output
      [1] "a, b, c, d, e, ..."
    Code
      ansi_collapse(l10, trunc = 4, style = "head")
    Output
      [1] "a, b, c, d, ..."
    Code
      ansi_collapse(l10, trunc = 1, style = "head")
    Output
      [1] "a, ..."
    Code
      ansi_collapse(l10, sep = "; ", style = "head")
    Output
      [1] "a; b; c; d; e; f; g; h; i, and j"
    Code
      ansi_collapse(l10, sep = "; ", last = "; or ", style = "head")
    Output
      [1] "a; b; c; d; e; f; g; h; i; or j"
    Code
      ansi_collapse(l10, sep = "; ", last = "; or ", trunc = 6, style = "head")
    Output
      [1] "a; b; c; d; e; f; ..."

# ansi_collapse with width trimming

    Code
      ansi_collapse(l10, width = 1, style = "head")
    Output
      <cli_ansi_string>
      [1] .
    Code
      ansi_collapse(l10, width = 2, style = "head")
    Output
      <cli_ansi_string>
      [1] ..
    Code
      ansi_collapse(l10, width = 3, style = "head")
    Output
      [1] "..."
    Code
      ansi_collapse(l10, width = 4, style = "head")
    Output
      [1] "a..."
    Code
      ansi_collapse(l10, width = 5, style = "head")
    Output
      [1] "a..."
    Code
      ansi_collapse(l10, width = 6, style = "head")
    Output
      [1] "a, ..."
    Code
      ansi_collapse(l10, width = 7, style = "head")
    Output
      [1] "a, ..."
    Code
      ansi_collapse(l10, width = 8, style = "head")
    Output
      [1] "a, ..."
    Code
      ansi_collapse(l10, width = 9, style = "head")
    Output
      [1] "a, b, ..."
    Code
      ansi_collapse(l10, width = 30, style = "head")
    Output
      [1] "a, b, c, d, e, f, g, h, i, ..."
    Code
      ansi_collapse(l10, width = 31, style = "head")
    Output
      [1] "a, b, c, d, e, f, g, h, i, ..."
    Code
      ansi_collapse(l10, width = 32, style = "head")
    Output
      [1] "a, b, c, d, e, f, g, h, i, and j"
    Code
      ansi_collapse(l10, width = 40, style = "head")
    Output
      [1] "a, b, c, d, e, f, g, h, i, and j"

---

    Code
      ansi_collapse(l10, width = 10, style = "both-ends")
    Condition
      Warning in `collapse_both_ends()`:
      ! finite `width` is not implemented in `cli::ansi_collapse()`.
      i `width = Inf` is used instead.
    Output
      [1] "a, b, c, d, e, f, g, h, i, and j"

# Issue #681

    Code
      v <- cli::cli_vec(c("foo", "bar", "foobar"), style = list(`vec-last` = ", or "))
      cli::cli_text("Must be one of: {v}.")
    Message
      Must be one of: foo, bar, or foobar.
    Code
      v <- cli::cli_vec(c("foo", "bar"), style = list(`vec-last` = " or "))
      cli::cli_text("Must be one of: {v}.")
    Message
      Must be one of: foo or bar.
    Code
      v <- cli::cli_vec(c("foo", "bar"), style = list(`vec-last` = " or ",
        `vec-sep2` = " xor "))
      cli::cli_text("Must be one of: {v}.")
    Message
      Must be one of: foo xor bar.

# cli_progress_bar

    Code
      capture_cli_messages(fun())
    Output
      [1] "\\ name status 1\n"

# removes previous progress bar

    Code
      capture_cli_messages(fun())
    Output
      [1] "first\n"       "first done\n"  "\n"            "second\n"     
      [5] "second done\n" "\n"           

# cli_progress_update can update status

    Code
      capture_cli_messages(fun())
    Output
      [1] "\\ name status 1\n"    "| name new status 2\n"

# cli_progress_update can update extra data

    Code
      capture_cli_messages(fun())
    Output
      [1] "Extra: bar\n" "Extra: baz\n"

# update set

    Code
      capture_cli_messages(fun())
    Output
      [1] "name 1/100\n"  "name 50/100\n"

# format changes if we (un)learn total

    Code
      out
    Output
      [1] "\\ name 1 done (100/s) | 3ms\n"                         
      [2] "name ===============>---------------   50% | ETA:  1s\n"
      [3] "/ name 75 done (100/s) | 3ms\n"                         

# auto-terminate

    Code
      capture_cli_messages(fun())
    Output
      [1] "first\n"                 "first done\n"           
      [3] "\n"                      "First is done by now.\n"
      [5] "second\n"                "second done\n"          
      [7] "\n"                     

# cli_progress_output

    Code
      capture_cli_messages(fun())
    Output
      [1] "\rfirst\033[K\r" "\r\033[K"        "just 1 text\n"   "first\r"        
      [5] "\rfirst\033[K\r" "\r\033[K"       

---

    Code
      capture_cli_messages(fun())
    Output
      [1] "\rfirst\r"     "\r     \r"     "just 1 text\n" "first\r"      
      [5] "\rfirst\r"     "\r     \r"    

# errors

    ! `options("cli.width")` must be an integer scalar.
    i `options("cli.width")` is a character vector.

---

    ! `options("cli.width")` cannot be `NA`.

---

    ! `options("cli.width")` must be a positive integer and not -100.

# format_error [plain]

    Code
      local({
        n <- "boo"
        stop(format_error(c("{.var n} must be a numeric vector", x = "You've supplied a {.cls {class(n)}} vector.")))
      })
    Condition
      Error:
      ! `n` must be a numeric vector
      x You've supplied a <character> vector.

---

    Code
      local({
        len <- 26
        idx <- 100
        stop(format_error(c("Must index an existing element:", i = "There {?is/are} {len} element{?s}.",
          x = "You've tried to subset element {idx}.")))
      })
    Condition
      Error:
      ! Must index an existing element:
      i There are 26 elements.
      x You've tried to subset element 100.

# format_error [ansi]

    Code
      local({
        n <- "boo"
        stop(format_error(c("{.var n} must be a numeric vector", x = "You've supplied a {.cls {class(n)}} vector.")))
      })
    Condition
      [1m[33mError[39m:[22m
      [33m![39m [1m[22m`n` must be a numeric vector
      [31mx[39m You've supplied a [34m<character>[39m vector.

---

    Code
      local({
        len <- 26
        idx <- 100
        stop(format_error(c("Must index an existing element:", i = "There {?is/are} {len} element{?s}.",
          x = "You've tried to subset element {idx}.")))
      })
    Condition
      [1m[33mError[39m:[22m
      [33m![39m [1m[22mMust index an existing element:
      [36mi[39m There are 26 elements.
      [31mx[39m You've tried to subset element 100.

# format_error [unicode]

    Code
      local({
        n <- "boo"
        stop(format_error(c("{.var n} must be a numeric vector", x = "You've supplied a {.cls {class(n)}} vector.")))
      })
    Condition
      Error:
      ! `n` must be a numeric vector
      ✖ You've supplied a <character> vector.

---

    Code
      local({
        len <- 26
        idx <- 100
        stop(format_error(c("Must index an existing element:", i = "There {?is/are} {len} element{?s}.",
          x = "You've tried to subset element {idx}.")))
      })
    Condition
      Error:
      ! Must index an existing element:
      ℹ There are 26 elements.
      ✖ You've tried to subset element 100.

# format_error [fancy]

    Code
      local({
        n <- "boo"
        stop(format_error(c("{.var n} must be a numeric vector", x = "You've supplied a {.cls {class(n)}} vector.")))
      })
    Condition
      [1m[33mError[39m:[22m
      [33m![39m [1m[22m`n` must be a numeric vector
      [31m✖[39m You've supplied a [34m<character>[39m vector.

---

    Code
      local({
        len <- 26
        idx <- 100
        stop(format_error(c("Must index an existing element:", i = "There {?is/are} {len} element{?s}.",
          x = "You've tried to subset element {idx}.")))
      })
    Condition
      [1m[33mError[39m:[22m
      [33m![39m [1m[22mMust index an existing element:
      [36mℹ[39m There are 26 elements.
      [31m✖[39m You've tried to subset element 100.

# format_warning [plain]

    Code
      n <- "boo"
      warning(format_warning(c("{.var n} must be a numeric vector", x = "You've supplied a {.cls {class(n)}} vector.")))
    Condition
      Warning:
      `n` must be a numeric vector
      x You've supplied a <character> vector.

---

    Code
      local({
        len <- 26
        idx <- 100
        warning(format_warning(c("Must index an existing element:", i = "There {?is/are} {len} element{?s}.",
          x = "You've tried to subset element {idx}.")))
      })
    Condition
      Warning:
      Must index an existing element:
      i There are 26 elements.
      x You've tried to subset element 100.

# format_warning [ansi]

    Code
      n <- "boo"
      warning(format_warning(c("{.var n} must be a numeric vector", x = "You've supplied a {.cls {class(n)}} vector.")))
    Condition
      [1m[33mWarning[39m:[22m
      [1m[22m`n` must be a numeric vector
      [31mx[39m You've supplied a [34m<character>[39m vector.

---

    Code
      local({
        len <- 26
        idx <- 100
        warning(format_warning(c("Must index an existing element:", i = "There {?is/are} {len} element{?s}.",
          x = "You've tried to subset element {idx}.")))
      })
    Condition
      [1m[33mWarning[39m:[22m
      [1m[22mMust index an existing element:
      [36mi[39m There are 26 elements.
      [31mx[39m You've tried to subset element 100.

# format_warning [unicode]

    Code
      n <- "boo"
      warning(format_warning(c("{.var n} must be a numeric vector", x = "You've supplied a {.cls {class(n)}} vector.")))
    Condition
      Warning:
      `n` must be a numeric vector
      ✖ You've supplied a <character> vector.

---

    Code
      local({
        len <- 26
        idx <- 100
        warning(format_warning(c("Must index an existing element:", i = "There {?is/are} {len} element{?s}.",
          x = "You've tried to subset element {idx}.")))
      })
    Condition
      Warning:
      Must index an existing element:
      ℹ There are 26 elements.
      ✖ You've tried to subset element 100.

# format_warning [fancy]

    Code
      n <- "boo"
      warning(format_warning(c("{.var n} must be a numeric vector", x = "You've supplied a {.cls {class(n)}} vector.")))
    Condition
      [1m[33mWarning[39m:[22m
      [1m[22m`n` must be a numeric vector
      [31m✖[39m You've supplied a [34m<character>[39m vector.

---

    Code
      local({
        len <- 26
        idx <- 100
        warning(format_warning(c("Must index an existing element:", i = "There {?is/are} {len} element{?s}.",
          x = "You've tried to subset element {idx}.")))
      })
    Condition
      [1m[33mWarning[39m:[22m
      [1m[22mMust index an existing element:
      [36mℹ[39m There are 26 elements.
      [31m✖[39m You've tried to subset element 100.

# format_message [plain]

    Code
      n <- "boo"
      message(format_message(c("{.var n} must be a numeric vector", x = "You've supplied a {.cls {class(n)}} vector.")))
    Message
      `n` must be a numeric vector
      x You've supplied a <character> vector.

---

    Code
      local({
        len <- 26
        idx <- 100
        message(format_message(c("Must index an existing element:", i = "There {?is/are} {len} element{?s}.",
          x = "You've tried to subset element {idx}.")))
      })
    Message
      Must index an existing element:
      i There are 26 elements.
      x You've tried to subset element 100.

# format_message [ansi]

    Code
      n <- "boo"
      message(format_message(c("{.var n} must be a numeric vector", x = "You've supplied a {.cls {class(n)}} vector.")))
    Message
      [1m[22m`n` must be a numeric vector
      [31mx[39m You've supplied a [34m<character>[39m vector.

---

    Code
      local({
        len <- 26
        idx <- 100
        message(format_message(c("Must index an existing element:", i = "There {?is/are} {len} element{?s}.",
          x = "You've tried to subset element {idx}.")))
      })
    Message
      [1m[22mMust index an existing element:
      [36mi[39m There are 26 elements.
      [31mx[39m You've tried to subset element 100.

# format_message [unicode]

    Code
      n <- "boo"
      message(format_message(c("{.var n} must be a numeric vector", x = "You've supplied a {.cls {class(n)}} vector.")))
    Message
      `n` must be a numeric vector
      ✖ You've supplied a <character> vector.

---

    Code
      local({
        len <- 26
        idx <- 100
        message(format_message(c("Must index an existing element:", i = "There {?is/are} {len} element{?s}.",
          x = "You've tried to subset element {idx}.")))
      })
    Message
      Must index an existing element:
      ℹ There are 26 elements.
      ✖ You've tried to subset element 100.

# format_message [fancy]

    Code
      n <- "boo"
      message(format_message(c("{.var n} must be a numeric vector", x = "You've supplied a {.cls {class(n)}} vector.")))
    Message
      [1m[22m`n` must be a numeric vector
      [31m✖[39m You've supplied a [34m<character>[39m vector.

---

    Code
      local({
        len <- 26
        idx <- 100
        message(format_message(c("Must index an existing element:", i = "There {?is/are} {len} element{?s}.",
          x = "You've tried to subset element {idx}.")))
      })
    Message
      [1m[22mMust index an existing element:
      [36mℹ[39m There are 26 elements.
      [31m✖[39m You've tried to subset element 100.

# color in RStudio [ansi]

    Code
      col <- get_rstudio_fg_color0()
      cat(col("this is the new color"))
    Output
      [30mthis is the new color[39m

# update_rstudio_color [ansi]

    Code
      cat(update_rstudio_color("color me interested"))
    Output
      [32mcolor me interested[39m

# named first element

    Code
      format_error(c(`*` = "foo", `*` = "bar"))
    Output
      [1] "* foo\n* bar"

---

    Code
      format_warning(c(`*` = "foo", `*` = "bar"))
    Output
      [1] "* foo\n* bar"

# cli.condition_width

    Code
      format_error(msg)
    Output
      [1] "1234567890 1234567890 1234567890\n1234567890 1234567890 1234567890\n1234567890 1234567890"
    Code
      format_warning(msg)
    Output
      [1] "1234567890 1234567890 1234567890\n1234567890 1234567890 1234567890\n1234567890 1234567890"
    Code
      format_message(msg)
    Output
      [1] "1234567890 1234567890 1234567890\n1234567890 1234567890 1234567890\n1234567890 1234567890"

---

    Code
      format_error(msg)
    Output
      [1] "1234567890 1234567890 1234567890 1234567890 1234567890 1234567890 1234567890 1234567890"
    Code
      format_warning(msg)
    Output
      [1] "1234567890 1234567890 1234567890 1234567890 1234567890 1234567890 1234567890 1234567890"
    Code
      format_message(msg)
    Output
      [1] "1234567890 1234567890 1234567890 1234567890 1234567890 1234567890 1234567890 1234567890"

# suppressing Unicode bullets [plain]

    Code
      local({
        n <- "boo"
        stop(format_error(c("{.var n} must be a numeric vector", x = "You've supplied a {.cls {class(n)}} vector.",
          v = "Success.", i = "Info.", `*` = "Bullet", `>` = "Arrow")))
      })
    Condition
      Error:
      ! `n` must be a numeric vector
      x You've supplied a <character> vector.
      v Success.
      i Info.
      * Bullet
      > Arrow

# suppressing Unicode bullets [ansi]

    Code
      local({
        n <- "boo"
        stop(format_error(c("{.var n} must be a numeric vector", x = "You've supplied a {.cls {class(n)}} vector.",
          v = "Success.", i = "Info.", `*` = "Bullet", `>` = "Arrow")))
      })
    Condition
      [1m[33mError[39m:[22m
      [33m![39m [1m[22m`n` must be a numeric vector
      [31mx[39m You've supplied a [34m<character>[39m vector.
      [32mv[39m Success.
      [36mi[39m Info.
      [36m*[39m Bullet
      > Arrow

# suppressing Unicode bullets [unicode]

    Code
      local({
        n <- "boo"
        stop(format_error(c("{.var n} must be a numeric vector", x = "You've supplied a {.cls {class(n)}} vector.",
          v = "Success.", i = "Info.", `*` = "Bullet", `>` = "Arrow")))
      })
    Condition
      Error:
      ! `n` must be a numeric vector
      x You've supplied a <character> vector.
      v Success.
      i Info.
      * Bullet
      > Arrow

# suppressing Unicode bullets [fancy]

    Code
      local({
        n <- "boo"
        stop(format_error(c("{.var n} must be a numeric vector", x = "You've supplied a {.cls {class(n)}} vector.",
          v = "Success.", i = "Info.", `*` = "Bullet", `>` = "Arrow")))
      })
    Condition
      [1m[33mError[39m:[22m
      [33m![39m [1m[22m`n` must be a numeric vector
      [31mx[39m You've supplied a [34m<character>[39m vector.
      [32mv[39m Success.
      [36mi[39m Info.
      [36m*[39m Bullet
      > Arrow

# cli_abort [plain]

    Code
      local({
        n <- "boo"
        cli_abort(c("{.var n} must be a numeric vector", x = "You've supplied a {.cls {class(n)}} vector."))
      })
    Condition
      Error:
      ! `n` must be a numeric vector
      x You've supplied a <character> vector.

---

    Code
      local({
        len <- 26
        idx <- 100
        cli_abort(c("Must index an existing element:", i = "There {?is/are} {len} element{?s}.",
          x = "You've tried to subset element {idx}."))
      })
    Condition
      Error:
      ! Must index an existing element:
      i There are 26 elements.
      x You've tried to subset element 100.

---

    Code
      c(err$message, err$body)
    Output
                                                                                    x 
               "`n` must be a numeric vector" "You've supplied a <character> vector." 

# cli_abort [ansi]

    Code
      local({
        n <- "boo"
        cli_abort(c("{.var n} must be a numeric vector", x = "You've supplied a {.cls {class(n)}} vector."))
      })
    Condition
      [1m[33mError[39m:[22m
      [1m[22m[33m![39m `n` must be a numeric vector
      [31mx[39m You've supplied a [34m<character>[39m vector.

---

    Code
      local({
        len <- 26
        idx <- 100
        cli_abort(c("Must index an existing element:", i = "There {?is/are} {len} element{?s}.",
          x = "You've tried to subset element {idx}."))
      })
    Condition
      [1m[33mError[39m:[22m
      [1m[22m[33m![39m Must index an existing element:
      [36mi[39m There are 26 elements.
      [31mx[39m You've tried to subset element 100.

---

    Code
      c(err$message, err$body)
    Output
                                                              
                               "`n` must be a numeric vector" 
                                                            x 
      "You've supplied a \033[34m<character>\033[39m vector." 

# cli_abort [unicode]

    Code
      local({
        n <- "boo"
        cli_abort(c("{.var n} must be a numeric vector", x = "You've supplied a {.cls {class(n)}} vector."))
      })
    Condition
      Error:
      ! `n` must be a numeric vector
      ✖ You've supplied a <character> vector.

---

    Code
      local({
        len <- 26
        idx <- 100
        cli_abort(c("Must index an existing element:", i = "There {?is/are} {len} element{?s}.",
          x = "You've tried to subset element {idx}."))
      })
    Condition
      Error:
      ! Must index an existing element:
      ℹ There are 26 elements.
      ✖ You've tried to subset element 100.

---

    Code
      c(err$message, err$body)
    Output
                                                                                    x 
               "`n` must be a numeric vector" "You've supplied a <character> vector." 

# cli_abort [fancy]

    Code
      local({
        n <- "boo"
        cli_abort(c("{.var n} must be a numeric vector", x = "You've supplied a {.cls {class(n)}} vector."))
      })
    Condition
      [1m[33mError[39m:[22m
      [1m[22m[33m![39m `n` must be a numeric vector
      [31m✖[39m You've supplied a [34m<character>[39m vector.

---

    Code
      local({
        len <- 26
        idx <- 100
        cli_abort(c("Must index an existing element:", i = "There {?is/are} {len} element{?s}.",
          x = "You've tried to subset element {idx}."))
      })
    Condition
      [1m[33mError[39m:[22m
      [1m[22m[33m![39m Must index an existing element:
      [36mℹ[39m There are 26 elements.
      [31m✖[39m You've tried to subset element 100.

---

    Code
      c(err$message, err$body)
    Output
                                                              
                               "`n` must be a numeric vector" 
                                                            x 
      "You've supplied a \033[34m<character>\033[39m vector." 

# cli_warn [plain]

    Code
      n <- "boo"
      cli_warn(c("{.var n} must be a numeric vector", x = "You've supplied a {.cls {class(n)}} vector."))
    Condition
      Warning:
      `n` must be a numeric vector
      x You've supplied a <character> vector.

---

    Code
      local({
        len <- 26
        idx <- 100
        cli_warn(c("Must index an existing element:", i = "There {?is/are} {len} element{?s}.",
          x = "You've tried to subset element {idx}."))
      })
    Condition
      Warning:
      Must index an existing element:
      i There are 26 elements.
      x You've tried to subset element 100.

# cli_warn [ansi]

    Code
      n <- "boo"
      cli_warn(c("{.var n} must be a numeric vector", x = "You've supplied a {.cls {class(n)}} vector."))
    Condition
      [1m[33mWarning[39m:[22m
      [1m[22m`n` must be a numeric vector
      [31mx[39m You've supplied a [34m<character>[39m vector.

---

    Code
      local({
        len <- 26
        idx <- 100
        cli_warn(c("Must index an existing element:", i = "There {?is/are} {len} element{?s}.",
          x = "You've tried to subset element {idx}."))
      })
    Condition
      [1m[33mWarning[39m:[22m
      [1m[22mMust index an existing element:
      [36mi[39m There are 26 elements.
      [31mx[39m You've tried to subset element 100.

# cli_warn [unicode]

    Code
      n <- "boo"
      cli_warn(c("{.var n} must be a numeric vector", x = "You've supplied a {.cls {class(n)}} vector."))
    Condition
      Warning:
      `n` must be a numeric vector
      ✖ You've supplied a <character> vector.

---

    Code
      local({
        len <- 26
        idx <- 100
        cli_warn(c("Must index an existing element:", i = "There {?is/are} {len} element{?s}.",
          x = "You've tried to subset element {idx}."))
      })
    Condition
      Warning:
      Must index an existing element:
      ℹ There are 26 elements.
      ✖ You've tried to subset element 100.

# cli_warn [fancy]

    Code
      n <- "boo"
      cli_warn(c("{.var n} must be a numeric vector", x = "You've supplied a {.cls {class(n)}} vector."))
    Condition
      [1m[33mWarning[39m:[22m
      [1m[22m`n` must be a numeric vector
      [31m✖[39m You've supplied a [34m<character>[39m vector.

---

    Code
      local({
        len <- 26
        idx <- 100
        cli_warn(c("Must index an existing element:", i = "There {?is/are} {len} element{?s}.",
          x = "You've tried to subset element {idx}."))
      })
    Condition
      [1m[33mWarning[39m:[22m
      [1m[22mMust index an existing element:
      [36mℹ[39m There are 26 elements.
      [31m✖[39m You've tried to subset element 100.

# cli_inform [plain]

    Code
      n <- "boo"
      cli_inform(c("{.var n} must be a numeric vector", x = "You've supplied a {.cls {class(n)}} vector."))
    Message
      `n` must be a numeric vector
      x You've supplied a <character> vector.

---

    Code
      local({
        len <- 26
        idx <- 100
        cli_inform(c("Must index an existing element:", i = "There {?is/are} {len} element{?s}.",
          x = "You've tried to subset element {idx}."))
      })
    Message
      Must index an existing element:
      i There are 26 elements.
      x You've tried to subset element 100.

# cli_inform [ansi]

    Code
      n <- "boo"
      cli_inform(c("{.var n} must be a numeric vector", x = "You've supplied a {.cls {class(n)}} vector."))
    Message
      [1m[22m`n` must be a numeric vector
      [31mx[39m You've supplied a [34m<character>[39m vector.

---

    Code
      local({
        len <- 26
        idx <- 100
        cli_inform(c("Must index an existing element:", i = "There {?is/are} {len} element{?s}.",
          x = "You've tried to subset element {idx}."))
      })
    Message
      [1m[22mMust index an existing element:
      [36mi[39m There are 26 elements.
      [31mx[39m You've tried to subset element 100.

# cli_inform [unicode]

    Code
      n <- "boo"
      cli_inform(c("{.var n} must be a numeric vector", x = "You've supplied a {.cls {class(n)}} vector."))
    Message
      `n` must be a numeric vector
      ✖ You've supplied a <character> vector.

---

    Code
      local({
        len <- 26
        idx <- 100
        cli_inform(c("Must index an existing element:", i = "There {?is/are} {len} element{?s}.",
          x = "You've tried to subset element {idx}."))
      })
    Message
      Must index an existing element:
      ℹ There are 26 elements.
      ✖ You've tried to subset element 100.

# cli_inform [fancy]

    Code
      n <- "boo"
      cli_inform(c("{.var n} must be a numeric vector", x = "You've supplied a {.cls {class(n)}} vector."))
    Message
      [1m[22m`n` must be a numeric vector
      [31m✖[39m You've supplied a [34m<character>[39m vector.

---

    Code
      local({
        len <- 26
        idx <- 100
        cli_inform(c("Must index an existing element:", i = "There {?is/are} {len} element{?s}.",
          x = "You've tried to subset element {idx}."))
      })
    Message
      [1m[22mMust index an existing element:
      [36mℹ[39m There are 26 elements.
      [31m✖[39m You've tried to subset element 100.

# cli_abort width in RStudio

    Code
      local({
        len <- 26
        idx <- 100
        cli_abort(c(lorem_ipsum(1, 3), i = lorem_ipsum(1, 3), x = lorem_ipsum(1, 3)))
      })
    Condition
      Error:
      ! Duis quis magna incididunt nulla commodo minim non exercitation
        nostrud ullamco dolor exercitation ut veniam. Fugiat irure tempor commodo
        voluptate ut. In et tempor excepteur quis.
      i Et nisi ad quis ad cupidatat tempor laborum est excepteur aliqua veniam ex.
        Sunt magna veniam Lorem elit enim et pariatur aliqua occaecat mollit
        consequat dolore in mollit. Officia labore reprehenderit culpa dolore quis
        nisi do aliqua commodo deserunt fugiat cupidatat nostrud ad.
      x Ad laboris consectetur esse minim pariatur irure do anim anim. Mollit ad
        cupidatat ullamco ullamco nulla elit in.

# color in RStudio [ansi]

    Code
      col <- get_rstudio_fg_color0()
      cat(col("this is the new color"))
    Output
      [30mthis is the new color[39m

# update_rstudio_color [ansi]

    Code
      cat(update_rstudio_color("color me interested"))
    Output
      [32mcolor me interested[39m

# generic

    Code
      local({
        cli_div(theme = list(.alert = list(before = "GENERIC! ")))
        cli_alert("wow")
      })
    Message
      GENERIC! wow

# success [plain]

    Code
      local({
        cli_alert_success("wow")
      })
    Message
      v wow

# success [ansi]

    Code
      local({
        cli_alert_success("wow")
      })
    Message
      [32mv[39m wow

# success [unicode]

    Code
      local({
        cli_alert_success("wow")
      })
    Message
      ✔ wow

# success [fancy]

    Code
      local({
        cli_alert_success("wow")
      })
    Message
      [32m✔[39m wow

# danger [plain]

    Code
      local({
        cli_alert_danger("wow")
      })
    Message
      x wow

# danger [ansi]

    Code
      local({
        cli_alert_danger("wow")
      })
    Message
      [31mx[39m wow

# danger [unicode]

    Code
      local({
        cli_alert_danger("wow")
      })
    Message
      ✖ wow

# danger [fancy]

    Code
      local({
        cli_alert_danger("wow")
      })
    Message
      [31m✖[39m wow

# warning [plain]

    Code
      local({
        cli_alert_warning("wow")
      })
    Message
      ! wow

# warning [ansi]

    Code
      local({
        cli_alert_warning("wow")
      })
    Message
      [33m![39m wow

# warning [unicode]

    Code
      local({
        cli_alert_warning("wow")
      })
    Message
      ! wow

# warning [fancy]

    Code
      local({
        cli_alert_warning("wow")
      })
    Message
      [33m![39m wow

# info [plain]

    Code
      local({
        cli_alert_info("wow")
      })
    Message
      i wow

# info [ansi]

    Code
      local({
        cli_alert_info("wow")
      })
    Message
      [36mi[39m wow

# info [unicode]

    Code
      local({
        cli_alert_info("wow")
      })
    Message
      ℹ wow

# info [fancy]

    Code
      local({
        cli_alert_info("wow")
      })
    Message
      [36mℹ[39m wow

# before and after can have spaces

    Code
      local({
        cli_div(theme = list(.alert = list(before = "x  ", after = "  x")))
        cli_alert("continuing that first alert", wrap = TRUE)
      })
    Message
      x  continuing that first alert  x

# inline classes [plain]

    Code
      invisible(lapply(classes, do))
    Message
      This is <<<it>>> really
      This is <<<it>>> really
      This is <<<it>>> really
      This is <<<it>>> really
      This is <<<it>>> really
      This is <<<it>>> really
      This is <<<it>>> really
      This is <<<it>>> really
      This is <<<it>>> really
      This is <<<it>>> really
      This is <<<it>>> really
      This is <<<it>>> really
      This is <<<it>>> really
      This is <<<it>>> really

# inline classes [ansi]

    Code
      invisible(lapply(classes, do))
    Message
      This is [36m<<<it>>>[39m really
      This is [36m<<<it>>>[39m really
      This is [36m<<<it>>>[39m really
      This is [36m<<<it>>>[39m really
      This is [36m<<<it>>>[39m really
      This is [36m<<<it>>>[39m really
      This is [36m<<<it>>>[39m really
      This is [36m<<<it>>>[39m really
      This is [36m<<<it>>>[39m really
      This is [36m<<<it>>>[39m really
      This is [36m<<<it>>>[39m really
      This is [36m<<<it>>>[39m really
      This is [36m<<<it>>>[39m really
      This is [36m<<<it>>>[39m really

# {{ and }} can be used for comments

    Code
      local({
        cli_text("Escaping {{ works")
        cli_text("Escaping }} works")
        cli_text("Escaping {{ and }} works")
        cli_text("Escaping {{{{ works")
        cli_text("Escaping }}}} works")
        cli_text("Escaping {{{{ and }} works")
        cli_text("Escaping {{{{ and }}}} works")
        cli_text("Escaping {{ and }}}} works")
      })
    Message
      Escaping { works
      Escaping } works
      Escaping { and } works
      Escaping {{ works
      Escaping }} works
      Escaping {{ and } works
      Escaping {{ and }} works
      Escaping { and }} works

# no glue substitution in expressions that evaluate to a string

    Code
      local({
        msg <- "Message with special characters like } { }} {{"
        cli_text("{msg}")
        cli_text("{.emph {msg}}")
      })
    Message
      Message with special characters like } { }} {{
      Message with special characters like } { }} {{

# S3 class is used for styling

    Code
      local({
        cli_div(theme = list(div = list(`class-map` = list(foo = "bar")), .bar = list(
          before = "::")))
        obj <- structure("yep", class = "foo")
        cli_text("This is {obj}.")
      })
    Message
      This is ::yep.

# various errors

    Code
      cli_text("xx {__cannot-parse-this__} yy")
    Condition
      Error:
      ! Could not parse cli `{}` expression: `__cannot-parse-th...`.
      Caused by error:
      ! <text>:1:2: unexpected input
      1: __
           ^

# cli_progress_message

    Code
      capture_cli_messages(fun())
    Output
      [1] "Simplest progress 'bar', `fn()` 2 twos\n"

# cli_progress_message error

    Code
      readLines(outfile)
    Output
      [1] "Simplest progress 'bar', `fn()` 2 twos"
      [2] "Error in (function ()  : oopsie"       

---

    Code
      win2unix(out)
    Output
      [1] "\033[?25l\rSimplest progress 'bar', `fn()` 2 twos\033[K\r\r\033[K\033[?25hError in (function ()  : oopsie\n"

# cli_progress_step

    Code
      msgs
    Output
      [1] "\ri First step\033[K\r"       "\rv First step [1s]\033[K\r" 
      [3] "\n"                           "\ri Second step\033[K\r"     
      [5] "\rv Second step [1s]\033[K\r" "\n"                          

# cli_progress_step error

    Code
      win2unix(out)
    Output
      [1] "i First step\nv First step [1s]\n\ni Second step\nx Second step [1s]\n\nError in (function ()  : oopsie\n"

# errors

    Code
      make_ansi_style(1:10)
    Condition
      Error:
      ! `style` must be an ANSI style
      i an ANSI style is a character scalar (cli style name, RGB or R color name), or a [3x1] or [4x1] numeric RGB matrix
      i `style` is an integer vector

# make_ansi_style

    Code
      make_ansi_style(1:10)
    Condition
      [1m[33mError[39m:[22m
      [33m![39m [1m[22m`style` must be an ANSI style
      [36mi[39m an ANSI style is a character scalar (cli style name, RGB or R color name), or a [3x1] or [4x1] numeric RGB matrix
      [36mi[39m `style` is an integer vector

---

    Code
      make_ansi_style("foobar")
    Condition
      [1m[33mError[39m:[22m
      [33m![39m [1m[22mUnknown style specification: [34m"style"[39m, it must be one of
      [36m*[39m a builtin cli style, e.g. [34m"bold"[39m or [34m"red"[39m,
      [36m*[39m an R color name, see `?grDevices::colors()`.
      [36m*[39m a [3x1] or [4x1] numeric RGB matrix with, range 0-255.

# quoting phrases that don't start or end with letter or number [plain]

    Code
      local({
        x0 <- "good-name"
        cli_text("The name is {.file {x0}}.")
        x <- "weird-name "
        cli_text("The name is {.file {x}}.")
        cli_text("The name is {.path {x}}.")
        cli_text("The name is {.email {x}}.")
      })
    Message
      The name is 'good-name'.
      The name is 'weird-name '.
      The name is 'weird-name '.
      The name is 'weird-name '.

# quoting phrases that don't start or end with letter or number [ansi]

    Code
      local({
        x0 <- "good-name"
        cli_text("The name is {.file {x0}}.")
        x <- "weird-name "
        cli_text("The name is {.file {x}}.")
        cli_text("The name is {.path {x}}.")
        cli_text("The name is {.email {x}}.")
      })
    Message
      The name is [34mgood-name[39m.
      The name is '[34mweird-name[39m[44m [49m'.
      The name is '[34mweird-name[39m[44m [49m'.
      The name is '[34mweird-name[39m[44m [49m'.

# quoting weird names, still [plain]

    Code
      local({
        cat_line(nb(quote_weird_name("good")))
        cat_line(nb(quote_weird_name("  bad")))
        cat_line(nb(quote_weird_name("bad  ")))
        cat_line(nb(quote_weird_name("  bad  ")))
      })
    Output
      'good'
      '  bad'
      'bad  '
      '  bad  '

# quoting weird names, still [ansi]

    Code
      local({
        cat_line(nb(quote_weird_name("good")))
        cat_line(nb(quote_weird_name("  bad")))
        cat_line(nb(quote_weird_name("bad  ")))
        cat_line(nb(quote_weird_name("  bad  ")))
      })
    Output
      good
      '[44m  [49mbad'
      'bad[44m  [49m'
      '[44m  [49mbad[44m  [49m'

# ~/ files are not weird [ansi]

    Code
      local({
        cat_line(nb(quote_weird_name("~/good")))
        cat_line(nb(quote_weird_name("~~bad")))
        cat_line(nb(quote_weird_name("bad~  ")))
        cat_line(nb(quote_weird_name(" ~ bad ~ ")))
      })
    Output
      ~/good
      '~~bad'
      'bad~[44m  [49m'
      '[44m [49m~ bad ~[44m [49m'

# custom truncation [plain]

    Code
      x <- cli_vec(1:100, list(`vec-trunc` = 5))
      cli_text("Some numbers: {x}.")
    Message
      Some numbers: 1, 2, 3, ..., 99, and 100.
    Code
      cli_text("Some numbers: {.val {x}}.")
    Message
      Some numbers: 1, 2, 3, ..., 99, and 100.

# custom truncation [ansi]

    Code
      x <- cli_vec(1:100, list(`vec-trunc` = 5))
      cli_text("Some numbers: {x}.")
    Message
      Some numbers: 1, 2, 3, ..., 99, and 100.
    Code
      cli_text("Some numbers: {.val {x}}.")
    Message
      Some numbers: [34m1[39m, [34m2[39m, [34m3[39m, ..., [34m99[39m, and [34m100[39m.

# custom truncation [unicode]

    Code
      x <- cli_vec(1:100, list(`vec-trunc` = 5))
      cli_text("Some numbers: {x}.")
    Message
      Some numbers: 1, 2, 3, …, 99, and 100.
    Code
      cli_text("Some numbers: {.val {x}}.")
    Message
      Some numbers: 1, 2, 3, …, 99, and 100.

# custom truncation [fancy]

    Code
      x <- cli_vec(1:100, list(`vec-trunc` = 5))
      cli_text("Some numbers: {x}.")
    Message
      Some numbers: 1, 2, 3, …, 99, and 100.
    Code
      cli_text("Some numbers: {.val {x}}.")
    Message
      Some numbers: [34m1[39m, [34m2[39m, [34m3[39m, …, [34m99[39m, and [34m100[39m.

# collapsing class names [plain]

    Code
      local({
        cc <- c("one", "two")
        cli_text("this is a class: {.cls myclass}")
        cli_text("multiple classes: {.cls {cc}}")
      })
    Message
      this is a class: <myclass>
      multiple classes: <one/two>

# collapsing class names [ansi]

    Code
      local({
        cc <- c("one", "two")
        cli_text("this is a class: {.cls myclass}")
        cli_text("multiple classes: {.cls {cc}}")
      })
    Message
      this is a class: [34m<myclass>[39m
      multiple classes: [34m<one/two>[39m

# transform [plain]

    Code
      local({
        cli_text("This is a {.field field} (before)")
        foo <- (function(x) toupper(x))
        cli_div(theme = list(span.field = list(transform = foo)))
        cli_text("This is a {.field field} (during)")
        cli_end()
        cli_text("This is a {.field field} (after)")
      })
    Message
      This is a field (before)
      This is a FIELD (during)
      This is a field (after)

# transform [ansi]

    Code
      local({
        cli_text("This is a {.field field} (before)")
        foo <- (function(x) toupper(x))
        cli_div(theme = list(span.field = list(transform = foo)))
        cli_text("This is a {.field field} (during)")
        cli_end()
        cli_text("This is a {.field field} (after)")
      })
    Message
      This is a [32mfield[39m (before)
      This is a [32mFIELD[39m (during)
      This is a [32mfield[39m (after)

# cli_format

    Code
      cli_format(1:4 / 7, list(digits = 2))
    Output
      [1] 0.14 0.29 0.43 0.57

# cli_format() is used for .val

    Code
      local({
        cli_div(theme = list(.val = list(digits = 2)))
        cli_text("Some random numbers: {.val {runif(4)}}.")
      })
    Message
      Some random numbers: 0.91, 0.94, 0.29, and 0.83.

# .q always double quotes

    Code
      cli_text("just a {.q string}, nothing more")
    Message
      just a "string", nothing more

# .or

    Code
      cli_text("{.or {letters[1:5]}}")
    Message
      a, b, c, d, or e

---

    Code
      cli_text("{.or {letters[1:2]}}")
    Message
      a or b

# line breaks

    Code
      ansi_strwrap(txt2, width = 60)
    Output
      <cli_ansi_string>
      [1] Cupidatat deserunt culpa enim deserunt minim aliqua tempor
      [2] fugiat cupidatat laboris officia esse ex aliqua. Ullamco  
      [3] mollit adipisicing anim.                                  
      [4] Cupidatat deserunt culpa enim deserunt minim aliqua tempor
      [5] fugiat cupidatat laboris officia esse ex aliqua. Ullamco  
      [6] mollit adipisicing anim.                                  

# double ticks [ansi]

    Code
      format_inline("{.code {x}}")
    Output
      [1] "\033[31m`a`\033[39m, \033[31m`` `x` ``\033[39m, and \033[31m`b`\033[39m"

---

    Code
      format_inline("{.fun {x}}")
    Output
      [1] "\033[31m`a()`\033[39m, \033[31m`` `x` ()``\033[39m, and \033[31m`b()`\033[39m"

# do not inherit 'transform' issue #422

    Code
      d <- deparse(c("cli", "glue"))
      cli::cli_alert_info("To install, run {.code install.packages({d})}")
    Message
      i To install, run `install.packages(c("cli", "glue"))`

---

    Code
      cli::cli_text("{.code foo({1+1})}")
    Message
      `foo(2)`

# no inherit color, issue #474 [plain]

    Code
      cli::cli_text("pre {.val x {'foo'} y} post")
    Message
      pre "x foo y" post

# no inherit color, issue #474 [ansi]

    Code
      cli::cli_text("pre {.val x {'foo'} y} post")
    Message
      pre [34m"x foo y"[39m post

# \f at the end, issue #491 [plain]

    Code
      cli_fmt(cli::cli_text("{.val a}{.val b}"))
    Output
      [1] "\"a\"\"b\""
    Code
      cli_fmt(cli::cli_text("\f{.val a}{.val b}"))
    Output
      [1] ""           "\"a\"\"b\""
    Code
      cli_fmt(cli::cli_text("\f\f{.val a}{.val b}"))
    Output
      [1] ""           ""           "\"a\"\"b\""
    Code
      cli_fmt(cli::cli_text("{.val a}\f{.val b}"))
    Output
      [1] "\"a\"" "\"b\""
    Code
      cli_fmt(cli::cli_text("{.val a}\f\f{.val b}"))
    Output
      [1] "\"a\"" ""      "\"b\""
    Code
      cli_fmt(cli::cli_text("{.val a}{.val b}\f"))
    Output
      [1] "\"a\"\"b\"" ""          
    Code
      cli_fmt(cli::cli_text("{.val a}{.val b}\f\f"))
    Output
      [1] "\"a\"\"b\"" ""           ""          
    Code
      cli_fmt(cli::cli_text("\f\f\f{.val a}\f\f\f{.val b}\f\f\f"))
    Output
       [1] ""      ""      ""      "\"a\"" ""      ""      "\"b\"" ""      ""     
      [10] ""     

# \f at the end, issue #491 [ansi]

    Code
      cli_fmt(cli::cli_text("{.val a}{.val b}"))
    Output
      [1] "\033[34m\"a\"\"b\"\033[39m"
    Code
      cli_fmt(cli::cli_text("\f{.val a}{.val b}"))
    Output
      [1] ""                           "\033[34m\"a\"\"b\"\033[39m"
    Code
      cli_fmt(cli::cli_text("\f\f{.val a}{.val b}"))
    Output
      [1] ""                           ""                          
      [3] "\033[34m\"a\"\"b\"\033[39m"
    Code
      cli_fmt(cli::cli_text("{.val a}\f{.val b}"))
    Output
      [1] "\033[34m\"a\"\033[39m" "\033[34m\"b\"\033[39m"
    Code
      cli_fmt(cli::cli_text("{.val a}\f\f{.val b}"))
    Output
      [1] "\033[34m\"a\"\033[39m" ""                      "\033[34m\"b\"\033[39m"
    Code
      cli_fmt(cli::cli_text("{.val a}{.val b}\f"))
    Output
      [1] "\033[34m\"a\"\"b\"\033[39m" ""                          
    Code
      cli_fmt(cli::cli_text("{.val a}{.val b}\f\f"))
    Output
      [1] "\033[34m\"a\"\"b\"\033[39m" ""                          
      [3] ""                          
    Code
      cli_fmt(cli::cli_text("\f\f\f{.val a}\f\f\f{.val b}\f\f\f"))
    Output
       [1] ""                      ""                      ""                     
       [4] "\033[34m\"a\"\033[39m" ""                      ""                     
       [7] "\033[34m\"b\"\033[39m" ""                      ""                     
      [10] ""                     

# truncate vectors at 20

    Code
      cli::cli_text("Some letters: {letters}")
    Message
      Some letters: a, b, c, d, e, f, g, h, i, j, k, l, m, n, o, p, q, r, ..., y, and
      z

# brace expresssion edge cases [ansi]

    Code
      cli_text("{.code {foo} and {bar}}")
    Message
      `foo and bar`
    Code
      cli_text("{.emph {foo} and {bar}}")
    Message
      [3mfoo and bar[23m
    Code
      cli_text("{.q {foo} and {bar}}")
    Message
      "foo and bar"

# various errors

    ! Invalid cli literal: `{.foobar}` starts with a dot.
    i Interpreted literals must not start with a dot in cli >= 3.4.0.
    i `{}` expressions starting with a dot are now only used for cli styles.
    i To avoid this error, put a space character after the starting `{` or use parentheses: `{(.foobar)}`.

---

    ! Invalid cli literal: `{.someve...}` starts with a dot.
    i Interpreted literals must not start with a dot in cli >= 3.4.0.
    i `{}` expressions starting with a dot are now only used for cli styles.
    i To avoid this error, put a space character after the starting `{` or use parentheses: `{(.someve...)}`.

---

    Code
      cli_text("xx {1 + 'a'} yy")
    Condition
      Error:
      ! Could not evaluate cli `{}` expression: `1 + 'a'`.
      Caused by error:
      ! non-numeric argument to binary operator

# format_inline and newlines

    Code
      format_inline("foo\nbar")
    Output
      [1] "foo\nbar"
    Code
      format_inline("\nfoo\n\nbar\n")
    Output
      [1] "\nfoo\n\nbar\n"
    Code
      format_inline("foo\fbar")
    Output
      [1] "foo\fbar"
    Code
      format_inline("\ffoo\f\fbar\f")
    Output
      [1] "\ffoo\f\fbar\f"

---

    Code
      format_inline("foo\nbar", keep_whitespace = FALSE)
    Output
      [1] "foo bar"
    Code
      format_inline("\nfoo\n\nbar\n", keep_whitespace = FALSE)
    Output
      [1] "foo\n\nbar"
    Code
      format_inline("foo\fbar", keep_whitespace = FALSE)
    Output
      [1] "foo\nbar"
    Code
      format_inline("\ffoo\f\fbar\f", keep_whitespace = FALSE)
    Output
      [1] "\nfoo\n\nbar\n"

# ruler

    Code
      ruler(20)
    Output
      ----+----1----+----2
      12345678901234567890

# na.omit

    Code
      na.omit(character())
    Output
      character(0)
    Code
      na.omit(integer())
    Output
      integer(0)
    Code
      na.omit(1:5)
    Output
      [1] 1 2 3 4 5
    Code
      na.omit(c(1, NA, 2, NA))
    Output
      [1] 1 2
    Code
      na.omit(c(NA_integer_, NA_integer_))
    Output
      integer(0)
    Code
      na.omit(list(1, 2, 3))
    Output
      [[1]]
      [1] 1
      
      [[2]]
      [1] 2
      
      [[3]]
      [1] 3
      

# str_trim

    Code
      str_trim("foo")
    Output
      [1] "foo"
    Code
      str_trim(character())
    Output
      character(0)
    Code
      str_trim("   foo")
    Output
      [1] "foo"
    Code
      str_trim("foo  ")
    Output
      [1] "foo"
    Code
      str_trim("   foo  ")
    Output
      [1] "foo"
    Code
      str_trim(c(NA_character_, " foo  ", NA_character_, "  bar  "))
    Output
      [1] NA    "foo" NA    "bar"

# leading_space

    Code
      paste0("-", leading_space("foo"), "-")
    Output
      [1] "--"
    Code
      paste0("-", leading_space("  foo"), "-")
    Output
      [1] "-  -"
    Code
      paste0("-", leading_space("  foo  "), "-")
    Output
      [1] "-  -"
    Code
      paste0("-", leading_space(" \t foo  "), "-")
    Output
      [1] "- \t -"
    Code
      paste0("-", leading_space(" foo  "), "-")
    Output
      [1] "- -"
    Code
      paste0("-", leading_space("   foo  "), "-")
    Output
      [1] "-   -"

# trailing_space

    Code
      paste0("-", trailing_space("foo"), "-")
    Output
      [1] "--"
    Code
      paste0("-", trailing_space("foo  "), "-")
    Output
      [1] "-  -"
    Code
      paste0("-", trailing_space("  foo  "), "-")
    Output
      [1] "-  -"
    Code
      paste0("-", trailing_space("  foo \t "), "-")
    Output
      [1] "- \t -"
    Code
      paste0("-", trailing_space("  foo "), "-")
    Output
      [1] "- -"
    Code
      paste0("-", trailing_space("   foo   "), "-")
    Output
      [1] "-   -"

# abbrev

    Code
      abbrev("123456789012345")
    Output
      <cli_ansi_string>
      [1] 1234567...
    Code
      abbrev("12345678901")
    Output
      <cli_ansi_string>
      [1] 1234567...
    Code
      abbrev("1234567890")
    Output
      <cli_ansi_string>
      [1] 1234567890
    Code
      abbrev("123456789")
    Output
      <cli_ansi_string>
      [1] 123456789
    Code
      abbrev("12345")
    Output
      <cli_ansi_string>
      [1] 12345
    Code
      abbrev("1")
    Output
      <cli_ansi_string>
      [1] 1
    Code
      abbrev("")
    Output
      <cli_ansi_string>
      [1] 
    Code
      abbrev("\033[31m1234567890\033[39m")
    Output
      <cli_ansi_string>
      [1] [31m1234567890[39m
    Code
      abbrev(c("\033[31m1234567890\033[39m", "", "1234567890123"), 5)
    Output
      <cli_ansi_string>
      [1] [31m12[39m...
      [2]                      
      [3] 12...                
    Code
      abbrev(rep("\033[31m1234567890\033[39m", 5), 5)
    Output
      <cli_ansi_string>
      [1] [31m12[39m...
      [2] [31m12[39m...
      [3] [31m12[39m...
      [4] [31m12[39m...
      [5] [31m12[39m...

# re_table

    Code
      tbl2
    Output
            start end   length                                
      text1 "19"  "23"  "5"    "\033[31m"                     
      text2 "27"  "31"  "5"    "\033[39m"                     
      text3 "48"  "52"  "5"    "\033[32m"                     
      text4 "58"  "62"  "5"    "\033[39m"                     
      text5 "74"  "98"  "25"   "\033]8;;https://example.com\a"
      text6 "103" "108" "6"    "\033]8;;\a"                   

---

    Code
      non_matching(list(tbl), txt)
    Output
      [[1]]
           start end length
      [1,]     1  18     18
      [2,]    24  26      3
      [3,]    32  47     16
      [4,]    53  57      5
      [5,]    63  73     11
      [6,]    99 102      4
      

# re_table special cases

    Code
      tbl
    Output
           start end length

---

    Code
      non_matching(list(tbl), txt)
    Output
      [[1]]
           start end length
      [1,]     1   6      6
      

---

    Code
      non_matching(list(tbl), txt, empty = TRUE)
    Output
      [[1]]
           start end length
      [1,]     1   6      6
      

---

    Code
      tbl
    Output
           start end length
      [1,]     1   5      5
      [2,]    12  16      5

---

    Code
      non_matching(list(tbl), txt)
    Output
      [[1]]
           start end length
      [1,]     6  11      6
      

---

    Code
      non_matching(list(tbl), txt, empty = TRUE)
    Output
      [[1]]
           start end length
      [1,]     1   0      0
      [2,]     6  11      6
      [3,]    17  16      0
      

---

    Code
      tbl
    Output
           start end length
      [1,]     5   9      5
      [2,]    10  14      5

---

    Code
      non_matching(list(tbl), txt)
    Output
      [[1]]
           start end length
      [1,]     1   4      4
      [2,]    15  18      4
      

---

    Code
      non_matching(list(tbl), txt, empty = TRUE)
    Output
      [[1]]
           start end length
      [1,]     1   4      4
      [2,]    10   9      0
      [3,]    15  18      4
      

# myseq

    Code
      myseq(1, 5)
    Output
      [1] 1 2 3 4 5
    Code
      myseq(1, 1)
    Output
      [1] 1
    Code
      myseq(1, 0)
    Output
      integer(0)
    Code
      myseq(1, 5, 2)
    Output
      [1] 1 3 5
    Code
      myseq(1, 6, 2)
    Output
      [1] 1 3 5
    Code
      myseq(1, 1, 2)
    Output
      [1] 1
    Code
      myseq(1, 2, -1)
    Output
      integer(0)
    Code
      myseq(10, 1, -1)
    Output
       [1] 10  9  8  7  6  5  4  3  2  1
    Code
      myseq(10, 1, -2)
    Output
      [1] 10  8  6  4  2
    Code
      myseq(1, 5, -2)
    Output
      integer(0)

# ansi_substr bad input

    Code
      ansi_substr("foobar", NULL, 10)
    Condition
      Error:
      ! `ansi_substr()` must have non-empty `start` and `stop` arguments
      i `start` has length 0

---

    Code
      ansi_substr("foobar", 10, NULL)
    Condition
      Error:
      ! `ansi_substr()` must have non-empty `start` and `stop` arguments
      i `stop` has length 0

---

    Code
      ansi_substr("foobar", "bad", "bad")
    Condition
      Error:
      ! `start` and `stop` must not have `NA` values
      i `start` has 1 `NA` value, after coercion to integer
      i `stop` has 1 `NA` value, after coercion to integer

# ansi_substr corner cases

    Code
      ansi_substr("abc", "hello", 1)
    Condition
      [1m[33mError[39m:[22m
      [33m![39m [1m[22m`start` and `stop` must not have `NA` values
      [36mi[39m `start` has 1 `NA` value, after coercion to integer

# Weird length 'split'

    Code
      ansi_strsplit(c("ab", "bd"), c("b", "d"))
    Condition
      [1m[33mError[39m:[22m
      [33m![39m [1m[22m`split` must be character of length <= 1, or must coerce to that
      [36mi[39m `split` is (or was coerced to) a character vector

# ansi_strtrim with zero-length ellipsis

    Code
      ansi_strtrim("12345", 1, ellipsis = "")
    Output
      <cli_ansi_string>
      [1] 1
    Code
      ansi_strtrim("12345", 3, ellipsis = "")
    Output
      <cli_ansi_string>
      [1] 123
    Code
      ansi_strtrim("12345", 5, ellipsis = "")
    Output
      <cli_ansi_string>
      [1] 12345

# ansi_strtrim errors

    Code
      ansi_strtrim("foobar", -1)
    Condition
      Error:
      ! `width` must be non-negative in `cli::ansi_strtrim()`.

# ansi_strtrim edge cases

    Code
      ansi_strtrim("foobar", width = 3, ellipsis = "...")
    Output
      [1] "..."
    Code
      ansi_strtrim("foobar", width = 2, ellipsis = "...")
    Output
      [1] ".."
    Code
      ansi_strtrim("foobar", width = 1, ellipsis = "...")
    Output
      [1] "."
    Code
      ansi_strtrim("foobar", width = 0, ellipsis = "...")
    Output
      [1] ""

# ansi_columns

    foo 1     foo 2     foo 3     foo 4     
    foo 5     foo 6     foo 7     foo 8     
    foo 9     foo 10                        

---

    123456789012...

# ansi_toupper [plain]

    Code
      local({
        cat_line(x)
        cat_line(ansi_toupper(x))
      })
    Output
      Red normal green
      RED NORMAL GREEN

# ansi_toupper [ansi]

    Code
      local({
        cat_line(x)
        cat_line(ansi_toupper(x))
      })
    Output
      [31mRed [39mnormal [1m[32mgreen[39m[22m
      [31mRED [39mNORMAL [1m[32mGREEN[39m[22m

# ansi_tolower [plain]

    Code
      local({
        cat_line(x)
        cat_line(ansi_tolower(x))
      })
    Output
      Red NORMAL grEeN
      red normal green

# ansi_tolower [ansi]

    Code
      local({
        cat_line(x)
        cat_line(ansi_tolower(x))
      })
    Output
      [31mRed [39mNORMAL [1m[32mgrEeN[39m[22m
      [31mred [39mnormal [1m[32mgreen[39m[22m

# ansi_chartr [plain]

    Code
      local({
        cat_line(x)
        cat_line(ansi_chartr(" R_", "-r*", x))
      })
    Output
      Red normal green
      red-normal-green

# ansi_chartr [ansi]

    Code
      local({
        cat_line(x)
        cat_line(ansi_chartr(" R_", "-r*", x))
      })
    Output
      [31mRed [39mnormal [1m[32mgreen[39m[22m
      [31mred-[39mnormal-[1m[32mgreen[39m[22m

# NA

    Code
      ansi_html(s)
    Output
      [1] "foo"                                         
      [2] NA                                            
      [3] "bar"                                         
      [4] "<span class=\"ansi ansi-bold\">foobar</span>"

# does not break

    Code
      local({
        withr::local_options(cli.width = 40)
        str30 <- "123456789 123456789 1234567890"
        cli_text(c(str30, "this is not breaking"))
      })
    Message
      123456789 123456789
      1234567890this is not breaking

# {.email} [plain-none]

    Code
      cli_text("{.email bugs.bunny@acme.com}")
    Message
      'bugs.bunny@acme.com'

# {.email} [fancy-none]

    Code
      cli_text("{.email bugs.bunny@acme.com}")
    Message
      [34mbugs.bunny@acme.com[39m

# {.email} [plain-all]

    Code
      cli_text("{.email bugs.bunny@acme.com}")
    Message
      ']8;;mailto:bugs.bunny@acme.combugs.bunny@acme.com]8;;'

# {.email} [fancy-all]

    Code
      cli_text("{.email bugs.bunny@acme.com}")
    Message
      ]8;;mailto:bugs.bunny@acme.com[34mbugs.bunny@acme.com[39m]8;;

# {.email} vectors [plain-none]

    Code
      emails <- paste0("bugs.bunny-", 1:3, "@acme.com")
      cli_text("{.email {emails}}")
    Message
      'bugs.bunny-1@acme.com', 'bugs.bunny-2@acme.com', and 'bugs.bunny-3@acme.com'

# {.email} vectors [fancy-none]

    Code
      emails <- paste0("bugs.bunny-", 1:3, "@acme.com")
      cli_text("{.email {emails}}")
    Message
      [34mbugs.bunny-1@acme.com[39m, [34mbugs.bunny-2@acme.com[39m, and [34mbugs.bunny-3@acme.com[39m

# {.email} vectors [plain-all]

    Code
      emails <- paste0("bugs.bunny-", 1:3, "@acme.com")
      cli_text("{.email {emails}}")
    Message
      ']8;;mailto:bugs.bunny-1@acme.combugs.bunny-1@acme.com]8;;', ']8;;mailto:bugs.bunny-2@acme.combugs.bunny-2@acme.com]8;;', and ']8;;mailto:bugs.bunny-3@acme.combugs.bunny-3@acme.com]8;;'

# {.email} vectors [fancy-all]

    Code
      emails <- paste0("bugs.bunny-", 1:3, "@acme.com")
      cli_text("{.email {emails}}")
    Message
      ]8;;mailto:bugs.bunny-1@acme.com[34mbugs.bunny-1@acme.com[39m]8;;, ]8;;mailto:bugs.bunny-2@acme.com[34mbugs.bunny-2@acme.com[39m]8;;, and ]8;;mailto:bugs.bunny-3@acme.com[34mbugs.bunny-3@acme.com[39m]8;;

# {.file} and {.path} [plain-none]

    Code
      cli_text("{.file /absolute/path}")
    Message
      '/absolute/path'
    Code
      cli_text("{.file file:///absolute/path}")
    Message
      'file:///absolute/path'
    Code
      cli_text("{.path /absolute/path}")
    Message
      '/absolute/path'
    Code
      cli_text("{.path file:///absolute/path}")
    Message
      'file:///absolute/path'

---

    Code
      cli_text("{.file relative/path}")
    Message
      'relative/path'
    Code
      cli_text("{.file ./relative/path}")
    Message
      './relative/path'
    Code
      cli_text("{.path relative/path}")
    Message
      'relative/path'
    Code
      cli_text("{.path ./relative/path}")
    Message
      './relative/path'

---

    Code
      cli_text("{.file ~/relative/path}")
    Message
      '~/relative/path'
    Code
      cli_text("{.path ~/relative/path}")
    Message
      '~/relative/path'

---

    Code
      paths <- c("~/foo", "bar", "file:///abs")
      cli_text("{.file {paths}}")
    Message
      '~/foo', 'bar', and 'file:///abs'

---

    Code
      paths <- c("foo  ", " bar ", "file:///a bs ")
      cli_text("{.file {paths}}")
    Message
      'foo  ', ' bar ', and 'file:///a bs '

---

    Code
      name <- cli::style_hyperlink("/foo/bar", "/foo/bar")
      cli_text("{.file {name}}")
    Message
      '/foo/bar'

---

    Code
      cli_text("{.file /absolute/path:12}")
    Message
      '/absolute/path:12'
    Code
      cli_text("{.file file:///absolute/path:5}")
    Message
      'file:///absolute/path:5'
    Code
      cli_text("{.path /absolute/path:123}")
    Message
      '/absolute/path:123'
    Code
      cli_text("{.path file:///absolute/path:51}")
    Message
      'file:///absolute/path:51'

---

    Code
      cli_text("{.file relative/path:12}")
    Message
      'relative/path:12'
    Code
      cli_text("{.file ./relative/path:5}")
    Message
      './relative/path:5'
    Code
      cli_text("{.path relative/path:123}")
    Message
      'relative/path:123'
    Code
      cli_text("{.path ./relative/path:51}")
    Message
      './relative/path:51'

---

    Code
      cli_text("{.file ~/relative/path:12}")
    Message
      '~/relative/path:12'
    Code
      cli_text("{.path ~/relative/path:5}")
    Message
      '~/relative/path:5'

---

    Code
      cli_text("{.file /absolute/path:12:5}")
    Message
      '/absolute/path:12:5'
    Code
      cli_text("{.file file:///absolute/path:5:100}")
    Message
      'file:///absolute/path:5:100'
    Code
      cli_text("{.path /absolute/path:123:1}")
    Message
      '/absolute/path:123:1'
    Code
      cli_text("{.path file:///absolute/path:51:6}")
    Message
      'file:///absolute/path:51:6'

---

    Code
      cli_text("{.file relative/path:12:13}")
    Message
      'relative/path:12:13'
    Code
      cli_text("{.file ./relative/path:5:20}")
    Message
      './relative/path:5:20'
    Code
      cli_text("{.path relative/path:123:21}")
    Message
      'relative/path:123:21'
    Code
      cli_text("{.path ./relative/path:51:2}")
    Message
      './relative/path:51:2'

---

    Code
      cli_text("{.file ~/relative/path:12:23}")
    Message
      '~/relative/path:12:23'
    Code
      cli_text("{.path ~/relative/path:5:2}")
    Message
      '~/relative/path:5:2'

---

    Code
      paths <- c("~/foo", "bar:10", "file:///abs:10:20")
      cli_text("{.file {paths}}")
    Message
      '~/foo', 'bar:10', and 'file:///abs:10:20'

# {.file} and {.path} [fancy-none]

    Code
      cli_text("{.file /absolute/path}")
    Message
      [34m/absolute/path[39m
    Code
      cli_text("{.file file:///absolute/path}")
    Message
      [34mfile:///absolute/path[39m
    Code
      cli_text("{.path /absolute/path}")
    Message
      [34m/absolute/path[39m
    Code
      cli_text("{.path file:///absolute/path}")
    Message
      [34mfile:///absolute/path[39m

---

    Code
      cli_text("{.file relative/path}")
    Message
      [34mrelative/path[39m
    Code
      cli_text("{.file ./relative/path}")
    Message
      [34m./relative/path[39m
    Code
      cli_text("{.path relative/path}")
    Message
      [34mrelative/path[39m
    Code
      cli_text("{.path ./relative/path}")
    Message
      [34m./relative/path[39m

---

    Code
      cli_text("{.file ~/relative/path}")
    Message
      [34m~/relative/path[39m
    Code
      cli_text("{.path ~/relative/path}")
    Message
      [34m~/relative/path[39m

---

    Code
      paths <- c("~/foo", "bar", "file:///abs")
      cli_text("{.file {paths}}")
    Message
      [34m~/foo[39m, [34mbar[39m, and [34mfile:///abs[39m

---

    Code
      paths <- c("foo  ", " bar ", "file:///a bs ")
      cli_text("{.file {paths}}")
    Message
      '[34mfoo[39m[44m  [49m', '[44m [49m[34mbar[39m[44m [49m', and '[34mfile:///a bs[39m[44m [49m'

---

    Code
      name <- cli::style_hyperlink("/foo/bar", "/foo/bar")
      cli_text("{.file {name}}")
    Message
      [34m/foo/bar[39m

---

    Code
      cli_text("{.file /absolute/path:12}")
    Message
      [34m/absolute/path:12[39m
    Code
      cli_text("{.file file:///absolute/path:5}")
    Message
      [34mfile:///absolute/path:5[39m
    Code
      cli_text("{.path /absolute/path:123}")
    Message
      [34m/absolute/path:123[39m
    Code
      cli_text("{.path file:///absolute/path:51}")
    Message
      [34mfile:///absolute/path:51[39m

---

    Code
      cli_text("{.file relative/path:12}")
    Message
      [34mrelative/path:12[39m
    Code
      cli_text("{.file ./relative/path:5}")
    Message
      [34m./relative/path:5[39m
    Code
      cli_text("{.path relative/path:123}")
    Message
      [34mrelative/path:123[39m
    Code
      cli_text("{.path ./relative/path:51}")
    Message
      [34m./relative/path:51[39m

---

    Code
      cli_text("{.file ~/relative/path:12}")
    Message
      [34m~/relative/path:12[39m
    Code
      cli_text("{.path ~/relative/path:5}")
    Message
      [34m~/relative/path:5[39m

---

    Code
      cli_text("{.file /absolute/path:12:5}")
    Message
      [34m/absolute/path:12:5[39m
    Code
      cli_text("{.file file:///absolute/path:5:100}")
    Message
      [34mfile:///absolute/path:5:100[39m
    Code
      cli_text("{.path /absolute/path:123:1}")
    Message
      [34m/absolute/path:123:1[39m
    Code
      cli_text("{.path file:///absolute/path:51:6}")
    Message
      [34mfile:///absolute/path:51:6[39m

---

    Code
      cli_text("{.file relative/path:12:13}")
    Message
      [34mrelative/path:12:13[39m
    Code
      cli_text("{.file ./relative/path:5:20}")
    Message
      [34m./relative/path:5:20[39m
    Code
      cli_text("{.path relative/path:123:21}")
    Message
      [34mrelative/path:123:21[39m
    Code
      cli_text("{.path ./relative/path:51:2}")
    Message
      [34m./relative/path:51:2[39m

---

    Code
      cli_text("{.file ~/relative/path:12:23}")
    Message
      [34m~/relative/path:12:23[39m
    Code
      cli_text("{.path ~/relative/path:5:2}")
    Message
      [34m~/relative/path:5:2[39m

---

    Code
      paths <- c("~/foo", "bar:10", "file:///abs:10:20")
      cli_text("{.file {paths}}")
    Message
      [34m~/foo[39m, [34mbar:10[39m, and [34mfile:///abs:10:20[39m

# {.file} and {.path} [plain-all]

    Code
      cli_text("{.file /absolute/path}")
    Message
      ']8;;file:///absolute/path/absolute/path]8;;'
    Code
      cli_text("{.file file:///absolute/path}")
    Message
      ']8;;file:///absolute/pathfile:///absolute/path]8;;'
    Code
      cli_text("{.path /absolute/path}")
    Message
      ']8;;file:///absolute/path/absolute/path]8;;'
    Code
      cli_text("{.path file:///absolute/path}")
    Message
      ']8;;file:///absolute/pathfile:///absolute/path]8;;'

---

    Code
      cli_text("{.file relative/path}")
    Message
      ']8;;file:///testthat/home/relative/pathrelative/path]8;;'
    Code
      cli_text("{.file ./relative/path}")
    Message
      ']8;;file:///testthat/home/./relative/path./relative/path]8;;'
    Code
      cli_text("{.path relative/path}")
    Message
      ']8;;file:///testthat/home/relative/pathrelative/path]8;;'
    Code
      cli_text("{.path ./relative/path}")
    Message
      ']8;;file:///testthat/home/./relative/path./relative/path]8;;'

---

    Code
      cli_text("{.file ~/relative/path}")
    Message
      ']8;;file:///my/home/relative/path~/relative/path]8;;'
    Code
      cli_text("{.path ~/relative/path}")
    Message
      ']8;;file:///my/home/relative/path~/relative/path]8;;'

---

    Code
      paths <- c("~/foo", "bar", "file:///abs")
      cli_text("{.file {paths}}")
    Message
      ']8;;file:///my/home/foo~/foo]8;;', ']8;;file:///testthat/home/barbar]8;;', and ']8;;file:///absfile:///abs]8;;'

---

    Code
      paths <- c("foo  ", " bar ", "file:///a bs ")
      cli_text("{.file {paths}}")
    Message
      ']8;;file:///testthat/home/foo  foo]8;;  ', ' ]8;;file:///testthat/home/ bar bar]8;; ', and ']8;;file:///a bs file:///a bs]8;; '

---

    Code
      name <- cli::style_hyperlink("/foo/bar", "/foo/bar")
      cli_text("{.file {name}}")
    Message
      ']8;;/foo/bar/foo/bar]8;;'

---

    Code
      cli_text("{.file /absolute/path:12}")
    Message
      ']8;line = 12:col = 1;file:///absolute/path/absolute/path:12]8;;'
    Code
      cli_text("{.file file:///absolute/path:5}")
    Message
      ']8;line = 5:col = 1;file:///absolute/pathfile:///absolute/path:5]8;;'
    Code
      cli_text("{.path /absolute/path:123}")
    Message
      ']8;line = 123:col = 1;file:///absolute/path/absolute/path:123]8;;'
    Code
      cli_text("{.path file:///absolute/path:51}")
    Message
      ']8;line = 51:col = 1;file:///absolute/pathfile:///absolute/path:51]8;;'

---

    Code
      cli_text("{.file relative/path:12}")
    Message
      ']8;line = 12:col = 1;file:///testthat/home/relative/pathrelative/path:12]8;;'
    Code
      cli_text("{.file ./relative/path:5}")
    Message
      ']8;line = 5:col = 1;file:///testthat/home/./relative/path./relative/path:5]8;;'
    Code
      cli_text("{.path relative/path:123}")
    Message
      ']8;line = 123:col = 1;file:///testthat/home/relative/pathrelative/path:123]8;;'
    Code
      cli_text("{.path ./relative/path:51}")
    Message
      ']8;line = 51:col = 1;file:///testthat/home/./relative/path./relative/path:51]8;;'

---

    Code
      cli_text("{.file ~/relative/path:12}")
    Message
      ']8;line = 12:col = 1;file:///my/home/relative/path~/relative/path:12]8;;'
    Code
      cli_text("{.path ~/relative/path:5}")
    Message
      ']8;line = 5:col = 1;file:///my/home/relative/path~/relative/path:5]8;;'

---

    Code
      cli_text("{.file /absolute/path:12:5}")
    Message
      ']8;line = 12:col = 5;file:///absolute/path/absolute/path:12:5]8;;'
    Code
      cli_text("{.file file:///absolute/path:5:100}")
    Message
      ']8;line = 5:col = 100;file:///absolute/pathfile:///absolute/path:5:100]8;;'
    Code
      cli_text("{.path /absolute/path:123:1}")
    Message
      ']8;line = 123:col = 1;file:///absolute/path/absolute/path:123:1]8;;'
    Code
      cli_text("{.path file:///absolute/path:51:6}")
    Message
      ']8;line = 51:col = 6;file:///absolute/pathfile:///absolute/path:51:6]8;;'

---

    Code
      cli_text("{.file relative/path:12:13}")
    Message
      ']8;line = 12:col = 13;file:///testthat/home/relative/pathrelative/path:12:13]8;;'
    Code
      cli_text("{.file ./relative/path:5:20}")
    Message
      ']8;line = 5:col = 20;file:///testthat/home/./relative/path./relative/path:5:20]8;;'
    Code
      cli_text("{.path relative/path:123:21}")
    Message
      ']8;line = 123:col = 21;file:///testthat/home/relative/pathrelative/path:123:21]8;;'
    Code
      cli_text("{.path ./relative/path:51:2}")
    Message
      ']8;line = 51:col = 2;file:///testthat/home/./relative/path./relative/path:51:2]8;;'

---

    Code
      cli_text("{.file ~/relative/path:12:23}")
    Message
      ']8;line = 12:col = 23;file:///my/home/relative/path~/relative/path:12:23]8;;'
    Code
      cli_text("{.path ~/relative/path:5:2}")
    Message
      ']8;line = 5:col = 2;file:///my/home/relative/path~/relative/path:5:2]8;;'

---

    Code
      paths <- c("~/foo", "bar:10", "file:///abs:10:20")
      cli_text("{.file {paths}}")
    Message
      ']8;;file:///my/home/foo~/foo]8;;', ']8;line = 10:col = 1;file:///testthat/home/barbar:10]8;;', and ']8;line = 10:col = 20;file:///absfile:///abs:10:20]8;;'

# {.file} and {.path} [fancy-all]

    Code
      cli_text("{.file /absolute/path}")
    Message
      ]8;;file:///absolute/path[34m/absolute/path[39m]8;;
    Code
      cli_text("{.file file:///absolute/path}")
    Message
      ]8;;file:///absolute/path[34mfile:///absolute/path[39m]8;;
    Code
      cli_text("{.path /absolute/path}")
    Message
      ]8;;file:///absolute/path[34m/absolute/path[39m]8;;
    Code
      cli_text("{.path file:///absolute/path}")
    Message
      ]8;;file:///absolute/path[34mfile:///absolute/path[39m]8;;

---

    Code
      cli_text("{.file relative/path}")
    Message
      ]8;;file:///testthat/home/relative/path[34mrelative/path[39m]8;;
    Code
      cli_text("{.file ./relative/path}")
    Message
      ]8;;file:///testthat/home/./relative/path[34m./relative/path[39m]8;;
    Code
      cli_text("{.path relative/path}")
    Message
      ]8;;file:///testthat/home/relative/path[34mrelative/path[39m]8;;
    Code
      cli_text("{.path ./relative/path}")
    Message
      ]8;;file:///testthat/home/./relative/path[34m./relative/path[39m]8;;

---

    Code
      cli_text("{.file ~/relative/path}")
    Message
      ]8;;file:///my/home/relative/path[34m~/relative/path[39m]8;;
    Code
      cli_text("{.path ~/relative/path}")
    Message
      ]8;;file:///my/home/relative/path[34m~/relative/path[39m]8;;

---

    Code
      paths <- c("~/foo", "bar", "file:///abs")
      cli_text("{.file {paths}}")
    Message
      ]8;;file:///my/home/foo[34m~/foo[39m]8;;, ]8;;file:///testthat/home/bar[34mbar[39m]8;;, and ]8;;file:///abs[34mfile:///abs[39m]8;;

---

    Code
      paths <- c("foo  ", " bar ", "file:///a bs ")
      cli_text("{.file {paths}}")
    Message
      ']8;;file:///testthat/home/foo  [34mfoo[39m]8;;[44m  [49m', '[44m [49m]8;;file:///testthat/home/ bar [34mbar[39m]8;;[44m [49m', and ']8;;file:///a bs [34mfile:///a bs[39m]8;;[44m [49m'

---

    Code
      name <- cli::style_hyperlink("/foo/bar", "/foo/bar")
      cli_text("{.file {name}}")
    Message
      ]8;;/foo/bar[34m/foo/bar[39m]8;;

---

    Code
      cli_text("{.file /absolute/path:12}")
    Message
      ]8;line = 12:col = 1;file:///absolute/path[34m/absolute/path:12[39m]8;;
    Code
      cli_text("{.file file:///absolute/path:5}")
    Message
      ]8;line = 5:col = 1;file:///absolute/path[34mfile:///absolute/path:5[39m]8;;
    Code
      cli_text("{.path /absolute/path:123}")
    Message
      ]8;line = 123:col = 1;file:///absolute/path[34m/absolute/path:123[39m]8;;
    Code
      cli_text("{.path file:///absolute/path:51}")
    Message
      ]8;line = 51:col = 1;file:///absolute/path[34mfile:///absolute/path:51[39m]8;;

---

    Code
      cli_text("{.file relative/path:12}")
    Message
      ]8;line = 12:col = 1;file:///testthat/home/relative/path[34mrelative/path:12[39m]8;;
    Code
      cli_text("{.file ./relative/path:5}")
    Message
      ]8;line = 5:col = 1;file:///testthat/home/./relative/path[34m./relative/path:5[39m]8;;
    Code
      cli_text("{.path relative/path:123}")
    Message
      ]8;line = 123:col = 1;file:///testthat/home/relative/path[34mrelative/path:123[39m]8;;
    Code
      cli_text("{.path ./relative/path:51}")
    Message
      ]8;line = 51:col = 1;file:///testthat/home/./relative/path[34m./relative/path:51[39m]8;;

---

    Code
      cli_text("{.file ~/relative/path:12}")
    Message
      ]8;line = 12:col = 1;file:///my/home/relative/path[34m~/relative/path:12[39m]8;;
    Code
      cli_text("{.path ~/relative/path:5}")
    Message
      ]8;line = 5:col = 1;file:///my/home/relative/path[34m~/relative/path:5[39m]8;;

---

    Code
      cli_text("{.file /absolute/path:12:5}")
    Message
      ]8;line = 12:col = 5;file:///absolute/path[34m/absolute/path:12:5[39m]8;;
    Code
      cli_text("{.file file:///absolute/path:5:100}")
    Message
      ]8;line = 5:col = 100;file:///absolute/path[34mfile:///absolute/path:5:100[39m]8;;
    Code
      cli_text("{.path /absolute/path:123:1}")
    Message
      ]8;line = 123:col = 1;file:///absolute/path[34m/absolute/path:123:1[39m]8;;
    Code
      cli_text("{.path file:///absolute/path:51:6}")
    Message
      ]8;line = 51:col = 6;file:///absolute/path[34mfile:///absolute/path:51:6[39m]8;;

---

    Code
      cli_text("{.file relative/path:12:13}")
    Message
      ]8;line = 12:col = 13;file:///testthat/home/relative/path[34mrelative/path:12:13[39m]8;;
    Code
      cli_text("{.file ./relative/path:5:20}")
    Message
      ]8;line = 5:col = 20;file:///testthat/home/./relative/path[34m./relative/path:5:20[39m]8;;
    Code
      cli_text("{.path relative/path:123:21}")
    Message
      ]8;line = 123:col = 21;file:///testthat/home/relative/path[34mrelative/path:123:21[39m]8;;
    Code
      cli_text("{.path ./relative/path:51:2}")
    Message
      ]8;line = 51:col = 2;file:///testthat/home/./relative/path[34m./relative/path:51:2[39m]8;;

---

    Code
      cli_text("{.file ~/relative/path:12:23}")
    Message
      ]8;line = 12:col = 23;file:///my/home/relative/path[34m~/relative/path:12:23[39m]8;;
    Code
      cli_text("{.path ~/relative/path:5:2}")
    Message
      ]8;line = 5:col = 2;file:///my/home/relative/path[34m~/relative/path:5:2[39m]8;;

---

    Code
      paths <- c("~/foo", "bar:10", "file:///abs:10:20")
      cli_text("{.file {paths}}")
    Message
      ]8;;file:///my/home/foo[34m~/foo[39m]8;;, ]8;line = 10:col = 1;file:///testthat/home/bar[34mbar:10[39m]8;;, and ]8;line = 10:col = 20;file:///abs[34mfile:///abs:10:20[39m]8;;

# {.fun} [plain-none]

    Code
      cli_text("{.fun myfun}")
    Message
      `myfun()`
    Code
      cli_text("{.fun mypackage::myfun}")
    Message
      `mypackage::myfun()`

---

    Code
      funs <- paste0("mypkg::myfun", 1:3)
      cli_text("{.fun {funs}}")
    Message
      `mypkg::myfun1()`, `mypkg::myfun2()`, and `mypkg::myfun3()`

# {.fun} [plain-all]

    Code
      cli_text("{.fun myfun}")
    Message
      `myfun()`
    Code
      cli_text("{.fun mypackage::myfun}")
    Message
      `]8;;x-r-help:mypackage::myfunmypackage::myfun]8;;()`

---

    Code
      funs <- paste0("mypkg::myfun", 1:3)
      cli_text("{.fun {funs}}")
    Message
      `]8;;x-r-help:mypkg::myfun1mypkg::myfun1]8;;()`, `]8;;x-r-help:mypkg::myfun2mypkg::myfun2]8;;()`, and `]8;;x-r-help:mypkg::myfun3mypkg::myfun3]8;;()`

# turning off help [plain-all]

    Code
      cli_text("{.fun pkg::func}")
    Message
      `pkg::func()`

# .fun with custom format [plain-all]

    Code
      cli_text("{.fun pkg::func}")
    Message
      `]8;;aaa-pkg::func-zzzpkg::func]8;;()`

# {.help} [plain-none]

    Code
      cli_text("{.help pkg::fun}")
    Message
      `?pkg::fun()`
    Code
      cli_text("{.help [link text](pkg::fun)}")
    Message
      link text (`?pkg::fun()`)

---

    Code
      funcs <- paste0("pkg::fun", 1:3)
      cli_text("{.help {funcs}}")
    Message
      `?pkg::fun1()`, `?pkg::fun2()`, and `?pkg::fun3()`

# {.help} [plain-all]

    Code
      cli_text("{.help pkg::fun}")
    Message
      ]8;;x-r-help:pkg::funpkg::fun]8;;
    Code
      cli_text("{.help [link text](pkg::fun)}")
    Message
      ]8;;x-r-help:pkg::funlink text]8;;

---

    Code
      funcs <- paste0("pkg::fun", 1:3)
      cli_text("{.help {funcs}}")
    Message
      ]8;;x-r-help:pkg::fun1pkg::fun1]8;;, ]8;;x-r-help:pkg::fun2pkg::fun2]8;;, and ]8;;x-r-help:pkg::fun3pkg::fun3]8;;

# .help with custom format [plain-all]

    Code
      cli_text("{.help pkg::fun}")
    Message
      ]8;;aaa-pkg::fun-zzzpkg::fun]8;;

# {.href} [plain-none]

    Code
      cli_text("{.href https://cli.r-lib.org}")
    Message
      <https://cli.r-lib.org>
    Code
      cli_text("{.href [linktext](https://cli.r-lib.org)}")
    Message
      linktext (<https://cli.r-lib.org>)
    Code
      cli_text("{.href [link text](https://cli.r-lib.org)}")
    Message
      link text (<https://cli.r-lib.org>)

# {.href} [plain-all]

    Code
      cli_text("{.href https://cli.r-lib.org}")
    Message
      <]8;;https://cli.r-lib.orghttps://cli.r-lib.org]8;;>
    Code
      cli_text("{.href [linktext](https://cli.r-lib.org)}")
    Message
      ]8;;https://cli.r-lib.orglinktext]8;;
    Code
      cli_text("{.href [link text](https://cli.r-lib.org)}")
    Message
      ]8;;https://cli.r-lib.orglink text]8;;

# {.href} vectors [plain-none]

    Code
      url <- paste0("https://cli.r-lib.org/", 1:3)
      cli_text("{.href {url}}")
    Message
      <https://cli.r-lib.org/1>, <https://cli.r-lib.org/2>, and
      <https://cli.r-lib.org/3>

# {.href} vectors [plain-all]

    Code
      url <- paste0("https://cli.r-lib.org/", 1:3)
      cli_text("{.href {url}}")
    Message
      <]8;;https://cli.r-lib.org/1https://cli.r-lib.org/1]8;;>, <]8;;https://cli.r-lib.org/2https://cli.r-lib.org/2]8;;>, and
      <]8;;https://cli.r-lib.org/3https://cli.r-lib.org/3]8;;>

# {.run} [plain-none]

    Code
      cli_text("{.run pkg::fun(param)}")
    Message
      `pkg::fun(param)`
    Code
      cli_text("{.run [run(p1, p2)](pkg::fun(p1, p2, other = 'foo'))}")
    Message
      `run(p1, p2)`

# {.run} [plain-all]

    Code
      cli_text("{.run pkg::fun(param)}")
    Message
      ]8;;x-r-run:pkg::fun(param)pkg::fun(param)]8;;
    Code
      cli_text("{.run [run(p1, p2)](pkg::fun(p1, p2, other = 'foo'))}")
    Message
      ]8;;x-r-run:pkg::fun(p1, p2, other = 'foo')run(p1, p2)]8;;

# {.run} vectors [plain-none]

    Code
      codes <- paste0("pkg::fun", 1:3, "()")
      cli_text("{.run {codes}}")
    Message
      `pkg::fun1()`, `pkg::fun2()`, and `pkg::fun3()`

# {.run} vectors [plain-all]

    Code
      codes <- paste0("pkg::fun", 1:3, "()")
      cli_text("{.run {codes}}")
    Message
      ]8;;x-r-run:pkg::fun1()pkg::fun1()]8;;, ]8;;x-r-run:pkg::fun2()pkg::fun2()]8;;, and ]8;;x-r-run:pkg::fun3()pkg::fun3()]8;;

# .run with custom format [plain-all]

    Code
      cli_text("{.run devtools::document()}")
    Message
      ]8;;aaa-devtools::document()-zzzdevtools::document()]8;;

# {.topic} [plain-none]

    Code
      cli_text("{.topic pkg::topic}")
    Message
      `?pkg::topic`
    Code
      cli_text("{.topic [link text](pkg::topic)}")
    Message
      link text (`?pkg::topic`)

---

    Code
      topics <- paste0("pkg::topic", 1:3)
      cli_text("{.topic {topics}}")
    Message
      `?pkg::topic1`, `?pkg::topic2`, and `?pkg::topic3`

# {.topic} [plain-all]

    Code
      cli_text("{.topic pkg::topic}")
    Message
      ]8;;x-r-help:pkg::topicpkg::topic]8;;
    Code
      cli_text("{.topic [link text](pkg::topic)}")
    Message
      ]8;;x-r-help:pkg::topiclink text]8;;

---

    Code
      topics <- paste0("pkg::topic", 1:3)
      cli_text("{.topic {topics}}")
    Message
      ]8;;x-r-help:pkg::topic1pkg::topic1]8;;, ]8;;x-r-help:pkg::topic2pkg::topic2]8;;, and ]8;;x-r-help:pkg::topic3pkg::topic3]8;;

# .topic with custom format [plain-all]

    Code
      cli_text("{.topic pkg::fun}")
    Message
      ]8;;aaa-pkg::fun-zzzpkg::fun]8;;

# {.url} [plain-none]

    Code
      cli_text("{.url https://cli.r-lib.org}")
    Message
      <https://cli.r-lib.org>

# {.url} [fancy-none]

    Code
      cli_text("{.url https://cli.r-lib.org}")
    Message
      [3m[34m<https://cli.r-lib.org>[39m[23m

# {.url} [plain-all]

    Code
      cli_text("{.url https://cli.r-lib.org}")
    Message
      <]8;;https://cli.r-lib.orghttps://cli.r-lib.org]8;;>

# {.url} [fancy-all]

    Code
      cli_text("{.url https://cli.r-lib.org}")
    Message
      [3m[34m<]8;;https://cli.r-lib.orghttps://cli.r-lib.org]8;;>[39m[23m

# {.url} vector [plain-none]

    Code
      urls <- paste0("https://cli.r-lib.org/", 1:3)
      cli_text("{.url {urls}}")
    Message
      <https://cli.r-lib.org/1>, <https://cli.r-lib.org/2>, and
      <https://cli.r-lib.org/3>

# {.url} vector [fancy-none]

    Code
      urls <- paste0("https://cli.r-lib.org/", 1:3)
      cli_text("{.url {urls}}")
    Message
      [3m[34m<https://cli.r-lib.org/1>[39m[23m, [3m[34m<https://cli.r-lib.org/2>[39m[23m, and
      [3m[34m<https://cli.r-lib.org/3>[39m[23m

# {.url} vector [plain-all]

    Code
      urls <- paste0("https://cli.r-lib.org/", 1:3)
      cli_text("{.url {urls}}")
    Message
      <]8;;https://cli.r-lib.org/1https://cli.r-lib.org/1]8;;>, <]8;;https://cli.r-lib.org/2https://cli.r-lib.org/2]8;;>, and
      <]8;;https://cli.r-lib.org/3https://cli.r-lib.org/3]8;;>

# {.url} vector [fancy-all]

    Code
      urls <- paste0("https://cli.r-lib.org/", 1:3)
      cli_text("{.url {urls}}")
    Message
      [3m[34m<]8;;https://cli.r-lib.org/1https://cli.r-lib.org/1]8;;>[39m[23m, [3m[34m<]8;;https://cli.r-lib.org/2https://cli.r-lib.org/2]8;;>[39m[23m, and
      [3m[34m<]8;;https://cli.r-lib.org/3https://cli.r-lib.org/3]8;;>[39m[23m

# linked {.url} [plain-all]

    Code
      link <- c("https://cli.r-lib.org", style_hyperlink("text",
        "https://cli.r-lib.org"))
      cli_text("{.url {link}}")
    Message
      <]8;;https://cli.r-lib.orghttps://cli.r-lib.org]8;;> and <]8;;https://cli.r-lib.orgtext]8;;>

# {.vignette} [plain-none]

    Code
      cli_text("{.vignette pkg::name}")
    Message
      `vignette(pkg::name)`
    Code
      cli_text("{.vignette [link text](pkg::name)}")
    Message
      link text (`vignette(pkg::name)`)

---

    Code
      vignettes <- paste0("pkg::topic", 1:3)
      cli_text("{.vignette {vignettes}}")
    Message
      `vignette(pkg::topic1)`, `vignette(pkg::topic2)`, and `vignette(pkg::topic3)`

# {.vignette} [plain-all]

    Code
      cli_text("{.vignette pkg::name}")
    Message
      ]8;;x-r-vignette:pkg::namepkg::name]8;;
    Code
      cli_text("{.vignette [link text](pkg::name)}")
    Message
      ]8;;x-r-vignette:pkg::namelink text]8;;

---

    Code
      vignettes <- paste0("pkg::topic", 1:3)
      cli_text("{.vignette {vignettes}}")
    Message
      ]8;;x-r-vignette:pkg::topic1pkg::topic1]8;;, ]8;;x-r-vignette:pkg::topic2pkg::topic2]8;;, and ]8;;x-r-vignette:pkg::topic3pkg::topic3]8;;

# .vignette with custom format [plain-all]

    Code
      cli_text("{.vignette pkgdown::accessibility}")
    Message
      ]8;;aaa-pkgdown::accessibility-zzzpkgdown::accessibility]8;;

# print.cli_ansi_style

    Code
      print(col_red)
    Output
      <cli_ansi_style>
      Example output

# print.cli_ansi_string

    Code
      print(col_red("red"))
    Output
      <cli_ansi_string>
      [1] [31mred[39m

# ansi-scale

    Code
      ansi_scale(c(0, 0, 0))
    Output
      [1] 0 0 0
    Code
      ansi_scale(c(255, 100, 0))
    Output
      [1] 5 2 0
    Code
      ansi_scale(c(255, 100, 0), round = FALSE)
    Output
      [1] 5.000000 1.960784 0.000000

# reserved [plain]

    Code
      cat(code_highlight("function () { }", list(reserved = "bold")))
    Output
      function () { }
    Code
      cat(code_highlight("if (1) NULL else NULL", list(reserved = "bold")))
    Output
      if (1) NULL else NULL
    Code
      cat(code_highlight("repeat {}", list(reserved = "bold")))
    Output
      repeat {}
    Code
      cat(code_highlight("while (1) {}", list(reserved = "bold")))
    Output
      while (1) {}
    Code
      cat(code_highlight("for (i in x) next", list(reserved = "bold")))
    Output
      for (i in x) next
    Code
      cat(code_highlight("for (i in x) break", list(reserved = "bold")))
    Output
      for (i in x) break

# reserved [ansi]

    Code
      cat(code_highlight("function () { }", list(reserved = "bold")))
    Output
      [1mfunction[22m () { }
    Code
      cat(code_highlight("if (1) NULL else NULL", list(reserved = "bold")))
    Output
      [1mif[22m (1) NULL [1melse[22m NULL
    Code
      cat(code_highlight("repeat {}", list(reserved = "bold")))
    Output
      [1mrepeat[22m {}
    Code
      cat(code_highlight("while (1) {}", list(reserved = "bold")))
    Output
      [1mwhile[22m (1) {}
    Code
      cat(code_highlight("for (i in x) next", list(reserved = "bold")))
    Output
      [1mfor[22m (i [1min[22m x) [1mnext[22m
    Code
      cat(code_highlight("for (i in x) break", list(reserved = "bold")))
    Output
      [1mfor[22m (i [1min[22m x) [1mbreak[22m

# number [plain]

    Code
      cat(code_highlight("1 + 1.0 + -1 + 2L + Inf", list(number = "bold")))
    Output
      1 + 1.0 + -1 + 2L + Inf
    Code
      cat(code_highlight("NA + NA_real_ + NA_integer_ + NA_character_", list(number = "bold")))
    Output
      NA + NA_real_ + NA_integer_ + NA_character_
    Code
      cat(code_highlight("TRUE + FALSE", list(number = "bold")))
    Output
      TRUE + FALSE

# number [ansi]

    Code
      cat(code_highlight("1 + 1.0 + -1 + 2L + Inf", list(number = "bold")))
    Output
      [1m1[22m + [1m1.0[22m + -[1m1[22m + [1m2L[22m + [1mInf[22m
    Code
      cat(code_highlight("NA + NA_real_ + NA_integer_ + NA_character_", list(number = "bold")))
    Output
      [1mNA[22m + [1mNA_real_[22m + [1mNA_integer_[22m + [1mNA_character_[22m
    Code
      cat(code_highlight("TRUE + FALSE", list(number = "bold")))
    Output
      [1mTRUE[22m + [1mFALSE[22m

# null [plain]

    Code
      cat(code_highlight("NULL", list(null = "bold")))
    Output
      NULL

# null [ansi]

    Code
      cat(code_highlight("NULL", list(null = "bold")))
    Output
      [1mNULL[22m

# operator [plain]

    Code
      cat(code_highlight("~ ! 1 - 2 + 3:4 * 5 / 6 ^ 7", list(operator = "bold")))
    Output
      ~ ! 1 - 2 + 3:4 * 5 / 6 ^ 7
    Code
      cat(code_highlight("? 1 %% 2 %+% 2 < 3 & 4 > 5 && 6 == 7 | 8 <= 9 || 10 >= 11",
        list(operator = "bold")))
    Output
      ? 1 %% 2 %+% 2 < 3 & 4 > 5 && 6 == 7 | 8 <= 9 || 10 >= 11
    Code
      cat(code_highlight("a <- 10; 20 -> b; c = 30; a$b; a@b", list(operator = "bold")))
    Output
      a <- 10; 20 -> b; c = 30; a$b; a@b

# operator [ansi]

    Code
      cat(code_highlight("~ ! 1 - 2 + 3:4 * 5 / 6 ^ 7", list(operator = "bold")))
    Output
      [1m~[22m [1m![22m 1 [1m-[22m 2 [1m+[22m 3[1m:[22m4 [1m*[22m 5 [1m/[22m 6 [1m^[22m 7
    Code
      cat(code_highlight("? 1 %% 2 %+% 2 < 3 & 4 > 5 && 6 == 7 | 8 <= 9 || 10 >= 11",
        list(operator = "bold")))
    Output
      [1m?[22m 1 [1m%%[22m 2 [1m%+%[22m 2 [1m<[22m 3 [1m&[22m 4 [1m>[22m 5 [1m&&[22m 6 [1m==[22m 7 [1m|[22m 8 [1m<=[22m 9 [1m||[22m 10 [1m>=[22m 11
    Code
      cat(code_highlight("a <- 10; 20 -> b; c = 30; a$b; a@b", list(operator = "bold")))
    Output
      a [1m<-[22m 10; 20 [1m->[22m b; c [1m=[22m 30; a[1m$[22mb; a[1m@[22mb

# call [plain]

    Code
      cat(code_highlight("ls(2)", list(call = "bold")))
    Output
      ls(2)

# call [ansi]

    Code
      cat(code_highlight("ls(2)", list(call = "bold")))
    Output
      [1mls[22m(2)

# string [plain]

    Code
      cat(code_highlight("'s' + \"s\"", list(string = "bold")))
    Output
      's' + "s"

# string [ansi]

    Code
      cat(code_highlight("'s' + \"s\"", list(string = "bold")))
    Output
      [1m's'[22m + [1m"s"[22m

# comment [plain]

    Code
      cat(code_highlight(c("# COM", " ls() ## ANOT"), list(comment = "bold")))
    Output
      # COM  ls() ## ANOT

# comment [ansi]

    Code
      cat(code_highlight(c("# COM", " ls() ## ANOT"), list(comment = "bold")))
    Output
      [1m# COM[22m  ls() [1m## ANOT[22m

# bracket [plain]

    Code
      cat(code_highlight("foo <- function(x){x}", list(bracket = list("bold"))))
    Output
      foo <- function(x){x}

# bracket [ansi]

    Code
      cat(code_highlight("foo <- function(x){x}", list(bracket = list("bold"))))
    Output
      foo <- function[1m([22mx[1m)[22m[1m{[22mx[1m}[22m

# code_theme_list

    Code
      code_theme_list()
    Output
       [1] "Ambiance"              "Chaos"                 "Chrome"               
       [4] "Clouds"                "Clouds Midnight"       "Cobalt"               
       [7] "Crimson Editor"        "Dawn"                  "Dracula"              
      [10] "Dreamweaver"           "Eclipse"               "Idle Fingers"         
      [13] "Katzenmilch"           "Kr Theme"              "Material"             
      [16] "Merbivore"             "Merbivore Soft"        "Mono Industrial"      
      [19] "Monokai"               "Pastel On Dark"        "Solarized Dark"       
      [22] "Solarized Light"       "Textmate (default)"    "Tomorrow"             
      [25] "Tomorrow Night"        "Tomorrow Night Blue"   "Tomorrow Night Bright"
      [28] "Tomorrow Night 80s"    "Twilight"              "Vibrant Ink"          
      [31] "Xcode"                

# new language features, raw strings [ansi]

    Code
      cat(code_highlight("\"old\" + r\"(\"new\"\"\")\"", list(string = "bold",
        reserved = "italic")))
    Output
      [1m"old"[22m + [3mr[23m[1m"("new""")"[22m

# new language features, pipe [ansi]

    Code
      cat(code_highlight("dir() |> toupper()", list(operator = "bold")))
    Output
      dir() [1m|>[22m toupper()

# new language features, lambda functions [ansi]

    Code
      cat(code_highlight("\\(x) x * 2", list(reserved = "bold")))
    Output
      [1m\[22m(x) x * 2

# verbatim text is correctly styled

    Code
      local({
        theme <- list(.padded = list(`margin-left` = 4))
        cli_div(class = "padded", theme = theme)
        lines <- c("first", "second", "third")
        cli_verbatim(lines)
        cli_verbatim(paste0(lines, collapse = "\n"))
      })
    Message
          first
          second
          third
          first
          second
          third

# simplest

    Code
      for (n in 0:2) cli_text("{n} package{?s}")
    Message
      0 packages
      1 package
      2 packages
    Code
      for (n in 0:2) print(pluralize("{n} package{?s}"))
    Output
      0 packages
      1 package
      2 packages

# irregular

    Code
      for (n in 0:2) cli_text("{n} dictionar{?y/ies}")
    Message
      0 dictionaries
      1 dictionary
      2 dictionaries
    Code
      for (n in 0:2) print(pluralize("{n} dictionar{?y/ies}"))
    Output
      0 dictionaries
      1 dictionary
      2 dictionaries

# multiple substitutions

    Code
      for (n in 0:2) cli_text("{n} package{?s} {?is/are} ...")
    Message
      0 packages are ...
      1 package is ...
      2 packages are ...
    Code
      for (n in 0:2) print(pluralize("{n} package{?s} {?is/are} ..."))
    Output
      0 packages are ...
      1 package is ...
      2 packages are ...

# multiple quantities

    Code
      for (m in 0:2) for (n in 0:2) cli_text("{m} package{?s} and {n} folder{?s}")
    Message
      0 packages and 0 folders
      0 packages and 1 folder
      0 packages and 2 folders
      1 package and 0 folders
      1 package and 1 folder
      1 package and 2 folders
      2 packages and 0 folders
      2 packages and 1 folder
      2 packages and 2 folders
    Code
      for (m in 0:2) for (n in 0:2) print(pluralize(
        "{m} package{?s} and {n} folder{?s}"))
    Output
      0 packages and 0 folders
      0 packages and 1 folder
      0 packages and 2 folders
      1 package and 0 folders
      1 package and 1 folder
      1 package and 2 folders
      2 packages and 0 folders
      2 packages and 1 folder
      2 packages and 2 folders

# no()

    Code
      for (n in 0:2) cli_text("{no(n)} package{?s}")
    Message
      no packages
      1 package
      2 packages
    Code
      for (n in 0:2) print(pluralize("{no(n)} package{?s}"))
    Output
      no packages
      1 package
      2 packages

# set qty() explicitly

    Code
      for (n in 0:2) cli_text("{qty(n)}There {?is/are} {n} package{?s}")
    Message
      There are 0 packages
      There is 1 package
      There are 2 packages
    Code
      for (n in 0:2) print(pluralize("{qty(n)}There {?is/are} {n} package{?s}"))
    Output
      There are 0 packages
      There is 1 package
      There are 2 packages

# collapsing vectors

    Code
      pkgs <- (function(n) glue::glue("pkg{seq_len(n)}"))
      for (n in 1:3) cli_text("The {pkgs(n)} package{?s}")
    Message
      The pkg1 package
      The pkg1 and pkg2 packages
      The pkg1, pkg2, and pkg3 packages
    Code
      for (n in 1:3) print(pluralize("The {pkgs(n)} package{?s}"))
    Output
      The pkg1 package
      The pkg1 and pkg2 packages
      The pkg1, pkg2, and pkg3 packages

# pluralization and style

    Code
      special_style <- list(span.foo = list(before = "<", after = ">"))
      cli_div(theme = special_style)
      for (n in 0:2) cli_text("{n} {.foo package{?s}}")
    Message
      0 packages
      1 package
      2 packages

---

    Code
      pkgs <- (function(n) glue::glue("pkg{seq_len(n)}"))
      for (n in 1:3) cli_text("The {.foo {pkgs(n)}} package{?s}")
    Message
      The pkg1 package
      The pkg1 and pkg2 packages
      The pkg1, pkg2, and pkg3 packages

# post-processing

    Code
      for (n in 0:2) cli_text("Package{?s}: {n}")
    Message
      Packages: 0
      Package: 1
      Packages: 2

---

    Code
      pkgs <- (function(n) glue::glue("pkg{seq_len(n)}"))
      for (n in 1:2) cli_text("Package{?s}: {pkgs(n)}")
    Message
      Package: pkg1
      Packages: pkg1 and pkg2
    Code
      for (n in 1:2) print(pluralize("Package{?s}: {pkgs(n)}"))
    Output
      Package: pkg1
      Packages: pkg1 and pkg2

# issue 158

    Code
      print(pluralize("{0} word{?A/B/}"))
    Output
      0 wordA
    Code
      print(pluralize("{1} word{?A/B/}"))
    Output
      1 wordB
    Code
      print(pluralize("{9} word{?A/B/}"))
    Output
      9 word

# Edge cases for pluralize() (#701)

    Code
      print(pluralize("{NA} file{?s} expected"))
    Output
      NA file expected
    Code
      print(pluralize("{NA_character_} file{?s} expected"))
    Output
      NA file expected
    Code
      print(pluralize("{NA_real_} file{?s} expected"))
    Output
      NA files expected
    Code
      print(pluralize("{NA_integer_} file{?s} expected"))
    Output
      NA files expected
    Code
      print(pluralize("{NaN} file{?s} expected"))
    Output
      NaN files expected
    Code
      print(pluralize("{Inf} file{?s} expected"))
    Output
      Inf files expected
    Code
      print(pluralize("{-Inf} file{?s} expected"))
    Output
      -Inf files expected

---

    Code
      print(pluralize("Found {NA} director{?y/ies}."))
    Output
      Found NA directory.
    Code
      print(pluralize("Found {NA_character_} director{?y/ies}."))
    Output
      Found NA directory.
    Code
      print(pluralize("Found {NA_real_} director{?y/ies}."))
    Output
      Found NA directories.
    Code
      print(pluralize("Found {NA_integer_} director{?y/ies}."))
    Output
      Found NA directories.
    Code
      print(pluralize("Found {NaN} director{?y/ies}."))
    Output
      Found NaN directories.
    Code
      print(pluralize("Found {Inf} director{?y/ies}."))
    Output
      Found Inf directories.
    Code
      print(pluralize("Found {-Inf} director{?y/ies}."))
    Output
      Found -Inf directories.

---

    Code
      print(pluralize("Will remove {?no/the/the} {NA} package{?s}."))
    Output
      Will remove the NA package.
    Code
      print(pluralize("Will remove {?no/the/the} {NA_character_} package{?s}."))
    Output
      Will remove the NA package.
    Code
      print(pluralize("Will remove {?no/the/the} {NA_real_} package{?s}."))
    Output
      Will remove the NA packages.
    Code
      print(pluralize("Will remove {?no/the/the} {NA_integer_} package{?s}."))
    Output
      Will remove the NA packages.
    Code
      print(pluralize("Will remove {?no/the/the} {NaN} package{?s}."))
    Output
      Will remove the NaN packages.
    Code
      print(pluralize("Will remove {?no/the/the} {Inf} package{?s}."))
    Output
      Will remove the Inf packages.
    Code
      print(pluralize("Will remove {?no/the/the} {-Inf} package{?s}."))
    Output
      Will remove the -Inf packages.

# tree [plain]

    Code
      tree(data)
    Output
      processx
      +-assertthat
      +-crayon
      +-debugme
      | \-crayon
      \-R6

---

    Code
      tree(data, root = "desc")
    Output
      desc
      +-assertthat
      +-R6
      +-crayon
      \-rprojroot
        \-backports

# tree [ansi]

    Code
      tree(data)
    Output
      processx
      +-assertthat
      +-crayon
      +-debugme
      | \-crayon
      \-R6

---

    Code
      tree(data, root = "desc")
    Output
      desc
      +-assertthat
      +-R6
      +-crayon
      \-rprojroot
        \-backports

# tree [unicode]

    Code
      tree(data)
    Output
      processx
      ├─assertthat
      ├─crayon
      ├─debugme
      │ └─crayon
      └─R6

---

    Code
      tree(data, root = "desc")
    Output
      desc
      ├─assertthat
      ├─R6
      ├─crayon
      └─rprojroot
        └─backports

# tree [fancy]

    Code
      tree(data)
    Output
      processx
      ├─assertthat
      ├─crayon
      ├─debugme
      │ └─crayon
      └─R6

---

    Code
      tree(data, root = "desc")
    Output
      desc
      ├─assertthat
      ├─R6
      ├─crayon
      └─rprojroot
        └─backports

# trimming [plain]

    Code
      tree(pkgs, trim = TRUE)
    Output
      dplyr@0.8.3
      +-assertthat@0.2.1
      +-glue@1.3.1
      +-magrittr@1.5
      +-R6@2.4.0
      +-Rcpp@1.0.2
      +-rlang@0.4.0
      +-tibble@2.1.3
      | +-cli@1.1.0
      | | +-assertthat@0.2.1  (trimmed)
      | | \-crayon@1.3.4
      | +-crayon@1.3.4  (trimmed)
      | +-fansi@0.4.0
      | +-pillar@1.4.2
      | | +-cli@1.1.0  (trimmed)
      | | +-crayon@1.3.4  (trimmed)
      | | +-fansi@0.4.0  (trimmed)
      | | +-rlang@0.4.0  (trimmed)
      | | +-utf8@1.1.4
      | | \-vctrs@0.2.0
      | |   +-backports@1.1.5
      | |   +-ellipsis@0.3.0
      | |   | \-rlang@0.4.0  (trimmed)
      | |   +-digest@0.6.21
      | |   +-glue@1.3.1  (trimmed)
      | |   +-rlang@0.4.0  (trimmed)
      | |   \-zeallot@0.1.0
      | +-pkgconfig@2.0.3
      | \-rlang@0.4.0  (trimmed)
      \-tidyselect@0.2.5
        +-glue@1.3.1  (trimmed)
        +-rlang@0.4.0  (trimmed)
        \-Rcpp@1.0.2  (trimmed)

# trimming [ansi]

    Code
      tree(pkgs, trim = TRUE)
    Output
      dplyr@0.8.3
      +-assertthat@0.2.1
      +-glue@1.3.1
      +-magrittr@1.5
      +-R6@2.4.0
      +-Rcpp@1.0.2
      +-rlang@0.4.0
      +-tibble@2.1.3
      | +-cli@1.1.0
      | | +-assertthat@0.2.1  (trimmed)
      | | \-crayon@1.3.4
      | +-crayon@1.3.4  (trimmed)
      | +-fansi@0.4.0
      | +-pillar@1.4.2
      | | +-cli@1.1.0  (trimmed)
      | | +-crayon@1.3.4  (trimmed)
      | | +-fansi@0.4.0  (trimmed)
      | | +-rlang@0.4.0  (trimmed)
      | | +-utf8@1.1.4
      | | \-vctrs@0.2.0
      | |   +-backports@1.1.5
      | |   +-ellipsis@0.3.0
      | |   | \-rlang@0.4.0  (trimmed)
      | |   +-digest@0.6.21
      | |   +-glue@1.3.1  (trimmed)
      | |   +-rlang@0.4.0  (trimmed)
      | |   \-zeallot@0.1.0
      | +-pkgconfig@2.0.3
      | \-rlang@0.4.0  (trimmed)
      \-tidyselect@0.2.5
        +-glue@1.3.1  (trimmed)
        +-rlang@0.4.0  (trimmed)
        \-Rcpp@1.0.2  (trimmed)

# trimming [unicode]

    Code
      tree(pkgs, trim = TRUE)
    Output
      dplyr@0.8.3
      ├─assertthat@0.2.1
      ├─glue@1.3.1
      ├─magrittr@1.5
      ├─R6@2.4.0
      ├─Rcpp@1.0.2
      ├─rlang@0.4.0
      ├─tibble@2.1.3
      │ ├─cli@1.1.0
      │ │ ├─assertthat@0.2.1  (trimmed)
      │ │ └─crayon@1.3.4
      │ ├─crayon@1.3.4  (trimmed)
      │ ├─fansi@0.4.0
      │ ├─pillar@1.4.2
      │ │ ├─cli@1.1.0  (trimmed)
      │ │ ├─crayon@1.3.4  (trimmed)
      │ │ ├─fansi@0.4.0  (trimmed)
      │ │ ├─rlang@0.4.0  (trimmed)
      │ │ ├─utf8@1.1.4
      │ │ └─vctrs@0.2.0
      │ │   ├─backports@1.1.5
      │ │   ├─ellipsis@0.3.0
      │ │   │ └─rlang@0.4.0  (trimmed)
      │ │   ├─digest@0.6.21
      │ │   ├─glue@1.3.1  (trimmed)
      │ │   ├─rlang@0.4.0  (trimmed)
      │ │   └─zeallot@0.1.0
      │ ├─pkgconfig@2.0.3
      │ └─rlang@0.4.0  (trimmed)
      └─tidyselect@0.2.5
        ├─glue@1.3.1  (trimmed)
        ├─rlang@0.4.0  (trimmed)
        └─Rcpp@1.0.2  (trimmed)

# trimming [fancy]

    Code
      tree(pkgs, trim = TRUE)
    Output
      dplyr@0.8.3
      ├─assertthat@0.2.1
      ├─glue@1.3.1
      ├─magrittr@1.5
      ├─R6@2.4.0
      ├─Rcpp@1.0.2
      ├─rlang@0.4.0
      ├─tibble@2.1.3
      │ ├─cli@1.1.0
      │ │ ├─assertthat@0.2.1  (trimmed)
      │ │ └─crayon@1.3.4
      │ ├─crayon@1.3.4  (trimmed)
      │ ├─fansi@0.4.0
      │ ├─pillar@1.4.2
      │ │ ├─cli@1.1.0  (trimmed)
      │ │ ├─crayon@1.3.4  (trimmed)
      │ │ ├─fansi@0.4.0  (trimmed)
      │ │ ├─rlang@0.4.0  (trimmed)
      │ │ ├─utf8@1.1.4
      │ │ └─vctrs@0.2.0
      │ │   ├─backports@1.1.5
      │ │   ├─ellipsis@0.3.0
      │ │   │ └─rlang@0.4.0  (trimmed)
      │ │   ├─digest@0.6.21
      │ │   ├─glue@1.3.1  (trimmed)
      │ │   ├─rlang@0.4.0  (trimmed)
      │ │   └─zeallot@0.1.0
      │ ├─pkgconfig@2.0.3
      │ └─rlang@0.4.0  (trimmed)
      └─tidyselect@0.2.5
        ├─glue@1.3.1  (trimmed)
        ├─rlang@0.4.0  (trimmed)
        └─Rcpp@1.0.2  (trimmed)

# ul [plain]

    Code
      local({
        lid <- cli_ul()
        cli_li("foo")
        cli_li(c("bar", "foobar"))
        cli_end(lid)
      })
    Message
      * foo
      * bar
      * foobar

# ul [unicode]

    Code
      local({
        lid <- cli_ul()
        cli_li("foo")
        cli_li(c("bar", "foobar"))
        cli_end(lid)
      })
    Message
      • foo
      • bar
      • foobar

# ol [plain]

    Code
      local({
        cli_div(theme = list(ol = list()))
        lid <- cli_ol()
        cli_li("foo")
        cli_li(c("bar", "foobar"))
        cli_end(lid)
      })
    Message
      1. foo
      2. bar
      3. foobar

# ol [unicode]

    Code
      local({
        cli_div(theme = list(ol = list()))
        lid <- cli_ol()
        cli_li("foo")
        cli_li(c("bar", "foobar"))
        cli_end(lid)
      })
    Message
      1. foo
      2. bar
      3. foobar

# ul ul [plain]

    Code
      local({
        cli_div(theme = list(`ul ul` = list(`list-style-type` = "-", `margin-left` = 2)))
        lid <- cli_ul()
        cli_li("1")
        lid2 <- cli_ul()
        cli_li("1 1")
        cli_li(c("1 2", "1 3"))
        cli_end(lid2)
        cli_li("2")
        cli_end(lid)
      })
    Message
      * 1
          - 1 1
          - 1 2
          - 1 3
      * 2

# ul ul [unicode]

    Code
      local({
        cli_div(theme = list(`ul ul` = list(`list-style-type` = "-", `margin-left` = 2)))
        lid <- cli_ul()
        cli_li("1")
        lid2 <- cli_ul()
        cli_li("1 1")
        cli_li(c("1 2", "1 3"))
        cli_end(lid2)
        cli_li("2")
        cli_end(lid)
      })
    Message
      • 1
          - 1 1
          - 1 2
          - 1 3
      • 2

# ul ol [plain]

    Code
      local({
        cli_div(theme = list(li = list(`margin-left` = 2)))
        lid <- cli_ul()
        cli_li("1")
        lid2 <- cli_ol()
        cli_li("1 1")
        cli_li(c("1 2", "1 3"))
        cli_end(lid2)
        cli_li("2")
        cli_end(lid)
      })
    Message
        * 1
          1. 1 1
          2. 1 2
          3. 1 3
        * 2

# ul ol [unicode]

    Code
      local({
        cli_div(theme = list(li = list(`margin-left` = 2)))
        lid <- cli_ul()
        cli_li("1")
        lid2 <- cli_ol()
        cli_li("1 1")
        cli_li(c("1 2", "1 3"))
        cli_end(lid2)
        cli_li("2")
        cli_end(lid)
      })
    Message
        • 1
          1. 1 1
          2. 1 2
          3. 1 3
        • 2

# ol ol [plain]

    Code
      local({
        cli_div(theme = list(li = list(`margin-left` = 2), `li li` = list(
          `margin-left` = 2)))
        lid <- cli_ol()
        cli_li("1")
        lid2 <- cli_ol()
        cli_li("1 1")
        cli_li(c("1 2", "1 3"))
        cli_end(lid2)
        cli_li("2")
        cli_end(lid)
      })
    Message
        1. 1
          1. 1 1
          2. 1 2
          3. 1 3
        2. 2

# ol ol [unicode]

    Code
      local({
        cli_div(theme = list(li = list(`margin-left` = 2), `li li` = list(
          `margin-left` = 2)))
        lid <- cli_ol()
        cli_li("1")
        lid2 <- cli_ol()
        cli_li("1 1")
        cli_li(c("1 2", "1 3"))
        cli_end(lid2)
        cli_li("2")
        cli_end(lid)
      })
    Message
        1. 1
          1. 1 1
          2. 1 2
          3. 1 3
        2. 2

# ol ul [plain]

    Code
      local({
        cli_div(theme = list(ul = list(`margin-left` = 2)))
        lid <- cli_ol()
        cli_li("1")
        lid2 <- cli_ul()
        cli_li("1 1")
        cli_li(c("1 2", "1 3"))
        cli_end(lid2)
        cli_li("2")
        cli_end(lid)
      })
    Message
      1. 1
          * 1 1
          * 1 2
          * 1 3
      2. 2

# ol ul [unicode]

    Code
      local({
        cli_div(theme = list(ul = list(`margin-left` = 2)))
        lid <- cli_ol()
        cli_li("1")
        lid2 <- cli_ul()
        cli_li("1 1")
        cli_li(c("1 2", "1 3"))
        cli_end(lid2)
        cli_li("2")
        cli_end(lid)
      })
    Message
      1. 1
          • 1 1
          • 1 2
          • 1 3
      2. 2

# starting with an item [plain]

    Code
      local({
        cli_li("foo")
        cli_li(c("bar", "foobar"))
        cli_end()
        cli_end()
      })
    Message
      * foo
      * bar
      * foobar

# starting with an item [unicode]

    Code
      local({
        cli_li("foo")
        cli_li(c("bar", "foobar"))
        cli_end()
        cli_end()
      })
    Message
      • foo
      • bar
      • foobar

# ol, with first item [plain]

    Code
      local({
        cli_div(theme = list(ol = list()))
        lid <- cli_ol("foo", .close = FALSE)
        cli_li(c("bar", "foobar"))
        cli_end(lid)
      })
    Message
      1. foo
      2. bar
      3. foobar

# ol, with first item [unicode]

    Code
      local({
        cli_div(theme = list(ol = list()))
        lid <- cli_ol("foo", .close = FALSE)
        cli_li(c("bar", "foobar"))
        cli_end(lid)
      })
    Message
      1. foo
      2. bar
      3. foobar

# ul, with first item [plain]

    Code
      local({
        lid <- cli_ul("foo", .close = FALSE)
        cli_li(c("bar", "foobar"))
        cli_end(lid)
      })
    Message
      * foo
      * bar
      * foobar

# ul, with first item [unicode]

    Code
      local({
        lid <- cli_ul("foo", .close = FALSE)
        cli_li(c("bar", "foobar"))
        cli_end(lid)
      })
    Message
      • foo
      • bar
      • foobar

# dl [plain]

    Code
      local({
        cli_div(theme = list(ul = list()))
        lid <- cli_dl()
        cli_li(c(this = "foo"))
        cli_li(c(that = "bar", other = "foobar"))
        cli_end(lid)
      })
    Message
      this: foo
      that: bar
      other: foobar

# dl [unicode]

    Code
      local({
        cli_div(theme = list(ul = list()))
        lid <- cli_dl()
        cli_li(c(this = "foo"))
        cli_li(c(that = "bar", other = "foobar"))
        cli_end(lid)
      })
    Message
      this: foo
      that: bar
      other: foobar

# dl dl [plain]

    Code
      local({
        cli_div(theme = list(li = list(`margin-left` = 2)))
        lid <- cli_dl()
        cli_li(c(a = "1"))
        lid2 <- cli_dl()
        cli_li(c(`a-a` = "1 1"))
        cli_li(c(`a-b` = "1 2", `a-c` = "1 3"))
        cli_end(lid2)
        cli_li(c(b = "2"))
        cli_end(lid)
      })
    Message
        a: 1
          a-a: 1 1
          a-b: 1 2
          a-c: 1 3
        b: 2

# dl dl [unicode]

    Code
      local({
        cli_div(theme = list(li = list(`margin-left` = 2)))
        lid <- cli_dl()
        cli_li(c(a = "1"))
        lid2 <- cli_dl()
        cli_li(c(`a-a` = "1 1"))
        cli_li(c(`a-b` = "1 2", `a-c` = "1 3"))
        cli_end(lid2)
        cli_li(c(b = "2"))
        cli_end(lid)
      })
    Message
        a: 1
          a-a: 1 1
          a-b: 1 2
          a-c: 1 3
        b: 2

# dl ol [plain]

    Code
      local({
        cli_div(theme = list(li = list(`margin-left` = 2)))
        lid <- cli_dl()
        cli_li(c(a = "1"))
        lid2 <- cli_ol()
        cli_li(c("1 1"))
        cli_li(c("1 2", "1 3"))
        cli_end(lid2)
        cli_li(c(b = "2"))
        cli_end(lid)
      })
    Message
        a: 1
          1. 1 1
          2. 1 2
          3. 1 3
        b: 2

# dl ol [unicode]

    Code
      local({
        cli_div(theme = list(li = list(`margin-left` = 2)))
        lid <- cli_dl()
        cli_li(c(a = "1"))
        lid2 <- cli_ol()
        cli_li(c("1 1"))
        cli_li(c("1 2", "1 3"))
        cli_end(lid2)
        cli_li(c(b = "2"))
        cli_end(lid)
      })
    Message
        a: 1
          1. 1 1
          2. 1 2
          3. 1 3
        b: 2

# dl ul [plain]

    Code
      local({
        cli_div(theme = list(li = list(`margin-left` = 2)))
        lid <- cli_dl()
        cli_li(c(a = "1"))
        lid2 <- cli_ul()
        cli_li(c("1 1"))
        cli_li(c("1 2", "1 3"))
        cli_end(lid2)
        cli_li(c(b = "2"))
        cli_end(lid)
      })
    Message
        a: 1
          * 1 1
          * 1 2
          * 1 3
        b: 2

# dl ul [unicode]

    Code
      local({
        cli_div(theme = list(li = list(`margin-left` = 2)))
        lid <- cli_dl()
        cli_li(c(a = "1"))
        lid2 <- cli_ul()
        cli_li(c("1 1"))
        cli_li(c("1 2", "1 3"))
        cli_end(lid2)
        cli_li(c(b = "2"))
        cli_end(lid)
      })
    Message
        a: 1
          • 1 1
          • 1 2
          • 1 3
        b: 2

# ol dl [plain]

    Code
      local({
        cli_div(theme = list(li = list(`margin-left` = 2)))
        lid <- cli_ol()
        cli_li("1")
        lid2 <- cli_dl()
        cli_li(c(`a-a` = "1 1"))
        cli_li(c(`a-b` = "1 2", `a-c` = "1 3"))
        cli_end(lid2)
        cli_li("2")
        cli_end(lid)
      })
    Message
        1. 1
          a-a: 1 1
          a-b: 1 2
          a-c: 1 3
        2. 2

# ol dl [unicode]

    Code
      local({
        cli_div(theme = list(li = list(`margin-left` = 2)))
        lid <- cli_ol()
        cli_li("1")
        lid2 <- cli_dl()
        cli_li(c(`a-a` = "1 1"))
        cli_li(c(`a-b` = "1 2", `a-c` = "1 3"))
        cli_end(lid2)
        cli_li("2")
        cli_end(lid)
      })
    Message
        1. 1
          a-a: 1 1
          a-b: 1 2
          a-c: 1 3
        2. 2

# ul dl [plain]

    Code
      local({
        cli_div(theme = list(li = list(`margin-left` = 2)))
        lid <- cli_ul()
        cli_li("1")
        lid2 <- cli_dl()
        cli_li(c(`a-a` = "1 1"))
        cli_li(c(`a-b` = "1 2", `a-c` = "1 3"))
        cli_end(lid2)
        cli_li("2")
        cli_end(lid)
      })
    Message
        * 1
          a-a: 1 1
          a-b: 1 2
          a-c: 1 3
        * 2

# ul dl [unicode]

    Code
      local({
        cli_div(theme = list(li = list(`margin-left` = 2)))
        lid <- cli_ul()
        cli_li("1")
        lid2 <- cli_dl()
        cli_li(c(`a-a` = "1 1"))
        cli_li(c(`a-b` = "1 2", `a-c` = "1 3"))
        cli_end(lid2)
        cli_li("2")
        cli_end(lid)
      })
    Message
        • 1
          a-a: 1 1
          a-b: 1 2
          a-c: 1 3
        • 2

# dl, with first item [plain]

    Code
      local({
        cli_div(theme = list(ul = list()))
        lid <- cli_dl(c(this = "foo"), .close = FALSE)
        cli_li(c(that = "bar", other = "foobar"))
        cli_end(lid)
      })
    Message
      this: foo
      that: bar
      other: foobar

# dl, with first item [unicode]

    Code
      local({
        cli_div(theme = list(ul = list()))
        lid <- cli_dl(c(this = "foo"), .close = FALSE)
        cli_li(c(that = "bar", other = "foobar"))
        cli_end(lid)
      })
    Message
      this: foo
      that: bar
      other: foobar

# styling pieces of a dl [ansi]

    Code
      local({
        cli_div(theme = list(.dt = list(postfix = " -> "), .dd = list(color = "blue")))
        cli_dl(c(foo = "bar", bar = "baz"))
      })
    Message
      foo -> [34mbar[39m
      bar -> [34mbaz[39m

# cli_dl edge cases

    Code
      cli_dl("foo", "must be a named character vector")
    Condition
      Error:
      ! `items` must be a named character vector
      i `items` is not named

---

    Code
      cli_dl(c(abc = "foo", empty = "", def = "bar"))
    Message
      abc: foo
      empty:
      def: bar

# cli_dl label style

    Code
      local({
        cli_dl(c(`{.code code}` = "{.code this is code too}", `{.str strin}` = "{.url https://x.com}"))
      })
    Message
      `code`: `this is code too`
      "strin": <https://x.com>

# type style

    Code
      cli_text("{.obj_type_friendly {mtcars}}")
    Message
      a data frame
    Code
      cli_text("{.obj_type_friendly {tibble::as_tibble(mtcars)}}")
    Message
      a tibble
    Code
      cli_text("{.obj_type_friendly {rlang::quo(1)}}")
    Message
      a <quosure> object
    Code
      cli_text("{.obj_type_friendly {list()}}")
    Message
      an empty list
    Code
      cli_text("{.obj_type_friendly {matrix(list(1, 2))}}")
    Message
      a list matrix
    Code
      cli_text("{.obj_type_friendly {array(list(1, 2, 3))}}")
    Message
      a list
    Code
      cli_text("{.obj_type_friendly {integer()}}")
    Message
      an empty integer vector
    Code
      cli_text("{.obj_type_friendly {matrix(1:3)}}")
    Message
      an integer matrix
    Code
      cli_text("{.obj_type_friendly {array(1:3, dim = 1:3)}}")
    Message
      an integer array
    Code
      cli_text("{.obj_type_friendly {character()}}")
    Message
      an empty character vector
    Code
      cli_text("{.obj_type_friendly {matrix(letters)}}")
    Message
      a character matrix
    Code
      cli_text("{.obj_type_friendly {array(letters[1:3], dim = 1:3)}}")
    Message
      a character array
    Code
      typename(quote(expr = ))
    Output
      [1] "absent"
    Code
      cli_text("{.obj_type_friendly {NA}}")
    Message
      `NA`
    Code
      cli_text("{.obj_type_friendly {NA_integer_}}")
    Message
      an integer `NA`
    Code
      cli_text("{.obj_type_friendly {NA_real_}}")
    Message
      a numeric `NA`
    Code
      cli_text("{.obj_type_friendly {NA_complex_}}")
    Message
      a complex `NA`
    Code
      cli_text("{.obj_type_friendly {NA_character_}}")
    Message
      a character `NA`
    Code
      cli_text("{.obj_type_friendly {TRUE}}")
    Message
      `TRUE`
    Code
      cli_text("{.obj_type_friendly {FALSE}}")
    Message
      `FALSE`
    Code
      cli_text("{.obj_type_friendly {1L}}")
    Message
      an integer
    Code
      cli_text("{.obj_type_friendly {1.0}}")
    Message
      a number
    Code
      cli_text("{.obj_type_friendly {1i}}")
    Message
      a complex number
    Code
      cli_text("{.obj_type_friendly {as.raw(1L)}}")
    Message
      a raw value
    Code
      cli_text("{.obj_type_friendly {'foo'}}")
    Message
      a string
    Code
      cli_text("{.obj_type_friendly {''}}")
    Message
      `""`
    Code
      cli_text("{.obj_type_friendly {list(1)}}")
    Message
      a list
    Code
      cli_text("{.obj_type_friendly {matrix(NA)}}")
    Message
      a logical matrix
    Code
      cli_text("{.obj_type_friendly {matrix(1)}}")
    Message
      a double matrix
    Code
      cli_text("{.obj_type_friendly {logical()}}")
    Message
      an empty logical vector
    Code
      cli_text("{.obj_type_friendly {integer()}}")
    Message
      an empty integer vector
    Code
      cli_text("{.obj_type_friendly {double()}}")
    Message
      an empty numeric vector
    Code
      cli_text("{.obj_type_friendly {complex()}}")
    Message
      an empty complex vector
    Code
      cli_text("{.obj_type_friendly {character()}}")
    Message
      an empty character vector
    Code
      cli_text("{.obj_type_friendly {raw()}}")
    Message
      an empty raw vector
    Code
      cli_text("{.obj_type_friendly {list()}}")
    Message
      an empty list
    Code
      cli_text("{.obj_type_friendly {structure(1, class = 'igraph')}}")
    Message
      an <igraph> object

# auto closing

    Code
      local({
        cli_div(theme = list(`.xx .emph` = list(before = "itsu:")))
        f <- (function() {
          cli_par(class = "xx")
          cli_text("foo {.emph blah} bar")
        })
        f()
        cli_text("foo {.emph blah} bar")
      })
    Message
      foo itsu:blah bar
      
      foo blah bar

# opt out of auto closing

    Code
      local({
        cli_div(theme = list(`.xx .emph` = list(before = "itsu:")))
        id <- NULL
        f <- (function() {
          id <<- cli_par(class = "xx", .auto_close = FALSE)
          cli_text("foo {.emph blah} bar")
        })
        f()
        cli_text("foo {.emph blah} bar")
        expect_false(is.null(id))
        cli_end(id)
        cli_text("foo {.emph blah} bar")
      })
    Message
      foo itsu:blah bar
      foo itsu:blah bar
      
      foo blah bar

# auto closing with special env

    Code
      local({
        cli_div(theme = list(`.xx .emph` = list(before = "itsu:")))
        f <- (function() {
          g()
          cli_text("foo {.emph blah} bar")
        })
        g <- (function() {
          cli_par(class = "xx", .auto_close = TRUE, .envir = parent.frame())
          cli_text("foo {.emph blah} bar")
        })
        f()
        cli_text("foo {.emph blah} bar")
      })
    Message
      foo itsu:blah bar
      foo itsu:blah bar
      
      foo blah bar

# div with special style

    Code
      f <- (function() {
        cli_div(theme = list(`.xx .emph` = list(before = "itsu:")))
        cli_par(class = "xx")
        cli_text("foo {.emph blah} bar")
      })
      f()
    Message
      foo itsu:blah bar
      
    Code
      cli_text("foo {.emph blah} bar")
    Message
      foo blah bar

# margin is squashed

    Code
      local({
        cli_div(theme = list(par = list(`margin-top` = 4, `margin-bottom` = 4)))
        cli_text("three lines")
        cli_par()
        cli_par()
        cli_par()
        cli_text("until here")
        cli_end()
        cli_end()
        cli_end()
        cli_par()
        cli_par()
        cli_par()
        cli_text("no space, still")
        cli_end()
        cli_end()
        cli_end()
        cli_text("three lines again")
      })
    Message
      three lines
      
      
      
      
      until here
      
      
      
      
      no space, still
      
      
      
      
      three lines again

# before and after work properly

    Code
      local({
        cli_div(theme = list(`div.alert-success` = list(before = "!!!")))
        cli_alert_success("{.pkg foobar} is good")
      })
    Message
      !!!foobar is good

```{r, include = FALSE}
knitr::opts_chunk$set(
  R.options = list(
    cli.num_colors = 1L,
    cli.unicode = FALSE
  ),
  results = "hold",
  comment = "#>",
  cache = TRUE
)
```

## My platform supports ANSI colors, why does cli not use them?

It is probably a mistake in the ANSI support detection algorithm.
Please open an issue at https://github.com/r-lib/cli/issues and do not forget to tell us the details of your platform and terminal or GUI.

## How do I turn off ANSI colors and styles?

Set the `NO_COLOR` environment variable to a non-empty value.
You can do this in your `.Renviron` file (use `usethis::edit_r_environ()`).

If you want to do this for testthat tests, then consider using the 3rd edition on testthat, which does turn off ANSI styling automatically inside `test_that()`.

## cli does not show the output before `file.choose()`

Try calling `flush.console()` to flush the console, before `file.choose()`.
If flushing does not work and you are in RStudio, then it is probably this RStudio bug: https://github.com/rstudio/rstudio/issues/8040
See more details at https://github.com/r-lib/cli/issues/151

## Why are heading separators wider than my screen in RStudio?

The display width of some Unicode characters ambiguous in the Unicode standard.
Some software treats them as narrow (one column on the screen), other as wide (two columns).
In some terminal emulators (for example iTerm2), you can configure the preferred behavior.

Unfortunately the box drawing characters that cli uses also have ambiguous width.

In RStudio the behavior depends on the font.
In particular, Consolas, Courier and Inconsolata treats them as wide characters, so cli output will not look great with these.
Some good, modern fonts that look good include Menlo, Fira Code and Source Code Pro.

If you do not want to change your font, you can also turn off Unicode output, by setting the `cli.unicode` option:
```r
options(cli.unicode = FALSE)
```

A related issue: https://github.com/r-lib/cli/issues/320

## Is there a suggested font to use with cli?

In modern terminals, cli output usually looks good.

If you see too wide heading separators in RStudio, then see the previous question: Why are heading separators wider than my screen in RStudio?.

If some output is garbled, then cli probably misdetected Unicode support for your terminal or font.
You can try choosing a different font.
In our experience output looks good with Menlo, Fira Code and Source Code Pro.
Alternatively you can turn off Unicode output:
```r
options(cli.unicode = FALSE)
```
If you think this is our fault, then please also file an issue at https://github.com/r-lib/cli/issues

---
editor_options: 
  markdown: 
    wrap: sentence
---

```{r, include = FALSE, cache = FALSE}
knitr::opts_chunk$set(
  R.options = list(
    cli.num_colors = 1L,
    cli.unicode = FALSE
  ),
  results = "hold",
  comment = "#>",
  cache = TRUE
)
```

# Introduction

cli has tools to create messages that are printed correctly in singular and plural forms.
This usually requires minimal extra work, and increases the quality of the messages greatly.
In this document we first show some pluralization examples that you can use as guidelines.
Hopefully these are intuitive enough, so that they can be used without knowing the exact cli pluralization rules.

If you need pluralization without the semantic cli functions, see the `pluralize()` function.

# Examples

## Pluralization markup

In the simplest case the message contains a single `{}` glue substitution, which specifies the quantity that is used to select between the singular and plural forms.
Pluralization uses markup that is similar to glue, but uses the `{?` and `}` delimiters:

```{r}
library(cli)
nfile <- 0; cli_text("Found {nfile} file{?s}.")
nfile <- 1; cli_text("Found {nfile} file{?s}.")
nfile <- 2; cli_text("Found {nfile} file{?s}.")
```

Here the value of `nfile` is used to decide whether the singular or plural form of `file` is used.
This is the most common case for English messages.

## Irregular plurals

If the plural form is more difficult than a simple `s` suffix, then the singular and plural forms can be given, separated with a forward slash:

```{r}
ndir <- 1; cli_text("Found {ndir} director{?y/ies}.")
ndir <- 5; cli_text("Found {ndir} director{?y/ies}.")
```

## Use `"no"` instead of zero

For readability, it is better to use the `no()` helper function to include a count in a message.
`no()` prints the word `"no"` if the count is zero, and prints the numeric count otherwise:

```{r}
nfile <- 0; cli_text("Found {no(nfile)} file{?s}.")
nfile <- 1; cli_text("Found {no(nfile)} file{?s}.")
nfile <- 2; cli_text("Found {no(nfile)} file{?s}.")
```

## Use the length of character vectors

With the auto-collapsing feature of cli it is easy to include a list of objects in a message.
When cli interprets a character vector as a pluralization quantity, it takes the length of the vector:

```{r}
pkgs <- "pkg1"
cli_text("Will remove the {.pkg {pkgs}} package{?s}.")
pkgs <- c("pkg1", "pkg2", "pkg3")
cli_text("Will remove the {.pkg {pkgs}} package{?s}.")
```

Note that the length is only used for non-numeric vectors (when `is.numeric(x)` return `FALSE`).
If you want to use the length of a numeric vector, convert it to character via `as.character()`.

You can combine collapsed vectors with `"no"`, like this:

```{r}
pkgs <- character()
cli_text("Will remove {?no/the/the} {.pkg {pkgs}} package{?s}.")
pkgs <- c("pkg1", "pkg2", "pkg3")
cli_text("Will remove {?no/the/the} {.pkg {pkgs}} package{?s}.")
```

When the pluralization markup contains three alternatives, like above, the first one is used for zero, the second for one, and the third one for larger quantities.

## Choosing the right quantity

When the text contains multiple glue `{}` substitutions, the one right before the pluralization markup is used.
For example:

```{r}
nfiles <- 3; ndirs <- 1
cli_text("Found {nfiles} file{?s} and {ndirs} director{?y/ies}")
```

This is sometimes not the the correct one.
You can explicitly specify the correct quantity using the `qty()` function.
This sets that quantity without printing anything:

```{r}
nupd <- 3; ntotal <- 10
cli_text("{nupd}/{ntotal} {qty(nupd)} file{?s} {?needs/need} updates")
```

Note that if the message only contains a single `{}` substitution, then this may appear before or after the pluralization markup.
If the message contains multiple `{}` substitutions *after* pluralization markup, an error is thrown.

Similarly, if the message contains no `{}` substitutions at all, but has pluralization markup, an error is thrown.

# Rules

The exact rules of cli pluralization.
There are two sets of rules.
The first set specifies how a quantity is associated with a `{?}` pluralization markup.
The second set describes how the `{?}` is parsed and interpreted.

## Quantities

1.  `{}` substitutions define quantities.
    If the value of a `{}` substitution is numeric (when `is.numeric(x)` holds), then it has to have length one to define a quantity.
    This is only enforced if the `{}` substitution is used for pluralization.
    The quantity is defined as the value of `{}` then, rounded with `as.integer()`.
    If the value of `{}` is not numeric, then its quantity is defined as its length.

2.  If a message has `{?}` markup but no `{}` substitution, an error is thrown.

3.  If a message has exactly one `{}` substitution, its value is used as the pluralization quantity for all `{?}` markup in the message.

4.  If a message has multiple `{}` substitutions, then for each `{?}` markup cli uses the quantity of the `{}` substitution that precedes it.

5.  If a message has multiple `{}` substitutions and has pluralization markup without a preceding `{}` substitution, an error is thrown.

## Pluralization markup

1.  Pluralization markup starts with `{?` and ends with `}`.
    It may not contain `{` and `}` characters, so it may not contain `{}` substitutions either.

2.  Alternative words or suffixes are separated by `/`.

3.  If there is a single alternative, then *nothing* is used if `quantity == 1` and this single alternative is used if `quantity != 1`.

4.  If there are two alternatives, the first one is used for `quantity == 1`, the second one for `quantity != 1` (including \``quantity == 0`).

5.  If there are three alternatives, the first one is used for `quantity == 0`, the second one for `quantity == 1`, and the third one otherwise.

# quoting phrases that don't start or end with letter or number [plain]

    Code
      local({
        x0 <- "good-name"
        cli_text("The name is {.file {x0}}.")
        x <- "weird-name "
        cli_text("The name is {.file {x}}.")
        cli_text("The name is {.path {x}}.")
        cli_text("The name is {.email {x}}.")
      })
    Message <cliMessage>
      The name is 'good-name'.
      The name is 'weird-name '.
      The name is 'weird-name '.
      The name is 'weird-name '.

# quoting phrases that don't start or end with letter or number [ansi]

    Code
      local({
        x0 <- "good-name"
        cli_text("The name is {.file {x0}}.")
        x <- "weird-name "
        cli_text("The name is {.file {x}}.")
        cli_text("The name is {.path {x}}.")
        cli_text("The name is {.email {x}}.")
      })
    Message <cliMessage>
      The name is [34mgood-name[39m.
      The name is '[34mweird-name[39m[44m [49m'.
      The name is '[34mweird-name[39m[44m [49m'.
      The name is '[34mweird-name[39m[44m [49m'.

# quoting weird names, still [plain]

    Code
      local({
        cat_line(nb(quote_weird_name("good")))
        cat_line(nb(quote_weird_name("  bad")))
        cat_line(nb(quote_weird_name("bad  ")))
        cat_line(nb(quote_weird_name("  bad  ")))
      })
    Output
      'good'
      '  bad'
      'bad  '
      '  bad  '

# quoting weird names, still [ansi]

    Code
      local({
        cat_line(nb(quote_weird_name("good")))
        cat_line(nb(quote_weird_name("  bad")))
        cat_line(nb(quote_weird_name("bad  ")))
        cat_line(nb(quote_weird_name("  bad  ")))
      })
    Output
      good
      '[44m  [49mbad'
      'bad[44m  [49m'
      '[44m  [49mbad[44m  [49m'

# ~/ files are not weird [ansi]

    Code
      local({
        cat_line(nb(quote_weird_name("~/good")))
        cat_line(nb(quote_weird_name("~~bad")))
        cat_line(nb(quote_weird_name("bad~  ")))
        cat_line(nb(quote_weird_name(" ~ bad ~ ")))
      })
    Output
      ~/good
      '~~bad'
      'bad~[44m  [49m'
      '[44m [49m~ bad ~[44m [49m'

# custom truncation [plain]

    Code
      x <- cli_vec(1:100, list(vec_trunc = 5))
      cli_text("Some numbers: {x}.")
    Message <cliMessage>
      Some numbers: 1, 2, 3, 4, 5, ....
    Code
      cli_text("Some numbers: {.val {x}}.")
    Message <cliMessage>
      Some numbers: 1, 2, 3, 4, 5, ....

# custom truncation [ansi]

    Code
      x <- cli_vec(1:100, list(vec_trunc = 5))
      cli_text("Some numbers: {x}.")
    Message <cliMessage>
      Some numbers: 1, 2, 3, 4, 5, ....
    Code
      cli_text("Some numbers: {.val {x}}.")
    Message <cliMessage>
      Some numbers: [34m1[39m, [34m2[39m, [34m3[39m, [34m4[39m, [34m5[39m, ....

# custom truncation [unicode]

    Code
      x <- cli_vec(1:100, list(vec_trunc = 5))
      cli_text("Some numbers: {x}.")
    Message <cliMessage>
      Some numbers: 1, 2, 3, 4, 5, ….
    Code
      cli_text("Some numbers: {.val {x}}.")
    Message <cliMessage>
      Some numbers: 1, 2, 3, 4, 5, ….

# custom truncation [fancy]

    Code
      x <- cli_vec(1:100, list(vec_trunc = 5))
      cli_text("Some numbers: {x}.")
    Message <cliMessage>
      Some numbers: 1, 2, 3, 4, 5, ….
    Code
      cli_text("Some numbers: {.val {x}}.")
    Message <cliMessage>
      Some numbers: [34m1[39m, [34m2[39m, [34m3[39m, [34m4[39m, [34m5[39m, ….

# collapsing class names [plain]

    Code
      local({
        cc <- c("one", "two")
        cli_text("this is a class: {.cls myclass}")
        cli_text("multiple classes: {.cls {cc}}")
      })
    Message <cliMessage>
      this is a class: <myclass>
      multiple classes: <one/two>

# collapsing class names [ansi]

    Code
      local({
        cc <- c("one", "two")
        cli_text("this is a class: {.cls myclass}")
        cli_text("multiple classes: {.cls {cc}}")
      })
    Message <cliMessage>
      this is a class: [34m<myclass>[39m
      multiple classes: [34m<one/two>[39m

# transform [plain]

    Code
      local({
        cli_text("This is a {.field field} (before)")
        foo <- (function(x) toupper(x))
        cli_div(theme = list(span.field = list(transform = foo)))
        cli_text("This is a {.field field} (during)")
        cli_end()
        cli_text("This is a {.field field} (after)")
      })
    Message <cliMessage>
      This is a field (before)
      This is a FIELD (during)
      This is a field (after)

# transform [ansi]

    Code
      local({
        cli_text("This is a {.field field} (before)")
        foo <- (function(x) toupper(x))
        cli_div(theme = list(span.field = list(transform = foo)))
        cli_text("This is a {.field field} (during)")
        cli_end()
        cli_text("This is a {.field field} (after)")
      })
    Message <cliMessage>
      This is a [32mfield[39m (before)
      This is a [32mFIELD[39m (during)
      This is a [32mfield[39m (after)

# cli_format

    Code
      cli_format(1:4 / 7, list(digits = 2))
    Output
      [1] 0.14 0.29 0.43 0.57

# cli_format() is used for .val

    Code
      cli_div(theme = list(.val = list(digits = 2)))
      cli_text("Some random numbers: {.val {runif(4)}}.")
    Message <cliMessage>
      Some random numbers: 0.91, 0.94, 0.29, and 0.83.

# .q always double quotes

    Code
      cli_text("just a {.q string}, nothing more")
    Message <cliMessage>
      just a "string", nothing more

# .or

    Code
      cli_text("{.or {letters[1:5]}}")
    Message <cliMessage>
      a, b, c, d, or e

---

    Code
      cli_text("{.or {letters[1:2]}}")
    Message <cliMessage>
      a or b

# line breaks

    Code
      ansi_strwrap(txt2, width = 60)
    Output
      <cli_ansi_string>
      [1] Cupidatat deserunt culpa enim deserunt minim aliqua tempor
      [2] fugiat cupidatat laboris officia esse ex aliqua. Ullamco  
      [3] mollit adipisicing anim.                                  
      [4] Cupidatat deserunt culpa enim deserunt minim aliqua tempor
      [5] fugiat cupidatat laboris officia esse ex aliqua. Ullamco  
      [6] mollit adipisicing anim.                                  

# double ticks [ansi]

    Code
      format_inline("{.code {x}}")
    Output
      [1] "\033[31m`a`\033[39m, \033[31m`` `x` ``\033[39m, and \033[31m`b`\033[39m"

---

    Code
      format_inline("{.fun {x}}")
    Output
      [1] "\033[31m`a()`\033[39m, \033[31m`` `x` ()``\033[39m, and \033[31m`b()`\033[39m"

# ansi_columns

    foo 1     foo 2     foo 3     foo 4     
    foo 5     foo 6     foo 7     foo 8     
    foo 9     foo 10                        

---

    123456789012...

# ansi_toupper [plain]

    Code
      local({
        cat_line(x)
        cat_line(ansi_toupper(x))
      })
    Output
      Red normal green
      RED NORMAL GREEN

# ansi_toupper [ansi]

    Code
      local({
        cat_line(x)
        cat_line(ansi_toupper(x))
      })
    Output
      [31mRed [39mnormal [1m[32mgreen[39m[22m
      [31mRED [39mNORMAL [1m[32mGREEN[39m[22m

# ansi_tolower [plain]

    Code
      local({
        cat_line(x)
        cat_line(ansi_tolower(x))
      })
    Output
      Red NORMAL grEeN
      red normal green

# ansi_tolower [ansi]

    Code
      local({
        cat_line(x)
        cat_line(ansi_tolower(x))
      })
    Output
      [31mRed [39mNORMAL [1m[32mgrEeN[39m[22m
      [31mred [39mnormal [1m[32mgreen[39m[22m

# ansi_chartr [plain]

    Code
      local({
        cat_line(x)
        cat_line(ansi_chartr(" R_", "-r*", x))
      })
    Output
      Red normal green
      red-normal-green

# ansi_chartr [ansi]

    Code
      local({
        cat_line(x)
        cat_line(ansi_chartr(" R_", "-r*", x))
      })
    Output
      [31mRed [39mnormal [1m[32mgreen[39m[22m
      [31mred-[39mnormal-[1m[32mgreen[39m[22m

# cli_progress_demo

    Code
      out
    Output
      \ 50 done (100/s) | 3ms[K

---

    Code
      out
    Output
      ==========>--------------------   33% | ETA:  1s[K
      ====================>----------   67% | ETA:  1s[K
      ==============================>  100% | ETA:  1s[K
      [K

---

    Code
      out
    Output
      \ 1 done (100/s) | 3ms[K
      | 2 done (100/s) | 3ms[K
      / 3 done (100/s) | 3ms[K
      - 4 done (100/s) | 3ms[K
      \ 5 done (100/s) | 3ms[K
      [K

---

    Code
      out
    Output
      ==========>--------------------   33% | ETA: 10s[K
      ====================>----------   67% | ETA:  3s[K
      ==============================>  100% | ETA:  0s[K
      [K

---

    Code
      msgs
    Output
      [1] "\r==========>--------------------   33% | ETA: 10s\033[K\r"
      [2] "\r====================>----------   67% | ETA:  3s\033[K\r"
      [3] "\r==============================>  100% | ETA:  0s\033[K\r"
      [4] "\r\033[K"                                                  

# pb_bar

    Code
      cli_text("-{cli::pb_bar}-")
    Message <cliMessage>
      --

---

    Code
      cli_text("-{cli::pb_bar}-")
    Message <cliMessage>
      -===============>--------------- -

# pb_current_bytes

    Code
      cli__pb_current_bytes(list(current = 0))
    Output
      [1] "0.0 kB"
    Code
      cli__pb_current_bytes(list(current = 1))
    Output
      [1] "0.0 kB"
    Code
      cli__pb_current_bytes(list(current = 1000))
    Output
      [1] "1.0 kB"
    Code
      cli__pb_current_bytes(list(current = 1000 * 23))
    Output
      [1] " 23 kB"
    Code
      cli__pb_current_bytes(list(current = 1000 * 1000 * 23))
    Output
      [1] " 23 MB"

# pb_elapsed

    Code
      cli__pb_elapsed()
    Output
      [1] "1s"

---

    Code
      cli__pb_elapsed()
    Output
      [1] "21s"

---

    Code
      cli__pb_elapsed()
    Output
      [1] "58s"

---

    Code
      cli__pb_elapsed()
    Output
      [1] "1h 5m"

# pb_elapsed_clock

    Code
      cli__pb_elapsed_clock()
    Output
      [1] "00:00:01"

---

    Code
      cli__pb_elapsed_clock()
    Output
      [1] "00:00:21"

---

    Code
      cli__pb_elapsed_clock()
    Output
      [1] "00:00:58"

---

    Code
      cli__pb_elapsed_clock()
    Output
      [1] "01:05:00"

# pb_elapsed_raw

    Code
      cli__pb_elapsed_raw()
    Output
      [1] 1

---

    Code
      cli__pb_elapsed_raw()
    Output
      [1] 21

---

    Code
      cli__pb_elapsed_raw()
    Output
      [1] 58

---

    Code
      cli__pb_elapsed_raw()
    Output
      [1] 3900

# pb_eta

    Code
      cli__pb_eta(list())
    Output
      [1] "?"

---

    Code
      cli__pb_eta(list())
    Output
      [1] "12s"

# pb_eta_raw

    Code
      cli__pb_eta_raw()
    Output
      [1] NA

---

    Code
      cli__pb_eta_raw()
    Output
      Time difference of 40 secs

---

    Code
      cli__pb_eta_raw()
    Output
      Time difference of 10 secs

---

    Code
      cli__pb_eta_raw()
    Output
      Time difference of 0 secs

# pb_eta_str

    Code
      cli__pb_eta_str(list())
    Output
      [1] ""

---

    Code
      cli__pb_eta_str(list())
    Output
      [1] "ETA:  1s"

# pb_percent

    Code
      cli__pb_percent(list(current = 0, total = 99))
    Output
      [1] "  0%"
    Code
      cli__pb_percent(list(current = 5, total = 99))
    Output
      [1] "  5%"
    Code
      cli__pb_percent(list(current = 10, total = 99))
    Output
      [1] " 10%"
    Code
      cli__pb_percent(list(current = 25, total = 99))
    Output
      [1] " 25%"
    Code
      cli__pb_percent(list(current = 99, total = 99))
    Output
      [1] "100%"
    Code
      cli__pb_percent(list(current = 100, total = 99))
    Output
      [1] "101%"

# pb_rate

    Code
      cli__pb_rate(list())
    Output
      [1] "?/s"

---

    Code
      cli__pb_rate(list())
    Output
      [1] "?/s"

---

    Code
      cli__pb_rate(list())
    Output
      [1] "0.1/s"

---

    Code
      cli__pb_rate(list())
    Output
      [1] "12/s"

# pb_rate_bytes

    Code
      cli__pb_rate_bytes(list())
    Output
      [1] "NaN kB/s"

---

    Code
      cli__pb_rate_bytes(list())
    Output
      [1] "Inf YB/s"

---

    Code
      cli__pb_rate_bytes(list())
    Output
      [1] "0.0 kB/s"

---

    Code
      cli__pb_rate_bytes(list())
    Output
      [1] "1.0 kB/s"

---

    Code
      cli__pb_rate_bytes(list())
    Output
      [1] " 24 kB/s"

---

    Code
      cli__pb_rate_bytes(list())
    Output
      [1] " 24 MB/s"

# pb_spin

    Code
      cli_text("-{cli::pb_spin}-{cli::pb_spin}-")
    Message <cliMessage>
      -\-\-

---

    Code
      cli_text("-{cli::pb_spin}-{cli::pb_spin}-")
    Message <cliMessage>
      -|-|-

---

    Code
      cli_text("-{cli::pb_spin}-{cli::pb_spin}-")
    Message <cliMessage>
      -|-|-

# pb_timestamp

    Code
      cli__pb_timestamp(list())
    Output
      [1] "2021-06-18T00:09:14+00:00"

---

    Code
      cli__pb_timestamp(list())
    Output
      [1] "2021-06-18T00:09:34+00:00"

# pb_total_bytes

    Code
      cli__pb_total_bytes(list(total = 0))
    Output
      [1] "0.0 kB"
    Code
      cli__pb_total_bytes(list(total = 1))
    Output
      [1] "0.0 kB"
    Code
      cli__pb_total_bytes(list(total = 1000))
    Output
      [1] "1.0 kB"
    Code
      cli__pb_total_bytes(list(total = 1000 * 23))
    Output
      [1] " 23 kB"
    Code
      cli__pb_total_bytes(list(total = 1000 * 1000 * 23))
    Output
      [1] " 23 MB"

# cli_abort [plain]

    Code
      local({
        n <- "boo"
        cli_abort(c("{.var n} must be a numeric vector", x = "You've supplied a {.cls {class(n)}} vector."))
      })
    Error <rlang_error>
      `n` must be a numeric vector
      x You've supplied a <character> vector.

---

    Code
      local({
        len <- 26
        idx <- 100
        cli_abort(c("Must index an existing element:", i = "There {?is/are} {len} element{?s}.",
          x = "You've tried to subset element {idx}."))
      })
    Error <rlang_error>
      Must index an existing element:
      i There are 26 elements.
      x You've tried to subset element 100.

---

    Code
      c(err$message, err$body)
    Output
                                                                                    x 
               "`n` must be a numeric vector" "You've supplied a <character> vector." 

# cli_abort [ansi]

    Code
      local({
        n <- "boo"
        cli_abort(c("{.var n} must be a numeric vector", x = "You've supplied a {.cls {class(n)}} vector."))
      })
    Error <rlang_error>
      [1m[22m`n` must be a numeric vector
      [31mx[39m You've supplied a [34m<character>[39m vector.

---

    Code
      local({
        len <- 26
        idx <- 100
        cli_abort(c("Must index an existing element:", i = "There {?is/are} {len} element{?s}.",
          x = "You've tried to subset element {idx}."))
      })
    Error <rlang_error>
      [1m[22mMust index an existing element:
      [36mi[39m There are 26 elements.
      [31mx[39m You've tried to subset element 100.

---

    Code
      c(err$message, err$body)
    Output
                                                              
                               "`n` must be a numeric vector" 
                                                            x 
      "You've supplied a \033[34m<character>\033[39m vector." 

# cli_abort [unicode]

    Code
      local({
        n <- "boo"
        cli_abort(c("{.var n} must be a numeric vector", x = "You've supplied a {.cls {class(n)}} vector."))
      })
    Error <rlang_error>
      `n` must be a numeric vector
      ✖ You've supplied a <character> vector.

---

    Code
      local({
        len <- 26
        idx <- 100
        cli_abort(c("Must index an existing element:", i = "There {?is/are} {len} element{?s}.",
          x = "You've tried to subset element {idx}."))
      })
    Error <rlang_error>
      Must index an existing element:
      ℹ There are 26 elements.
      ✖ You've tried to subset element 100.

---

    Code
      c(err$message, err$body)
    Output
                                                                                    x 
               "`n` must be a numeric vector" "You've supplied a <character> vector." 

# cli_abort [fancy]

    Code
      local({
        n <- "boo"
        cli_abort(c("{.var n} must be a numeric vector", x = "You've supplied a {.cls {class(n)}} vector."))
      })
    Error <rlang_error>
      [1m[22m`n` must be a numeric vector
      [31m✖[39m You've supplied a [34m<character>[39m vector.

---

    Code
      local({
        len <- 26
        idx <- 100
        cli_abort(c("Must index an existing element:", i = "There {?is/are} {len} element{?s}.",
          x = "You've tried to subset element {idx}."))
      })
    Error <rlang_error>
      [1m[22mMust index an existing element:
      [36mℹ[39m There are 26 elements.
      [31m✖[39m You've tried to subset element 100.

---

    Code
      c(err$message, err$body)
    Output
                                                              
                               "`n` must be a numeric vector" 
                                                            x 
      "You've supplied a \033[34m<character>\033[39m vector." 

# cli_warn [plain]

    Code
      n <- "boo"
      cli_warn(c("{.var n} must be a numeric vector", x = "You've supplied a {.cls {class(n)}} vector."))
    Warning <rlang_warning>
      `n` must be a numeric vector
      x You've supplied a <character> vector.

---

    Code
      local({
        len <- 26
        idx <- 100
        cli_warn(c("Must index an existing element:", i = "There {?is/are} {len} element{?s}.",
          x = "You've tried to subset element {idx}."))
      })
    Warning <rlang_warning>
      Must index an existing element:
      i There are 26 elements.
      x You've tried to subset element 100.

---

    Code
      wrn$cli_bullets
    Output
                                                                                i 
          "Must index an existing element:"              "There are 26 elements." 
                                          x 
      "You've tried to subset element 100." 

# cli_warn [ansi]

    Code
      n <- "boo"
      cli_warn(c("{.var n} must be a numeric vector", x = "You've supplied a {.cls {class(n)}} vector."))
    Warning <rlang_warning>
      [1m[22m`n` must be a numeric vector
      [31mx[39m You've supplied a [34m<character>[39m vector.

---

    Code
      local({
        len <- 26
        idx <- 100
        cli_warn(c("Must index an existing element:", i = "There {?is/are} {len} element{?s}.",
          x = "You've tried to subset element {idx}."))
      })
    Warning <rlang_warning>
      [1m[22mMust index an existing element:
      [36mi[39m There are 26 elements.
      [31mx[39m You've tried to subset element 100.

---

    Code
      wrn$cli_bullets
    Output
                                                                                i 
          "Must index an existing element:"              "There are 26 elements." 
                                          x 
      "You've tried to subset element 100." 

# cli_warn [unicode]

    Code
      n <- "boo"
      cli_warn(c("{.var n} must be a numeric vector", x = "You've supplied a {.cls {class(n)}} vector."))
    Warning <rlang_warning>
      `n` must be a numeric vector
      ✖ You've supplied a <character> vector.

---

    Code
      local({
        len <- 26
        idx <- 100
        cli_warn(c("Must index an existing element:", i = "There {?is/are} {len} element{?s}.",
          x = "You've tried to subset element {idx}."))
      })
    Warning <rlang_warning>
      Must index an existing element:
      ℹ There are 26 elements.
      ✖ You've tried to subset element 100.

---

    Code
      wrn$cli_bullets
    Output
                                                                                i 
          "Must index an existing element:"              "There are 26 elements." 
                                          x 
      "You've tried to subset element 100." 

# cli_warn [fancy]

    Code
      n <- "boo"
      cli_warn(c("{.var n} must be a numeric vector", x = "You've supplied a {.cls {class(n)}} vector."))
    Warning <rlang_warning>
      [1m[22m`n` must be a numeric vector
      [31m✖[39m You've supplied a [34m<character>[39m vector.

---

    Code
      local({
        len <- 26
        idx <- 100
        cli_warn(c("Must index an existing element:", i = "There {?is/are} {len} element{?s}.",
          x = "You've tried to subset element {idx}."))
      })
    Warning <rlang_warning>
      [1m[22mMust index an existing element:
      [36mℹ[39m There are 26 elements.
      [31m✖[39m You've tried to subset element 100.

---

    Code
      wrn$cli_bullets
    Output
                                                                                i 
          "Must index an existing element:"              "There are 26 elements." 
                                          x 
      "You've tried to subset element 100." 

# cli_inform [plain]

    Code
      n <- "boo"
      cli_inform(c("{.var n} must be a numeric vector", x = "You've supplied a {.cls {class(n)}} vector."))
    Message <cliMessage>
      
    Message <rlang_message>
      `n` must be a numeric vector
      x You've supplied a <character> vector.

---

    Code
      local({
        len <- 26
        idx <- 100
        cli_inform(c("Must index an existing element:", i = "There {?is/are} {len} element{?s}.",
          x = "You've tried to subset element {idx}."))
      })
    Message <cliMessage>
      
    Message <rlang_message>
      Must index an existing element:
      i There are 26 elements.
      x You've tried to subset element 100.

---

    Code
      tail(inf, 1)[[1]]$cli_bullets
    Output
                                                                                i 
          "Must index an existing element:"              "There are 26 elements." 
                                          x 
      "You've tried to subset element 100." 

# cli_inform [ansi]

    Code
      n <- "boo"
      cli_inform(c("{.var n} must be a numeric vector", x = "You've supplied a {.cls {class(n)}} vector."))
    Message <cliMessage>
      
    Message <rlang_message>
      [1m[22m`n` must be a numeric vector
      [31mx[39m You've supplied a [34m<character>[39m vector.

---

    Code
      local({
        len <- 26
        idx <- 100
        cli_inform(c("Must index an existing element:", i = "There {?is/are} {len} element{?s}.",
          x = "You've tried to subset element {idx}."))
      })
    Message <cliMessage>
      
    Message <rlang_message>
      [1m[22mMust index an existing element:
      [36mi[39m There are 26 elements.
      [31mx[39m You've tried to subset element 100.

---

    Code
      tail(inf, 1)[[1]]$cli_bullets
    Output
                                                                                i 
          "Must index an existing element:"              "There are 26 elements." 
                                          x 
      "You've tried to subset element 100." 

# cli_inform [unicode]

    Code
      n <- "boo"
      cli_inform(c("{.var n} must be a numeric vector", x = "You've supplied a {.cls {class(n)}} vector."))
    Message <cliMessage>
      
    Message <rlang_message>
      `n` must be a numeric vector
      ✖ You've supplied a <character> vector.

---

    Code
      local({
        len <- 26
        idx <- 100
        cli_inform(c("Must index an existing element:", i = "There {?is/are} {len} element{?s}.",
          x = "You've tried to subset element {idx}."))
      })
    Message <cliMessage>
      
    Message <rlang_message>
      Must index an existing element:
      ℹ There are 26 elements.
      ✖ You've tried to subset element 100.

---

    Code
      tail(inf, 1)[[1]]$cli_bullets
    Output
                                                                                i 
          "Must index an existing element:"              "There are 26 elements." 
                                          x 
      "You've tried to subset element 100." 

# cli_inform [fancy]

    Code
      n <- "boo"
      cli_inform(c("{.var n} must be a numeric vector", x = "You've supplied a {.cls {class(n)}} vector."))
    Message <cliMessage>
      
    Message <rlang_message>
      [1m[22m`n` must be a numeric vector
      [31m✖[39m You've supplied a [34m<character>[39m vector.

---

    Code
      local({
        len <- 26
        idx <- 100
        cli_inform(c("Must index an existing element:", i = "There {?is/are} {len} element{?s}.",
          x = "You've tried to subset element {idx}."))
      })
    Message <cliMessage>
      
    Message <rlang_message>
      [1m[22mMust index an existing element:
      [36mℹ[39m There are 26 elements.
      [31m✖[39m You've tried to subset element 100.

---

    Code
      tail(inf, 1)[[1]]$cli_bullets
    Output
                                                                                i 
          "Must index an existing element:"              "There are 26 elements." 
                                          x 
      "You've tried to subset element 100." 

# cli_abort width in RStudio

    Code
      local({
        len <- 26
        idx <- 100
        cli_abort(c(lorem_ipsum(1, 3), i = lorem_ipsum(1, 3), x = lorem_ipsum(1, 3)))
      })
    Error <rlang_error>
      Duis quis magna incididunt nulla commodo minim non exercitation nostrud
      ullamco dolor exercitation ut veniam. Fugiat irure tempor commodo voluptate ut.
      In et tempor excepteur quis.
      i Et nisi ad quis ad cupidatat tempor laborum est excepteur aliqua veniam ex.
        Sunt magna veniam Lorem elit enim et pariatur aliqua occaecat mollit
        consequat dolore in mollit. Officia labore reprehenderit culpa dolore quis
        nisi do aliqua commodo deserunt fugiat cupidatat nostrud ad.
      x Ad laboris consectetur esse minim pariatur irure do anim anim. Mollit ad
        cupidatat ullamco ullamco nulla elit in.

# color in RStudio [ansi]

    Code
      col <- get_rstudio_fg_color0()
      cat(col("this is the new color"))
    Output
      [30mthis is the new color[39m

# update_rstudio_color [ansi]

    Code
      cat(update_rstudio_color("color me interested"))
    Output
      [32mcolor me interested[39m

# cli_abort() captures correct call and backtrace

    Code
      print(expect_error(f()))
    Output
      <error/rlang_error>
      Error in `h()`:
      ! foo
      Backtrace:
        1. base::print(expect_error(f()))
        8. cli f()
        9. cli g()
       10. cli h()

---

    Code
      print(expect_error(f(list())))
    Output
      <error/cli_my_class>
      Error in `h()`:
      ! `x` can't be empty.
      Backtrace:
        1. base::print(expect_error(f(list())))
        8. cli f(list())
        9. cli g(x)
       10. cli h(x)

# verbatim text is correctly styled

    Code
      local({
        theme <- list(.padded = list(`margin-left` = 4))
        cli_div(class = "padded", theme = theme)
        lines <- c("first", "second", "third")
        cli_verbatim(lines)
        cli_verbatim(paste0(lines, collapse = "\n"))
      })
    Message <cliMessage>
          first
          second
          third
          first
          second
          third

# inline classes [plain]

    Code
      invisible(lapply(classes, do))
    Message <cliMessage>
      This is <<<it>>> really
      This is <<<it>>> really
      This is <<<it>>> really
      This is <<<it>>> really
      This is <<<it>>> really
      This is <<<it>>> really
      This is <<<it>>> really
      This is <<<it>>> really
      This is <<<it>>> really
      This is <<<it>>> really
      This is <<<it>>> really
      This is <<<it>>> really
      This is <<<it>>> really
      This is <<<it>>> really

# inline classes [ansi]

    Code
      invisible(lapply(classes, do))
    Message <cliMessage>
      This is [36m<<<it>>>[39m really
      This is [36m<<<it>>>[39m really
      This is [36m<<<it>>>[39m really
      This is [36m<<<it>>>[39m really
      This is [36m<<<it>>>[39m really
      This is [36m<<<it>>>[39m really
      This is [36m<<<it>>>[39m really
      This is [36m<<<it>>>[39m really
      This is [36m<<<it>>>[39m really
      This is [36m<<<it>>>[39m really
      This is [36m<<<it>>>[39m really
      This is [36m<<<it>>>[39m really
      This is [36m<<<it>>>[39m really
      This is [36m<<<it>>>[39m really

# {{ and }} can be used for comments

    Code
      local({
        cli_text("Escaping {{ works")
        cli_text("Escaping }} works")
        cli_text("Escaping {{ and }} works")
        cli_text("Escaping {{{{ works")
        cli_text("Escaping }}}} works")
        cli_text("Escaping {{{{ and }} works")
        cli_text("Escaping {{{{ and }}}} works")
        cli_text("Escaping {{ and }}}} works")
      })
    Message <cliMessage>
      Escaping { works
      Escaping } works
      Escaping { and } works
      Escaping {{ works
      Escaping }} works
      Escaping {{ and } works
      Escaping {{ and }} works
      Escaping { and }} works

# no glue substitution in expressions that evaluate to a string

    Code
      local({
        msg <- "Message with special characters like } { }} {{"
        cli_text("{msg}")
        cli_text("{.emph {msg}}")
      })
    Message <cliMessage>
      Message with special characters like } { }} {{
      Message with special characters like } { }} {{

# S3 class is used for styling

    Code
      local({
        cli_div(theme = list(div = list(`class-map` = list(foo = "bar")), .bar = list(
          before = "::")))
        obj <- structure("yep", class = "foo")
        cli_text("This is {obj}.")
      })
    Message <cliMessage>
      This is ::yep.

---
title: Deprecated Docker Engine features
linkTitle: Deprecated features
aliases: ["/engine/misc/deprecated/"]
description: "Deprecated Features."
keywords: "docker, documentation, about, technology, deprecate"
---

<!-- This file is maintained within the docker/cli GitHub
     repository at https://github.com/docker/cli/. Make all
     pull requests against that repo. If you see this file in
     another repository, consider it read-only there, as it will
     periodically be overwritten by the definitive file. Pull
     requests which include edits to this file in other repositories
     will be rejected.
-->

This page provides an overview of features that are deprecated in Engine. Changes
in packaging, and supported (Linux) distributions are not included. To learn
about end of support for Linux distributions, refer to the
[release notes](https://docs.docker.com/engine/release-notes/).

## Feature deprecation policy

As changes are made to Docker there may be times when existing features need to
be removed or replaced with newer features. Before an existing feature is removed
it is labeled as "deprecated" within the documentation and remains in Docker for
at least one stable release unless specified explicitly otherwise. After that time
it may be removed.

Users are expected to take note of the list of deprecated features each release
and plan their migration away from those features, and (if applicable) towards
the replacement features as soon as possible.

## Deprecated engine features

The following table provides an overview of the current status of deprecated features:

- **Deprecated**: the feature is marked "deprecated" and should no longer be used.

  The feature may be removed, disabled, or change behavior in a future release.
  The _"Deprecated"_ column contains the release in which the feature was marked
  deprecated, whereas the _"Remove"_ column contains a tentative release in which
  the feature is to be removed. If no release is included in the _"Remove"_ column,
  the release is yet to be decided on.

- **Removed**: the feature was removed, disabled, or hidden.

  Refer to the linked section for details. Some features are "soft" deprecated,
  which means that they remain functional for backward compatibility, and to
  allow users to migrate to alternatives. In such cases, a warning may be
  printed, and users should not rely on this feature.

| Status     | Feature                                                                                                                            | Deprecated | Remove |
|------------|------------------------------------------------------------------------------------------------------------------------------------|------------|--------|
| Deprecated | [Legacy links environment variables](#legacy-links-environment-variables)                                                          | v28.4      | v30.0  |
| Deprecated | [Special handling for quoted values for TLS flags](#special-handling-for-quoted-values-for-tls-flags)                              | v28.4      | v29.0  |
| Deprecated | [Empty/nil fields in image Config from inspect API](#emptynil-fields-in-image-config-from-inspect-api)                             | v28.3      | v29.0  |
| Deprecated | [Configuration for pushing  non-distributable artifacts](#configuration-for-pushing-non-distributable-artifacts)                   | v28.0      | v29.0  |
| Deprecated | [`--time` option on `docker stop` and `docker restart`](#--time-option-on-docker-stop-and-docker-restart)                          | v28.0      | -      |
| Removed    | [Non-standard fields in image inspect](#non-standard-fields-in-image-inspect)                                                      | v27.0      | v28.2  |
| Removed    | [API CORS headers](#api-cors-headers)                                                                                              | v27.0      | v28.0  |
| Removed    | [Graphdriver plugins (experimental)](#graphdriver-plugins-experimental)                                                            | v27.0      | v28.0  |
| Deprecated | [Unauthenticated TCP connections](#unauthenticated-tcp-connections)                                                                | v26.0      | v28.0  |
| Removed    | [`Container` and `ContainerConfig` fields in Image inspect](#container-and-containerconfig-fields-in-image-inspect)                | v25.0      | v26.0  |
| Removed    | [Deprecate legacy API versions](#deprecate-legacy-api-versions)                                                                    | v25.0      | v26.0  |
| Removed    | [Container short ID in network Aliases field](#container-short-id-in-network-aliases-field)                                        | v25.0      | v26.0  |
| Deprecated | [Mount `bind-nonrecursive` option](#mount-bind-nonrecursive-option)                                                                | v25.0      | v29.0  |
| Removed    | [IsAutomated field, and `is-automated` filter on `docker search`](#isautomated-field-and-is-automated-filter-on-docker-search)     | v25.0      | v28.2  |
| Removed    | [logentries logging driver](#logentries-logging-driver)                                                                            | v24.0      | v25.0  |
| Removed    | [OOM-score adjust for the daemon](#oom-score-adjust-for-the-daemon)                                                                | v24.0      | v25.0  |
| Removed    | [BuildKit build information](#buildkit-build-information)                                                                          | v23.0      | v24.0  |
| Deprecated | [Legacy builder for Linux images](#legacy-builder-for-linux-images)                                                                | v23.0      | -      |
| Deprecated | [Legacy builder fallback](#legacy-builder-fallback)                                                                                | v23.0      | -      |
| Removed    | [Btrfs storage driver on CentOS 7 and RHEL 7](#btrfs-storage-driver-on-centos-7-and-rhel-7)                                        | v20.10     | v23.0  |
| Removed    | [Support for encrypted TLS private keys](#support-for-encrypted-tls-private-keys)                                                  | v20.10     | v23.0  |
| Removed    | [Kubernetes stack and context support](#kubernetes-stack-and-context-support)                                                      | v20.10     | v23.0  |
| Removed    | [Pulling images from non-compliant image registries](#pulling-images-from-non-compliant-image-registries)                          | v20.10     | v28.2  |
| Removed    | [Linux containers on Windows (LCOW)](#linux-containers-on-windows-lcow-experimental)                                               | v20.10     | v23.0  |
| Deprecated | [BLKIO weight options with cgroups v1](#blkio-weight-options-with-cgroups-v1)                                                      | v20.10     | -      |
| Removed    | [Kernel memory limit](#kernel-memory-limit)                                                                                        | v20.10     | v23.0  |
| Removed    | [Classic Swarm and overlay networks using external key/value stores](#classic-swarm-and-overlay-networks-using-cluster-store)      | v20.10     | v23.0  |
| Removed    | [Support for the legacy `~/.dockercfg` configuration file for authentication](#support-for-legacy-dockercfg-configuration-files)   | v20.10     | v23.0  |
| Deprecated | [CLI plugins support](#cli-plugins-support)                                                                                        | v20.10     | -      |
| Deprecated | [Dockerfile legacy `ENV name value` syntax](#dockerfile-legacy-env-name-value-syntax)                                              | v20.10     | -      |
| Removed    | [`docker build --stream` flag (experimental)](#docker-build---stream-flag-experimental)                                            | v20.10     | v20.10 |
| Removed    | [`fluentd-async-connect` log opt](#fluentd-async-connect-log-opt)                                                                  | v20.10     | v28.0  |
| Removed    | [Configuration options for experimental CLI features](#configuration-options-for-experimental-cli-features)                        | v19.03     | v23.0  |
| Removed    | [Pushing and pulling with image manifest v2 schema 1](#pushing-and-pulling-with-image-manifest-v2-schema-1)                        | v19.03     | v28.2  |
| Removed    | [`docker engine` subcommands](#docker-engine-subcommands)                                                                          | v19.03     | v20.10 |
| Removed    | [Top-level `docker deploy` subcommand (experimental)](#top-level-docker-deploy-subcommand-experimental)                            | v19.03     | v20.10 |
| Removed    | [`docker stack deploy` using "dab" files (experimental)](#docker-stack-deploy-using-dab-files-experimental)                        | v19.03     | v20.10 |
| Removed    | [Support for the `overlay2.override_kernel_check` storage option](#support-for-the-overlay2override_kernel_check-storage-option)   | v19.03     | v24.0  |
| Removed    | [AuFS storage driver](#aufs-storage-driver)                                                                                        | v19.03     | v24.0  |
| Removed    | [Legacy "overlay" storage driver](#legacy-overlay-storage-driver)                                                                  | v18.09     | v24.0  |
| Removed    | [Device mapper storage driver](#device-mapper-storage-driver)                                                                      | v18.09     | v25.0  |
| Removed    | [Use of reserved namespaces in engine labels](#use-of-reserved-namespaces-in-engine-labels)                                        | v18.06     | v20.10 |
| Removed    | [`--disable-legacy-registry` override daemon option](#--disable-legacy-registry-override-daemon-option)                            | v17.12     | v19.03 |
| Removed    | [Interacting with V1 registries](#interacting-with-v1-registries)                                                                  | v17.06     | v17.12 |
| Removed    | [Asynchronous `service create` and `service update` as default](#asynchronous-service-create-and-service-update-as-default)        | v17.05     | v17.10 |
| Removed    | [`-g` and `--graph` flags on `dockerd`](#-g-and---graph-flags-on-dockerd)                                                          | v17.05     | v23.0  |
| Deprecated | [Top-level network properties in NetworkSettings](#top-level-network-properties-in-networksettings)                                | v1.13      | v17.12 |
| Removed    | [`filter` option for `/images/json` endpoint](#filter-option-for-imagesjson-endpoint)                                              | v1.13      | v20.10 |
| Removed    | [`repository:shortid` image references](#repositoryshortid-image-references)                                                       | v1.13      | v17.12 |
| Removed    | [`docker daemon` subcommand](#docker-daemon-subcommand)                                                                            | v1.13      | v17.12 |
| Removed    | [Duplicate keys with conflicting values in engine labels](#duplicate-keys-with-conflicting-values-in-engine-labels)                | v1.13      | v17.12 |
| Deprecated | [`MAINTAINER` in Dockerfile](#maintainer-in-dockerfile)                                                                            | v1.13      | -      |
| Deprecated | [API calls without a version](#api-calls-without-a-version)                                                                        | v1.13      | v17.12 |
| Removed    | [Backing filesystem without `d_type` support for overlay/overlay2](#backing-filesystem-without-d_type-support-for-overlayoverlay2) | v1.13      | v17.12 |
| Removed    | [`--automated` and `--stars` flags on `docker search`](#--automated-and---stars-flags-on-docker-search)                            | v1.12      | v20.10 |
| Deprecated | [`-h` shorthand for `--help`](#-h-shorthand-for---help)                                                                            | v1.12      | v17.09 |
| Removed    | [`-e` and `--email` flags on `docker login`](#-e-and---email-flags-on-docker-login)                                                | v1.11      | v17.06 |
| Deprecated | [Separator (`:`) of `--security-opt` flag on `docker run`](#separator--of---security-opt-flag-on-docker-run)                       | v1.11      | v17.06 |
| Deprecated | [Ambiguous event fields in API](#ambiguous-event-fields-in-api)                                                                    | v1.10      | -      |
| Removed    | [`-f` flag on `docker tag`](#-f-flag-on-docker-tag)                                                                                | v1.10      | v1.12  |
| Removed    | [HostConfig at API container start](#hostconfig-at-api-container-start)                                                            | v1.10      | v1.12  |
| Removed    | [`--before` and `--since` flags on `docker ps`](#--before-and---since-flags-on-docker-ps)                                          | v1.10      | v1.12  |
| Removed    | [Driver-specific log tags](#driver-specific-log-tags)                                                                              | v1.9       | v1.12  |
| Removed    | [Docker Content Trust `ENV` passphrase variables name change](#docker-content-trust-env-passphrase-variables-name-change)          | v1.9       | v1.12  |
| Removed    | [`/containers/(id or name)/copy` endpoint](#containersid-or-namecopy-endpoint)                                                     | v1.8       | v1.12  |
| Removed    | [LXC built-in exec driver](#lxc-built-in-exec-driver)                                                                              | v1.8       | v1.10  |
| Removed    | [Old Command Line Options](#old-command-line-options)                                                                              | v1.8       | v1.10  |
| Removed    | [`--api-enable-cors` flag on `dockerd`](#--api-enable-cors-flag-on-dockerd)                                                        | v1.6       | v17.09 |
| Removed    | [`--run` flag on `docker commit`](#--run-flag-on-docker-commit)                                                                    | v0.10      | v1.13  |
| Removed    | [Three arguments form in `docker import`](#three-arguments-form-in-docker-import)                                                  | v0.6.7     | v1.12  |

### Legacy links environment variables

**Deprecated in release: v28.4**

**Disabled by default in release: v29.0**

**Target for removal in release: v30.0**

Containers attached to the default bridge network can specify "legacy links" (e.g.
using `--links` on the CLI) to get access to other containers attached to that
network. The linking container (i.e., the container created with `--links`) automatically
gets environment variables that specify the IP address and port mappings of the linked
container. However, these environment variables are prefixed with the linked
container's names, making them impractical.

Starting with Docker v29.0, these environment variables are no longer set by
default. Users who still depend on them can start Docker Engine with the
environment variable `DOCKER_KEEP_DEPRECATED_LEGACY_LINKS_ENV_VARS=1` set.

Support for legacy links environment variables, as well as the `DOCKER_KEEP_DEPRECATED_LEGACY_LINKS_ENV_VARS`
will be removed in Docker Engine v30.0.

### Special handling for quoted values for TLS flags

**Deprecated in release: v28.4**

**Target for removal in release: v29.0**

The `--tlscacert`, `--tlscert`, and `--tlskey` command-line flags had
non-standard behavior for handling values contained in quotes (`"` or `'`).
Normally, quotes are handled by the shell, for example, in the following
example, the shell takes care of handling quotes before passing the values
to the `docker` CLI:

```console
docker --some-option "some-value-in-quotes" ...
```

However, when passing values using an equal sign (`=`), this may not happen
and values may be handled including quotes;

```console
docker --some-option="some-value-in-quotes" ...
```

This caused issues with "Docker Machine", which used this format as part
of its `docker-machine config` output, and the CLI carried special, non-standard
handling for these flags.

Docker Machine reached EOL, and this special handling made the processing
of flag values inconsistent with other flags used, so this behavior is
deprecated. Users depending on this behavior are recommended to specify
the quoted values using a space between the flag and its value, as illustrated
above.

### Empty/nil fields in image Config from inspect API

**Deprecated in release: v28.3**

**Target for removal in release: v29.0**

The `Config` field returned by `docker image inspect` (and the `GET /images/{name}/json`
API endpoint) currently includes certain fields even when they are empty or nil.
Starting in Docker v29.0, the following fields will be omitted from the API response
when they contain empty or default values:

- `Cmd`
- `Entrypoint`
- `Env`
- `Labels`
- `OnBuild`
- `User`
- `Volumes`
- `WorkingDir`

Applications consuming the image inspect API should be updated to handle the
absence of these fields gracefully, treating missing fields as having their
default/empty values.

For API version corresponding to Docker v29.0, these fields will be omitted when
empty. They will continue to be included when using clients that request an older
API version for backward compatibility.

### Configuration for pushing non-distributable artifacts

**Deprecated in release: v28.0**

**Target for removal in release: v29.0**

Non-distributable artifacts (also called foreign layers) were introduced in
docker v1.12 to accommodate Windows images for which the EULA did not allow
layers to be distributed through registries other than those hosted by Microsoft.
The concept of foreign / non-distributable layers was adopted by the OCI distribution
spec in [oci#233]. These restrictions were relaxed later to allow distributing
these images through non-public registries, for which a configuration was added
in Docker v17.0.6.0.

In 2022, Microsoft updated the EULA and [removed these restrictions][msft-3645201],
followed by the OCI distribution specification deprecating foreign layers in [oci#965].
In 2023, Microsoft [removed the use of foreign data layers][msft-3846833] for their images,
making this functionality obsolete.

Docker v28.0 deprecates the `--allow-nondistributable-artifacts` daemon flag and
corresponding `allow-nondistributable-artifacts` field in `daemon.json`. Setting
either option no longer takes an effect, but a deprecation warning log is added
to raise awareness about the deprecation. This warning is planned to become an
error in the Docker v29.0.

Users currently using these options are therefore recommended to remove this
option from their configuration to prevent the daemon from starting when
upgrading to Docker v29.0.

The `AllowNondistributableArtifactsCIDRs` and `AllowNondistributableArtifactsHostnames`
fields in the `RegistryConfig` of the `GET /info` API response are also deprecated.
For API version v1.48 and lower, the fields are still included in the response
but always `null`. In API version v1.49 and higher, the field will be omitted
entirely.

[oci#233]: https://github.com/opencontainers/image-spec/pull/233
[oci#965]: https://github.com/opencontainers/image-spec/pull/965
[msft-3645201]: https://techcommunity.microsoft.com/blog/containers/announcing-windows-container-base-image-redistribution-rights-change/3645201
[msft-3846833]: https://techcommunity.microsoft.com/blog/containers/announcing-removal-of-foreign-layers-from-windows-container-images/3846833

### `--time` option on `docker stop` and `docker restart`

**Deprecated in release: v28.0**

The `--time` option for the `docker stop`, `docker container stop`, `docker restart`,
and `docker container restart` commands has been renamed to `--timeout` for
consistency with other uses of timeout options. The `--time` option is now
deprecated and hidden, but remains functional for backward compatibility.
Users are encouraged to migrate to using the `--timeout` option instead.

### Non-standard fields in image inspect

**Deprecated in release: v27.0**

**Removed in release: v28.2**

The `Config` field returned shown in `docker image inspect` (and as returned by
the `GET /images/{name}/json` API endpoint) returns additional fields that are
not part of the image's configuration and not part of the [Docker image specification]
and [OCI image specification].

These fields are never set (and always return the default value for the type),
but are not omitted in the response when left empty. As these fields were not
intended to be part of the image configuration response, they are deprecated,
and will be removed from the API in thee next release.

The following fields are not part of the underlying image's `Config` field, and
removed in the API response for API v1.50 and newer, corresponding with v28.2.
They continue to be included when using clients that use an older API version:

- `Hostname`
- `Domainname`
- `AttachStdin`
- `AttachStdout`
- `AttachStderr`
- `Tty`
- `OpenStdin`
- `StdinOnce`
- `Image`
- `NetworkDisabled` (omitted unless set on older API versions)
- `MacAddress` (omitted unless set on older API versions)
- `StopTimeout` (omitted unless set on older API versions)

[Docker image specification]: https://github.com/moby/docker-image-spec/blob/v1.3.1/specs-go/v1/image.go#L19-L32
[OCI image specification]: https://github.com/opencontainers/image-spec/blob/v1.1.0/specs-go/v1/config.go#L24-L62

### Graphdriver plugins (experimental)

**Deprecated in**: v27.0**. 

**Disabled by default in release: v27.0**

**Target for removal in release: v28.0**

[Graphdriver plugins](https://github.com/docker/cli/blob/v26.1.4/docs/extend/plugins_graphdriver.md)
were an experimental feature that allowed extending the Docker Engine with custom
storage drivers for storing images and containers. This feature was not
maintained since its inception.

Support for graphdriver plugins was disabled by default in v27.0, and removed
in v28.0. Users of this feature are recommended to instead configure the Docker
Engine to use the [containerd image store](https://docs.docker.com/storage/containerd/)
and a custom [snapshotter](https://github.com/containerd/containerd/tree/v1.7.18/docs/snapshotters)

### API CORS headers

**Deprecated in release: v27.0**

**Disabled by default in release: v27.0**

**Removed in release: v28.0**

The `api-cors-header` configuration option for the Docker daemon is insecure,
and is therefore deprecated and scheduled for removal.
Incorrectly setting this option could leave a window of opportunity
for unauthenticated cross-origin requests to be accepted by the daemon.

In Docker Engine v27.0, this flag can still be set,
but it has no effect unless the environment variable
`DOCKERD_DEPRECATED_CORS_HEADER` is also set to a non-empty value.

This flag has been removed altogether in v28.0.

This is a breaking change for authorization plugins and other programs
that depend on this option for accessing the Docker API from a browser.
If you need to access the API through a browser, use a reverse proxy.

### Unauthenticated TCP connections

**Deprecated in release: v26.0**

**Target for removal in release: v28.0**

Configuring the Docker daemon to listen on a TCP address will require mandatory
TLS verification. This change aims to ensure secure communication by preventing
unauthorized access to the Docker daemon over potentially insecure networks.
This mandatory TLS requirement applies to all TCP addresses except `tcp://localhost`.

In version 27.0 and later, specifying `--tls=false` or `--tlsverify=false` CLI flags
causes the daemon to fail to start if it's also configured to accept remote connections over TCP.
This also applies to the equivalent configuration options in `daemon.json`.

To facilitate remote access to the Docker daemon over TCP, you'll need to
implement TLS verification. This secures the connection by encrypting data in
transit and providing a mechanism for mutual authentication.

For environments remote daemon access isn't required,
we recommend binding the Docker daemon to a Unix socket.
For daemons where remote access is required and where TLS encryption is not feasible,
you may want to consider using SSH as an alternative solution.

For further information, assistance, and step-by-step instructions on
configuring TLS (or SSH) for the Docker daemon, refer to
[Protect the Docker daemon socket](https://docs.docker.com/engine/security/protect-access/).

### `Container` and `ContainerConfig` fields in Image inspect

**Deprecated in release: v25.0**

**Removed in release: v26.0**

The `Container` and `ContainerConfig` fields returned by `docker inspect` are
mostly an implementation detail of the classic (non-BuildKit) image builder.
These fields are not portable and are empty when using the
BuildKit-based builder (enabled by default since v23.0).
These fields are deprecated in v25.0 and are omitted starting from v26.0 (
API version v1.45 and up). If image configuration of an image is needed,
you can obtain it from the `Config` field.

### Deprecate legacy API versions

**Deprecated in release: v25.0**

**Target for removal in release: v26.0**

The Docker daemon provides a versioned API for backward compatibility with old
clients. Docker clients can perform API-version negotiation to select the most
recent API version supported by the daemon (downgrading to and older version of
the API when necessary). API version negotiation was introduced in Docker v1.12.0
(API 1.24), and clients before that used a fixed API version.

Docker Engine versions through v25.0 provide support for all [API versions](https://docs.docker.com/engine/api/#api-version-matrix)
included in stable releases for a given platform. For Docker daemons on Linux,
the earliest supported API version is 1.12 (corresponding with Docker Engine
v1.0.0), whereas for Docker daemons on Windows, the earliest supported API
version is 1.24 (corresponding with Docker Engine v1.12.0).

Support for legacy API versions (providing old API versions on current versions
of the Docker Engine) is primarily intended to provide compatibility with recent,
but still supported versions of the client, which is a common scenario (the Docker
daemon may be updated to the latest release, but not all clients may be up-to-date
or vice versa). Support for API versions before that (API versions provided by
EOL versions of the Docker Daemon) is provided on a "best effort" basis.

Use of old API versions is rare, and support for legacy API versions
involves significant complexity (Docker 1.0.0 having been released 10 years ago).
Because of this, we'll start deprecating support for legacy API versions.

Docker Engine v25.0 by default disables API version older than 1.24 (aligning
the minimum supported API version between Linux and Windows daemons). When
connecting with a client that uses an API version older than 1.24,
the daemon returns an error. The following example configures the Docker
CLI to use API version 1.23, which produces an error:

```console
DOCKER_API_VERSION=1.23 docker version
Error response from daemon: client version 1.23 is too old. Minimum supported API version is 1.24,
upgrade your client to a newer version
```

Support for API versions lower than `1.24` has been permanently removed in Docker
Engine v26, and the minimum supported API version will be incrementally raised
in releases following that.

<!-- keeping the paragraphs below for when we incrementally raise the minimum API version -->
<!--
An environment variable (`DOCKER_MIN_API_VERSION`) is introduced that allows
re-enabling older API versions in the daemon. This environment variable must
be set in the daemon's environment (for example, through a [systemd override
file](https://docs.docker.com/config/daemon/systemd/)), and the specified
API version must be supported by the daemon (`1.24` or higher).

We do not recommend depending on the `DOCKER_MIN_API_VERSION` environment
variable other than for exceptional cases where it's not possible to update
old clients, and those clients must be supported.
-->

### Container short ID in network Aliases field

**Deprecated in release: v25.0**

**Removed in release: v26.0**

The `Aliases` field returned by `docker inspect` contains the container short
ID once the container is started. This behavior is deprecated in v25.0 but
kept until the next release, v26.0. Starting with that version, the `Aliases`
field will only contain the aliases set through the `docker container create`
and `docker run` flag `--network-alias`.

A new field `DNSNames` containing the container name (if one was specified),
the hostname, the network aliases, as well as the container short ID, has been
introduced in v25.0 and should be used instead of the `Aliases` field.

### Mount `bind-nonrecursive` option

**Deprecated in release: v25.0**

**Scheduled for removal in release: v29.0**

The `bind-nonrecursive` option was replaced with the [`bind-recursive`]
option (see [cli-4316], [cli-4671]). The option was still accepted, but
printed a deprecation warning:

```console
bind-nonrecursive is deprecated, use bind-recursive=disabled instead
```

In the v29.0 release, this warning will be removed and returned as an error.
Users should use the equivalent `bind-recursive=disabled` option instead.

[`bind-recursive`]: https://docs.docker.com/engine/storage/bind-mounts/#recursive-mounts
[cli-4316]: https://github.com/docker/cli/pull/4316
[cli-4671]: https://github.com/docker/cli/pull/4671

### IsAutomated field, and `is-automated` filter on `docker search`

**Deprecated in release: v25.0**

**Removed in release: v28.2**

The `is_automated` field has been deprecated by Docker Hub's search API.
Consequently, the `IsAutomated` field in image search will always be set
to `false` in future, and searching for "is-automated=true" will yield no
results.

The `AUTOMATED` column has been removed from the default `docker search`
and `docker image search` output in v25.0, and the corresponding `IsAutomated`
templating has been removed in v28.2.

### Logentries logging driver

**Deprecated in release: v24.0**

**Removed in release: v25.0**

The logentries service SaaS was shut down on November 15, 2022, rendering
this logging driver non-functional. Users should no longer use this logging
driver, and the driver has been removed in Docker 25.0. Existing containers
using this logging-driver are migrated to use the "local" logging driver
after upgrading.

### OOM-score adjust for the daemon

**Deprecated in release: v24.0**

**Removed in release: v25.0**

The `oom-score-adjust` option was added to prevent the daemon from being
OOM-killed before other processes. This option was mostly added as a
convenience, as running the daemon as a systemd unit was not yet common.

Having the daemon set its own limits is not best-practice, and something
better handled by the process-manager starting the daemon.

Docker v20.10 and newer no longer adjust the daemon's OOM score by default,
instead setting the OOM-score to the systemd unit (OOMScoreAdjust) that's
shipped with the packages.

Users currently depending on this feature are recommended to adjust the
daemon's OOM score using systemd or through other means, when starting
the daemon.

### BuildKit build information

**Deprecated in release: v23.0**

**Removed in release: v24.0**

[Build information](https://github.com/moby/buildkit/blob/v0.11/docs/buildinfo.md)
structures have been introduced in [BuildKit v0.10.0](https://github.com/moby/buildkit/releases/tag/v0.10.0)
and are generated with build metadata that allows you to see all the sources
(images, Git repositories) that were used by the build with their exact
versions and also the configuration that was passed to the build. This
information is also embedded into the image configuration if one is generated.

### Legacy builder for Linux images

**Deprecated in release: v23.0**

Docker v23.0 now uses BuildKit by default to build Linux images, and uses the
[Buildx](https://docs.docker.com/buildx/working-with-buildx/) CLI component for
`docker build`. With this change, `docker build` now exposes all advanced features
that BuildKit provides and which were previously only available through the
`docker buildx` subcommands.

The Buildx component is installed automatically when installing the `docker` CLI
using our `.deb` or `.rpm` packages, and statically linked binaries are provided
both on `download.docker.com`, and through the [`docker/buildx-bin` image](https://hub.docker.com/r/docker/buildx-bin)
on Docker Hub. Refer the [Buildx section](http://docs.docker.com/go/buildx/) for
detailed instructions on installing the Buildx component.

This release marks the beginning of the deprecation cycle of the classic ("legacy")
builder for Linux images. No active development will happen on the classic builder
(except for bugfixes). BuildKit development started five Years ago, left the
"experimental" phase since Docker 18.09, and is already the default builder for
[Docker Desktop](https://docs.docker.com/desktop/previous-versions/3.x-mac/#docker-desktop-320).
While we're comfortable that BuildKit is stable for general use, there may be
some changes in behavior. If you encounter issues with BuildKit, we encourage
you to report issues in the [BuildKit issue tracker on GitHub](https://github.com/moby/buildkit/){:target="_blank" rel="noopener" class="_"}

> Classic builder for building Windows images
>
> BuildKit does not (yet) provide support for building Windows images, and
> `docker build` continues to use the classic builder to build native Windows
> images on Windows daemons.

### Legacy builder fallback

**Deprecated in release: v23.0**

[Docker v23.0 now uses BuildKit by default to build Linux images](#legacy-builder-for-linux-images),
which requires the Buildx component to build images with BuildKit. There may be
situations where the Buildx component is not available, and BuildKit cannot be
used.

To provide a smooth transition to BuildKit as the default builder, Docker v23.0
has an automatic fallback for some situations, or produces an error to assist
users to resolve the problem.

In situations where the user did not explicitly opt-in to use BuildKit (i.e.,
`DOCKER_BUILDKIT=1` is not set), the CLI automatically falls back to the classic
builder, but prints a deprecation warning:

```text
DEPRECATED: The legacy builder is deprecated and will be removed in a future release.
            Install the buildx component to build images with BuildKit:
            https://docs.docker.com/go/buildx/
```

This situation may occur if the `docker` CLI is installed using the static binaries,
and the Buildx component is not installed or not installed correctly. This fallback
will be removed in a future release, therefore we recommend to [install the Buildx component](https://docs.docker.com/go/buildx/)
and use BuildKit for your builds, or opt-out of using BuildKit with `DOCKER_BUILDKIT=0`.

If you opted-in to use BuildKit (`DOCKER_BUILDKIT=1`), but the Buildx component
is missing, an error is printed instead, and the `docker build` command fails:

```text
ERROR: BuildKit is enabled but the buildx component is missing or broken.
       Install the buildx component to build images with BuildKit:
       https://docs.docker.com/go/buildx/
```

We recommend to [install the Buildx component](https://docs.docker.com/go/buildx/)
to continue using BuildKit for your builds, but alternatively, users can either
unset the `DOCKER_BUILDKIT` environment variable to fall back to the legacy builder,
or opt-out of using BuildKit with `DOCKER_BUILDKIT=0`.

Be aware that the [classic builder is deprecated](#legacy-builder-for-linux-images)
so both the automatic fallback and opting-out of using BuildKit will no longer
be possible in a future release.

### Btrfs storage driver on CentOS 7 and RHEL 7

**Removed in release: v23.0**

The `btrfs` storage driver on CentOS and RHEL was provided as a technology preview
by CentOS and RHEL, but has been deprecated since the [Red Hat Enterprise Linux 7.4 release](https://access.redhat.com/documentation/en-us/red_hat_enterprise_linux/7/html/storage_administration_guide/ch-btrfs),
and removed in CentOS 8 and RHEL 8. Users of the `btrfs` storage driver on CentOS
are recommended to migrate to a different storage driver, such as `overlay2`, which
is now the default storage driver. Docker 23.0 continues to provide the `btrfs`
storage driver to allow users to migrate to an alternative driver. The next release
of Docker will no longer provide this driver.

### Support for encrypted TLS private keys

**Deprecated in release: v20.10**

**Removed in release: v23.0**

Use of encrypted TLS private keys has been deprecated, and has been removed.
Golang has deprecated support for legacy PEM encryption (as specified in
[RFC 1423](https://datatracker.ietf.org/doc/html/rfc1423)), as it is insecure by
design (see [https://go-review.googlesource.com/c/go/+/264159](https://go-review.googlesource.com/c/go/+/264159)).

This feature allowed using an encrypted private key with a supplied password,
but did not provide additional security as the encryption is known to be broken,
and the key is sitting next to the password in the filesystem. Users are recommended
to decrypt the private key, and store it un-encrypted to continue using it.

### Kubernetes stack and context support

**Deprecated in release: v20.10**

**Removed in release: v23.0**

Following the deprecation of [Compose on Kubernetes](https://github.com/docker/compose-on-kubernetes),
support for Kubernetes in the `stack` and `context` commands has been removed from
the CLI, and options related to this functionality are now either ignored, or may
produce an error.

The following command-line flags are removed from the `docker context` subcommands:

- `--default-stack-orchestrator` - swarm is now the only (and default) orchestrator for stacks.
- `--kubernetes` - the Kubernetes endpoint can no longer be stored in `docker context`.
- `--kubeconfig` - exporting a context as a kubeconfig file is no longer supported.

The output produced by the `docker context inspect` subcommand no longer contains
information about `StackOrchestrator` and `Kubernetes` endpoints for new contexts.

The following command-line flags are removed from the `docker stack` subcommands:

- `--kubeconfig` - using a kubeconfig file as context is no longer supported.
- `--namespace` - configuring the Kubernetes namespace for stacks is no longer supported.
- `--orchestrator` - swarm is now the only (and default) orchestrator for stacks.

The `DOCKER_STACK_ORCHESTRATOR`, `DOCKER_ORCHESTRATOR`, and `KUBECONFIG` environment
variables, as well as the `stackOrchestrator` option in the `~/.docker/config.json`
CLI configuration file are no longer used, and ignored.

### Pulling images from non-compliant image registries

**Deprecated in release: v20.10**

**Removed in release: v28.2**

Docker Engine v20.10 and up includes optimizations to verify if images in the
local image cache need updating before pulling, preventing the Docker Engine
from making unnecessary API requests. These optimizations require the container
image registry to conform to the [Open Container Initiative Distribution Specification](https://github.com/opencontainers/distribution-spec).

While most registries conform to the specification, we encountered some registries
to be non-compliant, resulting in `docker pull` to fail.

As a temporary solution, Docker Engine v20.10 added a fallback mechanism to
allow `docker pull` to be functional when using a non-compliant registry. A
warning message is printed in this situation:

    WARNING Failed to pull manifest by the resolved digest. This registry does not
            appear to conform to the distribution registry specification; falling back to
            pull by tag. This fallback is DEPRECATED, and will be removed in a future
            release.

The fallback was added to allow users to either migrate their images to a
compliant registry, or for these registries to become compliant.

GitHub deprecated the legacy `docker.pkg.github.com` registry, and it was
[sunset on Feb 24th, 2025](https://github.blog/changelog/2025-01-23-legacy-docker-registry-closing-down/)
in favor of GitHub Container Registry (GHCR, ghcr.io), making this fallback
no longer needed.

### Linux containers on Windows (LCOW) (experimental)

**Deprecated in release: v20.10**

**Removed in release: v23.0**

The experimental feature to run Linux containers on Windows (LCOW) was introduced
as a technical preview in Docker 17.09. While many enhancements were made after
its introduction, the feature never reached completeness, and development has
now stopped in favor of running Docker natively on Linux in WSL2.

Developers who want to run Linux workloads on a Windows host are encouraged to use
[Docker Desktop with WSL2](https://docs.docker.com/docker-for-windows/wsl/) instead.

### BLKIO weight options with cgroups v1

**Deprecated in release: v20.10**

Specifying blkio weight (`docker run --blkio-weight` and `docker run --blkio-weight-device`)
is now marked as deprecated when using cgroups v1 because the corresponding features
were [removed in Linux kernel v5.0 and up](https://github.com/torvalds/linux/commit/f382fb0bcef4c37dc049e9f6963e3baf204d815c).
When using cgroups v2, the `--blkio-weight` options are implemented using
[`io.weight](https://github.com/torvalds/linux/blob/v5.0/Documentation/admin-guide/cgroup-v2.rst#io).

### Kernel memory limit

**Deprecated in release: v20.10**

**Removed in release: v23.0**

Specifying kernel memory limit (`docker run --kernel-memory`) is no longer supported
because the [Linux kernel deprecated `kmem.limit_in_bytes` in v5.4](https://github.com/torvalds/linux/commit/0158115f702b0ba208ab0b5adf44cae99b3ebcc7).
The OCI runtime specification now marks this option (as well as `--kernel-memory-tcp`)
as ["NOT RECOMMENDED"](https://github.com/opencontainers/runtime-spec/pull/1093),
and OCI runtimes such as `runc` no longer support this option.

Docker API v1.42 and up now ignores this option when set. Older versions of the
API continue to accept the option, but depending on the OCI runtime used, may
take no effect.

> [!NOTE]
> While not deprecated (yet) in Docker, the OCI runtime specification also
> deprecated the `memory.kmem.tcp.limit_in_bytes` option. When using `runc` as
> runtime, this option takes no effect. The Linux kernel did not explicitly
> deprecate this feature, and there is a tracking ticket in the `runc` issue
> tracker to determine if this option should be reinstated or if this was an
> oversight of the Linux kernel maintainers (see [opencontainers/runc#3174](https://github.com/opencontainers/runc/issues/3174)).
>
> The `memory.kmem.tcp.limit_in_bytes` option is only supported with cgroups v1,
> and not available on installations running with cgroups v2. This option is
> only supported by the API, and not exposed on the `docker` command-line.

### Classic Swarm and overlay networks using cluster store

**Deprecated in release: v20.10**

**Removed in release: v23.0**

Standalone ("classic") Swarm has been deprecated, and with that the use of overlay
networks using an external key/value store. The corresponding`--cluster-advertise`,
`--cluster-store`, and `--cluster-store-opt` daemon options have been removed.

### Support for legacy `~/.dockercfg` configuration files

**Deprecated in release: v20.10**

**Removed in release: v23.0**

The Docker CLI up until v1.7.0 used the `~/.dockercfg` file to store credentials
after authenticating to a registry (`docker login`). Docker v1.7.0 replaced this
file with a new CLI configuration file, located in `~/.docker/config.json`. When
implementing the new configuration file, the old file (and file-format) was kept
as a fall-back, to assist existing users with migrating to the new file.

Given that the old file format encourages insecure storage of credentials
(credentials are stored unencrypted), and that no version of the CLI since
Docker v1.7.0 has created this file, support for this file, and its format has
been removed.

### Configuration options for experimental CLI features

**Deprecated in release: v19.03**

**Removed in release: v23.0**

The `DOCKER_CLI_EXPERIMENTAL` environment variable and the corresponding `experimental`
field in the CLI configuration file are deprecated. Experimental features are
enabled by default, and these configuration options are no longer functional.

Starting with v23.0, the Docker CLI no longer prints `Experimental` for the client
in the output of `docker version`, and the field has been removed from the JSON
format.

### CLI plugins support

**Deprecated in release: v20.10**

CLI Plugin API is now marked as deprecated.

### Dockerfile legacy `ENV name value` syntax

**Deprecated in release: v20.10**

The Dockerfile `ENV` instruction allows values to be set using either `ENV name=value`
or `ENV name value`. The latter (`ENV name value`) form can be ambiguous, for example,
the following defines a single env-variable (`ONE`) with value `"TWO= THREE=world"`,
but may have intended to be setting three env-vars:

```dockerfile
ENV ONE TWO= THREE=world
```

This format also does not allow setting multiple environment-variables in a single
`ENV` line in the Dockerfile.

Use of the `ENV name value` syntax is discouraged, and may be removed in a future
release. Users are encouraged to update their Dockerfiles to use the `ENV name=value`
syntax, for example:

```dockerfile
ENV ONE="" TWO="" THREE="world"
```

### `docker build --stream` flag (experimental)

**Deprecated in release: v20.10**

**Removed in release: v20.10**

Docker v17.07 introduced an experimental `--stream` flag on `docker build` which
allowed the build-context to be incrementally sent to the daemon, instead of
unconditionally sending the whole build-context.

This functionality has been reimplemented as part of BuildKit, which uses streaming
by default and the `--stream` option will be ignored when using the classic builder,
printing a deprecation warning instead.

Users that want to use this feature are encouraged to enable BuildKit by setting
the `DOCKER_BUILDKIT=1` environment variable or through the daemon or CLI configuration
files.

### `fluentd-async-connect` log opt

**Deprecated in release: v20.10**

**Removed in release: v28.0**

The `--log-opt fluentd-async-connect` option for the fluentd logging driver is
[deprecated in favor of `--log-opt fluentd-async`](https://github.com/moby/moby/pull/39086).
A deprecation message is logged in the daemon logs if the old option is used:

```console
fluent#New: AsyncConnect is now deprecated, use Async instead
```

Users are encouraged to use the `fluentd-async` option going forward, as support
for the old option has been removed.

### Pushing and pulling with image manifest v2 schema 1

**Deprecated in release: v19.03**

**Disabled by default in release: v26.0**

**Removed in release: v28.2**


The image manifest [v2 schema 1](https://distribution.github.io/distribution/spec/deprecated-schema-v1/)
and "Docker Image v1" formats were deprecated in favor of the
[v2 schema 2](https://distribution.github.io/distribution/spec/manifest-v2-2/)
and [OCI image spec](https://github.com/opencontainers/image-spec/tree/v1.1.0)
formats.

These legacy formats should no longer be used, and users are recommended to
update images to use current formats, or to upgrade to more current images.
Starting with Docker v26.0, pulling these images is disabled by default, and
support has been removed in v28.2. Attempting to pull a legacy image now
produces an error:

```console
$ docker pull ubuntu:10.04
Error response from daemon:
Docker Image Format v1 and Docker Image manifest version 2, schema 1 support has been removed.
Suggest the author of docker.io/library/ubuntu:10.04 to upgrade the image to the OCI Format or Docker Image manifest v2, schema 2.
More information at https://docs.docker.com/go/deprecated-image-specs/
```

### `docker engine` subcommands

**Deprecated in release: v19.03**

**Removed in release: v20.10**

The `docker engine activate`, `docker engine check`, and `docker engine update`
provided an alternative installation method to upgrade Docker Community engines
to Docker Enterprise, using an image-based distribution of the Docker Engine.

This feature was only available on Linux, and only when executed on a local node.
Given the limitations of this feature, and the feature not getting widely adopted,
the `docker engine` subcommands will be removed, in favor of installation through
standard package managers.

### Top-level `docker deploy` subcommand (experimental)

**Deprecated in release: v19.03**

**Removed in release: v20.10**

The top-level `docker deploy` command (using the "Docker Application Bundle"
(.dab) file format was introduced as an experimental feature in Docker 1.13 /
17.03, but superseded by support for Docker Compose files using the `docker stack deploy`
subcommand.

### `docker stack deploy` using "dab" files (experimental)

**Deprecated in release: v19.03**

**Removed in release: v20.10**

With no development being done on this feature, and no active use of the file
format, support for the DAB file format and the top-level `docker deploy` command
(hidden by default in 19.03), will be removed, in favour of `docker stack deploy`
using compose files.

### Support for the `overlay2.override_kernel_check` storage option

**Deprecated in release: v19.03**

**Removed in release: v24.0**

This daemon configuration option disabled the Linux kernel version check used
to detect if the kernel supported OverlayFS with multiple lower dirs, which is
required for the overlay2 storage driver. Starting with Docker v19.03.7, the
detection was improved to no longer depend on the kernel _version_, so this
option was no longer used.

### AuFS storage driver

**Deprecated in release: v19.03**

**Removed in release: v24.0**

The `aufs` storage driver is deprecated in favor of `overlay2`, and has been
removed in a Docker Engine v24.0. Users of the `aufs` storage driver must
migrate to a different storage driver, such as `overlay2`, before upgrading
to Docker Engine v24.0.

The `aufs` storage driver facilitated running Docker on distros that have no
support for OverlayFS, such as Ubuntu 14.04 LTS, which originally shipped with
a 3.14 kernel.

Now that Ubuntu 14.04 is no longer a supported distro for Docker, and `overlay2`
is available to all supported distros (as they are either on kernel 4.x, or have
support for multiple lowerdirs backported), there is no reason to continue
maintenance of the `aufs` storage driver.

### Legacy overlay storage driver

**Deprecated in release: v18.09**

**Removed in release: v24.0**

The `overlay` storage driver is deprecated in favor of the `overlay2` storage
driver, which has all the benefits of `overlay`, without its limitations (excessive
inode consumption). The legacy `overlay` storage driver has been removed in
Docker Engine v24.0. Users of the `overlay` storage driver should migrate to the
`overlay2` storage driver before upgrading to Docker Engine v24.0.

The legacy `overlay` storage driver allowed using overlayFS-backed filesystems
on kernels older than v4.x. Now that all supported distributions are able to run `overlay2`
(as they are either on kernel 4.x, or have support for multiple lowerdirs
backported), there is no reason to keep maintaining the `overlay` storage driver.

### Device mapper storage driver

**Deprecated in release: v18.09**

**Disabled by default in release: v23.0**

**Removed in release: v25.0**

The `devicemapper` storage driver is deprecated in favor of `overlay2`, and has
been removed in Docker Engine v25.0. Users of the `devicemapper` storage driver
must migrate to a different storage driver, such as `overlay2`, before upgrading
to Docker Engine v25.0.

The `devicemapper` storage driver facilitates running Docker on older (3.x) kernels
that have no support for other storage drivers (such as overlay2, or btrfs).

Now that support for `overlay2` is added to all supported distros (as they are
either on kernel 4.x, or have support for multiple lowerdirs backported), there
is no reason to continue maintenance of the `devicemapper` storage driver.

### Use of reserved namespaces in engine labels

**Deprecated in release: v18.06**

**Removed in release: v20.10**

The namespaces `com.docker.*`, `io.docker.*`, and `org.dockerproject.*` in engine labels
were always documented to be reserved, but there was never any enforcement.

Usage of these namespaces will now cause a warning in the engine logs to discourage their
use, and will error instead in v20.10 and above.

### `--disable-legacy-registry` override daemon option

**Disabled In Release: v17.12**

**Removed in release: v19.03**

The `--disable-legacy-registry` flag was disabled in Docker 17.12 and will print
an error when used. For this error to be printed, the flag itself is still present,
but hidden. The flag has been removed in Docker 19.03.

### Interacting with V1 registries

**Disabled by default in release: v17.06**

**Removed in release: v17.12**

Version 1.8.3 added a flag (`--disable-legacy-registry=false`) which prevents the
Docker daemon from `pull`, `push`, and `login` operations against v1
registries.  Though enabled by default, this signals the intent to deprecate
the v1 protocol.

Support for the v1 protocol to the public registry was removed in 1.13. Any
mirror configurations using v1 should be updated to use a
[v2 registry mirror](https://docs.docker.com/registry/recipes/mirror/).

Starting with Docker 17.12, support for V1 registries has been removed, and the
`--disable-legacy-registry` flag can no longer be used, and `dockerd` will fail to
start when set.

### Asynchronous `service create` and `service update` as default

**Deprecated in release: v17.05**

**Disabled by default in release: [v17.10](https://github.com/docker/docker-ce/releases/tag/v17.10.0-ce)**

Docker 17.05 added an optional `--detach=false` option to make the
`docker service create` and `docker service update` work synchronously. This
option will be enabled by default in Docker 17.10, at which point the `--detach`
flag can be used to use the previous (asynchronous) behavior.

The default for this option will also be changed accordingly for `docker service rollback`
and `docker service scale` in Docker 17.10.

### `-g` and `--graph` flags on `dockerd`

**Deprecated in release: v17.05**

**Removed in release: v23.0**

The `-g` or `--graph` flag for the `dockerd` or `docker daemon` command was
used to indicate the directory in which to store persistent data and resource
configuration and has been replaced with the more descriptive `--data-root`
flag. These flags were deprecated and hidden in v17.05, and removed in v23.0.

### Top-level network properties in NetworkSettings

**Deprecated in release: [v1.13.0](https://github.com/docker/docker/releases/tag/v1.13.0)**

**Target for removal in release: v17.12**

When inspecting a container, `NetworkSettings` contains top-level information
about the default ("bridge") network;

`EndpointID`, `Gateway`, `GlobalIPv6Address`, `GlobalIPv6PrefixLen`, `IPAddress`,
`IPPrefixLen`, `IPv6Gateway`, and `MacAddress`.

These properties are deprecated in favor of per-network properties in
`NetworkSettings.Networks`. These properties were already "deprecated" in
Docker 1.9, but kept around for backward compatibility.

Refer to [#17538](https://github.com/docker/docker/pull/17538) for further
information.

### `filter` option for `/images/json` endpoint

**Deprecated in release: [v1.13.0](https://github.com/docker/docker/releases/tag/v1.13.0)**

**Removed in release: v20.10**

The `filter` option to filter the list of image by reference (name or name:tag)
is now implemented as a regular filter, named `reference`.

### `repository:shortid` image references

**Deprecated in release: [v1.13.0](https://github.com/docker/docker/releases/tag/v1.13.0)**

**Removed in release: v17.12**

The `repository:shortid` syntax for referencing images is very little used,
collides with tag references, and can be confused with digest references.

Support for the `repository:shortid` notation to reference images was removed
in Docker 17.12.

### `docker daemon` subcommand

**Deprecated in release: [v1.13.0](https://github.com/docker/docker/releases/tag/v1.13.0)**

**Removed in release: v17.12**

The daemon is moved to a separate binary (`dockerd`), and should be used instead.

### Duplicate keys with conflicting values in engine labels

**Deprecated in release: [v1.13.0](https://github.com/docker/docker/releases/tag/v1.13.0)**

**Removed in release: v17.12**

When setting duplicate keys with conflicting values, an error will be produced, and the daemon
will fail to start.

### `MAINTAINER` in Dockerfile

**Deprecated in release: [v1.13.0](https://github.com/docker/docker/releases/tag/v1.13.0)**

`MAINTAINER` was an early very limited form of `LABEL` which should be used instead.

### API calls without a version

**Deprecated in release: [v1.13.0](https://github.com/docker/docker/releases/tag/v1.13.0)**

**Target for removal in release: v17.12**

API versions should be supplied to all API calls to ensure compatibility with
future Engine versions. Instead of just requesting, for example, the URL
`/containers/json`, you must now request `/v1.25/containers/json`.

### Backing filesystem without `d_type` support for overlay/overlay2

**Deprecated in release: [v1.13.0](https://github.com/docker/docker/releases/tag/v1.13.0)**

**Removed in release: v17.12**

The overlay and overlay2 storage driver does not work as expected if the backing
filesystem does not support `d_type`. For example, XFS does not support `d_type`
if it is formatted with the `ftype=0` option.

Support for these setups has been removed, and Docker v23.0 and up now fails to
start when attempting to use the `overlay2` or `overlay` storage driver on a
backing filesystem without `d_type` support.

Refer to [#27358](https://github.com/docker/docker/issues/27358) for details.

### `--automated` and `--stars` flags on `docker search`

**Deprecated in release: [v1.12.0](https://github.com/docker/docker/releases/tag/v1.12.0)**

**Removed in release: v20.10**

The `docker search --automated` and `docker search --stars` options are deprecated.
Use `docker search --filter=is-automated=<true|false>` and `docker search --filter=stars=...` instead.

### `-h` shorthand for `--help`

**Deprecated in release: [v1.12.0](https://github.com/docker/docker/releases/tag/v1.12.0)**

**Target for removal in release: v17.09**

The shorthand (`-h`) is less common than `--help` on Linux and cannot be used
on all subcommands (due to it conflicting with, e.g. `-h` / `--hostname` on
`docker create`). For this reason, the `-h` shorthand was not printed in the
"usage" output of subcommands, nor documented, and is now marked "deprecated".

### `-e` and `--email` flags on `docker login`

**Deprecated in release: [v1.11.0](https://github.com/docker/docker/releases/tag/v1.11.0)**

**Removed in release: [v17.06](https://github.com/docker/docker-ce/releases/tag/v17.06.0-ce)**

The `docker login` no longer automatically registers an account with the target registry if the given username doesn't exist. Due to this change, the email flag is no longer required, and will be deprecated.

### Separator (`:`) of `--security-opt` flag on `docker run`

**Deprecated in release: [v1.11.0](https://github.com/docker/docker/releases/tag/v1.11.0)**

**Target for removal in release: v17.06**

The flag `--security-opt` doesn't use the colon separator (`:`) anymore to divide keys and values, it uses the equal symbol (`=`) for consistency with other similar flags, like `--storage-opt`.

### Ambiguous event fields in API

**Deprecated in release: [v1.10.0](https://github.com/docker/docker/releases/tag/v1.10.0)**

The fields `ID`, `Status` and `From` in the events API have been deprecated in favor of a more rich structure.
See the events API documentation for the new format.

### `-f` flag on `docker tag`

**Deprecated in release: [v1.10.0](https://github.com/docker/docker/releases/tag/v1.10.0)**

**Removed in release: [v1.12.0](https://github.com/docker/docker/releases/tag/v1.12.0)**

To make tagging consistent across the various `docker` commands, the `-f` flag on the `docker tag` command is deprecated. It is no longer necessary to specify `-f` to move a tag from one image to another. Nor will `docker` generate an error if the `-f` flag is missing and the specified tag is already in use.

### HostConfig at API container start

**Deprecated in release: [v1.10.0](https://github.com/docker/docker/releases/tag/v1.10.0)**

**Removed in release: [v1.12.0](https://github.com/docker/docker/releases/tag/v1.12.0)**

Passing an `HostConfig` to `POST /containers/{name}/start` is deprecated in favor of
defining it at container creation (`POST /containers/create`).

### `--before` and `--since` flags on `docker ps`

**Deprecated in release: [v1.10.0](https://github.com/docker/docker/releases/tag/v1.10.0)**

**Removed in release: [v1.12.0](https://github.com/docker/docker/releases/tag/v1.12.0)**

The `docker ps --before` and `docker ps --since` options are deprecated.
Use `docker ps --filter=before=...` and `docker ps --filter=since=...` instead.

### Driver-specific log tags

**Deprecated in release: [v1.9.0](https://github.com/docker/docker/releases/tag/v1.9.0)**

**Removed in release: [v1.12.0](https://github.com/docker/docker/releases/tag/v1.12.0)**

Log tags are now generated in a standard way across different logging drivers.
Because of which, the driver specific log tag options `syslog-tag`, `gelf-tag` and
`fluentd-tag` have been deprecated in favor of the generic `tag` option.

```console
$ docker --log-driver=syslog --log-opt tag="{{.ImageName}}/{{.Name}}/{{.ID}}"
```

### Docker Content Trust ENV passphrase variables name change

**Deprecated in release: [v1.9.0](https://github.com/docker/docker/releases/tag/v1.9.0)**

**Removed in release: [v1.12.0](https://github.com/docker/docker/releases/tag/v1.12.0)**

Since 1.9, Docker Content Trust Offline key has been renamed to Root key and the Tagging key has been renamed to Repository key. Due to this renaming, we're also changing the corresponding environment variables

- DOCKER_CONTENT_TRUST_OFFLINE_PASSPHRASE is now named DOCKER_CONTENT_TRUST_ROOT_PASSPHRASE
- DOCKER_CONTENT_TRUST_TAGGING_PASSPHRASE is now named DOCKER_CONTENT_TRUST_REPOSITORY_PASSPHRASE

### `/containers/(id or name)/copy` endpoint

**Deprecated in release: [v1.8.0](https://github.com/docker/docker/releases/tag/v1.8.0)**

**Removed in release: [v1.12.0](https://github.com/docker/docker/releases/tag/v1.12.0)**

The endpoint `/containers/(id or name)/copy` is deprecated in favor of `/containers/(id or name)/archive`.

### LXC built-in exec driver

**Deprecated in release: [v1.8.0](https://github.com/docker/docker/releases/tag/v1.8.0)**

**Removed in release: [v1.10.0](https://github.com/docker/docker/releases/tag/v1.10.0)**

The built-in LXC execution driver, the lxc-conf flag, and API fields have been removed.

### Old Command Line Options

**Deprecated in release: [v1.8.0](https://github.com/docker/docker/releases/tag/v1.8.0)**

**Removed in release: [v1.10.0](https://github.com/docker/docker/releases/tag/v1.10.0)**

The flags `-d` and `--daemon` are deprecated. Use the separate `dockerd` binary instead.

The following single-dash (`-opt`) variant of certain command line options
are deprecated and replaced with double-dash options (`--opt`):

- `docker attach -nostdin`
- `docker attach -sig-proxy`
- `docker build -no-cache`
- `docker build -rm`
- `docker commit -author`
- `docker commit -run`
- `docker events -since`
- `docker history -notrunc`
- `docker images -notrunc`
- `docker inspect -format`
- `docker ps -beforeId`
- `docker ps -notrunc`
- `docker ps -sinceId`
- `docker rm -link`
- `docker run -cidfile`
- `docker run -dns`
- `docker run -entrypoint`
- `docker run -expose`
- `docker run -link`
- `docker run -lxc-conf`
- `docker run -n`
- `docker run -privileged`
- `docker run -volumes-from`
- `docker search -notrunc`
- `docker search -stars`
- `docker search -t`
- `docker search -trusted`
- `docker tag -force`

The following double-dash options are deprecated and have no replacement:

- `docker run --cpuset`
- `docker run --networking`
- `docker ps --since-id`
- `docker ps --before-id`
- `docker search --trusted`

**Deprecated in release: [v1.5.0](https://github.com/docker/docker/releases/tag/v1.5.0)**

**Removed in release: [v1.12.0](https://github.com/docker/docker/releases/tag/v1.12.0)**

The single-dash (`-help`) was removed, in favor of the double-dash `--help`

### `--api-enable-cors` flag on `dockerd`

**Deprecated in release: [v1.6.0](https://github.com/docker/docker/releases/tag/v1.6.0)**

**Removed in release: [v17.09](https://github.com/docker/docker-ce/releases/tag/v17.09.0-ce)**

The flag `--api-enable-cors` is deprecated since v1.6.0. Use the flag
`--api-cors-header` instead.

### `--run` flag on `docker commit`

**Deprecated in release: [v0.10.0](https://github.com/docker/docker/releases/tag/v0.10.0)**

**Removed in release: [v1.13.0](https://github.com/docker/docker/releases/tag/v1.13.0)**

The flag `--run` of the `docker commit` command (and its short version `-run`) were deprecated in favor
of the `--changes` flag that allows to pass `Dockerfile` commands.

### Three arguments form in `docker import`

**Deprecated in release: [v0.6.7](https://github.com/docker/docker/releases/tag/v0.6.7)**

**Removed in release: [v1.12.0](https://github.com/docker/docker/releases/tag/v1.12.0)**

The `docker import` command format `file|URL|- [REPOSITORY [TAG]]` is deprecated since November 2013. It's no longer supported.

# meta basics [plain]

    Code
      cli::cli({
        message("This is before")
        cli_alert_info("First message")
        message("This as well")
        cli_alert_success("Success!")
      })
    Message <simpleMessage>
      This is before
      This as well
    Message <cliMessage>
      i First message
      v Success!

# meta basics [ansi]

    Code
      cli::cli({
        message("This is before")
        cli_alert_info("First message")
        message("This as well")
        cli_alert_success("Success!")
      })
    Message <simpleMessage>
      This is before
      This as well
    Message <cliMessage>
      [36mi[39m First message
      [32mv[39m Success!

# meta basics [unicode]

    Code
      cli::cli({
        message("This is before")
        cli_alert_info("First message")
        message("This as well")
        cli_alert_success("Success!")
      })
    Message <simpleMessage>
      This is before
      This as well
    Message <cliMessage>
      ℹ First message
      ✔ Success!

# meta basics [fancy]

    Code
      cli::cli({
        message("This is before")
        cli_alert_info("First message")
        message("This as well")
        cli_alert_success("Success!")
      })
    Message <simpleMessage>
      This is before
      This as well
    Message <cliMessage>
      [36mℹ[39m First message
      [32m✔[39m Success!

# meta is single cli_message [plain]

    Code
      cli_server_default(msgs[[1]])
    Message <cliMessage>
      i First message
      v Success!
    Output
      NULL

# meta is single cli_message [ansi]

    Code
      cli_server_default(msgs[[1]])
    Message <cliMessage>
      [36mi[39m First message
      [32mv[39m Success!
    Output
      NULL

# meta is single cli_message [unicode]

    Code
      cli_server_default(msgs[[1]])
    Message <cliMessage>
      ℹ First message
      ✔ Success!
    Output
      NULL

# meta is single cli_message [fancy]

    Code
      cli_server_default(msgs[[1]])
    Message <cliMessage>
      [36mℹ[39m First message
      [32m✔[39m Success!
    Output
      NULL

# meta is single cliMessage [plain]

    Code
      withCallingHandlers(cli::cli({
        cli_alert_info("First message")
        cli_alert_success("Success!")
      }), cliMessage = function(msg) {
        msgs <<- c(msgs, list(msg))
      })
    Message <cliMessage>
      i First message
      v Success!

# meta is single cliMessage [ansi]

    Code
      withCallingHandlers(cli::cli({
        cli_alert_info("First message")
        cli_alert_success("Success!")
      }), cliMessage = function(msg) {
        msgs <<- c(msgs, list(msg))
      })
    Message <cliMessage>
      [36mi[39m First message
      [32mv[39m Success!

# meta is single cliMessage [unicode]

    Code
      withCallingHandlers(cli::cli({
        cli_alert_info("First message")
        cli_alert_success("Success!")
      }), cliMessage = function(msg) {
        msgs <<- c(msgs, list(msg))
      })
    Message <cliMessage>
      ℹ First message
      ✔ Success!

# meta is single cliMessage [fancy]

    Code
      withCallingHandlers(cli::cli({
        cli_alert_info("First message")
        cli_alert_success("Success!")
      }), cliMessage = function(msg) {
        msgs <<- c(msgs, list(msg))
      })
    Message <cliMessage>
      [36mℹ[39m First message
      [32m✔[39m Success!

# substitution [plain]

    Code
      local({
        x <- 1:3
        cli({
          title <- "My title"
          cli_h1("Title: {.emph {title}}")
          cli_text("And {.emph some} more: {.val {x}}")
        })
      })
    Message <cliMessage>
      
      -- Title: My title -------------------------------------------------------------
      And some more: 1, 2, and 3

# substitution [ansi]

    Code
      local({
        x <- 1:3
        cli({
          title <- "My title"
          cli_h1("Title: {.emph {title}}")
          cli_text("And {.emph some} more: {.val {x}}")
        })
      })
    Message <cliMessage>
      
      [36m--[39m [1mTitle: [3mMy title[23m[22m [36m-------------------------------------------------------------[39m
      And [3msome[23m more: [34m1[39m, [34m2[39m, and [34m3[39m

# substitution [unicode]

    Code
      local({
        x <- 1:3
        cli({
          title <- "My title"
          cli_h1("Title: {.emph {title}}")
          cli_text("And {.emph some} more: {.val {x}}")
        })
      })
    Message <cliMessage>
      
      ── Title: My title ─────────────────────────────────────────────────────────────
      And some more: 1, 2, and 3

# substitution [fancy]

    Code
      local({
        x <- 1:3
        cli({
          title <- "My title"
          cli_h1("Title: {.emph {title}}")
          cli_text("And {.emph some} more: {.val {x}}")
        })
      })
    Message <cliMessage>
      
      [36m──[39m [1mTitle: [3mMy title[23m[22m [36m─────────────────────────────────────────────────────────────[39m
      And [3msome[23m more: [34m1[39m, [34m2[39m, and [34m3[39m

# ansi_html

    Code
      ansi_html(str)
    Output
       [1] "<span class=\"ansi ansi-bold\">bold</span>"                          
       [2] "<span class=\"ansi ansi-faint\">faint</span>"                        
       [3] "<span class=\"ansi ansi-italic\">italic</span>"                      
       [4] "<span class=\"ansi ansi-underline\">underline</span>"                
       [5] "<span class=\"ansi ansi-blink\">blink</span>"                        
       [6] "<span class=\"ansi ansi-inverse\">inverse</span>"                    
       [7] "<span class=\"ansi ansi-hide\">hide</span>"                          
       [8] "<span class=\"ansi ansi-crossedout\">crossedout</span>"              
       [9] "<span class=\"ansi ansi-color-0\">black</span>"                      
      [10] "<span class=\"ansi ansi-color-1\">red</span>"                        
      [11] "<span class=\"ansi ansi-color-2\">green</span>"                      
      [12] "<span class=\"ansi ansi-color-3\">yellow</span>"                     
      [13] "<span class=\"ansi ansi-color-4\">blue</span>"                       
      [14] "<span class=\"ansi ansi-color-5\">magenta</span>"                    
      [15] "<span class=\"ansi ansi-color-6\">cyan</span>"                       
      [16] "<span class=\"ansi ansi-color-7\">white</span>"                      
      [17] "<span class=\"ansi ansi-color-0\">bblack</span>"                     
      [18] "<span class=\"ansi ansi-color-1\">bred</span>"                       
      [19] "<span class=\"ansi ansi-color-2\">bgreen</span>"                     
      [20] "<span class=\"ansi ansi-color-3\">byellow</span>"                    
      [21] "<span class=\"ansi ansi-color-4\">bblue</span>"                      
      [22] "<span class=\"ansi ansi-color-5\">bmagenta</span>"                   
      [23] "<span class=\"ansi ansi-color-6\">bcyan</span>"                      
      [24] "<span class=\"ansi ansi-color-7\">bwhite</span>"                     
      [25] "<span class=\"ansi ansi-color-156\">color-156</span>"                
      [26] "<span class=\"ansi ansi-color-1-22-255\">color-1-22-255</span>"      
      [27] "<span class=\"ansi ansi-bg-color-0\">bg-black</span>"                
      [28] "<span class=\"ansi ansi-bg-color-1\">bg-red</span>"                  
      [29] "<span class=\"ansi ansi-bg-color-2\">bg-green</span>"                
      [30] "<span class=\"ansi ansi-bg-color-3\">bg-yellow</span>"               
      [31] "<span class=\"ansi ansi-bg-color-4\">bg-blue</span>"                 
      [32] "<span class=\"ansi ansi-bg-color-5\">bg-magenta</span>"              
      [33] "<span class=\"ansi ansi-bg-color-6\">bg-cyan</span>"                 
      [34] "<span class=\"ansi ansi-bg-color-7\">bg-white</span>"                
      [35] "<span class=\"ansi ansi-bg-color-0\">bg-bblack</span>"               
      [36] "<span class=\"ansi ansi-bg-color-1\">bg-bred</span>"                 
      [37] "<span class=\"ansi ansi-bg-color-2\">bg-bgreen</span>"               
      [38] "<span class=\"ansi ansi-bg-color-3\">bg-byellow</span>"              
      [39] "<span class=\"ansi ansi-bg-color-4\">bg-bblue</span>"                
      [40] "<span class=\"ansi ansi-bg-color-5\">bg-bmagenta</span>"             
      [41] "<span class=\"ansi ansi-bg-color-6\">bg-bcyan</span>"                
      [42] "<span class=\"ansi ansi-bg-color-7\">bg-bwhite</span>"               
      [43] "<span class=\"ansi ansi-bg-color-156\">bg-color-156</span>"          
      [44] "<span class=\"ansi ansi-bg-color-1-22-255\">bg-color-1-22-255</span>"

# multiple styles

    Code
      ansi_html("\033[1;2;35;45mmultiple")
    Output
      [1] "<span class=\"ansi ansi-bold ansi-faint ansi-color-5 ansi-bg-color-5\">multiple</span>"

# ansi_html_style

    Code
      ansi_html_style(colors = 8)
    Output
      .ansi-bold        { font-weight: bold;             }
      .ansi-italic      { font-style: italic;            }
      .ansi-underline   { text-decoration: underline;    }
      .ansi-blink       { text-decoration: blink;        }
      .ansi-hide        { visibility: hidden;            }
      .ansi-crossedout  { text-decoration: line-through; }
      .ansi-color-0     { color: #000000 }
      .ansi-color-1     { color: #cd3131 }
      .ansi-color-2     { color: #0dbc79 }
      .ansi-color-3     { color: #e5e510 }
      .ansi-color-4     { color: #2472c8 }
      .ansi-color-5     { color: #bc3fbc }
      .ansi-color-6     { color: #11a8cd }
      .ansi-color-7     { color: #e5e5e5 }
      .ansi-color-8     { color: #666666 }
      .ansi-color-9     { color: #f14c4c }
      .ansi-color-10    { color: #23d18b }
      .ansi-color-11    { color: #f5f543 }
      .ansi-color-12    { color: #3b8eea }
      .ansi-color-13    { color: #d670d6 }
      .ansi-color-14    { color: #29b8db }
      .ansi-color-15    { color: #e5e5e5 }
      .ansi-bg-color-0  { background-color: #000000 }
      .ansi-bg-color-1  { background-color: #cd3131 }
      .ansi-bg-color-2  { background-color: #0dbc79 }
      .ansi-bg-color-3  { background-color: #e5e510 }
      .ansi-bg-color-4  { background-color: #2472c8 }
      .ansi-bg-color-5  { background-color: #bc3fbc }
      .ansi-bg-color-6  { background-color: #11a8cd }
      .ansi-bg-color-7  { background-color: #e5e5e5 }
      .ansi-bg-color-8  { background-color: #666666 }
      .ansi-bg-color-9  { background-color: #f14c4c }
      .ansi-bg-color-10 { background-color: #23d18b }
      .ansi-bg-color-11 { background-color: #f5f543 }
      .ansi-bg-color-12 { background-color: #3b8eea }
      .ansi-bg-color-13 { background-color: #d670d6 }
      .ansi-bg-color-14 { background-color: #29b8db }
      .ansi-bg-color-15 { background-color: #e5e5e5 }

---

    Code
      ansi_html_style(colors = 256, palette = "ubuntu")
    Output
      .ansi-bold         { font-weight: bold;             }
      .ansi-italic       { font-style: italic;            }
      .ansi-underline    { text-decoration: underline;    }
      .ansi-blink        { text-decoration: blink;        }
      .ansi-hide         { visibility: hidden;            }
      .ansi-crossedout   { text-decoration: line-through; }
      .ansi-color-0      { color: #010101 }
      .ansi-color-1      { color: #de382b }
      .ansi-color-2      { color: #39b54a }
      .ansi-color-3      { color: #ffc706 }
      .ansi-color-4      { color: #006fb8 }
      .ansi-color-5      { color: #762671 }
      .ansi-color-6      { color: #2cb5e9 }
      .ansi-color-7      { color: #cccccc }
      .ansi-color-8      { color: #808080 }
      .ansi-color-9      { color: #ff0000 }
      .ansi-color-10     { color: #00ff00 }
      .ansi-color-11     { color: #ffff00 }
      .ansi-color-12     { color: #0000ff }
      .ansi-color-13     { color: #ff00ff }
      .ansi-color-14     { color: #00ffff }
      .ansi-color-15     { color: #ffffff }
      .ansi-bg-color-0   { background-color: #010101 }
      .ansi-bg-color-1   { background-color: #de382b }
      .ansi-bg-color-2   { background-color: #39b54a }
      .ansi-bg-color-3   { background-color: #ffc706 }
      .ansi-bg-color-4   { background-color: #006fb8 }
      .ansi-bg-color-5   { background-color: #762671 }
      .ansi-bg-color-6   { background-color: #2cb5e9 }
      .ansi-bg-color-7   { background-color: #cccccc }
      .ansi-bg-color-8   { background-color: #808080 }
      .ansi-bg-color-9   { background-color: #ff0000 }
      .ansi-bg-color-10  { background-color: #00ff00 }
      .ansi-bg-color-11  { background-color: #ffff00 }
      .ansi-bg-color-12  { background-color: #0000ff }
      .ansi-bg-color-13  { background-color: #ff00ff }
      .ansi-bg-color-14  { background-color: #00ffff }
      .ansi-bg-color-15  { background-color: #ffffff }
      .ansi-color-16     { color: #000000 }
      .ansi-color-52     { color: #330000 }
      .ansi-color-88     { color: #660000 }
      .ansi-color-124    { color: #990000 }
      .ansi-color-160    { color: #cc0000 }
      .ansi-color-196    { color: #ff0000 }
      .ansi-color-22     { color: #003300 }
      .ansi-color-58     { color: #333300 }
      .ansi-color-94     { color: #663300 }
      .ansi-color-130    { color: #993300 }
      .ansi-color-166    { color: #cc3300 }
      .ansi-color-202    { color: #ff3300 }
      .ansi-color-28     { color: #006600 }
      .ansi-color-64     { color: #336600 }
      .ansi-color-100    { color: #666600 }
      .ansi-color-136    { color: #996600 }
      .ansi-color-172    { color: #cc6600 }
      .ansi-color-208    { color: #ff6600 }
      .ansi-color-34     { color: #009900 }
      .ansi-color-70     { color: #339900 }
      .ansi-color-106    { color: #669900 }
      .ansi-color-142    { color: #999900 }
      .ansi-color-178    { color: #cc9900 }
      .ansi-color-214    { color: #ff9900 }
      .ansi-color-40     { color: #00cc00 }
      .ansi-color-76     { color: #33cc00 }
      .ansi-color-112    { color: #66cc00 }
      .ansi-color-148    { color: #99cc00 }
      .ansi-color-184    { color: #cccc00 }
      .ansi-color-220    { color: #ffcc00 }
      .ansi-color-46     { color: #00ff00 }
      .ansi-color-82     { color: #33ff00 }
      .ansi-color-118    { color: #66ff00 }
      .ansi-color-154    { color: #99ff00 }
      .ansi-color-190    { color: #ccff00 }
      .ansi-color-226    { color: #ffff00 }
      .ansi-color-17     { color: #000033 }
      .ansi-color-53     { color: #330033 }
      .ansi-color-89     { color: #660033 }
      .ansi-color-125    { color: #990033 }
      .ansi-color-161    { color: #cc0033 }
      .ansi-color-197    { color: #ff0033 }
      .ansi-color-23     { color: #003333 }
      .ansi-color-59     { color: #333333 }
      .ansi-color-95     { color: #663333 }
      .ansi-color-131    { color: #993333 }
      .ansi-color-167    { color: #cc3333 }
      .ansi-color-203    { color: #ff3333 }
      .ansi-color-29     { color: #006633 }
      .ansi-color-65     { color: #336633 }
      .ansi-color-101    { color: #666633 }
      .ansi-color-137    { color: #996633 }
      .ansi-color-173    { color: #cc6633 }
      .ansi-color-209    { color: #ff6633 }
      .ansi-color-35     { color: #009933 }
      .ansi-color-71     { color: #339933 }
      .ansi-color-107    { color: #669933 }
      .ansi-color-143    { color: #999933 }
      .ansi-color-179    { color: #cc9933 }
      .ansi-color-215    { color: #ff9933 }
      .ansi-color-41     { color: #00cc33 }
      .ansi-color-77     { color: #33cc33 }
      .ansi-color-113    { color: #66cc33 }
      .ansi-color-149    { color: #99cc33 }
      .ansi-color-185    { color: #cccc33 }
      .ansi-color-221    { color: #ffcc33 }
      .ansi-color-47     { color: #00ff33 }
      .ansi-color-83     { color: #33ff33 }
      .ansi-color-119    { color: #66ff33 }
      .ansi-color-155    { color: #99ff33 }
      .ansi-color-191    { color: #ccff33 }
      .ansi-color-227    { color: #ffff33 }
      .ansi-color-18     { color: #000066 }
      .ansi-color-54     { color: #330066 }
      .ansi-color-90     { color: #660066 }
      .ansi-color-126    { color: #990066 }
      .ansi-color-162    { color: #cc0066 }
      .ansi-color-198    { color: #ff0066 }
      .ansi-color-24     { color: #003366 }
      .ansi-color-60     { color: #333366 }
      .ansi-color-96     { color: #663366 }
      .ansi-color-132    { color: #993366 }
      .ansi-color-168    { color: #cc3366 }
      .ansi-color-204    { color: #ff3366 }
      .ansi-color-30     { color: #006666 }
      .ansi-color-66     { color: #336666 }
      .ansi-color-102    { color: #666666 }
      .ansi-color-138    { color: #996666 }
      .ansi-color-174    { color: #cc6666 }
      .ansi-color-210    { color: #ff6666 }
      .ansi-color-36     { color: #009966 }
      .ansi-color-72     { color: #339966 }
      .ansi-color-108    { color: #669966 }
      .ansi-color-144    { color: #999966 }
      .ansi-color-180    { color: #cc9966 }
      .ansi-color-216    { color: #ff9966 }
      .ansi-color-42     { color: #00cc66 }
      .ansi-color-78     { color: #33cc66 }
      .ansi-color-114    { color: #66cc66 }
      .ansi-color-150    { color: #99cc66 }
      .ansi-color-186    { color: #cccc66 }
      .ansi-color-222    { color: #ffcc66 }
      .ansi-color-48     { color: #00ff66 }
      .ansi-color-84     { color: #33ff66 }
      .ansi-color-120    { color: #66ff66 }
      .ansi-color-156    { color: #99ff66 }
      .ansi-color-192    { color: #ccff66 }
      .ansi-color-228    { color: #ffff66 }
      .ansi-color-19     { color: #000099 }
      .ansi-color-55     { color: #330099 }
      .ansi-color-91     { color: #660099 }
      .ansi-color-127    { color: #990099 }
      .ansi-color-163    { color: #cc0099 }
      .ansi-color-199    { color: #ff0099 }
      .ansi-color-25     { color: #003399 }
      .ansi-color-61     { color: #333399 }
      .ansi-color-97     { color: #663399 }
      .ansi-color-133    { color: #993399 }
      .ansi-color-169    { color: #cc3399 }
      .ansi-color-205    { color: #ff3399 }
      .ansi-color-31     { color: #006699 }
      .ansi-color-67     { color: #336699 }
      .ansi-color-103    { color: #666699 }
      .ansi-color-139    { color: #996699 }
      .ansi-color-175    { color: #cc6699 }
      .ansi-color-211    { color: #ff6699 }
      .ansi-color-37     { color: #009999 }
      .ansi-color-73     { color: #339999 }
      .ansi-color-109    { color: #669999 }
      .ansi-color-145    { color: #999999 }
      .ansi-color-181    { color: #cc9999 }
      .ansi-color-217    { color: #ff9999 }
      .ansi-color-43     { color: #00cc99 }
      .ansi-color-79     { color: #33cc99 }
      .ansi-color-115    { color: #66cc99 }
      .ansi-color-151    { color: #99cc99 }
      .ansi-color-187    { color: #cccc99 }
      .ansi-color-223    { color: #ffcc99 }
      .ansi-color-49     { color: #00ff99 }
      .ansi-color-85     { color: #33ff99 }
      .ansi-color-121    { color: #66ff99 }
      .ansi-color-157    { color: #99ff99 }
      .ansi-color-193    { color: #ccff99 }
      .ansi-color-229    { color: #ffff99 }
      .ansi-color-20     { color: #0000cc }
      .ansi-color-56     { color: #3300cc }
      .ansi-color-92     { color: #6600cc }
      .ansi-color-128    { color: #9900cc }
      .ansi-color-164    { color: #cc00cc }
      .ansi-color-200    { color: #ff00cc }
      .ansi-color-26     { color: #0033cc }
      .ansi-color-62     { color: #3333cc }
      .ansi-color-98     { color: #6633cc }
      .ansi-color-134    { color: #9933cc }
      .ansi-color-170    { color: #cc33cc }
      .ansi-color-206    { color: #ff33cc }
      .ansi-color-32     { color: #0066cc }
      .ansi-color-68     { color: #3366cc }
      .ansi-color-104    { color: #6666cc }
      .ansi-color-140    { color: #9966cc }
      .ansi-color-176    { color: #cc66cc }
      .ansi-color-212    { color: #ff66cc }
      .ansi-color-38     { color: #0099cc }
      .ansi-color-74     { color: #3399cc }
      .ansi-color-110    { color: #6699cc }
      .ansi-color-146    { color: #9999cc }
      .ansi-color-182    { color: #cc99cc }
      .ansi-color-218    { color: #ff99cc }
      .ansi-color-44     { color: #00cccc }
      .ansi-color-80     { color: #33cccc }
      .ansi-color-116    { color: #66cccc }
      .ansi-color-152    { color: #99cccc }
      .ansi-color-188    { color: #cccccc }
      .ansi-color-224    { color: #ffcccc }
      .ansi-color-50     { color: #00ffcc }
      .ansi-color-86     { color: #33ffcc }
      .ansi-color-122    { color: #66ffcc }
      .ansi-color-158    { color: #99ffcc }
      .ansi-color-194    { color: #ccffcc }
      .ansi-color-230    { color: #ffffcc }
      .ansi-color-21     { color: #0000ff }
      .ansi-color-57     { color: #3300ff }
      .ansi-color-93     { color: #6600ff }
      .ansi-color-129    { color: #9900ff }
      .ansi-color-165    { color: #cc00ff }
      .ansi-color-201    { color: #ff00ff }
      .ansi-color-27     { color: #0033ff }
      .ansi-color-63     { color: #3333ff }
      .ansi-color-99     { color: #6633ff }
      .ansi-color-135    { color: #9933ff }
      .ansi-color-171    { color: #cc33ff }
      .ansi-color-207    { color: #ff33ff }
      .ansi-color-33     { color: #0066ff }
      .ansi-color-69     { color: #3366ff }
      .ansi-color-105    { color: #6666ff }
      .ansi-color-141    { color: #9966ff }
      .ansi-color-177    { color: #cc66ff }
      .ansi-color-213    { color: #ff66ff }
      .ansi-color-39     { color: #0099ff }
      .ansi-color-75     { color: #3399ff }
      .ansi-color-111    { color: #6699ff }
      .ansi-color-147    { color: #9999ff }
      .ansi-color-183    { color: #cc99ff }
      .ansi-color-219    { color: #ff99ff }
      .ansi-color-45     { color: #00ccff }
      .ansi-color-81     { color: #33ccff }
      .ansi-color-117    { color: #66ccff }
      .ansi-color-153    { color: #99ccff }
      .ansi-color-189    { color: #ccccff }
      .ansi-color-225    { color: #ffccff }
      .ansi-color-51     { color: #00ffff }
      .ansi-color-87     { color: #33ffff }
      .ansi-color-123    { color: #66ffff }
      .ansi-color-159    { color: #99ffff }
      .ansi-color-195    { color: #ccffff }
      .ansi-color-231    { color: #ffffff }
      .ansi-color-232    { color: #0a0a0a }
      .ansi-color-233    { color: #141414 }
      .ansi-color-234    { color: #1f1f1f }
      .ansi-color-235    { color: #292929 }
      .ansi-color-236    { color: #333333 }
      .ansi-color-237    { color: #3d3d3d }
      .ansi-color-238    { color: #474747 }
      .ansi-color-239    { color: #525252 }
      .ansi-color-240    { color: #5c5c5c }
      .ansi-color-241    { color: #666666 }
      .ansi-color-242    { color: #707070 }
      .ansi-color-243    { color: #7a7a7a }
      .ansi-color-244    { color: #858585 }
      .ansi-color-245    { color: #8f8f8f }
      .ansi-color-246    { color: #999999 }
      .ansi-color-247    { color: #a3a3a3 }
      .ansi-color-248    { color: #adadad }
      .ansi-color-249    { color: #b8b8b8 }
      .ansi-color-250    { color: #c2c2c2 }
      .ansi-color-251    { color: #cccccc }
      .ansi-color-252    { color: #d6d6d6 }
      .ansi-color-253    { color: #e0e0e0 }
      .ansi-color-254    { color: #ebebeb }
      .ansi-color-255    { color: #f5f5f5 }
      .ansi-bg-color-16  { background-color: #000000 }
      .ansi-bg-color-52  { background-color: #330000 }
      .ansi-bg-color-88  { background-color: #660000 }
      .ansi-bg-color-124 { background-color: #990000 }
      .ansi-bg-color-160 { background-color: #cc0000 }
      .ansi-bg-color-196 { background-color: #ff0000 }
      .ansi-bg-color-22  { background-color: #003300 }
      .ansi-bg-color-58  { background-color: #333300 }
      .ansi-bg-color-94  { background-color: #663300 }
      .ansi-bg-color-130 { background-color: #993300 }
      .ansi-bg-color-166 { background-color: #cc3300 }
      .ansi-bg-color-202 { background-color: #ff3300 }
      .ansi-bg-color-28  { background-color: #006600 }
      .ansi-bg-color-64  { background-color: #336600 }
      .ansi-bg-color-100 { background-color: #666600 }
      .ansi-bg-color-136 { background-color: #996600 }
      .ansi-bg-color-172 { background-color: #cc6600 }
      .ansi-bg-color-208 { background-color: #ff6600 }
      .ansi-bg-color-34  { background-color: #009900 }
      .ansi-bg-color-70  { background-color: #339900 }
      .ansi-bg-color-106 { background-color: #669900 }
      .ansi-bg-color-142 { background-color: #999900 }
      .ansi-bg-color-178 { background-color: #cc9900 }
      .ansi-bg-color-214 { background-color: #ff9900 }
      .ansi-bg-color-40  { background-color: #00cc00 }
      .ansi-bg-color-76  { background-color: #33cc00 }
      .ansi-bg-color-112 { background-color: #66cc00 }
      .ansi-bg-color-148 { background-color: #99cc00 }
      .ansi-bg-color-184 { background-color: #cccc00 }
      .ansi-bg-color-220 { background-color: #ffcc00 }
      .ansi-bg-color-46  { background-color: #00ff00 }
      .ansi-bg-color-82  { background-color: #33ff00 }
      .ansi-bg-color-118 { background-color: #66ff00 }
      .ansi-bg-color-154 { background-color: #99ff00 }
      .ansi-bg-color-190 { background-color: #ccff00 }
      .ansi-bg-color-226 { background-color: #ffff00 }
      .ansi-bg-color-17  { background-color: #000033 }
      .ansi-bg-color-53  { background-color: #330033 }
      .ansi-bg-color-89  { background-color: #660033 }
      .ansi-bg-color-125 { background-color: #990033 }
      .ansi-bg-color-161 { background-color: #cc0033 }
      .ansi-bg-color-197 { background-color: #ff0033 }
      .ansi-bg-color-23  { background-color: #003333 }
      .ansi-bg-color-59  { background-color: #333333 }
      .ansi-bg-color-95  { background-color: #663333 }
      .ansi-bg-color-131 { background-color: #993333 }
      .ansi-bg-color-167 { background-color: #cc3333 }
      .ansi-bg-color-203 { background-color: #ff3333 }
      .ansi-bg-color-29  { background-color: #006633 }
      .ansi-bg-color-65  { background-color: #336633 }
      .ansi-bg-color-101 { background-color: #666633 }
      .ansi-bg-color-137 { background-color: #996633 }
      .ansi-bg-color-173 { background-color: #cc6633 }
      .ansi-bg-color-209 { background-color: #ff6633 }
      .ansi-bg-color-35  { background-color: #009933 }
      .ansi-bg-color-71  { background-color: #339933 }
      .ansi-bg-color-107 { background-color: #669933 }
      .ansi-bg-color-143 { background-color: #999933 }
      .ansi-bg-color-179 { background-color: #cc9933 }
      .ansi-bg-color-215 { background-color: #ff9933 }
      .ansi-bg-color-41  { background-color: #00cc33 }
      .ansi-bg-color-77  { background-color: #33cc33 }
      .ansi-bg-color-113 { background-color: #66cc33 }
      .ansi-bg-color-149 { background-color: #99cc33 }
      .ansi-bg-color-185 { background-color: #cccc33 }
      .ansi-bg-color-221 { background-color: #ffcc33 }
      .ansi-bg-color-47  { background-color: #00ff33 }
      .ansi-bg-color-83  { background-color: #33ff33 }
      .ansi-bg-color-119 { background-color: #66ff33 }
      .ansi-bg-color-155 { background-color: #99ff33 }
      .ansi-bg-color-191 { background-color: #ccff33 }
      .ansi-bg-color-227 { background-color: #ffff33 }
      .ansi-bg-color-18  { background-color: #000066 }
      .ansi-bg-color-54  { background-color: #330066 }
      .ansi-bg-color-90  { background-color: #660066 }
      .ansi-bg-color-126 { background-color: #990066 }
      .ansi-bg-color-162 { background-color: #cc0066 }
      .ansi-bg-color-198 { background-color: #ff0066 }
      .ansi-bg-color-24  { background-color: #003366 }
      .ansi-bg-color-60  { background-color: #333366 }
      .ansi-bg-color-96  { background-color: #663366 }
      .ansi-bg-color-132 { background-color: #993366 }
      .ansi-bg-color-168 { background-color: #cc3366 }
      .ansi-bg-color-204 { background-color: #ff3366 }
      .ansi-bg-color-30  { background-color: #006666 }
      .ansi-bg-color-66  { background-color: #336666 }
      .ansi-bg-color-102 { background-color: #666666 }
      .ansi-bg-color-138 { background-color: #996666 }
      .ansi-bg-color-174 { background-color: #cc6666 }
      .ansi-bg-color-210 { background-color: #ff6666 }
      .ansi-bg-color-36  { background-color: #009966 }
      .ansi-bg-color-72  { background-color: #339966 }
      .ansi-bg-color-108 { background-color: #669966 }
      .ansi-bg-color-144 { background-color: #999966 }
      .ansi-bg-color-180 { background-color: #cc9966 }
      .ansi-bg-color-216 { background-color: #ff9966 }
      .ansi-bg-color-42  { background-color: #00cc66 }
      .ansi-bg-color-78  { background-color: #33cc66 }
      .ansi-bg-color-114 { background-color: #66cc66 }
      .ansi-bg-color-150 { background-color: #99cc66 }
      .ansi-bg-color-186 { background-color: #cccc66 }
      .ansi-bg-color-222 { background-color: #ffcc66 }
      .ansi-bg-color-48  { background-color: #00ff66 }
      .ansi-bg-color-84  { background-color: #33ff66 }
      .ansi-bg-color-120 { background-color: #66ff66 }
      .ansi-bg-color-156 { background-color: #99ff66 }
      .ansi-bg-color-192 { background-color: #ccff66 }
      .ansi-bg-color-228 { background-color: #ffff66 }
      .ansi-bg-color-19  { background-color: #000099 }
      .ansi-bg-color-55  { background-color: #330099 }
      .ansi-bg-color-91  { background-color: #660099 }
      .ansi-bg-color-127 { background-color: #990099 }
      .ansi-bg-color-163 { background-color: #cc0099 }
      .ansi-bg-color-199 { background-color: #ff0099 }
      .ansi-bg-color-25  { background-color: #003399 }
      .ansi-bg-color-61  { background-color: #333399 }
      .ansi-bg-color-97  { background-color: #663399 }
      .ansi-bg-color-133 { background-color: #993399 }
      .ansi-bg-color-169 { background-color: #cc3399 }
      .ansi-bg-color-205 { background-color: #ff3399 }
      .ansi-bg-color-31  { background-color: #006699 }
      .ansi-bg-color-67  { background-color: #336699 }
      .ansi-bg-color-103 { background-color: #666699 }
      .ansi-bg-color-139 { background-color: #996699 }
      .ansi-bg-color-175 { background-color: #cc6699 }
      .ansi-bg-color-211 { background-color: #ff6699 }
      .ansi-bg-color-37  { background-color: #009999 }
      .ansi-bg-color-73  { background-color: #339999 }
      .ansi-bg-color-109 { background-color: #669999 }
      .ansi-bg-color-145 { background-color: #999999 }
      .ansi-bg-color-181 { background-color: #cc9999 }
      .ansi-bg-color-217 { background-color: #ff9999 }
      .ansi-bg-color-43  { background-color: #00cc99 }
      .ansi-bg-color-79  { background-color: #33cc99 }
      .ansi-bg-color-115 { background-color: #66cc99 }
      .ansi-bg-color-151 { background-color: #99cc99 }
      .ansi-bg-color-187 { background-color: #cccc99 }
      .ansi-bg-color-223 { background-color: #ffcc99 }
      .ansi-bg-color-49  { background-color: #00ff99 }
      .ansi-bg-color-85  { background-color: #33ff99 }
      .ansi-bg-color-121 { background-color: #66ff99 }
      .ansi-bg-color-157 { background-color: #99ff99 }
      .ansi-bg-color-193 { background-color: #ccff99 }
      .ansi-bg-color-229 { background-color: #ffff99 }
      .ansi-bg-color-20  { background-color: #0000cc }
      .ansi-bg-color-56  { background-color: #3300cc }
      .ansi-bg-color-92  { background-color: #6600cc }
      .ansi-bg-color-128 { background-color: #9900cc }
      .ansi-bg-color-164 { background-color: #cc00cc }
      .ansi-bg-color-200 { background-color: #ff00cc }
      .ansi-bg-color-26  { background-color: #0033cc }
      .ansi-bg-color-62  { background-color: #3333cc }
      .ansi-bg-color-98  { background-color: #6633cc }
      .ansi-bg-color-134 { background-color: #9933cc }
      .ansi-bg-color-170 { background-color: #cc33cc }
      .ansi-bg-color-206 { background-color: #ff33cc }
      .ansi-bg-color-32  { background-color: #0066cc }
      .ansi-bg-color-68  { background-color: #3366cc }
      .ansi-bg-color-104 { background-color: #6666cc }
      .ansi-bg-color-140 { background-color: #9966cc }
      .ansi-bg-color-176 { background-color: #cc66cc }
      .ansi-bg-color-212 { background-color: #ff66cc }
      .ansi-bg-color-38  { background-color: #0099cc }
      .ansi-bg-color-74  { background-color: #3399cc }
      .ansi-bg-color-110 { background-color: #6699cc }
      .ansi-bg-color-146 { background-color: #9999cc }
      .ansi-bg-color-182 { background-color: #cc99cc }
      .ansi-bg-color-218 { background-color: #ff99cc }
      .ansi-bg-color-44  { background-color: #00cccc }
      .ansi-bg-color-80  { background-color: #33cccc }
      .ansi-bg-color-116 { background-color: #66cccc }
      .ansi-bg-color-152 { background-color: #99cccc }
      .ansi-bg-color-188 { background-color: #cccccc }
      .ansi-bg-color-224 { background-color: #ffcccc }
      .ansi-bg-color-50  { background-color: #00ffcc }
      .ansi-bg-color-86  { background-color: #33ffcc }
      .ansi-bg-color-122 { background-color: #66ffcc }
      .ansi-bg-color-158 { background-color: #99ffcc }
      .ansi-bg-color-194 { background-color: #ccffcc }
      .ansi-bg-color-230 { background-color: #ffffcc }
      .ansi-bg-color-21  { background-color: #0000ff }
      .ansi-bg-color-57  { background-color: #3300ff }
      .ansi-bg-color-93  { background-color: #6600ff }
      .ansi-bg-color-129 { background-color: #9900ff }
      .ansi-bg-color-165 { background-color: #cc00ff }
      .ansi-bg-color-201 { background-color: #ff00ff }
      .ansi-bg-color-27  { background-color: #0033ff }
      .ansi-bg-color-63  { background-color: #3333ff }
      .ansi-bg-color-99  { background-color: #6633ff }
      .ansi-bg-color-135 { background-color: #9933ff }
      .ansi-bg-color-171 { background-color: #cc33ff }
      .ansi-bg-color-207 { background-color: #ff33ff }
      .ansi-bg-color-33  { background-color: #0066ff }
      .ansi-bg-color-69  { background-color: #3366ff }
      .ansi-bg-color-105 { background-color: #6666ff }
      .ansi-bg-color-141 { background-color: #9966ff }
      .ansi-bg-color-177 { background-color: #cc66ff }
      .ansi-bg-color-213 { background-color: #ff66ff }
      .ansi-bg-color-39  { background-color: #0099ff }
      .ansi-bg-color-75  { background-color: #3399ff }
      .ansi-bg-color-111 { background-color: #6699ff }
      .ansi-bg-color-147 { background-color: #9999ff }
      .ansi-bg-color-183 { background-color: #cc99ff }
      .ansi-bg-color-219 { background-color: #ff99ff }
      .ansi-bg-color-45  { background-color: #00ccff }
      .ansi-bg-color-81  { background-color: #33ccff }
      .ansi-bg-color-117 { background-color: #66ccff }
      .ansi-bg-color-153 { background-color: #99ccff }
      .ansi-bg-color-189 { background-color: #ccccff }
      .ansi-bg-color-225 { background-color: #ffccff }
      .ansi-bg-color-51  { background-color: #00ffff }
      .ansi-bg-color-87  { background-color: #33ffff }
      .ansi-bg-color-123 { background-color: #66ffff }
      .ansi-bg-color-159 { background-color: #99ffff }
      .ansi-bg-color-195 { background-color: #ccffff }
      .ansi-bg-color-231 { background-color: #ffffff }
      .ansi-bg-color-232 { background-color: #0a0a0a }
      .ansi-bg-color-233 { background-color: #141414 }
      .ansi-bg-color-234 { background-color: #1f1f1f }
      .ansi-bg-color-235 { background-color: #292929 }
      .ansi-bg-color-236 { background-color: #333333 }
      .ansi-bg-color-237 { background-color: #3d3d3d }
      .ansi-bg-color-238 { background-color: #474747 }
      .ansi-bg-color-239 { background-color: #525252 }
      .ansi-bg-color-240 { background-color: #5c5c5c }
      .ansi-bg-color-241 { background-color: #666666 }
      .ansi-bg-color-242 { background-color: #707070 }
      .ansi-bg-color-243 { background-color: #7a7a7a }
      .ansi-bg-color-244 { background-color: #858585 }
      .ansi-bg-color-245 { background-color: #8f8f8f }
      .ansi-bg-color-246 { background-color: #999999 }
      .ansi-bg-color-247 { background-color: #a3a3a3 }
      .ansi-bg-color-248 { background-color: #adadad }
      .ansi-bg-color-249 { background-color: #b8b8b8 }
      .ansi-bg-color-250 { background-color: #c2c2c2 }
      .ansi-bg-color-251 { background-color: #cccccc }
      .ansi-bg-color-252 { background-color: #d6d6d6 }
      .ansi-bg-color-253 { background-color: #e0e0e0 }
      .ansi-bg-color-254 { background-color: #ebebeb }
      .ansi-bg-color-255 { background-color: #f5f5f5 }

# generic

    Code
      local({
        cli_div(theme = list(.alert = list(before = "GENERIC! ")))
        cli_alert("wow")
      })
    Message <cliMessage>
      GENERIC! wow

# success [plain]

    Code
      local({
        cli_alert_success("wow")
      })
    Message <cliMessage>
      v wow

# success [ansi]

    Code
      local({
        cli_alert_success("wow")
      })
    Message <cliMessage>
      [32mv[39m wow

# success [unicode]

    Code
      local({
        cli_alert_success("wow")
      })
    Message <cliMessage>
      ✔ wow

# success [fancy]

    Code
      local({
        cli_alert_success("wow")
      })
    Message <cliMessage>
      [32m✔[39m wow

# danger [plain]

    Code
      local({
        cli_alert_danger("wow")
      })
    Message <cliMessage>
      x wow

# danger [ansi]

    Code
      local({
        cli_alert_danger("wow")
      })
    Message <cliMessage>
      [31mx[39m wow

# danger [unicode]

    Code
      local({
        cli_alert_danger("wow")
      })
    Message <cliMessage>
      ✖ wow

# danger [fancy]

    Code
      local({
        cli_alert_danger("wow")
      })
    Message <cliMessage>
      [31m✖[39m wow

# warning [plain]

    Code
      local({
        cli_alert_warning("wow")
      })
    Message <cliMessage>
      ! wow

# warning [ansi]

    Code
      local({
        cli_alert_warning("wow")
      })
    Message <cliMessage>
      [33m![39m wow

# warning [unicode]

    Code
      local({
        cli_alert_warning("wow")
      })
    Message <cliMessage>
      ! wow

# warning [fancy]

    Code
      local({
        cli_alert_warning("wow")
      })
    Message <cliMessage>
      [33m![39m wow

# info [plain]

    Code
      local({
        cli_alert_info("wow")
      })
    Message <cliMessage>
      i wow

# info [ansi]

    Code
      local({
        cli_alert_info("wow")
      })
    Message <cliMessage>
      [36mi[39m wow

# info [unicode]

    Code
      local({
        cli_alert_info("wow")
      })
    Message <cliMessage>
      ℹ wow

# info [fancy]

    Code
      local({
        cli_alert_info("wow")
      })
    Message <cliMessage>
      [36mℹ[39m wow

# before and after can have spaces

    Code
      local({
        cli_div(theme = list(.alert = list(before = "x  ", after = "  x")))
        cli_alert("continuing that first alert", wrap = TRUE)
      })
    Message <cliMessage>
      x  continuing that first alert  x

# format_error [plain]

    Code
      local({
        n <- "boo"
        stop(format_error(c("{.var n} must be a numeric vector", x = "You've supplied a {.cls {class(n)}} vector.")))
      })
    Error <simpleError>
      `n` must be a numeric vector
      x You've supplied a <character> vector.

---

    Code
      local({
        len <- 26
        idx <- 100
        stop(format_error(c("Must index an existing element:", i = "There {?is/are} {len} element{?s}.",
          x = "You've tried to subset element {idx}.")))
      })
    Error <simpleError>
      Must index an existing element:
      i There are 26 elements.
      x You've tried to subset element 100.

# format_error [ansi]

    Code
      local({
        n <- "boo"
        stop(format_error(c("{.var n} must be a numeric vector", x = "You've supplied a {.cls {class(n)}} vector.")))
      })
    Error <simpleError>
      [1m[22m`n` must be a numeric vector
      [31mx[39m You've supplied a [34m<character>[39m vector.

---

    Code
      local({
        len <- 26
        idx <- 100
        stop(format_error(c("Must index an existing element:", i = "There {?is/are} {len} element{?s}.",
          x = "You've tried to subset element {idx}.")))
      })
    Error <simpleError>
      [1m[22mMust index an existing element:
      [36mi[39m There are 26 elements.
      [31mx[39m You've tried to subset element 100.

# format_error [unicode]

    Code
      local({
        n <- "boo"
        stop(format_error(c("{.var n} must be a numeric vector", x = "You've supplied a {.cls {class(n)}} vector.")))
      })
    Error <simpleError>
      `n` must be a numeric vector
      ✖ You've supplied a <character> vector.

---

    Code
      local({
        len <- 26
        idx <- 100
        stop(format_error(c("Must index an existing element:", i = "There {?is/are} {len} element{?s}.",
          x = "You've tried to subset element {idx}.")))
      })
    Error <simpleError>
      Must index an existing element:
      ℹ There are 26 elements.
      ✖ You've tried to subset element 100.

# format_error [fancy]

    Code
      local({
        n <- "boo"
        stop(format_error(c("{.var n} must be a numeric vector", x = "You've supplied a {.cls {class(n)}} vector.")))
      })
    Error <simpleError>
      [1m[22m`n` must be a numeric vector
      [31m✖[39m You've supplied a [34m<character>[39m vector.

---

    Code
      local({
        len <- 26
        idx <- 100
        stop(format_error(c("Must index an existing element:", i = "There {?is/are} {len} element{?s}.",
          x = "You've tried to subset element {idx}.")))
      })
    Error <simpleError>
      [1m[22mMust index an existing element:
      [36mℹ[39m There are 26 elements.
      [31m✖[39m You've tried to subset element 100.

# format_warning [plain]

    Code
      n <- "boo"
      warning(format_warning(c("{.var n} must be a numeric vector", x = "You've supplied a {.cls {class(n)}} vector.")))
    Warning <simpleWarning>
      `n` must be a numeric vector
      x You've supplied a <character> vector.

---

    Code
      local({
        len <- 26
        idx <- 100
        warning(format_warning(c("Must index an existing element:", i = "There {?is/are} {len} element{?s}.",
          x = "You've tried to subset element {idx}.")))
      })
    Warning <simpleWarning>
      Must index an existing element:
      i There are 26 elements.
      x You've tried to subset element 100.

# format_warning [ansi]

    Code
      n <- "boo"
      warning(format_warning(c("{.var n} must be a numeric vector", x = "You've supplied a {.cls {class(n)}} vector.")))
    Warning <simpleWarning>
      [1m[22m`n` must be a numeric vector
      [31mx[39m You've supplied a [34m<character>[39m vector.

---

    Code
      local({
        len <- 26
        idx <- 100
        warning(format_warning(c("Must index an existing element:", i = "There {?is/are} {len} element{?s}.",
          x = "You've tried to subset element {idx}.")))
      })
    Warning <simpleWarning>
      [1m[22mMust index an existing element:
      [36mi[39m There are 26 elements.
      [31mx[39m You've tried to subset element 100.

# format_warning [unicode]

    Code
      n <- "boo"
      warning(format_warning(c("{.var n} must be a numeric vector", x = "You've supplied a {.cls {class(n)}} vector.")))
    Warning <simpleWarning>
      `n` must be a numeric vector
      ✖ You've supplied a <character> vector.

---

    Code
      local({
        len <- 26
        idx <- 100
        warning(format_warning(c("Must index an existing element:", i = "There {?is/are} {len} element{?s}.",
          x = "You've tried to subset element {idx}.")))
      })
    Warning <simpleWarning>
      Must index an existing element:
      ℹ There are 26 elements.
      ✖ You've tried to subset element 100.

# format_warning [fancy]

    Code
      n <- "boo"
      warning(format_warning(c("{.var n} must be a numeric vector", x = "You've supplied a {.cls {class(n)}} vector.")))
    Warning <simpleWarning>
      [1m[22m`n` must be a numeric vector
      [31m✖[39m You've supplied a [34m<character>[39m vector.

---

    Code
      local({
        len <- 26
        idx <- 100
        warning(format_warning(c("Must index an existing element:", i = "There {?is/are} {len} element{?s}.",
          x = "You've tried to subset element {idx}.")))
      })
    Warning <simpleWarning>
      [1m[22mMust index an existing element:
      [36mℹ[39m There are 26 elements.
      [31m✖[39m You've tried to subset element 100.

# format_message [plain]

    Code
      n <- "boo"
      message(format_message(c("{.var n} must be a numeric vector", x = "You've supplied a {.cls {class(n)}} vector.")))
    Message <simpleMessage>
      `n` must be a numeric vector
      x You've supplied a <character> vector.

---

    Code
      local({
        len <- 26
        idx <- 100
        message(format_message(c("Must index an existing element:", i = "There {?is/are} {len} element{?s}.",
          x = "You've tried to subset element {idx}.")))
      })
    Message <simpleMessage>
      Must index an existing element:
      i There are 26 elements.
      x You've tried to subset element 100.

# format_message [ansi]

    Code
      n <- "boo"
      message(format_message(c("{.var n} must be a numeric vector", x = "You've supplied a {.cls {class(n)}} vector.")))
    Message <simpleMessage>
      [1m[22m`n` must be a numeric vector
      [31mx[39m You've supplied a [34m<character>[39m vector.

---

    Code
      local({
        len <- 26
        idx <- 100
        message(format_message(c("Must index an existing element:", i = "There {?is/are} {len} element{?s}.",
          x = "You've tried to subset element {idx}.")))
      })
    Message <simpleMessage>
      [1m[22mMust index an existing element:
      [36mi[39m There are 26 elements.
      [31mx[39m You've tried to subset element 100.

# format_message [unicode]

    Code
      n <- "boo"
      message(format_message(c("{.var n} must be a numeric vector", x = "You've supplied a {.cls {class(n)}} vector.")))
    Message <simpleMessage>
      `n` must be a numeric vector
      ✖ You've supplied a <character> vector.

---

    Code
      local({
        len <- 26
        idx <- 100
        message(format_message(c("Must index an existing element:", i = "There {?is/are} {len} element{?s}.",
          x = "You've tried to subset element {idx}.")))
      })
    Message <simpleMessage>
      Must index an existing element:
      ℹ There are 26 elements.
      ✖ You've tried to subset element 100.

# format_message [fancy]

    Code
      n <- "boo"
      message(format_message(c("{.var n} must be a numeric vector", x = "You've supplied a {.cls {class(n)}} vector.")))
    Message <simpleMessage>
      [1m[22m`n` must be a numeric vector
      [31m✖[39m You've supplied a [34m<character>[39m vector.

---

    Code
      local({
        len <- 26
        idx <- 100
        message(format_message(c("Must index an existing element:", i = "There {?is/are} {len} element{?s}.",
          x = "You've tried to subset element {idx}.")))
      })
    Message <simpleMessage>
      [1m[22mMust index an existing element:
      [36mℹ[39m There are 26 elements.
      [31m✖[39m You've tried to subset element 100.

# format_error width in RStudio

    Code
      local({
        len <- 26
        idx <- 100
        stop(format_error(c(lorem_ipsum(1, 3), i = lorem_ipsum(1, 3), x = lorem_ipsum(
          1, 3))))
      })
    Error <simpleError>
      Duis quis magna incididunt nulla commodo minim non
      exercitation nostrud ullamco dolor exercitation ut veniam.
      Fugiat irure tempor commodo voluptate ut. In et tempor excepteur
      quis.
      i Et nisi ad quis ad cupidatat tempor laborum est excepteur
        aliqua veniam ex. Sunt magna veniam Lorem elit enim et
        pariatur aliqua occaecat mollit consequat dolore in mollit.
        Officia labore reprehenderit culpa dolore quis nisi do aliqua
        commodo deserunt fugiat cupidatat nostrud ad.
      x Ad laboris consectetur esse minim pariatur irure do anim anim.
        Mollit ad cupidatat ullamco ullamco nulla elit in.

# color in RStudio [ansi]

    Code
      col <- get_rstudio_fg_color0()
      cat(col("this is the new color"))
    Output
      [30mthis is the new color[39m

# update_rstudio_color [ansi]

    Code
      cat(update_rstudio_color("color me interested"))
    Output
      [32mcolor me interested[39m

# named first element

    Code
      format_error(c(`*` = "foo", `*` = "bar"))
    Output
      [1] "* foo\n* bar"

---

    Code
      format_warning(c(`*` = "foo", `*` = "bar"))
    Output
      [1] "* foo\n* bar"

# cli.condition_width

    Code
      format_error(msg)
    Output
      [1] "1234567890 1234567890 1234567890\n1234567890 1234567890 1234567890\n1234567890 1234567890"
    Code
      format_warning(msg)
    Output
      [1] "1234567890 1234567890 1234567890\n1234567890 1234567890 1234567890\n1234567890 1234567890"
    Code
      format_message(msg)
    Output
      [1] "1234567890 1234567890 1234567890\n1234567890 1234567890 1234567890\n1234567890 1234567890"

---

    Code
      format_error(msg)
    Output
      [1] "1234567890 1234567890 1234567890 1234567890 1234567890 1234567890 1234567890 1234567890"
    Code
      format_warning(msg)
    Output
      [1] "1234567890 1234567890 1234567890 1234567890 1234567890 1234567890 1234567890 1234567890"
    Code
      format_message(msg)
    Output
      [1] "1234567890 1234567890 1234567890 1234567890 1234567890 1234567890 1234567890 1234567890"

# suppressing Unicode bullets [plain]

    Code
      local({
        n <- "boo"
        stop(format_error(c("{.var n} must be a numeric vector", x = "You've supplied a {.cls {class(n)}} vector.",
          v = "Success.", i = "Info.", `*` = "Bullet", `>` = "Arrow")))
      })
    Error <simpleError>
      `n` must be a numeric vector
      x You've supplied a <character> vector.
      v Success.
      i Info.
      * Bullet
      > Arrow

# suppressing Unicode bullets [ansi]

    Code
      local({
        n <- "boo"
        stop(format_error(c("{.var n} must be a numeric vector", x = "You've supplied a {.cls {class(n)}} vector.",
          v = "Success.", i = "Info.", `*` = "Bullet", `>` = "Arrow")))
      })
    Error <simpleError>
      [1m[22m`n` must be a numeric vector
      [31mx[39m You've supplied a [34m<character>[39m vector.
      [32mv[39m Success.
      [36mi[39m Info.
      [36m*[39m Bullet
      > Arrow

# suppressing Unicode bullets [unicode]

    Code
      local({
        n <- "boo"
        stop(format_error(c("{.var n} must be a numeric vector", x = "You've supplied a {.cls {class(n)}} vector.",
          v = "Success.", i = "Info.", `*` = "Bullet", `>` = "Arrow")))
      })
    Error <simpleError>
      `n` must be a numeric vector
      x You've supplied a <character> vector.
      v Success.
      i Info.
      * Bullet
      > Arrow

# suppressing Unicode bullets [fancy]

    Code
      local({
        n <- "boo"
        stop(format_error(c("{.var n} must be a numeric vector", x = "You've supplied a {.cls {class(n)}} vector.",
          v = "Success.", i = "Info.", `*` = "Bullet", `>` = "Arrow")))
      })
    Error <simpleError>
      [1m[22m`n` must be a numeric vector
      [31mx[39m You've supplied a [34m<character>[39m vector.
      [32mv[39m Success.
      [36mi[39m Info.
      [36m*[39m Bullet
      > Arrow

# glue quotes and comments

    Code
      cli_dl(c(test_1 = "all good", test_2 = "not #good"))
    Message <cliMessage>
      test_1: all good
      test_2: not #good
    Code
      cli::cli_text("{.url https://example.com/#section}")
    Message <cliMessage>
      <https://example.com/#section>

# quotes, etc. within expressions are still OK

    Code
      cli::cli_text("{.url URL} {1 + 1 # + 1} {1 + 1}")
    Message <cliMessage>
      <URL> 2 2

# hash_emoji

    Code
      hash_emoji(character())$names
    Output
      list()
    Code
      hash_emoji("")$names
    Output
      [[1]]
      [1] "teacup without handle" "rhinoceros"            "flushed face"         
      
    Code
      hash_emoji("x")$names
    Output
      [[1]]
      [1] "mage: dark skin tone" "spoon"                "fuel pump"           
      
    Code
      hash_emoji(NA_character_)$names
    Output
      [[1]]
      [1] NA NA NA
      
    Code
      hash_emoji(NA)$names
    Output
      [[1]]
      [1] NA NA NA
      
    Code
      hash_emoji(c(NA, "", "foo", NA))$names
    Output
      [[1]]
      [1] NA NA NA
      
      [[2]]
      [1] "teacup without handle" "rhinoceros"            "flushed face"         
      
      [[3]]
      [1] "sun behind cloud"                    "raised back of hand: dark skin tone"
      [3] "children crossing"                  
      
      [[4]]
      [1] NA NA NA
      

# hash_raw_emoji

    Code
      hash_raw_emoji(raw())$names
    Output
      [1] "teacup without handle" "rhinoceros"            "flushed face"         
    Code
      hash_raw_emoji(as.raw(0))$names
    Output
      [1] "diamond with a dot"    "raised fist"           "face with thermometer"
    Code
      hash_raw_emoji(charToRaw("foobar"))$names
    Output
      [1] "fishing pole"     "money with wings" "eagle"           

# hash_obj_emoji

    Code
      hash_obj_emoji("")$names
    Output
      [1] "woman pilot: medium skin tone" "cat with wry smile"           
      [3] "couple with heart"            
    Code
      hash_obj_emoji(raw(0))$names
    Output
      [1] "fire"                                "man teacher: medium-light skin tone"
      [3] "carp streamer"                      
    Code
      hash_obj_emoji(1:10)$names
    Output
      [1] "backhand index pointing right: medium-light skin tone"
      [2] "woman office worker: medium skin tone"                
      [3] "tiger face"                                           
    Code
      hash_obj_emoji(mtcars)$names
    Output
      [1] "woman judge: medium skin tone" "pause button"                 
      [3] "flag: North Korea"            

# hash_animal

    Code
      hash_animal(character())$words
    Output
      list()
    Code
      hash_animal("")$words
    Output
      [[1]]
      [1] "dogtired"     "conventional" "olingo"      
      
    Code
      hash_animal("x")$words
    Output
      [[1]]
      [1] "goodhearted" "lovelorn"    "amphiuma"   
      
    Code
      hash_animal(NA_character_)$words
    Output
      [[1]]
      [1] NA NA NA
      
    Code
      hash_animal(NA)$words
    Output
      [[1]]
      [1] NA NA NA
      
    Code
      hash_animal(c(NA, "", "foo", NA))$words
    Output
      [[1]]
      [1] NA NA NA
      
      [[2]]
      [1] "dogtired"     "conventional" "olingo"      
      
      [[3]]
      [1] "sacrilegious" "diet"         "lion"        
      
      [[4]]
      [1] NA NA NA
      

# hash_raw_animal

    Code
      hash_raw_animal(raw())$words
    Output
      [1] "dogtired"     "conventional" "olingo"      
    Code
      hash_raw_animal(as.raw(0))$words
    Output
      [1] "scarabaeiform" "surly"         "goldfish"     
    Code
      hash_raw_animal(charToRaw("foobar"))$words
    Output
      [1] "unevolving" "degrading"  "harrier"   

# hash_obj_animal

    Code
      hash_obj_animal("")$words
    Output
      [1] "pathworky" "wondrous"  "minibeast"
    Code
      hash_obj_animal(raw(0))$words
    Output
      [1] "undestructive"   "unequal"         "indianglassfish"
    Code
      hash_obj_animal(1:10)$words
    Output
      [1] "benignant" "profound"  "ambushbug"
    Code
      hash_obj_animal(mtcars)$words
    Output
      [1] "novice"             "flushed"            "australianshelduck"

# headers [plain]

    Code
      local({
        cli_div(class = "testcli", theme = test_style())
        cli_h1("HEADER")
        cli_h2("Header")
        cli_h3("Header")
        x <- "foobar"
        xx <- 100
        cli_h2("{xx}. header: {x}")
      })
    Message <cliMessage>
      
      HEADER
      
      Header
      
      Header
      
      100. header: foobar
      

# headers [ansi]

    Code
      local({
        cli_div(class = "testcli", theme = test_style())
        cli_h1("HEADER")
        cli_h2("Header")
        cli_h3("Header")
        x <- "foobar"
        xx <- 100
        cli_h2("{xx}. header: {x}")
      })
    Message <cliMessage>
      [1m[3m[1m[3mHEADER[3m[1m[23m[22m
      
      [1m[1mHeader[1m[22m
      
      [4m[4mHeader[4m[24m
      
      [1m[1m[1m100[1m. header: [1mfoobar[1m[1m[22m
      

# headers [unicode]

    Code
      local({
        cli_div(class = "testcli", theme = test_style())
        cli_h1("HEADER")
        cli_h2("Header")
        cli_h3("Header")
        x <- "foobar"
        xx <- 100
        cli_h2("{xx}. header: {x}")
      })
    Message <cliMessage>
      HEADER
      
      Header
      
      Header
      
      100. header: foobar
      

# headers [fancy]

    Code
      local({
        cli_div(class = "testcli", theme = test_style())
        cli_h1("HEADER")
        cli_h2("Header")
        cli_h3("Header")
        x <- "foobar"
        xx <- 100
        cli_h2("{xx}. header: {x}")
      })
    Message <cliMessage>
      [1m[3m[1m[3mHEADER[3m[1m[23m[22m
      
      [1m[1mHeader[1m[22m
      
      [4m[4mHeader[4m[24m
      
      [1m[1m[1m100[1m. header: [1mfoobar[1m[1m[22m
      

# issue #218

    Code
      cli_h1("one {1} two {2} three {3}")
    Message <cliMessage>
      -- one 1 two 2 three 3 ---------------------------------------------------------
    Code
      cli_h2("one {1} two {2} three {3}")
    Message <cliMessage>
      
      -- one 1 two 2 three 3 --
      

# bullets [plain]

    Code
      cli_bullets(c("noindent", ` ` = "space", v = "success", x = "danger", `!` = "warning",
        i = "info", `*` = "bullet", `>` = "arrow"))
    Message <cliMessage>
      noindent
        space
      v success
      x danger
      ! warning
      i info
      * bullet
      > arrow

# bullets [ansi]

    Code
      cli_bullets(c("noindent", ` ` = "space", v = "success", x = "danger", `!` = "warning",
        i = "info", `*` = "bullet", `>` = "arrow"))
    Message <cliMessage>
      noindent
        space
      [32mv[39m success
      [31mx[39m danger
      [33m![39m warning
      [36mi[39m info
      [36m*[39m bullet
      > arrow

# bullets [unicode]

    Code
      cli_bullets(c("noindent", ` ` = "space", v = "success", x = "danger", `!` = "warning",
        i = "info", `*` = "bullet", `>` = "arrow"))
    Message <cliMessage>
      noindent
        space
      ✔ success
      ✖ danger
      ! warning
      ℹ info
      • bullet
      → arrow

# bullets [fancy]

    Code
      cli_bullets(c("noindent", ` ` = "space", v = "success", x = "danger", `!` = "warning",
        i = "info", `*` = "bullet", `>` = "arrow"))
    Message <cliMessage>
      noindent
        space
      [32m✔[39m success
      [31m✖[39m danger
      [33m![39m warning
      [36mℹ[39m info
      [36m•[39m bullet
      → arrow

# bullets glue [plain]

    Code
      cli_bullets(c("noindent {.key {1:3}}", ` ` = "space {.key {1:3}}", v = "success {.key {1:3}}",
        x = "danger {.key {1:3}}", `!` = "warning {.key {1:3}}", i = "info {.key {1:3}}",
        `*` = "bullet {.key {1:3}}", `>` = "arrow {.key {1:3}}"))
    Message <cliMessage>
      noindent [1], [2], and [3]
        space [1], [2], and [3]
      v success [1], [2], and [3]
      x danger [1], [2], and [3]
      ! warning [1], [2], and [3]
      i info [1], [2], and [3]
      * bullet [1], [2], and [3]
      > arrow [1], [2], and [3]

# bullets glue [ansi]

    Code
      cli_bullets(c("noindent {.key {1:3}}", ` ` = "space {.key {1:3}}", v = "success {.key {1:3}}",
        x = "danger {.key {1:3}}", `!` = "warning {.key {1:3}}", i = "info {.key {1:3}}",
        `*` = "bullet {.key {1:3}}", `>` = "arrow {.key {1:3}}"))
    Message <cliMessage>
      noindent [34m[1][39m, [34m[2][39m, and [34m[3][39m
        space [34m[1][39m, [34m[2][39m, and [34m[3][39m
      [32mv[39m success [34m[1][39m, [34m[2][39m, and [34m[3][39m
      [31mx[39m danger [34m[1][39m, [34m[2][39m, and [34m[3][39m
      [33m![39m warning [34m[1][39m, [34m[2][39m, and [34m[3][39m
      [36mi[39m info [34m[1][39m, [34m[2][39m, and [34m[3][39m
      [36m*[39m bullet [34m[1][39m, [34m[2][39m, and [34m[3][39m
      > arrow [34m[1][39m, [34m[2][39m, and [34m[3][39m

# bullets glue [unicode]

    Code
      cli_bullets(c("noindent {.key {1:3}}", ` ` = "space {.key {1:3}}", v = "success {.key {1:3}}",
        x = "danger {.key {1:3}}", `!` = "warning {.key {1:3}}", i = "info {.key {1:3}}",
        `*` = "bullet {.key {1:3}}", `>` = "arrow {.key {1:3}}"))
    Message <cliMessage>
      noindent [1], [2], and [3]
        space [1], [2], and [3]
      ✔ success [1], [2], and [3]
      ✖ danger [1], [2], and [3]
      ! warning [1], [2], and [3]
      ℹ info [1], [2], and [3]
      • bullet [1], [2], and [3]
      → arrow [1], [2], and [3]

# bullets glue [fancy]

    Code
      cli_bullets(c("noindent {.key {1:3}}", ` ` = "space {.key {1:3}}", v = "success {.key {1:3}}",
        x = "danger {.key {1:3}}", `!` = "warning {.key {1:3}}", i = "info {.key {1:3}}",
        `*` = "bullet {.key {1:3}}", `>` = "arrow {.key {1:3}}"))
    Message <cliMessage>
      noindent [34m[1][39m, [34m[2][39m, and [34m[3][39m
        space [34m[1][39m, [34m[2][39m, and [34m[3][39m
      [32m✔[39m success [34m[1][39m, [34m[2][39m, and [34m[3][39m
      [31m✖[39m danger [34m[1][39m, [34m[2][39m, and [34m[3][39m
      [33m![39m warning [34m[1][39m, [34m[2][39m, and [34m[3][39m
      [36mℹ[39m info [34m[1][39m, [34m[2][39m, and [34m[3][39m
      [36m•[39m bullet [34m[1][39m, [34m[2][39m, and [34m[3][39m
      → arrow [34m[1][39m, [34m[2][39m, and [34m[3][39m

# bullets wrapping [plain]

    Code
      cli_bullets(c(txt, ` ` = txt, v = txt, x = txt, `!` = txt, i = txt, `*` = txt,
        `>` = txt))
    Message <cliMessage>
      This is some text that is longer than the width. This is some text that is
      longer than the width. This is some text that is longer than the width.
        This is some text that is longer than the width. This is some text that is
        longer than the width. This is some text that is longer than the width.
      v This is some text that is longer than the width. This is some text that is
        longer than the width. This is some text that is longer than the width.
      x This is some text that is longer than the width. This is some text that is
        longer than the width. This is some text that is longer than the width.
      ! This is some text that is longer than the width. This is some text that is
        longer than the width. This is some text that is longer than the width.
      i This is some text that is longer than the width. This is some text that is
        longer than the width. This is some text that is longer than the width.
      * This is some text that is longer than the width. This is some text that is
        longer than the width. This is some text that is longer than the width.
      > This is some text that is longer than the width. This is some text that is
        longer than the width. This is some text that is longer than the width.

# bullets wrapping [ansi]

    Code
      cli_bullets(c(txt, ` ` = txt, v = txt, x = txt, `!` = txt, i = txt, `*` = txt,
        `>` = txt))
    Message <cliMessage>
      This is some text that is longer than the width. This is some text that is
      longer than the width. This is some text that is longer than the width.
        This is some text that is longer than the width. This is some text that is
        longer than the width. This is some text that is longer than the width.
      [32mv[39m This is some text that is longer than the width. This is some text that is
        longer than the width. This is some text that is longer than the width.
      [31mx[39m This is some text that is longer than the width. This is some text that is
        longer than the width. This is some text that is longer than the width.
      [33m![39m This is some text that is longer than the width. This is some text that is
        longer than the width. This is some text that is longer than the width.
      [36mi[39m This is some text that is longer than the width. This is some text that is
        longer than the width. This is some text that is longer than the width.
      [36m*[39m This is some text that is longer than the width. This is some text that is
        longer than the width. This is some text that is longer than the width.
      > This is some text that is longer than the width. This is some text that is
        longer than the width. This is some text that is longer than the width.

# bullets wrapping [unicode]

    Code
      cli_bullets(c(txt, ` ` = txt, v = txt, x = txt, `!` = txt, i = txt, `*` = txt,
        `>` = txt))
    Message <cliMessage>
      This is some text that is longer than the width. This is some text that is
      longer than the width. This is some text that is longer than the width.
        This is some text that is longer than the width. This is some text that is
        longer than the width. This is some text that is longer than the width.
      ✔ This is some text that is longer than the width. This is some text that is
        longer than the width. This is some text that is longer than the width.
      ✖ This is some text that is longer than the width. This is some text that is
        longer than the width. This is some text that is longer than the width.
      ! This is some text that is longer than the width. This is some text that is
        longer than the width. This is some text that is longer than the width.
      ℹ This is some text that is longer than the width. This is some text that is
        longer than the width. This is some text that is longer than the width.
      • This is some text that is longer than the width. This is some text that is
        longer than the width. This is some text that is longer than the width.
      → This is some text that is longer than the width. This is some text that is
        longer than the width. This is some text that is longer than the width.

# bullets wrapping [fancy]

    Code
      cli_bullets(c(txt, ` ` = txt, v = txt, x = txt, `!` = txt, i = txt, `*` = txt,
        `>` = txt))
    Message <cliMessage>
      This is some text that is longer than the width. This is some text that is
      longer than the width. This is some text that is longer than the width.
        This is some text that is longer than the width. This is some text that is
        longer than the width. This is some text that is longer than the width.
      [32m✔[39m This is some text that is longer than the width. This is some text that is
        longer than the width. This is some text that is longer than the width.
      [31m✖[39m This is some text that is longer than the width. This is some text that is
        longer than the width. This is some text that is longer than the width.
      [33m![39m This is some text that is longer than the width. This is some text that is
        longer than the width. This is some text that is longer than the width.
      [36mℹ[39m This is some text that is longer than the width. This is some text that is
        longer than the width. This is some text that is longer than the width.
      [36m•[39m This is some text that is longer than the width. This is some text that is
        longer than the width. This is some text that is longer than the width.
      → This is some text that is longer than the width. This is some text that is
        longer than the width. This is some text that is longer than the width.

# {.email} [plain-none]

    Code
      cli_text("{.email bugs.bunny@acme.com}")
    Message
      'bugs.bunny@acme.com'

# {.email} [fancy-none]

    Code
      cli_text("{.email bugs.bunny@acme.com}")
    Message
      [34mbugs.bunny@acme.com[39m

# {.email} [plain-all]

    Code
      cli_text("{.email bugs.bunny@acme.com}")
    Message
      ']8;;mailto:bugs.bunny@acme.combugs.bunny@acme.com]8;;'

# {.email} [fancy-all]

    Code
      cli_text("{.email bugs.bunny@acme.com}")
    Message
      ]8;;mailto:bugs.bunny@acme.com[34mbugs.bunny@acme.com[39m]8;;

# {.email} vectors [plain-none]

    Code
      emails <- paste0("bugs.bunny-", 1:3, "@acme.com")
      cli_text("{.email {emails}}")
    Message
      'bugs.bunny-1@acme.com', 'bugs.bunny-2@acme.com', and 'bugs.bunny-3@acme.com'

# {.email} vectors [fancy-none]

    Code
      emails <- paste0("bugs.bunny-", 1:3, "@acme.com")
      cli_text("{.email {emails}}")
    Message
      [34mbugs.bunny-1@acme.com[39m, [34mbugs.bunny-2@acme.com[39m, and [34mbugs.bunny-3@acme.com[39m

# {.email} vectors [plain-all]

    Code
      emails <- paste0("bugs.bunny-", 1:3, "@acme.com")
      cli_text("{.email {emails}}")
    Message
      ']8;;mailto:bugs.bunny-1@acme.combugs.bunny-1@acme.com]8;;', ']8;;mailto:bugs.bunny-2@acme.combugs.bunny-2@acme.com]8;;', and ']8;;mailto:bugs.bunny-3@acme.combugs.bunny-3@acme.com]8;;'

# {.email} vectors [fancy-all]

    Code
      emails <- paste0("bugs.bunny-", 1:3, "@acme.com")
      cli_text("{.email {emails}}")
    Message
      ]8;;mailto:bugs.bunny-1@acme.com[34mbugs.bunny-1@acme.com[39m]8;;, ]8;;mailto:bugs.bunny-2@acme.com[34mbugs.bunny-2@acme.com[39m]8;;, and ]8;;mailto:bugs.bunny-3@acme.com[34mbugs.bunny-3@acme.com[39m]8;;

# {.file} and {.path} [plain-none]

    Code
      cli_text("{.file /absolute/path}")
    Message
      '/absolute/path'
    Code
      cli_text("{.file file:///absolute/path}")
    Message
      'file:///absolute/path'
    Code
      cli_text("{.path /absolute/path}")
    Message
      '/absolute/path'
    Code
      cli_text("{.path file:///absolute/path}")
    Message
      'file:///absolute/path'

---

    Code
      cli_text("{.file relative/path}")
    Message
      'relative/path'
    Code
      cli_text("{.file ./relative/path}")
    Message
      './relative/path'
    Code
      cli_text("{.path relative/path}")
    Message
      'relative/path'
    Code
      cli_text("{.path ./relative/path}")
    Message
      './relative/path'

---

    Code
      cli_text("{.file ~/relative/path}")
    Message
      '~/relative/path'
    Code
      cli_text("{.path ~/relative/path}")
    Message
      '~/relative/path'

---

    Code
      paths <- c("~/foo", "bar", "file:///abs")
      cli_text("{.file {paths}}")
    Message
      '~/foo', 'bar', and 'file:///abs'

---

    Code
      paths <- c("foo  ", " bar ", "file:///a bs ")
      cli_text("{.file {paths}}")
    Message
      'foo  ', ' bar ', and 'file:///a bs '

---

    Code
      name <- cli::style_hyperlink("/foo/bar", "/foo/bar")
      cli_text("{.file {name}}")
    Message
      '/foo/bar'

---

    Code
      cli_text("{.file /absolute/path:12}")
    Message
      '/absolute/path:12'
    Code
      cli_text("{.file file:///absolute/path:5}")
    Message
      'file:///absolute/path:5'
    Code
      cli_text("{.path /absolute/path:123}")
    Message
      '/absolute/path:123'
    Code
      cli_text("{.path file:///absolute/path:51}")
    Message
      'file:///absolute/path:51'

---

    Code
      cli_text("{.file relative/path:12}")
    Message
      'relative/path:12'
    Code
      cli_text("{.file ./relative/path:5}")
    Message
      './relative/path:5'
    Code
      cli_text("{.path relative/path:123}")
    Message
      'relative/path:123'
    Code
      cli_text("{.path ./relative/path:51}")
    Message
      './relative/path:51'

---

    Code
      cli_text("{.file ~/relative/path:12}")
    Message
      '~/relative/path:12'
    Code
      cli_text("{.path ~/relative/path:5}")
    Message
      '~/relative/path:5'

---

    Code
      cli_text("{.file /absolute/path:12:5}")
    Message
      '/absolute/path:12:5'
    Code
      cli_text("{.file file:///absolute/path:5:100}")
    Message
      'file:///absolute/path:5:100'
    Code
      cli_text("{.path /absolute/path:123:1}")
    Message
      '/absolute/path:123:1'
    Code
      cli_text("{.path file:///absolute/path:51:6}")
    Message
      'file:///absolute/path:51:6'

---

    Code
      cli_text("{.file relative/path:12:13}")
    Message
      'relative/path:12:13'
    Code
      cli_text("{.file ./relative/path:5:20}")
    Message
      './relative/path:5:20'
    Code
      cli_text("{.path relative/path:123:21}")
    Message
      'relative/path:123:21'
    Code
      cli_text("{.path ./relative/path:51:2}")
    Message
      './relative/path:51:2'

---

    Code
      cli_text("{.file ~/relative/path:12:23}")
    Message
      '~/relative/path:12:23'
    Code
      cli_text("{.path ~/relative/path:5:2}")
    Message
      '~/relative/path:5:2'

---

    Code
      paths <- c("~/foo", "bar:10", "file:///abs:10:20")
      cli_text("{.file {paths}}")
    Message
      '~/foo', 'bar:10', and 'file:///abs:10:20'

# {.file} and {.path} [fancy-none]

    Code
      cli_text("{.file /absolute/path}")
    Message
      [34m/absolute/path[39m
    Code
      cli_text("{.file file:///absolute/path}")
    Message
      [34mfile:///absolute/path[39m
    Code
      cli_text("{.path /absolute/path}")
    Message
      [34m/absolute/path[39m
    Code
      cli_text("{.path file:///absolute/path}")
    Message
      [34mfile:///absolute/path[39m

---

    Code
      cli_text("{.file relative/path}")
    Message
      [34mrelative/path[39m
    Code
      cli_text("{.file ./relative/path}")
    Message
      [34m./relative/path[39m
    Code
      cli_text("{.path relative/path}")
    Message
      [34mrelative/path[39m
    Code
      cli_text("{.path ./relative/path}")
    Message
      [34m./relative/path[39m

---

    Code
      cli_text("{.file ~/relative/path}")
    Message
      [34m~/relative/path[39m
    Code
      cli_text("{.path ~/relative/path}")
    Message
      [34m~/relative/path[39m

---

    Code
      paths <- c("~/foo", "bar", "file:///abs")
      cli_text("{.file {paths}}")
    Message
      [34m~/foo[39m, [34mbar[39m, and [34mfile:///abs[39m

---

    Code
      paths <- c("foo  ", " bar ", "file:///a bs ")
      cli_text("{.file {paths}}")
    Message
      '[34mfoo[39m[44m  [49m', '[44m [49m[34mbar[39m[44m [49m', and '[34mfile:///a bs[39m[44m [49m'

---

    Code
      name <- cli::style_hyperlink("/foo/bar", "/foo/bar")
      cli_text("{.file {name}}")
    Message
      [34m/foo/bar[39m

---

    Code
      cli_text("{.file /absolute/path:12}")
    Message
      [34m/absolute/path:12[39m
    Code
      cli_text("{.file file:///absolute/path:5}")
    Message
      [34mfile:///absolute/path:5[39m
    Code
      cli_text("{.path /absolute/path:123}")
    Message
      [34m/absolute/path:123[39m
    Code
      cli_text("{.path file:///absolute/path:51}")
    Message
      [34mfile:///absolute/path:51[39m

---

    Code
      cli_text("{.file relative/path:12}")
    Message
      [34mrelative/path:12[39m
    Code
      cli_text("{.file ./relative/path:5}")
    Message
      [34m./relative/path:5[39m
    Code
      cli_text("{.path relative/path:123}")
    Message
      [34mrelative/path:123[39m
    Code
      cli_text("{.path ./relative/path:51}")
    Message
      [34m./relative/path:51[39m

---

    Code
      cli_text("{.file ~/relative/path:12}")
    Message
      [34m~/relative/path:12[39m
    Code
      cli_text("{.path ~/relative/path:5}")
    Message
      [34m~/relative/path:5[39m

---

    Code
      cli_text("{.file /absolute/path:12:5}")
    Message
      [34m/absolute/path:12:5[39m
    Code
      cli_text("{.file file:///absolute/path:5:100}")
    Message
      [34mfile:///absolute/path:5:100[39m
    Code
      cli_text("{.path /absolute/path:123:1}")
    Message
      [34m/absolute/path:123:1[39m
    Code
      cli_text("{.path file:///absolute/path:51:6}")
    Message
      [34mfile:///absolute/path:51:6[39m

---

    Code
      cli_text("{.file relative/path:12:13}")
    Message
      [34mrelative/path:12:13[39m
    Code
      cli_text("{.file ./relative/path:5:20}")
    Message
      [34m./relative/path:5:20[39m
    Code
      cli_text("{.path relative/path:123:21}")
    Message
      [34mrelative/path:123:21[39m
    Code
      cli_text("{.path ./relative/path:51:2}")
    Message
      [34m./relative/path:51:2[39m

---

    Code
      cli_text("{.file ~/relative/path:12:23}")
    Message
      [34m~/relative/path:12:23[39m
    Code
      cli_text("{.path ~/relative/path:5:2}")
    Message
      [34m~/relative/path:5:2[39m

---

    Code
      paths <- c("~/foo", "bar:10", "file:///abs:10:20")
      cli_text("{.file {paths}}")
    Message
      [34m~/foo[39m, [34mbar:10[39m, and [34mfile:///abs:10:20[39m

# {.file} and {.path} [plain-all]

    Code
      cli_text("{.file /absolute/path}")
    Message
      ']8;;file:///absolute/path/absolute/path]8;;'
    Code
      cli_text("{.file file:///absolute/path}")
    Message
      ']8;;file:///absolute/pathfile:///absolute/path]8;;'
    Code
      cli_text("{.path /absolute/path}")
    Message
      ']8;;file:///absolute/path/absolute/path]8;;'
    Code
      cli_text("{.path file:///absolute/path}")
    Message
      ']8;;file:///absolute/pathfile:///absolute/path]8;;'

---

    Code
      cli_text("{.file relative/path}")
    Message
      ']8;;file:///testthat/home/relative/pathrelative/path]8;;'
    Code
      cli_text("{.file ./relative/path}")
    Message
      ']8;;file:///testthat/home/./relative/path./relative/path]8;;'
    Code
      cli_text("{.path relative/path}")
    Message
      ']8;;file:///testthat/home/relative/pathrelative/path]8;;'
    Code
      cli_text("{.path ./relative/path}")
    Message
      ']8;;file:///testthat/home/./relative/path./relative/path]8;;'

---

    Code
      cli_text("{.file ~/relative/path}")
    Message
      ']8;;file:///my/home/relative/path~/relative/path]8;;'
    Code
      cli_text("{.path ~/relative/path}")
    Message
      ']8;;file:///my/home/relative/path~/relative/path]8;;'

---

    Code
      paths <- c("~/foo", "bar", "file:///abs")
      cli_text("{.file {paths}}")
    Message
      ']8;;file:///my/home/foo~/foo]8;;', ']8;;file:///testthat/home/barbar]8;;', and ']8;;file:///absfile:///abs]8;;'

---

    Code
      paths <- c("foo  ", " bar ", "file:///a bs ")
      cli_text("{.file {paths}}")
    Message
      ']8;;file:///testthat/home/foo  foo]8;;  ', ' ]8;;file:///testthat/home/ bar bar]8;; ', and ']8;;file:///a bs file:///a bs]8;; '

---

    Code
      name <- cli::style_hyperlink("/foo/bar", "/foo/bar")
      cli_text("{.file {name}}")
    Message
      ']8;;/foo/bar/foo/bar]8;;'

---

    Code
      cli_text("{.file /absolute/path:12}")
    Message
      ']8;line = 12:col = 1;file:///absolute/path/absolute/path:12]8;;'
    Code
      cli_text("{.file file:///absolute/path:5}")
    Message
      ']8;line = 5:col = 1;file:///absolute/pathfile:///absolute/path:5]8;;'
    Code
      cli_text("{.path /absolute/path:123}")
    Message
      ']8;line = 123:col = 1;file:///absolute/path/absolute/path:123]8;;'
    Code
      cli_text("{.path file:///absolute/path:51}")
    Message
      ']8;line = 51:col = 1;file:///absolute/pathfile:///absolute/path:51]8;;'

---

    Code
      cli_text("{.file relative/path:12}")
    Message
      ']8;line = 12:col = 1;file:///testthat/home/relative/pathrelative/path:12]8;;'
    Code
      cli_text("{.file ./relative/path:5}")
    Message
      ']8;line = 5:col = 1;file:///testthat/home/./relative/path./relative/path:5]8;;'
    Code
      cli_text("{.path relative/path:123}")
    Message
      ']8;line = 123:col = 1;file:///testthat/home/relative/pathrelative/path:123]8;;'
    Code
      cli_text("{.path ./relative/path:51}")
    Message
      ']8;line = 51:col = 1;file:///testthat/home/./relative/path./relative/path:51]8;;'

---

    Code
      cli_text("{.file ~/relative/path:12}")
    Message
      ']8;line = 12:col = 1;file:///my/home/relative/path~/relative/path:12]8;;'
    Code
      cli_text("{.path ~/relative/path:5}")
    Message
      ']8;line = 5:col = 1;file:///my/home/relative/path~/relative/path:5]8;;'

---

    Code
      cli_text("{.file /absolute/path:12:5}")
    Message
      ']8;line = 12:col = 5;file:///absolute/path/absolute/path:12:5]8;;'
    Code
      cli_text("{.file file:///absolute/path:5:100}")
    Message
      ']8;line = 5:col = 100;file:///absolute/pathfile:///absolute/path:5:100]8;;'
    Code
      cli_text("{.path /absolute/path:123:1}")
    Message
      ']8;line = 123:col = 1;file:///absolute/path/absolute/path:123:1]8;;'
    Code
      cli_text("{.path file:///absolute/path:51:6}")
    Message
      ']8;line = 51:col = 6;file:///absolute/pathfile:///absolute/path:51:6]8;;'

---

    Code
      cli_text("{.file relative/path:12:13}")
    Message
      ']8;line = 12:col = 13;file:///testthat/home/relative/pathrelative/path:12:13]8;;'
    Code
      cli_text("{.file ./relative/path:5:20}")
    Message
      ']8;line = 5:col = 20;file:///testthat/home/./relative/path./relative/path:5:20]8;;'
    Code
      cli_text("{.path relative/path:123:21}")
    Message
      ']8;line = 123:col = 21;file:///testthat/home/relative/pathrelative/path:123:21]8;;'
    Code
      cli_text("{.path ./relative/path:51:2}")
    Message
      ']8;line = 51:col = 2;file:///testthat/home/./relative/path./relative/path:51:2]8;;'

---

    Code
      cli_text("{.file ~/relative/path:12:23}")
    Message
      ']8;line = 12:col = 23;file:///my/home/relative/path~/relative/path:12:23]8;;'
    Code
      cli_text("{.path ~/relative/path:5:2}")
    Message
      ']8;line = 5:col = 2;file:///my/home/relative/path~/relative/path:5:2]8;;'

---

    Code
      paths <- c("~/foo", "bar:10", "file:///abs:10:20")
      cli_text("{.file {paths}}")
    Message
      ']8;;file:///my/home/foo~/foo]8;;', ']8;line = 10:col = 1;file:///testthat/home/barbar:10]8;;', and ']8;line = 10:col = 20;file:///absfile:///abs:10:20]8;;'

# {.file} and {.path} [fancy-all]

    Code
      cli_text("{.file /absolute/path}")
    Message
      ]8;;file:///absolute/path[34m/absolute/path[39m]8;;
    Code
      cli_text("{.file file:///absolute/path}")
    Message
      ]8;;file:///absolute/path[34mfile:///absolute/path[39m]8;;
    Code
      cli_text("{.path /absolute/path}")
    Message
      ]8;;file:///absolute/path[34m/absolute/path[39m]8;;
    Code
      cli_text("{.path file:///absolute/path}")
    Message
      ]8;;file:///absolute/path[34mfile:///absolute/path[39m]8;;

---

    Code
      cli_text("{.file relative/path}")
    Message
      ]8;;file:///testthat/home/relative/path[34mrelative/path[39m]8;;
    Code
      cli_text("{.file ./relative/path}")
    Message
      ]8;;file:///testthat/home/./relative/path[34m./relative/path[39m]8;;
    Code
      cli_text("{.path relative/path}")
    Message
      ]8;;file:///testthat/home/relative/path[34mrelative/path[39m]8;;
    Code
      cli_text("{.path ./relative/path}")
    Message
      ]8;;file:///testthat/home/./relative/path[34m./relative/path[39m]8;;

---

    Code
      cli_text("{.file ~/relative/path}")
    Message
      ]8;;file:///my/home/relative/path[34m~/relative/path[39m]8;;
    Code
      cli_text("{.path ~/relative/path}")
    Message
      ]8;;file:///my/home/relative/path[34m~/relative/path[39m]8;;

---

    Code
      paths <- c("~/foo", "bar", "file:///abs")
      cli_text("{.file {paths}}")
    Message
      ]8;;file:///my/home/foo[34m~/foo[39m]8;;, ]8;;file:///testthat/home/bar[34mbar[39m]8;;, and ]8;;file:///abs[34mfile:///abs[39m]8;;

---

    Code
      paths <- c("foo  ", " bar ", "file:///a bs ")
      cli_text("{.file {paths}}")
    Message
      ']8;;file:///testthat/home/foo  [34mfoo[39m]8;;[44m  [49m', '[44m [49m]8;;file:///testthat/home/ bar [34mbar[39m]8;;[44m [49m', and ']8;;file:///a bs [34mfile:///a bs[39m]8;;[44m [49m'

---

    Code
      name <- cli::style_hyperlink("/foo/bar", "/foo/bar")
      cli_text("{.file {name}}")
    Message
      ]8;;/foo/bar[34m/foo/bar[39m]8;;

---

    Code
      cli_text("{.file /absolute/path:12}")
    Message
      ]8;line = 12:col = 1;file:///absolute/path[34m/absolute/path:12[39m]8;;
    Code
      cli_text("{.file file:///absolute/path:5}")
    Message
      ]8;line = 5:col = 1;file:///absolute/path[34mfile:///absolute/path:5[39m]8;;
    Code
      cli_text("{.path /absolute/path:123}")
    Message
      ]8;line = 123:col = 1;file:///absolute/path[34m/absolute/path:123[39m]8;;
    Code
      cli_text("{.path file:///absolute/path:51}")
    Message
      ]8;line = 51:col = 1;file:///absolute/path[34mfile:///absolute/path:51[39m]8;;

---

    Code
      cli_text("{.file relative/path:12}")
    Message
      ]8;line = 12:col = 1;file:///testthat/home/relative/path[34mrelative/path:12[39m]8;;
    Code
      cli_text("{.file ./relative/path:5}")
    Message
      ]8;line = 5:col = 1;file:///testthat/home/./relative/path[34m./relative/path:5[39m]8;;
    Code
      cli_text("{.path relative/path:123}")
    Message
      ]8;line = 123:col = 1;file:///testthat/home/relative/path[34mrelative/path:123[39m]8;;
    Code
      cli_text("{.path ./relative/path:51}")
    Message
      ]8;line = 51:col = 1;file:///testthat/home/./relative/path[34m./relative/path:51[39m]8;;

---

    Code
      cli_text("{.file ~/relative/path:12}")
    Message
      ]8;line = 12:col = 1;file:///my/home/relative/path[34m~/relative/path:12[39m]8;;
    Code
      cli_text("{.path ~/relative/path:5}")
    Message
      ]8;line = 5:col = 1;file:///my/home/relative/path[34m~/relative/path:5[39m]8;;

---

    Code
      cli_text("{.file /absolute/path:12:5}")
    Message
      ]8;line = 12:col = 5;file:///absolute/path[34m/absolute/path:12:5[39m]8;;
    Code
      cli_text("{.file file:///absolute/path:5:100}")
    Message
      ]8;line = 5:col = 100;file:///absolute/path[34mfile:///absolute/path:5:100[39m]8;;
    Code
      cli_text("{.path /absolute/path:123:1}")
    Message
      ]8;line = 123:col = 1;file:///absolute/path[34m/absolute/path:123:1[39m]8;;
    Code
      cli_text("{.path file:///absolute/path:51:6}")
    Message
      ]8;line = 51:col = 6;file:///absolute/path[34mfile:///absolute/path:51:6[39m]8;;

---

    Code
      cli_text("{.file relative/path:12:13}")
    Message
      ]8;line = 12:col = 13;file:///testthat/home/relative/path[34mrelative/path:12:13[39m]8;;
    Code
      cli_text("{.file ./relative/path:5:20}")
    Message
      ]8;line = 5:col = 20;file:///testthat/home/./relative/path[34m./relative/path:5:20[39m]8;;
    Code
      cli_text("{.path relative/path:123:21}")
    Message
      ]8;line = 123:col = 21;file:///testthat/home/relative/path[34mrelative/path:123:21[39m]8;;
    Code
      cli_text("{.path ./relative/path:51:2}")
    Message
      ]8;line = 51:col = 2;file:///testthat/home/./relative/path[34m./relative/path:51:2[39m]8;;

---

    Code
      cli_text("{.file ~/relative/path:12:23}")
    Message
      ]8;line = 12:col = 23;file:///my/home/relative/path[34m~/relative/path:12:23[39m]8;;
    Code
      cli_text("{.path ~/relative/path:5:2}")
    Message
      ]8;line = 5:col = 2;file:///my/home/relative/path[34m~/relative/path:5:2[39m]8;;

---

    Code
      paths <- c("~/foo", "bar:10", "file:///abs:10:20")
      cli_text("{.file {paths}}")
    Message
      ]8;;file:///my/home/foo[34m~/foo[39m]8;;, ]8;line = 10:col = 1;file:///testthat/home/bar[34mbar:10[39m]8;;, and ]8;line = 10:col = 20;file:///abs[34mfile:///abs:10:20[39m]8;;

# {.fun} [plain-none]

    Code
      cli_text("{.fun myfun}")
    Message
      `myfun()`
    Code
      cli_text("{.fun mypackage::myfun}")
    Message
      `mypackage::myfun()`

---

    Code
      funs <- paste0("mypkg::myfun", 1:3)
      cli_text("{.fun {funs}}")
    Message
      `mypkg::myfun1()`, `mypkg::myfun2()`, and `mypkg::myfun3()`

# {.fun} [plain-all]

    Code
      cli_text("{.fun myfun}")
    Message
      `myfun()`
    Code
      cli_text("{.fun mypackage::myfun}")
    Message
      `]8;;x-r-help:mypackage::myfunmypackage::myfun]8;;()`

---

    Code
      funs <- paste0("mypkg::myfun", 1:3)
      cli_text("{.fun {funs}}")
    Message
      `]8;;x-r-help:mypkg::myfun1mypkg::myfun1]8;;()`, `]8;;x-r-help:mypkg::myfun2mypkg::myfun2]8;;()`, and `]8;;x-r-help:mypkg::myfun3mypkg::myfun3]8;;()`

# turning off help [plain-all]

    Code
      cli_text("{.fun pkg::func}")
    Message
      `pkg::func()`

# {.help} [plain-none]

    Code
      cli_text("{.help pkg::fun}")
    Message
      `?pkg::fun()`
    Code
      cli_text("{.help [link text](pkg::fun)}")
    Message
      link text (`?pkg::fun()`)

---

    Code
      funcs <- paste0("pkg::fun", 1:3)
      cli_text("{.help {funcs}}")
    Message
      `?pkg::fun1()`, `?pkg::fun2()`, and `?pkg::fun3()`

# {.help} [plain-all]

    Code
      cli_text("{.help pkg::fun}")
    Message
      ]8;;x-r-help:pkg::funpkg::fun]8;;
    Code
      cli_text("{.help [link text](pkg::fun)}")
    Message
      ]8;;x-r-help:pkg::funlink text]8;;

---

    Code
      funcs <- paste0("pkg::fun", 1:3)
      cli_text("{.help {funcs}}")
    Message
      ]8;;x-r-help:pkg::fun1pkg::fun1]8;;, ]8;;x-r-help:pkg::fun2pkg::fun2]8;;, and ]8;;x-r-help:pkg::fun3pkg::fun3]8;;

# {.href} [plain-none]

    Code
      cli_text("{.href https://cli.r-lib.org}")
    Message
      <https://cli.r-lib.org>
    Code
      cli_text("{.href [linktext](https://cli.r-lib.org)}")
    Message
      linktext (<https://cli.r-lib.org>)
    Code
      cli_text("{.href [link text](https://cli.r-lib.org)}")
    Message
      link text (<https://cli.r-lib.org>)

# {.href} [plain-all]

    Code
      cli_text("{.href https://cli.r-lib.org}")
    Message
      <]8;;https://cli.r-lib.orghttps://cli.r-lib.org]8;;>
    Code
      cli_text("{.href [linktext](https://cli.r-lib.org)}")
    Message
      ]8;;https://cli.r-lib.orglinktext]8;;
    Code
      cli_text("{.href [link text](https://cli.r-lib.org)}")
    Message
      ]8;;https://cli.r-lib.orglink text]8;;

# {.href} vectors [plain-none]

    Code
      url <- paste0("https://cli.r-lib.org/", 1:3)
      cli_text("{.href {url}}")
    Message
      <https://cli.r-lib.org/1>, <https://cli.r-lib.org/2>, and
      <https://cli.r-lib.org/3>

# {.href} vectors [plain-all]

    Code
      url <- paste0("https://cli.r-lib.org/", 1:3)
      cli_text("{.href {url}}")
    Message
      <]8;;https://cli.r-lib.org/1https://cli.r-lib.org/1]8;;>, <]8;;https://cli.r-lib.org/2https://cli.r-lib.org/2]8;;>, and
      <]8;;https://cli.r-lib.org/3https://cli.r-lib.org/3]8;;>

# {.run} [plain-none]

    Code
      cli_text("{.run pkg::fun(param)}")
    Message
      `pkg::fun(param)`
    Code
      cli_text("{.run [run(p1, p2)](pkg::fun(p1, p2, other = 'foo'))}")
    Message
      `run(p1, p2)`

# {.run} [plain-all]

    Code
      cli_text("{.run pkg::fun(param)}")
    Message
      ]8;;x-r-run:pkg::fun(param)pkg::fun(param)]8;;
    Code
      cli_text("{.run [run(p1, p2)](pkg::fun(p1, p2, other = 'foo'))}")
    Message
      ]8;;x-r-run:pkg::fun(p1, p2, other = 'foo')run(p1, p2)]8;;

# {.run} vectors [plain-none]

    Code
      codes <- paste0("pkg::fun", 1:3, "()")
      cli_text("{.run {codes}}")
    Message
      `pkg::fun1()`, `pkg::fun2()`, and `pkg::fun3()`

# {.run} vectors [plain-all]

    Code
      codes <- paste0("pkg::fun", 1:3, "()")
      cli_text("{.run {codes}}")
    Message
      ]8;;x-r-run:pkg::fun1()pkg::fun1()]8;;, ]8;;x-r-run:pkg::fun2()pkg::fun2()]8;;, and ]8;;x-r-run:pkg::fun3()pkg::fun3()]8;;

# {.topic} [plain-none]

    Code
      cli_text("{.topic pkg::topic}")
    Message
      `?pkg::topic`
    Code
      cli_text("{.topic [link text](pkg::topic)}")
    Message
      link text (`?pkg::topic`)

---

    Code
      topics <- paste0("pkg::topic", 1:3)
      cli_text("{.topic {topics}}")
    Message
      `?pkg::topic1`, `?pkg::topic2`, and `?pkg::topic3`

# {.topic} [plain-all]

    Code
      cli_text("{.topic pkg::topic}")
    Message
      ]8;;x-r-help:pkg::topicpkg::topic]8;;
    Code
      cli_text("{.topic [link text](pkg::topic)}")
    Message
      ]8;;x-r-help:pkg::topiclink text]8;;

---

    Code
      topics <- paste0("pkg::topic", 1:3)
      cli_text("{.topic {topics}}")
    Message
      ]8;;x-r-help:pkg::topic1pkg::topic1]8;;, ]8;;x-r-help:pkg::topic2pkg::topic2]8;;, and ]8;;x-r-help:pkg::topic3pkg::topic3]8;;

# {.url} [plain-none]

    Code
      cli_text("{.url https://cli.r-lib.org}")
    Message
      <https://cli.r-lib.org>

# {.url} [fancy-none]

    Code
      cli_text("{.url https://cli.r-lib.org}")
    Message
      [3m[34m<https://cli.r-lib.org>[39m[23m

# {.url} [plain-all]

    Code
      cli_text("{.url https://cli.r-lib.org}")
    Message
      <]8;;https://cli.r-lib.orghttps://cli.r-lib.org]8;;>

# {.url} [fancy-all]

    Code
      cli_text("{.url https://cli.r-lib.org}")
    Message
      [3m[34m<]8;;https://cli.r-lib.orghttps://cli.r-lib.org]8;;>[39m[23m

# {.url} vector [plain-none]

    Code
      urls <- paste0("https://cli.r-lib.org/", 1:3)
      cli_text("{.url {urls}}")
    Message
      <https://cli.r-lib.org/1>, <https://cli.r-lib.org/2>, and
      <https://cli.r-lib.org/3>

# {.url} vector [fancy-none]

    Code
      urls <- paste0("https://cli.r-lib.org/", 1:3)
      cli_text("{.url {urls}}")
    Message
      [3m[34m<https://cli.r-lib.org/1>[39m[23m, [3m[34m<https://cli.r-lib.org/2>[39m[23m, and
      [3m[34m<https://cli.r-lib.org/3>[39m[23m

# {.url} vector [plain-all]

    Code
      urls <- paste0("https://cli.r-lib.org/", 1:3)
      cli_text("{.url {urls}}")
    Message
      <]8;;https://cli.r-lib.org/1https://cli.r-lib.org/1]8;;>, <]8;;https://cli.r-lib.org/2https://cli.r-lib.org/2]8;;>, and
      <]8;;https://cli.r-lib.org/3https://cli.r-lib.org/3]8;;>

# {.url} vector [fancy-all]

    Code
      urls <- paste0("https://cli.r-lib.org/", 1:3)
      cli_text("{.url {urls}}")
    Message
      [3m[34m<]8;;https://cli.r-lib.org/1https://cli.r-lib.org/1]8;;>[39m[23m, [3m[34m<]8;;https://cli.r-lib.org/2https://cli.r-lib.org/2]8;;>[39m[23m, and
      [3m[34m<]8;;https://cli.r-lib.org/3https://cli.r-lib.org/3]8;;>[39m[23m

# linked {.url} [plain-all]

    Code
      link <- c("https://cli.r-lib.org", style_hyperlink("text",
        "https://cli.r-lib.org"))
      cli_text("{.url {link}}")
    Message
      <]8;;https://cli.r-lib.orghttps://cli.r-lib.org]8;;> and <]8;;https://cli.r-lib.orgtext]8;;>

# {.vignette} [plain-none]

    Code
      cli_text("{.vignette pkg::name}")
    Message
      `vignette(pkg::name)`
    Code
      cli_text("{.vignette [link text](pkg::name)}")
    Message
      link text (`vignette(pkg::name)`)

---

    Code
      vignettes <- paste0("pkg::topic", 1:3)
      cli_text("{.vignette {vignettes}}")
    Message
      `vignette(pkg::topic1)`, `vignette(pkg::topic2)`, and `vignette(pkg::topic3)`

# {.vignette} [plain-all]

    Code
      cli_text("{.vignette pkg::name}")
    Message
      ]8;;x-r-vignette:pkg::namepkg::name]8;;
    Code
      cli_text("{.vignette [link text](pkg::name)}")
    Message
      ]8;;x-r-vignette:pkg::namelink text]8;;

---

    Code
      vignettes <- paste0("pkg::topic", 1:3)
      cli_text("{.vignette {vignettes}}")
    Message
      ]8;;x-r-vignette:pkg::topic1pkg::topic1]8;;, ]8;;x-r-vignette:pkg::topic2pkg::topic2]8;;, and ]8;;x-r-vignette:pkg::topic3pkg::topic3]8;;

# deep lists ul

    Code
      for (i in 1:4) test_ul(i)
    Message <cliMessage>
      * Level 1
      * Level 1
        * Level 2
      * Level 1
        * Level 2
          * Level 3
      * Level 1
        * Level 2
          * Level 3
            * Level 4

# deep lists ol

    Code
      for (i in 1:4) test_ol(i)
    Message <cliMessage>
      1. Level 1
      1. Level 1
        1. Level 2
      1. Level 1
        1. Level 2
          1. Level 3
      1. Level 1
        1. Level 2
          1. Level 3
            1. Level 4

# deep lists ol ul

    Code
      for (i in 1:4) test_ol_ul(i)
    Message <cliMessage>
      1. Level 1
        * Level 2
      1. Level 1
        * Level 2
          1. Level 3
            * Level 4
      1. Level 1
        * Level 2
          1. Level 3
            * Level 4
              1. Level 5
                * Level 6
      1. Level 1
        * Level 2
          1. Level 3
            * Level 4
              1. Level 5
                * Level 6
                  1. Level 7
                    * Level 8

# deep lists ul ol

    Code
      for (i in 1:4) test_ul_ol(i)
    Message <cliMessage>
      * Level 1
        1. Level 2
      * Level 1
        1. Level 2
          * Level 3
            1. Level 4
      * Level 1
        1. Level 2
          * Level 3
            1. Level 4
              * Level 5
                1. Level 6
      * Level 1
        1. Level 2
          * Level 3
            1. Level 4
              * Level 5
                1. Level 6
                  * Level 7
                    1. Level 8

# simplest

    Code
      for (n in 0:2) cli_text("{n} package{?s}")
    Message <cliMessage>
      0 packages
      1 package
      2 packages
    Code
      for (n in 0:2) print(pluralize("{n} package{?s}"))
    Output
      0 packages
      1 package
      2 packages

# irregular

    Code
      for (n in 0:2) cli_text("{n} dictionar{?y/ies}")
    Message <cliMessage>
      0 dictionaries
      1 dictionary
      2 dictionaries
    Code
      for (n in 0:2) print(pluralize("{n} dictionar{?y/ies}"))
    Output
      0 dictionaries
      1 dictionary
      2 dictionaries

# multiple substitutions

    Code
      for (n in 0:2) cli_text("{n} package{?s} {?is/are} ...")
    Message <cliMessage>
      0 packages are ...
      1 package is ...
      2 packages are ...
    Code
      for (n in 0:2) print(pluralize("{n} package{?s} {?is/are} ..."))
    Output
      0 packages are ...
      1 package is ...
      2 packages are ...

# multiple quantities

    Code
      for (m in 0:2) for (n in 0:2) cli_text("{m} package{?s} and {n} folder{?s}")
    Message <cliMessage>
      0 packages and 0 folders
      0 packages and 1 folder
      0 packages and 2 folders
      1 package and 0 folders
      1 package and 1 folder
      1 package and 2 folders
      2 packages and 0 folders
      2 packages and 1 folder
      2 packages and 2 folders
    Code
      for (m in 0:2) for (n in 0:2) print(pluralize(
        "{m} package{?s} and {n} folder{?s}"))
    Output
      0 packages and 0 folders
      0 packages and 1 folder
      0 packages and 2 folders
      1 package and 0 folders
      1 package and 1 folder
      1 package and 2 folders
      2 packages and 0 folders
      2 packages and 1 folder
      2 packages and 2 folders

# no()

    Code
      for (n in 0:2) cli_text("{no(n)} package{?s}")
    Message <cliMessage>
      no packages
      1 package
      2 packages
    Code
      for (n in 0:2) print(pluralize("{no(n)} package{?s}"))
    Output
      no packages
      1 package
      2 packages

# set qty() explicitly

    Code
      for (n in 0:2) cli_text("{qty(n)}There {?is/are} {n} package{?s}")
    Message <cliMessage>
      There are 0 packages
      There is 1 package
      There are 2 packages
    Code
      for (n in 0:2) print(pluralize("{qty(n)}There {?is/are} {n} package{?s}"))
    Output
      There are 0 packages
      There is 1 package
      There are 2 packages

# collapsing vectors

    Code
      pkgs <- (function(n) glue::glue("pkg{seq_len(n)}"))
      for (n in 1:3) cli_text("The {pkgs(n)} package{?s}")
    Message <cliMessage>
      The pkg1 package
      The pkg1 and pkg2 packages
      The pkg1, pkg2, and pkg3 packages
    Code
      for (n in 1:3) print(pluralize("The {pkgs(n)} package{?s}"))
    Output
      The pkg1 package
      The pkg1 and pkg2 packages
      The pkg1, pkg2, and pkg3 packages

# pluralization and style

    Code
      special_style <- list(span.foo = list(before = "<", after = ">"))
      cli_div(theme = special_style)
      for (n in 0:2) cli_text("{n} {.foo package{?s}}")
    Message <cliMessage>
      0 packages
      1 package
      2 packages

---

    Code
      pkgs <- (function(n) glue::glue("pkg{seq_len(n)}"))
      for (n in 1:3) cli_text("The {.foo {pkgs(n)}} package{?s}")
    Message <cliMessage>
      The pkg1 package
      The pkg1 and pkg2 packages
      The pkg1, pkg2, and pkg3 packages

# post-processing

    Code
      for (n in 0:2) cli_text("Package{?s}: {n}")
    Message <cliMessage>
      Packages: 0
      Package: 1
      Packages: 2

---

    Code
      pkgs <- (function(n) glue::glue("pkg{seq_len(n)}"))
      for (n in 1:2) cli_text("Package{?s}: {pkgs(n)}")
    Message <cliMessage>
      Package: pkg1
      Packages: pkg1 and pkg2
    Code
      for (n in 1:2) print(pluralize("Package{?s}: {pkgs(n)}"))
    Output
      Package: pkg1
      Packages: pkg1 and pkg2

# issue 158

    Code
      print(pluralize("{0} word{?A/B/}"))
    Output
      0 wordA
    Code
      print(pluralize("{1} word{?A/B/}"))
    Output
      1 wordB
    Code
      print(pluralize("{9} word{?A/B/}"))
    Output
      9 word

# issue #154 [plain]

    Code
      cli_code("a\nb\nc")
    Message <cliMessage>
      a
      b
      c

# issue #154 [ansi]

    Code
      cli_code("a\nb\nc")
    Message <cliMessage>
      a
      b
      c

# issue #154 [unicode]

    Code
      cli_code("a\nb\nc")
    Message <cliMessage>
      a
      b
      c

# issue #154 [fancy]

    Code
      cli_code("a\nb\nc")
    Message <cliMessage>
      a
      b
      c

# collapsing without formatting, n>3

    Code
      pkgs <- paste0("pkg", 1:5)
      cli_text("Packages: {pkgs}.")
    Message
      Packages: pkg1, pkg2, pkg3, pkg4, and pkg5.

# collapsing without formatting, n<3

    Code
      pkgs <- paste0("pkg", 1:2)
      cli_text("Packages: {pkgs}.")
    Message
      Packages: pkg1 and pkg2.

# collapsing with formatting

    Code
      local({
        cli_div(theme = list(.pkg = list(fmt = function(x) paste0(x, " (P)"))))
        pkgs <- paste0("pkg", 1:5)
        cli_text("Packages: {.pkg {pkgs}}.")
      })
    Message
      Packages: pkg1 (P), pkg2 (P), pkg3 (P), pkg4 (P), and pkg5 (P).

# collapsing with formatting, custom seps

    Code
      local({
        cli_div(theme = list(div = list(`vec-sep` = " ... ")))
        pkgs <- paste0("pkg", 1:5)
        cli_text("Packages: {.pkg {pkgs}}.")
      })
    Message
      Packages: pkg1 ... pkg2 ... pkg3 ... pkg4, and pkg5.

# collapsing a cli_vec

    Code
      pkgs <- cli_vec(paste0("pkg", 1:5), style = list(`vec-sep` = " & ", `vec-last` = " & "))
      cli_text("Packages: {pkgs}.")
    Message
      Packages: pkg1 & pkg2 & pkg3 & pkg4 & pkg5.

# collapsing a cli_vec with styling [plain]

    Code
      local({
        cli_div(theme = list(body = list(`vec-sep` = " ... ")))
        pkgs <- cli_vec(paste0("pkg", 1:5), style = list(`vec-sep` = " & ",
          `vec-last` = " & ", color = "blue"))
        cli_text("Packages: {pkgs}.")
      })
    Message
      Packages: pkg1 & pkg2 & pkg3 & pkg4 & pkg5.

# collapsing a cli_vec with styling [ansi]

    Code
      local({
        cli_div(theme = list(body = list(`vec-sep` = " ... ")))
        pkgs <- cli_vec(paste0("pkg", 1:5), style = list(`vec-sep` = " & ",
          `vec-last` = " & ", color = "blue"))
        cli_text("Packages: {pkgs}.")
      })
    Message
      Packages: [34mpkg1[39m & [34mpkg2[39m & [34mpkg3[39m & [34mpkg4[39m & [34mpkg5[39m.

# head

    Code
      cli_text("{v(0,1)}")
    Message
      
    Code
      cli_text("{v(1,1)}")
    Message
      1
    Code
      cli_text("{v(2,1)}")
    Message
      1 and 2
    Code
      cli_text("{v(3,1)}")
    Message
      1, ...
    Code
      cli_text("{v(4,1)}")
    Message
      1, ...
    Code
      cli_text("{v(0,2)}")
    Message
      
    Code
      cli_text("{v(1,2)}")
    Message
      1
    Code
      cli_text("{v(2,2)}")
    Message
      1 and 2
    Code
      cli_text("{v(3,2)}")
    Message
      1, 2, and 3
    Code
      cli_text("{v(4,2)}")
    Message
      1, 2, ...
    Code
      cli_text("{v(0,3)}")
    Message
      
    Code
      cli_text("{v(1,3)}")
    Message
      1
    Code
      cli_text("{v(2,3)}")
    Message
      1 and 2
    Code
      cli_text("{v(3,3)}")
    Message
      1, 2, and 3
    Code
      cli_text("{v(4,3)}")
    Message
      1, 2, 3, and 4
    Code
      cli_text("{v(0,4)}")
    Message
      
    Code
      cli_text("{v(1,4)}")
    Message
      1
    Code
      cli_text("{v(2,4)}")
    Message
      1 and 2
    Code
      cli_text("{v(3,4)}")
    Message
      1, 2, and 3
    Code
      cli_text("{v(4,4)}")
    Message
      1, 2, 3, and 4
    Code
      cli_text("{v(0,5)}")
    Message
      
    Code
      cli_text("{v(1,5)}")
    Message
      1
    Code
      cli_text("{v(2,5)}")
    Message
      1 and 2
    Code
      cli_text("{v(3,5)}")
    Message
      1, 2, and 3
    Code
      cli_text("{v(4,5)}")
    Message
      1, 2, 3, and 4
    Code
      cli_text("{v(10,5)}")
    Message
      1, 2, 3, 4, 5, ...

# both-ends

    Code
      cli_text("{v(0,1)}")
    Message
      
    Code
      cli_text("{v(1,1)}")
    Message
      1
    Code
      cli_text("{v(2,1)}")
    Message
      1 and 2
    Code
      cli_text("{v(3,1)}")
    Message
      1, 2, and 3
    Code
      cli_text("{v(4,1)}")
    Message
      1, 2, 3, and 4
    Code
      cli_text("{v(5,1)}")
    Message
      1, 2, 3, 4, and 5
    Code
      cli_text("{v(6,1)}")
    Message
      1, 2, 3, ..., 5, and 6
    Code
      cli_text("{v(7,1)}")
    Message
      1, 2, 3, ..., 6, and 7
    Code
      cli_text("{v(10,1)}")
    Message
      1, 2, 3, ..., 9, and 10

# both-ends with formatting [plain]

    Code
      cli_text("{.val {v(0,1)}}")
    Message
      
    Code
      cli_text("{.val {v(1,1)}}")
    Message
      1
    Code
      cli_text("{.val {v(2,1)}}")
    Message
      1 and 2
    Code
      cli_text("{.val {v(3,1)}}")
    Message
      1, 2, and 3
    Code
      cli_text("{.val {v(4,1)}}")
    Message
      1, 2, 3, and 4
    Code
      cli_text("{.val {v(5,1)}}")
    Message
      1, 2, 3, 4, and 5
    Code
      cli_text("{.val {v(6,1)}}")
    Message
      1, 2, 3, ..., 5, and 6
    Code
      cli_text("{.val {v(7,1)}}")
    Message
      1, 2, 3, ..., 6, and 7
    Code
      cli_text("{.val {v(10,1)}}")
    Message
      1, 2, 3, ..., 9, and 10
    Code
      cli_text("{.val {v(10,6)}}")
    Message
      1, 2, 3, 4, ..., 9, and 10
    Code
      cli_text("{.val {v(10,10)}}")
    Message
      1, 2, 3, 4, 5, 6, 7, 8, 9, and 10
    Code
      cli_text("{.val {v(11,10)}}")
    Message
      1, 2, 3, 4, 5, 6, 7, 8, 9, 10, and 11

# both-ends with formatting [ansi]

    Code
      cli_text("{.val {v(0,1)}}")
    Message
      
    Code
      cli_text("{.val {v(1,1)}}")
    Message
      [34m1[39m
    Code
      cli_text("{.val {v(2,1)}}")
    Message
      [34m1[39m and [34m2[39m
    Code
      cli_text("{.val {v(3,1)}}")
    Message
      [34m1[39m, [34m2[39m, and [34m3[39m
    Code
      cli_text("{.val {v(4,1)}}")
    Message
      [34m1[39m, [34m2[39m, [34m3[39m, and [34m4[39m
    Code
      cli_text("{.val {v(5,1)}}")
    Message
      [34m1[39m, [34m2[39m, [34m3[39m, [34m4[39m, and [34m5[39m
    Code
      cli_text("{.val {v(6,1)}}")
    Message
      [34m1[39m, [34m2[39m, [34m3[39m, ..., [34m5[39m, and [34m6[39m
    Code
      cli_text("{.val {v(7,1)}}")
    Message
      [34m1[39m, [34m2[39m, [34m3[39m, ..., [34m6[39m, and [34m7[39m
    Code
      cli_text("{.val {v(10,1)}}")
    Message
      [34m1[39m, [34m2[39m, [34m3[39m, ..., [34m9[39m, and [34m10[39m
    Code
      cli_text("{.val {v(10,6)}}")
    Message
      [34m1[39m, [34m2[39m, [34m3[39m, [34m4[39m, ..., [34m9[39m, and [34m10[39m
    Code
      cli_text("{.val {v(10,10)}}")
    Message
      [34m1[39m, [34m2[39m, [34m3[39m, [34m4[39m, [34m5[39m, [34m6[39m, [34m7[39m, [34m8[39m, [34m9[39m, and [34m10[39m
    Code
      cli_text("{.val {v(11,10)}}")
    Message
      [34m1[39m, [34m2[39m, [34m3[39m, [34m4[39m, [34m5[39m, [34m6[39m, [34m7[39m, [34m8[39m, [34m9[39m, [34m10[39m, and [34m11[39m

# ansi_collapse

    Code
      ansi_collapse(l10)
    Output
      [1] "a, b, c, d, e, f, g, h, i, and j"
    Code
      ansi_collapse(l10, trunc = 6)
    Output
      [1] "a, b, c, d, ..., i, and j"
    Code
      ansi_collapse(l10, trunc = 5)
    Output
      [1] "a, b, c, ..., i, and j"
    Code
      ansi_collapse(l10, trunc = 4)
    Output
      [1] "a, b, c, ..., i, and j"
    Code
      ansi_collapse(l10, trunc = 1)
    Output
      [1] "a, b, c, ..., i, and j"
    Code
      ansi_collapse(l10, sep = "; ")
    Output
      [1] "a; b; c; d; e; f; g; h; i, and j"
    Code
      ansi_collapse(l10, sep = "; ", last = "; or ")
    Output
      [1] "a; b; c; d; e; f; g; h; i; or j"
    Code
      ansi_collapse(l10, sep = "; ")
    Output
      [1] "a; b; c; d; e; f; g; h; i, and j"
    Code
      ansi_collapse(l10, sep = "; ", last = "; or ", trunc = 6)
    Output
      [1] "a; b; c; d; ...; i; or j"
    Code
      ansi_collapse(l10, style = "head")
    Output
      [1] "a, b, c, d, e, f, g, h, i, and j"
    Code
      ansi_collapse(l10, trunc = 6, style = "head")
    Output
      [1] "a, b, c, d, e, f, ..."
    Code
      ansi_collapse(l10, trunc = 5, style = "head")
    Output
      [1] "a, b, c, d, e, ..."
    Code
      ansi_collapse(l10, trunc = 4, style = "head")
    Output
      [1] "a, b, c, d, ..."
    Code
      ansi_collapse(l10, trunc = 1, style = "head")
    Output
      [1] "a, ..."
    Code
      ansi_collapse(l10, sep = "; ", style = "head")
    Output
      [1] "a; b; c; d; e; f; g; h; i, and j"
    Code
      ansi_collapse(l10, sep = "; ", last = "; or ", style = "head")
    Output
      [1] "a; b; c; d; e; f; g; h; i; or j"
    Code
      ansi_collapse(l10, sep = "; ", last = "; or ", trunc = 6, style = "head")
    Output
      [1] "a; b; c; d; e; f; ..."

# ansi_collapse with width trimming

    Code
      ansi_collapse(l10, width = 1, style = "head")
    Output
      <cli_ansi_string>
      [1] .
    Code
      ansi_collapse(l10, width = 2, style = "head")
    Output
      <cli_ansi_string>
      [1] ..
    Code
      ansi_collapse(l10, width = 3, style = "head")
    Output
      [1] "..."
    Code
      ansi_collapse(l10, width = 4, style = "head")
    Output
      [1] "a..."
    Code
      ansi_collapse(l10, width = 5, style = "head")
    Output
      [1] "a..."
    Code
      ansi_collapse(l10, width = 6, style = "head")
    Output
      [1] "a, ..."
    Code
      ansi_collapse(l10, width = 7, style = "head")
    Output
      [1] "a, ..."
    Code
      ansi_collapse(l10, width = 8, style = "head")
    Output
      [1] "a, ..."
    Code
      ansi_collapse(l10, width = 9, style = "head")
    Output
      [1] "a, b, ..."
    Code
      ansi_collapse(l10, width = 30, style = "head")
    Output
      [1] "a, b, c, d, e, f, g, h, i, ..."
    Code
      ansi_collapse(l10, width = 31, style = "head")
    Output
      [1] "a, b, c, d, e, f, g, h, i, ..."
    Code
      ansi_collapse(l10, width = 32, style = "head")
    Output
      [1] "a, b, c, d, e, f, g, h, i, and j"
    Code
      ansi_collapse(l10, width = 40, style = "head")
    Output
      [1] "a, b, c, d, e, f, g, h, i, and j"

---

    Code
      ansi_collapse(l10, width = 10, style = "both-ends")
    Condition
      Warning in `collapse_both_ends()`:
      ! finite `width` is not implemented in `cli::ansi_collapse()`.
      i `width = Inf` is used instead.
    Output
      [1] "a, b, c, d, e, f, g, h, i, and j"

# auto closing

    Code
      local({
        cli_div(theme = list(`.xx .emph` = list(before = "itsu:")))
        f <- (function() {
          cli_par(class = "xx")
          cli_text("foo {.emph blah} bar")
        })
        f()
        cli_text("foo {.emph blah} bar")
      })
    Message <cliMessage>
      foo itsu:blah bar
      
      foo blah bar

# opt out of auto closing

    Code
      local({
        cli_div(theme = list(`.xx .emph` = list(before = "itsu:")))
        id <- NULL
        f <- (function() {
          id <<- cli_par(class = "xx", .auto_close = FALSE)
          cli_text("foo {.emph blah} bar")
        })
        f()
        cli_text("foo {.emph blah} bar")
        expect_false(is.null(id))
        cli_end(id)
        cli_text("foo {.emph blah} bar")
      })
    Message <cliMessage>
      foo itsu:blah bar
      foo itsu:blah bar
      
      foo blah bar

# auto closing with special env

    Code
      local({
        cli_div(theme = list(`.xx .emph` = list(before = "itsu:")))
        f <- (function() {
          g()
          cli_text("foo {.emph blah} bar")
        })
        g <- (function() {
          cli_par(class = "xx", .auto_close = TRUE, .envir = parent.frame())
          cli_text("foo {.emph blah} bar")
        })
        f()
        cli_text("foo {.emph blah} bar")
      })
    Message <cliMessage>
      foo itsu:blah bar
      foo itsu:blah bar
      
      foo blah bar

# div with special style

    Code
      f <- (function() {
        cli_div(theme = list(`.xx .emph` = list(before = "itsu:")))
        cli_par(class = "xx")
        cli_text("foo {.emph blah} bar")
      })
      f()
    Message <cliMessage>
      foo itsu:blah bar
      
    Code
      cli_text("foo {.emph blah} bar")
    Message <cliMessage>
      foo blah bar

# margin is squashed

    Code
      local({
        cli_div(theme = list(par = list(`margin-top` = 4, `margin-bottom` = 4)))
        cli_text("three lines")
        cli_par()
        cli_par()
        cli_par()
        cli_text("until here")
        cli_end()
        cli_end()
        cli_end()
        cli_par()
        cli_par()
        cli_par()
        cli_text("no space, still")
        cli_end()
        cli_end()
        cli_end()
        cli_text("three lines again")
      })
    Message <cliMessage>
      three lines
      
      
      
      
      until here
      
      
      
      
      no space, still
      
      
      
      
      three lines again

# before and after work properly

    Code
      local({
        cli_div(theme = list(`div.alert-success` = list(before = "!!!")))
        cli_alert_success("{.pkg foobar} is good")
      })
    Message <cliMessage>
      !!!foobar is good

# text is wrapped

    Code
      local({
        cli_div(class = "testcli", theme = test_style())
        cli_h1("Header")
        cli_text(lorem_ipsum())
      })
    Message <cliMessage>
      
      Header
      
      Duis quis magna incididunt nulla commodo minim non
      exercitation nostrud ullamco dolor exercitation ut veniam.
      Fugiat irure tempor commodo voluptate ut. In et tempor
      excepteur quis. Qui et nisi ad quis ad cupidatat tempor
      laborum est excepteur aliqua veniam. Velit sunt magna
      veniam Lorem elit enim et pariatur. Occaecat mollit
      consequat dolore in. Veniam officia labore reprehenderit
      culpa dolore quis nisi do aliqua commodo deserunt.
      Cupidatat nostrud ad est in ad laboris consectetur esse
      minim. Irure do anim anim ea mollit ad cupidatat ullamco
      ullamco nulla elit in. Lorem Lorem deserunt deserunt et ut
      velit nulla nulla ipsum ad laborum quis.

# add/remove/list themes [plain]

    Code
      cli_par(class = "green")
      cli_text(lorem_ipsum())
    Message <cliMessage>
      Non incididunt mollit ullamco duis officia proident. Laborum esse mollit mollit
      eiusmod tempor cupidatat. In commodo anim irure nostrud. Deserunt nisi amet
      laborum magna aliqua. Do esse consectetur ut deserunt nulla Lorem non. Fugiat
      est dolore deserunt aliqua amet et esse dolore elit exercitation sint
      exercitation non ipsum.
    Code
      cli_end()
    Message <cliMessage>
      

# add/remove/list themes [ansi]

    Code
      cli_par(class = "green")
      cli_text(lorem_ipsum())
    Message <cliMessage>
      [32mNon incididunt mollit ullamco duis officia proident. Laborum esse mollit mollit[39m
      [32meiusmod tempor cupidatat. In commodo anim irure nostrud. Deserunt nisi amet[39m
      [32mlaborum magna aliqua. Do esse consectetur ut deserunt nulla Lorem non. Fugiat[39m
      [32mest dolore deserunt aliqua amet et esse dolore elit exercitation sint[39m
      [32mexercitation non ipsum.[39m
    Code
      cli_end()
    Message <cliMessage>
      

# add/remove/list themes [unicode]

    Code
      cli_par(class = "green")
      cli_text(lorem_ipsum())
    Message <cliMessage>
      Non incididunt mollit ullamco duis officia proident. Laborum esse mollit mollit
      eiusmod tempor cupidatat. In commodo anim irure nostrud. Deserunt nisi amet
      laborum magna aliqua. Do esse consectetur ut deserunt nulla Lorem non. Fugiat
      est dolore deserunt aliqua amet et esse dolore elit exercitation sint
      exercitation non ipsum.
    Code
      cli_end()
    Message <cliMessage>
      

# add/remove/list themes [fancy]

    Code
      cli_par(class = "green")
      cli_text(lorem_ipsum())
    Message <cliMessage>
      [32mNon incididunt mollit ullamco duis officia proident. Laborum esse mollit mollit[39m
      [32meiusmod tempor cupidatat. In commodo anim irure nostrud. Deserunt nisi amet[39m
      [32mlaborum magna aliqua. Do esse consectetur ut deserunt nulla Lorem non. Fugiat[39m
      [32mest dolore deserunt aliqua amet et esse dolore elit exercitation sint[39m
      [32mexercitation non ipsum.[39m
    Code
      cli_end()
    Message <cliMessage>
      

# explicit formatter is used, and combined

    Code
      cli_text("this is {.emph it}, really")
    Message <cliMessage>
      this is (((<<it>>))), really

# user's override

    Code
      local({
        start_app(theme = custom, .auto_close = FALSE)
        cli_alert("Alert!")
        stop_app()
        withr::local_options(cli.user_theme = override)
        start_app(theme = custom, .auto_close = FALSE)
        cli_alert("Alert!")
        stop_app()
      })
    Message <cliMessage>
      custom:Alert!
      custom:Alert!override:

# NULL will undo a style property

    Code
      local({
        cli_alert("this has an arrow")
        cli_div(theme = list(.alert = list(before = NULL)))
        cli_alert("this does not")
      })
    Message <cliMessage>
      > this has an arrow
      this does not

# NULL will undo color [ansi]

    Code
      local({
        cli_alert("{.emph {.val this is blue}}")
        cli_div(theme = list(span.val = list(color = NULL)))
        cli_alert("{.emph {.val this is not}}")
      })
    Message <cliMessage>
      > [34m[3m[34m"this is blue"[34m[23m[39m
      > [0m[22m[23m[24m[27m[28m[29m[49m[3m[0m[22m[3m[24m[27m[28m[29m[49m"this is not"[0m[22m[23m[24m[27m[28m[29m[49m[23m[39m

---

    Code
      local({
        cli_alert("{.emph {.val this is blue}}")
        cli_div(theme = list(span.val = list(color = "none")))
        cli_alert("{.emph {.val this is not}}")
      })
    Message <cliMessage>
      > [34m[3m[34m"this is blue"[34m[23m[39m
      > [0m[22m[23m[24m[27m[28m[29m[49m[3m[0m[22m[3m[24m[27m[28m[29m[49m"this is not"[0m[22m[23m[24m[27m[28m[29m[49m[23m[39m

# NULL will undo background color [ansi]

    Code
      local({
        cli_alert("{.emph {.code this has bg color}}")
        cli_div(theme = list(span = list(`background-color` = NULL)))
        cli_alert("{.emph {.code this does not}}")
      })
    Message <cliMessage>
      > [3m`this has bg color`[23m
      > [0m[22m[23m[24m[27m[28m[29m[39m[3m[0m[22m[3m[24m[27m[28m[29m[39m`this does not`[0m[22m[23m[24m[27m[28m[29m[39m[23m[49m

---

    Code
      local({
        cli_alert("{.emph {.code this has bg color}}")
        cli_div(theme = list(span = list(`background-color` = "none")))
        cli_alert("{.emph {.code this does not}}")
      })
    Message <cliMessage>
      > [3m`this has bg color`[23m
      > [0m[22m[23m[24m[27m[28m[29m[39m[3m[0m[22m[3m[24m[27m[28m[29m[39m`this does not`[0m[22m[23m[24m[27m[28m[29m[39m[23m[49m

# ul [plain]

    Code
      local({
        lid <- cli_ul()
        cli_li("foo")
        cli_li(c("bar", "foobar"))
        cli_end(lid)
      })
    Message <cliMessage>
      * foo
      * bar
      * foobar

# ul [unicode]

    Code
      local({
        lid <- cli_ul()
        cli_li("foo")
        cli_li(c("bar", "foobar"))
        cli_end(lid)
      })
    Message <cliMessage>
      • foo
      • bar
      • foobar

# ol [plain]

    Code
      local({
        cli_div(theme = list(ol = list()))
        lid <- cli_ol()
        cli_li("foo")
        cli_li(c("bar", "foobar"))
        cli_end(lid)
      })
    Message <cliMessage>
      1. foo
      2. bar
      3. foobar

# ol [unicode]

    Code
      local({
        cli_div(theme = list(ol = list()))
        lid <- cli_ol()
        cli_li("foo")
        cli_li(c("bar", "foobar"))
        cli_end(lid)
      })
    Message <cliMessage>
      1. foo
      2. bar
      3. foobar

# ul ul [plain]

    Code
      local({
        cli_div(theme = list(`ul ul` = list(`list-style-type` = "-", `margin-left` = 2)))
        lid <- cli_ul()
        cli_li("1")
        lid2 <- cli_ul()
        cli_li("1 1")
        cli_li(c("1 2", "1 3"))
        cli_end(lid2)
        cli_li("2")
        cli_end(lid)
      })
    Message <cliMessage>
      * 1
          - 1 1
          - 1 2
          - 1 3
      * 2

# ul ul [unicode]

    Code
      local({
        cli_div(theme = list(`ul ul` = list(`list-style-type` = "-", `margin-left` = 2)))
        lid <- cli_ul()
        cli_li("1")
        lid2 <- cli_ul()
        cli_li("1 1")
        cli_li(c("1 2", "1 3"))
        cli_end(lid2)
        cli_li("2")
        cli_end(lid)
      })
    Message <cliMessage>
      • 1
          - 1 1
          - 1 2
          - 1 3
      • 2

# ul ol [plain]

    Code
      local({
        cli_div(theme = list(li = list(`margin-left` = 2)))
        lid <- cli_ul()
        cli_li("1")
        lid2 <- cli_ol()
        cli_li("1 1")
        cli_li(c("1 2", "1 3"))
        cli_end(lid2)
        cli_li("2")
        cli_end(lid)
      })
    Message <cliMessage>
        * 1
          1. 1 1
          2. 1 2
          3. 1 3
        * 2

# ul ol [unicode]

    Code
      local({
        cli_div(theme = list(li = list(`margin-left` = 2)))
        lid <- cli_ul()
        cli_li("1")
        lid2 <- cli_ol()
        cli_li("1 1")
        cli_li(c("1 2", "1 3"))
        cli_end(lid2)
        cli_li("2")
        cli_end(lid)
      })
    Message <cliMessage>
        • 1
          1. 1 1
          2. 1 2
          3. 1 3
        • 2

# ol ol [plain]

    Code
      local({
        cli_div(theme = list(li = list(`margin-left` = 2), `li li` = list(
          `margin-left` = 2)))
        lid <- cli_ol()
        cli_li("1")
        lid2 <- cli_ol()
        cli_li("1 1")
        cli_li(c("1 2", "1 3"))
        cli_end(lid2)
        cli_li("2")
        cli_end(lid)
      })
    Message <cliMessage>
        1. 1
          1. 1 1
          2. 1 2
          3. 1 3
        2. 2

# ol ol [unicode]

    Code
      local({
        cli_div(theme = list(li = list(`margin-left` = 2), `li li` = list(
          `margin-left` = 2)))
        lid <- cli_ol()
        cli_li("1")
        lid2 <- cli_ol()
        cli_li("1 1")
        cli_li(c("1 2", "1 3"))
        cli_end(lid2)
        cli_li("2")
        cli_end(lid)
      })
    Message <cliMessage>
        1. 1
          1. 1 1
          2. 1 2
          3. 1 3
        2. 2

# ol ul [plain]

    Code
      local({
        cli_div(theme = list(ul = list(`margin-left` = 2)))
        lid <- cli_ol()
        cli_li("1")
        lid2 <- cli_ul()
        cli_li("1 1")
        cli_li(c("1 2", "1 3"))
        cli_end(lid2)
        cli_li("2")
        cli_end(lid)
      })
    Message <cliMessage>
      1. 1
          * 1 1
          * 1 2
          * 1 3
      2. 2

# ol ul [unicode]

    Code
      local({
        cli_div(theme = list(ul = list(`margin-left` = 2)))
        lid <- cli_ol()
        cli_li("1")
        lid2 <- cli_ul()
        cli_li("1 1")
        cli_li(c("1 2", "1 3"))
        cli_end(lid2)
        cli_li("2")
        cli_end(lid)
      })
    Message <cliMessage>
      1. 1
          • 1 1
          • 1 2
          • 1 3
      2. 2

# starting with an item [plain]

    Code
      local({
        cli_li("foo")
        cli_li(c("bar", "foobar"))
        cli_end()
        cli_end()
      })
    Message <cliMessage>
      * foo
      * bar
      * foobar

# starting with an item [unicode]

    Code
      local({
        cli_li("foo")
        cli_li(c("bar", "foobar"))
        cli_end()
        cli_end()
      })
    Message <cliMessage>
      • foo
      • bar
      • foobar

# ol, with first item [plain]

    Code
      local({
        cli_div(theme = list(ol = list()))
        lid <- cli_ol("foo", .close = FALSE)
        cli_li(c("bar", "foobar"))
        cli_end(lid)
      })
    Message <cliMessage>
      1. foo
      2. bar
      3. foobar

# ol, with first item [unicode]

    Code
      local({
        cli_div(theme = list(ol = list()))
        lid <- cli_ol("foo", .close = FALSE)
        cli_li(c("bar", "foobar"))
        cli_end(lid)
      })
    Message <cliMessage>
      1. foo
      2. bar
      3. foobar

# ul, with first item [plain]

    Code
      local({
        lid <- cli_ul("foo", .close = FALSE)
        cli_li(c("bar", "foobar"))
        cli_end(lid)
      })
    Message <cliMessage>
      * foo
      * bar
      * foobar

# ul, with first item [unicode]

    Code
      local({
        lid <- cli_ul("foo", .close = FALSE)
        cli_li(c("bar", "foobar"))
        cli_end(lid)
      })
    Message <cliMessage>
      • foo
      • bar
      • foobar

# dl [plain]

    Code
      local({
        cli_div(theme = list(ul = list()))
        lid <- cli_dl()
        cli_li(c(this = "foo"))
        cli_li(c(that = "bar", other = "foobar"))
        cli_end(lid)
      })
    Message <cliMessage>
      this: foo
      that: bar
      other: foobar

# dl [unicode]

    Code
      local({
        cli_div(theme = list(ul = list()))
        lid <- cli_dl()
        cli_li(c(this = "foo"))
        cli_li(c(that = "bar", other = "foobar"))
        cli_end(lid)
      })
    Message <cliMessage>
      this: foo
      that: bar
      other: foobar

# dl dl [plain]

    Code
      local({
        cli_div(theme = list(li = list(`margin-left` = 2)))
        lid <- cli_dl()
        cli_li(c(a = "1"))
        lid2 <- cli_dl()
        cli_li(c(`a-a` = "1 1"))
        cli_li(c(`a-b` = "1 2", `a-c` = "1 3"))
        cli_end(lid2)
        cli_li(c(b = "2"))
        cli_end(lid)
      })
    Message <cliMessage>
        a: 1
          a-a: 1 1
          a-b: 1 2
          a-c: 1 3
        b: 2

# dl dl [unicode]

    Code
      local({
        cli_div(theme = list(li = list(`margin-left` = 2)))
        lid <- cli_dl()
        cli_li(c(a = "1"))
        lid2 <- cli_dl()
        cli_li(c(`a-a` = "1 1"))
        cli_li(c(`a-b` = "1 2", `a-c` = "1 3"))
        cli_end(lid2)
        cli_li(c(b = "2"))
        cli_end(lid)
      })
    Message <cliMessage>
        a: 1
          a-a: 1 1
          a-b: 1 2
          a-c: 1 3
        b: 2

# dl ol [plain]

    Code
      local({
        cli_div(theme = list(li = list(`margin-left` = 2)))
        lid <- cli_dl()
        cli_li(c(a = "1"))
        lid2 <- cli_ol()
        cli_li(c("1 1"))
        cli_li(c("1 2", "1 3"))
        cli_end(lid2)
        cli_li(c(b = "2"))
        cli_end(lid)
      })
    Message <cliMessage>
        a: 1
          1. 1 1
          2. 1 2
          3. 1 3
        b: 2

# dl ol [unicode]

    Code
      local({
        cli_div(theme = list(li = list(`margin-left` = 2)))
        lid <- cli_dl()
        cli_li(c(a = "1"))
        lid2 <- cli_ol()
        cli_li(c("1 1"))
        cli_li(c("1 2", "1 3"))
        cli_end(lid2)
        cli_li(c(b = "2"))
        cli_end(lid)
      })
    Message <cliMessage>
        a: 1
          1. 1 1
          2. 1 2
          3. 1 3
        b: 2

# dl ul [plain]

    Code
      local({
        cli_div(theme = list(li = list(`margin-left` = 2)))
        lid <- cli_dl()
        cli_li(c(a = "1"))
        lid2 <- cli_ul()
        cli_li(c("1 1"))
        cli_li(c("1 2", "1 3"))
        cli_end(lid2)
        cli_li(c(b = "2"))
        cli_end(lid)
      })
    Message <cliMessage>
        a: 1
          * 1 1
          * 1 2
          * 1 3
        b: 2

# dl ul [unicode]

    Code
      local({
        cli_div(theme = list(li = list(`margin-left` = 2)))
        lid <- cli_dl()
        cli_li(c(a = "1"))
        lid2 <- cli_ul()
        cli_li(c("1 1"))
        cli_li(c("1 2", "1 3"))
        cli_end(lid2)
        cli_li(c(b = "2"))
        cli_end(lid)
      })
    Message <cliMessage>
        a: 1
          • 1 1
          • 1 2
          • 1 3
        b: 2

# ol dl [plain]

    Code
      local({
        cli_div(theme = list(li = list(`margin-left` = 2)))
        lid <- cli_ol()
        cli_li("1")
        lid2 <- cli_dl()
        cli_li(c(`a-a` = "1 1"))
        cli_li(c(`a-b` = "1 2", `a-c` = "1 3"))
        cli_end(lid2)
        cli_li("2")
        cli_end(lid)
      })
    Message <cliMessage>
        1. 1
          a-a: 1 1
          a-b: 1 2
          a-c: 1 3
        2. 2

# ol dl [unicode]

    Code
      local({
        cli_div(theme = list(li = list(`margin-left` = 2)))
        lid <- cli_ol()
        cli_li("1")
        lid2 <- cli_dl()
        cli_li(c(`a-a` = "1 1"))
        cli_li(c(`a-b` = "1 2", `a-c` = "1 3"))
        cli_end(lid2)
        cli_li("2")
        cli_end(lid)
      })
    Message <cliMessage>
        1. 1
          a-a: 1 1
          a-b: 1 2
          a-c: 1 3
        2. 2

# ul dl [plain]

    Code
      local({
        cli_div(theme = list(li = list(`margin-left` = 2)))
        lid <- cli_ul()
        cli_li("1")
        lid2 <- cli_dl()
        cli_li(c(`a-a` = "1 1"))
        cli_li(c(`a-b` = "1 2", `a-c` = "1 3"))
        cli_end(lid2)
        cli_li("2")
        cli_end(lid)
      })
    Message <cliMessage>
        * 1
          a-a: 1 1
          a-b: 1 2
          a-c: 1 3
        * 2

# ul dl [unicode]

    Code
      local({
        cli_div(theme = list(li = list(`margin-left` = 2)))
        lid <- cli_ul()
        cli_li("1")
        lid2 <- cli_dl()
        cli_li(c(`a-a` = "1 1"))
        cli_li(c(`a-b` = "1 2", `a-c` = "1 3"))
        cli_end(lid2)
        cli_li("2")
        cli_end(lid)
      })
    Message <cliMessage>
        • 1
          a-a: 1 1
          a-b: 1 2
          a-c: 1 3
        • 2

# dl, with first item [plain]

    Code
      local({
        cli_div(theme = list(ul = list()))
        lid <- cli_dl(c(this = "foo"), .close = FALSE)
        cli_li(c(that = "bar", other = "foobar"))
        cli_end(lid)
      })
    Message <cliMessage>
      this: foo
      that: bar
      other: foobar

# dl, with first item [unicode]

    Code
      local({
        cli_div(theme = list(ul = list()))
        lid <- cli_dl(c(this = "foo"), .close = FALSE)
        cli_li(c(that = "bar", other = "foobar"))
        cli_end(lid)
      })
    Message <cliMessage>
      this: foo
      that: bar
      other: foobar

# styling pieces of a dl [ansi]

    Code
      local({
        cli_div(theme = list(.dt = list(after = " -> "), .dd = list(color = "blue")))
        cli_dl(c(foo = "bar", bar = "baz"))
      })
    Message <cliMessage>
      foo -> [34mbar[39m
      bar -> [34mbaz[39m

# cli_dl edge cases

    Code
      cli_dl(c(abc = "foo", empty = "", def = "bar"))
    Message <cliMessage>
      abc: foo
      empty:
      def: bar

# diff_chr

    Code
      d$lcs
    Output
        operation offset length old_offset new_offset
      1     match      0      8          8          8
      2    delete      8      1          9          8
      3    insert      8      1          9          9
      4     match      9      7         16         16
      5    insert     16      2         16         18
      6     match     16      1         17         19

---

    Code
      d$lcs
    Output
        operation offset length old_offset new_offset
      1     match      0      8          8          8
      2    delete      8      1          9          8
      3    insert      8      1          9          9
      4     match      9      7         16         16
      5    delete     16      2         18         16
      6     match     18      1         19         17

# diff_chr [plain]

    Code
      d
    Output
      @@ -6,7 +6,7 @@
       1
       1
       2
      -3
      +10
       4
       4
       4
      @@ -14,4 +14,6 @@
       4
       4
       4
      +6
      +7
       5

---

    Code
      d$lcs
    Output
        operation offset length old_offset new_offset
      1     match      0      8          8          8
      2    delete      8      1          9          8
      3    insert      8      1          9          9
      4     match      9      7         16         16
      5    insert     16      2         16         18
      6     match     16      1         17         19

# diff_chr [ansi]

    Code
      d
    Output
      @@ -6,7 +6,7 @@
       1
       1
       2
      [34m-3[39m
      [32m+10[39m
       4
       4
       4
      @@ -14,4 +14,6 @@
       4
       4
       4
      [32m+6[39m
      [32m+7[39m
       5

---

    Code
      d$lcs
    Output
        operation offset length old_offset new_offset
      1     match      0      8          8          8
      2    delete      8      1          9          8
      3    insert      8      1          9          9
      4     match      9      7         16         16
      5    insert     16      2         16         18
      6     match     16      1         17         19

# diff_chr edge cases

    Code
      diff_chr(character(), character())

---

    Code
      diff_chr(character(), character())$lcs
    Output
      [1] operation  offset     length     old_offset new_offset
      <0 rows> (or 0-length row.names)

---

    Code
      diff_chr("a", character())
    Output
      @@ -1 +0 @@
      -a

---

    Code
      diff_chr(character(), "b")
    Output
      @@ -0 +1 @@
      +b

---

    Code
      diff_chr("a", "a")

---

    Code
      diff_chr(letters, letters)

---

    Code
      diff_chr(c("a", NA, "a2"), "b")
    Output
      @@ -1,3 +1 @@
      -a
      -NA
      -a2
      +b

---

    Code
      diff_chr(NA_character_, "NA")
    Output
      @@ -1 +1 @@
      -NA
      +NA

# format.cli_diff_chr context

    Code
      print(d, context = 1)
    Output
      @@ -8,3 +8,3 @@
       2
      -3
      +10
       4
      @@ -16,2 +16,4 @@
       4
      +6
      +7
       5

---

    Code
      print(d, context = 0)
    Output
      @@ -9 +9 @@
      -3
      +10
      @@ -17,0 +17,2 @@
      +6
      +7

---

    Code
      print(d, context = Inf)
    Output
       1
       1
       1
       1
       1
       1
       1
       2
      -3
      +10
       4
       4
       4
       4
       4
       4
       4
      +6
      +7
       5

---

    Code
      print(d2, context = Inf)
    Output
       foo
       bar

# diff_str [plain]

    Code
      d
    Output
      {+PRE+}abcdefg[-hijklm-]{+MIDDLE+}nopqrstuvwxyz{+POST+}

# diff_str [ansi]

    Code
      d
    Output
      [42m[30mPRE[39m[49mabcdefg[44m[30mhijklm[39m[49m[42m[30mMIDDLE[39m[49mnopqrstuvwxyz[42m[30mPOST[39m[49m

# spark_bar [plain]

    Code
      spark_bar(seq(0, 1, length = 8))
    Output
      __,,**##
    Code
      spark_bar(c(0, NA, 0.5, NA, 1))
    Output
      _ , #

# spark_bar [unicode]

    Code
      spark_bar(seq(0, 1, length = 8))
    Output
      ▁▂▃▄▅▆▇█
    Code
      spark_bar(c(0, NA, 0.5, NA, 1))
    Output
      ▁ ▄ █

# spark_line [plain]

    Code
      spark_line(seq(0, 1, length = 10))
    Output
      _,,-^

# spark_line [unicode]

    Code
      spark_line(seq(0, 1, length = 10))
    Output
      ⣀⡠⠔⠊⠉

# is_string

    Code
      stopifnot(is_string(n))
    Condition
      Error:
      ! is_string(n) is not TRUE

---

    Code
      stopifnot(is_string(n))
    Condition
      Error:
      ! is_string(n) is not TRUE

---

    Code
      stopifnot(is_string(n))
    Condition
      Error:
      ! is_string(n) is not TRUE

---

    Code
      stopifnot(is_string(n))
    Condition
      Error:
      ! is_string(n) is not TRUE

---

    Code
      stopifnot(is_string(n))
    Condition
      Error:
      ! is_string(n) is not TRUE

---

    Code
      stopifnot(is_string(n))
    Condition
      Error:
      ! is_string(n) is not TRUE

---

    Code
      stopifnot(is_string(n))
    Condition
      Error:
      ! is_string(n) is not TRUE

# is_border_style

    Code
      stopifnot(is_border_style("blahblahxxx"))
    Condition
      Error:
      ! is_border_style("blahblahxxx") is not TRUE

# is_padding_or_margin

    Code
      stopifnot(is_padding_or_margin(b))
    Condition
      Error:
      ! is_padding_or_margin(b) is not TRUE

---

    Code
      stopifnot(is_padding_or_margin(b))
    Condition
      Error:
      ! is_padding_or_margin(b) is not TRUE

---

    Code
      stopifnot(is_padding_or_margin(b))
    Condition
      Error:
      ! is_padding_or_margin(b) is not TRUE

---

    Code
      stopifnot(is_padding_or_margin(b))
    Condition
      Error:
      ! is_padding_or_margin(b) is not TRUE

---

    Code
      stopifnot(is_padding_or_margin(b))
    Condition
      Error:
      ! is_padding_or_margin(b) is not TRUE

---

    Code
      stopifnot(is_padding_or_margin(b))
    Condition
      Error:
      ! is_padding_or_margin(b) is not TRUE

---

    Code
      stopifnot(is_padding_or_margin(b))
    Condition
      Error:
      ! is_padding_or_margin(b) is not TRUE

---

    Code
      stopifnot(is_padding_or_margin(b))
    Condition
      Error:
      ! is_padding_or_margin(b) is not TRUE

---

    Code
      stopifnot(is_padding_or_margin(b))
    Condition
      Error:
      ! is_padding_or_margin(b) is not TRUE

---

    Code
      stopifnot(is_padding_or_margin(b))
    Condition
      Error:
      ! is_padding_or_margin(b) is not TRUE

---

    Code
      stopifnot(is_padding_or_margin(b))
    Condition
      Error:
      ! is_padding_or_margin(b) is not TRUE

---

    Code
      stopifnot(is_padding_or_margin(b))
    Condition
      Error:
      ! is_padding_or_margin(b) is not TRUE

# is_col

    Code
      stopifnot(is_col(b))
    Condition
      Error:
      ! is_col(b) is not TRUE

---

    Code
      stopifnot(is_col(b))
    Condition
      Error:
      ! is_col(b) is not TRUE

---

    Code
      stopifnot(is_col(b))
    Condition
      Error:
      ! is_col(b) is not TRUE

# is_count

    Code
      stopifnot(is_count(n))
    Condition
      Error:
      ! is_count(n) is not TRUE

---

    Code
      stopifnot(is_count(n))
    Condition
      Error:
      ! is_count(n) is not TRUE

---

    Code
      stopifnot(is_count(n))
    Condition
      Error:
      ! is_count(n) is not TRUE

---

    Code
      stopifnot(is_count(n))
    Condition
      Error:
      ! is_count(n) is not TRUE

---

    Code
      stopifnot(is_count(n))
    Condition
      Error:
      ! is_count(n) is not TRUE

---

    Code
      stopifnot(is_count(n))
    Condition
      Error:
      ! is_count(n) is not TRUE

---

    Code
      stopifnot(is_count(n))
    Condition
      Error:
      ! is_count(n) is not TRUE

---

    Code
      stopifnot(is_count(n))
    Condition
      Error:
      ! is_count(n) is not TRUE

# cli_abort(.internal = TRUE) reports the correct function (r-lib/rlang#1386)

    Code
      (expect_error(fn()))
    Output
      <error/rlang_error>
      Error in `fn()`:
      ! Message.
      i This is an internal error that was detected in the base package.

# does not break

    Code
      local({
        withr::local_options(cli.width = 40)
        str30 <- "123456789 123456789 1234567890"
        cli_text(c(str30, "this is not breaking"))
      })
    Message <cliMessage>
      123456789 123456789
      1234567890this is not breaking

# errors

    Code
      fun()
    Condition
      Error:
      ! ! Error in a deferred `on.exit()` clause
      Caused by error in `1 + ""` at defer.R:209:7:
      ! non-numeric argument to binary operator

# glue errors

    Code
      cli_h1("foo { asdfasdfasdf } bar")
    Condition
      Error:
      ! ! Could not evaluate cli `{}` expression: ` asdfasdfasdf `.
      Caused by error in `eval(expr, envir = envir)` at inline.R:323:7:
      ! object 'asdfasdfasdf' not found
    Code
      cli_text("foo {cmd {dsfsdf()}}")
    Condition
      Error:
      ! ! Could not parse cli `{}` expression: `cmd {dsfsdf()}`.
      Caused by error in `parse(text = code, keep.source = FALSE)` at inline.R:317:7:
      ! <text>:1:5: unexpected '{'
      1: cmd {
              ^

# centered label

    Code
      rule(left = "label", center = "label")
    Condition
      Error in `rule()`:
      ! 'center' cannot be specified with 'left' or 'right'
    Code
      rule(center = "label", right = "label")
    Condition
      Error in `rule()`:
      ! 'center' cannot be specified with 'left' or 'right'

# print.cli_rule

    Code
      rule("foo")
    Output
      -- foo -------------

# various errors

    Code
      cli_text("xx {__cannot-parse-this__} yy")
    Condition
      Error:
      ! ! Could not parse cli `{}` expression: `__cannot-parse-th...`.
      Caused by error in `parse(text = code, keep.source = FALSE)`:
      ! <text>:1:2: unexpected input
      1: __
           ^

# trust inspect

<!---MARKER_GEN_START-->
Return low-level information about keys and signatures

### Options

| Name       | Type   | Default | Description                                      |
|:-----------|:-------|:--------|:-------------------------------------------------|
| `--pretty` | `bool` |         | Print the information in a human friendly format |


<!---MARKER_GEN_END-->

## Description

`docker trust inspect` provides low-level JSON information on signed repositories.
This includes all image tags that are signed, who signed them, and who can sign
new tags.

## Examples

### Get low-level details about signatures for a single image tag

Use the `docker trust inspect` to get trust information about an image. The
following example prints trust information for the `alpine:latest` image:

```console
$ docker trust inspect alpine:latest
```

The output is in JSON format, for example:

```json
[
  {
    "Name": "alpine:latest",
    "SignedTags": [
      {
        "SignedTag": "latest",
        "Digest": "d6bfc3baf615dc9618209a8d607ba2a8103d9c8a405b3bd8741d88b4bef36478",
        "Signers": [
          "Repo Admin"
        ]
      }
    ],
    "Signers": [],
    "AdministrativeKeys": [
      {
        "Name": "Repository",
        "Keys": [
            {
                "ID": "5a46c9aaa82ff150bb7305a2d17d0c521c2d784246807b2dc611f436a69041fd"
            }
        ]
      },
      {
        "Name": "Root",
        "Keys": [
            {
                "ID": "a2489bcac7a79aa67b19b96c4a3bf0c675ffdf00c6d2fabe1a5df1115e80adce"
            }
        ]
      }
    ]
  }
]
```

The `SignedTags` key will list the `SignedTag` name, its `Digest`,
and the `Signers` responsible for the signature.

`AdministrativeKeys` will list the `Repository` and `Root` keys.

If signers are set up for the repository via other `docker trust`
commands, `docker trust inspect` includes a `Signers` key:

```console
$ docker trust inspect my-image:purple
```

The output is in JSON format, for example:

```json
[
  {
    "Name": "my-image:purple",
    "SignedTags": [
      {
        "SignedTag": "purple",
        "Digest": "941d3dba358621ce3c41ef67b47cf80f701ff80cdf46b5cc86587eaebfe45557",
        "Signers": [
          "alice",
          "bob",
          "carol"
        ]
      }
    ],
    "Signers": [
      {
        "Name": "alice",
        "Keys": [
            {
                "ID": "04dd031411ed671ae1e12f47ddc8646d98f135090b01e54c3561e843084484a3"
            },
            {
                "ID": "6a11e4898a4014d400332ab0e096308c844584ff70943cdd1d6628d577f45fd8"
            }
        ]
      },
      {
        "Name": "bob",
        "Keys": [
            {
                "ID": "433e245c656ae9733cdcc504bfa560f90950104442c4528c9616daa45824ccba"
            }
        ]
      },
      {
        "Name": "carol",
        "Keys": [
            {
                "ID": "d32fa8b5ca08273a2880f455fcb318da3dc80aeae1a30610815140deef8f30d9"
            },
            {
                "ID": "9a8bbec6ba2af88a5fad6047d428d17e6d05dbdd03d15b4fc8a9a0e8049cd606"
            }
        ]
      }
    ],
    "AdministrativeKeys": [
      {
        "Name": "Repository",
        "Keys": [
            {
                "ID": "27df2c8187e7543345c2e0bf3a1262e0bc63a72754e9a7395eac3f747ec23a44"
            }
        ]
      },
      {
        "Name": "Root",
        "Keys": [
            {
                "ID": "40b66ccc8b176be8c7d365a17f3e046d1c3494e053dd57cfeacfe2e19c4f8e8f"
            }
        ]
      }
    ]
  }
]
```

If the image tag is unsigned or unavailable, `docker trust inspect` does not
display any signed tags.

```console
$ docker trust inspect unsigned-img

no signatures or cannot access unsigned-img
```

However, if other tags are signed in the same image repository,
`docker trust inspect` reports relevant key information:

```console
$ docker trust inspect alpine:unsigned
```

The output is in JSON format, for example:

```json
[
  {
    "Name": "alpine:unsigned",
    "Signers": [],
    "AdministrativeKeys": [
      {
        "Name": "Repository",
        "Keys": [
          {
            "ID": "5a46c9aaa82ff150bb7305a2d17d0c521c2d784246807b2dc611f436a69041fd"
          }
        ]
      },
      {
        "Name": "Root",
        "Keys": [
          {
            "ID": "a2489bcac7a79aa67b19b96c4a3bf0c675ffdf00c6d2fabe1a5df1115e80adce"
          }
        ]
      }
    ]
  }
]
```

### Get details about signatures for all image tags in a repository

If no tag is specified, `docker trust inspect` will report details for all
signed tags in the repository:

```console
$ docker trust inspect alpine
```

The output is in JSON format, for example:

```json
[
  {
    "Name": "alpine",
    "SignedTags": [
      {
        "SignedTag": "3.5",
        "Digest": "b007a354427e1880de9cdba533e8e57382b7f2853a68a478a17d447b302c219c",
        "Signers": [
          "Repo Admin"
        ]
      },
      {
        "SignedTag": "3.6",
        "Digest": "d6bfc3baf615dc9618209a8d607ba2a8103d9c8a405b3bd8741d88b4bef36478",
        "Signers": [
          "Repo Admin"
        ]
      },
      {
        "SignedTag": "edge",
        "Digest": "23e7d843e63a3eee29b6b8cfcd10e23dd1ef28f47251a985606a31040bf8e096",
        "Signers": [
          "Repo Admin"
        ]
      },
      {
        "SignedTag": "latest",
        "Digest": "d6bfc3baf615dc9618209a8d607ba2a8103d9c8a405b3bd8741d88b4bef36478",
        "Signers": [
          "Repo Admin"
        ]
      }
    ],
    "Signers": [],
    "AdministrativeKeys": [
      {
        "Name": "Repository",
        "Keys": [
          {
            "ID": "5a46c9aaa82ff150bb7305a2d17d0c521c2d784246807b2dc611f436a69041fd"
          }
        ]
      },
      {
        "Name": "Root",
        "Keys": [
          {
            "ID": "a2489bcac7a79aa67b19b96c4a3bf0c675ffdf00c6d2fabe1a5df1115e80adce"
          }
        ]
      }
    ]
  }
]
```


### Get details about signatures for multiple images

`docker trust inspect` can take multiple repositories and images as arguments,
and reports the results in an ordered list:

```console
$ docker trust inspect alpine notary
```

The output is in JSON format, for example:

```json
[
  {
    "Name": "alpine",
    "SignedTags": [
      {
        "SignedTag": "3.5",
        "Digest": "b007a354427e1880de9cdba533e8e57382b7f2853a68a478a17d447b302c219c",
        "Signers": [
          "Repo Admin"
        ]
      },
      {
        "SignedTag": "3.6",
        "Digest": "d6bfc3baf615dc9618209a8d607ba2a8103d9c8a405b3bd8741d88b4bef36478",
        "Signers": [
          "Repo Admin"
        ]
      },
      {
        "SignedTag": "edge",
        "Digest": "23e7d843e63a3eee29b6b8cfcd10e23dd1ef28f47251a985606a31040bf8e096",
        "Signers": [
          "Repo Admin"
        ]
      },
      {
        "SignedTag": "integ-test-base",
        "Digest": "3952dc48dcc4136ccdde37fbef7e250346538a55a0366e3fccc683336377e372",
        "Signers": [
          "Repo Admin"
        ]
      },
      {
        "SignedTag": "latest",
        "Digest": "d6bfc3baf615dc9618209a8d607ba2a8103d9c8a405b3bd8741d88b4bef36478",
        "Signers": [
          "Repo Admin"
        ]
      }
    ],
    "Signers": [],
    "AdministrativeKeys": [
      {
        "Name": "Repository",
        "Keys": [
          {
            "ID": "5a46c9aaa82ff150bb7305a2d17d0c521c2d784246807b2dc611f436a69041fd"
          }
        ]
      },
      {
        "Name": "Root",
        "Keys": [
          {
            "ID": "a2489bcac7a79aa67b19b96c4a3bf0c675ffdf00c6d2fabe1a5df1115e80adce"
          }
        ]
      }
    ]
  },
  {
    "Name": "notary",
    "SignedTags": [
      {
        "SignedTag": "server",
        "Digest": "71f64ab718a3331dee103bc5afc6bc492914738ce37c2d2f127a8133714ecf5c",
        "Signers": [
          "Repo Admin"
        ]
      },
      {
        "SignedTag": "signer",
        "Digest": "a6122d79b1e74f70b5dd933b18a6d1f99329a4728011079f06b245205f158fe8",
        "Signers": [
          "Repo Admin"
        ]
      }
    ],
    "Signers": [],
    "AdministrativeKeys": [
      {
        "Name": "Root",
        "Keys": [
          {
            "ID": "8cdcdef5bd039f4ab5a029126951b5985eebf57cabdcdc4d21f5b3be8bb4ce92"
          }
        ]
      },
      {
        "Name": "Repository",
        "Keys": [
          {
            "ID": "85bfd031017722f950d480a721f845a2944db26a3dc084040a70f1b0d9bbb3df"
          }
        ]
      }
    ]
  }
]
```

### Formatting

You can print the inspect output in a human-readable format instead of the default
JSON output, by using the `--pretty` option:

### Get details about signatures for a single image tag

```console
$ docker trust inspect --pretty alpine:latest

SIGNED TAG          DIGEST                                                             SIGNERS
latest              1072e499f3f655a032e88542330cf75b02e7bdf673278f701d7ba61629ee3ebe   (Repo Admin)

Administrative keys for alpine:latest:
Repository Key: 5a46c9aaa82ff150bb7305a2d17d0c521c2d784246807b2dc611f436a69041fd
Root Key:       a2489bcac7a79aa67b19b96c4a3bf0c675ffdf00c6d2fabe1a5df1115e80adce
```

The `SIGNED TAG` is the signed image tag with a unique content-addressable
`DIGEST`. `SIGNERS` lists all entities who have signed.

The administrative keys listed specify the root key of trust, as well as
the administrative repository key. These keys are responsible for modifying
signers, and rotating keys for the signed repository.

If signers are set up for the repository via other `docker trust` commands,
`docker trust inspect --pretty` displays them appropriately as a `SIGNER`
and specify their `KEYS`:

```console
$ docker trust inspect --pretty my-image:purple

SIGNED TAG          DIGEST                                                              SIGNERS
purple              941d3dba358621ce3c41ef67b47cf80f701ff80cdf46b5cc86587eaebfe45557    alice, bob, carol

List of signers and their keys:

SIGNER              KEYS
alice               47caae5b3e61, a85aab9d20a4
bob                 034370bcbd77, 82a66673242c
carol               b6f9f8e1aab0

Administrative keys for my-image:
Repository Key: 27df2c8187e7543345c2e0bf3a1262e0bc63a72754e9a7395eac3f747ec23a44
Root Key:       40b66ccc8b176be8c7d365a17f3e046d1c3494e053dd57cfeacfe2e19c4f8e8f
```

However, if other tags are signed in the same image repository,
`docker trust inspect` reports relevant key information.

```console
$ docker trust inspect --pretty alpine:unsigned

No signatures for alpine:unsigned


Administrative keys for alpine:unsigned:
Repository Key: 5a46c9aaa82ff150bb7305a2d17d0c521c2d784246807b2dc611f436a69041fd
Root Key:       a2489bcac7a79aa67b19b96c4a3bf0c675ffdf00c6d2fabe1a5df1115e80adce
```

### Get details about signatures for all image tags in a repository

```console
$ docker trust inspect --pretty alpine

SIGNED TAG          DIGEST                                                             SIGNERS
2.6                 9ace551613070689a12857d62c30ef0daa9a376107ec0fff0e34786cedb3399b   (Repo Admin)
2.7                 9f08005dff552038f0ad2f46b8e65ff3d25641747d3912e3ea8da6785046561a   (Repo Admin)
3.1                 d9477888b78e8c6392e0be8b2e73f8c67e2894ff9d4b8e467d1488fcceec21c8   (Repo Admin)
3.2                 19826d59171c2eb7e90ce52bfd822993bef6a6fe3ae6bb4a49f8c1d0a01e99c7   (Repo Admin)
3.3                 8fd4b76819e1e5baac82bd0a3d03abfe3906e034cc5ee32100d12aaaf3956dc7   (Repo Admin)
3.4                 833ad81ace8277324f3ca8c91c02bdcf1d13988d8ecf8a3f97ecdd69d0390ce9   (Repo Admin)
3.5                 af2a5bd2f8de8fc1ecabf1c76611cdc6a5f1ada1a2bdd7d3816e121b70300308   (Repo Admin)
3.6                 1072e499f3f655a032e88542330cf75b02e7bdf673278f701d7ba61629ee3ebe   (Repo Admin)
edge                79d50d15bd7ea48ea00cf3dd343b0e740c1afaa8e899bee475236ef338e1b53b   (Repo Admin)
latest              1072e499f3f655a032e88542330cf75b02e7bdf673278f701d7ba61629ee3ebe   (Repo Admin)

Administrative keys for alpine:
Repository Key: 5a46c9aaa82ff150bb7305a2d17d0c521c2d784246807b2dc611f436a69041fd
Root Key:       a2489bcac7a79aa67b19b96c4a3bf0c675ffdf00c6d2fabe1a5df1115e80adce
```

Here's an example with signers that are set up by `docker trust` commands:

```console
$ docker trust inspect --pretty my-image

SIGNED TAG          DIGEST                                                              SIGNERS
red                 852cc04935f930a857b630edc4ed6131e91b22073bcc216698842e44f64d2943    alice
blue                f1c38dbaeeb473c36716f6494d803fbfbe9d8a76916f7c0093f227821e378197    alice, bob
green               cae8fedc840f90c8057e1c24637d11865743ab1e61a972c1c9da06ec2de9a139    alice, bob
yellow              9cc65fc3126790e683d1b92f307a71f48f75fa7dd47a7b03145a123eaf0b45ba    carol
purple              941d3dba358621ce3c41ef67b47cf80f701ff80cdf46b5cc86587eaebfe45557    alice, bob, carol
orange              d6c271baa6d271bcc24ef1cbd65abf39123c17d2e83455bdab545a1a9093fc1c    alice

List of signers and their keys for my-image:

SIGNER              KEYS
alice               47caae5b3e61, a85aab9d20a4
bob                 034370bcbd77, 82a66673242c
carol               b6f9f8e1aab0

Administrative keys for my-image:
Repository Key: 27df2c8187e7543345c2e0bf3a1262e0bc63a72754e9a7395eac3f747ec23a44
Root Key:       40b66ccc8b176be8c7d365a17f3e046d1c3494e053dd57cfeacfe2e19c4f8e8f
```

# ps

<!---MARKER_GEN_START-->
List containers

### Aliases

`docker container ls`, `docker container list`, `docker container ps`, `docker ps`

### Options

| Name                                   | Type     | Default | Description                                                                                                                                                                                                                                                                                                                                                                                                                          |
|:---------------------------------------|:---------|:--------|:-------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------|
| [`-a`](#all), [`--all`](#all)          | `bool`   |         | Show all containers (default shows just running)                                                                                                                                                                                                                                                                                                                                                                                     |
| [`-f`](#filter), [`--filter`](#filter) | `filter` |         | Filter output based on conditions provided                                                                                                                                                                                                                                                                                                                                                                                           |
| [`--format`](#format)                  | `string` |         | Format output using a custom template:<br>'table':            Print output in table format with column headers (default)<br>'table TEMPLATE':   Print output in table format using the given Go template<br>'json':             Print in JSON format<br>'TEMPLATE':         Print output using the given Go template.<br>Refer to https://docs.docker.com/go/formatting/ for more information about formatting output with templates |
| `-n`, `--last`                         | `int`    | `-1`    | Show n last created containers (includes all states)                                                                                                                                                                                                                                                                                                                                                                                 |
| `-l`, `--latest`                       | `bool`   |         | Show the latest created container (includes all states)                                                                                                                                                                                                                                                                                                                                                                              |
| [`--no-trunc`](#no-trunc)              | `bool`   |         | Don't truncate output                                                                                                                                                                                                                                                                                                                                                                                                                |
| `-q`, `--quiet`                        | `bool`   |         | Only display container IDs                                                                                                                                                                                                                                                                                                                                                                                                           |
| [`-s`](#size), [`--size`](#size)       | `bool`   |         | Display total file sizes                                                                                                                                                                                                                                                                                                                                                                                                             |


<!---MARKER_GEN_END-->

## Examples

### <a name="no-trunc"></a> Do not truncate output (--no-trunc)

Running `docker ps --no-trunc` showing 2 linked containers.

```console
$ docker ps --no-trunc

CONTAINER ID                                                     IMAGE                        COMMAND                CREATED              STATUS              PORTS               NAMES
ca5534a51dd04bbcebe9b23ba05f389466cf0c190f1f8f182d7eea92a9671d00 ubuntu:24.04                 bash                   17 seconds ago       Up 16 seconds       3300-3310/tcp       webapp
9ca9747b233100676a48cc7806131586213fa5dab86dd1972d6a8732e3a84a4d crosbymichael/redis:latest   /redis-server --dir    33 minutes ago       Up 33 minutes       6379/tcp            redis,webapp/db
```

### <a name="all"></a> Show both running and stopped containers (-a, --all)

The `docker ps` command only shows running containers by default. To see all
containers, use the `--all` (or `-a`) flag:

```console
$ docker ps -a
```

`docker ps` groups exposed ports into a single range if possible. E.g., a
container that exposes TCP ports `100, 101, 102` displays `100-102/tcp` in
the `PORTS` column.

### <a name="size"></a> Show disk usage by container (--size)

The `docker ps --size` (or `-s`) command displays two different on-disk-sizes for each container:

```console
$ docker ps --size

CONTAINER ID   IMAGE          COMMAND                  CREATED        STATUS       PORTS   NAMES        SIZE
e90b8831a4b8   nginx          "/bin/bash -c 'mkdir "   11 weeks ago   Up 4 hours           my_nginx     35.58 kB (virtual 109.2 MB)
00c6131c5e30   telegraf:1.5   "/entrypoint.sh"         11 weeks ago   Up 11 weeks          my_telegraf  0 B (virtual 209.5 MB)
```
  * The "size" information shows the amount of data (on disk) that is used for the _writable_ layer of each container
  * The "virtual size" is the total amount of disk-space used for the read-only _image_ data used by the container and the writable layer.

For more information, refer to the [container size on disk](https://docs.docker.com/engine/storage/drivers/#container-size-on-disk) section.


### <a name="filter"></a> Filtering (--filter)

The `--filter` (or `-f`) flag format is a `key=value` pair. If there is more
than one filter, then pass multiple flags (e.g. `--filter "foo=bar" --filter "bif=baz"`).

The currently supported filters are:

| Filter                | Description                                                                                                                          |
|:----------------------|:-------------------------------------------------------------------------------------------------------------------------------------|
| `id`                  | Container's ID                                                                                                                       |
| `name`                | Container's name                                                                                                                     |
| `label`               | An arbitrary string representing either a key or a key-value pair. Expressed as `<key>` or `<key>=<value>`                           |
| `exited`              | An integer representing the container's exit code. Only useful with `--all`.                                                         |
| `status`              | One of `created`, `restarting`, `running`, `removing`, `paused`, `exited`, or `dead`                                                 |
| `ancestor`            | Filters containers which share a given image as an ancestor. Expressed as `<image-name>[:<tag>]`,  `<image id>`, or `<image@digest>` |
| `before` or `since`   | Filters containers created before or after a given container ID or name                                                              |
| `volume`              | Filters running containers which have mounted a given volume or bind mount.                                                          |
| `network`             | Filters running containers connected to a given network.                                                                             |
| `publish` or `expose` | Filters containers which publish or expose a given port. Expressed as `<port>[/<proto>]` or `<startport-endport>/[<proto>]`          |
| `health`              | Filters containers based on their healthcheck status. One of `starting`, `healthy`, `unhealthy` or `none`.                           |
| `isolation`           | Windows daemon only. One of `default`, `process`, or `hyperv`.                                                                       |
| `is-task`             | Filters containers that are a "task" for a service. Boolean option (`true` or `false`)                                               |


#### label

The `label` filter matches containers based on the presence of a `label` alone or a `label` and a
value.

The following filter matches containers with the `color` label regardless of its value.

```console
$ docker ps --filter "label=color"

CONTAINER ID        IMAGE               COMMAND             CREATED             STATUS              PORTS               NAMES
673394ef1d4c        busybox             "top"               47 seconds ago      Up 45 seconds                           nostalgic_shockley
d85756f57265        busybox             "top"               52 seconds ago      Up 51 seconds                           high_albattani
```

The following filter matches containers with the `color` label with the `blue` value.

```console
$ docker ps --filter "label=color=blue"

CONTAINER ID        IMAGE               COMMAND             CREATED              STATUS              PORTS               NAMES
d85756f57265        busybox             "top"               About a minute ago   Up About a minute                       high_albattani
```

#### name

The `name` filter matches on all or part of a container's name.

The following filter matches all containers with a name containing the `nostalgic_stallman` string.

```console
$ docker ps --filter "name=nostalgic_stallman"

CONTAINER ID        IMAGE               COMMAND             CREATED             STATUS              PORTS               NAMES
9b6247364a03        busybox             "top"               2 minutes ago       Up 2 minutes                            nostalgic_stallman
```

You can filter for a substring in a name as this shows:

```console
$ docker ps --filter "name=nostalgic"

CONTAINER ID        IMAGE               COMMAND             CREATED             STATUS              PORTS               NAMES
715ebfcee040        busybox             "top"               3 seconds ago       Up 1 second                             i_am_nostalgic
9b6247364a03        busybox             "top"               7 minutes ago       Up 7 minutes                            nostalgic_stallman
673394ef1d4c        busybox             "top"               38 minutes ago      Up 38 minutes                           nostalgic_shockley
```

#### exited

The `exited` filter matches containers by exist status code. For example, to
filter for containers that have exited successfully:

```console
$ docker ps -a --filter 'exited=0'

CONTAINER ID        IMAGE             COMMAND                CREATED             STATUS                   PORTS                      NAMES
ea09c3c82f6e        registry:latest   /srv/run.sh            2 weeks ago         Exited (0) 2 weeks ago   127.0.0.1:5000->5000/tcp   desperate_leakey
106ea823fe4e        fedora:latest     /bin/sh -c 'bash -l'   2 weeks ago         Exited (0) 2 weeks ago                              determined_albattani
48ee228c9464        fedora:20         bash                   2 weeks ago         Exited (0) 2 weeks ago                              tender_torvalds
```

#### Filter by exit signal

You can use a filter to locate containers that exited with status of `137`
meaning a `SIGKILL(9)` killed them.

```console
$ docker ps -a --filter 'exited=137'

CONTAINER ID        IMAGE               COMMAND                CREATED             STATUS                       PORTS               NAMES
b3e1c0ed5bfe        ubuntu:latest       "sleep 1000"           12 seconds ago      Exited (137) 5 seconds ago                       grave_kowalevski
a2eb5558d669        redis:latest        "/entrypoint.sh redi   2 hours ago         Exited (137) 2 hours ago                         sharp_lalande
```

Any of these events result in a `137` status:

* the `init` process of the container is killed manually
* `docker kill` kills the container
* Docker daemon restarts which kills all running containers

#### status

The `status` filter matches containers by status. The possible values for the container status are:

| Status       | Description                                                                                                                                                                                     |
| :----------- | :---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------- |
| `created`    | A container that has never been started.                                                                                                                                                        |
| `running`    | A running container, started by either `docker start` or `docker run`.                                                                                                                          |
| `paused`     | A paused container. See `docker pause`.                                                                                                                                                         |
| `restarting` | A container which is starting due to the designated restart policy for that container.                                                                                                          |
| `exited`     | A container which is no longer running. For example, the process inside the container completed or the container was stopped using the `docker stop` command.                                   |
| `removing`   | A container which is in the process of being removed. See `docker rm`.                                                                                                                          |
| `dead`       | A "defunct" container; for example, a container that was only partially removed because resources were kept busy by an external process. `dead` containers cannot be (re)started, only removed. |

For example, to filter for `running` containers:

```console
$ docker ps --filter status=running

CONTAINER ID        IMAGE                  COMMAND             CREATED             STATUS              PORTS               NAMES
715ebfcee040        busybox                "top"               16 minutes ago      Up 16 minutes                           i_am_nostalgic
d5c976d3c462        busybox                "top"               23 minutes ago      Up 23 minutes                           top
9b6247364a03        busybox                "top"               24 minutes ago      Up 24 minutes                           nostalgic_stallman
```

To filter for `paused` containers:

```console
$ docker ps --filter status=paused

CONTAINER ID        IMAGE               COMMAND             CREATED             STATUS                      PORTS               NAMES
673394ef1d4c        busybox             "top"               About an hour ago   Up About an hour (Paused)                       nostalgic_shockley
```

#### ancestor

The `ancestor` filter matches containers based on its image or a descendant of
it. The filter supports the following image representation:

- `image`
- `image:tag`
- `image:tag@digest`
- `short-id`
- `full-id`

If you don't specify a `tag`, the `latest` tag is used. For example, to filter
for containers that use the latest `ubuntu` image:

```console
$ docker ps --filter ancestor=ubuntu

CONTAINER ID        IMAGE               COMMAND             CREATED              STATUS              PORTS               NAMES
919e1179bdb8        ubuntu-c1           "top"               About a minute ago   Up About a minute                       admiring_lovelace
5d1e4a540723        ubuntu-c2           "top"               About a minute ago   Up About a minute                       admiring_sammet
82a598284012        ubuntu              "top"               3 minutes ago        Up 3 minutes                            sleepy_bose
bab2a34ba363        ubuntu              "top"               3 minutes ago        Up 3 minutes                            focused_yonath
```

Match containers based on the `ubuntu-c1` image which, in this case, is a child
of `ubuntu`:

```console
$ docker ps --filter ancestor=ubuntu-c1

CONTAINER ID        IMAGE               COMMAND             CREATED              STATUS              PORTS               NAMES
919e1179bdb8        ubuntu-c1           "top"               About a minute ago   Up About a minute                       admiring_lovelace
```

Match containers based on the `ubuntu` version `24.04` image:

```console
$ docker ps --filter ancestor=ubuntu:24.04

CONTAINER ID        IMAGE               COMMAND             CREATED              STATUS              PORTS               NAMES
82a598284012        ubuntu:24.04        "top"               3 minutes ago        Up 3 minutes                            sleepy_bose
```

The following matches containers based on the layer `d0e008c6cf02` or an image
that have this layer in its layer stack.

```console
$ docker ps --filter ancestor=d0e008c6cf02

CONTAINER ID        IMAGE               COMMAND             CREATED              STATUS              PORTS               NAMES
82a598284012        ubuntu:24.04        "top"               3 minutes ago        Up 3 minutes                            sleepy_bose
```

#### Create time

##### before

The `before` filter shows only containers created before the container with
a given ID or name. For example, having these containers created:

```console
$ docker ps

CONTAINER ID        IMAGE       COMMAND       CREATED              STATUS              PORTS              NAMES
9c3527ed70ce        busybox     "top"         14 seconds ago       Up 15 seconds                          desperate_dubinsky
4aace5031105        busybox     "top"         48 seconds ago       Up 49 seconds                          focused_hamilton
6e63f6ff38b0        busybox     "top"         About a minute ago   Up About a minute                      distracted_fermat
```

Filtering with `before` would give:

```console
$ docker ps -f before=9c3527ed70ce

CONTAINER ID        IMAGE       COMMAND       CREATED              STATUS              PORTS              NAMES
4aace5031105        busybox     "top"         About a minute ago   Up About a minute                      focused_hamilton
6e63f6ff38b0        busybox     "top"         About a minute ago   Up About a minute                      distracted_fermat
```

##### since

The `since` filter shows only containers created since the container with a given
ID or name. For example, with the same containers as in `before` filter:

```console
$ docker ps -f since=6e63f6ff38b0

CONTAINER ID        IMAGE       COMMAND       CREATED             STATUS              PORTS               NAMES
9c3527ed70ce        busybox     "top"         10 minutes ago      Up 10 minutes                           desperate_dubinsky
4aace5031105        busybox     "top"         10 minutes ago      Up 10 minutes                           focused_hamilton
```

#### volume

The `volume` filter shows only containers that mount a specific volume or have
a volume mounted in a specific path:

```console
$ docker ps --filter volume=remote-volume --format "table {{.ID}}\t{{.Mounts}}"

CONTAINER ID        MOUNTS
9c3527ed70ce        remote-volume

$ docker ps --filter volume=/data --format "table {{.ID}}\t{{.Mounts}}"

CONTAINER ID        MOUNTS
9c3527ed70ce        remote-volume
```

#### network

The `network` filter shows only containers that are connected to a network with
a given name or ID.

The following filter matches all containers that are connected to a network
with a name containing `net1`.

```console
$ docker run -d --net=net1 --name=test1 ubuntu top
$ docker run -d --net=net2 --name=test2 ubuntu top

$ docker ps --filter network=net1

CONTAINER ID        IMAGE       COMMAND       CREATED             STATUS              PORTS               NAMES
9d4893ed80fe        ubuntu      "top"         10 minutes ago      Up 10 minutes                           test1
```

The network filter matches on both the network's name and ID. The following
example shows all containers that are attached to the `net1` network, using
the network ID as a filter:

```console
$ docker network inspect --format "{{.ID}}" net1

8c0b4110ae930dbe26b258de9bc34a03f98056ed6f27f991d32919bfe401d7c5

$ docker ps --filter network=8c0b4110ae930dbe26b258de9bc34a03f98056ed6f27f991d32919bfe401d7c5

CONTAINER ID        IMAGE       COMMAND       CREATED             STATUS              PORTS               NAMES
9d4893ed80fe        ubuntu      "top"         10 minutes ago      Up 10 minutes                           test1
```

#### publish and expose

The `publish` and `expose` filters show only containers that have published or exposed port with a given port
number, port range, and/or protocol. The default protocol is `tcp` when not specified.

The following filter matches all containers that have published port of 80:

```console
$ docker run -d --publish=80 busybox top
$ docker run -d --expose=8080 busybox top

$ docker ps -a

CONTAINER ID        IMAGE               COMMAND             CREATED             STATUS              PORTS                   NAMES
9833437217a5        busybox             "top"               5 seconds ago       Up 4 seconds        8080/tcp                dreamy_mccarthy
fc7e477723b7        busybox             "top"               50 seconds ago      Up 50 seconds       0.0.0.0:32768->80/tcp   admiring_roentgen

$ docker ps --filter publish=80

CONTAINER ID        IMAGE               COMMAND             CREATED              STATUS              PORTS                   NAMES
fc7e477723b7        busybox             "top"               About a minute ago   Up About a minute   0.0.0.0:32768->80/tcp   admiring_roentgen
```

The following filter matches all containers that have exposed TCP port in the range of `8000-8080`:

```console
$ docker ps --filter expose=8000-8080/tcp

CONTAINER ID        IMAGE               COMMAND             CREATED             STATUS              PORTS               NAMES
9833437217a5        busybox             "top"               21 seconds ago      Up 19 seconds       8080/tcp            dreamy_mccarthy
```

The following filter matches all containers that have exposed UDP port `80`:

```console
$ docker ps --filter publish=80/udp

CONTAINER ID        IMAGE               COMMAND             CREATED             STATUS              PORTS               NAMES
```

### <a name="format"></a> Format the output (--format)

The formatting option (`--format`) pretty-prints container output using a Go
template.

Valid placeholders for the Go template are listed below:

| Placeholder     | Description                                                                                     |
|:----------------|:------------------------------------------------------------------------------------------------|
| `.ID`           | Container ID                                                                                    |
| `.Image`        | Image ID                                                                                        |
| `.Command`      | Quoted command                                                                                  |
| `.CreatedAt`    | Time when the container was created.                                                            |
| `.RunningFor`   | Elapsed time since the container was started.                                                   |
| `.Ports`        | Exposed ports.                                                                                  |
| `.State`        | Container status (for example; "created", "running", "exited").                                 |
| `.Status`       | Container status with details about duration and health-status.                                 |
| `.HealthStatus` | Container health status ("starting", "healthy", "unhealthy"; empty when unavailable).           |
| `.Size`         | Container disk size.                                                                            |
| `.Names`        | Container names.                                                                                |
| `.Labels`       | All labels assigned to the container.                                                           |
| `.Label`        | Value of a specific label for this container. For example `'{{.Label "com.docker.swarm.cpu"}}'` |
| `.Mounts`       | Names of the volumes mounted in this container.                                                 |
| `.Networks`     | Names of the networks attached to this container.                                               |

When using the `--format` option, the `ps` command will either output the data
exactly as the template declares or, when using the `table` directive, includes
column headers as well.

The following example uses a template without headers and outputs the `ID` and
`Command` entries separated by a colon (`:`) for all running containers:

```console
$ docker ps --format "{{.ID}}: {{.Command}}"

a87ecb4f327c: /bin/sh -c #(nop) MA
01946d9d34d8: /bin/sh -c #(nop) MA
c1d3b0166030: /bin/sh -c yum -y up
41d50ecd2f57: /bin/sh -c #(nop) MA
```

To list all running containers with their labels in a table format you can use:

```console
$ docker ps --format "table {{.ID}}\t{{.Labels}}"

CONTAINER ID        LABELS
a87ecb4f327c        com.docker.swarm.node=ubuntu,com.docker.swarm.storage=ssd
01946d9d34d8
c1d3b0166030        com.docker.swarm.node=debian,com.docker.swarm.cpu=6
41d50ecd2f57        com.docker.swarm.node=fedora,com.docker.swarm.cpu=3,com.docker.swarm.storage=ssd
```

To list all running containers in JSON format, use the `json` directive:

```console
$ docker ps --format json
{"Command":"\"/docker-entrypoint.…\"","CreatedAt":"2021-03-10 00:15:05 +0100 CET","ID":"a762a2b37a1d","Image":"nginx","Labels":"maintainer=NGINX Docker Maintainers \u003cdocker-maint@nginx.com\u003e","LocalVolumes":"0","Mounts":"","Names":"boring_keldysh","Networks":"bridge","Ports":"80/tcp","RunningFor":"4 seconds ago","Size":"0B","State":"running","Status":"Up 3 seconds"}
```

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/themes.R
\name{parse_selector}
\alias{parse_selector}
\title{Parse a CSS3-like selector}
\usage{
parse_selector(x)
}
\arguments{
\item{x}{CSS3-like selector string.}
}
\description{
This is the rather small subset of CSS3 that is supported:
}
\details{
Selectors:
\itemize{
\item Type selectors, e.g. \code{input} selects all \verb{<input>} elements.
\item Class selectors, e.g. \code{.index} selects any element that has a class
of "index".
\item ID selector. \verb{#toc} will match the element that has the ID \code{"toc"}.
}

Combinators:
\itemize{
\item Descendant combinator, i.e. the space, that combinator selects nodes
that are descendants of the first element. E.g. \verb{div span} will match
all \verb{<span>} elements that are inside a \verb{<div>} element.
}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hash.R
\name{hash_sha1}
\alias{hash_sha1}
\alias{hash_raw_sha1}
\alias{hash_obj_sha1}
\alias{hash_file_sha1}
\title{SHA-1 hash}
\usage{
hash_sha1(x)

hash_raw_sha1(x)

hash_obj_sha1(x, serialize_version = 2)

hash_file_sha1(paths)
}
\arguments{
\item{x}{Character vector. If not a character vector, then
\code{\link[=as.character]{as.character()}} is used to try to coerce it into one. \code{NA} entries
will have an \code{NA} hash.}

\item{serialize_version}{Workspace format version to use, see
\code{\link[base:serialize]{base::serialize()}}.}

\item{paths}{Character vector of file names.}
}
\value{
\code{hash_sha1()} returns a character vector of hexadecimal
SHA-1 hashes.

\code{hash_raw_sha1()} returns a character scalar.

\code{hash_obj_sha1()} returns a character scalar.

\code{hash_file_sha1()} returns a character vector of SHA-1
hashes.
}
\description{
Calculate the SHA-1 hash of each element of a character vector.
}
\details{
\code{hash_raw_sha1()} calculates the SHA-1 hash of the bytes
of a raw vector.

\code{hash_obj_sha1()} calculates the SHA-1 hash of an R
object. The object is serialized into a binary vector first.

\code{hash_file_sha1()} calculates the SHA-1 hash of one or
more files.
}
\examples{
hash_sha1(c("foo", NA, "bar", ""))
}
\seealso{
Other hash functions: 
\code{\link{hash_animal}()},
\code{\link{hash_emoji}()},
\code{\link{hash_md5}()},
\code{\link{hash_sha256}()},
\code{\link{hash_xxhash}()}
}
\concept{hash functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/status-bar.R
\name{cli_status}
\alias{cli_status}
\title{Update the status bar (superseded)}
\usage{
cli_status(
  msg,
  msg_done = paste(msg, "... done"),
  msg_failed = paste(msg, "... failed"),
  .keep = FALSE,
  .auto_close = TRUE,
  .envir = parent.frame(),
  .auto_result = c("clear", "done", "failed", "auto")
)
}
\arguments{
\item{msg}{The text to show, a character vector. It will be
collapsed into a single string, and the first line is kept and cut to
\code{\link[=console_width]{console_width()}}. The message is often associated with the start of
a calculation.}

\item{msg_done}{The message to use when the message is cleared, when
the calculation finishes successfully. If \code{.auto_close} is \code{TRUE}
and \code{.auto_result} is \code{"done"}, then this is printed automatically
when the calling function (or \code{.envir}) finishes.}

\item{msg_failed}{The message to use when the message is cleared, when
the calculation finishes unsuccessfully. If \code{.auto_close} is \code{TRUE}
and \code{.auto_result} is \code{"failed"}, then this is printed automatically
when the calling function (or \code{.envir}) finishes.}

\item{.keep}{What to do when this status bar is cleared. If \code{TRUE} then
the content of this status bar is kept, as regular cli output (the
screen is scrolled up if needed). If \code{FALSE}, then this status bar
is deleted.}

\item{.auto_close}{Whether to clear the status bar when the calling
function finishes (or \code{.envir} is removed from the stack, if
specified).}

\item{.envir}{Environment to evaluate the glue expressions in. It is
also used to auto-clear the status bar if \code{.auto_close} is \code{TRUE}.}

\item{.auto_result}{What to do when auto-closing the status bar.}
}
\value{
The id of the new status bar container element, invisibly.
}
\description{
\strong{The \verb{cli_status_*()} functions are superseded by
the \code{\link[=cli_progress_message]{cli_progress_message()}} and \code{\link[=cli_progress_step]{cli_progress_step()}} functions,
because they have a better default behavior.}

The status bar is the last line of the terminal. cli apps can use this
to show status information, progress bars, etc. The status bar is kept
intact by all semantic cli output.
}
\details{
Use \code{\link[=cli_status_clear]{cli_status_clear()}} to clear the status bar.

Often status messages are associated with processes. E.g. the app starts
downloading a large file, so it sets the status bar accordingly. Once the
download is done (or has failed), the app typically updates the status bar
again. cli automates much of this, via the \code{msg_done}, \code{msg_failed}, and
\code{.auto_result} arguments. See examples below.
}
\seealso{
Status bars support \link[=inline-markup]{inline markup}.

The \code{\link[=cli_progress_message]{cli_progress_message()}} and \code{\link[=cli_progress_step]{cli_progress_step()}}
functions, for a superior API.

Other status bar: 
\code{\link{cli_process_start}()},
\code{\link{cli_status_clear}()},
\code{\link{cli_status_update}()}

Other functions supporting inline markup: 
\code{\link{cli_abort}()},
\code{\link{cli_alert}()},
\code{\link{cli_blockquote}()},
\code{\link{cli_bullets}()},
\code{\link{cli_bullets_raw}()},
\code{\link{cli_dl}()},
\code{\link{cli_h1}()},
\code{\link{cli_li}()},
\code{\link{cli_ol}()},
\code{\link{cli_process_start}()},
\code{\link{cli_progress_along}()},
\code{\link{cli_progress_bar}()},
\code{\link{cli_progress_message}()},
\code{\link{cli_progress_output}()},
\code{\link{cli_progress_step}()},
\code{\link{cli_rule}},
\code{\link{cli_status_update}()},
\code{\link{cli_text}()},
\code{\link{cli_ul}()},
\code{\link{format_error}()},
\code{\link{format_inline}()}
}
\concept{functions supporting inline markup}
\concept{status bar}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cli.R
\name{format_inline}
\alias{format_inline}
\title{Format and returns a line of text}
\usage{
format_inline(
  ...,
  .envir = parent.frame(),
  collapse = TRUE,
  keep_whitespace = TRUE
)
}
\arguments{
\item{...}{Passed to \code{\link[=cli_text]{cli_text()}}.}

\item{.envir}{Environment to evaluate the expressions in.}

\item{collapse}{Whether to collapse the result if it has multiple
lines, e.g. because of \verb{\\f} characters.}

\item{keep_whitespace}{Whether to keep all whitepace (spaces, newlines
and form feeds) as is in the input.}
}
\value{
Character scalar, the formatted string.
}
\description{
You can use this function to format a line of cli text, without emitting
it to the screen. It uses \code{\link[=cli_text]{cli_text()}} internally.
}
\details{
\code{format_inline()} performs no width-wrapping.
}
\examples{
format_inline("A message for {.emph later}, thanks {.fn format_inline}.")
}
\seealso{
This function supports \link[=inline-markup]{inline markup}.

Other functions supporting inline markup: 
\code{\link{cli_abort}()},
\code{\link{cli_alert}()},
\code{\link{cli_blockquote}()},
\code{\link{cli_bullets}()},
\code{\link{cli_bullets_raw}()},
\code{\link{cli_dl}()},
\code{\link{cli_h1}()},
\code{\link{cli_li}()},
\code{\link{cli_ol}()},
\code{\link{cli_process_start}()},
\code{\link{cli_progress_along}()},
\code{\link{cli_progress_bar}()},
\code{\link{cli_progress_message}()},
\code{\link{cli_progress_output}()},
\code{\link{cli_progress_step}()},
\code{\link{cli_rule}},
\code{\link{cli_status}()},
\code{\link{cli_status_update}()},
\code{\link{cli_text}()},
\code{\link{cli_ul}()},
\code{\link{format_error}()}
}
\concept{functions supporting inline markup}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ansiex.R
\name{ansi_html}
\alias{ansi_html}
\title{Convert ANSI styled text to HTML}
\usage{
ansi_html(x, escape_reserved = TRUE, csi = c("drop", "keep"))
}
\arguments{
\item{x}{Input character vector.}

\item{escape_reserved}{Whether to escape characters that are reserved
in HTML (\code{&}, \code{<} and \code{>}).}

\item{csi}{What to do with non-SGR ANSI sequences, either \code{"keep"},
or \code{"drop"} them.}
}
\value{
Character vector of HTML.
}
\description{
Convert ANSI styled text to HTML
}
\examples{
\dontshow{if (cli:::has_packages(c("htmltools", "withr"))) (if (getRversion() >= "3.4") withAutoprint else force)(\{ # examplesIf}
## Syntax highlight the source code of an R function with ANSI tags,
## and export it to a HTML file.
code <- withr::with_options(
  list(ansi.num_colors = 256),
  code_highlight(format(ansi_html))
)
hcode <- paste(ansi_html(code), collapse = "\n")
css <- paste(format(ansi_html_style()), collapse=  "\n")
page <- htmltools::tagList(
  htmltools::tags$head(htmltools::tags$style(css)),
  htmltools::tags$pre(htmltools::HTML(hcode))
)

if (interactive()) htmltools::html_print(page)
\dontshow{\}) # examplesIf}
}
\seealso{
Other ANSI to HTML conversion: 
\code{\link{ansi_html_style}()}
}
\concept{ANSI to HTML conversion}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cliapp-docs.R
\name{containers}
\alias{containers}
\title{About cli containers}
\description{
Container elements may contain other elements. Currently the following
commands create container elements: \code{\link[=cli_div]{cli_div()}}, \code{\link[=cli_par]{cli_par()}}, the list
elements: \code{\link[=cli_ul]{cli_ul()}}, \code{\link[=cli_ol]{cli_ol()}}, \code{\link[=cli_dl]{cli_dl()}}, and list items are
containers as well: \code{\link[=cli_li]{cli_li()}}.
}
\details{
\subsection{Themes}{

A container can add a new theme, which is removed when the container
exits.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{d <- cli_div(theme = list(h1 = list(color = "blue",
                                    "font-weight" = "bold")))
cli_h1("Custom title")
cli_end(d)
}\if{html}{\out{</div>}}\if{html}{\out{
<div class="asciicast" style="color: #172431;font-family: 'Fira Code',Monaco,Consolas,Menlo,'Bitstream Vera Sans Mono','Powerline Symbols',monospace;line-height: 1.300000"><pre>
#>                                                                                 
#> <span style="font-weight: bold;color: #268BD2;">Custom title</span>                                                                    
</pre></div>
}}

}

\subsection{Auto-closing}{

Container elements are closed with \code{\link[=cli_end]{cli_end()}}. For convenience,
by default they are closed automatically when the function that created
them terminated (either regularly or with an error). The default
behavior can be changed with the \code{.auto_close} argument.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{div <- function() \{
  cli_div(class = "tmp", theme = list(.tmp = list(color = "yellow")))
  cli_text("This is yellow")
\}
div()
cli_text("This is not yellow any more")
}\if{html}{\out{</div>}}\if{html}{\out{
<div class="asciicast" style="color: #172431;font-family: 'Fira Code',Monaco,Consolas,Menlo,'Bitstream Vera Sans Mono','Powerline Symbols',monospace;line-height: 1.300000"><pre>
#> <span style="color: #B58900;">This is yellow</span>                                                                  
#> This is not yellow any more                                                     
</pre></div>
}}

}

\subsection{Debugging}{

You can use the internal \code{cli:::cli_debug_doc()} function to see the
currently open containers.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{fun <- function() \{
  cli_div(id = "mydiv")
  cli_par(class = "myclass")
  cli:::cli_debug_doc()
\}
fun()
}\if{html}{\out{</div>}}\if{html}{\out{
<div class="asciicast" style="color: #172431;font-family: 'Fira Code',Monaco,Consolas,Menlo,'Bitstream Vera Sans Mono','Powerline Symbols',monospace;line-height: 1.300000"><pre>
#>                                                                                 
#> &lt;cli document&gt;                                                                  
#> &lt;body id="body"&gt;                                                                
#> &lt;div id="mydiv"&gt; +theme                                                         
#> &lt;par id="cli-82040-226" class="myclass"&gt;                                        
</pre></div>
}}

}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ansiex.R
\name{ansi_strtrim}
\alias{ansi_strtrim}
\title{Truncate an ANSI string}
\usage{
ansi_strtrim(x, width = console_width(), ellipsis = symbol$ellipsis)
}
\arguments{
\item{x}{Character vector of ANSI strings.}

\item{width}{The width to truncate to.}

\item{ellipsis}{The string to append to truncated strings. Supply an
empty string if you don't want a marker.}
}
\description{
This function is similar to \code{\link[base:strtrim]{base::strtrim()}}, but works correctly with
ANSI styled strings. It also adds \code{...} (or the corresponding Unicode
character if Unicode characters are allowed) to the end of truncated
strings.
}
\details{
Note: \code{ansi_strtrim()} does not support NA values currently.
}
\examples{
text <- cli::col_red(cli:::lorem_ipsum())
ansi_strtrim(c(text, "foobar"), 40)
}
\seealso{
Other ANSI string operations: 
\code{\link{ansi_align}()},
\code{\link{ansi_columns}()},
\code{\link{ansi_nchar}()},
\code{\link{ansi_strsplit}()},
\code{\link{ansi_strwrap}()},
\code{\link{ansi_substr}()},
\code{\link{ansi_substring}()},
\code{\link{ansi_toupper}()},
\code{\link{ansi_trimws}()}
}
\concept{ANSI string operations}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ansiex.R
\name{ansi_strwrap}
\alias{ansi_strwrap}
\title{Wrap an ANSI styled string to a certain width}
\usage{
ansi_strwrap(
  x,
  width = console_width(),
  indent = 0,
  exdent = 0,
  simplify = TRUE
)
}
\arguments{
\item{x}{ANSI string.}

\item{width}{Width to wrap to.}

\item{indent}{Indentation of the first line of each paragraph.}

\item{exdent}{Indentation of the subsequent lines of each paragraph.}

\item{simplify}{Whether to return all wrapped strings in a single
character vector, or wrap each element of \code{x} independently and return
a list.}
}
\value{
If \code{simplify} is \code{FALSE}, then a list of character vectors,
each an ANSI string. Otherwise a single ANSI string vector.
}
\description{
This function is similar to \code{\link[base:strwrap]{base::strwrap()}}, but works on ANSI
styled strings, and leaves the styling intact.
}
\examples{
text <- cli:::lorem_ipsum()
# Highlight some words, that start with 's'
rexp <- gregexpr("\\\\b([sS][a-zA-Z]+)\\\\b", text)
regmatches(text, rexp) <- lapply(regmatches(text, rexp), col_red)
cat(text)

wrp <- ansi_strwrap(text, width = 40)
cat(wrp, sep = "\n")
}
\seealso{
Other ANSI string operations: 
\code{\link{ansi_align}()},
\code{\link{ansi_columns}()},
\code{\link{ansi_nchar}()},
\code{\link{ansi_strsplit}()},
\code{\link{ansi_strtrim}()},
\code{\link{ansi_substring}()},
\code{\link{ansi_substr}()},
\code{\link{ansi_toupper}()},
\code{\link{ansi_trimws}()}
}
\concept{ANSI string operations}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/box-styles.R, R/boxes.R
\name{list_border_styles}
\alias{list_border_styles}
\alias{boxx}
\title{Draw a banner-like box in the console}
\usage{
list_border_styles()

boxx(
  label,
  header = "",
  footer = "",
  border_style = "single",
  padding = 1,
  margin = 0,
  float = c("left", "center", "right"),
  col = NULL,
  background_col = NULL,
  border_col = col,
  align = c("left", "center", "right"),
  width = console_width()
)
}
\arguments{
\item{label}{Label to show, a character vector. Each element will be
in a new line. You can color it using the \verb{col_*}, \verb{bg_*} and
\verb{style_*} functions, see \link[=ansi-styles]{ANSI styles} and the examples
below.}

\item{header}{Text to show on top border of the box. If too long,
it will be cut.}

\item{footer}{Text to show on the bottom border of the box. If too long,
it will be cut.}

\item{border_style}{String that specifies the border style.
\code{list_border_styles} lists all current styles.}

\item{padding}{Padding within the box. Either an integer vector of
four numbers (bottom, left, top, right), or a single number \code{x}, which
is interpreted as \code{c(x, 3*x, x, 3*x)}.}

\item{margin}{Margin around the box. Either an integer vector of four
numbers (bottom, left, top, right), or a single number \code{x}, which is
interpreted as \code{c(x, 3*x, x, 3*x)}.}

\item{float}{Whether to display the box on the \code{"left"}, \code{"center"}, or
the \code{"right"} of the screen.}

\item{col}{Color of text, and default border color. Either a style
function (see \link[=ansi-styles]{ANSI styles}) or a color name that is
passed to \code{\link[=make_ansi_style]{make_ansi_style()}}.}

\item{background_col}{Background color of the inside of the box.
Either a style function (see \link[=ansi-styles]{ANSI styles}), or a color
name which will be used in \code{\link[=make_ansi_style]{make_ansi_style()}} to create a
\emph{background} style (i.e. \code{bg = TRUE} is used).}

\item{border_col}{Color of the border. Either a style function
(see \link[=ansi-styles]{ANSI styles}) or a color name that is passed to
\code{\link[=make_ansi_style]{make_ansi_style()}}.}

\item{align}{Alignment of the label within the box: \code{"left"},
\code{"center"}, or \code{"right"}.}

\item{width}{Width of the screen, defaults to \code{\link[=console_width]{console_width()}}.}
}
\description{
Draw a banner-like box in the console
}
\details{
\subsection{Defaults}{

\if{html}{\out{<div class="sourceCode r">}}\preformatted{boxx("Hello there!")
}\if{html}{\out{</div>}}\if{html}{\out{
<div class="asciicast" style="color: #172431;font-family: 'Fira Code',Monaco,Consolas,Menlo,'Bitstream Vera Sans Mono','Powerline Symbols',monospace;line-height: 1.300000"><pre>
#> ┌──────────────────┐                                                            
#> │                  │                                                            
#> │   Hello there!   │                                                            
#> │                  │                                                            
#> └──────────────────┘                                                            
</pre></div>
}}

}

\subsection{Change border style}{

\if{html}{\out{<div class="sourceCode r">}}\preformatted{boxx("Hello there!", border_style = "double")
}\if{html}{\out{</div>}}\if{html}{\out{
<div class="asciicast" style="color: #172431;font-family: 'Fira Code',Monaco,Consolas,Menlo,'Bitstream Vera Sans Mono','Powerline Symbols',monospace;line-height: 1.300000"><pre>
#> ╔══════════════════╗                                                            
#> ║                  ║                                                            
#> ║   Hello there!   ║                                                            
#> ║                  ║                                                            
#> ╚══════════════════╝                                                            
</pre></div>
}}

}

\subsection{Multiple lines}{

\if{html}{\out{<div class="sourceCode r">}}\preformatted{boxx(c("Hello", "there!"), padding = 1)
}\if{html}{\out{</div>}}\if{html}{\out{
<div class="asciicast" style="color: #172431;font-family: 'Fira Code',Monaco,Consolas,Menlo,'Bitstream Vera Sans Mono','Powerline Symbols',monospace;line-height: 1.300000"><pre>
#> ┌────────────┐                                                                  
#> │            │                                                                  
#> │   Hello    │                                                                  
#> │   there!   │                                                                  
#> │            │                                                                  
#> └────────────┘                                                                  
</pre></div>
}}

}

\subsection{Padding}{

\if{html}{\out{<div class="sourceCode r">}}\preformatted{boxx("Hello there!", padding = 1)
boxx("Hello there!", padding = c(1, 5, 1, 5))
}\if{html}{\out{</div>}}\if{html}{\out{
<div class="asciicast" style="color: #172431;font-family: 'Fira Code',Monaco,Consolas,Menlo,'Bitstream Vera Sans Mono','Powerline Symbols',monospace;line-height: 1.300000"><pre>
#> ┌──────────────────┐                                                            
#> │                  │                                                            
#> │   Hello there!   │                                                            
#> │                  │                                                            
#> └──────────────────┘                                                            
#> ┌──────────────────────┐                                                        
#> │                      │                                                        
#> │     Hello there!     │                                                        
#> │                      │                                                        
#> └──────────────────────┘                                                        
</pre></div>
}}

}

\subsection{Floating}{

\if{html}{\out{<div class="sourceCode r">}}\preformatted{boxx("Hello there!", padding = 1, float = "center")
boxx("Hello there!", padding = 1, float = "right")
}\if{html}{\out{</div>}}\if{html}{\out{
<div class="asciicast" style="color: #172431;font-family: 'Fira Code',Monaco,Consolas,Menlo,'Bitstream Vera Sans Mono','Powerline Symbols',monospace;line-height: 1.300000"><pre>
#>                           ┌──────────────────┐                                  
#>                           │                  │                                  
#>                           │   Hello there!   │                                  
#>                           │                  │                                  
#>                           └──────────────────┘                                  
#>                                                   ┌──────────────────┐          
#>                                                   │                  │          
#>                                                   │   Hello there!   │          
#>                                                   │                  │          
#>                                                   └──────────────────┘          
</pre></div>
}}

}

\subsection{Text color}{

\if{html}{\out{<div class="sourceCode r">}}\preformatted{boxx(col_cyan("Hello there!"), padding = 1, float = "center")
}\if{html}{\out{</div>}}\if{html}{\out{
<div class="asciicast" style="color: #172431;font-family: 'Fira Code',Monaco,Consolas,Menlo,'Bitstream Vera Sans Mono','Powerline Symbols',monospace;line-height: 1.300000"><pre>
#>                           ┌──────────────────┐                                  
#>                           │                  │                                  
#>                           │   <span style="color: #2AA198;">Hello there!</span>   │                                  
#>                           │                  │                                  
#>                           └──────────────────┘                                  
</pre></div>
}}

}

\subsection{Background color}{

\if{html}{\out{<div class="sourceCode r">}}\preformatted{boxx("Hello there!", padding = 1, background_col = "brown")
boxx("Hello there!", padding = 1, background_col = bg_red)
}\if{html}{\out{</div>}}\if{html}{\out{
<div class="asciicast" style="color: #172431;font-family: 'Fira Code',Monaco,Consolas,Menlo,'Bitstream Vera Sans Mono','Powerline Symbols',monospace;line-height: 1.300000"><pre>
#> ┌──────────────────┐                                                            
#> │<span style="background-color: #990000;">                  </span>│                                                            
#> │<span style="background-color: #990000;">   Hello there!   </span>│                                                            
#> │<span style="background-color: #990000;">                  </span>│                                                            
#> └──────────────────┘                                                            
#> ┌──────────────────┐                                                            
#> │<span style="background-color: #DC322F;">                  </span>│                                                            
#> │<span style="background-color: #DC322F;">   Hello there!   </span>│                                                            
#> │<span style="background-color: #DC322F;">                  </span>│                                                            
#> └──────────────────┘                                                            
</pre></div>
}}

}

\subsection{Border color}{

\if{html}{\out{<div class="sourceCode r">}}\preformatted{boxx("Hello there!", padding = 1, border_col = "green")
boxx("Hello there!", padding = 1, border_col = col_red)
}\if{html}{\out{</div>}}\if{html}{\out{
<div class="asciicast" style="color: #172431;font-family: 'Fira Code',Monaco,Consolas,Menlo,'Bitstream Vera Sans Mono','Powerline Symbols',monospace;line-height: 1.300000"><pre>
#> <span style="color: #859900;">┌──────────────────┐</span>                                                            
#> <span style="color: #859900;">│</span>                  <span style="color: #859900;">│</span>                                                            
#> <span style="color: #859900;">│</span>   Hello there!   <span style="color: #859900;">│</span>                                                            
#> <span style="color: #859900;">│</span>                  <span style="color: #859900;">│</span>                                                            
#> <span style="color: #859900;">└──────────────────┘</span>                                                            
#> <span style="color: #DC322F;">┌──────────────────┐</span>                                                            
#> <span style="color: #DC322F;">│</span>                  <span style="color: #DC322F;">│</span>                                                            
#> <span style="color: #DC322F;">│</span>   Hello there!   <span style="color: #DC322F;">│</span>                                                            
#> <span style="color: #DC322F;">│</span>                  <span style="color: #DC322F;">│</span>                                                            
#> <span style="color: #DC322F;">└──────────────────┘</span>                                                            
</pre></div>
}}

}

\subsection{Label alignment}{

\if{html}{\out{<div class="sourceCode r">}}\preformatted{boxx(c("Hi", "there", "you!"), padding = 1, align = "left")
boxx(c("Hi", "there", "you!"), padding = 1, align = "center")
boxx(c("Hi", "there", "you!"), padding = 1, align = "right")
}\if{html}{\out{</div>}}\if{html}{\out{
<div class="asciicast" style="color: #172431;font-family: 'Fira Code',Monaco,Consolas,Menlo,'Bitstream Vera Sans Mono','Powerline Symbols',monospace;line-height: 1.300000"><pre>
#> ┌───────────┐                                                                   
#> │           │                                                                   
#> │   Hi      │                                                                   
#> │   there   │                                                                   
#> │   you!    │                                                                   
#> │           │                                                                   
#> └───────────┘                                                                   
#> ┌───────────┐                                                                   
#> │           │                                                                   
#> │     Hi    │                                                                   
#> │   there   │                                                                   
#> │    you!   │                                                                   
#> │           │                                                                   
#> └───────────┘                                                                   
#> ┌───────────┐                                                                   
#> │           │                                                                   
#> │      Hi   │                                                                   
#> │   there   │                                                                   
#> │    you!   │                                                                   
#> │           │                                                                   
#> └───────────┘                                                                   
</pre></div>
}}

}

\subsection{A very customized box}{

\if{html}{\out{<div class="sourceCode r">}}\preformatted{star <- symbol$star
label <- c(paste(star, "Hello", star), "  there!")
boxx(
  col_white(label),
  border_style="round",
  padding = 1,
  float = "center",
  border_col = "tomato3",
  background_col="darkolivegreen"
)
}\if{html}{\out{</div>}}\if{html}{\out{
<div class="asciicast" style="color: #172431;font-family: 'Fira Code',Monaco,Consolas,Menlo,'Bitstream Vera Sans Mono','Powerline Symbols',monospace;line-height: 1.300000"><pre>
#> <span style="color: #cc3333;">                           ╭───────────────╮</span>                                    
#>                            <span style="color: #cc3333;">│</span><span style="background-color: #336633;">               </span><span style="color: #cc3333;">│</span>                                    
#>                            <span style="color: #cc3333;">│</span><span style="background-color: #336633;">   </span><span style="color: #EEE8D5;background-color: #336633;">★ Hello ★</span><span style="background-color: #336633;">   </span><span style="color: #cc3333;">│</span>                                    
#>                            <span style="color: #cc3333;">│</span><span style="background-color: #336633;">   </span><span style="color: #EEE8D5;background-color: #336633;">  there!</span><span style="background-color: #336633;">    </span><span style="color: #cc3333;">│</span>                                    
#>                            <span style="color: #cc3333;">│</span><span style="background-color: #336633;">               </span><span style="color: #cc3333;">│</span>                                    
#> <span style="color: #cc3333;">                           ╰───────────────╯</span>                                    
</pre></div>
}}

}
}
\section{About fonts and terminal settings}{

The boxes might or might not look great in your terminal, depending
on the box style you use and the font the terminal uses. We found that
the Menlo font looks nice in most terminals an also in Emacs.

RStudio currently has a line height greater than one for console output,
which makes the boxes ugly.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cat.R
\name{cat_line}
\alias{cat_line}
\alias{cat_bullet}
\alias{cat_boxx}
\alias{cat_rule}
\alias{cat_print}
\title{\code{cat()} helpers}
\usage{
cat_line(..., col = NULL, background_col = NULL, file = stdout())

cat_bullet(
  ...,
  col = NULL,
  background_col = NULL,
  bullet = "bullet",
  bullet_col = NULL,
  file = stdout()
)

cat_boxx(..., file = stdout())

cat_rule(..., file = stdout())

cat_print(x, file = "")
}
\arguments{
\item{...}{For \code{cat_line()} and \code{cat_bullet()}, pasted together with
\code{collapse = "\\n"}. For \code{cat_rule()} and \code{cat_boxx()} passed on to
\code{\link[=rule]{rule()}} and \code{\link[=boxx]{boxx()}} respectively.}

\item{col, background_col, bullet_col}{Colors for text, background, and
bullets respectively.}

\item{file}{Output destination. Defaults to standard output.}

\item{bullet}{Name of bullet character. Indexes into \link{symbol}}

\item{x}{An object to print.}
}
\description{
These helpers provide useful wrappers around \code{\link[=cat]{cat()}}: most importantly
they all set \code{sep = ""}, and \code{cat_line()} automatically adds a newline.
}
\examples{
cat_line("This is ", "a ", "line of text.", col = "red")
cat_bullet(letters[1:5])
cat_bullet(letters[1:5], bullet = "tick", bullet_col = "green")
cat_rule()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/diff.R
\name{diff_chr}
\alias{diff_chr}
\title{Compare two character vectors elementwise}
\usage{
diff_chr(old, new, max_dist = Inf)
}
\arguments{
\item{old}{First character vector.}

\item{new}{Second character vector.}

\item{max_dist}{Maximum distance to consider, or \code{Inf} for no limit.
If the LCS edit distance is larger than this, then the function
throws an error with class \code{"cli_diff_max_dist"}. (If you specify
\code{Inf} the real limit is \code{.Machine$integer.max} but to reach this the
function would have to run a very long time.)}
}
\value{
A list that is a \code{cli_diff_chr} object, with a \code{format()} and a
\code{print()} method. You can also access its members:
\itemize{
\item \code{old} and \code{new} are the original inputs,
\item \code{lcs} is a data frame of LCS edit that transform \code{old} into \code{new}.
}

The \code{lcs} data frame has the following columns:
\itemize{
\item \code{operation}: one of \code{"match"}, \code{"delete"} or \code{"insert"}.
\item \code{offset}: offset in \code{old} for matches and deletions, offset in \code{new}
for insertions.
\item \code{length}: length of the operation, i.e. number of matching, deleted
or inserted elements.
\item \code{old_offset}: offset in \code{old} \emph{after} the operation.
\item \code{new_offset}: offset in \code{new} \emph{after} the operation.
}
}
\description{
Its printed output is similar to calling \code{diff -u} at the command
line.
}
\examples{
letters2 <- c("P", "R", "E", letters, "P", "O", "S", "T")
letters2[11:16] <- c("M", "I", "D", "D", "L", "E")
diff_chr(letters, letters2)
}
\seealso{
The diffobj package for a much more comprehensive set of
\code{diff}-like tools.

Other diff functions in cli: 
\code{\link{diff_str}()}
}
\concept{diff functions in cli}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/symbol.R
\name{symbol}
\alias{symbol}
\alias{list_symbols}
\title{Various handy symbols to use in a command line UI}
\format{
A named list, see \code{names(symbol)} for all sign names.
}
\usage{
symbol

list_symbols()
}
\description{
Various handy symbols to use in a command line UI
}
\details{
On Windows they have a fallback to less fancy symbols.

\code{list_symbols()} prints a table with all symbols to the screen.
}
\examples{
cat(symbol$tick, " SUCCESS\n", symbol$cross, " FAILURE\n", sep = "")

## All symbols
cat(paste(format(names(symbol), width = 20),
  unlist(symbol)), sep = "\n")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/width.R
\name{console_width}
\alias{console_width}
\title{Determine the width of the console}
\usage{
console_width()
}
\value{
Integer scalar, the console with, in number of characters.
}
\description{
It uses the \code{cli.width} option, if set. Otherwise it tries to
determine the size of the terminal or console window.
}
\details{
These are the exact rules:
\itemize{
\item If the \code{cli.width} option is set to a positive integer, it is used.
\item If the \code{cli.width} option is set, but it is not a positive integer,
and error is thrown.
}

Then we try to determine the size of the terminal or console window:
\itemize{
\item If we are not in RStudio, or we are in an RStudio terminal,
then we try to use the \code{tty_size()} function to query the
terminal size. This might fail if R is not running in a terminal,
but failures are ignored.
\item If we are in the RStudio build pane, then the \code{RSTUDIO_CONSOLE_WIDTH}
environment variable is used. If the build pane is resized, then this
environment variable is not accurate any more, and the output might
get garbled.
\item We are \emph{not} using the \code{RSTUDIO_CONSOLE_WIDTH} environment variable
if we are in the RStudio console.
}

If we cannot determine the size of the terminal or console window, then
we use the \code{width} option. If the \code{width} option is not set, then
we return 80L.
}
\examples{
console_width()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ansiex.R
\name{ansi_string}
\alias{ansi_string}
\title{Labels a character vector as containing ANSI control codes.}
\usage{
ansi_string(x)
}
\arguments{
\item{x}{A character vector or something that can be
coerced into one.}
}
\value{
A \code{cli_ansi_string} object, a subclass of
\code{character}, with the same length and contents
as \code{x}.
}
\description{
This function sets the class of its argument, activating
ANSI-string-specific methods such as for printing.
}
\seealso{
Other low level ANSI functions: 
\code{\link{ansi_has_any}()},
\code{\link{ansi_hide_cursor}()},
\code{\link{ansi_regex}()},
\code{\link{ansi_strip}()}
}
\concept{low level ANSI functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/spinner.R
\name{make_spinner}
\alias{make_spinner}
\title{Create a spinner}
\usage{
make_spinner(
  which = NULL,
  stream = "auto",
  template = "{spin}",
  static = c("dots", "print", "print_line", "silent")
)
}
\arguments{
\item{which}{The name of the chosen spinner. If \code{NULL}, then the default
is used, which can be customized via the \code{cli.spinner_unicode},
\code{cli.spinner_ascii} and \code{cli.spinner} options. (The latter applies to
both Unicode and ASCII displays. These options can be set to the name
of a built-in spinner, or to a list that has an entry called \code{frames},
a character vector of frames.}

\item{stream}{The stream to use for the spinner. Typically this is
standard error, or maybe the standard output stream.
It can also be a string, one of \code{"auto"}, \code{"message"}, \code{"stdout"},
\code{"stderr"}. \code{"auto"} will select \code{stdout()} if the session is
interactive and there are no sinks, otherwise it will select
\code{stderr()}.}

\item{template}{A template string, that will contain the spinner. The
spinner itself will be substituted for \code{{spin}}. See example below.}

\item{static}{What to do if the terminal does not support dynamic
displays:
\itemize{
\item \code{"dots"}: show a dot for each \verb{$spin()} call.
\item \code{"print"}: just print the frames of the spinner, one after another.
\item \code{"print_line"}: print the frames of the spinner, each on its own line.
\item \code{"silent"} do not print anything, just the \code{template}.
}}
}
\value{
A \code{cli_spinner} object, which is a list of functions. See
its methods below.

\code{cli_spinner} methods:
\itemize{
\item \verb{$spin()}: output the next frame of the spinner.
\item \verb{$finish()}: terminate the spinner. Depending on terminal capabilities
this removes the spinner from the screen. Spinners can be reused,
you can start calling the \verb{$spin()} method again.
}

All methods return the spinner object itself, invisibly.

The spinner is automatically throttled to its ideal update frequency.
}
\description{
Create a spinner
}
\section{Examples}{

\subsection{Default spinner}{

\if{html}{\out{<div class="sourceCode r">}}\preformatted{sp1 <- make_spinner()
fun_with_spinner <- function() \{
  lapply(1:100, function(x) \{ sp1$spin(); Sys.sleep(0.05) \})
  sp1$finish()
\}
ansi_with_hidden_cursor(fun_with_spinner())
}\if{html}{\out{</div>}}

\if{html}{\figure{make-spinner-default.svg}}
}

\subsection{Spinner with a template}{

\if{html}{\out{<div class="sourceCode r">}}\preformatted{sp2 <- make_spinner(template = "Computing \{spin\}")
fun_with_spinner2 <- function() \{
  lapply(1:100, function(x) \{ sp2$spin(); Sys.sleep(0.05) \})
  sp2$finish()
\}
ansi_with_hidden_cursor(fun_with_spinner2())
}\if{html}{\out{</div>}}

\if{html}{\figure{make-spinner-template.svg}}
}

\subsection{Custom spinner}{

\if{html}{\out{<div class="sourceCode r">}}\preformatted{sp3 <- make_spinner("simpleDotsScrolling", template = "Downloading \{spin\}")
fun_with_spinner3 <- function() \{
  lapply(1:100, function(x) \{ sp3$spin(); Sys.sleep(0.05) \})
  sp3$finish()
\}
ansi_with_hidden_cursor(fun_with_spinner3())
}\if{html}{\out{</div>}}

\if{html}{\figure{make-spinner-custom.svg}}
}
}

\seealso{
Other spinners: 
\code{\link{demo_spinners}()},
\code{\link{get_spinner}()},
\code{\link{list_spinners}()}
}
\concept{spinners}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/diff.R
\name{diff_str}
\alias{diff_str}
\title{Compare two character strings, character by character}
\usage{
diff_str(old, new, max_dist = Inf)
}
\arguments{
\item{old}{First string, must not be \code{NA}.}

\item{new}{Second string, must not be \code{NA}.}

\item{max_dist}{Maximum distance to consider, or \code{Inf} for no limit.
If the LCS edit distance is larger than this, then the function
throws an error with class \code{"cli_diff_max_dist"}. (If you specify
\code{Inf} the real limit is \code{.Machine$integer.max} but to reach this the
function would have to run a very long time.)}
}
\value{
A list that is a \code{cli_diff_str} object and also a
\code{cli_diff_chr} object, see \link{diff_str} for the details about its
structure.
}
\description{
Characters are defined by UTF-8 graphemes.
}
\examples{
str1 <- "abcdefghijklmnopqrstuvwxyz"
str2 <- "PREabcdefgMIDDLEnopqrstuvwxyzPOST"
diff_str(str1, str2)
}
\seealso{
The diffobj package for a much more comprehensive set of
\code{diff}-like tools.

Other diff functions in cli: 
\code{\link{diff_chr}()}
}
\concept{diff functions in cli}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ansiex.R
\name{ansi_grep}
\alias{ansi_grep}
\alias{ansi_grepl}
\title{Like \code{\link[base:grep]{base::grep()}} and \code{\link[base:grep]{base::grepl()}}, but for ANSI strings}
\usage{
ansi_grep(pattern, x, ignore.case = FALSE, perl = FALSE, value = FALSE, ...)

ansi_grepl(pattern, x, ...)
}
\arguments{
\item{pattern}{Character scalar, regular expression or fixed string
(if \code{fixed = TRUE}), the pattern to search for. Other objects will be
coerced using \code{\link[=as.character]{as.character()}}.}

\item{x}{Character vector to search in. Other objects will be coerced
using \code{\link[=as.character]{as.character()}}.}

\item{ignore.case, perl, value}{Passed to \code{\link[base:grep]{base::grep()}}.}

\item{...}{Extra arguments are passed to \code{\link[base:grep]{base::grep()}} or \code{\link[base:grep]{base::grepl()}}.}
}
\value{
The same as \code{\link[base:grep]{base::grep()}} and \code{\link[base:grep]{base::grepl()}}, respectively.
}
\description{
First ANSI sequences will be stripped with \code{\link[=ansi_strip]{ansi_strip()}}, both
}
\details{
Note that these functions work on code points (or bytes if
\code{useBytes = TRUE}), and not graphemes.

Unlike \code{\link[base:grep]{base::grep()}} and \code{\link[base:grep]{base::grepl()}} these functions do not special
case factors.

Both \code{pattern} and \code{x} are converted to UTF-8.
}
\examples{
red_needle <- col_red("needle")
haystack <- c("foo", "needle", "foo")
green_haystack <- col_green(haystack)
ansi_grepl(red_needle, haystack)
ansi_grepl(red_needle, green_haystack)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hash.R
\name{hash_sha256}
\alias{hash_sha256}
\alias{hash_raw_sha256}
\alias{hash_obj_sha256}
\alias{hash_file_sha256}
\title{SHA-256 hash}
\usage{
hash_sha256(x)

hash_raw_sha256(x)

hash_obj_sha256(x, serialize_version = 2)

hash_file_sha256(paths)
}
\arguments{
\item{x}{Character vector. If not a character vector, then
\code{\link[=as.character]{as.character()}} is used to try to coerce it into one. \code{NA} entries
will have an \code{NA} hash.}

\item{serialize_version}{Workspace format version to use, see
\code{\link[base:serialize]{base::serialize()}}.}

\item{paths}{Character vector of file names.}
}
\value{
\code{hash_sha256()} returns aharacter vector of hexadecimal
SHA-256 hashes.

\code{hash_raw_sha256()} returns a character scalar.

\code{hash_obj_sha256()} returns a character scalar.

\code{hash_file_sha256()} returns a character vector of SHA-256
hashes.
}
\description{
Calculate the SHA-256 hash of each element of a character vector.
}
\details{
\code{hash_raw_sha256()} calculates the SHA-256 hash of the bytes
of a raw vector.

\code{hash_obj_sha256()} calculates the SHA-256 hash of an R
object. The object is serialized into a binary vector first.

\code{hash_file_sha256()} calculates the SHA-256 hash of one or
more files.
}
\examples{
hash_sha256(c("foo", NA, "bar", ""))
}
\seealso{
Other hash functions: 
\code{\link{hash_animal}()},
\code{\link{hash_emoji}()},
\code{\link{hash_md5}()}
}
\concept{hash functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ansiex.R
\name{ansi_align}
\alias{ansi_align}
\title{Align an ANSI colored string}
\usage{
ansi_align(
  text,
  width = console_width(),
  align = c("left", "center", "right"),
  type = "width"
)
}
\arguments{
\item{text}{The character vector to align.}

\item{width}{Width of the field to align in.}

\item{align}{Whether to align \code{"left"}, \code{"center"} or \code{"right"}.}

\item{type}{Passed on to \code{\link[=ansi_nchar]{ansi_nchar()}} and there to \code{\link[=nchar]{nchar()}}}
}
\value{
The aligned character vector.
}
\description{
Align an ANSI colored string
}
\details{
\if{html}{\out{<div class="sourceCode r">}}\preformatted{str <- c(
  col_red("This is red"),
  style_bold("This is bold")
)
astr <- ansi_align(str, width = 30)
boxx(astr)
}\if{html}{\out{</div>}}\if{html}{\out{
<div class="asciicast" style="color: #172431;font-family: 'Fira Code',Monaco,Consolas,Menlo,'Bitstream Vera Sans Mono','Powerline Symbols',monospace;line-height: 1.300000"><pre>
#> ┌────────────────────────────────────┐                                          
#> │                                    │                                          
#> │   <span style="color: #DC322F;">This is red</span>                      │                                          
#> │   <span style="font-weight: bold;">This is bold</span>                     │                                          
#> │                                    │                                          
#> └────────────────────────────────────┘                                          
</pre></div>
}}


\if{html}{\out{<div class="sourceCode r">}}\preformatted{str <- c(
  col_red("This is red"),
  style_bold("This is bold")
)
astr <- ansi_align(str, align = "center", width = 30)
boxx(astr)
}\if{html}{\out{</div>}}\if{html}{\out{
<div class="asciicast" style="color: #172431;font-family: 'Fira Code',Monaco,Consolas,Menlo,'Bitstream Vera Sans Mono','Powerline Symbols',monospace;line-height: 1.300000"><pre>
#> ┌────────────────────────────────────┐                                          
#> │                                    │                                          
#> │             <span style="color: #DC322F;">This is red</span>            │                                          
#> │            <span style="font-weight: bold;">This is bold</span>            │                                          
#> │                                    │                                          
#> └────────────────────────────────────┘                                          
</pre></div>
}}


\if{html}{\out{<div class="sourceCode r">}}\preformatted{str <- c(
  col_red("This is red"),
  style_bold("This is bold")
)
astr <- ansi_align(str, align = "right", width = 30)
boxx(astr)
}\if{html}{\out{</div>}}\if{html}{\out{
<div class="asciicast" style="color: #172431;font-family: 'Fira Code',Monaco,Consolas,Menlo,'Bitstream Vera Sans Mono','Powerline Symbols',monospace;line-height: 1.300000"><pre>
#> ┌────────────────────────────────────┐                                          
#> │                                    │                                          
#> │                      <span style="color: #DC322F;">This is red</span>   │                                          
#> │                     <span style="font-weight: bold;">This is bold</span>   │                                          
#> │                                    │                                          
#> └────────────────────────────────────┘                                          
</pre></div>
}}
}
\seealso{
Other ANSI string operations: 
\code{\link{ansi_columns}()},
\code{\link{ansi_nchar}()},
\code{\link{ansi_strsplit}()},
\code{\link{ansi_strtrim}()},
\code{\link{ansi_strwrap}()},
\code{\link{ansi_substr}()},
\code{\link{ansi_substring}()},
\code{\link{ansi_toupper}()},
\code{\link{ansi_trimws}()}
}
\concept{ANSI string operations}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/vt.R
\name{vt_output}
\alias{vt_output}
\title{Simulate (a subset of) a VT-5xx ANSI terminal}
\usage{
vt_output(output, width = 80L, height = 25L)
}
\arguments{
\item{output}{Character vector or raw vector. Character vectors are
collapsed (without a separater), and converted to a raw vector using
\code{\link[base:rawConversion]{base::charToRaw()}}.}

\item{width}{Terminal width.}

\item{height}{Terminal height.}
}
\value{
Data frame with columns \code{lineno}, \code{segmentno}, \code{segment},
\code{attributes}.
}
\description{
This is utility function that calculates the state of a VT-5xx screen
after a certain set of output.
}
\details{
Currently it supports:
\itemize{
\item configurable terminal width and height
\item ASCII printable characters.
\item \verb{\\n}, \verb{\\r}.
\item ANSI SGR colors, 8 color mode, 256 color mode and true color mode.
\item Other ANSI SGR features: bold, italic, underline, strikethrough,
blink, inverse.
}

It does \emph{not} currently supports other features, mode notably:
\itemize{
\item Other ANSI control sequences and features. Other control sequences
are silently ignored.
\item Wide Unicode characters. Their width is not taken into account
correctly.
\item Unicode graphemes.
}
}
\note{
This function is experimental, and the virtual temrinal API will
likely change in future versions of cli.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pluralize.R
\name{no}
\alias{no}
\alias{qty}
\title{Pluralization helper functions}
\usage{
no(expr)

qty(expr)
}
\arguments{
\item{expr}{For \code{no()} it is an expression that is printed as "no" in
cli expressions, it is interpreted as a zero quantity. For \code{qty()}
an expression that sets the pluralization quantity without printing
anything. See examples below.}
}
\description{
Pluralization helper functions
}
\seealso{
Other pluralization: 
\code{\link{pluralization}},
\code{\link{pluralize}()}
}
\concept{pluralization}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/status-bar.R
\name{cli_process_start}
\alias{cli_process_start}
\alias{cli_process_done}
\alias{cli_process_failed}
\title{Indicate the start and termination of some computation in the status bar
(superseded)}
\usage{
cli_process_start(
  msg,
  msg_done = paste(msg, "... done"),
  msg_failed = paste(msg, "... failed"),
  on_exit = c("auto", "failed", "done"),
  msg_class = "alert-info",
  done_class = "alert-success",
  failed_class = "alert-danger",
  .auto_close = TRUE,
  .envir = parent.frame()
)

cli_process_done(
  id = NULL,
  msg_done = NULL,
  .envir = parent.frame(),
  done_class = "alert-success"
)

cli_process_failed(
  id = NULL,
  msg = NULL,
  msg_failed = NULL,
  .envir = parent.frame(),
  failed_class = "alert-danger"
)
}
\arguments{
\item{msg}{The message to show to indicate the start of the process or
computation. It will be collapsed into a single string, and the first
line is kept and cut to \code{\link[=console_width]{console_width()}}.}

\item{msg_done}{The message to use for successful termination.}

\item{msg_failed}{The message to use for unsuccessful termination.}

\item{on_exit}{Whether this process should fail or terminate
successfully when the calling function (or the environment in \code{.envir})
exits.}

\item{msg_class}{The style class to add to the message. Use an empty
string to suppress styling.}

\item{done_class}{The style class to add to the successful termination
message. Use an empty string to suppress styling.a}

\item{failed_class}{The style class to add to the unsuccessful
termination message. Use an empty string to suppress styling.a}

\item{.auto_close}{Whether to clear the status bar when the calling
function finishes (or \code{.envir} is removed from the stack, if
specified).}

\item{.envir}{Environment to evaluate the glue expressions in. It is
also used to auto-clear the status bar if \code{.auto_close} is \code{TRUE}.}

\item{id}{Id of the status bar container to clear. If \code{id} is not the id
of the current status bar (because it was overwritten by another
status bar container), then the status bar is not cleared. If \code{NULL}
(the default) then the status bar is always cleared.}
}
\value{
Id of the status bar container.
}
\description{
\strong{The \verb{cli_process_*()} functions are superseded by
the \code{\link[=cli_progress_message]{cli_progress_message()}} and \code{\link[=cli_progress_step]{cli_progress_step()}} functions,
because they have a better default behavior.}

Typically you call \code{cli_process_start()} to start the process, and then
\code{cli_process_done()} when it is done. If an error happens before
\code{cli_process_done()} is called, then cli automatically shows the message
for unsuccessful termination.
}
\details{
If you handle the errors of the process or computation, then you can do
the opposite: call \code{cli_process_start()} with \code{on_exit = "done"}, and
in the error handler call \code{cli_process_failed()}. cli will automatically
call \code{cli_process_done()} on successful termination, when the calling
function finishes.

See examples below.
}
\examples{

## Failure by default
fun <- function() {
  cli_process_start("Calculating")
  if (interactive()) Sys.sleep(1)
  if (runif(1) < 0.5) stop("Failed")
  cli_process_done()
}
tryCatch(fun(), error = function(err) err)

## Success by default
fun2 <- function() {
  cli_process_start("Calculating", on_exit = "done")
  tryCatch({
    if (interactive()) Sys.sleep(1)
    if (runif(1) < 0.5) stop("Failed")
  }, error = function(err) cli_process_failed())
}
fun2()
}
\seealso{
The \code{\link[=cli_progress_message]{cli_progress_message()}} and \code{\link[=cli_progress_step]{cli_progress_step()}}
functions, for a superior API.

Other status bar: 
\code{\link{cli_status_clear}()},
\code{\link{cli_status_update}()},
\code{\link{cli_status}()}
}
\concept{status bar}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utf8.R
\name{is_utf8_output}
\alias{is_utf8_output}
\title{Whether cli is emitting UTF-8 characters}
\usage{
is_utf8_output()
}
\value{
Flag, whether cli uses UTF-8 characters.
}
\description{
UTF-8 cli characters can be turned on by setting the \code{cli.unicode}
option to \code{TRUE}. They can be turned off by setting if to \code{FALSE}.
If this option is not set, then \code{\link[base:l10n_info]{base::l10n_info()}} is used to detect
UTF-8 support.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rules.R
\name{rule}
\alias{rule}
\title{Make a rule with one or two text labels}
\usage{
rule(
  left = "",
  center = "",
  right = "",
  line = 1,
  col = NULL,
  line_col = col,
  background_col = NULL,
  width = console_width()
)
}
\arguments{
\item{left}{Label to show on the left. It interferes with the \code{center}
label, only at most one of them can be present.}

\item{center}{Label to show at the center. It interferes  with the
\code{left} and \code{right} labels.}

\item{right}{Label to show on the right. It interferes with the \code{center}
label, only at most one of them can be present.}

\item{line}{The character or string that is used to draw the line.
It can also \code{1} or \code{2}, to request a single line (Unicode, if
available), or a double line. Some strings are interpreted specially,
see \emph{Line styles} below.}

\item{col}{Color of text, and default line color. Either an ANSI style
function (see \link[=ansi-styles]{ANSI styles}), or a color name that is
passed to \code{\link[=make_ansi_style]{make_ansi_style()}}.}

\item{line_col, background_col}{Either a color name (used in
\code{\link[=make_ansi_style]{make_ansi_style()}}), or a style function (see
\link[=ansi-styles]{ANSI styles}), to color the line and background.}

\item{width}{Width of the rule. Defaults to the \code{width} option, see
\code{\link[base:options]{base::options()}}.}
}
\value{
Character scalar, the rule.
}
\description{
The rule can include either a centered text label, or labels on the
left and right side.

To color the labels, use the functions \verb{col_*}, \verb{bg_*} and \verb{style_*}
functions, see \link[=ansi-styles]{ANSI styles}, and the examples below.
To color the line, either these functions directly, or the \code{line_col}
option.
}
\details{
\subsection{Simple rule}{

\if{html}{\out{<div class="sourceCode r">}}\preformatted{rule()
}\if{html}{\out{</div>}}\if{html}{\out{
<div class="asciicast" style="color: #172431;font-family: 'Fira Code',Monaco,Consolas,Menlo,'Bitstream Vera Sans Mono','Powerline Symbols',monospace;line-height: 1.300000"><pre>
#> ──────────────────────────────────────────────────────────────────────          
</pre></div>
}}

}

\subsection{Line styles}{

Some strings for the \code{line} argument are interpreted specially:
\itemize{
\item \code{"single"}: (same as \code{1}), a single line,
\item \code{"double"}: (same as \code{2}), a double line,
\item \code{"bar1"}, \code{"bar2"}, \code{"bar3"}, etc., \code{"bar8"} uses varying height bars.
}
\subsection{Double rule}{

\if{html}{\out{<div class="sourceCode r">}}\preformatted{rule(line = 2)
}\if{html}{\out{</div>}}\if{html}{\out{
<div class="asciicast" style="color: #172431;font-family: 'Fira Code',Monaco,Consolas,Menlo,'Bitstream Vera Sans Mono','Powerline Symbols',monospace;line-height: 1.300000"><pre>
#> ══════════════════════════════════════════════════════════════════════          
</pre></div>
}}

}

\subsection{Bars}{

\if{html}{\out{<div class="sourceCode r">}}\preformatted{rule(line = "bar2")
rule(line = "bar5")
}\if{html}{\out{</div>}}\if{html}{\out{
<div class="asciicast" style="color: #172431;font-family: 'Fira Code',Monaco,Consolas,Menlo,'Bitstream Vera Sans Mono','Powerline Symbols',monospace;line-height: 1.300000"><pre>
#> ▂▂▂▂▂▂▂▂▂▂▂▂▂▂▂▂▂▂▂▂▂▂▂▂▂▂▂▂▂▂▂▂▂▂▂▂▂▂▂▂▂▂▂▂▂▂▂▂▂▂▂▂▂▂▂▂▂▂▂▂▂▂▂▂▂▂▂▂▂▂          
#> ▅▅▅▅▅▅▅▅▅▅▅▅▅▅▅▅▅▅▅▅▅▅▅▅▅▅▅▅▅▅▅▅▅▅▅▅▅▅▅▅▅▅▅▅▅▅▅▅▅▅▅▅▅▅▅▅▅▅▅▅▅▅▅▅▅▅▅▅▅▅          
</pre></div>
}}

}

\subsection{Custom lines}{

\if{html}{\out{<div class="sourceCode r">}}\preformatted{rule(center = "TITLE", line = "~")
}\if{html}{\out{</div>}}\if{html}{\out{
<div class="asciicast" style="color: #172431;font-family: 'Fira Code',Monaco,Consolas,Menlo,'Bitstream Vera Sans Mono','Powerline Symbols',monospace;line-height: 1.300000"><pre>
#> ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ TITLE ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~          
</pre></div>
}}


\if{html}{\out{<div class="sourceCode r">}}\preformatted{rule(center = "TITLE", line = col_blue("~-"))
}\if{html}{\out{</div>}}\if{html}{\out{
<div class="asciicast" style="color: #172431;font-family: 'Fira Code',Monaco,Consolas,Menlo,'Bitstream Vera Sans Mono','Powerline Symbols',monospace;line-height: 1.300000"><pre>
#> <span style="color: #268BD2;">~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-</span> TITLE <span style="color: #268BD2;">~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~</span>          
</pre></div>
}}


\if{html}{\out{<div class="sourceCode r">}}\preformatted{rule(center = bg_red(" ", symbol$star, "TITLE",
  symbol$star, " "),
  line = "\\u2582",
  line_col = "orange")
}\if{html}{\out{</div>}}\if{html}{\out{
<div class="asciicast" style="color: #172431;font-family: 'Fira Code',Monaco,Consolas,Menlo,'Bitstream Vera Sans Mono','Powerline Symbols',monospace;line-height: 1.300000"><pre>
#> <span style="color: #ff9900;">▂▂▂▂▂▂▂▂▂▂▂▂▂▂▂▂▂▂▂▂▂▂▂▂▂▂▂▂▂▂</span> <span style="background-color: #DC322F;"> ★TITLE★ </span> <span style="color: #ff9900;">▂▂▂▂▂▂▂▂▂▂▂▂▂▂▂▂▂▂▂▂▂▂▂▂▂▂▂▂▂</span>          
</pre></div>
}}

}

}

\subsection{Left label}{

\if{html}{\out{<div class="sourceCode r">}}\preformatted{rule(left = "Results")
}\if{html}{\out{</div>}}\if{html}{\out{
<div class="asciicast" style="color: #172431;font-family: 'Fira Code',Monaco,Consolas,Menlo,'Bitstream Vera Sans Mono','Powerline Symbols',monospace;line-height: 1.300000"><pre>
#> ── Results ───────────────────────────────────────────────────────────          
</pre></div>
}}

}

\subsection{Centered label}{

\if{html}{\out{<div class="sourceCode r">}}\preformatted{rule(center = " * RESULTS * ")
}\if{html}{\out{</div>}}\if{html}{\out{
<div class="asciicast" style="color: #172431;font-family: 'Fira Code',Monaco,Consolas,Menlo,'Bitstream Vera Sans Mono','Powerline Symbols',monospace;line-height: 1.300000"><pre>
#> ────────────────────────────  * RESULTS *  ───────────────────────────          
</pre></div>
}}

}

\subsection{Colored labels}{

\if{html}{\out{<div class="sourceCode r">}}\preformatted{rule(center = col_red(" * RESULTS * "))
}\if{html}{\out{</div>}}\if{html}{\out{
<div class="asciicast" style="color: #172431;font-family: 'Fira Code',Monaco,Consolas,Menlo,'Bitstream Vera Sans Mono','Powerline Symbols',monospace;line-height: 1.300000"><pre>
#> ──────────────────────────── <span style="color: #DC322F;"> * RESULTS * </span> ───────────────────────────          
</pre></div>
}}

}

\subsection{Colored line}{

\if{html}{\out{<div class="sourceCode r">}}\preformatted{rule(center = col_red(" * RESULTS * "), line_col = "red")
}\if{html}{\out{</div>}}\if{html}{\out{
<div class="asciicast" style="color: #172431;font-family: 'Fira Code',Monaco,Consolas,Menlo,'Bitstream Vera Sans Mono','Powerline Symbols',monospace;line-height: 1.300000"><pre>
#> <span style="color: #DC322F;">────────────────────────────</span> <span style="color: #DC322F;"> * RESULTS * </span> <span style="color: #DC322F;">───────────────────────────</span>          
</pre></div>
}}

}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ansi.R
\name{make_ansi_style}
\alias{make_ansi_style}
\title{Create a new ANSI style}
\usage{
make_ansi_style(..., bg = FALSE, grey = FALSE, colors = num_ansi_colors())
}
\arguments{
\item{...}{The style to create. See details and examples below.}

\item{bg}{Whether the color applies to the background.}

\item{grey}{Whether to specifically create a grey color.
This flag is included, because ANSI 256 has a finer color scale
for greys, then the usual 0:5 scale for red, green and blue components.
It is only used for RGB color specifications (either numerically
or via a hexa string), and it is ignored on eigth color ANSI
terminals.}

\item{colors}{Number of colors, detected automatically
by default.}
}
\value{
A function that can be used to color (style) strings.
}
\description{
Create a function that can be used to add ANSI styles to text.
}
\details{
The \code{...} style argument can be any of the following:
\itemize{
\item A cli ANSI style function of class \code{ansi_style}. This is returned
as is, without looking at the other arguments.
\item An R color name, see \code{\link[grDevices:colors]{grDevices::colors()}}.
\item A 6- or 8-digit hexa color string, e.g. \verb{#ff0000} means
red. Transparency (alpha channel) values are ignored.
\item A one-column matrix with three rows for the red, green
and blue channels, as returned by \code{\link[grDevices:col2rgb]{grDevices::col2rgb()}}.
}

\code{make_ansi_style()} detects the number of colors to use
automatically (this can be overridden using the \code{colors}
argument). If the number of colors is less than 256 (detected or given),
then it falls back to the color in the ANSI eight color mode that
is closest to the specified (RGB or R) color.
}
\examples{
make_ansi_style("orange")
make_ansi_style("#123456")
make_ansi_style("orange", bg = TRUE)

orange <- make_ansi_style("orange")
orange("foobar")
cat(orange("foobar"))
}
\seealso{
Other ANSI styling: 
\code{\link{ansi-styles}},
\code{\link{combine_ansi_styles}()},
\code{\link{num_ansi_colors}()}
}
\concept{ANSI styling}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/progress-along.R
\name{cli_progress_along}
\alias{cli_progress_along}
\title{Add a progress bar to a mapping function or for loop}
\usage{
cli_progress_along(
  x,
  name = NULL,
  total = length(x),
  ...,
  .envir = parent.frame()
)
}
\arguments{
\item{x}{Sequence to add the progress bar to.}

\item{name}{Name of the progress bar, a label, passed to
\code{\link[=cli_progress_bar]{cli_progress_bar()}}.}

\item{total}{Passed to \code{\link[=cli_progress_bar]{cli_progress_bar()}}.}

\item{...}{Passed to \code{\link[=cli_progress_bar]{cli_progress_bar()}}.}

\item{.envir}{Passed to \code{\link[=cli_progress_bar]{cli_progress_bar()}}.}
}
\value{
An index vector from 1 to \code{length(x)} that triggers progress
updates as you iterate over it.
}
\description{
Note that this function is currently experimental!

Use \code{cli_progress_along()} in a mapping function or in a for loop, to add a
progress bar. It uses \code{\link[=cli_progress_bar]{cli_progress_bar()}} internally.
}
\details{
\subsection{\code{for} loop}{

A \code{for} loop with \code{cli_progress_along()} looks like this:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{for (i in cli_progress_along(seq)) \{
  ...
\}
}\if{html}{\out{</div>}}

A complete example:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{clifun <- function() \{
  for (i in cli_progress_along(1:100, "Downloading")) \{
     Sys.sleep(4/100)
  \}
\}
clifun()
}\if{html}{\out{</div>}}

\if{html}{\figure{progress-along-1.svg}}
}

\subsection{\code{lapply()} and other mapping functions}{

They will look like this:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{lapply(cli_progress_along(X), function(i) ...)
}\if{html}{\out{</div>}}

A complete example:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{res <- lapply(cli_progress_along(1:100, "Downloading"), function(i) \{
  Sys.sleep(4/100)
\})
}\if{html}{\out{</div>}}

\if{html}{\figure{progress-along-2.svg}}
}

\subsection{Custom format string}{

\if{html}{\out{<div class="sourceCode r">}}\preformatted{clifun <- function() \{
  for (i in cli_progress_along(1:100,
      format = "Downloading data file \{cli::pb_current\}")) \{
     Sys.sleep(4/100)
  \}
\}
clifun()
}\if{html}{\out{</div>}}

\if{html}{\figure{progress-along-3.svg}}
}

\subsection{Breaking out of loops}{

Note that if you use \code{break} in the \code{for} loop, you probably want to
terminate the progress bar explicitly when breaking out of the loop,
or right after the loop:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{for (i in cli_progress_along(seq)) \{
  ...
  if (cond) cli_progress_done() && break
  ...
\}
}\if{html}{\out{</div>}}
}
}
\seealso{
This function supports \link[=inline-markup]{inline markup}.

\code{\link[=cli_progress_bar]{cli_progress_bar()}} and the traditional progress bar API.

Other progress bar functions: 
\code{\link{cli_progress_bar}()},
\code{\link{cli_progress_builtin_handlers}()},
\code{\link{cli_progress_message}()},
\code{\link{cli_progress_num}()},
\code{\link{cli_progress_output}()},
\code{\link{cli_progress_step}()},
\code{\link{cli_progress_styles}()},
\code{\link{progress-variables}}

Other functions supporting inline markup: 
\code{\link{cli_abort}()},
\code{\link{cli_alert}()},
\code{\link{cli_blockquote}()},
\code{\link{cli_bullets}()},
\code{\link{cli_bullets_raw}()},
\code{\link{cli_dl}()},
\code{\link{cli_h1}()},
\code{\link{cli_li}()},
\code{\link{cli_ol}()},
\code{\link{cli_process_start}()},
\code{\link{cli_progress_bar}()},
\code{\link{cli_progress_message}()},
\code{\link{cli_progress_output}()},
\code{\link{cli_progress_step}()},
\code{\link{cli_rule}},
\code{\link{cli_status}()},
\code{\link{cli_status_update}()},
\code{\link{cli_text}()},
\code{\link{cli_ul}()},
\code{\link{format_error}()},
\code{\link{format_inline}()}
}
\concept{functions supporting inline markup}
\concept{progress bar functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ansiex.R
\name{ansi_html_style}
\alias{ansi_html_style}
\title{CSS styles for the output of \code{ansi_html()}}
\usage{
ansi_html_style(
  colors = TRUE,
  palette = c("vscode", "dichro", "vga", "winxp", "win10", "macos", "putty", "mirc",
    "xterm", "ubuntu", "eclipse", "iterm", "iterm-pastel", "iterm-smoooooth",
    "iterm-snazzy", "iterm-solarized", "iterm-tango")
)
}
\arguments{
\item{colors}{Whether or not to include colors. \code{FALSE} will not include
colors, \code{TRUE} or \code{8} will include eight colors (plus their bright
variants), \code{256} will include 256 colors.}

\item{palette}{Character scalar, palette to use for the first eight colors
plus their bright variants. Terminals define these colors differently,
and cli includes a couple of examples. Sources of palettes:
\itemize{
\item https://en.wikipedia.org/wiki/ANSI_escape_code#3-bit_and_4-bit
\item iTerm2 builtin palettes
\item \url{https://github.com/sindresorhus/iterm2-snazzy}
}}
}
\value{
Named list of CSS declaration blocks, where the names are
CSS selectors. It has a \code{format()} and \code{print()} methods, which you
can use to write the output to a CSS or HTML file.
}
\description{
CSS styles for the output of \code{ansi_html()}
}
\examples{
ansi_html_style(colors = FALSE)
ansi_html_style(colors = 8, palette = "iterm-snazzy")
}
\seealso{
Other ANSI to HTML conversion: 
\code{\link{ansi_html}()}
}
\concept{ANSI to HTML conversion}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/format.R
\name{cli_vec}
\alias{cli_vec}
\title{Add custom cli style to a vector}
\usage{
cli_vec(x, style = list())
}
\arguments{
\item{x}{Vector that will be collapsed by cli.}

\item{style}{Style to apply to the vector. It is used as a theme on
a \code{span} element that is created for the vector. You can set \code{vec-sep}
and \code{vec-last} to modify the separator and the last separator.}
}
\description{
Add custom cli style to a vector
}
\details{
You can use this function to change the default parameters of
collapsing the vector into a string, see an example below.

The style is added as an attribute, so operations that remove
attributes will remove the style as well.
\subsection{Custom collapsing separator}{

\if{html}{\out{<div class="sourceCode r">}}\preformatted{v <- cli_vec(
  c("foo", "bar", "foobar"),
  style = list("vec-sep" = " & ", "vec-last" = " & ")
)
cli_text("My list: \{v\}.")
}\if{html}{\out{</div>}}\if{html}{\out{
<div class="asciicast" style="color: #172431;font-family: 'Fira Code',Monaco,Consolas,Menlo,'Bitstream Vera Sans Mono','Powerline Symbols',monospace;line-height: 1.300000"><pre>
#> My list: foo &amp; bar &amp; foobar.                                                    
</pre></div>
}}

}

\subsection{Custom truncation}{

\if{html}{\out{<div class="sourceCode r">}}\preformatted{x <- cli_vec(names(mtcars), list("vec-trunc" = 3))
cli_text("Column names: \{x\}.")
}\if{html}{\out{</div>}}\if{html}{\out{
<div class="asciicast" style="color: #172431;font-family: 'Fira Code',Monaco,Consolas,Menlo,'Bitstream Vera Sans Mono','Powerline Symbols',monospace;line-height: 1.300000"><pre>
#> Column names: mpg, cyl, disp, …, gear, and carb.                                
</pre></div>
}}

}
}
\seealso{
\code{\link[=cli_format]{cli_format()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/themes.R
\name{cli_list_themes}
\alias{cli_list_themes}
\title{List the currently active themes}
\usage{
cli_list_themes()
}
\value{
A list of data frames with the active themes.
Each data frame row is a style that applies to selected CLI tree nodes.
Each data frame has columns:
\itemize{
\item \code{selector}: The original CSS-like selector string. See \link{themes}.
\item \code{parsed}: The parsed selector, as used by cli for matching to nodes.
\item \code{style}: The original style.
\item \code{cnt}: The id of the container the style is currently applied to, or
\code{NA} if the style is not used.
}
}
\description{
If there is no active app, then it calls \code{\link[=start_app]{start_app()}}.
}
\seealso{
\link{themes}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ansiex.R
\name{ansi_strsplit}
\alias{ansi_strsplit}
\title{Split an ANSI colored string}
\usage{
ansi_strsplit(x, split, ...)
}
\arguments{
\item{x}{Character vector, potentially ANSI styled, or a vector to
coerced to character.}

\item{split}{Character vector of length 1 (or object which can be coerced to
such) containing regular expression(s) (unless \code{fixed = TRUE}) to use
for splitting.  If empty matches occur, in particular if \code{split} has
zero characters, \code{x} is split into single characters.}

\item{...}{Extra arguments are passed to \code{base::strsplit()}.}
}
\value{
A list of the same length as \code{x}, the \code{i}-th element of
which contains the vector of splits of \code{x[i]}. ANSI styles are
retained.
}
\description{
This is the color-aware counterpart of \code{\link[base:strsplit]{base::strsplit()}}.
It works almost exactly like the original, but keeps the colors in the
substrings.
}
\examples{
str <- paste0(
  col_red("I am red---"),
  col_green("and I am green-"),
  style_underline("I underlined")
)

cat(str, "\n")

# split at dashes, keep color
cat(ansi_strsplit(str, "[-]+")[[1]], sep = "\n")
strsplit(ansi_strip(str), "[-]+")

# split to characters, keep color
cat(ansi_strsplit(str, "")[[1]], "\n", sep = " ")
strsplit(ansi_strip(str), "")
}
\seealso{
Other ANSI string operations: 
\code{\link{ansi_align}()},
\code{\link{ansi_columns}()},
\code{\link{ansi_nchar}()},
\code{\link{ansi_strtrim}()},
\code{\link{ansi_strwrap}()},
\code{\link{ansi_substr}()},
\code{\link{ansi_substring}()},
\code{\link{ansi_toupper}()},
\code{\link{ansi_trimws}()}
}
\concept{ANSI string operations}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/spinner.R
\name{get_spinner}
\alias{get_spinner}
\title{Character vector to put a spinner on the screen}
\usage{
get_spinner(which = NULL)
}
\arguments{
\item{which}{The name of the chosen spinner. If \code{NULL}, then the default
is used, which can be customized via the \code{cli.spinner_unicode},
\code{cli.spinner_ascii} and \code{cli.spinner} options. (The latter applies to
both Unicode and ASCII displays. These options can be set to the name
of a built-in spinner, or to a list that has an entry called \code{frames},
a character vector of frames.}
}
\value{
A list with entries: \code{name}, \code{interval}: the suggested update
interval in milliseconds and \code{frames}: the character vector of the
spinner's frames.
}
\description{
\code{cli} contains many different spinners, you choose one according to your
taste.
}
\details{
\if{html}{\out{<div class="sourceCode r">}}\preformatted{options(cli.spinner = "hearts")
fun <- function() \{
  cli_progress_bar("Spinning")
  for (i in 1:100) \{
    Sys.sleep(4/100)
    cli_progress_update()
  \}
\}
fun()
options(cli.spinner = NULL)
}\if{html}{\out{</div>}}

\if{html}{\figure{get-spinner.svg}}
}
\seealso{
Other spinners: 
\code{\link{demo_spinners}()},
\code{\link{list_spinners}()},
\code{\link{make_spinner}()}
}
\concept{spinners}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cli.R
\name{cli_h1}
\alias{cli_h1}
\alias{cli_h2}
\alias{cli_h3}
\title{CLI headings}
\usage{
cli_h1(text, id = NULL, class = NULL, .envir = parent.frame())

cli_h2(text, id = NULL, class = NULL, .envir = parent.frame())

cli_h3(text, id = NULL, class = NULL, .envir = parent.frame())
}
\arguments{
\item{text}{Text of the heading. It can contain inline markup.}

\item{id}{Id of the heading element, string. It can be used in themes.}

\item{class}{Class of the heading element, string. It can be used in
themes.}

\item{.envir}{Environment to evaluate the glue expressions in.}
}
\description{
cli has three levels of headings.
}
\details{
This is how the headings look with the default builtin theme.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{cli_h1("Header \{.emph 1\}")
cli_h2("Header \{.emph 2\}")
cli_h3("Header \{.emph 3\}")
}\if{html}{\out{</div>}}\if{html}{\out{
<div class="asciicast" style="color: #172431;font-family: 'Fira Code',Monaco,Consolas,Menlo,'Bitstream Vera Sans Mono','Powerline Symbols',monospace;line-height: 1.300000"><pre>
#>                                                                                 
#> <span style="color: #2AA198;">──</span> <span style="font-weight: bold;">Header </span><span style="font-weight: bold;font-style: italic;">1</span> <span style="color: #2AA198;">──────────────────────────────────────────────────────────</span>          
#>                                                                                 
#> ── <span style="font-weight: bold;">Header </span><span style="font-weight: bold;font-style: italic;">2</span> ──                                                                  
#>                                                                                 
#> ── Header <span style="font-style: italic;">3</span>                                                                     
</pre></div>
}}
}
\seealso{
These functions supports \link[=inline-markup]{inline markup}.

Other functions supporting inline markup: 
\code{\link{cli_abort}()},
\code{\link{cli_alert}()},
\code{\link{cli_blockquote}()},
\code{\link{cli_bullets}()},
\code{\link{cli_bullets_raw}()},
\code{\link{cli_dl}()},
\code{\link{cli_li}()},
\code{\link{cli_ol}()},
\code{\link{cli_process_start}()},
\code{\link{cli_progress_along}()},
\code{\link{cli_progress_bar}()},
\code{\link{cli_progress_message}()},
\code{\link{cli_progress_output}()},
\code{\link{cli_progress_step}()},
\code{\link{cli_rule}},
\code{\link{cli_status}()},
\code{\link{cli_status_update}()},
\code{\link{cli_text}()},
\code{\link{cli_ul}()},
\code{\link{format_error}()},
\code{\link{format_inline}()}
}
\concept{functions supporting inline markup}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hash.R
\name{hash_md5}
\alias{hash_md5}
\alias{hash_raw_md5}
\alias{hash_obj_md5}
\alias{hash_file_md5}
\title{MD5 hash}
\usage{
hash_md5(x)

hash_raw_md5(x)

hash_obj_md5(x, serialize_version = 2)

hash_file_md5(paths)
}
\arguments{
\item{x}{Character vector. If not a character vector, then
\code{\link[=as.character]{as.character()}} is used to try to coerce it into one. \code{NA} entries
will have an \code{NA} hash.}

\item{serialize_version}{Workspace format version to use, see
\code{\link[base:serialize]{base::serialize()}}.}

\item{paths}{Character vector of file names.}
}
\value{
\code{hash_md5()} returns a character vector of hexadecimal MD5
hashes.

\code{hash_raw_md5()} returns a character scalar.

\code{hash_obj_md5()} returns a character scalar.
}
\description{
Calculate the MD5 hash of each element of a character vector.
}
\details{
\code{hash_raw_md5()} calculates the MD5 hash of the bytes
of a raw vector.

\code{hash_obj_md5()} calculates the MD5 hash of an R
object. The object is serialized into a binary vector first.

\code{hash_file_md5()} calculates the MD5 hash of one or more
files.
}
\examples{
hash_md5(c("foo", NA, "bar", ""))
}
\seealso{
\code{\link[tools:md5sum]{tools::md5sum()}} for a base R MD5 function that works on
files.

Other hash functions: 
\code{\link{hash_animal}()},
\code{\link{hash_emoji}()},
\code{\link{hash_sha1}()},
\code{\link{hash_sha256}()},
\code{\link{hash_xxhash}()}
}
\concept{hash functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cli-package.R
\docType{package}
\name{cli-package}
\alias{cli-package}
\title{cli: Helpers for Developing Command Line Interfaces}
\description{
A suite of tools to build attractive command line interfaces ('CLIs'), from semantic elements: headings, lists, alerts, paragraphs, etc. Supports custom themes via a 'CSS'-like language. It also contains a number of lower level 'CLI' elements: rules, boxes, trees, and 'Unicode' symbols with 'ASCII' alternatives. It support ANSI colors and text styles as well.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://cli.r-lib.org}
  \item \url{https://github.com/r-lib/cli}
  \item Report bugs at \url{https://github.com/r-lib/cli/issues}
}

}
\author{
\strong{Maintainer}: Gábor Csárdi \email{gabor@posit.co}

Other contributors:
\itemize{
  \item Hadley Wickham [contributor]
  \item Kirill Müller [contributor]
  \item Salim Brüggemann \email{salim-b@pm.me} (\href{https://orcid.org/0000-0002-5329-5987}{ORCID}) [contributor]
  \item Posit Software, PBC [copyright holder, funder]
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cli.R
\name{cli_li}
\alias{cli_li}
\title{CLI list item(s)}
\usage{
cli_li(
  items = NULL,
  labels = names(items),
  id = NULL,
  class = NULL,
  .auto_close = TRUE,
  .envir = parent.frame()
)
}
\arguments{
\item{items}{Character vector of items, or \code{NULL}.}

\item{labels}{For definition lists the item labels.}

\item{id}{Id of the new container. Can be used for closing it with
\code{\link[=cli_end]{cli_end()}} or in themes. If \code{NULL}, then an id is generated and
returned invisibly.}

\item{class}{Class of the item container. Can be used in themes.}

\item{.auto_close}{Whether to close the container, when the calling
function finishes (or \code{.envir} is removed, if specified).}

\item{.envir}{Environment to evaluate the glue expressions in. It is
also used to auto-close the container if \code{.auto_close} is \code{TRUE}.}
}
\value{
The id of the new container element, invisibly.
}
\description{
A list item is a container, see \link{containers}.
}
\details{
\subsection{Nested lists}{

\if{html}{\out{<div class="sourceCode r">}}\preformatted{fun <- function() \{
  ul <- cli_ul()
  cli_li("one:")
  cli_ol(letters[1:3])
  cli_li("two:")
  cli_li("three")
  cli_end(ul)
\}
fun()
}\if{html}{\out{</div>}}\if{html}{\out{
<div class="asciicast" style="color: #172431;font-family: 'Fira Code',Monaco,Consolas,Menlo,'Bitstream Vera Sans Mono','Powerline Symbols',monospace;line-height: 1.300000"><pre>
#> • one:                                                                          
#>   1. a                                                                          
#>   2. b                                                                          
#>   3. c                                                                          
#> • two:                                                                          
#> • three                                                                         
</pre></div>
}}

}
}
\seealso{
This function supports \link[=inline-markup]{inline markup}.

Other functions supporting inline markup: 
\code{\link{cli_abort}()},
\code{\link{cli_alert}()},
\code{\link{cli_blockquote}()},
\code{\link{cli_bullets}()},
\code{\link{cli_bullets_raw}()},
\code{\link{cli_dl}()},
\code{\link{cli_h1}()},
\code{\link{cli_ol}()},
\code{\link{cli_process_start}()},
\code{\link{cli_progress_along}()},
\code{\link{cli_progress_bar}()},
\code{\link{cli_progress_message}()},
\code{\link{cli_progress_output}()},
\code{\link{cli_progress_step}()},
\code{\link{cli_rule}},
\code{\link{cli_status}()},
\code{\link{cli_status_update}()},
\code{\link{cli_text}()},
\code{\link{cli_ul}()},
\code{\link{format_error}()},
\code{\link{format_inline}()}
}
\concept{functions supporting inline markup}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tty.R
\name{is_ansi_tty}
\alias{is_ansi_tty}
\title{Detect if a stream support ANSI escape characters}
\usage{
is_ansi_tty(stream = "auto")
}
\arguments{
\item{stream}{The stream to inspect or manipulate, an R connection
object. It can also be a string, one of \code{"auto"}, \code{"message"},
\code{"stdout"}, \code{"stderr"}. \code{"auto"} will select \code{stdout()} if the session is
interactive and there are no sinks, otherwise it will select \code{stderr()}.}
}
\value{
\code{TRUE} or \code{FALSE}.
}
\description{
We check that all of the following hold:
\itemize{
\item The stream is a terminal.
\item The platform is Unix.
\item R is not running inside R.app (the macOS GUI).
\item R is not running inside RStudio.
\item R is not running inside Emacs.
\item The terminal is not "dumb".
\item \code{stream} is either the standard output or the standard error stream.
}
}
\examples{
is_ansi_tty()
}
\seealso{
Other terminal capabilities: 
\code{\link{ansi_hide_cursor}()},
\code{\link{is_dynamic_tty}()}
}
\concept{terminal capabilities}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ansi.R
\name{combine_ansi_styles}
\alias{combine_ansi_styles}
\title{Combine two or more ANSI styles}
\usage{
combine_ansi_styles(...)
}
\arguments{
\item{...}{The styles to combine. For character strings, the
\code{\link[=make_ansi_style]{make_ansi_style()}} function is used to create a style first.
They will be applied from right to left.}
}
\value{
The combined style function.
}
\description{
Combine two or more styles or style functions into a new style function
that can be called on strings to style them.
}
\details{
It does not usually make sense to combine two foreground
colors (or two background colors), because only the first one
applied will be used.

It does make sense to combine different kind of styles,
e.g. background color, foreground color, bold font.
}
\examples{
## Use style names
alert <- combine_ansi_styles("bold", "red4")
cat(alert("Warning!"), "\n")

## Or style functions
alert <- combine_ansi_styles(style_bold, col_red, bg_cyan)
cat(alert("Warning!"), "\n")

## Combine a composite style
alert <- combine_ansi_styles(
  "bold",
  combine_ansi_styles("red", bg_cyan))
cat(alert("Warning!"), "\n")
}
\seealso{
Other ANSI styling: 
\code{\link{ansi-styles}},
\code{\link{make_ansi_style}()},
\code{\link{num_ansi_colors}()}
}
\concept{ANSI styling}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bullets.R
\name{cli_bullets_raw}
\alias{cli_bullets_raw}
\alias{format_bullets_raw}
\title{List of verbatim items}
\usage{
cli_bullets_raw(text, id = NULL, class = NULL)

format_bullets_raw(text, id = NULL, class = NULL)
}
\arguments{
\item{text}{Character vector of items. See details below on how names
are interpreted.}

\item{id}{Optional id of the \code{div.bullets} element, can be used in themes.}

\item{class}{Optional additional class(es) for the \code{div.bullets} element.}
}
\description{
\code{cli_format_bullets_raw()} is similar to \code{\link[=cli_bullets]{cli_bullets()}}, but it does
not perform any inline styling or glue substitutions in the input.
}
\details{
\code{format_bullets_raw()} returns the output instead of printing it.
}
\seealso{
These functions support \link[=inline-markup]{inline markup}.

See \code{\link[=cli_bullets]{cli_bullets()}} for examples.

Other functions supporting inline markup: 
\code{\link{cli_abort}()},
\code{\link{cli_alert}()},
\code{\link{cli_blockquote}()},
\code{\link{cli_bullets}()},
\code{\link{cli_dl}()},
\code{\link{cli_h1}()},
\code{\link{cli_li}()},
\code{\link{cli_ol}()},
\code{\link{cli_process_start}()},
\code{\link{cli_progress_along}()},
\code{\link{cli_progress_bar}()},
\code{\link{cli_progress_message}()},
\code{\link{cli_progress_output}()},
\code{\link{cli_progress_step}()},
\code{\link{cli_rule}},
\code{\link{cli_status}()},
\code{\link{cli_status_update}()},
\code{\link{cli_text}()},
\code{\link{cli_ul}()},
\code{\link{format_error}()},
\code{\link{format_inline}()}
}
\concept{functions supporting inline markup}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sitrep.R
\name{cli_sitrep}
\alias{cli_sitrep}
\title{cli situation report}
\usage{
cli_sitrep()
}
\value{
Named list with entries listed above. It has a \code{cli_sitrep}
class, with a \code{print()} and \code{format()} method.
}
\description{
Contains currently:
\itemize{
\item \code{cli_unicode_option}: whether the \code{cli.unicode} option is set and its
value. See \code{\link[=is_utf8_output]{is_utf8_output()}}.
\item \code{symbol_charset}: the selected character set for \link{symbol}, UTF-8,
Windows, or ASCII.
\item \code{console_utf8}: whether the console supports UTF-8. See
\code{\link[base:l10n_info]{base::l10n_info()}}.
\item \code{latex_active}: whether we are inside knitr, creating a LaTeX
document.
\item \code{num_colors}: number of ANSI colors. See \code{\link[=num_ansi_colors]{num_ansi_colors()}}.
\item \code{console_with}: detected console width.
}
}
\examples{
cli_sitrep()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rlang.R
\name{cli_abort}
\alias{cli_abort}
\alias{cli_warn}
\alias{cli_inform}
\title{Signal an error, warning or message with a cli formatted
message}
\usage{
cli_abort(
  message,
  ...,
  call = .envir,
  .envir = parent.frame(),
  .frame = .envir
)

cli_warn(message, ..., .envir = parent.frame())

cli_inform(message, ..., .envir = parent.frame())
}
\arguments{
\item{message}{It is formatted via a call to \code{\link[=cli_bullets]{cli_bullets()}}.}

\item{...}{Passed to \code{\link[rlang:abort]{rlang::abort()}}, \code{\link[rlang:abort]{rlang::warn()}} or
\code{\link[rlang:abort]{rlang::inform()}}.}

\item{call}{The execution environment of a currently running
function, e.g. \code{call = caller_env()}. The corresponding function
call is retrieved and mentioned in error messages as the source
of the error.

You only need to supply \code{call} when throwing a condition from a
helper function which wouldn't be relevant to mention in the
message.

Can also be \code{NULL} or a \link[rlang:topic-defuse]{defused function call} to
respectively not display any call or hard-code a code to display.

For more information about error calls, see \ifelse{html}{\link[rlang:topic-error-call]{Including function calls in error messages}}{\link[rlang:topic-error-call]{Including function calls in error messages}}.}

\item{.envir}{Environment to evaluate the glue expressions in.}

\item{.frame}{The throwing context. Used as default for
\code{.trace_bottom}, and to determine the internal package to mention
in internal errors when \code{.internal} is \code{TRUE}.}
}
\description{
These functions let you create error, warning or diagnostic
messages with cli formatting, including inline styling,
pluralization and glue substitutions.
}
\details{
\if{html}{\out{<div class="sourceCode r">}}\preformatted{n <- "boo"
cli_abort(c(
        "\{.var n\} must be a numeric vector",
  "x" = "You've supplied a \{.cls \{class(n)\}\} vector."
))
}\if{html}{\out{</div>}}\if{html}{\out{
<div class="asciicast" style="color: #172431;font-family: 'Fira Code',Monaco,Consolas,Menlo,'Bitstream Vera Sans Mono','Powerline Symbols',monospace;line-height: 1.300000"><pre>
#> <span style="font-weight: bold;color: #B58900;">Error</span><span style="font-weight: bold;">:</span>                                                                          
#> <span style="color: #B58900;">!</span> `n` must be a numeric vector                                                  
#> <span style="color: #DC322F;">✖</span> You've supplied a <span style="color: #268BD2;">&lt;character&gt;</span> vector.                                         
#> <span style="color: #002B36;">Run `rlang::last_error()` to see where the error occurred.</span>                      
</pre></div>
}}


\if{html}{\out{<div class="sourceCode r">}}\preformatted{len <- 26
idx <- 100
cli_abort(c(
        "Must index an existing element:",
  "i" = "There \{?is/are\} \{len\} element\{?s\}.",
  "x" = "You've tried to subset element \{idx\}."
))
}\if{html}{\out{</div>}}\if{html}{\out{
<div class="asciicast" style="color: #172431;font-family: 'Fira Code',Monaco,Consolas,Menlo,'Bitstream Vera Sans Mono','Powerline Symbols',monospace;line-height: 1.300000"><pre>
#> <span style="font-weight: bold;color: #B58900;">Error</span><span style="font-weight: bold;">:</span>                                                                          
#> <span style="color: #B58900;">!</span> Must index an existing element:                                               
#> <span style="color: #2AA198;">ℹ</span> There are 26 elements.                                                        
#> <span style="color: #DC322F;">✖</span> You've tried to subset element 100.                                           
#> <span style="color: #002B36;">Run `rlang::last_error()` to see where the error occurred.</span>                      
</pre></div>
}}
}
\seealso{
These functions support \link[=inline-markup]{inline markup}.

Other functions supporting inline markup: 
\code{\link{cli_alert}()},
\code{\link{cli_blockquote}()},
\code{\link{cli_bullets}()},
\code{\link{cli_bullets_raw}()},
\code{\link{cli_dl}()},
\code{\link{cli_h1}()},
\code{\link{cli_li}()},
\code{\link{cli_ol}()},
\code{\link{cli_process_start}()},
\code{\link{cli_progress_along}()},
\code{\link{cli_progress_bar}()},
\code{\link{cli_progress_message}()},
\code{\link{cli_progress_output}()},
\code{\link{cli_progress_step}()},
\code{\link{cli_rule}},
\code{\link{cli_status}()},
\code{\link{cli_status_update}()},
\code{\link{cli_text}()},
\code{\link{cli_ul}()},
\code{\link{format_error}()},
\code{\link{format_inline}()}
}
\concept{functions supporting inline markup}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cli.R
\name{cli_text}
\alias{cli_text}
\title{CLI text}
\usage{
cli_text(..., .envir = parent.frame())
}
\arguments{
\item{...}{The text to show, in character vectors. They will be
concatenated into a single string. Newlines are \emph{not} preserved.}

\item{.envir}{Environment to evaluate the glue expressions in.}
}
\description{
Write some text to the screen. This function is most appropriate for
longer paragraphs. See \code{\link[=cli_alert]{cli_alert()}} for shorter status messages.
}
\details{
\subsection{Text wrapping}{

Text is wrapped to the console width, see \code{\link[=console_width]{console_width()}}.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{cli_text(cli:::lorem_ipsum())
}\if{html}{\out{</div>}}\if{html}{\out{
<div class="asciicast" style="color: #172431;font-family: 'Fira Code',Monaco,Consolas,Menlo,'Bitstream Vera Sans Mono','Powerline Symbols',monospace;line-height: 1.300000"><pre>
#> Lorem ad ipsum veniam esse nisi deserunt duis. Qui incididunt elit              
#> elit mollit sint nulla consectetur aute commodo do elit laboris minim           
#> et. Laboris ipsum mollit voluptate et non do incididunt eiusmod. Anim           
#> consectetur mollit laborum occaecat eiusmod excepteur. Ullamco non              
#> tempor esse anim tempor magna non.                                              
</pre></div>
}}

}

\subsection{New lines}{

A \code{cli_text()} call always appends a newline character to the end.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{cli_text("First line.")
cli_text("Second line.")
}\if{html}{\out{</div>}}\if{html}{\out{
<div class="asciicast" style="color: #172431;font-family: 'Fira Code',Monaco,Consolas,Menlo,'Bitstream Vera Sans Mono','Powerline Symbols',monospace;line-height: 1.300000"><pre>
#> First line.                                                                     
#> Second line.                                                                    
</pre></div>
}}

}

\subsection{Styling}{

You can use \link[=inline-markup]{inline markup}, as usual.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{cli_text("The \{.fn cli_text\} function in the \{.pkg cli\} package.")
}\if{html}{\out{</div>}}\if{html}{\out{
<div class="asciicast" style="color: #172431;font-family: 'Fira Code',Monaco,Consolas,Menlo,'Bitstream Vera Sans Mono','Powerline Symbols',monospace;line-height: 1.300000"><pre>
#> The `cli_text()` function in the <span style="color: #268BD2;">cli</span> package.                                   
</pre></div>
}}

}

\subsection{Interpolation}{

String interpolation via glue works as usual. Interpolated vectors
are collapsed.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{pos <- c(5, 14, 25, 26)
cli_text("We have \{length(pos)\} missing measurements: \{pos\}.")
}\if{html}{\out{</div>}}\if{html}{\out{
<div class="asciicast" style="color: #172431;font-family: 'Fira Code',Monaco,Consolas,Menlo,'Bitstream Vera Sans Mono','Powerline Symbols',monospace;line-height: 1.300000"><pre>
#> We have 4 missing measurements: 5, 14, 25, and 26.                              
</pre></div>
}}

}

\subsection{Styling and interpolation}{

Use double braces to combine styling and string interpolation.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{fun <- "cli-text"
pkg <- "cli"
cli_text("The \{.fn \{fun\}\} function in the \{.pkg \{pkg\}\} package.")
}\if{html}{\out{</div>}}\if{html}{\out{
<div class="asciicast" style="color: #172431;font-family: 'Fira Code',Monaco,Consolas,Menlo,'Bitstream Vera Sans Mono','Powerline Symbols',monospace;line-height: 1.300000"><pre>
#> The `cli-text()` function in the <span style="color: #268BD2;">cli</span> package.                                   
</pre></div>
}}

}

\subsection{Multiple arguments}{

Arguments are concatenated.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{cli_text(c("This ", "will ", "all "), "be ", "one ", "sentence.")
}\if{html}{\out{</div>}}\if{html}{\out{
<div class="asciicast" style="color: #172431;font-family: 'Fira Code',Monaco,Consolas,Menlo,'Bitstream Vera Sans Mono','Powerline Symbols',monospace;line-height: 1.300000"><pre>
#> This will all be one sentence.                                                  
</pre></div>
}}

}

\subsection{Containers}{

You can use \code{cli_text()} within cli \link{containers}.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{ul <- cli_ul()
cli_li("First item.")
cli_text("Still the \{.emph first\} item")
cli_li("Second item.")
cli_text("Still the \{.emph second\} item")
cli_end(ul)
}\if{html}{\out{</div>}}\if{html}{\out{
<div class="asciicast" style="color: #172431;font-family: 'Fira Code',Monaco,Consolas,Menlo,'Bitstream Vera Sans Mono','Powerline Symbols',monospace;line-height: 1.300000"><pre>
#> • First item.                                                                   
#> Still the <span style="font-style: italic;">first</span> item                                                            
#> • Second item.                                                                  
#> Still the <span style="font-style: italic;">second</span> item                                                           
</pre></div>
}}

}
}
\seealso{
This function supports \link[=inline-markup]{inline markup}.

Other functions supporting inline markup: 
\code{\link{cli_abort}()},
\code{\link{cli_alert}()},
\code{\link{cli_blockquote}()},
\code{\link{cli_bullets}()},
\code{\link{cli_bullets_raw}()},
\code{\link{cli_dl}()},
\code{\link{cli_h1}()},
\code{\link{cli_li}()},
\code{\link{cli_ol}()},
\code{\link{cli_process_start}()},
\code{\link{cli_progress_along}()},
\code{\link{cli_progress_bar}()},
\code{\link{cli_progress_message}()},
\code{\link{cli_progress_output}()},
\code{\link{cli_progress_step}()},
\code{\link{cli_rule}},
\code{\link{cli_status}()},
\code{\link{cli_status_update}()},
\code{\link{cli_ul}()},
\code{\link{format_error}()},
\code{\link{format_inline}()}
}
\concept{functions supporting inline markup}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/app.R
\name{start_app}
\alias{start_app}
\alias{stop_app}
\alias{default_app}
\title{Start, stop, query the default cli application}
\usage{
start_app(
  theme = getOption("cli.theme"),
  output = c("auto", "message", "stdout", "stderr"),
  .auto_close = TRUE,
  .envir = parent.frame()
)

stop_app(app = NULL)

default_app()
}
\arguments{
\item{theme}{Theme to use.}

\item{output}{How to print the output.}

\item{.auto_close}{Whether to stop the app, when the calling frame
is destroyed.}

\item{.envir}{The environment to use, instead of the calling frame,
to trigger the stop of the app.}

\item{app}{App to stop. If \code{NULL}, the current default app is stopped.
Otherwise we find the supplied app in the app stack, and remote it,
together with all the apps above it.}
}
\value{
\code{start_app} returns the new app, \code{default_app} returns the default app.
\code{stop_app} does not return anything.
}
\description{
\code{start_app} creates an app, and places it on the top of the app stack.
}
\details{
\code{stop_app} removes the top app, or multiple apps from the app stack.

\code{default_app} returns the default app, the one on the top of the stack.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cli.R
\name{cli_fmt}
\alias{cli_fmt}
\title{Capture the output of cli functions instead of printing it}
\usage{
cli_fmt(expr, collapse = FALSE, strip_newline = FALSE)
}
\arguments{
\item{expr}{Expression to evaluate, containing \verb{cli_*()} calls,
typically.}

\item{collapse}{Whether to collapse the output into a single character
scalar, or return a character vector with one element for each line.}

\item{strip_newline}{Whether to strip the trailing newline.}
}
\description{
Capture the output of cli functions instead of printing it
}
\examples{
cli_fmt({
  cli_alert_info("Loading data file")
  cli_alert_success("Loaded data file")
})
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/progress-client.R
\name{cli_progress_output}
\alias{cli_progress_output}
\title{Add text output to a progress bar}
\usage{
cli_progress_output(text, id = NULL, .envir = parent.frame())
}
\arguments{
\item{text}{Text to output. It is formatted via \code{\link[=cli_text]{cli_text()}}.}

\item{id}{Progress bar id. The default is the current progress bar.}

\item{.envir}{Environment to use for glue interpolation of \code{text}.}
}
\value{
\code{TRUE}, always.
}
\description{
The text is calculated via \code{\link[=cli_text]{cli_text()}}, so all cli features can be
used here, including progress variables.
}
\details{
The text is passed to the progress handler(s), that may or may not be
able to print it.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{fun <- function() \{
  cli_alert_info("Before the progress bar")
  cli_progress_bar("Calculating", total = 100)
  for (i in 1:50) \{
    Sys.sleep(4/100)
    cli_progress_update()
  \}
  cli_progress_output("Already half way!")
  for (i in 1:50) \{
    Sys.sleep(4/100)
    cli_progress_update()
  \}
  cli_alert_info("All done")
\}
fun()
}\if{html}{\out{</div>}}

\if{html}{\figure{progress-output2.svg}}
}
\seealso{
This function supports \link[=inline-markup]{inline markup}.

Other progress bar functions: 
\code{\link{cli_progress_along}()},
\code{\link{cli_progress_bar}()},
\code{\link{cli_progress_builtin_handlers}()},
\code{\link{cli_progress_message}()},
\code{\link{cli_progress_num}()},
\code{\link{cli_progress_step}()},
\code{\link{cli_progress_styles}()},
\code{\link{progress-variables}}

Other functions supporting inline markup: 
\code{\link{cli_abort}()},
\code{\link{cli_alert}()},
\code{\link{cli_blockquote}()},
\code{\link{cli_bullets}()},
\code{\link{cli_bullets_raw}()},
\code{\link{cli_dl}()},
\code{\link{cli_h1}()},
\code{\link{cli_li}()},
\code{\link{cli_ol}()},
\code{\link{cli_process_start}()},
\code{\link{cli_progress_along}()},
\code{\link{cli_progress_bar}()},
\code{\link{cli_progress_message}()},
\code{\link{cli_progress_step}()},
\code{\link{cli_rule}},
\code{\link{cli_status}()},
\code{\link{cli_status_update}()},
\code{\link{cli_text}()},
\code{\link{cli_ul}()},
\code{\link{format_error}()},
\code{\link{format_inline}()}
}
\concept{functions supporting inline markup}
\concept{progress bar functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/prettycode.R
\name{code_theme_list}
\alias{code_theme_list}
\title{Syntax highlighting themes}
\usage{
code_theme_list()
}
\value{
Character vector of the built-in code theme names.
}
\description{
\code{code_theme_list()} lists the built-in code themes.
}
\section{Code themes}{
A theme is a list of character vectors, except for \code{bracket}, see below.
Each character vector must contain RGB colors (e.g. \code{"#a9a9a9"}),
and cli styles, e.g. \code{"bold"}. Entries in the list:
\itemize{
\item \code{reserved}: reserved words
\item \code{number}: numeric literals
\item \code{null}: the \code{NULL} constant
\item \code{operator}: operators, including assignment
\item \code{call}: function calls
\item \code{string}: character literals
\item \code{comment}: comments
\item \code{bracket}: brackets: \code{(){}[]} This is a list of character vectors,
to create "rainbow" brackets. It is recycled for deeply nested lists.
}
}

\section{The default code theme}{
In RStudio, it matches the current theme of the IDE.

You can use three options to customize the code theme:
\itemize{
\item If \code{cli.code_theme} is set, it is used.
\item Otherwise if R is running in RStudio and \code{cli.code_theme_rstudio} is
set, then it is used.
\item Otherwise if T is not running in RStudio and \code{cli.code_theme_terminal}
is set, then it is used.
}

You can set these options to the name of a built-in theme, or to list
that specifies a custom theme. See \code{\link[=code_theme_list]{code_theme_list()}} for the list
of the built-in themes.
}

\examples{
code_theme_list()
code_highlight(deparse(get), code_theme = "Solarized Dark")
}
\seealso{
Other syntax highlighting: 
\code{\link{code_highlight}()}
}
\concept{syntax highlighting}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/debug.R
\name{cli_debug_doc}
\alias{cli_debug_doc}
\title{Debug cli internals}
\usage{
cli_debug_doc(app = default_app() \%||\% start_app())
}
\arguments{
\item{app}{The cli app to debug. Defaults to the current app.
if there is no app, then it creates one by calling \code{\link[=start_app]{start_app()}}.}
}
\value{
Data frame with columns: \code{tag}, \code{id}, \code{class} (space separated),
theme (id of the theme the element added), \code{styles} (computed styles
for the element).
}
\description{
Return the current state of a cli app. It includes the currently
open tags, their ids, classes and their computed styles.
}
\details{
The returned data frame has a print method, and if you want to create
a plain data frame from it, index it with an empty bracket:
\code{cli_debug_doc()[]}.

To see all currently active themes, use \code{app$themes}, e.g. for the
default app: \code{default_app()$themes}.
}
\examples{
\dontrun{
cli_debug_doc()

olid <- cli_ol()
cli_li()
cli_debug_doc()
cli_debug_doc()[]

cli_end(olid)
cli_debug_doc()
}
}
\seealso{
\code{\link[=cli_sitrep]{cli_sitrep()}}. To debug containers, you can set the
\code{CLI-DEBUG_BAD_END} environment variable to \code{true}, and then cli will
warn when it cannot find the specified container to close (or any
contained at all).
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/status-bar.R
\name{cli_status_update}
\alias{cli_status_update}
\title{Update the status bar (superseded)}
\usage{
cli_status_update(
  id = NULL,
  msg = NULL,
  msg_done = NULL,
  msg_failed = NULL,
  .envir = parent.frame()
)
}
\arguments{
\item{id}{Id of the status bar to update. Defaults to the current
status bar container.}

\item{msg}{Text to update the status bar with. \code{NULL} if you don't want
to change it.}

\item{msg_done}{Updated "done" message. \code{NULL} if you don't want to
change it.}

\item{msg_failed}{Updated "failed" message. \code{NULL} if you don't want to
change it.}

\item{.envir}{Environment to evaluate the glue expressions in.}
}
\value{
Id of the status bar container.
}
\description{
\strong{The \verb{cli_status_*()} functions are superseded by
the \code{\link[=cli_progress_message]{cli_progress_message()}} and \code{\link[=cli_progress_step]{cli_progress_step()}} functions,
because they have a better default behavior.}

Update the status bar
}
\seealso{
This function supports \link[=inline-markup]{inline markup}.

The \code{\link[=cli_progress_message]{cli_progress_message()}} and \code{\link[=cli_progress_step]{cli_progress_step()}}
functions, for a superior API.

Other status bar: 
\code{\link{cli_process_start}()},
\code{\link{cli_status}()},
\code{\link{cli_status_clear}()}

Other functions supporting inline markup: 
\code{\link{cli_abort}()},
\code{\link{cli_alert}()},
\code{\link{cli_blockquote}()},
\code{\link{cli_bullets}()},
\code{\link{cli_bullets_raw}()},
\code{\link{cli_dl}()},
\code{\link{cli_h1}()},
\code{\link{cli_li}()},
\code{\link{cli_ol}()},
\code{\link{cli_process_start}()},
\code{\link{cli_progress_along}()},
\code{\link{cli_progress_bar}()},
\code{\link{cli_progress_message}()},
\code{\link{cli_progress_output}()},
\code{\link{cli_progress_step}()},
\code{\link{cli_rule}},
\code{\link{cli_status}()},
\code{\link{cli_text}()},
\code{\link{cli_ul}()},
\code{\link{format_error}()},
\code{\link{format_inline}()}
}
\concept{functions supporting inline markup}
\concept{status bar}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ansiex.R
\name{ansi_toupper}
\alias{ansi_toupper}
\alias{ansi_tolower}
\alias{ansi_chartr}
\title{ANSI character translation and case folding}
\usage{
ansi_toupper(x)

ansi_tolower(x)

ansi_chartr(old, new, x)
}
\arguments{
\item{x}{Input string. May have ANSI colors and styles.}

\item{old}{a character string specifying the characters to be
    translated.  If a character vector of length 2 or more is supplied,
    the first element is used with a warning.}

\item{new}{a character string specifying the translations. If a
    character vector of length 2 or more is supplied, the first element
    is used with a warning.}
}
\value{
Character vector of the same length as \code{x}, containing
the translated strings. ANSI styles are retained.
}
\description{
There functions are similar to \code{\link[=toupper]{toupper()}}, \code{\link[=tolower]{tolower()}} and
\code{\link[=chartr]{chartr()}}, but they keep the ANSI colors of the string.
}
\examples{
ansi_toupper(col_red("Uppercase"))

ansi_tolower(col_red("LowerCase"))

x <- paste0(col_green("MiXeD"), col_red(" cAsE 123"))
ansi_chartr("iXs", "why", x)
}
\seealso{
Other ANSI string operations: 
\code{\link{ansi_align}()},
\code{\link{ansi_columns}()},
\code{\link{ansi_nchar}()},
\code{\link{ansi_strsplit}()},
\code{\link{ansi_strtrim}()},
\code{\link{ansi_strwrap}()},
\code{\link{ansi_substr}()},
\code{\link{ansi_substring}()},
\code{\link{ansi_trimws}()}

Other ANSI string operations: 
\code{\link{ansi_align}()},
\code{\link{ansi_columns}()},
\code{\link{ansi_nchar}()},
\code{\link{ansi_strsplit}()},
\code{\link{ansi_strtrim}()},
\code{\link{ansi_strwrap}()},
\code{\link{ansi_substr}()},
\code{\link{ansi_substring}()},
\code{\link{ansi_trimws}()}

Other ANSI string operations: 
\code{\link{ansi_align}()},
\code{\link{ansi_columns}()},
\code{\link{ansi_nchar}()},
\code{\link{ansi_strsplit}()},
\code{\link{ansi_strtrim}()},
\code{\link{ansi_strwrap}()},
\code{\link{ansi_substr}()},
\code{\link{ansi_substring}()},
\code{\link{ansi_trimws}()}
}
\concept{ANSI string operations}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utf8.R
\name{utf8_nchar}
\alias{utf8_nchar}
\title{Count the number of characters in a character vector}
\usage{
utf8_nchar(x, type = c("chars", "bytes", "width", "graphemes", "codepoints"))
}
\arguments{
\item{x}{Character vector, it is converted to UTF-8.}

\item{type}{Whether to count graphemes (characters), code points,
bytes, or calculate the display width of the string.}
}
\value{
Numeric vector, the length of the strings in the character
vector.
}
\description{
By default it counts Unicode grapheme clusters, instead of code points.
}
\examples{
# Grapheme example, emoji with combining characters. This is a single
# grapheme, consisting of five Unicode code points:
# * `\U0001f477` is the construction worker emoji
# * `\U0001f3fb` is emoji modifier that changes the skin color
# * `\u200d` is the zero width joiner
# * `\u2640` is the female sign
# * `\ufe0f` is variation selector 16, requesting an emoji style glyph
emo <- "\U0001f477\U0001f3fb\u200d\u2640\ufe0f"
cat(emo)

utf8_nchar(emo, "chars") # = graphemes
utf8_nchar(emo, "bytes")
utf8_nchar(emo, "width")
utf8_nchar(emo, "codepoints")

# For comparision, the output for width depends on the R version used:
nchar(emo, "chars")
nchar(emo, "bytes")
nchar(emo, "width")
}
\seealso{
Other UTF-8 string manipulation: 
\code{\link{utf8_graphemes}()},
\code{\link{utf8_substr}()}
}
\concept{UTF-8 string manipulation}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ansiex.R
\name{ansi_nzchar}
\alias{ansi_nzchar}
\title{Like \code{\link[base:nchar]{base::nzchar()}}, but for ANSI strings}
\usage{
ansi_nzchar(x, ...)
}
\arguments{
\item{x}{Charcater vector. Other objects are coarced using
\code{\link[base:character]{base::as.character()}}.}

\item{...}{Passed to \code{\link[base:nchar]{base::nzchar()}}.}
}
\description{
Like \code{\link[base:nchar]{base::nzchar()}}, but for ANSI strings
}
\examples{
ansi_nzchar("")
ansi_nzchar(col_red(""))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cli.R
\name{cli_ol}
\alias{cli_ol}
\title{Ordered CLI list}
\usage{
cli_ol(
  items = NULL,
  id = NULL,
  class = NULL,
  .close = TRUE,
  .auto_close = TRUE,
  .envir = parent.frame()
)
}
\arguments{
\item{items}{If not \code{NULL}, then a character vector. Each element of
the vector will be one list item, and the list container will be
closed by default (see the \code{.close} argument).}

\item{id}{Id of the list container. Can be used for closing it with
\code{\link[=cli_end]{cli_end()}} or in themes. If \code{NULL}, then an id is generated and
returned invisibly.}

\item{class}{Class of the list container. Can be used in themes.}

\item{.close}{Whether to close the list container if the \code{items} were
specified. If \code{FALSE} then new items can be added to the list.}

\item{.auto_close}{Whether to close the container, when the calling
function finishes (or \code{.envir} is removed, if specified).}

\item{.envir}{Environment to evaluate the glue expressions in. It is
also used to auto-close the container if \code{.auto_close} is \code{TRUE}.}
}
\value{
The id of the new container element, invisibly.
}
\description{
An ordered list is a container, see \link{containers}.
}
\details{
\subsection{Adding all items at once}{

\if{html}{\out{<div class="sourceCode r">}}\preformatted{fun <- function() \{
  cli_ol(c("one", "two", "three"))
\}
fun()
}\if{html}{\out{</div>}}\if{html}{\out{
<div class="asciicast" style="color: #172431;font-family: 'Fira Code',Monaco,Consolas,Menlo,'Bitstream Vera Sans Mono','Powerline Symbols',monospace;line-height: 1.300000"><pre>
#> 1. one                                                                          
#> 2. two                                                                          
#> 3. three                                                                        
</pre></div>
}}

}

\subsection{Adding items one by one}{

\if{html}{\out{<div class="sourceCode r">}}\preformatted{## Adding items one by one
fun <- function() \{
  cli_ol()
  cli_li("\{.emph one\}")
  cli_li("\{.emph two\}")
  cli_li("\{.emph three\}")
  cli_end()
\}
fun()
}\if{html}{\out{</div>}}\if{html}{\out{
<div class="asciicast" style="color: #172431;font-family: 'Fira Code',Monaco,Consolas,Menlo,'Bitstream Vera Sans Mono','Powerline Symbols',monospace;line-height: 1.300000"><pre>
#> 1. <span style="font-style: italic;">one</span>                                                                          
#> 2. <span style="font-style: italic;">two</span>                                                                          
#> 3. <span style="font-style: italic;">three</span>                                                                        
</pre></div>
}}

}

\subsection{Nested lists}{

\if{html}{\out{<div class="sourceCode r">}}\preformatted{fun <- function() \{
  cli_div(theme = list(ol = list("margin-left" = 2)))
  cli_ul()
  cli_li("one")
  cli_ol(c("foo", "bar", "foobar"))
  cli_li("two")
  cli_end()
  cli_end()
\}
fun()
}\if{html}{\out{</div>}}\if{html}{\out{
<div class="asciicast" style="color: #172431;font-family: 'Fira Code',Monaco,Consolas,Menlo,'Bitstream Vera Sans Mono','Powerline Symbols',monospace;line-height: 1.300000"><pre>
#> • one                                                                           
#>     1. foo                                                                      
#>     2. bar                                                                      
#>     3. foobar                                                                   
#> • two                                                                           
</pre></div>
}}

}
}
\seealso{
This function supports \link[=inline-markup]{inline markup}.

Other functions supporting inline markup: 
\code{\link{cli_abort}()},
\code{\link{cli_alert}()},
\code{\link{cli_blockquote}()},
\code{\link{cli_bullets}()},
\code{\link{cli_bullets_raw}()},
\code{\link{cli_dl}()},
\code{\link{cli_h1}()},
\code{\link{cli_li}()},
\code{\link{cli_process_start}()},
\code{\link{cli_progress_along}()},
\code{\link{cli_progress_bar}()},
\code{\link{cli_progress_message}()},
\code{\link{cli_progress_output}()},
\code{\link{cli_progress_step}()},
\code{\link{cli_rule}},
\code{\link{cli_status}()},
\code{\link{cli_status_update}()},
\code{\link{cli_text}()},
\code{\link{cli_ul}()},
\code{\link{format_error}()},
\code{\link{format_inline}()}
}
\concept{functions supporting inline markup}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/progress-server.R
\name{cli_progress_builtin_handlers}
\alias{cli_progress_builtin_handlers}
\title{cli progress handlers}
\usage{
cli_progress_builtin_handlers()
}
\value{
\code{cli_progress_builtin_handlers()} returns the names of the
currently supported progress handlers.
}
\description{
The progress handler(s) to use can be selected with global options.
}
\details{
There are three options that specify which handlers will be selected,
but most of the time you only need to use one of them. You can set these
options to a character vector, the names of the built-in cli handlers you
want to use:
\itemize{
\item If \code{cli.progress_handlers_only} is set, then these handlers are used,
without considering others and without checking if they are able to
handle a progress bar. This option is mainly intended for testing
purposes.
\item The handlers named in \code{cli.progress_handlers} are checked if they are
able to handle the progress bar, and from the ones that are, the first
one is selected. This is usually the option that the end use would want
to set.
\item The handlers named in \code{cli.progress_handlers_force} are always appended
to the ones selected via \code{cli.progress_handlers}. This option is useful
to add an additional handler, e.g. a logger that writes to a file.
}
}
\section{The built-in progress handlers}{
\subsection{\code{cli}}{

Use cli's internal status bar, the last line of the screen, to show the
progress bar. This handler is always able to handle all progress bars.
}

\subsection{\code{logger}}{

Log progress updates to the screen, with one line for each update and with
time stamps. This handler is always able to handle all progress bars.
}

\subsection{\code{progressr}}{

Use the progressr package to create
progress bars. This handler is always able to handle all progress bars.
(The progressr package needs to be installed.)
}

\subsection{\code{rstudio}}{

Use \href{https://posit.co/blog/rstudio-1-2-jobs/}{RStudio's job panel}
to show the progress bars. This handler is available at the RStudio console,
in recent versions of RStudio.
}

\subsection{\code{say}}{

Use the macOS \code{say} command to announce progress events in speech (type
\verb{man say} on a terminal for more info). Set the \code{cli.progress_say_frequency}
option to set the minimum delay between \code{say} invocations, the default is
three seconds. This handler is available on macOS, if the \code{say} command is
on the path.

The external command and its arguments can be configured with options:
\itemize{
\item \code{cli_progress_say_args}: command line arguments, e.g. you can use this
to select a voice on macOS,
\item \code{cli_progress_say_command}: external command to run,
\item \code{cli_progress_say_frequency}: wait at least this many seconds between
calling the external command.
}
}

\subsection{\code{shiny}}{

Use \href{https://shiny.rstudio.com/articles/progress.html}{shiny's progress bars}.
This handler is available if a shiny app is running.
}
}

\seealso{
Other progress bar functions: 
\code{\link{cli_progress_along}()},
\code{\link{cli_progress_bar}()},
\code{\link{cli_progress_message}()},
\code{\link{cli_progress_num}()},
\code{\link{cli_progress_output}()},
\code{\link{cli_progress_step}()},
\code{\link{cli_progress_styles}()},
\code{\link{progress-variables}}
}
\concept{progress bar functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utf8.R
\name{utf8_substr}
\alias{utf8_substr}
\title{Substring of an UTF-8 string}
\usage{
utf8_substr(x, start, stop)
}
\arguments{
\item{x}{Character vector.}

\item{start}{Starting index or indices, recycled to match the length
of \code{x}.}

\item{stop}{Ending index or indices, recycled to match the length of
\code{x}.}
}
\value{
Character vector of the same length as \code{x}, containing
the requested substrings.
}
\description{
This function uses grapheme clusters instead of Unicode code points in
UTF-8 strings.
}
\examples{
# Five grapheme clusters, select the middle three
str <- paste0(
  "\U0001f477\U0001f3ff\u200d\u2640\ufe0f",
  "\U0001f477\U0001f3ff",
  "\U0001f477\u200d\u2640\ufe0f",
  "\U0001f477\U0001f3fb",
  "\U0001f477\U0001f3ff")
cat(str)
str24 <- utf8_substr(str, 2, 4)
cat(str24)
}
\seealso{
Other UTF-8 string manipulation: 
\code{\link{utf8_graphemes}()},
\code{\link{utf8_nchar}()}
}
\concept{UTF-8 string manipulation}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/themes.R
\name{match_selector_node}
\alias{match_selector_node}
\title{Match a selector node to a container}
\usage{
match_selector_node(node, cnt)
}
\arguments{
\item{node}{Selector node, as parsed by \code{parse_selector_node()}.}

\item{cnt}{Container node, has elements \code{tag}, \code{id}, \code{class}.

The selector node matches the container, if all these hold:
\itemize{
\item The id of the selector is missing or unique.
\item The tag of the selector is missing or unique.
\item The id of the container is missing or unique.
\item The tag of the container is unique.
\item If the selector specifies an id, it matches the id of the container.
\item If the selector specifies a tag, it matches the tag of the container.
\item If the selector specifies class names, the container has all these
classes.
}}
}
\description{
Match a selector node to a container
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cli.R
\name{cli}
\alias{cli}
\title{Compose multiple cli functions}
\usage{
cli(expr)
}
\arguments{
\item{expr}{Expression that contains \verb{cli_*} calls. Their output is
collected and sent as a single message.}
}
\value{
Nothing.
}
\description{
\code{cli()} will record all \verb{cli_*} calls in \code{expr}, and emit them together
in a single message. This is useful if you want to built a larger
piece of output from multiple \verb{cli_*} calls.
}
\details{
Use this function to build a more complex piece of CLI that would not
make sense to show in pieces.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{cli(\{
  cli_h1("Title")
  cli_h2("Subtitle")
  cli_ul(c("this", "that", "end"))
\})
}\if{html}{\out{</div>}}\if{html}{\out{
<div class="asciicast" style="color: #172431;font-family: 'Fira Code',Monaco,Consolas,Menlo,'Bitstream Vera Sans Mono','Powerline Symbols',monospace;line-height: 1.300000"><pre>
#>                                                                                 
#> <span style="color: #2AA198;">──</span> <span style="font-weight: bold;">Title</span> <span style="color: #2AA198;">─────────────────────────────────────────────────────────────</span>          
#>                                                                                 
#> ── <span style="font-weight: bold;">Subtitle</span> ──                                                                  
#>                                                                                 
#> • this                                                                          
#> • that                                                                          
#> • end                                                                           
</pre></div>
}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/docs.R
\name{progress-c}
\alias{progress-c}
\title{The cli progress C API}
\description{
The cli progress C API
}
\section{The cli progress C API}{

\subsection{\code{CLI_SHOULD_TICK}}{

A macro that evaluates to (int) 1 if a cli progress bar update is due,
and to (int) 0 otherwise. If the timer hasn't been initialized in this
compilation unit yet, then it is always 0. To initialize the timer,
call \code{cli_progress_init_timer()} or create a progress bar with
\code{cli_progress_bar()}.
}

\subsection{\code{cli_progress_add()}}{

\if{html}{\out{<div class="sourceCode c">}}\preformatted{void cli_progress_add(SEXP bar, double inc);
}\if{html}{\out{</div>}}

Add a number of progress units to the progress bar. It will also
trigger an update if an update is due.
\itemize{
\item \code{bar}: progress bar object.
\item \code{inc}: progress increment.
}
}

\subsection{\code{cli_progress_bar()}}{

\if{html}{\out{<div class="sourceCode c">}}\preformatted{SEXP cli_progress_bar(double total, SEXP config);
}\if{html}{\out{</div>}}

Create a new progress bar object. The returned progress bar object
must be \code{PROTECT()}-ed.
\itemize{
\item \code{total}: Total number of progress units. Use \code{NA_REAL} if it is not
known.
\item \code{config}: R named list object of additional parameters. May be \code{NULL}
(the C \verb{NULL~) or }R_NilValue\verb{(the R}NULL`) for the defaults.
}

\code{config} may contain the following entries:
\itemize{
\item \code{name}: progress bar name.
\item \code{status}: (initial) progress bar status.
\item \code{type}: progress bar type.
\item \code{total}: total number of progress units.
\item \code{show_after}: show the progress bar after the specified number of
seconds. This overrides the global \code{show_after} option.
\item \code{format}: format string, must be specified for custom progress bars.
\item \code{format_done}: format string for successful termination.
\item \code{format_failed}: format string for unsuccessful termination.
\item \code{clear}: whether to remove the progress bar from the screen after
termination.
\item \code{auto_terminate}: whether to terminate the progress bar when the
number of current units equals the number of total progress units.
}
\subsection{Example}{

\if{html}{\out{<div class="sourceCode c">}}\preformatted{#include <cli/progress.h>
SEXP progress_test1() \{
  int i;
  SEXP bar = PROTECT(cli_progress_bar(1000, NULL));
  for (i = 0; i < 1000; i++) \{
    cli_progress_sleep(0, 4 * 1000 * 1000);
    if (CLI_SHOULD_TICK) cli_progress_set(bar, i);
  \}
  cli_progress_done(bar);
  UNPROTECT(1);
  return Rf_ScalarInteger(i);
\}
}\if{html}{\out{</div>}}
}

}

\subsection{\code{cli_progress_done()}}{

\if{html}{\out{<div class="sourceCode c">}}\preformatted{void cli_progress_done(SEXP bar);
}\if{html}{\out{</div>}}

Terminate the progress bar.
\itemize{
\item \code{bar}: progress bar object.
}
}

\subsection{\code{cli_progress_init_timer()}}{

\if{html}{\out{<div class="sourceCode c">}}\preformatted{void cli_progress_init_timer();
}\if{html}{\out{</div>}}

Initialize the cli timer without creating a progress bar.
}

\subsection{\code{cli_progress_num()}}{

\if{html}{\out{<div class="sourceCode c">}}\preformatted{int cli_progress_num();
}\if{html}{\out{</div>}}

Returns the number of currently active progress bars.
}

\subsection{\code{cli_progress_set()}}{

\if{html}{\out{<div class="sourceCode c">}}\preformatted{void cli_progress_set(SEXP bar, double set);
}\if{html}{\out{</div>}}

Set the progress bar to the specified number of progress units.
\itemize{
\item \code{bar}: progress bar object.
\item \code{set}: number of current progress progress units.
}
}

\subsection{\code{cli_progress_set_clear()}}{

\if{html}{\out{<div class="sourceCode c">}}\preformatted{void cli_progress_set_clear(SEXP bar, int clear);
}\if{html}{\out{</div>}}

Set whether to remove the progress bar from the screen. You can call
this any time before \code{cli_progress_done()} is called.
\itemize{
\item \code{bar}: progress bar object.
\item \code{clear}: whether to remove the progress bar from the screen, zero or
one.
}
}

\subsection{\code{cli_progress_set_format()}}{

\if{html}{\out{<div class="sourceCode c">}}\preformatted{void cli_progress_set_format(SEXP bar, const char *format, ...);
}\if{html}{\out{</div>}}

Set a custom format string for the progress bar. This call does not
try to update the progress bar. If you want to request an update,
call \code{cli_progress_add()}, \code{cli_progress_set()} or
\code{cli_progress_update()}.
\itemize{
\item \code{bar}: progress bar object.
\item \code{format}: format string.
\item \code{...}: values to substitute into \code{format}.
}

\code{format} and \code{...} are passed to \code{vsnprintf()} to create a format
string.

Format strings may contain glue substitutions, referring to
\href{https://cli.r-lib.org/dev/reference/progress-variables.html}{progress variables}, pluralization, and cli
styling.
}

\subsection{\code{cli_progress_set_name()}}{

\if{html}{\out{<div class="sourceCode c">}}\preformatted{void cli_progress_set_name(SEXP bar, const char *name);
}\if{html}{\out{</div>}}

Set the name of the progress bar.
\itemize{
\item \code{bar}; progress bar object.
\item \code{name}: progress bar name.
}
}

\subsection{\code{cli_progress_set_status()}}{

\if{html}{\out{<div class="sourceCode c">}}\preformatted{void cli_progress_set_status(SEXP bar, const char *status);
}\if{html}{\out{</div>}}

Set the status of the progress bar.
\itemize{
\item \code{bar}: progress bar object.
\item \code{status }: progress bar status.
}
}

\subsection{\code{cli_progress_set_type()}}{

\if{html}{\out{<div class="sourceCode c">}}\preformatted{void cli_progress_set_type(SEXP bar, const char *type);
}\if{html}{\out{</div>}}

Set the progress bar type. Call this function right after creating
the progress bar with \code{cli_progress_bar()}. Otherwise the behavior is
undefined.
\itemize{
\item \code{bar}: progress bar object.
\item \code{type}: progress bar type. Possible progress bar types:
\code{iterator}, \code{tasks}, \code{download} and \code{custom}.
}
}

\subsection{\code{cli_progress_update()}}{

\if{html}{\out{<div class="sourceCode c">}}\preformatted{void cli_progress_update(SEXP bar, double set, double inc, int force);
}\if{html}{\out{</div>}}

Update the progress bar. Unlike the simpler \code{cli_progress_add()} and
\code{cli_progress_set()} function, it can force an update if \code{force} is
set to 1.
\itemize{
\item \code{bar}: progress bar object.
\item \code{set}: the number of current progress units. It is ignored if
negative.
\item \code{inc}: increment to add to the current number of progress units.
It is ignored if \code{set} is not negative.
\item \code{force}: whether to force an update, even if no update is due.
}

To force an update without changing the current number of progress units,
supply \code{set = -1}, \code{inc = 0} and \code{force = 1}.
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ansiex.R
\name{ansi_substring}
\alias{ansi_substring}
\title{Substring(s) of an ANSI colored string}
\usage{
ansi_substring(text, first, last = 1000000L)
}
\arguments{
\item{text}{Character vector, potentially ANSI styled, or a vector to
coerced to character. It is recycled to the longest of \code{first}
and \code{last}.}

\item{first}{Starting index or indices, recycled to match the length
of \code{x}.}

\item{last}{Ending index or indices, recycled to match the length
of \code{x}.}
}
\value{
Character vector of the same length as \code{x}, containing
the requested substrings. ANSI styles are retained.
}
\description{
This is the color-aware counterpart of \code{\link[base:substr]{base::substring()}}.
It works exactly like the original, but keeps the colors in the
substrings. The ANSI escape sequences are ignored when
calculating the positions within the string.
}
\examples{
str <- paste(
  col_red("red"),
  "default",
  col_green("green")
)

cat(str, "\n")
cat(ansi_substring(str, 1, 5), "\n")
cat(ansi_substring(str, 1, 15), "\n")
cat(ansi_substring(str, 3, 7), "\n")

substring(ansi_strip(str), 1, 5)
substring(ansi_strip(str), 1, 15)
substring(ansi_strip(str), 3, 7)

str2 <- paste(
  "another",
  col_red("multi-", style_underline("style")),
  "text"
)

cat(str2, "\n")
cat(ansi_substring(str2, c(3,5), c(7, 18)), sep = "\n")
substring(ansi_strip(str2), c(3,5), c(7, 18))
}
\seealso{
Other ANSI string operations: 
\code{\link{ansi_align}()},
\code{\link{ansi_columns}()},
\code{\link{ansi_nchar}()},
\code{\link{ansi_strsplit}()},
\code{\link{ansi_strtrim}()},
\code{\link{ansi_strwrap}()},
\code{\link{ansi_substr}()},
\code{\link{ansi_toupper}()},
\code{\link{ansi_trimws}()}
}
\concept{ANSI string operations}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hash.R
\name{hash_animal}
\alias{hash_animal}
\alias{hash_raw_animal}
\alias{hash_obj_animal}
\title{Adjective-animal hash}
\usage{
hash_animal(x, n_adj = 2)

hash_raw_animal(x, n_adj = 2)

hash_obj_animal(x, n_adj = 2, serialize_version = 2)
}
\arguments{
\item{x}{Character vector. \code{NA} entries will have an \code{NA} hash.}

\item{n_adj}{Number of adjectives to use. It must be from 0 through 3.}

\item{serialize_version}{Workspace format version to use, see
\code{\link[base:serialize]{base::serialize()}}.}
}
\value{
A data frame with columns
\itemize{
\item \code{hash}: the hash value, a string.
\item \code{words}: list column with the adjectives and the animal name in a
character vector.
}

\code{hash_raw_animal()} and \code{hash_obj_animal()} return a list
with entries:
\itemize{
\item \code{hash}: the hash value, a string,
\item `words: the adjectives and the animal name in a character vector.
}
}
\description{
Adjective-animal hash
}
\details{
It uses the first 13 hexadecimal characters (out of the 32) of the MD5
hash of the input, and converts them into an adjective-animal form to
create a human readable hash.
\subsection{Number of possible hash values}{

\code{hash_animal()} uses 1748 animal names and
8946 different adjectives. The number of
different hashes you can get for different values of \code{n_adj}:\tabular{rr}{
   \code{n_adj} \tab size of the hash table space \cr
   0 \tab 1,748 \cr
   1 \tab 15,637,608 \cr
   2 \tab 139,894,041,168 \cr
   3 \tab 1,251,492,092,288,928 \cr
}

}

\subsection{Source}{

The list of adjectives and animals comes from the ids package,
and in turn from
\url{https://github.com/a-type/adjective-adjective-animal}, and
from \verb{https://gfycat.com} (now gone).
}

\code{hash_raw_animal()} calculates the adjective-animal hash of
the bytes of a raw vector.

\code{hash_obj_animal()} calculates the adjective-animal hash of
an R object. The object is serialized into a binary vector first.
}
\examples{
hash_animal(c("foo", "bar"))

# if you increase `n_adj`, the shorter hash is a suffix of the longer:
hash_animal("cli package", 0)$hash
hash_animal("cli package", 1)$hash
hash_animal("cli package", 2)$hash
hash_animal("cli package", 3)$hash
}
\seealso{
the ids package for generating random adjective-animal ids

Other hash functions: 
\code{\link{hash_emoji}()},
\code{\link{hash_md5}()},
\code{\link{hash_sha1}()},
\code{\link{hash_sha256}()},
\code{\link{hash_xxhash}()}
}
\concept{hash functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bullets.R
\name{cli_bullets}
\alias{cli_bullets}
\title{List of items}
\usage{
cli_bullets(text, id = NULL, class = NULL, .envir = parent.frame())
}
\arguments{
\item{text}{Character vector of items. See details below on how names
are interpreted.}

\item{id}{Optional id of the \code{div.bullets} element, can be used in themes.}

\item{class}{Optional additional class(es) for the \code{div.bullets} element.}

\item{.envir}{Environment to evaluate the glue expressions in.}
}
\description{
It is often useful to print out a list of items, tasks a function or
package performs, or a list of notes.
}
\details{
Items may be formatted differently, e.g. they can have a prefix symbol.
Formatting is specified by the names of \code{text}, and can be themed.
cli creates a \code{div} element of class \code{bullets} for the whole bullet list.
Each item is another \code{div} element of class \verb{bullet-<name>}, where
\verb{<name>} is the name of the entry in \code{text}. Entries in \code{text} without
a name create a \code{div} element of class \code{bullet-empty}, and if the
name is a single space character, the class is \code{bullet-space}.

The built-in theme defines the following item types:
\itemize{
\item No name: Item without a prefix.
\item \verb{ }: Indented item.
\item \code{*}: Item with a bullet.
\item \code{>}: Item with an arrow or pointer.
\item \code{v}: Item with a green "tick" symbol, like \code{\link[=cli_alert_success]{cli_alert_success()}}.
\item \code{x}: Item with a ref cross, like \code{\link[=cli_alert_danger]{cli_alert_danger()}}.
\item \code{!}: Item with a yellow exclamation mark, like \code{\link[=cli_alert_warning]{cli_alert_warning()}}.
\item \code{i}: Info item, like \code{\link[=cli_alert_info]{cli_alert_info()}}.
}

You can define new item type by simply defining theming for the
corresponding \verb{bullet-<name>} classes.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{cli_bullets(c(
        "noindent",
  " " = "indent",
  "*" = "bullet",
  ">" = "arrow",
  "v" = "success",
  "x" = "danger",
  "!" = "warning",
  "i" = "info"
))
}\if{html}{\out{</div>}}\if{html}{\out{
<div class="asciicast" style="color: #172431;font-family: 'Fira Code',Monaco,Consolas,Menlo,'Bitstream Vera Sans Mono','Powerline Symbols',monospace;line-height: 1.300000"><pre>
#> noindent                                                                        
#>   indent                                                                        
#> <span style="color: #2AA198;">•</span> bullet                                                                        
#> → arrow                                                                         
#> <span style="color: #859900;">✔</span> success                                                                       
#> <span style="color: #DC322F;">✖</span> danger                                                                        
#> <span style="color: #B58900;">!</span> warning                                                                       
#> <span style="color: #2AA198;">ℹ</span> info                                                                          
</pre></div>
}}
}
\seealso{
This function supports \link[=inline-markup]{inline markup}.

Other functions supporting inline markup: 
\code{\link{cli_abort}()},
\code{\link{cli_alert}()},
\code{\link{cli_blockquote}()},
\code{\link{cli_bullets_raw}()},
\code{\link{cli_dl}()},
\code{\link{cli_h1}()},
\code{\link{cli_li}()},
\code{\link{cli_ol}()},
\code{\link{cli_process_start}()},
\code{\link{cli_progress_along}()},
\code{\link{cli_progress_bar}()},
\code{\link{cli_progress_message}()},
\code{\link{cli_progress_output}()},
\code{\link{cli_progress_step}()},
\code{\link{cli_rule}},
\code{\link{cli_status}()},
\code{\link{cli_status_update}()},
\code{\link{cli_text}()},
\code{\link{cli_ul}()},
\code{\link{format_error}()},
\code{\link{format_inline}()}
}
\concept{functions supporting inline markup}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/simple-theme.R
\name{simple_theme}
\alias{simple_theme}
\title{A simple CLI theme}
\usage{
simple_theme(dark = getOption("cli.theme_dark", "auto"))
}
\arguments{
\item{dark}{Whether the theme should be optimized for a dark
background. If \code{"auto"}, then cli will try to detect this.
Detection usually works in recent RStudio versions, and in iTerm
on macOS, but not on other platforms.}
}
\description{
To use this theme, you can set it as the \code{cli.theme} option.
Note that this is in addition to the builtin theme, which is still in
effect.
}
\details{
\if{html}{\out{<div class="sourceCode r">}}\preformatted{options(cli.theme = cli::simple_theme())
}\if{html}{\out{</div>}}

and then CLI apps started after this will use it as the default theme.
You can also use it temporarily, in a div element:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{cli_div(theme = cli::simple_theme())
}\if{html}{\out{</div>}}
}
\section{Showcase}{
\if{html}{\out{<div class="sourceCode r">}}\preformatted{show <- cli_div(theme = cli::simple_theme())

cli_h1("Heading 1")
cli_h2("Heading 2")
cli_h3("Heading 3")

cli_par()
cli_alert_danger("Danger alert")
cli_alert_warning("Warning alert")
cli_alert_info("Info alert")
cli_alert_success("Success alert")
cli_alert("Alert for starting a process or computation",
  class = "alert-start")
cli_end()

cli_text("Packages and versions: \{.pkg cli\} \{.version 1.0.0\}.")
cli_text("Time intervals: \{.timestamp 3.4s\}")

cli_text("\{.emph Emphasis\} and  \{.strong strong emphasis\}")

cli_text("This is a piece of code: \{.code sum(x) / length(x)\}")
cli_text("Function names: \{.fn cli::simple_theme\}")

cli_text("Files: \{.file /usr/bin/env\}")
cli_text("URLs: \{.url https://r-project.org\}")

cli_h2("Longer code chunk")
cli_par(class = "code R")
cli_verbatim(
  '# window functions are useful for grouped mutates',
  'mtcars \%>\%',
  '  group_by(cyl) \%>\%',
  '  mutate(rank = min_rank(desc(mpg)))')

cli_end(show)
}\if{html}{\out{</div>}}\if{html}{\out{
<div class="asciicast" style="color: #172431;font-family: 'Fira Code',Monaco,Consolas,Menlo,'Bitstream Vera Sans Mono','Powerline Symbols',monospace;line-height: 1.300000"><pre>
#>                                                                                 
#> <span style="color: #2AA198;">──</span> <span style="font-weight: bold;color: #2AA198;">Heading 1</span> <span style="color: #2AA198;">─────────────────────────────────────────────────────────</span>          
#>                                                                                 
#> ─ <span style="font-weight: bold;color: #2AA198;">Heading 2</span> ──                                                                  
#>                                                                                 
#> <span style="color: #2AA198;">Heading 3</span>                                                                       
#> <span style="color: #EEE8D5;background-color: #DC322F;">✖ Danger alert</span>                                                                  
#> <span style="font-weight: bold;color: #ff9900;">! Warning alert</span>                                                                 
#> <span style="color: #2AA198;">ℹ</span> Info alert                                                                    
#> <span style="color: #859900;">✔</span> Success alert                                                                 
#> → Alert for starting a process or computation                                   
#>                                                                                 
#> Packages and versions: <span style="font-weight: bold;color: #268BD2;">cli</span> <span style="color: #268BD2;">1.0.0</span>.                                               
#> Time intervals: <span style="color: #b8b8b8;">[3.4s]</span>                                                          
#> <span style="font-style: italic;">Emphasis</span> and <span style="font-weight: bold;font-style: italic;">strong emphasis</span>                                                    
#> This is a piece of code: `sum(x) / length(x)`                                   
#> Function names: <span style="color: #292929;background-color: #f5f5f5;">`cli::simple_theme()`()</span>                                         
#> Files: <span style="color: #268BD2;">/usr/bin/env</span>                                                             
#> URLs: <span style="font-style: italic;color: #268BD2;">&lt;https://r-project.org&gt;</span>                                                   
#>                                                                                 
#> ─ <span style="font-weight: bold;color: #2AA198;">Longer code chunk</span> ──                                                          
#> <span style="font-style: italic;color: #a3a3a3;"># window functions are useful for grouped mutates</span>                               
#> mtcars <span style="color: #859900;">&#37;&gt;&#37;</span>                                                                      
#>   <span style="color: #2AA198;">group_by</span><span style="color: #B58900;">(</span>cyl<span style="color: #B58900;">)</span> <span style="color: #859900;">&#37;&gt;&#37;</span>                                                             
#>   <span style="color: #2AA198;">mutate</span><span style="color: #B58900;">(</span>rank = <span style="color: #2AA198;">min_rank</span><span style="color: #268BD2;">(</span><span style="color: #2AA198;">desc(</span>mpg<span style="color: #2AA198;">)</span><span style="color: #268BD2;">)</span><span style="color: #B58900;">)</span>                                            
#>                                                                                 
</pre></div>
}}
}

\seealso{
\link{themes}, \code{\link[=builtin_theme]{builtin_theme()}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/themes.R
\name{match_selector}
\alias{match_selector}
\title{Match a selector to a container stack}
\usage{
match_selector(sels, cnts)
}
\arguments{
\item{sels}{A list of selector nodes.}

\item{cnts}{A list of container nodes.

The last selector in the list must match the last container, so we
do the matching from the back. This is because we use this function
to calculate the style of newly encountered containers.}
}
\description{
Match a selector to a container stack
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ansiex.R
\name{ansi_regex}
\alias{ansi_regex}
\title{Perl compatible regular expression that matches ANSI escape
sequences}
\usage{
ansi_regex()
}
\value{
String scalar, the regular expression.
}
\description{
Don't forget to use \code{perl = TRUE} when using this with \code{\link[=grepl]{grepl()}} and
friends.
}
\seealso{
Other low level ANSI functions: 
\code{\link{ansi_has_any}()},
\code{\link{ansi_hide_cursor}()},
\code{\link{ansi_strip}()}
}
\concept{low level ANSI functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cliapp-docs.R
\name{inline-markup}
\alias{inline-markup}
\title{About inline markup in the semantic cli}
\description{
To learn how to use cli’s semantic markup, start with the ‘Building
a semantic CLI’ article at \url{https://cli.r-lib.org}.
}
\section{Command substitution}{


All text emitted by cli supports glue interpolation. Expressions
enclosed by braces will be evaluated as R code. See \code{\link[glue:glue]{glue::glue()}} for
details.

In addition to regular glue interpolation, cli can also add classes
to parts of the text, and these classes can be used in themes. For
example

\if{html}{\out{<div class="sourceCode r">}}\preformatted{cli_text("This is \{.emph important\}.")
}\if{html}{\out{</div>}}\if{html}{\out{
<div class="asciicast" style="color: #172431;font-family: 'Fira Code',Monaco,Consolas,Menlo,'Bitstream Vera Sans Mono','Powerline Symbols',monospace;line-height: 1.300000"><pre>
#> This is <span style="font-style: italic;">important</span>.                                                              
</pre></div>
}}


adds a class to the "important" word, class \code{"emph"}. Note that in this
case the string within the braces is usually not a valid R expression.
If you want to mix classes with interpolation, add another pair of
braces:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{adjective <- "great"
cli_text("This is \{.emph \{adjective\}\}.")
}\if{html}{\out{</div>}}\if{html}{\out{
<div class="asciicast" style="color: #172431;font-family: 'Fira Code',Monaco,Consolas,Menlo,'Bitstream Vera Sans Mono','Powerline Symbols',monospace;line-height: 1.300000"><pre>
#> This is <span style="font-style: italic;">great</span>.                                                                  
</pre></div>
}}


An inline class will always create a \code{span} element internally. So in
themes, you can use the \code{span.emph} CSS selector to change how inline
text is emphasized:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{cli_div(theme = list(span.emph = list(color = "red")))
adjective <- "nice and red"
cli_text("This is \{.emph \{adjective\}\}.")
}\if{html}{\out{</div>}}\if{html}{\out{
<div class="asciicast" style="color: #172431;font-family: 'Fira Code',Monaco,Consolas,Menlo,'Bitstream Vera Sans Mono','Powerline Symbols',monospace;line-height: 1.300000"><pre>
#> This is <span style="font-style: italic;color: #DC322F;">nice and red</span>.                                                           
</pre></div>
}}
}

\section{Classes}{


The default theme defines the following inline classes:
\itemize{
\item \code{arg} for a function argument.
\item \code{cls} for an S3, S4, R6 or other class name.
\item \code{code} for a piece of code.
\item \code{dt} is used for the terms in a definition list (\code{\link[=cli_dl]{cli_dl()}}).
\item \code{dd} is used for the descriptions in a definition list (\code{\link[=cli_dl]{cli_dl()}}).
\item \code{email} for an email address.
If the terminal supports ANSI hyperlinks (e.g. RStudio, iTerm2, etc.),
then cli creates a clickable link.
See \link{links} for more information about cli hyperlinks.
\item \code{emph} for emphasized text.
\item \code{envvar} for the name of an environment variable.
\item \code{field} for a generic field, e.g. in a named list.
\item \code{file} for a file name. If the terminal supports ANSI hyperlinks (e.g.
RStudio, iTerm2, etc.), then cli creates a clickable link that opens
the file in RStudio or with the default app for the file type.
See \link{links} for more information about cli hyperlinks.
\item \code{fn} for a function name. If it is in the \code{package::function_name}
form, and the terminal supports ANSI hyperlinks (e.g. RStudio,
iTerm2, etc.), then cli creates a clickable link.
See \link{links} for more information about cli hyperlinks.
\item \code{fun} same as \code{fn}.
\item \code{help} is a help page of a \emph{function}.
If the terminal supports ANSI hyperlinks to help pages (e.g. RStudio),
then cli creates a clickable link. It supports link text.
See \link{links} for more information about cli hyperlinks.
\item \code{href} creates a hyperlink, potentially with a link text.
If the terminal supports ANSI hyperlinks (e.g. RStudio, iTerm2, etc.),
then cli creates a clickable link.
See \link{links} for more information about cli hyperlinks.
\item \code{kbd} for a keyboard key.
\item \code{key} same as \code{kbd}.
\item \code{obj_type_friendly} formats the type of an R object in a readable way,
and it should be used with \code{{}}, see an example below.
\item \code{or} changes the string that separates the last two elements of
collapsed vectors (see below) from "and" to "or".
\item \code{path} for a path (the same as \code{file} in the default theme).
\item \code{pkg} for a package name.
\item \code{run} is an R expression, that is potentially clickable if the terminal
supports ANSI hyperlinks to runnable code (e.g. RStudio).
It supports link text. See \link{links} for more information about cli hyperlinks.
\item \code{str} for a double quoted string escaped by \code{\link[base:encodeString]{base::encodeString()}}.
\item \code{strong} for strong importance.
\item \code{topic} is a help page of a \emph{topic}.
If the terminal supports ANSI hyperlinks to help pages (e.g. RStudio),
then cli creates a clickable link. It supports link text.
See \link{links} for more information about cli hyperlinks.
\item \code{type} formats the type of an R object in a readable way, and it
should be used with \code{{}}, see an example below.
\item \code{url} for a URL. If the terminal supports ANSI hyperlinks (e.g.
RStudio, iTerm2, etc.), then cli creates a clickable link.
See \link{links} for more information about cli hyperlinks.
\item \code{var} for a variable name.
\item \code{val} for a generic "value".
\item \code{vignette} is a vignette.
If the terminal supports ANSI hyperlinks to help pages (e.g. RStudio),
then cli creates a clickable link. It supports link text.
See \link{links} for more information about cli hyperlinks.
}

\if{html}{\out{<div class="sourceCode r">}}\preformatted{ul <- cli_ul()
cli_li("\{.emph Emphasized\} text.")
cli_li("\{.strong Strong\} importance.")
cli_li("A piece of code: \{.code sum(a) / length(a)\}.")
cli_li("A package name: \{.pkg cli\}.")
cli_li("A function name: \{.fn cli_text\}.")
cli_li("A keyboard key: press \{.kbd ENTER\}.")
cli_li("A file name: \{.file /usr/bin/env\}.")
cli_li("An email address: \{.email bugs.bunny@acme.com\}.")
cli_li("A URL: \{.url https://example.com\}.")
cli_li("An environment variable: \{.envvar R_LIBS\}.")
cli_li("`mtcars` is \{.obj_type_friendly \{mtcars\}\}")
cli_end(ul)
}\if{html}{\out{</div>}}\if{html}{\out{
<div class="asciicast" style="color: #172431;font-family: 'Fira Code',Monaco,Consolas,Menlo,'Bitstream Vera Sans Mono','Powerline Symbols',monospace;line-height: 1.300000"><pre>
#> • <span style="font-style: italic;">Emphasized</span> text.                                                              
#> • <span style="font-weight: bold;">Strong</span> importance.                                                            
#> • A piece of code: `sum(a) / length(a)`.                                        
#> • A package name: <span style="color: #268BD2;">cli</span>.                                                          
#> • A function name: `cli_text()`.                                                
#> • A keyboard key: press <span style="color: #268BD2;">[ENTER]</span>.                                                
#> • A file name: <span style="color: #268BD2;">/usr/bin/env</span>.                                                    
#> • An email address: <span style="color: #268BD2;">bugs.bunny@acme.com</span>.                                        
#> • A URL: <span style="font-style: italic;color: #268BD2;">&lt;https://example.com&gt;</span>.                                                 
#> • An environment variable: `R_LIBS`.                                            
#> • `mtcars` is a data frame                                                      
</pre></div>
}}


You can add new classes by defining them in the theme, and then using
them.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{cli_div(theme = list(
  span.myclass = list(color = "lightgrey"),
  "span.myclass" = list(before = "<<"),
  "span.myclass" = list(after = ">>")))
cli_text("This is \{.myclass in angle brackets\}.")
cli_end()
}\if{html}{\out{</div>}}\if{html}{\out{
<div class="asciicast" style="color: #172431;font-family: 'Fira Code',Monaco,Consolas,Menlo,'Bitstream Vera Sans Mono','Powerline Symbols',monospace;line-height: 1.300000"><pre>
#> This is <span style="color: #cccccc;">&lt;&lt;in angle brackets&gt;&gt;</span>.                                                  
</pre></div>
}}

\subsection{Highlighting weird-looking values}{

Often it is useful to highlight a weird file or path name, e.g. one
that starts or ends with space characters. The built-in theme does this
for \code{.file}, \code{.path} and \code{.email} by default. You can highlight
any string inline by adding the \code{.q} class to it.

The current highlighting algorithm
\itemize{
\item adds single quotes to the string if it does not start or end with an
alphanumeric character, underscore, dot or forward slash.
\item Highlights the background colors of leading and trailing spaces on
terminals that support ANSI colors.
}
}
}

\section{Collapsing inline vectors}{


When cli performs inline text formatting, it automatically collapses
glue substitutions, after formatting. This is handy to create lists of
files, packages, etc.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{pkgs <- c("pkg1", "pkg2", "pkg3")
cli_text("Packages: \{pkgs\}.")
cli_text("Packages: \{.pkg \{pkgs\}\}.")
}\if{html}{\out{</div>}}\if{html}{\out{
<div class="asciicast" style="color: #172431;font-family: 'Fira Code',Monaco,Consolas,Menlo,'Bitstream Vera Sans Mono','Powerline Symbols',monospace;line-height: 1.300000"><pre>
#> Packages: pkg1, pkg2, and pkg3.                                                 
#> Packages: <span style="color: #268BD2;">pkg1</span>, <span style="color: #268BD2;">pkg2</span>, and <span style="color: #268BD2;">pkg3</span>.                                                 
</pre></div>
}}


Class names are collapsed differently by default

\if{html}{\out{<div class="sourceCode r">}}\preformatted{x <- Sys.time()
cli_text("Hey, \{.var x\} has class \{.cls \{class(x)\}\}.")
}\if{html}{\out{</div>}}\if{html}{\out{
<div class="asciicast" style="color: #172431;font-family: 'Fira Code',Monaco,Consolas,Menlo,'Bitstream Vera Sans Mono','Powerline Symbols',monospace;line-height: 1.300000"><pre>
#> Hey, `x` has class <span style="color: #268BD2;">&lt;POSIXct/POSIXt&gt;</span>.                                            
</pre></div>
}}


By default cli truncates long vectors. The truncation limit is by default
twenty elements, but you can change it with the \code{vec-trunc} style.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{nms <- cli_vec(names(mtcars), list("vec-trunc" = 5))
cli_text("Column names: \{nms\}.")
}\if{html}{\out{</div>}}\if{html}{\out{
<div class="asciicast" style="color: #172431;font-family: 'Fira Code',Monaco,Consolas,Menlo,'Bitstream Vera Sans Mono','Powerline Symbols',monospace;line-height: 1.300000"><pre>
#> Column names: mpg, cyl, disp, …, gear, and carb.                                
</pre></div>
}}
}

\section{Formatting values}{


The \code{val} inline class formats values. By default (c.f. the built-in
theme), it calls the \code{\link[=cli_format]{cli_format()}} generic function, with the current
style as the argument. See \code{\link[=cli_format]{cli_format()}} for examples.

\code{str} is for formatting strings, it uses \code{\link[base:encodeString]{base::encodeString()}} with
double quotes.
}

\section{Escaping \verb{\{} and \verb{\}}}{


It might happen that you want to pass a string to \verb{cli_*} functions,
and you do \emph{not} want command substitution in that string, because it
might contain \verb{\{} and \verb{\}} characters. The simplest solution for this is
to refer to the string from a template:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{msg <- "Error in if (ncol(dat$y)) \{: argument is of length zero"
cli_alert_warning("\{msg\}")
}\if{html}{\out{</div>}}\if{html}{\out{
<div class="asciicast" style="color: #172431;font-family: 'Fira Code',Monaco,Consolas,Menlo,'Bitstream Vera Sans Mono','Powerline Symbols',monospace;line-height: 1.300000"><pre>
#> <span style="color: #B58900;">!</span> Error in if (ncol(dat$y)) \{: argument is of length zero                       
</pre></div>
}}


If you want to explicitly escape \verb{\{} and \verb{\}} characters, just double
them:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{cli_alert_warning("A warning with \{\{ braces \}\}.")
}\if{html}{\out{</div>}}\if{html}{\out{
<div class="asciicast" style="color: #172431;font-family: 'Fira Code',Monaco,Consolas,Menlo,'Bitstream Vera Sans Mono','Powerline Symbols',monospace;line-height: 1.300000"><pre>
#> <span style="color: #B58900;">!</span> A warning with \{ braces \}.                                                    
</pre></div>
}}


See also examples below.
}

\section{Pluralization}{


All cli commands that emit text support pluralization. Some examples:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{ndirs <- 1
nfiles <- 13
pkgs <- c("pkg1", "pkg2", "pkg3")
cli_alert_info("Found \{ndirs\} director\{?y/ies\} and \{nfiles\} file\{?s\}.")
cli_text("Will install \{length(pkgs)\} package\{?s\}: \{.pkg \{pkgs\}\}")
}\if{html}{\out{</div>}}\if{html}{\out{
<div class="asciicast" style="color: #172431;font-family: 'Fira Code',Monaco,Consolas,Menlo,'Bitstream Vera Sans Mono','Powerline Symbols',monospace;line-height: 1.300000"><pre>
#> <span style="color: #2AA198;">ℹ</span> Found 1 directory and 13 files.                                               
#> Will install 3 packages: <span style="color: #268BD2;">pkg1</span>, <span style="color: #268BD2;">pkg2</span>, and <span style="color: #268BD2;">pkg3</span>                                   
</pre></div>
}}


See \link{pluralization} for details.
}

\section{Wrapping}{


Most cli containers wrap the text to width the container's width,
while observing margins requested by the theme.

To avoid a line break, you can use the UTF_8 non-breaking space
character: \verb{\\u00a0}. cli will not break a line here.

To force a line break, insert a form feed character: \verb{\\f} or
\verb{\\u000c}. cli will insert a line break there.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/print.R
\name{cli_format_method}
\alias{cli_format_method}
\title{Create a format method for an object using cli tools}
\usage{
cli_format_method(expr, theme = getOption("cli.theme"))
}
\arguments{
\item{expr}{Expression that calls \verb{cli_*} methods, \code{\link[base:cat]{base::cat()}} or
\code{\link[base:print]{base::print()}} to format an object's printout.}

\item{theme}{Theme to use for the formatting.}
}
\value{
Character vector, one element for each line of the printout.
}
\description{
This method can be typically used in \code{format()} S3 methods. Then the
\code{print()} method of the class can be easily defined in terms of such a
\code{format()} method. See examples below.
}
\examples{

# Let's create format and print methods for a new S3 class that
# represents the an installed R package: `r_package`

# An `r_package` will contain the DESCRIPTION metadata of the package
# and also its installation path.
new_r_package <- function(pkg) {
  tryCatch(
    desc <- packageDescription(pkg),
    warning = function(e) stop("Cannot find R package `", pkg, "`")
  )
  file <- dirname(attr(desc, "file"))
  if (basename(file) != pkg) file <- dirname(file)
  structure(
    list(desc = unclass(desc), lib = dirname(file)),
    class = "r_package"
  )
}

format.r_package <- function(x, ...) {
  cli_format_method({
    cli_h1("{.pkg {x$desc$Package}} {cli::symbol$line} {x$desc$Title}")
    cli_text("{x$desc$Description}")
    cli_ul(c(
      "Version: {x$desc$Version}",
      if (!is.null(x$desc$Maintainer)) "Maintainer: {x$desc$Maintainer}",
      "License: {x$desc$License}"
    ))
    if (!is.na(x$desc$URL)) cli_text("See more at {.url {x$desc$URL}}")
  })
}

# Now the print method is easy:
print.r_package <- function(x, ...) {
  cat(format(x, ...), sep = "\n")
}

# Try it out
new_r_package("cli")

# The formatting of the output depends on the current theme:
opt <- options(cli.theme = simple_theme())
print(new_r_package("cli"))
options(opt)  # <- restore theme
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ansi-hyperlink.R
\name{style_hyperlink}
\alias{style_hyperlink}
\alias{ansi_has_hyperlink_support}
\alias{ansi_hyperlink_types}
\title{Terminal Hyperlinks}
\usage{
style_hyperlink(text, url, params = NULL)

ansi_has_hyperlink_support()

ansi_hyperlink_types()
}
\arguments{
\item{text}{Text to show. \code{text} and \code{url} are recycled to match their
length, via a \code{paste0()} call.}

\item{url}{URL to link to.}

\item{params}{A named character vector of additional parameters, or \code{NULL}.}
}
\value{
Styled \code{cli_ansi_string} for \code{style_hyperlink()}.
Logical scalar for \code{ansi_has_hyperlink_support()}.
}
\description{
\code{ansi_hyperlink()} creates an ANSI hyperlink.
}
\details{
This function is currently experimental. In particular, many of the
\verb{ansi_*()} functions do not support it properly.

\code{ansi_has_hyperlink_support()} checks if the current \code{stdout()}
supports hyperlinks.

See also
\url{https://gist.github.com/egmontkob/eb114294efbcd5adb1944c9f3cb5feda}.

\code{ansi_hyperlink_types()} checks if current \code{stdout()} supports various
types of hyperlinks. It returns a list with entries \code{href}, \code{run},
\code{help} and \code{vignettes}.
}
\examples{
cat("This is an", style_hyperlink("R", "https://r-project.org"), "link.\n")
ansi_has_hyperlink_support()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/progress-client.R
\name{cli_progress_step}
\alias{cli_progress_step}
\title{Simplified cli progress messages, with styling}
\usage{
cli_progress_step(
  msg,
  msg_done = msg,
  msg_failed = msg,
  spinner = FALSE,
  class = if (!spinner) ".alert-info",
  current = TRUE,
  .auto_close = TRUE,
  .envir = parent.frame(),
  ...
)
}
\arguments{
\item{msg}{Message to show. It may contain glue substitution and cli
styling. It can be updated via \code{\link[=cli_progress_update]{cli_progress_update()}}, as usual.
It is style as a cli info alert (see \code{\link[=cli_alert_info]{cli_alert_info()}}).}

\item{msg_done}{Message to show on successful termination. By default
this it is the same as \code{msg} and it is styled as a cli success alert
(see \code{\link[=cli_alert_success]{cli_alert_success()}}).}

\item{msg_failed}{Message to show on unsuccessful termination. By
default it is the same as \code{msg} and it is styled as a cli danger alert
(see \code{\link[=cli_alert_danger]{cli_alert_danger()}}).}

\item{spinner}{Whether to show a spinner at the beginning of the line.
To make the spinner spin, you'll need to call \code{cli_progress_update()}
regularly.}

\item{class}{cli class to add to the message. By default there is no
class for steps with a spinner.}

\item{current}{Passed to \code{\link[=cli_progress_bar]{cli_progress_bar()}}.}

\item{.auto_close}{Passed to \code{\link[=cli_progress_bar]{cli_progress_bar()}}.}

\item{.envir}{Passed to \code{\link[=cli_progress_bar]{cli_progress_bar()}}.}

\item{...}{Passed to \code{\link[=cli_progress_bar]{cli_progress_bar()}}.}
}
\description{
This is a simplified progress bar, a single (dynamic)
message, without progress units.
}
\details{
\code{cli_progress_step()} always shows the progress message,
even if no update is due.
\subsection{Basic use}{

\if{html}{\out{<div class="sourceCode r">}}\preformatted{f <- function() \{
  cli_progress_step("Downloading data")
  Sys.sleep(2)
  cli_progress_step("Importing data")
  Sys.sleep(1)
  cli_progress_step("Cleaning data")
  Sys.sleep(2)
  cli_progress_step("Fitting model")
  Sys.sleep(3)
\}
f()
}\if{html}{\out{</div>}}

\if{html}{\figure{progress-step.svg}}
}

\subsection{Spinner}{

You can add a spinner to some or all steps with \code{spinner = TRUE},
but note that this will only work if you call \code{\link[=cli_progress_update]{cli_progress_update()}}
regularly.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{f <- function() \{
  cli_progress_step("Downloading data", spinner = TRUE)
  for (i in 1:100) \{ Sys.sleep(2/100); cli_progress_update() \}
  cli_progress_step("Importing data")
  Sys.sleep(1)
  cli_progress_step("Cleaning data")
  Sys.sleep(2)
  cli_progress_step("Fitting model", spinner = TRUE)
  for (i in 1:100) \{ Sys.sleep(3/100); cli_progress_update() \}
\}
f()
}\if{html}{\out{</div>}}

\if{html}{\figure{progress-step-spin.svg}}
}

\subsection{Dynamic messages}{

You can make the step messages dynamic, using glue templates.
Since \code{cli_progress_step()} show that message immediately, we need
to initialize \code{msg} first.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{f <- function() \{
  msg <- ""
  cli_progress_step("Downloading data\{msg\}", spinner = TRUE)
  for (i in 1:100) \{
    Sys.sleep(2/100)
    msg <- glue::glue(", got file \{i\}/100")
    cli_progress_update()
  \}
  cli_progress_step("Importing data")
  Sys.sleep(1)
  cli_progress_step("Cleaning data")
  Sys.sleep(2)
  cli_progress_step("Fitting model", spinner = TRUE)
  for (i in 1:100) \{ Sys.sleep(3/100); cli_progress_update() \}
\}
f()
}\if{html}{\out{</div>}}

\if{html}{\figure{progress-step-dynamic.svg}}
}

\subsection{Termination messages}{

You can specify a different message for successful and/or
unsuccessful termination:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{f <- function() \{
  size <- 0L
  cli_progress_step(
    "Downloading data.",
    msg_done = "Downloaded \{prettyunits::pretty_bytes(size)\}.",
    spinner = TRUE
  )
  for (i in 1:100) \{
    Sys.sleep(3/100)
    size <- size + 8192
    cli_progress_update()
  \}
\}
f()
}\if{html}{\out{</div>}}

\if{html}{\figure{progress-step-msg.svg}}
}
}
\seealso{
This function supports \link[=inline-markup]{inline markup}.

Other progress bar functions: 
\code{\link{cli_progress_along}()},
\code{\link{cli_progress_bar}()},
\code{\link{cli_progress_builtin_handlers}()},
\code{\link{cli_progress_message}()},
\code{\link{cli_progress_num}()},
\code{\link{cli_progress_output}()},
\code{\link{cli_progress_styles}()},
\code{\link{progress-variables}}

Other functions supporting inline markup: 
\code{\link{cli_abort}()},
\code{\link{cli_alert}()},
\code{\link{cli_blockquote}()},
\code{\link{cli_bullets}()},
\code{\link{cli_bullets_raw}()},
\code{\link{cli_dl}()},
\code{\link{cli_h1}()},
\code{\link{cli_li}()},
\code{\link{cli_ol}()},
\code{\link{cli_process_start}()},
\code{\link{cli_progress_along}()},
\code{\link{cli_progress_bar}()},
\code{\link{cli_progress_message}()},
\code{\link{cli_progress_output}()},
\code{\link{cli_rule}},
\code{\link{cli_status}()},
\code{\link{cli_status_update}()},
\code{\link{cli_text}()},
\code{\link{cli_ul}()},
\code{\link{format_error}()},
\code{\link{format_inline}()}
}
\concept{functions supporting inline markup}
\concept{progress bar functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/themes.R
\name{builtin_theme}
\alias{builtin_theme}
\title{The built-in CLI theme}
\usage{
builtin_theme(dark = getOption("cli.theme_dark", "auto"))
}
\arguments{
\item{dark}{Whether to use a dark theme. The \code{cli.theme_dark} option
can be used to request a dark theme explicitly. If this is not set,
or set to \code{"auto"}, then cli tries to detect a dark theme, this
works in recent RStudio versions and in iTerm on macOS.}
}
\value{
A named list, a CLI theme.
}
\description{
This theme is always active, and it is at the bottom of the theme
stack. See \link{themes}.
}
\section{Showcase}{
\if{html}{\out{<div class="sourceCode r">}}\preformatted{cli_h1("Heading 1")
cli_h2("Heading 2")
cli_h3("Heading 3")

cli_par()
cli_alert_danger("Danger alert")
cli_alert_warning("Warning alert")
cli_alert_info("Info alert")
cli_alert_success("Success alert")
cli_alert("Alert for starting a process or computation",
  class = "alert-start")
cli_end()

cli_text("Packages and versions: \{.pkg cli\} \{.version 1.0.0\}.")
cli_text("Time intervals: \{.timestamp 3.4s\}")

cli_text("\{.emph Emphasis\} and  \{.strong strong emphasis\}")

cli_text("This is a piece of code: \{.code sum(x) / length(x)\}")
cli_text("Function names: \{.fn cli::simple_theme\}")

cli_text("Files: \{.file /usr/bin/env\}")
cli_text("URLs: \{.url https://r-project.org\}")

cli_h2("Longer code chunk")
cli_par(class = "code R")
cli_verbatim(
  '# window functions are useful for grouped mutates',
  'mtcars |>',
  '  group_by(cyl) |>',
  '  mutate(rank = min_rank(desc(mpg)))')
}\if{html}{\out{</div>}}\if{html}{\out{
<div class="asciicast" style="color: #172431;font-family: 'Fira Code',Monaco,Consolas,Menlo,'Bitstream Vera Sans Mono','Powerline Symbols',monospace;line-height: 1.300000"><pre>
#> <span style="color: #2AA198;">──</span> <span style="font-weight: bold;">Heading 1</span> <span style="color: #2AA198;">─────────────────────────────────────────────────────────</span>          
#>                                                                                 
#> ── <span style="font-weight: bold;">Heading 2</span> ──                                                                 
#>                                                                                 
#> ── Heading 3                                                                    
#> <span style="color: #DC322F;">✖</span> Danger alert                                                                  
#> <span style="color: #B58900;">!</span> Warning alert                                                                 
#> <span style="color: #2AA198;">ℹ</span> Info alert                                                                    
#> <span style="color: #859900;">✔</span> Success alert                                                                 
#> → Alert for starting a process or computation                                   
#>                                                                                 
#> Packages and versions: <span style="color: #268BD2;">cli</span> 1.0.0.                                               
#> Time intervals: <span style="color: #b8b8b8;">[3.4s]</span>                                                          
#> <span style="font-style: italic;">Emphasis</span> and <span style="font-weight: bold;">strong emphasis</span>                                                    
#> This is a piece of code: `sum(x) / length(x)`                                   
#> Function names: `<a href="x-r-help:cli::simple_theme">cli::simple_theme</a>()`                                           
#> Files: <span style="color: #268BD2;"><a href="file:///usr/bin/env">/usr/bin/env</a></span>                                                             
#> URLs: <span style="font-style: italic;color: #268BD2;">&lt;</span><span style="font-style: italic;color: #268BD2;"><a href="https://r-project.org">https://r-project.org</a></span><span style="font-style: italic;color: #268BD2;">&gt;</span>                                                   
#>                                                                                 
#> ── <span style="font-weight: bold;">Longer code chunk</span> ──                                                         
#>                                                                                 
#> <span style="color: #666699;"># window functions are useful for grouped mutates</span>                               
#> mtcars <span style="color: #cc9900;">|&gt;</span>                                                                       
#>   <span style="font-weight: bold;">group_by</span><span style="color: #cc9900;">(</span>cyl<span style="color: #cc9900;">)</span> <span style="color: #cc9900;">|&gt;</span>                                                              
#>   <span style="font-weight: bold;">mutate</span><span style="color: #cc9900;">(</span>rank = <span style="font-weight: bold;">min_rank</span><span style="color: #B58900;">(</span><span style="font-weight: bold;">desc</span><span style="color: #268BD2;">(</span>mpg<span style="color: #268BD2;">)</span><span style="color: #B58900;">)</span><span style="color: #cc9900;">)</span>                                            
</pre></div>
}}
}

\seealso{
\link{themes}, \code{\link[=simple_theme]{simple_theme()}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/prettycode.R
\name{pretty_print_code}
\alias{pretty_print_code}
\title{Turn on pretty-printing functions at the R console}
\usage{
pretty_print_code()
}
\description{
Defines a print method for functions, in the current session, that supports
syntax highlighting.
}
\details{
The new print method takes priority over the built-in one. Use
\code{\link[base:message]{base::suppressMessages()}} to suppress the alert message.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cli.R
\name{cli_div}
\alias{cli_div}
\title{Generic CLI container}
\usage{
cli_div(
  id = NULL,
  class = NULL,
  theme = NULL,
  .auto_close = TRUE,
  .envir = parent.frame()
)
}
\arguments{
\item{id}{Element id, a string. If \code{NULL}, then a new id is generated
and returned.}

\item{class}{Class name, sting. Can be used in themes.}

\item{theme}{A custom theme for the container. See \link{themes}.}

\item{.auto_close}{Whether to close the container, when the calling
function finishes (or \code{.envir} is removed, if specified).}

\item{.envir}{Environment to evaluate the glue expressions in. It is
also used to auto-close the container if \code{.auto_close} is \code{TRUE}.}
}
\value{
The id of the new container element, invisibly.
}
\description{
See \link{containers}. A \code{cli_div} container is special, because it may
add new themes, that are valid within the container.
}
\examples{
## div with custom theme
d <- cli_div(theme = list(h1 = list(color = "blue",
                                    "font-weight" = "bold")))
cli_h1("Custom title")
cli_end(d)

## Close automatically
div <- function() {
  cli_div(class = "tmp", theme = list(.tmp = list(color = "yellow")))
  cli_text("This is yellow")
}
div()
cli_text("This is not yellow any more")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ansiex.R
\name{ansi_nchar}
\alias{ansi_nchar}
\title{Count number of characters in an ANSI colored string}
\usage{
ansi_nchar(x, type = c("chars", "bytes", "width", "graphemes", "codepoints"))
}
\arguments{
\item{x}{Character vector, potentially ANSI styled, or a vector to be
coerced to character. If it converted to UTF-8.}

\item{type}{Whether to count graphemes (characters), code points,
bytes, or calculate the display width of the string.}
}
\value{
Numeric vector, the length of the strings in the character
vector.
}
\description{
This is a color-aware counterpart of \code{\link[=utf8_nchar]{utf8_nchar()}}. By default it
counts Unicode grapheme clusters, instead of code points.
}
\examples{
str <- paste(
  col_red("red"),
  "default",
  col_green("green")
)

cat(str, "\n")
nchar(str)
ansi_nchar(str)
nchar(ansi_strip(str))
}
\seealso{
Other ANSI string operations: 
\code{\link{ansi_align}()},
\code{\link{ansi_columns}()},
\code{\link{ansi_strsplit}()},
\code{\link{ansi_strtrim}()},
\code{\link{ansi_strwrap}()},
\code{\link{ansi_substr}()},
\code{\link{ansi_substring}()},
\code{\link{ansi_toupper}()},
\code{\link{ansi_trimws}()}
}
\concept{ANSI string operations}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tty.R
\name{cli_output_connection}
\alias{cli_output_connection}
\title{The connection option that cli would use}
\usage{
cli_output_connection()
}
\value{
Connection object.
}
\description{
Note that this only refers to the current R process. If the output
is produced in another process, then it is not relevant.
}
\details{
In interactive sessions the standard output is chosen, otherwise the
standard error is used. This is to avoid painting output messages red
in the R GUIs.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ansiex.R
\name{ansi_simplify}
\alias{ansi_simplify}
\title{Simplify ANSI styling tags}
\usage{
ansi_simplify(x, csi = c("keep", "drop"))
}
\arguments{
\item{x}{Input string}

\item{csi}{What to do with non-SGR ANSI sequences, either \code{"keep"},
or \code{"drop"} them.}
}
\value{
Simplified \code{cli_ansi_string} vector.
}
\description{
It creates an equivalent, but possibly shorter ANSI styled string, by
removing duplicate and empty tags.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cli.R
\name{cli_ul}
\alias{cli_ul}
\title{Unordered CLI list}
\usage{
cli_ul(
  items = NULL,
  id = NULL,
  class = NULL,
  .close = TRUE,
  .auto_close = TRUE,
  .envir = parent.frame()
)
}
\arguments{
\item{items}{If not \code{NULL}, then a character vector. Each element of
the vector will be one list item, and the list container will be
closed by default (see the \code{.close} argument).}

\item{id}{Id of the list container. Can be used for closing it with
\code{\link[=cli_end]{cli_end()}} or in themes. If \code{NULL}, then an id is generated and
returned invisibly.}

\item{class}{Class of the list container. Can be used in themes.}

\item{.close}{Whether to close the list container if the \code{items} were
specified. If \code{FALSE} then new items can be added to the list.}

\item{.auto_close}{Whether to close the container, when the calling
function finishes (or \code{.envir} is removed, if specified).}

\item{.envir}{Environment to evaluate the glue expressions in. It is
also used to auto-close the container if \code{.auto_close} is \code{TRUE}.}
}
\value{
The id of the new container element, invisibly.
}
\description{
An unordered list is a container, see \link{containers}.
}
\details{
\subsection{Adding all items at once}{

\if{html}{\out{<div class="sourceCode r">}}\preformatted{fun <- function() \{
  cli_ul(c("one", "two", "three"))
\}
fun()
}\if{html}{\out{</div>}}\if{html}{\out{
<div class="asciicast" style="color: #172431;font-family: 'Fira Code',Monaco,Consolas,Menlo,'Bitstream Vera Sans Mono','Powerline Symbols',monospace;line-height: 1.300000"><pre>
#> • one                                                                           
#> • two                                                                           
#> • three                                                                         
</pre></div>
}}

}

\subsection{Adding items one by one}{

\if{html}{\out{<div class="sourceCode r">}}\preformatted{fun <- function() \{
  cli_ul()
  cli_li("\{.emph one\}")
  cli_li("\{.emph two\}")
  cli_li("\{.emph three\}")
  cli_end()
\}
fun()
}\if{html}{\out{</div>}}\if{html}{\out{
<div class="asciicast" style="color: #172431;font-family: 'Fira Code',Monaco,Consolas,Menlo,'Bitstream Vera Sans Mono','Powerline Symbols',monospace;line-height: 1.300000"><pre>
#> • <span style="font-style: italic;">one</span>                                                                           
#> • <span style="font-style: italic;">two</span>                                                                           
#> • <span style="font-style: italic;">three</span>                                                                         
</pre></div>
}}

}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ruler.R
\name{ruler}
\alias{ruler}
\title{Print the helpful ruler to the screen}
\usage{
ruler(width = console_width())
}
\arguments{
\item{width}{Ruler width.}
}
\description{
Print the helpful ruler to the screen
}
\examples{
ruler()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cli.R
\name{cli_end}
\alias{cli_end}
\title{Close a CLI container}
\usage{
cli_end(id = NULL)
}
\arguments{
\item{id}{Id of the container to close. If missing, the current
container is closed, if any.}
}
\description{
Containers aut0-close by default, but sometimes you need to explicitly
close them. Closing a container also closes all of its nested
containers.
}
\details{
\subsection{Explicit closing}{

\if{html}{\out{<div class="sourceCode r">}}\preformatted{cnt <- cli_par()
cli_text("First paragraph.")
cli_end(cnt)
cnt <- cli_par()
cli_text("Second paragraph.")
cli_end(cnt)
}\if{html}{\out{</div>}}\if{html}{\out{
<div class="asciicast" style="color: #172431;font-family: 'Fira Code',Monaco,Consolas,Menlo,'Bitstream Vera Sans Mono','Powerline Symbols',monospace;line-height: 1.300000"><pre>
#> First paragraph.                                                                
#>                                                                                 
#> Second paragraph.                                                               
#>                                                                                 
</pre></div>
}}

}

\subsection{Closing a stack of containers}{

\if{html}{\out{<div class="sourceCode r">}}\preformatted{list <- cli_ul()
cli_li("Item one:")
cli_li("Item two:")
cli_par()
cli_text("Still item two.")
cli_end(list)
cli_text("Not in the list any more")
}\if{html}{\out{</div>}}\if{html}{\out{
<div class="asciicast" style="color: #172431;font-family: 'Fira Code',Monaco,Consolas,Menlo,'Bitstream Vera Sans Mono','Powerline Symbols',monospace;line-height: 1.300000"><pre>
#> • Item one:                                                                     
#> • Item two:                                                                     
#> Still item two.                                                                 
#>                                                                                 
#> Not in the list any more                                                        
</pre></div>
}}

}

\subsection{Omitting \code{id}}{

If \code{id} is omitted, the container that was opened last will be closed.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{cli_par()
cli_text("First paragraph")
cli_end()
cli_par()
cli_text("Second paragraph")
cli_end()
}\if{html}{\out{</div>}}\if{html}{\out{
<div class="asciicast" style="color: #172431;font-family: 'Fira Code',Monaco,Consolas,Menlo,'Bitstream Vera Sans Mono','Powerline Symbols',monospace;line-height: 1.300000"><pre>
#> First paragraph                                                                 
#>                                                                                 
#> Second paragraph                                                                
#>                                                                                 
</pre></div>
}}

}

\subsection{Debugging containers}{

You can use the internal \code{cli:::cli_debug_doc()} function to see the
currently open containers.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{fun <- function() \{
  cli_div(id = "mydiv")
  cli_par(class = "myclass")
  cli:::cli_debug_doc()
\}
fun()
}\if{html}{\out{</div>}}\if{html}{\out{
<div class="asciicast" style="color: #172431;font-family: 'Fira Code',Monaco,Consolas,Menlo,'Bitstream Vera Sans Mono','Powerline Symbols',monospace;line-height: 1.300000"><pre>
#> &lt;cli document&gt;                                                                  
#> &lt;body id="body"&gt;                                                                
#> &lt;div id="mydiv"&gt; +theme                                                         
#> &lt;par id="cli-82040-64" class="myclass"&gt;                                         
</pre></div>
}}

}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/status-bar.R
\name{cli_status_clear}
\alias{cli_status_clear}
\title{Clear the status bar (superseded)}
\usage{
cli_status_clear(
  id = NULL,
  result = c("clear", "done", "failed"),
  msg_done = NULL,
  msg_failed = NULL,
  .envir = parent.frame()
)
}
\arguments{
\item{id}{Id of the status bar container to clear. If \code{id} is not the id
of the current status bar (because it was overwritten by another
status bar container), then the status bar is not cleared. If \code{NULL}
(the default) then the status bar is always cleared.}

\item{result}{Whether to show a message for success or failure or just
clear the status bar.}

\item{msg_done}{If not \code{NULL}, then the message to use for successful
process termination. This overrides the message given when the status
bar was created.}

\item{msg_failed}{If not \code{NULL}, then the message to use for failed
process termination. This overrides the message give when the status
bar was created.}

\item{.envir}{Environment to evaluate the glue expressions in. It is
also used to auto-clear the status bar if \code{.auto_close} is \code{TRUE}.}
}
\description{
\strong{The \verb{cli_status_*()} functions are superseded by
the \code{\link[=cli_progress_message]{cli_progress_message()}} and \code{\link[=cli_progress_step]{cli_progress_step()}} functions,
because they have a better default behavior.}

Clear the status bar
}
\seealso{
The \code{\link[=cli_progress_message]{cli_progress_message()}} and \code{\link[=cli_progress_step]{cli_progress_step()}}
functions, for a superior API.

Other status bar: 
\code{\link{cli_process_start}()},
\code{\link{cli_status}()},
\code{\link{cli_status_update}()}
}
\concept{status bar}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/spark.R
\name{spark_bar}
\alias{spark_bar}
\title{Draw a sparkline bar graph with unicode block characters}
\usage{
spark_bar(x)
}
\arguments{
\item{x}{A numeric vector between 0 and 1}
}
\description{
Rendered using \href{https://en.wikipedia.org/wiki/Block_Elements}{block elements}.
In most common fixed width fonts these are rendered wider than regular
characters which means they are not suitable if you need precise alignment.

You might want to avoid sparklines on non-UTF-8 systems, because they
do not look good. You can use \code{\link[=is_utf8_output]{is_utf8_output()}} to test for support
for them.
}
\details{
\if{html}{\out{<div class="sourceCode r">}}\preformatted{x <- seq(0, 1, length = 6)
spark_bar(x)
}\if{html}{\out{</div>}}\if{html}{\out{
<div class="asciicast" style="color: #172431;font-family: 'Fira Code',Monaco,Consolas,Menlo,'Bitstream Vera Sans Mono','Powerline Symbols',monospace;line-height: 1.300000"><pre>
#> ▁▂▄▅▇█                                                                          
</pre></div>
}}


\if{html}{\out{<div class="sourceCode r">}}\preformatted{x <- seq(0, 1, length = 6)
spark_bar(sample(x))
}\if{html}{\out{</div>}}\if{html}{\out{
<div class="asciicast" style="color: #172431;font-family: 'Fira Code',Monaco,Consolas,Menlo,'Bitstream Vera Sans Mono','Powerline Symbols',monospace;line-height: 1.300000"><pre>
#> ▅▁█▄▇▂                                                                          
</pre></div>
}}


\if{html}{\out{<div class="sourceCode r">}}\preformatted{spark_bar(seq(0, 1, length = 8))
}\if{html}{\out{</div>}}\if{html}{\out{
<div class="asciicast" style="color: #172431;font-family: 'Fira Code',Monaco,Consolas,Menlo,'Bitstream Vera Sans Mono','Powerline Symbols',monospace;line-height: 1.300000"><pre>
#> ▁▂▃▄▅▆▇█                                                                        
</pre></div>
}}


\code{NA}s are left out:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{spark_bar(c(0, NA, 0.5, NA, 1))
}\if{html}{\out{</div>}}\if{html}{\out{
<div class="asciicast" style="color: #172431;font-family: 'Fira Code',Monaco,Consolas,Menlo,'Bitstream Vera Sans Mono','Powerline Symbols',monospace;line-height: 1.300000"><pre>
#> ▁ ▄ █                                                                           
</pre></div>
}}
}
\seealso{
\code{\link[=spark_line]{spark_line()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hash.R
\name{hash_emoji}
\alias{hash_emoji}
\alias{hash_raw_emoji}
\alias{hash_obj_emoji}
\title{Emoji hash}
\usage{
hash_emoji(x, size = 3)

hash_raw_emoji(x, size = 3)

hash_obj_emoji(x, size = 3, serialize_version = 2)
}
\arguments{
\item{x}{Character vector. \code{NA} entries will have an \code{NA} hash.}

\item{size}{Number of emojis to use in a hash. Currently it has to
be from 1 through 4.}

\item{serialize_version}{Workspace format version to use, see
\code{\link[base:serialize]{base::serialize()}}.}
}
\value{
\code{hash_emoji()} returns a data frame with columns
\itemize{
\item \code{hash}: the emoji hash, a string of the requested size.
\item \code{emojis}: list column with the emoji characters in character
vectors. Note that an emoji might have multiple code points.
\item \code{text}: text representation of \code{hash}, comma separated.
\item \code{names}: list column with the text representations of \code{emojis}, in
character vectors.
}

\code{hash_raw_emoji()} and \code{hash_obj_emoji()} return a list with
entries:
\itemize{
\item \code{hash}: the emoji hash, a string of requested size,
\item \code{emojis}: the individual emoji characters in a character vector,
\item \code{text}: text representation of \code{hash}, comma separated,
\item \code{names}: names of the emojis, in a character vector.
}
}
\description{
Emoji hash
}
\details{
It uses the first 13 hexadecimal characters (out of the 32) of the MD5
hash of the input, and converts them into an emoji representation.
It uses a manually selected subset of all emojis, that tend to be
displayed correctly.
\subsection{Number of possible hash values}{

cli uses 2280 possible emojis. This is the number of
different hashes you can get for different values of \code{size}:\tabular{rr}{
   \code{size} \tab size of hash table space \cr
   1 \tab 2,280 \cr
   2 \tab 5,198,400 \cr
   3 \tab 11,852,352,000 \cr
   4 \tab 27,023,362,560,000 \cr
}

}

\code{hash_raw_emoji()} calculates the emoji hash of the bytes
of a raw vector.

\code{hash_obj_emoji()} calculates the emoji hash of an R
object. The object is serialized into a binary vector first.
}
\examples{
hash_emoji(c("foo", NA, "bar", ""))$text

# if you increase `size`, the shorter hash is a prefix of the longer:
hash_emoji("foobar", 1)$text
hash_emoji("foobar", 2)$text
hash_emoji("foobar", 3)$text
hash_emoji("foobar", 4)$text
}
\seealso{
the emoji package for a comprehensive list of emojis

Other hash functions: 
\code{\link{hash_animal}()},
\code{\link{hash_md5}()},
\code{\link{hash_sha1}()},
\code{\link{hash_sha256}()},
\code{\link{hash_xxhash}()}
}
\concept{hash functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/keypress.R
\name{keypress}
\alias{keypress}
\title{Read a single keypress at the terminal}
\usage{
keypress(block = TRUE)
}
\arguments{
\item{block}{Whether to wait for a key press, if there is none
available now.}
}
\value{
The key pressed, a character scalar. For non-blocking reads
\code{NA} is returned if no keys are available.
}
\description{
It currently only works at Linux/Unix and OSX terminals,
and at the Windows command line. see \code{\link{has_keypress_support}}.
}
\details{
The following special keys are supported:
\itemize{
\item Arrow keys: 'up', 'down', 'right', 'left'.
\item Function keys: from 'f1' to 'f12'.
\item Others: 'home', 'end', 'insert', 'delete', 'pageup', 'pagedown',
'tab', 'enter', 'backspace' (same as 'delete' on OSX keyboards),
'escape'.
\item Control with one of the following keys: 'a', 'b', 'c', 'd', 'e', 'f',
'h', 'k', 'l', 'n', 'p', 't', 'u', 'w'.
}
}
\examples{
\dontshow{if (FALSE) (if (getRversion() >= "3.4") withAutoprint else force)(\{ # examplesIf}
x <- keypress()
cat("You pressed key", x, "\n")
\dontshow{\}) # examplesIf}
}
\seealso{
Other keypress function: 
\code{\link{has_keypress_support}()}
}
\concept{keypress function}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cli.R
\name{cli_code}
\alias{cli_code}
\title{A block of code}
\usage{
cli_code(
  lines = NULL,
  ...,
  language = "R",
  .auto_close = TRUE,
  .envir = environment()
)
}
\arguments{
\item{lines}{Character vector, each line will be a line of code, and
newline characters also create new lines. Note that \emph{no} glue
substitution is performed on the code.}

\item{...}{More character vectors, they are appended to \code{lines}.}

\item{language}{Programming language. This is also added as a class,
in addition to \code{code}.}

\item{.auto_close}{Passed to \code{cli_div()} when creating the container of
the code. By default the code container is closed after emitting
\code{lines} and \code{...} via \code{cli_verbatim()}. You can keep that container
open with \code{.auto_close} and/or \code{.envir}, and then calling
\code{cli_verbatim()} to add (more) code. Note that the code will be
formatted and syntax highlighted separately for each \code{cli_verbatim()}
call.}

\item{.envir}{Passed to \code{cli_div()} when creating the container of the
code.}
}
\value{
The id of the container that contains the code.
}
\description{
A helper function that creates a \code{div} with class \code{code} and then calls
\code{cli_verbatim()} to output code lines. The builtin theme formats these
containers specially. In particular, it adds syntax highlighting to
valid R code.
}
\details{
\if{html}{\out{<div class="sourceCode r">}}\preformatted{myfun <- function() \{
  message("Just an example function")
  graphics::pairs(iris, col = 1:4)
\}
cli_code(format(myfun))
}\if{html}{\out{</div>}}\if{html}{\out{
<div class="asciicast" style="color: #172431;font-family: 'Fira Code',Monaco,Consolas,Menlo,'Bitstream Vera Sans Mono','Powerline Symbols',monospace;line-height: 1.300000"><pre>
#> <span style="color: #DC322F;">function</span> <span style="color: #B58900;">()</span>                                                                     
#> <span style="color: #B58900;">\{</span>                                                                               
#>     <span style="color: #2AA198;">message</span><span style="color: #268BD2;">(</span><span style="color: #B58900;">"Just an example function"</span><span style="color: #268BD2;">)</span>                                         
#>     graphics::<span style="color: #2AA198;">pairs</span><span style="color: #268BD2;">(</span>iris, col = <span style="color: #268BD2;">1</span><span style="color: #859900;">:</span><span style="color: #268BD2;">4)</span>                                            
#> <span style="color: #B58900;">\}</span>                                                                               
</pre></div>
}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ansiex.R
\name{ansi_columns}
\alias{ansi_columns}
\title{Format a character vector in multiple columns}
\usage{
ansi_columns(
  text,
  width = console_width(),
  sep = " ",
  fill = c("rows", "cols"),
  max_cols = 4,
  align = c("left", "center", "right"),
  type = "width",
  ellipsis = symbol$ellipsis
)
}
\arguments{
\item{text}{Character vector to format. Each element will formatted
as a cell of a table.}

\item{width}{Width of the screen.}

\item{sep}{Separator between the columns. It may have ANSI styles.}

\item{fill}{Whether to fill the columns row-wise or column-wise.}

\item{max_cols}{Maximum number of columns to use. Will not use more,
even if there is space for it.}

\item{align}{Alignment within the columns.}

\item{type}{Passed to \code{\link[=ansi_nchar]{ansi_nchar()}} and \code{\link[=ansi_align]{ansi_align()}}. Most probably
you want the default, \code{"width"}.}

\item{ellipsis}{The string to append to truncated strings. Supply an
empty string if you don't want a marker.}
}
\value{
ANSI string vector.
}
\description{
This function helps with multi-column output of ANSI styles strings.
It works well together with \code{\link[=boxx]{boxx()}}, see the example below.
}
\details{
If a string does not fit into the specified \code{width}, it will be
truncated using \code{\link[=ansi_strtrim]{ansi_strtrim()}}.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{fmt <- ansi_columns(
  paste(col_red("foo"), 1:10),
  width = 50,
  fill = "rows",
  max_cols=10,
  align = "center",
  sep = "   "
)
boxx(fmt, padding = c(0,1,0,1), header = col_cyan("Columns"))
}\if{html}{\out{</div>}}\if{html}{\out{
<div class="asciicast" style="color: #172431;font-family: 'Fira Code',Monaco,Consolas,Menlo,'Bitstream Vera Sans Mono','Powerline Symbols',monospace;line-height: 1.300000"><pre>
#> ┌ <span style="color: #2AA198;">Columns</span> ───────────────────────────────────────────┐                          
#> │  <span style="color: #DC322F;">foo</span> 1     <span style="color: #DC322F;">foo</span> 2     <span style="color: #DC322F;">foo</span> 3     <span style="color: #DC322F;">foo</span> 4     <span style="color: #DC322F;">foo</span> 5     │                          
#> │  <span style="color: #DC322F;">foo</span> 6     <span style="color: #DC322F;">foo</span> 7     <span style="color: #DC322F;">foo</span> 8     <span style="color: #DC322F;">foo</span> 9     <span style="color: #DC322F;">foo</span> 10    │                          
#> └────────────────────────────────────────────────────┘                          
</pre></div>
}}
}
\seealso{
Other ANSI string operations: 
\code{\link{ansi_align}()},
\code{\link{ansi_nchar}()},
\code{\link{ansi_strsplit}()},
\code{\link{ansi_strtrim}()},
\code{\link{ansi_strwrap}()},
\code{\link{ansi_substr}()},
\code{\link{ansi_substring}()},
\code{\link{ansi_toupper}()},
\code{\link{ansi_trimws}()}
}
\concept{ANSI string operations}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cli.R
\name{cli_alert}
\alias{cli_alert}
\alias{cli_alert_success}
\alias{cli_alert_danger}
\alias{cli_alert_warning}
\alias{cli_alert_info}
\title{CLI alerts}
\usage{
cli_alert(text, id = NULL, class = NULL, wrap = FALSE, .envir = parent.frame())

cli_alert_success(
  text,
  id = NULL,
  class = NULL,
  wrap = FALSE,
  .envir = parent.frame()
)

cli_alert_danger(
  text,
  id = NULL,
  class = NULL,
  wrap = FALSE,
  .envir = parent.frame()
)

cli_alert_warning(
  text,
  id = NULL,
  class = NULL,
  wrap = FALSE,
  .envir = parent.frame()
)

cli_alert_info(
  text,
  id = NULL,
  class = NULL,
  wrap = FALSE,
  .envir = parent.frame()
)
}
\arguments{
\item{text}{Text of the alert.}

\item{id}{Id of the alert element. Can be used in themes.}

\item{class}{Class of the alert element. Can be used in themes.}

\item{wrap}{Whether to auto-wrap the text of the alert.}

\item{.envir}{Environment to evaluate the glue expressions in.}
}
\description{
Alerts are typically short status messages.
}
\details{
\subsection{Success}{

\if{html}{\out{<div class="sourceCode r">}}\preformatted{nbld <- 11
tbld <- prettyunits::pretty_sec(5.6)
cli_alert_success("Built \{.emph \{nbld\}\} status report\{?s\} in \{tbld\}.")
}\if{html}{\out{</div>}}\if{html}{\out{
<div class="asciicast" style="color: #172431;font-family: 'Fira Code',Monaco,Consolas,Menlo,'Bitstream Vera Sans Mono','Powerline Symbols',monospace;line-height: 1.300000"><pre>
#> <span style="color: #859900;">✔</span> Built <span style="font-style: italic;">11</span> status reports in 5.6s.                                              
</pre></div>
}}

}

\subsection{Info}{

\if{html}{\out{<div class="sourceCode r">}}\preformatted{cfl <- "~/.cache/files/latest.cache"
cli_alert_info("Updating cache file \{.path \{cfl\}\}.")
}\if{html}{\out{</div>}}\if{html}{\out{
<div class="asciicast" style="color: #172431;font-family: 'Fira Code',Monaco,Consolas,Menlo,'Bitstream Vera Sans Mono','Powerline Symbols',monospace;line-height: 1.300000"><pre>
#> <span style="color: #2AA198;">ℹ</span> Updating cache file <span style="color: #268BD2;">~/.cache/files/latest.cache</span>.                              
</pre></div>
}}

}

\subsection{Warning}{

\if{html}{\out{<div class="sourceCode r">}}\preformatted{cfl <- "~/.cache/files/latest.cache"
cli_alert_warning("Failed to update cache file \{.path \{cfl\}\}.")
}\if{html}{\out{</div>}}\if{html}{\out{
<div class="asciicast" style="color: #172431;font-family: 'Fira Code',Monaco,Consolas,Menlo,'Bitstream Vera Sans Mono','Powerline Symbols',monospace;line-height: 1.300000"><pre>
#> <span style="color: #B58900;">!</span> Failed to update cache file <span style="color: #268BD2;">~/.cache/files/latest.cache</span>.                      
</pre></div>
}}

}

\subsection{Danger}{

\if{html}{\out{<div class="sourceCode r">}}\preformatted{cfl <- "~/.config/report.yaml"
cli_alert_danger("Cannot validate config file at \{.path \{cfl\}\}.")
}\if{html}{\out{</div>}}\if{html}{\out{
<div class="asciicast" style="color: #172431;font-family: 'Fira Code',Monaco,Consolas,Menlo,'Bitstream Vera Sans Mono','Powerline Symbols',monospace;line-height: 1.300000"><pre>
#> <span style="color: #DC322F;">✖</span> Cannot validate config file at <span style="color: #268BD2;">~/.config/report.yaml</span>.                         
</pre></div>
}}

}

\subsection{Text wrapping}{

Alerts are printed without wrapping, unless you set \code{wrap = TRUE}:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{cli_alert_info("Data columns: \{.val \{names(mtcars)\}\}.")
cli_alert_info("Data columns: \{.val \{names(mtcars)\}\}.", wrap = TRUE)
}\if{html}{\out{</div>}}\if{html}{\out{
<div class="asciicast" style="color: #172431;font-family: 'Fira Code',Monaco,Consolas,Menlo,'Bitstream Vera Sans Mono','Powerline Symbols',monospace;line-height: 1.300000"><pre>
#> <span style="color: #2AA198;">ℹ</span> Data columns: <span style="color: #268BD2;">"mpg"</span>, <span style="color: #268BD2;">"cyl"</span>, <span style="color: #268BD2;">"disp"</span>, <span style="color: #268BD2;">"hp"</span>, <span style="color: #268BD2;">"drat"</span>, <span style="color: #268BD2;">"wt"</span>, <span style="color: #268BD2;">"qsec"</span>, <span style="color: #268BD2;">"vs"</span>, <span style="color: #268BD2;">"am"</span>, <span style="color: #268BD2;">"g</span>
#> <span style="color: #268BD2;">ear"</span>, and <span style="color: #268BD2;">"carb"</span>.                                                               
#> <span style="color: #2AA198;">ℹ</span> Data columns: <span style="color: #268BD2;">"mpg"</span>, <span style="color: #268BD2;">"cyl"</span>, <span style="color: #268BD2;">"disp"</span>, <span style="color: #268BD2;">"hp"</span>, <span style="color: #268BD2;">"drat"</span>, <span style="color: #268BD2;">"wt"</span>, <span style="color: #268BD2;">"qsec"</span>,               
#> <span style="color: #268BD2;">"vs"</span>, <span style="color: #268BD2;">"am"</span>, <span style="color: #268BD2;">"gear"</span>, and <span style="color: #268BD2;">"carb"</span>.                                                 
</pre></div>
}}

}
}
\seealso{
These functions supports \link[=inline-markup]{inline markup}.

Other functions supporting inline markup: 
\code{\link{cli_abort}()},
\code{\link{cli_blockquote}()},
\code{\link{cli_bullets}()},
\code{\link{cli_bullets_raw}()},
\code{\link{cli_dl}()},
\code{\link{cli_h1}()},
\code{\link{cli_li}()},
\code{\link{cli_ol}()},
\code{\link{cli_process_start}()},
\code{\link{cli_progress_along}()},
\code{\link{cli_progress_bar}()},
\code{\link{cli_progress_message}()},
\code{\link{cli_progress_output}()},
\code{\link{cli_progress_step}()},
\code{\link{cli_rule}},
\code{\link{cli_status}()},
\code{\link{cli_status_update}()},
\code{\link{cli_text}()},
\code{\link{cli_ul}()},
\code{\link{format_error}()},
\code{\link{format_inline}()}
}
\concept{functions supporting inline markup}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cli.R
\name{cli_par}
\alias{cli_par}
\title{CLI paragraph}
\usage{
cli_par(id = NULL, class = NULL, .auto_close = TRUE, .envir = parent.frame())
}
\arguments{
\item{id}{Element id, a string. If \code{NULL}, then a new id is generated
and returned.}

\item{class}{Class name, sting. Can be used in themes.}

\item{.auto_close}{Whether to close the container, when the calling
function finishes (or \code{.envir} is removed, if specified).}

\item{.envir}{Environment to evaluate the glue expressions in. It is
also used to auto-close the container if \code{.auto_close} is \code{TRUE}.}
}
\value{
The id of the new container element, invisibly.
}
\description{
The builtin theme leaves an empty line between paragraphs.
See also \link{containers}.
}
\details{
\if{html}{\out{<div class="sourceCode r">}}\preformatted{clifun <- function() \{
  cli_par()
  cli_text(cli:::lorem_ipsum())
\}
clifun()
clifun()
}\if{html}{\out{</div>}}\if{html}{\out{
<div class="asciicast" style="color: #172431;font-family: 'Fira Code',Monaco,Consolas,Menlo,'Bitstream Vera Sans Mono','Powerline Symbols',monospace;line-height: 1.300000"><pre>
#> Sunt anim ullamco Lorem qui mollit anim est in deserunt adipisicing.            
#> Enim deserunt laborum ad qui qui. Anim esse non anim magna Lorem                
#> consequat dolore labore cupidatat magna et. Esse nulla eiusmod Lorem            
#> exercitation cupidatat velit enim exercitation excepteur non officia            
#> incididunt. Id laborum dolore commodo Lorem esse ea sint proident.              
#>                                                                                 
#> Fugiat mollit in Lorem velit qui exercitation ipsum consectetur ad              
#> nisi ut eu do ullamco. Mollit officia reprehenderit culpa Lorem est             
#> reprehenderit excepteur enim magna incididunt ea. Irure nisi ad                 
#> exercitation deserunt enim anim excepteur quis minim laboris veniam             
#> nulla pariatur. Enim irure aute nulla irure qui non. Minim velit                
#> proident sunt sint. Proident sit occaecat ex aute.                              
#>                                                                                 
</pre></div>
}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ansiex.R
\name{ansi_trimws}
\alias{ansi_trimws}
\title{Remove leading and/or trailing whitespace from an ANSI string}
\usage{
ansi_trimws(x, which = c("both", "left", "right"))
}
\arguments{
\item{x}{ANSI string vector.}

\item{which}{Whether to remove leading or trailing whitespace or both.}
}
\value{
ANSI string, with the whitespace removed.
}
\description{
This function is similar to \code{\link[base:trimws]{base::trimws()}} but works on ANSI strings,
and keeps color and other styling.
}
\examples{
trimws(paste0("   ", col_red("I am red"), "   "))
ansi_trimws(paste0("   ", col_red("I am red"), "   "))
trimws(col_red("   I am red   "))
ansi_trimws(col_red("   I am red   "))
}
\seealso{
Other ANSI string operations: 
\code{\link{ansi_align}()},
\code{\link{ansi_columns}()},
\code{\link{ansi_nchar}()},
\code{\link{ansi_strsplit}()},
\code{\link{ansi_strtrim}()},
\code{\link{ansi_strwrap}()},
\code{\link{ansi_substring}()},
\code{\link{ansi_substr}()},
\code{\link{ansi_toupper}()}
}
\concept{ANSI string operations}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cli.R
\name{cli_blockquote}
\alias{cli_blockquote}
\title{CLI block quote}
\usage{
cli_blockquote(
  quote,
  citation = NULL,
  id = NULL,
  class = NULL,
  .envir = parent.frame()
)
}
\arguments{
\item{quote}{Text of the quotation.}

\item{citation}{Source of the quotation, typically a link or the name
of a person.}

\item{id}{Element id, a string. If \code{NULL}, then a new id is generated
and returned.}

\item{class}{Class name, sting. Can be used in themes.}

\item{.envir}{Environment to evaluate the glue expressions in. It is
also used to auto-close the container if \code{.auto_close} is \code{TRUE}.}
}
\description{
A section that is quoted from another source. It is typically indented.
}
\details{
\if{html}{\out{<div class="sourceCode r">}}\preformatted{evil <- paste(
  "The real problem is that programmers have spent far too much time",
  "worrying about efficiency in the wrong places and at the wrong",
  "times; premature optimization is the root of all evil (or at least",
  "most of it) in programming.")
cli_blockquote(evil, citation = "Donald Ervin Knuth")
}\if{html}{\out{</div>}}\if{html}{\out{
<div class="asciicast" style="color: #172431;font-family: 'Fira Code',Monaco,Consolas,Menlo,'Bitstream Vera Sans Mono','Powerline Symbols',monospace;line-height: 1.300000"><pre>
#>                                                                                 
#>     <span style="font-style: italic;">“The real problem is that programmers have spent far</span>                        
#>     <span style="font-style: italic;">too much time worrying about efficiency in the wrong</span>                        
#>     <span style="font-style: italic;">places and at the wrong times; premature optimization</span>                       
#>     <span style="font-style: italic;">is the root of all evil (or at least most of it) in</span>                         
#>     <span style="font-style: italic;">programming.”</span>                                                               
#>     <span style="font-weight: bold;font-style: italic;">— Donald Ervin Knuth</span>                                                        
#>                                                                                 
</pre></div>
}}
}
\seealso{
This function supports \link[=inline-markup]{inline markup}.

Other functions supporting inline markup: 
\code{\link{cli_abort}()},
\code{\link{cli_alert}()},
\code{\link{cli_bullets}()},
\code{\link{cli_bullets_raw}()},
\code{\link{cli_dl}()},
\code{\link{cli_h1}()},
\code{\link{cli_li}()},
\code{\link{cli_ol}()},
\code{\link{cli_process_start}()},
\code{\link{cli_progress_along}()},
\code{\link{cli_progress_bar}()},
\code{\link{cli_progress_message}()},
\code{\link{cli_progress_output}()},
\code{\link{cli_progress_step}()},
\code{\link{cli_rule}},
\code{\link{cli_status}()},
\code{\link{cli_status_update}()},
\code{\link{cli_text}()},
\code{\link{cli_ul}()},
\code{\link{format_error}()},
\code{\link{format_inline}()}
}
\concept{functions supporting inline markup}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tty.R
\name{ansi_hide_cursor}
\alias{ansi_hide_cursor}
\alias{ansi_show_cursor}
\alias{ansi_with_hidden_cursor}
\title{Hide/show cursor in a terminal}
\usage{
ansi_hide_cursor(stream = "auto")

ansi_show_cursor(stream = "auto")

ansi_with_hidden_cursor(expr, stream = "auto")
}
\arguments{
\item{stream}{The stream to inspect or manipulate, an R connection
object. It can also be a string, one of \code{"auto"}, \code{"message"},
\code{"stdout"}, \code{"stderr"}. \code{"auto"} will select \code{stdout()} if the session is
interactive and there are no sinks, otherwise it will select \code{stderr()}.}

\item{expr}{R expression to evaluate.}
}
\description{
This only works in terminal emulators. In other environments, it
does nothing.
}
\details{
\code{ansi_hide_cursor()} hides the cursor.

\code{ansi_show_cursor()} shows the cursor.

\code{ansi_with_hidden_cursor()} temporarily hides the cursor for
evaluating an expression.
}
\seealso{
Other terminal capabilities: 
\code{\link{is_ansi_tty}()},
\code{\link{is_dynamic_tty}()}

Other low level ANSI functions: 
\code{\link{ansi_has_any}()},
\code{\link{ansi_regex}()},
\code{\link{ansi_string}()},
\code{\link{ansi_strip}()}
}
\concept{low level ANSI functions}
\concept{terminal capabilities}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/spinner.R
\name{demo_spinners}
\alias{demo_spinners}
\title{Show a demo of some (by default all) spinners}
\usage{
demo_spinners(which = NULL)
}
\arguments{
\item{which}{Character vector, which spinners to demo.}
}
\description{
Each spinner is shown for about 2-3 seconds.
}
\details{
\if{html}{\out{<div class="sourceCode r">}}\preformatted{demo_spinners("clock")
}\if{html}{\out{</div>}}

\if{html}{\figure{demo-spinners.svg}}
}
\seealso{
Other spinners: 
\code{\link{get_spinner}()},
\code{\link{list_spinners}()},
\code{\link{make_spinner}()}
}
\concept{spinners}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cli.R
\name{cli_dl}
\alias{cli_dl}
\title{Definition list}
\usage{
cli_dl(
  items = NULL,
  labels = names(items),
  id = NULL,
  class = NULL,
  .close = TRUE,
  .auto_close = TRUE,
  .envir = parent.frame()
)
}
\arguments{
\item{items}{Named character vector, or \code{NULL}. If not \code{NULL}, they
are used as list items.}

\item{labels}{Item labels. Defaults the names in \code{items}.}

\item{id}{Id of the list container. Can be used for closing it with
\code{\link[=cli_end]{cli_end()}} or in themes. If \code{NULL}, then an id is generated and
returned invisibly.}

\item{class}{Class of the list container. Can be used in themes.}

\item{.close}{Whether to close the list container if the \code{items} were
specified. If \code{FALSE} then new items can be added to the list.}

\item{.auto_close}{Whether to close the container, when the calling
function finishes (or \code{.envir} is removed, if specified).}

\item{.envir}{Environment to evaluate the glue expressions in. It is
also used to auto-close the container if \code{.auto_close} is \code{TRUE}.}
}
\value{
The id of the new container element, invisibly.
}
\description{
A definition list is a container, see \link{containers}.
}
\details{
\subsection{All items at once}{

\if{html}{\out{<div class="sourceCode r">}}\preformatted{fun <- function() \{
  cli_dl(c(foo = "one", bar = "two", baz = "three"))
\}
fun()
}\if{html}{\out{</div>}}\if{html}{\out{
<div class="asciicast" style="color: #172431;font-family: 'Fira Code',Monaco,Consolas,Menlo,'Bitstream Vera Sans Mono','Powerline Symbols',monospace;line-height: 1.300000"><pre>
#> foo: one                                                                        
#> bar: two                                                                        
#> baz: three                                                                      
</pre></div>
}}

}

\subsection{Items one by one}{

\if{html}{\out{<div class="sourceCode r">}}\preformatted{fun <- function() \{
  cli_dl()
  cli_li(c(foo = "\{.emph one\}"))
  cli_li(c(bar = "two"))
  cli_li(c(baz = "three"))
\}
fun()
}\if{html}{\out{</div>}}\if{html}{\out{
<div class="asciicast" style="color: #172431;font-family: 'Fira Code',Monaco,Consolas,Menlo,'Bitstream Vera Sans Mono','Powerline Symbols',monospace;line-height: 1.300000"><pre>
#> foo: <span style="font-style: italic;">one</span>                                                                        
#> bar: two                                                                        
#> baz: three                                                                      
</pre></div>
}}

}
}
\seealso{
This function supports \link[=inline-markup]{inline markup}.

Other functions supporting inline markup: 
\code{\link{cli_abort}()},
\code{\link{cli_alert}()},
\code{\link{cli_blockquote}()},
\code{\link{cli_bullets}()},
\code{\link{cli_bullets_raw}()},
\code{\link{cli_h1}()},
\code{\link{cli_li}()},
\code{\link{cli_ol}()},
\code{\link{cli_process_start}()},
\code{\link{cli_progress_along}()},
\code{\link{cli_progress_bar}()},
\code{\link{cli_progress_message}()},
\code{\link{cli_progress_output}()},
\code{\link{cli_progress_step}()},
\code{\link{cli_rule}},
\code{\link{cli_status}()},
\code{\link{cli_status_update}()},
\code{\link{cli_text}()},
\code{\link{cli_ul}()},
\code{\link{format_error}()},
\code{\link{format_inline}()}
}
\concept{functions supporting inline markup}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cliapp-docs.R
\name{themes}
\alias{themes}
\title{About cli themes}
\description{
CLI elements can be styled via a CSS-like language of selectors and
properties. Only a small subset of CSS3 is supported, and
a lot visual properties cannot be implemented on a terminal, so these
will be ignored as well.
}
\section{Adding themes}{

The style of an element is calculated from themes from four sources.
These form a stack, and the themes on the top of the stack take
precedence, over themes in the bottom.
\enumerate{
\item The cli package has a built-in theme. This is always active.
See \code{\link[=builtin_theme]{builtin_theme()}}.
\item When an app object is created via \code{\link[=start_app]{start_app()}}, the caller can
specify a theme, that is added to theme stack. If no theme is
specified for \code{\link[=start_app]{start_app()}}, the content of the \code{cli.theme} option
is used. Removed when the corresponding app stops.
\item The user may specify a theme in the \code{cli.user_theme} option. This
is added to the stack \emph{after} the app's theme (step 2.), so it can
override its settings. Removed when the app that added it stops.
\item Themes specified explicitly in \code{\link[=cli_div]{cli_div()}} elements. These are
removed from the theme stack, when the corresponding \code{\link[=cli_div]{cli_div()}}
elements are closed.
}
}

\section{Writing themes}{

A theme is a named list of lists. The name of each entry is a CSS
selector. Only a subset of CSS is supported:
\itemize{
\item Type selectors, e.g. \code{input} selects all \verb{<input>} elements.
\item Class selectors, e.g. \code{.index} selects any element that has a class
of "index".
\item ID selector. \verb{#toc} will match the element that has the ID "toc".
\item The descendant combinator, i.e. the space, that selects nodes
that are descendants of the first element. E.g. \verb{div span} will match
all \verb{<span>} elements that are inside a \verb{<div>} element.
}

The content of a theme list entry is another named list, where the
names are CSS properties, e.g. \code{color}, or \code{font-weight} or
\code{margin-left}, and the list entries themselves define the values of
the properties. See \code{\link[=builtin_theme]{builtin_theme()}} and \code{\link[=simple_theme]{simple_theme()}} for examples.
}

\section{Formatter callbacks}{

For flexibility, themes may also define formatter functions, with
property name \code{fmt}. These will be called once the other styles are
applied to an element. They are only called on elements that produce
output, i.e. \emph{not} on container elements.
}

\section{Supported properties}{

Right now only a limited set of properties are supported. These include
left, right, top and bottom margins, background and foreground colors,
bold and italic fonts, underlined text. The \code{before} and \code{after}
properties are supported to insert text before and after the
content of the element.

The current list of properties:
\itemize{
\item \code{after}: A string literal to insert after the element. It can also be
a function that returns a string literal. Supported by all inline
elements, list items, alerts and rules.
\item \code{background-color}: An R color name, or HTML hexadecimal color.
It can be applied to most elements (inline elements, rules, text,
etc.), but the background of containers is not colored properly
currently.
\item \code{before}: A string literal to insert before the element. It can also be
a function that returns a string literal. Supported by all inline
elements, list items, alerts and rules.
\item \code{class-map}: Its value can be a named list, and it specifies how
R (S3) class names are mapped to cli class names. E.g.
\code{list(fs_path = "file")} specifies that \code{fs_path} objects (from the fs
package) should always print as \code{.file} objects in cli.
\item \code{color}: Text color, an R color name or a HTML hexadecimal color. It
can be applied to most elements that are printed.
\item \code{collapse}: Specifies how to collapse a vector, before applying
styling. If a character string, then that is used as the separator.
If a function, then it is called, with the vector as the only
argument.
\item \code{digits}: Number of digits after the decimal point for numeric inline
element of class \code{.val}.
\item \code{fmt}: Generic formatter function that takes an input text and returns
formatted text. Can be applied to most elements. If colors are in use,
the input text provided to \code{fmt} already includes ANSI sequences.
\item \code{font-style}: If \code{"italic"} then the text is printed as cursive.
\item \code{font-weight}: If \code{"bold"}, then the text is printed in boldface.
\item \code{line-type}: Line type for \code{\link[=cli_rule]{cli_rule()}}.
\item \code{list-style-type}: String literal or functions that returns a string
literal, to be used as a list item marker in un-ordered lists.
\item \code{margin-bottom}, \code{margin-left}, \code{margin-right}, \code{margin-top}: Margins.
\item \code{padding-left}, \code{padding-right}: This is currently used the same way
as the margins, but this might change later.
\item \code{start}: Integer number, the first element in an ordered list.
\item \code{string-quote}: Quoting character for inline elements of class \code{.val}.
\item \code{text-decoration}: If \code{"underline"}, then underlined text is created.
\item \code{text-exdent}: Amount of indentation from the second line of wrapped
text.
\item \code{transform}: A function to call on glue substitutions, before
collapsing them. Note that \code{transform} is applied prior to
implementing color via ANSI sequences.
\item \code{vec-last}: The last separator when collapsing vectors.
\item \code{vec-sep}: The separator to use when collapsing vectors.
\item \code{vec-sep2}: The separator to use for two elements when collapsing
vectors. If not set, then \code{vec-sep} is used for these as well.
\item \code{vec-trunc}: Vectors longer than this will be truncated. Defaults to
100.
\item \code{vec-trunc-style}: Select between two ways of collapsing vectors:
\itemize{
\item \code{"both-ends"} is the current default and it shows the beginning and
the end of the vector.
\item \code{"head"} only shows the beginning of the vector.
}
}

More properties might be added later. If you think that a property is
not applied properly to an element, please open an issue about it in
the cli issue tracker.
}

\section{Examples}{

Color of headings, that are only active in paragraphs with an
'output' class:

\if{html}{\out{<div class="sourceCode">}}\preformatted{list(
  "par.output h1" = list("background-color" = "red", color = "#e0e0e0"),
  "par.output h2" = list("background-color" = "orange", color = "#e0e0e0"),
  "par.output h3" = list("background-color" = "blue", color = "#e0e0e0")
)
}\if{html}{\out{</div>}}

Create a custom alert type:

\if{html}{\out{<div class="sourceCode">}}\preformatted{list(
  ".alert-start" = list(before = symbol$play),
  ".alert-stop"  = list(before = symbol$stop)
)
}\if{html}{\out{</div>}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/progress-client.R
\name{cli_progress_message}
\alias{cli_progress_message}
\title{Simplified cli progress messages}
\usage{
cli_progress_message(
  msg,
  current = TRUE,
  .auto_close = TRUE,
  .envir = parent.frame(),
  ...
)
}
\arguments{
\item{msg}{Message to show. It may contain glue substitution and cli
styling. It can be updated via \code{\link[=cli_progress_update]{cli_progress_update()}}, as usual.}

\item{current}{Passed to \code{\link[=cli_progress_bar]{cli_progress_bar()}}.}

\item{.auto_close}{Passed to \code{\link[=cli_progress_bar]{cli_progress_bar()}}.}

\item{.envir}{Passed to \code{\link[=cli_progress_bar]{cli_progress_bar()}}.}

\item{...}{Passed to \code{\link[=cli_progress_bar]{cli_progress_bar()}}.}
}
\value{
The id of the new progress bar.
}
\description{
This is a simplified progress bar, a single (dynamic)
message, without progress units.
}
\details{
\code{cli_progress_message()} always shows the message, even if no
update is due. When the progress message is terminated, it is removed
from the screen by default.

Note that the message can be dynamic: if you update it with
\code{\link[=cli_progress_update]{cli_progress_update()}}, then cli uses the current values in the string
substitutions.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{fun <- function() \{
  cli_progress_message("Task one is running...")
  Sys.sleep(2)

  cli_progress_message("Task two is running...")
  Sys.sleep(2)

  step <- 1L
  cli_progress_message("Task three is underway: step \{step\}")
  for (step in 1:5) \{
    Sys.sleep(0.5)
    cli_progress_update()
  \}
\}
fun()
}\if{html}{\out{</div>}}

\if{html}{\figure{progress-message.svg}}
}
\seealso{
This function supports \link[=inline-markup]{inline markup}.

\code{\link[=cli_progress_bar]{cli_progress_bar()}} for the complete progress bar API.
\code{\link[=cli_progress_step]{cli_progress_step()}} for a similar display that is styled by default.

Other progress bar functions: 
\code{\link{cli_progress_along}()},
\code{\link{cli_progress_bar}()},
\code{\link{cli_progress_builtin_handlers}()},
\code{\link{cli_progress_num}()},
\code{\link{cli_progress_output}()},
\code{\link{cli_progress_step}()},
\code{\link{cli_progress_styles}()},
\code{\link{progress-variables}}

Other functions supporting inline markup: 
\code{\link{cli_abort}()},
\code{\link{cli_alert}()},
\code{\link{cli_blockquote}()},
\code{\link{cli_bullets}()},
\code{\link{cli_bullets_raw}()},
\code{\link{cli_dl}()},
\code{\link{cli_h1}()},
\code{\link{cli_li}()},
\code{\link{cli_ol}()},
\code{\link{cli_process_start}()},
\code{\link{cli_progress_along}()},
\code{\link{cli_progress_bar}()},
\code{\link{cli_progress_output}()},
\code{\link{cli_progress_step}()},
\code{\link{cli_rule}},
\code{\link{cli_status}()},
\code{\link{cli_status_update}()},
\code{\link{cli_text}()},
\code{\link{cli_ul}()},
\code{\link{format_error}()},
\code{\link{format_inline}()}
}
\concept{functions supporting inline markup}
\concept{progress bar functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/glue.R
\name{ansi_collapse}
\alias{ansi_collapse}
\title{Collapse a vector into a string scalar}
\usage{
ansi_collapse(
  x,
  sep = ", ",
  sep2 = sub("^,", "", last),
  last = ", and ",
  trunc = Inf,
  width = Inf,
  ellipsis = symbol$ellipsis,
  style = c("both-ends", "head")
)
}
\arguments{
\item{x}{Character vector, or an object with an \code{as.character()} method
to collapse.}

\item{sep}{Separator. A character string.}

\item{sep2}{Separator for the special case that \code{x} contains only two
elements. A character string. Defaults to the value of \code{last} without the
serial comma.}

\item{last}{Last separator, if there is no truncation. E.g. use
\code{", and "} for the \href{https://en.wikipedia.org/wiki/Serial_comma}{serial comma}. A character string.}

\item{trunc}{Maximum number of elements to show. For \code{style = "head"}
at least \code{trunc = 1} is used. For \code{style = "both-ends"} at least
\code{trunc = 5} is used, even if a smaller number is specified.}

\item{width}{Limit for the display width of the result, in characters.
This is a hard limit, and the output will never exceed it.
This argument is not implemented for the \code{"both-ends"} style, which
always uses \code{Inf}, with a warning if a finite \code{width} value is set.}

\item{ellipsis}{Character string to use at the place of the truncation.
By default, the Unicode ellipsis character is used if the console is
UTF-8, and three dots otherwise.}

\item{style}{Truncation style:
\itemize{
\item \code{both-ends}: the default, shows the beginning and end of the vector,
and skips elements in the middle if needed.
\item \code{head}: shows the beginning of the vector, and skips elements at the
end, if needed.
}}
}
\value{
Character scalar. It is \code{NA_character_} if any elements in \code{x}
are \code{NA}.
}
\description{
Features:
\itemize{
\item custom separator (\code{sep}),
\item custom separator for length-two input (\code{sep2}),
\item custom last separator (\code{last}),
\item adds ellipsis to truncated strings,
\item uses Unicode ellipsis character on UTF-8 console,
\item can collapse "from both ends", with \code{style = "both-ends"},
\item can consider a limit for the display width of the result, in characters,
\item handles ANSI control sequences correctly when measuring display width.
}
}
\examples{
ansi_collapse(letters)

# truncate
ansi_collapse(letters, trunc = 5)

# head style
ansi_collapse(letters, trunc = 5, style = "head")
}
\seealso{
\code{glue_collapse} in the glue package inspired this function.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/num-ansi-colors.R
\name{num_ansi_colors}
\alias{num_ansi_colors}
\alias{detect_tty_colors}
\title{Detect the number of ANSI colors to use}
\usage{
num_ansi_colors(stream = "auto")

detect_tty_colors()
}
\arguments{
\item{stream}{The stream that will be used for output, an R connection
object. It can also be a string, one of \code{"auto"}, \code{"message"},
\code{"stdout"}, \code{"stderr"}. \code{"auto"} will select \code{stdout()} if the session is
interactive and there are no sinks, otherwise it will select \code{stderr()}.}
}
\value{
Integer, the number of ANSI colors the current R session
supports for \code{stream}.
}
\description{
Certain Unix and Windows terminals, and also certain R GUIs, e.g.
RStudio, support styling terminal output using special control
sequences (ANSI sequences).

\code{num_ansi_colors()} detects if the current R session supports ANSI
sequences, and if it does how many colors are supported.
}
\details{
The detection mechanism is quite involved and it is designed to work
out of the box on most systems. If it does not work on your system,
please report a bug. Setting options and environment variables to turn
on ANSI support is error prone, because they are inherited in other
environments, e.g. knitr, that might not have ANSI support.

If you want to \emph{turn off} ANSI colors, set the \code{NO_COLOR} environment
variable to a non-empty value.

The exact detection mechanism is as follows:
\enumerate{
\item If the \code{cli.num_colors} options is set, that is returned.
\item If the \code{R_CLI_NUM_COLORS} environment variable is set to a
non-empty value, then it is used.
\item If the \code{crayon.enabled} option is set to \code{FALSE}, 1L is returned.
(This is for compatibility with code that uses the crayon package.)
\item If the \code{crayon.enabled} option is set to \code{TRUE} and the
\code{crayon.colors} option is not set, then the value of the
\code{cli.default_num_colors} option, or if it is unset, then 8L is
returned.
\item If the \code{crayon.enabled} option is set to \code{TRUE} and the
\code{crayon.colors} option is also set, then the latter is returned.
(This is for compatibility with code that uses the crayon package.)
\item If the \code{NO_COLOR} environment variable is set, then 1L is returned.
\item If we are in knitr, then 1L is returned, to turn off colors in
\code{.Rmd} chunks.
\item If \code{stream} is \code{"auto"} (the default) and there is an active
sink (either for \code{"output"} or \code{"message"}), then we return 1L.
(In theory we would only need to check the stream that will be
be actually used, but there is no easy way to tell that.)
\item If \code{stream} is not \code{"auto"}, but it is \code{stderr()} and there is an
active sink for it, then 1L is returned.
(If a sink is active for "output", then R changes the \code{stdout()}
stream, so this check is not needed.)
\item If the \code{cli.default_num_colors} option is set, then we use that.
\item If R is running inside RGui on Windows, or R.app on macOS, then we
return 1L.
\item If R is running inside RStudio, with color support, then the
appropriate number of colors is returned, usually 256L.
\item If R is running on Windows, inside an Emacs version that is recent
enough to support ANSI colors, then the value of the
\code{cli.default_num_colors} option, or if unset 8L is returned.
(On Windows, Emacs has \code{isatty(stdout()) == FALSE}, so we need to
check for this here before dealing with terminals.)
\item If \code{stream} is not the standard output or standard error  in a
terminal, then 1L is returned.
\item Otherwise we use and cache the result of the terminal color
detection (see below).
}

The terminal color detection algorithm:
\enumerate{
\item If the \code{COLORTERM} environment variable is set to \code{truecolor} or
\verb{24bit}, then we return 16 million colors.
\item If the \code{COLORTERM} environment variable is set to anything else,
then we return the value of the \code{cli.num_default_colors} option,
8L if unset.
\item If R is running on Unix, inside an Emacs version that is recent
enough to support ANSI colors, then the value of the
\code{cli.default_num_colors} option is returned, or 8L if unset.
\item If we are on Windows in an RStudio terminal, then apparently
we only have eight colors, but the \code{cli.default_num_colors} option
can be used to override this.
\item If we are in a recent enough Windows 10 terminal, then there
is either true color (from build 14931) or 256 color (from
build 10586) support. You can also use the \code{cli.default_num_colors}
option to override these.
\item If we are on Windows, under ConEmu or cmder, or ANSICON is loaded,
then the value of \code{cli.default_num_colors}, or 8L if unset, is
returned.
\item Otherwise if we are on Windows, return 1L.
\item Otherwise we are on Unix and try to run \verb{tput colors} to determine
the number of colors. If this succeeds, we return its return value.
If the \code{TERM} environment variable is \code{xterm} and \code{tput}
returned 8L, we return 256L, because xterm compatible terminals
tend to support 256 colors
(\url{https://github.com/r-lib/crayon/issues/17})
You can override this with the \code{cli.default_num_colors} option.
\item If \code{TERM} is set to \code{dumb}, we return 1L.
\item If \code{TERM} starts with \code{screen}, \code{xterm}, or \code{vt100}, we return 8L.
\item If \code{TERM} contains \code{color}, \code{ansi}, \code{cygwin} or \code{linux}, we return 8L.
\item Otherwise we return 1L.
}
}
\examples{
num_ansi_colors()

}
\seealso{
Other ANSI styling: 
\code{\link{ansi-styles}},
\code{\link{combine_ansi_styles}()},
\code{\link{make_ansi_style}()}
}
\concept{ANSI styling}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pluralize.R
\name{pluralize}
\alias{pluralize}
\title{String templating with pluralization}
\usage{
pluralize(
  ...,
  .envir = parent.frame(),
  .transformer = glue::identity_transformer
)
}
\arguments{
\item{..., .envir, .transformer}{All arguments are passed to \code{\link[glue:glue]{glue::glue()}}.}
}
\description{
\code{pluralize()} is similar to \code{\link[glue:glue]{glue::glue()}}, with two differences:
\itemize{
\item It supports cli's \link{pluralization} syntax, using \verb{\{?\}} markers.
\item It collapses substituted vectors into a comma separated string.
}
}
\details{
See \link{pluralization} and some examples below.

You need to install the glue package to use this function.
}
\examples{
\dontshow{if (requireNamespace("glue", quietly = TRUE)) (if (getRversion() >= "3.4") withAutoprint else force)(\{ # examplesIf}
# Regular plurals
nfile <- 0; pluralize("Found {nfile} file{?s}.")
nfile <- 1; pluralize("Found {nfile} file{?s}.")
nfile <- 2; pluralize("Found {nfile} file{?s}.")

# Irregular plurals
ndir <- 1; pluralize("Found {ndir} director{?y/ies}.")
ndir <- 5; pluralize("Found {ndir} director{?y/ies}.")

# Use 'no' instead of zero
nfile <- 0; pluralize("Found {no(nfile)} file{?s}.")
nfile <- 1; pluralize("Found {no(nfile)} file{?s}.")
nfile <- 2; pluralize("Found {no(nfile)} file{?s}.")

# Use the length of character vectors
pkgs <- "pkg1"
pluralize("Will remove the {pkgs} package{?s}.")
pkgs <- c("pkg1", "pkg2", "pkg3")
pluralize("Will remove the {pkgs} package{?s}.")

pkgs <- character()
pluralize("Will remove {?no/the/the} {pkgs} package{?s}.")
pkgs <- c("pkg1", "pkg2", "pkg3")
pluralize("Will remove {?no/the/the} {pkgs} package{?s}.")

# Multiple quantities
nfiles <- 3; ndirs <- 1
pluralize("Found {nfiles} file{?s} and {ndirs} director{?y/ies}")

# Explicit quantities
nupd <- 3; ntotal <- 10
cli_text("{nupd}/{ntotal} {qty(nupd)} file{?s} {?needs/need} updates")
\dontshow{\}) # examplesIf}
}
\seealso{
Other pluralization: 
\code{\link{no}()},
\code{\link{pluralization}}
}
\concept{pluralization}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/test.R
\name{test_that_cli}
\alias{test_that_cli}
\title{Test cli output with testthat}
\usage{
test_that_cli(
  desc,
  code,
  configs = c("plain", "ansi", "unicode", "fancy"),
  links = NULL
)
}
\arguments{
\item{desc}{Test description, passed to \code{\link[testthat:test_that]{testthat::test_that()}}, after
appending the name of the cli configuration to it.}

\item{code}{Test code, it is modified to set up the cli config, and
then passed to \code{\link[testthat:test_that]{testthat::test_that()}}}

\item{configs}{cli configurations to test \code{code} with. The default is
\code{NULL}, which includes all possible configurations. It can also be a
character vector, to restrict the tests to some configurations only.
See available configurations below.}

\item{links}{Whether to run the code with various hyperlinks allowed.
If \code{NULL} then hyperlinks are turned off. Otherwise it can be a character
vector with possible hyperlink configurations:
\itemize{
\item \code{"all"}: turn on all hyperlinks,
\item \code{"none"}: turn off all hyperlinks.
}}
}
\description{
Use this function in your testthat test files, to test cli output.
It requires testthat edition 3, and works best with snapshot tests.
}
\details{
\code{test_that_cli()} calls \code{\link[testthat:test_that]{testthat::test_that()}} multiple times, with
different cli configurations. This makes it simple to test cli output
with and without ANSI colors, with and without Unicode characters.

Currently available configurations:
\itemize{
\item \code{plain}: no ANSI colors, ASCII characters only.
\item \code{ansi}: ANSI colors, ASCII characters only.
\item \code{unicode}: no ANSI colors, Unicode characters.
\item \code{fancy}; ANSI colors, Unicode characters.
}

See examples below and in cli's own tests, e.g. in
\url{https://github.com/r-lib/cli/tree/main/tests/testthat}
and the corresponding snapshots at
\url{https://github.com/r-lib/cli/tree/main/tests/testthat/_snaps}
\subsection{Important note regarding Windows}{

Because of base R's limitation to record Unicode characters on Windows,
we suggest that you record your snapshots on Unix, or you restrict
your tests to ASCII configurations.

Unicode tests on Windows are automatically skipped by testthat
currently.
}
}
\examples{
# testthat cannot record or compare snapshots when you run these
# examples interactively, so you might want to copy them into a test
# file

# Default configurations
cli::test_that_cli("success", {
  testthat::local_edition(3)
  testthat::expect_snapshot({
    cli::cli_alert_success("wow")
  })
})

# Only use two configurations, because this output does not have colors
cli::test_that_cli(configs = c("plain", "unicode"), "cat_bullet", {
  testthat::local_edition(3)
  testthat::expect_snapshot({
    cli::cat_bullet(letters[1:5])
  })
})

# You often need to evaluate all cli calls of a test case in the same
# environment. Use `local()` to do that:
cli::test_that_cli("theming", {
  testthat::local_edition(3)
  testthat::expect_snapshot(local({
    cli::cli_div(theme = list(".alert" = list(before = "!!! ")))
    cli::cli_alert("wow")
  }))
})
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pluralize.R
\name{pluralization}
\alias{pluralization}
\title{About cli pluralization}
\description{
About cli pluralization
}
\section{Introduction}{
cli has tools to create messages that are printed correctly in singular
and plural forms. This usually requires minimal extra work, and
increases the quality of the messages greatly. In this document we first
show some pluralization examples that you can use as guidelines.
Hopefully these are intuitive enough, so that they can be used without
knowing the exact cli pluralization rules.

If you need pluralization without the semantic cli functions, see the
\code{pluralize()} function.
}

\section{Examples}{
\subsection{Pluralization markup}{

In the simplest case the message contains a single \code{{}} glue
substitution, which specifies the quantity that is used to select
between the singular and plural forms. Pluralization uses markup that is
similar to glue, but uses the \verb{\{?} and \verb{\}} delimiters:\if{html}{\out{<div class="sourceCode r">}}\preformatted{library(cli)
nfile <- 0; cli_text("Found \{nfile\} file\{?s\}.")
}\if{html}{\out{</div>}}\preformatted{#> Found 0 files.
}\if{html}{\out{<div class="sourceCode r">}}\preformatted{nfile <- 1; cli_text("Found \{nfile\} file\{?s\}.")
}\if{html}{\out{</div>}}\preformatted{#> Found 1 file.
}\if{html}{\out{<div class="sourceCode r">}}\preformatted{nfile <- 2; cli_text("Found \{nfile\} file\{?s\}.")
}\if{html}{\out{</div>}}\preformatted{#> Found 2 files.
}

Here the value of \code{nfile} is used to decide whether the singular or
plural form of \code{file} is used. This is the most common case for English
messages.
}

\subsection{Irregular plurals}{

If the plural form is more difficult than a simple \code{s} suffix, then the
singular and plural forms can be given, separated with a forward slash:\if{html}{\out{<div class="sourceCode r">}}\preformatted{ndir <- 1; cli_text("Found \{ndir\} director\{?y/ies\}.")
}\if{html}{\out{</div>}}\preformatted{#> Found 1 directory.
}\if{html}{\out{<div class="sourceCode r">}}\preformatted{ndir <- 5; cli_text("Found \{ndir\} director\{?y/ies\}.")
}\if{html}{\out{</div>}}\preformatted{#> Found 5 directories.
}
}

\subsection{Use \code{"no"} instead of zero}{

For readability, it is better to use the \code{no()} helper function to
include a count in a message. \code{no()} prints the word \code{"no"} if the count
is zero, and prints the numeric count otherwise:\if{html}{\out{<div class="sourceCode r">}}\preformatted{nfile <- 0; cli_text("Found \{no(nfile)\} file\{?s\}.")
}\if{html}{\out{</div>}}\preformatted{#> Found no files.
}\if{html}{\out{<div class="sourceCode r">}}\preformatted{nfile <- 1; cli_text("Found \{no(nfile)\} file\{?s\}.")
}\if{html}{\out{</div>}}\preformatted{#> Found 1 file.
}\if{html}{\out{<div class="sourceCode r">}}\preformatted{nfile <- 2; cli_text("Found \{no(nfile)\} file\{?s\}.")
}\if{html}{\out{</div>}}\preformatted{#> Found 2 files.
}
}

\subsection{Use the length of character vectors}{

With the auto-collapsing feature of cli it is easy to include a list of
objects in a message. When cli interprets a character vector as a
pluralization quantity, it takes the length of the vector:\if{html}{\out{<div class="sourceCode r">}}\preformatted{pkgs <- "pkg1"
cli_text("Will remove the \{.pkg \{pkgs\}\} package\{?s\}.")
}\if{html}{\out{</div>}}\preformatted{#> Will remove the pkg1 package.
}\if{html}{\out{<div class="sourceCode r">}}\preformatted{pkgs <- c("pkg1", "pkg2", "pkg3")
cli_text("Will remove the \{.pkg \{pkgs\}\} package\{?s\}.")
}\if{html}{\out{</div>}}\preformatted{#> Will remove the pkg1, pkg2, and pkg3 packages.
}

Note that the length is only used for non-numeric vectors (when
\code{is.numeric(x)} return \code{FALSE}). If you want to use the length of a
numeric vector, convert it to character via \code{as.character()}.

You can combine collapsed vectors with \code{"no"}, like this:\if{html}{\out{<div class="sourceCode r">}}\preformatted{pkgs <- character()
cli_text("Will remove \{?no/the/the\} \{.pkg \{pkgs\}\} package\{?s\}.")
}\if{html}{\out{</div>}}\preformatted{#> Will remove no packages.
}\if{html}{\out{<div class="sourceCode r">}}\preformatted{pkgs <- c("pkg1", "pkg2", "pkg3")
cli_text("Will remove \{?no/the/the\} \{.pkg \{pkgs\}\} package\{?s\}.")
}\if{html}{\out{</div>}}\preformatted{#> Will remove the pkg1, pkg2, and pkg3 packages.
}

When the pluralization markup contains three alternatives, like above,
the first one is used for zero, the second for one, and the third one
for larger quantities.
}

\subsection{Choosing the right quantity}{

When the text contains multiple glue \code{{}} substitutions, the one right
before the pluralization markup is used. For example:\if{html}{\out{<div class="sourceCode r">}}\preformatted{nfiles <- 3; ndirs <- 1
cli_text("Found \{nfiles\} file\{?s\} and \{ndirs\} director\{?y/ies\}")
}\if{html}{\out{</div>}}\preformatted{#> Found 3 files and 1 directory
}

This is sometimes not the the correct one. You can explicitly specify
the correct quantity using the \code{qty()} function. This sets that quantity
without printing anything:\if{html}{\out{<div class="sourceCode r">}}\preformatted{nupd <- 3; ntotal <- 10
cli_text("\{nupd\}/\{ntotal\} \{qty(nupd)\} file\{?s\} \{?needs/need\} updates")
}\if{html}{\out{</div>}}\preformatted{#> 3/10 files need updates
}

Note that if the message only contains a single \code{{}} substitution, then
this may appear before or after the pluralization markup. If the message
contains multiple \code{{}} substitutions \emph{after} pluralization markup, an
error is thrown.

Similarly, if the message contains no \code{{}} substitutions at all, but has
pluralization markup, an error is thrown.
}
}

\section{Rules}{
The exact rules of cli pluralization. There are two sets of rules. The
first set specifies how a quantity is associated with a \verb{\{?\}}
pluralization markup. The second set describes how the \verb{\{?\}} is parsed
and interpreted.
\subsection{Quantities}{
\enumerate{
\item \code{{}} substitutions define quantities. If the value of a \code{{}}
substitution is numeric (when \code{is.numeric(x)} holds), then it has to
have length one to define a quantity. This is only enforced if the
\code{{}} substitution is used for pluralization. The quantity is defined
as the value of \code{{}} then, rounded with \code{as.integer()}. If the value
of \code{{}} is not numeric, then its quantity is defined as its length.
\item If a message has \verb{\{?\}} markup but no \code{{}} substitution, an error is
thrown.
\item If a message has exactly one \code{{}} substitution, its value is used as
the pluralization quantity for all \verb{\{?\}} markup in the message.
\item If a message has multiple \code{{}} substitutions, then for each \verb{\{?\}}
markup cli uses the quantity of the \code{{}} substitution that precedes
it.
\item If a message has multiple \code{{}} substitutions and has pluralization
markup without a preceding \code{{}} substitution, an error is thrown.
}
}

\subsection{Pluralization markup}{
\enumerate{
\item Pluralization markup starts with \verb{\{?} and ends with \verb{\}}. It may not
contain \verb{\{} and \verb{\}} characters, so it may not contain \code{{}}
substitutions either.
\item Alternative words or suffixes are separated by \code{/}.
\item If there is a single alternative, then \emph{nothing} is used if
\code{quantity == 1} and this single alternative is used if
\code{quantity != 1}.
\item If there are two alternatives, the first one is used for
\code{quantity == 1}, the second one for \code{quantity != 1} (including
`\code{quantity == 0}).
\item If there are three alternatives, the first one is used for
\code{quantity == 0}, the second one for \code{quantity == 1}, and the third
one otherwise.
}
}
}

\seealso{
Other pluralization: 
\code{\link{no}()},
\code{\link{pluralize}()}
}
\concept{pluralization}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/progress-bar.R
\name{cli_progress_styles}
\alias{cli_progress_styles}
\title{List of built-in cli progress styles}
\usage{
cli_progress_styles()
}
\value{
A named list with sublists containing elements
\code{complete}, \code{incomplete} and potentially \code{current}.
}
\description{
The following options are used to select a style:
\itemize{
\item \code{cli_progress_bar_style}
\item \code{cli_progress_bar_style_ascii}
\item \code{cli_progress_bar_style_unicode}
}
}
\details{
On Unicode terminals (if \code{\link[=is_utf8_output]{is_utf8_output()}} is \code{TRUE}), the
\code{cli_progress_bar_style_unicode} and \code{cli_progress_bar_style}
options are used.

On ASCII terminals (if \code{\link[=is_utf8_output]{is_utf8_output()}} is \code{FALSE}), the
\code{cli_pgoress_bar_style_ascii} and \code{cli_progress_bar_style} options
are are used.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{for (style in names(cli_progress_styles())) \{
  options(cli.progress_bar_style = style)
  label <- ansi_align(paste0("Style '", style, "'"), 20)
  print(cli_progress_demo(label, live = FALSE, at = 66, total = 100))
\}
options(cli.progress_var_style = NULL)
}\if{html}{\out{</div>}}\if{html}{\out{
<div class="asciicast" style="color: #172431;font-family: 'Fira Code',Monaco,Consolas,Menlo,'Bitstream Vera Sans Mono','Powerline Symbols',monospace;line-height: 1.300000"><pre>
#> Style 'classic'      <span style="color: #859900;">#####################           </span>  66&#37; | ETA:  3s           
#> Style 'squares'      <span style="color: #859900;">■■■■■■■■■■■■■■■■■■■■■           </span>  66&#37; | ETA:  3s           
#> Style 'dot'          <span style="color: #002B36;">────────────────────</span><span style="color: #DC322F;">●</span><span style="color: #002B36;">──────────</span><span style="color: #859900;"> </span>  66&#37; | ETA:  3s           
#> Style 'fillsquares'  <span style="color: #859900;">■■■■■■■■■■■■■■■■■■■■■</span><span style="color: #002B36;">□□□□□□□□□□</span><span style="color: #859900;"> </span>  66&#37; | ETA:  3s           
#> Style 'bar'          <span style="color: #859900;">█████████████████████</span><span style="color: #002B36;">██████████</span><span style="color: #859900;"> </span>  66&#37; | ETA:  3s           
</pre></div>
}}
}
\seealso{
Other progress bar functions: 
\code{\link{cli_progress_along}()},
\code{\link{cli_progress_bar}()},
\code{\link{cli_progress_builtin_handlers}()},
\code{\link{cli_progress_message}()},
\code{\link{cli_progress_num}()},
\code{\link{cli_progress_output}()},
\code{\link{cli_progress_step}()},
\code{\link{progress-variables}}
}
\concept{progress bar functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/format.R
\name{cli_format}
\alias{cli_format}
\alias{cli_format.default}
\alias{cli_format.character}
\alias{cli_format.numeric}
\title{Format a value for printing}
\usage{
cli_format(x, style = NULL, ...)

\method{cli_format}{default}(x, style = NULL, ...)

\method{cli_format}{character}(x, style = NULL, ...)

\method{cli_format}{numeric}(x, style = NULL, ...)
}
\arguments{
\item{x}{The object to format.}

\item{style}{List of formatting options, see the individual methods
for the style options they support.}

\item{...}{Additional arguments for methods.}
}
\description{
This function can be used directly, or via the \verb{\{.val ...\}} inline
style. \verb{\{.val \{expr\}\}} calls \code{cli_format()} automatically on the value
of \code{expr}, before styling and collapsing it.
}
\details{
\subsection{Default style}{

\if{html}{\out{<div class="sourceCode r">}}\preformatted{months <- month.name[1:3]
cli_text("\{.val \{months\}\}")
}\if{html}{\out{</div>}}\if{html}{\out{
<div class="asciicast" style="color: #172431;font-family: 'Fira Code',Monaco,Consolas,Menlo,'Bitstream Vera Sans Mono','Powerline Symbols',monospace;line-height: 1.300000"><pre>
#> <span style="color: #268BD2;">"January"</span>, <span style="color: #268BD2;">"February"</span>, and <span style="color: #268BD2;">"March"</span>                                              
</pre></div>
}}


\if{html}{\out{<div class="sourceCode r">}}\preformatted{nums <- 1:5 / 7
cli_text("\{.val \{nums\}\}")
}\if{html}{\out{</div>}}\if{html}{\out{
<div class="asciicast" style="color: #172431;font-family: 'Fira Code',Monaco,Consolas,Menlo,'Bitstream Vera Sans Mono','Powerline Symbols',monospace;line-height: 1.300000"><pre>
#> <span style="color: #268BD2;">0.142857142857143</span>, <span style="color: #268BD2;">0.285714285714286</span>, <span style="color: #268BD2;">0.428571428571429</span>,                        
#> <span style="color: #268BD2;">0.571428571428571</span>, and <span style="color: #268BD2;">0.714285714285714</span>                                        
</pre></div>
}}

}

\subsection{Styling with themes}{

\if{html}{\out{<div class="sourceCode r">}}\preformatted{nums <- 1:5 / 7
divid <- cli_div(theme = list(.val = list(digits = 3)))
cli_text("\{.val \{nums\}\}")
cli_end(divid)
}\if{html}{\out{</div>}}\if{html}{\out{
<div class="asciicast" style="color: #172431;font-family: 'Fira Code',Monaco,Consolas,Menlo,'Bitstream Vera Sans Mono','Powerline Symbols',monospace;line-height: 1.300000"><pre>
#> <span style="color: #268BD2;">0.143</span>, <span style="color: #268BD2;">0.286</span>, <span style="color: #268BD2;">0.429</span>, <span style="color: #268BD2;">0.571</span>, and <span style="color: #268BD2;">0.714</span>                                           
</pre></div>
}}


It is possible to define new S3 methods for \code{cli_format} and then
these will be used automatically for \verb{\{.val ...\}} expressions.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{cli_format.month <- function(x, style = NULL, ...) \{
  x <- encodeString(substr(x, 1, 3), quote = "\\"")
  NextMethod("cli_format")
\}
registerS3method("cli_format", "month", cli_format.month)
months <- structure(month.name[1:3], class = "month")
cli_text("\{.val \{months\}\}")
}\if{html}{\out{</div>}}\if{html}{\out{
<div class="asciicast" style="color: #172431;font-family: 'Fira Code',Monaco,Consolas,Menlo,'Bitstream Vera Sans Mono','Powerline Symbols',monospace;line-height: 1.300000"><pre>
#> <span style="color: #268BD2;">"Jan"</span>, <span style="color: #268BD2;">"Feb"</span>, and <span style="color: #268BD2;">"Mar"</span>                                                         
</pre></div>
}}

}
}
\seealso{
\code{\link[=cli_vec]{cli_vec()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tty.R
\name{is_dynamic_tty}
\alias{is_dynamic_tty}
\title{Detect whether a stream supports \verb{\\\\r} (Carriage return)}
\usage{
is_dynamic_tty(stream = "auto")
}
\arguments{
\item{stream}{The stream to inspect or manipulate, an R connection
object. It can also be a string, one of \code{"auto"}, \code{"message"},
\code{"stdout"}, \code{"stderr"}. \code{"auto"} will select \code{stdout()} if the session is
interactive and there are no sinks, otherwise it will select \code{stderr()}.}
}
\description{
In a terminal, \verb{\\\\r} moves the cursor to the first position of the
same line. It is also supported by most R IDEs. \verb{\\\\r} is typically
used to achieve a more dynamic, less cluttered user interface, e.g.
to create progress bars.
}
\details{
If the output is directed to a file, then \verb{\\\\r} characters are typically
unwanted. This function detects if \verb{\\\\r} can be used for the given
stream or not.

The detection mechanism is as follows:
\enumerate{
\item If the \code{cli.dynamic} option is set to \code{TRUE}, \code{TRUE} is returned.
\item If the \code{cli.dynamic} option is set to anything else, \code{FALSE} is
returned.
\item If the \code{R_CLI_DYNAMIC} environment variable is not empty and set to
the string \code{"true"}, \code{"TRUE"} or \code{"True"}, \code{TRUE} is returned.
\item If \code{R_CLI_DYNAMIC} is not empty and set to anything else, \code{FALSE} is
returned.
\item If the stream is a terminal, then \code{TRUE} is returned.
\item If the stream is the standard output or error within RStudio,
the macOS R app, or RKWard IDE, \code{TRUE} is returned.
\item Otherwise \code{FALSE} is returned.
}
}
\examples{
is_dynamic_tty()
is_dynamic_tty(stdout())
}
\seealso{
Other terminal capabilities: 
\code{\link{ansi_hide_cursor}()},
\code{\link{is_ansi_tty}()}
}
\concept{terminal capabilities}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/zzz.R
\name{ansi-styles}
\alias{ansi-styles}
\alias{bg_black}
\alias{bg_blue}
\alias{bg_cyan}
\alias{bg_green}
\alias{bg_magenta}
\alias{bg_red}
\alias{bg_white}
\alias{bg_yellow}
\alias{bg_none}
\alias{bg_br_black}
\alias{bg_br_blue}
\alias{bg_br_cyan}
\alias{bg_br_green}
\alias{bg_br_magenta}
\alias{bg_br_red}
\alias{bg_br_white}
\alias{bg_br_yellow}
\alias{col_black}
\alias{col_blue}
\alias{col_cyan}
\alias{col_green}
\alias{col_magenta}
\alias{col_red}
\alias{col_white}
\alias{col_yellow}
\alias{col_grey}
\alias{col_silver}
\alias{col_none}
\alias{col_br_black}
\alias{col_br_blue}
\alias{col_br_cyan}
\alias{col_br_green}
\alias{col_br_magenta}
\alias{col_br_red}
\alias{col_br_white}
\alias{col_br_yellow}
\alias{style_dim}
\alias{style_blurred}
\alias{style_bold}
\alias{style_hidden}
\alias{style_inverse}
\alias{style_italic}
\alias{style_reset}
\alias{style_strikethrough}
\alias{style_underline}
\alias{style_no_bold}
\alias{style_no_blurred}
\alias{style_no_dim}
\alias{style_no_italic}
\alias{style_no_underline}
\alias{style_no_inverse}
\alias{style_no_hidden}
\alias{style_no_strikethrough}
\alias{style_no_color}
\alias{style_no_bg_color}
\title{ANSI colored text}
\usage{
bg_black(...)

bg_blue(...)

bg_cyan(...)

bg_green(...)

bg_magenta(...)

bg_red(...)

bg_white(...)

bg_yellow(...)

bg_none(...)

bg_br_black(...)

bg_br_blue(...)

bg_br_cyan(...)

bg_br_green(...)

bg_br_magenta(...)

bg_br_red(...)

bg_br_white(...)

bg_br_yellow(...)

col_black(...)

col_blue(...)

col_cyan(...)

col_green(...)

col_magenta(...)

col_red(...)

col_white(...)

col_yellow(...)

col_grey(...)

col_silver(...)

col_none(...)

col_br_black(...)

col_br_blue(...)

col_br_cyan(...)

col_br_green(...)

col_br_magenta(...)

col_br_red(...)

col_br_white(...)

col_br_yellow(...)

style_dim(...)

style_blurred(...)

style_bold(...)

style_hidden(...)

style_inverse(...)

style_italic(...)

style_reset(...)

style_strikethrough(...)

style_underline(...)

style_no_bold(...)

style_no_blurred(...)

style_no_dim(...)

style_no_italic(...)

style_no_underline(...)

style_no_inverse(...)

style_no_hidden(...)

style_no_strikethrough(...)

style_no_color(...)

style_no_bg_color(...)
}
\arguments{
\item{...}{Character strings, they will be pasted together with
\code{paste0()}, before applying the style function.}
}
\value{
An ANSI string (class \code{cli_ansi_string}), that contains ANSI
sequences, if the current platform supports them. You can simply
use \code{cat()} to print them to the terminal.
}
\description{
cli has a number of functions to color and style text at the command
line. They provide a more modern interface than the crayon package.
}
\details{
The \verb{col_*} functions change the (foreground) color to the text.
These are the eight original ANSI colors. Note that in some terminals,
they might actually look differently, as terminals have their own
settings for how to show them. \code{col_none()} is the default color, this
is useful in a substring of a colored string.

The \verb{col_br_*} functions are bright versions of the eight ANSI colors.
Note that on some terminal configurations and themes they might be the
same as the non-bright colors.

The \verb{bg_*} functions change the background color of the text.
These are the eight original ANSI background colors. These, too, can
vary in appearance, depending on terminal settings. \code{bg_none()} the
the default background color, this is useful in a substring of a
background-colored string.

The \verb{bg_br_*} functions are the bright versions of the eight ANSI
background colors. Note that on some terminal configurations and themes
they might be the same as the non-bright colors.

The \verb{style_*} functions apply other styling to the text. The currently
supported styling functions are:
\itemize{
\item \code{style_reset()} to remove any style, including color,
\item \code{style_bold()} for boldface / strong text, although some terminals
show a bright, high intensity text instead,
\item \code{style_dim()} (or \code{style_blurred()} reduced intensity text.
\item \code{style_italic()} (not widely supported).
\item \code{style_underline()},
\item \code{style_inverse()},
\item \code{style_hidden()},
\item \code{style_strikethrough()} (not widely supported).
}

The style functions take any number of character vectors as arguments,
and they concatenate them using \code{paste0()} before adding the style.

Styles can also be nested, and then inner style takes precedence, see
examples below.

Sometimes you want to revert back to the default text color, in the
middle of colored text, or you want to have a normal font in the middle
of italic text. You can use the \verb{style_no_*} functions for this. Every
\verb{style_*()} function has a \verb{style_no_*()} pair, which defends its
argument from taking on the style. See examples below.
}
\examples{
col_blue("Hello ", "world!")
cat(col_blue("Hello ", "world!"))

cat("... to highlight the", col_red("search term"),
    "in a block of text\n")

## Style stack properly
cat(col_green(
 "I am a green line ",
 col_blue(style_underline(style_bold("with a blue substring"))),
 " that becomes green again!"
))

error <- combine_ansi_styles("red", "bold")
warn <- combine_ansi_styles("magenta", "underline")
note <- col_cyan
cat(error("Error: subscript out of bounds!\n"))
cat(warn("Warning: shorter argument was recycled.\n"))
cat(note("Note: no such directory.\n"))

# style_no_* functions, note that the color is not removed
style_italic(col_green(paste0(
  "italic before, ",
  style_no_italic("normal here, "),
  "italic after"
)))

# avoiding  color for substring
style_italic(col_red(paste(
  "red before",
  col_none("not red between"),
  "red after"
)))
}
\seealso{
Other ANSI styling: 
\code{\link{combine_ansi_styles}()},
\code{\link{make_ansi_style}()},
\code{\link{num_ansi_colors}()}
}
\concept{ANSI styling}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ansiex.R
\name{ansi_has_any}
\alias{ansi_has_any}
\title{Check if a string has some ANSI styling}
\usage{
ansi_has_any(string, sgr = TRUE, csi = TRUE, link = TRUE)
}
\arguments{
\item{string}{The string to check. It can also be a character
vector.}

\item{sgr}{Whether to look for SGR (styling) control sequences.}

\item{csi}{Whether to look for non-SGR control sequences.}

\item{link}{Whether to look for ANSI hyperlinks.}
}
\value{
Logical vector, \code{TRUE} for the strings that have some
ANSI styling.
}
\description{
Check if a string has some ANSI styling
}
\examples{
## The second one has style if ANSI colors are supported
ansi_has_any("foobar")
ansi_has_any(col_red("foobar"))
}
\seealso{
Other low level ANSI functions: 
\code{\link{ansi_hide_cursor}()},
\code{\link{ansi_regex}()},
\code{\link{ansi_string}()},
\code{\link{ansi_strip}()}
}
\concept{low level ANSI functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/prettycode.R
\name{code_highlight}
\alias{code_highlight}
\title{Syntax highlight R code}
\usage{
code_highlight(code, code_theme = NULL, envir = NULL)
}
\arguments{
\item{code}{Character vector, each element is one line of code.}

\item{code_theme}{Theme see \code{\link[=code_theme_list]{code_theme_list()}}.}

\item{envir}{Environment to look up function calls for hyperlinks.
If \code{NULL}, then the global search path is used.}
}
\value{
Character vector, the highlighted code.
}
\description{
Syntax highlight R code
}
\details{
See \code{\link[=code_theme_list]{code_theme_list()}} for the default syntax highlighting theme and
how to change it.

If \code{code} does not parse, then it is returned unchanged and a
\code{cli_parse_failure} condition is thrown. Note that this is not an error,
and the condition is ignored, unless explicitly caught.
}
\examples{
code_highlight(deparse(ls))
cat(code_highlight(deparse(ls)), sep = "\n")
}
\seealso{
Other syntax highlighting: 
\code{\link{code_theme_list}()}
}
\concept{syntax highlighting}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utf8.R
\name{utf8_graphemes}
\alias{utf8_graphemes}
\title{Break an UTF-8 character vector into grapheme clusters}
\usage{
utf8_graphemes(x)
}
\arguments{
\item{x}{Character vector.}
}
\value{
List of characters vectors, the grapheme clusters of the input
string.
}
\description{
Break an UTF-8 character vector into grapheme clusters
}
\examples{
# Five grapheme clusters
str <- paste0(
  "\U0001f477\U0001f3ff\u200d\u2640\ufe0f",
  "\U0001f477\U0001f3ff",
  "\U0001f477\u200d\u2640\ufe0f",
  "\U0001f477\U0001f3fb",
  "\U0001f477\U0001f3ff")
cat(str, "\n")
chrs <- utf8_graphemes(str)
}
\seealso{
Other UTF-8 string manipulation: 
\code{\link{utf8_nchar}()},
\code{\link{utf8_substr}()}
}
\concept{UTF-8 string manipulation}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ansiex.R
\name{ansi_strip}
\alias{ansi_strip}
\title{Remove ANSI escape sequences from a string}
\usage{
ansi_strip(string, sgr = TRUE, csi = TRUE, link = TRUE)
}
\arguments{
\item{string}{The input string.}

\item{sgr}{Whether to remove for SGR (styling) control sequences.}

\item{csi}{Whether to remove for non-SGR control sequences.}

\item{link}{Whether to remove ANSI hyperlinks.}
}
\value{
The cleaned up string. Note that \code{ansi_strip()} always drops
the \code{cli_ansi_string} class, even if \code{sgr} and sci\code{are}FALSE`.
}
\description{
The input may be of class \code{cli_ansi_string} class, this is also dropped
from the result.
}
\examples{
ansi_strip(col_red("foobar")) == "foobar"
}
\seealso{
Other low level ANSI functions: 
\code{\link{ansi_has_any}()},
\code{\link{ansi_hide_cursor}()},
\code{\link{ansi_regex}()},
\code{\link{ansi_string}()}
}
\concept{low level ANSI functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/progress-utils.R
\name{cli_progress_num}
\alias{cli_progress_num}
\alias{cli_progress_cleanup}
\title{Progress bar utility functions.}
\usage{
cli_progress_num()

cli_progress_cleanup()
}
\value{
\code{cli_progress_num()} returns an integer scalar.

`cli_progress_cleanup() does not return anything.
}
\description{
Progress bar utility functions.
}
\details{
\code{cli_progress_num()} returns the number of currently
active progress bars. (These do not currently include the progress
bars created in C/C++ code.)

\code{cli_progress_cleanup()} terminates all active progress bars.
(It currently ignores progress bars created in the C/C++ code.)
}
\seealso{
Other progress bar functions: 
\code{\link{cli_progress_along}()},
\code{\link{cli_progress_bar}()},
\code{\link{cli_progress_builtin_handlers}()},
\code{\link{cli_progress_message}()},
\code{\link{cli_progress_output}()},
\code{\link{cli_progress_step}()},
\code{\link{cli_progress_styles}()},
\code{\link{progress-variables}}
}
\concept{progress bar functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/unicode.R
\name{unicode-width-workaround}
\alias{unicode-width-workaround}
\title{Working around the bad Unicode character widths}
\description{
R 3.6.2 and also the coming 3.6.3 and 4.0.0 versions use the Unicode 8
standard to calculate the display width of Unicode characters.
Unfortunately the widths of most emojis are incorrect in this standard,
and width 1 is reported instead of the correct 2 value.
}
\details{
cli implements a workaround for this. The package contains a table that
contains all Unicode ranges that have wide characters (display width 2).

On first use of one of the workaround wrappers (in \code{ansi_nchar()}, etc.)
we check what the current version of R thinks about the width of these
characters, and then create a regex that matches the ones that R
is wrong about (\code{re_bad_char_width}).

Then we use this regex to duplicate all of the problematic characters
in the input string to the wrapper function, before calling the real
string manipulation function (\code{nchar()}, \code{strwrap()}) etc. At end we
undo the duplication before we return the result.

This workaround is fine for \code{nchar()} and \code{strwrap()}, and consequently
\code{ansi_align()} and \code{ansi_strtrim()} as well.

The rest of the \verb{ansi_*()} functions work on characters, and do not
deal with character width.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/spark.R
\name{spark_line}
\alias{spark_line}
\title{Draw a sparkline line graph with Braille characters.}
\usage{
spark_line(x)
}
\arguments{
\item{x}{A numeric vector between 0 and 1}
}
\description{
You might want to avoid sparklines on non-UTF-8 systems, because they
do not look good. You can use \code{\link[=is_utf8_output]{is_utf8_output()}} to test for support
for them.
}
\details{
\if{html}{\out{<div class="sourceCode r">}}\preformatted{x <- seq(0, 1, length = 10)
spark_line(x)
}\if{html}{\out{</div>}}\if{html}{\out{
<div class="asciicast" style="color: #172431;font-family: 'Fira Code',Monaco,Consolas,Menlo,'Bitstream Vera Sans Mono','Powerline Symbols',monospace;line-height: 1.300000"><pre>
#> ⣀⡠⠔⠊⠉                                                                           
</pre></div>
}}
}
\seealso{
\code{\link[=spark_bar]{spark_bar()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/progress-variables.R
\name{progress-variables}
\alias{progress-variables}
\alias{cli__pb_bar}
\alias{pb_bar}
\alias{cli__pb_current}
\alias{pb_current}
\alias{cli__pb_current_bytes}
\alias{pb_current_bytes}
\alias{cli__pb_elapsed}
\alias{pb_elapsed}
\alias{cli__pb_elapsed_clock}
\alias{pb_elapsed_clock}
\alias{cli__pb_elapsed_raw}
\alias{pb_elapsed_raw}
\alias{cli__pb_eta}
\alias{pb_eta}
\alias{cli__pb_eta_raw}
\alias{pb_eta_raw}
\alias{cli__pb_eta_str}
\alias{pb_eta_str}
\alias{cli__pb_extra}
\alias{pb_extra}
\alias{cli__pb_id}
\alias{pb_id}
\alias{cli__pb_name}
\alias{pb_name}
\alias{cli__pb_percent}
\alias{pb_percent}
\alias{cli__pb_pid}
\alias{pb_pid}
\alias{cli__pb_rate}
\alias{pb_rate}
\alias{cli__pb_rate_raw}
\alias{pb_rate_raw}
\alias{cli__pb_rate_bytes}
\alias{pb_rate_bytes}
\alias{cli__pb_spin}
\alias{pb_spin}
\alias{cli__pb_status}
\alias{pb_status}
\alias{cli__pb_timestamp}
\alias{pb_timestamp}
\alias{cli__pb_total}
\alias{pb_total}
\alias{cli__pb_total_bytes}
\alias{pb_total_bytes}
\title{Progress bar variables}
\description{
Progress bar variables
}
\details{
These variables can be used in cli progress bar format
strings. They are calculated on demand. To use a variable, e.g. \code{pb_bar}
in a package, you either need to to import \code{pb_bar} from cli, or use
the qualified form in the format string: \code{cli::pb_bar}.

Similarly, in R scripts, you can use \code{pb_bar} after \code{library(cli)},
or \code{cli::pb_bar} if you do not attach the cli package.

\subsection{\code{pb_bar}}{

Creates a visual progress bar. If the number of total units
is unknown, then it will return an empty string.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{cli_progress_bar(
  total = 100,
  format = "Fitting model \{cli::pb_bar\} \{cli::pb_percent\}"
)
}\if{html}{\out{</div>}}\if{html}{\out{
<div class="asciicast" style="color: #172431;font-family: 'Fira Code',Monaco,Consolas,Menlo,'Bitstream Vera Sans Mono','Powerline Symbols',monospace;line-height: 1.300000"><pre>
#> Fitting model <span style="color: #859900;">█████████████████████</span><span style="color: #002B36;">██████████</span><span style="color: #859900;"> </span>  66&#37;                             
</pre></div>
}}

}

\subsection{\code{pb_current}}{

The number of current progress units.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{cli_progress_bar(
  total = 100,
  format = "\{cli::pb_spin\} Reading file \{cli::pb_current\}/\{cli::pb_total\}"
)
}\if{html}{\out{</div>}}\if{html}{\out{
<div class="asciicast" style="color: #172431;font-family: 'Fira Code',Monaco,Consolas,Menlo,'Bitstream Vera Sans Mono','Powerline Symbols',monospace;line-height: 1.300000"><pre>
#> ⠙ Reading file 66/100                                                           
</pre></div>
}}

}

\subsection{\code{pb_current_bytes}}{

The number of current progress units formatted as bytes.
The output has a constant width of six characters.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{cli_progress_bar(
  format = "Got \{cli::pb_current_bytes\} in \{cli::pb_elapsed\}"
)
}\if{html}{\out{</div>}}\if{html}{\out{
<div class="asciicast" style="color: #172431;font-family: 'Fira Code',Monaco,Consolas,Menlo,'Bitstream Vera Sans Mono','Powerline Symbols',monospace;line-height: 1.300000"><pre>
#> Got 524 kB in 5s                                                                
</pre></div>
}}

}

\subsection{\code{pb_elapsed}}{

The elapsed time since the start of the progress bar. The time is
measured since the progress bar was created with \code{\link[=cli_progress_bar]{cli_progress_bar()}}
or similar.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{cli_progress_bar(
  total = 100,
  format = "\{cli::pb_bar\} \{cli::pb_percent\} [\{cli::pb_elapsed\}]"
)
}\if{html}{\out{</div>}}\if{html}{\out{
<div class="asciicast" style="color: #172431;font-family: 'Fira Code',Monaco,Consolas,Menlo,'Bitstream Vera Sans Mono','Powerline Symbols',monospace;line-height: 1.300000"><pre>
#> <span style="color: #859900;">█████████████████████</span><span style="color: #002B36;">██████████</span><span style="color: #859900;"> </span>  66&#37; [5s]                                      
</pre></div>
}}

}

\subsection{\code{pb_elapsed_clock}}{

The elapsed time, in \verb{hh::mm::ss} format.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{cli_progress_bar(
  total = 100,
  format = "\{cli::pb_bar\} \{cli::pb_percent\} [\{cli::pb_elapsed_clock\}]"
)
}\if{html}{\out{</div>}}\if{html}{\out{
<div class="asciicast" style="color: #172431;font-family: 'Fira Code',Monaco,Consolas,Menlo,'Bitstream Vera Sans Mono','Powerline Symbols',monospace;line-height: 1.300000"><pre>
#> <span style="color: #859900;">█████████████████████</span><span style="color: #002B36;">██████████</span><span style="color: #859900;"> </span>  66&#37; [00:00:05]                                
</pre></div>
}}

}

\subsection{\code{pb_elapsed_raw}}{

The number of seconds since the start of the progress bar.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{cli_progress_bar(
  total = 100,
  format = "\{cli::pb_bar\} \{cli::pb_percent\} [\{round(cli::pb_elapsed_raw)\}s]"
)
}\if{html}{\out{</div>}}\if{html}{\out{
<div class="asciicast" style="color: #172431;font-family: 'Fira Code',Monaco,Consolas,Menlo,'Bitstream Vera Sans Mono','Powerline Symbols',monospace;line-height: 1.300000"><pre>
#> <span style="color: #859900;">█████████████████████</span><span style="color: #002B36;">██████████</span><span style="color: #859900;"> </span>  66&#37; [5s]                                      
</pre></div>
}}

}

\subsection{\code{pb_eta}}{

The estimated time until the end of the progress bar,
in human readable form.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{cli_progress_bar(
  total = 100,
  format = "\{cli::pb_bar\} \{cli::pb_percent\} | ETA: \{cli::pb_eta\}"
)
}\if{html}{\out{</div>}}\if{html}{\out{
<div class="asciicast" style="color: #172431;font-family: 'Fira Code',Monaco,Consolas,Menlo,'Bitstream Vera Sans Mono','Powerline Symbols',monospace;line-height: 1.300000"><pre>
#> <span style="color: #859900;">█████████████████████</span><span style="color: #002B36;">██████████</span><span style="color: #859900;"> </span>  66&#37; | ETA:  3s                                
</pre></div>
}}

}

\subsection{\code{pb_eta_raw}}{

The estimated time until the end of the progress
bar, in seconds. This is useful if you want to adjust the default
\code{pb_eta} display.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{cli_progress_bar(
  total = 100,
  format = "\{cli::pb_bar\} \{cli::pb_percent\} | ETA: \{round(cli::pb_eta_raw)\}s"
)
}\if{html}{\out{</div>}}\if{html}{\out{
<div class="asciicast" style="color: #172431;font-family: 'Fira Code',Monaco,Consolas,Menlo,'Bitstream Vera Sans Mono','Powerline Symbols',monospace;line-height: 1.300000"><pre>
#> <span style="color: #859900;">█████████████████████</span><span style="color: #002B36;">██████████</span><span style="color: #859900;"> </span>  66&#37; | ETA: 3s                                 
</pre></div>
}}

}

\subsection{\code{pb_eta_str}}{

The estimated time until the end of the progress bar.
It includes the \code{"ETA:"} prefix. It is only shown if the time can be
estimated, otherwise it is the empty string.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{cli_progress_bar(
  total = 100,
  format = "\{cli::pb_bar\} \{cli::pb_percent\} | \{cli::pb_eta_str\}"
)
}\if{html}{\out{</div>}}\if{html}{\out{
<div class="asciicast" style="color: #172431;font-family: 'Fira Code',Monaco,Consolas,Menlo,'Bitstream Vera Sans Mono','Powerline Symbols',monospace;line-height: 1.300000"><pre>
#> <span style="color: #859900;">█████████████████████</span><span style="color: #002B36;">██████████</span><span style="color: #859900;"> </span>  66&#37; | ETA:  3s                                
</pre></div>
}}

}

\subsection{\code{pb_extra}}{

\code{pb_extra} can be used to access extra data, see the \code{extra} argument
of \code{cli_progress_bar()} and \code{cli_progress_update()}.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{cli_progress_bar(
  total = 100,
  extra = list(user = whoami::username()),
  format = "Cleaning cache for user '\{cli::pb_extra$user\}': \{cli::pb_current_bytes\}"
)
}\if{html}{\out{</div>}}\if{html}{\out{
<div class="asciicast" style="color: #172431;font-family: 'Fira Code',Monaco,Consolas,Menlo,'Bitstream Vera Sans Mono','Powerline Symbols',monospace;line-height: 1.300000"><pre>
#> Cleaning cache for user 'gaborcsardi': 161 MB                                   
</pre></div>
}}

}

\subsection{\code{pb_id}}{

The id of the progress bar. The id has the format
\verb{cli-<pid>-<counter>} where \verb{<pid>} is the process id, and
\verb{<counter>} is an integer counter that is incremented every time
cli needs a new unique id.

This is useful for debugging progress bars.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{cli_progress_bar(
  format = "Progress bar '\{cli::pb_id\}' is at \{cli::pb_current\}"
)
}\if{html}{\out{</div>}}\if{html}{\out{
<div class="asciicast" style="color: #172431;font-family: 'Fira Code',Monaco,Consolas,Menlo,'Bitstream Vera Sans Mono','Powerline Symbols',monospace;line-height: 1.300000"><pre>
#> Progress bar 'cli-82040-1814' is at 64                                          
</pre></div>
}}

}

\subsection{\code{pb_name}}{

The name of the progress bar. This is supplied by the
developer, and it is by default the empty string. A space character
is added to non-empty names.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{cli_progress_bar(
  name = "Loading training data",
  total = 100,
  format = "\{cli::pb_name\} \{cli::pb_bar\} \{cli::pb_percent\}"
)
}\if{html}{\out{</div>}}\if{html}{\out{
<div class="asciicast" style="color: #172431;font-family: 'Fira Code',Monaco,Consolas,Menlo,'Bitstream Vera Sans Mono','Powerline Symbols',monospace;line-height: 1.300000"><pre>
#> Loading training data  <span style="color: #859900;">█████████████████████</span><span style="color: #002B36;">██████████</span><span style="color: #859900;"> </span>  66&#37;                    
</pre></div>
}}

}

\subsection{\code{pb_percent}}{

The percentage of the progress bar, always formatted
in three characters plus the percentage sign. If the total number of
units is unknown, then it is \code{" NA\%"}.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{cli_progress_bar(
  total = 100,
  format = "\{cli::pb_bar\} \{cli::pb_percent\}"
)
}\if{html}{\out{</div>}}\if{html}{\out{
<div class="asciicast" style="color: #172431;font-family: 'Fira Code',Monaco,Consolas,Menlo,'Bitstream Vera Sans Mono','Powerline Symbols',monospace;line-height: 1.300000"><pre>
#> <span style="color: #859900;">█████████████████████</span><span style="color: #002B36;">██████████</span><span style="color: #859900;"> </span>  66&#37;                                           
</pre></div>
}}

}

\subsection{\code{pb_pid}}{

The integer process id of the progress bar. This is useful if you are
aggregating logging output or progress results from multiple processes.
}

\subsection{\code{pb_rate}}{

The progress rate, in number of units per second, formatted in a string.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{cli_progress_bar(
  total = 156,
  format = "Reading input files \{pb_current\}/\{pb_total\} [\{pb_rate\}]"
)
}\if{html}{\out{</div>}}\if{html}{\out{
<div class="asciicast" style="color: #172431;font-family: 'Fira Code',Monaco,Consolas,Menlo,'Bitstream Vera Sans Mono','Powerline Symbols',monospace;line-height: 1.300000"><pre>
#> Reading input files 68/156 [14/s]                                               
</pre></div>
}}

}

\subsection{\code{pb_rate_raw}}{

The raw progress rate, in number of units per second.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{cli_progress_bar(
  total = 156,
  format = "Reading input files \{pb_current\}/\{pb_total\} [\{round(pb_rate_raw)\}/s]"
)
}\if{html}{\out{</div>}}\if{html}{\out{
<div class="asciicast" style="color: #172431;font-family: 'Fira Code',Monaco,Consolas,Menlo,'Bitstream Vera Sans Mono','Powerline Symbols',monospace;line-height: 1.300000"><pre>
#> Reading input files 68/156 [14/s]                                               
</pre></div>
}}

}

\subsection{\code{pb_rate_bytes}}{

The progress rate, formatted as bytes per second, in human readable form.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{cli_progress_bar(
  total = 256 * 1024 * 1014,
  format = paste0(
    "Reading data \{pb_current_bytes\}/\{pb_total_bytes\} ",
    "[\{ansi_trimws(pb_rate_bytes)\}]"
  )
}\if{html}{\out{</div>}}\if{html}{\out{
<div class="asciicast" style="color: #172431;font-family: 'Fira Code',Monaco,Consolas,Menlo,'Bitstream Vera Sans Mono','Powerline Symbols',monospace;line-height: 1.300000"><pre>
#> Reading data  70 MB/266 MB [14 MB/s]                                            
</pre></div>
}}

}

\subsection{\code{pb_spin}}{

A spinner. The default spinner is selected via a \code{\link[=get_spinner]{get_spinner()}} call.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{cli_progress_bar(
  total = 100,
  format = "\{cli::pb_spin\} Reading file \{cli::pb_current\}/\{cli::pb_total\}"
)
}\if{html}{\out{</div>}}\if{html}{\out{
<div class="asciicast" style="color: #172431;font-family: 'Fira Code',Monaco,Consolas,Menlo,'Bitstream Vera Sans Mono','Powerline Symbols',monospace;line-height: 1.300000"><pre>
#> ⠙ Reading file 66/100                                                           
</pre></div>
}}

}

\subsection{\code{pb_status}}{

The status string of the progress bar. By default this is an empty
string, but it is possible to set it in \code{\link[=cli_progress_bar]{cli_progress_bar()}}
and `cli_progress_update()].

\if{html}{\out{<div class="sourceCode r">}}\preformatted{cli_progress_bar(status = "Connecting...")
}\if{html}{\out{</div>}}\if{html}{\out{
<div class="asciicast" style="color: #172431;font-family: 'Fira Code',Monaco,Consolas,Menlo,'Bitstream Vera Sans Mono','Powerline Symbols',monospace;line-height: 1.300000"><pre>
#> ⠙ Connecting... 0 done (0/s) | 1s                                               
</pre></div>
}}

}

\subsection{\code{pb_timestamp}}{

A time stamp for the current time in ISO 8601 format.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{cli_progress_bar(
  "Loading training data files",
  format = "\{pb_timestamp\} \{pb_current\} (\{pb_rate\})"
}\if{html}{\out{</div>}}\if{html}{\out{
<div class="asciicast" style="color: #172431;font-family: 'Fira Code',Monaco,Consolas,Menlo,'Bitstream Vera Sans Mono','Powerline Symbols',monospace;line-height: 1.300000"><pre>
#> 2022-09-07T11:27:50+00:00 125 (25/s)                                            
</pre></div>
}}

}

\subsection{\code{pb_total}}{

The total number of progress units, or \code{NA} if the number of units is
unknown.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{cli_progress_bar(
  total = 100,
  format = "\{cli::pb_spin\} Reading file \{cli::pb_current\}/\{cli::pb_total\}"
)
}\if{html}{\out{</div>}}\if{html}{\out{
<div class="asciicast" style="color: #172431;font-family: 'Fira Code',Monaco,Consolas,Menlo,'Bitstream Vera Sans Mono','Powerline Symbols',monospace;line-height: 1.300000"><pre>
#> ⠙ Reading file 66/100                                                           
</pre></div>
}}

}

\subsection{\code{pb_total_bytes}}{

The total number of progress units, formatted as
bytes, in a human readable format.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{cli_progress_bar(
  total = 256 * 1024 * 1014,
  format = paste0(
    "Reading data \{pb_current_bytes\}/\{pb_total_bytes\} ",
    "[\{ansi_trimws(pb_rate_bytes)\}]"
  )
}\if{html}{\out{</div>}}\if{html}{\out{
<div class="asciicast" style="color: #172431;font-family: 'Fira Code',Monaco,Consolas,Menlo,'Bitstream Vera Sans Mono','Powerline Symbols',monospace;line-height: 1.300000"><pre>
#> Reading data  70 MB/266 MB [14 MB/s]                                            
</pre></div>
}}

}
}
\seealso{
Other progress bar functions: 
\code{\link{cli_progress_along}()},
\code{\link{cli_progress_bar}()},
\code{\link{cli_progress_builtin_handlers}()},
\code{\link{cli_progress_message}()},
\code{\link{cli_progress_num}()},
\code{\link{cli_progress_output}()},
\code{\link{cli_progress_step}()},
\code{\link{cli_progress_styles}()}
}
\concept{progress bar functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/spinner.R
\name{list_spinners}
\alias{list_spinners}
\title{List all available spinners}
\usage{
list_spinners()
}
\value{
Character vector of all available spinner names.
}
\description{
List all available spinners
}
\examples{
list_spinners()
get_spinner(list_spinners()[1])
}
\seealso{
Other spinners: 
\code{\link{demo_spinners}()},
\code{\link{get_spinner}()},
\code{\link{make_spinner}()}
}
\concept{spinners}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/progress-variables.R
\name{cli_progress_demo}
\alias{cli_progress_demo}
\title{cli progress bar demo}
\usage{
cli_progress_demo(
  name = NULL,
  status = NULL,
  type = c("iterator", "tasks", "download", "custom"),
  total = NA,
  .envir = parent.frame(),
  ...,
  at = if (is_interactive()) NULL else 50,
  show_after = 0,
  live = NULL,
  delay = 0,
  start = as.difftime(5, units = "secs")
)
}
\arguments{
\item{name}{Passed to \code{\link[=cli_progress_bar]{cli_progress_bar()}}.}

\item{status}{Passed to \code{\link[=cli_progress_bar]{cli_progress_bar()}}.}

\item{type}{Passed to \code{\link[=cli_progress_bar]{cli_progress_bar()}}.}

\item{total}{Passed to \code{\link[=cli_progress_bar]{cli_progress_bar()}}.}

\item{.envir}{Passed to \code{\link[=cli_progress_bar]{cli_progress_bar()}}.}

\item{...}{Passed to \code{\link[=cli_progress_bar]{cli_progress_bar()}}.}

\item{at}{The number of progress units to show and capture the progress
bar at. If \code{NULL}, then a sequence of states is generated to show the
progress from beginning to end.}

\item{show_after}{Delay to show the progress bar. Overrides the
\code{cli.progress_show_after} option.}

\item{live}{Whether to show the progress bat on the screen, or just
return the recorded updates. Defaults to the value of the
\code{cli.progress_demo_live} options. If unset, then it is \code{TRUE} in
interactive sessions.}

\item{delay}{Delay between progress bar updates.}

\item{start}{Time to subtract from the start time, to simulate a
progress bar that takes longer to run.}
}
\value{
List with class \code{cli_progress_demo}, which has a print and a
format method for pretty printing. The \code{lines} entry contains the
output lines, each corresponding to one update.
}
\description{
Useful for experimenting with format strings and for documentation.
It creates a progress bar, iterates it until it terminates and saves the
progress updates.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ansi-palette.R
\docType{data}
\name{truecolor}
\alias{truecolor}
\alias{ansi_palettes}
\alias{ansi_palette_show}
\title{ANSI colors palettes}
\format{
\code{truecolor} is an integer scalar.

\code{ansi_palettes} is a data frame with one row for each palette,
and one column for each base ANSI color. \code{attr(ansi_palettes, "info")}
contains a list with information about each palette.
}
\usage{
truecolor

ansi_palettes

ansi_palette_show(palette = NULL, colors = num_ansi_colors(), rows = 4)
}
\arguments{
\item{palette}{The palette to show, in the same format as for the
\code{cli.palette} option, so it can be the name of a built-in palette,
of a list of 16 colors.}

\item{colors}{Number of ANSI colors to use the show the palette. If the
platform does not have sufficient support, the output might have
a lower color resolution. Without color support it will have no color
at all.}

\item{rows}{The number of colored rows to print.}
}
\value{
\code{ansi_palette_show} returns a character vector, the rows that
are printed to the screen, invisibly.
}
\description{
If your platform supports at least 256 colors, then you can configure
the colors that cli uses for the eight base and the eight bright colors.
(I.e. the colors of \code{\link[=col_black]{col_black()}}, \code{\link[=col_red]{col_red()}}, and \code{\link[=col_br_black]{col_br_black()}},
\code{\link[=col_br_red]{col_br_red()}}, etc.
}
\details{
\code{truecolor} is an integer constant for the number of 24 bit ANSI colors.

To customize the default palette, set the \code{cli.palette} option to the
name of a built-in palette (see \code{ansi_palettes()}), or the list of
16 colors. Colors can be specified with RGB colors strings:
\verb{#rrggbb} or R color names (see the output of \code{\link[grDevices:colors]{grDevices::colors()}}).

For example, you can put this in your R profile:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{options(cli.palette = "vscode")
}\if{html}{\out{</div>}}

It is currently not possible to configure the background colors
separately, these will be always the same as the foreground colors.

If your platform only has 256 colors, then the colors specified in the
palette have to be interpolated. On true color platforms they RGB
values are used as-is.

\code{ansi_palettes} is a data frame of the built-in palettes, each row
is one palette.

\code{ansi_palette_show()} shows the colors of an ANSI palette on the screen.
}
\examples{
ansi_palettes
ansi_palette_show("dichro", colors = truecolor)
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/docs.R
\name{faq}
\alias{faq}
\title{Frequently Asked Questions}
\description{
Frequently Asked Questions
}
\details{
\subsection{My platform supports ANSI colors, why does cli not use them?}{

It is probably a mistake in the ANSI support detection algorithm. Please
open an issue at \url{https://github.com/r-lib/cli/issues} and do not forget
to tell us the details of your platform and terminal or GUI.
}

\subsection{How do I turn off ANSI colors and styles?}{

Set the \code{NO_COLOR} environment variable to a non-empty value. You can do
this in your \code{.Renviron} file (use \code{usethis::edit_r_environ()}).

If you want to do this for testthat tests, then consider using the 3rd
edition on testthat, which does turn off ANSI styling automatically
inside \code{test_that()}.
}

\subsection{cli does not show the output before \code{file.choose()}}{

Try calling \code{flush.console()} to flush the console, before
\code{file.choose()}. If flushing does not work and you are in RStudio, then
it is probably this RStudio bug:
\url{https://github.com/rstudio/rstudio/issues/8040} See more details at
\url{https://github.com/r-lib/cli/issues/151}
}

\subsection{Why are heading separators wider than my screen in RStudio?}{

The display width of some Unicode characters ambiguous in the Unicode
standard. Some software treats them as narrow (one column on the
screen), other as wide (two columns). In some terminal emulators (for
example iTerm2), you can configure the preferred behavior.

Unfortunately the box drawing characters that cli uses also have
ambiguous width.

In RStudio the behavior depends on the font. In particular, Consolas,
Courier and Inconsolata treats them as wide characters, so cli output
will not look great with these. Some good, modern fonts that look good
include Menlo, Fira Code and Source Code Pro.

If you do not want to change your font, you can also turn off Unicode
output, by setting the \code{cli.unicode} option:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{options(cli.unicode = FALSE)
}\if{html}{\out{</div>}}

A related issue: \url{https://github.com/r-lib/cli/issues/320}
}

\subsection{Is there a suggested font to use with cli?}{

In modern terminals, cli output usually looks good.

If you see too wide heading separators in RStudio, then see the previous
question: Why are heading separators wider than my screen in RStudio?.

If some output is garbled, then cli probably misdetected Unicode support
for your terminal or font. You can try choosing a different font. In our
experience output looks good with Menlo, Fira Code and Source Code Pro.
Alternatively you can turn off Unicode output:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{options(cli.unicode = FALSE)
}\if{html}{\out{</div>}}

If you think this is our fault, then please also file an issue at
\url{https://github.com/r-lib/cli/issues}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/docs.R
\name{cli-config}
\alias{cli-config}
\title{cli environment variables and options}
\description{
cli environment variables and options
}
\section{User facing configuration}{


These are environment variables and options that users may set, to modify
the behavior of cli.
\subsection{Hyperlinks}{

cli uses ANSI escape sequences to create hyperlinks.
Specifically, cli creates \href{https://gist.github.com/egmontkob/eb114294efbcd5adb1944c9f3cb5feda}{OSC 8 hyperlinks} that have this syntax:

\if{html}{\out{<div class="sourceCode">}}\preformatted{OSC 8 ; \{OPTIONAL PARAMS \} ; \{URI\} ST \{LINK TEXT\} OSC 8 ; ; ST
}\if{html}{\out{</div>}}

Under the hood, \code{\link[=style_hyperlink]{style_hyperlink()}} is the helper that forms these links, but it is more common to request them indirectly via inline markup (documented in \code{help("links")}).
\subsection{\code{R_CLI_HYPERLINK_MODE}}{

Set to \code{posix} to force generating POSIX compatible ANSI hyperlinks.
This is about the specific operating system command (OSC) and string terminator (ST) used in hyperlinks.

If not set, then RStudio compatible links are generated. This is a
temporary crutch until RStudio handles POSIX hyperlinks correctly, and
after that it will be removed.
}

\subsection{\code{cli.hyperlink} option and \code{R_CLI_HYPERLINKS} env var}{

Set this option or env var to \code{true}, \code{TRUE} or \code{True} to tell cli that the terminal supports ANSI hyperlinks.
Leave this configuration unset (or set to anything else) when there is no hyperlink support.
Specifically, this configuration indicates the support for URL and file hyperlinks, requested via markup like \verb{\{.href url\}} or \verb{\{.file path/file\}}.
(Below we describe the configuration that indicates support for even more specialized types of hyperlinks.)

The option \code{cli.hyperlink} takes precedence over the \code{R_CLI_HYPERLINKS} env var.
}

\subsection{\verb{cli.hyperlink_*} options and \verb{R_CLI_HYPERLINK_*} env vars}{

cli supports a few special types of hyperlink that go beyond pointing to, e.g., a webpage or a file.
These specialized hyperlinks cause R-specific actions to happen, such as executing a bit of R code or opening specific documentation.

Set the relevant option or env var to \code{true}, \code{TRUE}, or \code{True} to tell cli that the terminal is capable of implementing these specialized behaviours.
Leave this configuration unset (or set to anything else) when there is no support for a specific type of hyperlink.\tabular{llll}{
   Action \tab Example inline markup \tab Option \tab Env var \cr
   Run R code \tab \verb{\{.run testthat::snapshot_review()\}} \tab \code{cli.hyperlink_run} \tab \code{R_CLI_HYPERLINK_RUN} \cr
   Open a help topic \tab \verb{\{.fun stats::lm\}} \verb{\{.topic tibble::tibble_options\}} \verb{\{.help [\{.fun lm\}](stats::lm)\}} \tab \code{cli.hyperlink_help} \tab \code{R_CLI_HYPERLINK_HELP} \cr
   Open a vignette \tab \verb{\{.vignette tibble::types\}} \tab \code{cli.hyperlink_vignette} \tab \code{R_CLI_HYPERLINK_VIGNETTE} \cr
}


In all cases, the option takes priority over the corresponding env var.
}

\subsection{\verb{cli.hyperlink_*_url_format} options and \verb{R_CLI_HYPERLINK_*_URL_FORMAT} env vars}{

Recall the overall structure of cli's hyperlinks:

\if{html}{\out{<div class="sourceCode">}}\preformatted{OSC 8 ; \{OPTIONAL PARAMS \} ; \{URI\} ST \{LINK TEXT\} OSC 8 ; ; ST
}\if{html}{\out{</div>}}

The \code{URI} part has a default format for each type of hyperlink, but it is possible to provide a custom format via an option or an env var.
If defined, the option takes priority over the env var.\tabular{llll}{
   Action \tab Default URI format \tab Customize via option \tab Customize via env var \cr
   Open a file \tab (see below) \tab \code{cli.hyperlink_file_url_format} \tab \code{R_CLI_HYPERLINK_FILE_URL_FORMAT} \cr
   Run R code \tab \code{x-r-run:{code}} \tab \code{cli.hyperlink_run_url_format} \tab \code{R_CLI_HYPERLINK_RUN_URL_FORMAT} \cr
   Open a help topic \tab \code{x-r-help:{topic}} \tab \code{cli.hyperlink_help_url_format} \tab \code{R_CLI_HYPERLINK_HELP_URL_FORMAT} \cr
   Open a vignette \tab \code{x-r-vignette:{vignette}} \tab \code{cli.hyperlink_vignette_url_format} \tab \code{R_CLI_HYPERLINK_VIGNETTE_URL_FORMAT} \cr
}


A format must be a glue-like template with the relevant placeholder in curly braces (\code{code}, \code{topic} or \code{vignette}).

Here's an example of a custom URI format for runnable code, which is useful in an integrated Positron terminal:

\if{html}{\out{<div class="sourceCode">}}\preformatted{positron://positron.positron-r/cli?command=x-r-run:\{code\}
}\if{html}{\out{</div>}}

(For backwards compatibility with older versions of RStudio, in some contexts, a legacy format is used, e.g. \code{ide:run:{code}}.)

The default handling for file hyperlinks is geared towards the expectations of RStudio and can't really be expressed as a URI format, but it's approximately \verb{file://\{path\}}, plus possibly passing \code{line} and \code{col} (yes, it really is \code{col}, not \code{column}) in the \verb{OPTIONAL PARAMS} part of the OSC 8 hyperlink.

A custom format can be provided for file hyperlinks and the relevant placeholders are \code{path}, \code{line}, and \code{column}.
Examples of custom file hyperlink formats:

\if{html}{\out{<div class="sourceCode">}}\preformatted{positron://file\{path\}:\{line\}:\{column\}
vscode://file\{path\}:\{line\}:\{column\}
txmt://open?url=file://\{path\}&line=\{line\}&column=\{column\}
}\if{html}{\out{</div>}}

It's OK if the format includes only \code{path}.
}

}

\subsection{User facing environment variables}{
\subsection{\code{NO_COLOR}}{

Set to a nonempty value to turn off ANSI colors.
See \code{\link[=num_ansi_colors]{num_ansi_colors()}}.
}

\subsection{\code{ESS_BACKGROUND_MODE}}{

Set this environment variable to \code{light} or \code{dark} to indicate dark mode
in Emacs. Once https://github.com/emacs-ess/ESS/pull/1178 is merged, ESS
will set this automatically.
}

\subsection{\code{R_CLI_DYNAMIC}}{

Set to \code{true}, \code{TRUE} or \code{True} to assume a dynamic terminal, that supports \verb{\\r}.
Set to anything else to assume a non-dynamic terminal.
See \code{\link[=is_dynamic_tty]{is_dynamic_tty()}}.
}

\subsection{\code{R_CLI_NUM_COLORS}}{

Set to a positive integer to assume a given number of colors.
See \code{\link[=num_ansi_colors]{num_ansi_colors()}}.
}

}

\subsection{User facing options}{
\subsection{\code{cli.ansi}}{

Set to \code{true}, \code{TRUE} or \code{True} to assume a terminal that supports ANSI
control sequences.
Set to anything else to assume a non-ANSI terminal.
See \code{\link[=is_ansi_tty]{is_ansi_tty()}}.
}

\subsection{\code{cli.condition_unicode_bullets}}{

\code{TRUE} or \code{FALSE} to force turn on or off the Unicode symbols when printing
conditions. E.g. in \code{format_error()}, \code{format_warning()}, \code{format_message()} and
also in \code{cli_abort()}, \code{cli_warn()} and \code{cli_inform()}.
}

\subsection{\code{cli.condition_width}}{

Integer scalar (or \code{Inf}) to set the console width when cli is formatting
errors, warnings or messages in \code{format_error()}, \code{format_warning()} and
\code{format_message()}. When formatting conditions this option takes precedence
over \code{cli.width}.
}

\subsection{\code{cli.default_handler}}{

General handler function for all cli conditions.
See \url{https://cli.r-lib.org/articles/semantic-cli.html#cli-messages-1}
}

\subsection{\code{cli.default_num_colors}}{

Default number of ANSI colors. This value is used if the number of colors
is not already set by
\itemize{
\item the \code{cli.num_colors} option,
\item the \code{R_CLI_NUM_COLORS} environment variable,
\item the \code{crayon.enabled} and \code{crayon.colors} options,
\item the \code{NO_COLOR} environment variable,
\item the \code{knitr.in.progress} option,
\item a \code{sink()} call for the stream.
}

You can also use this option if color support is detected correctly, but
you want to adjust the number of colors. E.g.
\itemize{
\item if \code{crayon.enabled} is \code{TRUE}, but \code{crayon.colors} is not,
\item in Emacs on Windows,
\item in terminals.
}

See \code{\link[=num_ansi_colors]{num_ansi_colors()}}. See also the \code{cli.num_colors} option.
}

\subsection{\code{cli.dynamic}}{

Set to \code{TRUE} to assume a dynamic terminal, that supports \verb{\\r}.
Set to anything else to assume a non-dynamic terminal.
See \code{\link[=is_dynamic_tty]{is_dynamic_tty()}}.
}

\subsection{\code{cli.hide_cursor}}{

Whether the cli status bar should try to hide the cursor on terminals.
Set the \code{FALSE} if the hidden cursor causes issues.
}

\subsection{\code{cli.ignore_unknown_rstudio_theme}}{

Set to \code{TRUE} to omit a warning for an unknown RStudio theme in
\code{code_highlight()}.
}

\subsection{\code{cli.num_colors}}{

Number of ANSI colors. See \code{\link[=num_ansi_colors]{num_ansi_colors()}}. See also the
\code{cli.default_num_colors} option.
}

\subsection{\code{cli.message_class}}{

Character vector of classes to add to cli's conditions.
}

\subsection{\code{cli.progress_bar_style}}{

Progress bar style.
See \code{\link[=cli_progress_styles]{cli_progress_styles()}}.
}

\subsection{\code{cli.progress_bar_style_ascii}}{

Progress bar style on ASCII consoles.
See \code{\link[=cli_progress_styles]{cli_progress_styles()}}.
}

\subsection{\code{cli.progress_bar_style_unicode}}{

Progress bar style on Unicode (UTF-8) consoles;
See \code{\link[=cli_progress_styles]{cli_progress_styles()}}.
}

\subsection{\code{cli.progress_clear}}{

Whether to clear terminated progress bar from the screen on dynamic terminals.
See \code{\link[=cli_progress_bar]{cli_progress_bar()}}.
}

\subsection{\code{cli.progress_demo_live}}{

Whether \code{cli_progress_demo()} should show a live demo, or just record the
progress bar frames.
}

\subsection{\code{cli.progress_format_download}}{

Default format string for \code{download} progress bars.
}

\subsection{\code{cli.progress_format_download_nototal}}{

Default format string for \code{download} progress bars with unknown totals.
}

\subsection{\code{cli.progress_format_iterator}}{

Default format string for \code{iterator} progress bars.
}

\subsection{\code{cli.progress_format_iterator_nototal}}{

Default format string for \code{iterator} progress bars with unknown total number
of progress units.
}

\subsection{\code{cli.progress_format_tasks}}{

Default format string for \code{tasks} progress bars.
}

\subsection{\code{cli.progress_format_tasks_nototal}}{

Default format string for \code{tasks} progress bars with unknown totals.
}

\subsection{\code{cli.progress_handlers}}{

Progress handlers to try.
See \code{\link[=cli_progress_builtin_handlers]{cli_progress_builtin_handlers()}}.
}

\subsection{\code{cli.progress_handlers_force}}{

Progress handlers that will always be used, even if another handler
was already selected.
See \code{\link[=cli_progress_builtin_handlers]{cli_progress_builtin_handlers()}}.
}

\subsection{\code{cli.progress_handlers_only}}{

Progress handlers to force, ignoring handlers set in
\code{cli.progress_handlers} and \code{cli.progress_handlers_force}.
See \code{\link[=cli_progress_builtin_handlers]{cli_progress_builtin_handlers()}}.
}

\subsection{\code{cli.progress_say_args}}{

Command line arguments for the \code{say} progress handlers.
See \code{\link[=cli_progress_builtin_handlers]{cli_progress_builtin_handlers()}}.
}

\subsection{\code{cli.progress_say_command}}{

External command to use in the \code{say} progress handler.
See \code{\link[=cli_progress_builtin_handlers]{cli_progress_builtin_handlers()}}.
}

\subsection{\code{cli.progress_say_frequency}}{

Minimum delay between \code{say} calls in the \code{say} progress handler.
\code{say} ignores very frequent updates, to keep the speech comprehensible.
See \code{\link[=cli_progress_builtin_handlers]{cli_progress_builtin_handlers()}}.
}

\subsection{\code{cli.progress_show_after}}{

Delay before showing a progress bar, in seconds.
Progress bars that finish before this delay are not shown at all.
cli also shows progress bars that have more than 50\% to go after half of
this delay has passed.
}

\subsection{\code{cli.spinner}}{

Default spinner to use, see \code{\link[=get_spinner]{get_spinner()}}.
}

\subsection{\code{cli.spinner_ascii}}{

Default spinner to use on ASCII terminals, see \code{\link[=get_spinner]{get_spinner()}}.
}

\subsection{\code{cli.spinner_unicode}}{

Default spinner to use on Unicode terminals, see \code{\link[=get_spinner]{get_spinner()}}.
}

\subsection{\code{cli.theme}}{

Default cli theme, in addition to the built-in theme.
This option in intended for the package developers.
See \link{themes} and \code{\link[=start_app]{start_app()}}.
}

\subsection{\code{cli.theme_dark}}{

Whether cli should assume a dark theme for the builtin theme.
See \code{\link[=builtin_theme]{builtin_theme()}}.
}

\subsection{\code{cli.unicode}}{

Whether to assume a Unicode terminal.
If not set, then it is auto-detected.
See \code{\link[=is_utf8_output]{is_utf8_output()}}.
}

\subsection{\code{cli.user_theme}}{

cli user theme.
This option is intended for end users.
See \link{themes}.
}

\subsection{\code{cli.warn_inline_newlines}}{

Whether to emit a warning when cli replaces newline characters with
spaces within a \code{{.class }} inline style.
Defaults to \code{FALSE}.
}

\subsection{\code{cli.width}}{

Terminal width to assume.
If not set, then it is auto-detected.
See \code{\link[=console_width]{console_width()}}.
}

\subsection{\code{rlib_interactive}}{

Whether to assume an interactive R session.
If not set, then it is auto-detected.
}

\subsection{\code{width}}{

Terminal width.
This is used on some platforms, if \code{cli.width} is not set.
}

}
}

\section{Internal configuration}{


These environment variables and options are for cli developers. Users
should not rely on them as they may change between cli releases.
\subsection{Internal environment variables}{
\subsection{\code{ASCIICAST}}{

Used to detect an asciicast sub-process in RStudio.
}

\subsection{\code{ANSICON}}{

Used to detect ANSICON when detecting the number of ANSI colors.
}

\subsection{\code{CI}}{

Used to detect if the code is running on a CI.
If yes, we avoid ANSI hyperlinks.
}

\subsection{\code{CLI_DEBUG_BAD_END}}{

Whether to warn about \code{cli_end()} calls when there is no container to close.
}

\subsection{\code{CLI_NO_BUILTIN_THEME}}{

Set it to \code{true} to omit the builtin theme.
}

\subsection{\code{CLI_SPEED_TIME}}{

Can be used to speed up cli's timer.
It is a factor, e.g. setting it to 2 makes cli's time go twice as fast.
}

\subsection{\code{CLI_TICK_TIME}}{

How often the cli timer should alert, in milliseconds.
}

\subsection{\code{CMDER_ROOT}}{

Used to detect cmder when detecting the number of ANSI colors.
}

\subsection{\code{COLORTERM}}{

Used when detecting ANSI color support.
}

\subsection{\code{ConEmuANSI}}{

Used to detect ConEmu when detecting the number of ANSI colors.
}

\subsection{\code{EMACS}}{

Used to detect Emacs.
}

\subsection{\code{INSIDE_EMACS}}{

Used to detect Emacs.
}

\subsection{\code{NOT_CRAN}}{

Set to \code{true} to run tests / examples / checks, that do not run on CRAN.
}

\subsection{\verb{_R_CHECK_PACKAGE_NAME_}}{

Used to detect \verb{R CMD check}.
}

\subsection{\code{R_BROWSER}}{

Used to detect the RStudio build pane.
}

\subsection{\code{R_GUI_APP_VERSION}}{

Used to detect R.app on macOS, to decide if the console has ANSI control
sequences.
}

\subsection{\code{R_PACKAGE_DIR}}{

Used to detect if the code is running under \verb{R CMD INSTALL}.
}

\subsection{\code{R_PDFVIEWER}}{

Used to detect the RStudio build pane.
}

\subsection{\code{R_PROGRESS_NO_EXAMPLES}}{

Set to \code{true} to avoid running examples, outside of \verb{R CMD check}.
}

\subsection{\code{RSTUDIO}}{

Used to detect RStudio, in various functions.
}

\subsection{\code{RSTUDIO_CONSOLE_COLOR}}{

Used to detect the number of colors in RStudio.
See \code{\link[=num_ansi_colors]{num_ansi_colors()}}.
}

\subsection{\code{RSTUDIO_CONSOLE_WIDTH}}{

Used to auto-detect console width in RStudio.
}

\subsection{\code{RSTUDIO_TERM}}{

Used to detect the RStudio build pane.
}

\subsection{\code{TEAMCITY_VERSION}}{

Used to detect the TeamCity CI, to turn off ANSI hyperlinks.
}

\subsection{\code{TERM}}{

Used to detect if the console has ANSI control sequences, in a terminal.
}

\subsection{\code{TERM_PROGRAM}}{

Used to detect iTerm for the dark theme detection and the ANSI hyperlink
support detection.
}

\subsection{\code{TERM_PROGRAM_VERSION}}{

Used to detect a suitable iTerm version for ANSI hyperlink support.
}

\subsection{\code{TESTTHAT}}{

Used to detect running in testthat tests.
}

\subsection{\code{VTE_VERSION}}{

Used to detect a suitable VTE version for ANSI hyperlinks.
}

}

\subsection{Internal options}{
\subsection{\code{cli__pb}}{

This option is set to the progress bar that is being updated, when
interpolating the format string.
}

\subsection{\code{cli.record}}{

Internal option to mark the state that cli is recording messages.
}

\subsection{\code{crayon.colors}}{

Deprecated option for the number of ANSI colors, that is still supported
by cli, when the new options are not set.
See \code{\link[=num_ansi_colors]{num_ansi_colors()}}.
}

\subsection{\code{crayon.enabled}}{

Deprecated option to turn ANSI colors on/off.
This is still supported by cli when the new options are not set.
See \code{\link[=num_ansi_colors]{num_ansi_colors()}}.
}

\subsection{\code{crayon.hyperlink}}{

Whether to assume ANSI hyperlink support.
See \code{\link[=ansi_has_hyperlink_support]{ansi_has_hyperlink_support()}}.
}

\subsection{\code{knitr.in.progress}}{

Used to detect knitr when detecting interactive sessions and ANSI color
support.
}

\subsection{\code{rstudio.notebook.executing}}{

Used to detect knitr when detecting interactive sessions.
}

}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cliapp-docs.R
\name{links}
\alias{links}
\title{cli hyperlinks}
\description{
Certain cli styles create clickable links, if your IDE or terminal
supports them.
}
\section{Note: hyperlinks are currently experimental}{
The details of the styles that create hyperlinks will prrobably change
in the near future, based on user feedback.
}

\section{About the links in this manual page}{
The hyperlinks that are included in this manual are demonstrative
only, except for the \verb{https:} links. They look like a hyperlink, and
you can click on them, but they do nothing. I.e. a \code{.run} link will
not run the linked expression if you click on it.
}

\section{Hyperlink Support}{
As of today, the latest release of RStudio (version v2022.07.0+548)
supports all hyperlink types discussed here. Certain terminals, e.g.
iTerm on macOS, Linux terminals based on VTE (GNOME terminal) support
\code{.href}, \code{.email} and \code{.file} links.

You can use \code{\link[=ansi_has_hyperlink_support]{ansi_has_hyperlink_support()}} to check if your terminal or
IDE has hyperlink support in general, and \code{\link[=ansi_hyperlink_types]{ansi_hyperlink_types()}} to
check if various types of hyperlinks are supported.

If your hyperlink support is not detected properly in your IDE or
terminal, please open a cli issue at
\url{https://github.com/r-lib/cli/issues}.
}

\section{Link text}{
Before we delve into the various types of hyperlinks, a general comment
about link texts. Some link styles support a custom link text:
\itemize{
\item \code{.href}
\item \code{.help}
\item \code{.topic}
\item \code{.vignette}
\item \code{.run}
}

Others, i.e. \code{.email}, \code{.file}, \code{.fun} and \code{.url} do not support custom
link text.

The generic syntax for link text is the same as for Markdown hyperlinks:

\if{html}{\out{<div class="sourceCode">}}\preformatted{\{.style [link text](url)\}
}\if{html}{\out{</div>}}
\subsection{Vectorization}{

Note that it is not possible to add link text to a vector of URLs. E.g.
this will create a list of three URLs, all clickable:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{urls <- paste0("https://httpbin.org/status/", c(200, 403, 404))
cli::cli_text("Some httpbin URLs: \{.url \{urls\}\}.")
}\if{html}{\out{</div>}}\if{html}{\out{
<div class="asciicast" style="color: #172431;font-family: 'Fira Code',Monaco,Consolas,Menlo,'Bitstream Vera Sans Mono','Powerline Symbols',monospace;line-height: 1.300000"><pre>
#> Some httpbin URLs: <span style="font-style: italic;color: #268BD2;">&lt;</span><span style="font-style: italic;color: #268BD2;"><a href="https://httpbin.org/status/200">https://httpbin.org/status/200</a></span><span style="font-style: italic;color: #268BD2;">&gt;</span>,                            
#> <span style="font-style: italic;color: #268BD2;">&lt;</span><span style="font-style: italic;color: #268BD2;"><a href="https://httpbin.org/status/403">https://httpbin.org/status/403</a></span><span style="font-style: italic;color: #268BD2;">&gt;</span>, and                                           
#> <span style="font-style: italic;color: #268BD2;">&lt;</span><span style="font-style: italic;color: #268BD2;"><a href="https://httpbin.org/status/404">https://httpbin.org/status/404</a></span><span style="font-style: italic;color: #268BD2;">&gt;</span>.                                               
</pre></div>
But it is not possible to use a different link text for them.
}}

}

\subsection{What if hyperlinks are not available?}{

If ANSI hyperlinks are not available, then the link text for of these
styles outputs both the link text and the URL in a (hopefully) helpful
way. See examples below.
}
}

\section{URLs}{
There are two cli styles to link to generic URLs. \code{.url} does not
allow custom link text, but \verb{\\href} does.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{cli_text(
  "See the cli homepage at \{.url https://cli.r-lib.org\} for details."
)
}\if{html}{\out{</div>}}\if{html}{\out{
<div class="asciicast" style="color: #172431;font-family: 'Fira Code',Monaco,Consolas,Menlo,'Bitstream Vera Sans Mono','Powerline Symbols',monospace;line-height: 1.300000"><pre>
#> See the cli homepage at <span style="font-style: italic;color: #268BD2;">&lt;</span><span style="font-style: italic;color: #268BD2;"><a href="https://cli.r-lib.org">https://cli.r-lib.org</a></span><span style="font-style: italic;color: #268BD2;">&gt;</span> for details.                    
</pre></div>
}}


\if{html}{\out{<div class="sourceCode r">}}\preformatted{cli_text(
  "See the \{.href [cli homepage](https://cli.r-lib.org)\} for details."
)
}\if{html}{\out{</div>}}\if{html}{\out{
<div class="asciicast" style="color: #172431;font-family: 'Fira Code',Monaco,Consolas,Menlo,'Bitstream Vera Sans Mono','Powerline Symbols',monospace;line-height: 1.300000"><pre>
#> See the <a href="https://cli.r-lib.org">cli homepage</a> for details.                                               
</pre></div>
}}

\subsection{Without hyperlink support}{

This is how these links look without hyperlink support:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{local(\{
  withr::local_options(cli.hyperlink = FALSE)
  cli_text(
    "See the cli homepage at \{.url https://cli.r-lib.org\} for details."
  )
  cli_text(
    "See the \{.href [cli homepage](https://cli.r-lib.org)\} for details."
  )
\})
}\if{html}{\out{</div>}}\if{html}{\out{
<div class="asciicast" style="color: #172431;font-family: 'Fira Code',Monaco,Consolas,Menlo,'Bitstream Vera Sans Mono','Powerline Symbols',monospace;line-height: 1.300000"><pre>
#> See the cli homepage at <span style="font-style: italic;color: #268BD2;">&lt;https://cli.r-lib.org&gt;</span> for details.                    
#> See the cli homepage (<span style="font-style: italic;color: #268BD2;">&lt;https://cli.r-lib.org&gt;</span>) for details.                     
</pre></div>
}}

}

\subsection{URL encoding}{

Note that cli does not encode the url, so you might need to call
\code{utils::URLencode()} on it, especially, if it is substituted in
via \code{{}}.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{weirdurl <- utils::URLencode("https://example.com/has some spaces")
cli_text("See more at \{.url \{weirdurl\}\}.")
}\if{html}{\out{</div>}}\if{html}{\out{
<div class="asciicast" style="color: #172431;font-family: 'Fira Code',Monaco,Consolas,Menlo,'Bitstream Vera Sans Mono','Powerline Symbols',monospace;line-height: 1.300000"><pre>
#> See more at <span style="font-style: italic;color: #268BD2;">&lt;</span><span style="font-style: italic;color: #268BD2;"><a href="https://example.com/has&#37;20some&#37;20spaces">https://example.com/has&#37;20some&#37;20spaces</a></span><span style="font-style: italic;color: #268BD2;">&gt;</span>.                          
</pre></div>
}}

}
}

\section{Files}{
The \code{.file} style now automatically creates a \verb{file:} hyperlink.
Because \verb{file:} hyperlinks must contain an absolute path, cli tries to
convert relative paths, and paths starting with \code{~} to aboslute path.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{cli_text("... edit your \{.file ~/.Rprofile\} file.\}")
}\if{html}{\out{</div>}}\if{html}{\out{
<div class="asciicast" style="color: #172431;font-family: 'Fira Code',Monaco,Consolas,Menlo,'Bitstream Vera Sans Mono','Powerline Symbols',monospace;line-height: 1.300000"><pre>
#> ... edit your <span style="color: #268BD2;"><a href="file:///Users/gaborcsardi/.Rprofile">~/.Rprofile</a></span> file.\}                                                
</pre></div>
}}

\subsection{Link text}{

\code{.file} cannot use a custom link text. If you custom link text, then
you can use \code{.href} with a \verb{file:} URL.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{prof <- path.expand("~/.Rprofile")
cli_text("... edit your \{.href [R profile](file://\{prof\})\}.")
}\if{html}{\out{</div>}}\if{html}{\out{
<div class="asciicast" style="color: #172431;font-family: 'Fira Code',Monaco,Consolas,Menlo,'Bitstream Vera Sans Mono','Powerline Symbols',monospace;line-height: 1.300000"><pre>
#> ... edit your <a href="file:///Users/gaborcsardi/.Rprofile">R profile</a>.                                                        
</pre></div>
}}

}

\subsection{Line and column numbers}{

You may add a line number to a file name, separated by \code{:}. Handlers
typically place the cursor at that line after opening the file.
You may also add a column number, after the line number, separated by
another \code{:}.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{cli_text("... see line 5 in \{.file ~/.Rprofile:5\}.")
}\if{html}{\out{</div>}}\if{html}{\out{
<div class="asciicast" style="color: #172431;font-family: 'Fira Code',Monaco,Consolas,Menlo,'Bitstream Vera Sans Mono','Powerline Symbols',monospace;line-height: 1.300000"><pre>
#> ... see line 5 in <span style="color: #268BD2;"><a href="file:///Users/gaborcsardi/.Rprofile">~/.Rprofile:5</a></span>.                                                
</pre></div>
}}

}

\subsection{Default handler}{

In RStudio \verb{file:} URLs open within RStudio. If you click on a file
link outside of RStudio, typically the operating system is consulted
for the application to open it.
}

\subsection{Without hyperlink support}{

One issue with using \code{.href} file files is that it does not look great
if hyperlinks are not available. This will be improved in the future:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{local(\{
  withr::local_options(cli.hyperlink = FALSE)
  prof <- path.expand("~/.Rprofile")
  cli_text("... edit your \{.href [R profile](file://\{prof\})\}.")
\})
}\if{html}{\out{</div>}}\if{html}{\out{
<div class="asciicast" style="color: #172431;font-family: 'Fira Code',Monaco,Consolas,Menlo,'Bitstream Vera Sans Mono','Powerline Symbols',monospace;line-height: 1.300000"><pre>
#> ... edit your R profile (<span style="font-style: italic;color: #268BD2;">&lt;file:///Users/gaborcsardi/.Rprofile&gt;</span>).                
</pre></div>
}}

}
}

\section{Links to the manual}{
\code{.fun} automatically creates links to the manual page of the function,
provided the function name is in the \code{packagename::functionname} form:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{cli::cli_text("... see \{.fun stats::lm\} to learn more.")
}\if{html}{\out{</div>}}\if{html}{\out{
<div class="asciicast" style="color: #172431;font-family: 'Fira Code',Monaco,Consolas,Menlo,'Bitstream Vera Sans Mono','Powerline Symbols',monospace;line-height: 1.300000"><pre>
#> ... see `<a href="x-r-help:stats::lm">stats::lm</a>()` to learn more.                                            
</pre></div>
}}

\subsection{Link text}{

For a custom link text, use \code{.help} instead of \code{.fun}.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{cli::cli_text("... see \{.help [\{.fun lm\}](stats::lm)\} to learn more.")
}\if{html}{\out{</div>}}\if{html}{\out{
<div class="asciicast" style="color: #172431;font-family: 'Fira Code',Monaco,Consolas,Menlo,'Bitstream Vera Sans Mono','Powerline Symbols',monospace;line-height: 1.300000"><pre>
#> ... see <a href="x-r-help:stats::lm">`lm()`</a> to learn more.                                                   
</pre></div>
}}

}

\subsection{Without hyperlink support}{

The same message without hyperlink support looks like this:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{local(\{
  withr::local_options(cli.hyperlink = FALSE)
  cli::cli_text("... see \{.help [\{.fun lm\}](stats::lm)\} to learn more.")
\})
}\if{html}{\out{</div>}}\if{html}{\out{
<div class="asciicast" style="color: #172431;font-family: 'Fira Code',Monaco,Consolas,Menlo,'Bitstream Vera Sans Mono','Powerline Symbols',monospace;line-height: 1.300000"><pre>
#> ... see `lm()` (`?stats::lm()`) to learn more.                                  
</pre></div>
}}

}

\subsection{Topics}{

To link to a help topic that is not a function, use \code{.topic}:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{cli::cli_text("... the tibble options at \{.topic tibble::tibble_options\}.")
}\if{html}{\out{</div>}}\if{html}{\out{
<div class="asciicast" style="color: #172431;font-family: 'Fira Code',Monaco,Consolas,Menlo,'Bitstream Vera Sans Mono','Powerline Symbols',monospace;line-height: 1.300000"><pre>
#> ... the tibble options at <a href="x-r-help:tibble::tibble_options">tibble::tibble_options</a>.                               
</pre></div>
}}


\code{.topic} support link text.
}

\subsection{Vignettes}{

To link to a vignette, use \code{.vignette}:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{cli::cli_text("... see the \{.vignette tibble::types\} vignette.")
}\if{html}{\out{</div>}}\if{html}{\out{
<div class="asciicast" style="color: #172431;font-family: 'Fira Code',Monaco,Consolas,Menlo,'Bitstream Vera Sans Mono','Powerline Symbols',monospace;line-height: 1.300000"><pre>
#> ... see the <a href="x-r-vignette:tibble::types">tibble::types</a> vignette.                                             
</pre></div>
}}

}
}

\section{Click to run code}{
RStudio also supports a special link type that runs R code in the
current R session upon clicking.

You can create these links with \code{.run}:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{cli::cli_text("Run \{.run testthat::snapshot_review()\} to review")
}\if{html}{\out{</div>}}\if{html}{\out{
<div class="asciicast" style="color: #172431;font-family: 'Fira Code',Monaco,Consolas,Menlo,'Bitstream Vera Sans Mono','Powerline Symbols',monospace;line-height: 1.300000"><pre>
#> Run <a href="x-r-run:testthat::snapshot_review()">testthat::snapshot_review()</a> to review                                       
</pre></div>
}}

\subsection{Link text}{

Sometimes you want to show a slightly different expression in the link,
than the one that is evaluated. E.g. the evaluated expression probably
needs to qualify packages with \code{::}, but you might not want to show this:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{cli::cli_text(
  "Run \{.run [snapshot_review()](testthat::snapshot_review())\} to review"
)
}\if{html}{\out{</div>}}\if{html}{\out{
<div class="asciicast" style="color: #172431;font-family: 'Fira Code',Monaco,Consolas,Menlo,'Bitstream Vera Sans Mono','Powerline Symbols',monospace;line-height: 1.300000"><pre>
#> Run <a href="x-r-run:testthat::snapshot_review()">snapshot_review()</a> to review                                                 
</pre></div>
}}

}

\subsection{Security considerations}{

To make \code{.run} hyperlinks more secure, RStudio will not run code
\itemize{
\item that is not in the \code{pkg::fun(args)} form,
\item if \code{args} contains \code{(}, \verb{)} or \verb{;},
\item if it calls a core package (base, stats, etc.),
\item if it calls a package that is not loaded, and it is not one of
testthat, devtools, usethis, rlang, pkgload, or pkgdown which are explicitly allowed.
}

When RStudio does not run a \code{.run} hyperlink, then it shows the code
and the user can copy and paste it to the console, if they consider
it safe to run.

Note that depending on your version of RStudio, the behavior can change.
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/progress-client.R
\name{cli_progress_bar}
\alias{cli_progress_bar}
\alias{__cli_update_due}
\alias{cli_tick_reset}
\alias{ccli_tick_reset}
\alias{ticking}
\alias{cli_progress_update}
\alias{cli_progress_done}
\title{cli progress bars}
\usage{
cli_progress_bar(
  name = NULL,
  status = NULL,
  type = c("iterator", "tasks", "download", "custom"),
  total = NA,
  format = NULL,
  format_done = NULL,
  format_failed = NULL,
  clear = getOption("cli.progress_clear", TRUE),
  current = TRUE,
  auto_terminate = type != "download",
  extra = NULL,
  .auto_close = TRUE,
  .envir = parent.frame()
)

cli_progress_update(
  inc = NULL,
  set = NULL,
  total = NULL,
  status = NULL,
  extra = NULL,
  id = NULL,
  force = FALSE,
  .envir = parent.frame()
)

cli_progress_done(id = NULL, .envir = parent.frame(), result = "done")
}
\arguments{
\item{name}{This is typically used as a label, and should be short,
at most 20 characters.}

\item{status}{New status string of the progress bar, if not \code{NULL}.}

\item{type}{Type of the progress bar. It is used to select a default
display if \code{format} is not specified. Currently supported types:
\itemize{
\item \code{iterator}: e.g. a for loop or a mapping function,
\item \code{tasks}: a (typically small) number of tasks,
\item \code{download}: download of one file,
\item \code{custom}: custom type, \code{format} must not be \code{NULL} for this type.
}}

\item{total}{Total number of progress units, or \code{NA} if it is unknown.
\code{cli_progress_update()} can update the total number of units. This is
handy if you don't know the size of a download at the beginning, and
also in some other cases. If \code{format} is set to \code{NULL}, \code{format} (plus
\code{format_done} and \code{format_failed}) will be updated when you change
\code{total} from \code{NA} to a number. I.e. default format strings will be
updated, custom ones won't be.}

\item{format}{Format string. It has to be specified for custom progress
bars, otherwise it is optional, and a default display is selected
based on the progress bat type and whether the number of total units
is known. Format strings may contain glue substitution, the support
pluralization and cli styling. See \link{progress-variables} for special
variables that you can use in the custom format.}

\item{format_done}{Format string for successful termination. By default
the same as \code{format}.}

\item{format_failed}{Format string for unsuccessful termination. By
default the same as \code{format}.}

\item{clear}{Whether to remove the progress bar from the screen after
it has terminated. Defaults to the \code{cli.progress_clear} option, or
\code{TRUE} if unset.}

\item{current}{Whether to use this progress bar as the current progress
bar of the calling function. See more at 'The current progress bar'
below.}

\item{auto_terminate}{Whether to terminate the progress bar if the
number of current units reaches the number of total units.}

\item{extra}{Extra data to add to the progress bar. This can be
used in custom format strings for example. It should be a named list.
\code{cli_progress_update()} can update the extra data. Often you can get
away with referring to local variables in the format string, and
then you don't need to use this argument. Explicitly including these
constants or variables in \code{extra} can result in cleaner code. In
the rare cases when you need to refer to the same progress bar from
multiple functions, and you can them to \code{extra}.}

\item{.auto_close}{Whether to terminate the progress bar when the
calling function (or the one with execution environment in \code{.envir}
exits. (Auto termination does not work for progress bars created
from the global environment, e.g. from a script.)}

\item{.envir}{The environment to use for auto-termination and for glue
substitution. It is also used to find and set the current progress bar.}

\item{inc}{Increment in progress units. This is ignored if \code{set} is
not \code{NULL}.}

\item{set}{Set the current number of progress units to this value.
Ignored if \code{NULL}.}

\item{id}{Progress bar to update or terminate. If \code{NULL}, then the
current progress bar of the calling function (or \code{.envir} if
specified) is updated or terminated.}

\item{force}{Whether to force a display update, even if no update is
due.}

\item{result}{String to select successful or unsuccessful termination.
It is only used if the progress bar is not cleared from the screen.
It can be one of \code{"done"}, \code{"failed"}, \code{"clear"}, and \code{"auto"}.}
}
\value{
\code{cli_progress_bar()} returns the id of the new progress bar.
The id is a string constant.

\code{cli_progress_update()} returns the id of the progress bar,
invisibly.

\code{cli_progress_done()} returns \code{TRUE}, invisibly, always.
}
\description{
This is the reference manual of the three functions that create,
update and terminate progress bars. For a tutorial see the
\href{https://cli.r-lib.org/articles/progress.html}{cli progress bars}.

\code{cli_progress_bar()} creates a new progress bar.

\code{cli_progress_update()} updates the state of a progress bar, and
potentially the display as well.

\code{cli_progress_done()} terminates a progress bar.
}
\details{
\subsection{Basic usage}{

\code{cli_progress_bar()} creates a progress bar, \code{cli_progress_update()}
updates an existing progress bar, and \code{cli_progress_done()} terminates
it.

It is good practice to always set the \code{name} argument, to make the
progress bar more informative.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{clean <- function() \{
  cli_progress_bar("Cleaning data", total = 100)
  for (i in 1:100) \{
    Sys.sleep(5/100)
    cli_progress_update()
  \}
  cli_progress_done()
\}
clean()
}\if{html}{\out{</div>}}

\if{html}{\figure{progress-1.svg}}
}

\subsection{Progress bar types}{

There are three builtin types of progress bars, and a custom type.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{tasks <- function() \{
  cli_progress_bar("Tasks", total = 3, type = "tasks")
  for (i in 1:3) \{
    Sys.sleep(1)
    cli_progress_update()
  \}
  cli_progress_done()
\}
tasks()
}\if{html}{\out{</div>}}

\if{html}{\figure{progress-tasks.svg}}
}

\subsection{Unknown \code{total}}{

If \code{total} is not known, then cli shows a different progress bar.
Note that you can also set \code{total} in \code{cli_progress_update()}, if it
not known when the progress bar is created, but you learn it later.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{nototal <- function() \{
  cli_progress_bar("Parameter tuning")
  for (i in 1:100) \{
    Sys.sleep(3/100)
    cli_progress_update()
  \}
  cli_progress_done()
\}
nototal()
}\if{html}{\out{</div>}}

\if{html}{\figure{progress-natotal.svg}}
}

\subsection{Clearing the progress bar}{

By default cli removes terminated progress bars from the screen, if
the terminal supports this. If you want to change this, use the
\code{clear} argument of \code{cli_progress_bar()}, or the \code{cli.progress_clear}
global option (see \link{cli-config}) to change this.

(In the cli documentation we usually set \code{cli.progress_clear} to \code{FALSE},
so users can see how finished progress bars look.)

In this example the first progress bar is cleared, the second is not.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{fun <- function() \{
  cli_progress_bar("Data cleaning", total = 100, clear = TRUE)
  for (i in 1:100) \{
    Sys.sleep(3/100)
    cli_progress_update()
  \}
  cli_progress_bar("Parameter tuning", total = 100, clear = FALSE)
  for (i in 1:100) \{
    Sys.sleep(3/100)
    cli_progress_update()
  \}
\}
fun()
}\if{html}{\out{</div>}}

\if{html}{\figure{progress-clear.svg}}
}

\subsection{Initial delay}{

Updating a progress bar on the screen is costly, so cli tries to avoid
it for quick loops. By default a progress bar is only shown after two
seconds, or after half of that if less than 50\% of the iterations are
complete. You can change the two second default with the
\code{cli.progress_show_after} global option (see \link{cli-config}).

(In the cli documentation we usually set \code{cli.progress_show_after} to
\code{0} (zero seconds), so progress bars are shown immediately.)

In this example we only show the progress bar after one second, because
more than 50\% of the iterations remain after one second.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{fun <- function() \{
  cli_alert("Starting now, at \{Sys.time()\}")
  cli_progress_bar(
    total = 100,
    format = "\{cli::pb_bar\} \{pb_percent\} @ \{Sys.time()\}"
  )
  for (i in 1:100) \{
    Sys.sleep(4/100)
    cli_progress_update()
  \}
\}
options(cli.progress_show_after = 2)
fun()
}\if{html}{\out{</div>}}

\if{html}{\figure{progress-after.svg}}
}

\subsection{The \emph{current} progress bar}{

By default cli sets the new progress bar as the \emph{current} progress bar
of the calling function. The current progress bar is the default one
in cli progress bar operations. E.g. if no progress bar id is supplied
in \code{cli_progress_update()}, then the current progress bar is updated.

Every function can only have a single \emph{current} progress bar, and if a
new one is created, then the previous one (if any) is automatically
terminated. The current progress bar is also terminated when the function
that created it exits. Thanks to these rules, most often you don't need
to explicitly deal with progress bar ids, and you don't need to
explicitly call \code{cli_progress_done()}:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{fun <- function() \{
  cli_progress_bar("First step ", total = 100)
  for (i in 1:100) \{
    Sys.sleep(2/100)
    cli_progress_update()
  \}
  cli_progress_bar("Second step", total = 100)
  for (i in 1:100) \{
    Sys.sleep(2/100)
    cli_progress_update()
  \}
\}
fun()
}\if{html}{\out{</div>}}

\if{html}{\figure{progress-current.svg}}
}

\subsection{cli output while the progress bar is active}{

cli allows emitting regular cli output (alerts, headers, lists, etc.)
while a progress bar is active. On terminals that support this, cli
will remove the progress bar temporarily, emit the output, and then
restores the progress bar.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{fun <- function() \{
  cli_alert_info("Before the progress bar")
  cli_progress_bar("Calculating", total = 100)
  for (i in 1:50) \{
    Sys.sleep(4/100)
    cli_progress_update()
  \}
  cli_alert_info("Already half way!")
  for (i in 1:50) \{
    Sys.sleep(4/100)
    cli_progress_update()
  \}
  cli_alert_info("All done")
\}
fun()
}\if{html}{\out{</div>}}

\if{html}{\figure{progress-output.svg}}

See also \code{\link[=cli_progress_output]{cli_progress_output()}}, which sends text for the current
progress handler. E.g. in a Shiny app it will send the output to the
Shiny progress bar, as opposed to the \code{cli_alert()} etc. cli functions
which will print the text to the console.
}

\subsection{Custom formats}{

In addition to the builtin types, you can also specify a custom
format string. In this case \link[=progress-variables]{progress variables}
are probably useful to avoid calculating some progress bar quantities
like the elapsed time, of the ETA manually. You can also use your own
variables in the calling function:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{fun <- function(urls) \{
  cli_progress_bar(
    format = paste0(
      "\{pb_spin\} Downloading \{.path \{basename(url)\}\} ",
      "[\{pb_current\}/\{pb_total\}]   ETA:\{pb_eta\}"
    ),
    format_done = paste0(
      "\{col_green(symbol$tick)\} Downloaded \{pb_total\} files ",
      "in \{pb_elapsed\}."
    ),,
    total = length(urls)
  )
  for (url in urls) \{
    cli_progress_update()
    Sys.sleep(5/10)
  \}
\}
fun(paste0("https://acme.com/data-", 1:10, ".zip"))
}\if{html}{\out{</div>}}

\if{html}{\figure{progress-format.svg}}
}
}
\seealso{
These functions support \link[=inline-markup]{inline markup}.

\code{\link[=cli_progress_message]{cli_progress_message()}} and \code{\link[=cli_progress_step]{cli_progress_step()}} for simpler
progress messages.

Other progress bar functions: 
\code{\link{cli_progress_along}()},
\code{\link{cli_progress_builtin_handlers}()},
\code{\link{cli_progress_message}()},
\code{\link{cli_progress_num}()},
\code{\link{cli_progress_output}()},
\code{\link{cli_progress_step}()},
\code{\link{cli_progress_styles}()},
\code{\link{progress-variables}}

Other functions supporting inline markup: 
\code{\link{cli_abort}()},
\code{\link{cli_alert}()},
\code{\link{cli_blockquote}()},
\code{\link{cli_bullets}()},
\code{\link{cli_bullets_raw}()},
\code{\link{cli_dl}()},
\code{\link{cli_h1}()},
\code{\link{cli_li}()},
\code{\link{cli_ol}()},
\code{\link{cli_process_start}()},
\code{\link{cli_progress_along}()},
\code{\link{cli_progress_message}()},
\code{\link{cli_progress_output}()},
\code{\link{cli_progress_step}()},
\code{\link{cli_rule}},
\code{\link{cli_status}()},
\code{\link{cli_status_update}()},
\code{\link{cli_text}()},
\code{\link{cli_ul}()},
\code{\link{format_error}()},
\code{\link{format_inline}()}
}
\concept{functions supporting inline markup}
\concept{progress bar functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tree.R
\name{tree}
\alias{tree}
\title{Draw a tree}
\usage{
tree(
  data,
  root = data[[1]][[1]],
  style = NULL,
  width = console_width(),
  trim = FALSE
)
}
\arguments{
\item{data}{Data frame that contains the tree structure.
The first column is an id, and the second column is a list column,
that contains the ids of the child nodes. The optional third column
may contain the text to print to annotate the node.}

\item{root}{The name of the root node.}

\item{style}{Optional box style list.}

\item{width}{Maximum width of the output. Defaults to the \code{width}
option, see \code{\link[base:options]{base::options()}}.}

\item{trim}{Whether to avoid traversing the same nodes multiple times.
If \code{TRUE} and \code{data} has a \code{trimmed} column, then that is used for
printing repeated nodes.}
}
\value{
Character vector, the lines of the tree drawing.
}
\description{
Draw a tree using box drawing characters. Unicode characters are
used if available. (Set the \code{cli.unicode} option if auto-detection
fails.)
}
\details{
A node might appear multiple times in the tree, or might not appear
at all.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{data <- data.frame(
  stringsAsFactors = FALSE,
  package = c("processx", "backports", "assertthat", "Matrix",
    "magrittr", "rprojroot", "clisymbols", "prettyunits", "withr",
    "desc", "igraph", "R6", "crayon", "debugme", "digest", "irlba",
    "rcmdcheck", "callr", "pkgconfig", "lattice"),
  dependencies = I(list(
    c("assertthat", "crayon", "debugme", "R6"), character(0),
    character(0), "lattice", character(0), "backports", character(0),
    c("magrittr", "assertthat"), character(0),
    c("assertthat", "R6", "crayon", "rprojroot"),
    c("irlba", "magrittr", "Matrix", "pkgconfig"), character(0),
    character(0), "crayon", character(0), "Matrix",
    c("callr", "clisymbols", "crayon", "desc", "digest", "prettyunits",
      "R6", "rprojroot", "withr"),
    c("processx", "R6"), character(0), character(0)
  ))
)
tree(data)
}\if{html}{\out{</div>}}\if{html}{\out{
<div class="asciicast" style="color: #172431;font-family: 'Fira Code',Monaco,Consolas,Menlo,'Bitstream Vera Sans Mono','Powerline Symbols',monospace;line-height: 1.300000"><pre>
#> processx                                                                        
#> ├─assertthat                                                                    
#> ├─crayon                                                                        
#> ├─debugme                                                                       
#> │ └─crayon                                                                      
#> └─R6                                                                            
</pre></div>
}}


\if{html}{\out{<div class="sourceCode r">}}\preformatted{tree(data, root = "rcmdcheck")
}\if{html}{\out{</div>}}\if{html}{\out{
<div class="asciicast" style="color: #172431;font-family: 'Fira Code',Monaco,Consolas,Menlo,'Bitstream Vera Sans Mono','Powerline Symbols',monospace;line-height: 1.300000"><pre>
#> rcmdcheck                                                                       
#> ├─callr                                                                         
#> │ ├─processx                                                                    
#> │ │ ├─assertthat                                                                
#> │ │ ├─crayon                                                                    
#> │ │ ├─debugme                                                                   
#> │ │ │ └─crayon                                                                  
#> │ │ └─R6                                                                        
#> │ └─R6                                                                          
#> ├─clisymbols                                                                    
#> ├─crayon                                                                        
#> ├─desc                                                                          
#> │ ├─assertthat                                                                  
#> │ ├─R6                                                                          
#> │ ├─crayon                                                                      
#> │ └─rprojroot                                                                   
#> │   └─backports                                                                 
#> ├─digest                                                                        
#> ├─prettyunits                                                                   
#> │ ├─magrittr                                                                    
#> │ └─assertthat                                                                  
#> ├─R6                                                                            
#> ├─rprojroot                                                                     
#> │ └─backports                                                                   
#> └─withr                                                                         
</pre></div>
}}

\subsection{Colored nodes}{

\if{html}{\out{<div class="sourceCode r">}}\preformatted{data$label <- paste(data$package,
  style_dim(paste0("(", c("2.0.0.1", "1.1.1", "0.2.0", "1.2-11",
    "1.5", "1.2", "1.2.0", "1.0.2", "2.0.0", "1.1.1.9000", "1.1.2",
    "2.2.2", "1.3.4", "1.0.2", "0.6.12", "2.2.1", "1.2.1.9002",
    "1.0.0.9000", "2.0.1", "0.20-35"), ")"))
  )
roots <- ! data$package \%in\% unlist(data$dependencies)
data$label[roots] <- col_cyan(style_italic(data$label[roots]))
tree(data, root = "rcmdcheck")
}\if{html}{\out{</div>}}\if{html}{\out{
<div class="asciicast" style="color: #172431;font-family: 'Fira Code',Monaco,Consolas,Menlo,'Bitstream Vera Sans Mono','Powerline Symbols',monospace;line-height: 1.300000"><pre>
#> <span style="font-style: italic;color: #2AA198;">rcmdcheck (1.2.1.9002)</span>                                                          
#> ├─callr (1.0.0.9000)                                                            
#> │ ├─processx (2.0.0.1)                                                          
#> │ │ ├─assertthat (0.2.0)                                                        
#> │ │ ├─crayon (1.3.4)                                                            
#> │ │ ├─debugme (1.0.2)                                                           
#> │ │ │ └─crayon (1.3.4)                                                          
#> │ │ └─R6 (2.2.2)                                                                
#> │ └─R6 (2.2.2)                                                                  
#> ├─clisymbols (1.2.0)                                                            
#> ├─crayon (1.3.4)                                                                
#> ├─desc (1.1.1.9000)                                                             
#> │ ├─assertthat (0.2.0)                                                          
#> │ ├─R6 (2.2.2)                                                                  
#> │ ├─crayon (1.3.4)                                                              
#> │ └─rprojroot (1.2)                                                             
#> │   └─backports (1.1.1)                                                         
#> ├─digest (0.6.12)                                                               
#> ├─prettyunits (1.0.2)                                                           
#> │ ├─magrittr (1.5)                                                              
#> │ └─assertthat (0.2.0)                                                          
#> ├─R6 (2.2.2)                                                                    
#> ├─rprojroot (1.2)                                                               
#> │ └─backports (1.1.1)                                                           
#> └─withr (2.0.0)                                                                 
</pre></div>
}}

}

\subsection{Trimming}{

\if{html}{\out{<div class="sourceCode r">}}\preformatted{pkgdeps <- list(
  "dplyr@0.8.3" = c("assertthat@0.2.1", "glue@1.3.1", "magrittr@1.5",
    "R6@2.4.0", "Rcpp@1.0.2", "rlang@0.4.0", "tibble@2.1.3",
    "tidyselect@0.2.5"),
  "assertthat@0.2.1" = character(),
  "glue@1.3.1" = character(),
  "magrittr@1.5" = character(),
  "pkgconfig@2.0.3" = character(),
  "R6@2.4.0" = character(),
  "Rcpp@1.0.2" = character(),
  "rlang@0.4.0" = character(),
  "tibble@2.1.3" = c("cli@1.1.0", "crayon@1.3.4", "fansi@0.4.0",
     "pillar@1.4.2", "pkgconfig@2.0.3", "rlang@0.4.0"),
  "cli@1.1.0" = c("assertthat@0.2.1", "crayon@1.3.4"),
  "crayon@1.3.4" = character(),
  "fansi@0.4.0" = character(),
  "pillar@1.4.2" = c("cli@1.1.0", "crayon@1.3.4", "fansi@0.4.0",
     "rlang@0.4.0", "utf8@1.1.4", "vctrs@0.2.0"),
  "utf8@1.1.4" = character(),
  "vctrs@0.2.0" = c("backports@1.1.5", "ellipsis@0.3.0",
     "digest@0.6.21", "glue@1.3.1", "rlang@0.4.0", "zeallot@0.1.0"),
  "backports@1.1.5" = character(),
  "ellipsis@0.3.0" = c("rlang@0.4.0"),
  "digest@0.6.21" = character(),
  "glue@1.3.1" = character(),
  "zeallot@0.1.0" = character(),
  "tidyselect@0.2.5" = c("glue@1.3.1", "purrr@1.3.1", "rlang@0.4.0",
     "Rcpp@1.0.2"),
  "purrr@0.3.3" = c("magrittr@1.5", "rlang@0.4.0")
)

pkgs <- data.frame(
  stringsAsFactors = FALSE,
  name = names(pkgdeps),
  deps = I(unname(pkgdeps))
)

tree(pkgs, trim = TRUE)
}\if{html}{\out{</div>}}\if{html}{\out{
<div class="asciicast" style="color: #172431;font-family: 'Fira Code',Monaco,Consolas,Menlo,'Bitstream Vera Sans Mono','Powerline Symbols',monospace;line-height: 1.300000"><pre>
#> dplyr@0.8.3                                                                     
#> ├─assertthat@0.2.1                                                              
#> ├─glue@1.3.1                                                                    
#> ├─magrittr@1.5                                                                  
#> ├─R6@2.4.0                                                                      
#> ├─Rcpp@1.0.2                                                                    
#> ├─rlang@0.4.0                                                                   
#> ├─tibble@2.1.3                                                                  
#> │ ├─cli@1.1.0                                                                   
#> │ │ ├─assertthat@0.2.1                                                          
#> │ │ └─crayon@1.3.4                                                              
#> │ ├─crayon@1.3.4                                                                
#> │ ├─fansi@0.4.0                                                                 
#> │ ├─pillar@1.4.2                                                                
#> │ │ ├─cli@1.1.0                                                                 
#> │ │ ├─crayon@1.3.4                                                              
#> │ │ ├─fansi@0.4.0                                                               
#> │ │ ├─rlang@0.4.0                                                               
#> │ │ ├─utf8@1.1.4                                                                
#> │ │ └─vctrs@0.2.0                                                               
#> │ │   ├─backports@1.1.5                                                         
#> │ │   ├─ellipsis@0.3.0                                                          
#> │ │   │ └─rlang@0.4.0                                                           
#> │ │   ├─digest@0.6.21                                                           
#> │ │   ├─glue@1.3.1                                                              
#> │ │   ├─rlang@0.4.0                                                             
#> │ │   └─zeallot@0.1.0                                                           
#> │ ├─pkgconfig@2.0.3                                                             
#> │ └─rlang@0.4.0                                                                 
#> └─tidyselect@0.2.5                                                              
#>   ├─glue@1.3.1                                                                  
#>   ├─rlang@0.4.0                                                                 
#>   └─Rcpp@1.0.2                                                                  
</pre></div>
}}


\if{html}{\out{<div class="sourceCode r">}}\preformatted{# Mark the trimmed nodes
pkgs$label <- pkgs$name
pkgs$trimmed <- paste(pkgs$name, " (trimmed)")
tree(pkgs, trim = TRUE)
}\if{html}{\out{</div>}}\if{html}{\out{
<div class="asciicast" style="color: #172431;font-family: 'Fira Code',Monaco,Consolas,Menlo,'Bitstream Vera Sans Mono','Powerline Symbols',monospace;line-height: 1.300000"><pre>
#> dplyr@0.8.3                                                                     
#> ├─assertthat@0.2.1                                                              
#> ├─glue@1.3.1                                                                    
#> ├─magrittr@1.5                                                                  
#> ├─R6@2.4.0                                                                      
#> ├─Rcpp@1.0.2                                                                    
#> ├─rlang@0.4.0                                                                   
#> ├─tibble@2.1.3                                                                  
#> │ ├─cli@1.1.0                                                                   
#> │ │ ├─assertthat@0.2.1  (trimmed)                                               
#> │ │ └─crayon@1.3.4                                                              
#> │ ├─crayon@1.3.4  (trimmed)                                                     
#> │ ├─fansi@0.4.0                                                                 
#> │ ├─pillar@1.4.2                                                                
#> │ │ ├─cli@1.1.0  (trimmed)                                                      
#> │ │ ├─crayon@1.3.4  (trimmed)                                                   
#> │ │ ├─fansi@0.4.0  (trimmed)                                                    
#> │ │ ├─rlang@0.4.0  (trimmed)                                                    
#> │ │ ├─utf8@1.1.4                                                                
#> │ │ └─vctrs@0.2.0                                                               
#> │ │   ├─backports@1.1.5                                                         
#> │ │   ├─ellipsis@0.3.0                                                          
#> │ │   │ └─rlang@0.4.0  (trimmed)                                                
#> │ │   ├─digest@0.6.21                                                           
#> │ │   ├─glue@1.3.1  (trimmed)                                                   
#> │ │   ├─rlang@0.4.0  (trimmed)                                                  
#> │ │   └─zeallot@0.1.0                                                           
#> │ ├─pkgconfig@2.0.3                                                             
#> │ └─rlang@0.4.0  (trimmed)                                                      
#> └─tidyselect@0.2.5                                                              
#>   ├─glue@1.3.1  (trimmed)                                                       
#>   ├─rlang@0.4.0  (trimmed)                                                      
#>   └─Rcpp@1.0.2  (trimmed)                                                       
</pre></div>
}}

}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/format-conditions.R
\name{format_error}
\alias{format_error}
\alias{format_warning}
\alias{format_message}
\title{Format an error, warning or diagnostic message}
\usage{
format_error(message, .envir = parent.frame())

format_warning(message, .envir = parent.frame())

format_message(message, .envir = parent.frame())
}
\arguments{
\item{message}{It is formatted via a call to \code{\link[=cli_bullets]{cli_bullets()}}.}

\item{.envir}{Environment to evaluate the glue expressions in.}
}
\description{
You can then throw this message with \code{\link[=stop]{stop()}} or \code{rlang::abort()}.
}
\details{
The messages can use inline styling, pluralization and glue
substitutions.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{n <- "boo"
stop(format_error(c(
        "\{.var n\} must be a numeric vector",
  "x" = "You've supplied a \{.cls \{class(n)\}\} vector."
)))
}\if{html}{\out{</div>}}\if{html}{\out{
<div class="asciicast" style="color: #172431;font-family: 'Fira Code',Monaco,Consolas,Menlo,'Bitstream Vera Sans Mono','Powerline Symbols',monospace;line-height: 1.300000"><pre>
#> Error: `n` must be a numeric vector                                             
#> <span style="color: #DC322F;">✖</span> You've supplied a <span style="color: #268BD2;">&lt;character&gt;</span> vector.                                         
</pre></div>
}}


\if{html}{\out{<div class="sourceCode r">}}\preformatted{len <- 26
idx <- 100
stop(format_error(c(
        "Must index an existing element:",
  "i" = "There \{?is/are\} \{len\} element\{?s\}.",
  "x" = "You've tried to subset element \{idx\}."
)))
}\if{html}{\out{</div>}}\if{html}{\out{
<div class="asciicast" style="color: #172431;font-family: 'Fira Code',Monaco,Consolas,Menlo,'Bitstream Vera Sans Mono','Powerline Symbols',monospace;line-height: 1.300000"><pre>
#> Error: Must index an existing element:                                          
#> <span style="color: #2AA198;">ℹ</span> There are 26 elements.                                                        
#> <span style="color: #DC322F;">✖</span> You've tried to subset element 100.                                           
</pre></div>
}}
}
\seealso{
These functions support \link[=inline-markup]{inline markup}.

Other functions supporting inline markup: 
\code{\link{cli_abort}()},
\code{\link{cli_alert}()},
\code{\link{cli_blockquote}()},
\code{\link{cli_bullets}()},
\code{\link{cli_bullets_raw}()},
\code{\link{cli_dl}()},
\code{\link{cli_h1}()},
\code{\link{cli_li}()},
\code{\link{cli_ol}()},
\code{\link{cli_process_start}()},
\code{\link{cli_progress_along}()},
\code{\link{cli_progress_bar}()},
\code{\link{cli_progress_message}()},
\code{\link{cli_progress_output}()},
\code{\link{cli_progress_step}()},
\code{\link{cli_rule}},
\code{\link{cli_status}()},
\code{\link{cli_status_update}()},
\code{\link{cli_text}()},
\code{\link{cli_ul}()},
\code{\link{format_inline}()}
}
\concept{functions supporting inline markup}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hash.R
\name{hash_xxhash}
\alias{hash_xxhash}
\alias{hash_raw_xxhash}
\alias{hash_obj_xxhash}
\alias{hash_file_xxhash}
\alias{hash_xxhash64}
\alias{hash_raw_xxhash64}
\alias{hash_obj_xxhash64}
\alias{hash_file_xxhash64}
\title{xxHash}
\usage{
hash_xxhash(x)

hash_raw_xxhash(x)

hash_obj_xxhash(x, serialize_version = 2)

hash_file_xxhash(paths)

hash_xxhash64(x)

hash_raw_xxhash64(x)

hash_obj_xxhash64(x, serialize_version = 2)

hash_file_xxhash64(paths)
}
\arguments{
\item{x}{Character vector. If not a character vector, then
\code{\link[=as.character]{as.character()}} is used to try to coerce it into one. \code{NA} entries
will have an \code{NA} hash.}

\item{serialize_version}{Workspace format version to use, see
\code{\link[base:serialize]{base::serialize()}}.}

\item{paths}{Character vector of file names.}
}
\value{
\code{hash_xxhash()} returns a character vector of hexadecimal
xxHash hashes.

\code{hash_raw_xxhash()} returns a character scalar.

\code{hash_obj_xxhash()} returns a character scalar.

\code{hash_file_xxhash()} returns a character vector of xxHash
hashes.
}
\description{
Extremely fast hash algorithm.
}
\details{
\code{hash_raw_xxhash()} calculates the xxHash hash of the bytes
of a raw vector.

\code{hash_obj_xxhash()} calculates the xxHash hash of an R
object. The object is serialized into a binary vector first.

\code{hash_file_xxhash()} calculates the xxHash hash of one or
more files.

The \code{64} functions caculate the 64 bit variant
of xxHash. Otherwise they work the same.
}
\examples{
hash_xxhash(c("foo", NA, "bar", ""))
}
\seealso{
Other hash functions: 
\code{\link{hash_animal}()},
\code{\link{hash_emoji}()},
\code{\link{hash_md5}()},
\code{\link{hash_sha1}()},
\code{\link{hash_sha256}()}
}
\concept{hash functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/keypress.R
\name{has_keypress_support}
\alias{has_keypress_support}
\title{Check if the current platform/terminal supports reading
single keys.}
\usage{
has_keypress_support()
}
\value{
Whether there is support for waiting for individual
keypressses.
}
\description{
Check if the current platform/terminal supports reading
single keys.
}
\details{
Supported platforms:
\itemize{
\item Terminals in Windows and Unix.
\item RStudio terminal.
}

Not supported:
\itemize{
\item RStudio (if not in the RStudio terminal).
\item R.app on macOS.
\item Rgui on Windows.
\item Emacs ESS.
\item Others.
}
}
\examples{
has_keypress_support()
}
\seealso{
Other keypress function: 
\code{\link{keypress}()}
}
\concept{keypress function}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cli.R
\name{cli_verbatim}
\alias{cli_verbatim}
\title{CLI verbatim text}
\usage{
cli_verbatim(..., .envir = parent.frame())
}
\arguments{
\item{...}{The text to show, in character vectors. Each element is
printed on a new line.}

\item{.envir}{Environment to evaluate the glue expressions in.}
}
\description{
It is not wrapped, but printed as is. Long lines will overflow.
No glue substitution is performed on verbatim text.
}
\details{
\subsection{Line breaks}{

\if{html}{\out{<div class="sourceCode r">}}\preformatted{cli_verbatim("This has\\nthree\\nlines,")
}\if{html}{\out{</div>}}\if{html}{\out{
<div class="asciicast" style="color: #172431;font-family: 'Fira Code',Monaco,Consolas,Menlo,'Bitstream Vera Sans Mono','Powerline Symbols',monospace;line-height: 1.300000"><pre>
#> This has                                                                        
#> three                                                                           
#> lines,                                                                          
</pre></div>
}}

}

\subsection{Special characters}{

No glue substitution happens here.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{cli_verbatim("No string \{interpolation\} or \{.emph styling\} here")
}\if{html}{\out{</div>}}\if{html}{\out{
<div class="asciicast" style="color: #172431;font-family: 'Fira Code',Monaco,Consolas,Menlo,'Bitstream Vera Sans Mono','Powerline Symbols',monospace;line-height: 1.300000"><pre>
#> No string \{interpolation\} or \{.emph styling\} here                               
</pre></div>
}}

}
}
\seealso{
\code{\link[=cli_code]{cli_code()}} for printing R or other source code.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cli.R
\name{cli_rule}
\alias{cli_rule}
\title{CLI horizontal rule}
\usage{
cli_rule(
  left = "",
  center = "",
  right = "",
  id = NULL,
  .envir = parent.frame()
)
}
\arguments{
\item{left}{Label to show on the left. It interferes with the \code{center}
label, only at most one of them can be present.}

\item{center}{Label to show at the center. It interferes  with the
\code{left} and \code{right} labels.}

\item{right}{Label to show on the right. It interferes with the \code{center}
label, only at most one of them can be present.}

\item{id}{Element id, a string. If \code{NULL}, then a new id is generated
and returned.}

\item{.envir}{Environment to evaluate the glue expressions in.}
}
\description{
It can be used to separate parts of the output.
}
\details{
\subsection{Inline styling and interpolation}{

\if{html}{\out{<div class="sourceCode r">}}\preformatted{pkg <- "mypackage"
cli_rule(left = "\{.pkg \{pkg\}\} results")
}\if{html}{\out{</div>}}\if{html}{\out{
<div class="asciicast" style="color: #172431;font-family: 'Fira Code',Monaco,Consolas,Menlo,'Bitstream Vera Sans Mono','Powerline Symbols',monospace;line-height: 1.300000"><pre>
#> ── <span style="color: #268BD2;">mypackage</span> results ─────────────────────────────────────────────────          
</pre></div>
}}

}

\subsection{Theming}{

The line style of the rule can be changed via the the \code{line-type}
property. Possible values are:
\itemize{
\item \code{"single"}: (same as \code{1}), a single line,
\item \code{"double"}: (same as \code{2}), a double line,
\item \code{"bar1"}, \code{"bar2"}, \code{"bar3"}, etc., \code{"bar8"} uses varying height bars.
}

Colors and background colors can similarly changed via a theme.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{d <- cli_div(theme = list(rule = list(
  color = "cyan",
  "line-type" = "double")))
cli_rule("Summary", right = "\{.pkg mypackage\}")
cli_end(d)
}\if{html}{\out{</div>}}\if{html}{\out{
<div class="asciicast" style="color: #172431;font-family: 'Fira Code',Monaco,Consolas,Menlo,'Bitstream Vera Sans Mono','Powerline Symbols',monospace;line-height: 1.300000"><pre>
#> <span style="color: #2AA198;">══ Summary ══════════════════════════════════════════════ </span><span style="color: #268BD2;">mypackage</span><span style="color: #2AA198;"> ══</span>          
</pre></div>
}}

}
}
\seealso{
This function supports \link[=inline-markup]{inline markup}.

Other functions supporting inline markup: 
\code{\link{cli_abort}()},
\code{\link{cli_alert}()},
\code{\link{cli_blockquote}()},
\code{\link{cli_bullets}()},
\code{\link{cli_bullets_raw}()},
\code{\link{cli_dl}()},
\code{\link{cli_h1}()},
\code{\link{cli_li}()},
\code{\link{cli_ol}()},
\code{\link{cli_process_start}()},
\code{\link{cli_progress_along}()},
\code{\link{cli_progress_bar}()},
\code{\link{cli_progress_message}()},
\code{\link{cli_progress_output}()},
\code{\link{cli_progress_step}()},
\code{\link{cli_status}()},
\code{\link{cli_status_update}()},
\code{\link{cli_text}()},
\code{\link{cli_ul}()},
\code{\link{format_error}()},
\code{\link{format_inline}()}
}
\concept{functions supporting inline markup}
